#include <stdio.h>
#include <stdlib.h>

#include "ptg_gen.h"
#include "obstack.h"

/* -------------------------------------------------------- */
/*                      Memory Management                   */
/* -------------------------------------------------------- */

static Obstack _PTGObstack;
static void *_PTGFirstObj = NULL;

static void _PTGInit()
{
	if (_PTGFirstObj)
		return;
	obstack_init(&_PTGObstack);
	_PTGFirstObj = obstack_alloc(&_PTGObstack, 0);
}

void PTGFree()
{
	obstack_free(&_PTGObstack, _PTGFirstObj);
	_PTGFirstObj = obstack_alloc(&_PTGObstack, 0);
}


#if defined(__STDC__) || defined(__cplusplus)
static void *MALLOC(int size)
#else
static void *MALLOC(size) int size;
#endif
{
	if (! _PTGFirstObj)
		_PTGInit();
	return (obstack_alloc(&_PTGObstack, size));
}

/* -------------------------------------------------------- */
/*                       Output-functions                   */
/* -------------------------------------------------------- */

static PTG_OUTPUT_FILE f;
static char buffer[40];

#if PTG_OUTPUT_DEFAULT
#if defined(__STDC__) || defined(__cplusplus)
PTGNode PTGOut (PTGNode r)
#else
PTGNode PTGOut (r)
	PTGNode r;
#endif
{
	f = stdout;
	if (r) {
		(* (r->_print)) (r);
	}
	return (r);
}

#if defined(__STDC__) || defined(__cplusplus)
PTGNode PTGOutFile (char *fn, PTGNode r)
#else
PTGNode PTGOutFile (fn, r)
	char * fn; PTGNode r;
#endif
{
	if ((f = fopen(fn, "w"))  == (FILE *)NULL)
	{
		fprintf(stderr, "ERROR: PTGOutFile: output file '%s' can't be opened.\n",fn);
		exit(1);
	}
	if (r)
		(* (r->_print)) (r);
	fclose(f);
	return (r);
}

#if defined(__STDC__) || defined(__cplusplus)
PTGNode PTGOutFPtr(FILE *fptr, PTGNode r)
#else
PTGNode PTGOutFPtr(fptr, r)
	FILE *fptr; PTGNode r;
#endif
{
	if ((f = fptr) == (FILE *)NULL)
	{
		fprintf(stderr, "ERROR: PTGOutFPtr: output file not open.\n");
		exit(1);
	}
	if (r)
		(* (r->_print)) (r);
	return (r);
}

#endif

#if defined(__STDC__) || defined(__cplusplus)
PTGNode PTGProcess(PTG_OUTPUT_FILE file, PTGNode r)
#else
PTGNode PTGProcess(file, r)
	PTG_OUTPUT_FILE file; PTGNode r;
#endif
{
	f = file;
	if (r)
		(* (r->_print)) (r);
	return (r);
}
/* -------------------------------------------------------- */
/*                            PTGNULL                       */
/* -------------------------------------------------------- */

/* Define PTGNULL as a PTGNode that prints nothing. */

#if defined(__STDC__) || defined(__cplusplus)
static void _PrPTGNULL(_PPTG0 n)
#else
static void _PrPTGNULL(n)
_PPTG0 n;
#endif
{(void)n; /* function printing nothing */}

struct _SPTG0   _PTGNULL = { (_PTGProc) _PrPTGNULL };

/* -------------------------------------------------------- */
/*          Node-Construction and Print-functions           */
/* -------------------------------------------------------- */


/* Implementation of Pattern Frame */

typedef struct _SPTGFrame{
	_PTGProc _print;
	PTGNode p1;
	PTGNode p2;
} * _PPTGFrame;

#ifdef PROTO_OK
static void _PrPTGFrame(_PPTGFrame n)
#else
static void _PrPTGFrame(n)
	_PPTGFrame n;
#endif
{
	PTG_OUTPUT_STRING(f, "set mainFct {");
	n->p1->_print(n->p1);
	PTG_OUTPUT_STRING(f, "}\n\n");
	n->p2->_print(n->p2);
	PTG_OUTPUT_STRING(f, "\n");
}

#ifdef PROTO_OK
PTGNode PTGFrame(PTGNode p1, PTGNode p2)
#else
PTGNode PTGFrame(p1, p2)
PTGNode p1;
PTGNode p2;
#endif
{
	_PPTGFrame n;

	n = (_PPTGFrame)MALLOC(sizeof(struct _SPTGFrame));
	n->_print = (_PTGProc)_PrPTGFrame;
	n->p1 = p1;
	n->p2 = p2;
	return (PTGNode)n;
}

/* Implementation of Pattern DoubleNewlineSeq */

typedef struct _SPTGDoubleNewlineSeq{
	_PTGProc _print;
	PTGNode p1;
	PTGNode p2;
} * _PPTGDoubleNewlineSeq;

#ifdef PROTO_OK
static void _PrPTGDoubleNewlineSeq(_PPTGDoubleNewlineSeq n)
#else
static void _PrPTGDoubleNewlineSeq(n)
	_PPTGDoubleNewlineSeq n;
#endif
{
	n->p1->_print(n->p1);
	if (n->p1 != PTGNULL && n->p2 != PTGNULL)
	{
		PTG_OUTPUT_STRING(f, "\n\n");
	}
	n->p2->_print(n->p2);
}

#ifdef PROTO_OK
PTGNode PTGDoubleNewlineSeq(PTGNode p1, PTGNode p2)
#else
PTGNode PTGDoubleNewlineSeq(p1, p2)
PTGNode p1;
PTGNode p2;
#endif
{
	_PPTGDoubleNewlineSeq n;

	n = (_PPTGDoubleNewlineSeq)MALLOC(sizeof(struct _SPTGDoubleNewlineSeq));
	if(p1 == PTGNULL && p2 == PTGNULL)
		return PTGNULL;

	n->_print = (_PTGProc)_PrPTGDoubleNewlineSeq;
	n->p1 = p1;
	n->p2 = p2;
	return (PTGNode)n;
}

/* Implementation of Pattern EmptyName */

typedef struct _SPTGEmptyName{
	_PTGProc _print;
} * _PPTGEmptyName;

#ifdef PROTO_OK
static void _PrPTGEmptyName(_PPTGEmptyName n)
#else
static void _PrPTGEmptyName(n)
	_PPTGEmptyName n;
#endif
{
	PTG_OUTPUT_STRING(f, "EPS");
}

static struct _SPTGEmptyName _sptgEmptyName = { (_PTGProc) _PrPTGEmptyName };

#ifdef PROTO_OK
PTGNode PTGEmptyName(void)
#else
PTGNode PTGEmptyName()
#endif
{
	return (PTGNode)(&_sptgEmptyName);
}

/* Implementation of Pattern AltName */

typedef struct _SPTGAltName{
	_PTGProc _print;
	PTGNode p1;
	PTGNode p2;
} * _PPTGAltName;

#ifdef PROTO_OK
static void _PrPTGAltName(_PPTGAltName n)
#else
static void _PrPTGAltName(n)
	_PPTGAltName n;
#endif
{
	PTG_OUTPUT_STRING(f, "(");
	n->p1->_print(n->p1);
	PTG_OUTPUT_STRING(f, "|");
	n->p2->_print(n->p2);
	PTG_OUTPUT_STRING(f, ")");
}

#ifdef PROTO_OK
PTGNode PTGAltName(PTGNode p1, PTGNode p2)
#else
PTGNode PTGAltName(p1, p2)
PTGNode p1;
PTGNode p2;
#endif
{
	_PPTGAltName n;

	n = (_PPTGAltName)MALLOC(sizeof(struct _SPTGAltName));
	n->_print = (_PTGProc)_PrPTGAltName;
	n->p1 = p1;
	n->p2 = p2;
	return (PTGNode)n;
}

/* Implementation of Pattern ConcatName */

typedef struct _SPTGConcatName{
	_PTGProc _print;
	PTGNode p1;
	PTGNode p2;
} * _PPTGConcatName;

#ifdef PROTO_OK
static void _PrPTGConcatName(_PPTGConcatName n)
#else
static void _PrPTGConcatName(n)
	_PPTGConcatName n;
#endif
{
	PTG_OUTPUT_STRING(f, "(");
	n->p1->_print(n->p1);
	n->p2->_print(n->p2);
	PTG_OUTPUT_STRING(f, ")");
}

#ifdef PROTO_OK
PTGNode PTGConcatName(PTGNode p1, PTGNode p2)
#else
PTGNode PTGConcatName(p1, p2)
PTGNode p1;
PTGNode p2;
#endif
{
	_PPTGConcatName n;

	n = (_PPTGConcatName)MALLOC(sizeof(struct _SPTGConcatName));
	n->_print = (_PTGProc)_PrPTGConcatName;
	n->p1 = p1;
	n->p2 = p2;
	return (PTGNode)n;
}

/* Implementation of Pattern RepName */

typedef struct _SPTGRepName{
	_PTGProc _print;
	PTGNode p1;
} * _PPTGRepName;

#ifdef PROTO_OK
static void _PrPTGRepName(_PPTGRepName n)
#else
static void _PrPTGRepName(n)
	_PPTGRepName n;
#endif
{
	PTG_OUTPUT_STRING(f, "(");
	n->p1->_print(n->p1);
	PTG_OUTPUT_STRING(f, ")*");
}

#ifdef PROTO_OK
PTGNode PTGRepName(PTGNode p1)
#else
PTGNode PTGRepName(p1)
PTGNode p1;
#endif
{
	_PPTGRepName n;

	n = (_PPTGRepName)MALLOC(sizeof(struct _SPTGRepName));
	n->_print = (_PTGProc)_PrPTGRepName;
	n->p1 = p1;
	return (PTGNode)n;
}

/* Implementation of Pattern NonOptRepName */

typedef struct _SPTGNonOptRepName{
	_PTGProc _print;
	PTGNode p1;
} * _PPTGNonOptRepName;

#ifdef PROTO_OK
static void _PrPTGNonOptRepName(_PPTGNonOptRepName n)
#else
static void _PrPTGNonOptRepName(n)
	_PPTGNonOptRepName n;
#endif
{
	PTG_OUTPUT_STRING(f, "(");
	n->p1->_print(n->p1);
	PTG_OUTPUT_STRING(f, ")+");
}

#ifdef PROTO_OK
PTGNode PTGNonOptRepName(PTGNode p1)
#else
PTGNode PTGNonOptRepName(p1)
PTGNode p1;
#endif
{
	_PPTGNonOptRepName n;

	n = (_PPTGNonOptRepName)MALLOC(sizeof(struct _SPTGNonOptRepName));
	n->_print = (_PTGProc)_PrPTGNonOptRepName;
	n->p1 = p1;
	return (PTGNode)n;
}

/* Implementation of Pattern FieldSelName */

typedef struct _SPTGFieldSelName{
	_PTGProc _print;
	PTGNode p1;
} * _PPTGFieldSelName;

#ifdef PROTO_OK
static void _PrPTGFieldSelName(_PPTGFieldSelName n)
#else
static void _PrPTGFieldSelName(n)
	_PPTGFieldSelName n;
#endif
{
	PTG_OUTPUT_STRING(f, ".");
	n->p1->_print(n->p1);
}

#ifdef PROTO_OK
PTGNode PTGFieldSelName(PTGNode p1)
#else
PTGNode PTGFieldSelName(p1)
PTGNode p1;
#endif
{
	_PPTGFieldSelName n;

	n = (_PPTGFieldSelName)MALLOC(sizeof(struct _SPTGFieldSelName));
	n->_print = (_PTGProc)_PrPTGFieldSelName;
	n->p1 = p1;
	return (PTGNode)n;
}

/* Implementation of Pattern ConditionName */

typedef struct _SPTGConditionName{
	_PTGProc _print;
	PTGNode p1;
} * _PPTGConditionName;

#ifdef PROTO_OK
static void _PrPTGConditionName(_PPTGConditionName n)
#else
static void _PrPTGConditionName(n)
	_PPTGConditionName n;
#endif
{
	PTG_OUTPUT_STRING(f, "[");
	n->p1->_print(n->p1);
	PTG_OUTPUT_STRING(f, "]");
}

#ifdef PROTO_OK
PTGNode PTGConditionName(PTGNode p1)
#else
PTGNode PTGConditionName(p1)
PTGNode p1;
#endif
{
	_PPTGConditionName n;

	n = (_PPTGConditionName)MALLOC(sizeof(struct _SPTGConditionName));
	n->_print = (_PTGProc)_PrPTGConditionName;
	n->p1 = p1;
	return (PTGNode)n;
}

/* Implementation of Pattern AltFct */

typedef struct _SPTGAltFct{
	_PTGProc _print;
	PTGNode p1;
	PTGNode p2;
	PTGNode p3;
} * _PPTGAltFct;

#ifdef PROTO_OK
static void _PrPTGAltFct(_PPTGAltFct n)
#else
static void _PrPTGAltFct(n)
	_PPTGAltFct n;
#endif
{
	PTG_OUTPUT_STRING(f, "proc {");
	n->p1->_print(n->p1);
	PTG_OUTPUT_STRING(f, "} {nodeList} {\n");
	PTG_OUTPUT_STRING(f, "  return [union [{");
	n->p2->_print(n->p2);
	PTG_OUTPUT_STRING(f, "} $nodeList] [{");
	n->p3->_print(n->p3);
	PTG_OUTPUT_STRING(f, "} $nodeList]]\n");
	PTG_OUTPUT_STRING(f, "}");
}

#ifdef PROTO_OK
PTGNode PTGAltFct(PTGNode p1, PTGNode p2, PTGNode p3)
#else
PTGNode PTGAltFct(p1, p2, p3)
PTGNode p1;
PTGNode p2;
PTGNode p3;
#endif
{
	_PPTGAltFct n;

	n = (_PPTGAltFct)MALLOC(sizeof(struct _SPTGAltFct));
	n->_print = (_PTGProc)_PrPTGAltFct;
	n->p1 = p1;
	n->p2 = p2;
	n->p3 = p3;
	return (PTGNode)n;
}

/* Implementation of Pattern ConcatFct */

typedef struct _SPTGConcatFct{
	_PTGProc _print;
	PTGNode p1;
	PTGNode p2;
	PTGNode p3;
} * _PPTGConcatFct;

#ifdef PROTO_OK
static void _PrPTGConcatFct(_PPTGConcatFct n)
#else
static void _PrPTGConcatFct(n)
	_PPTGConcatFct n;
#endif
{
	PTG_OUTPUT_STRING(f, "proc {");
	n->p1->_print(n->p1);
	PTG_OUTPUT_STRING(f, "} {nodeList} {\n");
	PTG_OUTPUT_STRING(f, "  return [{");
	n->p3->_print(n->p3);
	PTG_OUTPUT_STRING(f, "} [{");
	n->p2->_print(n->p2);
	PTG_OUTPUT_STRING(f, "} $nodeList]]\n");
	PTG_OUTPUT_STRING(f, "}");
}

#ifdef PROTO_OK
PTGNode PTGConcatFct(PTGNode p1, PTGNode p2, PTGNode p3)
#else
PTGNode PTGConcatFct(p1, p2, p3)
PTGNode p1;
PTGNode p2;
PTGNode p3;
#endif
{
	_PPTGConcatFct n;

	n = (_PPTGConcatFct)MALLOC(sizeof(struct _SPTGConcatFct));
	n->_print = (_PTGProc)_PrPTGConcatFct;
	n->p1 = p1;
	n->p2 = p2;
	n->p3 = p3;
	return (PTGNode)n;
}

/* Implementation of Pattern RepFct */

typedef struct _SPTGRepFct{
	_PTGProc _print;
	PTGNode p1;
	PTGNode p2;
} * _PPTGRepFct;

#ifdef PROTO_OK
static void _PrPTGRepFct(_PPTGRepFct n)
#else
static void _PrPTGRepFct(n)
	_PPTGRepFct n;
#endif
{
	PTG_OUTPUT_STRING(f, "proc {");
	n->p1->_print(n->p1);
	PTG_OUTPUT_STRING(f, "} {nodeList} {\n");
	PTG_OUTPUT_STRING(f, "  while 1 {\n");
	PTG_OUTPUT_STRING(f, "    set len [llength $nodeList]\n");
	PTG_OUTPUT_STRING(f, "    set nodeList [union $nodeList [{");
	n->p2->_print(n->p2);
	PTG_OUTPUT_STRING(f, "} $nodeList]]\n");
	PTG_OUTPUT_STRING(f, "    if [llength $nodeList]==$len {\n");
	PTG_OUTPUT_STRING(f, "      return $nodeList\n");
	PTG_OUTPUT_STRING(f, "    }\n");
	PTG_OUTPUT_STRING(f, "  }\n");
	PTG_OUTPUT_STRING(f, "}");
}

#ifdef PROTO_OK
PTGNode PTGRepFct(PTGNode p1, PTGNode p2)
#else
PTGNode PTGRepFct(p1, p2)
PTGNode p1;
PTGNode p2;
#endif
{
	_PPTGRepFct n;

	n = (_PPTGRepFct)MALLOC(sizeof(struct _SPTGRepFct));
	n->_print = (_PTGProc)_PrPTGRepFct;
	n->p1 = p1;
	n->p2 = p2;
	return (PTGNode)n;
}

/* Implementation of Pattern NonOptRepFct */

typedef struct _SPTGNonOptRepFct{
	_PTGProc _print;
	PTGNode p1;
	PTGNode p2;
} * _PPTGNonOptRepFct;

#ifdef PROTO_OK
static void _PrPTGNonOptRepFct(_PPTGNonOptRepFct n)
#else
static void _PrPTGNonOptRepFct(n)
	_PPTGNonOptRepFct n;
#endif
{
	PTG_OUTPUT_STRING(f, "proc {");
	n->p1->_print(n->p1);
	PTG_OUTPUT_STRING(f, "} {nodeList} {\n");
	PTG_OUTPUT_STRING(f, "  set nodeList [{");
	n->p2->_print(n->p2);
	PTG_OUTPUT_STRING(f, "} $nodeList]\n");
	PTG_OUTPUT_STRING(f, "  while 1 {\n");
	PTG_OUTPUT_STRING(f, "    set len [llength $nodeList]\n");
	PTG_OUTPUT_STRING(f, "    set nodeList [union $nodeList [{");
	n->p2->_print(n->p2);
	PTG_OUTPUT_STRING(f, "} $nodeList]]\n");
	PTG_OUTPUT_STRING(f, "    if [llength $nodeList]==$len {\n");
	PTG_OUTPUT_STRING(f, "      return $nodeList\n");
	PTG_OUTPUT_STRING(f, "    }\n");
	PTG_OUTPUT_STRING(f, "  }\n");
	PTG_OUTPUT_STRING(f, "}");
}

#ifdef PROTO_OK
PTGNode PTGNonOptRepFct(PTGNode p1, PTGNode p2)
#else
PTGNode PTGNonOptRepFct(p1, p2)
PTGNode p1;
PTGNode p2;
#endif
{
	_PPTGNonOptRepFct n;

	n = (_PPTGNonOptRepFct)MALLOC(sizeof(struct _SPTGNonOptRepFct));
	n->_print = (_PTGProc)_PrPTGNonOptRepFct;
	n->p1 = p1;
	n->p2 = p2;
	return (PTGNode)n;
}

/* Implementation of Pattern FieldSelFct */

typedef struct _SPTGFieldSelFct{
	_PTGProc _print;
	PTGNode p1;
	PTGNode p2;
} * _PPTGFieldSelFct;

#ifdef PROTO_OK
static void _PrPTGFieldSelFct(_PPTGFieldSelFct n)
#else
static void _PrPTGFieldSelFct(n)
	_PPTGFieldSelFct n;
#endif
{
	PTG_OUTPUT_STRING(f, "proc {");
	n->p1->_print(n->p1);
	PTG_OUTPUT_STRING(f, "} {nodeList} {\n");
	PTG_OUTPUT_STRING(f, "  return [fieldSel {");
	n->p2->_print(n->p2);
	PTG_OUTPUT_STRING(f, "} $nodeList]\n");
	PTG_OUTPUT_STRING(f, "}");
}

#ifdef PROTO_OK
PTGNode PTGFieldSelFct(PTGNode p1, PTGNode p2)
#else
PTGNode PTGFieldSelFct(p1, p2)
PTGNode p1;
PTGNode p2;
#endif
{
	_PPTGFieldSelFct n;

	n = (_PPTGFieldSelFct)MALLOC(sizeof(struct _SPTGFieldSelFct));
	n->_print = (_PTGProc)_PrPTGFieldSelFct;
	n->p1 = p1;
	n->p2 = p2;
	return (PTGNode)n;
}

/* Implementation of Pattern ConditionFct */

typedef struct _SPTGConditionFct{
	_PTGProc _print;
	PTGNode p1;
	PTGNode p2;
} * _PPTGConditionFct;

#ifdef PROTO_OK
static void _PrPTGConditionFct(_PPTGConditionFct n)
#else
static void _PrPTGConditionFct(n)
	_PPTGConditionFct n;
#endif
{
	PTG_OUTPUT_STRING(f, "proc {");
	n->p1->_print(n->p1);
	PTG_OUTPUT_STRING(f, "} {nodeList} {\n");
	PTG_OUTPUT_STRING(f, "  return [filterType ");
	n->p2->_print(n->p2);
	PTG_OUTPUT_STRING(f, " $nodeList]\n");
	PTG_OUTPUT_STRING(f, "}");
}

#ifdef PROTO_OK
PTGNode PTGConditionFct(PTGNode p1, PTGNode p2)
#else
PTGNode PTGConditionFct(p1, p2)
PTGNode p1;
PTGNode p2;
#endif
{
	_PPTGConditionFct n;

	n = (_PPTGConditionFct)MALLOC(sizeof(struct _SPTGConditionFct));
	n->_print = (_PTGProc)_PrPTGConditionFct;
	n->p1 = p1;
	n->p2 = p2;
	return (PTGNode)n;
}

/* Implementation of Pattern Including */

typedef struct _SPTGIncluding{
	_PTGProc _print;
	PTGNode p1;
} * _PPTGIncluding;

#ifdef PROTO_OK
static void _PrPTGIncluding(_PPTGIncluding n)
#else
static void _PrPTGIncluding(n)
	_PPTGIncluding n;
#endif
{
	PTG_OUTPUT_STRING(f, "INCLUDING ");
	n->p1->_print(n->p1);
}

#ifdef PROTO_OK
PTGNode PTGIncluding(PTGNode p1)
#else
PTGNode PTGIncluding(p1)
PTGNode p1;
#endif
{
	_PPTGIncluding n;

	n = (_PPTGIncluding)MALLOC(sizeof(struct _SPTGIncluding));
	n->_print = (_PTGProc)_PrPTGIncluding;
	n->p1 = p1;
	return (PTGNode)n;
}

/* Implementation of Pattern Id */

typedef struct _SPTGId{
	_PTGProc _print;
	int p1;
} * _PPTGId;

#ifdef PROTO_OK
static void _PrPTGId(_PPTGId n)
#else
static void _PrPTGId(n)
	_PPTGId n;
#endif
{
	PtgOutId(f, n->p1);
}

#ifdef PROTO_OK
PTGNode PTGId(int p1)
#else
PTGNode PTGId(p1)
int p1;
#endif
{
	_PPTGId n;

	n = (_PPTGId)MALLOC(sizeof(struct _SPTGId));
	n->_print = (_PTGProc)_PrPTGId;
	n->p1 = p1;
	return (PTGNode)n;
}

/* Implementation of Pattern AsIs */

typedef struct _SPTGAsIs{
	_PTGProc _print;
	char * p1;
} * _PPTGAsIs;

#ifdef PROTO_OK
static void _PrPTGAsIs(_PPTGAsIs n)
#else
static void _PrPTGAsIs(n)
	_PPTGAsIs n;
#endif
{
	PTG_OUTPUT_STRING(f, n->p1);
}

#ifdef PROTO_OK
PTGNode PTGAsIs(char * p1)
#else
PTGNode PTGAsIs(p1)
char * p1;
#endif
{
	_PPTGAsIs n;

	n = (_PPTGAsIs)MALLOC(sizeof(struct _SPTGAsIs));
	n->_print = (_PTGProc)_PrPTGAsIs;
	n->p1 = p1;
	return (PTGNode)n;
}

/* Implementation of Pattern Numb */

typedef struct _SPTGNumb{
	_PTGProc _print;
	int p1;
} * _PPTGNumb;

#ifdef PROTO_OK
static void _PrPTGNumb(_PPTGNumb n)
#else
static void _PrPTGNumb(n)
	_PPTGNumb n;
#endif
{
	PTG_OUTPUT_INT(f, n->p1);
}

#ifdef PROTO_OK
PTGNode PTGNumb(int p1)
#else
PTGNode PTGNumb(p1)
int p1;
#endif
{
	_PPTGNumb n;

	n = (_PPTGNumb)MALLOC(sizeof(struct _SPTGNumb));
	n->_print = (_PTGProc)_PrPTGNumb;
	n->p1 = p1;
	return (PTGNode)n;
}

/* Implementation of Pattern CString */

typedef struct _SPTGCString{
	_PTGProc _print;
	char * p1;
} * _PPTGCString;

#ifdef PROTO_OK
static void _PrPTGCString(_PPTGCString n)
#else
static void _PrPTGCString(n)
	_PPTGCString n;
#endif
{
	CPtgOutstr(f, n->p1);
}

#ifdef PROTO_OK
PTGNode PTGCString(char * p1)
#else
PTGNode PTGCString(p1)
char * p1;
#endif
{
	_PPTGCString n;

	n = (_PPTGCString)MALLOC(sizeof(struct _SPTGCString));
	n->_print = (_PTGProc)_PrPTGCString;
	n->p1 = p1;
	return (PTGNode)n;
}

/* Implementation of Pattern CChar */

typedef struct _SPTGCChar{
	_PTGProc _print;
	int p1;
} * _PPTGCChar;

#ifdef PROTO_OK
static void _PrPTGCChar(_PPTGCChar n)
#else
static void _PrPTGCChar(n)
	_PPTGCChar n;
#endif
{
	CPtgOutchar(f, n->p1);
}

#ifdef PROTO_OK
PTGNode PTGCChar(int p1)
#else
PTGNode PTGCChar(p1)
int p1;
#endif
{
	_PPTGCChar n;

	n = (_PPTGCChar)MALLOC(sizeof(struct _SPTGCChar));
	n->_print = (_PTGProc)_PrPTGCChar;
	n->p1 = p1;
	return (PTGNode)n;
}

/* Implementation of Pattern PString */

typedef struct _SPTGPString{
	_PTGProc _print;
	char * p1;
} * _PPTGPString;

#ifdef PROTO_OK
static void _PrPTGPString(_PPTGPString n)
#else
static void _PrPTGPString(n)
	_PPTGPString n;
#endif
{
	PPtgOutstr(f, n->p1);
}

#ifdef PROTO_OK
PTGNode PTGPString(char * p1)
#else
PTGNode PTGPString(p1)
char * p1;
#endif
{
	_PPTGPString n;

	n = (_PPTGPString)MALLOC(sizeof(struct _SPTGPString));
	n->_print = (_PTGProc)_PrPTGPString;
	n->p1 = p1;
	return (PTGNode)n;
}

/* Implementation of Pattern Seq */

typedef struct _SPTGSeq{
	_PTGProc _print;
	PTGNode p1;
	PTGNode p2;
} * _PPTGSeq;

#ifdef PROTO_OK
static void _PrPTGSeq(_PPTGSeq n)
#else
static void _PrPTGSeq(n)
	_PPTGSeq n;
#endif
{
	n->p1->_print(n->p1);
	n->p2->_print(n->p2);
}

#ifdef PROTO_OK
PTGNode PTGSeq(PTGNode p1, PTGNode p2)
#else
PTGNode PTGSeq(p1, p2)
PTGNode p1;
PTGNode p2;
#endif
{
	_PPTGSeq n;

	n = (_PPTGSeq)MALLOC(sizeof(struct _SPTGSeq));
	if(p1 == PTGNULL && p2 == PTGNULL)
		return PTGNULL;

	n->_print = (_PTGProc)_PrPTGSeq;
	n->p1 = p1;
	n->p2 = p2;
	return (PTGNode)n;
}

/* Implementation of Pattern CommaSeq */

typedef struct _SPTGCommaSeq{
	_PTGProc _print;
	PTGNode p1;
	PTGNode p2;
} * _PPTGCommaSeq;

#ifdef PROTO_OK
static void _PrPTGCommaSeq(_PPTGCommaSeq n)
#else
static void _PrPTGCommaSeq(n)
	_PPTGCommaSeq n;
#endif
{
	n->p1->_print(n->p1);
	if (n->p1 != PTGNULL && n->p2 != PTGNULL)
	{
		PTG_OUTPUT_STRING(f, ", ");
	}
	n->p2->_print(n->p2);
}

#ifdef PROTO_OK
PTGNode PTGCommaSeq(PTGNode p1, PTGNode p2)
#else
PTGNode PTGCommaSeq(p1, p2)
PTGNode p1;
PTGNode p2;
#endif
{
	_PPTGCommaSeq n;

	n = (_PPTGCommaSeq)MALLOC(sizeof(struct _SPTGCommaSeq));
	if(p1 == PTGNULL && p2 == PTGNULL)
		return PTGNULL;

	n->_print = (_PTGProc)_PrPTGCommaSeq;
	n->p1 = p1;
	n->p2 = p2;
	return (PTGNode)n;
}

/* Implementation of Pattern Eol */

typedef struct _SPTGEol{
	_PTGProc _print;
	PTGNode p1;
} * _PPTGEol;

#ifdef PROTO_OK
static void _PrPTGEol(_PPTGEol n)
#else
static void _PrPTGEol(n)
	_PPTGEol n;
#endif
{
	n->p1->_print(n->p1);
	PTG_OUTPUT_STRING(f, "\n");
}

#ifdef PROTO_OK
PTGNode PTGEol(PTGNode p1)
#else
PTGNode PTGEol(p1)
PTGNode p1;
#endif
{
	_PPTGEol n;

	n = (_PPTGEol)MALLOC(sizeof(struct _SPTGEol));
	n->_print = (_PTGProc)_PrPTGEol;
	n->p1 = p1;
	return (PTGNode)n;
}


/* -------------------------------------------------------- */
/*                  Default Output Functions                */
/* -------------------------------------------------------- */


#ifdef PROTO_OK
void _PTGPrintInt(PTG_OUTPUT_FILE file, int param)
#else
void _PTGPrintInt(file, param)
	PTG_OUTPUT_FILE file; int param;
#endif
{    /* used for short and int */
	sprintf(buffer,"%d",param);
	PTG_OUTPUT_STRING(file,buffer);
}

#ifdef PROTO_OK
void _PTGPrintLong(PTG_OUTPUT_FILE file, long param)
#else
void _PTGPrintLong(file, param)
	PTG_OUTPUT_FILE file; long param;
#endif
{
	sprintf(buffer,"%ld",param);
	PTG_OUTPUT_STRING(file,buffer);
}

#ifdef PROTO_OK
void _PTGPrintDouble(PTG_OUTPUT_FILE file, double param)
#else
void _PTGPrintDouble(file, param)
	PTG_OUTPUT_FILE file; double param;
#endif
{    /* used for float and double */
	sprintf(buffer,"%g",param);
	PTG_OUTPUT_STRING(file,buffer);
}

#ifdef PROTO_OK
void _PTGPrintChar(PTG_OUTPUT_FILE file, char param)
#else
void _PTGPrintChar(file, param)
	PTG_OUTPUT_FILE file; char param;
#endif
{
	buffer[0] = param;
	buffer[1] = 0;
	PTG_OUTPUT_STRING(file,buffer);
}

