# ----------------------------------------------------------------------
# Routinen zur Anzeige von Standarddialogen
# ----------------------------------------------------------------------
# $Id: CDlg.tcl,v 1.9 2005/02/02 09:41:37 cschmidt Exp $

# nderungen
#
# -*-courier-medium-r-normal-*-12-*-*-*-*-*-*-*  ->[font::get "_textField"]
# "Courier -10" -> [font::get "_textField"]


namespace eval ::CDlg {
}


lappend "::main::initProc" "::CDlg::Init"

proc ::CDlg::Init {} {
    global tcl_platform

    # Text to appear as default in the Ok-Button of a ::CDlg::Information
    option add *Info*ok.text "Ok" widgetDefault

    # Text to appear as default in the Ok-Button of a ::CDlg::Error
    option add *Error*ok.text "Ok" widgetDefault

    # Make background white and foreground black for some widgets to
    # make them appear less tcltk-like and more natural.

    option add *Entry.background white widgetDefault
    option add *Entry.highlightbackground blue widgetDefault
    option add *Listbox.background white widgetDefault
    option add *Listbox.highlightbackground blue widgetDefault
    option add *Text.background white widgetDefault
    option add *Text.highlightbackground blue widgetDefault
    if {$tcl_platform(platform) == "unix"} {
	option add *Radiobutton.selectColor black widgetDefault
	option add *Checkbutton.selectColor black widgetDefault
	option add *Menu.selectColor black widgetDefault
    }

    # Raise the dialog if button is pressed anywhere (grab!) on the
    # screen. Requires a matching bindtags list.

    bind CDlgModalDialog <ButtonPress> {
	wm deiconify %W
	raise %W
    }
}

proc ::CDlg::Create {class {title ""} {win "auto"}} {
    # Creates an empty dialog, consisting of two frames, an info-frame
    # and a button-frame separated by a horizontal 3D-line.
    #
    # class      Classname of dialog, also used as window title, if title
    #            is empty.
    #            Under the classname, the user can modify window colors
    #            fonts, etc. by entering settings to his .Xdefaults-file.
    # win        Pathname of the window, optionally computed automagically.

    if {$win == "auto"} {
	set count 0
	set win ".dialog[incr count]"
	while {[winfo exists $win]} {
	    set win ".dialog[incr count]"
	}   
    }

    if {$title == ""} {
	set title $class
    }

    toplevel $win -class $class
    frame $win.info
    pack $win.info -expand yes -fill both -padx 2 -pady 2
    frame $win.sep -height 2 -borderwidth 1 -relief sunken
    pack $win.sep -fill x -pady 4
    frame $win.controls 
    pack $win.controls -fill x -padx 4 -pady 4

    wm title $win $title
    wm group $win .
    
    after idle [format {
	update idletasks
	wm minsize %s [winfo reqwidth %s] [winfo reqheight %s]
    } $win $win $win]
    	
    return $win
    
}

proc ::CDlg::Info {win} {
    # Return pathname of info-pane in a dialog created by CDlgCreate

    return "$win.info"
}

proc ::CDlg::Controls {win} {
    # Return pathname of controls-pane in a dialog created by CDlgCreate

    return "$win.controls"
}

proc ::CDlg::Center {window {parent ""}} {
    # Positions a window in the center of an existing one. Default is
    # Center of the screen.
    if {$parent == ""} {
	set X 0
	set Y 0
	set W [winfo screenwidth .]
	set H [winfo screenheight .]
    } else {
	set W [winfo width $parent]
	set H [winfo height $parent]
	set X [winfo rootx $parent]
	set Y [winfo rooty $parent]
    }

    wm withdraw $window
    update idletasks

    set w [winfo reqwidth $window]
    set h [winfo reqheight $window]
    
    wm geometry $window [format "+%d+%d" \
	    [expr $X + ($W - $w)/2] [expr $Y + ($H - $h)/2] ]
    wm deiconify $window
    
}


proc CDlg::grabNoWait {win} {
    wm deiconify $win
    if {[lsearch [bindtags $win] CDlgModalDialog] < 0} {
	bindtags $win [linsert [bindtags $win] 0 CDlgModalDialog]
    }
    tkwait visibility $win
    grab $win
    update idletasks
    focus [tk_focusNext $win]
}


proc ::CDlg::Wait { win {foc ""} } {
    # Waits for User Interaction with $win. Some routine must
    # destroy the window what will cause this procedure to terminate.
    # foc is the first input field in the dialog.

    set prevfocus [focus]
    wm deiconify $win    
    if {[lsearch [bindtags $win] CDlgModalDialog] < 0} {
	bindtags $win [linsert [bindtags $win] 0 CDlgModalDialog]
    }

    set prevgrab [grab current]
    if {$prevgrab != ""} {
	grab release $prevgrab
    }

    tkwait visibility $win
    grab $win
    if {"$foc" != ""} {
	focus $foc
    } else {
	focus $win
    }

    tkwait window $win
    catch {
        # catch is nessesary for windows. It seems, as if
        # $win == $prevfocus or something like that
        focus $prevfocus
    }
    if {$prevgrab != ""} {
	grab $prevgrab
    }
}
 
proc ::CDlg::NotifyDialog { class title bitmap message } {
    # Internal Routine to build Notification Dialogs. Called from
    # Information, Error, etc. The label of the button is not set and
    # must be assigned using an option.

    set win [Create $class $title]

    set info [Info $win]
    label $info.icon -bitmap $bitmap
    pack $info.icon -side left -padx 8 -pady 8
    label $info.mesg -text $message -wraplength 4i -justify left
    pack $info.mesg -side right -expand yes -fill both -padx 8 -pady 8
    
    set cntrls [Controls $win]
    button $cntrls.ok -text "Ok" -command "\
	    grab release $win\n\
	    destroy $win"
    pack $cntrls.ok -pady 4 -side left -expand yes

    after idle "CDlg::Center $win"

    return $win
}

# Some Ready to use Dialog Procedures.

proc ::CDlg::Information1 { message {title "Information"}} {
    tk_messageBox -icon info -title $title -message $message -type ok
}


proc ::CDlg::Information { message {title "Information"}} {
    # A Popup-Window with an i-icon that displays message and has
    # an ok-button.
    
    set win [NotifyDialog "Info" $title info $message]

    Wait $win
}


proc ::CDlg::Error1 { message {title "Error"}} {
    tk_messageBox -icon error -title $title -message $message -type ok
}


proc ::CDlg::Error { message {title "Error"}} {

    set win [NotifyDialog "Error" $title error $message]
    wm title $win "$title"

    Wait $win
}


proc ::CDlg::Question {question {title Question}} {
    return [tk_messageBox -title $title -icon question -type yesnocancel -message $question]
}


proc ::CDlg::Confirm { message {ok "OK"} {cancel "Cancel"}} {
    # Displays a dialog containing the message. The user chooses between
    # two buttons. The text on the button the user pressed is returned.

    variable result ""

    set CDlg(Confirm) ""

    set win [NotifyDialog "Confirm" "Confirm" questhead $message]
    set cntrls [Controls $win]
    $cntrls.ok configure -text $ok
    $cntrls.ok configure -command [format {
	grab release %s
	destroy %s
	set ::CDlg::result %s
    } $win $win $ok]

    button $cntrls.cancel -text $cancel -command [format {
	grab release %s
	destroy %s
	set ::CDlg::result "%s"
    } $win $win $cancel]
	
    pack $cntrls.cancel -side left -expand yes
        
    Wait $win

    return $result
}


proc CDlg::createDialog {title okCmd {cancelCmd ""}} {
    set win [::CDlg::Create myDialog $title]
    set info [::CDlg::Info $win]
    set controls [::CDlg::Controls $win]

    # Fill in control buttons first. 
    # If anything goes wrong, the window can be dismissed

    button $controls.ok -text "Ok" -command $okCmd
    bind $win <Return> "$controls.ok invoke"
    pack $controls.ok -side left -expand true

    if ![empty $cancelCmd] {
        button $controls.cancel -text "Cancel" -command $cancelCmd
        bind $win <Escape> "$controls.cancel invoke"
        pack $controls.cancel -side left -expand true
    }

    return $win
}


proc CDlg::waitDialog {win} {
    ::CDlg::Center $win
    ::CDlg::Wait $win $focus
}


proc CDlg::stdInputDialog {title elemList dataVar} {
    # data-Array:
    #  - XXX,value
    #  - XXX,label
    #  - XXX,editor (z.B. Text, Entry, ...)

    
    upvar $dataVar data
    set resultVar "::CDlg::resultVar[newint]"
    set dialog [createDialog $title "set $resultVar 1" "set $resultVar 0"]
    bind $dialog <Destroy> "set $resultVar 0"
    set info [::CDlg::Info $dialog]
    
    # Compute Maximum Label Width
    set width 0
    foreach elem $elemList {
        set lwidth [expr [string length $data($elem,label)] + 2]
        if {$width < $lwidth} { set width $lwidth }
    }

    # Insert DialogComponents
    foreach elem $elemList {
        set editor [lindex $data($elem,editor) 0]
        set options [lrange $data($elem,editor) 1 end]
        lappend options -label $data($elem,label) -labelWidth $width
        eval dcWithLabel::create $info.$elem "dc$editor" $options
        pack $info.$elem -fill both -expand [oop::call shouldExpand $info.$elem]
        oop::call setValue $info.$elem $data($elem,value)
    }
    Center $dialog
    grabNoWait $dialog

    set finish 0

    # Wait and check for errors
    while { !$finish } {
        tkwait variable $resultVar
        set result [set $resultVar]
        set finish 1
        if $result {
            foreach elem $elemList {
                if ![empty $data($elem,check)] {
                    set err [eval $data($elem,check) [list [oop::call getValue $info.$elem]]]
                    if ![empty $err] {
                        Error [i18n::text "Invalid value of attribute '%s':\n%s" $data($elem,label) $err]
                        set finish 0
                        break
                    }
                }
            }
        }
    }

    if $result {
        foreach elem $elemList {
            set data($elem,value) [oop::call getValue $info.$elem]
        }
    }
    destroy $dialog
    return $result
}


proc CDlg::getString {title varName} {
    upvar $varName var
    set data(a,value) ""
    set data(a,label) ""
    set data(a,editor) Entry
    set data(a,check) ""
    set result [stdInputDialog $title "a" data]
    if $result {
        set var $data(a,value)
        return 1
    } else {
        return 0
    }    
}



proc CDlg::mkScrollableY {frame content} {
    scrollbar $frame.yscroll -orient vertical -command "$frame.$content yview" -takefocus 0
    pack $frame.yscroll -side right -fill y

    $frame.$content configure  -yscrollcommand [list $frame.yscroll set]
    pack $frame.$content -fill both -expand true
}


proc CDlg::mkScrollableXY {frame content} {
    # frame: pfadname eines frames
    # content: name des Inhalts, der mit scrollbars versehen werden soll.
    # $frame.$content ist also ein gueltiger Pfadname.

    frame $frame.bottom
    pack $frame.bottom -side bottom -fill x

    scrollbar $frame.bottom.xscroll -orient horizontal -command "$frame.$content xview" -takefocus 0

    set size [expr "[$frame.bottom.xscroll cget -bd] + [$frame.bottom.xscroll cget -highlightthickness]"]
    set pad [expr "[$frame.bottom.xscroll cget -width] + 2 * $size"]
    frame $frame.bottom.pad -width $pad -height $pad
    pack $frame.bottom.pad -side right

    pack $frame.bottom.xscroll -side bottom -fill x

    scrollbar $frame.yscroll -orient vertical -command "$frame.$content yview" -takefocus 0
    pack $frame.yscroll -side right -fill y

    $frame.$content configure  -yscrollcommand "$frame.yscroll set" -xscrollcommand "$frame.bottom.xscroll set"

    pack $frame.$content -side right -fill both -expand 1
}
