
namespace eval ::specialDialogView {}


proc specialDialogView::updateTitle {id} {
}


proc specialDialogView::new {root viewName} {
    #set id [abstractView::new $root]
    set id [oop::new]
    oop::addClass $id "specialDialogView"
    oop::data $id
  
    set data(root) $root
    set data(file) [file::getFileOfNode $data(root)]
    set data(fileRoot) [c::getRoot $data(root)]
    set data(objects) ""
    set title "Edit [c::getClassName $root] properties"
    set data(toplevel) [CDlg::createDialog $title [list specialDialogView::saveAndClose $id] [list specialDialogView::cancel $id]]
    set data(workspace) [CDlg::Info $data(toplevel)]
    bind $data(workspace) <Destroy> [list ::specialDialogView::delete $id]
    set data(viewName) $viewName
    set data(viewType) $viewName
    set data(hidden) 1
    wm withdraw $data(toplevel)
    set data(initialEditState) [change::getEditState $data(fileRoot)]
    set data(normalTermination) 0

    file::addView $data(file) $id
    return $id
}

proc specialDialogView::update {id} {
    oop::data $id
    abstractView::update $id
    renewStart $id
    c::startProcessor $data(viewName) $data(root)
    renewEnd $id
}


proc specialDialogView::isReusable {id node type} {
    oop::data $id
    return [expr [equal $node $data(root)] && [equal $type $data(viewType)]]
}


proc specialDialogView::getFile {id} {
    oop::data $id
    return $data(file)
}


proc specialDialogView::delete {id} {
    # this function is automatically executed, after "destroy $data(toplevel)"
    oop::data $id
    
    file::removeView [getFile $id] $id
    if !$data(normalTermination) {
        # cancel button or window close button
        change::restoreState $data(fileRoot) $data(initialEditState)
    }
    oop::delete $id
}    


proc specialDialogView::beforeDeletion {id node} {
}


proc specialDialogView::save {id} {
    oop::data $id
    foreach obj $data(objects) {
        oop::call writeValue $data(workspace).[join $obj "-"]
    }
}


proc specialDialogView::saveAndClose {id} {
    oop::data $id
    change::clipCheckpoints $data(fileRoot) $data(initialEditState)
    save $id
    edit::forceUpdate $data(root)

    set data(normalTermination) 1
    destroy $data(toplevel)
}


proc specialDialogView::cancel {id} {
    oop::data $id
    destroy $data(toplevel)
}


proc specialDialogView::read {id} {
    oop::data $id
    foreach obj $data(objects) {
        oop::call readValue $data(workspace).[join $obj "-"]
    }    
}


# --- renew handling ---------------------------------------------------------------


proc specialDialogView::renewStart {id} {
    oop::data $id
    set data(renewObjects) ""
    set ::specialDialogView::id $id
}


proc specialDialogView::renewEnd {id} {
    unset ::specialDialogView::id
    oop::data $id
    set newObjects ""
    foreach {context kind optionList} $data(renewObjects) {
        set obj [list $context $kind]
        if ![elementOf $obj $data(objects)] {
            eval dcWithLabel::create $data(workspace).[join $obj "-"] $kind $optionList
            pack $data(workspace).[join $obj "-"] -fill both -expand [oop::call shouldExpand $data(workspace).[join $obj "-"]]
        } else {
            eval dcWithLabel::renew $data(workspace).[join $obj "-"] $optionList
        }
        lappend newObjects $obj
    }
    foreach old $data(objects) {
        if ![elementOf $old $newObjects] {
            destroy $data(workspace).[join $old "-"]
        }
    }
    set data(objects) $newObjects

    # reihenfolge anpassen
    set curr [lindex $newObjects 0]
    foreach obj [lrange $newObjects 1 end] {
        pack configure $data(workspace).[join $obj "-"] -after $data(workspace).[join $curr "-"]
        set curr $obj
    }

    read $id

    if $data(hidden) {
        set data(hidden) 0
        CDlg::Center $data(toplevel)
        CDlg::grabNoWait $data(toplevel)
    }
}


proc specialDialogView::renewCreate {context kind optionList} {
    if { [equal $kind "None"] } return

    set id $::specialDialogView::id
    oop::data $id
    set kind "dc$kind"
    if [c::isValue $context] {
        lappend optionList -setCmd [list edit::setValue $context] -getCmd [list c::getValue $context]
    }
    if [c::isReference $context] {
        set values [treedraw::getVLPointerList $context]
        lappend optionList -setCmd [list edit::setValue $context] -getCmd [list c::getValue $context] -values $values
    }
    if [c::isAggregate $context] {
        #set values "(empty) {}"
        #foreach value [c::getAcceptedClassesForChild $context] {
        #    lappend values $value $value
        #}
        lappend optionList -setCmd [list edit::setClassOfChild $context] -getCmd [list c::getClassOfChild $context]
    }

    lappend optionList -onInput "specialDialogView::save $id"
    lappend data(renewObjects) $context $kind $optionList
}

