
namespace eval ::syncutil {}
namespace eval ::synchook {}

proc syncutil::syncProgram {obj} {
    set context [c::getContext $obj]
    if [procExists synchook::$context] {
        synchook::$context $obj
    }
    foreach sub [c::getSubElements $obj] {
        syncProgram $sub
    }
}


proc syncutil::viewRepCreationProc {viewName model} {
    # returns the representation of object $model in views of type $viewName
    assert { [c::isLanguageObject $model] }

    set className [c::getRepClass [c::getClassName $model] $viewName]
    return [change::createObject $className [change::getFile $model]]
}


proc syncutil::createMissing {defList repList refAttr repCreationProc} {
    # find missing
    foreach def [c::getList {$defList.CHILDREN}] {
        set missing($def) 1
    }    
    foreach def [c::getList {$repList.CHILDREN.$refAttr.VALUE}] {
        set missing($def) 0
    }

    # createMissing
    foreach def [c::getList {$defList.CHILDREN}] {
        if $missing($def) {
            set newobj [eval $repCreationProc $def]
            change::insert [c::icBelow $repList] $newobj
            change::setValue [c::get $newobj.$refAttr] $def
        }
    }
}

proc syncutil::deleteSuperfluous {defList repList refAttr} {
    
    # find superfluous
    set superfluous ""
    foreach rep [c::getList {$repList.CHILDREN}] {
        if [nullptr [c::get $rep.$refAttr.VALUE]] {
            lappend superfluous $rep
        } elseif { ![equal [c::get $rep.$refAttr.VALUE.PARENT] $defList] } {
            lappend superfluous $rep
        }
    }

    # delete superfluous
    foreach obj $superfluous {
        change::delete $obj
    }
}


proc syncutil::rightOrder {a b refAttr order} {
    # private
    return [expr [lsearch $order [c::get $a.$refAttr.VALUE]] <= [lsearch $order [c::get $b.$refAttr.VALUE]]]
}


proc syncutil::order {defList repList refAttr} {

    # find elements to re-insert
    set reInsert ""
    set order [c::getList {$defList.CHILDREN}]
    set prev ""
    foreach rep [c::getList {$repList.CHILDREN}] {
        if { ![empty $prev] && ![rightOrder $prev $rep $refAttr $order] } {
            change::extract $rep
            lappend reInsert $rep
        } else {
            set prev $rep
        }
    }

    # re-insert
    foreach obj $reInsert {
        foreach rep [c::getList {$repList.CHILDREN}] {
            if { ![rightOrder $rep $obj $refAttr $order] } {
                change::insert $rep $obj
                break
            }
        }
    }
}


proc syncutil::computeModelToRep {repRoot modelToRepName} {
    upvar $modelToRepName modelToRep
    catch {unset modelToRep}
    set modelToRep(0x0) 0x0
    c::traverseTree $repRoot isStructureObject rep {
        if [c::getModel $rep model] {
            lappend modelToRep($model) $rep
            # former lappend2
        }
    }
}


proc syncutil::syncValue {obj} {
    # precondition: obj is a value object
    if ![getTerminalOption sync $obj] {
        return
    }
    if [c::getModel $obj model] {
        change::setValue $obj [c::get $model.VALUE]
    }
}


proc syncutil::syncReference {obj modelToRepName} {
    # precondition: obj is a reference object
    if ![getTerminalOption sync $obj] {
        return
    }
    upvar $modelToRepName modelToRep    
    if [c::getModel $obj model] {
        if ![equal [c::getOrNull $model.VALUE] [c::getOrNull $obj.VALUE.SEMANTIC]] {
            if ![info exists modelToRep([c::get $model.VALUE])] {
                set newValue 0x0
            } else {
                set newValue [lindex $modelToRep([c::get $model.VALUE]) 0]
            }
            change::setValue $obj $newValue
        }
    }
}

proc syncutil::moveToRightList {obj modelToRepName} {
    # precondition: obj is LanguageObject
    # postcondition: obj is in the "right" aggregate list, determined by the model
    if ![getLanguageObjectOption moveToRightList $obj] {
        return
    }
    upvar $modelToRepName modelToRep    
    if [c::getModel $obj model] {
        if ![equal [c::get $obj.PARENT.PARENT.modelRef.VALUE] [c::get $model.PARENT.PARENT]] {
            set destList [lindex $modelToRep([c::get $model.PARENT]) 0]
            change::insert [c::icBelow $destList] $obj
        }
    }
}      

proc syncutil::createMissingRep {rep viewName} {
    # precondition: rep is an aggreate object
    # creates missing subelements of "rep".
    if ![getAggregateOption createMissingRep $rep] {
        return
    }
    if [c::getModel $rep model] {
        createMissing $model $rep "modelRef" [list viewRepCreationProc $viewName]
    }
}


proc syncutil::orderRep {rep} {
    # precondition: rep is an aggreate object
    if ![getAggregateOption orderRep $rep] {
        return
    }
    if [c::getModel $rep model] {
        order $model $rep modelRef
    }
}


proc syncutil::removeAbandoned {rep} {
    # precondition: rep is LanguageObject
    if ![getLanguageObjectOption removeAbandoned $rep] {
        return 0
    }
    if [c::getSubElement $rep modelRef -var modelRef] {
        if [nullptr [c::getValue $modelRef]] {
            change::delete $rep
            return 1
        }
    }
    return 0
}

proc syncutil::removeMultipleRep {rep modelToRepName} {
    # precondition: rep is LanguageObject
    if ![getLanguageObjectOption removeMultipleRep $rep] {
        return 0
    }
    upvar $modelToRepName modelToRep
    if [c::getModel $rep model] {
        if [lsearch -exact $modelToRep($model) $rep]>0 {
            change::delete $rep
            return 1
        }
    }
    return 0
}

proc syncutil::syncView {viewName view} {

    # delete
    c::traverseTree $view isLanguageObject obj {
        if [removeAbandoned $obj] {
            set obj ""
        }
    }
    computeModelToRep $view modelToRep
    c::traverseTree $view isLanguageObject obj {
        if [removeMultipleRep $obj modelToRep] {
            set obj ""
        }
    }

    # move
    computeModelToRep $view modelToRep
    c::traverseTree $view isLanguageObject obj {
        if ![equal $obj $view] {
            moveToRightList $obj modelToRep
        }
    }

    # create
    c::traverseTree $view isAggregate obj {
        createMissingRep $obj $viewName
    }

    # non-structural
    c::traverseTree $view isAggregate obj {
        orderRep $obj
    }
    c::traverseTree $view isValue obj {
        syncValue $obj
    }
    computeModelToRep $view modelToRep
    c::traverseTree $view isReference obj {
        syncReference $obj modelToRep
    }
}


proc syncutil::getSynchookVar {name} {
    if [info exists ::synchook::$name] {
        return [namespace eval ::synchook "variable $name; set $name"]
    } else {
        return ""
    }
}

proc syncutil::getTerminalOption {option obj} {
    set args [getSynchookVar [c::getContext $obj]]
    handleArgs "-sync 1"
    return $arg($option)
}

proc syncutil::getLanguageObjectOption {option obj} {
    set args [getSynchookVar [c::getClassName $obj]]
    handleArgs "-moveToRightList 1 -removeAbandoned 1 -removeMultipleRep 1"
    return $arg($option)
}

proc syncutil::getAggregateOption {option obj} {
    set args [getSynchookVar [c::getContext $obj]]
    handleArgs "-createMissingRep 1 -orderRep 1"
    return $arg($option)
}

