
namespace eval treePath {}


proc treePath::eval {startNode expr} {
    variable exprCache
    regsub {^[\n ]*} $expr "" expr
    regsub {[\n ]*$} $expr "" expr

    if [getCacheValue $startNode $expr result] {
        return $result
    }

    if ![info exists exprCache($expr)] {
        loadExprCode $expr
        set exprCache($expr) $treePath::mainFct
    }

    if [catch {
        set result [$exprCache($expr) [list $startNode]]
        setCacheValue $startNode $expr $result
    } errMessage] {
        set savedInfo $::errorInfo
        error "Error while evaluating '$expr': $errMessage" $savedInfo
    }
    return $result    
}

proc treePath::resetCache {} {
    variable valueCache
    catch {unset valueCache}
}

proc treePath::getCacheValue {node expr resultName} {
    variable valueCache
    upvar $resultName result
    if [info exists valueCache($node,$expr)] {
        set result $valueCache($node,$expr)
        return 1
    } else {
        return 0
    }
}

proc treePath::setCacheValue {node expr value} {
    variable valueCache
    set valueCache($node,$expr) $value
}

proc treePath::loadExprCode {expr} {
    set code [exec "$main::appHome/treePathProcessor.exe" << $expr]
    uplevel #0 [list namespace eval treePath $code]
}

proc treePath::union {nodeList1 nodeList2} {
    foreach elem $nodeList1 {
        set arr($elem) 1
    }
    foreach elem $nodeList2 {
        set arr($elem) 1
    }
    return [array names arr]
}

proc treePath::filterType {type nodeList} {
    if [empty $type] {
        return $nodeList
    }
    set result ""
    foreach node $nodeList {
        if [equal [c::getContext $node] $type] {
            lappend result $node
        }
    }
    return $result
}

proc treePath::fieldSel {field nodeList} {

    # Ersetzen veralteter Bezeichnungen
    switch $field {
        CHILD { set field CHILDREN }
    }

    switch -glob $field {
        PARENT {
            foreach node $nodeList {
                if ![nullptr $node] {
                    set parent [c::getParent $node]
                    if ![nullptr $parent] {
                        set arr($parent) 1
                    }
                }
            }
            set result [array names arr]
        }
        VALUE {
            foreach node $nodeList {
                if { ![nullptr $node] && [c::isTerminal $node] } {
                    set arr([c::getValue $node]) 1
                }
            }
            set result [array names arr]
        }
        CHILDREN {
            set result ""
            foreach node $nodeList {
                if ![nullptr $node] {
                    set result [concat $result [c::getSubElements $node]]
                }
            }
        }
        IVALUE {
            if [empty $nodeList] {
                return ""
            }
            set root [c::getRoot [lindex $nodeList 0]]
            foreach node $nodeList {
                if ![nullptr $node] {
                    foreach elem [c::findReferences $node $root] {
                        set arr($elem) 1
                    }
                }
            }
            set result [array names arr]
        }
        "INCLUDING *" {
            foreach node $nodeList {
                set arr([c::findIncluding [lindex $field 1] $node]) 1
            }
            set result [array names arr]
        }
        SEMANTIC {
            foreach node $nodeList {
                if [c::getModel $node smElem] {
                    set arr($smElem) 1
                }
            }
            set result [array names arr]
        }
        SEMANTICREF {
            set result ""
            foreach node $nodeList {
                if [c::getSubElement $node "modelRef" -var sub] {
                    lappend result $sub
                }
            }
        }
        default {
            set result ""
            foreach node $nodeList {
                if ![c::isLanguageObject $node] {
                    error "Node class [c::getContext $node] has no Attribute '$field'"
                }
                if [c::getSubElement $node $field -var sub] {
                    lappend result $sub
                } else {
                    error "Node class [c::getContext $node] has no Attribute '$field'"
                }
            }
        }
    }
    return $result
}


proc treePath::EPS {nodeList} {
    return $nodeList    
}
