
# treeView
#
# Baumsicht auf das abstrakte Modell


namespace eval ::treeView {}


### Constructors

proc treeView::new {root} {
    set id [abstractView::new $root]
    oop::addClass $id "treeView"
    oop::data $id

    frame $data(toplevel).sCanvas
    pack $data(toplevel).sCanvas -fill both -expand 1

    canvas $data(toplevel).sCanvas.workspace -background white
    CDlg::mkScrollableXY $data(toplevel).sCanvas workspace

    set data(canvas) "$data(toplevel).sCanvas.workspace"
    set data(viewType) "Tree view"
    set data(selection) ""
    return $id
}


#### Instance methods


proc treeView::update {id} {
    oop::data $id
    abstractView::update $id
    $data(canvas) delete all
    paintStructure $id $data(root)
    foreach {x1 y1 x2 y2} [$data(canvas) bbox all] {
        $data(canvas) config -scrollregion "0 0 [expr $x2+10] [expr $y2+10]"
    }
}

proc treeView::highlightNode {id node} {
    oop::data $id

    set n [c::getParent $node]
    while { ![nullptr $n] } {
        set ::treeView::showDetail($id,$n) 1
        set n [c::getParent $n]
    }
    set data(selection) $node
    #edit::forceUpdate $node
    update $id
}


proc  treeView::toggleShowDetail {id elem} {
    set ::treeView::showDetail($id,$elem) [expr ! $::treeView::showDetail($id,$elem)]
    #edit::forceUpdate $elem
    update $id
    #c::propagateChange $elem
}


proc  treeView::paintItem {id elem color1 text1 color2 text2 showAlways} {
    oop::data $id
    if $showAlways {
        # do nothing
    } elseif { [llength [c::getSubElements $elem]]>0 } {
        if $::treeView::showDetail($id,$elem) {
            $data(canvas) create image $data(xpos) $data(ypos) -anchor w -image img::minus -tag showDetail-$elem
            $data(canvas) bind showDetail-$elem <Button-1> "treeView::toggleShowDetail $id $elem"
        } else {
            $data(canvas) create image $data(xpos) $data(ypos) -anchor w -image img::plus -tag showDetail-$elem
            $data(canvas) bind showDetail-$elem <Button-1> "treeView::toggleShowDetail $id $elem"
        }
    } else {
        $data(canvas) create image $data(xpos) $data(ypos) -anchor w -image img::leer
    }

    set tid [$data(canvas) create text [expr $data(xpos)+$data(icondist)] $data(ypos) -anchor w -text $text1 -fill $color1 -font $data(font) -tag elem-$elem]
    if { [equal $elem $data(selection)] } {
        $data(canvas) create rectangle [ $data(canvas) bbox $tid] -fill yellow -outline {}
        $data(canvas) raise $tid
        set data(selection) ""
    }

    set width [font measure $data(font) "$text1 "]
    $data(canvas) create text [expr $data(xpos)+$data(icondist)+$width] $data(ypos) -anchor w -text $text2 -fill $color2 -font $data(font) -tag elem-$elem
    $data(canvas) bind elem-$elem <Button-3> "treedraw::contextMenu $elem %X %Y"

    incr data(ypos) $data(deltay)
}



proc  treeView::paintStructureRecursive {id elem} {
    oop::data $id
    
    set attrColor darkgray
    set objColor blue
    set valueColor red
    
    if ![info exists ::treeView::showDetail($id,$elem)] {
        set ::treeView::showDetail($id,$elem) 0
    }

    if [c::isLanguageObject $elem] {
        set showAlways 0
        paintItem $id $elem $objColor [c::getClassName $elem] $objColor "([treedraw::getShortName $elem])" $showAlways
    } elseif [c::isValue $elem] {
        set showAlways 0
        paintItem $id $elem $attrColor "[c::getRole $elem]:" $valueColor [treedraw::getValue $elem] $showAlways
    } elseif [c::isReference $elem] {
        set showAlways 0
        paintItem $id $elem $attrColor "[c::getRole $elem]:" $valueColor "REF [treedraw::getValue $elem]" $showAlways
    } elseif [c::isAggregate $elem] {
        set showAlways 0
        paintItem $id $elem $attrColor "[c::getRole $elem]:" $valueColor "..." $showAlways
    } else {
        error "impossible branch"
        set showAlways 0
        paintItem $id $elem $attrColor "[c::getRole $elem] ([c::getClassName $elem])" $valueColor "" $showAlways
    }

    if { !$::treeView::showDetail($id,$elem) && !$showAlways } {
        return
    }

    if [c::isLanguageObject $elem] {
        set lineColor $objColor
    } else {
        set lineColor $attrColor
    }

    incr data(xpos) $data(deltax)
    set start [expr $data(ypos)-$data(deltay)/2]
    set end $start
    foreach sub [c::getSubElements $elem] {
        set end [expr $data(ypos)+[image height img::leer]/2]
        paintStructureRecursive $id $sub
    }
    set linex [expr $data(xpos)+[image width img::leer]/2]
    $data(canvas) lower [$data(canvas) create line $linex $start $linex $end -fill $lineColor -width 1] all
    incr data(xpos) -$data(deltax)
}


proc  treeView::paintStructure {id elem} {
    oop::data $id

    set data(font) [font::get "_treeView"]

    set fontx [font measure $data(font) "M"]
    set fonty [font metrics $data(font) -linespace]
    set iconx [image width img::leer]
    set icony [image height img::leer]

    set data(xpos) 10
    set data(ypos) 10
    set data(deltax) [expr 2*$fontx]
    set data(deltay) [expr $fonty*4/3]
    set data(icondist) [expr $iconx+$fontx/2]

    paintStructureRecursive $id $elem
}
