
namespace eval ::vec {
}

proc vec::round {vec gridSize} {
    assign "x y" $vec
    set x [expr (round($x) + $gridSize/2) / $gridSize * $gridSize]
    set y [expr (round($y) + $gridSize/2) / $gridSize * $gridSize]
    return "$x $y"
}

proc vec::add {vec1 vec2} {
    if { [llength $vec1] != [llength $vec2] } {
        error "invalid add: $vec1 + $vec2"
    }
    set result ""
    foreach i [range 0 [llength $vec1]] {
        lappend result [expr { [lindex $vec1 $i] + [lindex $vec2 $i] }]
    }
    return $result
}

proc vec::sub {vec1 vec2} {
    if { [llength $vec1] != [llength $vec2] } {
        error "invalid sub: $vec1 + $vec2"
    }
    set result ""
    foreach i [range 0 [llength $vec1]] {
        lappend result [expr { [lindex $vec1 $i] - [lindex $vec2 $i] }]
    }
    return $result
}

proc vec::move {box vec} {
    return [add $box "$vec $vec"]
}

proc vec::moveInside {box suroundingBox} {
    # changes the position of $box, so that it is completely inside $suroundingBox.

    assign "x y x2 y2" $box
    set w [expr $x2-$x]
    set h [expr $y2-$y]
    assign "sx1 sy1 sx2 sy2" $suroundingBox

    if { $x < $sx1 } {
        set x $sx1
    }
    if { $x > $sx2-$w } {
        set x [expr $sx2-$w]
    }
    if { $y < $sy1 } {
        set y $sy1
    }
    if { $y > $sy2-$h } {
        set y [expr $sy2-$h]
    }
    return "$x $y [expr $x+$w] [expr $y+$h]"
}



proc vec::distance {vec1 vec2} {
    assign "x1 y1" $vec1
    assign "x2 y2" $vec2
    return [expr "sqrt(pow($x2-$x1,2)+pow($y2-$y1,2))"]
}


proc vec::distancePointRect {point rect} {
    assign "bx1 by1 bx2 by2" $rect
    assign "x y" $point

    if {$bx1 < $x && $x < $bx2} {
        set dx 0
    } elseif {$x <= $bx1} {
        set dx [expr $bx1 - $x]
    } else {
        set dx [expr $x - $bx2]
    }

    if {$by1 < $y && $y < $by2} {
        set dy 0
    } elseif {$y <= $by1} {
        set dy [expr $by1 - $y]
    } else {
        set dy [expr $y - $by2]
    }
    set distance [expr sqrt($dx * $dx + $dy * $dy)]
    return $distance
}


proc vec::makeSquare {pos size} {
    assign "x y" $pos
    set r [expr $size/2]
    set result "[expr $x-$r] [expr $y-$r] [expr $x+$r] [expr $y+$r]"
    return $result    
}

proc vec::isInside {point region} {
    assign "x y" $point
    assign "x1 y1 x2 y2" $region
    return [expr ($x1 <= $x && $x <= $x2)  && ($y1 <= $y && $y <= $y2)]
}


proc vec::normalize {vector length} {
    assign "x y" $vector
    set l [expr sqrt($x*$x+$y*$y)]
    return [list [expr $length*$x/$l] [expr $length*$y/$l]]
}


proc vec::rotateRight90 {vector} {
    assign "x y" $vector
    return [list $y [expr -$x]]
}


# Funktionalitaet??? (Wird von canvasitems.tcl verwendet)
proc vec::transform {drawing rvec qvec shift} {
  foreach {primtype coords options} $drawing {
    foreach {x y} $coords {
      set x1 [lindex $rvec 0]
      set y1 [lindex $rvec 1]
      set x2 [lindex $qvec 0]
      set y2 [lindex $qvec 1]
      set sx [lindex $shift 0]
      set sy [lindex $shift 1]
      lappend newcoords \
        [expr $sx + $x*$x2 + $y*$x1] \
        [expr $sy + $x*$y2 + $y*$y1]
    }
    lappend newdrawing $primtype $newcoords $options
  }
  return $newdrawing
}
