
namespace eval xmlio {}

#package require xml
# Nicht benoetigt. Ist in der Datei tclxml2.5.tcl enthalten.


# Modul: xmlio.tcl
#
# Speichert und ldt Bume als XML-Datei.
# Zugriff auf die Baum-Datenstruktur erfolgt ber
# das Modul "c".


# ------------------------------------------------------------------ Save

# PRIVATE
proc xmlio::mask {string} {
    regsub -all "&" $string "\\&amp;" string
    regsub -all "\n" $string "\\&#10;" string
    regsub -all "<" $string "\\&lt;" string
    regsub -all ">" $string "\\&gt;" string
    regsub -all "\"" $string "\\&quot;" string
    return $string
}

# PRIVATE
proc xmlio::getFullName {nodePtr} {
    if [nullptr $nodePtr] {
        return 0x0
    } else {
        return [mask [basiccomp::getFullName $nodePtr]]
    }
}

# PRIVATE
proc xmlio::write {elem out prefix} {
    # writes tree representation of $elem, but excludes
    # elementes where "isLibraryAttribute" returns true.

    if [isLibraryAttribute $elem] {
        return
    }

    if [c::isLanguageObject $elem] {
        puts $out "$prefix<[c::getClassName $elem] id=\"[getFullName $elem]\" >"
        foreach sub [c::getSubElements $elem] {
            write $sub $out "$prefix  "
        }
        puts $out "$prefix</[c::getClassName $elem]>"
    } elseif [equal [c::getClassName $elem] ObservedType] {
        if [equal [c::getTypeName $elem] VLPointer] {
            puts $out "$prefix<[c::getRole $elem] ref=\"[getFullName [c::getValue $elem]]\" />"
        } else {
            puts -nonewline $out "$prefix<[c::getRole $elem]>"
            puts -nonewline $out [mask [c::getValue $elem]]
            puts $out "</[c::getRole $elem]>"
        }
    } else {
        puts $out "$prefix<[c::getRole $elem]>"
        foreach sub [c::getSubElements $elem] {
            write $sub $out "$prefix  "
        }
        puts $out "$prefix</[c::getRole $elem]>"
    }
}


# PUBLIC
proc xmlio::save {root filename} {
    # stores the (sub)tree $root in the file $filename

    set out [open $filename w]
    puts $out "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>"
    puts $out ""
    puts $out "<!-- This file stores a Visual Language Expression. -->"
    puts $out "<!-- The VL-Editor was generated by DEViL Version $main::devilVersion. -->"
    puts $out ""
    puts $out "<VLX editor=\"$cfg::appName\" grammarVersion=\"$cfg::grammarVersion\" editorVersion=\"$cfg::editorVersion\" >"
    xmlio::write $root $out "  "
    puts $out "</VLX>"
    close $out
}


# ----------------------------------------------------------------- Load Structure


# PRIVATE
proc xmlio::cdataEvent {value args} {
    variable data
    set data(text) "$data(text)$value"
}


# PRIVATE
proc xmlio::checkAppName {fileAppName} {
    if ![equal $fileAppName $cfg::appName] {
        set res [::CDlg::Confirm "File has wrong type! \n\nType of file: $fileAppName \nType of editor: $cfg::appName \n\nReally try to load file?"]
        if ![equal $res "OK"] {
            error "Aborted by user"
        }
    }
}


# PRIVATE
proc xmlio::checkGrammarVersion {fileVersion} {
    regexp {^([0-9]+).([0-9]+)$} $fileVersion dummy fMajor fMinor
    regexp {^([0-9]+).([0-9]+)$} $cfg::grammarVersion dummy eMajor eMinor
    
    set info "\nFile grammar version: $fileVersion\nEditor grammar version: $cfg::grammarVersion\nReally try to load file?"

    if { $fMajor>$eMajor || $fMajor==$eMajor && $fMinor>$eMinor } {
        set res [::CDlg::Confirm  "Old editor version! $info"]
        if ![equal $res "OK"] {
            error "Aborted by user"
        }   
    }
    if { $fMajor<$eMajor } {
        set res [::CDlg::Confirm "Old file version! $info"]
        if ![equal $res "OK"] {
            error "Aborted by user"
        }
    }
    if { $fMajor==$eMajor && $fMinor<$eMinor } {
        # No action, this is ok.
    }
}


# PRIVATE
proc xmlio::elemStartEvent {name attlist args} {
    variable data
    array set attr $attlist
    if [empty $data(stack)] {
        if ![equal $name "VLX"] {
            error "Invalid root tag name '$name', must be VLX."
        }
        checkAppName $attr(editor)
        checkGrammarVersion $attr(grammarVersion)
        lpush data(stack) "dummy"
    } elseif { [llength $data(stack)]==1 } {
        if [equal $name "Root"] {
            set newNode [d::createRoot]
            set data(file) $newNode
        } else {
            if ![info exists data(file)] {
                error "Tag name mismatch: 'Root' expected, but found '$name'"
            }
            set newNode [d::createObject $name $data(file)]
            set data(tree) $newNode
        }
        lpush data(stack) $newNode
    } elseif { [equal [ltop data(stack)] "ignore"] } {
        lpush data(stack) "ignore"
    } elseif { [c::isLanguageObject [ltop data(stack)]] } {
        if { ![c::getSubElement [ltop data(stack)] $name -var sub] } {
            log::error "Class '[c::getClassName [ltop data(stack)]]' has no attribute '$name'"
            lpush data(stack) "ignore"
        } else {
            lpush data(stack) $sub
            set data(text) ""
            if { [equal [c::getClassName [ltop data(stack)]] ObservedType] && [equal [c::getTypeName [ltop data(stack)]] VLPointer] } {
                lappend data(crossref) [ltop data(stack)] $attr(ref)
            }
        }
    } elseif { [c::isAggregate [ltop data(stack)]] } {
        if { ![c::hasClass $name] } {
            log::error "No such class: '$name'"
            lpush data(stack) "ignore"
        } else {
            set newobj [d::createObject $name $data(file)]
            if { ![c::canInsert [c::icBelow [ltop data(stack)]] $newobj] } {
                log::error "Attribute [c::getContext [ltop data(stack)]] must not contain objects of type '$name'"
                lpush data(stack) "ignore"
            } else {
                d::insert [c::icBelow [ltop data(stack)]] $newobj
                lpush data(stack) $newobj
                if [info exists attr(id)] {
                    set data(obj,$attr(id)) $newobj
                    if { [regexp {::obj([0-9]+)$} $attr(id) dummy persId] } {
                        c::setPersistentId $data(file) $newobj $persId
                    }
                }
            }
        }
    } else {
        error "Unknown error"
    }
}

proc xmlio::elemEndEvent {name args} {
    variable data
    if [llength $data(stack)]<2 {
        # nop
    } elseif { ![equal [ltop data(stack)] "ignore"] && [equal [c::getClassName [ltop data(stack)]] ObservedType] } {
        if [equal [c::getTypeName [ltop data(stack)]] VLPointer] {
            # nop
        } else {
            d::setValue [ltop data(stack)] $data(text)
        }
    }
    lpop data(stack)
}


proc xmlio::loadStructure {filename} {
    # loads structure using the data-object.
    # Two useful informations, taht remain in the data-object:
    #
    # - crossref: List of crossreferences to be processed
    # - obj,$X: objId -> nodePtr
    # - file: file of the loaded structure (object of class Root)
    # - tree: Root of the loaded tree

    variable data
    set parser [::xml::parser \
      -characterdatacommand xmlio::cdataEvent \
      -elementstartcommand xmlio::elemStartEvent \
      -elementendcommand xmlio::elemEndEvent \
    ]
    set data(stack) ""
    set data(text) ""
    set input [open $filename]
    if [catch {
        $parser parse [read $input]
    } errMsg] {
        set savedInfo $::errorInfo
        close $input
        error $errMsg $savedInfo
    }
    close $input
}



# ---- load and connect ----------------------------------------------------------------

proc xmlio::connect {} {
    variable data
    set data(obj,0x0) 0x0
    set data(obj,obj_0x0) 0x0
    # obj_0x0 is only for compatibility with old files
    foreach {obj refname} $data(crossref) {
        if [info exists data(obj,$refname)] {
            d::setValue $obj $data(obj,$refname)
        } else {
            log::error "Object id '$refname' does not exist! Ignoring this reference."
        }
    }
}

proc xmlio::load {filename} {
    variable data

    log::process "LOAD" "Loading $filename"
    catch {unset data}
    set data(crossref) ""
    if [catch {
        loadStructure $filename
        loadLibraries
        connect
    } errMsg] {
        set savedInfo $::errorInfo
        log::process "" "Loading aborted"
        error "Can't load XML file: $errMsg" $savedInfo
    }
    log::process "" "Loading ready"
    return $data(file)
}


# ---- Library and template stuff ---------------------------------------------------------------------

# PUBLIC
proc xmlio::getNewFile {} {
    set templateFilePath [file join $::main::appHome $::cfg::newFileTemplate]
    if [file exists $templateFilePath] {
        set root [load $templateFilePath]
    } else {
        set root [edit::createRoot]
        variable data
        catch {unset data}
        set data(crossref) ""
        set data(file) $root
        loadLibraries
        connect
    }
    return $root    
}


proc xmlio::loadLibraries {} {
    variable data
    if [c::getSubElement $data(file) library -var libNode] {
        set libraryFilePath [file join $::main::appHome "library"]
        if [file exists $libraryFilePath] {
            loadStructure $libraryFilePath
            d::insert [c::icBelow $libNode] $data(tree)
        }
    }
}


proc xmlio::exportSubtreeInteractive {node} {
    set types ""
    lappend types [list "XML file" ".xml"]
    lappend types [list "All Files" "*"]

    set filename [tk_getSaveFile -title "Export subtree" -filetypes $types -defaultextension ".xml" -initialdir [file::getInitialDir "exportFiles"]]
    if {$filename == ""} {return 0}
    file::setInitialDir "exportFiles" [file dirname $filename]
    save $node $filename
}


proc xmlio::isLibraryAttribute {node} {
    if [equal [c::getContext $node] "Root.library"] {        
        return 1
    } else {
        return 0
    }    
}

