
/* implementation of tree construction functions */

#include "node.h"

#include "nodecode.h"

#include "attrpredef.h"

#include "visitmap.h"

#include "treeact.h"

#ifdef MONITOR
#include "attr_mon_dapto.h"
#endif

#include <stdlib.h>

#define _USE_OBSTACK 1

/* use of obstack: */

#if _USE_OBSTACK

#include "obstack.h"
static struct obstack TreeSpace;
static void *_TreeBase;

#ifdef __cplusplus
void* NODEPTR_struct::operator new(size_t size)
{
	return obstack_alloc(&TreeSpace, size);
}
#else
#if defined(__STDC__) || defined(__cplusplus)
char* TreeNodeAlloc(int size)
#else
char* TreeNodeAlloc(size) int size;
#endif
{
	return (char *)(obstack_alloc(&TreeSpace, size));
}
#endif

void InitTree()
{
	obstack_init(&TreeSpace);
	_TreeBase=obstack_alloc(&TreeSpace,0);
}

void FreeTree()
{
	obstack_free(&TreeSpace, _TreeBase);
	_TreeBase=obstack_alloc(&TreeSpace,0);
}

#else

#include <stdio.h>

#ifdef __cplusplus
void* NODEPTR_struct::operator new(size_t size)
{
	void *retval = malloc(size);
	if (retval) return retval;
	fprintf(stderr, "*** DEADLY: No more memory.\n");
	exit(1);
}
#else
#if defined(__STDC__) || defined(__cplusplus)
char* TreeNodeAlloc(int size)
#else
char* TreeNodeAlloc(size) int size;
#endif
{
	char *retval = (char *) malloc(size);
	if (retval) return retval;
	fprintf(stderr, "*** DEADLY: No more memory.\n");
	exit(1);
}
#endif

void InitTree() { }

void FreeTree() { }

#endif

#ifdef MONITOR
#define _SETCOORD(node) \
        node->_coord = _coordref ? *_coordref : NoCoord;
#define _COPYCOORD(node) \
        node->_coord = _currn->_desc1->_coord;
#else
#define _SETCOORD(node)
#define _COPYCOORD(node)
#endif
#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkVIEWROOT (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkVIEWROOT (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBVIEWROOT)) return (_currn);
if (IsSymb (_currn, SYMBwebPageView_WebPage)) return (MkpVIEWROOT_webPageView(_coordref, _currn));
if (IsSymb (_currn, SYMBwebActionView_WebAction)) return (MkpVIEWROOT_webActionView(_coordref, _currn));
if (IsSymb (_currn, SYMBstylesheetView_StyleSheet)) return (MkpVIEWROOT_stylesheetView(_coordref, _currn));
if (IsSymb (_currn, SYMBsspDlg_StyleSheetPair)) return (MkpVIEWROOT_sspDlg(_coordref, _currn));
if (IsSymb (_currn, SYMBrootView_Root)) return (MkpVIEWROOT_rootView(_coordref, _currn));
if (IsSymb (_currn, SYMBnavigationBarView_NavigationBar)) return (MkpVIEWROOT_navigationBarView(_coordref, _currn));
if (IsSymb (_currn, SYMBmetaInformation_MetaInformation)) return (MkpVIEWROOT_metaInformation(_coordref, _currn));
if (IsSymb (_currn, SYMBglobalFrameView_GlobalFrame)) return (MkpVIEWROOT_globalFrameView(_coordref, _currn));
if (IsSymb (_currn, SYMBfcDlg_FormularCondition)) return (MkpVIEWROOT_fcDlg(_coordref, _currn));
if (IsSymb (_currn, SYMBdataBaseView_DataBase)) return (MkpVIEWROOT_dataBaseView(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_Root)) return (MkpVIEWROOT_codeGen(_coordref, _currn));
return(NULLNODEPTR);
}/* MkVIEWROOT */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Root (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Root (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Root)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_Root */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceDB (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceDB (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceDB)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceDB */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceDB_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceDB_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceDB_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceDB_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceDB_position (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceDB_position (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceDB_position)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceDB_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceDB_columns (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceDB_columns (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceDB_columns)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceDB_columns */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DBColumn (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DBColumn (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DBColumn)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DBColumn */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataBase (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataBase (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataBase)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataBase */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataBase_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataBase_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataBase_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataBase_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataBase_userName (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataBase_userName (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataBase_userName)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataBase_userName */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataBase_userPassword (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataBase_userPassword (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataBase_userPassword)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataBase_userPassword */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataBase_hostName (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataBase_hostName (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataBase_hostName)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataBase_hostName */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataBase_port (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataBase_port (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataBase_port)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataBase_port */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataBase_dataBaseType (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataBase_dataBaseType (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataBase_dataBaseType)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataBase_dataBaseType */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataBase_dataSources (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataBase_dataSources (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataBase_dataSources)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataBase_dataSources */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataBase_dataSourceModifications (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataBase_dataSourceModifications (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataBase_dataSourceModifications)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataBase_dataSourceModifications */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataBase_setSize (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataBase_setSize (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataBase_setSize)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataBase_setSize */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataBase_modifySetSize (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataBase_modifySetSize (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataBase_modifySetSize)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataBase_modifySetSize */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceJoin (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceJoin (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceJoin)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceJoin */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_LimitDataSource (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_LimitDataSource (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_LimitDataSource)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_LimitDataSource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceProjectionFilter (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceProjectionFilter (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceProjectionFilter)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceProjectionFilter */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceDelete (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceDelete (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceDelete)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceDelete */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceUpdate (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceUpdate (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceUpdate)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceUpdate */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceInsert (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceInsert (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceInsert)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceInsert */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DBColumn_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DBColumn_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DBColumn_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DBColumn_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DBColumn_type (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DBColumn_type (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DBColumn_type)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_DBString)) return (Mkrule_9(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBInteger)) return (Mkrule_8(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBFloat)) return (Mkrule_7(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBDouble)) return (Mkrule_6(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBBigInt)) return (Mkrule_5(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBDate)) return (Mkrule_4(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBTimeStamp)) return (Mkrule_3(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBBoolean)) return (Mkrule_2(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBBlob)) return (Mkrule_1(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_DBColumn_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DBBlob (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DBBlob (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DBBlob)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DBBlob */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DBBoolean (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DBBoolean (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DBBoolean)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DBBoolean */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DBTimeStamp (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DBTimeStamp (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DBTimeStamp)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DBTimeStamp */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DBDate (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DBDate (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DBDate)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DBDate */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DBBigInt (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DBBigInt (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DBBigInt)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DBBigInt */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DBDouble (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DBDouble (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DBDouble)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DBDouble */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DBFloat (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DBFloat (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DBFloat)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DBFloat */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DBInteger (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DBInteger (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DBInteger)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DBInteger */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DBString (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DBString (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DBString)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DBString */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceProjectionFilter_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceProjectionFilter_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceProjectionFilter_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceProjectionFilter_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceProjectionFilter_position (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceProjectionFilter_position (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceProjectionFilter_position)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceProjectionFilter_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceProjectionFilter_sourceDataSource (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceProjectionFilter_sourceDataSource (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceProjectionFilter_sourceDataSource)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceProjectionFilter_sourceDataSource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceProjectionFilter_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceProjectionFilter_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceProjectionFilter_condition)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_BinaryExpression)) return (Mkrule_15(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_UnaryExpression)) return (Mkrule_14(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ConstantExpression)) return (Mkrule_13(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ColumnValueExpression)) return (Mkrule_12(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ParameterValueExpression)) return (Mkrule_11(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ComputationExpression)) return (Mkrule_10(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceProjectionFilter_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceProjectionFilter_formalParameters (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceProjectionFilter_formalParameters (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceProjectionFilter_formalParameters)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceProjectionFilter_formalParameters */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceProjectionFilter_projectionColumns (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceProjectionFilter_projectionColumns (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceProjectionFilter_projectionColumns)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceProjectionFilter_projectionColumns */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ComputationExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ComputationExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ComputationExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ComputationExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ParameterValueExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ParameterValueExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ParameterValueExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ParameterValueExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ColumnValueExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ColumnValueExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ColumnValueExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ColumnValueExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ConstantExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ConstantExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ConstantExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ConstantExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_UnaryExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_UnaryExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_UnaryExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_UnaryExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_BinaryExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_BinaryExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_BinaryExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_BinaryExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceFormalParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceFormalParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceFormalParam)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceFormalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ComputationProjection (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ComputationProjection (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ComputationProjection)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ComputationProjection */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ColumnProjection (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ColumnProjection (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ColumnProjection)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ColumnProjection */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ColumnProjection_column (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ColumnProjection_column (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ColumnProjection_column)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ColumnProjection_column */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ComputationProjection_column (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ComputationProjection_column (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ComputationProjection_column)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ComputationProjection_column */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ComputationProjection_computation (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ComputationProjection_computation (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ComputationProjection_computation)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ComputationProjection_computation */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_LimitDataSource_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_LimitDataSource_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_LimitDataSource_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_LimitDataSource_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_LimitDataSource_position (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_LimitDataSource_position (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_LimitDataSource_position)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_LimitDataSource_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_LimitDataSource_sourceDataSouce (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_LimitDataSource_sourceDataSouce (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_LimitDataSource_sourceDataSouce)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_LimitDataSource_sourceDataSouce */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_LimitDataSource_minRow (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_LimitDataSource_minRow (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_LimitDataSource_minRow)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_BinaryExpression)) return (Mkrule_21(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_UnaryExpression)) return (Mkrule_20(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ConstantExpression)) return (Mkrule_19(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ColumnValueExpression)) return (Mkrule_18(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ParameterValueExpression)) return (Mkrule_17(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ComputationExpression)) return (Mkrule_16(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_LimitDataSource_minRow */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_LimitDataSource_maxRow (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_LimitDataSource_maxRow (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_LimitDataSource_maxRow)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_BinaryExpression)) return (Mkrule_27(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_UnaryExpression)) return (Mkrule_26(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ConstantExpression)) return (Mkrule_25(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ColumnValueExpression)) return (Mkrule_24(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ParameterValueExpression)) return (Mkrule_23(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ComputationExpression)) return (Mkrule_22(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_LimitDataSource_maxRow */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_LimitDataSource_formalParameters (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_LimitDataSource_formalParameters (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_LimitDataSource_formalParameters)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_LimitDataSource_formalParameters */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_BinaryExpression_leftOperand (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_BinaryExpression_leftOperand (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_BinaryExpression_leftOperand)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_BinaryExpression)) return (Mkrule_33(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_UnaryExpression)) return (Mkrule_32(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ConstantExpression)) return (Mkrule_31(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ColumnValueExpression)) return (Mkrule_30(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ParameterValueExpression)) return (Mkrule_29(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ComputationExpression)) return (Mkrule_28(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_BinaryExpression_leftOperand */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_BinaryExpression_rightOperand (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_BinaryExpression_rightOperand (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_BinaryExpression_rightOperand)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_BinaryExpression)) return (Mkrule_39(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_UnaryExpression)) return (Mkrule_38(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ConstantExpression)) return (Mkrule_37(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ColumnValueExpression)) return (Mkrule_36(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ParameterValueExpression)) return (Mkrule_35(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ComputationExpression)) return (Mkrule_34(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_BinaryExpression_rightOperand */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_BinaryExpression_operator (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_BinaryExpression_operator (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_BinaryExpression_operator)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_BinaryExpression_operator */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_UnaryExpression_operand (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_UnaryExpression_operand (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_UnaryExpression_operand)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_BinaryExpression)) return (Mkrule_45(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_UnaryExpression)) return (Mkrule_44(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ConstantExpression)) return (Mkrule_43(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ColumnValueExpression)) return (Mkrule_42(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ParameterValueExpression)) return (Mkrule_41(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ComputationExpression)) return (Mkrule_40(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_UnaryExpression_operand */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_UnaryExpression_operator (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_UnaryExpression_operator (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_UnaryExpression_operator)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_UnaryExpression_operator */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ConstantExpression_type (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ConstantExpression_type (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ConstantExpression_type)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_DBString)) return (Mkrule_54(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBInteger)) return (Mkrule_53(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBFloat)) return (Mkrule_52(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBDouble)) return (Mkrule_51(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBBigInt)) return (Mkrule_50(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBDate)) return (Mkrule_49(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBTimeStamp)) return (Mkrule_48(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBBoolean)) return (Mkrule_47(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBBlob)) return (Mkrule_46(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_ConstantExpression_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ConstantExpression_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ConstantExpression_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ConstantExpression_value)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ConstantExpression_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ColumnValueExpression_column (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ColumnValueExpression_column (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ColumnValueExpression_column)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ColumnValueExpression_column */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ParameterValueExpression_parameter (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ParameterValueExpression_parameter (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ParameterValueExpression_parameter)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ParameterValueExpression_parameter */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ComputationExpression_column (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ComputationExpression_column (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ComputationExpression_column)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ComputationExpression_column */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ComputationExpression_computation (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ComputationExpression_computation (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ComputationExpression_computation)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ComputationExpression_computation */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceValue)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceValue_column (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceValue_column (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceValue_column)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceValue_column */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceValue_expression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceValue_expression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceValue_expression)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_BinaryExpression)) return (Mkrule_60(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_UnaryExpression)) return (Mkrule_59(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ConstantExpression)) return (Mkrule_58(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ColumnValueExpression)) return (Mkrule_57(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ParameterValueExpression)) return (Mkrule_56(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ComputationExpression)) return (Mkrule_55(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceValue_expression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceJoin_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceJoin_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceJoin_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceJoin_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceJoin_position (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceJoin_position (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceJoin_position)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceJoin_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceJoin_dataSource1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceJoin_dataSource1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceJoin_dataSource1)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceJoin_dataSource1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceJoin_dataSource2 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceJoin_dataSource2 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceJoin_dataSource2)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceJoin_dataSource2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceJoin_columnRef1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceJoin_columnRef1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceJoin_columnRef1)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceJoin_columnRef1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceJoin_columnRef2 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceJoin_columnRef2 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceJoin_columnRef2)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceJoin_columnRef2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceJoin_joinOperator (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceJoin_joinOperator (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceJoin_joinOperator)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceJoin_joinOperator */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceInsert_targetDataSource (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceInsert_targetDataSource (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceInsert_targetDataSource)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceInsert_targetDataSource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceInsert_formalParams (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceInsert_formalParams (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceInsert_formalParams)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceInsert_formalParams */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceInsert_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceInsert_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceInsert_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceInsert_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceInsert_position (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceInsert_position (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceInsert_position)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceInsert_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceInsert_values (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceInsert_values (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceInsert_values)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceInsert_values */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceUpdate_targetDataSource (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceUpdate_targetDataSource (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceUpdate_targetDataSource)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceUpdate_targetDataSource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceUpdate_formalParams (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceUpdate_formalParams (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceUpdate_formalParams)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceUpdate_formalParams */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceUpdate_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceUpdate_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceUpdate_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceUpdate_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceUpdate_position (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceUpdate_position (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceUpdate_position)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceUpdate_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceUpdate_conditions (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceUpdate_conditions (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceUpdate_conditions)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_BinaryExpression)) return (Mkrule_66(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_UnaryExpression)) return (Mkrule_65(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ConstantExpression)) return (Mkrule_64(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ColumnValueExpression)) return (Mkrule_63(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ParameterValueExpression)) return (Mkrule_62(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ComputationExpression)) return (Mkrule_61(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceUpdate_conditions */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceUpdate_values (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceUpdate_values (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceUpdate_values)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceUpdate_values */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceDelete_targetDataSource (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceDelete_targetDataSource (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceDelete_targetDataSource)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceDelete_targetDataSource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceDelete_formalParams (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceDelete_formalParams (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceDelete_formalParams)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceDelete_formalParams */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceDelete_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceDelete_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceDelete_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceDelete_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceDelete_position (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceDelete_position (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceDelete_position)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceDelete_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceDelete_conditions (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceDelete_conditions (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceDelete_conditions)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_BinaryExpression)) return (Mkrule_72(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_UnaryExpression)) return (Mkrule_71(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ConstantExpression)) return (Mkrule_70(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ColumnValueExpression)) return (Mkrule_69(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ParameterValueExpression)) return (Mkrule_68(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ComputationExpression)) return (Mkrule_67(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceDelete_conditions */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceActualParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceActualParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceActualParam)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceActualParam_formalParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceActualParam_formalParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceActualParam_formalParam)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceActualParam_formalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceActualParam_expressions (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceActualParam_expressions)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_WebNodeConstantExpression)) return (Mkrule_75(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_WebNodeVariableExpression)) return (Mkrule_74(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_WebNodeColumnExpression)) return (Mkrule_73(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceActualParam_expressions */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebNodeColumnExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebNodeColumnExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebNodeColumnExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebNodeColumnExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebNodeVariableExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebNodeVariableExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebNodeVariableExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebNodeVariableExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebNodeConstantExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebNodeConstantExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebNodeConstantExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebNodeConstantExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceFormalParam_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceFormalParam_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceFormalParam_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceFormalParam_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataSourceFormalParam_type (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataSourceFormalParam_type (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataSourceFormalParam_type)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_DBString)) return (Mkrule_84(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBInteger)) return (Mkrule_83(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBFloat)) return (Mkrule_82(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBDouble)) return (Mkrule_81(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBBigInt)) return (Mkrule_80(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBDate)) return (Mkrule_79(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBTimeStamp)) return (Mkrule_78(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBBoolean)) return (Mkrule_77(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBBlob)) return (Mkrule_76(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_DataSourceFormalParam_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Formular (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Formular (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Formular)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_Formular */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Formular_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Formular_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Formular_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_Formular_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Formular_action (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Formular_action (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Formular_action)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_Formular_action */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Formular_elements (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Formular_elements (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Formular_elements)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_Formular_elements */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Formular_datasource (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Formular_datasource (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Formular_datasource)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_Formular_datasource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Formular_actualParams (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Formular_actualParams (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Formular_actualParams)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_Formular_actualParams */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularSelectElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularSelectElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularSelectElement)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularSelectElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularCheckboxElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularCheckboxElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularCheckboxElement)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularCheckboxElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularPasswordElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularPasswordElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularPasswordElement)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularPasswordElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularTextAreaElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularTextAreaElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularTextAreaElement)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularTextAreaElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularTextElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularTextElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularTextElement)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularTextElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularRadioGroupElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularRadioGroupElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularRadioGroupElement)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularRadioGroupElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebNodeActualParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebNodeActualParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebNodeActualParam)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebNodeActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularRadioGroupElement_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularRadioGroupElement_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularRadioGroupElement_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularRadioGroupElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularRadioGroupElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularRadioGroupElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularRadioGroupElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularRadioGroupElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularRadioGroupElement_description (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularRadioGroupElement_description (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularRadioGroupElement_description)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularRadioGroupElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularRadioGroupElement_pvalue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularRadioGroupElement_pvalue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularRadioGroupElement_pvalue)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularRadioGroupElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularRadioGroupElement_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularRadioGroupElement_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularRadioGroupElement_condition)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_FormularCondition)) return (Mkrule_85(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_FormularRadioGroupElement_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularRadioGroupElement_elements (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularRadioGroupElement_elements (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularRadioGroupElement_elements)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularRadioGroupElement_elements */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularRadioGroupElement_checked (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularRadioGroupElement_checked (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularRadioGroupElement_checked)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularRadioGroupElement_checked */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularCondition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularCondition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularCondition)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularCondition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularRadioElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularRadioElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularRadioElement)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularRadioElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularTextElement_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularTextElement_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularTextElement_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularTextElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularTextElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularTextElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularTextElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularTextElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularTextElement_description (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularTextElement_description (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularTextElement_description)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularTextElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularTextElement_pvalue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularTextElement_pvalue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularTextElement_pvalue)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularTextElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularTextElement_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularTextElement_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularTextElement_condition)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_FormularCondition)) return (Mkrule_86(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_FormularTextElement_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularTextAreaElement_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularTextAreaElement_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularTextAreaElement_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularTextAreaElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularTextAreaElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularTextAreaElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularTextAreaElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularTextAreaElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularTextAreaElement_description (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularTextAreaElement_description (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularTextAreaElement_description)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularTextAreaElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularTextAreaElement_pvalue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularTextAreaElement_pvalue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularTextAreaElement_pvalue)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularTextAreaElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularTextAreaElement_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularTextAreaElement_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularTextAreaElement_condition)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_FormularCondition)) return (Mkrule_87(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_FormularTextAreaElement_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularPasswordElement_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularPasswordElement_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularPasswordElement_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularPasswordElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularPasswordElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularPasswordElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularPasswordElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularPasswordElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularPasswordElement_description (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularPasswordElement_description (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularPasswordElement_description)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularPasswordElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularPasswordElement_pvalue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularPasswordElement_pvalue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularPasswordElement_pvalue)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularPasswordElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularPasswordElement_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularPasswordElement_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularPasswordElement_condition)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_FormularCondition)) return (Mkrule_88(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_FormularPasswordElement_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularCheckboxElement_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularCheckboxElement_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularCheckboxElement_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularCheckboxElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularCheckboxElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularCheckboxElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularCheckboxElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularCheckboxElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularCheckboxElement_description (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularCheckboxElement_description (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularCheckboxElement_description)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularCheckboxElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularCheckboxElement_pvalue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularCheckboxElement_pvalue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularCheckboxElement_pvalue)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularCheckboxElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularCheckboxElement_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularCheckboxElement_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularCheckboxElement_condition)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_FormularCondition)) return (Mkrule_89(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_FormularCheckboxElement_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularCheckboxElement_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularCheckboxElement_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularCheckboxElement_value)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularCheckboxElement_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularCheckboxElement_checked (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularCheckboxElement_checked (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularCheckboxElement_checked)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularCheckboxElement_checked */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularSelectElement_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularSelectElement_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularSelectElement_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularSelectElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularSelectElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularSelectElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularSelectElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularSelectElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularSelectElement_description (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularSelectElement_description (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularSelectElement_description)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularSelectElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularSelectElement_pvalue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularSelectElement_pvalue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularSelectElement_pvalue)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularSelectElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularSelectElement_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularSelectElement_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularSelectElement_condition)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_FormularCondition)) return (Mkrule_90(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_FormularSelectElement_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularSelectElement_options (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularSelectElement_options (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularSelectElement_options)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_FormularConstantSelectOption)) return (Mkrule_93(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_FormularDataSourceSelectOption)) return (Mkrule_92(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_FormularPHPScriptSelectOption)) return (Mkrule_91(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_FormularSelectElement_options */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularSelectElement_size (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularSelectElement_size (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularSelectElement_size)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularSelectElement_size */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularSelectElement_multiple (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularSelectElement_multiple (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularSelectElement_multiple)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularSelectElement_multiple */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularPHPScriptSelectOption (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularPHPScriptSelectOption (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularPHPScriptSelectOption)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularPHPScriptSelectOption */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularDataSourceSelectOption (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularDataSourceSelectOption (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularDataSourceSelectOption)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularDataSourceSelectOption */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularConstantSelectOption (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularConstantSelectOption (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularConstantSelectOption)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularConstantSelectOption */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularConstantSelectOption_values (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularConstantSelectOption_values (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularConstantSelectOption_values)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularConstantSelectOption_values */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularDataSourceSelectOption_values (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularDataSourceSelectOption_values (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularDataSourceSelectOption_values)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularDataSourceSelectOption_values */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularDataSourceSelectOption_params (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularDataSourceSelectOption_params (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularDataSourceSelectOption_params)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularDataSourceSelectOption_params */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularPHPScriptSelectOption_script (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularPHPScriptSelectOption_script (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularPHPScriptSelectOption_script)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularPHPScriptSelectOption_script */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularRadioElement_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularRadioElement_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularRadioElement_value)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularRadioElement_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularRadioElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularRadioElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularRadioElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularRadioElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularRadioElement_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularRadioElement_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularRadioElement_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularRadioElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularCondition_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularCondition_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularCondition_condition)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularCondition_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularCondition_params (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularCondition_params (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularCondition_params)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularCondition_params */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularConditionActualParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularConditionActualParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularConditionActualParam)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularConditionActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularConditionDefinition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularConditionDefinition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularConditionDefinition)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularConditionDefinition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularConditionDefinition_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularConditionDefinition_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularConditionDefinition_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularConditionDefinition_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularConditionDefinition_params (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularConditionDefinition_params (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularConditionDefinition_params)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularConditionDefinition_params */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularConditionDefinition_body (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularConditionDefinition_body (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularConditionDefinition_body)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularConditionDefinition_body */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularConditionFormalParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularConditionFormalParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularConditionFormalParam)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularConditionFormalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularConditionFormalParam_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularConditionFormalParam_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularConditionFormalParam_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularConditionFormalParam_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularConditionActualParam_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularConditionActualParam_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularConditionActualParam_value)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularConditionActualParam_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FormularConditionActualParam_formalparam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FormularConditionActualParam_formalparam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FormularConditionActualParam_formalparam)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FormularConditionActualParam_formalparam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebNodeActualParam_formalParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebNodeActualParam_formalParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebNodeActualParam_formalParam)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebNodeActualParam_formalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebNodeActualParam_expressions (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebNodeActualParam_expressions)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_WebNodeConstantExpression)) return (Mkrule_96(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_WebNodeVariableExpression)) return (Mkrule_95(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_WebNodeColumnExpression)) return (Mkrule_94(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_WebNodeActualParam_expressions */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebNodeFormalParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebNodeFormalParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebNodeFormalParam)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebNodeFormalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebNodeFormalParam_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebNodeFormalParam_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebNodeFormalParam_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebNodeFormalParam_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebNodeFormalParam_type (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebNodeFormalParam_type)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_DBString)) return (Mkrule_105(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBInteger)) return (Mkrule_104(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBFloat)) return (Mkrule_103(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBDouble)) return (Mkrule_102(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBBigInt)) return (Mkrule_101(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBDate)) return (Mkrule_100(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBTimeStamp)) return (Mkrule_99(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBBoolean)) return (Mkrule_98(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_DBBlob)) return (Mkrule_97(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_WebNodeFormalParam_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebNodeConstantExpression_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebNodeConstantExpression_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebNodeConstantExpression_value)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebNodeConstantExpression_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebNodeVariableExpression_content (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebNodeVariableExpression_content (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebNodeVariableExpression_content)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebNodeVariableExpression_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebNodeColumnExpression_content (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebNodeColumnExpression_content (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebNodeColumnExpression_content)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebNodeColumnExpression_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Paragraph (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Paragraph (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Paragraph)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_Paragraph */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Paragraph_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Paragraph_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Paragraph_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_Paragraph_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Paragraph_fragments (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Paragraph_fragments (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Paragraph_fragments)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_Paragraph_fragments */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_OutputElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_OutputElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_OutputElement)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_OutputElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ImageFragment (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ImageFragment (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ImageFragment)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ImageFragment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_TextFragment (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_TextFragment (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_TextFragment)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_TextFragment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_TextFragment_link (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_TextFragment_link (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_TextFragment_link)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_TextFragment_link */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_TextFragment_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_TextFragment_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_TextFragment_value)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_TextFragment_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ImageFragment_link (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ImageFragment_link (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ImageFragment_link)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ImageFragment_link */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ImageFragment_width (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ImageFragment_width (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ImageFragment_width)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ImageFragment_width */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ImageFragment_height (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ImageFragment_height (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ImageFragment_height)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ImageFragment_height */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ImageFragment_text (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ImageFragment_text (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ImageFragment_text)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ImageFragment_text */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ImageFragment_imageName (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ImageFragment_imageName (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ImageFragment_imageName)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ImageFragment_imageName */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Heading (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Heading (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Heading)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_Heading */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Heading_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Heading_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Heading_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_Heading_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Heading_text (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Heading_text (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Heading_text)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_Heading_text */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Heading_level (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Heading_level (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Heading_level)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_Heading_level */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Table (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Table (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Table)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_Table */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Table_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Table_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Table_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_Table_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Table_rows (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Table_rows (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Table_rows)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_Table_rows */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Table_columns (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Table_columns (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Table_columns)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_Table_columns */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_TableRow (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_TableRow (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_TableRow)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_TableRow */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_TableColumn (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_TableColumn (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_TableColumn)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_TableColumn */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_TableRow_cells (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_TableRow_cells (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_TableRow_cells)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_TableRow_cells */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_TableCell (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_TableCell (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_TableCell)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_TableCell */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_TableCell_columnRef (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_TableCell_columnRef (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_TableCell_columnRef)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_TableCell_columnRef */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_TableCell_content (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_TableCell_content (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_TableCell_content)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_Frame)) return (Mkrule_106(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_TableCell_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Frame (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Frame (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Frame)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_Frame */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_LinkTargetExtern (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_LinkTargetExtern (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_LinkTargetExtern)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_LinkTargetExtern */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_LinkTargetExtern_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_LinkTargetExtern_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_LinkTargetExtern_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_LinkTargetExtern_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_LinkTargetExtern_url (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_LinkTargetExtern_url (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_LinkTargetExtern_url)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_LinkTargetExtern_url */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Root_webApp (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Root_webApp (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Root_webApp)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_WebApplication)) return (Mkrule_107(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_Root_webApp */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Root_library (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Root_library (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Root_library)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_Root_library */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebApplication (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebApplication (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebApplication)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebApplication */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Library (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Library (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Library)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_Library */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Library_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Library_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Library_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_Library_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Library_formularconditions (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Library_formularconditions (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Library_formularconditions)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_Library_formularconditions */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Library_navigationElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Library_navigationElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Library_navigationElement)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_Library_navigationElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_NavigationElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_NavigationElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_NavigationElement)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_NavigationElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_NavigationElement_code (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_NavigationElement_code (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_NavigationElement_code)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_NavigationElement_code */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_NavigationElement_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_NavigationElement_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_NavigationElement_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_NavigationElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebApplication_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebApplication_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebApplication_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebApplication_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebApplication_webZone (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebApplication_webZone (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebApplication_webZone)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_RootWebZone)) return (Mkrule_108(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_WebApplication_webZone */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebApplication_navigationBar (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebApplication_navigationBar (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebApplication_navigationBar)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_NavigationBar)) return (Mkrule_109(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_WebApplication_navigationBar */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebApplication_styleSheet (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebApplication_styleSheet (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebApplication_styleSheet)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_StyleSheet)) return (Mkrule_110(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_WebApplication_styleSheet */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebApplication_frames (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebApplication_frames (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebApplication_frames)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebApplication_frames */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebApplication_meta (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebApplication_meta (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebApplication_meta)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_MetaInformation)) return (Mkrule_111(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_WebApplication_meta */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebApplication_database (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebApplication_database (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebApplication_database)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_DataBase)) return (Mkrule_112(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_WebApplication_database */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebApplication_roles (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebApplication_roles (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebApplication_roles)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebApplication_roles */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebApplication_linksExtern (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebApplication_linksExtern (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebApplication_linksExtern)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebApplication_linksExtern */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebApplication_imagePath (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebApplication_imagePath (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebApplication_imagePath)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebApplication_imagePath */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_RootWebZone (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_RootWebZone (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_RootWebZone)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_RootWebZone */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_NavigationBar (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_NavigationBar (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_NavigationBar)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_NavigationBar */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_StyleSheet (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_StyleSheet (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_StyleSheet)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_StyleSheet */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_GlobalFrame (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_GlobalFrame (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_GlobalFrame)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_GlobalFrame */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_MetaInformation (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_MetaInformation (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_MetaInformation)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_MetaInformation */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebRole (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebRole (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebRole)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebRole */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_RootWebZone_webTargets (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_RootWebZone_webTargets (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_RootWebZone_webTargets)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_RootWebZone_webTargets */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_RootWebZone_errorLink (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_RootWebZone_errorLink (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_RootWebZone_errorLink)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_RootWebZone_errorLink */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_RootWebZone_roles (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_RootWebZone_roles (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_RootWebZone_roles)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_RootWebZone_roles */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_RootWebZone_setSize (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_RootWebZone_setSize (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_RootWebZone_setSize)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_RootWebZone_setSize */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebAction (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebAction (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebAction)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebAction */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebPage (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebPage (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebPage)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebPage */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_SubWebZone (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_SubWebZone (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_SubWebZone)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_SubWebZone */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebRoleRef (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebRoleRef (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebRoleRef)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebRoleRef */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_SubWebZone_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_SubWebZone_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_SubWebZone_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_SubWebZone_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_SubWebZone_webTargets (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_SubWebZone_webTargets (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_SubWebZone_webTargets)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_SubWebZone_webTargets */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_SubWebZone_errorLink (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_SubWebZone_errorLink (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_SubWebZone_errorLink)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_SubWebZone_errorLink */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_SubWebZone_globalparams (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_SubWebZone_globalparams (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_SubWebZone_globalparams)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_SubWebZone_globalparams */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_SubWebZone_roles (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_SubWebZone_roles (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_SubWebZone_roles)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_SubWebZone_roles */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_SubWebZone_setSize (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_SubWebZone_setSize (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_SubWebZone_setSize)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_SubWebZone_setSize */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_SubWebZone_position (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_SubWebZone_position (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_SubWebZone_position)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_SubWebZone_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebRole_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebRole_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebRole_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebRole_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebRole_color (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebRole_color (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebRole_color)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebRole_color */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebRoleRef_role (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebRoleRef_role (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebRoleRef_role)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebRoleRef_role */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebPage_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebPage_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebPage_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebPage_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebPage_formalParameter (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebPage_formalParameter (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebPage_formalParameter)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebPage_formalParameter */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebPage_title (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebPage_title (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebPage_title)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebPage_title */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebPage_frame (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebPage_frame (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebPage_frame)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_Frame)) return (Mkrule_113(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_WebPage_frame */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebPage_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebPage_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebPage_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebPage_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebPage_position (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebPage_position (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebPage_position)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebPage_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Frame_components (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Frame_components (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Frame_components)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_Frame_components */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_Frame_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_Frame_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_Frame_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_Frame_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ListOutputElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ListOutputElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ListOutputElement)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ListOutputElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FrameRef (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FrameRef (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FrameRef)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FrameRef */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_GlobalFrame_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_GlobalFrame_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_GlobalFrame_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_GlobalFrame_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_GlobalFrame_components (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_GlobalFrame_components (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_GlobalFrame_components)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_GlobalFrame_components */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_GlobalFrame_setSize (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_GlobalFrame_setSize (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_GlobalFrame_setSize)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_GlobalFrame_setSize */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_NavigationBar_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_NavigationBar_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_NavigationBar_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_NavigationBar_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_NavigationBar_elements (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_NavigationBar_elements (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_NavigationBar_elements)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_NavigationBar_elements */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_NavigationBar_alignment (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_NavigationBar_alignment (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_NavigationBar_alignment)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_NavigationBar_alignment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_NavigationBarElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_NavigationBarElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_NavigationBarElement)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_NavigationBarElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_NavigationBarElement_target (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_NavigationBarElement_target (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_NavigationBarElement_target)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_NavigationBarElement_target */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FrameRef_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FrameRef_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FrameRef_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FrameRef_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FrameRef_frame (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FrameRef_frame (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FrameRef_frame)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FrameRef_frame */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_MetaInformation_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_MetaInformation_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_MetaInformation_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_MetaInformation_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_MetaInformation_elements (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_MetaInformation_elements (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_MetaInformation_elements)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_MetaInformation_elements */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_MetaInformationElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_MetaInformationElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_MetaInformationElement)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_MetaInformationElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_MetaInformationElement_property (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_MetaInformationElement_property (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_MetaInformationElement_property)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_MetaInformationElement_property */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_MetaInformationElement_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_MetaInformationElement_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_MetaInformationElement_value)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_MetaInformationElement_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_OutputElement_link (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_OutputElement_link (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_OutputElement_link)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_OutputElement_link */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_OutputElement_content (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_OutputElement_content (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_OutputElement_content)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_OutputElement_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ListOutputElement_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ListOutputElement_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ListOutputElement_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ListOutputElement_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ListOutputElement_dataSourceActualParams (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ListOutputElement_dataSourceActualParams (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ListOutputElement_dataSourceActualParams)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ListOutputElement_dataSourceActualParams */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ListOutputElement_dataSource (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ListOutputElement_dataSource (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ListOutputElement_dataSource)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ListOutputElement_dataSource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ListOutputElement_maxNumberOfElements (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ListOutputElement_maxNumberOfElements (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ListOutputElement_maxNumberOfElements)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ListOutputElement_maxNumberOfElements */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ListOutputElement_navigationElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ListOutputElement_navigationElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ListOutputElement_navigationElement)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ListOutputElement_navigationElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_StyleSheet_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_StyleSheet_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_StyleSheet_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_StyleSheet_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_StyleSheet_custom (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_StyleSheet_custom (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_StyleSheet_custom)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_StyleSheet_custom */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_StyleSheet_default (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_StyleSheet_default (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_StyleSheet_default)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_StyleSheet_default */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_CustomStyleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_CustomStyleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_CustomStyleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_CustomStyleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_HtmlStyleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_HtmlStyleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_HtmlStyleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_HtmlStyleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_CustomStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_CustomStyleSheetBlock_pairs (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_CustomStyleSheetBlock_pairs)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_CustomStyleSheetBlock_pairs */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_CustomStyleSheetBlock_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_CustomStyleSheetBlock_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_CustomStyleSheetBlock_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_CustomStyleSheetBlock_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_StyleSheetPair (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_StyleSheetPair (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_StyleSheetPair)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_StyleSheetPair */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_StyleSheetPair_category (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_StyleSheetPair_category (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_StyleSheetPair_category)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_StyleSheetPair_category */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_StyleSheetPair_attribute (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_StyleSheetPair_attribute (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_StyleSheetPair_attribute)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_StyleSheetPair_attribute */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_StyleSheetPair_valueCategory (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_StyleSheetPair_valueCategory (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_StyleSheetPair_valueCategory)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_StyleSheetPair_valueCategory */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_StyleSheetPair_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_StyleSheetPair_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_StyleSheetPair_value)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_SizeValue)) return (Mkrule_121(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ColorValue)) return (Mkrule_120(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_FamilyValue)) return (Mkrule_119(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_StyleValue)) return (Mkrule_118(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_HeadingValue)) return (Mkrule_117(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_AlignmentValue)) return (Mkrule_116(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_ListValue)) return (Mkrule_115(_coordref, _currn));
if (IsSymb (_currn, SYMBcodeGen_CustomValue)) return (Mkrule_114(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_StyleSheetPair_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_CustomValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_CustomValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_CustomValue)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_CustomValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ListValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ListValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ListValue)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ListValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_AlignmentValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_AlignmentValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_AlignmentValue)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_AlignmentValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_HeadingValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_HeadingValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_HeadingValue)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_HeadingValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_StyleValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_StyleValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_StyleValue)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_StyleValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FamilyValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FamilyValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FamilyValue)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FamilyValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ColorValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ColorValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ColorValue)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ColorValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_SizeValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_SizeValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_SizeValue)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_SizeValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_SizeValue_measure (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_SizeValue_measure (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_SizeValue_measure)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_SizeValue_measure */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_SizeValue_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_SizeValue_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_SizeValue_value)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_SizeValue_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ColorValue_color (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ColorValue_color (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ColorValue_color)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ColorValue_color */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_FamilyValue_family (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_FamilyValue_family (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_FamilyValue_family)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_FamilyValue_family */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_StyleValue_style (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_StyleValue_style (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_StyleValue_style)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_StyleValue_style */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_HeadingValue_heading (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_HeadingValue_heading (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_HeadingValue_heading)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_HeadingValue_heading */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_AlignmentValue_alignment (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_AlignmentValue_alignment (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_AlignmentValue_alignment)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_AlignmentValue_alignment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ListValue_listed (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ListValue_listed (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ListValue_listed)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ListValue_listed */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_CustomValue_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_CustomValue_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_CustomValue_value)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_CustomValue_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_HtmlStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_HtmlStyleSheetBlock_pairs (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_HtmlStyleSheetBlock_pairs)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_HtmlStyleSheetBlock_pairs */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_HtmlStyleSheetBlock_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_HtmlStyleSheetBlock_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_HtmlStyleSheetBlock_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_HtmlStyleSheetBlock_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ActionHandler (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ActionHandler (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ActionHandler)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ActionHandler */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ActionHandler_targetpage (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ActionHandler_targetpage (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ActionHandler_targetpage)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ActionHandler_targetpage */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_ActionHandler_params (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_ActionHandler_params (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_ActionHandler_params)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_ActionHandler_params */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebAction_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebAction_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebAction_name)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebAction_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebAction_formalParameter (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebAction_formalParameter (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebAction_formalParameter)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebAction_formalParameter */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebAction_successpage (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebAction_successpage (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebAction_successpage)) return (_currn);
if (IsSymb (_currn, SYMBcodeGen_ActionHandler)) return (Mkrule_122(_coordref, _currn));
return(NULLNODEPTR);
}/* MkcodeGen_WebAction_successpage */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebAction_operations (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebAction_operations (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebAction_operations)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebAction_operations */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_WebAction_position (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_WebAction_position (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_WebAction_position)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_WebAction_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataBaseOperation (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataBaseOperation (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataBaseOperation)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataBaseOperation */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_PHPOperation (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_PHPOperation (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_PHPOperation)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_PHPOperation */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_PHPOperation_errorhandler (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_PHPOperation_errorhandler (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_PHPOperation_errorhandler)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_PHPOperation_errorhandler */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_PHPOperation_content (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_PHPOperation_content (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_PHPOperation_content)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_PHPOperation_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataBaseOperation_errorhandler (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataBaseOperation_errorhandler (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataBaseOperation_errorhandler)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataBaseOperation_errorhandler */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataBaseOperation_filter (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataBaseOperation_filter (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataBaseOperation_filter)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataBaseOperation_filter */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkcodeGen_DataBaseOperation_params (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkcodeGen_DataBaseOperation_params (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBcodeGen_DataBaseOperation_params)) return (_currn);
return(NULLNODEPTR);
}/* MkcodeGen_DataBaseOperation_params */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataBase (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataBase (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataBase)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataBase */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DBColumn (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DBColumn (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DBColumn)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DBColumn */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DBColumn_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DBColumn_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DBColumn_name)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DBColumn_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DBColumn_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DBColumn_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DBColumn_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DBColumn_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DBColumn_type (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DBColumn_type (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DBColumn_type)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DBColumn_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DBBlob (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DBBlob (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DBBlob)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DBBlob */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DBBoolean (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DBBoolean (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DBBoolean)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DBBoolean */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DBTimeStamp (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DBTimeStamp (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DBTimeStamp)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DBTimeStamp */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DBDate (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DBDate (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DBDate)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DBDate */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DBBigInt (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DBBigInt (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DBBigInt)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DBBigInt */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DBDouble (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DBDouble (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DBDouble)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DBDouble */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DBFloat (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DBFloat (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DBFloat)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DBFloat */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DBInteger (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DBInteger (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DBInteger)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DBInteger */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DBString (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DBString (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DBString)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DBString */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceFormalParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceFormalParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceFormalParam)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceFormalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceFormalParam_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceFormalParam_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceFormalParam_name)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceFormalParam_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceFormalParam_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceFormalParam_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceFormalParam_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceFormalParam_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceFormalParam_type (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceFormalParam_type (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceFormalParam_type)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceFormalParam_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_ColumnProjection (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_ColumnProjection (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_ColumnProjection)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_ColumnProjection */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_ColumnProjection_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_ColumnProjection_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_ColumnProjection_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_ColumnProjection_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_ColumnProjection_column (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_ColumnProjection_column (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_ColumnProjection_column)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_ColumnProjection_column */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_ComputationProjection (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_ComputationProjection (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_ComputationProjection)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_ComputationProjection */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_ComputationProjection_computation (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_ComputationProjection_computation (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_ComputationProjection_computation)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_ComputationProjection_computation */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_ComputationProjection_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_ComputationProjection_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_ComputationProjection_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_ComputationProjection_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_ComputationProjection_column (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_ComputationProjection_column (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_ComputationProjection_column)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_ComputationProjection_column */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_ComputationProjection_keyword2 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_ComputationProjection_keyword2 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_ComputationProjection_keyword2)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_ComputationProjection_keyword2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_BinaryExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_BinaryExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_BinaryExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_BinaryExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_BinaryExpression_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_BinaryExpression_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_BinaryExpression_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_BinaryExpression_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_BinaryExpression_leftOperand (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_BinaryExpression_leftOperand (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_BinaryExpression_leftOperand)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_BinaryExpression_leftOperand */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_BinaryExpression_operator (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_BinaryExpression_operator (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_BinaryExpression_operator)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_BinaryExpression_operator */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_BinaryExpression_rightOperand (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_BinaryExpression_rightOperand (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_BinaryExpression_rightOperand)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_BinaryExpression_rightOperand */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_BinaryExpression_keyword2 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_BinaryExpression_keyword2 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_BinaryExpression_keyword2)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_BinaryExpression_keyword2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_ComputationExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_ComputationExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_ComputationExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_ComputationExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_ParameterValueExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_ParameterValueExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_ParameterValueExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_ParameterValueExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_ColumnValueExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_ColumnValueExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_ColumnValueExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_ColumnValueExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_ConstantExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_ConstantExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_ConstantExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_ConstantExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_UnaryExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_UnaryExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_UnaryExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_UnaryExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_UnaryExpression_operator (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_UnaryExpression_operator (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_UnaryExpression_operator)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_UnaryExpression_operator */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_UnaryExpression_operand (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_UnaryExpression_operand (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_UnaryExpression_operand)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_UnaryExpression_operand */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_ConstantExpression_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_ConstantExpression_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_ConstantExpression_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_ConstantExpression_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_ConstantExpression_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_ConstantExpression_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_ConstantExpression_value)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_ConstantExpression_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_ConstantExpression_keyword2 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_ConstantExpression_keyword2 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_ConstantExpression_keyword2)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_ConstantExpression_keyword2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_ColumnValueExpression_column (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_ColumnValueExpression_column (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_ColumnValueExpression_column)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_ColumnValueExpression_column */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_ParameterValueExpression_parameter (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_ParameterValueExpression_parameter (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_ParameterValueExpression_parameter)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_ParameterValueExpression_parameter */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_ComputationExpression_computation (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_ComputationExpression_computation (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_ComputationExpression_computation)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_ComputationExpression_computation */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_ComputationExpression_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_ComputationExpression_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_ComputationExpression_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_ComputationExpression_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_ComputationExpression_column (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_ComputationExpression_column (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_ComputationExpression_column)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_ComputationExpression_column */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_ComputationExpression_keyword2 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_ComputationExpression_keyword2 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_ComputationExpression_keyword2)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_ComputationExpression_keyword2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DBString_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DBString_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DBString_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DBString_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DBInteger_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DBInteger_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DBInteger_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DBInteger_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DBFloat_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DBFloat_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DBFloat_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DBFloat_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DBDouble_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DBDouble_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DBDouble_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DBDouble_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DBBigInt_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DBBigInt_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DBBigInt_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DBBigInt_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DBDate_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DBDate_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DBDate_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DBDate_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DBTimeStamp_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DBTimeStamp_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DBTimeStamp_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DBTimeStamp_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DBBoolean_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DBBoolean_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DBBoolean_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DBBoolean_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DBBlob_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DBBlob_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DBBlob_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DBBlob_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceValue)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceValue_column (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceValue_column (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceValue_column)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceValue_column */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceValue_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceValue_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceValue_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceValue_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceValue_expression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceValue_expression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceValue_expression)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceValue_expression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceProjectionFilter (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceProjectionFilter (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceProjectionFilter)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceProjectionFilter */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceProjectionFilter_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceProjectionFilter_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceProjectionFilter_name)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceProjectionFilter_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceProjectionFilter_formalParameters (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceProjectionFilter_formalParameters (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceProjectionFilter_formalParameters)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceProjectionFilter_formalParameters */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceProjectionFilter_projectionColumns (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceProjectionFilter_projectionColumns (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceProjectionFilter_projectionColumns)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceProjectionFilter_projectionColumns */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceProjectionFilter_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceProjectionFilter_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceProjectionFilter_condition)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceProjectionFilter_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceProjectionFilter_connection (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceProjectionFilter_connection (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceProjectionFilter_connection)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceProjectionFilter_connection */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceProjectionFilter_from (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceProjectionFilter_from (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceProjectionFilter_from)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceProjectionFilter_from */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceProjectionFilter_sourceDataSource (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceProjectionFilter_sourceDataSource (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceProjectionFilter_sourceDataSource)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceProjectionFilter_sourceDataSource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_LimitDataSource (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_LimitDataSource (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_LimitDataSource)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_LimitDataSource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_LimitDataSource_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_LimitDataSource_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_LimitDataSource_name)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_LimitDataSource_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_LimitDataSource_formalParameters (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_LimitDataSource_formalParameters (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_LimitDataSource_formalParameters)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_LimitDataSource_formalParameters */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_LimitDataSource_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_LimitDataSource_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_LimitDataSource_condition)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_LimitDataSource_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_LimitDataSource_connection (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_LimitDataSource_connection (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_LimitDataSource_connection)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_LimitDataSource_connection */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_LimitDataSource_minRow (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_LimitDataSource_minRow (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_LimitDataSource_minRow)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_LimitDataSource_minRow */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_LimitDataSource_dots (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_LimitDataSource_dots (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_LimitDataSource_dots)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_LimitDataSource_dots */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_LimitDataSource_maxRow (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_LimitDataSource_maxRow (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_LimitDataSource_maxRow)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_LimitDataSource_maxRow */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_LimitDataSource_from (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_LimitDataSource_from (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_LimitDataSource_from)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_LimitDataSource_from */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_LimitDataSource_sourceDataSouce (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_LimitDataSource_sourceDataSouce (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_LimitDataSource_sourceDataSouce)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_LimitDataSource_sourceDataSouce */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceJoin (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceJoin (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceJoin)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceJoin */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceJoin_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceJoin_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceJoin_name)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceJoin_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceJoin_joinOperator (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceJoin_joinOperator (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceJoin_joinOperator)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceJoin_joinOperator */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceJoin_connection1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceJoin_connection1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceJoin_connection1)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceJoin_connection1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceJoin_connection2 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceJoin_connection2 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceJoin_connection2)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceJoin_connection2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceJoin_from1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceJoin_from1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceJoin_from1)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceJoin_from1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceJoin_dataSource1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceJoin_dataSource1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceJoin_dataSource1)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceJoin_dataSource1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceJoin_from2 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceJoin_from2 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceJoin_from2)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceJoin_from2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceJoin_dataSource2 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceJoin_dataSource2 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceJoin_dataSource2)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceJoin_dataSource2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceInsert (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceInsert (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceInsert)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceInsert */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceInsert_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceInsert_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceInsert_name)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceInsert_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceInsert_formalParams (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceInsert_formalParams (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceInsert_formalParams)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceInsert_formalParams */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceInsert_values (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceInsert_values (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceInsert_values)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceInsert_values */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceInsert_connection (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceInsert_connection (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceInsert_connection)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceInsert_connection */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceInsert_from (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceInsert_from (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceInsert_from)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceInsert_from */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceInsert_targetDataSource (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceInsert_targetDataSource (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceInsert_targetDataSource)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceInsert_targetDataSource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceUpdate (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceUpdate (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceUpdate)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceUpdate */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceUpdate_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceUpdate_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceUpdate_name)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceUpdate_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceUpdate_formalParams (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceUpdate_formalParams (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceUpdate_formalParams)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceUpdate_formalParams */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceUpdate_values (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceUpdate_values (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceUpdate_values)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceUpdate_values */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceUpdate_conditions (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceUpdate_conditions (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceUpdate_conditions)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceUpdate_conditions */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceUpdate_connection (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceUpdate_connection (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceUpdate_connection)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceUpdate_connection */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceUpdate_from (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceUpdate_from (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceUpdate_from)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceUpdate_from */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceUpdate_targetDataSource (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceUpdate_targetDataSource (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceUpdate_targetDataSource)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceUpdate_targetDataSource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceDelete (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceDelete (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceDelete)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceDelete */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceDelete_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceDelete_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceDelete_name)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceDelete_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceDelete_formalParams (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceDelete_formalParams (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceDelete_formalParams)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceDelete_formalParams */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceDelete_conditions (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceDelete_conditions (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceDelete_conditions)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceDelete_conditions */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceDelete_connection (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceDelete_connection (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceDelete_connection)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceDelete_connection */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceDelete_from (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceDelete_from (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceDelete_from)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceDelete_from */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceDelete_targetDataSource (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceDelete_targetDataSource (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceDelete_targetDataSource)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceDelete_targetDataSource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceDB (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceDB (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceDB)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceDB */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceDB_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceDB_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceDB_name)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceDB_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceDB_position (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceDB_position (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceDB_position)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceDB_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataSourceDB_columns (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataSourceDB_columns (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataSourceDB_columns)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataSourceDB_columns */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataBase_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataBase_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataBase_name)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataBase_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataBase_userName (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataBase_userName (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataBase_userName)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataBase_userName */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataBase_userPassword (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataBase_userPassword (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataBase_userPassword)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataBase_userPassword */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataBase_hostName (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataBase_hostName (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataBase_hostName)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataBase_hostName */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataBase_port (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataBase_port (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataBase_port)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataBase_port */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataBase_dataBaseType (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataBase_dataBaseType (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataBase_dataBaseType)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataBase_dataBaseType */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataBase_dataSources (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataBase_dataSources (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataBase_dataSources)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataBase_dataSources */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataBase_dataSourceModifications (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataBase_dataSourceModifications (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataBase_dataSourceModifications)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataBase_dataSourceModifications */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataBase_setSize (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataBase_setSize (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataBase_setSize)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataBase_setSize */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkdataBaseView_DataBase_modifySetSize (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkdataBaseView_DataBase_modifySetSize (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBdataBaseView_DataBase_modifySetSize)) return (_currn);
return(NULLNODEPTR);
}/* MkdataBaseView_DataBase_modifySetSize */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkfcDlg_FormularCondition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkfcDlg_FormularCondition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBfcDlg_FormularCondition)) return (_currn);
return(NULLNODEPTR);
}/* MkfcDlg_FormularCondition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkfcDlg_FormularCondition_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkfcDlg_FormularCondition_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBfcDlg_FormularCondition_condition)) return (_currn);
return(NULLNODEPTR);
}/* MkfcDlg_FormularCondition_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkfcDlg_FormularCondition_params (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkfcDlg_FormularCondition_params (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBfcDlg_FormularCondition_params)) return (_currn);
return(NULLNODEPTR);
}/* MkfcDlg_FormularCondition_params */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkfcDlg_FormularConditionActualParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkfcDlg_FormularConditionActualParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBfcDlg_FormularConditionActualParam)) return (_currn);
return(NULLNODEPTR);
}/* MkfcDlg_FormularConditionActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkfcDlg_FormularConditionActualParam_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkfcDlg_FormularConditionActualParam_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBfcDlg_FormularConditionActualParam_value)) return (_currn);
return(NULLNODEPTR);
}/* MkfcDlg_FormularConditionActualParam_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkfcDlg_FormularConditionActualParam_formalparam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkfcDlg_FormularConditionActualParam_formalparam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBfcDlg_FormularConditionActualParam_formalparam)) return (_currn);
return(NULLNODEPTR);
}/* MkfcDlg_FormularConditionActualParam_formalparam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_GlobalFrame (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_GlobalFrame (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_GlobalFrame)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_GlobalFrame */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularCheckboxElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularCheckboxElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularCheckboxElement)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularCheckboxElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularCheckboxElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularCheckboxElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularCheckboxElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularCheckboxElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularCheckboxElement_type (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularCheckboxElement_type (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularCheckboxElement_type)) return (_currn);
if (IsSymb (_currn, SYMBglobalFrameView_FormularCheckboxElement_name)) return (MkpglobalFrameView_FormularCheckboxElement_type(_coordref, _currn));
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularCheckboxElement_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularCheckboxElement_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularCheckboxElement_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularCheckboxElement_condition)) return (_currn);
if (IsSymb (_currn, SYMBglobalFrameView_FormularCondition)) return (Mkrule_123(_coordref, _currn));
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularCheckboxElement_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularCheckboxElement_description (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularCheckboxElement_description (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularCheckboxElement_description)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularCheckboxElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularCheckboxElement_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularCheckboxElement_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularCheckboxElement_name)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularCheckboxElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularCondition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularCondition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularCondition)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularCondition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularTextElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularTextElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularTextElement)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularTextElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularTextElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularTextElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularTextElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularTextElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularTextElement_type (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularTextElement_type (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularTextElement_type)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularTextElement_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularTextElement_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularTextElement_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularTextElement_condition)) return (_currn);
if (IsSymb (_currn, SYMBglobalFrameView_FormularCondition)) return (Mkrule_124(_coordref, _currn));
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularTextElement_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularTextElement_description (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularTextElement_description (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularTextElement_description)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularTextElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularPasswordElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularPasswordElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularPasswordElement)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularPasswordElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularPasswordElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularPasswordElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularPasswordElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularPasswordElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularPasswordElement_type (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularPasswordElement_type (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularPasswordElement_type)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularPasswordElement_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularPasswordElement_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularPasswordElement_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularPasswordElement_condition)) return (_currn);
if (IsSymb (_currn, SYMBglobalFrameView_FormularCondition)) return (Mkrule_125(_coordref, _currn));
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularPasswordElement_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularPasswordElement_description (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularPasswordElement_description (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularPasswordElement_description)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularPasswordElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularTextAreaElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularTextAreaElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularTextAreaElement)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularTextAreaElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularTextAreaElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularTextAreaElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularTextAreaElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularTextAreaElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularTextAreaElement_type (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularTextAreaElement_type (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularTextAreaElement_type)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularTextAreaElement_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularTextAreaElement_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularTextAreaElement_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularTextAreaElement_condition)) return (_currn);
if (IsSymb (_currn, SYMBglobalFrameView_FormularCondition)) return (Mkrule_126(_coordref, _currn));
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularTextAreaElement_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularTextAreaElement_description (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularTextAreaElement_description (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularTextAreaElement_description)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularTextAreaElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularSelectElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularSelectElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularSelectElement)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularSelectElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularSelectElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularSelectElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularSelectElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularSelectElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularSelectElement_multiple (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularSelectElement_multiple (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularSelectElement_multiple)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularSelectElement_multiple */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularSelectElement_options (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularSelectElement_options (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularSelectElement_options)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularSelectElement_options */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularSelectElement_description (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularSelectElement_description (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularSelectElement_description)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularSelectElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularPHPScriptSelectOption (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularPHPScriptSelectOption (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularPHPScriptSelectOption)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularPHPScriptSelectOption */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularDataSourceSelectOption (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularDataSourceSelectOption (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularDataSourceSelectOption)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularDataSourceSelectOption */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularConstantSelectOption (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularConstantSelectOption (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularConstantSelectOption)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularConstantSelectOption */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularRadioElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularRadioElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularRadioElement)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularRadioElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularRadioElement_type (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularRadioElement_type (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularRadioElement_type)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularRadioElement_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularRadioElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularRadioElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularRadioElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularRadioElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularRadioElement_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularRadioElement_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularRadioElement_name)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularRadioElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_DataSourceActualParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_DataSourceActualParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_DataSourceActualParam)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_DataSourceActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_DataSourceActualParam_formalParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_DataSourceActualParam_formalParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_DataSourceActualParam_formalParam)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_DataSourceActualParam_formalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_DataSourceActualParam_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_DataSourceActualParam_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_DataSourceActualParam_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_DataSourceActualParam_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_DataSourceActualParam_expressions (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_DataSourceActualParam_expressions)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_DataSourceActualParam_expressions */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_WebNodeColumnExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_WebNodeColumnExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_WebNodeColumnExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_WebNodeColumnExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_WebNodeVariableExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_WebNodeVariableExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_WebNodeVariableExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_WebNodeVariableExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_WebNodeConstantExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_WebNodeConstantExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_WebNodeConstantExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_WebNodeConstantExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_WebNodeActualParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_WebNodeActualParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_WebNodeActualParam)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_WebNodeActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_WebNodeActualParam_formalParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_WebNodeActualParam_formalParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_WebNodeActualParam_formalParam)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_WebNodeActualParam_formalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_WebNodeActualParam_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_WebNodeActualParam_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_WebNodeActualParam_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_WebNodeActualParam_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_WebNodeActualParam_expressions (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_WebNodeActualParam_expressions)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_WebNodeActualParam_expressions */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_WebNodeConstantExpression_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_WebNodeConstantExpression_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_WebNodeConstantExpression_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_WebNodeConstantExpression_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_WebNodeConstantExpression_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_WebNodeConstantExpression_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_WebNodeConstantExpression_value)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_WebNodeConstantExpression_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_WebNodeConstantExpression_keyword2 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_WebNodeConstantExpression_keyword2 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_WebNodeConstantExpression_keyword2)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_WebNodeConstantExpression_keyword2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_WebNodeVariableExpression_content (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_WebNodeVariableExpression_content (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_WebNodeVariableExpression_content)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_WebNodeVariableExpression_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_WebNodeColumnExpression_content (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_WebNodeColumnExpression_content (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_WebNodeColumnExpression_content)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_WebNodeColumnExpression_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_Formular (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_Formular (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_Formular)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_Formular */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_Formular_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_Formular_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_Formular_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_Formular_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_Formular_action (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_Formular_action (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_Formular_action)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_Formular_action */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_Formular_elements (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_Formular_elements (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_Formular_elements)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_Formular_elements */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_Formular_datasource (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_Formular_datasource (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_Formular_datasource)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_Formular_datasource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_Formular_actualParams (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_Formular_actualParams (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_Formular_actualParams)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_Formular_actualParams */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularRadioGroupElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularRadioGroupElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularRadioGroupElement)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularRadioGroupElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularRadioGroupElement_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularRadioGroupElement_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularRadioGroupElement_name)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularRadioGroupElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularRadioGroupElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularRadioGroupElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularRadioGroupElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularRadioGroupElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularRadioGroupElement_description (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularRadioGroupElement_description (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularRadioGroupElement_description)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularRadioGroupElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularRadioGroupElement_pvalue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularRadioGroupElement_pvalue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularRadioGroupElement_pvalue)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularRadioGroupElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularRadioGroupElement_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularRadioGroupElement_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularRadioGroupElement_condition)) return (_currn);
if (IsSymb (_currn, SYMBglobalFrameView_FormularCondition)) return (Mkrule_127(_coordref, _currn));
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularRadioGroupElement_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularRadioGroupElement_elements (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularRadioGroupElement_elements (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularRadioGroupElement_elements)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularRadioGroupElement_elements */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularRadioGroupElement_checked (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularRadioGroupElement_checked (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularRadioGroupElement_checked)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularRadioGroupElement_checked */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularConstantSelectOption_values (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularConstantSelectOption_values (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularConstantSelectOption_values)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularConstantSelectOption_values */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularDataSourceSelectOption_values (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularDataSourceSelectOption_values (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularDataSourceSelectOption_values)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularDataSourceSelectOption_values */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularDataSourceSelectOption_params (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularDataSourceSelectOption_params (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularDataSourceSelectOption_params)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularDataSourceSelectOption_params */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularPHPScriptSelectOption_script (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularPHPScriptSelectOption_script (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularPHPScriptSelectOption_script)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularPHPScriptSelectOption_script */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularCondition_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularCondition_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularCondition_condition)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularCondition_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularCondition_params (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularCondition_params (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularCondition_params)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularCondition_params */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularConditionActualParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularConditionActualParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularConditionActualParam)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularConditionActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularConditionActualParam_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularConditionActualParam_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularConditionActualParam_value)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularConditionActualParam_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FormularConditionActualParam_formalparam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FormularConditionActualParam_formalparam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FormularConditionActualParam_formalparam)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FormularConditionActualParam_formalparam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_Paragraph (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_Paragraph (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_Paragraph)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_Paragraph */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_Paragraph_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_Paragraph_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_Paragraph_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_Paragraph_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_Paragraph_fragments (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_Paragraph_fragments (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_Paragraph_fragments)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_Paragraph_fragments */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_OutputElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_OutputElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_OutputElement)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_OutputElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_ImageFragment (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_ImageFragment (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_ImageFragment)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_ImageFragment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_TextFragment (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_TextFragment (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_TextFragment)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_TextFragment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_TextFragment_link (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_TextFragment_link (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_TextFragment_link)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_TextFragment_link */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_TextFragment_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_TextFragment_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_TextFragment_value)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_TextFragment_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_ImageFragment_link (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_ImageFragment_link (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_ImageFragment_link)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_ImageFragment_link */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_ImageFragment_width (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_ImageFragment_width (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_ImageFragment_width)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_ImageFragment_width */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_ImageFragment_height (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_ImageFragment_height (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_ImageFragment_height)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_ImageFragment_height */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_ImageFragment_text (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_ImageFragment_text (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_ImageFragment_text)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_ImageFragment_text */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_ImageFragment_imageName (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_ImageFragment_imageName (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_ImageFragment_imageName)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_ImageFragment_imageName */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_Heading (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_Heading (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_Heading)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_Heading */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_Heading_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_Heading_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_Heading_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_Heading_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_Heading_text (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_Heading_text (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_Heading_text)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_Heading_text */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_Heading_level (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_Heading_level (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_Heading_level)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_Heading_level */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_Table (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_Table (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_Table)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_Table */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_Table_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_Table_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_Table_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_Table_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_Table_rows (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_Table_rows (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_Table_rows)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_Table_rows */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_Table_columns (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_Table_columns (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_Table_columns)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_Table_columns */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_TableRow (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_TableRow (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_TableRow)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_TableRow */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_TableColumn (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_TableColumn (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_TableColumn)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_TableColumn */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_TableRow_cells (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_TableRow_cells (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_TableRow_cells)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_TableRow_cells */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_TableCell (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_TableCell (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_TableCell)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_TableCell */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_TableCell_columnRef (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_TableCell_columnRef (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_TableCell_columnRef)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_TableCell_columnRef */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_TableCell_content (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_TableCell_content (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_TableCell_content)) return (_currn);
if (IsSymb (_currn, SYMBglobalFrameView_Frame)) return (Mkrule_128(_coordref, _currn));
return(NULLNODEPTR);
}/* MkglobalFrameView_TableCell_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_Frame (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_Frame (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_Frame)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_Frame */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_Frame_components (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_Frame_components (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_Frame_components)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_Frame_components */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_Frame_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_Frame_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_Frame_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_Frame_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_ListOutputElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_ListOutputElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_ListOutputElement)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_ListOutputElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FrameRef (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FrameRef (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FrameRef)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FrameRef */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_GlobalFrame_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_GlobalFrame_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_GlobalFrame_name)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_GlobalFrame_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_GlobalFrame_components (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_GlobalFrame_components (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_GlobalFrame_components)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_GlobalFrame_components */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_GlobalFrame_setSize (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_GlobalFrame_setSize (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_GlobalFrame_setSize)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_GlobalFrame_setSize */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FrameRef_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FrameRef_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FrameRef_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FrameRef_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_FrameRef_frame (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_FrameRef_frame (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_FrameRef_frame)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_FrameRef_frame */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_OutputElement_link (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_OutputElement_link (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_OutputElement_link)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_OutputElement_link */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_OutputElement_content (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_OutputElement_content (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_OutputElement_content)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_OutputElement_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_ListOutputElement_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_ListOutputElement_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_ListOutputElement_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_ListOutputElement_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_ListOutputElement_dataSourceActualParams (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_ListOutputElement_dataSourceActualParams (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_ListOutputElement_dataSourceActualParams)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_ListOutputElement_dataSourceActualParams */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_ListOutputElement_dataSource (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_ListOutputElement_dataSource (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_ListOutputElement_dataSource)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_ListOutputElement_dataSource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_ListOutputElement_maxNumberOfElements (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_ListOutputElement_maxNumberOfElements (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_ListOutputElement_maxNumberOfElements)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_ListOutputElement_maxNumberOfElements */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkglobalFrameView_ListOutputElement_navigationElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkglobalFrameView_ListOutputElement_navigationElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBglobalFrameView_ListOutputElement_navigationElement)) return (_currn);
return(NULLNODEPTR);
}/* MkglobalFrameView_ListOutputElement_navigationElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkmetaInformation_MetaInformation (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkmetaInformation_MetaInformation (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBmetaInformation_MetaInformation)) return (_currn);
return(NULLNODEPTR);
}/* MkmetaInformation_MetaInformation */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkmetaInformation_MetaInformation_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkmetaInformation_MetaInformation_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBmetaInformation_MetaInformation_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkmetaInformation_MetaInformation_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkmetaInformation_MetaInformation_keyword2 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkmetaInformation_MetaInformation_keyword2 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBmetaInformation_MetaInformation_keyword2)) return (_currn);
return(NULLNODEPTR);
}/* MkmetaInformation_MetaInformation_keyword2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkmetaInformation_MetaInformation_keyword3 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkmetaInformation_MetaInformation_keyword3 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBmetaInformation_MetaInformation_keyword3)) return (_currn);
return(NULLNODEPTR);
}/* MkmetaInformation_MetaInformation_keyword3 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkmetaInformation_MetaInformation_elements (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkmetaInformation_MetaInformation_elements (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBmetaInformation_MetaInformation_elements)) return (_currn);
return(NULLNODEPTR);
}/* MkmetaInformation_MetaInformation_elements */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkmetaInformation_MetaInformation_keyword4 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkmetaInformation_MetaInformation_keyword4 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBmetaInformation_MetaInformation_keyword4)) return (_currn);
return(NULLNODEPTR);
}/* MkmetaInformation_MetaInformation_keyword4 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkmetaInformation_MetaInformationElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkmetaInformation_MetaInformationElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBmetaInformation_MetaInformationElement)) return (_currn);
return(NULLNODEPTR);
}/* MkmetaInformation_MetaInformationElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkmetaInformation_MetaInformationElement_property (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkmetaInformation_MetaInformationElement_property (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBmetaInformation_MetaInformationElement_property)) return (_currn);
return(NULLNODEPTR);
}/* MkmetaInformation_MetaInformationElement_property */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkmetaInformation_MetaInformationElement_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkmetaInformation_MetaInformationElement_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBmetaInformation_MetaInformationElement_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkmetaInformation_MetaInformationElement_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkmetaInformation_MetaInformationElement_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkmetaInformation_MetaInformationElement_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBmetaInformation_MetaInformationElement_value)) return (_currn);
return(NULLNODEPTR);
}/* MkmetaInformation_MetaInformationElement_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkmetaInformation_MetaInformationElement_keyword2 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkmetaInformation_MetaInformationElement_keyword2 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBmetaInformation_MetaInformationElement_keyword2)) return (_currn);
return(NULLNODEPTR);
}/* MkmetaInformation_MetaInformationElement_keyword2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MknavigationBarView_NavigationBar (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MknavigationBarView_NavigationBar (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBnavigationBarView_NavigationBar)) return (_currn);
return(NULLNODEPTR);
}/* MknavigationBarView_NavigationBar */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MknavigationBarView_NavigationBar_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MknavigationBarView_NavigationBar_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBnavigationBarView_NavigationBar_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MknavigationBarView_NavigationBar_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MknavigationBarView_NavigationBar_elements (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MknavigationBarView_NavigationBar_elements (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBnavigationBarView_NavigationBar_elements)) return (_currn);
return(NULLNODEPTR);
}/* MknavigationBarView_NavigationBar_elements */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MknavigationBarView_NavigationBar_alignment (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MknavigationBarView_NavigationBar_alignment (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBnavigationBarView_NavigationBar_alignment)) return (_currn);
return(NULLNODEPTR);
}/* MknavigationBarView_NavigationBar_alignment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MknavigationBarView_NavigationBarElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MknavigationBarView_NavigationBarElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBnavigationBarView_NavigationBarElement)) return (_currn);
return(NULLNODEPTR);
}/* MknavigationBarView_NavigationBarElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MknavigationBarView_NavigationBarElement_target (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MknavigationBarView_NavigationBarElement_target (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBnavigationBarView_NavigationBarElement_target)) return (_currn);
return(NULLNODEPTR);
}/* MknavigationBarView_NavigationBarElement_target */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Root (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Root (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Root)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_Root */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebNodeFormalParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebNodeFormalParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebNodeFormalParam)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebNodeFormalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebNodeFormalParam_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebNodeFormalParam_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebNodeFormalParam_name)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebNodeFormalParam_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebNodeFormalParam_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebNodeFormalParam_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebNodeFormalParam_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebNodeFormalParam_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebNodeFormalParam_type (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebNodeFormalParam_type)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebNodeFormalParam_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_DBBlob (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_DBBlob (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_DBBlob)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_DBBlob */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_DBBoolean (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_DBBoolean (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_DBBoolean)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_DBBoolean */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_DBTimeStamp (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_DBTimeStamp (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_DBTimeStamp)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_DBTimeStamp */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_DBDate (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_DBDate (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_DBDate)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_DBDate */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_DBBigInt (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_DBBigInt (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_DBBigInt)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_DBBigInt */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_DBDouble (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_DBDouble (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_DBDouble)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_DBDouble */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_DBFloat (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_DBFloat (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_DBFloat)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_DBFloat */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_DBInteger (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_DBInteger (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_DBInteger)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_DBInteger */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_DBString (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_DBString (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_DBString)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_DBString */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_DataBase (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_DataBase (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_DataBase)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_DataBase */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_DBString_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_DBString_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_DBString_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_DBString_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_DBInteger_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_DBInteger_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_DBInteger_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_DBInteger_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_DBFloat_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_DBFloat_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_DBFloat_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_DBFloat_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_DBDouble_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_DBDouble_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_DBDouble_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_DBDouble_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_DBBigInt_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_DBBigInt_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_DBBigInt_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_DBBigInt_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_DBDate_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_DBDate_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_DBDate_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_DBDate_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_DBTimeStamp_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_DBTimeStamp_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_DBTimeStamp_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_DBTimeStamp_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_DBBoolean_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_DBBoolean_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_DBBoolean_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_DBBoolean_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_DBBlob_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_DBBlob_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_DBBlob_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_DBBlob_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_DataSourceActualParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_DataSourceActualParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_DataSourceActualParam)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_DataSourceActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_DataSourceActualParam_formalParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_DataSourceActualParam_formalParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_DataSourceActualParam_formalParam)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_DataSourceActualParam_formalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_DataSourceActualParam_expressions (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_DataSourceActualParam_expressions)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_DataSourceActualParam_expressions */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebNodeColumnExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebNodeColumnExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebNodeColumnExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebNodeColumnExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebNodeVariableExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebNodeVariableExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebNodeVariableExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebNodeVariableExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebNodeConstantExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebNodeConstantExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebNodeConstantExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebNodeConstantExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Formular (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Formular (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Formular)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_Formular */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Formular_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Formular_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Formular_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_Formular_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Formular_action (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Formular_action (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Formular_action)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_Formular_action */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Formular_elements (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Formular_elements (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Formular_elements)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_Formular_elements */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Formular_datasource (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Formular_datasource (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Formular_datasource)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_Formular_datasource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Formular_actualParams (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Formular_actualParams (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Formular_actualParams)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_Formular_actualParams */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularSelectElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularSelectElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularSelectElement)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularSelectElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularCheckboxElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularCheckboxElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularCheckboxElement)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularCheckboxElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularPasswordElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularPasswordElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularPasswordElement)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularPasswordElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularTextAreaElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularTextAreaElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularTextAreaElement)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularTextAreaElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularTextElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularTextElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularTextElement)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularTextElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularRadioGroupElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularRadioGroupElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularRadioGroupElement)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularRadioGroupElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebNodeActualParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebNodeActualParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebNodeActualParam)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebNodeActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularRadioGroupElement_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularRadioGroupElement_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularRadioGroupElement_name)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularRadioGroupElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularRadioGroupElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularRadioGroupElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularRadioGroupElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularRadioGroupElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularRadioGroupElement_description (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularRadioGroupElement_description (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularRadioGroupElement_description)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularRadioGroupElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularRadioGroupElement_pvalue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularRadioGroupElement_pvalue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularRadioGroupElement_pvalue)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularRadioGroupElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularRadioGroupElement_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularRadioGroupElement_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularRadioGroupElement_condition)) return (_currn);
if (IsSymb (_currn, SYMBrootView_FormularCondition)) return (Mkrule_129(_coordref, _currn));
return(NULLNODEPTR);
}/* MkrootView_FormularRadioGroupElement_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularRadioGroupElement_elements (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularRadioGroupElement_elements (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularRadioGroupElement_elements)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularRadioGroupElement_elements */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularRadioGroupElement_checked (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularRadioGroupElement_checked (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularRadioGroupElement_checked)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularRadioGroupElement_checked */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularCondition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularCondition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularCondition)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularCondition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularRadioElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularRadioElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularRadioElement)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularRadioElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularTextElement_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularTextElement_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularTextElement_name)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularTextElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularTextElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularTextElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularTextElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularTextElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularTextElement_description (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularTextElement_description (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularTextElement_description)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularTextElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularTextElement_pvalue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularTextElement_pvalue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularTextElement_pvalue)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularTextElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularTextElement_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularTextElement_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularTextElement_condition)) return (_currn);
if (IsSymb (_currn, SYMBrootView_FormularCondition)) return (Mkrule_130(_coordref, _currn));
return(NULLNODEPTR);
}/* MkrootView_FormularTextElement_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularTextAreaElement_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularTextAreaElement_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularTextAreaElement_name)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularTextAreaElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularTextAreaElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularTextAreaElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularTextAreaElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularTextAreaElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularTextAreaElement_description (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularTextAreaElement_description (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularTextAreaElement_description)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularTextAreaElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularTextAreaElement_pvalue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularTextAreaElement_pvalue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularTextAreaElement_pvalue)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularTextAreaElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularTextAreaElement_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularTextAreaElement_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularTextAreaElement_condition)) return (_currn);
if (IsSymb (_currn, SYMBrootView_FormularCondition)) return (Mkrule_131(_coordref, _currn));
return(NULLNODEPTR);
}/* MkrootView_FormularTextAreaElement_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularPasswordElement_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularPasswordElement_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularPasswordElement_name)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularPasswordElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularPasswordElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularPasswordElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularPasswordElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularPasswordElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularPasswordElement_description (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularPasswordElement_description (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularPasswordElement_description)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularPasswordElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularPasswordElement_pvalue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularPasswordElement_pvalue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularPasswordElement_pvalue)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularPasswordElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularPasswordElement_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularPasswordElement_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularPasswordElement_condition)) return (_currn);
if (IsSymb (_currn, SYMBrootView_FormularCondition)) return (Mkrule_132(_coordref, _currn));
return(NULLNODEPTR);
}/* MkrootView_FormularPasswordElement_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularCheckboxElement_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularCheckboxElement_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularCheckboxElement_name)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularCheckboxElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularCheckboxElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularCheckboxElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularCheckboxElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularCheckboxElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularCheckboxElement_description (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularCheckboxElement_description (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularCheckboxElement_description)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularCheckboxElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularCheckboxElement_pvalue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularCheckboxElement_pvalue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularCheckboxElement_pvalue)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularCheckboxElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularCheckboxElement_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularCheckboxElement_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularCheckboxElement_condition)) return (_currn);
if (IsSymb (_currn, SYMBrootView_FormularCondition)) return (Mkrule_133(_coordref, _currn));
return(NULLNODEPTR);
}/* MkrootView_FormularCheckboxElement_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularCheckboxElement_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularCheckboxElement_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularCheckboxElement_value)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularCheckboxElement_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularCheckboxElement_checked (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularCheckboxElement_checked (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularCheckboxElement_checked)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularCheckboxElement_checked */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularSelectElement_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularSelectElement_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularSelectElement_name)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularSelectElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularSelectElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularSelectElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularSelectElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularSelectElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularSelectElement_description (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularSelectElement_description (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularSelectElement_description)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularSelectElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularSelectElement_pvalue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularSelectElement_pvalue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularSelectElement_pvalue)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularSelectElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularSelectElement_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularSelectElement_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularSelectElement_condition)) return (_currn);
if (IsSymb (_currn, SYMBrootView_FormularCondition)) return (Mkrule_134(_coordref, _currn));
return(NULLNODEPTR);
}/* MkrootView_FormularSelectElement_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularSelectElement_options (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularSelectElement_options (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularSelectElement_options)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularSelectElement_options */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularSelectElement_size (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularSelectElement_size (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularSelectElement_size)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularSelectElement_size */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularSelectElement_multiple (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularSelectElement_multiple (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularSelectElement_multiple)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularSelectElement_multiple */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularPHPScriptSelectOption (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularPHPScriptSelectOption (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularPHPScriptSelectOption)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularPHPScriptSelectOption */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularDataSourceSelectOption (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularDataSourceSelectOption (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularDataSourceSelectOption)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularDataSourceSelectOption */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularConstantSelectOption (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularConstantSelectOption (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularConstantSelectOption)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularConstantSelectOption */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularConstantSelectOption_values (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularConstantSelectOption_values (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularConstantSelectOption_values)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularConstantSelectOption_values */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularDataSourceSelectOption_values (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularDataSourceSelectOption_values (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularDataSourceSelectOption_values)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularDataSourceSelectOption_values */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularDataSourceSelectOption_params (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularDataSourceSelectOption_params (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularDataSourceSelectOption_params)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularDataSourceSelectOption_params */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularPHPScriptSelectOption_script (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularPHPScriptSelectOption_script (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularPHPScriptSelectOption_script)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularPHPScriptSelectOption_script */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularRadioElement_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularRadioElement_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularRadioElement_value)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularRadioElement_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularRadioElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularRadioElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularRadioElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularRadioElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularRadioElement_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularRadioElement_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularRadioElement_name)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularRadioElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularCondition_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularCondition_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularCondition_condition)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularCondition_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularCondition_params (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularCondition_params (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularCondition_params)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularCondition_params */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularConditionActualParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularConditionActualParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularConditionActualParam)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularConditionActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularConditionDefinition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularConditionDefinition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularConditionDefinition)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularConditionDefinition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularConditionDefinition_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularConditionDefinition_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularConditionDefinition_name)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularConditionDefinition_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularConditionDefinition_params (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularConditionDefinition_params (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularConditionDefinition_params)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularConditionDefinition_params */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularConditionDefinition_body (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularConditionDefinition_body (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularConditionDefinition_body)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularConditionDefinition_body */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularConditionFormalParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularConditionFormalParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularConditionFormalParam)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularConditionFormalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularConditionFormalParam_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularConditionFormalParam_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularConditionFormalParam_name)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularConditionFormalParam_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularConditionActualParam_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularConditionActualParam_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularConditionActualParam_value)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularConditionActualParam_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FormularConditionActualParam_formalparam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FormularConditionActualParam_formalparam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FormularConditionActualParam_formalparam)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FormularConditionActualParam_formalparam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebNodeActualParam_formalParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebNodeActualParam_formalParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebNodeActualParam_formalParam)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebNodeActualParam_formalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebNodeActualParam_expressions (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebNodeActualParam_expressions)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebNodeActualParam_expressions */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebNodeConstantExpression_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebNodeConstantExpression_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebNodeConstantExpression_value)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebNodeConstantExpression_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebNodeVariableExpression_content (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebNodeVariableExpression_content (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebNodeVariableExpression_content)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebNodeVariableExpression_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebNodeColumnExpression_content (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebNodeColumnExpression_content (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebNodeColumnExpression_content)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebNodeColumnExpression_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Paragraph (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Paragraph (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Paragraph)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_Paragraph */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Paragraph_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Paragraph_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Paragraph_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_Paragraph_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Paragraph_fragments (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Paragraph_fragments (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Paragraph_fragments)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_Paragraph_fragments */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_OutputElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_OutputElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_OutputElement)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_OutputElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_ImageFragment (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_ImageFragment (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_ImageFragment)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_ImageFragment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_TextFragment (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_TextFragment (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_TextFragment)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_TextFragment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_TextFragment_link (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_TextFragment_link (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_TextFragment_link)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_TextFragment_link */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_TextFragment_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_TextFragment_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_TextFragment_value)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_TextFragment_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_ImageFragment_link (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_ImageFragment_link (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_ImageFragment_link)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_ImageFragment_link */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_ImageFragment_width (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_ImageFragment_width (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_ImageFragment_width)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_ImageFragment_width */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_ImageFragment_height (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_ImageFragment_height (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_ImageFragment_height)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_ImageFragment_height */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_ImageFragment_text (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_ImageFragment_text (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_ImageFragment_text)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_ImageFragment_text */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_ImageFragment_imageName (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_ImageFragment_imageName (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_ImageFragment_imageName)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_ImageFragment_imageName */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Heading (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Heading (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Heading)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_Heading */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Heading_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Heading_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Heading_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_Heading_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Heading_text (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Heading_text (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Heading_text)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_Heading_text */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Heading_level (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Heading_level (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Heading_level)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_Heading_level */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Table (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Table (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Table)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_Table */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Table_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Table_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Table_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_Table_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Table_rows (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Table_rows (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Table_rows)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_Table_rows */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Table_columns (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Table_columns (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Table_columns)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_Table_columns */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_TableRow (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_TableRow (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_TableRow)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_TableRow */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_TableColumn (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_TableColumn (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_TableColumn)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_TableColumn */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_TableRow_cells (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_TableRow_cells (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_TableRow_cells)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_TableRow_cells */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_TableCell (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_TableCell (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_TableCell)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_TableCell */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_TableCell_columnRef (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_TableCell_columnRef (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_TableCell_columnRef)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_TableCell_columnRef */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_TableCell_content (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_TableCell_content (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_TableCell_content)) return (_currn);
if (IsSymb (_currn, SYMBrootView_Frame)) return (Mkrule_135(_coordref, _currn));
return(NULLNODEPTR);
}/* MkrootView_TableCell_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Frame (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Frame (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Frame)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_Frame */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_LinkTargetExtern (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_LinkTargetExtern (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_LinkTargetExtern)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_LinkTargetExtern */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_LinkTargetExtern_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_LinkTargetExtern_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_LinkTargetExtern_name)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_LinkTargetExtern_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_LinkTargetExtern_url (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_LinkTargetExtern_url (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_LinkTargetExtern_url)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_LinkTargetExtern_url */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Root_webApp (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Root_webApp (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Root_webApp)) return (_currn);
if (IsSymb (_currn, SYMBrootView_WebApplication)) return (Mkrule_136(_coordref, _currn));
return(NULLNODEPTR);
}/* MkrootView_Root_webApp */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Root_library (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Root_library (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Root_library)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_Root_library */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebApplication (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebApplication (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebApplication)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebApplication */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Library (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Library (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Library)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_Library */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Library_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Library_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Library_name)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_Library_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Library_formularconditions (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Library_formularconditions (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Library_formularconditions)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_Library_formularconditions */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Library_navigationElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Library_navigationElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Library_navigationElement)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_Library_navigationElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_NavigationElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_NavigationElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_NavigationElement)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_NavigationElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_NavigationElement_code (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_NavigationElement_code (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_NavigationElement_code)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_NavigationElement_code */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_NavigationElement_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_NavigationElement_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_NavigationElement_name)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_NavigationElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebApplication_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebApplication_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebApplication_name)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebApplication_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebApplication_webZone (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebApplication_webZone (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebApplication_webZone)) return (_currn);
if (IsSymb (_currn, SYMBrootView_RootWebZone)) return (Mkrule_137(_coordref, _currn));
return(NULLNODEPTR);
}/* MkrootView_WebApplication_webZone */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebApplication_navigationBar (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebApplication_navigationBar (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebApplication_navigationBar)) return (_currn);
if (IsSymb (_currn, SYMBrootView_NavigationBar)) return (Mkrule_138(_coordref, _currn));
return(NULLNODEPTR);
}/* MkrootView_WebApplication_navigationBar */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebApplication_styleSheet (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebApplication_styleSheet (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebApplication_styleSheet)) return (_currn);
if (IsSymb (_currn, SYMBrootView_StyleSheet)) return (Mkrule_139(_coordref, _currn));
return(NULLNODEPTR);
}/* MkrootView_WebApplication_styleSheet */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebApplication_frames (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebApplication_frames (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebApplication_frames)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebApplication_frames */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebApplication_meta (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebApplication_meta (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebApplication_meta)) return (_currn);
if (IsSymb (_currn, SYMBrootView_MetaInformation)) return (Mkrule_140(_coordref, _currn));
return(NULLNODEPTR);
}/* MkrootView_WebApplication_meta */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebApplication_database (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebApplication_database (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebApplication_database)) return (_currn);
if (IsSymb (_currn, SYMBrootView_DataBase)) return (Mkrule_141(_coordref, _currn));
return(NULLNODEPTR);
}/* MkrootView_WebApplication_database */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebApplication_roles (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebApplication_roles (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebApplication_roles)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebApplication_roles */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebApplication_linksExtern (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebApplication_linksExtern (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebApplication_linksExtern)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebApplication_linksExtern */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebApplication_imagePath (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebApplication_imagePath (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebApplication_imagePath)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebApplication_imagePath */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_RootWebZone (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_RootWebZone (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_RootWebZone)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_RootWebZone */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_NavigationBar (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_NavigationBar (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_NavigationBar)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_NavigationBar */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_StyleSheet (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_StyleSheet (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_StyleSheet)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_StyleSheet */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_GlobalFrame (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_GlobalFrame (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_GlobalFrame)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_GlobalFrame */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_MetaInformation (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_MetaInformation (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_MetaInformation)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_MetaInformation */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebRole (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebRole (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebRole)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebRole */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_RootWebZone_webTargets (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_RootWebZone_webTargets (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_RootWebZone_webTargets)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_RootWebZone_webTargets */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_RootWebZone_errorLink (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_RootWebZone_errorLink (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_RootWebZone_errorLink)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_RootWebZone_errorLink */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_RootWebZone_roles (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_RootWebZone_roles (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_RootWebZone_roles)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_RootWebZone_roles */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_RootWebZone_setSize (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_RootWebZone_setSize (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_RootWebZone_setSize)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_RootWebZone_setSize */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebAction (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebAction (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebAction)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebAction */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebPage (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebPage (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebPage)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebPage */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_SubWebZone (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_SubWebZone (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_SubWebZone)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_SubWebZone */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebRoleRef (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebRoleRef (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebRoleRef)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebRoleRef */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_SubWebZone_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_SubWebZone_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_SubWebZone_name)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_SubWebZone_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_SubWebZone_webTargets (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_SubWebZone_webTargets (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_SubWebZone_webTargets)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_SubWebZone_webTargets */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_SubWebZone_errorLink (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_SubWebZone_errorLink (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_SubWebZone_errorLink)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_SubWebZone_errorLink */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_SubWebZone_globalparams (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_SubWebZone_globalparams (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_SubWebZone_globalparams)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_SubWebZone_globalparams */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_SubWebZone_roles (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_SubWebZone_roles (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_SubWebZone_roles)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_SubWebZone_roles */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_SubWebZone_setSize (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_SubWebZone_setSize (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_SubWebZone_setSize)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_SubWebZone_setSize */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_SubWebZone_position (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_SubWebZone_position (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_SubWebZone_position)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_SubWebZone_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebRole_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebRole_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebRole_name)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebRole_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebRole_color (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebRole_color (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebRole_color)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebRole_color */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebRoleRef_role (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebRoleRef_role (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebRoleRef_role)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebRoleRef_role */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebPage_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebPage_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebPage_name)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebPage_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebPage_formalParameter (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebPage_formalParameter (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebPage_formalParameter)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebPage_formalParameter */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebPage_title (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebPage_title (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebPage_title)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebPage_title */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebPage_frame (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebPage_frame (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebPage_frame)) return (_currn);
if (IsSymb (_currn, SYMBrootView_Frame)) return (Mkrule_142(_coordref, _currn));
return(NULLNODEPTR);
}/* MkrootView_WebPage_frame */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebPage_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebPage_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebPage_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebPage_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebPage_position (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebPage_position (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebPage_position)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebPage_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Frame_components (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Frame_components (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Frame_components)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_Frame_components */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_Frame_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_Frame_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_Frame_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_Frame_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_ListOutputElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_ListOutputElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_ListOutputElement)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_ListOutputElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FrameRef (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FrameRef (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FrameRef)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FrameRef */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_GlobalFrame_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_GlobalFrame_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_GlobalFrame_name)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_GlobalFrame_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_GlobalFrame_components (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_GlobalFrame_components (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_GlobalFrame_components)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_GlobalFrame_components */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_GlobalFrame_setSize (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_GlobalFrame_setSize (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_GlobalFrame_setSize)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_GlobalFrame_setSize */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_NavigationBar_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_NavigationBar_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_NavigationBar_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_NavigationBar_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_NavigationBar_elements (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_NavigationBar_elements (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_NavigationBar_elements)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_NavigationBar_elements */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_NavigationBar_alignment (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_NavigationBar_alignment (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_NavigationBar_alignment)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_NavigationBar_alignment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_NavigationBarElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_NavigationBarElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_NavigationBarElement)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_NavigationBarElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_NavigationBarElement_target (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_NavigationBarElement_target (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_NavigationBarElement_target)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_NavigationBarElement_target */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FrameRef_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FrameRef_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FrameRef_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FrameRef_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FrameRef_frame (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FrameRef_frame (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FrameRef_frame)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FrameRef_frame */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_MetaInformation_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_MetaInformation_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_MetaInformation_name)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_MetaInformation_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_MetaInformation_elements (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_MetaInformation_elements (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_MetaInformation_elements)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_MetaInformation_elements */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_MetaInformationElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_MetaInformationElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_MetaInformationElement)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_MetaInformationElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_MetaInformationElement_property (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_MetaInformationElement_property (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_MetaInformationElement_property)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_MetaInformationElement_property */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_MetaInformationElement_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_MetaInformationElement_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_MetaInformationElement_value)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_MetaInformationElement_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_OutputElement_link (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_OutputElement_link (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_OutputElement_link)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_OutputElement_link */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_OutputElement_content (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_OutputElement_content (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_OutputElement_content)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_OutputElement_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_ListOutputElement_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_ListOutputElement_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_ListOutputElement_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_ListOutputElement_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_ListOutputElement_dataSourceActualParams (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_ListOutputElement_dataSourceActualParams (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_ListOutputElement_dataSourceActualParams)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_ListOutputElement_dataSourceActualParams */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_ListOutputElement_dataSource (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_ListOutputElement_dataSource (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_ListOutputElement_dataSource)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_ListOutputElement_dataSource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_ListOutputElement_maxNumberOfElements (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_ListOutputElement_maxNumberOfElements (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_ListOutputElement_maxNumberOfElements)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_ListOutputElement_maxNumberOfElements */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_ListOutputElement_navigationElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_ListOutputElement_navigationElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_ListOutputElement_navigationElement)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_ListOutputElement_navigationElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_StyleSheet_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_StyleSheet_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_StyleSheet_name)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_StyleSheet_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_StyleSheet_custom (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_StyleSheet_custom (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_StyleSheet_custom)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_StyleSheet_custom */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_StyleSheet_default (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_StyleSheet_default (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_StyleSheet_default)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_StyleSheet_default */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_CustomStyleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_CustomStyleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_CustomStyleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_CustomStyleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_HtmlStyleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_HtmlStyleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_HtmlStyleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_HtmlStyleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_CustomStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_CustomStyleSheetBlock_pairs (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_CustomStyleSheetBlock_pairs)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_CustomStyleSheetBlock_pairs */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_CustomStyleSheetBlock_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_CustomStyleSheetBlock_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_CustomStyleSheetBlock_name)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_CustomStyleSheetBlock_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_StyleSheetPair (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_StyleSheetPair (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_StyleSheetPair)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_StyleSheetPair */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_StyleSheetPair_category (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_StyleSheetPair_category (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_StyleSheetPair_category)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_StyleSheetPair_category */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_StyleSheetPair_attribute (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_StyleSheetPair_attribute (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_StyleSheetPair_attribute)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_StyleSheetPair_attribute */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_StyleSheetPair_valueCategory (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_StyleSheetPair_valueCategory (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_StyleSheetPair_valueCategory)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_StyleSheetPair_valueCategory */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_StyleSheetPair_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_StyleSheetPair_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_StyleSheetPair_value)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_StyleSheetPair_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_CustomValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_CustomValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_CustomValue)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_CustomValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_ListValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_ListValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_ListValue)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_ListValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_AlignmentValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_AlignmentValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_AlignmentValue)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_AlignmentValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_HeadingValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_HeadingValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_HeadingValue)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_HeadingValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_StyleValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_StyleValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_StyleValue)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_StyleValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FamilyValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FamilyValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FamilyValue)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FamilyValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_ColorValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_ColorValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_ColorValue)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_ColorValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_SizeValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_SizeValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_SizeValue)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_SizeValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_SizeValue_measure (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_SizeValue_measure (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_SizeValue_measure)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_SizeValue_measure */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_SizeValue_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_SizeValue_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_SizeValue_value)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_SizeValue_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_ColorValue_color (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_ColorValue_color (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_ColorValue_color)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_ColorValue_color */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_FamilyValue_family (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_FamilyValue_family (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_FamilyValue_family)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_FamilyValue_family */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_StyleValue_style (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_StyleValue_style (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_StyleValue_style)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_StyleValue_style */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_HeadingValue_heading (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_HeadingValue_heading (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_HeadingValue_heading)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_HeadingValue_heading */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_AlignmentValue_alignment (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_AlignmentValue_alignment (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_AlignmentValue_alignment)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_AlignmentValue_alignment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_ListValue_listed (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_ListValue_listed (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_ListValue_listed)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_ListValue_listed */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_CustomValue_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_CustomValue_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_CustomValue_value)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_CustomValue_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_HtmlStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_HtmlStyleSheetBlock_pairs (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_HtmlStyleSheetBlock_pairs)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_HtmlStyleSheetBlock_pairs */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_HtmlStyleSheetBlock_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_HtmlStyleSheetBlock_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_HtmlStyleSheetBlock_name)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_HtmlStyleSheetBlock_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_ActionHandler (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_ActionHandler (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_ActionHandler)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_ActionHandler */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_ActionHandler_targetpage (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_ActionHandler_targetpage (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_ActionHandler_targetpage)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_ActionHandler_targetpage */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_ActionHandler_params (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_ActionHandler_params (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_ActionHandler_params)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_ActionHandler_params */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebAction_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebAction_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebAction_name)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebAction_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebAction_formalParameter (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebAction_formalParameter (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebAction_formalParameter)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebAction_formalParameter */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebAction_successpage (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebAction_successpage (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebAction_successpage)) return (_currn);
if (IsSymb (_currn, SYMBrootView_ActionHandler)) return (Mkrule_143(_coordref, _currn));
return(NULLNODEPTR);
}/* MkrootView_WebAction_successpage */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebAction_operations (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebAction_operations (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebAction_operations)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebAction_operations */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_WebAction_position (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_WebAction_position (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_WebAction_position)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_WebAction_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_DataBaseOperation (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_DataBaseOperation (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_DataBaseOperation)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_DataBaseOperation */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_PHPOperation (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_PHPOperation (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_PHPOperation)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_PHPOperation */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_PHPOperation_errorhandler (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_PHPOperation_errorhandler (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_PHPOperation_errorhandler)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_PHPOperation_errorhandler */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_PHPOperation_content (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_PHPOperation_content (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_PHPOperation_content)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_PHPOperation_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_DataBaseOperation_errorhandler (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_DataBaseOperation_errorhandler (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_DataBaseOperation_errorhandler)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_DataBaseOperation_errorhandler */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_DataBaseOperation_filter (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_DataBaseOperation_filter (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_DataBaseOperation_filter)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_DataBaseOperation_filter */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkrootView_DataBaseOperation_params (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkrootView_DataBaseOperation_params (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBrootView_DataBaseOperation_params)) return (_currn);
return(NULLNODEPTR);
}/* MkrootView_DataBaseOperation_params */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MksspDlg_StyleSheetPair (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MksspDlg_StyleSheetPair (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBsspDlg_StyleSheetPair)) return (_currn);
return(NULLNODEPTR);
}/* MksspDlg_StyleSheetPair */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MksspDlg_StyleSheetPair_category (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MksspDlg_StyleSheetPair_category (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBsspDlg_StyleSheetPair_category)) return (_currn);
return(NULLNODEPTR);
}/* MksspDlg_StyleSheetPair_category */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MksspDlg_StyleSheetPair_attribute (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MksspDlg_StyleSheetPair_attribute (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBsspDlg_StyleSheetPair_attribute)) return (_currn);
return(NULLNODEPTR);
}/* MksspDlg_StyleSheetPair_attribute */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MksspDlg_StyleSheetPair_valueCategory (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MksspDlg_StyleSheetPair_valueCategory (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBsspDlg_StyleSheetPair_valueCategory)) return (_currn);
return(NULLNODEPTR);
}/* MksspDlg_StyleSheetPair_valueCategory */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MksspDlg_StyleSheetPair_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MksspDlg_StyleSheetPair_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBsspDlg_StyleSheetPair_value)) return (_currn);
return(NULLNODEPTR);
}/* MksspDlg_StyleSheetPair_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MksspDlg_CustomValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MksspDlg_CustomValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBsspDlg_CustomValue)) return (_currn);
return(NULLNODEPTR);
}/* MksspDlg_CustomValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MksspDlg_ListValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MksspDlg_ListValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBsspDlg_ListValue)) return (_currn);
return(NULLNODEPTR);
}/* MksspDlg_ListValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MksspDlg_AlignmentValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MksspDlg_AlignmentValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBsspDlg_AlignmentValue)) return (_currn);
return(NULLNODEPTR);
}/* MksspDlg_AlignmentValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MksspDlg_HeadingValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MksspDlg_HeadingValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBsspDlg_HeadingValue)) return (_currn);
return(NULLNODEPTR);
}/* MksspDlg_HeadingValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MksspDlg_StyleValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MksspDlg_StyleValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBsspDlg_StyleValue)) return (_currn);
return(NULLNODEPTR);
}/* MksspDlg_StyleValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MksspDlg_FamilyValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MksspDlg_FamilyValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBsspDlg_FamilyValue)) return (_currn);
return(NULLNODEPTR);
}/* MksspDlg_FamilyValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MksspDlg_ColorValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MksspDlg_ColorValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBsspDlg_ColorValue)) return (_currn);
return(NULLNODEPTR);
}/* MksspDlg_ColorValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MksspDlg_SizeValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MksspDlg_SizeValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBsspDlg_SizeValue)) return (_currn);
return(NULLNODEPTR);
}/* MksspDlg_SizeValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MksspDlg_SizeValue_measure (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MksspDlg_SizeValue_measure (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBsspDlg_SizeValue_measure)) return (_currn);
return(NULLNODEPTR);
}/* MksspDlg_SizeValue_measure */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MksspDlg_SizeValue_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MksspDlg_SizeValue_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBsspDlg_SizeValue_value)) return (_currn);
return(NULLNODEPTR);
}/* MksspDlg_SizeValue_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MksspDlg_ColorValue_color (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MksspDlg_ColorValue_color (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBsspDlg_ColorValue_color)) return (_currn);
return(NULLNODEPTR);
}/* MksspDlg_ColorValue_color */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MksspDlg_FamilyValue_family (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MksspDlg_FamilyValue_family (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBsspDlg_FamilyValue_family)) return (_currn);
return(NULLNODEPTR);
}/* MksspDlg_FamilyValue_family */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MksspDlg_StyleValue_style (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MksspDlg_StyleValue_style (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBsspDlg_StyleValue_style)) return (_currn);
return(NULLNODEPTR);
}/* MksspDlg_StyleValue_style */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MksspDlg_HeadingValue_heading (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MksspDlg_HeadingValue_heading (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBsspDlg_HeadingValue_heading)) return (_currn);
return(NULLNODEPTR);
}/* MksspDlg_HeadingValue_heading */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MksspDlg_AlignmentValue_alignment (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MksspDlg_AlignmentValue_alignment (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBsspDlg_AlignmentValue_alignment)) return (_currn);
return(NULLNODEPTR);
}/* MksspDlg_AlignmentValue_alignment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MksspDlg_ListValue_listed (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MksspDlg_ListValue_listed (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBsspDlg_ListValue_listed)) return (_currn);
return(NULLNODEPTR);
}/* MksspDlg_ListValue_listed */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MksspDlg_CustomValue_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MksspDlg_CustomValue_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBsspDlg_CustomValue_value)) return (_currn);
return(NULLNODEPTR);
}/* MksspDlg_CustomValue_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_StyleSheet (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_StyleSheet (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_StyleSheet)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_StyleSheet */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_StyleSheet_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_StyleSheet_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_StyleSheet_name)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_StyleSheet_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_StyleSheet_custom (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_StyleSheet_custom (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_StyleSheet_custom)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_StyleSheet_custom */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_StyleSheet_default (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_StyleSheet_default (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_StyleSheet_default)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_StyleSheet_default */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_CustomStyleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_CustomStyleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_CustomStyleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_CustomStyleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_HtmlStyleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_HtmlStyleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_HtmlStyleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_HtmlStyleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_CustomStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_CustomStyleSheetBlock_pairs (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_CustomStyleSheetBlock_pairs)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_CustomStyleSheetBlock_pairs */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_CustomStyleSheetBlock_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_CustomStyleSheetBlock_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_CustomStyleSheetBlock_name)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_CustomStyleSheetBlock_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_StyleSheetPair (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_StyleSheetPair (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_StyleSheetPair)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_StyleSheetPair */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_StyleSheetPair_category (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_StyleSheetPair_category (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_StyleSheetPair_category)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_StyleSheetPair_category */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_StyleSheetPair_attribute (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_StyleSheetPair_attribute (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_StyleSheetPair_attribute)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_StyleSheetPair_attribute */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_StyleSheetPair_valueCategory (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_StyleSheetPair_valueCategory (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_StyleSheetPair_valueCategory)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_StyleSheetPair_valueCategory */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_StyleSheetPair_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_StyleSheetPair_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_StyleSheetPair_value)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_StyleSheetPair_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_CustomValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_CustomValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_CustomValue)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_CustomValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_ListValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_ListValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_ListValue)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_ListValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_AlignmentValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_AlignmentValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_AlignmentValue)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_AlignmentValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_HeadingValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_HeadingValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_HeadingValue)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_HeadingValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_StyleValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_StyleValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_StyleValue)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_StyleValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_FamilyValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_FamilyValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_FamilyValue)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_FamilyValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_ColorValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_ColorValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_ColorValue)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_ColorValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_SizeValue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_SizeValue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_SizeValue)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_SizeValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_SizeValue_measure (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_SizeValue_measure (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_SizeValue_measure)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_SizeValue_measure */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_SizeValue_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_SizeValue_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_SizeValue_value)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_SizeValue_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_ColorValue_color (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_ColorValue_color (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_ColorValue_color)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_ColorValue_color */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_FamilyValue_family (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_FamilyValue_family (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_FamilyValue_family)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_FamilyValue_family */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_StyleValue_style (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_StyleValue_style (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_StyleValue_style)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_StyleValue_style */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_HeadingValue_heading (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_HeadingValue_heading (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_HeadingValue_heading)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_HeadingValue_heading */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_AlignmentValue_alignment (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_AlignmentValue_alignment (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_AlignmentValue_alignment)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_AlignmentValue_alignment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_ListValue_listed (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_ListValue_listed (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_ListValue_listed)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_ListValue_listed */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_CustomValue_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_CustomValue_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_CustomValue_value)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_CustomValue_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_HtmlStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_HtmlStyleSheetBlock_pairs (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_HtmlStyleSheetBlock_pairs)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_HtmlStyleSheetBlock_pairs */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkstylesheetView_HtmlStyleSheetBlock_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkstylesheetView_HtmlStyleSheetBlock_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBstylesheetView_HtmlStyleSheetBlock_name)) return (_currn);
return(NULLNODEPTR);
}/* MkstylesheetView_HtmlStyleSheetBlock_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_WebAction (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_WebAction (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_WebAction)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_WebAction */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_WebNodeFormalParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_WebNodeFormalParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_WebNodeFormalParam)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_WebNodeFormalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_WebNodeFormalParam_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_WebNodeFormalParam_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_WebNodeFormalParam_name)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_WebNodeFormalParam_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_WebNodeFormalParam_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_WebNodeFormalParam_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_WebNodeFormalParam_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_WebNodeFormalParam_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_WebNodeFormalParam_type (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_WebNodeFormalParam_type)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_WebNodeFormalParam_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_DBBlob (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_DBBlob (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_DBBlob)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_DBBlob */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_DBBoolean (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_DBBoolean (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_DBBoolean)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_DBBoolean */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_DBTimeStamp (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_DBTimeStamp (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_DBTimeStamp)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_DBTimeStamp */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_DBDate (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_DBDate (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_DBDate)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_DBDate */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_DBBigInt (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_DBBigInt (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_DBBigInt)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_DBBigInt */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_DBDouble (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_DBDouble (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_DBDouble)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_DBDouble */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_DBFloat (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_DBFloat (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_DBFloat)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_DBFloat */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_DBInteger (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_DBInteger (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_DBInteger)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_DBInteger */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_DBString (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_DBString (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_DBString)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_DBString */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_WebNodeActualParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_WebNodeActualParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_WebNodeActualParam)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_WebNodeActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_WebNodeActualParam_formalParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_WebNodeActualParam_formalParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_WebNodeActualParam_formalParam)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_WebNodeActualParam_formalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_WebNodeActualParam_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_WebNodeActualParam_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_WebNodeActualParam_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_WebNodeActualParam_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_WebNodeActualParam_expressions (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_WebNodeActualParam_expressions)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_WebNodeActualParam_expressions */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_WebNodeColumnExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_WebNodeColumnExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_WebNodeColumnExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_WebNodeColumnExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_WebNodeVariableExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_WebNodeVariableExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_WebNodeVariableExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_WebNodeVariableExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_WebNodeConstantExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_WebNodeConstantExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_WebNodeConstantExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_WebNodeConstantExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_WebNodeConstantExpression_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_WebNodeConstantExpression_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_WebNodeConstantExpression_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_WebNodeConstantExpression_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_WebNodeConstantExpression_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_WebNodeConstantExpression_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_WebNodeConstantExpression_value)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_WebNodeConstantExpression_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_WebNodeConstantExpression_keyword2 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_WebNodeConstantExpression_keyword2 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_WebNodeConstantExpression_keyword2)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_WebNodeConstantExpression_keyword2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_WebNodeVariableExpression_content (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_WebNodeVariableExpression_content (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_WebNodeVariableExpression_content)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_WebNodeVariableExpression_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_WebNodeColumnExpression_content (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_WebNodeColumnExpression_content (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_WebNodeColumnExpression_content)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_WebNodeColumnExpression_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_DBString_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_DBString_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_DBString_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_DBString_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_DBInteger_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_DBInteger_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_DBInteger_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_DBInteger_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_DBFloat_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_DBFloat_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_DBFloat_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_DBFloat_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_DBDouble_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_DBDouble_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_DBDouble_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_DBDouble_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_DBBigInt_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_DBBigInt_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_DBBigInt_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_DBBigInt_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_DBDate_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_DBDate_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_DBDate_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_DBDate_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_DBTimeStamp_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_DBTimeStamp_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_DBTimeStamp_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_DBTimeStamp_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_DBBoolean_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_DBBoolean_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_DBBoolean_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_DBBoolean_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_DBBlob_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_DBBlob_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_DBBlob_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_DBBlob_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_DataSourceActualParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_DataSourceActualParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_DataSourceActualParam)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_DataSourceActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_DataSourceActualParam_formalParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_DataSourceActualParam_formalParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_DataSourceActualParam_formalParam)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_DataSourceActualParam_formalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_DataSourceActualParam_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_DataSourceActualParam_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_DataSourceActualParam_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_DataSourceActualParam_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_DataSourceActualParam_expressions (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_DataSourceActualParam_expressions)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_DataSourceActualParam_expressions */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_ActionHandler (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_ActionHandler (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_ActionHandler)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_ActionHandler */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_ActionHandler_targetpage (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_ActionHandler_targetpage (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_ActionHandler_targetpage)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_ActionHandler_targetpage */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_ActionHandler_params (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_ActionHandler_params (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_ActionHandler_params)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_ActionHandler_params */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_WebAction_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_WebAction_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_WebAction_name)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_WebAction_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_WebAction_formalParameter (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_WebAction_formalParameter (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_WebAction_formalParameter)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_WebAction_formalParameter */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_WebAction_successpage (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_WebAction_successpage (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_WebAction_successpage)) return (_currn);
if (IsSymb (_currn, SYMBwebActionView_ActionHandler)) return (Mkrule_144(_coordref, _currn));
return(NULLNODEPTR);
}/* MkwebActionView_WebAction_successpage */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_WebAction_operations (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_WebAction_operations (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_WebAction_operations)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_WebAction_operations */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_WebAction_position (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_WebAction_position (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_WebAction_position)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_WebAction_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_DataBaseOperation (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_DataBaseOperation (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_DataBaseOperation)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_DataBaseOperation */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_PHPOperation (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_PHPOperation (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_PHPOperation)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_PHPOperation */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_PHPOperation_errorhandler (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_PHPOperation_errorhandler (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_PHPOperation_errorhandler)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_PHPOperation_errorhandler */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_PHPOperation_content (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_PHPOperation_content (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_PHPOperation_content)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_PHPOperation_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_DataBaseOperation_errorhandler (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_DataBaseOperation_errorhandler (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_DataBaseOperation_errorhandler)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_DataBaseOperation_errorhandler */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_DataBaseOperation_filter (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_DataBaseOperation_filter (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_DataBaseOperation_filter)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_DataBaseOperation_filter */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebActionView_DataBaseOperation_params (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebActionView_DataBaseOperation_params (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebActionView_DataBaseOperation_params)) return (_currn);
return(NULLNODEPTR);
}/* MkwebActionView_DataBaseOperation_params */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_WebPage (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_WebPage (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_WebPage)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_WebPage */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularCheckboxElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularCheckboxElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularCheckboxElement)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularCheckboxElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularCheckboxElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularCheckboxElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularCheckboxElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularCheckboxElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularCheckboxElement_type (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularCheckboxElement_type (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularCheckboxElement_type)) return (_currn);
if (IsSymb (_currn, SYMBwebPageView_FormularCheckboxElement_name)) return (MkpwebPageView_FormularCheckboxElement_type(_coordref, _currn));
return(NULLNODEPTR);
}/* MkwebPageView_FormularCheckboxElement_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularCheckboxElement_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularCheckboxElement_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularCheckboxElement_condition)) return (_currn);
if (IsSymb (_currn, SYMBwebPageView_FormularCondition)) return (Mkrule_145(_coordref, _currn));
return(NULLNODEPTR);
}/* MkwebPageView_FormularCheckboxElement_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularCheckboxElement_description (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularCheckboxElement_description (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularCheckboxElement_description)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularCheckboxElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularCheckboxElement_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularCheckboxElement_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularCheckboxElement_name)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularCheckboxElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularCondition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularCondition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularCondition)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularCondition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularTextElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularTextElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularTextElement)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularTextElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularTextElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularTextElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularTextElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularTextElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularTextElement_type (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularTextElement_type (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularTextElement_type)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularTextElement_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularTextElement_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularTextElement_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularTextElement_condition)) return (_currn);
if (IsSymb (_currn, SYMBwebPageView_FormularCondition)) return (Mkrule_146(_coordref, _currn));
return(NULLNODEPTR);
}/* MkwebPageView_FormularTextElement_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularTextElement_description (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularTextElement_description (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularTextElement_description)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularTextElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularPasswordElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularPasswordElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularPasswordElement)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularPasswordElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularPasswordElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularPasswordElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularPasswordElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularPasswordElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularPasswordElement_type (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularPasswordElement_type (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularPasswordElement_type)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularPasswordElement_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularPasswordElement_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularPasswordElement_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularPasswordElement_condition)) return (_currn);
if (IsSymb (_currn, SYMBwebPageView_FormularCondition)) return (Mkrule_147(_coordref, _currn));
return(NULLNODEPTR);
}/* MkwebPageView_FormularPasswordElement_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularPasswordElement_description (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularPasswordElement_description (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularPasswordElement_description)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularPasswordElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularTextAreaElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularTextAreaElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularTextAreaElement)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularTextAreaElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularTextAreaElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularTextAreaElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularTextAreaElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularTextAreaElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularTextAreaElement_type (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularTextAreaElement_type (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularTextAreaElement_type)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularTextAreaElement_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularTextAreaElement_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularTextAreaElement_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularTextAreaElement_condition)) return (_currn);
if (IsSymb (_currn, SYMBwebPageView_FormularCondition)) return (Mkrule_148(_coordref, _currn));
return(NULLNODEPTR);
}/* MkwebPageView_FormularTextAreaElement_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularTextAreaElement_description (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularTextAreaElement_description (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularTextAreaElement_description)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularTextAreaElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularSelectElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularSelectElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularSelectElement)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularSelectElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularSelectElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularSelectElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularSelectElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularSelectElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularSelectElement_multiple (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularSelectElement_multiple (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularSelectElement_multiple)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularSelectElement_multiple */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularSelectElement_options (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularSelectElement_options (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularSelectElement_options)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularSelectElement_options */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularSelectElement_description (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularSelectElement_description (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularSelectElement_description)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularSelectElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularPHPScriptSelectOption (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularPHPScriptSelectOption (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularPHPScriptSelectOption)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularPHPScriptSelectOption */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularDataSourceSelectOption (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularDataSourceSelectOption (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularDataSourceSelectOption)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularDataSourceSelectOption */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularConstantSelectOption (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularConstantSelectOption (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularConstantSelectOption)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularConstantSelectOption */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularRadioElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularRadioElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularRadioElement)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularRadioElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularRadioElement_type (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularRadioElement_type (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularRadioElement_type)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularRadioElement_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularRadioElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularRadioElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularRadioElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularRadioElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularRadioElement_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularRadioElement_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularRadioElement_name)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularRadioElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_DataSourceActualParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_DataSourceActualParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_DataSourceActualParam)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_DataSourceActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_DataSourceActualParam_formalParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_DataSourceActualParam_formalParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_DataSourceActualParam_formalParam)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_DataSourceActualParam_formalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_DataSourceActualParam_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_DataSourceActualParam_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_DataSourceActualParam_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_DataSourceActualParam_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_DataSourceActualParam_expressions (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_DataSourceActualParam_expressions)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_DataSourceActualParam_expressions */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_WebNodeColumnExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_WebNodeColumnExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_WebNodeColumnExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_WebNodeColumnExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_WebNodeVariableExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_WebNodeVariableExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_WebNodeVariableExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_WebNodeVariableExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_WebNodeConstantExpression (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_WebNodeConstantExpression (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_WebNodeConstantExpression)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_WebNodeConstantExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_WebNodeActualParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_WebNodeActualParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_WebNodeActualParam)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_WebNodeActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_WebNodeActualParam_formalParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_WebNodeActualParam_formalParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_WebNodeActualParam_formalParam)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_WebNodeActualParam_formalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_WebNodeActualParam_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_WebNodeActualParam_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_WebNodeActualParam_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_WebNodeActualParam_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_WebNodeActualParam_expressions (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_WebNodeActualParam_expressions)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_WebNodeActualParam_expressions */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_WebNodeConstantExpression_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_WebNodeConstantExpression_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_WebNodeConstantExpression_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_WebNodeConstantExpression_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_WebNodeConstantExpression_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_WebNodeConstantExpression_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_WebNodeConstantExpression_value)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_WebNodeConstantExpression_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_WebNodeConstantExpression_keyword2 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_WebNodeConstantExpression_keyword2 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_WebNodeConstantExpression_keyword2)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_WebNodeConstantExpression_keyword2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_WebNodeVariableExpression_content (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_WebNodeVariableExpression_content (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_WebNodeVariableExpression_content)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_WebNodeVariableExpression_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_WebNodeColumnExpression_content (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_WebNodeColumnExpression_content (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_WebNodeColumnExpression_content)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_WebNodeColumnExpression_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_WebNodeFormalParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_WebNodeFormalParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_WebNodeFormalParam)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_WebNodeFormalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_WebNodeFormalParam_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_WebNodeFormalParam_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_WebNodeFormalParam_name)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_WebNodeFormalParam_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_WebNodeFormalParam_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_WebNodeFormalParam_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_WebNodeFormalParam_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_WebNodeFormalParam_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_WebNodeFormalParam_type (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_WebNodeFormalParam_type)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_WebNodeFormalParam_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_DBBlob (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_DBBlob (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_DBBlob)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_DBBlob */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_DBBoolean (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_DBBoolean (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_DBBoolean)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_DBBoolean */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_DBTimeStamp (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_DBTimeStamp (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_DBTimeStamp)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_DBTimeStamp */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_DBDate (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_DBDate (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_DBDate)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_DBDate */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_DBBigInt (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_DBBigInt (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_DBBigInt)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_DBBigInt */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_DBDouble (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_DBDouble (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_DBDouble)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_DBDouble */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_DBFloat (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_DBFloat (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_DBFloat)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_DBFloat */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_DBInteger (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_DBInteger (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_DBInteger)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_DBInteger */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_DBString (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_DBString (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_DBString)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_DBString */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_DBString_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_DBString_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_DBString_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_DBString_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_DBInteger_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_DBInteger_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_DBInteger_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_DBInteger_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_DBFloat_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_DBFloat_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_DBFloat_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_DBFloat_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_DBDouble_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_DBDouble_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_DBDouble_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_DBDouble_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_DBBigInt_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_DBBigInt_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_DBBigInt_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_DBBigInt_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_DBDate_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_DBDate_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_DBDate_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_DBDate_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_DBTimeStamp_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_DBTimeStamp_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_DBTimeStamp_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_DBTimeStamp_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_DBBoolean_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_DBBoolean_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_DBBoolean_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_DBBoolean_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_DBBlob_keyword1 (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_DBBlob_keyword1 (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_DBBlob_keyword1)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_DBBlob_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_Formular (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_Formular (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_Formular)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_Formular */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_Formular_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_Formular_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_Formular_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_Formular_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_Formular_action (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_Formular_action (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_Formular_action)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_Formular_action */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_Formular_elements (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_Formular_elements (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_Formular_elements)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_Formular_elements */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_Formular_datasource (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_Formular_datasource (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_Formular_datasource)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_Formular_datasource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_Formular_actualParams (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_Formular_actualParams (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_Formular_actualParams)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_Formular_actualParams */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularRadioGroupElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularRadioGroupElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularRadioGroupElement)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularRadioGroupElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularRadioGroupElement_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularRadioGroupElement_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularRadioGroupElement_name)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularRadioGroupElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularRadioGroupElement_label (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularRadioGroupElement_label (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularRadioGroupElement_label)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularRadioGroupElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularRadioGroupElement_description (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularRadioGroupElement_description (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularRadioGroupElement_description)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularRadioGroupElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularRadioGroupElement_pvalue (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularRadioGroupElement_pvalue (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularRadioGroupElement_pvalue)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularRadioGroupElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularRadioGroupElement_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularRadioGroupElement_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularRadioGroupElement_condition)) return (_currn);
if (IsSymb (_currn, SYMBwebPageView_FormularCondition)) return (Mkrule_149(_coordref, _currn));
return(NULLNODEPTR);
}/* MkwebPageView_FormularRadioGroupElement_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularRadioGroupElement_elements (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularRadioGroupElement_elements (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularRadioGroupElement_elements)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularRadioGroupElement_elements */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularRadioGroupElement_checked (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularRadioGroupElement_checked (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularRadioGroupElement_checked)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularRadioGroupElement_checked */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularConstantSelectOption_values (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularConstantSelectOption_values (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularConstantSelectOption_values)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularConstantSelectOption_values */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularDataSourceSelectOption_values (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularDataSourceSelectOption_values (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularDataSourceSelectOption_values)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularDataSourceSelectOption_values */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularDataSourceSelectOption_params (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularDataSourceSelectOption_params (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularDataSourceSelectOption_params)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularDataSourceSelectOption_params */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularPHPScriptSelectOption_script (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularPHPScriptSelectOption_script (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularPHPScriptSelectOption_script)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularPHPScriptSelectOption_script */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularCondition_condition (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularCondition_condition (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularCondition_condition)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularCondition_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularCondition_params (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularCondition_params (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularCondition_params)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularCondition_params */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularConditionActualParam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularConditionActualParam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularConditionActualParam)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularConditionActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularConditionActualParam_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularConditionActualParam_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularConditionActualParam_value)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularConditionActualParam_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FormularConditionActualParam_formalparam (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FormularConditionActualParam_formalparam (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FormularConditionActualParam_formalparam)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FormularConditionActualParam_formalparam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_Paragraph (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_Paragraph (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_Paragraph)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_Paragraph */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_Paragraph_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_Paragraph_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_Paragraph_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_Paragraph_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_Paragraph_fragments (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_Paragraph_fragments (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_Paragraph_fragments)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_Paragraph_fragments */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_OutputElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_OutputElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_OutputElement)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_OutputElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_ImageFragment (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_ImageFragment (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_ImageFragment)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_ImageFragment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_TextFragment (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_TextFragment (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_TextFragment)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_TextFragment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_TextFragment_link (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_TextFragment_link (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_TextFragment_link)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_TextFragment_link */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_TextFragment_value (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_TextFragment_value (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_TextFragment_value)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_TextFragment_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_ImageFragment_link (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_ImageFragment_link (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_ImageFragment_link)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_ImageFragment_link */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_ImageFragment_width (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_ImageFragment_width (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_ImageFragment_width)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_ImageFragment_width */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_ImageFragment_height (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_ImageFragment_height (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_ImageFragment_height)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_ImageFragment_height */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_ImageFragment_text (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_ImageFragment_text (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_ImageFragment_text)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_ImageFragment_text */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_ImageFragment_imageName (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_ImageFragment_imageName (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_ImageFragment_imageName)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_ImageFragment_imageName */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_Heading (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_Heading (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_Heading)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_Heading */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_Heading_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_Heading_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_Heading_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_Heading_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_Heading_text (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_Heading_text (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_Heading_text)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_Heading_text */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_Heading_level (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_Heading_level (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_Heading_level)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_Heading_level */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_Table (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_Table (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_Table)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_Table */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_Table_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_Table_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_Table_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_Table_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_Table_rows (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_Table_rows (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_Table_rows)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_Table_rows */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_Table_columns (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_Table_columns (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_Table_columns)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_Table_columns */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_TableRow (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_TableRow (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_TableRow)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_TableRow */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_TableColumn (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_TableColumn (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_TableColumn)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_TableColumn */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_TableRow_cells (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_TableRow_cells (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_TableRow_cells)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_TableRow_cells */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_TableCell (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_TableCell (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_TableCell)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_TableCell */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_TableCell_columnRef (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_TableCell_columnRef (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_TableCell_columnRef)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_TableCell_columnRef */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_TableCell_content (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_TableCell_content (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_TableCell_content)) return (_currn);
if (IsSymb (_currn, SYMBwebPageView_Frame)) return (Mkrule_150(_coordref, _currn));
return(NULLNODEPTR);
}/* MkwebPageView_TableCell_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_Frame (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_Frame (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_Frame)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_Frame */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_WebPage_name (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_WebPage_name (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_WebPage_name)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_WebPage_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_WebPage_formalParameter (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_WebPage_formalParameter (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_WebPage_formalParameter)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_WebPage_formalParameter */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_WebPage_title (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_WebPage_title (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_WebPage_title)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_WebPage_title */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_WebPage_frame (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_WebPage_frame (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_WebPage_frame)) return (_currn);
if (IsSymb (_currn, SYMBwebPageView_Frame)) return (Mkrule_151(_coordref, _currn));
return(NULLNODEPTR);
}/* MkwebPageView_WebPage_frame */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_WebPage_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_WebPage_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_WebPage_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_WebPage_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_WebPage_position (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_WebPage_position (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_WebPage_position)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_WebPage_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_Frame_components (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_Frame_components (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_Frame_components)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_Frame_components */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_Frame_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_Frame_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_Frame_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_Frame_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_ListOutputElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_ListOutputElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_ListOutputElement)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_ListOutputElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FrameRef (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FrameRef (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FrameRef)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FrameRef */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FrameRef_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FrameRef_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FrameRef_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FrameRef_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_FrameRef_frame (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_FrameRef_frame (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_FrameRef_frame)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_FrameRef_frame */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_OutputElement_link (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_OutputElement_link (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_OutputElement_link)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_OutputElement_link */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_OutputElement_content (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_OutputElement_content (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_OutputElement_content)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_OutputElement_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_ListOutputElement_styleSheetBlock (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_ListOutputElement_styleSheetBlock (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_ListOutputElement_styleSheetBlock)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_ListOutputElement_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_ListOutputElement_dataSourceActualParams (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_ListOutputElement_dataSourceActualParams (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_ListOutputElement_dataSourceActualParams)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_ListOutputElement_dataSourceActualParams */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_ListOutputElement_dataSource (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_ListOutputElement_dataSource (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_ListOutputElement_dataSource)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_ListOutputElement_dataSource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_ListOutputElement_maxNumberOfElements (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_ListOutputElement_maxNumberOfElements (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_ListOutputElement_maxNumberOfElements)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_ListOutputElement_maxNumberOfElements */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkwebPageView_ListOutputElement_navigationElement (POSITION *_coordref, NODEPTR _currn)
#else
NODEPTR MkwebPageView_ListOutputElement_navigationElement (_coordref, _currn)
	POSITION *_coordref; NODEPTR _currn;
#endif
{
if (_currn == NULLNODEPTR) return NULLNODEPTR;
if (IsSymb (_currn, SYMBwebPageView_ListOutputElement_navigationElement)) return (_currn);
return(NULLNODEPTR);
}/* MkwebPageView_ListOutputElement_navigationElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpEmpty (POSITION *_coordref)
#else
NODEPTR MkpEmpty (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpEmpty _currn;
#ifdef __cplusplus
_currn = new _TPpEmpty;
#else
_currn = (_TPPpEmpty) TreeNodeAlloc (sizeof (struct _TPpEmpty));
#endif
_currn->_prod = RULEpEmpty;
_SETCOORD(_currn)
_TERMACT_pEmpty;
return ( (NODEPTR) _currn);
}/* MkpEmpty */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpVIEWROOT_codeGen (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR MkpVIEWROOT_codeGen (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPpVIEWROOT_codeGen _currn;
#ifdef __cplusplus
_currn = new _TPpVIEWROOT_codeGen;
#else
_currn = (_TPPpVIEWROOT_codeGen) TreeNodeAlloc (sizeof (struct _TPpVIEWROOT_codeGen));
#endif
_currn->_prod = RULEpVIEWROOT_codeGen;
_currn->_desc1 = (_TSPcodeGen_Root) MkcodeGen_Root (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pVIEWROOT_codeGen: root of subtree no. 1 can not be made a codeGen_Root node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_pVIEWROOT_codeGen;
return ( (NODEPTR) _currn);
}/* MkpVIEWROOT_codeGen */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceDB (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpcodeGen_DataSourceDB (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpcodeGen_DataSourceDB _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceDB;
#else
_currn = (_TPPpcodeGen_DataSourceDB) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceDB));
#endif
_currn->_prod = RULEpcodeGen_DataSourceDB;
_currn->_desc1 = (_TSPcodeGen_DataSourceDB_name) MkcodeGen_DataSourceDB_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceDB: root of subtree no. 1 can not be made a codeGen_DataSourceDB_name node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_DataSourceDB_position) MkcodeGen_DataSourceDB_position (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceDB: root of subtree no. 2 can not be made a codeGen_DataSourceDB_position node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_DataSourceDB_columns) MkcodeGen_DataSourceDB_columns (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceDB: root of subtree no. 3 can not be made a codeGen_DataSourceDB_columns node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceDB;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceDB */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceDB_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataSourceDB_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataSourceDB_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceDB_name;
#else
_currn = (_TPPpcodeGen_DataSourceDB_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceDB_name));
#endif
_currn->_prod = RULEpcodeGen_DataSourceDB_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceDB_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceDB_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceDB_position (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataSourceDB_position (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataSourceDB_position _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceDB_position;
#else
_currn = (_TPPpcodeGen_DataSourceDB_position) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceDB_position));
#endif
_currn->_prod = RULEpcodeGen_DataSourceDB_position;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceDB_position;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceDB_position */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_DBColumnpcodeGen_DataSourceDB_columns (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_DBColumnpcodeGen_DataSourceDB_columns (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_DBColumnpcodeGen_DataSourceDB_columns _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_DBColumnpcodeGen_DataSourceDB_columns;
#else
	_new = (_TPPLST_codeGen_DBColumnpcodeGen_DataSourceDB_columns) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_DBColumnpcodeGen_DataSourceDB_columns));
#endif
	_new->_prod = RULELST_codeGen_DBColumnpcodeGen_DataSourceDB_columns;
	_new->_desc1 = (_TSPcodeGen_DBColumn)elem;
	_new->_desc2 = (_TSPLST_codeGen_DataSourceDB_columns)_new;
	{_TPPLST_codeGen_DBColumnpcodeGen_DataSourceDB_columns _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_DBColumnpcodeGen_DataSourceDB_columns;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_DataSourceDB_columns (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_DataSourceDB_columns (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_DBColumn)) 
		return _MkcodeGen_DBColumnpcodeGen_DataSourceDB_columns (_coordref, MkcodeGen_DBColumn (_coordref, t));
	message (DEADLY, "a pcodeGen_DataSourceDB_columns list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_DataSourceDB_columns (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_DataSourceDB_columns (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_DataSourceDB_columns)
	l =_ElempcodeGen_DataSourceDB_columns (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_DataSourceDB_columns)
	r = _ElempcodeGen_DataSourceDB_columns (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_DataSourceDB_columns)r)->_desc2);
((_TPPLST_0pcodeGen_DataSourceDB_columns)r)->_desc2 = ((_TPPLST_0pcodeGen_DataSourceDB_columns)l)->_desc2;
((_TPPLST_0pcodeGen_DataSourceDB_columns)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceDB_columns (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_DataSourceDB_columns (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_DataSourceDB_columns root;
	_TPPLST_0pcodeGen_DataSourceDB_columns list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_DataSourceDB_columns)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_DataSourceDB_columns)
		t = _ElempcodeGen_DataSourceDB_columns (_coordref, t);
	remain = (_TPPLST_0pcodeGen_DataSourceDB_columns)(((_TPPLST_0pcodeGen_DataSourceDB_columns) t)->_desc2);
	((_TPPLST_0pcodeGen_DataSourceDB_columns) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_DataSourceDB_columns;
#else
list = (_TPPLST_0pcodeGen_DataSourceDB_columns) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_DataSourceDB_columns));
#endif
list->_prod = RULELST_0pcodeGen_DataSourceDB_columns;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_DataSourceDB_columns)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_DBColumnpcodeGen_DataSourceDB_columns)
	{_TPPLST_codeGen_DBColumnpcodeGen_DataSourceDB_columns _currn = _currn;
	 _TERMACT_LST_codeGen_DBColumnpcodeGen_DataSourceDB_columns;
	} this code has been moved to _ElempcodeGen_DataSourceDB_columns*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_DataSourceDB_columns;
#else
root = (_TPPpcodeGen_DataSourceDB_columns) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceDB_columns));
#endif
root->_prod = RULEpcodeGen_DataSourceDB_columns;
root->_desc1 = (_TSPLST_codeGen_DataSourceDB_columns)list;
{ _TPPpcodeGen_DataSourceDB_columns _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_DataSourceDB_columns;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataBase (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5, NODEPTR _desc6, NODEPTR _desc7, NODEPTR _desc8, NODEPTR _desc9, NODEPTR _desc10)
#else
NODEPTR MkpcodeGen_DataBase (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5,_desc6,_desc7,_desc8,_desc9,_desc10)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
	NODEPTR _desc6;
	NODEPTR _desc7;
	NODEPTR _desc8;
	NODEPTR _desc9;
	NODEPTR _desc10;
#endif
{	_TPPpcodeGen_DataBase _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataBase;
#else
_currn = (_TPPpcodeGen_DataBase) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataBase));
#endif
_currn->_prod = RULEpcodeGen_DataBase;
_currn->_desc1 = (_TSPcodeGen_DataBase_name) MkcodeGen_DataBase_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataBase: root of subtree no. 1 can not be made a codeGen_DataBase_name node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_DataBase_userName) MkcodeGen_DataBase_userName (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataBase: root of subtree no. 2 can not be made a codeGen_DataBase_userName node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_DataBase_userPassword) MkcodeGen_DataBase_userPassword (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataBase: root of subtree no. 3 can not be made a codeGen_DataBase_userPassword node ", 0, _coordref);
_currn->_desc4 = (_TSPcodeGen_DataBase_hostName) MkcodeGen_DataBase_hostName (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataBase: root of subtree no. 4 can not be made a codeGen_DataBase_hostName node ", 0, _coordref);
_currn->_desc5 = (_TSPcodeGen_DataBase_port) MkcodeGen_DataBase_port (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataBase: root of subtree no. 5 can not be made a codeGen_DataBase_port node ", 0, _coordref);
_currn->_desc6 = (_TSPcodeGen_DataBase_dataBaseType) MkcodeGen_DataBase_dataBaseType (_coordref, _desc6);	
if (((NODEPTR)_currn->_desc6) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataBase: root of subtree no. 6 can not be made a codeGen_DataBase_dataBaseType node ", 0, _coordref);
_currn->_desc7 = (_TSPcodeGen_DataBase_dataSources) MkcodeGen_DataBase_dataSources (_coordref, _desc7);	
if (((NODEPTR)_currn->_desc7) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataBase: root of subtree no. 7 can not be made a codeGen_DataBase_dataSources node ", 0, _coordref);
_currn->_desc8 = (_TSPcodeGen_DataBase_dataSourceModifications) MkcodeGen_DataBase_dataSourceModifications (_coordref, _desc8);	
if (((NODEPTR)_currn->_desc8) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataBase: root of subtree no. 8 can not be made a codeGen_DataBase_dataSourceModifications node ", 0, _coordref);
_currn->_desc9 = (_TSPcodeGen_DataBase_setSize) MkcodeGen_DataBase_setSize (_coordref, _desc9);	
if (((NODEPTR)_currn->_desc9) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataBase: root of subtree no. 9 can not be made a codeGen_DataBase_setSize node ", 0, _coordref);
_currn->_desc10 = (_TSPcodeGen_DataBase_modifySetSize) MkcodeGen_DataBase_modifySetSize (_coordref, _desc10);	
if (((NODEPTR)_currn->_desc10) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataBase: root of subtree no. 10 can not be made a codeGen_DataBase_modifySetSize node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataBase;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataBase */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataBase_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataBase_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataBase_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataBase_name;
#else
_currn = (_TPPpcodeGen_DataBase_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataBase_name));
#endif
_currn->_prod = RULEpcodeGen_DataBase_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataBase_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataBase_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataBase_userName (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataBase_userName (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataBase_userName _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataBase_userName;
#else
_currn = (_TPPpcodeGen_DataBase_userName) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataBase_userName));
#endif
_currn->_prod = RULEpcodeGen_DataBase_userName;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataBase_userName;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataBase_userName */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataBase_userPassword (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataBase_userPassword (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataBase_userPassword _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataBase_userPassword;
#else
_currn = (_TPPpcodeGen_DataBase_userPassword) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataBase_userPassword));
#endif
_currn->_prod = RULEpcodeGen_DataBase_userPassword;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataBase_userPassword;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataBase_userPassword */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataBase_hostName (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataBase_hostName (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataBase_hostName _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataBase_hostName;
#else
_currn = (_TPPpcodeGen_DataBase_hostName) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataBase_hostName));
#endif
_currn->_prod = RULEpcodeGen_DataBase_hostName;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataBase_hostName;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataBase_hostName */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataBase_port (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataBase_port (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataBase_port _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataBase_port;
#else
_currn = (_TPPpcodeGen_DataBase_port) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataBase_port));
#endif
_currn->_prod = RULEpcodeGen_DataBase_port;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataBase_port;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataBase_port */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataBase_dataBaseType (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataBase_dataBaseType (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataBase_dataBaseType _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataBase_dataBaseType;
#else
_currn = (_TPPpcodeGen_DataBase_dataBaseType) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataBase_dataBaseType));
#endif
_currn->_prod = RULEpcodeGen_DataBase_dataBaseType;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataBase_dataBaseType;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataBase_dataBaseType */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_DataSourceDBpcodeGen_DataBase_dataSources (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_DataSourceDBpcodeGen_DataBase_dataSources (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_DataSourceDBpcodeGen_DataBase_dataSources _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_DataSourceDBpcodeGen_DataBase_dataSources;
#else
	_new = (_TPPLST_codeGen_DataSourceDBpcodeGen_DataBase_dataSources) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_DataSourceDBpcodeGen_DataBase_dataSources));
#endif
	_new->_prod = RULELST_codeGen_DataSourceDBpcodeGen_DataBase_dataSources;
	_new->_desc1 = (_TSPcodeGen_DataSourceDB)elem;
	_new->_desc2 = (_TSPLST_codeGen_DataBase_dataSources)_new;
	{_TPPLST_codeGen_DataSourceDBpcodeGen_DataBase_dataSources _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_DataSourceDBpcodeGen_DataBase_dataSources;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_DataSourceProjectionFilterpcodeGen_DataBase_dataSources (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_DataSourceProjectionFilterpcodeGen_DataBase_dataSources (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_DataSourceProjectionFilterpcodeGen_DataBase_dataSources _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_DataSourceProjectionFilterpcodeGen_DataBase_dataSources;
#else
	_new = (_TPPLST_codeGen_DataSourceProjectionFilterpcodeGen_DataBase_dataSources) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_DataSourceProjectionFilterpcodeGen_DataBase_dataSources));
#endif
	_new->_prod = RULELST_codeGen_DataSourceProjectionFilterpcodeGen_DataBase_dataSources;
	_new->_desc1 = (_TSPcodeGen_DataSourceProjectionFilter)elem;
	_new->_desc2 = (_TSPLST_codeGen_DataBase_dataSources)_new;
	{_TPPLST_codeGen_DataSourceProjectionFilterpcodeGen_DataBase_dataSources _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_DataSourceProjectionFilterpcodeGen_DataBase_dataSources;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_LimitDataSourcepcodeGen_DataBase_dataSources (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_LimitDataSourcepcodeGen_DataBase_dataSources (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_LimitDataSourcepcodeGen_DataBase_dataSources _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_LimitDataSourcepcodeGen_DataBase_dataSources;
#else
	_new = (_TPPLST_codeGen_LimitDataSourcepcodeGen_DataBase_dataSources) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_LimitDataSourcepcodeGen_DataBase_dataSources));
#endif
	_new->_prod = RULELST_codeGen_LimitDataSourcepcodeGen_DataBase_dataSources;
	_new->_desc1 = (_TSPcodeGen_LimitDataSource)elem;
	_new->_desc2 = (_TSPLST_codeGen_DataBase_dataSources)_new;
	{_TPPLST_codeGen_LimitDataSourcepcodeGen_DataBase_dataSources _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_LimitDataSourcepcodeGen_DataBase_dataSources;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_DataSourceJoinpcodeGen_DataBase_dataSources (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_DataSourceJoinpcodeGen_DataBase_dataSources (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_DataSourceJoinpcodeGen_DataBase_dataSources _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_DataSourceJoinpcodeGen_DataBase_dataSources;
#else
	_new = (_TPPLST_codeGen_DataSourceJoinpcodeGen_DataBase_dataSources) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_DataSourceJoinpcodeGen_DataBase_dataSources));
#endif
	_new->_prod = RULELST_codeGen_DataSourceJoinpcodeGen_DataBase_dataSources;
	_new->_desc1 = (_TSPcodeGen_DataSourceJoin)elem;
	_new->_desc2 = (_TSPLST_codeGen_DataBase_dataSources)_new;
	{_TPPLST_codeGen_DataSourceJoinpcodeGen_DataBase_dataSources _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_DataSourceJoinpcodeGen_DataBase_dataSources;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_DataBase_dataSources (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_DataBase_dataSources (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_DataSourceDB)) 
		return _MkcodeGen_DataSourceDBpcodeGen_DataBase_dataSources (_coordref, MkcodeGen_DataSourceDB (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_DataSourceProjectionFilter)) 
		return _MkcodeGen_DataSourceProjectionFilterpcodeGen_DataBase_dataSources (_coordref, MkcodeGen_DataSourceProjectionFilter (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_LimitDataSource)) 
		return _MkcodeGen_LimitDataSourcepcodeGen_DataBase_dataSources (_coordref, MkcodeGen_LimitDataSource (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_DataSourceJoin)) 
		return _MkcodeGen_DataSourceJoinpcodeGen_DataBase_dataSources (_coordref, MkcodeGen_DataSourceJoin (_coordref, t));
	message (DEADLY, "a pcodeGen_DataBase_dataSources list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_DataBase_dataSources (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_DataBase_dataSources (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_DataBase_dataSources)
	l =_ElempcodeGen_DataBase_dataSources (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_DataBase_dataSources)
	r = _ElempcodeGen_DataBase_dataSources (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_DataBase_dataSources)r)->_desc2);
((_TPPLST_0pcodeGen_DataBase_dataSources)r)->_desc2 = ((_TPPLST_0pcodeGen_DataBase_dataSources)l)->_desc2;
((_TPPLST_0pcodeGen_DataBase_dataSources)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataBase_dataSources (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_DataBase_dataSources (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_DataBase_dataSources root;
	_TPPLST_0pcodeGen_DataBase_dataSources list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_DataBase_dataSources)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_DataBase_dataSources)
		t = _ElempcodeGen_DataBase_dataSources (_coordref, t);
	remain = (_TPPLST_0pcodeGen_DataBase_dataSources)(((_TPPLST_0pcodeGen_DataBase_dataSources) t)->_desc2);
	((_TPPLST_0pcodeGen_DataBase_dataSources) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_DataBase_dataSources;
#else
list = (_TPPLST_0pcodeGen_DataBase_dataSources) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_DataBase_dataSources));
#endif
list->_prod = RULELST_0pcodeGen_DataBase_dataSources;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_DataBase_dataSources)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_DataSourceDBpcodeGen_DataBase_dataSources)
	{_TPPLST_codeGen_DataSourceDBpcodeGen_DataBase_dataSources _currn = _currn;
	 _TERMACT_LST_codeGen_DataSourceDBpcodeGen_DataBase_dataSources;
	} this code has been moved to _ElempcodeGen_DataBase_dataSources*/
	/*if (_currn->_prod == RULELST_codeGen_DataSourceProjectionFilterpcodeGen_DataBase_dataSources)
	{_TPPLST_codeGen_DataSourceProjectionFilterpcodeGen_DataBase_dataSources _currn = _currn;
	 _TERMACT_LST_codeGen_DataSourceProjectionFilterpcodeGen_DataBase_dataSources;
	} this code has been moved to _ElempcodeGen_DataBase_dataSources*/
	/*if (_currn->_prod == RULELST_codeGen_LimitDataSourcepcodeGen_DataBase_dataSources)
	{_TPPLST_codeGen_LimitDataSourcepcodeGen_DataBase_dataSources _currn = _currn;
	 _TERMACT_LST_codeGen_LimitDataSourcepcodeGen_DataBase_dataSources;
	} this code has been moved to _ElempcodeGen_DataBase_dataSources*/
	/*if (_currn->_prod == RULELST_codeGen_DataSourceJoinpcodeGen_DataBase_dataSources)
	{_TPPLST_codeGen_DataSourceJoinpcodeGen_DataBase_dataSources _currn = _currn;
	 _TERMACT_LST_codeGen_DataSourceJoinpcodeGen_DataBase_dataSources;
	} this code has been moved to _ElempcodeGen_DataBase_dataSources*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_DataBase_dataSources;
#else
root = (_TPPpcodeGen_DataBase_dataSources) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataBase_dataSources));
#endif
root->_prod = RULEpcodeGen_DataBase_dataSources;
root->_desc1 = (_TSPLST_codeGen_DataBase_dataSources)list;
{ _TPPpcodeGen_DataBase_dataSources _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_DataBase_dataSources;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_DataSourceInsertpcodeGen_DataBase_dataSourceModifications (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_DataSourceInsertpcodeGen_DataBase_dataSourceModifications (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_DataSourceInsertpcodeGen_DataBase_dataSourceModifications _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_DataSourceInsertpcodeGen_DataBase_dataSourceModifications;
#else
	_new = (_TPPLST_codeGen_DataSourceInsertpcodeGen_DataBase_dataSourceModifications) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_DataSourceInsertpcodeGen_DataBase_dataSourceModifications));
#endif
	_new->_prod = RULELST_codeGen_DataSourceInsertpcodeGen_DataBase_dataSourceModifications;
	_new->_desc1 = (_TSPcodeGen_DataSourceInsert)elem;
	_new->_desc2 = (_TSPLST_codeGen_DataBase_dataSourceModifications)_new;
	{_TPPLST_codeGen_DataSourceInsertpcodeGen_DataBase_dataSourceModifications _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_DataSourceInsertpcodeGen_DataBase_dataSourceModifications;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_DataSourceUpdatepcodeGen_DataBase_dataSourceModifications (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_DataSourceUpdatepcodeGen_DataBase_dataSourceModifications (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_DataSourceUpdatepcodeGen_DataBase_dataSourceModifications _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_DataSourceUpdatepcodeGen_DataBase_dataSourceModifications;
#else
	_new = (_TPPLST_codeGen_DataSourceUpdatepcodeGen_DataBase_dataSourceModifications) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_DataSourceUpdatepcodeGen_DataBase_dataSourceModifications));
#endif
	_new->_prod = RULELST_codeGen_DataSourceUpdatepcodeGen_DataBase_dataSourceModifications;
	_new->_desc1 = (_TSPcodeGen_DataSourceUpdate)elem;
	_new->_desc2 = (_TSPLST_codeGen_DataBase_dataSourceModifications)_new;
	{_TPPLST_codeGen_DataSourceUpdatepcodeGen_DataBase_dataSourceModifications _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_DataSourceUpdatepcodeGen_DataBase_dataSourceModifications;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_DataSourceDeletepcodeGen_DataBase_dataSourceModifications (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_DataSourceDeletepcodeGen_DataBase_dataSourceModifications (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_DataSourceDeletepcodeGen_DataBase_dataSourceModifications _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_DataSourceDeletepcodeGen_DataBase_dataSourceModifications;
#else
	_new = (_TPPLST_codeGen_DataSourceDeletepcodeGen_DataBase_dataSourceModifications) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_DataSourceDeletepcodeGen_DataBase_dataSourceModifications));
#endif
	_new->_prod = RULELST_codeGen_DataSourceDeletepcodeGen_DataBase_dataSourceModifications;
	_new->_desc1 = (_TSPcodeGen_DataSourceDelete)elem;
	_new->_desc2 = (_TSPLST_codeGen_DataBase_dataSourceModifications)_new;
	{_TPPLST_codeGen_DataSourceDeletepcodeGen_DataBase_dataSourceModifications _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_DataSourceDeletepcodeGen_DataBase_dataSourceModifications;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_DataBase_dataSourceModifications (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_DataBase_dataSourceModifications (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_DataSourceInsert)) 
		return _MkcodeGen_DataSourceInsertpcodeGen_DataBase_dataSourceModifications (_coordref, MkcodeGen_DataSourceInsert (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_DataSourceUpdate)) 
		return _MkcodeGen_DataSourceUpdatepcodeGen_DataBase_dataSourceModifications (_coordref, MkcodeGen_DataSourceUpdate (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_DataSourceDelete)) 
		return _MkcodeGen_DataSourceDeletepcodeGen_DataBase_dataSourceModifications (_coordref, MkcodeGen_DataSourceDelete (_coordref, t));
	message (DEADLY, "a pcodeGen_DataBase_dataSourceModifications list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_DataBase_dataSourceModifications (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_DataBase_dataSourceModifications (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_DataBase_dataSourceModifications)
	l =_ElempcodeGen_DataBase_dataSourceModifications (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_DataBase_dataSourceModifications)
	r = _ElempcodeGen_DataBase_dataSourceModifications (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_DataBase_dataSourceModifications)r)->_desc2);
((_TPPLST_0pcodeGen_DataBase_dataSourceModifications)r)->_desc2 = ((_TPPLST_0pcodeGen_DataBase_dataSourceModifications)l)->_desc2;
((_TPPLST_0pcodeGen_DataBase_dataSourceModifications)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataBase_dataSourceModifications (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_DataBase_dataSourceModifications (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_DataBase_dataSourceModifications root;
	_TPPLST_0pcodeGen_DataBase_dataSourceModifications list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_DataBase_dataSourceModifications)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_DataBase_dataSourceModifications)
		t = _ElempcodeGen_DataBase_dataSourceModifications (_coordref, t);
	remain = (_TPPLST_0pcodeGen_DataBase_dataSourceModifications)(((_TPPLST_0pcodeGen_DataBase_dataSourceModifications) t)->_desc2);
	((_TPPLST_0pcodeGen_DataBase_dataSourceModifications) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_DataBase_dataSourceModifications;
#else
list = (_TPPLST_0pcodeGen_DataBase_dataSourceModifications) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_DataBase_dataSourceModifications));
#endif
list->_prod = RULELST_0pcodeGen_DataBase_dataSourceModifications;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_DataBase_dataSourceModifications)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_DataSourceInsertpcodeGen_DataBase_dataSourceModifications)
	{_TPPLST_codeGen_DataSourceInsertpcodeGen_DataBase_dataSourceModifications _currn = _currn;
	 _TERMACT_LST_codeGen_DataSourceInsertpcodeGen_DataBase_dataSourceModifications;
	} this code has been moved to _ElempcodeGen_DataBase_dataSourceModifications*/
	/*if (_currn->_prod == RULELST_codeGen_DataSourceUpdatepcodeGen_DataBase_dataSourceModifications)
	{_TPPLST_codeGen_DataSourceUpdatepcodeGen_DataBase_dataSourceModifications _currn = _currn;
	 _TERMACT_LST_codeGen_DataSourceUpdatepcodeGen_DataBase_dataSourceModifications;
	} this code has been moved to _ElempcodeGen_DataBase_dataSourceModifications*/
	/*if (_currn->_prod == RULELST_codeGen_DataSourceDeletepcodeGen_DataBase_dataSourceModifications)
	{_TPPLST_codeGen_DataSourceDeletepcodeGen_DataBase_dataSourceModifications _currn = _currn;
	 _TERMACT_LST_codeGen_DataSourceDeletepcodeGen_DataBase_dataSourceModifications;
	} this code has been moved to _ElempcodeGen_DataBase_dataSourceModifications*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_DataBase_dataSourceModifications;
#else
root = (_TPPpcodeGen_DataBase_dataSourceModifications) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataBase_dataSourceModifications));
#endif
root->_prod = RULEpcodeGen_DataBase_dataSourceModifications;
root->_desc1 = (_TSPLST_codeGen_DataBase_dataSourceModifications)list;
{ _TPPpcodeGen_DataBase_dataSourceModifications _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_DataBase_dataSourceModifications;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataBase_setSize (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataBase_setSize (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataBase_setSize _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataBase_setSize;
#else
_currn = (_TPPpcodeGen_DataBase_setSize) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataBase_setSize));
#endif
_currn->_prod = RULEpcodeGen_DataBase_setSize;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataBase_setSize;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataBase_setSize */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataBase_modifySetSize (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataBase_modifySetSize (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataBase_modifySetSize _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataBase_modifySetSize;
#else
_currn = (_TPPpcodeGen_DataBase_modifySetSize) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataBase_modifySetSize));
#endif
_currn->_prod = RULEpcodeGen_DataBase_modifySetSize;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataBase_modifySetSize;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataBase_modifySetSize */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DBColumn (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_DBColumn (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_DBColumn _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DBColumn;
#else
_currn = (_TPPpcodeGen_DBColumn) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DBColumn));
#endif
_currn->_prod = RULEpcodeGen_DBColumn;
_currn->_desc1 = (_TSPcodeGen_DBColumn_name) MkcodeGen_DBColumn_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DBColumn: root of subtree no. 1 can not be made a codeGen_DBColumn_name node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_DBColumn_type) MkcodeGen_DBColumn_type (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DBColumn: root of subtree no. 2 can not be made a codeGen_DBColumn_type node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_DBColumn;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_DBColumn */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DBColumn_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DBColumn_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DBColumn_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DBColumn_name;
#else
_currn = (_TPPpcodeGen_DBColumn_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DBColumn_name));
#endif
_currn->_prod = RULEpcodeGen_DBColumn_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DBColumn_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DBColumn_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DBString (POSITION *_coordref, int _TERM1)
#else
NODEPTR MkpcodeGen_DBString (_coordref, _TERM1)
	POSITION *_coordref;
	int _TERM1;
#endif
{	_TPPpcodeGen_DBString _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DBString;
#else
_currn = (_TPPpcodeGen_DBString) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DBString));
#endif
_currn->_prod = RULEpcodeGen_DBString;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DBString;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_DBString */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DBInteger (POSITION *_coordref, int _TERM1)
#else
NODEPTR MkpcodeGen_DBInteger (_coordref, _TERM1)
	POSITION *_coordref;
	int _TERM1;
#endif
{	_TPPpcodeGen_DBInteger _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DBInteger;
#else
_currn = (_TPPpcodeGen_DBInteger) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DBInteger));
#endif
_currn->_prod = RULEpcodeGen_DBInteger;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DBInteger;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_DBInteger */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DBFloat (POSITION *_coordref, int _TERM1)
#else
NODEPTR MkpcodeGen_DBFloat (_coordref, _TERM1)
	POSITION *_coordref;
	int _TERM1;
#endif
{	_TPPpcodeGen_DBFloat _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DBFloat;
#else
_currn = (_TPPpcodeGen_DBFloat) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DBFloat));
#endif
_currn->_prod = RULEpcodeGen_DBFloat;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DBFloat;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_DBFloat */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DBDouble (POSITION *_coordref, int _TERM1)
#else
NODEPTR MkpcodeGen_DBDouble (_coordref, _TERM1)
	POSITION *_coordref;
	int _TERM1;
#endif
{	_TPPpcodeGen_DBDouble _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DBDouble;
#else
_currn = (_TPPpcodeGen_DBDouble) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DBDouble));
#endif
_currn->_prod = RULEpcodeGen_DBDouble;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DBDouble;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_DBDouble */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DBBigInt (POSITION *_coordref, int _TERM1)
#else
NODEPTR MkpcodeGen_DBBigInt (_coordref, _TERM1)
	POSITION *_coordref;
	int _TERM1;
#endif
{	_TPPpcodeGen_DBBigInt _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DBBigInt;
#else
_currn = (_TPPpcodeGen_DBBigInt) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DBBigInt));
#endif
_currn->_prod = RULEpcodeGen_DBBigInt;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DBBigInt;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_DBBigInt */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DBDate (POSITION *_coordref, int _TERM1)
#else
NODEPTR MkpcodeGen_DBDate (_coordref, _TERM1)
	POSITION *_coordref;
	int _TERM1;
#endif
{	_TPPpcodeGen_DBDate _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DBDate;
#else
_currn = (_TPPpcodeGen_DBDate) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DBDate));
#endif
_currn->_prod = RULEpcodeGen_DBDate;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DBDate;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_DBDate */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DBTimeStamp (POSITION *_coordref, int _TERM1)
#else
NODEPTR MkpcodeGen_DBTimeStamp (_coordref, _TERM1)
	POSITION *_coordref;
	int _TERM1;
#endif
{	_TPPpcodeGen_DBTimeStamp _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DBTimeStamp;
#else
_currn = (_TPPpcodeGen_DBTimeStamp) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DBTimeStamp));
#endif
_currn->_prod = RULEpcodeGen_DBTimeStamp;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DBTimeStamp;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_DBTimeStamp */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DBBoolean (POSITION *_coordref, int _TERM1)
#else
NODEPTR MkpcodeGen_DBBoolean (_coordref, _TERM1)
	POSITION *_coordref;
	int _TERM1;
#endif
{	_TPPpcodeGen_DBBoolean _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DBBoolean;
#else
_currn = (_TPPpcodeGen_DBBoolean) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DBBoolean));
#endif
_currn->_prod = RULEpcodeGen_DBBoolean;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DBBoolean;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_DBBoolean */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DBBlob (POSITION *_coordref, int _TERM1)
#else
NODEPTR MkpcodeGen_DBBlob (_coordref, _TERM1)
	POSITION *_coordref;
	int _TERM1;
#endif
{	_TPPpcodeGen_DBBlob _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DBBlob;
#else
_currn = (_TPPpcodeGen_DBBlob) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DBBlob));
#endif
_currn->_prod = RULEpcodeGen_DBBlob;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DBBlob;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_DBBlob */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceProjectionFilter (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5, NODEPTR _desc6)
#else
NODEPTR MkpcodeGen_DataSourceProjectionFilter (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5,_desc6)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
	NODEPTR _desc6;
#endif
{	_TPPpcodeGen_DataSourceProjectionFilter _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceProjectionFilter;
#else
_currn = (_TPPpcodeGen_DataSourceProjectionFilter) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceProjectionFilter));
#endif
_currn->_prod = RULEpcodeGen_DataSourceProjectionFilter;
_currn->_desc1 = (_TSPcodeGen_DataSourceProjectionFilter_name) MkcodeGen_DataSourceProjectionFilter_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceProjectionFilter: root of subtree no. 1 can not be made a codeGen_DataSourceProjectionFilter_name node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_DataSourceProjectionFilter_position) MkcodeGen_DataSourceProjectionFilter_position (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceProjectionFilter: root of subtree no. 2 can not be made a codeGen_DataSourceProjectionFilter_position node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_DataSourceProjectionFilter_sourceDataSource) MkcodeGen_DataSourceProjectionFilter_sourceDataSource (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceProjectionFilter: root of subtree no. 3 can not be made a codeGen_DataSourceProjectionFilter_sourceDataSource node ", 0, _coordref);
_currn->_desc4 = (_TSPcodeGen_DataSourceProjectionFilter_condition) MkcodeGen_DataSourceProjectionFilter_condition (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceProjectionFilter: root of subtree no. 4 can not be made a codeGen_DataSourceProjectionFilter_condition node ", 0, _coordref);
_currn->_desc5 = (_TSPcodeGen_DataSourceProjectionFilter_formalParameters) MkcodeGen_DataSourceProjectionFilter_formalParameters (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceProjectionFilter: root of subtree no. 5 can not be made a codeGen_DataSourceProjectionFilter_formalParameters node ", 0, _coordref);
_currn->_desc6 = (_TSPcodeGen_DataSourceProjectionFilter_projectionColumns) MkcodeGen_DataSourceProjectionFilter_projectionColumns (_coordref, _desc6);	
if (((NODEPTR)_currn->_desc6) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceProjectionFilter: root of subtree no. 6 can not be made a codeGen_DataSourceProjectionFilter_projectionColumns node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceProjectionFilter;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceProjectionFilter */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceProjectionFilter_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataSourceProjectionFilter_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataSourceProjectionFilter_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceProjectionFilter_name;
#else
_currn = (_TPPpcodeGen_DataSourceProjectionFilter_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceProjectionFilter_name));
#endif
_currn->_prod = RULEpcodeGen_DataSourceProjectionFilter_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceProjectionFilter_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceProjectionFilter_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceProjectionFilter_position (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataSourceProjectionFilter_position (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataSourceProjectionFilter_position _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceProjectionFilter_position;
#else
_currn = (_TPPpcodeGen_DataSourceProjectionFilter_position) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceProjectionFilter_position));
#endif
_currn->_prod = RULEpcodeGen_DataSourceProjectionFilter_position;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceProjectionFilter_position;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceProjectionFilter_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceProjectionFilter_sourceDataSource (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataSourceProjectionFilter_sourceDataSource (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataSourceProjectionFilter_sourceDataSource _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceProjectionFilter_sourceDataSource;
#else
_currn = (_TPPpcodeGen_DataSourceProjectionFilter_sourceDataSource) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceProjectionFilter_sourceDataSource));
#endif
_currn->_prod = RULEpcodeGen_DataSourceProjectionFilter_sourceDataSource;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceProjectionFilter_sourceDataSource;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceProjectionFilter_sourceDataSource */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_DataSourceFormalParampcodeGen_DataSourceProjectionFilter_formalParameters (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_DataSourceFormalParampcodeGen_DataSourceProjectionFilter_formalParameters (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_DataSourceFormalParampcodeGen_DataSourceProjectionFilter_formalParameters _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_DataSourceFormalParampcodeGen_DataSourceProjectionFilter_formalParameters;
#else
	_new = (_TPPLST_codeGen_DataSourceFormalParampcodeGen_DataSourceProjectionFilter_formalParameters) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_DataSourceFormalParampcodeGen_DataSourceProjectionFilter_formalParameters));
#endif
	_new->_prod = RULELST_codeGen_DataSourceFormalParampcodeGen_DataSourceProjectionFilter_formalParameters;
	_new->_desc1 = (_TSPcodeGen_DataSourceFormalParam)elem;
	_new->_desc2 = (_TSPLST_codeGen_DataSourceProjectionFilter_formalParameters)_new;
	{_TPPLST_codeGen_DataSourceFormalParampcodeGen_DataSourceProjectionFilter_formalParameters _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_DataSourceFormalParampcodeGen_DataSourceProjectionFilter_formalParameters;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_DataSourceProjectionFilter_formalParameters (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_DataSourceProjectionFilter_formalParameters (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_DataSourceFormalParam)) 
		return _MkcodeGen_DataSourceFormalParampcodeGen_DataSourceProjectionFilter_formalParameters (_coordref, MkcodeGen_DataSourceFormalParam (_coordref, t));
	message (DEADLY, "a pcodeGen_DataSourceProjectionFilter_formalParameters list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_DataSourceProjectionFilter_formalParameters (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_DataSourceProjectionFilter_formalParameters (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_DataSourceProjectionFilter_formalParameters)
	l =_ElempcodeGen_DataSourceProjectionFilter_formalParameters (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_DataSourceProjectionFilter_formalParameters)
	r = _ElempcodeGen_DataSourceProjectionFilter_formalParameters (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_DataSourceProjectionFilter_formalParameters)r)->_desc2);
((_TPPLST_0pcodeGen_DataSourceProjectionFilter_formalParameters)r)->_desc2 = ((_TPPLST_0pcodeGen_DataSourceProjectionFilter_formalParameters)l)->_desc2;
((_TPPLST_0pcodeGen_DataSourceProjectionFilter_formalParameters)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceProjectionFilter_formalParameters (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_DataSourceProjectionFilter_formalParameters (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_DataSourceProjectionFilter_formalParameters root;
	_TPPLST_0pcodeGen_DataSourceProjectionFilter_formalParameters list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_DataSourceProjectionFilter_formalParameters)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_DataSourceProjectionFilter_formalParameters)
		t = _ElempcodeGen_DataSourceProjectionFilter_formalParameters (_coordref, t);
	remain = (_TPPLST_0pcodeGen_DataSourceProjectionFilter_formalParameters)(((_TPPLST_0pcodeGen_DataSourceProjectionFilter_formalParameters) t)->_desc2);
	((_TPPLST_0pcodeGen_DataSourceProjectionFilter_formalParameters) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_DataSourceProjectionFilter_formalParameters;
#else
list = (_TPPLST_0pcodeGen_DataSourceProjectionFilter_formalParameters) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_DataSourceProjectionFilter_formalParameters));
#endif
list->_prod = RULELST_0pcodeGen_DataSourceProjectionFilter_formalParameters;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_DataSourceProjectionFilter_formalParameters)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_DataSourceFormalParampcodeGen_DataSourceProjectionFilter_formalParameters)
	{_TPPLST_codeGen_DataSourceFormalParampcodeGen_DataSourceProjectionFilter_formalParameters _currn = _currn;
	 _TERMACT_LST_codeGen_DataSourceFormalParampcodeGen_DataSourceProjectionFilter_formalParameters;
	} this code has been moved to _ElempcodeGen_DataSourceProjectionFilter_formalParameters*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_DataSourceProjectionFilter_formalParameters;
#else
root = (_TPPpcodeGen_DataSourceProjectionFilter_formalParameters) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceProjectionFilter_formalParameters));
#endif
root->_prod = RULEpcodeGen_DataSourceProjectionFilter_formalParameters;
root->_desc1 = (_TSPLST_codeGen_DataSourceProjectionFilter_formalParameters)list;
{ _TPPpcodeGen_DataSourceProjectionFilter_formalParameters _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_DataSourceProjectionFilter_formalParameters;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_ColumnProjectionpcodeGen_DataSourceProjectionFilter_projectionColumns (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_ColumnProjectionpcodeGen_DataSourceProjectionFilter_projectionColumns (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_ColumnProjectionpcodeGen_DataSourceProjectionFilter_projectionColumns _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_ColumnProjectionpcodeGen_DataSourceProjectionFilter_projectionColumns;
#else
	_new = (_TPPLST_codeGen_ColumnProjectionpcodeGen_DataSourceProjectionFilter_projectionColumns) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_ColumnProjectionpcodeGen_DataSourceProjectionFilter_projectionColumns));
#endif
	_new->_prod = RULELST_codeGen_ColumnProjectionpcodeGen_DataSourceProjectionFilter_projectionColumns;
	_new->_desc1 = (_TSPcodeGen_ColumnProjection)elem;
	_new->_desc2 = (_TSPLST_codeGen_DataSourceProjectionFilter_projectionColumns)_new;
	{_TPPLST_codeGen_ColumnProjectionpcodeGen_DataSourceProjectionFilter_projectionColumns _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_ColumnProjectionpcodeGen_DataSourceProjectionFilter_projectionColumns;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_ComputationProjectionpcodeGen_DataSourceProjectionFilter_projectionColumns (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_ComputationProjectionpcodeGen_DataSourceProjectionFilter_projectionColumns (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_ComputationProjectionpcodeGen_DataSourceProjectionFilter_projectionColumns _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_ComputationProjectionpcodeGen_DataSourceProjectionFilter_projectionColumns;
#else
	_new = (_TPPLST_codeGen_ComputationProjectionpcodeGen_DataSourceProjectionFilter_projectionColumns) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_ComputationProjectionpcodeGen_DataSourceProjectionFilter_projectionColumns));
#endif
	_new->_prod = RULELST_codeGen_ComputationProjectionpcodeGen_DataSourceProjectionFilter_projectionColumns;
	_new->_desc1 = (_TSPcodeGen_ComputationProjection)elem;
	_new->_desc2 = (_TSPLST_codeGen_DataSourceProjectionFilter_projectionColumns)_new;
	{_TPPLST_codeGen_ComputationProjectionpcodeGen_DataSourceProjectionFilter_projectionColumns _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_ComputationProjectionpcodeGen_DataSourceProjectionFilter_projectionColumns;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_DataSourceProjectionFilter_projectionColumns (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_DataSourceProjectionFilter_projectionColumns (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_ColumnProjection)) 
		return _MkcodeGen_ColumnProjectionpcodeGen_DataSourceProjectionFilter_projectionColumns (_coordref, MkcodeGen_ColumnProjection (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_ComputationProjection)) 
		return _MkcodeGen_ComputationProjectionpcodeGen_DataSourceProjectionFilter_projectionColumns (_coordref, MkcodeGen_ComputationProjection (_coordref, t));
	message (DEADLY, "a pcodeGen_DataSourceProjectionFilter_projectionColumns list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_DataSourceProjectionFilter_projectionColumns (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_DataSourceProjectionFilter_projectionColumns (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_DataSourceProjectionFilter_projectionColumns)
	l =_ElempcodeGen_DataSourceProjectionFilter_projectionColumns (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_DataSourceProjectionFilter_projectionColumns)
	r = _ElempcodeGen_DataSourceProjectionFilter_projectionColumns (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_DataSourceProjectionFilter_projectionColumns)r)->_desc2);
((_TPPLST_0pcodeGen_DataSourceProjectionFilter_projectionColumns)r)->_desc2 = ((_TPPLST_0pcodeGen_DataSourceProjectionFilter_projectionColumns)l)->_desc2;
((_TPPLST_0pcodeGen_DataSourceProjectionFilter_projectionColumns)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceProjectionFilter_projectionColumns (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_DataSourceProjectionFilter_projectionColumns (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_DataSourceProjectionFilter_projectionColumns root;
	_TPPLST_0pcodeGen_DataSourceProjectionFilter_projectionColumns list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_DataSourceProjectionFilter_projectionColumns)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_DataSourceProjectionFilter_projectionColumns)
		t = _ElempcodeGen_DataSourceProjectionFilter_projectionColumns (_coordref, t);
	remain = (_TPPLST_0pcodeGen_DataSourceProjectionFilter_projectionColumns)(((_TPPLST_0pcodeGen_DataSourceProjectionFilter_projectionColumns) t)->_desc2);
	((_TPPLST_0pcodeGen_DataSourceProjectionFilter_projectionColumns) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_DataSourceProjectionFilter_projectionColumns;
#else
list = (_TPPLST_0pcodeGen_DataSourceProjectionFilter_projectionColumns) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_DataSourceProjectionFilter_projectionColumns));
#endif
list->_prod = RULELST_0pcodeGen_DataSourceProjectionFilter_projectionColumns;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_DataSourceProjectionFilter_projectionColumns)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_ColumnProjectionpcodeGen_DataSourceProjectionFilter_projectionColumns)
	{_TPPLST_codeGen_ColumnProjectionpcodeGen_DataSourceProjectionFilter_projectionColumns _currn = _currn;
	 _TERMACT_LST_codeGen_ColumnProjectionpcodeGen_DataSourceProjectionFilter_projectionColumns;
	} this code has been moved to _ElempcodeGen_DataSourceProjectionFilter_projectionColumns*/
	/*if (_currn->_prod == RULELST_codeGen_ComputationProjectionpcodeGen_DataSourceProjectionFilter_projectionColumns)
	{_TPPLST_codeGen_ComputationProjectionpcodeGen_DataSourceProjectionFilter_projectionColumns _currn = _currn;
	 _TERMACT_LST_codeGen_ComputationProjectionpcodeGen_DataSourceProjectionFilter_projectionColumns;
	} this code has been moved to _ElempcodeGen_DataSourceProjectionFilter_projectionColumns*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_DataSourceProjectionFilter_projectionColumns;
#else
root = (_TPPpcodeGen_DataSourceProjectionFilter_projectionColumns) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceProjectionFilter_projectionColumns));
#endif
root->_prod = RULEpcodeGen_DataSourceProjectionFilter_projectionColumns;
root->_desc1 = (_TSPLST_codeGen_DataSourceProjectionFilter_projectionColumns)list;
{ _TPPpcodeGen_DataSourceProjectionFilter_projectionColumns _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_DataSourceProjectionFilter_projectionColumns;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ColumnProjection (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpcodeGen_ColumnProjection (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpcodeGen_ColumnProjection _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ColumnProjection;
#else
_currn = (_TPPpcodeGen_ColumnProjection) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ColumnProjection));
#endif
_currn->_prod = RULEpcodeGen_ColumnProjection;
_currn->_desc1 = (_TSPcodeGen_ColumnProjection_column) MkcodeGen_ColumnProjection_column (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_ColumnProjection: root of subtree no. 1 can not be made a codeGen_ColumnProjection_column node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_ColumnProjection;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_ColumnProjection */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ColumnProjection_column (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_ColumnProjection_column (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_ColumnProjection_column _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ColumnProjection_column;
#else
_currn = (_TPPpcodeGen_ColumnProjection_column) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ColumnProjection_column));
#endif
_currn->_prod = RULEpcodeGen_ColumnProjection_column;
_SETCOORD(_currn)
_TERMACT_pcodeGen_ColumnProjection_column;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_ColumnProjection_column */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ComputationProjection (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_ComputationProjection (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_ComputationProjection _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ComputationProjection;
#else
_currn = (_TPPpcodeGen_ComputationProjection) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ComputationProjection));
#endif
_currn->_prod = RULEpcodeGen_ComputationProjection;
_currn->_desc1 = (_TSPcodeGen_ComputationProjection_column) MkcodeGen_ComputationProjection_column (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_ComputationProjection: root of subtree no. 1 can not be made a codeGen_ComputationProjection_column node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_ComputationProjection_computation) MkcodeGen_ComputationProjection_computation (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_ComputationProjection: root of subtree no. 2 can not be made a codeGen_ComputationProjection_computation node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_ComputationProjection;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_ComputationProjection */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ComputationProjection_column (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_ComputationProjection_column (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_ComputationProjection_column _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ComputationProjection_column;
#else
_currn = (_TPPpcodeGen_ComputationProjection_column) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ComputationProjection_column));
#endif
_currn->_prod = RULEpcodeGen_ComputationProjection_column;
_SETCOORD(_currn)
_TERMACT_pcodeGen_ComputationProjection_column;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_ComputationProjection_column */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ComputationProjection_computation (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_ComputationProjection_computation (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_ComputationProjection_computation _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ComputationProjection_computation;
#else
_currn = (_TPPpcodeGen_ComputationProjection_computation) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ComputationProjection_computation));
#endif
_currn->_prod = RULEpcodeGen_ComputationProjection_computation;
_SETCOORD(_currn)
_TERMACT_pcodeGen_ComputationProjection_computation;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_ComputationProjection_computation */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_LimitDataSource (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5, NODEPTR _desc6)
#else
NODEPTR MkpcodeGen_LimitDataSource (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5,_desc6)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
	NODEPTR _desc6;
#endif
{	_TPPpcodeGen_LimitDataSource _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_LimitDataSource;
#else
_currn = (_TPPpcodeGen_LimitDataSource) TreeNodeAlloc (sizeof (struct _TPpcodeGen_LimitDataSource));
#endif
_currn->_prod = RULEpcodeGen_LimitDataSource;
_currn->_desc1 = (_TSPcodeGen_LimitDataSource_name) MkcodeGen_LimitDataSource_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_LimitDataSource: root of subtree no. 1 can not be made a codeGen_LimitDataSource_name node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_LimitDataSource_position) MkcodeGen_LimitDataSource_position (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_LimitDataSource: root of subtree no. 2 can not be made a codeGen_LimitDataSource_position node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_LimitDataSource_sourceDataSouce) MkcodeGen_LimitDataSource_sourceDataSouce (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_LimitDataSource: root of subtree no. 3 can not be made a codeGen_LimitDataSource_sourceDataSouce node ", 0, _coordref);
_currn->_desc4 = (_TSPcodeGen_LimitDataSource_minRow) MkcodeGen_LimitDataSource_minRow (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_LimitDataSource: root of subtree no. 4 can not be made a codeGen_LimitDataSource_minRow node ", 0, _coordref);
_currn->_desc5 = (_TSPcodeGen_LimitDataSource_maxRow) MkcodeGen_LimitDataSource_maxRow (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_LimitDataSource: root of subtree no. 5 can not be made a codeGen_LimitDataSource_maxRow node ", 0, _coordref);
_currn->_desc6 = (_TSPcodeGen_LimitDataSource_formalParameters) MkcodeGen_LimitDataSource_formalParameters (_coordref, _desc6);	
if (((NODEPTR)_currn->_desc6) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_LimitDataSource: root of subtree no. 6 can not be made a codeGen_LimitDataSource_formalParameters node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_LimitDataSource;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_LimitDataSource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_LimitDataSource_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_LimitDataSource_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_LimitDataSource_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_LimitDataSource_name;
#else
_currn = (_TPPpcodeGen_LimitDataSource_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_LimitDataSource_name));
#endif
_currn->_prod = RULEpcodeGen_LimitDataSource_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_LimitDataSource_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_LimitDataSource_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_LimitDataSource_position (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_LimitDataSource_position (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_LimitDataSource_position _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_LimitDataSource_position;
#else
_currn = (_TPPpcodeGen_LimitDataSource_position) TreeNodeAlloc (sizeof (struct _TPpcodeGen_LimitDataSource_position));
#endif
_currn->_prod = RULEpcodeGen_LimitDataSource_position;
_SETCOORD(_currn)
_TERMACT_pcodeGen_LimitDataSource_position;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_LimitDataSource_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_LimitDataSource_sourceDataSouce (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_LimitDataSource_sourceDataSouce (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_LimitDataSource_sourceDataSouce _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_LimitDataSource_sourceDataSouce;
#else
_currn = (_TPPpcodeGen_LimitDataSource_sourceDataSouce) TreeNodeAlloc (sizeof (struct _TPpcodeGen_LimitDataSource_sourceDataSouce));
#endif
_currn->_prod = RULEpcodeGen_LimitDataSource_sourceDataSouce;
_SETCOORD(_currn)
_TERMACT_pcodeGen_LimitDataSource_sourceDataSouce;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_LimitDataSource_sourceDataSouce */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_DataSourceFormalParampcodeGen_LimitDataSource_formalParameters (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_DataSourceFormalParampcodeGen_LimitDataSource_formalParameters (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_DataSourceFormalParampcodeGen_LimitDataSource_formalParameters _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_DataSourceFormalParampcodeGen_LimitDataSource_formalParameters;
#else
	_new = (_TPPLST_codeGen_DataSourceFormalParampcodeGen_LimitDataSource_formalParameters) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_DataSourceFormalParampcodeGen_LimitDataSource_formalParameters));
#endif
	_new->_prod = RULELST_codeGen_DataSourceFormalParampcodeGen_LimitDataSource_formalParameters;
	_new->_desc1 = (_TSPcodeGen_DataSourceFormalParam)elem;
	_new->_desc2 = (_TSPLST_codeGen_LimitDataSource_formalParameters)_new;
	{_TPPLST_codeGen_DataSourceFormalParampcodeGen_LimitDataSource_formalParameters _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_DataSourceFormalParampcodeGen_LimitDataSource_formalParameters;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_LimitDataSource_formalParameters (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_LimitDataSource_formalParameters (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_DataSourceFormalParam)) 
		return _MkcodeGen_DataSourceFormalParampcodeGen_LimitDataSource_formalParameters (_coordref, MkcodeGen_DataSourceFormalParam (_coordref, t));
	message (DEADLY, "a pcodeGen_LimitDataSource_formalParameters list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_LimitDataSource_formalParameters (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_LimitDataSource_formalParameters (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_LimitDataSource_formalParameters)
	l =_ElempcodeGen_LimitDataSource_formalParameters (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_LimitDataSource_formalParameters)
	r = _ElempcodeGen_LimitDataSource_formalParameters (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_LimitDataSource_formalParameters)r)->_desc2);
((_TPPLST_0pcodeGen_LimitDataSource_formalParameters)r)->_desc2 = ((_TPPLST_0pcodeGen_LimitDataSource_formalParameters)l)->_desc2;
((_TPPLST_0pcodeGen_LimitDataSource_formalParameters)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_LimitDataSource_formalParameters (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_LimitDataSource_formalParameters (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_LimitDataSource_formalParameters root;
	_TPPLST_0pcodeGen_LimitDataSource_formalParameters list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_LimitDataSource_formalParameters)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_LimitDataSource_formalParameters)
		t = _ElempcodeGen_LimitDataSource_formalParameters (_coordref, t);
	remain = (_TPPLST_0pcodeGen_LimitDataSource_formalParameters)(((_TPPLST_0pcodeGen_LimitDataSource_formalParameters) t)->_desc2);
	((_TPPLST_0pcodeGen_LimitDataSource_formalParameters) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_LimitDataSource_formalParameters;
#else
list = (_TPPLST_0pcodeGen_LimitDataSource_formalParameters) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_LimitDataSource_formalParameters));
#endif
list->_prod = RULELST_0pcodeGen_LimitDataSource_formalParameters;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_LimitDataSource_formalParameters)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_DataSourceFormalParampcodeGen_LimitDataSource_formalParameters)
	{_TPPLST_codeGen_DataSourceFormalParampcodeGen_LimitDataSource_formalParameters _currn = _currn;
	 _TERMACT_LST_codeGen_DataSourceFormalParampcodeGen_LimitDataSource_formalParameters;
	} this code has been moved to _ElempcodeGen_LimitDataSource_formalParameters*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_LimitDataSource_formalParameters;
#else
root = (_TPPpcodeGen_LimitDataSource_formalParameters) TreeNodeAlloc (sizeof (struct _TPpcodeGen_LimitDataSource_formalParameters));
#endif
root->_prod = RULEpcodeGen_LimitDataSource_formalParameters;
root->_desc1 = (_TSPLST_codeGen_LimitDataSource_formalParameters)list;
{ _TPPpcodeGen_LimitDataSource_formalParameters _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_LimitDataSource_formalParameters;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_BinaryExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpcodeGen_BinaryExpression (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpcodeGen_BinaryExpression _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_BinaryExpression;
#else
_currn = (_TPPpcodeGen_BinaryExpression) TreeNodeAlloc (sizeof (struct _TPpcodeGen_BinaryExpression));
#endif
_currn->_prod = RULEpcodeGen_BinaryExpression;
_currn->_desc1 = (_TSPcodeGen_BinaryExpression_leftOperand) MkcodeGen_BinaryExpression_leftOperand (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_BinaryExpression: root of subtree no. 1 can not be made a codeGen_BinaryExpression_leftOperand node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_BinaryExpression_rightOperand) MkcodeGen_BinaryExpression_rightOperand (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_BinaryExpression: root of subtree no. 2 can not be made a codeGen_BinaryExpression_rightOperand node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_BinaryExpression_operator) MkcodeGen_BinaryExpression_operator (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_BinaryExpression: root of subtree no. 3 can not be made a codeGen_BinaryExpression_operator node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_BinaryExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_BinaryExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_BinaryExpression_operator (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_BinaryExpression_operator (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_BinaryExpression_operator _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_BinaryExpression_operator;
#else
_currn = (_TPPpcodeGen_BinaryExpression_operator) TreeNodeAlloc (sizeof (struct _TPpcodeGen_BinaryExpression_operator));
#endif
_currn->_prod = RULEpcodeGen_BinaryExpression_operator;
_SETCOORD(_currn)
_TERMACT_pcodeGen_BinaryExpression_operator;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_BinaryExpression_operator */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_UnaryExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_UnaryExpression (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_UnaryExpression _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_UnaryExpression;
#else
_currn = (_TPPpcodeGen_UnaryExpression) TreeNodeAlloc (sizeof (struct _TPpcodeGen_UnaryExpression));
#endif
_currn->_prod = RULEpcodeGen_UnaryExpression;
_currn->_desc1 = (_TSPcodeGen_UnaryExpression_operand) MkcodeGen_UnaryExpression_operand (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_UnaryExpression: root of subtree no. 1 can not be made a codeGen_UnaryExpression_operand node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_UnaryExpression_operator) MkcodeGen_UnaryExpression_operator (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_UnaryExpression: root of subtree no. 2 can not be made a codeGen_UnaryExpression_operator node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_UnaryExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_UnaryExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_UnaryExpression_operator (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_UnaryExpression_operator (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_UnaryExpression_operator _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_UnaryExpression_operator;
#else
_currn = (_TPPpcodeGen_UnaryExpression_operator) TreeNodeAlloc (sizeof (struct _TPpcodeGen_UnaryExpression_operator));
#endif
_currn->_prod = RULEpcodeGen_UnaryExpression_operator;
_SETCOORD(_currn)
_TERMACT_pcodeGen_UnaryExpression_operator;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_UnaryExpression_operator */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ConstantExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_ConstantExpression (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_ConstantExpression _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ConstantExpression;
#else
_currn = (_TPPpcodeGen_ConstantExpression) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ConstantExpression));
#endif
_currn->_prod = RULEpcodeGen_ConstantExpression;
_currn->_desc1 = (_TSPcodeGen_ConstantExpression_type) MkcodeGen_ConstantExpression_type (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_ConstantExpression: root of subtree no. 1 can not be made a codeGen_ConstantExpression_type node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_ConstantExpression_value) MkcodeGen_ConstantExpression_value (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_ConstantExpression: root of subtree no. 2 can not be made a codeGen_ConstantExpression_value node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_ConstantExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_ConstantExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ConstantExpression_value (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_ConstantExpression_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_ConstantExpression_value _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ConstantExpression_value;
#else
_currn = (_TPPpcodeGen_ConstantExpression_value) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ConstantExpression_value));
#endif
_currn->_prod = RULEpcodeGen_ConstantExpression_value;
_SETCOORD(_currn)
_TERMACT_pcodeGen_ConstantExpression_value;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_ConstantExpression_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ColumnValueExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpcodeGen_ColumnValueExpression (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpcodeGen_ColumnValueExpression _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ColumnValueExpression;
#else
_currn = (_TPPpcodeGen_ColumnValueExpression) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ColumnValueExpression));
#endif
_currn->_prod = RULEpcodeGen_ColumnValueExpression;
_currn->_desc1 = (_TSPcodeGen_ColumnValueExpression_column) MkcodeGen_ColumnValueExpression_column (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_ColumnValueExpression: root of subtree no. 1 can not be made a codeGen_ColumnValueExpression_column node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_ColumnValueExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_ColumnValueExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ColumnValueExpression_column (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_ColumnValueExpression_column (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_ColumnValueExpression_column _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ColumnValueExpression_column;
#else
_currn = (_TPPpcodeGen_ColumnValueExpression_column) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ColumnValueExpression_column));
#endif
_currn->_prod = RULEpcodeGen_ColumnValueExpression_column;
_SETCOORD(_currn)
_TERMACT_pcodeGen_ColumnValueExpression_column;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_ColumnValueExpression_column */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ParameterValueExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpcodeGen_ParameterValueExpression (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpcodeGen_ParameterValueExpression _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ParameterValueExpression;
#else
_currn = (_TPPpcodeGen_ParameterValueExpression) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ParameterValueExpression));
#endif
_currn->_prod = RULEpcodeGen_ParameterValueExpression;
_currn->_desc1 = (_TSPcodeGen_ParameterValueExpression_parameter) MkcodeGen_ParameterValueExpression_parameter (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_ParameterValueExpression: root of subtree no. 1 can not be made a codeGen_ParameterValueExpression_parameter node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_ParameterValueExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_ParameterValueExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ParameterValueExpression_parameter (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_ParameterValueExpression_parameter (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_ParameterValueExpression_parameter _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ParameterValueExpression_parameter;
#else
_currn = (_TPPpcodeGen_ParameterValueExpression_parameter) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ParameterValueExpression_parameter));
#endif
_currn->_prod = RULEpcodeGen_ParameterValueExpression_parameter;
_SETCOORD(_currn)
_TERMACT_pcodeGen_ParameterValueExpression_parameter;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_ParameterValueExpression_parameter */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ComputationExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_ComputationExpression (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_ComputationExpression _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ComputationExpression;
#else
_currn = (_TPPpcodeGen_ComputationExpression) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ComputationExpression));
#endif
_currn->_prod = RULEpcodeGen_ComputationExpression;
_currn->_desc1 = (_TSPcodeGen_ComputationExpression_column) MkcodeGen_ComputationExpression_column (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_ComputationExpression: root of subtree no. 1 can not be made a codeGen_ComputationExpression_column node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_ComputationExpression_computation) MkcodeGen_ComputationExpression_computation (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_ComputationExpression: root of subtree no. 2 can not be made a codeGen_ComputationExpression_computation node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_ComputationExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_ComputationExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ComputationExpression_column (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_ComputationExpression_column (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_ComputationExpression_column _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ComputationExpression_column;
#else
_currn = (_TPPpcodeGen_ComputationExpression_column) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ComputationExpression_column));
#endif
_currn->_prod = RULEpcodeGen_ComputationExpression_column;
_SETCOORD(_currn)
_TERMACT_pcodeGen_ComputationExpression_column;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_ComputationExpression_column */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ComputationExpression_computation (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_ComputationExpression_computation (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_ComputationExpression_computation _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ComputationExpression_computation;
#else
_currn = (_TPPpcodeGen_ComputationExpression_computation) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ComputationExpression_computation));
#endif
_currn->_prod = RULEpcodeGen_ComputationExpression_computation;
_SETCOORD(_currn)
_TERMACT_pcodeGen_ComputationExpression_computation;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_ComputationExpression_computation */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_DataSourceValue (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_DataSourceValue _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceValue;
#else
_currn = (_TPPpcodeGen_DataSourceValue) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceValue));
#endif
_currn->_prod = RULEpcodeGen_DataSourceValue;
_currn->_desc1 = (_TSPcodeGen_DataSourceValue_column) MkcodeGen_DataSourceValue_column (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceValue: root of subtree no. 1 can not be made a codeGen_DataSourceValue_column node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_DataSourceValue_expression) MkcodeGen_DataSourceValue_expression (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceValue: root of subtree no. 2 can not be made a codeGen_DataSourceValue_expression node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceValue_column (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataSourceValue_column (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataSourceValue_column _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceValue_column;
#else
_currn = (_TPPpcodeGen_DataSourceValue_column) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceValue_column));
#endif
_currn->_prod = RULEpcodeGen_DataSourceValue_column;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceValue_column;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceValue_column */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceJoin (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5, NODEPTR _desc6, NODEPTR _desc7)
#else
NODEPTR MkpcodeGen_DataSourceJoin (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5,_desc6,_desc7)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
	NODEPTR _desc6;
	NODEPTR _desc7;
#endif
{	_TPPpcodeGen_DataSourceJoin _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceJoin;
#else
_currn = (_TPPpcodeGen_DataSourceJoin) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceJoin));
#endif
_currn->_prod = RULEpcodeGen_DataSourceJoin;
_currn->_desc1 = (_TSPcodeGen_DataSourceJoin_name) MkcodeGen_DataSourceJoin_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceJoin: root of subtree no. 1 can not be made a codeGen_DataSourceJoin_name node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_DataSourceJoin_position) MkcodeGen_DataSourceJoin_position (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceJoin: root of subtree no. 2 can not be made a codeGen_DataSourceJoin_position node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_DataSourceJoin_dataSource1) MkcodeGen_DataSourceJoin_dataSource1 (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceJoin: root of subtree no. 3 can not be made a codeGen_DataSourceJoin_dataSource1 node ", 0, _coordref);
_currn->_desc4 = (_TSPcodeGen_DataSourceJoin_dataSource2) MkcodeGen_DataSourceJoin_dataSource2 (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceJoin: root of subtree no. 4 can not be made a codeGen_DataSourceJoin_dataSource2 node ", 0, _coordref);
_currn->_desc5 = (_TSPcodeGen_DataSourceJoin_columnRef1) MkcodeGen_DataSourceJoin_columnRef1 (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceJoin: root of subtree no. 5 can not be made a codeGen_DataSourceJoin_columnRef1 node ", 0, _coordref);
_currn->_desc6 = (_TSPcodeGen_DataSourceJoin_columnRef2) MkcodeGen_DataSourceJoin_columnRef2 (_coordref, _desc6);	
if (((NODEPTR)_currn->_desc6) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceJoin: root of subtree no. 6 can not be made a codeGen_DataSourceJoin_columnRef2 node ", 0, _coordref);
_currn->_desc7 = (_TSPcodeGen_DataSourceJoin_joinOperator) MkcodeGen_DataSourceJoin_joinOperator (_coordref, _desc7);	
if (((NODEPTR)_currn->_desc7) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceJoin: root of subtree no. 7 can not be made a codeGen_DataSourceJoin_joinOperator node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceJoin;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceJoin */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceJoin_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataSourceJoin_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataSourceJoin_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceJoin_name;
#else
_currn = (_TPPpcodeGen_DataSourceJoin_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceJoin_name));
#endif
_currn->_prod = RULEpcodeGen_DataSourceJoin_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceJoin_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceJoin_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceJoin_position (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataSourceJoin_position (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataSourceJoin_position _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceJoin_position;
#else
_currn = (_TPPpcodeGen_DataSourceJoin_position) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceJoin_position));
#endif
_currn->_prod = RULEpcodeGen_DataSourceJoin_position;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceJoin_position;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceJoin_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceJoin_dataSource1 (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataSourceJoin_dataSource1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataSourceJoin_dataSource1 _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceJoin_dataSource1;
#else
_currn = (_TPPpcodeGen_DataSourceJoin_dataSource1) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceJoin_dataSource1));
#endif
_currn->_prod = RULEpcodeGen_DataSourceJoin_dataSource1;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceJoin_dataSource1;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceJoin_dataSource1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceJoin_dataSource2 (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataSourceJoin_dataSource2 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataSourceJoin_dataSource2 _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceJoin_dataSource2;
#else
_currn = (_TPPpcodeGen_DataSourceJoin_dataSource2) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceJoin_dataSource2));
#endif
_currn->_prod = RULEpcodeGen_DataSourceJoin_dataSource2;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceJoin_dataSource2;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceJoin_dataSource2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceJoin_columnRef1 (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataSourceJoin_columnRef1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataSourceJoin_columnRef1 _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceJoin_columnRef1;
#else
_currn = (_TPPpcodeGen_DataSourceJoin_columnRef1) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceJoin_columnRef1));
#endif
_currn->_prod = RULEpcodeGen_DataSourceJoin_columnRef1;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceJoin_columnRef1;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceJoin_columnRef1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceJoin_columnRef2 (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataSourceJoin_columnRef2 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataSourceJoin_columnRef2 _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceJoin_columnRef2;
#else
_currn = (_TPPpcodeGen_DataSourceJoin_columnRef2) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceJoin_columnRef2));
#endif
_currn->_prod = RULEpcodeGen_DataSourceJoin_columnRef2;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceJoin_columnRef2;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceJoin_columnRef2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceJoin_joinOperator (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataSourceJoin_joinOperator (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataSourceJoin_joinOperator _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceJoin_joinOperator;
#else
_currn = (_TPPpcodeGen_DataSourceJoin_joinOperator) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceJoin_joinOperator));
#endif
_currn->_prod = RULEpcodeGen_DataSourceJoin_joinOperator;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceJoin_joinOperator;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceJoin_joinOperator */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceInsert (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkpcodeGen_DataSourceInsert (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPpcodeGen_DataSourceInsert _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceInsert;
#else
_currn = (_TPPpcodeGen_DataSourceInsert) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceInsert));
#endif
_currn->_prod = RULEpcodeGen_DataSourceInsert;
_currn->_desc1 = (_TSPcodeGen_DataSourceInsert_targetDataSource) MkcodeGen_DataSourceInsert_targetDataSource (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceInsert: root of subtree no. 1 can not be made a codeGen_DataSourceInsert_targetDataSource node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_DataSourceInsert_formalParams) MkcodeGen_DataSourceInsert_formalParams (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceInsert: root of subtree no. 2 can not be made a codeGen_DataSourceInsert_formalParams node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_DataSourceInsert_name) MkcodeGen_DataSourceInsert_name (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceInsert: root of subtree no. 3 can not be made a codeGen_DataSourceInsert_name node ", 0, _coordref);
_currn->_desc4 = (_TSPcodeGen_DataSourceInsert_position) MkcodeGen_DataSourceInsert_position (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceInsert: root of subtree no. 4 can not be made a codeGen_DataSourceInsert_position node ", 0, _coordref);
_currn->_desc5 = (_TSPcodeGen_DataSourceInsert_values) MkcodeGen_DataSourceInsert_values (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceInsert: root of subtree no. 5 can not be made a codeGen_DataSourceInsert_values node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceInsert;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceInsert */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceInsert_targetDataSource (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataSourceInsert_targetDataSource (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataSourceInsert_targetDataSource _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceInsert_targetDataSource;
#else
_currn = (_TPPpcodeGen_DataSourceInsert_targetDataSource) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceInsert_targetDataSource));
#endif
_currn->_prod = RULEpcodeGen_DataSourceInsert_targetDataSource;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceInsert_targetDataSource;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceInsert_targetDataSource */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_DataSourceFormalParampcodeGen_DataSourceInsert_formalParams (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_DataSourceFormalParampcodeGen_DataSourceInsert_formalParams (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_DataSourceFormalParampcodeGen_DataSourceInsert_formalParams _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_DataSourceFormalParampcodeGen_DataSourceInsert_formalParams;
#else
	_new = (_TPPLST_codeGen_DataSourceFormalParampcodeGen_DataSourceInsert_formalParams) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_DataSourceFormalParampcodeGen_DataSourceInsert_formalParams));
#endif
	_new->_prod = RULELST_codeGen_DataSourceFormalParampcodeGen_DataSourceInsert_formalParams;
	_new->_desc1 = (_TSPcodeGen_DataSourceFormalParam)elem;
	_new->_desc2 = (_TSPLST_codeGen_DataSourceInsert_formalParams)_new;
	{_TPPLST_codeGen_DataSourceFormalParampcodeGen_DataSourceInsert_formalParams _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_DataSourceFormalParampcodeGen_DataSourceInsert_formalParams;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_DataSourceInsert_formalParams (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_DataSourceInsert_formalParams (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_DataSourceFormalParam)) 
		return _MkcodeGen_DataSourceFormalParampcodeGen_DataSourceInsert_formalParams (_coordref, MkcodeGen_DataSourceFormalParam (_coordref, t));
	message (DEADLY, "a pcodeGen_DataSourceInsert_formalParams list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_DataSourceInsert_formalParams (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_DataSourceInsert_formalParams (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_DataSourceInsert_formalParams)
	l =_ElempcodeGen_DataSourceInsert_formalParams (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_DataSourceInsert_formalParams)
	r = _ElempcodeGen_DataSourceInsert_formalParams (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_DataSourceInsert_formalParams)r)->_desc2);
((_TPPLST_0pcodeGen_DataSourceInsert_formalParams)r)->_desc2 = ((_TPPLST_0pcodeGen_DataSourceInsert_formalParams)l)->_desc2;
((_TPPLST_0pcodeGen_DataSourceInsert_formalParams)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceInsert_formalParams (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_DataSourceInsert_formalParams (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_DataSourceInsert_formalParams root;
	_TPPLST_0pcodeGen_DataSourceInsert_formalParams list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_DataSourceInsert_formalParams)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_DataSourceInsert_formalParams)
		t = _ElempcodeGen_DataSourceInsert_formalParams (_coordref, t);
	remain = (_TPPLST_0pcodeGen_DataSourceInsert_formalParams)(((_TPPLST_0pcodeGen_DataSourceInsert_formalParams) t)->_desc2);
	((_TPPLST_0pcodeGen_DataSourceInsert_formalParams) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_DataSourceInsert_formalParams;
#else
list = (_TPPLST_0pcodeGen_DataSourceInsert_formalParams) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_DataSourceInsert_formalParams));
#endif
list->_prod = RULELST_0pcodeGen_DataSourceInsert_formalParams;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_DataSourceInsert_formalParams)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_DataSourceFormalParampcodeGen_DataSourceInsert_formalParams)
	{_TPPLST_codeGen_DataSourceFormalParampcodeGen_DataSourceInsert_formalParams _currn = _currn;
	 _TERMACT_LST_codeGen_DataSourceFormalParampcodeGen_DataSourceInsert_formalParams;
	} this code has been moved to _ElempcodeGen_DataSourceInsert_formalParams*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_DataSourceInsert_formalParams;
#else
root = (_TPPpcodeGen_DataSourceInsert_formalParams) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceInsert_formalParams));
#endif
root->_prod = RULEpcodeGen_DataSourceInsert_formalParams;
root->_desc1 = (_TSPLST_codeGen_DataSourceInsert_formalParams)list;
{ _TPPpcodeGen_DataSourceInsert_formalParams _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_DataSourceInsert_formalParams;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceInsert_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataSourceInsert_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataSourceInsert_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceInsert_name;
#else
_currn = (_TPPpcodeGen_DataSourceInsert_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceInsert_name));
#endif
_currn->_prod = RULEpcodeGen_DataSourceInsert_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceInsert_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceInsert_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceInsert_position (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataSourceInsert_position (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataSourceInsert_position _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceInsert_position;
#else
_currn = (_TPPpcodeGen_DataSourceInsert_position) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceInsert_position));
#endif
_currn->_prod = RULEpcodeGen_DataSourceInsert_position;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceInsert_position;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceInsert_position */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_DataSourceValuepcodeGen_DataSourceInsert_values (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_DataSourceValuepcodeGen_DataSourceInsert_values (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_DataSourceValuepcodeGen_DataSourceInsert_values _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_DataSourceValuepcodeGen_DataSourceInsert_values;
#else
	_new = (_TPPLST_codeGen_DataSourceValuepcodeGen_DataSourceInsert_values) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_DataSourceValuepcodeGen_DataSourceInsert_values));
#endif
	_new->_prod = RULELST_codeGen_DataSourceValuepcodeGen_DataSourceInsert_values;
	_new->_desc1 = (_TSPcodeGen_DataSourceValue)elem;
	_new->_desc2 = (_TSPLST_codeGen_DataSourceInsert_values)_new;
	{_TPPLST_codeGen_DataSourceValuepcodeGen_DataSourceInsert_values _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_DataSourceValuepcodeGen_DataSourceInsert_values;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_DataSourceInsert_values (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_DataSourceInsert_values (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_DataSourceValue)) 
		return _MkcodeGen_DataSourceValuepcodeGen_DataSourceInsert_values (_coordref, MkcodeGen_DataSourceValue (_coordref, t));
	message (DEADLY, "a pcodeGen_DataSourceInsert_values list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_DataSourceInsert_values (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_DataSourceInsert_values (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_DataSourceInsert_values)
	l =_ElempcodeGen_DataSourceInsert_values (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_DataSourceInsert_values)
	r = _ElempcodeGen_DataSourceInsert_values (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_DataSourceInsert_values)r)->_desc2);
((_TPPLST_0pcodeGen_DataSourceInsert_values)r)->_desc2 = ((_TPPLST_0pcodeGen_DataSourceInsert_values)l)->_desc2;
((_TPPLST_0pcodeGen_DataSourceInsert_values)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceInsert_values (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_DataSourceInsert_values (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_DataSourceInsert_values root;
	_TPPLST_0pcodeGen_DataSourceInsert_values list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_DataSourceInsert_values)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_DataSourceInsert_values)
		t = _ElempcodeGen_DataSourceInsert_values (_coordref, t);
	remain = (_TPPLST_0pcodeGen_DataSourceInsert_values)(((_TPPLST_0pcodeGen_DataSourceInsert_values) t)->_desc2);
	((_TPPLST_0pcodeGen_DataSourceInsert_values) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_DataSourceInsert_values;
#else
list = (_TPPLST_0pcodeGen_DataSourceInsert_values) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_DataSourceInsert_values));
#endif
list->_prod = RULELST_0pcodeGen_DataSourceInsert_values;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_DataSourceInsert_values)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_DataSourceValuepcodeGen_DataSourceInsert_values)
	{_TPPLST_codeGen_DataSourceValuepcodeGen_DataSourceInsert_values _currn = _currn;
	 _TERMACT_LST_codeGen_DataSourceValuepcodeGen_DataSourceInsert_values;
	} this code has been moved to _ElempcodeGen_DataSourceInsert_values*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_DataSourceInsert_values;
#else
root = (_TPPpcodeGen_DataSourceInsert_values) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceInsert_values));
#endif
root->_prod = RULEpcodeGen_DataSourceInsert_values;
root->_desc1 = (_TSPLST_codeGen_DataSourceInsert_values)list;
{ _TPPpcodeGen_DataSourceInsert_values _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_DataSourceInsert_values;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceUpdate (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5, NODEPTR _desc6)
#else
NODEPTR MkpcodeGen_DataSourceUpdate (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5,_desc6)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
	NODEPTR _desc6;
#endif
{	_TPPpcodeGen_DataSourceUpdate _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceUpdate;
#else
_currn = (_TPPpcodeGen_DataSourceUpdate) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceUpdate));
#endif
_currn->_prod = RULEpcodeGen_DataSourceUpdate;
_currn->_desc1 = (_TSPcodeGen_DataSourceUpdate_targetDataSource) MkcodeGen_DataSourceUpdate_targetDataSource (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceUpdate: root of subtree no. 1 can not be made a codeGen_DataSourceUpdate_targetDataSource node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_DataSourceUpdate_formalParams) MkcodeGen_DataSourceUpdate_formalParams (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceUpdate: root of subtree no. 2 can not be made a codeGen_DataSourceUpdate_formalParams node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_DataSourceUpdate_name) MkcodeGen_DataSourceUpdate_name (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceUpdate: root of subtree no. 3 can not be made a codeGen_DataSourceUpdate_name node ", 0, _coordref);
_currn->_desc4 = (_TSPcodeGen_DataSourceUpdate_position) MkcodeGen_DataSourceUpdate_position (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceUpdate: root of subtree no. 4 can not be made a codeGen_DataSourceUpdate_position node ", 0, _coordref);
_currn->_desc5 = (_TSPcodeGen_DataSourceUpdate_conditions) MkcodeGen_DataSourceUpdate_conditions (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceUpdate: root of subtree no. 5 can not be made a codeGen_DataSourceUpdate_conditions node ", 0, _coordref);
_currn->_desc6 = (_TSPcodeGen_DataSourceUpdate_values) MkcodeGen_DataSourceUpdate_values (_coordref, _desc6);	
if (((NODEPTR)_currn->_desc6) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceUpdate: root of subtree no. 6 can not be made a codeGen_DataSourceUpdate_values node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceUpdate;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceUpdate */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceUpdate_targetDataSource (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataSourceUpdate_targetDataSource (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataSourceUpdate_targetDataSource _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceUpdate_targetDataSource;
#else
_currn = (_TPPpcodeGen_DataSourceUpdate_targetDataSource) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceUpdate_targetDataSource));
#endif
_currn->_prod = RULEpcodeGen_DataSourceUpdate_targetDataSource;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceUpdate_targetDataSource;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceUpdate_targetDataSource */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_DataSourceFormalParampcodeGen_DataSourceUpdate_formalParams (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_DataSourceFormalParampcodeGen_DataSourceUpdate_formalParams (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_DataSourceFormalParampcodeGen_DataSourceUpdate_formalParams _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_DataSourceFormalParampcodeGen_DataSourceUpdate_formalParams;
#else
	_new = (_TPPLST_codeGen_DataSourceFormalParampcodeGen_DataSourceUpdate_formalParams) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_DataSourceFormalParampcodeGen_DataSourceUpdate_formalParams));
#endif
	_new->_prod = RULELST_codeGen_DataSourceFormalParampcodeGen_DataSourceUpdate_formalParams;
	_new->_desc1 = (_TSPcodeGen_DataSourceFormalParam)elem;
	_new->_desc2 = (_TSPLST_codeGen_DataSourceUpdate_formalParams)_new;
	{_TPPLST_codeGen_DataSourceFormalParampcodeGen_DataSourceUpdate_formalParams _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_DataSourceFormalParampcodeGen_DataSourceUpdate_formalParams;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_DataSourceUpdate_formalParams (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_DataSourceUpdate_formalParams (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_DataSourceFormalParam)) 
		return _MkcodeGen_DataSourceFormalParampcodeGen_DataSourceUpdate_formalParams (_coordref, MkcodeGen_DataSourceFormalParam (_coordref, t));
	message (DEADLY, "a pcodeGen_DataSourceUpdate_formalParams list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_DataSourceUpdate_formalParams (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_DataSourceUpdate_formalParams (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_DataSourceUpdate_formalParams)
	l =_ElempcodeGen_DataSourceUpdate_formalParams (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_DataSourceUpdate_formalParams)
	r = _ElempcodeGen_DataSourceUpdate_formalParams (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_DataSourceUpdate_formalParams)r)->_desc2);
((_TPPLST_0pcodeGen_DataSourceUpdate_formalParams)r)->_desc2 = ((_TPPLST_0pcodeGen_DataSourceUpdate_formalParams)l)->_desc2;
((_TPPLST_0pcodeGen_DataSourceUpdate_formalParams)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceUpdate_formalParams (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_DataSourceUpdate_formalParams (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_DataSourceUpdate_formalParams root;
	_TPPLST_0pcodeGen_DataSourceUpdate_formalParams list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_DataSourceUpdate_formalParams)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_DataSourceUpdate_formalParams)
		t = _ElempcodeGen_DataSourceUpdate_formalParams (_coordref, t);
	remain = (_TPPLST_0pcodeGen_DataSourceUpdate_formalParams)(((_TPPLST_0pcodeGen_DataSourceUpdate_formalParams) t)->_desc2);
	((_TPPLST_0pcodeGen_DataSourceUpdate_formalParams) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_DataSourceUpdate_formalParams;
#else
list = (_TPPLST_0pcodeGen_DataSourceUpdate_formalParams) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_DataSourceUpdate_formalParams));
#endif
list->_prod = RULELST_0pcodeGen_DataSourceUpdate_formalParams;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_DataSourceUpdate_formalParams)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_DataSourceFormalParampcodeGen_DataSourceUpdate_formalParams)
	{_TPPLST_codeGen_DataSourceFormalParampcodeGen_DataSourceUpdate_formalParams _currn = _currn;
	 _TERMACT_LST_codeGen_DataSourceFormalParampcodeGen_DataSourceUpdate_formalParams;
	} this code has been moved to _ElempcodeGen_DataSourceUpdate_formalParams*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_DataSourceUpdate_formalParams;
#else
root = (_TPPpcodeGen_DataSourceUpdate_formalParams) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceUpdate_formalParams));
#endif
root->_prod = RULEpcodeGen_DataSourceUpdate_formalParams;
root->_desc1 = (_TSPLST_codeGen_DataSourceUpdate_formalParams)list;
{ _TPPpcodeGen_DataSourceUpdate_formalParams _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_DataSourceUpdate_formalParams;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceUpdate_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataSourceUpdate_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataSourceUpdate_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceUpdate_name;
#else
_currn = (_TPPpcodeGen_DataSourceUpdate_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceUpdate_name));
#endif
_currn->_prod = RULEpcodeGen_DataSourceUpdate_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceUpdate_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceUpdate_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceUpdate_position (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataSourceUpdate_position (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataSourceUpdate_position _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceUpdate_position;
#else
_currn = (_TPPpcodeGen_DataSourceUpdate_position) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceUpdate_position));
#endif
_currn->_prod = RULEpcodeGen_DataSourceUpdate_position;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceUpdate_position;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceUpdate_position */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_DataSourceValuepcodeGen_DataSourceUpdate_values (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_DataSourceValuepcodeGen_DataSourceUpdate_values (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_DataSourceValuepcodeGen_DataSourceUpdate_values _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_DataSourceValuepcodeGen_DataSourceUpdate_values;
#else
	_new = (_TPPLST_codeGen_DataSourceValuepcodeGen_DataSourceUpdate_values) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_DataSourceValuepcodeGen_DataSourceUpdate_values));
#endif
	_new->_prod = RULELST_codeGen_DataSourceValuepcodeGen_DataSourceUpdate_values;
	_new->_desc1 = (_TSPcodeGen_DataSourceValue)elem;
	_new->_desc2 = (_TSPLST_codeGen_DataSourceUpdate_values)_new;
	{_TPPLST_codeGen_DataSourceValuepcodeGen_DataSourceUpdate_values _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_DataSourceValuepcodeGen_DataSourceUpdate_values;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_DataSourceUpdate_values (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_DataSourceUpdate_values (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_DataSourceValue)) 
		return _MkcodeGen_DataSourceValuepcodeGen_DataSourceUpdate_values (_coordref, MkcodeGen_DataSourceValue (_coordref, t));
	message (DEADLY, "a pcodeGen_DataSourceUpdate_values list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_DataSourceUpdate_values (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_DataSourceUpdate_values (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_DataSourceUpdate_values)
	l =_ElempcodeGen_DataSourceUpdate_values (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_DataSourceUpdate_values)
	r = _ElempcodeGen_DataSourceUpdate_values (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_DataSourceUpdate_values)r)->_desc2);
((_TPPLST_0pcodeGen_DataSourceUpdate_values)r)->_desc2 = ((_TPPLST_0pcodeGen_DataSourceUpdate_values)l)->_desc2;
((_TPPLST_0pcodeGen_DataSourceUpdate_values)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceUpdate_values (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_DataSourceUpdate_values (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_DataSourceUpdate_values root;
	_TPPLST_0pcodeGen_DataSourceUpdate_values list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_DataSourceUpdate_values)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_DataSourceUpdate_values)
		t = _ElempcodeGen_DataSourceUpdate_values (_coordref, t);
	remain = (_TPPLST_0pcodeGen_DataSourceUpdate_values)(((_TPPLST_0pcodeGen_DataSourceUpdate_values) t)->_desc2);
	((_TPPLST_0pcodeGen_DataSourceUpdate_values) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_DataSourceUpdate_values;
#else
list = (_TPPLST_0pcodeGen_DataSourceUpdate_values) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_DataSourceUpdate_values));
#endif
list->_prod = RULELST_0pcodeGen_DataSourceUpdate_values;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_DataSourceUpdate_values)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_DataSourceValuepcodeGen_DataSourceUpdate_values)
	{_TPPLST_codeGen_DataSourceValuepcodeGen_DataSourceUpdate_values _currn = _currn;
	 _TERMACT_LST_codeGen_DataSourceValuepcodeGen_DataSourceUpdate_values;
	} this code has been moved to _ElempcodeGen_DataSourceUpdate_values*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_DataSourceUpdate_values;
#else
root = (_TPPpcodeGen_DataSourceUpdate_values) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceUpdate_values));
#endif
root->_prod = RULEpcodeGen_DataSourceUpdate_values;
root->_desc1 = (_TSPLST_codeGen_DataSourceUpdate_values)list;
{ _TPPpcodeGen_DataSourceUpdate_values _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_DataSourceUpdate_values;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceDelete (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkpcodeGen_DataSourceDelete (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPpcodeGen_DataSourceDelete _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceDelete;
#else
_currn = (_TPPpcodeGen_DataSourceDelete) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceDelete));
#endif
_currn->_prod = RULEpcodeGen_DataSourceDelete;
_currn->_desc1 = (_TSPcodeGen_DataSourceDelete_targetDataSource) MkcodeGen_DataSourceDelete_targetDataSource (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceDelete: root of subtree no. 1 can not be made a codeGen_DataSourceDelete_targetDataSource node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_DataSourceDelete_formalParams) MkcodeGen_DataSourceDelete_formalParams (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceDelete: root of subtree no. 2 can not be made a codeGen_DataSourceDelete_formalParams node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_DataSourceDelete_name) MkcodeGen_DataSourceDelete_name (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceDelete: root of subtree no. 3 can not be made a codeGen_DataSourceDelete_name node ", 0, _coordref);
_currn->_desc4 = (_TSPcodeGen_DataSourceDelete_position) MkcodeGen_DataSourceDelete_position (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceDelete: root of subtree no. 4 can not be made a codeGen_DataSourceDelete_position node ", 0, _coordref);
_currn->_desc5 = (_TSPcodeGen_DataSourceDelete_conditions) MkcodeGen_DataSourceDelete_conditions (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceDelete: root of subtree no. 5 can not be made a codeGen_DataSourceDelete_conditions node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceDelete;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceDelete */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceDelete_targetDataSource (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataSourceDelete_targetDataSource (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataSourceDelete_targetDataSource _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceDelete_targetDataSource;
#else
_currn = (_TPPpcodeGen_DataSourceDelete_targetDataSource) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceDelete_targetDataSource));
#endif
_currn->_prod = RULEpcodeGen_DataSourceDelete_targetDataSource;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceDelete_targetDataSource;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceDelete_targetDataSource */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_DataSourceFormalParampcodeGen_DataSourceDelete_formalParams (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_DataSourceFormalParampcodeGen_DataSourceDelete_formalParams (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_DataSourceFormalParampcodeGen_DataSourceDelete_formalParams _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_DataSourceFormalParampcodeGen_DataSourceDelete_formalParams;
#else
	_new = (_TPPLST_codeGen_DataSourceFormalParampcodeGen_DataSourceDelete_formalParams) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_DataSourceFormalParampcodeGen_DataSourceDelete_formalParams));
#endif
	_new->_prod = RULELST_codeGen_DataSourceFormalParampcodeGen_DataSourceDelete_formalParams;
	_new->_desc1 = (_TSPcodeGen_DataSourceFormalParam)elem;
	_new->_desc2 = (_TSPLST_codeGen_DataSourceDelete_formalParams)_new;
	{_TPPLST_codeGen_DataSourceFormalParampcodeGen_DataSourceDelete_formalParams _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_DataSourceFormalParampcodeGen_DataSourceDelete_formalParams;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_DataSourceDelete_formalParams (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_DataSourceDelete_formalParams (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_DataSourceFormalParam)) 
		return _MkcodeGen_DataSourceFormalParampcodeGen_DataSourceDelete_formalParams (_coordref, MkcodeGen_DataSourceFormalParam (_coordref, t));
	message (DEADLY, "a pcodeGen_DataSourceDelete_formalParams list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_DataSourceDelete_formalParams (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_DataSourceDelete_formalParams (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_DataSourceDelete_formalParams)
	l =_ElempcodeGen_DataSourceDelete_formalParams (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_DataSourceDelete_formalParams)
	r = _ElempcodeGen_DataSourceDelete_formalParams (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_DataSourceDelete_formalParams)r)->_desc2);
((_TPPLST_0pcodeGen_DataSourceDelete_formalParams)r)->_desc2 = ((_TPPLST_0pcodeGen_DataSourceDelete_formalParams)l)->_desc2;
((_TPPLST_0pcodeGen_DataSourceDelete_formalParams)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceDelete_formalParams (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_DataSourceDelete_formalParams (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_DataSourceDelete_formalParams root;
	_TPPLST_0pcodeGen_DataSourceDelete_formalParams list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_DataSourceDelete_formalParams)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_DataSourceDelete_formalParams)
		t = _ElempcodeGen_DataSourceDelete_formalParams (_coordref, t);
	remain = (_TPPLST_0pcodeGen_DataSourceDelete_formalParams)(((_TPPLST_0pcodeGen_DataSourceDelete_formalParams) t)->_desc2);
	((_TPPLST_0pcodeGen_DataSourceDelete_formalParams) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_DataSourceDelete_formalParams;
#else
list = (_TPPLST_0pcodeGen_DataSourceDelete_formalParams) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_DataSourceDelete_formalParams));
#endif
list->_prod = RULELST_0pcodeGen_DataSourceDelete_formalParams;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_DataSourceDelete_formalParams)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_DataSourceFormalParampcodeGen_DataSourceDelete_formalParams)
	{_TPPLST_codeGen_DataSourceFormalParampcodeGen_DataSourceDelete_formalParams _currn = _currn;
	 _TERMACT_LST_codeGen_DataSourceFormalParampcodeGen_DataSourceDelete_formalParams;
	} this code has been moved to _ElempcodeGen_DataSourceDelete_formalParams*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_DataSourceDelete_formalParams;
#else
root = (_TPPpcodeGen_DataSourceDelete_formalParams) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceDelete_formalParams));
#endif
root->_prod = RULEpcodeGen_DataSourceDelete_formalParams;
root->_desc1 = (_TSPLST_codeGen_DataSourceDelete_formalParams)list;
{ _TPPpcodeGen_DataSourceDelete_formalParams _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_DataSourceDelete_formalParams;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceDelete_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataSourceDelete_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataSourceDelete_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceDelete_name;
#else
_currn = (_TPPpcodeGen_DataSourceDelete_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceDelete_name));
#endif
_currn->_prod = RULEpcodeGen_DataSourceDelete_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceDelete_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceDelete_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceDelete_position (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataSourceDelete_position (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataSourceDelete_position _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceDelete_position;
#else
_currn = (_TPPpcodeGen_DataSourceDelete_position) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceDelete_position));
#endif
_currn->_prod = RULEpcodeGen_DataSourceDelete_position;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceDelete_position;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceDelete_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceActualParam (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_DataSourceActualParam (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_DataSourceActualParam _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceActualParam;
#else
_currn = (_TPPpcodeGen_DataSourceActualParam) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceActualParam));
#endif
_currn->_prod = RULEpcodeGen_DataSourceActualParam;
_currn->_desc1 = (_TSPcodeGen_DataSourceActualParam_formalParam) MkcodeGen_DataSourceActualParam_formalParam (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceActualParam: root of subtree no. 1 can not be made a codeGen_DataSourceActualParam_formalParam node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_DataSourceActualParam_expressions) MkcodeGen_DataSourceActualParam_expressions (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceActualParam: root of subtree no. 2 can not be made a codeGen_DataSourceActualParam_expressions node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceActualParam;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceActualParam_formalParam (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataSourceActualParam_formalParam (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataSourceActualParam_formalParam _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceActualParam_formalParam;
#else
_currn = (_TPPpcodeGen_DataSourceActualParam_formalParam) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceActualParam_formalParam));
#endif
_currn->_prod = RULEpcodeGen_DataSourceActualParam_formalParam;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceActualParam_formalParam;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceActualParam_formalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceFormalParam (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_DataSourceFormalParam (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_DataSourceFormalParam _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceFormalParam;
#else
_currn = (_TPPpcodeGen_DataSourceFormalParam) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceFormalParam));
#endif
_currn->_prod = RULEpcodeGen_DataSourceFormalParam;
_currn->_desc1 = (_TSPcodeGen_DataSourceFormalParam_name) MkcodeGen_DataSourceFormalParam_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceFormalParam: root of subtree no. 1 can not be made a codeGen_DataSourceFormalParam_name node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_DataSourceFormalParam_type) MkcodeGen_DataSourceFormalParam_type (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataSourceFormalParam: root of subtree no. 2 can not be made a codeGen_DataSourceFormalParam_type node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceFormalParam;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceFormalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataSourceFormalParam_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataSourceFormalParam_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataSourceFormalParam_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataSourceFormalParam_name;
#else
_currn = (_TPPpcodeGen_DataSourceFormalParam_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataSourceFormalParam_name));
#endif
_currn->_prod = RULEpcodeGen_DataSourceFormalParam_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataSourceFormalParam_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataSourceFormalParam_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_Formular (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkpcodeGen_Formular (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPpcodeGen_Formular _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_Formular;
#else
_currn = (_TPPpcodeGen_Formular) TreeNodeAlloc (sizeof (struct _TPpcodeGen_Formular));
#endif
_currn->_prod = RULEpcodeGen_Formular;
_currn->_desc1 = (_TSPcodeGen_Formular_styleSheetBlock) MkcodeGen_Formular_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_Formular: root of subtree no. 1 can not be made a codeGen_Formular_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_Formular_action) MkcodeGen_Formular_action (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_Formular: root of subtree no. 2 can not be made a codeGen_Formular_action node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_Formular_elements) MkcodeGen_Formular_elements (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_Formular: root of subtree no. 3 can not be made a codeGen_Formular_elements node ", 0, _coordref);
_currn->_desc4 = (_TSPcodeGen_Formular_datasource) MkcodeGen_Formular_datasource (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_Formular: root of subtree no. 4 can not be made a codeGen_Formular_datasource node ", 0, _coordref);
_currn->_desc5 = (_TSPcodeGen_Formular_actualParams) MkcodeGen_Formular_actualParams (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_Formular: root of subtree no. 5 can not be made a codeGen_Formular_actualParams node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_Formular;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_Formular */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_Formular_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_Formular_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_Formular_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_Formular_styleSheetBlock;
#else
_currn = (_TPPpcodeGen_Formular_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpcodeGen_Formular_styleSheetBlock));
#endif
_currn->_prod = RULEpcodeGen_Formular_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_pcodeGen_Formular_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_Formular_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_Formular_action (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_Formular_action (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_Formular_action _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_Formular_action;
#else
_currn = (_TPPpcodeGen_Formular_action) TreeNodeAlloc (sizeof (struct _TPpcodeGen_Formular_action));
#endif
_currn->_prod = RULEpcodeGen_Formular_action;
_SETCOORD(_currn)
_TERMACT_pcodeGen_Formular_action;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_Formular_action */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_FormularRadioGroupElementpcodeGen_Formular_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_FormularRadioGroupElementpcodeGen_Formular_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_FormularRadioGroupElementpcodeGen_Formular_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_FormularRadioGroupElementpcodeGen_Formular_elements;
#else
	_new = (_TPPLST_codeGen_FormularRadioGroupElementpcodeGen_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_FormularRadioGroupElementpcodeGen_Formular_elements));
#endif
	_new->_prod = RULELST_codeGen_FormularRadioGroupElementpcodeGen_Formular_elements;
	_new->_desc1 = (_TSPcodeGen_FormularRadioGroupElement)elem;
	_new->_desc2 = (_TSPLST_codeGen_Formular_elements)_new;
	{_TPPLST_codeGen_FormularRadioGroupElementpcodeGen_Formular_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_FormularRadioGroupElementpcodeGen_Formular_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_FormularTextElementpcodeGen_Formular_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_FormularTextElementpcodeGen_Formular_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_FormularTextElementpcodeGen_Formular_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_FormularTextElementpcodeGen_Formular_elements;
#else
	_new = (_TPPLST_codeGen_FormularTextElementpcodeGen_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_FormularTextElementpcodeGen_Formular_elements));
#endif
	_new->_prod = RULELST_codeGen_FormularTextElementpcodeGen_Formular_elements;
	_new->_desc1 = (_TSPcodeGen_FormularTextElement)elem;
	_new->_desc2 = (_TSPLST_codeGen_Formular_elements)_new;
	{_TPPLST_codeGen_FormularTextElementpcodeGen_Formular_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_FormularTextElementpcodeGen_Formular_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_FormularTextAreaElementpcodeGen_Formular_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_FormularTextAreaElementpcodeGen_Formular_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_FormularTextAreaElementpcodeGen_Formular_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_FormularTextAreaElementpcodeGen_Formular_elements;
#else
	_new = (_TPPLST_codeGen_FormularTextAreaElementpcodeGen_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_FormularTextAreaElementpcodeGen_Formular_elements));
#endif
	_new->_prod = RULELST_codeGen_FormularTextAreaElementpcodeGen_Formular_elements;
	_new->_desc1 = (_TSPcodeGen_FormularTextAreaElement)elem;
	_new->_desc2 = (_TSPLST_codeGen_Formular_elements)_new;
	{_TPPLST_codeGen_FormularTextAreaElementpcodeGen_Formular_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_FormularTextAreaElementpcodeGen_Formular_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_FormularPasswordElementpcodeGen_Formular_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_FormularPasswordElementpcodeGen_Formular_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_FormularPasswordElementpcodeGen_Formular_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_FormularPasswordElementpcodeGen_Formular_elements;
#else
	_new = (_TPPLST_codeGen_FormularPasswordElementpcodeGen_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_FormularPasswordElementpcodeGen_Formular_elements));
#endif
	_new->_prod = RULELST_codeGen_FormularPasswordElementpcodeGen_Formular_elements;
	_new->_desc1 = (_TSPcodeGen_FormularPasswordElement)elem;
	_new->_desc2 = (_TSPLST_codeGen_Formular_elements)_new;
	{_TPPLST_codeGen_FormularPasswordElementpcodeGen_Formular_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_FormularPasswordElementpcodeGen_Formular_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_FormularCheckboxElementpcodeGen_Formular_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_FormularCheckboxElementpcodeGen_Formular_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_FormularCheckboxElementpcodeGen_Formular_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_FormularCheckboxElementpcodeGen_Formular_elements;
#else
	_new = (_TPPLST_codeGen_FormularCheckboxElementpcodeGen_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_FormularCheckboxElementpcodeGen_Formular_elements));
#endif
	_new->_prod = RULELST_codeGen_FormularCheckboxElementpcodeGen_Formular_elements;
	_new->_desc1 = (_TSPcodeGen_FormularCheckboxElement)elem;
	_new->_desc2 = (_TSPLST_codeGen_Formular_elements)_new;
	{_TPPLST_codeGen_FormularCheckboxElementpcodeGen_Formular_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_FormularCheckboxElementpcodeGen_Formular_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_FormularSelectElementpcodeGen_Formular_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_FormularSelectElementpcodeGen_Formular_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_FormularSelectElementpcodeGen_Formular_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_FormularSelectElementpcodeGen_Formular_elements;
#else
	_new = (_TPPLST_codeGen_FormularSelectElementpcodeGen_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_FormularSelectElementpcodeGen_Formular_elements));
#endif
	_new->_prod = RULELST_codeGen_FormularSelectElementpcodeGen_Formular_elements;
	_new->_desc1 = (_TSPcodeGen_FormularSelectElement)elem;
	_new->_desc2 = (_TSPLST_codeGen_Formular_elements)_new;
	{_TPPLST_codeGen_FormularSelectElementpcodeGen_Formular_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_FormularSelectElementpcodeGen_Formular_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_Formular_elements (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_Formular_elements (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_FormularRadioGroupElement)) 
		return _MkcodeGen_FormularRadioGroupElementpcodeGen_Formular_elements (_coordref, MkcodeGen_FormularRadioGroupElement (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_FormularTextElement)) 
		return _MkcodeGen_FormularTextElementpcodeGen_Formular_elements (_coordref, MkcodeGen_FormularTextElement (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_FormularTextAreaElement)) 
		return _MkcodeGen_FormularTextAreaElementpcodeGen_Formular_elements (_coordref, MkcodeGen_FormularTextAreaElement (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_FormularPasswordElement)) 
		return _MkcodeGen_FormularPasswordElementpcodeGen_Formular_elements (_coordref, MkcodeGen_FormularPasswordElement (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_FormularCheckboxElement)) 
		return _MkcodeGen_FormularCheckboxElementpcodeGen_Formular_elements (_coordref, MkcodeGen_FormularCheckboxElement (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_FormularSelectElement)) 
		return _MkcodeGen_FormularSelectElementpcodeGen_Formular_elements (_coordref, MkcodeGen_FormularSelectElement (_coordref, t));
	message (DEADLY, "a pcodeGen_Formular_elements list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_Formular_elements (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_Formular_elements (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_Formular_elements)
	l =_ElempcodeGen_Formular_elements (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_Formular_elements)
	r = _ElempcodeGen_Formular_elements (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_Formular_elements)r)->_desc2);
((_TPPLST_0pcodeGen_Formular_elements)r)->_desc2 = ((_TPPLST_0pcodeGen_Formular_elements)l)->_desc2;
((_TPPLST_0pcodeGen_Formular_elements)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_Formular_elements (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_Formular_elements (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_Formular_elements root;
	_TPPLST_0pcodeGen_Formular_elements list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_Formular_elements)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_Formular_elements)
		t = _ElempcodeGen_Formular_elements (_coordref, t);
	remain = (_TPPLST_0pcodeGen_Formular_elements)(((_TPPLST_0pcodeGen_Formular_elements) t)->_desc2);
	((_TPPLST_0pcodeGen_Formular_elements) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_Formular_elements;
#else
list = (_TPPLST_0pcodeGen_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_Formular_elements));
#endif
list->_prod = RULELST_0pcodeGen_Formular_elements;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_Formular_elements)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_FormularRadioGroupElementpcodeGen_Formular_elements)
	{_TPPLST_codeGen_FormularRadioGroupElementpcodeGen_Formular_elements _currn = _currn;
	 _TERMACT_LST_codeGen_FormularRadioGroupElementpcodeGen_Formular_elements;
	} this code has been moved to _ElempcodeGen_Formular_elements*/
	/*if (_currn->_prod == RULELST_codeGen_FormularTextElementpcodeGen_Formular_elements)
	{_TPPLST_codeGen_FormularTextElementpcodeGen_Formular_elements _currn = _currn;
	 _TERMACT_LST_codeGen_FormularTextElementpcodeGen_Formular_elements;
	} this code has been moved to _ElempcodeGen_Formular_elements*/
	/*if (_currn->_prod == RULELST_codeGen_FormularTextAreaElementpcodeGen_Formular_elements)
	{_TPPLST_codeGen_FormularTextAreaElementpcodeGen_Formular_elements _currn = _currn;
	 _TERMACT_LST_codeGen_FormularTextAreaElementpcodeGen_Formular_elements;
	} this code has been moved to _ElempcodeGen_Formular_elements*/
	/*if (_currn->_prod == RULELST_codeGen_FormularPasswordElementpcodeGen_Formular_elements)
	{_TPPLST_codeGen_FormularPasswordElementpcodeGen_Formular_elements _currn = _currn;
	 _TERMACT_LST_codeGen_FormularPasswordElementpcodeGen_Formular_elements;
	} this code has been moved to _ElempcodeGen_Formular_elements*/
	/*if (_currn->_prod == RULELST_codeGen_FormularCheckboxElementpcodeGen_Formular_elements)
	{_TPPLST_codeGen_FormularCheckboxElementpcodeGen_Formular_elements _currn = _currn;
	 _TERMACT_LST_codeGen_FormularCheckboxElementpcodeGen_Formular_elements;
	} this code has been moved to _ElempcodeGen_Formular_elements*/
	/*if (_currn->_prod == RULELST_codeGen_FormularSelectElementpcodeGen_Formular_elements)
	{_TPPLST_codeGen_FormularSelectElementpcodeGen_Formular_elements _currn = _currn;
	 _TERMACT_LST_codeGen_FormularSelectElementpcodeGen_Formular_elements;
	} this code has been moved to _ElempcodeGen_Formular_elements*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_Formular_elements;
#else
root = (_TPPpcodeGen_Formular_elements) TreeNodeAlloc (sizeof (struct _TPpcodeGen_Formular_elements));
#endif
root->_prod = RULEpcodeGen_Formular_elements;
root->_desc1 = (_TSPLST_codeGen_Formular_elements)list;
{ _TPPpcodeGen_Formular_elements _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_Formular_elements;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_Formular_datasource (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_Formular_datasource (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_Formular_datasource _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_Formular_datasource;
#else
_currn = (_TPPpcodeGen_Formular_datasource) TreeNodeAlloc (sizeof (struct _TPpcodeGen_Formular_datasource));
#endif
_currn->_prod = RULEpcodeGen_Formular_datasource;
_SETCOORD(_currn)
_TERMACT_pcodeGen_Formular_datasource;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_Formular_datasource */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_WebNodeActualParampcodeGen_Formular_actualParams (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_WebNodeActualParampcodeGen_Formular_actualParams (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_WebNodeActualParampcodeGen_Formular_actualParams _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_WebNodeActualParampcodeGen_Formular_actualParams;
#else
	_new = (_TPPLST_codeGen_WebNodeActualParampcodeGen_Formular_actualParams) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_WebNodeActualParampcodeGen_Formular_actualParams));
#endif
	_new->_prod = RULELST_codeGen_WebNodeActualParampcodeGen_Formular_actualParams;
	_new->_desc1 = (_TSPcodeGen_WebNodeActualParam)elem;
	_new->_desc2 = (_TSPLST_codeGen_Formular_actualParams)_new;
	{_TPPLST_codeGen_WebNodeActualParampcodeGen_Formular_actualParams _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_WebNodeActualParampcodeGen_Formular_actualParams;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_Formular_actualParams (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_Formular_actualParams (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_WebNodeActualParam)) 
		return _MkcodeGen_WebNodeActualParampcodeGen_Formular_actualParams (_coordref, MkcodeGen_WebNodeActualParam (_coordref, t));
	message (DEADLY, "a pcodeGen_Formular_actualParams list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_Formular_actualParams (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_Formular_actualParams (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_Formular_actualParams)
	l =_ElempcodeGen_Formular_actualParams (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_Formular_actualParams)
	r = _ElempcodeGen_Formular_actualParams (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_Formular_actualParams)r)->_desc2);
((_TPPLST_0pcodeGen_Formular_actualParams)r)->_desc2 = ((_TPPLST_0pcodeGen_Formular_actualParams)l)->_desc2;
((_TPPLST_0pcodeGen_Formular_actualParams)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_Formular_actualParams (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_Formular_actualParams (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_Formular_actualParams root;
	_TPPLST_0pcodeGen_Formular_actualParams list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_Formular_actualParams)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_Formular_actualParams)
		t = _ElempcodeGen_Formular_actualParams (_coordref, t);
	remain = (_TPPLST_0pcodeGen_Formular_actualParams)(((_TPPLST_0pcodeGen_Formular_actualParams) t)->_desc2);
	((_TPPLST_0pcodeGen_Formular_actualParams) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_Formular_actualParams;
#else
list = (_TPPLST_0pcodeGen_Formular_actualParams) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_Formular_actualParams));
#endif
list->_prod = RULELST_0pcodeGen_Formular_actualParams;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_Formular_actualParams)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_WebNodeActualParampcodeGen_Formular_actualParams)
	{_TPPLST_codeGen_WebNodeActualParampcodeGen_Formular_actualParams _currn = _currn;
	 _TERMACT_LST_codeGen_WebNodeActualParampcodeGen_Formular_actualParams;
	} this code has been moved to _ElempcodeGen_Formular_actualParams*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_Formular_actualParams;
#else
root = (_TPPpcodeGen_Formular_actualParams) TreeNodeAlloc (sizeof (struct _TPpcodeGen_Formular_actualParams));
#endif
root->_prod = RULEpcodeGen_Formular_actualParams;
root->_desc1 = (_TSPLST_codeGen_Formular_actualParams)list;
{ _TPPpcodeGen_Formular_actualParams _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_Formular_actualParams;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularRadioGroupElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5, NODEPTR _desc6, NODEPTR _desc7)
#else
NODEPTR MkpcodeGen_FormularRadioGroupElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5,_desc6,_desc7)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
	NODEPTR _desc6;
	NODEPTR _desc7;
#endif
{	_TPPpcodeGen_FormularRadioGroupElement _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularRadioGroupElement;
#else
_currn = (_TPPpcodeGen_FormularRadioGroupElement) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularRadioGroupElement));
#endif
_currn->_prod = RULEpcodeGen_FormularRadioGroupElement;
_currn->_desc1 = (_TSPcodeGen_FormularRadioGroupElement_name) MkcodeGen_FormularRadioGroupElement_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularRadioGroupElement: root of subtree no. 1 can not be made a codeGen_FormularRadioGroupElement_name node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_FormularRadioGroupElement_label) MkcodeGen_FormularRadioGroupElement_label (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularRadioGroupElement: root of subtree no. 2 can not be made a codeGen_FormularRadioGroupElement_label node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_FormularRadioGroupElement_description) MkcodeGen_FormularRadioGroupElement_description (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularRadioGroupElement: root of subtree no. 3 can not be made a codeGen_FormularRadioGroupElement_description node ", 0, _coordref);
_currn->_desc4 = (_TSPcodeGen_FormularRadioGroupElement_pvalue) MkcodeGen_FormularRadioGroupElement_pvalue (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularRadioGroupElement: root of subtree no. 4 can not be made a codeGen_FormularRadioGroupElement_pvalue node ", 0, _coordref);
_currn->_desc5 = (_TSPcodeGen_FormularRadioGroupElement_condition) MkcodeGen_FormularRadioGroupElement_condition (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularRadioGroupElement: root of subtree no. 5 can not be made a codeGen_FormularRadioGroupElement_condition node ", 0, _coordref);
_currn->_desc6 = (_TSPcodeGen_FormularRadioGroupElement_elements) MkcodeGen_FormularRadioGroupElement_elements (_coordref, _desc6);	
if (((NODEPTR)_currn->_desc6) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularRadioGroupElement: root of subtree no. 6 can not be made a codeGen_FormularRadioGroupElement_elements node ", 0, _coordref);
_currn->_desc7 = (_TSPcodeGen_FormularRadioGroupElement_checked) MkcodeGen_FormularRadioGroupElement_checked (_coordref, _desc7);	
if (((NODEPTR)_currn->_desc7) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularRadioGroupElement: root of subtree no. 7 can not be made a codeGen_FormularRadioGroupElement_checked node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularRadioGroupElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularRadioGroupElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularRadioGroupElement_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularRadioGroupElement_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularRadioGroupElement_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularRadioGroupElement_name;
#else
_currn = (_TPPpcodeGen_FormularRadioGroupElement_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularRadioGroupElement_name));
#endif
_currn->_prod = RULEpcodeGen_FormularRadioGroupElement_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularRadioGroupElement_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularRadioGroupElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularRadioGroupElement_label (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularRadioGroupElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularRadioGroupElement_label _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularRadioGroupElement_label;
#else
_currn = (_TPPpcodeGen_FormularRadioGroupElement_label) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularRadioGroupElement_label));
#endif
_currn->_prod = RULEpcodeGen_FormularRadioGroupElement_label;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularRadioGroupElement_label;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularRadioGroupElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularRadioGroupElement_description (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularRadioGroupElement_description (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularRadioGroupElement_description _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularRadioGroupElement_description;
#else
_currn = (_TPPpcodeGen_FormularRadioGroupElement_description) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularRadioGroupElement_description));
#endif
_currn->_prod = RULEpcodeGen_FormularRadioGroupElement_description;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularRadioGroupElement_description;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularRadioGroupElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularRadioGroupElement_pvalue (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularRadioGroupElement_pvalue (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularRadioGroupElement_pvalue _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularRadioGroupElement_pvalue;
#else
_currn = (_TPPpcodeGen_FormularRadioGroupElement_pvalue) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularRadioGroupElement_pvalue));
#endif
_currn->_prod = RULEpcodeGen_FormularRadioGroupElement_pvalue;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularRadioGroupElement_pvalue;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularRadioGroupElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_FormularRadioElementpcodeGen_FormularRadioGroupElement_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_FormularRadioElementpcodeGen_FormularRadioGroupElement_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_FormularRadioElementpcodeGen_FormularRadioGroupElement_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_FormularRadioElementpcodeGen_FormularRadioGroupElement_elements;
#else
	_new = (_TPPLST_codeGen_FormularRadioElementpcodeGen_FormularRadioGroupElement_elements) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_FormularRadioElementpcodeGen_FormularRadioGroupElement_elements));
#endif
	_new->_prod = RULELST_codeGen_FormularRadioElementpcodeGen_FormularRadioGroupElement_elements;
	_new->_desc1 = (_TSPcodeGen_FormularRadioElement)elem;
	_new->_desc2 = (_TSPLST_codeGen_FormularRadioGroupElement_elements)_new;
	{_TPPLST_codeGen_FormularRadioElementpcodeGen_FormularRadioGroupElement_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_FormularRadioElementpcodeGen_FormularRadioGroupElement_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_FormularRadioGroupElement_elements (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_FormularRadioGroupElement_elements (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_FormularRadioElement)) 
		return _MkcodeGen_FormularRadioElementpcodeGen_FormularRadioGroupElement_elements (_coordref, MkcodeGen_FormularRadioElement (_coordref, t));
	message (DEADLY, "a pcodeGen_FormularRadioGroupElement_elements list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_FormularRadioGroupElement_elements (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_FormularRadioGroupElement_elements (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_FormularRadioGroupElement_elements)
	l =_ElempcodeGen_FormularRadioGroupElement_elements (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_FormularRadioGroupElement_elements)
	r = _ElempcodeGen_FormularRadioGroupElement_elements (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_FormularRadioGroupElement_elements)r)->_desc2);
((_TPPLST_0pcodeGen_FormularRadioGroupElement_elements)r)->_desc2 = ((_TPPLST_0pcodeGen_FormularRadioGroupElement_elements)l)->_desc2;
((_TPPLST_0pcodeGen_FormularRadioGroupElement_elements)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularRadioGroupElement_elements (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_FormularRadioGroupElement_elements (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_FormularRadioGroupElement_elements root;
	_TPPLST_0pcodeGen_FormularRadioGroupElement_elements list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_FormularRadioGroupElement_elements)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_FormularRadioGroupElement_elements)
		t = _ElempcodeGen_FormularRadioGroupElement_elements (_coordref, t);
	remain = (_TPPLST_0pcodeGen_FormularRadioGroupElement_elements)(((_TPPLST_0pcodeGen_FormularRadioGroupElement_elements) t)->_desc2);
	((_TPPLST_0pcodeGen_FormularRadioGroupElement_elements) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_FormularRadioGroupElement_elements;
#else
list = (_TPPLST_0pcodeGen_FormularRadioGroupElement_elements) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_FormularRadioGroupElement_elements));
#endif
list->_prod = RULELST_0pcodeGen_FormularRadioGroupElement_elements;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_FormularRadioGroupElement_elements)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_FormularRadioElementpcodeGen_FormularRadioGroupElement_elements)
	{_TPPLST_codeGen_FormularRadioElementpcodeGen_FormularRadioGroupElement_elements _currn = _currn;
	 _TERMACT_LST_codeGen_FormularRadioElementpcodeGen_FormularRadioGroupElement_elements;
	} this code has been moved to _ElempcodeGen_FormularRadioGroupElement_elements*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_FormularRadioGroupElement_elements;
#else
root = (_TPPpcodeGen_FormularRadioGroupElement_elements) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularRadioGroupElement_elements));
#endif
root->_prod = RULEpcodeGen_FormularRadioGroupElement_elements;
root->_desc1 = (_TSPLST_codeGen_FormularRadioGroupElement_elements)list;
{ _TPPpcodeGen_FormularRadioGroupElement_elements _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_FormularRadioGroupElement_elements;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularRadioGroupElement_checked (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularRadioGroupElement_checked (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularRadioGroupElement_checked _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularRadioGroupElement_checked;
#else
_currn = (_TPPpcodeGen_FormularRadioGroupElement_checked) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularRadioGroupElement_checked));
#endif
_currn->_prod = RULEpcodeGen_FormularRadioGroupElement_checked;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularRadioGroupElement_checked;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularRadioGroupElement_checked */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularTextElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkpcodeGen_FormularTextElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPpcodeGen_FormularTextElement _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularTextElement;
#else
_currn = (_TPPpcodeGen_FormularTextElement) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularTextElement));
#endif
_currn->_prod = RULEpcodeGen_FormularTextElement;
_currn->_desc1 = (_TSPcodeGen_FormularTextElement_name) MkcodeGen_FormularTextElement_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularTextElement: root of subtree no. 1 can not be made a codeGen_FormularTextElement_name node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_FormularTextElement_label) MkcodeGen_FormularTextElement_label (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularTextElement: root of subtree no. 2 can not be made a codeGen_FormularTextElement_label node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_FormularTextElement_description) MkcodeGen_FormularTextElement_description (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularTextElement: root of subtree no. 3 can not be made a codeGen_FormularTextElement_description node ", 0, _coordref);
_currn->_desc4 = (_TSPcodeGen_FormularTextElement_pvalue) MkcodeGen_FormularTextElement_pvalue (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularTextElement: root of subtree no. 4 can not be made a codeGen_FormularTextElement_pvalue node ", 0, _coordref);
_currn->_desc5 = (_TSPcodeGen_FormularTextElement_condition) MkcodeGen_FormularTextElement_condition (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularTextElement: root of subtree no. 5 can not be made a codeGen_FormularTextElement_condition node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularTextElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularTextElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularTextElement_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularTextElement_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularTextElement_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularTextElement_name;
#else
_currn = (_TPPpcodeGen_FormularTextElement_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularTextElement_name));
#endif
_currn->_prod = RULEpcodeGen_FormularTextElement_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularTextElement_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularTextElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularTextElement_label (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularTextElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularTextElement_label _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularTextElement_label;
#else
_currn = (_TPPpcodeGen_FormularTextElement_label) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularTextElement_label));
#endif
_currn->_prod = RULEpcodeGen_FormularTextElement_label;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularTextElement_label;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularTextElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularTextElement_description (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularTextElement_description (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularTextElement_description _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularTextElement_description;
#else
_currn = (_TPPpcodeGen_FormularTextElement_description) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularTextElement_description));
#endif
_currn->_prod = RULEpcodeGen_FormularTextElement_description;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularTextElement_description;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularTextElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularTextElement_pvalue (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularTextElement_pvalue (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularTextElement_pvalue _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularTextElement_pvalue;
#else
_currn = (_TPPpcodeGen_FormularTextElement_pvalue) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularTextElement_pvalue));
#endif
_currn->_prod = RULEpcodeGen_FormularTextElement_pvalue;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularTextElement_pvalue;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularTextElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularTextAreaElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkpcodeGen_FormularTextAreaElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPpcodeGen_FormularTextAreaElement _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularTextAreaElement;
#else
_currn = (_TPPpcodeGen_FormularTextAreaElement) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularTextAreaElement));
#endif
_currn->_prod = RULEpcodeGen_FormularTextAreaElement;
_currn->_desc1 = (_TSPcodeGen_FormularTextAreaElement_name) MkcodeGen_FormularTextAreaElement_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularTextAreaElement: root of subtree no. 1 can not be made a codeGen_FormularTextAreaElement_name node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_FormularTextAreaElement_label) MkcodeGen_FormularTextAreaElement_label (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularTextAreaElement: root of subtree no. 2 can not be made a codeGen_FormularTextAreaElement_label node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_FormularTextAreaElement_description) MkcodeGen_FormularTextAreaElement_description (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularTextAreaElement: root of subtree no. 3 can not be made a codeGen_FormularTextAreaElement_description node ", 0, _coordref);
_currn->_desc4 = (_TSPcodeGen_FormularTextAreaElement_pvalue) MkcodeGen_FormularTextAreaElement_pvalue (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularTextAreaElement: root of subtree no. 4 can not be made a codeGen_FormularTextAreaElement_pvalue node ", 0, _coordref);
_currn->_desc5 = (_TSPcodeGen_FormularTextAreaElement_condition) MkcodeGen_FormularTextAreaElement_condition (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularTextAreaElement: root of subtree no. 5 can not be made a codeGen_FormularTextAreaElement_condition node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularTextAreaElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularTextAreaElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularTextAreaElement_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularTextAreaElement_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularTextAreaElement_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularTextAreaElement_name;
#else
_currn = (_TPPpcodeGen_FormularTextAreaElement_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularTextAreaElement_name));
#endif
_currn->_prod = RULEpcodeGen_FormularTextAreaElement_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularTextAreaElement_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularTextAreaElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularTextAreaElement_label (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularTextAreaElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularTextAreaElement_label _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularTextAreaElement_label;
#else
_currn = (_TPPpcodeGen_FormularTextAreaElement_label) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularTextAreaElement_label));
#endif
_currn->_prod = RULEpcodeGen_FormularTextAreaElement_label;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularTextAreaElement_label;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularTextAreaElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularTextAreaElement_description (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularTextAreaElement_description (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularTextAreaElement_description _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularTextAreaElement_description;
#else
_currn = (_TPPpcodeGen_FormularTextAreaElement_description) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularTextAreaElement_description));
#endif
_currn->_prod = RULEpcodeGen_FormularTextAreaElement_description;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularTextAreaElement_description;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularTextAreaElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularTextAreaElement_pvalue (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularTextAreaElement_pvalue (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularTextAreaElement_pvalue _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularTextAreaElement_pvalue;
#else
_currn = (_TPPpcodeGen_FormularTextAreaElement_pvalue) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularTextAreaElement_pvalue));
#endif
_currn->_prod = RULEpcodeGen_FormularTextAreaElement_pvalue;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularTextAreaElement_pvalue;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularTextAreaElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularPasswordElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkpcodeGen_FormularPasswordElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPpcodeGen_FormularPasswordElement _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularPasswordElement;
#else
_currn = (_TPPpcodeGen_FormularPasswordElement) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularPasswordElement));
#endif
_currn->_prod = RULEpcodeGen_FormularPasswordElement;
_currn->_desc1 = (_TSPcodeGen_FormularPasswordElement_name) MkcodeGen_FormularPasswordElement_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularPasswordElement: root of subtree no. 1 can not be made a codeGen_FormularPasswordElement_name node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_FormularPasswordElement_label) MkcodeGen_FormularPasswordElement_label (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularPasswordElement: root of subtree no. 2 can not be made a codeGen_FormularPasswordElement_label node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_FormularPasswordElement_description) MkcodeGen_FormularPasswordElement_description (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularPasswordElement: root of subtree no. 3 can not be made a codeGen_FormularPasswordElement_description node ", 0, _coordref);
_currn->_desc4 = (_TSPcodeGen_FormularPasswordElement_pvalue) MkcodeGen_FormularPasswordElement_pvalue (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularPasswordElement: root of subtree no. 4 can not be made a codeGen_FormularPasswordElement_pvalue node ", 0, _coordref);
_currn->_desc5 = (_TSPcodeGen_FormularPasswordElement_condition) MkcodeGen_FormularPasswordElement_condition (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularPasswordElement: root of subtree no. 5 can not be made a codeGen_FormularPasswordElement_condition node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularPasswordElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularPasswordElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularPasswordElement_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularPasswordElement_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularPasswordElement_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularPasswordElement_name;
#else
_currn = (_TPPpcodeGen_FormularPasswordElement_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularPasswordElement_name));
#endif
_currn->_prod = RULEpcodeGen_FormularPasswordElement_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularPasswordElement_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularPasswordElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularPasswordElement_label (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularPasswordElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularPasswordElement_label _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularPasswordElement_label;
#else
_currn = (_TPPpcodeGen_FormularPasswordElement_label) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularPasswordElement_label));
#endif
_currn->_prod = RULEpcodeGen_FormularPasswordElement_label;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularPasswordElement_label;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularPasswordElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularPasswordElement_description (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularPasswordElement_description (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularPasswordElement_description _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularPasswordElement_description;
#else
_currn = (_TPPpcodeGen_FormularPasswordElement_description) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularPasswordElement_description));
#endif
_currn->_prod = RULEpcodeGen_FormularPasswordElement_description;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularPasswordElement_description;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularPasswordElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularPasswordElement_pvalue (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularPasswordElement_pvalue (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularPasswordElement_pvalue _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularPasswordElement_pvalue;
#else
_currn = (_TPPpcodeGen_FormularPasswordElement_pvalue) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularPasswordElement_pvalue));
#endif
_currn->_prod = RULEpcodeGen_FormularPasswordElement_pvalue;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularPasswordElement_pvalue;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularPasswordElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularCheckboxElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5, NODEPTR _desc6, NODEPTR _desc7)
#else
NODEPTR MkpcodeGen_FormularCheckboxElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5,_desc6,_desc7)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
	NODEPTR _desc6;
	NODEPTR _desc7;
#endif
{	_TPPpcodeGen_FormularCheckboxElement _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularCheckboxElement;
#else
_currn = (_TPPpcodeGen_FormularCheckboxElement) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularCheckboxElement));
#endif
_currn->_prod = RULEpcodeGen_FormularCheckboxElement;
_currn->_desc1 = (_TSPcodeGen_FormularCheckboxElement_name) MkcodeGen_FormularCheckboxElement_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularCheckboxElement: root of subtree no. 1 can not be made a codeGen_FormularCheckboxElement_name node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_FormularCheckboxElement_label) MkcodeGen_FormularCheckboxElement_label (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularCheckboxElement: root of subtree no. 2 can not be made a codeGen_FormularCheckboxElement_label node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_FormularCheckboxElement_description) MkcodeGen_FormularCheckboxElement_description (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularCheckboxElement: root of subtree no. 3 can not be made a codeGen_FormularCheckboxElement_description node ", 0, _coordref);
_currn->_desc4 = (_TSPcodeGen_FormularCheckboxElement_pvalue) MkcodeGen_FormularCheckboxElement_pvalue (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularCheckboxElement: root of subtree no. 4 can not be made a codeGen_FormularCheckboxElement_pvalue node ", 0, _coordref);
_currn->_desc5 = (_TSPcodeGen_FormularCheckboxElement_condition) MkcodeGen_FormularCheckboxElement_condition (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularCheckboxElement: root of subtree no. 5 can not be made a codeGen_FormularCheckboxElement_condition node ", 0, _coordref);
_currn->_desc6 = (_TSPcodeGen_FormularCheckboxElement_value) MkcodeGen_FormularCheckboxElement_value (_coordref, _desc6);	
if (((NODEPTR)_currn->_desc6) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularCheckboxElement: root of subtree no. 6 can not be made a codeGen_FormularCheckboxElement_value node ", 0, _coordref);
_currn->_desc7 = (_TSPcodeGen_FormularCheckboxElement_checked) MkcodeGen_FormularCheckboxElement_checked (_coordref, _desc7);	
if (((NODEPTR)_currn->_desc7) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularCheckboxElement: root of subtree no. 7 can not be made a codeGen_FormularCheckboxElement_checked node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularCheckboxElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularCheckboxElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularCheckboxElement_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularCheckboxElement_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularCheckboxElement_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularCheckboxElement_name;
#else
_currn = (_TPPpcodeGen_FormularCheckboxElement_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularCheckboxElement_name));
#endif
_currn->_prod = RULEpcodeGen_FormularCheckboxElement_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularCheckboxElement_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularCheckboxElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularCheckboxElement_label (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularCheckboxElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularCheckboxElement_label _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularCheckboxElement_label;
#else
_currn = (_TPPpcodeGen_FormularCheckboxElement_label) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularCheckboxElement_label));
#endif
_currn->_prod = RULEpcodeGen_FormularCheckboxElement_label;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularCheckboxElement_label;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularCheckboxElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularCheckboxElement_description (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularCheckboxElement_description (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularCheckboxElement_description _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularCheckboxElement_description;
#else
_currn = (_TPPpcodeGen_FormularCheckboxElement_description) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularCheckboxElement_description));
#endif
_currn->_prod = RULEpcodeGen_FormularCheckboxElement_description;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularCheckboxElement_description;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularCheckboxElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularCheckboxElement_pvalue (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularCheckboxElement_pvalue (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularCheckboxElement_pvalue _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularCheckboxElement_pvalue;
#else
_currn = (_TPPpcodeGen_FormularCheckboxElement_pvalue) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularCheckboxElement_pvalue));
#endif
_currn->_prod = RULEpcodeGen_FormularCheckboxElement_pvalue;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularCheckboxElement_pvalue;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularCheckboxElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularCheckboxElement_value (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularCheckboxElement_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularCheckboxElement_value _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularCheckboxElement_value;
#else
_currn = (_TPPpcodeGen_FormularCheckboxElement_value) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularCheckboxElement_value));
#endif
_currn->_prod = RULEpcodeGen_FormularCheckboxElement_value;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularCheckboxElement_value;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularCheckboxElement_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularCheckboxElement_checked (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularCheckboxElement_checked (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularCheckboxElement_checked _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularCheckboxElement_checked;
#else
_currn = (_TPPpcodeGen_FormularCheckboxElement_checked) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularCheckboxElement_checked));
#endif
_currn->_prod = RULEpcodeGen_FormularCheckboxElement_checked;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularCheckboxElement_checked;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularCheckboxElement_checked */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularSelectElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5, NODEPTR _desc6, NODEPTR _desc7, NODEPTR _desc8)
#else
NODEPTR MkpcodeGen_FormularSelectElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5,_desc6,_desc7,_desc8)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
	NODEPTR _desc6;
	NODEPTR _desc7;
	NODEPTR _desc8;
#endif
{	_TPPpcodeGen_FormularSelectElement _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularSelectElement;
#else
_currn = (_TPPpcodeGen_FormularSelectElement) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularSelectElement));
#endif
_currn->_prod = RULEpcodeGen_FormularSelectElement;
_currn->_desc1 = (_TSPcodeGen_FormularSelectElement_name) MkcodeGen_FormularSelectElement_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularSelectElement: root of subtree no. 1 can not be made a codeGen_FormularSelectElement_name node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_FormularSelectElement_label) MkcodeGen_FormularSelectElement_label (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularSelectElement: root of subtree no. 2 can not be made a codeGen_FormularSelectElement_label node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_FormularSelectElement_description) MkcodeGen_FormularSelectElement_description (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularSelectElement: root of subtree no. 3 can not be made a codeGen_FormularSelectElement_description node ", 0, _coordref);
_currn->_desc4 = (_TSPcodeGen_FormularSelectElement_pvalue) MkcodeGen_FormularSelectElement_pvalue (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularSelectElement: root of subtree no. 4 can not be made a codeGen_FormularSelectElement_pvalue node ", 0, _coordref);
_currn->_desc5 = (_TSPcodeGen_FormularSelectElement_condition) MkcodeGen_FormularSelectElement_condition (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularSelectElement: root of subtree no. 5 can not be made a codeGen_FormularSelectElement_condition node ", 0, _coordref);
_currn->_desc6 = (_TSPcodeGen_FormularSelectElement_options) MkcodeGen_FormularSelectElement_options (_coordref, _desc6);	
if (((NODEPTR)_currn->_desc6) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularSelectElement: root of subtree no. 6 can not be made a codeGen_FormularSelectElement_options node ", 0, _coordref);
_currn->_desc7 = (_TSPcodeGen_FormularSelectElement_size) MkcodeGen_FormularSelectElement_size (_coordref, _desc7);	
if (((NODEPTR)_currn->_desc7) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularSelectElement: root of subtree no. 7 can not be made a codeGen_FormularSelectElement_size node ", 0, _coordref);
_currn->_desc8 = (_TSPcodeGen_FormularSelectElement_multiple) MkcodeGen_FormularSelectElement_multiple (_coordref, _desc8);	
if (((NODEPTR)_currn->_desc8) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularSelectElement: root of subtree no. 8 can not be made a codeGen_FormularSelectElement_multiple node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularSelectElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularSelectElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularSelectElement_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularSelectElement_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularSelectElement_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularSelectElement_name;
#else
_currn = (_TPPpcodeGen_FormularSelectElement_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularSelectElement_name));
#endif
_currn->_prod = RULEpcodeGen_FormularSelectElement_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularSelectElement_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularSelectElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularSelectElement_label (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularSelectElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularSelectElement_label _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularSelectElement_label;
#else
_currn = (_TPPpcodeGen_FormularSelectElement_label) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularSelectElement_label));
#endif
_currn->_prod = RULEpcodeGen_FormularSelectElement_label;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularSelectElement_label;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularSelectElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularSelectElement_description (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularSelectElement_description (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularSelectElement_description _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularSelectElement_description;
#else
_currn = (_TPPpcodeGen_FormularSelectElement_description) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularSelectElement_description));
#endif
_currn->_prod = RULEpcodeGen_FormularSelectElement_description;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularSelectElement_description;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularSelectElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularSelectElement_pvalue (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularSelectElement_pvalue (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularSelectElement_pvalue _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularSelectElement_pvalue;
#else
_currn = (_TPPpcodeGen_FormularSelectElement_pvalue) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularSelectElement_pvalue));
#endif
_currn->_prod = RULEpcodeGen_FormularSelectElement_pvalue;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularSelectElement_pvalue;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularSelectElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularSelectElement_size (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularSelectElement_size (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularSelectElement_size _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularSelectElement_size;
#else
_currn = (_TPPpcodeGen_FormularSelectElement_size) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularSelectElement_size));
#endif
_currn->_prod = RULEpcodeGen_FormularSelectElement_size;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularSelectElement_size;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularSelectElement_size */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularSelectElement_multiple (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularSelectElement_multiple (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularSelectElement_multiple _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularSelectElement_multiple;
#else
_currn = (_TPPpcodeGen_FormularSelectElement_multiple) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularSelectElement_multiple));
#endif
_currn->_prod = RULEpcodeGen_FormularSelectElement_multiple;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularSelectElement_multiple;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularSelectElement_multiple */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularConstantSelectOption (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpcodeGen_FormularConstantSelectOption (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpcodeGen_FormularConstantSelectOption _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularConstantSelectOption;
#else
_currn = (_TPPpcodeGen_FormularConstantSelectOption) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularConstantSelectOption));
#endif
_currn->_prod = RULEpcodeGen_FormularConstantSelectOption;
_currn->_desc1 = (_TSPcodeGen_FormularConstantSelectOption_values) MkcodeGen_FormularConstantSelectOption_values (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularConstantSelectOption: root of subtree no. 1 can not be made a codeGen_FormularConstantSelectOption_values node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularConstantSelectOption;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularConstantSelectOption */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularConstantSelectOption_values (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularConstantSelectOption_values (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularConstantSelectOption_values _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularConstantSelectOption_values;
#else
_currn = (_TPPpcodeGen_FormularConstantSelectOption_values) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularConstantSelectOption_values));
#endif
_currn->_prod = RULEpcodeGen_FormularConstantSelectOption_values;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularConstantSelectOption_values;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularConstantSelectOption_values */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularDataSourceSelectOption (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_FormularDataSourceSelectOption (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_FormularDataSourceSelectOption _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularDataSourceSelectOption;
#else
_currn = (_TPPpcodeGen_FormularDataSourceSelectOption) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularDataSourceSelectOption));
#endif
_currn->_prod = RULEpcodeGen_FormularDataSourceSelectOption;
_currn->_desc1 = (_TSPcodeGen_FormularDataSourceSelectOption_values) MkcodeGen_FormularDataSourceSelectOption_values (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularDataSourceSelectOption: root of subtree no. 1 can not be made a codeGen_FormularDataSourceSelectOption_values node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_FormularDataSourceSelectOption_params) MkcodeGen_FormularDataSourceSelectOption_params (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularDataSourceSelectOption: root of subtree no. 2 can not be made a codeGen_FormularDataSourceSelectOption_params node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularDataSourceSelectOption;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularDataSourceSelectOption */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularDataSourceSelectOption_values (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularDataSourceSelectOption_values (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularDataSourceSelectOption_values _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularDataSourceSelectOption_values;
#else
_currn = (_TPPpcodeGen_FormularDataSourceSelectOption_values) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularDataSourceSelectOption_values));
#endif
_currn->_prod = RULEpcodeGen_FormularDataSourceSelectOption_values;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularDataSourceSelectOption_values;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularDataSourceSelectOption_values */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_DataSourceActualParampcodeGen_FormularDataSourceSelectOption_params (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_DataSourceActualParampcodeGen_FormularDataSourceSelectOption_params (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_DataSourceActualParampcodeGen_FormularDataSourceSelectOption_params _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_DataSourceActualParampcodeGen_FormularDataSourceSelectOption_params;
#else
	_new = (_TPPLST_codeGen_DataSourceActualParampcodeGen_FormularDataSourceSelectOption_params) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_DataSourceActualParampcodeGen_FormularDataSourceSelectOption_params));
#endif
	_new->_prod = RULELST_codeGen_DataSourceActualParampcodeGen_FormularDataSourceSelectOption_params;
	_new->_desc1 = (_TSPcodeGen_DataSourceActualParam)elem;
	_new->_desc2 = (_TSPLST_codeGen_FormularDataSourceSelectOption_params)_new;
	{_TPPLST_codeGen_DataSourceActualParampcodeGen_FormularDataSourceSelectOption_params _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_DataSourceActualParampcodeGen_FormularDataSourceSelectOption_params;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_FormularDataSourceSelectOption_params (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_FormularDataSourceSelectOption_params (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_DataSourceActualParam)) 
		return _MkcodeGen_DataSourceActualParampcodeGen_FormularDataSourceSelectOption_params (_coordref, MkcodeGen_DataSourceActualParam (_coordref, t));
	message (DEADLY, "a pcodeGen_FormularDataSourceSelectOption_params list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_FormularDataSourceSelectOption_params (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_FormularDataSourceSelectOption_params (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_FormularDataSourceSelectOption_params)
	l =_ElempcodeGen_FormularDataSourceSelectOption_params (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_FormularDataSourceSelectOption_params)
	r = _ElempcodeGen_FormularDataSourceSelectOption_params (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_FormularDataSourceSelectOption_params)r)->_desc2);
((_TPPLST_0pcodeGen_FormularDataSourceSelectOption_params)r)->_desc2 = ((_TPPLST_0pcodeGen_FormularDataSourceSelectOption_params)l)->_desc2;
((_TPPLST_0pcodeGen_FormularDataSourceSelectOption_params)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularDataSourceSelectOption_params (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_FormularDataSourceSelectOption_params (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_FormularDataSourceSelectOption_params root;
	_TPPLST_0pcodeGen_FormularDataSourceSelectOption_params list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_FormularDataSourceSelectOption_params)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_FormularDataSourceSelectOption_params)
		t = _ElempcodeGen_FormularDataSourceSelectOption_params (_coordref, t);
	remain = (_TPPLST_0pcodeGen_FormularDataSourceSelectOption_params)(((_TPPLST_0pcodeGen_FormularDataSourceSelectOption_params) t)->_desc2);
	((_TPPLST_0pcodeGen_FormularDataSourceSelectOption_params) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_FormularDataSourceSelectOption_params;
#else
list = (_TPPLST_0pcodeGen_FormularDataSourceSelectOption_params) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_FormularDataSourceSelectOption_params));
#endif
list->_prod = RULELST_0pcodeGen_FormularDataSourceSelectOption_params;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_FormularDataSourceSelectOption_params)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_DataSourceActualParampcodeGen_FormularDataSourceSelectOption_params)
	{_TPPLST_codeGen_DataSourceActualParampcodeGen_FormularDataSourceSelectOption_params _currn = _currn;
	 _TERMACT_LST_codeGen_DataSourceActualParampcodeGen_FormularDataSourceSelectOption_params;
	} this code has been moved to _ElempcodeGen_FormularDataSourceSelectOption_params*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_FormularDataSourceSelectOption_params;
#else
root = (_TPPpcodeGen_FormularDataSourceSelectOption_params) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularDataSourceSelectOption_params));
#endif
root->_prod = RULEpcodeGen_FormularDataSourceSelectOption_params;
root->_desc1 = (_TSPLST_codeGen_FormularDataSourceSelectOption_params)list;
{ _TPPpcodeGen_FormularDataSourceSelectOption_params _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_FormularDataSourceSelectOption_params;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularPHPScriptSelectOption (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpcodeGen_FormularPHPScriptSelectOption (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpcodeGen_FormularPHPScriptSelectOption _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularPHPScriptSelectOption;
#else
_currn = (_TPPpcodeGen_FormularPHPScriptSelectOption) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularPHPScriptSelectOption));
#endif
_currn->_prod = RULEpcodeGen_FormularPHPScriptSelectOption;
_currn->_desc1 = (_TSPcodeGen_FormularPHPScriptSelectOption_script) MkcodeGen_FormularPHPScriptSelectOption_script (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularPHPScriptSelectOption: root of subtree no. 1 can not be made a codeGen_FormularPHPScriptSelectOption_script node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularPHPScriptSelectOption;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularPHPScriptSelectOption */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularPHPScriptSelectOption_script (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularPHPScriptSelectOption_script (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularPHPScriptSelectOption_script _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularPHPScriptSelectOption_script;
#else
_currn = (_TPPpcodeGen_FormularPHPScriptSelectOption_script) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularPHPScriptSelectOption_script));
#endif
_currn->_prod = RULEpcodeGen_FormularPHPScriptSelectOption_script;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularPHPScriptSelectOption_script;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularPHPScriptSelectOption_script */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularRadioElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpcodeGen_FormularRadioElement (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpcodeGen_FormularRadioElement _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularRadioElement;
#else
_currn = (_TPPpcodeGen_FormularRadioElement) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularRadioElement));
#endif
_currn->_prod = RULEpcodeGen_FormularRadioElement;
_currn->_desc1 = (_TSPcodeGen_FormularRadioElement_value) MkcodeGen_FormularRadioElement_value (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularRadioElement: root of subtree no. 1 can not be made a codeGen_FormularRadioElement_value node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_FormularRadioElement_label) MkcodeGen_FormularRadioElement_label (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularRadioElement: root of subtree no. 2 can not be made a codeGen_FormularRadioElement_label node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_FormularRadioElement_name) MkcodeGen_FormularRadioElement_name (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularRadioElement: root of subtree no. 3 can not be made a codeGen_FormularRadioElement_name node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularRadioElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularRadioElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularRadioElement_value (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularRadioElement_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularRadioElement_value _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularRadioElement_value;
#else
_currn = (_TPPpcodeGen_FormularRadioElement_value) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularRadioElement_value));
#endif
_currn->_prod = RULEpcodeGen_FormularRadioElement_value;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularRadioElement_value;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularRadioElement_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularRadioElement_label (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularRadioElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularRadioElement_label _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularRadioElement_label;
#else
_currn = (_TPPpcodeGen_FormularRadioElement_label) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularRadioElement_label));
#endif
_currn->_prod = RULEpcodeGen_FormularRadioElement_label;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularRadioElement_label;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularRadioElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularRadioElement_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularRadioElement_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularRadioElement_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularRadioElement_name;
#else
_currn = (_TPPpcodeGen_FormularRadioElement_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularRadioElement_name));
#endif
_currn->_prod = RULEpcodeGen_FormularRadioElement_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularRadioElement_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularRadioElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularCondition (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_FormularCondition (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_FormularCondition _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularCondition;
#else
_currn = (_TPPpcodeGen_FormularCondition) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularCondition));
#endif
_currn->_prod = RULEpcodeGen_FormularCondition;
_currn->_desc1 = (_TSPcodeGen_FormularCondition_condition) MkcodeGen_FormularCondition_condition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularCondition: root of subtree no. 1 can not be made a codeGen_FormularCondition_condition node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_FormularCondition_params) MkcodeGen_FormularCondition_params (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularCondition: root of subtree no. 2 can not be made a codeGen_FormularCondition_params node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularCondition;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularCondition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularCondition_condition (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularCondition_condition (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularCondition_condition _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularCondition_condition;
#else
_currn = (_TPPpcodeGen_FormularCondition_condition) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularCondition_condition));
#endif
_currn->_prod = RULEpcodeGen_FormularCondition_condition;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularCondition_condition;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularCondition_condition */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_FormularConditionActualParampcodeGen_FormularCondition_params (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_FormularConditionActualParampcodeGen_FormularCondition_params (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_FormularConditionActualParampcodeGen_FormularCondition_params _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_FormularConditionActualParampcodeGen_FormularCondition_params;
#else
	_new = (_TPPLST_codeGen_FormularConditionActualParampcodeGen_FormularCondition_params) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_FormularConditionActualParampcodeGen_FormularCondition_params));
#endif
	_new->_prod = RULELST_codeGen_FormularConditionActualParampcodeGen_FormularCondition_params;
	_new->_desc1 = (_TSPcodeGen_FormularConditionActualParam)elem;
	_new->_desc2 = (_TSPLST_codeGen_FormularCondition_params)_new;
	{_TPPLST_codeGen_FormularConditionActualParampcodeGen_FormularCondition_params _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_FormularConditionActualParampcodeGen_FormularCondition_params;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_FormularCondition_params (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_FormularCondition_params (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_FormularConditionActualParam)) 
		return _MkcodeGen_FormularConditionActualParampcodeGen_FormularCondition_params (_coordref, MkcodeGen_FormularConditionActualParam (_coordref, t));
	message (DEADLY, "a pcodeGen_FormularCondition_params list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_FormularCondition_params (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_FormularCondition_params (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_FormularCondition_params)
	l =_ElempcodeGen_FormularCondition_params (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_FormularCondition_params)
	r = _ElempcodeGen_FormularCondition_params (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_FormularCondition_params)r)->_desc2);
((_TPPLST_0pcodeGen_FormularCondition_params)r)->_desc2 = ((_TPPLST_0pcodeGen_FormularCondition_params)l)->_desc2;
((_TPPLST_0pcodeGen_FormularCondition_params)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularCondition_params (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_FormularCondition_params (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_FormularCondition_params root;
	_TPPLST_0pcodeGen_FormularCondition_params list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_FormularCondition_params)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_FormularCondition_params)
		t = _ElempcodeGen_FormularCondition_params (_coordref, t);
	remain = (_TPPLST_0pcodeGen_FormularCondition_params)(((_TPPLST_0pcodeGen_FormularCondition_params) t)->_desc2);
	((_TPPLST_0pcodeGen_FormularCondition_params) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_FormularCondition_params;
#else
list = (_TPPLST_0pcodeGen_FormularCondition_params) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_FormularCondition_params));
#endif
list->_prod = RULELST_0pcodeGen_FormularCondition_params;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_FormularCondition_params)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_FormularConditionActualParampcodeGen_FormularCondition_params)
	{_TPPLST_codeGen_FormularConditionActualParampcodeGen_FormularCondition_params _currn = _currn;
	 _TERMACT_LST_codeGen_FormularConditionActualParampcodeGen_FormularCondition_params;
	} this code has been moved to _ElempcodeGen_FormularCondition_params*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_FormularCondition_params;
#else
root = (_TPPpcodeGen_FormularCondition_params) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularCondition_params));
#endif
root->_prod = RULEpcodeGen_FormularCondition_params;
root->_desc1 = (_TSPLST_codeGen_FormularCondition_params)list;
{ _TPPpcodeGen_FormularCondition_params _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_FormularCondition_params;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularConditionDefinition (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpcodeGen_FormularConditionDefinition (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpcodeGen_FormularConditionDefinition _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularConditionDefinition;
#else
_currn = (_TPPpcodeGen_FormularConditionDefinition) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularConditionDefinition));
#endif
_currn->_prod = RULEpcodeGen_FormularConditionDefinition;
_currn->_desc1 = (_TSPcodeGen_FormularConditionDefinition_name) MkcodeGen_FormularConditionDefinition_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularConditionDefinition: root of subtree no. 1 can not be made a codeGen_FormularConditionDefinition_name node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_FormularConditionDefinition_params) MkcodeGen_FormularConditionDefinition_params (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularConditionDefinition: root of subtree no. 2 can not be made a codeGen_FormularConditionDefinition_params node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_FormularConditionDefinition_body) MkcodeGen_FormularConditionDefinition_body (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularConditionDefinition: root of subtree no. 3 can not be made a codeGen_FormularConditionDefinition_body node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularConditionDefinition;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularConditionDefinition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularConditionDefinition_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularConditionDefinition_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularConditionDefinition_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularConditionDefinition_name;
#else
_currn = (_TPPpcodeGen_FormularConditionDefinition_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularConditionDefinition_name));
#endif
_currn->_prod = RULEpcodeGen_FormularConditionDefinition_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularConditionDefinition_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularConditionDefinition_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_FormularConditionFormalParampcodeGen_FormularConditionDefinition_params (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_FormularConditionFormalParampcodeGen_FormularConditionDefinition_params (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_FormularConditionFormalParampcodeGen_FormularConditionDefinition_params _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_FormularConditionFormalParampcodeGen_FormularConditionDefinition_params;
#else
	_new = (_TPPLST_codeGen_FormularConditionFormalParampcodeGen_FormularConditionDefinition_params) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_FormularConditionFormalParampcodeGen_FormularConditionDefinition_params));
#endif
	_new->_prod = RULELST_codeGen_FormularConditionFormalParampcodeGen_FormularConditionDefinition_params;
	_new->_desc1 = (_TSPcodeGen_FormularConditionFormalParam)elem;
	_new->_desc2 = (_TSPLST_codeGen_FormularConditionDefinition_params)_new;
	{_TPPLST_codeGen_FormularConditionFormalParampcodeGen_FormularConditionDefinition_params _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_FormularConditionFormalParampcodeGen_FormularConditionDefinition_params;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_FormularConditionDefinition_params (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_FormularConditionDefinition_params (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_FormularConditionFormalParam)) 
		return _MkcodeGen_FormularConditionFormalParampcodeGen_FormularConditionDefinition_params (_coordref, MkcodeGen_FormularConditionFormalParam (_coordref, t));
	message (DEADLY, "a pcodeGen_FormularConditionDefinition_params list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_FormularConditionDefinition_params (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_FormularConditionDefinition_params (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_FormularConditionDefinition_params)
	l =_ElempcodeGen_FormularConditionDefinition_params (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_FormularConditionDefinition_params)
	r = _ElempcodeGen_FormularConditionDefinition_params (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_FormularConditionDefinition_params)r)->_desc2);
((_TPPLST_0pcodeGen_FormularConditionDefinition_params)r)->_desc2 = ((_TPPLST_0pcodeGen_FormularConditionDefinition_params)l)->_desc2;
((_TPPLST_0pcodeGen_FormularConditionDefinition_params)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularConditionDefinition_params (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_FormularConditionDefinition_params (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_FormularConditionDefinition_params root;
	_TPPLST_0pcodeGen_FormularConditionDefinition_params list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_FormularConditionDefinition_params)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_FormularConditionDefinition_params)
		t = _ElempcodeGen_FormularConditionDefinition_params (_coordref, t);
	remain = (_TPPLST_0pcodeGen_FormularConditionDefinition_params)(((_TPPLST_0pcodeGen_FormularConditionDefinition_params) t)->_desc2);
	((_TPPLST_0pcodeGen_FormularConditionDefinition_params) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_FormularConditionDefinition_params;
#else
list = (_TPPLST_0pcodeGen_FormularConditionDefinition_params) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_FormularConditionDefinition_params));
#endif
list->_prod = RULELST_0pcodeGen_FormularConditionDefinition_params;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_FormularConditionDefinition_params)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_FormularConditionFormalParampcodeGen_FormularConditionDefinition_params)
	{_TPPLST_codeGen_FormularConditionFormalParampcodeGen_FormularConditionDefinition_params _currn = _currn;
	 _TERMACT_LST_codeGen_FormularConditionFormalParampcodeGen_FormularConditionDefinition_params;
	} this code has been moved to _ElempcodeGen_FormularConditionDefinition_params*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_FormularConditionDefinition_params;
#else
root = (_TPPpcodeGen_FormularConditionDefinition_params) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularConditionDefinition_params));
#endif
root->_prod = RULEpcodeGen_FormularConditionDefinition_params;
root->_desc1 = (_TSPLST_codeGen_FormularConditionDefinition_params)list;
{ _TPPpcodeGen_FormularConditionDefinition_params _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_FormularConditionDefinition_params;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularConditionDefinition_body (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularConditionDefinition_body (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularConditionDefinition_body _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularConditionDefinition_body;
#else
_currn = (_TPPpcodeGen_FormularConditionDefinition_body) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularConditionDefinition_body));
#endif
_currn->_prod = RULEpcodeGen_FormularConditionDefinition_body;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularConditionDefinition_body;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularConditionDefinition_body */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularConditionFormalParam (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpcodeGen_FormularConditionFormalParam (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpcodeGen_FormularConditionFormalParam _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularConditionFormalParam;
#else
_currn = (_TPPpcodeGen_FormularConditionFormalParam) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularConditionFormalParam));
#endif
_currn->_prod = RULEpcodeGen_FormularConditionFormalParam;
_currn->_desc1 = (_TSPcodeGen_FormularConditionFormalParam_name) MkcodeGen_FormularConditionFormalParam_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularConditionFormalParam: root of subtree no. 1 can not be made a codeGen_FormularConditionFormalParam_name node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularConditionFormalParam;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularConditionFormalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularConditionFormalParam_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularConditionFormalParam_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularConditionFormalParam_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularConditionFormalParam_name;
#else
_currn = (_TPPpcodeGen_FormularConditionFormalParam_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularConditionFormalParam_name));
#endif
_currn->_prod = RULEpcodeGen_FormularConditionFormalParam_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularConditionFormalParam_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularConditionFormalParam_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularConditionActualParam (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_FormularConditionActualParam (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_FormularConditionActualParam _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularConditionActualParam;
#else
_currn = (_TPPpcodeGen_FormularConditionActualParam) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularConditionActualParam));
#endif
_currn->_prod = RULEpcodeGen_FormularConditionActualParam;
_currn->_desc1 = (_TSPcodeGen_FormularConditionActualParam_value) MkcodeGen_FormularConditionActualParam_value (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularConditionActualParam: root of subtree no. 1 can not be made a codeGen_FormularConditionActualParam_value node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_FormularConditionActualParam_formalparam) MkcodeGen_FormularConditionActualParam_formalparam (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FormularConditionActualParam: root of subtree no. 2 can not be made a codeGen_FormularConditionActualParam_formalparam node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularConditionActualParam;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularConditionActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularConditionActualParam_value (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularConditionActualParam_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularConditionActualParam_value _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularConditionActualParam_value;
#else
_currn = (_TPPpcodeGen_FormularConditionActualParam_value) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularConditionActualParam_value));
#endif
_currn->_prod = RULEpcodeGen_FormularConditionActualParam_value;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularConditionActualParam_value;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularConditionActualParam_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FormularConditionActualParam_formalparam (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FormularConditionActualParam_formalparam (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FormularConditionActualParam_formalparam _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FormularConditionActualParam_formalparam;
#else
_currn = (_TPPpcodeGen_FormularConditionActualParam_formalparam) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FormularConditionActualParam_formalparam));
#endif
_currn->_prod = RULEpcodeGen_FormularConditionActualParam_formalparam;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FormularConditionActualParam_formalparam;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FormularConditionActualParam_formalparam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebNodeActualParam (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_WebNodeActualParam (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_WebNodeActualParam _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_WebNodeActualParam;
#else
_currn = (_TPPpcodeGen_WebNodeActualParam) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebNodeActualParam));
#endif
_currn->_prod = RULEpcodeGen_WebNodeActualParam;
_currn->_desc1 = (_TSPcodeGen_WebNodeActualParam_formalParam) MkcodeGen_WebNodeActualParam_formalParam (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebNodeActualParam: root of subtree no. 1 can not be made a codeGen_WebNodeActualParam_formalParam node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_WebNodeActualParam_expressions) MkcodeGen_WebNodeActualParam_expressions (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebNodeActualParam: root of subtree no. 2 can not be made a codeGen_WebNodeActualParam_expressions node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_WebNodeActualParam;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_WebNodeActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebNodeActualParam_formalParam (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_WebNodeActualParam_formalParam (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_WebNodeActualParam_formalParam _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_WebNodeActualParam_formalParam;
#else
_currn = (_TPPpcodeGen_WebNodeActualParam_formalParam) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebNodeActualParam_formalParam));
#endif
_currn->_prod = RULEpcodeGen_WebNodeActualParam_formalParam;
_SETCOORD(_currn)
_TERMACT_pcodeGen_WebNodeActualParam_formalParam;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_WebNodeActualParam_formalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebNodeFormalParam (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_WebNodeFormalParam (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_WebNodeFormalParam _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_WebNodeFormalParam;
#else
_currn = (_TPPpcodeGen_WebNodeFormalParam) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebNodeFormalParam));
#endif
_currn->_prod = RULEpcodeGen_WebNodeFormalParam;
_currn->_desc1 = (_TSPcodeGen_WebNodeFormalParam_name) MkcodeGen_WebNodeFormalParam_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebNodeFormalParam: root of subtree no. 1 can not be made a codeGen_WebNodeFormalParam_name node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_WebNodeFormalParam_type) MkcodeGen_WebNodeFormalParam_type (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebNodeFormalParam: root of subtree no. 2 can not be made a codeGen_WebNodeFormalParam_type node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_WebNodeFormalParam;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_WebNodeFormalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebNodeFormalParam_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_WebNodeFormalParam_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_WebNodeFormalParam_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_WebNodeFormalParam_name;
#else
_currn = (_TPPpcodeGen_WebNodeFormalParam_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebNodeFormalParam_name));
#endif
_currn->_prod = RULEpcodeGen_WebNodeFormalParam_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_WebNodeFormalParam_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_WebNodeFormalParam_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebNodeConstantExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpcodeGen_WebNodeConstantExpression (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpcodeGen_WebNodeConstantExpression _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_WebNodeConstantExpression;
#else
_currn = (_TPPpcodeGen_WebNodeConstantExpression) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebNodeConstantExpression));
#endif
_currn->_prod = RULEpcodeGen_WebNodeConstantExpression;
_currn->_desc1 = (_TSPcodeGen_WebNodeConstantExpression_value) MkcodeGen_WebNodeConstantExpression_value (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebNodeConstantExpression: root of subtree no. 1 can not be made a codeGen_WebNodeConstantExpression_value node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_WebNodeConstantExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_WebNodeConstantExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebNodeConstantExpression_value (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_WebNodeConstantExpression_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_WebNodeConstantExpression_value _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_WebNodeConstantExpression_value;
#else
_currn = (_TPPpcodeGen_WebNodeConstantExpression_value) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebNodeConstantExpression_value));
#endif
_currn->_prod = RULEpcodeGen_WebNodeConstantExpression_value;
_SETCOORD(_currn)
_TERMACT_pcodeGen_WebNodeConstantExpression_value;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_WebNodeConstantExpression_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebNodeVariableExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpcodeGen_WebNodeVariableExpression (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpcodeGen_WebNodeVariableExpression _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_WebNodeVariableExpression;
#else
_currn = (_TPPpcodeGen_WebNodeVariableExpression) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebNodeVariableExpression));
#endif
_currn->_prod = RULEpcodeGen_WebNodeVariableExpression;
_currn->_desc1 = (_TSPcodeGen_WebNodeVariableExpression_content) MkcodeGen_WebNodeVariableExpression_content (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebNodeVariableExpression: root of subtree no. 1 can not be made a codeGen_WebNodeVariableExpression_content node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_WebNodeVariableExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_WebNodeVariableExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebNodeVariableExpression_content (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_WebNodeVariableExpression_content (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_WebNodeVariableExpression_content _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_WebNodeVariableExpression_content;
#else
_currn = (_TPPpcodeGen_WebNodeVariableExpression_content) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebNodeVariableExpression_content));
#endif
_currn->_prod = RULEpcodeGen_WebNodeVariableExpression_content;
_SETCOORD(_currn)
_TERMACT_pcodeGen_WebNodeVariableExpression_content;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_WebNodeVariableExpression_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebNodeColumnExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpcodeGen_WebNodeColumnExpression (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpcodeGen_WebNodeColumnExpression _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_WebNodeColumnExpression;
#else
_currn = (_TPPpcodeGen_WebNodeColumnExpression) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebNodeColumnExpression));
#endif
_currn->_prod = RULEpcodeGen_WebNodeColumnExpression;
_currn->_desc1 = (_TSPcodeGen_WebNodeColumnExpression_content) MkcodeGen_WebNodeColumnExpression_content (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebNodeColumnExpression: root of subtree no. 1 can not be made a codeGen_WebNodeColumnExpression_content node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_WebNodeColumnExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_WebNodeColumnExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebNodeColumnExpression_content (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_WebNodeColumnExpression_content (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_WebNodeColumnExpression_content _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_WebNodeColumnExpression_content;
#else
_currn = (_TPPpcodeGen_WebNodeColumnExpression_content) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebNodeColumnExpression_content));
#endif
_currn->_prod = RULEpcodeGen_WebNodeColumnExpression_content;
_SETCOORD(_currn)
_TERMACT_pcodeGen_WebNodeColumnExpression_content;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_WebNodeColumnExpression_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_Paragraph (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_Paragraph (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_Paragraph _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_Paragraph;
#else
_currn = (_TPPpcodeGen_Paragraph) TreeNodeAlloc (sizeof (struct _TPpcodeGen_Paragraph));
#endif
_currn->_prod = RULEpcodeGen_Paragraph;
_currn->_desc1 = (_TSPcodeGen_Paragraph_styleSheetBlock) MkcodeGen_Paragraph_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_Paragraph: root of subtree no. 1 can not be made a codeGen_Paragraph_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_Paragraph_fragments) MkcodeGen_Paragraph_fragments (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_Paragraph: root of subtree no. 2 can not be made a codeGen_Paragraph_fragments node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_Paragraph;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_Paragraph */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_Paragraph_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_Paragraph_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_Paragraph_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_Paragraph_styleSheetBlock;
#else
_currn = (_TPPpcodeGen_Paragraph_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpcodeGen_Paragraph_styleSheetBlock));
#endif
_currn->_prod = RULEpcodeGen_Paragraph_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_pcodeGen_Paragraph_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_Paragraph_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_TextFragmentpcodeGen_Paragraph_fragments (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_TextFragmentpcodeGen_Paragraph_fragments (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_TextFragmentpcodeGen_Paragraph_fragments _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_TextFragmentpcodeGen_Paragraph_fragments;
#else
	_new = (_TPPLST_codeGen_TextFragmentpcodeGen_Paragraph_fragments) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_TextFragmentpcodeGen_Paragraph_fragments));
#endif
	_new->_prod = RULELST_codeGen_TextFragmentpcodeGen_Paragraph_fragments;
	_new->_desc1 = (_TSPcodeGen_TextFragment)elem;
	_new->_desc2 = (_TSPLST_codeGen_Paragraph_fragments)_new;
	{_TPPLST_codeGen_TextFragmentpcodeGen_Paragraph_fragments _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_TextFragmentpcodeGen_Paragraph_fragments;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_ImageFragmentpcodeGen_Paragraph_fragments (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_ImageFragmentpcodeGen_Paragraph_fragments (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_ImageFragmentpcodeGen_Paragraph_fragments _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_ImageFragmentpcodeGen_Paragraph_fragments;
#else
	_new = (_TPPLST_codeGen_ImageFragmentpcodeGen_Paragraph_fragments) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_ImageFragmentpcodeGen_Paragraph_fragments));
#endif
	_new->_prod = RULELST_codeGen_ImageFragmentpcodeGen_Paragraph_fragments;
	_new->_desc1 = (_TSPcodeGen_ImageFragment)elem;
	_new->_desc2 = (_TSPLST_codeGen_Paragraph_fragments)_new;
	{_TPPLST_codeGen_ImageFragmentpcodeGen_Paragraph_fragments _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_ImageFragmentpcodeGen_Paragraph_fragments;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_OutputElementpcodeGen_Paragraph_fragments (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_OutputElementpcodeGen_Paragraph_fragments (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_OutputElementpcodeGen_Paragraph_fragments _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_OutputElementpcodeGen_Paragraph_fragments;
#else
	_new = (_TPPLST_codeGen_OutputElementpcodeGen_Paragraph_fragments) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_OutputElementpcodeGen_Paragraph_fragments));
#endif
	_new->_prod = RULELST_codeGen_OutputElementpcodeGen_Paragraph_fragments;
	_new->_desc1 = (_TSPcodeGen_OutputElement)elem;
	_new->_desc2 = (_TSPLST_codeGen_Paragraph_fragments)_new;
	{_TPPLST_codeGen_OutputElementpcodeGen_Paragraph_fragments _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_OutputElementpcodeGen_Paragraph_fragments;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_Paragraph_fragments (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_Paragraph_fragments (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_TextFragment)) 
		return _MkcodeGen_TextFragmentpcodeGen_Paragraph_fragments (_coordref, MkcodeGen_TextFragment (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_ImageFragment)) 
		return _MkcodeGen_ImageFragmentpcodeGen_Paragraph_fragments (_coordref, MkcodeGen_ImageFragment (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_OutputElement)) 
		return _MkcodeGen_OutputElementpcodeGen_Paragraph_fragments (_coordref, MkcodeGen_OutputElement (_coordref, t));
	message (DEADLY, "a pcodeGen_Paragraph_fragments list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_Paragraph_fragments (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_Paragraph_fragments (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_Paragraph_fragments)
	l =_ElempcodeGen_Paragraph_fragments (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_Paragraph_fragments)
	r = _ElempcodeGen_Paragraph_fragments (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_Paragraph_fragments)r)->_desc2);
((_TPPLST_0pcodeGen_Paragraph_fragments)r)->_desc2 = ((_TPPLST_0pcodeGen_Paragraph_fragments)l)->_desc2;
((_TPPLST_0pcodeGen_Paragraph_fragments)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_Paragraph_fragments (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_Paragraph_fragments (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_Paragraph_fragments root;
	_TPPLST_0pcodeGen_Paragraph_fragments list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_Paragraph_fragments)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_Paragraph_fragments)
		t = _ElempcodeGen_Paragraph_fragments (_coordref, t);
	remain = (_TPPLST_0pcodeGen_Paragraph_fragments)(((_TPPLST_0pcodeGen_Paragraph_fragments) t)->_desc2);
	((_TPPLST_0pcodeGen_Paragraph_fragments) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_Paragraph_fragments;
#else
list = (_TPPLST_0pcodeGen_Paragraph_fragments) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_Paragraph_fragments));
#endif
list->_prod = RULELST_0pcodeGen_Paragraph_fragments;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_Paragraph_fragments)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_TextFragmentpcodeGen_Paragraph_fragments)
	{_TPPLST_codeGen_TextFragmentpcodeGen_Paragraph_fragments _currn = _currn;
	 _TERMACT_LST_codeGen_TextFragmentpcodeGen_Paragraph_fragments;
	} this code has been moved to _ElempcodeGen_Paragraph_fragments*/
	/*if (_currn->_prod == RULELST_codeGen_ImageFragmentpcodeGen_Paragraph_fragments)
	{_TPPLST_codeGen_ImageFragmentpcodeGen_Paragraph_fragments _currn = _currn;
	 _TERMACT_LST_codeGen_ImageFragmentpcodeGen_Paragraph_fragments;
	} this code has been moved to _ElempcodeGen_Paragraph_fragments*/
	/*if (_currn->_prod == RULELST_codeGen_OutputElementpcodeGen_Paragraph_fragments)
	{_TPPLST_codeGen_OutputElementpcodeGen_Paragraph_fragments _currn = _currn;
	 _TERMACT_LST_codeGen_OutputElementpcodeGen_Paragraph_fragments;
	} this code has been moved to _ElempcodeGen_Paragraph_fragments*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_Paragraph_fragments;
#else
root = (_TPPpcodeGen_Paragraph_fragments) TreeNodeAlloc (sizeof (struct _TPpcodeGen_Paragraph_fragments));
#endif
root->_prod = RULEpcodeGen_Paragraph_fragments;
root->_desc1 = (_TSPLST_codeGen_Paragraph_fragments)list;
{ _TPPpcodeGen_Paragraph_fragments _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_Paragraph_fragments;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_TextFragment (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_TextFragment (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_TextFragment _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_TextFragment;
#else
_currn = (_TPPpcodeGen_TextFragment) TreeNodeAlloc (sizeof (struct _TPpcodeGen_TextFragment));
#endif
_currn->_prod = RULEpcodeGen_TextFragment;
_currn->_desc1 = (_TSPcodeGen_TextFragment_link) MkcodeGen_TextFragment_link (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_TextFragment: root of subtree no. 1 can not be made a codeGen_TextFragment_link node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_TextFragment_value) MkcodeGen_TextFragment_value (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_TextFragment: root of subtree no. 2 can not be made a codeGen_TextFragment_value node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_TextFragment;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_TextFragment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_TextFragment_link (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_TextFragment_link (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_TextFragment_link _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_TextFragment_link;
#else
_currn = (_TPPpcodeGen_TextFragment_link) TreeNodeAlloc (sizeof (struct _TPpcodeGen_TextFragment_link));
#endif
_currn->_prod = RULEpcodeGen_TextFragment_link;
_SETCOORD(_currn)
_TERMACT_pcodeGen_TextFragment_link;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_TextFragment_link */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_TextFragment_value (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_TextFragment_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_TextFragment_value _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_TextFragment_value;
#else
_currn = (_TPPpcodeGen_TextFragment_value) TreeNodeAlloc (sizeof (struct _TPpcodeGen_TextFragment_value));
#endif
_currn->_prod = RULEpcodeGen_TextFragment_value;
_SETCOORD(_currn)
_TERMACT_pcodeGen_TextFragment_value;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_TextFragment_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ImageFragment (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkpcodeGen_ImageFragment (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPpcodeGen_ImageFragment _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ImageFragment;
#else
_currn = (_TPPpcodeGen_ImageFragment) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ImageFragment));
#endif
_currn->_prod = RULEpcodeGen_ImageFragment;
_currn->_desc1 = (_TSPcodeGen_ImageFragment_link) MkcodeGen_ImageFragment_link (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_ImageFragment: root of subtree no. 1 can not be made a codeGen_ImageFragment_link node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_ImageFragment_width) MkcodeGen_ImageFragment_width (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_ImageFragment: root of subtree no. 2 can not be made a codeGen_ImageFragment_width node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_ImageFragment_height) MkcodeGen_ImageFragment_height (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_ImageFragment: root of subtree no. 3 can not be made a codeGen_ImageFragment_height node ", 0, _coordref);
_currn->_desc4 = (_TSPcodeGen_ImageFragment_text) MkcodeGen_ImageFragment_text (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_ImageFragment: root of subtree no. 4 can not be made a codeGen_ImageFragment_text node ", 0, _coordref);
_currn->_desc5 = (_TSPcodeGen_ImageFragment_imageName) MkcodeGen_ImageFragment_imageName (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_ImageFragment: root of subtree no. 5 can not be made a codeGen_ImageFragment_imageName node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_ImageFragment;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_ImageFragment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ImageFragment_link (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_ImageFragment_link (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_ImageFragment_link _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ImageFragment_link;
#else
_currn = (_TPPpcodeGen_ImageFragment_link) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ImageFragment_link));
#endif
_currn->_prod = RULEpcodeGen_ImageFragment_link;
_SETCOORD(_currn)
_TERMACT_pcodeGen_ImageFragment_link;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_ImageFragment_link */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ImageFragment_width (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_ImageFragment_width (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_ImageFragment_width _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ImageFragment_width;
#else
_currn = (_TPPpcodeGen_ImageFragment_width) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ImageFragment_width));
#endif
_currn->_prod = RULEpcodeGen_ImageFragment_width;
_SETCOORD(_currn)
_TERMACT_pcodeGen_ImageFragment_width;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_ImageFragment_width */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ImageFragment_height (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_ImageFragment_height (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_ImageFragment_height _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ImageFragment_height;
#else
_currn = (_TPPpcodeGen_ImageFragment_height) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ImageFragment_height));
#endif
_currn->_prod = RULEpcodeGen_ImageFragment_height;
_SETCOORD(_currn)
_TERMACT_pcodeGen_ImageFragment_height;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_ImageFragment_height */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ImageFragment_text (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_ImageFragment_text (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_ImageFragment_text _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ImageFragment_text;
#else
_currn = (_TPPpcodeGen_ImageFragment_text) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ImageFragment_text));
#endif
_currn->_prod = RULEpcodeGen_ImageFragment_text;
_SETCOORD(_currn)
_TERMACT_pcodeGen_ImageFragment_text;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_ImageFragment_text */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ImageFragment_imageName (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_ImageFragment_imageName (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_ImageFragment_imageName _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ImageFragment_imageName;
#else
_currn = (_TPPpcodeGen_ImageFragment_imageName) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ImageFragment_imageName));
#endif
_currn->_prod = RULEpcodeGen_ImageFragment_imageName;
_SETCOORD(_currn)
_TERMACT_pcodeGen_ImageFragment_imageName;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_ImageFragment_imageName */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_Heading (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpcodeGen_Heading (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpcodeGen_Heading _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_Heading;
#else
_currn = (_TPPpcodeGen_Heading) TreeNodeAlloc (sizeof (struct _TPpcodeGen_Heading));
#endif
_currn->_prod = RULEpcodeGen_Heading;
_currn->_desc1 = (_TSPcodeGen_Heading_styleSheetBlock) MkcodeGen_Heading_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_Heading: root of subtree no. 1 can not be made a codeGen_Heading_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_Heading_text) MkcodeGen_Heading_text (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_Heading: root of subtree no. 2 can not be made a codeGen_Heading_text node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_Heading_level) MkcodeGen_Heading_level (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_Heading: root of subtree no. 3 can not be made a codeGen_Heading_level node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_Heading;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_Heading */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_Heading_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_Heading_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_Heading_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_Heading_styleSheetBlock;
#else
_currn = (_TPPpcodeGen_Heading_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpcodeGen_Heading_styleSheetBlock));
#endif
_currn->_prod = RULEpcodeGen_Heading_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_pcodeGen_Heading_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_Heading_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_Heading_text (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_Heading_text (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_Heading_text _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_Heading_text;
#else
_currn = (_TPPpcodeGen_Heading_text) TreeNodeAlloc (sizeof (struct _TPpcodeGen_Heading_text));
#endif
_currn->_prod = RULEpcodeGen_Heading_text;
_SETCOORD(_currn)
_TERMACT_pcodeGen_Heading_text;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_Heading_text */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_Heading_level (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_Heading_level (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_Heading_level _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_Heading_level;
#else
_currn = (_TPPpcodeGen_Heading_level) TreeNodeAlloc (sizeof (struct _TPpcodeGen_Heading_level));
#endif
_currn->_prod = RULEpcodeGen_Heading_level;
_SETCOORD(_currn)
_TERMACT_pcodeGen_Heading_level;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_Heading_level */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_Table (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpcodeGen_Table (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpcodeGen_Table _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_Table;
#else
_currn = (_TPPpcodeGen_Table) TreeNodeAlloc (sizeof (struct _TPpcodeGen_Table));
#endif
_currn->_prod = RULEpcodeGen_Table;
_currn->_desc1 = (_TSPcodeGen_Table_styleSheetBlock) MkcodeGen_Table_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_Table: root of subtree no. 1 can not be made a codeGen_Table_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_Table_rows) MkcodeGen_Table_rows (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_Table: root of subtree no. 2 can not be made a codeGen_Table_rows node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_Table_columns) MkcodeGen_Table_columns (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_Table: root of subtree no. 3 can not be made a codeGen_Table_columns node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_Table;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_Table */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_Table_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_Table_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_Table_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_Table_styleSheetBlock;
#else
_currn = (_TPPpcodeGen_Table_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpcodeGen_Table_styleSheetBlock));
#endif
_currn->_prod = RULEpcodeGen_Table_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_pcodeGen_Table_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_Table_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_TableRowpcodeGen_Table_rows (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_TableRowpcodeGen_Table_rows (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_TableRowpcodeGen_Table_rows _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_TableRowpcodeGen_Table_rows;
#else
	_new = (_TPPLST_codeGen_TableRowpcodeGen_Table_rows) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_TableRowpcodeGen_Table_rows));
#endif
	_new->_prod = RULELST_codeGen_TableRowpcodeGen_Table_rows;
	_new->_desc1 = (_TSPcodeGen_TableRow)elem;
	_new->_desc2 = (_TSPLST_codeGen_Table_rows)_new;
	{_TPPLST_codeGen_TableRowpcodeGen_Table_rows _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_TableRowpcodeGen_Table_rows;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_Table_rows (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_Table_rows (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_TableRow)) 
		return _MkcodeGen_TableRowpcodeGen_Table_rows (_coordref, MkcodeGen_TableRow (_coordref, t));
	message (DEADLY, "a pcodeGen_Table_rows list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_Table_rows (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_Table_rows (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_Table_rows)
	l =_ElempcodeGen_Table_rows (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_Table_rows)
	r = _ElempcodeGen_Table_rows (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_Table_rows)r)->_desc2);
((_TPPLST_0pcodeGen_Table_rows)r)->_desc2 = ((_TPPLST_0pcodeGen_Table_rows)l)->_desc2;
((_TPPLST_0pcodeGen_Table_rows)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_Table_rows (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_Table_rows (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_Table_rows root;
	_TPPLST_0pcodeGen_Table_rows list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_Table_rows)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_Table_rows)
		t = _ElempcodeGen_Table_rows (_coordref, t);
	remain = (_TPPLST_0pcodeGen_Table_rows)(((_TPPLST_0pcodeGen_Table_rows) t)->_desc2);
	((_TPPLST_0pcodeGen_Table_rows) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_Table_rows;
#else
list = (_TPPLST_0pcodeGen_Table_rows) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_Table_rows));
#endif
list->_prod = RULELST_0pcodeGen_Table_rows;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_Table_rows)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_TableRowpcodeGen_Table_rows)
	{_TPPLST_codeGen_TableRowpcodeGen_Table_rows _currn = _currn;
	 _TERMACT_LST_codeGen_TableRowpcodeGen_Table_rows;
	} this code has been moved to _ElempcodeGen_Table_rows*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_Table_rows;
#else
root = (_TPPpcodeGen_Table_rows) TreeNodeAlloc (sizeof (struct _TPpcodeGen_Table_rows));
#endif
root->_prod = RULEpcodeGen_Table_rows;
root->_desc1 = (_TSPLST_codeGen_Table_rows)list;
{ _TPPpcodeGen_Table_rows _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_Table_rows;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_TableColumnpcodeGen_Table_columns (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_TableColumnpcodeGen_Table_columns (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_TableColumnpcodeGen_Table_columns _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_TableColumnpcodeGen_Table_columns;
#else
	_new = (_TPPLST_codeGen_TableColumnpcodeGen_Table_columns) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_TableColumnpcodeGen_Table_columns));
#endif
	_new->_prod = RULELST_codeGen_TableColumnpcodeGen_Table_columns;
	_new->_desc1 = (_TSPcodeGen_TableColumn)elem;
	_new->_desc2 = (_TSPLST_codeGen_Table_columns)_new;
	{_TPPLST_codeGen_TableColumnpcodeGen_Table_columns _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_TableColumnpcodeGen_Table_columns;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_Table_columns (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_Table_columns (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_TableColumn)) 
		return _MkcodeGen_TableColumnpcodeGen_Table_columns (_coordref, MkcodeGen_TableColumn (_coordref, t));
	message (DEADLY, "a pcodeGen_Table_columns list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_Table_columns (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_Table_columns (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_Table_columns)
	l =_ElempcodeGen_Table_columns (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_Table_columns)
	r = _ElempcodeGen_Table_columns (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_Table_columns)r)->_desc2);
((_TPPLST_0pcodeGen_Table_columns)r)->_desc2 = ((_TPPLST_0pcodeGen_Table_columns)l)->_desc2;
((_TPPLST_0pcodeGen_Table_columns)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_Table_columns (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_Table_columns (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_Table_columns root;
	_TPPLST_0pcodeGen_Table_columns list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_Table_columns)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_Table_columns)
		t = _ElempcodeGen_Table_columns (_coordref, t);
	remain = (_TPPLST_0pcodeGen_Table_columns)(((_TPPLST_0pcodeGen_Table_columns) t)->_desc2);
	((_TPPLST_0pcodeGen_Table_columns) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_Table_columns;
#else
list = (_TPPLST_0pcodeGen_Table_columns) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_Table_columns));
#endif
list->_prod = RULELST_0pcodeGen_Table_columns;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_Table_columns)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_TableColumnpcodeGen_Table_columns)
	{_TPPLST_codeGen_TableColumnpcodeGen_Table_columns _currn = _currn;
	 _TERMACT_LST_codeGen_TableColumnpcodeGen_Table_columns;
	} this code has been moved to _ElempcodeGen_Table_columns*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_Table_columns;
#else
root = (_TPPpcodeGen_Table_columns) TreeNodeAlloc (sizeof (struct _TPpcodeGen_Table_columns));
#endif
root->_prod = RULEpcodeGen_Table_columns;
root->_desc1 = (_TSPLST_codeGen_Table_columns)list;
{ _TPPpcodeGen_Table_columns _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_Table_columns;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_TableRow (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpcodeGen_TableRow (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpcodeGen_TableRow _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_TableRow;
#else
_currn = (_TPPpcodeGen_TableRow) TreeNodeAlloc (sizeof (struct _TPpcodeGen_TableRow));
#endif
_currn->_prod = RULEpcodeGen_TableRow;
_currn->_desc1 = (_TSPcodeGen_TableRow_cells) MkcodeGen_TableRow_cells (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_TableRow: root of subtree no. 1 can not be made a codeGen_TableRow_cells node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_TableRow;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_TableRow */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_TableCellpcodeGen_TableRow_cells (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_TableCellpcodeGen_TableRow_cells (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_TableCellpcodeGen_TableRow_cells _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_TableCellpcodeGen_TableRow_cells;
#else
	_new = (_TPPLST_codeGen_TableCellpcodeGen_TableRow_cells) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_TableCellpcodeGen_TableRow_cells));
#endif
	_new->_prod = RULELST_codeGen_TableCellpcodeGen_TableRow_cells;
	_new->_desc1 = (_TSPcodeGen_TableCell)elem;
	_new->_desc2 = (_TSPLST_codeGen_TableRow_cells)_new;
	{_TPPLST_codeGen_TableCellpcodeGen_TableRow_cells _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_TableCellpcodeGen_TableRow_cells;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_TableRow_cells (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_TableRow_cells (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_TableCell)) 
		return _MkcodeGen_TableCellpcodeGen_TableRow_cells (_coordref, MkcodeGen_TableCell (_coordref, t));
	message (DEADLY, "a pcodeGen_TableRow_cells list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_TableRow_cells (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_TableRow_cells (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_TableRow_cells)
	l =_ElempcodeGen_TableRow_cells (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_TableRow_cells)
	r = _ElempcodeGen_TableRow_cells (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_TableRow_cells)r)->_desc2);
((_TPPLST_0pcodeGen_TableRow_cells)r)->_desc2 = ((_TPPLST_0pcodeGen_TableRow_cells)l)->_desc2;
((_TPPLST_0pcodeGen_TableRow_cells)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_TableRow_cells (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_TableRow_cells (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_TableRow_cells root;
	_TPPLST_0pcodeGen_TableRow_cells list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_TableRow_cells)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_TableRow_cells)
		t = _ElempcodeGen_TableRow_cells (_coordref, t);
	remain = (_TPPLST_0pcodeGen_TableRow_cells)(((_TPPLST_0pcodeGen_TableRow_cells) t)->_desc2);
	((_TPPLST_0pcodeGen_TableRow_cells) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_TableRow_cells;
#else
list = (_TPPLST_0pcodeGen_TableRow_cells) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_TableRow_cells));
#endif
list->_prod = RULELST_0pcodeGen_TableRow_cells;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_TableRow_cells)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_TableCellpcodeGen_TableRow_cells)
	{_TPPLST_codeGen_TableCellpcodeGen_TableRow_cells _currn = _currn;
	 _TERMACT_LST_codeGen_TableCellpcodeGen_TableRow_cells;
	} this code has been moved to _ElempcodeGen_TableRow_cells*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_TableRow_cells;
#else
root = (_TPPpcodeGen_TableRow_cells) TreeNodeAlloc (sizeof (struct _TPpcodeGen_TableRow_cells));
#endif
root->_prod = RULEpcodeGen_TableRow_cells;
root->_desc1 = (_TSPLST_codeGen_TableRow_cells)list;
{ _TPPpcodeGen_TableRow_cells _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_TableRow_cells;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_TableColumn (POSITION *_coordref, int _TERM1)
#else
NODEPTR MkpcodeGen_TableColumn (_coordref, _TERM1)
	POSITION *_coordref;
	int _TERM1;
#endif
{	_TPPpcodeGen_TableColumn _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_TableColumn;
#else
_currn = (_TPPpcodeGen_TableColumn) TreeNodeAlloc (sizeof (struct _TPpcodeGen_TableColumn));
#endif
_currn->_prod = RULEpcodeGen_TableColumn;
_SETCOORD(_currn)
_TERMACT_pcodeGen_TableColumn;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_TableColumn */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_TableCell (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_TableCell (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_TableCell _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_TableCell;
#else
_currn = (_TPPpcodeGen_TableCell) TreeNodeAlloc (sizeof (struct _TPpcodeGen_TableCell));
#endif
_currn->_prod = RULEpcodeGen_TableCell;
_currn->_desc1 = (_TSPcodeGen_TableCell_columnRef) MkcodeGen_TableCell_columnRef (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_TableCell: root of subtree no. 1 can not be made a codeGen_TableCell_columnRef node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_TableCell_content) MkcodeGen_TableCell_content (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_TableCell: root of subtree no. 2 can not be made a codeGen_TableCell_content node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_TableCell;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_TableCell */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_TableCell_columnRef (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_TableCell_columnRef (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_TableCell_columnRef _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_TableCell_columnRef;
#else
_currn = (_TPPpcodeGen_TableCell_columnRef) TreeNodeAlloc (sizeof (struct _TPpcodeGen_TableCell_columnRef));
#endif
_currn->_prod = RULEpcodeGen_TableCell_columnRef;
_SETCOORD(_currn)
_TERMACT_pcodeGen_TableCell_columnRef;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_TableCell_columnRef */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_LinkTargetExtern (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_LinkTargetExtern (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_LinkTargetExtern _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_LinkTargetExtern;
#else
_currn = (_TPPpcodeGen_LinkTargetExtern) TreeNodeAlloc (sizeof (struct _TPpcodeGen_LinkTargetExtern));
#endif
_currn->_prod = RULEpcodeGen_LinkTargetExtern;
_currn->_desc1 = (_TSPcodeGen_LinkTargetExtern_name) MkcodeGen_LinkTargetExtern_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_LinkTargetExtern: root of subtree no. 1 can not be made a codeGen_LinkTargetExtern_name node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_LinkTargetExtern_url) MkcodeGen_LinkTargetExtern_url (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_LinkTargetExtern: root of subtree no. 2 can not be made a codeGen_LinkTargetExtern_url node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_LinkTargetExtern;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_LinkTargetExtern */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_LinkTargetExtern_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_LinkTargetExtern_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_LinkTargetExtern_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_LinkTargetExtern_name;
#else
_currn = (_TPPpcodeGen_LinkTargetExtern_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_LinkTargetExtern_name));
#endif
_currn->_prod = RULEpcodeGen_LinkTargetExtern_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_LinkTargetExtern_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_LinkTargetExtern_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_LinkTargetExtern_url (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_LinkTargetExtern_url (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_LinkTargetExtern_url _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_LinkTargetExtern_url;
#else
_currn = (_TPPpcodeGen_LinkTargetExtern_url) TreeNodeAlloc (sizeof (struct _TPpcodeGen_LinkTargetExtern_url));
#endif
_currn->_prod = RULEpcodeGen_LinkTargetExtern_url;
_SETCOORD(_currn)
_TERMACT_pcodeGen_LinkTargetExtern_url;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_LinkTargetExtern_url */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_Root (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_Root (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_Root _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_Root;
#else
_currn = (_TPPpcodeGen_Root) TreeNodeAlloc (sizeof (struct _TPpcodeGen_Root));
#endif
_currn->_prod = RULEpcodeGen_Root;
_currn->_desc1 = (_TSPcodeGen_Root_webApp) MkcodeGen_Root_webApp (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_Root: root of subtree no. 1 can not be made a codeGen_Root_webApp node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_Root_library) MkcodeGen_Root_library (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_Root: root of subtree no. 2 can not be made a codeGen_Root_library node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_Root;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_Root */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_LibrarypcodeGen_Root_library (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_LibrarypcodeGen_Root_library (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_LibrarypcodeGen_Root_library _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_LibrarypcodeGen_Root_library;
#else
	_new = (_TPPLST_codeGen_LibrarypcodeGen_Root_library) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_LibrarypcodeGen_Root_library));
#endif
	_new->_prod = RULELST_codeGen_LibrarypcodeGen_Root_library;
	_new->_desc1 = (_TSPcodeGen_Library)elem;
	_new->_desc2 = (_TSPLST_codeGen_Root_library)_new;
	{_TPPLST_codeGen_LibrarypcodeGen_Root_library _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_LibrarypcodeGen_Root_library;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_Root_library (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_Root_library (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_Library)) 
		return _MkcodeGen_LibrarypcodeGen_Root_library (_coordref, MkcodeGen_Library (_coordref, t));
	message (DEADLY, "a pcodeGen_Root_library list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_Root_library (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_Root_library (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_Root_library)
	l =_ElempcodeGen_Root_library (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_Root_library)
	r = _ElempcodeGen_Root_library (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_Root_library)r)->_desc2);
((_TPPLST_0pcodeGen_Root_library)r)->_desc2 = ((_TPPLST_0pcodeGen_Root_library)l)->_desc2;
((_TPPLST_0pcodeGen_Root_library)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_Root_library (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_Root_library (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_Root_library root;
	_TPPLST_0pcodeGen_Root_library list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_Root_library)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_Root_library)
		t = _ElempcodeGen_Root_library (_coordref, t);
	remain = (_TPPLST_0pcodeGen_Root_library)(((_TPPLST_0pcodeGen_Root_library) t)->_desc2);
	((_TPPLST_0pcodeGen_Root_library) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_Root_library;
#else
list = (_TPPLST_0pcodeGen_Root_library) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_Root_library));
#endif
list->_prod = RULELST_0pcodeGen_Root_library;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_Root_library)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_LibrarypcodeGen_Root_library)
	{_TPPLST_codeGen_LibrarypcodeGen_Root_library _currn = _currn;
	 _TERMACT_LST_codeGen_LibrarypcodeGen_Root_library;
	} this code has been moved to _ElempcodeGen_Root_library*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_Root_library;
#else
root = (_TPPpcodeGen_Root_library) TreeNodeAlloc (sizeof (struct _TPpcodeGen_Root_library));
#endif
root->_prod = RULEpcodeGen_Root_library;
root->_desc1 = (_TSPLST_codeGen_Root_library)list;
{ _TPPpcodeGen_Root_library _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_Root_library;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_Library (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpcodeGen_Library (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpcodeGen_Library _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_Library;
#else
_currn = (_TPPpcodeGen_Library) TreeNodeAlloc (sizeof (struct _TPpcodeGen_Library));
#endif
_currn->_prod = RULEpcodeGen_Library;
_currn->_desc1 = (_TSPcodeGen_Library_name) MkcodeGen_Library_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_Library: root of subtree no. 1 can not be made a codeGen_Library_name node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_Library_formularconditions) MkcodeGen_Library_formularconditions (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_Library: root of subtree no. 2 can not be made a codeGen_Library_formularconditions node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_Library_navigationElement) MkcodeGen_Library_navigationElement (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_Library: root of subtree no. 3 can not be made a codeGen_Library_navigationElement node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_Library;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_Library */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_Library_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_Library_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_Library_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_Library_name;
#else
_currn = (_TPPpcodeGen_Library_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_Library_name));
#endif
_currn->_prod = RULEpcodeGen_Library_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_Library_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_Library_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_FormularConditionDefinitionpcodeGen_Library_formularconditions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_FormularConditionDefinitionpcodeGen_Library_formularconditions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_FormularConditionDefinitionpcodeGen_Library_formularconditions _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_FormularConditionDefinitionpcodeGen_Library_formularconditions;
#else
	_new = (_TPPLST_codeGen_FormularConditionDefinitionpcodeGen_Library_formularconditions) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_FormularConditionDefinitionpcodeGen_Library_formularconditions));
#endif
	_new->_prod = RULELST_codeGen_FormularConditionDefinitionpcodeGen_Library_formularconditions;
	_new->_desc1 = (_TSPcodeGen_FormularConditionDefinition)elem;
	_new->_desc2 = (_TSPLST_codeGen_Library_formularconditions)_new;
	{_TPPLST_codeGen_FormularConditionDefinitionpcodeGen_Library_formularconditions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_FormularConditionDefinitionpcodeGen_Library_formularconditions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_Library_formularconditions (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_Library_formularconditions (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_FormularConditionDefinition)) 
		return _MkcodeGen_FormularConditionDefinitionpcodeGen_Library_formularconditions (_coordref, MkcodeGen_FormularConditionDefinition (_coordref, t));
	message (DEADLY, "a pcodeGen_Library_formularconditions list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_Library_formularconditions (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_Library_formularconditions (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_Library_formularconditions)
	l =_ElempcodeGen_Library_formularconditions (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_Library_formularconditions)
	r = _ElempcodeGen_Library_formularconditions (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_Library_formularconditions)r)->_desc2);
((_TPPLST_0pcodeGen_Library_formularconditions)r)->_desc2 = ((_TPPLST_0pcodeGen_Library_formularconditions)l)->_desc2;
((_TPPLST_0pcodeGen_Library_formularconditions)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_Library_formularconditions (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_Library_formularconditions (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_Library_formularconditions root;
	_TPPLST_0pcodeGen_Library_formularconditions list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_Library_formularconditions)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_Library_formularconditions)
		t = _ElempcodeGen_Library_formularconditions (_coordref, t);
	remain = (_TPPLST_0pcodeGen_Library_formularconditions)(((_TPPLST_0pcodeGen_Library_formularconditions) t)->_desc2);
	((_TPPLST_0pcodeGen_Library_formularconditions) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_Library_formularconditions;
#else
list = (_TPPLST_0pcodeGen_Library_formularconditions) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_Library_formularconditions));
#endif
list->_prod = RULELST_0pcodeGen_Library_formularconditions;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_Library_formularconditions)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_FormularConditionDefinitionpcodeGen_Library_formularconditions)
	{_TPPLST_codeGen_FormularConditionDefinitionpcodeGen_Library_formularconditions _currn = _currn;
	 _TERMACT_LST_codeGen_FormularConditionDefinitionpcodeGen_Library_formularconditions;
	} this code has been moved to _ElempcodeGen_Library_formularconditions*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_Library_formularconditions;
#else
root = (_TPPpcodeGen_Library_formularconditions) TreeNodeAlloc (sizeof (struct _TPpcodeGen_Library_formularconditions));
#endif
root->_prod = RULEpcodeGen_Library_formularconditions;
root->_desc1 = (_TSPLST_codeGen_Library_formularconditions)list;
{ _TPPpcodeGen_Library_formularconditions _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_Library_formularconditions;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_NavigationElementpcodeGen_Library_navigationElement (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_NavigationElementpcodeGen_Library_navigationElement (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_NavigationElementpcodeGen_Library_navigationElement _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_NavigationElementpcodeGen_Library_navigationElement;
#else
	_new = (_TPPLST_codeGen_NavigationElementpcodeGen_Library_navigationElement) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_NavigationElementpcodeGen_Library_navigationElement));
#endif
	_new->_prod = RULELST_codeGen_NavigationElementpcodeGen_Library_navigationElement;
	_new->_desc1 = (_TSPcodeGen_NavigationElement)elem;
	_new->_desc2 = (_TSPLST_codeGen_Library_navigationElement)_new;
	{_TPPLST_codeGen_NavigationElementpcodeGen_Library_navigationElement _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_NavigationElementpcodeGen_Library_navigationElement;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_Library_navigationElement (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_Library_navigationElement (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_NavigationElement)) 
		return _MkcodeGen_NavigationElementpcodeGen_Library_navigationElement (_coordref, MkcodeGen_NavigationElement (_coordref, t));
	message (DEADLY, "a pcodeGen_Library_navigationElement list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_Library_navigationElement (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_Library_navigationElement (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_Library_navigationElement)
	l =_ElempcodeGen_Library_navigationElement (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_Library_navigationElement)
	r = _ElempcodeGen_Library_navigationElement (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_Library_navigationElement)r)->_desc2);
((_TPPLST_0pcodeGen_Library_navigationElement)r)->_desc2 = ((_TPPLST_0pcodeGen_Library_navigationElement)l)->_desc2;
((_TPPLST_0pcodeGen_Library_navigationElement)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_Library_navigationElement (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_Library_navigationElement (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_Library_navigationElement root;
	_TPPLST_0pcodeGen_Library_navigationElement list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_Library_navigationElement)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_Library_navigationElement)
		t = _ElempcodeGen_Library_navigationElement (_coordref, t);
	remain = (_TPPLST_0pcodeGen_Library_navigationElement)(((_TPPLST_0pcodeGen_Library_navigationElement) t)->_desc2);
	((_TPPLST_0pcodeGen_Library_navigationElement) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_Library_navigationElement;
#else
list = (_TPPLST_0pcodeGen_Library_navigationElement) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_Library_navigationElement));
#endif
list->_prod = RULELST_0pcodeGen_Library_navigationElement;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_Library_navigationElement)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_NavigationElementpcodeGen_Library_navigationElement)
	{_TPPLST_codeGen_NavigationElementpcodeGen_Library_navigationElement _currn = _currn;
	 _TERMACT_LST_codeGen_NavigationElementpcodeGen_Library_navigationElement;
	} this code has been moved to _ElempcodeGen_Library_navigationElement*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_Library_navigationElement;
#else
root = (_TPPpcodeGen_Library_navigationElement) TreeNodeAlloc (sizeof (struct _TPpcodeGen_Library_navigationElement));
#endif
root->_prod = RULEpcodeGen_Library_navigationElement;
root->_desc1 = (_TSPLST_codeGen_Library_navigationElement)list;
{ _TPPpcodeGen_Library_navigationElement _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_Library_navigationElement;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_NavigationElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_NavigationElement (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_NavigationElement _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_NavigationElement;
#else
_currn = (_TPPpcodeGen_NavigationElement) TreeNodeAlloc (sizeof (struct _TPpcodeGen_NavigationElement));
#endif
_currn->_prod = RULEpcodeGen_NavigationElement;
_currn->_desc1 = (_TSPcodeGen_NavigationElement_code) MkcodeGen_NavigationElement_code (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_NavigationElement: root of subtree no. 1 can not be made a codeGen_NavigationElement_code node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_NavigationElement_name) MkcodeGen_NavigationElement_name (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_NavigationElement: root of subtree no. 2 can not be made a codeGen_NavigationElement_name node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_NavigationElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_NavigationElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_NavigationElement_code (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_NavigationElement_code (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_NavigationElement_code _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_NavigationElement_code;
#else
_currn = (_TPPpcodeGen_NavigationElement_code) TreeNodeAlloc (sizeof (struct _TPpcodeGen_NavigationElement_code));
#endif
_currn->_prod = RULEpcodeGen_NavigationElement_code;
_SETCOORD(_currn)
_TERMACT_pcodeGen_NavigationElement_code;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_NavigationElement_code */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_NavigationElement_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_NavigationElement_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_NavigationElement_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_NavigationElement_name;
#else
_currn = (_TPPpcodeGen_NavigationElement_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_NavigationElement_name));
#endif
_currn->_prod = RULEpcodeGen_NavigationElement_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_NavigationElement_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_NavigationElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebApplication (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5, NODEPTR _desc6, NODEPTR _desc7, NODEPTR _desc8, NODEPTR _desc9, NODEPTR _desc10)
#else
NODEPTR MkpcodeGen_WebApplication (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5,_desc6,_desc7,_desc8,_desc9,_desc10)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
	NODEPTR _desc6;
	NODEPTR _desc7;
	NODEPTR _desc8;
	NODEPTR _desc9;
	NODEPTR _desc10;
#endif
{	_TPPpcodeGen_WebApplication _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_WebApplication;
#else
_currn = (_TPPpcodeGen_WebApplication) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebApplication));
#endif
_currn->_prod = RULEpcodeGen_WebApplication;
_currn->_desc1 = (_TSPcodeGen_WebApplication_name) MkcodeGen_WebApplication_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebApplication: root of subtree no. 1 can not be made a codeGen_WebApplication_name node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_WebApplication_webZone) MkcodeGen_WebApplication_webZone (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebApplication: root of subtree no. 2 can not be made a codeGen_WebApplication_webZone node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_WebApplication_navigationBar) MkcodeGen_WebApplication_navigationBar (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebApplication: root of subtree no. 3 can not be made a codeGen_WebApplication_navigationBar node ", 0, _coordref);
_currn->_desc4 = (_TSPcodeGen_WebApplication_styleSheet) MkcodeGen_WebApplication_styleSheet (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebApplication: root of subtree no. 4 can not be made a codeGen_WebApplication_styleSheet node ", 0, _coordref);
_currn->_desc5 = (_TSPcodeGen_WebApplication_frames) MkcodeGen_WebApplication_frames (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebApplication: root of subtree no. 5 can not be made a codeGen_WebApplication_frames node ", 0, _coordref);
_currn->_desc6 = (_TSPcodeGen_WebApplication_meta) MkcodeGen_WebApplication_meta (_coordref, _desc6);	
if (((NODEPTR)_currn->_desc6) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebApplication: root of subtree no. 6 can not be made a codeGen_WebApplication_meta node ", 0, _coordref);
_currn->_desc7 = (_TSPcodeGen_WebApplication_database) MkcodeGen_WebApplication_database (_coordref, _desc7);	
if (((NODEPTR)_currn->_desc7) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebApplication: root of subtree no. 7 can not be made a codeGen_WebApplication_database node ", 0, _coordref);
_currn->_desc8 = (_TSPcodeGen_WebApplication_roles) MkcodeGen_WebApplication_roles (_coordref, _desc8);	
if (((NODEPTR)_currn->_desc8) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebApplication: root of subtree no. 8 can not be made a codeGen_WebApplication_roles node ", 0, _coordref);
_currn->_desc9 = (_TSPcodeGen_WebApplication_linksExtern) MkcodeGen_WebApplication_linksExtern (_coordref, _desc9);	
if (((NODEPTR)_currn->_desc9) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebApplication: root of subtree no. 9 can not be made a codeGen_WebApplication_linksExtern node ", 0, _coordref);
_currn->_desc10 = (_TSPcodeGen_WebApplication_imagePath) MkcodeGen_WebApplication_imagePath (_coordref, _desc10);	
if (((NODEPTR)_currn->_desc10) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebApplication: root of subtree no. 10 can not be made a codeGen_WebApplication_imagePath node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_WebApplication;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_WebApplication */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebApplication_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_WebApplication_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_WebApplication_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_WebApplication_name;
#else
_currn = (_TPPpcodeGen_WebApplication_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebApplication_name));
#endif
_currn->_prod = RULEpcodeGen_WebApplication_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_WebApplication_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_WebApplication_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_GlobalFramepcodeGen_WebApplication_frames (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_GlobalFramepcodeGen_WebApplication_frames (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_GlobalFramepcodeGen_WebApplication_frames _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_GlobalFramepcodeGen_WebApplication_frames;
#else
	_new = (_TPPLST_codeGen_GlobalFramepcodeGen_WebApplication_frames) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_GlobalFramepcodeGen_WebApplication_frames));
#endif
	_new->_prod = RULELST_codeGen_GlobalFramepcodeGen_WebApplication_frames;
	_new->_desc1 = (_TSPcodeGen_GlobalFrame)elem;
	_new->_desc2 = (_TSPLST_codeGen_WebApplication_frames)_new;
	{_TPPLST_codeGen_GlobalFramepcodeGen_WebApplication_frames _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_GlobalFramepcodeGen_WebApplication_frames;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_WebApplication_frames (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_WebApplication_frames (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_GlobalFrame)) 
		return _MkcodeGen_GlobalFramepcodeGen_WebApplication_frames (_coordref, MkcodeGen_GlobalFrame (_coordref, t));
	message (DEADLY, "a pcodeGen_WebApplication_frames list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_WebApplication_frames (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_WebApplication_frames (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_WebApplication_frames)
	l =_ElempcodeGen_WebApplication_frames (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_WebApplication_frames)
	r = _ElempcodeGen_WebApplication_frames (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_WebApplication_frames)r)->_desc2);
((_TPPLST_0pcodeGen_WebApplication_frames)r)->_desc2 = ((_TPPLST_0pcodeGen_WebApplication_frames)l)->_desc2;
((_TPPLST_0pcodeGen_WebApplication_frames)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebApplication_frames (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_WebApplication_frames (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_WebApplication_frames root;
	_TPPLST_0pcodeGen_WebApplication_frames list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_WebApplication_frames)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_WebApplication_frames)
		t = _ElempcodeGen_WebApplication_frames (_coordref, t);
	remain = (_TPPLST_0pcodeGen_WebApplication_frames)(((_TPPLST_0pcodeGen_WebApplication_frames) t)->_desc2);
	((_TPPLST_0pcodeGen_WebApplication_frames) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_WebApplication_frames;
#else
list = (_TPPLST_0pcodeGen_WebApplication_frames) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_WebApplication_frames));
#endif
list->_prod = RULELST_0pcodeGen_WebApplication_frames;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_WebApplication_frames)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_GlobalFramepcodeGen_WebApplication_frames)
	{_TPPLST_codeGen_GlobalFramepcodeGen_WebApplication_frames _currn = _currn;
	 _TERMACT_LST_codeGen_GlobalFramepcodeGen_WebApplication_frames;
	} this code has been moved to _ElempcodeGen_WebApplication_frames*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_WebApplication_frames;
#else
root = (_TPPpcodeGen_WebApplication_frames) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebApplication_frames));
#endif
root->_prod = RULEpcodeGen_WebApplication_frames;
root->_desc1 = (_TSPLST_codeGen_WebApplication_frames)list;
{ _TPPpcodeGen_WebApplication_frames _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_WebApplication_frames;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_WebRolepcodeGen_WebApplication_roles (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_WebRolepcodeGen_WebApplication_roles (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_WebRolepcodeGen_WebApplication_roles _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_WebRolepcodeGen_WebApplication_roles;
#else
	_new = (_TPPLST_codeGen_WebRolepcodeGen_WebApplication_roles) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_WebRolepcodeGen_WebApplication_roles));
#endif
	_new->_prod = RULELST_codeGen_WebRolepcodeGen_WebApplication_roles;
	_new->_desc1 = (_TSPcodeGen_WebRole)elem;
	_new->_desc2 = (_TSPLST_codeGen_WebApplication_roles)_new;
	{_TPPLST_codeGen_WebRolepcodeGen_WebApplication_roles _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_WebRolepcodeGen_WebApplication_roles;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_WebApplication_roles (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_WebApplication_roles (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_WebRole)) 
		return _MkcodeGen_WebRolepcodeGen_WebApplication_roles (_coordref, MkcodeGen_WebRole (_coordref, t));
	message (DEADLY, "a pcodeGen_WebApplication_roles list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_WebApplication_roles (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_WebApplication_roles (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_WebApplication_roles)
	l =_ElempcodeGen_WebApplication_roles (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_WebApplication_roles)
	r = _ElempcodeGen_WebApplication_roles (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_WebApplication_roles)r)->_desc2);
((_TPPLST_0pcodeGen_WebApplication_roles)r)->_desc2 = ((_TPPLST_0pcodeGen_WebApplication_roles)l)->_desc2;
((_TPPLST_0pcodeGen_WebApplication_roles)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebApplication_roles (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_WebApplication_roles (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_WebApplication_roles root;
	_TPPLST_0pcodeGen_WebApplication_roles list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_WebApplication_roles)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_WebApplication_roles)
		t = _ElempcodeGen_WebApplication_roles (_coordref, t);
	remain = (_TPPLST_0pcodeGen_WebApplication_roles)(((_TPPLST_0pcodeGen_WebApplication_roles) t)->_desc2);
	((_TPPLST_0pcodeGen_WebApplication_roles) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_WebApplication_roles;
#else
list = (_TPPLST_0pcodeGen_WebApplication_roles) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_WebApplication_roles));
#endif
list->_prod = RULELST_0pcodeGen_WebApplication_roles;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_WebApplication_roles)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_WebRolepcodeGen_WebApplication_roles)
	{_TPPLST_codeGen_WebRolepcodeGen_WebApplication_roles _currn = _currn;
	 _TERMACT_LST_codeGen_WebRolepcodeGen_WebApplication_roles;
	} this code has been moved to _ElempcodeGen_WebApplication_roles*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_WebApplication_roles;
#else
root = (_TPPpcodeGen_WebApplication_roles) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebApplication_roles));
#endif
root->_prod = RULEpcodeGen_WebApplication_roles;
root->_desc1 = (_TSPLST_codeGen_WebApplication_roles)list;
{ _TPPpcodeGen_WebApplication_roles _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_WebApplication_roles;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_LinkTargetExternpcodeGen_WebApplication_linksExtern (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_LinkTargetExternpcodeGen_WebApplication_linksExtern (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_LinkTargetExternpcodeGen_WebApplication_linksExtern _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_LinkTargetExternpcodeGen_WebApplication_linksExtern;
#else
	_new = (_TPPLST_codeGen_LinkTargetExternpcodeGen_WebApplication_linksExtern) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_LinkTargetExternpcodeGen_WebApplication_linksExtern));
#endif
	_new->_prod = RULELST_codeGen_LinkTargetExternpcodeGen_WebApplication_linksExtern;
	_new->_desc1 = (_TSPcodeGen_LinkTargetExtern)elem;
	_new->_desc2 = (_TSPLST_codeGen_WebApplication_linksExtern)_new;
	{_TPPLST_codeGen_LinkTargetExternpcodeGen_WebApplication_linksExtern _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_LinkTargetExternpcodeGen_WebApplication_linksExtern;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_WebApplication_linksExtern (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_WebApplication_linksExtern (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_LinkTargetExtern)) 
		return _MkcodeGen_LinkTargetExternpcodeGen_WebApplication_linksExtern (_coordref, MkcodeGen_LinkTargetExtern (_coordref, t));
	message (DEADLY, "a pcodeGen_WebApplication_linksExtern list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_WebApplication_linksExtern (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_WebApplication_linksExtern (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_WebApplication_linksExtern)
	l =_ElempcodeGen_WebApplication_linksExtern (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_WebApplication_linksExtern)
	r = _ElempcodeGen_WebApplication_linksExtern (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_WebApplication_linksExtern)r)->_desc2);
((_TPPLST_0pcodeGen_WebApplication_linksExtern)r)->_desc2 = ((_TPPLST_0pcodeGen_WebApplication_linksExtern)l)->_desc2;
((_TPPLST_0pcodeGen_WebApplication_linksExtern)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebApplication_linksExtern (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_WebApplication_linksExtern (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_WebApplication_linksExtern root;
	_TPPLST_0pcodeGen_WebApplication_linksExtern list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_WebApplication_linksExtern)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_WebApplication_linksExtern)
		t = _ElempcodeGen_WebApplication_linksExtern (_coordref, t);
	remain = (_TPPLST_0pcodeGen_WebApplication_linksExtern)(((_TPPLST_0pcodeGen_WebApplication_linksExtern) t)->_desc2);
	((_TPPLST_0pcodeGen_WebApplication_linksExtern) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_WebApplication_linksExtern;
#else
list = (_TPPLST_0pcodeGen_WebApplication_linksExtern) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_WebApplication_linksExtern));
#endif
list->_prod = RULELST_0pcodeGen_WebApplication_linksExtern;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_WebApplication_linksExtern)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_LinkTargetExternpcodeGen_WebApplication_linksExtern)
	{_TPPLST_codeGen_LinkTargetExternpcodeGen_WebApplication_linksExtern _currn = _currn;
	 _TERMACT_LST_codeGen_LinkTargetExternpcodeGen_WebApplication_linksExtern;
	} this code has been moved to _ElempcodeGen_WebApplication_linksExtern*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_WebApplication_linksExtern;
#else
root = (_TPPpcodeGen_WebApplication_linksExtern) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebApplication_linksExtern));
#endif
root->_prod = RULEpcodeGen_WebApplication_linksExtern;
root->_desc1 = (_TSPLST_codeGen_WebApplication_linksExtern)list;
{ _TPPpcodeGen_WebApplication_linksExtern _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_WebApplication_linksExtern;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebApplication_imagePath (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_WebApplication_imagePath (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_WebApplication_imagePath _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_WebApplication_imagePath;
#else
_currn = (_TPPpcodeGen_WebApplication_imagePath) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebApplication_imagePath));
#endif
_currn->_prod = RULEpcodeGen_WebApplication_imagePath;
_SETCOORD(_currn)
_TERMACT_pcodeGen_WebApplication_imagePath;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_WebApplication_imagePath */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_RootWebZone (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4)
#else
NODEPTR MkpcodeGen_RootWebZone (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
#endif
{	_TPPpcodeGen_RootWebZone _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_RootWebZone;
#else
_currn = (_TPPpcodeGen_RootWebZone) TreeNodeAlloc (sizeof (struct _TPpcodeGen_RootWebZone));
#endif
_currn->_prod = RULEpcodeGen_RootWebZone;
_currn->_desc1 = (_TSPcodeGen_RootWebZone_webTargets) MkcodeGen_RootWebZone_webTargets (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_RootWebZone: root of subtree no. 1 can not be made a codeGen_RootWebZone_webTargets node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_RootWebZone_errorLink) MkcodeGen_RootWebZone_errorLink (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_RootWebZone: root of subtree no. 2 can not be made a codeGen_RootWebZone_errorLink node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_RootWebZone_roles) MkcodeGen_RootWebZone_roles (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_RootWebZone: root of subtree no. 3 can not be made a codeGen_RootWebZone_roles node ", 0, _coordref);
_currn->_desc4 = (_TSPcodeGen_RootWebZone_setSize) MkcodeGen_RootWebZone_setSize (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_RootWebZone: root of subtree no. 4 can not be made a codeGen_RootWebZone_setSize node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_RootWebZone;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_RootWebZone */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_SubWebZonepcodeGen_RootWebZone_webTargets (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_SubWebZonepcodeGen_RootWebZone_webTargets (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_SubWebZonepcodeGen_RootWebZone_webTargets _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_SubWebZonepcodeGen_RootWebZone_webTargets;
#else
	_new = (_TPPLST_codeGen_SubWebZonepcodeGen_RootWebZone_webTargets) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_SubWebZonepcodeGen_RootWebZone_webTargets));
#endif
	_new->_prod = RULELST_codeGen_SubWebZonepcodeGen_RootWebZone_webTargets;
	_new->_desc1 = (_TSPcodeGen_SubWebZone)elem;
	_new->_desc2 = (_TSPLST_codeGen_RootWebZone_webTargets)_new;
	{_TPPLST_codeGen_SubWebZonepcodeGen_RootWebZone_webTargets _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_SubWebZonepcodeGen_RootWebZone_webTargets;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_WebPagepcodeGen_RootWebZone_webTargets (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_WebPagepcodeGen_RootWebZone_webTargets (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_WebPagepcodeGen_RootWebZone_webTargets _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_WebPagepcodeGen_RootWebZone_webTargets;
#else
	_new = (_TPPLST_codeGen_WebPagepcodeGen_RootWebZone_webTargets) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_WebPagepcodeGen_RootWebZone_webTargets));
#endif
	_new->_prod = RULELST_codeGen_WebPagepcodeGen_RootWebZone_webTargets;
	_new->_desc1 = (_TSPcodeGen_WebPage)elem;
	_new->_desc2 = (_TSPLST_codeGen_RootWebZone_webTargets)_new;
	{_TPPLST_codeGen_WebPagepcodeGen_RootWebZone_webTargets _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_WebPagepcodeGen_RootWebZone_webTargets;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_WebActionpcodeGen_RootWebZone_webTargets (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_WebActionpcodeGen_RootWebZone_webTargets (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_WebActionpcodeGen_RootWebZone_webTargets _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_WebActionpcodeGen_RootWebZone_webTargets;
#else
	_new = (_TPPLST_codeGen_WebActionpcodeGen_RootWebZone_webTargets) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_WebActionpcodeGen_RootWebZone_webTargets));
#endif
	_new->_prod = RULELST_codeGen_WebActionpcodeGen_RootWebZone_webTargets;
	_new->_desc1 = (_TSPcodeGen_WebAction)elem;
	_new->_desc2 = (_TSPLST_codeGen_RootWebZone_webTargets)_new;
	{_TPPLST_codeGen_WebActionpcodeGen_RootWebZone_webTargets _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_WebActionpcodeGen_RootWebZone_webTargets;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_RootWebZone_webTargets (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_RootWebZone_webTargets (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_SubWebZone)) 
		return _MkcodeGen_SubWebZonepcodeGen_RootWebZone_webTargets (_coordref, MkcodeGen_SubWebZone (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_WebPage)) 
		return _MkcodeGen_WebPagepcodeGen_RootWebZone_webTargets (_coordref, MkcodeGen_WebPage (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_WebAction)) 
		return _MkcodeGen_WebActionpcodeGen_RootWebZone_webTargets (_coordref, MkcodeGen_WebAction (_coordref, t));
	message (DEADLY, "a pcodeGen_RootWebZone_webTargets list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_RootWebZone_webTargets (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_RootWebZone_webTargets (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_RootWebZone_webTargets)
	l =_ElempcodeGen_RootWebZone_webTargets (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_RootWebZone_webTargets)
	r = _ElempcodeGen_RootWebZone_webTargets (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_RootWebZone_webTargets)r)->_desc2);
((_TPPLST_0pcodeGen_RootWebZone_webTargets)r)->_desc2 = ((_TPPLST_0pcodeGen_RootWebZone_webTargets)l)->_desc2;
((_TPPLST_0pcodeGen_RootWebZone_webTargets)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_RootWebZone_webTargets (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_RootWebZone_webTargets (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_RootWebZone_webTargets root;
	_TPPLST_0pcodeGen_RootWebZone_webTargets list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_RootWebZone_webTargets)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_RootWebZone_webTargets)
		t = _ElempcodeGen_RootWebZone_webTargets (_coordref, t);
	remain = (_TPPLST_0pcodeGen_RootWebZone_webTargets)(((_TPPLST_0pcodeGen_RootWebZone_webTargets) t)->_desc2);
	((_TPPLST_0pcodeGen_RootWebZone_webTargets) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_RootWebZone_webTargets;
#else
list = (_TPPLST_0pcodeGen_RootWebZone_webTargets) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_RootWebZone_webTargets));
#endif
list->_prod = RULELST_0pcodeGen_RootWebZone_webTargets;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_RootWebZone_webTargets)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_SubWebZonepcodeGen_RootWebZone_webTargets)
	{_TPPLST_codeGen_SubWebZonepcodeGen_RootWebZone_webTargets _currn = _currn;
	 _TERMACT_LST_codeGen_SubWebZonepcodeGen_RootWebZone_webTargets;
	} this code has been moved to _ElempcodeGen_RootWebZone_webTargets*/
	/*if (_currn->_prod == RULELST_codeGen_WebPagepcodeGen_RootWebZone_webTargets)
	{_TPPLST_codeGen_WebPagepcodeGen_RootWebZone_webTargets _currn = _currn;
	 _TERMACT_LST_codeGen_WebPagepcodeGen_RootWebZone_webTargets;
	} this code has been moved to _ElempcodeGen_RootWebZone_webTargets*/
	/*if (_currn->_prod == RULELST_codeGen_WebActionpcodeGen_RootWebZone_webTargets)
	{_TPPLST_codeGen_WebActionpcodeGen_RootWebZone_webTargets _currn = _currn;
	 _TERMACT_LST_codeGen_WebActionpcodeGen_RootWebZone_webTargets;
	} this code has been moved to _ElempcodeGen_RootWebZone_webTargets*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_RootWebZone_webTargets;
#else
root = (_TPPpcodeGen_RootWebZone_webTargets) TreeNodeAlloc (sizeof (struct _TPpcodeGen_RootWebZone_webTargets));
#endif
root->_prod = RULEpcodeGen_RootWebZone_webTargets;
root->_desc1 = (_TSPLST_codeGen_RootWebZone_webTargets)list;
{ _TPPpcodeGen_RootWebZone_webTargets _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_RootWebZone_webTargets;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_RootWebZone_errorLink (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_RootWebZone_errorLink (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_RootWebZone_errorLink _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_RootWebZone_errorLink;
#else
_currn = (_TPPpcodeGen_RootWebZone_errorLink) TreeNodeAlloc (sizeof (struct _TPpcodeGen_RootWebZone_errorLink));
#endif
_currn->_prod = RULEpcodeGen_RootWebZone_errorLink;
_SETCOORD(_currn)
_TERMACT_pcodeGen_RootWebZone_errorLink;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_RootWebZone_errorLink */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_WebRoleRefpcodeGen_RootWebZone_roles (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_WebRoleRefpcodeGen_RootWebZone_roles (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_WebRoleRefpcodeGen_RootWebZone_roles _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_WebRoleRefpcodeGen_RootWebZone_roles;
#else
	_new = (_TPPLST_codeGen_WebRoleRefpcodeGen_RootWebZone_roles) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_WebRoleRefpcodeGen_RootWebZone_roles));
#endif
	_new->_prod = RULELST_codeGen_WebRoleRefpcodeGen_RootWebZone_roles;
	_new->_desc1 = (_TSPcodeGen_WebRoleRef)elem;
	_new->_desc2 = (_TSPLST_codeGen_RootWebZone_roles)_new;
	{_TPPLST_codeGen_WebRoleRefpcodeGen_RootWebZone_roles _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_WebRoleRefpcodeGen_RootWebZone_roles;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_RootWebZone_roles (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_RootWebZone_roles (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_WebRoleRef)) 
		return _MkcodeGen_WebRoleRefpcodeGen_RootWebZone_roles (_coordref, MkcodeGen_WebRoleRef (_coordref, t));
	message (DEADLY, "a pcodeGen_RootWebZone_roles list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_RootWebZone_roles (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_RootWebZone_roles (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_RootWebZone_roles)
	l =_ElempcodeGen_RootWebZone_roles (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_RootWebZone_roles)
	r = _ElempcodeGen_RootWebZone_roles (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_RootWebZone_roles)r)->_desc2);
((_TPPLST_0pcodeGen_RootWebZone_roles)r)->_desc2 = ((_TPPLST_0pcodeGen_RootWebZone_roles)l)->_desc2;
((_TPPLST_0pcodeGen_RootWebZone_roles)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_RootWebZone_roles (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_RootWebZone_roles (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_RootWebZone_roles root;
	_TPPLST_0pcodeGen_RootWebZone_roles list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_RootWebZone_roles)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_RootWebZone_roles)
		t = _ElempcodeGen_RootWebZone_roles (_coordref, t);
	remain = (_TPPLST_0pcodeGen_RootWebZone_roles)(((_TPPLST_0pcodeGen_RootWebZone_roles) t)->_desc2);
	((_TPPLST_0pcodeGen_RootWebZone_roles) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_RootWebZone_roles;
#else
list = (_TPPLST_0pcodeGen_RootWebZone_roles) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_RootWebZone_roles));
#endif
list->_prod = RULELST_0pcodeGen_RootWebZone_roles;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_RootWebZone_roles)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_WebRoleRefpcodeGen_RootWebZone_roles)
	{_TPPLST_codeGen_WebRoleRefpcodeGen_RootWebZone_roles _currn = _currn;
	 _TERMACT_LST_codeGen_WebRoleRefpcodeGen_RootWebZone_roles;
	} this code has been moved to _ElempcodeGen_RootWebZone_roles*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_RootWebZone_roles;
#else
root = (_TPPpcodeGen_RootWebZone_roles) TreeNodeAlloc (sizeof (struct _TPpcodeGen_RootWebZone_roles));
#endif
root->_prod = RULEpcodeGen_RootWebZone_roles;
root->_desc1 = (_TSPLST_codeGen_RootWebZone_roles)list;
{ _TPPpcodeGen_RootWebZone_roles _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_RootWebZone_roles;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_RootWebZone_setSize (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_RootWebZone_setSize (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_RootWebZone_setSize _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_RootWebZone_setSize;
#else
_currn = (_TPPpcodeGen_RootWebZone_setSize) TreeNodeAlloc (sizeof (struct _TPpcodeGen_RootWebZone_setSize));
#endif
_currn->_prod = RULEpcodeGen_RootWebZone_setSize;
_SETCOORD(_currn)
_TERMACT_pcodeGen_RootWebZone_setSize;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_RootWebZone_setSize */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_SubWebZone (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5, NODEPTR _desc6, NODEPTR _desc7)
#else
NODEPTR MkpcodeGen_SubWebZone (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5,_desc6,_desc7)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
	NODEPTR _desc6;
	NODEPTR _desc7;
#endif
{	_TPPpcodeGen_SubWebZone _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_SubWebZone;
#else
_currn = (_TPPpcodeGen_SubWebZone) TreeNodeAlloc (sizeof (struct _TPpcodeGen_SubWebZone));
#endif
_currn->_prod = RULEpcodeGen_SubWebZone;
_currn->_desc1 = (_TSPcodeGen_SubWebZone_name) MkcodeGen_SubWebZone_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_SubWebZone: root of subtree no. 1 can not be made a codeGen_SubWebZone_name node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_SubWebZone_webTargets) MkcodeGen_SubWebZone_webTargets (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_SubWebZone: root of subtree no. 2 can not be made a codeGen_SubWebZone_webTargets node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_SubWebZone_errorLink) MkcodeGen_SubWebZone_errorLink (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_SubWebZone: root of subtree no. 3 can not be made a codeGen_SubWebZone_errorLink node ", 0, _coordref);
_currn->_desc4 = (_TSPcodeGen_SubWebZone_globalparams) MkcodeGen_SubWebZone_globalparams (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_SubWebZone: root of subtree no. 4 can not be made a codeGen_SubWebZone_globalparams node ", 0, _coordref);
_currn->_desc5 = (_TSPcodeGen_SubWebZone_roles) MkcodeGen_SubWebZone_roles (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_SubWebZone: root of subtree no. 5 can not be made a codeGen_SubWebZone_roles node ", 0, _coordref);
_currn->_desc6 = (_TSPcodeGen_SubWebZone_setSize) MkcodeGen_SubWebZone_setSize (_coordref, _desc6);	
if (((NODEPTR)_currn->_desc6) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_SubWebZone: root of subtree no. 6 can not be made a codeGen_SubWebZone_setSize node ", 0, _coordref);
_currn->_desc7 = (_TSPcodeGen_SubWebZone_position) MkcodeGen_SubWebZone_position (_coordref, _desc7);	
if (((NODEPTR)_currn->_desc7) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_SubWebZone: root of subtree no. 7 can not be made a codeGen_SubWebZone_position node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_SubWebZone;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_SubWebZone */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_SubWebZone_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_SubWebZone_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_SubWebZone_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_SubWebZone_name;
#else
_currn = (_TPPpcodeGen_SubWebZone_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_SubWebZone_name));
#endif
_currn->_prod = RULEpcodeGen_SubWebZone_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_SubWebZone_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_SubWebZone_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_SubWebZonepcodeGen_SubWebZone_webTargets (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_SubWebZonepcodeGen_SubWebZone_webTargets (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_SubWebZonepcodeGen_SubWebZone_webTargets _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_SubWebZonepcodeGen_SubWebZone_webTargets;
#else
	_new = (_TPPLST_codeGen_SubWebZonepcodeGen_SubWebZone_webTargets) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_SubWebZonepcodeGen_SubWebZone_webTargets));
#endif
	_new->_prod = RULELST_codeGen_SubWebZonepcodeGen_SubWebZone_webTargets;
	_new->_desc1 = (_TSPcodeGen_SubWebZone)elem;
	_new->_desc2 = (_TSPLST_codeGen_SubWebZone_webTargets)_new;
	{_TPPLST_codeGen_SubWebZonepcodeGen_SubWebZone_webTargets _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_SubWebZonepcodeGen_SubWebZone_webTargets;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_WebPagepcodeGen_SubWebZone_webTargets (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_WebPagepcodeGen_SubWebZone_webTargets (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_WebPagepcodeGen_SubWebZone_webTargets _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_WebPagepcodeGen_SubWebZone_webTargets;
#else
	_new = (_TPPLST_codeGen_WebPagepcodeGen_SubWebZone_webTargets) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_WebPagepcodeGen_SubWebZone_webTargets));
#endif
	_new->_prod = RULELST_codeGen_WebPagepcodeGen_SubWebZone_webTargets;
	_new->_desc1 = (_TSPcodeGen_WebPage)elem;
	_new->_desc2 = (_TSPLST_codeGen_SubWebZone_webTargets)_new;
	{_TPPLST_codeGen_WebPagepcodeGen_SubWebZone_webTargets _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_WebPagepcodeGen_SubWebZone_webTargets;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_WebActionpcodeGen_SubWebZone_webTargets (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_WebActionpcodeGen_SubWebZone_webTargets (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_WebActionpcodeGen_SubWebZone_webTargets _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_WebActionpcodeGen_SubWebZone_webTargets;
#else
	_new = (_TPPLST_codeGen_WebActionpcodeGen_SubWebZone_webTargets) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_WebActionpcodeGen_SubWebZone_webTargets));
#endif
	_new->_prod = RULELST_codeGen_WebActionpcodeGen_SubWebZone_webTargets;
	_new->_desc1 = (_TSPcodeGen_WebAction)elem;
	_new->_desc2 = (_TSPLST_codeGen_SubWebZone_webTargets)_new;
	{_TPPLST_codeGen_WebActionpcodeGen_SubWebZone_webTargets _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_WebActionpcodeGen_SubWebZone_webTargets;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_SubWebZone_webTargets (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_SubWebZone_webTargets (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_SubWebZone)) 
		return _MkcodeGen_SubWebZonepcodeGen_SubWebZone_webTargets (_coordref, MkcodeGen_SubWebZone (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_WebPage)) 
		return _MkcodeGen_WebPagepcodeGen_SubWebZone_webTargets (_coordref, MkcodeGen_WebPage (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_WebAction)) 
		return _MkcodeGen_WebActionpcodeGen_SubWebZone_webTargets (_coordref, MkcodeGen_WebAction (_coordref, t));
	message (DEADLY, "a pcodeGen_SubWebZone_webTargets list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_SubWebZone_webTargets (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_SubWebZone_webTargets (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_SubWebZone_webTargets)
	l =_ElempcodeGen_SubWebZone_webTargets (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_SubWebZone_webTargets)
	r = _ElempcodeGen_SubWebZone_webTargets (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_SubWebZone_webTargets)r)->_desc2);
((_TPPLST_0pcodeGen_SubWebZone_webTargets)r)->_desc2 = ((_TPPLST_0pcodeGen_SubWebZone_webTargets)l)->_desc2;
((_TPPLST_0pcodeGen_SubWebZone_webTargets)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_SubWebZone_webTargets (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_SubWebZone_webTargets (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_SubWebZone_webTargets root;
	_TPPLST_0pcodeGen_SubWebZone_webTargets list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_SubWebZone_webTargets)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_SubWebZone_webTargets)
		t = _ElempcodeGen_SubWebZone_webTargets (_coordref, t);
	remain = (_TPPLST_0pcodeGen_SubWebZone_webTargets)(((_TPPLST_0pcodeGen_SubWebZone_webTargets) t)->_desc2);
	((_TPPLST_0pcodeGen_SubWebZone_webTargets) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_SubWebZone_webTargets;
#else
list = (_TPPLST_0pcodeGen_SubWebZone_webTargets) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_SubWebZone_webTargets));
#endif
list->_prod = RULELST_0pcodeGen_SubWebZone_webTargets;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_SubWebZone_webTargets)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_SubWebZonepcodeGen_SubWebZone_webTargets)
	{_TPPLST_codeGen_SubWebZonepcodeGen_SubWebZone_webTargets _currn = _currn;
	 _TERMACT_LST_codeGen_SubWebZonepcodeGen_SubWebZone_webTargets;
	} this code has been moved to _ElempcodeGen_SubWebZone_webTargets*/
	/*if (_currn->_prod == RULELST_codeGen_WebPagepcodeGen_SubWebZone_webTargets)
	{_TPPLST_codeGen_WebPagepcodeGen_SubWebZone_webTargets _currn = _currn;
	 _TERMACT_LST_codeGen_WebPagepcodeGen_SubWebZone_webTargets;
	} this code has been moved to _ElempcodeGen_SubWebZone_webTargets*/
	/*if (_currn->_prod == RULELST_codeGen_WebActionpcodeGen_SubWebZone_webTargets)
	{_TPPLST_codeGen_WebActionpcodeGen_SubWebZone_webTargets _currn = _currn;
	 _TERMACT_LST_codeGen_WebActionpcodeGen_SubWebZone_webTargets;
	} this code has been moved to _ElempcodeGen_SubWebZone_webTargets*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_SubWebZone_webTargets;
#else
root = (_TPPpcodeGen_SubWebZone_webTargets) TreeNodeAlloc (sizeof (struct _TPpcodeGen_SubWebZone_webTargets));
#endif
root->_prod = RULEpcodeGen_SubWebZone_webTargets;
root->_desc1 = (_TSPLST_codeGen_SubWebZone_webTargets)list;
{ _TPPpcodeGen_SubWebZone_webTargets _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_SubWebZone_webTargets;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_SubWebZone_errorLink (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_SubWebZone_errorLink (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_SubWebZone_errorLink _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_SubWebZone_errorLink;
#else
_currn = (_TPPpcodeGen_SubWebZone_errorLink) TreeNodeAlloc (sizeof (struct _TPpcodeGen_SubWebZone_errorLink));
#endif
_currn->_prod = RULEpcodeGen_SubWebZone_errorLink;
_SETCOORD(_currn)
_TERMACT_pcodeGen_SubWebZone_errorLink;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_SubWebZone_errorLink */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_WebNodeFormalParampcodeGen_SubWebZone_globalparams (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_WebNodeFormalParampcodeGen_SubWebZone_globalparams (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_WebNodeFormalParampcodeGen_SubWebZone_globalparams _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_WebNodeFormalParampcodeGen_SubWebZone_globalparams;
#else
	_new = (_TPPLST_codeGen_WebNodeFormalParampcodeGen_SubWebZone_globalparams) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_WebNodeFormalParampcodeGen_SubWebZone_globalparams));
#endif
	_new->_prod = RULELST_codeGen_WebNodeFormalParampcodeGen_SubWebZone_globalparams;
	_new->_desc1 = (_TSPcodeGen_WebNodeFormalParam)elem;
	_new->_desc2 = (_TSPLST_codeGen_SubWebZone_globalparams)_new;
	{_TPPLST_codeGen_WebNodeFormalParampcodeGen_SubWebZone_globalparams _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_WebNodeFormalParampcodeGen_SubWebZone_globalparams;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_SubWebZone_globalparams (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_SubWebZone_globalparams (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_WebNodeFormalParam)) 
		return _MkcodeGen_WebNodeFormalParampcodeGen_SubWebZone_globalparams (_coordref, MkcodeGen_WebNodeFormalParam (_coordref, t));
	message (DEADLY, "a pcodeGen_SubWebZone_globalparams list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_SubWebZone_globalparams (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_SubWebZone_globalparams (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_SubWebZone_globalparams)
	l =_ElempcodeGen_SubWebZone_globalparams (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_SubWebZone_globalparams)
	r = _ElempcodeGen_SubWebZone_globalparams (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_SubWebZone_globalparams)r)->_desc2);
((_TPPLST_0pcodeGen_SubWebZone_globalparams)r)->_desc2 = ((_TPPLST_0pcodeGen_SubWebZone_globalparams)l)->_desc2;
((_TPPLST_0pcodeGen_SubWebZone_globalparams)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_SubWebZone_globalparams (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_SubWebZone_globalparams (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_SubWebZone_globalparams root;
	_TPPLST_0pcodeGen_SubWebZone_globalparams list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_SubWebZone_globalparams)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_SubWebZone_globalparams)
		t = _ElempcodeGen_SubWebZone_globalparams (_coordref, t);
	remain = (_TPPLST_0pcodeGen_SubWebZone_globalparams)(((_TPPLST_0pcodeGen_SubWebZone_globalparams) t)->_desc2);
	((_TPPLST_0pcodeGen_SubWebZone_globalparams) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_SubWebZone_globalparams;
#else
list = (_TPPLST_0pcodeGen_SubWebZone_globalparams) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_SubWebZone_globalparams));
#endif
list->_prod = RULELST_0pcodeGen_SubWebZone_globalparams;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_SubWebZone_globalparams)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_WebNodeFormalParampcodeGen_SubWebZone_globalparams)
	{_TPPLST_codeGen_WebNodeFormalParampcodeGen_SubWebZone_globalparams _currn = _currn;
	 _TERMACT_LST_codeGen_WebNodeFormalParampcodeGen_SubWebZone_globalparams;
	} this code has been moved to _ElempcodeGen_SubWebZone_globalparams*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_SubWebZone_globalparams;
#else
root = (_TPPpcodeGen_SubWebZone_globalparams) TreeNodeAlloc (sizeof (struct _TPpcodeGen_SubWebZone_globalparams));
#endif
root->_prod = RULEpcodeGen_SubWebZone_globalparams;
root->_desc1 = (_TSPLST_codeGen_SubWebZone_globalparams)list;
{ _TPPpcodeGen_SubWebZone_globalparams _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_SubWebZone_globalparams;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_WebRoleRefpcodeGen_SubWebZone_roles (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_WebRoleRefpcodeGen_SubWebZone_roles (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_WebRoleRefpcodeGen_SubWebZone_roles _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_WebRoleRefpcodeGen_SubWebZone_roles;
#else
	_new = (_TPPLST_codeGen_WebRoleRefpcodeGen_SubWebZone_roles) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_WebRoleRefpcodeGen_SubWebZone_roles));
#endif
	_new->_prod = RULELST_codeGen_WebRoleRefpcodeGen_SubWebZone_roles;
	_new->_desc1 = (_TSPcodeGen_WebRoleRef)elem;
	_new->_desc2 = (_TSPLST_codeGen_SubWebZone_roles)_new;
	{_TPPLST_codeGen_WebRoleRefpcodeGen_SubWebZone_roles _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_WebRoleRefpcodeGen_SubWebZone_roles;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_SubWebZone_roles (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_SubWebZone_roles (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_WebRoleRef)) 
		return _MkcodeGen_WebRoleRefpcodeGen_SubWebZone_roles (_coordref, MkcodeGen_WebRoleRef (_coordref, t));
	message (DEADLY, "a pcodeGen_SubWebZone_roles list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_SubWebZone_roles (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_SubWebZone_roles (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_SubWebZone_roles)
	l =_ElempcodeGen_SubWebZone_roles (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_SubWebZone_roles)
	r = _ElempcodeGen_SubWebZone_roles (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_SubWebZone_roles)r)->_desc2);
((_TPPLST_0pcodeGen_SubWebZone_roles)r)->_desc2 = ((_TPPLST_0pcodeGen_SubWebZone_roles)l)->_desc2;
((_TPPLST_0pcodeGen_SubWebZone_roles)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_SubWebZone_roles (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_SubWebZone_roles (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_SubWebZone_roles root;
	_TPPLST_0pcodeGen_SubWebZone_roles list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_SubWebZone_roles)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_SubWebZone_roles)
		t = _ElempcodeGen_SubWebZone_roles (_coordref, t);
	remain = (_TPPLST_0pcodeGen_SubWebZone_roles)(((_TPPLST_0pcodeGen_SubWebZone_roles) t)->_desc2);
	((_TPPLST_0pcodeGen_SubWebZone_roles) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_SubWebZone_roles;
#else
list = (_TPPLST_0pcodeGen_SubWebZone_roles) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_SubWebZone_roles));
#endif
list->_prod = RULELST_0pcodeGen_SubWebZone_roles;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_SubWebZone_roles)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_WebRoleRefpcodeGen_SubWebZone_roles)
	{_TPPLST_codeGen_WebRoleRefpcodeGen_SubWebZone_roles _currn = _currn;
	 _TERMACT_LST_codeGen_WebRoleRefpcodeGen_SubWebZone_roles;
	} this code has been moved to _ElempcodeGen_SubWebZone_roles*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_SubWebZone_roles;
#else
root = (_TPPpcodeGen_SubWebZone_roles) TreeNodeAlloc (sizeof (struct _TPpcodeGen_SubWebZone_roles));
#endif
root->_prod = RULEpcodeGen_SubWebZone_roles;
root->_desc1 = (_TSPLST_codeGen_SubWebZone_roles)list;
{ _TPPpcodeGen_SubWebZone_roles _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_SubWebZone_roles;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_SubWebZone_setSize (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_SubWebZone_setSize (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_SubWebZone_setSize _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_SubWebZone_setSize;
#else
_currn = (_TPPpcodeGen_SubWebZone_setSize) TreeNodeAlloc (sizeof (struct _TPpcodeGen_SubWebZone_setSize));
#endif
_currn->_prod = RULEpcodeGen_SubWebZone_setSize;
_SETCOORD(_currn)
_TERMACT_pcodeGen_SubWebZone_setSize;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_SubWebZone_setSize */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_SubWebZone_position (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_SubWebZone_position (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_SubWebZone_position _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_SubWebZone_position;
#else
_currn = (_TPPpcodeGen_SubWebZone_position) TreeNodeAlloc (sizeof (struct _TPpcodeGen_SubWebZone_position));
#endif
_currn->_prod = RULEpcodeGen_SubWebZone_position;
_SETCOORD(_currn)
_TERMACT_pcodeGen_SubWebZone_position;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_SubWebZone_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebRole (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_WebRole (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_WebRole _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_WebRole;
#else
_currn = (_TPPpcodeGen_WebRole) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebRole));
#endif
_currn->_prod = RULEpcodeGen_WebRole;
_currn->_desc1 = (_TSPcodeGen_WebRole_name) MkcodeGen_WebRole_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebRole: root of subtree no. 1 can not be made a codeGen_WebRole_name node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_WebRole_color) MkcodeGen_WebRole_color (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebRole: root of subtree no. 2 can not be made a codeGen_WebRole_color node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_WebRole;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_WebRole */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebRole_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_WebRole_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_WebRole_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_WebRole_name;
#else
_currn = (_TPPpcodeGen_WebRole_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebRole_name));
#endif
_currn->_prod = RULEpcodeGen_WebRole_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_WebRole_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_WebRole_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebRole_color (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_WebRole_color (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_WebRole_color _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_WebRole_color;
#else
_currn = (_TPPpcodeGen_WebRole_color) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebRole_color));
#endif
_currn->_prod = RULEpcodeGen_WebRole_color;
_SETCOORD(_currn)
_TERMACT_pcodeGen_WebRole_color;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_WebRole_color */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebRoleRef (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpcodeGen_WebRoleRef (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpcodeGen_WebRoleRef _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_WebRoleRef;
#else
_currn = (_TPPpcodeGen_WebRoleRef) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebRoleRef));
#endif
_currn->_prod = RULEpcodeGen_WebRoleRef;
_currn->_desc1 = (_TSPcodeGen_WebRoleRef_role) MkcodeGen_WebRoleRef_role (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebRoleRef: root of subtree no. 1 can not be made a codeGen_WebRoleRef_role node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_WebRoleRef;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_WebRoleRef */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebRoleRef_role (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_WebRoleRef_role (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_WebRoleRef_role _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_WebRoleRef_role;
#else
_currn = (_TPPpcodeGen_WebRoleRef_role) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebRoleRef_role));
#endif
_currn->_prod = RULEpcodeGen_WebRoleRef_role;
_SETCOORD(_currn)
_TERMACT_pcodeGen_WebRoleRef_role;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_WebRoleRef_role */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebPage (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5, NODEPTR _desc6)
#else
NODEPTR MkpcodeGen_WebPage (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5,_desc6)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
	NODEPTR _desc6;
#endif
{	_TPPpcodeGen_WebPage _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_WebPage;
#else
_currn = (_TPPpcodeGen_WebPage) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebPage));
#endif
_currn->_prod = RULEpcodeGen_WebPage;
_currn->_desc1 = (_TSPcodeGen_WebPage_name) MkcodeGen_WebPage_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebPage: root of subtree no. 1 can not be made a codeGen_WebPage_name node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_WebPage_formalParameter) MkcodeGen_WebPage_formalParameter (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebPage: root of subtree no. 2 can not be made a codeGen_WebPage_formalParameter node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_WebPage_title) MkcodeGen_WebPage_title (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebPage: root of subtree no. 3 can not be made a codeGen_WebPage_title node ", 0, _coordref);
_currn->_desc4 = (_TSPcodeGen_WebPage_frame) MkcodeGen_WebPage_frame (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebPage: root of subtree no. 4 can not be made a codeGen_WebPage_frame node ", 0, _coordref);
_currn->_desc5 = (_TSPcodeGen_WebPage_styleSheetBlock) MkcodeGen_WebPage_styleSheetBlock (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebPage: root of subtree no. 5 can not be made a codeGen_WebPage_styleSheetBlock node ", 0, _coordref);
_currn->_desc6 = (_TSPcodeGen_WebPage_position) MkcodeGen_WebPage_position (_coordref, _desc6);	
if (((NODEPTR)_currn->_desc6) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebPage: root of subtree no. 6 can not be made a codeGen_WebPage_position node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_WebPage;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_WebPage */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebPage_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_WebPage_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_WebPage_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_WebPage_name;
#else
_currn = (_TPPpcodeGen_WebPage_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebPage_name));
#endif
_currn->_prod = RULEpcodeGen_WebPage_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_WebPage_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_WebPage_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_WebNodeFormalParampcodeGen_WebPage_formalParameter (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_WebNodeFormalParampcodeGen_WebPage_formalParameter (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_WebNodeFormalParampcodeGen_WebPage_formalParameter _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_WebNodeFormalParampcodeGen_WebPage_formalParameter;
#else
	_new = (_TPPLST_codeGen_WebNodeFormalParampcodeGen_WebPage_formalParameter) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_WebNodeFormalParampcodeGen_WebPage_formalParameter));
#endif
	_new->_prod = RULELST_codeGen_WebNodeFormalParampcodeGen_WebPage_formalParameter;
	_new->_desc1 = (_TSPcodeGen_WebNodeFormalParam)elem;
	_new->_desc2 = (_TSPLST_codeGen_WebPage_formalParameter)_new;
	{_TPPLST_codeGen_WebNodeFormalParampcodeGen_WebPage_formalParameter _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_WebNodeFormalParampcodeGen_WebPage_formalParameter;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_WebPage_formalParameter (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_WebPage_formalParameter (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_WebNodeFormalParam)) 
		return _MkcodeGen_WebNodeFormalParampcodeGen_WebPage_formalParameter (_coordref, MkcodeGen_WebNodeFormalParam (_coordref, t));
	message (DEADLY, "a pcodeGen_WebPage_formalParameter list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_WebPage_formalParameter (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_WebPage_formalParameter (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_WebPage_formalParameter)
	l =_ElempcodeGen_WebPage_formalParameter (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_WebPage_formalParameter)
	r = _ElempcodeGen_WebPage_formalParameter (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_WebPage_formalParameter)r)->_desc2);
((_TPPLST_0pcodeGen_WebPage_formalParameter)r)->_desc2 = ((_TPPLST_0pcodeGen_WebPage_formalParameter)l)->_desc2;
((_TPPLST_0pcodeGen_WebPage_formalParameter)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebPage_formalParameter (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_WebPage_formalParameter (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_WebPage_formalParameter root;
	_TPPLST_0pcodeGen_WebPage_formalParameter list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_WebPage_formalParameter)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_WebPage_formalParameter)
		t = _ElempcodeGen_WebPage_formalParameter (_coordref, t);
	remain = (_TPPLST_0pcodeGen_WebPage_formalParameter)(((_TPPLST_0pcodeGen_WebPage_formalParameter) t)->_desc2);
	((_TPPLST_0pcodeGen_WebPage_formalParameter) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_WebPage_formalParameter;
#else
list = (_TPPLST_0pcodeGen_WebPage_formalParameter) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_WebPage_formalParameter));
#endif
list->_prod = RULELST_0pcodeGen_WebPage_formalParameter;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_WebPage_formalParameter)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_WebNodeFormalParampcodeGen_WebPage_formalParameter)
	{_TPPLST_codeGen_WebNodeFormalParampcodeGen_WebPage_formalParameter _currn = _currn;
	 _TERMACT_LST_codeGen_WebNodeFormalParampcodeGen_WebPage_formalParameter;
	} this code has been moved to _ElempcodeGen_WebPage_formalParameter*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_WebPage_formalParameter;
#else
root = (_TPPpcodeGen_WebPage_formalParameter) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebPage_formalParameter));
#endif
root->_prod = RULEpcodeGen_WebPage_formalParameter;
root->_desc1 = (_TSPLST_codeGen_WebPage_formalParameter)list;
{ _TPPpcodeGen_WebPage_formalParameter _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_WebPage_formalParameter;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebPage_title (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_WebPage_title (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_WebPage_title _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_WebPage_title;
#else
_currn = (_TPPpcodeGen_WebPage_title) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebPage_title));
#endif
_currn->_prod = RULEpcodeGen_WebPage_title;
_SETCOORD(_currn)
_TERMACT_pcodeGen_WebPage_title;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_WebPage_title */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebPage_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_WebPage_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_WebPage_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_WebPage_styleSheetBlock;
#else
_currn = (_TPPpcodeGen_WebPage_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebPage_styleSheetBlock));
#endif
_currn->_prod = RULEpcodeGen_WebPage_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_pcodeGen_WebPage_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_WebPage_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebPage_position (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_WebPage_position (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_WebPage_position _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_WebPage_position;
#else
_currn = (_TPPpcodeGen_WebPage_position) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebPage_position));
#endif
_currn->_prod = RULEpcodeGen_WebPage_position;
_SETCOORD(_currn)
_TERMACT_pcodeGen_WebPage_position;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_WebPage_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_Frame (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_Frame (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_Frame _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_Frame;
#else
_currn = (_TPPpcodeGen_Frame) TreeNodeAlloc (sizeof (struct _TPpcodeGen_Frame));
#endif
_currn->_prod = RULEpcodeGen_Frame;
_currn->_desc1 = (_TSPcodeGen_Frame_components) MkcodeGen_Frame_components (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_Frame: root of subtree no. 1 can not be made a codeGen_Frame_components node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_Frame_styleSheetBlock) MkcodeGen_Frame_styleSheetBlock (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_Frame: root of subtree no. 2 can not be made a codeGen_Frame_styleSheetBlock node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_Frame;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_Frame */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_FormularpcodeGen_Frame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_FormularpcodeGen_Frame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_FormularpcodeGen_Frame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_FormularpcodeGen_Frame_components;
#else
	_new = (_TPPLST_codeGen_FormularpcodeGen_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_FormularpcodeGen_Frame_components));
#endif
	_new->_prod = RULELST_codeGen_FormularpcodeGen_Frame_components;
	_new->_desc1 = (_TSPcodeGen_Formular)elem;
	_new->_desc2 = (_TSPLST_codeGen_Frame_components)_new;
	{_TPPLST_codeGen_FormularpcodeGen_Frame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_FormularpcodeGen_Frame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_ParagraphpcodeGen_Frame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_ParagraphpcodeGen_Frame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_ParagraphpcodeGen_Frame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_ParagraphpcodeGen_Frame_components;
#else
	_new = (_TPPLST_codeGen_ParagraphpcodeGen_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_ParagraphpcodeGen_Frame_components));
#endif
	_new->_prod = RULELST_codeGen_ParagraphpcodeGen_Frame_components;
	_new->_desc1 = (_TSPcodeGen_Paragraph)elem;
	_new->_desc2 = (_TSPLST_codeGen_Frame_components)_new;
	{_TPPLST_codeGen_ParagraphpcodeGen_Frame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_ParagraphpcodeGen_Frame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_HeadingpcodeGen_Frame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_HeadingpcodeGen_Frame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_HeadingpcodeGen_Frame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_HeadingpcodeGen_Frame_components;
#else
	_new = (_TPPLST_codeGen_HeadingpcodeGen_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_HeadingpcodeGen_Frame_components));
#endif
	_new->_prod = RULELST_codeGen_HeadingpcodeGen_Frame_components;
	_new->_desc1 = (_TSPcodeGen_Heading)elem;
	_new->_desc2 = (_TSPLST_codeGen_Frame_components)_new;
	{_TPPLST_codeGen_HeadingpcodeGen_Frame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_HeadingpcodeGen_Frame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_TablepcodeGen_Frame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_TablepcodeGen_Frame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_TablepcodeGen_Frame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_TablepcodeGen_Frame_components;
#else
	_new = (_TPPLST_codeGen_TablepcodeGen_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_TablepcodeGen_Frame_components));
#endif
	_new->_prod = RULELST_codeGen_TablepcodeGen_Frame_components;
	_new->_desc1 = (_TSPcodeGen_Table)elem;
	_new->_desc2 = (_TSPLST_codeGen_Frame_components)_new;
	{_TPPLST_codeGen_TablepcodeGen_Frame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_TablepcodeGen_Frame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_FrameRefpcodeGen_Frame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_FrameRefpcodeGen_Frame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_FrameRefpcodeGen_Frame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_FrameRefpcodeGen_Frame_components;
#else
	_new = (_TPPLST_codeGen_FrameRefpcodeGen_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_FrameRefpcodeGen_Frame_components));
#endif
	_new->_prod = RULELST_codeGen_FrameRefpcodeGen_Frame_components;
	_new->_desc1 = (_TSPcodeGen_FrameRef)elem;
	_new->_desc2 = (_TSPLST_codeGen_Frame_components)_new;
	{_TPPLST_codeGen_FrameRefpcodeGen_Frame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_FrameRefpcodeGen_Frame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_ListOutputElementpcodeGen_Frame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_ListOutputElementpcodeGen_Frame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_ListOutputElementpcodeGen_Frame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_ListOutputElementpcodeGen_Frame_components;
#else
	_new = (_TPPLST_codeGen_ListOutputElementpcodeGen_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_ListOutputElementpcodeGen_Frame_components));
#endif
	_new->_prod = RULELST_codeGen_ListOutputElementpcodeGen_Frame_components;
	_new->_desc1 = (_TSPcodeGen_ListOutputElement)elem;
	_new->_desc2 = (_TSPLST_codeGen_Frame_components)_new;
	{_TPPLST_codeGen_ListOutputElementpcodeGen_Frame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_ListOutputElementpcodeGen_Frame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_Frame_components (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_Frame_components (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_Formular)) 
		return _MkcodeGen_FormularpcodeGen_Frame_components (_coordref, MkcodeGen_Formular (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_Paragraph)) 
		return _MkcodeGen_ParagraphpcodeGen_Frame_components (_coordref, MkcodeGen_Paragraph (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_Heading)) 
		return _MkcodeGen_HeadingpcodeGen_Frame_components (_coordref, MkcodeGen_Heading (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_Table)) 
		return _MkcodeGen_TablepcodeGen_Frame_components (_coordref, MkcodeGen_Table (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_FrameRef)) 
		return _MkcodeGen_FrameRefpcodeGen_Frame_components (_coordref, MkcodeGen_FrameRef (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_ListOutputElement)) 
		return _MkcodeGen_ListOutputElementpcodeGen_Frame_components (_coordref, MkcodeGen_ListOutputElement (_coordref, t));
	message (DEADLY, "a pcodeGen_Frame_components list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_Frame_components (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_Frame_components (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_Frame_components)
	l =_ElempcodeGen_Frame_components (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_Frame_components)
	r = _ElempcodeGen_Frame_components (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_Frame_components)r)->_desc2);
((_TPPLST_0pcodeGen_Frame_components)r)->_desc2 = ((_TPPLST_0pcodeGen_Frame_components)l)->_desc2;
((_TPPLST_0pcodeGen_Frame_components)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_Frame_components (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_Frame_components (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_Frame_components root;
	_TPPLST_0pcodeGen_Frame_components list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_Frame_components)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_Frame_components)
		t = _ElempcodeGen_Frame_components (_coordref, t);
	remain = (_TPPLST_0pcodeGen_Frame_components)(((_TPPLST_0pcodeGen_Frame_components) t)->_desc2);
	((_TPPLST_0pcodeGen_Frame_components) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_Frame_components;
#else
list = (_TPPLST_0pcodeGen_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_Frame_components));
#endif
list->_prod = RULELST_0pcodeGen_Frame_components;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_Frame_components)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_FormularpcodeGen_Frame_components)
	{_TPPLST_codeGen_FormularpcodeGen_Frame_components _currn = _currn;
	 _TERMACT_LST_codeGen_FormularpcodeGen_Frame_components;
	} this code has been moved to _ElempcodeGen_Frame_components*/
	/*if (_currn->_prod == RULELST_codeGen_ParagraphpcodeGen_Frame_components)
	{_TPPLST_codeGen_ParagraphpcodeGen_Frame_components _currn = _currn;
	 _TERMACT_LST_codeGen_ParagraphpcodeGen_Frame_components;
	} this code has been moved to _ElempcodeGen_Frame_components*/
	/*if (_currn->_prod == RULELST_codeGen_HeadingpcodeGen_Frame_components)
	{_TPPLST_codeGen_HeadingpcodeGen_Frame_components _currn = _currn;
	 _TERMACT_LST_codeGen_HeadingpcodeGen_Frame_components;
	} this code has been moved to _ElempcodeGen_Frame_components*/
	/*if (_currn->_prod == RULELST_codeGen_TablepcodeGen_Frame_components)
	{_TPPLST_codeGen_TablepcodeGen_Frame_components _currn = _currn;
	 _TERMACT_LST_codeGen_TablepcodeGen_Frame_components;
	} this code has been moved to _ElempcodeGen_Frame_components*/
	/*if (_currn->_prod == RULELST_codeGen_FrameRefpcodeGen_Frame_components)
	{_TPPLST_codeGen_FrameRefpcodeGen_Frame_components _currn = _currn;
	 _TERMACT_LST_codeGen_FrameRefpcodeGen_Frame_components;
	} this code has been moved to _ElempcodeGen_Frame_components*/
	/*if (_currn->_prod == RULELST_codeGen_ListOutputElementpcodeGen_Frame_components)
	{_TPPLST_codeGen_ListOutputElementpcodeGen_Frame_components _currn = _currn;
	 _TERMACT_LST_codeGen_ListOutputElementpcodeGen_Frame_components;
	} this code has been moved to _ElempcodeGen_Frame_components*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_Frame_components;
#else
root = (_TPPpcodeGen_Frame_components) TreeNodeAlloc (sizeof (struct _TPpcodeGen_Frame_components));
#endif
root->_prod = RULEpcodeGen_Frame_components;
root->_desc1 = (_TSPLST_codeGen_Frame_components)list;
{ _TPPpcodeGen_Frame_components _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_Frame_components;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_Frame_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_Frame_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_Frame_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_Frame_styleSheetBlock;
#else
_currn = (_TPPpcodeGen_Frame_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpcodeGen_Frame_styleSheetBlock));
#endif
_currn->_prod = RULEpcodeGen_Frame_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_pcodeGen_Frame_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_Frame_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_GlobalFrame (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpcodeGen_GlobalFrame (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpcodeGen_GlobalFrame _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_GlobalFrame;
#else
_currn = (_TPPpcodeGen_GlobalFrame) TreeNodeAlloc (sizeof (struct _TPpcodeGen_GlobalFrame));
#endif
_currn->_prod = RULEpcodeGen_GlobalFrame;
_currn->_desc1 = (_TSPcodeGen_GlobalFrame_name) MkcodeGen_GlobalFrame_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_GlobalFrame: root of subtree no. 1 can not be made a codeGen_GlobalFrame_name node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_GlobalFrame_components) MkcodeGen_GlobalFrame_components (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_GlobalFrame: root of subtree no. 2 can not be made a codeGen_GlobalFrame_components node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_GlobalFrame_setSize) MkcodeGen_GlobalFrame_setSize (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_GlobalFrame: root of subtree no. 3 can not be made a codeGen_GlobalFrame_setSize node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_GlobalFrame;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_GlobalFrame */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_GlobalFrame_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_GlobalFrame_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_GlobalFrame_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_GlobalFrame_name;
#else
_currn = (_TPPpcodeGen_GlobalFrame_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_GlobalFrame_name));
#endif
_currn->_prod = RULEpcodeGen_GlobalFrame_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_GlobalFrame_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_GlobalFrame_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_FormularpcodeGen_GlobalFrame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_FormularpcodeGen_GlobalFrame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_FormularpcodeGen_GlobalFrame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_FormularpcodeGen_GlobalFrame_components;
#else
	_new = (_TPPLST_codeGen_FormularpcodeGen_GlobalFrame_components) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_FormularpcodeGen_GlobalFrame_components));
#endif
	_new->_prod = RULELST_codeGen_FormularpcodeGen_GlobalFrame_components;
	_new->_desc1 = (_TSPcodeGen_Formular)elem;
	_new->_desc2 = (_TSPLST_codeGen_GlobalFrame_components)_new;
	{_TPPLST_codeGen_FormularpcodeGen_GlobalFrame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_FormularpcodeGen_GlobalFrame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_ParagraphpcodeGen_GlobalFrame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_ParagraphpcodeGen_GlobalFrame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_ParagraphpcodeGen_GlobalFrame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_ParagraphpcodeGen_GlobalFrame_components;
#else
	_new = (_TPPLST_codeGen_ParagraphpcodeGen_GlobalFrame_components) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_ParagraphpcodeGen_GlobalFrame_components));
#endif
	_new->_prod = RULELST_codeGen_ParagraphpcodeGen_GlobalFrame_components;
	_new->_desc1 = (_TSPcodeGen_Paragraph)elem;
	_new->_desc2 = (_TSPLST_codeGen_GlobalFrame_components)_new;
	{_TPPLST_codeGen_ParagraphpcodeGen_GlobalFrame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_ParagraphpcodeGen_GlobalFrame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_HeadingpcodeGen_GlobalFrame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_HeadingpcodeGen_GlobalFrame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_HeadingpcodeGen_GlobalFrame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_HeadingpcodeGen_GlobalFrame_components;
#else
	_new = (_TPPLST_codeGen_HeadingpcodeGen_GlobalFrame_components) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_HeadingpcodeGen_GlobalFrame_components));
#endif
	_new->_prod = RULELST_codeGen_HeadingpcodeGen_GlobalFrame_components;
	_new->_desc1 = (_TSPcodeGen_Heading)elem;
	_new->_desc2 = (_TSPLST_codeGen_GlobalFrame_components)_new;
	{_TPPLST_codeGen_HeadingpcodeGen_GlobalFrame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_HeadingpcodeGen_GlobalFrame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_TablepcodeGen_GlobalFrame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_TablepcodeGen_GlobalFrame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_TablepcodeGen_GlobalFrame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_TablepcodeGen_GlobalFrame_components;
#else
	_new = (_TPPLST_codeGen_TablepcodeGen_GlobalFrame_components) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_TablepcodeGen_GlobalFrame_components));
#endif
	_new->_prod = RULELST_codeGen_TablepcodeGen_GlobalFrame_components;
	_new->_desc1 = (_TSPcodeGen_Table)elem;
	_new->_desc2 = (_TSPLST_codeGen_GlobalFrame_components)_new;
	{_TPPLST_codeGen_TablepcodeGen_GlobalFrame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_TablepcodeGen_GlobalFrame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_FrameRefpcodeGen_GlobalFrame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_FrameRefpcodeGen_GlobalFrame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_FrameRefpcodeGen_GlobalFrame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_FrameRefpcodeGen_GlobalFrame_components;
#else
	_new = (_TPPLST_codeGen_FrameRefpcodeGen_GlobalFrame_components) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_FrameRefpcodeGen_GlobalFrame_components));
#endif
	_new->_prod = RULELST_codeGen_FrameRefpcodeGen_GlobalFrame_components;
	_new->_desc1 = (_TSPcodeGen_FrameRef)elem;
	_new->_desc2 = (_TSPLST_codeGen_GlobalFrame_components)_new;
	{_TPPLST_codeGen_FrameRefpcodeGen_GlobalFrame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_FrameRefpcodeGen_GlobalFrame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_ListOutputElementpcodeGen_GlobalFrame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_ListOutputElementpcodeGen_GlobalFrame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_ListOutputElementpcodeGen_GlobalFrame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_ListOutputElementpcodeGen_GlobalFrame_components;
#else
	_new = (_TPPLST_codeGen_ListOutputElementpcodeGen_GlobalFrame_components) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_ListOutputElementpcodeGen_GlobalFrame_components));
#endif
	_new->_prod = RULELST_codeGen_ListOutputElementpcodeGen_GlobalFrame_components;
	_new->_desc1 = (_TSPcodeGen_ListOutputElement)elem;
	_new->_desc2 = (_TSPLST_codeGen_GlobalFrame_components)_new;
	{_TPPLST_codeGen_ListOutputElementpcodeGen_GlobalFrame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_ListOutputElementpcodeGen_GlobalFrame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_GlobalFrame_components (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_GlobalFrame_components (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_Formular)) 
		return _MkcodeGen_FormularpcodeGen_GlobalFrame_components (_coordref, MkcodeGen_Formular (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_Paragraph)) 
		return _MkcodeGen_ParagraphpcodeGen_GlobalFrame_components (_coordref, MkcodeGen_Paragraph (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_Heading)) 
		return _MkcodeGen_HeadingpcodeGen_GlobalFrame_components (_coordref, MkcodeGen_Heading (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_Table)) 
		return _MkcodeGen_TablepcodeGen_GlobalFrame_components (_coordref, MkcodeGen_Table (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_FrameRef)) 
		return _MkcodeGen_FrameRefpcodeGen_GlobalFrame_components (_coordref, MkcodeGen_FrameRef (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_ListOutputElement)) 
		return _MkcodeGen_ListOutputElementpcodeGen_GlobalFrame_components (_coordref, MkcodeGen_ListOutputElement (_coordref, t));
	message (DEADLY, "a pcodeGen_GlobalFrame_components list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_GlobalFrame_components (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_GlobalFrame_components (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_GlobalFrame_components)
	l =_ElempcodeGen_GlobalFrame_components (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_GlobalFrame_components)
	r = _ElempcodeGen_GlobalFrame_components (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_GlobalFrame_components)r)->_desc2);
((_TPPLST_0pcodeGen_GlobalFrame_components)r)->_desc2 = ((_TPPLST_0pcodeGen_GlobalFrame_components)l)->_desc2;
((_TPPLST_0pcodeGen_GlobalFrame_components)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_GlobalFrame_components (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_GlobalFrame_components (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_GlobalFrame_components root;
	_TPPLST_0pcodeGen_GlobalFrame_components list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_GlobalFrame_components)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_GlobalFrame_components)
		t = _ElempcodeGen_GlobalFrame_components (_coordref, t);
	remain = (_TPPLST_0pcodeGen_GlobalFrame_components)(((_TPPLST_0pcodeGen_GlobalFrame_components) t)->_desc2);
	((_TPPLST_0pcodeGen_GlobalFrame_components) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_GlobalFrame_components;
#else
list = (_TPPLST_0pcodeGen_GlobalFrame_components) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_GlobalFrame_components));
#endif
list->_prod = RULELST_0pcodeGen_GlobalFrame_components;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_GlobalFrame_components)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_FormularpcodeGen_GlobalFrame_components)
	{_TPPLST_codeGen_FormularpcodeGen_GlobalFrame_components _currn = _currn;
	 _TERMACT_LST_codeGen_FormularpcodeGen_GlobalFrame_components;
	} this code has been moved to _ElempcodeGen_GlobalFrame_components*/
	/*if (_currn->_prod == RULELST_codeGen_ParagraphpcodeGen_GlobalFrame_components)
	{_TPPLST_codeGen_ParagraphpcodeGen_GlobalFrame_components _currn = _currn;
	 _TERMACT_LST_codeGen_ParagraphpcodeGen_GlobalFrame_components;
	} this code has been moved to _ElempcodeGen_GlobalFrame_components*/
	/*if (_currn->_prod == RULELST_codeGen_HeadingpcodeGen_GlobalFrame_components)
	{_TPPLST_codeGen_HeadingpcodeGen_GlobalFrame_components _currn = _currn;
	 _TERMACT_LST_codeGen_HeadingpcodeGen_GlobalFrame_components;
	} this code has been moved to _ElempcodeGen_GlobalFrame_components*/
	/*if (_currn->_prod == RULELST_codeGen_TablepcodeGen_GlobalFrame_components)
	{_TPPLST_codeGen_TablepcodeGen_GlobalFrame_components _currn = _currn;
	 _TERMACT_LST_codeGen_TablepcodeGen_GlobalFrame_components;
	} this code has been moved to _ElempcodeGen_GlobalFrame_components*/
	/*if (_currn->_prod == RULELST_codeGen_FrameRefpcodeGen_GlobalFrame_components)
	{_TPPLST_codeGen_FrameRefpcodeGen_GlobalFrame_components _currn = _currn;
	 _TERMACT_LST_codeGen_FrameRefpcodeGen_GlobalFrame_components;
	} this code has been moved to _ElempcodeGen_GlobalFrame_components*/
	/*if (_currn->_prod == RULELST_codeGen_ListOutputElementpcodeGen_GlobalFrame_components)
	{_TPPLST_codeGen_ListOutputElementpcodeGen_GlobalFrame_components _currn = _currn;
	 _TERMACT_LST_codeGen_ListOutputElementpcodeGen_GlobalFrame_components;
	} this code has been moved to _ElempcodeGen_GlobalFrame_components*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_GlobalFrame_components;
#else
root = (_TPPpcodeGen_GlobalFrame_components) TreeNodeAlloc (sizeof (struct _TPpcodeGen_GlobalFrame_components));
#endif
root->_prod = RULEpcodeGen_GlobalFrame_components;
root->_desc1 = (_TSPLST_codeGen_GlobalFrame_components)list;
{ _TPPpcodeGen_GlobalFrame_components _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_GlobalFrame_components;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_GlobalFrame_setSize (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_GlobalFrame_setSize (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_GlobalFrame_setSize _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_GlobalFrame_setSize;
#else
_currn = (_TPPpcodeGen_GlobalFrame_setSize) TreeNodeAlloc (sizeof (struct _TPpcodeGen_GlobalFrame_setSize));
#endif
_currn->_prod = RULEpcodeGen_GlobalFrame_setSize;
_SETCOORD(_currn)
_TERMACT_pcodeGen_GlobalFrame_setSize;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_GlobalFrame_setSize */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_NavigationBar (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpcodeGen_NavigationBar (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpcodeGen_NavigationBar _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_NavigationBar;
#else
_currn = (_TPPpcodeGen_NavigationBar) TreeNodeAlloc (sizeof (struct _TPpcodeGen_NavigationBar));
#endif
_currn->_prod = RULEpcodeGen_NavigationBar;
_currn->_desc1 = (_TSPcodeGen_NavigationBar_styleSheetBlock) MkcodeGen_NavigationBar_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_NavigationBar: root of subtree no. 1 can not be made a codeGen_NavigationBar_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_NavigationBar_elements) MkcodeGen_NavigationBar_elements (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_NavigationBar: root of subtree no. 2 can not be made a codeGen_NavigationBar_elements node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_NavigationBar_alignment) MkcodeGen_NavigationBar_alignment (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_NavigationBar: root of subtree no. 3 can not be made a codeGen_NavigationBar_alignment node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_NavigationBar;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_NavigationBar */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_NavigationBar_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_NavigationBar_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_NavigationBar_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_NavigationBar_styleSheetBlock;
#else
_currn = (_TPPpcodeGen_NavigationBar_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpcodeGen_NavigationBar_styleSheetBlock));
#endif
_currn->_prod = RULEpcodeGen_NavigationBar_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_pcodeGen_NavigationBar_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_NavigationBar_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_NavigationBarElementpcodeGen_NavigationBar_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_NavigationBarElementpcodeGen_NavigationBar_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_NavigationBarElementpcodeGen_NavigationBar_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_NavigationBarElementpcodeGen_NavigationBar_elements;
#else
	_new = (_TPPLST_codeGen_NavigationBarElementpcodeGen_NavigationBar_elements) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_NavigationBarElementpcodeGen_NavigationBar_elements));
#endif
	_new->_prod = RULELST_codeGen_NavigationBarElementpcodeGen_NavigationBar_elements;
	_new->_desc1 = (_TSPcodeGen_NavigationBarElement)elem;
	_new->_desc2 = (_TSPLST_codeGen_NavigationBar_elements)_new;
	{_TPPLST_codeGen_NavigationBarElementpcodeGen_NavigationBar_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_NavigationBarElementpcodeGen_NavigationBar_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_NavigationBar_elements (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_NavigationBar_elements (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_NavigationBarElement)) 
		return _MkcodeGen_NavigationBarElementpcodeGen_NavigationBar_elements (_coordref, MkcodeGen_NavigationBarElement (_coordref, t));
	message (DEADLY, "a pcodeGen_NavigationBar_elements list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_NavigationBar_elements (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_NavigationBar_elements (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_NavigationBar_elements)
	l =_ElempcodeGen_NavigationBar_elements (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_NavigationBar_elements)
	r = _ElempcodeGen_NavigationBar_elements (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_NavigationBar_elements)r)->_desc2);
((_TPPLST_0pcodeGen_NavigationBar_elements)r)->_desc2 = ((_TPPLST_0pcodeGen_NavigationBar_elements)l)->_desc2;
((_TPPLST_0pcodeGen_NavigationBar_elements)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_NavigationBar_elements (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_NavigationBar_elements (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_NavigationBar_elements root;
	_TPPLST_0pcodeGen_NavigationBar_elements list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_NavigationBar_elements)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_NavigationBar_elements)
		t = _ElempcodeGen_NavigationBar_elements (_coordref, t);
	remain = (_TPPLST_0pcodeGen_NavigationBar_elements)(((_TPPLST_0pcodeGen_NavigationBar_elements) t)->_desc2);
	((_TPPLST_0pcodeGen_NavigationBar_elements) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_NavigationBar_elements;
#else
list = (_TPPLST_0pcodeGen_NavigationBar_elements) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_NavigationBar_elements));
#endif
list->_prod = RULELST_0pcodeGen_NavigationBar_elements;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_NavigationBar_elements)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_NavigationBarElementpcodeGen_NavigationBar_elements)
	{_TPPLST_codeGen_NavigationBarElementpcodeGen_NavigationBar_elements _currn = _currn;
	 _TERMACT_LST_codeGen_NavigationBarElementpcodeGen_NavigationBar_elements;
	} this code has been moved to _ElempcodeGen_NavigationBar_elements*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_NavigationBar_elements;
#else
root = (_TPPpcodeGen_NavigationBar_elements) TreeNodeAlloc (sizeof (struct _TPpcodeGen_NavigationBar_elements));
#endif
root->_prod = RULEpcodeGen_NavigationBar_elements;
root->_desc1 = (_TSPLST_codeGen_NavigationBar_elements)list;
{ _TPPpcodeGen_NavigationBar_elements _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_NavigationBar_elements;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_NavigationBar_alignment (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_NavigationBar_alignment (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_NavigationBar_alignment _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_NavigationBar_alignment;
#else
_currn = (_TPPpcodeGen_NavigationBar_alignment) TreeNodeAlloc (sizeof (struct _TPpcodeGen_NavigationBar_alignment));
#endif
_currn->_prod = RULEpcodeGen_NavigationBar_alignment;
_SETCOORD(_currn)
_TERMACT_pcodeGen_NavigationBar_alignment;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_NavigationBar_alignment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_NavigationBarElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpcodeGen_NavigationBarElement (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpcodeGen_NavigationBarElement _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_NavigationBarElement;
#else
_currn = (_TPPpcodeGen_NavigationBarElement) TreeNodeAlloc (sizeof (struct _TPpcodeGen_NavigationBarElement));
#endif
_currn->_prod = RULEpcodeGen_NavigationBarElement;
_currn->_desc1 = (_TSPcodeGen_NavigationBarElement_target) MkcodeGen_NavigationBarElement_target (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_NavigationBarElement: root of subtree no. 1 can not be made a codeGen_NavigationBarElement_target node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_NavigationBarElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_NavigationBarElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_NavigationBarElement_target (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_NavigationBarElement_target (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_NavigationBarElement_target _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_NavigationBarElement_target;
#else
_currn = (_TPPpcodeGen_NavigationBarElement_target) TreeNodeAlloc (sizeof (struct _TPpcodeGen_NavigationBarElement_target));
#endif
_currn->_prod = RULEpcodeGen_NavigationBarElement_target;
_SETCOORD(_currn)
_TERMACT_pcodeGen_NavigationBarElement_target;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_NavigationBarElement_target */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FrameRef (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_FrameRef (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_FrameRef _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FrameRef;
#else
_currn = (_TPPpcodeGen_FrameRef) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FrameRef));
#endif
_currn->_prod = RULEpcodeGen_FrameRef;
_currn->_desc1 = (_TSPcodeGen_FrameRef_styleSheetBlock) MkcodeGen_FrameRef_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FrameRef: root of subtree no. 1 can not be made a codeGen_FrameRef_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_FrameRef_frame) MkcodeGen_FrameRef_frame (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FrameRef: root of subtree no. 2 can not be made a codeGen_FrameRef_frame node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_FrameRef;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_FrameRef */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FrameRef_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FrameRef_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FrameRef_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FrameRef_styleSheetBlock;
#else
_currn = (_TPPpcodeGen_FrameRef_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FrameRef_styleSheetBlock));
#endif
_currn->_prod = RULEpcodeGen_FrameRef_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FrameRef_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FrameRef_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FrameRef_frame (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FrameRef_frame (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FrameRef_frame _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FrameRef_frame;
#else
_currn = (_TPPpcodeGen_FrameRef_frame) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FrameRef_frame));
#endif
_currn->_prod = RULEpcodeGen_FrameRef_frame;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FrameRef_frame;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FrameRef_frame */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_MetaInformation (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_MetaInformation (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_MetaInformation _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_MetaInformation;
#else
_currn = (_TPPpcodeGen_MetaInformation) TreeNodeAlloc (sizeof (struct _TPpcodeGen_MetaInformation));
#endif
_currn->_prod = RULEpcodeGen_MetaInformation;
_currn->_desc1 = (_TSPcodeGen_MetaInformation_name) MkcodeGen_MetaInformation_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_MetaInformation: root of subtree no. 1 can not be made a codeGen_MetaInformation_name node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_MetaInformation_elements) MkcodeGen_MetaInformation_elements (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_MetaInformation: root of subtree no. 2 can not be made a codeGen_MetaInformation_elements node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_MetaInformation;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_MetaInformation */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_MetaInformation_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_MetaInformation_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_MetaInformation_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_MetaInformation_name;
#else
_currn = (_TPPpcodeGen_MetaInformation_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_MetaInformation_name));
#endif
_currn->_prod = RULEpcodeGen_MetaInformation_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_MetaInformation_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_MetaInformation_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_MetaInformationElementpcodeGen_MetaInformation_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_MetaInformationElementpcodeGen_MetaInformation_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_MetaInformationElementpcodeGen_MetaInformation_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_MetaInformationElementpcodeGen_MetaInformation_elements;
#else
	_new = (_TPPLST_codeGen_MetaInformationElementpcodeGen_MetaInformation_elements) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_MetaInformationElementpcodeGen_MetaInformation_elements));
#endif
	_new->_prod = RULELST_codeGen_MetaInformationElementpcodeGen_MetaInformation_elements;
	_new->_desc1 = (_TSPcodeGen_MetaInformationElement)elem;
	_new->_desc2 = (_TSPLST_codeGen_MetaInformation_elements)_new;
	{_TPPLST_codeGen_MetaInformationElementpcodeGen_MetaInformation_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_MetaInformationElementpcodeGen_MetaInformation_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_MetaInformation_elements (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_MetaInformation_elements (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_MetaInformationElement)) 
		return _MkcodeGen_MetaInformationElementpcodeGen_MetaInformation_elements (_coordref, MkcodeGen_MetaInformationElement (_coordref, t));
	message (DEADLY, "a pcodeGen_MetaInformation_elements list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_MetaInformation_elements (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_MetaInformation_elements (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_MetaInformation_elements)
	l =_ElempcodeGen_MetaInformation_elements (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_MetaInformation_elements)
	r = _ElempcodeGen_MetaInformation_elements (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_MetaInformation_elements)r)->_desc2);
((_TPPLST_0pcodeGen_MetaInformation_elements)r)->_desc2 = ((_TPPLST_0pcodeGen_MetaInformation_elements)l)->_desc2;
((_TPPLST_0pcodeGen_MetaInformation_elements)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_MetaInformation_elements (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_MetaInformation_elements (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_MetaInformation_elements root;
	_TPPLST_0pcodeGen_MetaInformation_elements list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_MetaInformation_elements)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_MetaInformation_elements)
		t = _ElempcodeGen_MetaInformation_elements (_coordref, t);
	remain = (_TPPLST_0pcodeGen_MetaInformation_elements)(((_TPPLST_0pcodeGen_MetaInformation_elements) t)->_desc2);
	((_TPPLST_0pcodeGen_MetaInformation_elements) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_MetaInformation_elements;
#else
list = (_TPPLST_0pcodeGen_MetaInformation_elements) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_MetaInformation_elements));
#endif
list->_prod = RULELST_0pcodeGen_MetaInformation_elements;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_MetaInformation_elements)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_MetaInformationElementpcodeGen_MetaInformation_elements)
	{_TPPLST_codeGen_MetaInformationElementpcodeGen_MetaInformation_elements _currn = _currn;
	 _TERMACT_LST_codeGen_MetaInformationElementpcodeGen_MetaInformation_elements;
	} this code has been moved to _ElempcodeGen_MetaInformation_elements*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_MetaInformation_elements;
#else
root = (_TPPpcodeGen_MetaInformation_elements) TreeNodeAlloc (sizeof (struct _TPpcodeGen_MetaInformation_elements));
#endif
root->_prod = RULEpcodeGen_MetaInformation_elements;
root->_desc1 = (_TSPLST_codeGen_MetaInformation_elements)list;
{ _TPPpcodeGen_MetaInformation_elements _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_MetaInformation_elements;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_MetaInformationElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_MetaInformationElement (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_MetaInformationElement _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_MetaInformationElement;
#else
_currn = (_TPPpcodeGen_MetaInformationElement) TreeNodeAlloc (sizeof (struct _TPpcodeGen_MetaInformationElement));
#endif
_currn->_prod = RULEpcodeGen_MetaInformationElement;
_currn->_desc1 = (_TSPcodeGen_MetaInformationElement_property) MkcodeGen_MetaInformationElement_property (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_MetaInformationElement: root of subtree no. 1 can not be made a codeGen_MetaInformationElement_property node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_MetaInformationElement_value) MkcodeGen_MetaInformationElement_value (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_MetaInformationElement: root of subtree no. 2 can not be made a codeGen_MetaInformationElement_value node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_MetaInformationElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_MetaInformationElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_MetaInformationElement_property (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_MetaInformationElement_property (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_MetaInformationElement_property _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_MetaInformationElement_property;
#else
_currn = (_TPPpcodeGen_MetaInformationElement_property) TreeNodeAlloc (sizeof (struct _TPpcodeGen_MetaInformationElement_property));
#endif
_currn->_prod = RULEpcodeGen_MetaInformationElement_property;
_SETCOORD(_currn)
_TERMACT_pcodeGen_MetaInformationElement_property;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_MetaInformationElement_property */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_MetaInformationElement_value (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_MetaInformationElement_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_MetaInformationElement_value _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_MetaInformationElement_value;
#else
_currn = (_TPPpcodeGen_MetaInformationElement_value) TreeNodeAlloc (sizeof (struct _TPpcodeGen_MetaInformationElement_value));
#endif
_currn->_prod = RULEpcodeGen_MetaInformationElement_value;
_SETCOORD(_currn)
_TERMACT_pcodeGen_MetaInformationElement_value;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_MetaInformationElement_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_OutputElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_OutputElement (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_OutputElement _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_OutputElement;
#else
_currn = (_TPPpcodeGen_OutputElement) TreeNodeAlloc (sizeof (struct _TPpcodeGen_OutputElement));
#endif
_currn->_prod = RULEpcodeGen_OutputElement;
_currn->_desc1 = (_TSPcodeGen_OutputElement_link) MkcodeGen_OutputElement_link (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_OutputElement: root of subtree no. 1 can not be made a codeGen_OutputElement_link node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_OutputElement_content) MkcodeGen_OutputElement_content (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_OutputElement: root of subtree no. 2 can not be made a codeGen_OutputElement_content node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_OutputElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_OutputElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_OutputElement_link (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_OutputElement_link (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_OutputElement_link _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_OutputElement_link;
#else
_currn = (_TPPpcodeGen_OutputElement_link) TreeNodeAlloc (sizeof (struct _TPpcodeGen_OutputElement_link));
#endif
_currn->_prod = RULEpcodeGen_OutputElement_link;
_SETCOORD(_currn)
_TERMACT_pcodeGen_OutputElement_link;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_OutputElement_link */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_OutputElement_content (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_OutputElement_content (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_OutputElement_content _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_OutputElement_content;
#else
_currn = (_TPPpcodeGen_OutputElement_content) TreeNodeAlloc (sizeof (struct _TPpcodeGen_OutputElement_content));
#endif
_currn->_prod = RULEpcodeGen_OutputElement_content;
_SETCOORD(_currn)
_TERMACT_pcodeGen_OutputElement_content;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_OutputElement_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ListOutputElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkpcodeGen_ListOutputElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPpcodeGen_ListOutputElement _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ListOutputElement;
#else
_currn = (_TPPpcodeGen_ListOutputElement) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ListOutputElement));
#endif
_currn->_prod = RULEpcodeGen_ListOutputElement;
_currn->_desc1 = (_TSPcodeGen_ListOutputElement_styleSheetBlock) MkcodeGen_ListOutputElement_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_ListOutputElement: root of subtree no. 1 can not be made a codeGen_ListOutputElement_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_ListOutputElement_dataSourceActualParams) MkcodeGen_ListOutputElement_dataSourceActualParams (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_ListOutputElement: root of subtree no. 2 can not be made a codeGen_ListOutputElement_dataSourceActualParams node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_ListOutputElement_dataSource) MkcodeGen_ListOutputElement_dataSource (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_ListOutputElement: root of subtree no. 3 can not be made a codeGen_ListOutputElement_dataSource node ", 0, _coordref);
_currn->_desc4 = (_TSPcodeGen_ListOutputElement_maxNumberOfElements) MkcodeGen_ListOutputElement_maxNumberOfElements (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_ListOutputElement: root of subtree no. 4 can not be made a codeGen_ListOutputElement_maxNumberOfElements node ", 0, _coordref);
_currn->_desc5 = (_TSPcodeGen_ListOutputElement_navigationElement) MkcodeGen_ListOutputElement_navigationElement (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_ListOutputElement: root of subtree no. 5 can not be made a codeGen_ListOutputElement_navigationElement node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_ListOutputElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_ListOutputElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ListOutputElement_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_ListOutputElement_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_ListOutputElement_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ListOutputElement_styleSheetBlock;
#else
_currn = (_TPPpcodeGen_ListOutputElement_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ListOutputElement_styleSheetBlock));
#endif
_currn->_prod = RULEpcodeGen_ListOutputElement_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_pcodeGen_ListOutputElement_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_ListOutputElement_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_DataSourceActualParampcodeGen_ListOutputElement_dataSourceActualParams (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_DataSourceActualParampcodeGen_ListOutputElement_dataSourceActualParams (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_DataSourceActualParampcodeGen_ListOutputElement_dataSourceActualParams _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_DataSourceActualParampcodeGen_ListOutputElement_dataSourceActualParams;
#else
	_new = (_TPPLST_codeGen_DataSourceActualParampcodeGen_ListOutputElement_dataSourceActualParams) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_DataSourceActualParampcodeGen_ListOutputElement_dataSourceActualParams));
#endif
	_new->_prod = RULELST_codeGen_DataSourceActualParampcodeGen_ListOutputElement_dataSourceActualParams;
	_new->_desc1 = (_TSPcodeGen_DataSourceActualParam)elem;
	_new->_desc2 = (_TSPLST_codeGen_ListOutputElement_dataSourceActualParams)_new;
	{_TPPLST_codeGen_DataSourceActualParampcodeGen_ListOutputElement_dataSourceActualParams _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_DataSourceActualParampcodeGen_ListOutputElement_dataSourceActualParams;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_ListOutputElement_dataSourceActualParams (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_ListOutputElement_dataSourceActualParams (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_DataSourceActualParam)) 
		return _MkcodeGen_DataSourceActualParampcodeGen_ListOutputElement_dataSourceActualParams (_coordref, MkcodeGen_DataSourceActualParam (_coordref, t));
	message (DEADLY, "a pcodeGen_ListOutputElement_dataSourceActualParams list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_ListOutputElement_dataSourceActualParams (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_ListOutputElement_dataSourceActualParams (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_ListOutputElement_dataSourceActualParams)
	l =_ElempcodeGen_ListOutputElement_dataSourceActualParams (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_ListOutputElement_dataSourceActualParams)
	r = _ElempcodeGen_ListOutputElement_dataSourceActualParams (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_ListOutputElement_dataSourceActualParams)r)->_desc2);
((_TPPLST_0pcodeGen_ListOutputElement_dataSourceActualParams)r)->_desc2 = ((_TPPLST_0pcodeGen_ListOutputElement_dataSourceActualParams)l)->_desc2;
((_TPPLST_0pcodeGen_ListOutputElement_dataSourceActualParams)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ListOutputElement_dataSourceActualParams (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_ListOutputElement_dataSourceActualParams (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_ListOutputElement_dataSourceActualParams root;
	_TPPLST_0pcodeGen_ListOutputElement_dataSourceActualParams list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_ListOutputElement_dataSourceActualParams)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_ListOutputElement_dataSourceActualParams)
		t = _ElempcodeGen_ListOutputElement_dataSourceActualParams (_coordref, t);
	remain = (_TPPLST_0pcodeGen_ListOutputElement_dataSourceActualParams)(((_TPPLST_0pcodeGen_ListOutputElement_dataSourceActualParams) t)->_desc2);
	((_TPPLST_0pcodeGen_ListOutputElement_dataSourceActualParams) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_ListOutputElement_dataSourceActualParams;
#else
list = (_TPPLST_0pcodeGen_ListOutputElement_dataSourceActualParams) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_ListOutputElement_dataSourceActualParams));
#endif
list->_prod = RULELST_0pcodeGen_ListOutputElement_dataSourceActualParams;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_ListOutputElement_dataSourceActualParams)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_DataSourceActualParampcodeGen_ListOutputElement_dataSourceActualParams)
	{_TPPLST_codeGen_DataSourceActualParampcodeGen_ListOutputElement_dataSourceActualParams _currn = _currn;
	 _TERMACT_LST_codeGen_DataSourceActualParampcodeGen_ListOutputElement_dataSourceActualParams;
	} this code has been moved to _ElempcodeGen_ListOutputElement_dataSourceActualParams*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_ListOutputElement_dataSourceActualParams;
#else
root = (_TPPpcodeGen_ListOutputElement_dataSourceActualParams) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ListOutputElement_dataSourceActualParams));
#endif
root->_prod = RULEpcodeGen_ListOutputElement_dataSourceActualParams;
root->_desc1 = (_TSPLST_codeGen_ListOutputElement_dataSourceActualParams)list;
{ _TPPpcodeGen_ListOutputElement_dataSourceActualParams _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_ListOutputElement_dataSourceActualParams;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ListOutputElement_dataSource (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_ListOutputElement_dataSource (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_ListOutputElement_dataSource _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ListOutputElement_dataSource;
#else
_currn = (_TPPpcodeGen_ListOutputElement_dataSource) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ListOutputElement_dataSource));
#endif
_currn->_prod = RULEpcodeGen_ListOutputElement_dataSource;
_SETCOORD(_currn)
_TERMACT_pcodeGen_ListOutputElement_dataSource;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_ListOutputElement_dataSource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ListOutputElement_maxNumberOfElements (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_ListOutputElement_maxNumberOfElements (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_ListOutputElement_maxNumberOfElements _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ListOutputElement_maxNumberOfElements;
#else
_currn = (_TPPpcodeGen_ListOutputElement_maxNumberOfElements) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ListOutputElement_maxNumberOfElements));
#endif
_currn->_prod = RULEpcodeGen_ListOutputElement_maxNumberOfElements;
_SETCOORD(_currn)
_TERMACT_pcodeGen_ListOutputElement_maxNumberOfElements;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_ListOutputElement_maxNumberOfElements */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ListOutputElement_navigationElement (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_ListOutputElement_navigationElement (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_ListOutputElement_navigationElement _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ListOutputElement_navigationElement;
#else
_currn = (_TPPpcodeGen_ListOutputElement_navigationElement) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ListOutputElement_navigationElement));
#endif
_currn->_prod = RULEpcodeGen_ListOutputElement_navigationElement;
_SETCOORD(_currn)
_TERMACT_pcodeGen_ListOutputElement_navigationElement;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_ListOutputElement_navigationElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_StyleSheet (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpcodeGen_StyleSheet (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpcodeGen_StyleSheet _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_StyleSheet;
#else
_currn = (_TPPpcodeGen_StyleSheet) TreeNodeAlloc (sizeof (struct _TPpcodeGen_StyleSheet));
#endif
_currn->_prod = RULEpcodeGen_StyleSheet;
_currn->_desc1 = (_TSPcodeGen_StyleSheet_name) MkcodeGen_StyleSheet_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_StyleSheet: root of subtree no. 1 can not be made a codeGen_StyleSheet_name node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_StyleSheet_custom) MkcodeGen_StyleSheet_custom (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_StyleSheet: root of subtree no. 2 can not be made a codeGen_StyleSheet_custom node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_StyleSheet_default) MkcodeGen_StyleSheet_default (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_StyleSheet: root of subtree no. 3 can not be made a codeGen_StyleSheet_default node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_StyleSheet;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_StyleSheet */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_StyleSheet_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_StyleSheet_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_StyleSheet_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_StyleSheet_name;
#else
_currn = (_TPPpcodeGen_StyleSheet_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_StyleSheet_name));
#endif
_currn->_prod = RULEpcodeGen_StyleSheet_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_StyleSheet_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_StyleSheet_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_CustomStyleSheetBlockpcodeGen_StyleSheet_custom (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_CustomStyleSheetBlockpcodeGen_StyleSheet_custom (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_CustomStyleSheetBlockpcodeGen_StyleSheet_custom _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_CustomStyleSheetBlockpcodeGen_StyleSheet_custom;
#else
	_new = (_TPPLST_codeGen_CustomStyleSheetBlockpcodeGen_StyleSheet_custom) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_CustomStyleSheetBlockpcodeGen_StyleSheet_custom));
#endif
	_new->_prod = RULELST_codeGen_CustomStyleSheetBlockpcodeGen_StyleSheet_custom;
	_new->_desc1 = (_TSPcodeGen_CustomStyleSheetBlock)elem;
	_new->_desc2 = (_TSPLST_codeGen_StyleSheet_custom)_new;
	{_TPPLST_codeGen_CustomStyleSheetBlockpcodeGen_StyleSheet_custom _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_CustomStyleSheetBlockpcodeGen_StyleSheet_custom;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_StyleSheet_custom (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_StyleSheet_custom (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_CustomStyleSheetBlock)) 
		return _MkcodeGen_CustomStyleSheetBlockpcodeGen_StyleSheet_custom (_coordref, MkcodeGen_CustomStyleSheetBlock (_coordref, t));
	message (DEADLY, "a pcodeGen_StyleSheet_custom list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_StyleSheet_custom (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_StyleSheet_custom (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_StyleSheet_custom)
	l =_ElempcodeGen_StyleSheet_custom (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_StyleSheet_custom)
	r = _ElempcodeGen_StyleSheet_custom (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_StyleSheet_custom)r)->_desc2);
((_TPPLST_0pcodeGen_StyleSheet_custom)r)->_desc2 = ((_TPPLST_0pcodeGen_StyleSheet_custom)l)->_desc2;
((_TPPLST_0pcodeGen_StyleSheet_custom)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_StyleSheet_custom (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_StyleSheet_custom (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_StyleSheet_custom root;
	_TPPLST_0pcodeGen_StyleSheet_custom list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_StyleSheet_custom)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_StyleSheet_custom)
		t = _ElempcodeGen_StyleSheet_custom (_coordref, t);
	remain = (_TPPLST_0pcodeGen_StyleSheet_custom)(((_TPPLST_0pcodeGen_StyleSheet_custom) t)->_desc2);
	((_TPPLST_0pcodeGen_StyleSheet_custom) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_StyleSheet_custom;
#else
list = (_TPPLST_0pcodeGen_StyleSheet_custom) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_StyleSheet_custom));
#endif
list->_prod = RULELST_0pcodeGen_StyleSheet_custom;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_StyleSheet_custom)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_CustomStyleSheetBlockpcodeGen_StyleSheet_custom)
	{_TPPLST_codeGen_CustomStyleSheetBlockpcodeGen_StyleSheet_custom _currn = _currn;
	 _TERMACT_LST_codeGen_CustomStyleSheetBlockpcodeGen_StyleSheet_custom;
	} this code has been moved to _ElempcodeGen_StyleSheet_custom*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_StyleSheet_custom;
#else
root = (_TPPpcodeGen_StyleSheet_custom) TreeNodeAlloc (sizeof (struct _TPpcodeGen_StyleSheet_custom));
#endif
root->_prod = RULEpcodeGen_StyleSheet_custom;
root->_desc1 = (_TSPLST_codeGen_StyleSheet_custom)list;
{ _TPPpcodeGen_StyleSheet_custom _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_StyleSheet_custom;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_HtmlStyleSheetBlockpcodeGen_StyleSheet_default (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_HtmlStyleSheetBlockpcodeGen_StyleSheet_default (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_HtmlStyleSheetBlockpcodeGen_StyleSheet_default _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_HtmlStyleSheetBlockpcodeGen_StyleSheet_default;
#else
	_new = (_TPPLST_codeGen_HtmlStyleSheetBlockpcodeGen_StyleSheet_default) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_HtmlStyleSheetBlockpcodeGen_StyleSheet_default));
#endif
	_new->_prod = RULELST_codeGen_HtmlStyleSheetBlockpcodeGen_StyleSheet_default;
	_new->_desc1 = (_TSPcodeGen_HtmlStyleSheetBlock)elem;
	_new->_desc2 = (_TSPLST_codeGen_StyleSheet_default)_new;
	{_TPPLST_codeGen_HtmlStyleSheetBlockpcodeGen_StyleSheet_default _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_HtmlStyleSheetBlockpcodeGen_StyleSheet_default;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_StyleSheet_default (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_StyleSheet_default (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_HtmlStyleSheetBlock)) 
		return _MkcodeGen_HtmlStyleSheetBlockpcodeGen_StyleSheet_default (_coordref, MkcodeGen_HtmlStyleSheetBlock (_coordref, t));
	message (DEADLY, "a pcodeGen_StyleSheet_default list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_StyleSheet_default (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_StyleSheet_default (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_StyleSheet_default)
	l =_ElempcodeGen_StyleSheet_default (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_StyleSheet_default)
	r = _ElempcodeGen_StyleSheet_default (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_StyleSheet_default)r)->_desc2);
((_TPPLST_0pcodeGen_StyleSheet_default)r)->_desc2 = ((_TPPLST_0pcodeGen_StyleSheet_default)l)->_desc2;
((_TPPLST_0pcodeGen_StyleSheet_default)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_StyleSheet_default (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_StyleSheet_default (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_StyleSheet_default root;
	_TPPLST_0pcodeGen_StyleSheet_default list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_StyleSheet_default)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_StyleSheet_default)
		t = _ElempcodeGen_StyleSheet_default (_coordref, t);
	remain = (_TPPLST_0pcodeGen_StyleSheet_default)(((_TPPLST_0pcodeGen_StyleSheet_default) t)->_desc2);
	((_TPPLST_0pcodeGen_StyleSheet_default) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_StyleSheet_default;
#else
list = (_TPPLST_0pcodeGen_StyleSheet_default) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_StyleSheet_default));
#endif
list->_prod = RULELST_0pcodeGen_StyleSheet_default;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_StyleSheet_default)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_HtmlStyleSheetBlockpcodeGen_StyleSheet_default)
	{_TPPLST_codeGen_HtmlStyleSheetBlockpcodeGen_StyleSheet_default _currn = _currn;
	 _TERMACT_LST_codeGen_HtmlStyleSheetBlockpcodeGen_StyleSheet_default;
	} this code has been moved to _ElempcodeGen_StyleSheet_default*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_StyleSheet_default;
#else
root = (_TPPpcodeGen_StyleSheet_default) TreeNodeAlloc (sizeof (struct _TPpcodeGen_StyleSheet_default));
#endif
root->_prod = RULEpcodeGen_StyleSheet_default;
root->_desc1 = (_TSPLST_codeGen_StyleSheet_default)list;
{ _TPPpcodeGen_StyleSheet_default _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_StyleSheet_default;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_CustomStyleSheetBlock (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_CustomStyleSheetBlock (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_CustomStyleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_CustomStyleSheetBlock;
#else
_currn = (_TPPpcodeGen_CustomStyleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpcodeGen_CustomStyleSheetBlock));
#endif
_currn->_prod = RULEpcodeGen_CustomStyleSheetBlock;
_currn->_desc1 = (_TSPcodeGen_CustomStyleSheetBlock_pairs) MkcodeGen_CustomStyleSheetBlock_pairs (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_CustomStyleSheetBlock: root of subtree no. 1 can not be made a codeGen_CustomStyleSheetBlock_pairs node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_CustomStyleSheetBlock_name) MkcodeGen_CustomStyleSheetBlock_name (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_CustomStyleSheetBlock: root of subtree no. 2 can not be made a codeGen_CustomStyleSheetBlock_name node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_CustomStyleSheetBlock;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_CustomStyleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_StyleSheetPairpcodeGen_CustomStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_StyleSheetPairpcodeGen_CustomStyleSheetBlock_pairs (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_StyleSheetPairpcodeGen_CustomStyleSheetBlock_pairs _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_StyleSheetPairpcodeGen_CustomStyleSheetBlock_pairs;
#else
	_new = (_TPPLST_codeGen_StyleSheetPairpcodeGen_CustomStyleSheetBlock_pairs) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_StyleSheetPairpcodeGen_CustomStyleSheetBlock_pairs));
#endif
	_new->_prod = RULELST_codeGen_StyleSheetPairpcodeGen_CustomStyleSheetBlock_pairs;
	_new->_desc1 = (_TSPcodeGen_StyleSheetPair)elem;
	_new->_desc2 = (_TSPLST_codeGen_CustomStyleSheetBlock_pairs)_new;
	{_TPPLST_codeGen_StyleSheetPairpcodeGen_CustomStyleSheetBlock_pairs _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_StyleSheetPairpcodeGen_CustomStyleSheetBlock_pairs;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_CustomStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_CustomStyleSheetBlock_pairs (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_StyleSheetPair)) 
		return _MkcodeGen_StyleSheetPairpcodeGen_CustomStyleSheetBlock_pairs (_coordref, MkcodeGen_StyleSheetPair (_coordref, t));
	message (DEADLY, "a pcodeGen_CustomStyleSheetBlock_pairs list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_CustomStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_CustomStyleSheetBlock_pairs (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_CustomStyleSheetBlock_pairs)
	l =_ElempcodeGen_CustomStyleSheetBlock_pairs (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_CustomStyleSheetBlock_pairs)
	r = _ElempcodeGen_CustomStyleSheetBlock_pairs (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_CustomStyleSheetBlock_pairs)r)->_desc2);
((_TPPLST_0pcodeGen_CustomStyleSheetBlock_pairs)r)->_desc2 = ((_TPPLST_0pcodeGen_CustomStyleSheetBlock_pairs)l)->_desc2;
((_TPPLST_0pcodeGen_CustomStyleSheetBlock_pairs)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_CustomStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_CustomStyleSheetBlock_pairs (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_CustomStyleSheetBlock_pairs root;
	_TPPLST_0pcodeGen_CustomStyleSheetBlock_pairs list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_CustomStyleSheetBlock_pairs)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_CustomStyleSheetBlock_pairs)
		t = _ElempcodeGen_CustomStyleSheetBlock_pairs (_coordref, t);
	remain = (_TPPLST_0pcodeGen_CustomStyleSheetBlock_pairs)(((_TPPLST_0pcodeGen_CustomStyleSheetBlock_pairs) t)->_desc2);
	((_TPPLST_0pcodeGen_CustomStyleSheetBlock_pairs) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_CustomStyleSheetBlock_pairs;
#else
list = (_TPPLST_0pcodeGen_CustomStyleSheetBlock_pairs) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_CustomStyleSheetBlock_pairs));
#endif
list->_prod = RULELST_0pcodeGen_CustomStyleSheetBlock_pairs;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_CustomStyleSheetBlock_pairs)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_StyleSheetPairpcodeGen_CustomStyleSheetBlock_pairs)
	{_TPPLST_codeGen_StyleSheetPairpcodeGen_CustomStyleSheetBlock_pairs _currn = _currn;
	 _TERMACT_LST_codeGen_StyleSheetPairpcodeGen_CustomStyleSheetBlock_pairs;
	} this code has been moved to _ElempcodeGen_CustomStyleSheetBlock_pairs*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_CustomStyleSheetBlock_pairs;
#else
root = (_TPPpcodeGen_CustomStyleSheetBlock_pairs) TreeNodeAlloc (sizeof (struct _TPpcodeGen_CustomStyleSheetBlock_pairs));
#endif
root->_prod = RULEpcodeGen_CustomStyleSheetBlock_pairs;
root->_desc1 = (_TSPLST_codeGen_CustomStyleSheetBlock_pairs)list;
{ _TPPpcodeGen_CustomStyleSheetBlock_pairs _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_CustomStyleSheetBlock_pairs;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_CustomStyleSheetBlock_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_CustomStyleSheetBlock_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_CustomStyleSheetBlock_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_CustomStyleSheetBlock_name;
#else
_currn = (_TPPpcodeGen_CustomStyleSheetBlock_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_CustomStyleSheetBlock_name));
#endif
_currn->_prod = RULEpcodeGen_CustomStyleSheetBlock_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_CustomStyleSheetBlock_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_CustomStyleSheetBlock_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_StyleSheetPair (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4)
#else
NODEPTR MkpcodeGen_StyleSheetPair (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
#endif
{	_TPPpcodeGen_StyleSheetPair _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_StyleSheetPair;
#else
_currn = (_TPPpcodeGen_StyleSheetPair) TreeNodeAlloc (sizeof (struct _TPpcodeGen_StyleSheetPair));
#endif
_currn->_prod = RULEpcodeGen_StyleSheetPair;
_currn->_desc1 = (_TSPcodeGen_StyleSheetPair_category) MkcodeGen_StyleSheetPair_category (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_StyleSheetPair: root of subtree no. 1 can not be made a codeGen_StyleSheetPair_category node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_StyleSheetPair_attribute) MkcodeGen_StyleSheetPair_attribute (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_StyleSheetPair: root of subtree no. 2 can not be made a codeGen_StyleSheetPair_attribute node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_StyleSheetPair_valueCategory) MkcodeGen_StyleSheetPair_valueCategory (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_StyleSheetPair: root of subtree no. 3 can not be made a codeGen_StyleSheetPair_valueCategory node ", 0, _coordref);
_currn->_desc4 = (_TSPcodeGen_StyleSheetPair_value) MkcodeGen_StyleSheetPair_value (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_StyleSheetPair: root of subtree no. 4 can not be made a codeGen_StyleSheetPair_value node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_StyleSheetPair;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_StyleSheetPair */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_StyleSheetPair_category (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_StyleSheetPair_category (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_StyleSheetPair_category _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_StyleSheetPair_category;
#else
_currn = (_TPPpcodeGen_StyleSheetPair_category) TreeNodeAlloc (sizeof (struct _TPpcodeGen_StyleSheetPair_category));
#endif
_currn->_prod = RULEpcodeGen_StyleSheetPair_category;
_SETCOORD(_currn)
_TERMACT_pcodeGen_StyleSheetPair_category;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_StyleSheetPair_category */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_StyleSheetPair_attribute (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_StyleSheetPair_attribute (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_StyleSheetPair_attribute _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_StyleSheetPair_attribute;
#else
_currn = (_TPPpcodeGen_StyleSheetPair_attribute) TreeNodeAlloc (sizeof (struct _TPpcodeGen_StyleSheetPair_attribute));
#endif
_currn->_prod = RULEpcodeGen_StyleSheetPair_attribute;
_SETCOORD(_currn)
_TERMACT_pcodeGen_StyleSheetPair_attribute;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_StyleSheetPair_attribute */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_StyleSheetPair_valueCategory (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_StyleSheetPair_valueCategory (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_StyleSheetPair_valueCategory _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_StyleSheetPair_valueCategory;
#else
_currn = (_TPPpcodeGen_StyleSheetPair_valueCategory) TreeNodeAlloc (sizeof (struct _TPpcodeGen_StyleSheetPair_valueCategory));
#endif
_currn->_prod = RULEpcodeGen_StyleSheetPair_valueCategory;
_SETCOORD(_currn)
_TERMACT_pcodeGen_StyleSheetPair_valueCategory;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_StyleSheetPair_valueCategory */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_SizeValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_SizeValue (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_SizeValue _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_SizeValue;
#else
_currn = (_TPPpcodeGen_SizeValue) TreeNodeAlloc (sizeof (struct _TPpcodeGen_SizeValue));
#endif
_currn->_prod = RULEpcodeGen_SizeValue;
_currn->_desc1 = (_TSPcodeGen_SizeValue_measure) MkcodeGen_SizeValue_measure (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_SizeValue: root of subtree no. 1 can not be made a codeGen_SizeValue_measure node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_SizeValue_value) MkcodeGen_SizeValue_value (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_SizeValue: root of subtree no. 2 can not be made a codeGen_SizeValue_value node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_SizeValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_SizeValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_SizeValue_measure (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_SizeValue_measure (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_SizeValue_measure _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_SizeValue_measure;
#else
_currn = (_TPPpcodeGen_SizeValue_measure) TreeNodeAlloc (sizeof (struct _TPpcodeGen_SizeValue_measure));
#endif
_currn->_prod = RULEpcodeGen_SizeValue_measure;
_SETCOORD(_currn)
_TERMACT_pcodeGen_SizeValue_measure;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_SizeValue_measure */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_SizeValue_value (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_SizeValue_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_SizeValue_value _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_SizeValue_value;
#else
_currn = (_TPPpcodeGen_SizeValue_value) TreeNodeAlloc (sizeof (struct _TPpcodeGen_SizeValue_value));
#endif
_currn->_prod = RULEpcodeGen_SizeValue_value;
_SETCOORD(_currn)
_TERMACT_pcodeGen_SizeValue_value;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_SizeValue_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ColorValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpcodeGen_ColorValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpcodeGen_ColorValue _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ColorValue;
#else
_currn = (_TPPpcodeGen_ColorValue) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ColorValue));
#endif
_currn->_prod = RULEpcodeGen_ColorValue;
_currn->_desc1 = (_TSPcodeGen_ColorValue_color) MkcodeGen_ColorValue_color (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_ColorValue: root of subtree no. 1 can not be made a codeGen_ColorValue_color node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_ColorValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_ColorValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ColorValue_color (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_ColorValue_color (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_ColorValue_color _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ColorValue_color;
#else
_currn = (_TPPpcodeGen_ColorValue_color) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ColorValue_color));
#endif
_currn->_prod = RULEpcodeGen_ColorValue_color;
_SETCOORD(_currn)
_TERMACT_pcodeGen_ColorValue_color;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_ColorValue_color */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FamilyValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpcodeGen_FamilyValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpcodeGen_FamilyValue _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FamilyValue;
#else
_currn = (_TPPpcodeGen_FamilyValue) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FamilyValue));
#endif
_currn->_prod = RULEpcodeGen_FamilyValue;
_currn->_desc1 = (_TSPcodeGen_FamilyValue_family) MkcodeGen_FamilyValue_family (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_FamilyValue: root of subtree no. 1 can not be made a codeGen_FamilyValue_family node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_FamilyValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_FamilyValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_FamilyValue_family (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_FamilyValue_family (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_FamilyValue_family _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_FamilyValue_family;
#else
_currn = (_TPPpcodeGen_FamilyValue_family) TreeNodeAlloc (sizeof (struct _TPpcodeGen_FamilyValue_family));
#endif
_currn->_prod = RULEpcodeGen_FamilyValue_family;
_SETCOORD(_currn)
_TERMACT_pcodeGen_FamilyValue_family;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_FamilyValue_family */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_StyleValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpcodeGen_StyleValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpcodeGen_StyleValue _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_StyleValue;
#else
_currn = (_TPPpcodeGen_StyleValue) TreeNodeAlloc (sizeof (struct _TPpcodeGen_StyleValue));
#endif
_currn->_prod = RULEpcodeGen_StyleValue;
_currn->_desc1 = (_TSPcodeGen_StyleValue_style) MkcodeGen_StyleValue_style (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_StyleValue: root of subtree no. 1 can not be made a codeGen_StyleValue_style node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_StyleValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_StyleValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_StyleValue_style (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_StyleValue_style (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_StyleValue_style _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_StyleValue_style;
#else
_currn = (_TPPpcodeGen_StyleValue_style) TreeNodeAlloc (sizeof (struct _TPpcodeGen_StyleValue_style));
#endif
_currn->_prod = RULEpcodeGen_StyleValue_style;
_SETCOORD(_currn)
_TERMACT_pcodeGen_StyleValue_style;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_StyleValue_style */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_HeadingValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpcodeGen_HeadingValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpcodeGen_HeadingValue _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_HeadingValue;
#else
_currn = (_TPPpcodeGen_HeadingValue) TreeNodeAlloc (sizeof (struct _TPpcodeGen_HeadingValue));
#endif
_currn->_prod = RULEpcodeGen_HeadingValue;
_currn->_desc1 = (_TSPcodeGen_HeadingValue_heading) MkcodeGen_HeadingValue_heading (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_HeadingValue: root of subtree no. 1 can not be made a codeGen_HeadingValue_heading node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_HeadingValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_HeadingValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_HeadingValue_heading (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_HeadingValue_heading (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_HeadingValue_heading _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_HeadingValue_heading;
#else
_currn = (_TPPpcodeGen_HeadingValue_heading) TreeNodeAlloc (sizeof (struct _TPpcodeGen_HeadingValue_heading));
#endif
_currn->_prod = RULEpcodeGen_HeadingValue_heading;
_SETCOORD(_currn)
_TERMACT_pcodeGen_HeadingValue_heading;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_HeadingValue_heading */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_AlignmentValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpcodeGen_AlignmentValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpcodeGen_AlignmentValue _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_AlignmentValue;
#else
_currn = (_TPPpcodeGen_AlignmentValue) TreeNodeAlloc (sizeof (struct _TPpcodeGen_AlignmentValue));
#endif
_currn->_prod = RULEpcodeGen_AlignmentValue;
_currn->_desc1 = (_TSPcodeGen_AlignmentValue_alignment) MkcodeGen_AlignmentValue_alignment (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_AlignmentValue: root of subtree no. 1 can not be made a codeGen_AlignmentValue_alignment node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_AlignmentValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_AlignmentValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_AlignmentValue_alignment (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_AlignmentValue_alignment (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_AlignmentValue_alignment _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_AlignmentValue_alignment;
#else
_currn = (_TPPpcodeGen_AlignmentValue_alignment) TreeNodeAlloc (sizeof (struct _TPpcodeGen_AlignmentValue_alignment));
#endif
_currn->_prod = RULEpcodeGen_AlignmentValue_alignment;
_SETCOORD(_currn)
_TERMACT_pcodeGen_AlignmentValue_alignment;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_AlignmentValue_alignment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ListValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpcodeGen_ListValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpcodeGen_ListValue _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ListValue;
#else
_currn = (_TPPpcodeGen_ListValue) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ListValue));
#endif
_currn->_prod = RULEpcodeGen_ListValue;
_currn->_desc1 = (_TSPcodeGen_ListValue_listed) MkcodeGen_ListValue_listed (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_ListValue: root of subtree no. 1 can not be made a codeGen_ListValue_listed node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_ListValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_ListValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ListValue_listed (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_ListValue_listed (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_ListValue_listed _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ListValue_listed;
#else
_currn = (_TPPpcodeGen_ListValue_listed) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ListValue_listed));
#endif
_currn->_prod = RULEpcodeGen_ListValue_listed;
_SETCOORD(_currn)
_TERMACT_pcodeGen_ListValue_listed;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_ListValue_listed */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_CustomValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpcodeGen_CustomValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpcodeGen_CustomValue _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_CustomValue;
#else
_currn = (_TPPpcodeGen_CustomValue) TreeNodeAlloc (sizeof (struct _TPpcodeGen_CustomValue));
#endif
_currn->_prod = RULEpcodeGen_CustomValue;
_currn->_desc1 = (_TSPcodeGen_CustomValue_value) MkcodeGen_CustomValue_value (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_CustomValue: root of subtree no. 1 can not be made a codeGen_CustomValue_value node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_CustomValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_CustomValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_CustomValue_value (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_CustomValue_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_CustomValue_value _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_CustomValue_value;
#else
_currn = (_TPPpcodeGen_CustomValue_value) TreeNodeAlloc (sizeof (struct _TPpcodeGen_CustomValue_value));
#endif
_currn->_prod = RULEpcodeGen_CustomValue_value;
_SETCOORD(_currn)
_TERMACT_pcodeGen_CustomValue_value;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_CustomValue_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_HtmlStyleSheetBlock (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_HtmlStyleSheetBlock (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_HtmlStyleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_HtmlStyleSheetBlock;
#else
_currn = (_TPPpcodeGen_HtmlStyleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpcodeGen_HtmlStyleSheetBlock));
#endif
_currn->_prod = RULEpcodeGen_HtmlStyleSheetBlock;
_currn->_desc1 = (_TSPcodeGen_HtmlStyleSheetBlock_pairs) MkcodeGen_HtmlStyleSheetBlock_pairs (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_HtmlStyleSheetBlock: root of subtree no. 1 can not be made a codeGen_HtmlStyleSheetBlock_pairs node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_HtmlStyleSheetBlock_name) MkcodeGen_HtmlStyleSheetBlock_name (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_HtmlStyleSheetBlock: root of subtree no. 2 can not be made a codeGen_HtmlStyleSheetBlock_name node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_HtmlStyleSheetBlock;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_HtmlStyleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_StyleSheetPairpcodeGen_HtmlStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_StyleSheetPairpcodeGen_HtmlStyleSheetBlock_pairs (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_StyleSheetPairpcodeGen_HtmlStyleSheetBlock_pairs _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_StyleSheetPairpcodeGen_HtmlStyleSheetBlock_pairs;
#else
	_new = (_TPPLST_codeGen_StyleSheetPairpcodeGen_HtmlStyleSheetBlock_pairs) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_StyleSheetPairpcodeGen_HtmlStyleSheetBlock_pairs));
#endif
	_new->_prod = RULELST_codeGen_StyleSheetPairpcodeGen_HtmlStyleSheetBlock_pairs;
	_new->_desc1 = (_TSPcodeGen_StyleSheetPair)elem;
	_new->_desc2 = (_TSPLST_codeGen_HtmlStyleSheetBlock_pairs)_new;
	{_TPPLST_codeGen_StyleSheetPairpcodeGen_HtmlStyleSheetBlock_pairs _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_StyleSheetPairpcodeGen_HtmlStyleSheetBlock_pairs;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_HtmlStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_HtmlStyleSheetBlock_pairs (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_StyleSheetPair)) 
		return _MkcodeGen_StyleSheetPairpcodeGen_HtmlStyleSheetBlock_pairs (_coordref, MkcodeGen_StyleSheetPair (_coordref, t));
	message (DEADLY, "a pcodeGen_HtmlStyleSheetBlock_pairs list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_HtmlStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_HtmlStyleSheetBlock_pairs (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_HtmlStyleSheetBlock_pairs)
	l =_ElempcodeGen_HtmlStyleSheetBlock_pairs (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_HtmlStyleSheetBlock_pairs)
	r = _ElempcodeGen_HtmlStyleSheetBlock_pairs (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_HtmlStyleSheetBlock_pairs)r)->_desc2);
((_TPPLST_0pcodeGen_HtmlStyleSheetBlock_pairs)r)->_desc2 = ((_TPPLST_0pcodeGen_HtmlStyleSheetBlock_pairs)l)->_desc2;
((_TPPLST_0pcodeGen_HtmlStyleSheetBlock_pairs)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_HtmlStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_HtmlStyleSheetBlock_pairs (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_HtmlStyleSheetBlock_pairs root;
	_TPPLST_0pcodeGen_HtmlStyleSheetBlock_pairs list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_HtmlStyleSheetBlock_pairs)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_HtmlStyleSheetBlock_pairs)
		t = _ElempcodeGen_HtmlStyleSheetBlock_pairs (_coordref, t);
	remain = (_TPPLST_0pcodeGen_HtmlStyleSheetBlock_pairs)(((_TPPLST_0pcodeGen_HtmlStyleSheetBlock_pairs) t)->_desc2);
	((_TPPLST_0pcodeGen_HtmlStyleSheetBlock_pairs) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_HtmlStyleSheetBlock_pairs;
#else
list = (_TPPLST_0pcodeGen_HtmlStyleSheetBlock_pairs) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_HtmlStyleSheetBlock_pairs));
#endif
list->_prod = RULELST_0pcodeGen_HtmlStyleSheetBlock_pairs;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_HtmlStyleSheetBlock_pairs)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_StyleSheetPairpcodeGen_HtmlStyleSheetBlock_pairs)
	{_TPPLST_codeGen_StyleSheetPairpcodeGen_HtmlStyleSheetBlock_pairs _currn = _currn;
	 _TERMACT_LST_codeGen_StyleSheetPairpcodeGen_HtmlStyleSheetBlock_pairs;
	} this code has been moved to _ElempcodeGen_HtmlStyleSheetBlock_pairs*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_HtmlStyleSheetBlock_pairs;
#else
root = (_TPPpcodeGen_HtmlStyleSheetBlock_pairs) TreeNodeAlloc (sizeof (struct _TPpcodeGen_HtmlStyleSheetBlock_pairs));
#endif
root->_prod = RULEpcodeGen_HtmlStyleSheetBlock_pairs;
root->_desc1 = (_TSPLST_codeGen_HtmlStyleSheetBlock_pairs)list;
{ _TPPpcodeGen_HtmlStyleSheetBlock_pairs _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_HtmlStyleSheetBlock_pairs;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_HtmlStyleSheetBlock_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_HtmlStyleSheetBlock_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_HtmlStyleSheetBlock_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_HtmlStyleSheetBlock_name;
#else
_currn = (_TPPpcodeGen_HtmlStyleSheetBlock_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_HtmlStyleSheetBlock_name));
#endif
_currn->_prod = RULEpcodeGen_HtmlStyleSheetBlock_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_HtmlStyleSheetBlock_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_HtmlStyleSheetBlock_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ActionHandler (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_ActionHandler (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_ActionHandler _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ActionHandler;
#else
_currn = (_TPPpcodeGen_ActionHandler) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ActionHandler));
#endif
_currn->_prod = RULEpcodeGen_ActionHandler;
_currn->_desc1 = (_TSPcodeGen_ActionHandler_targetpage) MkcodeGen_ActionHandler_targetpage (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_ActionHandler: root of subtree no. 1 can not be made a codeGen_ActionHandler_targetpage node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_ActionHandler_params) MkcodeGen_ActionHandler_params (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_ActionHandler: root of subtree no. 2 can not be made a codeGen_ActionHandler_params node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_ActionHandler;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_ActionHandler */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ActionHandler_targetpage (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_ActionHandler_targetpage (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_ActionHandler_targetpage _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_ActionHandler_targetpage;
#else
_currn = (_TPPpcodeGen_ActionHandler_targetpage) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ActionHandler_targetpage));
#endif
_currn->_prod = RULEpcodeGen_ActionHandler_targetpage;
_SETCOORD(_currn)
_TERMACT_pcodeGen_ActionHandler_targetpage;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_ActionHandler_targetpage */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_WebNodeActualParampcodeGen_ActionHandler_params (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_WebNodeActualParampcodeGen_ActionHandler_params (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_WebNodeActualParampcodeGen_ActionHandler_params _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_WebNodeActualParampcodeGen_ActionHandler_params;
#else
	_new = (_TPPLST_codeGen_WebNodeActualParampcodeGen_ActionHandler_params) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_WebNodeActualParampcodeGen_ActionHandler_params));
#endif
	_new->_prod = RULELST_codeGen_WebNodeActualParampcodeGen_ActionHandler_params;
	_new->_desc1 = (_TSPcodeGen_WebNodeActualParam)elem;
	_new->_desc2 = (_TSPLST_codeGen_ActionHandler_params)_new;
	{_TPPLST_codeGen_WebNodeActualParampcodeGen_ActionHandler_params _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_WebNodeActualParampcodeGen_ActionHandler_params;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_ActionHandler_params (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_ActionHandler_params (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_WebNodeActualParam)) 
		return _MkcodeGen_WebNodeActualParampcodeGen_ActionHandler_params (_coordref, MkcodeGen_WebNodeActualParam (_coordref, t));
	message (DEADLY, "a pcodeGen_ActionHandler_params list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_ActionHandler_params (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_ActionHandler_params (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_ActionHandler_params)
	l =_ElempcodeGen_ActionHandler_params (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_ActionHandler_params)
	r = _ElempcodeGen_ActionHandler_params (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_ActionHandler_params)r)->_desc2);
((_TPPLST_0pcodeGen_ActionHandler_params)r)->_desc2 = ((_TPPLST_0pcodeGen_ActionHandler_params)l)->_desc2;
((_TPPLST_0pcodeGen_ActionHandler_params)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_ActionHandler_params (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_ActionHandler_params (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_ActionHandler_params root;
	_TPPLST_0pcodeGen_ActionHandler_params list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_ActionHandler_params)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_ActionHandler_params)
		t = _ElempcodeGen_ActionHandler_params (_coordref, t);
	remain = (_TPPLST_0pcodeGen_ActionHandler_params)(((_TPPLST_0pcodeGen_ActionHandler_params) t)->_desc2);
	((_TPPLST_0pcodeGen_ActionHandler_params) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_ActionHandler_params;
#else
list = (_TPPLST_0pcodeGen_ActionHandler_params) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_ActionHandler_params));
#endif
list->_prod = RULELST_0pcodeGen_ActionHandler_params;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_ActionHandler_params)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_WebNodeActualParampcodeGen_ActionHandler_params)
	{_TPPLST_codeGen_WebNodeActualParampcodeGen_ActionHandler_params _currn = _currn;
	 _TERMACT_LST_codeGen_WebNodeActualParampcodeGen_ActionHandler_params;
	} this code has been moved to _ElempcodeGen_ActionHandler_params*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_ActionHandler_params;
#else
root = (_TPPpcodeGen_ActionHandler_params) TreeNodeAlloc (sizeof (struct _TPpcodeGen_ActionHandler_params));
#endif
root->_prod = RULEpcodeGen_ActionHandler_params;
root->_desc1 = (_TSPLST_codeGen_ActionHandler_params)list;
{ _TPPpcodeGen_ActionHandler_params _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_ActionHandler_params;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebAction (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkpcodeGen_WebAction (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPpcodeGen_WebAction _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_WebAction;
#else
_currn = (_TPPpcodeGen_WebAction) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebAction));
#endif
_currn->_prod = RULEpcodeGen_WebAction;
_currn->_desc1 = (_TSPcodeGen_WebAction_name) MkcodeGen_WebAction_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebAction: root of subtree no. 1 can not be made a codeGen_WebAction_name node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_WebAction_formalParameter) MkcodeGen_WebAction_formalParameter (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebAction: root of subtree no. 2 can not be made a codeGen_WebAction_formalParameter node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_WebAction_successpage) MkcodeGen_WebAction_successpage (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebAction: root of subtree no. 3 can not be made a codeGen_WebAction_successpage node ", 0, _coordref);
_currn->_desc4 = (_TSPcodeGen_WebAction_operations) MkcodeGen_WebAction_operations (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebAction: root of subtree no. 4 can not be made a codeGen_WebAction_operations node ", 0, _coordref);
_currn->_desc5 = (_TSPcodeGen_WebAction_position) MkcodeGen_WebAction_position (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_WebAction: root of subtree no. 5 can not be made a codeGen_WebAction_position node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_WebAction;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_WebAction */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebAction_name (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_WebAction_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_WebAction_name _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_WebAction_name;
#else
_currn = (_TPPpcodeGen_WebAction_name) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebAction_name));
#endif
_currn->_prod = RULEpcodeGen_WebAction_name;
_SETCOORD(_currn)
_TERMACT_pcodeGen_WebAction_name;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_WebAction_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_WebNodeFormalParampcodeGen_WebAction_formalParameter (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_WebNodeFormalParampcodeGen_WebAction_formalParameter (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_WebNodeFormalParampcodeGen_WebAction_formalParameter _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_WebNodeFormalParampcodeGen_WebAction_formalParameter;
#else
	_new = (_TPPLST_codeGen_WebNodeFormalParampcodeGen_WebAction_formalParameter) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_WebNodeFormalParampcodeGen_WebAction_formalParameter));
#endif
	_new->_prod = RULELST_codeGen_WebNodeFormalParampcodeGen_WebAction_formalParameter;
	_new->_desc1 = (_TSPcodeGen_WebNodeFormalParam)elem;
	_new->_desc2 = (_TSPLST_codeGen_WebAction_formalParameter)_new;
	{_TPPLST_codeGen_WebNodeFormalParampcodeGen_WebAction_formalParameter _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_WebNodeFormalParampcodeGen_WebAction_formalParameter;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_WebAction_formalParameter (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_WebAction_formalParameter (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_WebNodeFormalParam)) 
		return _MkcodeGen_WebNodeFormalParampcodeGen_WebAction_formalParameter (_coordref, MkcodeGen_WebNodeFormalParam (_coordref, t));
	message (DEADLY, "a pcodeGen_WebAction_formalParameter list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_WebAction_formalParameter (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_WebAction_formalParameter (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_WebAction_formalParameter)
	l =_ElempcodeGen_WebAction_formalParameter (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_WebAction_formalParameter)
	r = _ElempcodeGen_WebAction_formalParameter (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_WebAction_formalParameter)r)->_desc2);
((_TPPLST_0pcodeGen_WebAction_formalParameter)r)->_desc2 = ((_TPPLST_0pcodeGen_WebAction_formalParameter)l)->_desc2;
((_TPPLST_0pcodeGen_WebAction_formalParameter)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebAction_formalParameter (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_WebAction_formalParameter (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_WebAction_formalParameter root;
	_TPPLST_0pcodeGen_WebAction_formalParameter list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_WebAction_formalParameter)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_WebAction_formalParameter)
		t = _ElempcodeGen_WebAction_formalParameter (_coordref, t);
	remain = (_TPPLST_0pcodeGen_WebAction_formalParameter)(((_TPPLST_0pcodeGen_WebAction_formalParameter) t)->_desc2);
	((_TPPLST_0pcodeGen_WebAction_formalParameter) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_WebAction_formalParameter;
#else
list = (_TPPLST_0pcodeGen_WebAction_formalParameter) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_WebAction_formalParameter));
#endif
list->_prod = RULELST_0pcodeGen_WebAction_formalParameter;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_WebAction_formalParameter)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_WebNodeFormalParampcodeGen_WebAction_formalParameter)
	{_TPPLST_codeGen_WebNodeFormalParampcodeGen_WebAction_formalParameter _currn = _currn;
	 _TERMACT_LST_codeGen_WebNodeFormalParampcodeGen_WebAction_formalParameter;
	} this code has been moved to _ElempcodeGen_WebAction_formalParameter*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_WebAction_formalParameter;
#else
root = (_TPPpcodeGen_WebAction_formalParameter) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebAction_formalParameter));
#endif
root->_prod = RULEpcodeGen_WebAction_formalParameter;
root->_desc1 = (_TSPLST_codeGen_WebAction_formalParameter)list;
{ _TPPpcodeGen_WebAction_formalParameter _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_WebAction_formalParameter;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_PHPOperationpcodeGen_WebAction_operations (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_PHPOperationpcodeGen_WebAction_operations (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_PHPOperationpcodeGen_WebAction_operations _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_PHPOperationpcodeGen_WebAction_operations;
#else
	_new = (_TPPLST_codeGen_PHPOperationpcodeGen_WebAction_operations) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_PHPOperationpcodeGen_WebAction_operations));
#endif
	_new->_prod = RULELST_codeGen_PHPOperationpcodeGen_WebAction_operations;
	_new->_desc1 = (_TSPcodeGen_PHPOperation)elem;
	_new->_desc2 = (_TSPLST_codeGen_WebAction_operations)_new;
	{_TPPLST_codeGen_PHPOperationpcodeGen_WebAction_operations _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_PHPOperationpcodeGen_WebAction_operations;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_DataBaseOperationpcodeGen_WebAction_operations (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_DataBaseOperationpcodeGen_WebAction_operations (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_DataBaseOperationpcodeGen_WebAction_operations _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_DataBaseOperationpcodeGen_WebAction_operations;
#else
	_new = (_TPPLST_codeGen_DataBaseOperationpcodeGen_WebAction_operations) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_DataBaseOperationpcodeGen_WebAction_operations));
#endif
	_new->_prod = RULELST_codeGen_DataBaseOperationpcodeGen_WebAction_operations;
	_new->_desc1 = (_TSPcodeGen_DataBaseOperation)elem;
	_new->_desc2 = (_TSPLST_codeGen_WebAction_operations)_new;
	{_TPPLST_codeGen_DataBaseOperationpcodeGen_WebAction_operations _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_DataBaseOperationpcodeGen_WebAction_operations;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_WebAction_operations (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_WebAction_operations (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_PHPOperation)) 
		return _MkcodeGen_PHPOperationpcodeGen_WebAction_operations (_coordref, MkcodeGen_PHPOperation (_coordref, t));
	if (IsSymb (t, SYMBcodeGen_DataBaseOperation)) 
		return _MkcodeGen_DataBaseOperationpcodeGen_WebAction_operations (_coordref, MkcodeGen_DataBaseOperation (_coordref, t));
	message (DEADLY, "a pcodeGen_WebAction_operations list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_WebAction_operations (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_WebAction_operations (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_WebAction_operations)
	l =_ElempcodeGen_WebAction_operations (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_WebAction_operations)
	r = _ElempcodeGen_WebAction_operations (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_WebAction_operations)r)->_desc2);
((_TPPLST_0pcodeGen_WebAction_operations)r)->_desc2 = ((_TPPLST_0pcodeGen_WebAction_operations)l)->_desc2;
((_TPPLST_0pcodeGen_WebAction_operations)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebAction_operations (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_WebAction_operations (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_WebAction_operations root;
	_TPPLST_0pcodeGen_WebAction_operations list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_WebAction_operations)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_WebAction_operations)
		t = _ElempcodeGen_WebAction_operations (_coordref, t);
	remain = (_TPPLST_0pcodeGen_WebAction_operations)(((_TPPLST_0pcodeGen_WebAction_operations) t)->_desc2);
	((_TPPLST_0pcodeGen_WebAction_operations) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_WebAction_operations;
#else
list = (_TPPLST_0pcodeGen_WebAction_operations) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_WebAction_operations));
#endif
list->_prod = RULELST_0pcodeGen_WebAction_operations;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_WebAction_operations)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_PHPOperationpcodeGen_WebAction_operations)
	{_TPPLST_codeGen_PHPOperationpcodeGen_WebAction_operations _currn = _currn;
	 _TERMACT_LST_codeGen_PHPOperationpcodeGen_WebAction_operations;
	} this code has been moved to _ElempcodeGen_WebAction_operations*/
	/*if (_currn->_prod == RULELST_codeGen_DataBaseOperationpcodeGen_WebAction_operations)
	{_TPPLST_codeGen_DataBaseOperationpcodeGen_WebAction_operations _currn = _currn;
	 _TERMACT_LST_codeGen_DataBaseOperationpcodeGen_WebAction_operations;
	} this code has been moved to _ElempcodeGen_WebAction_operations*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_WebAction_operations;
#else
root = (_TPPpcodeGen_WebAction_operations) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebAction_operations));
#endif
root->_prod = RULEpcodeGen_WebAction_operations;
root->_desc1 = (_TSPLST_codeGen_WebAction_operations)list;
{ _TPPpcodeGen_WebAction_operations _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_WebAction_operations;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_WebAction_position (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_WebAction_position (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_WebAction_position _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_WebAction_position;
#else
_currn = (_TPPpcodeGen_WebAction_position) TreeNodeAlloc (sizeof (struct _TPpcodeGen_WebAction_position));
#endif
_currn->_prod = RULEpcodeGen_WebAction_position;
_SETCOORD(_currn)
_TERMACT_pcodeGen_WebAction_position;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_WebAction_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_PHPOperation (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpcodeGen_PHPOperation (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpcodeGen_PHPOperation _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_PHPOperation;
#else
_currn = (_TPPpcodeGen_PHPOperation) TreeNodeAlloc (sizeof (struct _TPpcodeGen_PHPOperation));
#endif
_currn->_prod = RULEpcodeGen_PHPOperation;
_currn->_desc1 = (_TSPcodeGen_PHPOperation_errorhandler) MkcodeGen_PHPOperation_errorhandler (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_PHPOperation: root of subtree no. 1 can not be made a codeGen_PHPOperation_errorhandler node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_PHPOperation_content) MkcodeGen_PHPOperation_content (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_PHPOperation: root of subtree no. 2 can not be made a codeGen_PHPOperation_content node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_PHPOperation;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_PHPOperation */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_ActionHandlerpcodeGen_PHPOperation_errorhandler (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_ActionHandlerpcodeGen_PHPOperation_errorhandler (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_ActionHandlerpcodeGen_PHPOperation_errorhandler _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_ActionHandlerpcodeGen_PHPOperation_errorhandler;
#else
	_new = (_TPPLST_codeGen_ActionHandlerpcodeGen_PHPOperation_errorhandler) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_ActionHandlerpcodeGen_PHPOperation_errorhandler));
#endif
	_new->_prod = RULELST_codeGen_ActionHandlerpcodeGen_PHPOperation_errorhandler;
	_new->_desc1 = (_TSPcodeGen_ActionHandler)elem;
	_new->_desc2 = (_TSPLST_codeGen_PHPOperation_errorhandler)_new;
	{_TPPLST_codeGen_ActionHandlerpcodeGen_PHPOperation_errorhandler _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_ActionHandlerpcodeGen_PHPOperation_errorhandler;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_PHPOperation_errorhandler (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_PHPOperation_errorhandler (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_ActionHandler)) 
		return _MkcodeGen_ActionHandlerpcodeGen_PHPOperation_errorhandler (_coordref, MkcodeGen_ActionHandler (_coordref, t));
	message (DEADLY, "a pcodeGen_PHPOperation_errorhandler list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_PHPOperation_errorhandler (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_PHPOperation_errorhandler (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_PHPOperation_errorhandler)
	l =_ElempcodeGen_PHPOperation_errorhandler (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_PHPOperation_errorhandler)
	r = _ElempcodeGen_PHPOperation_errorhandler (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_PHPOperation_errorhandler)r)->_desc2);
((_TPPLST_0pcodeGen_PHPOperation_errorhandler)r)->_desc2 = ((_TPPLST_0pcodeGen_PHPOperation_errorhandler)l)->_desc2;
((_TPPLST_0pcodeGen_PHPOperation_errorhandler)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_PHPOperation_errorhandler (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_PHPOperation_errorhandler (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_PHPOperation_errorhandler root;
	_TPPLST_0pcodeGen_PHPOperation_errorhandler list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_PHPOperation_errorhandler)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_PHPOperation_errorhandler)
		t = _ElempcodeGen_PHPOperation_errorhandler (_coordref, t);
	remain = (_TPPLST_0pcodeGen_PHPOperation_errorhandler)(((_TPPLST_0pcodeGen_PHPOperation_errorhandler) t)->_desc2);
	((_TPPLST_0pcodeGen_PHPOperation_errorhandler) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_PHPOperation_errorhandler;
#else
list = (_TPPLST_0pcodeGen_PHPOperation_errorhandler) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_PHPOperation_errorhandler));
#endif
list->_prod = RULELST_0pcodeGen_PHPOperation_errorhandler;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_PHPOperation_errorhandler)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_ActionHandlerpcodeGen_PHPOperation_errorhandler)
	{_TPPLST_codeGen_ActionHandlerpcodeGen_PHPOperation_errorhandler _currn = _currn;
	 _TERMACT_LST_codeGen_ActionHandlerpcodeGen_PHPOperation_errorhandler;
	} this code has been moved to _ElempcodeGen_PHPOperation_errorhandler*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_PHPOperation_errorhandler;
#else
root = (_TPPpcodeGen_PHPOperation_errorhandler) TreeNodeAlloc (sizeof (struct _TPpcodeGen_PHPOperation_errorhandler));
#endif
root->_prod = RULEpcodeGen_PHPOperation_errorhandler;
root->_desc1 = (_TSPLST_codeGen_PHPOperation_errorhandler)list;
{ _TPPpcodeGen_PHPOperation_errorhandler _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_PHPOperation_errorhandler;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_PHPOperation_content (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_PHPOperation_content (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_PHPOperation_content _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_PHPOperation_content;
#else
_currn = (_TPPpcodeGen_PHPOperation_content) TreeNodeAlloc (sizeof (struct _TPpcodeGen_PHPOperation_content));
#endif
_currn->_prod = RULEpcodeGen_PHPOperation_content;
_SETCOORD(_currn)
_TERMACT_pcodeGen_PHPOperation_content;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_PHPOperation_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataBaseOperation (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpcodeGen_DataBaseOperation (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpcodeGen_DataBaseOperation _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataBaseOperation;
#else
_currn = (_TPPpcodeGen_DataBaseOperation) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataBaseOperation));
#endif
_currn->_prod = RULEpcodeGen_DataBaseOperation;
_currn->_desc1 = (_TSPcodeGen_DataBaseOperation_errorhandler) MkcodeGen_DataBaseOperation_errorhandler (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataBaseOperation: root of subtree no. 1 can not be made a codeGen_DataBaseOperation_errorhandler node ", 0, _coordref);
_currn->_desc2 = (_TSPcodeGen_DataBaseOperation_filter) MkcodeGen_DataBaseOperation_filter (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataBaseOperation: root of subtree no. 2 can not be made a codeGen_DataBaseOperation_filter node ", 0, _coordref);
_currn->_desc3 = (_TSPcodeGen_DataBaseOperation_params) MkcodeGen_DataBaseOperation_params (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pcodeGen_DataBaseOperation: root of subtree no. 3 can not be made a codeGen_DataBaseOperation_params node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataBaseOperation;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataBaseOperation */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_ActionHandlerpcodeGen_DataBaseOperation_errorhandler (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_ActionHandlerpcodeGen_DataBaseOperation_errorhandler (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_ActionHandlerpcodeGen_DataBaseOperation_errorhandler _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_ActionHandlerpcodeGen_DataBaseOperation_errorhandler;
#else
	_new = (_TPPLST_codeGen_ActionHandlerpcodeGen_DataBaseOperation_errorhandler) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_ActionHandlerpcodeGen_DataBaseOperation_errorhandler));
#endif
	_new->_prod = RULELST_codeGen_ActionHandlerpcodeGen_DataBaseOperation_errorhandler;
	_new->_desc1 = (_TSPcodeGen_ActionHandler)elem;
	_new->_desc2 = (_TSPLST_codeGen_DataBaseOperation_errorhandler)_new;
	{_TPPLST_codeGen_ActionHandlerpcodeGen_DataBaseOperation_errorhandler _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_ActionHandlerpcodeGen_DataBaseOperation_errorhandler;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_DataBaseOperation_errorhandler (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_DataBaseOperation_errorhandler (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_ActionHandler)) 
		return _MkcodeGen_ActionHandlerpcodeGen_DataBaseOperation_errorhandler (_coordref, MkcodeGen_ActionHandler (_coordref, t));
	message (DEADLY, "a pcodeGen_DataBaseOperation_errorhandler list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_DataBaseOperation_errorhandler (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_DataBaseOperation_errorhandler (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_DataBaseOperation_errorhandler)
	l =_ElempcodeGen_DataBaseOperation_errorhandler (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_DataBaseOperation_errorhandler)
	r = _ElempcodeGen_DataBaseOperation_errorhandler (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_DataBaseOperation_errorhandler)r)->_desc2);
((_TPPLST_0pcodeGen_DataBaseOperation_errorhandler)r)->_desc2 = ((_TPPLST_0pcodeGen_DataBaseOperation_errorhandler)l)->_desc2;
((_TPPLST_0pcodeGen_DataBaseOperation_errorhandler)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataBaseOperation_errorhandler (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_DataBaseOperation_errorhandler (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_DataBaseOperation_errorhandler root;
	_TPPLST_0pcodeGen_DataBaseOperation_errorhandler list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_DataBaseOperation_errorhandler)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_DataBaseOperation_errorhandler)
		t = _ElempcodeGen_DataBaseOperation_errorhandler (_coordref, t);
	remain = (_TPPLST_0pcodeGen_DataBaseOperation_errorhandler)(((_TPPLST_0pcodeGen_DataBaseOperation_errorhandler) t)->_desc2);
	((_TPPLST_0pcodeGen_DataBaseOperation_errorhandler) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_DataBaseOperation_errorhandler;
#else
list = (_TPPLST_0pcodeGen_DataBaseOperation_errorhandler) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_DataBaseOperation_errorhandler));
#endif
list->_prod = RULELST_0pcodeGen_DataBaseOperation_errorhandler;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_DataBaseOperation_errorhandler)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_ActionHandlerpcodeGen_DataBaseOperation_errorhandler)
	{_TPPLST_codeGen_ActionHandlerpcodeGen_DataBaseOperation_errorhandler _currn = _currn;
	 _TERMACT_LST_codeGen_ActionHandlerpcodeGen_DataBaseOperation_errorhandler;
	} this code has been moved to _ElempcodeGen_DataBaseOperation_errorhandler*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_DataBaseOperation_errorhandler;
#else
root = (_TPPpcodeGen_DataBaseOperation_errorhandler) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataBaseOperation_errorhandler));
#endif
root->_prod = RULEpcodeGen_DataBaseOperation_errorhandler;
root->_desc1 = (_TSPLST_codeGen_DataBaseOperation_errorhandler)list;
{ _TPPpcodeGen_DataBaseOperation_errorhandler _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_DataBaseOperation_errorhandler;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataBaseOperation_filter (POSITION *_coordref)
#else
NODEPTR MkpcodeGen_DataBaseOperation_filter (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpcodeGen_DataBaseOperation_filter _currn;
#ifdef __cplusplus
_currn = new _TPpcodeGen_DataBaseOperation_filter;
#else
_currn = (_TPPpcodeGen_DataBaseOperation_filter) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataBaseOperation_filter));
#endif
_currn->_prod = RULEpcodeGen_DataBaseOperation_filter;
_SETCOORD(_currn)
_TERMACT_pcodeGen_DataBaseOperation_filter;
return ( (NODEPTR) _currn);
}/* MkpcodeGen_DataBaseOperation_filter */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkcodeGen_DataSourceActualParampcodeGen_DataBaseOperation_params (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkcodeGen_DataSourceActualParampcodeGen_DataBaseOperation_params (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_codeGen_DataSourceActualParampcodeGen_DataBaseOperation_params _new;
#ifdef __cplusplus
	_new = new _TPLST_codeGen_DataSourceActualParampcodeGen_DataBaseOperation_params;
#else
	_new = (_TPPLST_codeGen_DataSourceActualParampcodeGen_DataBaseOperation_params) TreeNodeAlloc (sizeof (struct _TPLST_codeGen_DataSourceActualParampcodeGen_DataBaseOperation_params));
#endif
	_new->_prod = RULELST_codeGen_DataSourceActualParampcodeGen_DataBaseOperation_params;
	_new->_desc1 = (_TSPcodeGen_DataSourceActualParam)elem;
	_new->_desc2 = (_TSPLST_codeGen_DataBaseOperation_params)_new;
	{_TPPLST_codeGen_DataSourceActualParampcodeGen_DataBaseOperation_params _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_codeGen_DataSourceActualParampcodeGen_DataBaseOperation_params;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempcodeGen_DataBaseOperation_params (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempcodeGen_DataBaseOperation_params (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBcodeGen_DataSourceActualParam)) 
		return _MkcodeGen_DataSourceActualParampcodeGen_DataBaseOperation_params (_coordref, MkcodeGen_DataSourceActualParam (_coordref, t));
	message (DEADLY, "a pcodeGen_DataBaseOperation_params list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pcodeGen_DataBaseOperation_params (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pcodeGen_DataBaseOperation_params (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_codeGen_DataBaseOperation_params)
	l =_ElempcodeGen_DataBaseOperation_params (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_codeGen_DataBaseOperation_params)
	r = _ElempcodeGen_DataBaseOperation_params (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pcodeGen_DataBaseOperation_params)r)->_desc2);
((_TPPLST_0pcodeGen_DataBaseOperation_params)r)->_desc2 = ((_TPPLST_0pcodeGen_DataBaseOperation_params)l)->_desc2;
((_TPPLST_0pcodeGen_DataBaseOperation_params)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpcodeGen_DataBaseOperation_params (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpcodeGen_DataBaseOperation_params (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpcodeGen_DataBaseOperation_params root;
	_TPPLST_0pcodeGen_DataBaseOperation_params list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pcodeGen_DataBaseOperation_params)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_codeGen_DataBaseOperation_params)
		t = _ElempcodeGen_DataBaseOperation_params (_coordref, t);
	remain = (_TPPLST_0pcodeGen_DataBaseOperation_params)(((_TPPLST_0pcodeGen_DataBaseOperation_params) t)->_desc2);
	((_TPPLST_0pcodeGen_DataBaseOperation_params) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pcodeGen_DataBaseOperation_params;
#else
list = (_TPPLST_0pcodeGen_DataBaseOperation_params) TreeNodeAlloc (sizeof (struct _TPLST_0pcodeGen_DataBaseOperation_params));
#endif
list->_prod = RULELST_0pcodeGen_DataBaseOperation_params;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pcodeGen_DataBaseOperation_params)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_codeGen_DataSourceActualParampcodeGen_DataBaseOperation_params)
	{_TPPLST_codeGen_DataSourceActualParampcodeGen_DataBaseOperation_params _currn = _currn;
	 _TERMACT_LST_codeGen_DataSourceActualParampcodeGen_DataBaseOperation_params;
	} this code has been moved to _ElempcodeGen_DataBaseOperation_params*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpcodeGen_DataBaseOperation_params;
#else
root = (_TPPpcodeGen_DataBaseOperation_params) TreeNodeAlloc (sizeof (struct _TPpcodeGen_DataBaseOperation_params));
#endif
root->_prod = RULEpcodeGen_DataBaseOperation_params;
root->_desc1 = (_TSPLST_codeGen_DataBaseOperation_params)list;
{ _TPPpcodeGen_DataBaseOperation_params _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pcodeGen_DataBaseOperation_params;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpVIEWROOT_dataBaseView (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR MkpVIEWROOT_dataBaseView (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPpVIEWROOT_dataBaseView _currn;
#ifdef __cplusplus
_currn = new _TPpVIEWROOT_dataBaseView;
#else
_currn = (_TPPpVIEWROOT_dataBaseView) TreeNodeAlloc (sizeof (struct _TPpVIEWROOT_dataBaseView));
#endif
_currn->_prod = RULEpVIEWROOT_dataBaseView;
_currn->_desc1 = (_TSPdataBaseView_DataBase) MkdataBaseView_DataBase (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pVIEWROOT_dataBaseView: root of subtree no. 1 can not be made a dataBaseView_DataBase node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_pVIEWROOT_dataBaseView;
return ( (NODEPTR) _currn);
}/* MkpVIEWROOT_dataBaseView */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DBColumn (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpdataBaseView_DBColumn (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpdataBaseView_DBColumn _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DBColumn;
#else
_currn = (_TPPpdataBaseView_DBColumn) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DBColumn));
#endif
_currn->_prod = RULEpdataBaseView_DBColumn;
_currn->_desc1 = (_TSPdataBaseView_DBColumn_name) MkdataBaseView_DBColumn_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DBColumn: root of subtree no. 1 can not be made a dataBaseView_DBColumn_name node ", 0, _coordref);
_currn->_desc2 = (_TSPdataBaseView_DBColumn_keyword1) MkdataBaseView_DBColumn_keyword1 (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DBColumn: root of subtree no. 2 can not be made a dataBaseView_DBColumn_keyword1 node ", 0, _coordref);
_currn->_desc3 = (_TSPdataBaseView_DBColumn_type) MkdataBaseView_DBColumn_type (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DBColumn: root of subtree no. 3 can not be made a dataBaseView_DBColumn_type node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DBColumn;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DBColumn */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DBColumn_name (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DBColumn_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DBColumn_name _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DBColumn_name;
#else
_currn = (_TPPpdataBaseView_DBColumn_name) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DBColumn_name));
#endif
_currn->_prod = RULEpdataBaseView_DBColumn_name;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DBColumn_name;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DBColumn_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DBColumn_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DBColumn_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DBColumn_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DBColumn_keyword1;
#else
_currn = (_TPPpdataBaseView_DBColumn_keyword1) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DBColumn_keyword1));
#endif
_currn->_prod = RULEpdataBaseView_DBColumn_keyword1;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DBColumn_keyword1;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DBColumn_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DBStringpdataBaseView_DBColumn_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DBStringpdataBaseView_DBColumn_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DBStringpdataBaseView_DBColumn_type _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DBStringpdataBaseView_DBColumn_type;
#else
	_new = (_TPPLST_dataBaseView_DBStringpdataBaseView_DBColumn_type) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DBStringpdataBaseView_DBColumn_type));
#endif
	_new->_prod = RULELST_dataBaseView_DBStringpdataBaseView_DBColumn_type;
	_new->_desc1 = (_TSPdataBaseView_DBString)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DBColumn_type)_new;
	{_TPPLST_dataBaseView_DBStringpdataBaseView_DBColumn_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DBStringpdataBaseView_DBColumn_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DBIntegerpdataBaseView_DBColumn_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DBIntegerpdataBaseView_DBColumn_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DBIntegerpdataBaseView_DBColumn_type _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DBIntegerpdataBaseView_DBColumn_type;
#else
	_new = (_TPPLST_dataBaseView_DBIntegerpdataBaseView_DBColumn_type) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DBIntegerpdataBaseView_DBColumn_type));
#endif
	_new->_prod = RULELST_dataBaseView_DBIntegerpdataBaseView_DBColumn_type;
	_new->_desc1 = (_TSPdataBaseView_DBInteger)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DBColumn_type)_new;
	{_TPPLST_dataBaseView_DBIntegerpdataBaseView_DBColumn_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DBIntegerpdataBaseView_DBColumn_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DBFloatpdataBaseView_DBColumn_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DBFloatpdataBaseView_DBColumn_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DBFloatpdataBaseView_DBColumn_type _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DBFloatpdataBaseView_DBColumn_type;
#else
	_new = (_TPPLST_dataBaseView_DBFloatpdataBaseView_DBColumn_type) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DBFloatpdataBaseView_DBColumn_type));
#endif
	_new->_prod = RULELST_dataBaseView_DBFloatpdataBaseView_DBColumn_type;
	_new->_desc1 = (_TSPdataBaseView_DBFloat)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DBColumn_type)_new;
	{_TPPLST_dataBaseView_DBFloatpdataBaseView_DBColumn_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DBFloatpdataBaseView_DBColumn_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DBDoublepdataBaseView_DBColumn_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DBDoublepdataBaseView_DBColumn_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DBDoublepdataBaseView_DBColumn_type _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DBDoublepdataBaseView_DBColumn_type;
#else
	_new = (_TPPLST_dataBaseView_DBDoublepdataBaseView_DBColumn_type) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DBDoublepdataBaseView_DBColumn_type));
#endif
	_new->_prod = RULELST_dataBaseView_DBDoublepdataBaseView_DBColumn_type;
	_new->_desc1 = (_TSPdataBaseView_DBDouble)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DBColumn_type)_new;
	{_TPPLST_dataBaseView_DBDoublepdataBaseView_DBColumn_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DBDoublepdataBaseView_DBColumn_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DBBigIntpdataBaseView_DBColumn_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DBBigIntpdataBaseView_DBColumn_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DBBigIntpdataBaseView_DBColumn_type _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DBBigIntpdataBaseView_DBColumn_type;
#else
	_new = (_TPPLST_dataBaseView_DBBigIntpdataBaseView_DBColumn_type) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DBBigIntpdataBaseView_DBColumn_type));
#endif
	_new->_prod = RULELST_dataBaseView_DBBigIntpdataBaseView_DBColumn_type;
	_new->_desc1 = (_TSPdataBaseView_DBBigInt)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DBColumn_type)_new;
	{_TPPLST_dataBaseView_DBBigIntpdataBaseView_DBColumn_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DBBigIntpdataBaseView_DBColumn_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DBDatepdataBaseView_DBColumn_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DBDatepdataBaseView_DBColumn_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DBDatepdataBaseView_DBColumn_type _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DBDatepdataBaseView_DBColumn_type;
#else
	_new = (_TPPLST_dataBaseView_DBDatepdataBaseView_DBColumn_type) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DBDatepdataBaseView_DBColumn_type));
#endif
	_new->_prod = RULELST_dataBaseView_DBDatepdataBaseView_DBColumn_type;
	_new->_desc1 = (_TSPdataBaseView_DBDate)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DBColumn_type)_new;
	{_TPPLST_dataBaseView_DBDatepdataBaseView_DBColumn_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DBDatepdataBaseView_DBColumn_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DBTimeStamppdataBaseView_DBColumn_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DBTimeStamppdataBaseView_DBColumn_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DBTimeStamppdataBaseView_DBColumn_type _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DBTimeStamppdataBaseView_DBColumn_type;
#else
	_new = (_TPPLST_dataBaseView_DBTimeStamppdataBaseView_DBColumn_type) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DBTimeStamppdataBaseView_DBColumn_type));
#endif
	_new->_prod = RULELST_dataBaseView_DBTimeStamppdataBaseView_DBColumn_type;
	_new->_desc1 = (_TSPdataBaseView_DBTimeStamp)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DBColumn_type)_new;
	{_TPPLST_dataBaseView_DBTimeStamppdataBaseView_DBColumn_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DBTimeStamppdataBaseView_DBColumn_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DBBooleanpdataBaseView_DBColumn_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DBBooleanpdataBaseView_DBColumn_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DBBooleanpdataBaseView_DBColumn_type _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DBBooleanpdataBaseView_DBColumn_type;
#else
	_new = (_TPPLST_dataBaseView_DBBooleanpdataBaseView_DBColumn_type) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DBBooleanpdataBaseView_DBColumn_type));
#endif
	_new->_prod = RULELST_dataBaseView_DBBooleanpdataBaseView_DBColumn_type;
	_new->_desc1 = (_TSPdataBaseView_DBBoolean)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DBColumn_type)_new;
	{_TPPLST_dataBaseView_DBBooleanpdataBaseView_DBColumn_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DBBooleanpdataBaseView_DBColumn_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DBBlobpdataBaseView_DBColumn_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DBBlobpdataBaseView_DBColumn_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DBBlobpdataBaseView_DBColumn_type _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DBBlobpdataBaseView_DBColumn_type;
#else
	_new = (_TPPLST_dataBaseView_DBBlobpdataBaseView_DBColumn_type) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DBBlobpdataBaseView_DBColumn_type));
#endif
	_new->_prod = RULELST_dataBaseView_DBBlobpdataBaseView_DBColumn_type;
	_new->_desc1 = (_TSPdataBaseView_DBBlob)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DBColumn_type)_new;
	{_TPPLST_dataBaseView_DBBlobpdataBaseView_DBColumn_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DBBlobpdataBaseView_DBColumn_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempdataBaseView_DBColumn_type (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempdataBaseView_DBColumn_type (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBdataBaseView_DBString)) 
		return _MkdataBaseView_DBStringpdataBaseView_DBColumn_type (_coordref, MkdataBaseView_DBString (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_DBInteger)) 
		return _MkdataBaseView_DBIntegerpdataBaseView_DBColumn_type (_coordref, MkdataBaseView_DBInteger (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_DBFloat)) 
		return _MkdataBaseView_DBFloatpdataBaseView_DBColumn_type (_coordref, MkdataBaseView_DBFloat (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_DBDouble)) 
		return _MkdataBaseView_DBDoublepdataBaseView_DBColumn_type (_coordref, MkdataBaseView_DBDouble (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_DBBigInt)) 
		return _MkdataBaseView_DBBigIntpdataBaseView_DBColumn_type (_coordref, MkdataBaseView_DBBigInt (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_DBDate)) 
		return _MkdataBaseView_DBDatepdataBaseView_DBColumn_type (_coordref, MkdataBaseView_DBDate (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_DBTimeStamp)) 
		return _MkdataBaseView_DBTimeStamppdataBaseView_DBColumn_type (_coordref, MkdataBaseView_DBTimeStamp (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_DBBoolean)) 
		return _MkdataBaseView_DBBooleanpdataBaseView_DBColumn_type (_coordref, MkdataBaseView_DBBoolean (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_DBBlob)) 
		return _MkdataBaseView_DBBlobpdataBaseView_DBColumn_type (_coordref, MkdataBaseView_DBBlob (_coordref, t));
	message (DEADLY, "a pdataBaseView_DBColumn_type list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pdataBaseView_DBColumn_type (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pdataBaseView_DBColumn_type (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_dataBaseView_DBColumn_type)
	l =_ElempdataBaseView_DBColumn_type (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_dataBaseView_DBColumn_type)
	r = _ElempdataBaseView_DBColumn_type (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pdataBaseView_DBColumn_type)r)->_desc2);
((_TPPLST_0pdataBaseView_DBColumn_type)r)->_desc2 = ((_TPPLST_0pdataBaseView_DBColumn_type)l)->_desc2;
((_TPPLST_0pdataBaseView_DBColumn_type)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DBColumn_type (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpdataBaseView_DBColumn_type (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpdataBaseView_DBColumn_type root;
	_TPPLST_0pdataBaseView_DBColumn_type list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pdataBaseView_DBColumn_type)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_dataBaseView_DBColumn_type)
		t = _ElempdataBaseView_DBColumn_type (_coordref, t);
	remain = (_TPPLST_0pdataBaseView_DBColumn_type)(((_TPPLST_0pdataBaseView_DBColumn_type) t)->_desc2);
	((_TPPLST_0pdataBaseView_DBColumn_type) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pdataBaseView_DBColumn_type;
#else
list = (_TPPLST_0pdataBaseView_DBColumn_type) TreeNodeAlloc (sizeof (struct _TPLST_0pdataBaseView_DBColumn_type));
#endif
list->_prod = RULELST_0pdataBaseView_DBColumn_type;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pdataBaseView_DBColumn_type)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_dataBaseView_DBStringpdataBaseView_DBColumn_type)
	{_TPPLST_dataBaseView_DBStringpdataBaseView_DBColumn_type _currn = _currn;
	 _TERMACT_LST_dataBaseView_DBStringpdataBaseView_DBColumn_type;
	} this code has been moved to _ElempdataBaseView_DBColumn_type*/
	/*if (_currn->_prod == RULELST_dataBaseView_DBIntegerpdataBaseView_DBColumn_type)
	{_TPPLST_dataBaseView_DBIntegerpdataBaseView_DBColumn_type _currn = _currn;
	 _TERMACT_LST_dataBaseView_DBIntegerpdataBaseView_DBColumn_type;
	} this code has been moved to _ElempdataBaseView_DBColumn_type*/
	/*if (_currn->_prod == RULELST_dataBaseView_DBFloatpdataBaseView_DBColumn_type)
	{_TPPLST_dataBaseView_DBFloatpdataBaseView_DBColumn_type _currn = _currn;
	 _TERMACT_LST_dataBaseView_DBFloatpdataBaseView_DBColumn_type;
	} this code has been moved to _ElempdataBaseView_DBColumn_type*/
	/*if (_currn->_prod == RULELST_dataBaseView_DBDoublepdataBaseView_DBColumn_type)
	{_TPPLST_dataBaseView_DBDoublepdataBaseView_DBColumn_type _currn = _currn;
	 _TERMACT_LST_dataBaseView_DBDoublepdataBaseView_DBColumn_type;
	} this code has been moved to _ElempdataBaseView_DBColumn_type*/
	/*if (_currn->_prod == RULELST_dataBaseView_DBBigIntpdataBaseView_DBColumn_type)
	{_TPPLST_dataBaseView_DBBigIntpdataBaseView_DBColumn_type _currn = _currn;
	 _TERMACT_LST_dataBaseView_DBBigIntpdataBaseView_DBColumn_type;
	} this code has been moved to _ElempdataBaseView_DBColumn_type*/
	/*if (_currn->_prod == RULELST_dataBaseView_DBDatepdataBaseView_DBColumn_type)
	{_TPPLST_dataBaseView_DBDatepdataBaseView_DBColumn_type _currn = _currn;
	 _TERMACT_LST_dataBaseView_DBDatepdataBaseView_DBColumn_type;
	} this code has been moved to _ElempdataBaseView_DBColumn_type*/
	/*if (_currn->_prod == RULELST_dataBaseView_DBTimeStamppdataBaseView_DBColumn_type)
	{_TPPLST_dataBaseView_DBTimeStamppdataBaseView_DBColumn_type _currn = _currn;
	 _TERMACT_LST_dataBaseView_DBTimeStamppdataBaseView_DBColumn_type;
	} this code has been moved to _ElempdataBaseView_DBColumn_type*/
	/*if (_currn->_prod == RULELST_dataBaseView_DBBooleanpdataBaseView_DBColumn_type)
	{_TPPLST_dataBaseView_DBBooleanpdataBaseView_DBColumn_type _currn = _currn;
	 _TERMACT_LST_dataBaseView_DBBooleanpdataBaseView_DBColumn_type;
	} this code has been moved to _ElempdataBaseView_DBColumn_type*/
	/*if (_currn->_prod == RULELST_dataBaseView_DBBlobpdataBaseView_DBColumn_type)
	{_TPPLST_dataBaseView_DBBlobpdataBaseView_DBColumn_type _currn = _currn;
	 _TERMACT_LST_dataBaseView_DBBlobpdataBaseView_DBColumn_type;
	} this code has been moved to _ElempdataBaseView_DBColumn_type*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpdataBaseView_DBColumn_type;
#else
root = (_TPPpdataBaseView_DBColumn_type) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DBColumn_type));
#endif
root->_prod = RULEpdataBaseView_DBColumn_type;
root->_desc1 = (_TSPLST_dataBaseView_DBColumn_type)list;
{ _TPPpdataBaseView_DBColumn_type _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pdataBaseView_DBColumn_type;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceFormalParam (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpdataBaseView_DataSourceFormalParam (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpdataBaseView_DataSourceFormalParam _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceFormalParam;
#else
_currn = (_TPPpdataBaseView_DataSourceFormalParam) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceFormalParam));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceFormalParam;
_currn->_desc1 = (_TSPdataBaseView_DataSourceFormalParam_name) MkdataBaseView_DataSourceFormalParam_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceFormalParam: root of subtree no. 1 can not be made a dataBaseView_DataSourceFormalParam_name node ", 0, _coordref);
_currn->_desc2 = (_TSPdataBaseView_DataSourceFormalParam_keyword1) MkdataBaseView_DataSourceFormalParam_keyword1 (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceFormalParam: root of subtree no. 2 can not be made a dataBaseView_DataSourceFormalParam_keyword1 node ", 0, _coordref);
_currn->_desc3 = (_TSPdataBaseView_DataSourceFormalParam_type) MkdataBaseView_DataSourceFormalParam_type (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceFormalParam: root of subtree no. 3 can not be made a dataBaseView_DataSourceFormalParam_type node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceFormalParam;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceFormalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceFormalParam_name (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataSourceFormalParam_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataSourceFormalParam_name _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceFormalParam_name;
#else
_currn = (_TPPpdataBaseView_DataSourceFormalParam_name) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceFormalParam_name));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceFormalParam_name;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceFormalParam_name;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceFormalParam_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceFormalParam_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataSourceFormalParam_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataSourceFormalParam_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceFormalParam_keyword1;
#else
_currn = (_TPPpdataBaseView_DataSourceFormalParam_keyword1) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceFormalParam_keyword1));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceFormalParam_keyword1;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceFormalParam_keyword1;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceFormalParam_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DBStringpdataBaseView_DataSourceFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DBStringpdataBaseView_DataSourceFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DBStringpdataBaseView_DataSourceFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DBStringpdataBaseView_DataSourceFormalParam_type;
#else
	_new = (_TPPLST_dataBaseView_DBStringpdataBaseView_DataSourceFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DBStringpdataBaseView_DataSourceFormalParam_type));
#endif
	_new->_prod = RULELST_dataBaseView_DBStringpdataBaseView_DataSourceFormalParam_type;
	_new->_desc1 = (_TSPdataBaseView_DBString)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceFormalParam_type)_new;
	{_TPPLST_dataBaseView_DBStringpdataBaseView_DataSourceFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DBStringpdataBaseView_DataSourceFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DBIntegerpdataBaseView_DataSourceFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DBIntegerpdataBaseView_DataSourceFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DBIntegerpdataBaseView_DataSourceFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DBIntegerpdataBaseView_DataSourceFormalParam_type;
#else
	_new = (_TPPLST_dataBaseView_DBIntegerpdataBaseView_DataSourceFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DBIntegerpdataBaseView_DataSourceFormalParam_type));
#endif
	_new->_prod = RULELST_dataBaseView_DBIntegerpdataBaseView_DataSourceFormalParam_type;
	_new->_desc1 = (_TSPdataBaseView_DBInteger)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceFormalParam_type)_new;
	{_TPPLST_dataBaseView_DBIntegerpdataBaseView_DataSourceFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DBIntegerpdataBaseView_DataSourceFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DBFloatpdataBaseView_DataSourceFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DBFloatpdataBaseView_DataSourceFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DBFloatpdataBaseView_DataSourceFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DBFloatpdataBaseView_DataSourceFormalParam_type;
#else
	_new = (_TPPLST_dataBaseView_DBFloatpdataBaseView_DataSourceFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DBFloatpdataBaseView_DataSourceFormalParam_type));
#endif
	_new->_prod = RULELST_dataBaseView_DBFloatpdataBaseView_DataSourceFormalParam_type;
	_new->_desc1 = (_TSPdataBaseView_DBFloat)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceFormalParam_type)_new;
	{_TPPLST_dataBaseView_DBFloatpdataBaseView_DataSourceFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DBFloatpdataBaseView_DataSourceFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DBDoublepdataBaseView_DataSourceFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DBDoublepdataBaseView_DataSourceFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DBDoublepdataBaseView_DataSourceFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DBDoublepdataBaseView_DataSourceFormalParam_type;
#else
	_new = (_TPPLST_dataBaseView_DBDoublepdataBaseView_DataSourceFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DBDoublepdataBaseView_DataSourceFormalParam_type));
#endif
	_new->_prod = RULELST_dataBaseView_DBDoublepdataBaseView_DataSourceFormalParam_type;
	_new->_desc1 = (_TSPdataBaseView_DBDouble)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceFormalParam_type)_new;
	{_TPPLST_dataBaseView_DBDoublepdataBaseView_DataSourceFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DBDoublepdataBaseView_DataSourceFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DBBigIntpdataBaseView_DataSourceFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DBBigIntpdataBaseView_DataSourceFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DBBigIntpdataBaseView_DataSourceFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DBBigIntpdataBaseView_DataSourceFormalParam_type;
#else
	_new = (_TPPLST_dataBaseView_DBBigIntpdataBaseView_DataSourceFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DBBigIntpdataBaseView_DataSourceFormalParam_type));
#endif
	_new->_prod = RULELST_dataBaseView_DBBigIntpdataBaseView_DataSourceFormalParam_type;
	_new->_desc1 = (_TSPdataBaseView_DBBigInt)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceFormalParam_type)_new;
	{_TPPLST_dataBaseView_DBBigIntpdataBaseView_DataSourceFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DBBigIntpdataBaseView_DataSourceFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DBDatepdataBaseView_DataSourceFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DBDatepdataBaseView_DataSourceFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DBDatepdataBaseView_DataSourceFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DBDatepdataBaseView_DataSourceFormalParam_type;
#else
	_new = (_TPPLST_dataBaseView_DBDatepdataBaseView_DataSourceFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DBDatepdataBaseView_DataSourceFormalParam_type));
#endif
	_new->_prod = RULELST_dataBaseView_DBDatepdataBaseView_DataSourceFormalParam_type;
	_new->_desc1 = (_TSPdataBaseView_DBDate)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceFormalParam_type)_new;
	{_TPPLST_dataBaseView_DBDatepdataBaseView_DataSourceFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DBDatepdataBaseView_DataSourceFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DBTimeStamppdataBaseView_DataSourceFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DBTimeStamppdataBaseView_DataSourceFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DBTimeStamppdataBaseView_DataSourceFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DBTimeStamppdataBaseView_DataSourceFormalParam_type;
#else
	_new = (_TPPLST_dataBaseView_DBTimeStamppdataBaseView_DataSourceFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DBTimeStamppdataBaseView_DataSourceFormalParam_type));
#endif
	_new->_prod = RULELST_dataBaseView_DBTimeStamppdataBaseView_DataSourceFormalParam_type;
	_new->_desc1 = (_TSPdataBaseView_DBTimeStamp)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceFormalParam_type)_new;
	{_TPPLST_dataBaseView_DBTimeStamppdataBaseView_DataSourceFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DBTimeStamppdataBaseView_DataSourceFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DBBooleanpdataBaseView_DataSourceFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DBBooleanpdataBaseView_DataSourceFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DBBooleanpdataBaseView_DataSourceFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DBBooleanpdataBaseView_DataSourceFormalParam_type;
#else
	_new = (_TPPLST_dataBaseView_DBBooleanpdataBaseView_DataSourceFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DBBooleanpdataBaseView_DataSourceFormalParam_type));
#endif
	_new->_prod = RULELST_dataBaseView_DBBooleanpdataBaseView_DataSourceFormalParam_type;
	_new->_desc1 = (_TSPdataBaseView_DBBoolean)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceFormalParam_type)_new;
	{_TPPLST_dataBaseView_DBBooleanpdataBaseView_DataSourceFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DBBooleanpdataBaseView_DataSourceFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DBBlobpdataBaseView_DataSourceFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DBBlobpdataBaseView_DataSourceFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DBBlobpdataBaseView_DataSourceFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DBBlobpdataBaseView_DataSourceFormalParam_type;
#else
	_new = (_TPPLST_dataBaseView_DBBlobpdataBaseView_DataSourceFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DBBlobpdataBaseView_DataSourceFormalParam_type));
#endif
	_new->_prod = RULELST_dataBaseView_DBBlobpdataBaseView_DataSourceFormalParam_type;
	_new->_desc1 = (_TSPdataBaseView_DBBlob)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceFormalParam_type)_new;
	{_TPPLST_dataBaseView_DBBlobpdataBaseView_DataSourceFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DBBlobpdataBaseView_DataSourceFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempdataBaseView_DataSourceFormalParam_type (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempdataBaseView_DataSourceFormalParam_type (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBdataBaseView_DBString)) 
		return _MkdataBaseView_DBStringpdataBaseView_DataSourceFormalParam_type (_coordref, MkdataBaseView_DBString (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_DBInteger)) 
		return _MkdataBaseView_DBIntegerpdataBaseView_DataSourceFormalParam_type (_coordref, MkdataBaseView_DBInteger (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_DBFloat)) 
		return _MkdataBaseView_DBFloatpdataBaseView_DataSourceFormalParam_type (_coordref, MkdataBaseView_DBFloat (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_DBDouble)) 
		return _MkdataBaseView_DBDoublepdataBaseView_DataSourceFormalParam_type (_coordref, MkdataBaseView_DBDouble (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_DBBigInt)) 
		return _MkdataBaseView_DBBigIntpdataBaseView_DataSourceFormalParam_type (_coordref, MkdataBaseView_DBBigInt (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_DBDate)) 
		return _MkdataBaseView_DBDatepdataBaseView_DataSourceFormalParam_type (_coordref, MkdataBaseView_DBDate (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_DBTimeStamp)) 
		return _MkdataBaseView_DBTimeStamppdataBaseView_DataSourceFormalParam_type (_coordref, MkdataBaseView_DBTimeStamp (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_DBBoolean)) 
		return _MkdataBaseView_DBBooleanpdataBaseView_DataSourceFormalParam_type (_coordref, MkdataBaseView_DBBoolean (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_DBBlob)) 
		return _MkdataBaseView_DBBlobpdataBaseView_DataSourceFormalParam_type (_coordref, MkdataBaseView_DBBlob (_coordref, t));
	message (DEADLY, "a pdataBaseView_DataSourceFormalParam_type list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pdataBaseView_DataSourceFormalParam_type (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pdataBaseView_DataSourceFormalParam_type (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_dataBaseView_DataSourceFormalParam_type)
	l =_ElempdataBaseView_DataSourceFormalParam_type (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_dataBaseView_DataSourceFormalParam_type)
	r = _ElempdataBaseView_DataSourceFormalParam_type (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pdataBaseView_DataSourceFormalParam_type)r)->_desc2);
((_TPPLST_0pdataBaseView_DataSourceFormalParam_type)r)->_desc2 = ((_TPPLST_0pdataBaseView_DataSourceFormalParam_type)l)->_desc2;
((_TPPLST_0pdataBaseView_DataSourceFormalParam_type)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceFormalParam_type (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpdataBaseView_DataSourceFormalParam_type (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpdataBaseView_DataSourceFormalParam_type root;
	_TPPLST_0pdataBaseView_DataSourceFormalParam_type list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pdataBaseView_DataSourceFormalParam_type)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_dataBaseView_DataSourceFormalParam_type)
		t = _ElempdataBaseView_DataSourceFormalParam_type (_coordref, t);
	remain = (_TPPLST_0pdataBaseView_DataSourceFormalParam_type)(((_TPPLST_0pdataBaseView_DataSourceFormalParam_type) t)->_desc2);
	((_TPPLST_0pdataBaseView_DataSourceFormalParam_type) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pdataBaseView_DataSourceFormalParam_type;
#else
list = (_TPPLST_0pdataBaseView_DataSourceFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_0pdataBaseView_DataSourceFormalParam_type));
#endif
list->_prod = RULELST_0pdataBaseView_DataSourceFormalParam_type;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pdataBaseView_DataSourceFormalParam_type)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_dataBaseView_DBStringpdataBaseView_DataSourceFormalParam_type)
	{_TPPLST_dataBaseView_DBStringpdataBaseView_DataSourceFormalParam_type _currn = _currn;
	 _TERMACT_LST_dataBaseView_DBStringpdataBaseView_DataSourceFormalParam_type;
	} this code has been moved to _ElempdataBaseView_DataSourceFormalParam_type*/
	/*if (_currn->_prod == RULELST_dataBaseView_DBIntegerpdataBaseView_DataSourceFormalParam_type)
	{_TPPLST_dataBaseView_DBIntegerpdataBaseView_DataSourceFormalParam_type _currn = _currn;
	 _TERMACT_LST_dataBaseView_DBIntegerpdataBaseView_DataSourceFormalParam_type;
	} this code has been moved to _ElempdataBaseView_DataSourceFormalParam_type*/
	/*if (_currn->_prod == RULELST_dataBaseView_DBFloatpdataBaseView_DataSourceFormalParam_type)
	{_TPPLST_dataBaseView_DBFloatpdataBaseView_DataSourceFormalParam_type _currn = _currn;
	 _TERMACT_LST_dataBaseView_DBFloatpdataBaseView_DataSourceFormalParam_type;
	} this code has been moved to _ElempdataBaseView_DataSourceFormalParam_type*/
	/*if (_currn->_prod == RULELST_dataBaseView_DBDoublepdataBaseView_DataSourceFormalParam_type)
	{_TPPLST_dataBaseView_DBDoublepdataBaseView_DataSourceFormalParam_type _currn = _currn;
	 _TERMACT_LST_dataBaseView_DBDoublepdataBaseView_DataSourceFormalParam_type;
	} this code has been moved to _ElempdataBaseView_DataSourceFormalParam_type*/
	/*if (_currn->_prod == RULELST_dataBaseView_DBBigIntpdataBaseView_DataSourceFormalParam_type)
	{_TPPLST_dataBaseView_DBBigIntpdataBaseView_DataSourceFormalParam_type _currn = _currn;
	 _TERMACT_LST_dataBaseView_DBBigIntpdataBaseView_DataSourceFormalParam_type;
	} this code has been moved to _ElempdataBaseView_DataSourceFormalParam_type*/
	/*if (_currn->_prod == RULELST_dataBaseView_DBDatepdataBaseView_DataSourceFormalParam_type)
	{_TPPLST_dataBaseView_DBDatepdataBaseView_DataSourceFormalParam_type _currn = _currn;
	 _TERMACT_LST_dataBaseView_DBDatepdataBaseView_DataSourceFormalParam_type;
	} this code has been moved to _ElempdataBaseView_DataSourceFormalParam_type*/
	/*if (_currn->_prod == RULELST_dataBaseView_DBTimeStamppdataBaseView_DataSourceFormalParam_type)
	{_TPPLST_dataBaseView_DBTimeStamppdataBaseView_DataSourceFormalParam_type _currn = _currn;
	 _TERMACT_LST_dataBaseView_DBTimeStamppdataBaseView_DataSourceFormalParam_type;
	} this code has been moved to _ElempdataBaseView_DataSourceFormalParam_type*/
	/*if (_currn->_prod == RULELST_dataBaseView_DBBooleanpdataBaseView_DataSourceFormalParam_type)
	{_TPPLST_dataBaseView_DBBooleanpdataBaseView_DataSourceFormalParam_type _currn = _currn;
	 _TERMACT_LST_dataBaseView_DBBooleanpdataBaseView_DataSourceFormalParam_type;
	} this code has been moved to _ElempdataBaseView_DataSourceFormalParam_type*/
	/*if (_currn->_prod == RULELST_dataBaseView_DBBlobpdataBaseView_DataSourceFormalParam_type)
	{_TPPLST_dataBaseView_DBBlobpdataBaseView_DataSourceFormalParam_type _currn = _currn;
	 _TERMACT_LST_dataBaseView_DBBlobpdataBaseView_DataSourceFormalParam_type;
	} this code has been moved to _ElempdataBaseView_DataSourceFormalParam_type*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpdataBaseView_DataSourceFormalParam_type;
#else
root = (_TPPpdataBaseView_DataSourceFormalParam_type) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceFormalParam_type));
#endif
root->_prod = RULEpdataBaseView_DataSourceFormalParam_type;
root->_desc1 = (_TSPLST_dataBaseView_DataSourceFormalParam_type)list;
{ _TPPpdataBaseView_DataSourceFormalParam_type _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pdataBaseView_DataSourceFormalParam_type;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_ColumnProjection (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpdataBaseView_ColumnProjection (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpdataBaseView_ColumnProjection _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_ColumnProjection;
#else
_currn = (_TPPpdataBaseView_ColumnProjection) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_ColumnProjection));
#endif
_currn->_prod = RULEpdataBaseView_ColumnProjection;
_currn->_desc1 = (_TSPdataBaseView_ColumnProjection_keyword1) MkdataBaseView_ColumnProjection_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_ColumnProjection: root of subtree no. 1 can not be made a dataBaseView_ColumnProjection_keyword1 node ", 0, _coordref);
_currn->_desc2 = (_TSPdataBaseView_ColumnProjection_column) MkdataBaseView_ColumnProjection_column (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_ColumnProjection: root of subtree no. 2 can not be made a dataBaseView_ColumnProjection_column node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_ColumnProjection;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_ColumnProjection */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_ColumnProjection_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_ColumnProjection_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_ColumnProjection_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_ColumnProjection_keyword1;
#else
_currn = (_TPPpdataBaseView_ColumnProjection_keyword1) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_ColumnProjection_keyword1));
#endif
_currn->_prod = RULEpdataBaseView_ColumnProjection_keyword1;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_ColumnProjection_keyword1;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_ColumnProjection_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_ColumnProjection_column (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_ColumnProjection_column (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_ColumnProjection_column _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_ColumnProjection_column;
#else
_currn = (_TPPpdataBaseView_ColumnProjection_column) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_ColumnProjection_column));
#endif
_currn->_prod = RULEpdataBaseView_ColumnProjection_column;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_ColumnProjection_column;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_ColumnProjection_column */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_ComputationProjection (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4)
#else
NODEPTR MkpdataBaseView_ComputationProjection (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
#endif
{	_TPPpdataBaseView_ComputationProjection _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_ComputationProjection;
#else
_currn = (_TPPpdataBaseView_ComputationProjection) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_ComputationProjection));
#endif
_currn->_prod = RULEpdataBaseView_ComputationProjection;
_currn->_desc1 = (_TSPdataBaseView_ComputationProjection_computation) MkdataBaseView_ComputationProjection_computation (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_ComputationProjection: root of subtree no. 1 can not be made a dataBaseView_ComputationProjection_computation node ", 0, _coordref);
_currn->_desc2 = (_TSPdataBaseView_ComputationProjection_keyword1) MkdataBaseView_ComputationProjection_keyword1 (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_ComputationProjection: root of subtree no. 2 can not be made a dataBaseView_ComputationProjection_keyword1 node ", 0, _coordref);
_currn->_desc3 = (_TSPdataBaseView_ComputationProjection_column) MkdataBaseView_ComputationProjection_column (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_ComputationProjection: root of subtree no. 3 can not be made a dataBaseView_ComputationProjection_column node ", 0, _coordref);
_currn->_desc4 = (_TSPdataBaseView_ComputationProjection_keyword2) MkdataBaseView_ComputationProjection_keyword2 (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_ComputationProjection: root of subtree no. 4 can not be made a dataBaseView_ComputationProjection_keyword2 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_ComputationProjection;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_ComputationProjection */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_ComputationProjection_computation (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_ComputationProjection_computation (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_ComputationProjection_computation _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_ComputationProjection_computation;
#else
_currn = (_TPPpdataBaseView_ComputationProjection_computation) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_ComputationProjection_computation));
#endif
_currn->_prod = RULEpdataBaseView_ComputationProjection_computation;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_ComputationProjection_computation;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_ComputationProjection_computation */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_ComputationProjection_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_ComputationProjection_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_ComputationProjection_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_ComputationProjection_keyword1;
#else
_currn = (_TPPpdataBaseView_ComputationProjection_keyword1) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_ComputationProjection_keyword1));
#endif
_currn->_prod = RULEpdataBaseView_ComputationProjection_keyword1;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_ComputationProjection_keyword1;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_ComputationProjection_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_ComputationProjection_column (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_ComputationProjection_column (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_ComputationProjection_column _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_ComputationProjection_column;
#else
_currn = (_TPPpdataBaseView_ComputationProjection_column) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_ComputationProjection_column));
#endif
_currn->_prod = RULEpdataBaseView_ComputationProjection_column;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_ComputationProjection_column;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_ComputationProjection_column */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_ComputationProjection_keyword2 (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_ComputationProjection_keyword2 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_ComputationProjection_keyword2 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_ComputationProjection_keyword2;
#else
_currn = (_TPPpdataBaseView_ComputationProjection_keyword2) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_ComputationProjection_keyword2));
#endif
_currn->_prod = RULEpdataBaseView_ComputationProjection_keyword2;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_ComputationProjection_keyword2;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_ComputationProjection_keyword2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_BinaryExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkpdataBaseView_BinaryExpression (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPpdataBaseView_BinaryExpression _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_BinaryExpression;
#else
_currn = (_TPPpdataBaseView_BinaryExpression) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_BinaryExpression));
#endif
_currn->_prod = RULEpdataBaseView_BinaryExpression;
_currn->_desc1 = (_TSPdataBaseView_BinaryExpression_keyword1) MkdataBaseView_BinaryExpression_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_BinaryExpression: root of subtree no. 1 can not be made a dataBaseView_BinaryExpression_keyword1 node ", 0, _coordref);
_currn->_desc2 = (_TSPdataBaseView_BinaryExpression_leftOperand) MkdataBaseView_BinaryExpression_leftOperand (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_BinaryExpression: root of subtree no. 2 can not be made a dataBaseView_BinaryExpression_leftOperand node ", 0, _coordref);
_currn->_desc3 = (_TSPdataBaseView_BinaryExpression_operator) MkdataBaseView_BinaryExpression_operator (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_BinaryExpression: root of subtree no. 3 can not be made a dataBaseView_BinaryExpression_operator node ", 0, _coordref);
_currn->_desc4 = (_TSPdataBaseView_BinaryExpression_rightOperand) MkdataBaseView_BinaryExpression_rightOperand (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_BinaryExpression: root of subtree no. 4 can not be made a dataBaseView_BinaryExpression_rightOperand node ", 0, _coordref);
_currn->_desc5 = (_TSPdataBaseView_BinaryExpression_keyword2) MkdataBaseView_BinaryExpression_keyword2 (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_BinaryExpression: root of subtree no. 5 can not be made a dataBaseView_BinaryExpression_keyword2 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_BinaryExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_BinaryExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_BinaryExpression_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_BinaryExpression_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_BinaryExpression_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_BinaryExpression_keyword1;
#else
_currn = (_TPPpdataBaseView_BinaryExpression_keyword1) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_BinaryExpression_keyword1));
#endif
_currn->_prod = RULEpdataBaseView_BinaryExpression_keyword1;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_BinaryExpression_keyword1;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_BinaryExpression_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_BinaryExpressionpdataBaseView_BinaryExpression_leftOperand (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_BinaryExpressionpdataBaseView_BinaryExpression_leftOperand (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_BinaryExpression_leftOperand _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_BinaryExpressionpdataBaseView_BinaryExpression_leftOperand;
#else
	_new = (_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_BinaryExpression_leftOperand) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_BinaryExpressionpdataBaseView_BinaryExpression_leftOperand));
#endif
	_new->_prod = RULELST_dataBaseView_BinaryExpressionpdataBaseView_BinaryExpression_leftOperand;
	_new->_desc1 = (_TSPdataBaseView_BinaryExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_BinaryExpression_leftOperand)_new;
	{_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_BinaryExpression_leftOperand _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_BinaryExpressionpdataBaseView_BinaryExpression_leftOperand;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_UnaryExpressionpdataBaseView_BinaryExpression_leftOperand (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_UnaryExpressionpdataBaseView_BinaryExpression_leftOperand (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_BinaryExpression_leftOperand _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_UnaryExpressionpdataBaseView_BinaryExpression_leftOperand;
#else
	_new = (_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_BinaryExpression_leftOperand) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_UnaryExpressionpdataBaseView_BinaryExpression_leftOperand));
#endif
	_new->_prod = RULELST_dataBaseView_UnaryExpressionpdataBaseView_BinaryExpression_leftOperand;
	_new->_desc1 = (_TSPdataBaseView_UnaryExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_BinaryExpression_leftOperand)_new;
	{_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_BinaryExpression_leftOperand _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_UnaryExpressionpdataBaseView_BinaryExpression_leftOperand;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ConstantExpressionpdataBaseView_BinaryExpression_leftOperand (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ConstantExpressionpdataBaseView_BinaryExpression_leftOperand (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_BinaryExpression_leftOperand _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ConstantExpressionpdataBaseView_BinaryExpression_leftOperand;
#else
	_new = (_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_BinaryExpression_leftOperand) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ConstantExpressionpdataBaseView_BinaryExpression_leftOperand));
#endif
	_new->_prod = RULELST_dataBaseView_ConstantExpressionpdataBaseView_BinaryExpression_leftOperand;
	_new->_desc1 = (_TSPdataBaseView_ConstantExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_BinaryExpression_leftOperand)_new;
	{_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_BinaryExpression_leftOperand _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ConstantExpressionpdataBaseView_BinaryExpression_leftOperand;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ColumnValueExpressionpdataBaseView_BinaryExpression_leftOperand (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ColumnValueExpressionpdataBaseView_BinaryExpression_leftOperand (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_BinaryExpression_leftOperand _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ColumnValueExpressionpdataBaseView_BinaryExpression_leftOperand;
#else
	_new = (_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_BinaryExpression_leftOperand) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ColumnValueExpressionpdataBaseView_BinaryExpression_leftOperand));
#endif
	_new->_prod = RULELST_dataBaseView_ColumnValueExpressionpdataBaseView_BinaryExpression_leftOperand;
	_new->_desc1 = (_TSPdataBaseView_ColumnValueExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_BinaryExpression_leftOperand)_new;
	{_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_BinaryExpression_leftOperand _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ColumnValueExpressionpdataBaseView_BinaryExpression_leftOperand;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ParameterValueExpressionpdataBaseView_BinaryExpression_leftOperand (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ParameterValueExpressionpdataBaseView_BinaryExpression_leftOperand (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_BinaryExpression_leftOperand _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ParameterValueExpressionpdataBaseView_BinaryExpression_leftOperand;
#else
	_new = (_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_BinaryExpression_leftOperand) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ParameterValueExpressionpdataBaseView_BinaryExpression_leftOperand));
#endif
	_new->_prod = RULELST_dataBaseView_ParameterValueExpressionpdataBaseView_BinaryExpression_leftOperand;
	_new->_desc1 = (_TSPdataBaseView_ParameterValueExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_BinaryExpression_leftOperand)_new;
	{_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_BinaryExpression_leftOperand _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ParameterValueExpressionpdataBaseView_BinaryExpression_leftOperand;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ComputationExpressionpdataBaseView_BinaryExpression_leftOperand (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ComputationExpressionpdataBaseView_BinaryExpression_leftOperand (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_BinaryExpression_leftOperand _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ComputationExpressionpdataBaseView_BinaryExpression_leftOperand;
#else
	_new = (_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_BinaryExpression_leftOperand) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ComputationExpressionpdataBaseView_BinaryExpression_leftOperand));
#endif
	_new->_prod = RULELST_dataBaseView_ComputationExpressionpdataBaseView_BinaryExpression_leftOperand;
	_new->_desc1 = (_TSPdataBaseView_ComputationExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_BinaryExpression_leftOperand)_new;
	{_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_BinaryExpression_leftOperand _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ComputationExpressionpdataBaseView_BinaryExpression_leftOperand;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempdataBaseView_BinaryExpression_leftOperand (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempdataBaseView_BinaryExpression_leftOperand (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBdataBaseView_BinaryExpression)) 
		return _MkdataBaseView_BinaryExpressionpdataBaseView_BinaryExpression_leftOperand (_coordref, MkdataBaseView_BinaryExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_UnaryExpression)) 
		return _MkdataBaseView_UnaryExpressionpdataBaseView_BinaryExpression_leftOperand (_coordref, MkdataBaseView_UnaryExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ConstantExpression)) 
		return _MkdataBaseView_ConstantExpressionpdataBaseView_BinaryExpression_leftOperand (_coordref, MkdataBaseView_ConstantExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ColumnValueExpression)) 
		return _MkdataBaseView_ColumnValueExpressionpdataBaseView_BinaryExpression_leftOperand (_coordref, MkdataBaseView_ColumnValueExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ParameterValueExpression)) 
		return _MkdataBaseView_ParameterValueExpressionpdataBaseView_BinaryExpression_leftOperand (_coordref, MkdataBaseView_ParameterValueExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ComputationExpression)) 
		return _MkdataBaseView_ComputationExpressionpdataBaseView_BinaryExpression_leftOperand (_coordref, MkdataBaseView_ComputationExpression (_coordref, t));
	message (DEADLY, "a pdataBaseView_BinaryExpression_leftOperand list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pdataBaseView_BinaryExpression_leftOperand (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pdataBaseView_BinaryExpression_leftOperand (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_dataBaseView_BinaryExpression_leftOperand)
	l =_ElempdataBaseView_BinaryExpression_leftOperand (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_dataBaseView_BinaryExpression_leftOperand)
	r = _ElempdataBaseView_BinaryExpression_leftOperand (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pdataBaseView_BinaryExpression_leftOperand)r)->_desc2);
((_TPPLST_0pdataBaseView_BinaryExpression_leftOperand)r)->_desc2 = ((_TPPLST_0pdataBaseView_BinaryExpression_leftOperand)l)->_desc2;
((_TPPLST_0pdataBaseView_BinaryExpression_leftOperand)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_BinaryExpression_leftOperand (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpdataBaseView_BinaryExpression_leftOperand (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpdataBaseView_BinaryExpression_leftOperand root;
	_TPPLST_0pdataBaseView_BinaryExpression_leftOperand list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pdataBaseView_BinaryExpression_leftOperand)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_dataBaseView_BinaryExpression_leftOperand)
		t = _ElempdataBaseView_BinaryExpression_leftOperand (_coordref, t);
	remain = (_TPPLST_0pdataBaseView_BinaryExpression_leftOperand)(((_TPPLST_0pdataBaseView_BinaryExpression_leftOperand) t)->_desc2);
	((_TPPLST_0pdataBaseView_BinaryExpression_leftOperand) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pdataBaseView_BinaryExpression_leftOperand;
#else
list = (_TPPLST_0pdataBaseView_BinaryExpression_leftOperand) TreeNodeAlloc (sizeof (struct _TPLST_0pdataBaseView_BinaryExpression_leftOperand));
#endif
list->_prod = RULELST_0pdataBaseView_BinaryExpression_leftOperand;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pdataBaseView_BinaryExpression_leftOperand)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_dataBaseView_BinaryExpressionpdataBaseView_BinaryExpression_leftOperand)
	{_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_BinaryExpression_leftOperand _currn = _currn;
	 _TERMACT_LST_dataBaseView_BinaryExpressionpdataBaseView_BinaryExpression_leftOperand;
	} this code has been moved to _ElempdataBaseView_BinaryExpression_leftOperand*/
	/*if (_currn->_prod == RULELST_dataBaseView_UnaryExpressionpdataBaseView_BinaryExpression_leftOperand)
	{_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_BinaryExpression_leftOperand _currn = _currn;
	 _TERMACT_LST_dataBaseView_UnaryExpressionpdataBaseView_BinaryExpression_leftOperand;
	} this code has been moved to _ElempdataBaseView_BinaryExpression_leftOperand*/
	/*if (_currn->_prod == RULELST_dataBaseView_ConstantExpressionpdataBaseView_BinaryExpression_leftOperand)
	{_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_BinaryExpression_leftOperand _currn = _currn;
	 _TERMACT_LST_dataBaseView_ConstantExpressionpdataBaseView_BinaryExpression_leftOperand;
	} this code has been moved to _ElempdataBaseView_BinaryExpression_leftOperand*/
	/*if (_currn->_prod == RULELST_dataBaseView_ColumnValueExpressionpdataBaseView_BinaryExpression_leftOperand)
	{_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_BinaryExpression_leftOperand _currn = _currn;
	 _TERMACT_LST_dataBaseView_ColumnValueExpressionpdataBaseView_BinaryExpression_leftOperand;
	} this code has been moved to _ElempdataBaseView_BinaryExpression_leftOperand*/
	/*if (_currn->_prod == RULELST_dataBaseView_ParameterValueExpressionpdataBaseView_BinaryExpression_leftOperand)
	{_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_BinaryExpression_leftOperand _currn = _currn;
	 _TERMACT_LST_dataBaseView_ParameterValueExpressionpdataBaseView_BinaryExpression_leftOperand;
	} this code has been moved to _ElempdataBaseView_BinaryExpression_leftOperand*/
	/*if (_currn->_prod == RULELST_dataBaseView_ComputationExpressionpdataBaseView_BinaryExpression_leftOperand)
	{_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_BinaryExpression_leftOperand _currn = _currn;
	 _TERMACT_LST_dataBaseView_ComputationExpressionpdataBaseView_BinaryExpression_leftOperand;
	} this code has been moved to _ElempdataBaseView_BinaryExpression_leftOperand*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpdataBaseView_BinaryExpression_leftOperand;
#else
root = (_TPPpdataBaseView_BinaryExpression_leftOperand) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_BinaryExpression_leftOperand));
#endif
root->_prod = RULEpdataBaseView_BinaryExpression_leftOperand;
root->_desc1 = (_TSPLST_dataBaseView_BinaryExpression_leftOperand)list;
{ _TPPpdataBaseView_BinaryExpression_leftOperand _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pdataBaseView_BinaryExpression_leftOperand;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_BinaryExpression_operator (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_BinaryExpression_operator (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_BinaryExpression_operator _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_BinaryExpression_operator;
#else
_currn = (_TPPpdataBaseView_BinaryExpression_operator) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_BinaryExpression_operator));
#endif
_currn->_prod = RULEpdataBaseView_BinaryExpression_operator;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_BinaryExpression_operator;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_BinaryExpression_operator */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_BinaryExpressionpdataBaseView_BinaryExpression_rightOperand (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_BinaryExpressionpdataBaseView_BinaryExpression_rightOperand (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_BinaryExpression_rightOperand _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_BinaryExpressionpdataBaseView_BinaryExpression_rightOperand;
#else
	_new = (_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_BinaryExpression_rightOperand) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_BinaryExpressionpdataBaseView_BinaryExpression_rightOperand));
#endif
	_new->_prod = RULELST_dataBaseView_BinaryExpressionpdataBaseView_BinaryExpression_rightOperand;
	_new->_desc1 = (_TSPdataBaseView_BinaryExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_BinaryExpression_rightOperand)_new;
	{_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_BinaryExpression_rightOperand _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_BinaryExpressionpdataBaseView_BinaryExpression_rightOperand;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_UnaryExpressionpdataBaseView_BinaryExpression_rightOperand (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_UnaryExpressionpdataBaseView_BinaryExpression_rightOperand (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_BinaryExpression_rightOperand _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_UnaryExpressionpdataBaseView_BinaryExpression_rightOperand;
#else
	_new = (_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_BinaryExpression_rightOperand) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_UnaryExpressionpdataBaseView_BinaryExpression_rightOperand));
#endif
	_new->_prod = RULELST_dataBaseView_UnaryExpressionpdataBaseView_BinaryExpression_rightOperand;
	_new->_desc1 = (_TSPdataBaseView_UnaryExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_BinaryExpression_rightOperand)_new;
	{_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_BinaryExpression_rightOperand _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_UnaryExpressionpdataBaseView_BinaryExpression_rightOperand;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ConstantExpressionpdataBaseView_BinaryExpression_rightOperand (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ConstantExpressionpdataBaseView_BinaryExpression_rightOperand (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_BinaryExpression_rightOperand _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ConstantExpressionpdataBaseView_BinaryExpression_rightOperand;
#else
	_new = (_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_BinaryExpression_rightOperand) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ConstantExpressionpdataBaseView_BinaryExpression_rightOperand));
#endif
	_new->_prod = RULELST_dataBaseView_ConstantExpressionpdataBaseView_BinaryExpression_rightOperand;
	_new->_desc1 = (_TSPdataBaseView_ConstantExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_BinaryExpression_rightOperand)_new;
	{_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_BinaryExpression_rightOperand _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ConstantExpressionpdataBaseView_BinaryExpression_rightOperand;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ColumnValueExpressionpdataBaseView_BinaryExpression_rightOperand (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ColumnValueExpressionpdataBaseView_BinaryExpression_rightOperand (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_BinaryExpression_rightOperand _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ColumnValueExpressionpdataBaseView_BinaryExpression_rightOperand;
#else
	_new = (_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_BinaryExpression_rightOperand) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ColumnValueExpressionpdataBaseView_BinaryExpression_rightOperand));
#endif
	_new->_prod = RULELST_dataBaseView_ColumnValueExpressionpdataBaseView_BinaryExpression_rightOperand;
	_new->_desc1 = (_TSPdataBaseView_ColumnValueExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_BinaryExpression_rightOperand)_new;
	{_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_BinaryExpression_rightOperand _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ColumnValueExpressionpdataBaseView_BinaryExpression_rightOperand;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ParameterValueExpressionpdataBaseView_BinaryExpression_rightOperand (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ParameterValueExpressionpdataBaseView_BinaryExpression_rightOperand (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_BinaryExpression_rightOperand _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ParameterValueExpressionpdataBaseView_BinaryExpression_rightOperand;
#else
	_new = (_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_BinaryExpression_rightOperand) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ParameterValueExpressionpdataBaseView_BinaryExpression_rightOperand));
#endif
	_new->_prod = RULELST_dataBaseView_ParameterValueExpressionpdataBaseView_BinaryExpression_rightOperand;
	_new->_desc1 = (_TSPdataBaseView_ParameterValueExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_BinaryExpression_rightOperand)_new;
	{_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_BinaryExpression_rightOperand _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ParameterValueExpressionpdataBaseView_BinaryExpression_rightOperand;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ComputationExpressionpdataBaseView_BinaryExpression_rightOperand (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ComputationExpressionpdataBaseView_BinaryExpression_rightOperand (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_BinaryExpression_rightOperand _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ComputationExpressionpdataBaseView_BinaryExpression_rightOperand;
#else
	_new = (_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_BinaryExpression_rightOperand) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ComputationExpressionpdataBaseView_BinaryExpression_rightOperand));
#endif
	_new->_prod = RULELST_dataBaseView_ComputationExpressionpdataBaseView_BinaryExpression_rightOperand;
	_new->_desc1 = (_TSPdataBaseView_ComputationExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_BinaryExpression_rightOperand)_new;
	{_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_BinaryExpression_rightOperand _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ComputationExpressionpdataBaseView_BinaryExpression_rightOperand;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempdataBaseView_BinaryExpression_rightOperand (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempdataBaseView_BinaryExpression_rightOperand (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBdataBaseView_BinaryExpression)) 
		return _MkdataBaseView_BinaryExpressionpdataBaseView_BinaryExpression_rightOperand (_coordref, MkdataBaseView_BinaryExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_UnaryExpression)) 
		return _MkdataBaseView_UnaryExpressionpdataBaseView_BinaryExpression_rightOperand (_coordref, MkdataBaseView_UnaryExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ConstantExpression)) 
		return _MkdataBaseView_ConstantExpressionpdataBaseView_BinaryExpression_rightOperand (_coordref, MkdataBaseView_ConstantExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ColumnValueExpression)) 
		return _MkdataBaseView_ColumnValueExpressionpdataBaseView_BinaryExpression_rightOperand (_coordref, MkdataBaseView_ColumnValueExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ParameterValueExpression)) 
		return _MkdataBaseView_ParameterValueExpressionpdataBaseView_BinaryExpression_rightOperand (_coordref, MkdataBaseView_ParameterValueExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ComputationExpression)) 
		return _MkdataBaseView_ComputationExpressionpdataBaseView_BinaryExpression_rightOperand (_coordref, MkdataBaseView_ComputationExpression (_coordref, t));
	message (DEADLY, "a pdataBaseView_BinaryExpression_rightOperand list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pdataBaseView_BinaryExpression_rightOperand (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pdataBaseView_BinaryExpression_rightOperand (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_dataBaseView_BinaryExpression_rightOperand)
	l =_ElempdataBaseView_BinaryExpression_rightOperand (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_dataBaseView_BinaryExpression_rightOperand)
	r = _ElempdataBaseView_BinaryExpression_rightOperand (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pdataBaseView_BinaryExpression_rightOperand)r)->_desc2);
((_TPPLST_0pdataBaseView_BinaryExpression_rightOperand)r)->_desc2 = ((_TPPLST_0pdataBaseView_BinaryExpression_rightOperand)l)->_desc2;
((_TPPLST_0pdataBaseView_BinaryExpression_rightOperand)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_BinaryExpression_rightOperand (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpdataBaseView_BinaryExpression_rightOperand (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpdataBaseView_BinaryExpression_rightOperand root;
	_TPPLST_0pdataBaseView_BinaryExpression_rightOperand list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pdataBaseView_BinaryExpression_rightOperand)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_dataBaseView_BinaryExpression_rightOperand)
		t = _ElempdataBaseView_BinaryExpression_rightOperand (_coordref, t);
	remain = (_TPPLST_0pdataBaseView_BinaryExpression_rightOperand)(((_TPPLST_0pdataBaseView_BinaryExpression_rightOperand) t)->_desc2);
	((_TPPLST_0pdataBaseView_BinaryExpression_rightOperand) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pdataBaseView_BinaryExpression_rightOperand;
#else
list = (_TPPLST_0pdataBaseView_BinaryExpression_rightOperand) TreeNodeAlloc (sizeof (struct _TPLST_0pdataBaseView_BinaryExpression_rightOperand));
#endif
list->_prod = RULELST_0pdataBaseView_BinaryExpression_rightOperand;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pdataBaseView_BinaryExpression_rightOperand)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_dataBaseView_BinaryExpressionpdataBaseView_BinaryExpression_rightOperand)
	{_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_BinaryExpression_rightOperand _currn = _currn;
	 _TERMACT_LST_dataBaseView_BinaryExpressionpdataBaseView_BinaryExpression_rightOperand;
	} this code has been moved to _ElempdataBaseView_BinaryExpression_rightOperand*/
	/*if (_currn->_prod == RULELST_dataBaseView_UnaryExpressionpdataBaseView_BinaryExpression_rightOperand)
	{_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_BinaryExpression_rightOperand _currn = _currn;
	 _TERMACT_LST_dataBaseView_UnaryExpressionpdataBaseView_BinaryExpression_rightOperand;
	} this code has been moved to _ElempdataBaseView_BinaryExpression_rightOperand*/
	/*if (_currn->_prod == RULELST_dataBaseView_ConstantExpressionpdataBaseView_BinaryExpression_rightOperand)
	{_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_BinaryExpression_rightOperand _currn = _currn;
	 _TERMACT_LST_dataBaseView_ConstantExpressionpdataBaseView_BinaryExpression_rightOperand;
	} this code has been moved to _ElempdataBaseView_BinaryExpression_rightOperand*/
	/*if (_currn->_prod == RULELST_dataBaseView_ColumnValueExpressionpdataBaseView_BinaryExpression_rightOperand)
	{_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_BinaryExpression_rightOperand _currn = _currn;
	 _TERMACT_LST_dataBaseView_ColumnValueExpressionpdataBaseView_BinaryExpression_rightOperand;
	} this code has been moved to _ElempdataBaseView_BinaryExpression_rightOperand*/
	/*if (_currn->_prod == RULELST_dataBaseView_ParameterValueExpressionpdataBaseView_BinaryExpression_rightOperand)
	{_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_BinaryExpression_rightOperand _currn = _currn;
	 _TERMACT_LST_dataBaseView_ParameterValueExpressionpdataBaseView_BinaryExpression_rightOperand;
	} this code has been moved to _ElempdataBaseView_BinaryExpression_rightOperand*/
	/*if (_currn->_prod == RULELST_dataBaseView_ComputationExpressionpdataBaseView_BinaryExpression_rightOperand)
	{_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_BinaryExpression_rightOperand _currn = _currn;
	 _TERMACT_LST_dataBaseView_ComputationExpressionpdataBaseView_BinaryExpression_rightOperand;
	} this code has been moved to _ElempdataBaseView_BinaryExpression_rightOperand*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpdataBaseView_BinaryExpression_rightOperand;
#else
root = (_TPPpdataBaseView_BinaryExpression_rightOperand) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_BinaryExpression_rightOperand));
#endif
root->_prod = RULEpdataBaseView_BinaryExpression_rightOperand;
root->_desc1 = (_TSPLST_dataBaseView_BinaryExpression_rightOperand)list;
{ _TPPpdataBaseView_BinaryExpression_rightOperand _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pdataBaseView_BinaryExpression_rightOperand;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_BinaryExpression_keyword2 (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_BinaryExpression_keyword2 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_BinaryExpression_keyword2 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_BinaryExpression_keyword2;
#else
_currn = (_TPPpdataBaseView_BinaryExpression_keyword2) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_BinaryExpression_keyword2));
#endif
_currn->_prod = RULEpdataBaseView_BinaryExpression_keyword2;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_BinaryExpression_keyword2;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_BinaryExpression_keyword2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_UnaryExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpdataBaseView_UnaryExpression (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpdataBaseView_UnaryExpression _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_UnaryExpression;
#else
_currn = (_TPPpdataBaseView_UnaryExpression) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_UnaryExpression));
#endif
_currn->_prod = RULEpdataBaseView_UnaryExpression;
_currn->_desc1 = (_TSPdataBaseView_UnaryExpression_operator) MkdataBaseView_UnaryExpression_operator (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_UnaryExpression: root of subtree no. 1 can not be made a dataBaseView_UnaryExpression_operator node ", 0, _coordref);
_currn->_desc2 = (_TSPdataBaseView_UnaryExpression_operand) MkdataBaseView_UnaryExpression_operand (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_UnaryExpression: root of subtree no. 2 can not be made a dataBaseView_UnaryExpression_operand node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_UnaryExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_UnaryExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_UnaryExpression_operator (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_UnaryExpression_operator (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_UnaryExpression_operator _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_UnaryExpression_operator;
#else
_currn = (_TPPpdataBaseView_UnaryExpression_operator) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_UnaryExpression_operator));
#endif
_currn->_prod = RULEpdataBaseView_UnaryExpression_operator;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_UnaryExpression_operator;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_UnaryExpression_operator */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_BinaryExpressionpdataBaseView_UnaryExpression_operand (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_BinaryExpressionpdataBaseView_UnaryExpression_operand (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_UnaryExpression_operand _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_BinaryExpressionpdataBaseView_UnaryExpression_operand;
#else
	_new = (_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_UnaryExpression_operand) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_BinaryExpressionpdataBaseView_UnaryExpression_operand));
#endif
	_new->_prod = RULELST_dataBaseView_BinaryExpressionpdataBaseView_UnaryExpression_operand;
	_new->_desc1 = (_TSPdataBaseView_BinaryExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_UnaryExpression_operand)_new;
	{_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_UnaryExpression_operand _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_BinaryExpressionpdataBaseView_UnaryExpression_operand;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_UnaryExpressionpdataBaseView_UnaryExpression_operand (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_UnaryExpressionpdataBaseView_UnaryExpression_operand (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_UnaryExpression_operand _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_UnaryExpressionpdataBaseView_UnaryExpression_operand;
#else
	_new = (_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_UnaryExpression_operand) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_UnaryExpressionpdataBaseView_UnaryExpression_operand));
#endif
	_new->_prod = RULELST_dataBaseView_UnaryExpressionpdataBaseView_UnaryExpression_operand;
	_new->_desc1 = (_TSPdataBaseView_UnaryExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_UnaryExpression_operand)_new;
	{_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_UnaryExpression_operand _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_UnaryExpressionpdataBaseView_UnaryExpression_operand;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ConstantExpressionpdataBaseView_UnaryExpression_operand (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ConstantExpressionpdataBaseView_UnaryExpression_operand (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_UnaryExpression_operand _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ConstantExpressionpdataBaseView_UnaryExpression_operand;
#else
	_new = (_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_UnaryExpression_operand) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ConstantExpressionpdataBaseView_UnaryExpression_operand));
#endif
	_new->_prod = RULELST_dataBaseView_ConstantExpressionpdataBaseView_UnaryExpression_operand;
	_new->_desc1 = (_TSPdataBaseView_ConstantExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_UnaryExpression_operand)_new;
	{_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_UnaryExpression_operand _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ConstantExpressionpdataBaseView_UnaryExpression_operand;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ColumnValueExpressionpdataBaseView_UnaryExpression_operand (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ColumnValueExpressionpdataBaseView_UnaryExpression_operand (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_UnaryExpression_operand _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ColumnValueExpressionpdataBaseView_UnaryExpression_operand;
#else
	_new = (_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_UnaryExpression_operand) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ColumnValueExpressionpdataBaseView_UnaryExpression_operand));
#endif
	_new->_prod = RULELST_dataBaseView_ColumnValueExpressionpdataBaseView_UnaryExpression_operand;
	_new->_desc1 = (_TSPdataBaseView_ColumnValueExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_UnaryExpression_operand)_new;
	{_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_UnaryExpression_operand _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ColumnValueExpressionpdataBaseView_UnaryExpression_operand;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ParameterValueExpressionpdataBaseView_UnaryExpression_operand (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ParameterValueExpressionpdataBaseView_UnaryExpression_operand (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_UnaryExpression_operand _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ParameterValueExpressionpdataBaseView_UnaryExpression_operand;
#else
	_new = (_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_UnaryExpression_operand) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ParameterValueExpressionpdataBaseView_UnaryExpression_operand));
#endif
	_new->_prod = RULELST_dataBaseView_ParameterValueExpressionpdataBaseView_UnaryExpression_operand;
	_new->_desc1 = (_TSPdataBaseView_ParameterValueExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_UnaryExpression_operand)_new;
	{_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_UnaryExpression_operand _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ParameterValueExpressionpdataBaseView_UnaryExpression_operand;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ComputationExpressionpdataBaseView_UnaryExpression_operand (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ComputationExpressionpdataBaseView_UnaryExpression_operand (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_UnaryExpression_operand _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ComputationExpressionpdataBaseView_UnaryExpression_operand;
#else
	_new = (_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_UnaryExpression_operand) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ComputationExpressionpdataBaseView_UnaryExpression_operand));
#endif
	_new->_prod = RULELST_dataBaseView_ComputationExpressionpdataBaseView_UnaryExpression_operand;
	_new->_desc1 = (_TSPdataBaseView_ComputationExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_UnaryExpression_operand)_new;
	{_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_UnaryExpression_operand _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ComputationExpressionpdataBaseView_UnaryExpression_operand;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempdataBaseView_UnaryExpression_operand (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempdataBaseView_UnaryExpression_operand (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBdataBaseView_BinaryExpression)) 
		return _MkdataBaseView_BinaryExpressionpdataBaseView_UnaryExpression_operand (_coordref, MkdataBaseView_BinaryExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_UnaryExpression)) 
		return _MkdataBaseView_UnaryExpressionpdataBaseView_UnaryExpression_operand (_coordref, MkdataBaseView_UnaryExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ConstantExpression)) 
		return _MkdataBaseView_ConstantExpressionpdataBaseView_UnaryExpression_operand (_coordref, MkdataBaseView_ConstantExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ColumnValueExpression)) 
		return _MkdataBaseView_ColumnValueExpressionpdataBaseView_UnaryExpression_operand (_coordref, MkdataBaseView_ColumnValueExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ParameterValueExpression)) 
		return _MkdataBaseView_ParameterValueExpressionpdataBaseView_UnaryExpression_operand (_coordref, MkdataBaseView_ParameterValueExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ComputationExpression)) 
		return _MkdataBaseView_ComputationExpressionpdataBaseView_UnaryExpression_operand (_coordref, MkdataBaseView_ComputationExpression (_coordref, t));
	message (DEADLY, "a pdataBaseView_UnaryExpression_operand list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pdataBaseView_UnaryExpression_operand (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pdataBaseView_UnaryExpression_operand (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_dataBaseView_UnaryExpression_operand)
	l =_ElempdataBaseView_UnaryExpression_operand (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_dataBaseView_UnaryExpression_operand)
	r = _ElempdataBaseView_UnaryExpression_operand (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pdataBaseView_UnaryExpression_operand)r)->_desc2);
((_TPPLST_0pdataBaseView_UnaryExpression_operand)r)->_desc2 = ((_TPPLST_0pdataBaseView_UnaryExpression_operand)l)->_desc2;
((_TPPLST_0pdataBaseView_UnaryExpression_operand)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_UnaryExpression_operand (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpdataBaseView_UnaryExpression_operand (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpdataBaseView_UnaryExpression_operand root;
	_TPPLST_0pdataBaseView_UnaryExpression_operand list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pdataBaseView_UnaryExpression_operand)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_dataBaseView_UnaryExpression_operand)
		t = _ElempdataBaseView_UnaryExpression_operand (_coordref, t);
	remain = (_TPPLST_0pdataBaseView_UnaryExpression_operand)(((_TPPLST_0pdataBaseView_UnaryExpression_operand) t)->_desc2);
	((_TPPLST_0pdataBaseView_UnaryExpression_operand) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pdataBaseView_UnaryExpression_operand;
#else
list = (_TPPLST_0pdataBaseView_UnaryExpression_operand) TreeNodeAlloc (sizeof (struct _TPLST_0pdataBaseView_UnaryExpression_operand));
#endif
list->_prod = RULELST_0pdataBaseView_UnaryExpression_operand;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pdataBaseView_UnaryExpression_operand)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_dataBaseView_BinaryExpressionpdataBaseView_UnaryExpression_operand)
	{_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_UnaryExpression_operand _currn = _currn;
	 _TERMACT_LST_dataBaseView_BinaryExpressionpdataBaseView_UnaryExpression_operand;
	} this code has been moved to _ElempdataBaseView_UnaryExpression_operand*/
	/*if (_currn->_prod == RULELST_dataBaseView_UnaryExpressionpdataBaseView_UnaryExpression_operand)
	{_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_UnaryExpression_operand _currn = _currn;
	 _TERMACT_LST_dataBaseView_UnaryExpressionpdataBaseView_UnaryExpression_operand;
	} this code has been moved to _ElempdataBaseView_UnaryExpression_operand*/
	/*if (_currn->_prod == RULELST_dataBaseView_ConstantExpressionpdataBaseView_UnaryExpression_operand)
	{_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_UnaryExpression_operand _currn = _currn;
	 _TERMACT_LST_dataBaseView_ConstantExpressionpdataBaseView_UnaryExpression_operand;
	} this code has been moved to _ElempdataBaseView_UnaryExpression_operand*/
	/*if (_currn->_prod == RULELST_dataBaseView_ColumnValueExpressionpdataBaseView_UnaryExpression_operand)
	{_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_UnaryExpression_operand _currn = _currn;
	 _TERMACT_LST_dataBaseView_ColumnValueExpressionpdataBaseView_UnaryExpression_operand;
	} this code has been moved to _ElempdataBaseView_UnaryExpression_operand*/
	/*if (_currn->_prod == RULELST_dataBaseView_ParameterValueExpressionpdataBaseView_UnaryExpression_operand)
	{_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_UnaryExpression_operand _currn = _currn;
	 _TERMACT_LST_dataBaseView_ParameterValueExpressionpdataBaseView_UnaryExpression_operand;
	} this code has been moved to _ElempdataBaseView_UnaryExpression_operand*/
	/*if (_currn->_prod == RULELST_dataBaseView_ComputationExpressionpdataBaseView_UnaryExpression_operand)
	{_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_UnaryExpression_operand _currn = _currn;
	 _TERMACT_LST_dataBaseView_ComputationExpressionpdataBaseView_UnaryExpression_operand;
	} this code has been moved to _ElempdataBaseView_UnaryExpression_operand*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpdataBaseView_UnaryExpression_operand;
#else
root = (_TPPpdataBaseView_UnaryExpression_operand) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_UnaryExpression_operand));
#endif
root->_prod = RULEpdataBaseView_UnaryExpression_operand;
root->_desc1 = (_TSPLST_dataBaseView_UnaryExpression_operand)list;
{ _TPPpdataBaseView_UnaryExpression_operand _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pdataBaseView_UnaryExpression_operand;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_ConstantExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpdataBaseView_ConstantExpression (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpdataBaseView_ConstantExpression _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_ConstantExpression;
#else
_currn = (_TPPpdataBaseView_ConstantExpression) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_ConstantExpression));
#endif
_currn->_prod = RULEpdataBaseView_ConstantExpression;
_currn->_desc1 = (_TSPdataBaseView_ConstantExpression_keyword1) MkdataBaseView_ConstantExpression_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_ConstantExpression: root of subtree no. 1 can not be made a dataBaseView_ConstantExpression_keyword1 node ", 0, _coordref);
_currn->_desc2 = (_TSPdataBaseView_ConstantExpression_value) MkdataBaseView_ConstantExpression_value (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_ConstantExpression: root of subtree no. 2 can not be made a dataBaseView_ConstantExpression_value node ", 0, _coordref);
_currn->_desc3 = (_TSPdataBaseView_ConstantExpression_keyword2) MkdataBaseView_ConstantExpression_keyword2 (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_ConstantExpression: root of subtree no. 3 can not be made a dataBaseView_ConstantExpression_keyword2 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_ConstantExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_ConstantExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_ConstantExpression_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_ConstantExpression_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_ConstantExpression_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_ConstantExpression_keyword1;
#else
_currn = (_TPPpdataBaseView_ConstantExpression_keyword1) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_ConstantExpression_keyword1));
#endif
_currn->_prod = RULEpdataBaseView_ConstantExpression_keyword1;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_ConstantExpression_keyword1;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_ConstantExpression_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_ConstantExpression_value (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_ConstantExpression_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_ConstantExpression_value _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_ConstantExpression_value;
#else
_currn = (_TPPpdataBaseView_ConstantExpression_value) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_ConstantExpression_value));
#endif
_currn->_prod = RULEpdataBaseView_ConstantExpression_value;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_ConstantExpression_value;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_ConstantExpression_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_ConstantExpression_keyword2 (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_ConstantExpression_keyword2 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_ConstantExpression_keyword2 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_ConstantExpression_keyword2;
#else
_currn = (_TPPpdataBaseView_ConstantExpression_keyword2) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_ConstantExpression_keyword2));
#endif
_currn->_prod = RULEpdataBaseView_ConstantExpression_keyword2;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_ConstantExpression_keyword2;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_ConstantExpression_keyword2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_ColumnValueExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpdataBaseView_ColumnValueExpression (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpdataBaseView_ColumnValueExpression _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_ColumnValueExpression;
#else
_currn = (_TPPpdataBaseView_ColumnValueExpression) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_ColumnValueExpression));
#endif
_currn->_prod = RULEpdataBaseView_ColumnValueExpression;
_currn->_desc1 = (_TSPdataBaseView_ColumnValueExpression_column) MkdataBaseView_ColumnValueExpression_column (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_ColumnValueExpression: root of subtree no. 1 can not be made a dataBaseView_ColumnValueExpression_column node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_ColumnValueExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_ColumnValueExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_ColumnValueExpression_column (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_ColumnValueExpression_column (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_ColumnValueExpression_column _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_ColumnValueExpression_column;
#else
_currn = (_TPPpdataBaseView_ColumnValueExpression_column) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_ColumnValueExpression_column));
#endif
_currn->_prod = RULEpdataBaseView_ColumnValueExpression_column;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_ColumnValueExpression_column;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_ColumnValueExpression_column */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_ParameterValueExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpdataBaseView_ParameterValueExpression (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpdataBaseView_ParameterValueExpression _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_ParameterValueExpression;
#else
_currn = (_TPPpdataBaseView_ParameterValueExpression) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_ParameterValueExpression));
#endif
_currn->_prod = RULEpdataBaseView_ParameterValueExpression;
_currn->_desc1 = (_TSPdataBaseView_ParameterValueExpression_parameter) MkdataBaseView_ParameterValueExpression_parameter (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_ParameterValueExpression: root of subtree no. 1 can not be made a dataBaseView_ParameterValueExpression_parameter node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_ParameterValueExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_ParameterValueExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_ParameterValueExpression_parameter (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_ParameterValueExpression_parameter (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_ParameterValueExpression_parameter _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_ParameterValueExpression_parameter;
#else
_currn = (_TPPpdataBaseView_ParameterValueExpression_parameter) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_ParameterValueExpression_parameter));
#endif
_currn->_prod = RULEpdataBaseView_ParameterValueExpression_parameter;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_ParameterValueExpression_parameter;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_ParameterValueExpression_parameter */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_ComputationExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4)
#else
NODEPTR MkpdataBaseView_ComputationExpression (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
#endif
{	_TPPpdataBaseView_ComputationExpression _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_ComputationExpression;
#else
_currn = (_TPPpdataBaseView_ComputationExpression) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_ComputationExpression));
#endif
_currn->_prod = RULEpdataBaseView_ComputationExpression;
_currn->_desc1 = (_TSPdataBaseView_ComputationExpression_computation) MkdataBaseView_ComputationExpression_computation (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_ComputationExpression: root of subtree no. 1 can not be made a dataBaseView_ComputationExpression_computation node ", 0, _coordref);
_currn->_desc2 = (_TSPdataBaseView_ComputationExpression_keyword1) MkdataBaseView_ComputationExpression_keyword1 (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_ComputationExpression: root of subtree no. 2 can not be made a dataBaseView_ComputationExpression_keyword1 node ", 0, _coordref);
_currn->_desc3 = (_TSPdataBaseView_ComputationExpression_column) MkdataBaseView_ComputationExpression_column (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_ComputationExpression: root of subtree no. 3 can not be made a dataBaseView_ComputationExpression_column node ", 0, _coordref);
_currn->_desc4 = (_TSPdataBaseView_ComputationExpression_keyword2) MkdataBaseView_ComputationExpression_keyword2 (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_ComputationExpression: root of subtree no. 4 can not be made a dataBaseView_ComputationExpression_keyword2 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_ComputationExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_ComputationExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_ComputationExpression_computation (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_ComputationExpression_computation (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_ComputationExpression_computation _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_ComputationExpression_computation;
#else
_currn = (_TPPpdataBaseView_ComputationExpression_computation) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_ComputationExpression_computation));
#endif
_currn->_prod = RULEpdataBaseView_ComputationExpression_computation;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_ComputationExpression_computation;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_ComputationExpression_computation */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_ComputationExpression_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_ComputationExpression_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_ComputationExpression_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_ComputationExpression_keyword1;
#else
_currn = (_TPPpdataBaseView_ComputationExpression_keyword1) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_ComputationExpression_keyword1));
#endif
_currn->_prod = RULEpdataBaseView_ComputationExpression_keyword1;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_ComputationExpression_keyword1;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_ComputationExpression_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_ComputationExpression_column (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_ComputationExpression_column (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_ComputationExpression_column _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_ComputationExpression_column;
#else
_currn = (_TPPpdataBaseView_ComputationExpression_column) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_ComputationExpression_column));
#endif
_currn->_prod = RULEpdataBaseView_ComputationExpression_column;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_ComputationExpression_column;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_ComputationExpression_column */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_ComputationExpression_keyword2 (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_ComputationExpression_keyword2 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_ComputationExpression_keyword2 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_ComputationExpression_keyword2;
#else
_currn = (_TPPpdataBaseView_ComputationExpression_keyword2) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_ComputationExpression_keyword2));
#endif
_currn->_prod = RULEpdataBaseView_ComputationExpression_keyword2;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_ComputationExpression_keyword2;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_ComputationExpression_keyword2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DBString (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpdataBaseView_DBString (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpdataBaseView_DBString _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DBString;
#else
_currn = (_TPPpdataBaseView_DBString) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DBString));
#endif
_currn->_prod = RULEpdataBaseView_DBString;
_currn->_desc1 = (_TSPdataBaseView_DBString_keyword1) MkdataBaseView_DBString_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DBString: root of subtree no. 1 can not be made a dataBaseView_DBString_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DBString;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DBString */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DBString_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DBString_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DBString_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DBString_keyword1;
#else
_currn = (_TPPpdataBaseView_DBString_keyword1) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DBString_keyword1));
#endif
_currn->_prod = RULEpdataBaseView_DBString_keyword1;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DBString_keyword1;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DBString_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DBInteger (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpdataBaseView_DBInteger (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpdataBaseView_DBInteger _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DBInteger;
#else
_currn = (_TPPpdataBaseView_DBInteger) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DBInteger));
#endif
_currn->_prod = RULEpdataBaseView_DBInteger;
_currn->_desc1 = (_TSPdataBaseView_DBInteger_keyword1) MkdataBaseView_DBInteger_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DBInteger: root of subtree no. 1 can not be made a dataBaseView_DBInteger_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DBInteger;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DBInteger */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DBInteger_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DBInteger_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DBInteger_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DBInteger_keyword1;
#else
_currn = (_TPPpdataBaseView_DBInteger_keyword1) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DBInteger_keyword1));
#endif
_currn->_prod = RULEpdataBaseView_DBInteger_keyword1;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DBInteger_keyword1;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DBInteger_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DBFloat (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpdataBaseView_DBFloat (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpdataBaseView_DBFloat _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DBFloat;
#else
_currn = (_TPPpdataBaseView_DBFloat) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DBFloat));
#endif
_currn->_prod = RULEpdataBaseView_DBFloat;
_currn->_desc1 = (_TSPdataBaseView_DBFloat_keyword1) MkdataBaseView_DBFloat_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DBFloat: root of subtree no. 1 can not be made a dataBaseView_DBFloat_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DBFloat;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DBFloat */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DBFloat_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DBFloat_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DBFloat_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DBFloat_keyword1;
#else
_currn = (_TPPpdataBaseView_DBFloat_keyword1) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DBFloat_keyword1));
#endif
_currn->_prod = RULEpdataBaseView_DBFloat_keyword1;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DBFloat_keyword1;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DBFloat_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DBDouble (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpdataBaseView_DBDouble (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpdataBaseView_DBDouble _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DBDouble;
#else
_currn = (_TPPpdataBaseView_DBDouble) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DBDouble));
#endif
_currn->_prod = RULEpdataBaseView_DBDouble;
_currn->_desc1 = (_TSPdataBaseView_DBDouble_keyword1) MkdataBaseView_DBDouble_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DBDouble: root of subtree no. 1 can not be made a dataBaseView_DBDouble_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DBDouble;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DBDouble */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DBDouble_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DBDouble_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DBDouble_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DBDouble_keyword1;
#else
_currn = (_TPPpdataBaseView_DBDouble_keyword1) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DBDouble_keyword1));
#endif
_currn->_prod = RULEpdataBaseView_DBDouble_keyword1;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DBDouble_keyword1;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DBDouble_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DBBigInt (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpdataBaseView_DBBigInt (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpdataBaseView_DBBigInt _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DBBigInt;
#else
_currn = (_TPPpdataBaseView_DBBigInt) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DBBigInt));
#endif
_currn->_prod = RULEpdataBaseView_DBBigInt;
_currn->_desc1 = (_TSPdataBaseView_DBBigInt_keyword1) MkdataBaseView_DBBigInt_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DBBigInt: root of subtree no. 1 can not be made a dataBaseView_DBBigInt_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DBBigInt;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DBBigInt */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DBBigInt_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DBBigInt_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DBBigInt_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DBBigInt_keyword1;
#else
_currn = (_TPPpdataBaseView_DBBigInt_keyword1) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DBBigInt_keyword1));
#endif
_currn->_prod = RULEpdataBaseView_DBBigInt_keyword1;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DBBigInt_keyword1;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DBBigInt_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DBDate (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpdataBaseView_DBDate (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpdataBaseView_DBDate _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DBDate;
#else
_currn = (_TPPpdataBaseView_DBDate) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DBDate));
#endif
_currn->_prod = RULEpdataBaseView_DBDate;
_currn->_desc1 = (_TSPdataBaseView_DBDate_keyword1) MkdataBaseView_DBDate_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DBDate: root of subtree no. 1 can not be made a dataBaseView_DBDate_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DBDate;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DBDate */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DBDate_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DBDate_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DBDate_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DBDate_keyword1;
#else
_currn = (_TPPpdataBaseView_DBDate_keyword1) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DBDate_keyword1));
#endif
_currn->_prod = RULEpdataBaseView_DBDate_keyword1;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DBDate_keyword1;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DBDate_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DBTimeStamp (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpdataBaseView_DBTimeStamp (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpdataBaseView_DBTimeStamp _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DBTimeStamp;
#else
_currn = (_TPPpdataBaseView_DBTimeStamp) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DBTimeStamp));
#endif
_currn->_prod = RULEpdataBaseView_DBTimeStamp;
_currn->_desc1 = (_TSPdataBaseView_DBTimeStamp_keyword1) MkdataBaseView_DBTimeStamp_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DBTimeStamp: root of subtree no. 1 can not be made a dataBaseView_DBTimeStamp_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DBTimeStamp;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DBTimeStamp */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DBTimeStamp_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DBTimeStamp_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DBTimeStamp_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DBTimeStamp_keyword1;
#else
_currn = (_TPPpdataBaseView_DBTimeStamp_keyword1) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DBTimeStamp_keyword1));
#endif
_currn->_prod = RULEpdataBaseView_DBTimeStamp_keyword1;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DBTimeStamp_keyword1;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DBTimeStamp_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DBBoolean (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpdataBaseView_DBBoolean (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpdataBaseView_DBBoolean _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DBBoolean;
#else
_currn = (_TPPpdataBaseView_DBBoolean) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DBBoolean));
#endif
_currn->_prod = RULEpdataBaseView_DBBoolean;
_currn->_desc1 = (_TSPdataBaseView_DBBoolean_keyword1) MkdataBaseView_DBBoolean_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DBBoolean: root of subtree no. 1 can not be made a dataBaseView_DBBoolean_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DBBoolean;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DBBoolean */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DBBoolean_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DBBoolean_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DBBoolean_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DBBoolean_keyword1;
#else
_currn = (_TPPpdataBaseView_DBBoolean_keyword1) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DBBoolean_keyword1));
#endif
_currn->_prod = RULEpdataBaseView_DBBoolean_keyword1;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DBBoolean_keyword1;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DBBoolean_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DBBlob (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpdataBaseView_DBBlob (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpdataBaseView_DBBlob _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DBBlob;
#else
_currn = (_TPPpdataBaseView_DBBlob) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DBBlob));
#endif
_currn->_prod = RULEpdataBaseView_DBBlob;
_currn->_desc1 = (_TSPdataBaseView_DBBlob_keyword1) MkdataBaseView_DBBlob_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DBBlob: root of subtree no. 1 can not be made a dataBaseView_DBBlob_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DBBlob;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DBBlob */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DBBlob_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DBBlob_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DBBlob_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DBBlob_keyword1;
#else
_currn = (_TPPpdataBaseView_DBBlob_keyword1) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DBBlob_keyword1));
#endif
_currn->_prod = RULEpdataBaseView_DBBlob_keyword1;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DBBlob_keyword1;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DBBlob_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpdataBaseView_DataSourceValue (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpdataBaseView_DataSourceValue _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceValue;
#else
_currn = (_TPPpdataBaseView_DataSourceValue) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceValue));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceValue;
_currn->_desc1 = (_TSPdataBaseView_DataSourceValue_column) MkdataBaseView_DataSourceValue_column (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceValue: root of subtree no. 1 can not be made a dataBaseView_DataSourceValue_column node ", 0, _coordref);
_currn->_desc2 = (_TSPdataBaseView_DataSourceValue_keyword1) MkdataBaseView_DataSourceValue_keyword1 (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceValue: root of subtree no. 2 can not be made a dataBaseView_DataSourceValue_keyword1 node ", 0, _coordref);
_currn->_desc3 = (_TSPdataBaseView_DataSourceValue_expression) MkdataBaseView_DataSourceValue_expression (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceValue: root of subtree no. 3 can not be made a dataBaseView_DataSourceValue_expression node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceValue_column (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataSourceValue_column (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataSourceValue_column _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceValue_column;
#else
_currn = (_TPPpdataBaseView_DataSourceValue_column) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceValue_column));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceValue_column;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceValue_column;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceValue_column */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceValue_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataSourceValue_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataSourceValue_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceValue_keyword1;
#else
_currn = (_TPPpdataBaseView_DataSourceValue_keyword1) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceValue_keyword1));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceValue_keyword1;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceValue_keyword1;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceValue_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_BinaryExpressionpdataBaseView_DataSourceValue_expression (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_BinaryExpressionpdataBaseView_DataSourceValue_expression (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceValue_expression _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceValue_expression;
#else
	_new = (_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceValue_expression) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceValue_expression));
#endif
	_new->_prod = RULELST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceValue_expression;
	_new->_desc1 = (_TSPdataBaseView_BinaryExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceValue_expression)_new;
	{_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceValue_expression _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceValue_expression;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_UnaryExpressionpdataBaseView_DataSourceValue_expression (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_UnaryExpressionpdataBaseView_DataSourceValue_expression (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceValue_expression _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceValue_expression;
#else
	_new = (_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceValue_expression) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceValue_expression));
#endif
	_new->_prod = RULELST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceValue_expression;
	_new->_desc1 = (_TSPdataBaseView_UnaryExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceValue_expression)_new;
	{_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceValue_expression _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceValue_expression;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ConstantExpressionpdataBaseView_DataSourceValue_expression (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ConstantExpressionpdataBaseView_DataSourceValue_expression (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceValue_expression _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceValue_expression;
#else
	_new = (_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceValue_expression) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceValue_expression));
#endif
	_new->_prod = RULELST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceValue_expression;
	_new->_desc1 = (_TSPdataBaseView_ConstantExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceValue_expression)_new;
	{_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceValue_expression _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceValue_expression;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ColumnValueExpressionpdataBaseView_DataSourceValue_expression (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ColumnValueExpressionpdataBaseView_DataSourceValue_expression (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceValue_expression _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceValue_expression;
#else
	_new = (_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceValue_expression) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceValue_expression));
#endif
	_new->_prod = RULELST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceValue_expression;
	_new->_desc1 = (_TSPdataBaseView_ColumnValueExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceValue_expression)_new;
	{_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceValue_expression _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceValue_expression;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ParameterValueExpressionpdataBaseView_DataSourceValue_expression (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ParameterValueExpressionpdataBaseView_DataSourceValue_expression (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceValue_expression _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceValue_expression;
#else
	_new = (_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceValue_expression) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceValue_expression));
#endif
	_new->_prod = RULELST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceValue_expression;
	_new->_desc1 = (_TSPdataBaseView_ParameterValueExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceValue_expression)_new;
	{_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceValue_expression _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceValue_expression;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ComputationExpressionpdataBaseView_DataSourceValue_expression (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ComputationExpressionpdataBaseView_DataSourceValue_expression (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceValue_expression _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceValue_expression;
#else
	_new = (_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceValue_expression) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceValue_expression));
#endif
	_new->_prod = RULELST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceValue_expression;
	_new->_desc1 = (_TSPdataBaseView_ComputationExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceValue_expression)_new;
	{_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceValue_expression _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceValue_expression;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempdataBaseView_DataSourceValue_expression (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempdataBaseView_DataSourceValue_expression (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBdataBaseView_BinaryExpression)) 
		return _MkdataBaseView_BinaryExpressionpdataBaseView_DataSourceValue_expression (_coordref, MkdataBaseView_BinaryExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_UnaryExpression)) 
		return _MkdataBaseView_UnaryExpressionpdataBaseView_DataSourceValue_expression (_coordref, MkdataBaseView_UnaryExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ConstantExpression)) 
		return _MkdataBaseView_ConstantExpressionpdataBaseView_DataSourceValue_expression (_coordref, MkdataBaseView_ConstantExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ColumnValueExpression)) 
		return _MkdataBaseView_ColumnValueExpressionpdataBaseView_DataSourceValue_expression (_coordref, MkdataBaseView_ColumnValueExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ParameterValueExpression)) 
		return _MkdataBaseView_ParameterValueExpressionpdataBaseView_DataSourceValue_expression (_coordref, MkdataBaseView_ParameterValueExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ComputationExpression)) 
		return _MkdataBaseView_ComputationExpressionpdataBaseView_DataSourceValue_expression (_coordref, MkdataBaseView_ComputationExpression (_coordref, t));
	message (DEADLY, "a pdataBaseView_DataSourceValue_expression list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pdataBaseView_DataSourceValue_expression (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pdataBaseView_DataSourceValue_expression (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_dataBaseView_DataSourceValue_expression)
	l =_ElempdataBaseView_DataSourceValue_expression (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_dataBaseView_DataSourceValue_expression)
	r = _ElempdataBaseView_DataSourceValue_expression (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pdataBaseView_DataSourceValue_expression)r)->_desc2);
((_TPPLST_0pdataBaseView_DataSourceValue_expression)r)->_desc2 = ((_TPPLST_0pdataBaseView_DataSourceValue_expression)l)->_desc2;
((_TPPLST_0pdataBaseView_DataSourceValue_expression)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceValue_expression (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpdataBaseView_DataSourceValue_expression (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpdataBaseView_DataSourceValue_expression root;
	_TPPLST_0pdataBaseView_DataSourceValue_expression list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pdataBaseView_DataSourceValue_expression)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_dataBaseView_DataSourceValue_expression)
		t = _ElempdataBaseView_DataSourceValue_expression (_coordref, t);
	remain = (_TPPLST_0pdataBaseView_DataSourceValue_expression)(((_TPPLST_0pdataBaseView_DataSourceValue_expression) t)->_desc2);
	((_TPPLST_0pdataBaseView_DataSourceValue_expression) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pdataBaseView_DataSourceValue_expression;
#else
list = (_TPPLST_0pdataBaseView_DataSourceValue_expression) TreeNodeAlloc (sizeof (struct _TPLST_0pdataBaseView_DataSourceValue_expression));
#endif
list->_prod = RULELST_0pdataBaseView_DataSourceValue_expression;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pdataBaseView_DataSourceValue_expression)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceValue_expression)
	{_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceValue_expression _currn = _currn;
	 _TERMACT_LST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceValue_expression;
	} this code has been moved to _ElempdataBaseView_DataSourceValue_expression*/
	/*if (_currn->_prod == RULELST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceValue_expression)
	{_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceValue_expression _currn = _currn;
	 _TERMACT_LST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceValue_expression;
	} this code has been moved to _ElempdataBaseView_DataSourceValue_expression*/
	/*if (_currn->_prod == RULELST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceValue_expression)
	{_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceValue_expression _currn = _currn;
	 _TERMACT_LST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceValue_expression;
	} this code has been moved to _ElempdataBaseView_DataSourceValue_expression*/
	/*if (_currn->_prod == RULELST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceValue_expression)
	{_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceValue_expression _currn = _currn;
	 _TERMACT_LST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceValue_expression;
	} this code has been moved to _ElempdataBaseView_DataSourceValue_expression*/
	/*if (_currn->_prod == RULELST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceValue_expression)
	{_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceValue_expression _currn = _currn;
	 _TERMACT_LST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceValue_expression;
	} this code has been moved to _ElempdataBaseView_DataSourceValue_expression*/
	/*if (_currn->_prod == RULELST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceValue_expression)
	{_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceValue_expression _currn = _currn;
	 _TERMACT_LST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceValue_expression;
	} this code has been moved to _ElempdataBaseView_DataSourceValue_expression*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpdataBaseView_DataSourceValue_expression;
#else
root = (_TPPpdataBaseView_DataSourceValue_expression) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceValue_expression));
#endif
root->_prod = RULEpdataBaseView_DataSourceValue_expression;
root->_desc1 = (_TSPLST_dataBaseView_DataSourceValue_expression)list;
{ _TPPpdataBaseView_DataSourceValue_expression _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pdataBaseView_DataSourceValue_expression;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceProjectionFilter (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkpdataBaseView_DataSourceProjectionFilter (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPpdataBaseView_DataSourceProjectionFilter _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceProjectionFilter;
#else
_currn = (_TPPpdataBaseView_DataSourceProjectionFilter) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceProjectionFilter));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceProjectionFilter;
_currn->_desc1 = (_TSPdataBaseView_DataSourceProjectionFilter_name) MkdataBaseView_DataSourceProjectionFilter_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceProjectionFilter: root of subtree no. 1 can not be made a dataBaseView_DataSourceProjectionFilter_name node ", 0, _coordref);
_currn->_desc2 = (_TSPdataBaseView_DataSourceProjectionFilter_formalParameters) MkdataBaseView_DataSourceProjectionFilter_formalParameters (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceProjectionFilter: root of subtree no. 2 can not be made a dataBaseView_DataSourceProjectionFilter_formalParameters node ", 0, _coordref);
_currn->_desc3 = (_TSPdataBaseView_DataSourceProjectionFilter_projectionColumns) MkdataBaseView_DataSourceProjectionFilter_projectionColumns (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceProjectionFilter: root of subtree no. 3 can not be made a dataBaseView_DataSourceProjectionFilter_projectionColumns node ", 0, _coordref);
_currn->_desc4 = (_TSPdataBaseView_DataSourceProjectionFilter_condition) MkdataBaseView_DataSourceProjectionFilter_condition (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceProjectionFilter: root of subtree no. 4 can not be made a dataBaseView_DataSourceProjectionFilter_condition node ", 0, _coordref);
_currn->_desc5 = (_TSPdataBaseView_DataSourceProjectionFilter_connection) MkdataBaseView_DataSourceProjectionFilter_connection (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceProjectionFilter: root of subtree no. 5 can not be made a dataBaseView_DataSourceProjectionFilter_connection node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceProjectionFilter;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceProjectionFilter */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceProjectionFilter_name (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataSourceProjectionFilter_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataSourceProjectionFilter_name _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceProjectionFilter_name;
#else
_currn = (_TPPpdataBaseView_DataSourceProjectionFilter_name) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceProjectionFilter_name));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceProjectionFilter_name;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceProjectionFilter_name;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceProjectionFilter_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DataSourceFormalParampdataBaseView_DataSourceProjectionFilter_formalParameters (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DataSourceFormalParampdataBaseView_DataSourceProjectionFilter_formalParameters (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceProjectionFilter_formalParameters _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceProjectionFilter_formalParameters;
#else
	_new = (_TPPLST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceProjectionFilter_formalParameters) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceProjectionFilter_formalParameters));
#endif
	_new->_prod = RULELST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceProjectionFilter_formalParameters;
	_new->_desc1 = (_TSPdataBaseView_DataSourceFormalParam)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceProjectionFilter_formalParameters)_new;
	{_TPPLST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceProjectionFilter_formalParameters _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceProjectionFilter_formalParameters;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempdataBaseView_DataSourceProjectionFilter_formalParameters (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempdataBaseView_DataSourceProjectionFilter_formalParameters (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBdataBaseView_DataSourceFormalParam)) 
		return _MkdataBaseView_DataSourceFormalParampdataBaseView_DataSourceProjectionFilter_formalParameters (_coordref, MkdataBaseView_DataSourceFormalParam (_coordref, t));
	message (DEADLY, "a pdataBaseView_DataSourceProjectionFilter_formalParameters list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pdataBaseView_DataSourceProjectionFilter_formalParameters (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pdataBaseView_DataSourceProjectionFilter_formalParameters (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_dataBaseView_DataSourceProjectionFilter_formalParameters)
	l =_ElempdataBaseView_DataSourceProjectionFilter_formalParameters (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_dataBaseView_DataSourceProjectionFilter_formalParameters)
	r = _ElempdataBaseView_DataSourceProjectionFilter_formalParameters (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pdataBaseView_DataSourceProjectionFilter_formalParameters)r)->_desc2);
((_TPPLST_0pdataBaseView_DataSourceProjectionFilter_formalParameters)r)->_desc2 = ((_TPPLST_0pdataBaseView_DataSourceProjectionFilter_formalParameters)l)->_desc2;
((_TPPLST_0pdataBaseView_DataSourceProjectionFilter_formalParameters)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceProjectionFilter_formalParameters (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpdataBaseView_DataSourceProjectionFilter_formalParameters (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpdataBaseView_DataSourceProjectionFilter_formalParameters root;
	_TPPLST_0pdataBaseView_DataSourceProjectionFilter_formalParameters list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pdataBaseView_DataSourceProjectionFilter_formalParameters)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_dataBaseView_DataSourceProjectionFilter_formalParameters)
		t = _ElempdataBaseView_DataSourceProjectionFilter_formalParameters (_coordref, t);
	remain = (_TPPLST_0pdataBaseView_DataSourceProjectionFilter_formalParameters)(((_TPPLST_0pdataBaseView_DataSourceProjectionFilter_formalParameters) t)->_desc2);
	((_TPPLST_0pdataBaseView_DataSourceProjectionFilter_formalParameters) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pdataBaseView_DataSourceProjectionFilter_formalParameters;
#else
list = (_TPPLST_0pdataBaseView_DataSourceProjectionFilter_formalParameters) TreeNodeAlloc (sizeof (struct _TPLST_0pdataBaseView_DataSourceProjectionFilter_formalParameters));
#endif
list->_prod = RULELST_0pdataBaseView_DataSourceProjectionFilter_formalParameters;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pdataBaseView_DataSourceProjectionFilter_formalParameters)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceProjectionFilter_formalParameters)
	{_TPPLST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceProjectionFilter_formalParameters _currn = _currn;
	 _TERMACT_LST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceProjectionFilter_formalParameters;
	} this code has been moved to _ElempdataBaseView_DataSourceProjectionFilter_formalParameters*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpdataBaseView_DataSourceProjectionFilter_formalParameters;
#else
root = (_TPPpdataBaseView_DataSourceProjectionFilter_formalParameters) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceProjectionFilter_formalParameters));
#endif
root->_prod = RULEpdataBaseView_DataSourceProjectionFilter_formalParameters;
root->_desc1 = (_TSPLST_dataBaseView_DataSourceProjectionFilter_formalParameters)list;
{ _TPPpdataBaseView_DataSourceProjectionFilter_formalParameters _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pdataBaseView_DataSourceProjectionFilter_formalParameters;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ColumnProjectionpdataBaseView_DataSourceProjectionFilter_projectionColumns (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ColumnProjectionpdataBaseView_DataSourceProjectionFilter_projectionColumns (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ColumnProjectionpdataBaseView_DataSourceProjectionFilter_projectionColumns _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ColumnProjectionpdataBaseView_DataSourceProjectionFilter_projectionColumns;
#else
	_new = (_TPPLST_dataBaseView_ColumnProjectionpdataBaseView_DataSourceProjectionFilter_projectionColumns) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ColumnProjectionpdataBaseView_DataSourceProjectionFilter_projectionColumns));
#endif
	_new->_prod = RULELST_dataBaseView_ColumnProjectionpdataBaseView_DataSourceProjectionFilter_projectionColumns;
	_new->_desc1 = (_TSPdataBaseView_ColumnProjection)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceProjectionFilter_projectionColumns)_new;
	{_TPPLST_dataBaseView_ColumnProjectionpdataBaseView_DataSourceProjectionFilter_projectionColumns _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ColumnProjectionpdataBaseView_DataSourceProjectionFilter_projectionColumns;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ComputationProjectionpdataBaseView_DataSourceProjectionFilter_projectionColumns (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ComputationProjectionpdataBaseView_DataSourceProjectionFilter_projectionColumns (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ComputationProjectionpdataBaseView_DataSourceProjectionFilter_projectionColumns _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ComputationProjectionpdataBaseView_DataSourceProjectionFilter_projectionColumns;
#else
	_new = (_TPPLST_dataBaseView_ComputationProjectionpdataBaseView_DataSourceProjectionFilter_projectionColumns) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ComputationProjectionpdataBaseView_DataSourceProjectionFilter_projectionColumns));
#endif
	_new->_prod = RULELST_dataBaseView_ComputationProjectionpdataBaseView_DataSourceProjectionFilter_projectionColumns;
	_new->_desc1 = (_TSPdataBaseView_ComputationProjection)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceProjectionFilter_projectionColumns)_new;
	{_TPPLST_dataBaseView_ComputationProjectionpdataBaseView_DataSourceProjectionFilter_projectionColumns _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ComputationProjectionpdataBaseView_DataSourceProjectionFilter_projectionColumns;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempdataBaseView_DataSourceProjectionFilter_projectionColumns (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempdataBaseView_DataSourceProjectionFilter_projectionColumns (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBdataBaseView_ColumnProjection)) 
		return _MkdataBaseView_ColumnProjectionpdataBaseView_DataSourceProjectionFilter_projectionColumns (_coordref, MkdataBaseView_ColumnProjection (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ComputationProjection)) 
		return _MkdataBaseView_ComputationProjectionpdataBaseView_DataSourceProjectionFilter_projectionColumns (_coordref, MkdataBaseView_ComputationProjection (_coordref, t));
	message (DEADLY, "a pdataBaseView_DataSourceProjectionFilter_projectionColumns list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pdataBaseView_DataSourceProjectionFilter_projectionColumns (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pdataBaseView_DataSourceProjectionFilter_projectionColumns (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_dataBaseView_DataSourceProjectionFilter_projectionColumns)
	l =_ElempdataBaseView_DataSourceProjectionFilter_projectionColumns (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_dataBaseView_DataSourceProjectionFilter_projectionColumns)
	r = _ElempdataBaseView_DataSourceProjectionFilter_projectionColumns (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pdataBaseView_DataSourceProjectionFilter_projectionColumns)r)->_desc2);
((_TPPLST_0pdataBaseView_DataSourceProjectionFilter_projectionColumns)r)->_desc2 = ((_TPPLST_0pdataBaseView_DataSourceProjectionFilter_projectionColumns)l)->_desc2;
((_TPPLST_0pdataBaseView_DataSourceProjectionFilter_projectionColumns)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceProjectionFilter_projectionColumns (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpdataBaseView_DataSourceProjectionFilter_projectionColumns (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpdataBaseView_DataSourceProjectionFilter_projectionColumns root;
	_TPPLST_0pdataBaseView_DataSourceProjectionFilter_projectionColumns list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pdataBaseView_DataSourceProjectionFilter_projectionColumns)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_dataBaseView_DataSourceProjectionFilter_projectionColumns)
		t = _ElempdataBaseView_DataSourceProjectionFilter_projectionColumns (_coordref, t);
	remain = (_TPPLST_0pdataBaseView_DataSourceProjectionFilter_projectionColumns)(((_TPPLST_0pdataBaseView_DataSourceProjectionFilter_projectionColumns) t)->_desc2);
	((_TPPLST_0pdataBaseView_DataSourceProjectionFilter_projectionColumns) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pdataBaseView_DataSourceProjectionFilter_projectionColumns;
#else
list = (_TPPLST_0pdataBaseView_DataSourceProjectionFilter_projectionColumns) TreeNodeAlloc (sizeof (struct _TPLST_0pdataBaseView_DataSourceProjectionFilter_projectionColumns));
#endif
list->_prod = RULELST_0pdataBaseView_DataSourceProjectionFilter_projectionColumns;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pdataBaseView_DataSourceProjectionFilter_projectionColumns)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_dataBaseView_ColumnProjectionpdataBaseView_DataSourceProjectionFilter_projectionColumns)
	{_TPPLST_dataBaseView_ColumnProjectionpdataBaseView_DataSourceProjectionFilter_projectionColumns _currn = _currn;
	 _TERMACT_LST_dataBaseView_ColumnProjectionpdataBaseView_DataSourceProjectionFilter_projectionColumns;
	} this code has been moved to _ElempdataBaseView_DataSourceProjectionFilter_projectionColumns*/
	/*if (_currn->_prod == RULELST_dataBaseView_ComputationProjectionpdataBaseView_DataSourceProjectionFilter_projectionColumns)
	{_TPPLST_dataBaseView_ComputationProjectionpdataBaseView_DataSourceProjectionFilter_projectionColumns _currn = _currn;
	 _TERMACT_LST_dataBaseView_ComputationProjectionpdataBaseView_DataSourceProjectionFilter_projectionColumns;
	} this code has been moved to _ElempdataBaseView_DataSourceProjectionFilter_projectionColumns*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpdataBaseView_DataSourceProjectionFilter_projectionColumns;
#else
root = (_TPPpdataBaseView_DataSourceProjectionFilter_projectionColumns) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceProjectionFilter_projectionColumns));
#endif
root->_prod = RULEpdataBaseView_DataSourceProjectionFilter_projectionColumns;
root->_desc1 = (_TSPLST_dataBaseView_DataSourceProjectionFilter_projectionColumns)list;
{ _TPPpdataBaseView_DataSourceProjectionFilter_projectionColumns _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pdataBaseView_DataSourceProjectionFilter_projectionColumns;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_BinaryExpressionpdataBaseView_DataSourceProjectionFilter_condition (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_BinaryExpressionpdataBaseView_DataSourceProjectionFilter_condition (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceProjectionFilter_condition _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceProjectionFilter_condition;
#else
	_new = (_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceProjectionFilter_condition) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceProjectionFilter_condition));
#endif
	_new->_prod = RULELST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceProjectionFilter_condition;
	_new->_desc1 = (_TSPdataBaseView_BinaryExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceProjectionFilter_condition)_new;
	{_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceProjectionFilter_condition _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceProjectionFilter_condition;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_UnaryExpressionpdataBaseView_DataSourceProjectionFilter_condition (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_UnaryExpressionpdataBaseView_DataSourceProjectionFilter_condition (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceProjectionFilter_condition _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceProjectionFilter_condition;
#else
	_new = (_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceProjectionFilter_condition) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceProjectionFilter_condition));
#endif
	_new->_prod = RULELST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceProjectionFilter_condition;
	_new->_desc1 = (_TSPdataBaseView_UnaryExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceProjectionFilter_condition)_new;
	{_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceProjectionFilter_condition _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceProjectionFilter_condition;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ConstantExpressionpdataBaseView_DataSourceProjectionFilter_condition (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ConstantExpressionpdataBaseView_DataSourceProjectionFilter_condition (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceProjectionFilter_condition _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceProjectionFilter_condition;
#else
	_new = (_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceProjectionFilter_condition) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceProjectionFilter_condition));
#endif
	_new->_prod = RULELST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceProjectionFilter_condition;
	_new->_desc1 = (_TSPdataBaseView_ConstantExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceProjectionFilter_condition)_new;
	{_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceProjectionFilter_condition _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceProjectionFilter_condition;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ColumnValueExpressionpdataBaseView_DataSourceProjectionFilter_condition (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ColumnValueExpressionpdataBaseView_DataSourceProjectionFilter_condition (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceProjectionFilter_condition _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceProjectionFilter_condition;
#else
	_new = (_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceProjectionFilter_condition) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceProjectionFilter_condition));
#endif
	_new->_prod = RULELST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceProjectionFilter_condition;
	_new->_desc1 = (_TSPdataBaseView_ColumnValueExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceProjectionFilter_condition)_new;
	{_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceProjectionFilter_condition _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceProjectionFilter_condition;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ParameterValueExpressionpdataBaseView_DataSourceProjectionFilter_condition (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ParameterValueExpressionpdataBaseView_DataSourceProjectionFilter_condition (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceProjectionFilter_condition _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceProjectionFilter_condition;
#else
	_new = (_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceProjectionFilter_condition) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceProjectionFilter_condition));
#endif
	_new->_prod = RULELST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceProjectionFilter_condition;
	_new->_desc1 = (_TSPdataBaseView_ParameterValueExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceProjectionFilter_condition)_new;
	{_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceProjectionFilter_condition _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceProjectionFilter_condition;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ComputationExpressionpdataBaseView_DataSourceProjectionFilter_condition (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ComputationExpressionpdataBaseView_DataSourceProjectionFilter_condition (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceProjectionFilter_condition _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceProjectionFilter_condition;
#else
	_new = (_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceProjectionFilter_condition) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceProjectionFilter_condition));
#endif
	_new->_prod = RULELST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceProjectionFilter_condition;
	_new->_desc1 = (_TSPdataBaseView_ComputationExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceProjectionFilter_condition)_new;
	{_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceProjectionFilter_condition _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceProjectionFilter_condition;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempdataBaseView_DataSourceProjectionFilter_condition (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempdataBaseView_DataSourceProjectionFilter_condition (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBdataBaseView_BinaryExpression)) 
		return _MkdataBaseView_BinaryExpressionpdataBaseView_DataSourceProjectionFilter_condition (_coordref, MkdataBaseView_BinaryExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_UnaryExpression)) 
		return _MkdataBaseView_UnaryExpressionpdataBaseView_DataSourceProjectionFilter_condition (_coordref, MkdataBaseView_UnaryExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ConstantExpression)) 
		return _MkdataBaseView_ConstantExpressionpdataBaseView_DataSourceProjectionFilter_condition (_coordref, MkdataBaseView_ConstantExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ColumnValueExpression)) 
		return _MkdataBaseView_ColumnValueExpressionpdataBaseView_DataSourceProjectionFilter_condition (_coordref, MkdataBaseView_ColumnValueExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ParameterValueExpression)) 
		return _MkdataBaseView_ParameterValueExpressionpdataBaseView_DataSourceProjectionFilter_condition (_coordref, MkdataBaseView_ParameterValueExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ComputationExpression)) 
		return _MkdataBaseView_ComputationExpressionpdataBaseView_DataSourceProjectionFilter_condition (_coordref, MkdataBaseView_ComputationExpression (_coordref, t));
	message (DEADLY, "a pdataBaseView_DataSourceProjectionFilter_condition list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pdataBaseView_DataSourceProjectionFilter_condition (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pdataBaseView_DataSourceProjectionFilter_condition (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_dataBaseView_DataSourceProjectionFilter_condition)
	l =_ElempdataBaseView_DataSourceProjectionFilter_condition (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_dataBaseView_DataSourceProjectionFilter_condition)
	r = _ElempdataBaseView_DataSourceProjectionFilter_condition (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pdataBaseView_DataSourceProjectionFilter_condition)r)->_desc2);
((_TPPLST_0pdataBaseView_DataSourceProjectionFilter_condition)r)->_desc2 = ((_TPPLST_0pdataBaseView_DataSourceProjectionFilter_condition)l)->_desc2;
((_TPPLST_0pdataBaseView_DataSourceProjectionFilter_condition)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceProjectionFilter_condition (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpdataBaseView_DataSourceProjectionFilter_condition (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpdataBaseView_DataSourceProjectionFilter_condition root;
	_TPPLST_0pdataBaseView_DataSourceProjectionFilter_condition list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pdataBaseView_DataSourceProjectionFilter_condition)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_dataBaseView_DataSourceProjectionFilter_condition)
		t = _ElempdataBaseView_DataSourceProjectionFilter_condition (_coordref, t);
	remain = (_TPPLST_0pdataBaseView_DataSourceProjectionFilter_condition)(((_TPPLST_0pdataBaseView_DataSourceProjectionFilter_condition) t)->_desc2);
	((_TPPLST_0pdataBaseView_DataSourceProjectionFilter_condition) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pdataBaseView_DataSourceProjectionFilter_condition;
#else
list = (_TPPLST_0pdataBaseView_DataSourceProjectionFilter_condition) TreeNodeAlloc (sizeof (struct _TPLST_0pdataBaseView_DataSourceProjectionFilter_condition));
#endif
list->_prod = RULELST_0pdataBaseView_DataSourceProjectionFilter_condition;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pdataBaseView_DataSourceProjectionFilter_condition)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceProjectionFilter_condition)
	{_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceProjectionFilter_condition _currn = _currn;
	 _TERMACT_LST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceProjectionFilter_condition;
	} this code has been moved to _ElempdataBaseView_DataSourceProjectionFilter_condition*/
	/*if (_currn->_prod == RULELST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceProjectionFilter_condition)
	{_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceProjectionFilter_condition _currn = _currn;
	 _TERMACT_LST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceProjectionFilter_condition;
	} this code has been moved to _ElempdataBaseView_DataSourceProjectionFilter_condition*/
	/*if (_currn->_prod == RULELST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceProjectionFilter_condition)
	{_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceProjectionFilter_condition _currn = _currn;
	 _TERMACT_LST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceProjectionFilter_condition;
	} this code has been moved to _ElempdataBaseView_DataSourceProjectionFilter_condition*/
	/*if (_currn->_prod == RULELST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceProjectionFilter_condition)
	{_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceProjectionFilter_condition _currn = _currn;
	 _TERMACT_LST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceProjectionFilter_condition;
	} this code has been moved to _ElempdataBaseView_DataSourceProjectionFilter_condition*/
	/*if (_currn->_prod == RULELST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceProjectionFilter_condition)
	{_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceProjectionFilter_condition _currn = _currn;
	 _TERMACT_LST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceProjectionFilter_condition;
	} this code has been moved to _ElempdataBaseView_DataSourceProjectionFilter_condition*/
	/*if (_currn->_prod == RULELST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceProjectionFilter_condition)
	{_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceProjectionFilter_condition _currn = _currn;
	 _TERMACT_LST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceProjectionFilter_condition;
	} this code has been moved to _ElempdataBaseView_DataSourceProjectionFilter_condition*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpdataBaseView_DataSourceProjectionFilter_condition;
#else
root = (_TPPpdataBaseView_DataSourceProjectionFilter_condition) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceProjectionFilter_condition));
#endif
root->_prod = RULEpdataBaseView_DataSourceProjectionFilter_condition;
root->_desc1 = (_TSPLST_dataBaseView_DataSourceProjectionFilter_condition)list;
{ _TPPpdataBaseView_DataSourceProjectionFilter_condition _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pdataBaseView_DataSourceProjectionFilter_condition;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceProjectionFilter_from (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataSourceProjectionFilter_from (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataSourceProjectionFilter_from _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceProjectionFilter_from;
#else
_currn = (_TPPpdataBaseView_DataSourceProjectionFilter_from) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceProjectionFilter_from));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceProjectionFilter_from;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceProjectionFilter_from;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceProjectionFilter_from */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceProjectionFilter_sourceDataSource (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataSourceProjectionFilter_sourceDataSource (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataSourceProjectionFilter_sourceDataSource _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceProjectionFilter_sourceDataSource;
#else
_currn = (_TPPpdataBaseView_DataSourceProjectionFilter_sourceDataSource) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceProjectionFilter_sourceDataSource));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceProjectionFilter_sourceDataSource;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceProjectionFilter_sourceDataSource;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceProjectionFilter_sourceDataSource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceProjectionFilter_connection (POSITION *_coordref, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpdataBaseView_DataSourceProjectionFilter_connection (_coordref,_desc1,_desc2)
	POSITION *_coordref;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpdataBaseView_DataSourceProjectionFilter_connection _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceProjectionFilter_connection;
#else
_currn = (_TPPpdataBaseView_DataSourceProjectionFilter_connection) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceProjectionFilter_connection));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceProjectionFilter_connection;
_currn->_desc1 = (_TSPdataBaseView_DataSourceProjectionFilter_from) MkdataBaseView_DataSourceProjectionFilter_from (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceProjectionFilter_connection: root of subtree no. 1 can not be made a dataBaseView_DataSourceProjectionFilter_from node ", 0, _coordref);
_currn->_desc2 = (_TSPdataBaseView_DataSourceProjectionFilter_sourceDataSource) MkdataBaseView_DataSourceProjectionFilter_sourceDataSource (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceProjectionFilter_connection: root of subtree no. 2 can not be made a dataBaseView_DataSourceProjectionFilter_sourceDataSource node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceProjectionFilter_connection;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceProjectionFilter_connection */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_LimitDataSource (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4)
#else
NODEPTR MkpdataBaseView_LimitDataSource (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
#endif
{	_TPPpdataBaseView_LimitDataSource _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_LimitDataSource;
#else
_currn = (_TPPpdataBaseView_LimitDataSource) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_LimitDataSource));
#endif
_currn->_prod = RULEpdataBaseView_LimitDataSource;
_currn->_desc1 = (_TSPdataBaseView_LimitDataSource_name) MkdataBaseView_LimitDataSource_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_LimitDataSource: root of subtree no. 1 can not be made a dataBaseView_LimitDataSource_name node ", 0, _coordref);
_currn->_desc2 = (_TSPdataBaseView_LimitDataSource_formalParameters) MkdataBaseView_LimitDataSource_formalParameters (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_LimitDataSource: root of subtree no. 2 can not be made a dataBaseView_LimitDataSource_formalParameters node ", 0, _coordref);
_currn->_desc3 = (_TSPdataBaseView_LimitDataSource_condition) MkdataBaseView_LimitDataSource_condition (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_LimitDataSource: root of subtree no. 3 can not be made a dataBaseView_LimitDataSource_condition node ", 0, _coordref);
_currn->_desc4 = (_TSPdataBaseView_LimitDataSource_connection) MkdataBaseView_LimitDataSource_connection (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_LimitDataSource: root of subtree no. 4 can not be made a dataBaseView_LimitDataSource_connection node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_LimitDataSource;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_LimitDataSource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_LimitDataSource_name (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_LimitDataSource_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_LimitDataSource_name _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_LimitDataSource_name;
#else
_currn = (_TPPpdataBaseView_LimitDataSource_name) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_LimitDataSource_name));
#endif
_currn->_prod = RULEpdataBaseView_LimitDataSource_name;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_LimitDataSource_name;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_LimitDataSource_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DataSourceFormalParampdataBaseView_LimitDataSource_formalParameters (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DataSourceFormalParampdataBaseView_LimitDataSource_formalParameters (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DataSourceFormalParampdataBaseView_LimitDataSource_formalParameters _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DataSourceFormalParampdataBaseView_LimitDataSource_formalParameters;
#else
	_new = (_TPPLST_dataBaseView_DataSourceFormalParampdataBaseView_LimitDataSource_formalParameters) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DataSourceFormalParampdataBaseView_LimitDataSource_formalParameters));
#endif
	_new->_prod = RULELST_dataBaseView_DataSourceFormalParampdataBaseView_LimitDataSource_formalParameters;
	_new->_desc1 = (_TSPdataBaseView_DataSourceFormalParam)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_LimitDataSource_formalParameters)_new;
	{_TPPLST_dataBaseView_DataSourceFormalParampdataBaseView_LimitDataSource_formalParameters _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DataSourceFormalParampdataBaseView_LimitDataSource_formalParameters;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempdataBaseView_LimitDataSource_formalParameters (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempdataBaseView_LimitDataSource_formalParameters (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBdataBaseView_DataSourceFormalParam)) 
		return _MkdataBaseView_DataSourceFormalParampdataBaseView_LimitDataSource_formalParameters (_coordref, MkdataBaseView_DataSourceFormalParam (_coordref, t));
	message (DEADLY, "a pdataBaseView_LimitDataSource_formalParameters list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pdataBaseView_LimitDataSource_formalParameters (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pdataBaseView_LimitDataSource_formalParameters (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_dataBaseView_LimitDataSource_formalParameters)
	l =_ElempdataBaseView_LimitDataSource_formalParameters (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_dataBaseView_LimitDataSource_formalParameters)
	r = _ElempdataBaseView_LimitDataSource_formalParameters (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pdataBaseView_LimitDataSource_formalParameters)r)->_desc2);
((_TPPLST_0pdataBaseView_LimitDataSource_formalParameters)r)->_desc2 = ((_TPPLST_0pdataBaseView_LimitDataSource_formalParameters)l)->_desc2;
((_TPPLST_0pdataBaseView_LimitDataSource_formalParameters)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_LimitDataSource_formalParameters (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpdataBaseView_LimitDataSource_formalParameters (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpdataBaseView_LimitDataSource_formalParameters root;
	_TPPLST_0pdataBaseView_LimitDataSource_formalParameters list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pdataBaseView_LimitDataSource_formalParameters)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_dataBaseView_LimitDataSource_formalParameters)
		t = _ElempdataBaseView_LimitDataSource_formalParameters (_coordref, t);
	remain = (_TPPLST_0pdataBaseView_LimitDataSource_formalParameters)(((_TPPLST_0pdataBaseView_LimitDataSource_formalParameters) t)->_desc2);
	((_TPPLST_0pdataBaseView_LimitDataSource_formalParameters) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pdataBaseView_LimitDataSource_formalParameters;
#else
list = (_TPPLST_0pdataBaseView_LimitDataSource_formalParameters) TreeNodeAlloc (sizeof (struct _TPLST_0pdataBaseView_LimitDataSource_formalParameters));
#endif
list->_prod = RULELST_0pdataBaseView_LimitDataSource_formalParameters;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pdataBaseView_LimitDataSource_formalParameters)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_dataBaseView_DataSourceFormalParampdataBaseView_LimitDataSource_formalParameters)
	{_TPPLST_dataBaseView_DataSourceFormalParampdataBaseView_LimitDataSource_formalParameters _currn = _currn;
	 _TERMACT_LST_dataBaseView_DataSourceFormalParampdataBaseView_LimitDataSource_formalParameters;
	} this code has been moved to _ElempdataBaseView_LimitDataSource_formalParameters*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpdataBaseView_LimitDataSource_formalParameters;
#else
root = (_TPPpdataBaseView_LimitDataSource_formalParameters) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_LimitDataSource_formalParameters));
#endif
root->_prod = RULEpdataBaseView_LimitDataSource_formalParameters;
root->_desc1 = (_TSPLST_dataBaseView_LimitDataSource_formalParameters)list;
{ _TPPpdataBaseView_LimitDataSource_formalParameters _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pdataBaseView_LimitDataSource_formalParameters;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_BinaryExpressionpdataBaseView_LimitDataSource_minRow (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_BinaryExpressionpdataBaseView_LimitDataSource_minRow (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_LimitDataSource_minRow _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_BinaryExpressionpdataBaseView_LimitDataSource_minRow;
#else
	_new = (_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_LimitDataSource_minRow) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_BinaryExpressionpdataBaseView_LimitDataSource_minRow));
#endif
	_new->_prod = RULELST_dataBaseView_BinaryExpressionpdataBaseView_LimitDataSource_minRow;
	_new->_desc1 = (_TSPdataBaseView_BinaryExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_LimitDataSource_minRow)_new;
	{_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_LimitDataSource_minRow _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_BinaryExpressionpdataBaseView_LimitDataSource_minRow;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_UnaryExpressionpdataBaseView_LimitDataSource_minRow (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_UnaryExpressionpdataBaseView_LimitDataSource_minRow (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_LimitDataSource_minRow _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_UnaryExpressionpdataBaseView_LimitDataSource_minRow;
#else
	_new = (_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_LimitDataSource_minRow) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_UnaryExpressionpdataBaseView_LimitDataSource_minRow));
#endif
	_new->_prod = RULELST_dataBaseView_UnaryExpressionpdataBaseView_LimitDataSource_minRow;
	_new->_desc1 = (_TSPdataBaseView_UnaryExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_LimitDataSource_minRow)_new;
	{_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_LimitDataSource_minRow _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_UnaryExpressionpdataBaseView_LimitDataSource_minRow;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ConstantExpressionpdataBaseView_LimitDataSource_minRow (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ConstantExpressionpdataBaseView_LimitDataSource_minRow (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_LimitDataSource_minRow _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ConstantExpressionpdataBaseView_LimitDataSource_minRow;
#else
	_new = (_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_LimitDataSource_minRow) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ConstantExpressionpdataBaseView_LimitDataSource_minRow));
#endif
	_new->_prod = RULELST_dataBaseView_ConstantExpressionpdataBaseView_LimitDataSource_minRow;
	_new->_desc1 = (_TSPdataBaseView_ConstantExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_LimitDataSource_minRow)_new;
	{_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_LimitDataSource_minRow _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ConstantExpressionpdataBaseView_LimitDataSource_minRow;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ColumnValueExpressionpdataBaseView_LimitDataSource_minRow (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ColumnValueExpressionpdataBaseView_LimitDataSource_minRow (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_LimitDataSource_minRow _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ColumnValueExpressionpdataBaseView_LimitDataSource_minRow;
#else
	_new = (_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_LimitDataSource_minRow) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ColumnValueExpressionpdataBaseView_LimitDataSource_minRow));
#endif
	_new->_prod = RULELST_dataBaseView_ColumnValueExpressionpdataBaseView_LimitDataSource_minRow;
	_new->_desc1 = (_TSPdataBaseView_ColumnValueExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_LimitDataSource_minRow)_new;
	{_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_LimitDataSource_minRow _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ColumnValueExpressionpdataBaseView_LimitDataSource_minRow;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ParameterValueExpressionpdataBaseView_LimitDataSource_minRow (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ParameterValueExpressionpdataBaseView_LimitDataSource_minRow (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_LimitDataSource_minRow _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ParameterValueExpressionpdataBaseView_LimitDataSource_minRow;
#else
	_new = (_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_LimitDataSource_minRow) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ParameterValueExpressionpdataBaseView_LimitDataSource_minRow));
#endif
	_new->_prod = RULELST_dataBaseView_ParameterValueExpressionpdataBaseView_LimitDataSource_minRow;
	_new->_desc1 = (_TSPdataBaseView_ParameterValueExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_LimitDataSource_minRow)_new;
	{_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_LimitDataSource_minRow _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ParameterValueExpressionpdataBaseView_LimitDataSource_minRow;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ComputationExpressionpdataBaseView_LimitDataSource_minRow (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ComputationExpressionpdataBaseView_LimitDataSource_minRow (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_LimitDataSource_minRow _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ComputationExpressionpdataBaseView_LimitDataSource_minRow;
#else
	_new = (_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_LimitDataSource_minRow) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ComputationExpressionpdataBaseView_LimitDataSource_minRow));
#endif
	_new->_prod = RULELST_dataBaseView_ComputationExpressionpdataBaseView_LimitDataSource_minRow;
	_new->_desc1 = (_TSPdataBaseView_ComputationExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_LimitDataSource_minRow)_new;
	{_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_LimitDataSource_minRow _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ComputationExpressionpdataBaseView_LimitDataSource_minRow;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempdataBaseView_LimitDataSource_minRow (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempdataBaseView_LimitDataSource_minRow (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBdataBaseView_BinaryExpression)) 
		return _MkdataBaseView_BinaryExpressionpdataBaseView_LimitDataSource_minRow (_coordref, MkdataBaseView_BinaryExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_UnaryExpression)) 
		return _MkdataBaseView_UnaryExpressionpdataBaseView_LimitDataSource_minRow (_coordref, MkdataBaseView_UnaryExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ConstantExpression)) 
		return _MkdataBaseView_ConstantExpressionpdataBaseView_LimitDataSource_minRow (_coordref, MkdataBaseView_ConstantExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ColumnValueExpression)) 
		return _MkdataBaseView_ColumnValueExpressionpdataBaseView_LimitDataSource_minRow (_coordref, MkdataBaseView_ColumnValueExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ParameterValueExpression)) 
		return _MkdataBaseView_ParameterValueExpressionpdataBaseView_LimitDataSource_minRow (_coordref, MkdataBaseView_ParameterValueExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ComputationExpression)) 
		return _MkdataBaseView_ComputationExpressionpdataBaseView_LimitDataSource_minRow (_coordref, MkdataBaseView_ComputationExpression (_coordref, t));
	message (DEADLY, "a pdataBaseView_LimitDataSource_minRow list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pdataBaseView_LimitDataSource_minRow (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pdataBaseView_LimitDataSource_minRow (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_dataBaseView_LimitDataSource_minRow)
	l =_ElempdataBaseView_LimitDataSource_minRow (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_dataBaseView_LimitDataSource_minRow)
	r = _ElempdataBaseView_LimitDataSource_minRow (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pdataBaseView_LimitDataSource_minRow)r)->_desc2);
((_TPPLST_0pdataBaseView_LimitDataSource_minRow)r)->_desc2 = ((_TPPLST_0pdataBaseView_LimitDataSource_minRow)l)->_desc2;
((_TPPLST_0pdataBaseView_LimitDataSource_minRow)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_LimitDataSource_minRow (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpdataBaseView_LimitDataSource_minRow (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpdataBaseView_LimitDataSource_minRow root;
	_TPPLST_0pdataBaseView_LimitDataSource_minRow list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pdataBaseView_LimitDataSource_minRow)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_dataBaseView_LimitDataSource_minRow)
		t = _ElempdataBaseView_LimitDataSource_minRow (_coordref, t);
	remain = (_TPPLST_0pdataBaseView_LimitDataSource_minRow)(((_TPPLST_0pdataBaseView_LimitDataSource_minRow) t)->_desc2);
	((_TPPLST_0pdataBaseView_LimitDataSource_minRow) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pdataBaseView_LimitDataSource_minRow;
#else
list = (_TPPLST_0pdataBaseView_LimitDataSource_minRow) TreeNodeAlloc (sizeof (struct _TPLST_0pdataBaseView_LimitDataSource_minRow));
#endif
list->_prod = RULELST_0pdataBaseView_LimitDataSource_minRow;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pdataBaseView_LimitDataSource_minRow)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_dataBaseView_BinaryExpressionpdataBaseView_LimitDataSource_minRow)
	{_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_LimitDataSource_minRow _currn = _currn;
	 _TERMACT_LST_dataBaseView_BinaryExpressionpdataBaseView_LimitDataSource_minRow;
	} this code has been moved to _ElempdataBaseView_LimitDataSource_minRow*/
	/*if (_currn->_prod == RULELST_dataBaseView_UnaryExpressionpdataBaseView_LimitDataSource_minRow)
	{_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_LimitDataSource_minRow _currn = _currn;
	 _TERMACT_LST_dataBaseView_UnaryExpressionpdataBaseView_LimitDataSource_minRow;
	} this code has been moved to _ElempdataBaseView_LimitDataSource_minRow*/
	/*if (_currn->_prod == RULELST_dataBaseView_ConstantExpressionpdataBaseView_LimitDataSource_minRow)
	{_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_LimitDataSource_minRow _currn = _currn;
	 _TERMACT_LST_dataBaseView_ConstantExpressionpdataBaseView_LimitDataSource_minRow;
	} this code has been moved to _ElempdataBaseView_LimitDataSource_minRow*/
	/*if (_currn->_prod == RULELST_dataBaseView_ColumnValueExpressionpdataBaseView_LimitDataSource_minRow)
	{_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_LimitDataSource_minRow _currn = _currn;
	 _TERMACT_LST_dataBaseView_ColumnValueExpressionpdataBaseView_LimitDataSource_minRow;
	} this code has been moved to _ElempdataBaseView_LimitDataSource_minRow*/
	/*if (_currn->_prod == RULELST_dataBaseView_ParameterValueExpressionpdataBaseView_LimitDataSource_minRow)
	{_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_LimitDataSource_minRow _currn = _currn;
	 _TERMACT_LST_dataBaseView_ParameterValueExpressionpdataBaseView_LimitDataSource_minRow;
	} this code has been moved to _ElempdataBaseView_LimitDataSource_minRow*/
	/*if (_currn->_prod == RULELST_dataBaseView_ComputationExpressionpdataBaseView_LimitDataSource_minRow)
	{_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_LimitDataSource_minRow _currn = _currn;
	 _TERMACT_LST_dataBaseView_ComputationExpressionpdataBaseView_LimitDataSource_minRow;
	} this code has been moved to _ElempdataBaseView_LimitDataSource_minRow*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpdataBaseView_LimitDataSource_minRow;
#else
root = (_TPPpdataBaseView_LimitDataSource_minRow) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_LimitDataSource_minRow));
#endif
root->_prod = RULEpdataBaseView_LimitDataSource_minRow;
root->_desc1 = (_TSPLST_dataBaseView_LimitDataSource_minRow)list;
{ _TPPpdataBaseView_LimitDataSource_minRow _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pdataBaseView_LimitDataSource_minRow;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_LimitDataSource_dots (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_LimitDataSource_dots (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_LimitDataSource_dots _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_LimitDataSource_dots;
#else
_currn = (_TPPpdataBaseView_LimitDataSource_dots) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_LimitDataSource_dots));
#endif
_currn->_prod = RULEpdataBaseView_LimitDataSource_dots;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_LimitDataSource_dots;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_LimitDataSource_dots */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_BinaryExpressionpdataBaseView_LimitDataSource_maxRow (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_BinaryExpressionpdataBaseView_LimitDataSource_maxRow (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_LimitDataSource_maxRow _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_BinaryExpressionpdataBaseView_LimitDataSource_maxRow;
#else
	_new = (_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_LimitDataSource_maxRow) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_BinaryExpressionpdataBaseView_LimitDataSource_maxRow));
#endif
	_new->_prod = RULELST_dataBaseView_BinaryExpressionpdataBaseView_LimitDataSource_maxRow;
	_new->_desc1 = (_TSPdataBaseView_BinaryExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_LimitDataSource_maxRow)_new;
	{_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_LimitDataSource_maxRow _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_BinaryExpressionpdataBaseView_LimitDataSource_maxRow;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_UnaryExpressionpdataBaseView_LimitDataSource_maxRow (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_UnaryExpressionpdataBaseView_LimitDataSource_maxRow (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_LimitDataSource_maxRow _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_UnaryExpressionpdataBaseView_LimitDataSource_maxRow;
#else
	_new = (_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_LimitDataSource_maxRow) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_UnaryExpressionpdataBaseView_LimitDataSource_maxRow));
#endif
	_new->_prod = RULELST_dataBaseView_UnaryExpressionpdataBaseView_LimitDataSource_maxRow;
	_new->_desc1 = (_TSPdataBaseView_UnaryExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_LimitDataSource_maxRow)_new;
	{_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_LimitDataSource_maxRow _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_UnaryExpressionpdataBaseView_LimitDataSource_maxRow;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ConstantExpressionpdataBaseView_LimitDataSource_maxRow (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ConstantExpressionpdataBaseView_LimitDataSource_maxRow (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_LimitDataSource_maxRow _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ConstantExpressionpdataBaseView_LimitDataSource_maxRow;
#else
	_new = (_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_LimitDataSource_maxRow) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ConstantExpressionpdataBaseView_LimitDataSource_maxRow));
#endif
	_new->_prod = RULELST_dataBaseView_ConstantExpressionpdataBaseView_LimitDataSource_maxRow;
	_new->_desc1 = (_TSPdataBaseView_ConstantExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_LimitDataSource_maxRow)_new;
	{_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_LimitDataSource_maxRow _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ConstantExpressionpdataBaseView_LimitDataSource_maxRow;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ColumnValueExpressionpdataBaseView_LimitDataSource_maxRow (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ColumnValueExpressionpdataBaseView_LimitDataSource_maxRow (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_LimitDataSource_maxRow _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ColumnValueExpressionpdataBaseView_LimitDataSource_maxRow;
#else
	_new = (_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_LimitDataSource_maxRow) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ColumnValueExpressionpdataBaseView_LimitDataSource_maxRow));
#endif
	_new->_prod = RULELST_dataBaseView_ColumnValueExpressionpdataBaseView_LimitDataSource_maxRow;
	_new->_desc1 = (_TSPdataBaseView_ColumnValueExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_LimitDataSource_maxRow)_new;
	{_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_LimitDataSource_maxRow _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ColumnValueExpressionpdataBaseView_LimitDataSource_maxRow;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ParameterValueExpressionpdataBaseView_LimitDataSource_maxRow (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ParameterValueExpressionpdataBaseView_LimitDataSource_maxRow (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_LimitDataSource_maxRow _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ParameterValueExpressionpdataBaseView_LimitDataSource_maxRow;
#else
	_new = (_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_LimitDataSource_maxRow) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ParameterValueExpressionpdataBaseView_LimitDataSource_maxRow));
#endif
	_new->_prod = RULELST_dataBaseView_ParameterValueExpressionpdataBaseView_LimitDataSource_maxRow;
	_new->_desc1 = (_TSPdataBaseView_ParameterValueExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_LimitDataSource_maxRow)_new;
	{_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_LimitDataSource_maxRow _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ParameterValueExpressionpdataBaseView_LimitDataSource_maxRow;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ComputationExpressionpdataBaseView_LimitDataSource_maxRow (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ComputationExpressionpdataBaseView_LimitDataSource_maxRow (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_LimitDataSource_maxRow _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ComputationExpressionpdataBaseView_LimitDataSource_maxRow;
#else
	_new = (_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_LimitDataSource_maxRow) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ComputationExpressionpdataBaseView_LimitDataSource_maxRow));
#endif
	_new->_prod = RULELST_dataBaseView_ComputationExpressionpdataBaseView_LimitDataSource_maxRow;
	_new->_desc1 = (_TSPdataBaseView_ComputationExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_LimitDataSource_maxRow)_new;
	{_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_LimitDataSource_maxRow _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ComputationExpressionpdataBaseView_LimitDataSource_maxRow;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempdataBaseView_LimitDataSource_maxRow (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempdataBaseView_LimitDataSource_maxRow (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBdataBaseView_BinaryExpression)) 
		return _MkdataBaseView_BinaryExpressionpdataBaseView_LimitDataSource_maxRow (_coordref, MkdataBaseView_BinaryExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_UnaryExpression)) 
		return _MkdataBaseView_UnaryExpressionpdataBaseView_LimitDataSource_maxRow (_coordref, MkdataBaseView_UnaryExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ConstantExpression)) 
		return _MkdataBaseView_ConstantExpressionpdataBaseView_LimitDataSource_maxRow (_coordref, MkdataBaseView_ConstantExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ColumnValueExpression)) 
		return _MkdataBaseView_ColumnValueExpressionpdataBaseView_LimitDataSource_maxRow (_coordref, MkdataBaseView_ColumnValueExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ParameterValueExpression)) 
		return _MkdataBaseView_ParameterValueExpressionpdataBaseView_LimitDataSource_maxRow (_coordref, MkdataBaseView_ParameterValueExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ComputationExpression)) 
		return _MkdataBaseView_ComputationExpressionpdataBaseView_LimitDataSource_maxRow (_coordref, MkdataBaseView_ComputationExpression (_coordref, t));
	message (DEADLY, "a pdataBaseView_LimitDataSource_maxRow list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pdataBaseView_LimitDataSource_maxRow (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pdataBaseView_LimitDataSource_maxRow (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_dataBaseView_LimitDataSource_maxRow)
	l =_ElempdataBaseView_LimitDataSource_maxRow (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_dataBaseView_LimitDataSource_maxRow)
	r = _ElempdataBaseView_LimitDataSource_maxRow (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pdataBaseView_LimitDataSource_maxRow)r)->_desc2);
((_TPPLST_0pdataBaseView_LimitDataSource_maxRow)r)->_desc2 = ((_TPPLST_0pdataBaseView_LimitDataSource_maxRow)l)->_desc2;
((_TPPLST_0pdataBaseView_LimitDataSource_maxRow)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_LimitDataSource_maxRow (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpdataBaseView_LimitDataSource_maxRow (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpdataBaseView_LimitDataSource_maxRow root;
	_TPPLST_0pdataBaseView_LimitDataSource_maxRow list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pdataBaseView_LimitDataSource_maxRow)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_dataBaseView_LimitDataSource_maxRow)
		t = _ElempdataBaseView_LimitDataSource_maxRow (_coordref, t);
	remain = (_TPPLST_0pdataBaseView_LimitDataSource_maxRow)(((_TPPLST_0pdataBaseView_LimitDataSource_maxRow) t)->_desc2);
	((_TPPLST_0pdataBaseView_LimitDataSource_maxRow) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pdataBaseView_LimitDataSource_maxRow;
#else
list = (_TPPLST_0pdataBaseView_LimitDataSource_maxRow) TreeNodeAlloc (sizeof (struct _TPLST_0pdataBaseView_LimitDataSource_maxRow));
#endif
list->_prod = RULELST_0pdataBaseView_LimitDataSource_maxRow;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pdataBaseView_LimitDataSource_maxRow)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_dataBaseView_BinaryExpressionpdataBaseView_LimitDataSource_maxRow)
	{_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_LimitDataSource_maxRow _currn = _currn;
	 _TERMACT_LST_dataBaseView_BinaryExpressionpdataBaseView_LimitDataSource_maxRow;
	} this code has been moved to _ElempdataBaseView_LimitDataSource_maxRow*/
	/*if (_currn->_prod == RULELST_dataBaseView_UnaryExpressionpdataBaseView_LimitDataSource_maxRow)
	{_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_LimitDataSource_maxRow _currn = _currn;
	 _TERMACT_LST_dataBaseView_UnaryExpressionpdataBaseView_LimitDataSource_maxRow;
	} this code has been moved to _ElempdataBaseView_LimitDataSource_maxRow*/
	/*if (_currn->_prod == RULELST_dataBaseView_ConstantExpressionpdataBaseView_LimitDataSource_maxRow)
	{_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_LimitDataSource_maxRow _currn = _currn;
	 _TERMACT_LST_dataBaseView_ConstantExpressionpdataBaseView_LimitDataSource_maxRow;
	} this code has been moved to _ElempdataBaseView_LimitDataSource_maxRow*/
	/*if (_currn->_prod == RULELST_dataBaseView_ColumnValueExpressionpdataBaseView_LimitDataSource_maxRow)
	{_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_LimitDataSource_maxRow _currn = _currn;
	 _TERMACT_LST_dataBaseView_ColumnValueExpressionpdataBaseView_LimitDataSource_maxRow;
	} this code has been moved to _ElempdataBaseView_LimitDataSource_maxRow*/
	/*if (_currn->_prod == RULELST_dataBaseView_ParameterValueExpressionpdataBaseView_LimitDataSource_maxRow)
	{_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_LimitDataSource_maxRow _currn = _currn;
	 _TERMACT_LST_dataBaseView_ParameterValueExpressionpdataBaseView_LimitDataSource_maxRow;
	} this code has been moved to _ElempdataBaseView_LimitDataSource_maxRow*/
	/*if (_currn->_prod == RULELST_dataBaseView_ComputationExpressionpdataBaseView_LimitDataSource_maxRow)
	{_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_LimitDataSource_maxRow _currn = _currn;
	 _TERMACT_LST_dataBaseView_ComputationExpressionpdataBaseView_LimitDataSource_maxRow;
	} this code has been moved to _ElempdataBaseView_LimitDataSource_maxRow*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpdataBaseView_LimitDataSource_maxRow;
#else
root = (_TPPpdataBaseView_LimitDataSource_maxRow) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_LimitDataSource_maxRow));
#endif
root->_prod = RULEpdataBaseView_LimitDataSource_maxRow;
root->_desc1 = (_TSPLST_dataBaseView_LimitDataSource_maxRow)list;
{ _TPPpdataBaseView_LimitDataSource_maxRow _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pdataBaseView_LimitDataSource_maxRow;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_LimitDataSource_condition (POSITION *_coordref, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpdataBaseView_LimitDataSource_condition (_coordref,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpdataBaseView_LimitDataSource_condition _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_LimitDataSource_condition;
#else
_currn = (_TPPpdataBaseView_LimitDataSource_condition) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_LimitDataSource_condition));
#endif
_currn->_prod = RULEpdataBaseView_LimitDataSource_condition;
_currn->_desc1 = (_TSPdataBaseView_LimitDataSource_minRow) MkdataBaseView_LimitDataSource_minRow (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_LimitDataSource_condition: root of subtree no. 1 can not be made a dataBaseView_LimitDataSource_minRow node ", 0, _coordref);
_currn->_desc2 = (_TSPdataBaseView_LimitDataSource_dots) MkdataBaseView_LimitDataSource_dots (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_LimitDataSource_condition: root of subtree no. 2 can not be made a dataBaseView_LimitDataSource_dots node ", 0, _coordref);
_currn->_desc3 = (_TSPdataBaseView_LimitDataSource_maxRow) MkdataBaseView_LimitDataSource_maxRow (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_LimitDataSource_condition: root of subtree no. 3 can not be made a dataBaseView_LimitDataSource_maxRow node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_LimitDataSource_condition;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_LimitDataSource_condition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_LimitDataSource_from (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_LimitDataSource_from (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_LimitDataSource_from _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_LimitDataSource_from;
#else
_currn = (_TPPpdataBaseView_LimitDataSource_from) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_LimitDataSource_from));
#endif
_currn->_prod = RULEpdataBaseView_LimitDataSource_from;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_LimitDataSource_from;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_LimitDataSource_from */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_LimitDataSource_sourceDataSouce (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_LimitDataSource_sourceDataSouce (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_LimitDataSource_sourceDataSouce _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_LimitDataSource_sourceDataSouce;
#else
_currn = (_TPPpdataBaseView_LimitDataSource_sourceDataSouce) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_LimitDataSource_sourceDataSouce));
#endif
_currn->_prod = RULEpdataBaseView_LimitDataSource_sourceDataSouce;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_LimitDataSource_sourceDataSouce;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_LimitDataSource_sourceDataSouce */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_LimitDataSource_connection (POSITION *_coordref, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpdataBaseView_LimitDataSource_connection (_coordref,_desc1,_desc2)
	POSITION *_coordref;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpdataBaseView_LimitDataSource_connection _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_LimitDataSource_connection;
#else
_currn = (_TPPpdataBaseView_LimitDataSource_connection) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_LimitDataSource_connection));
#endif
_currn->_prod = RULEpdataBaseView_LimitDataSource_connection;
_currn->_desc1 = (_TSPdataBaseView_LimitDataSource_from) MkdataBaseView_LimitDataSource_from (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_LimitDataSource_connection: root of subtree no. 1 can not be made a dataBaseView_LimitDataSource_from node ", 0, _coordref);
_currn->_desc2 = (_TSPdataBaseView_LimitDataSource_sourceDataSouce) MkdataBaseView_LimitDataSource_sourceDataSouce (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_LimitDataSource_connection: root of subtree no. 2 can not be made a dataBaseView_LimitDataSource_sourceDataSouce node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_LimitDataSource_connection;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_LimitDataSource_connection */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceJoin (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4)
#else
NODEPTR MkpdataBaseView_DataSourceJoin (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
#endif
{	_TPPpdataBaseView_DataSourceJoin _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceJoin;
#else
_currn = (_TPPpdataBaseView_DataSourceJoin) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceJoin));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceJoin;
_currn->_desc1 = (_TSPdataBaseView_DataSourceJoin_name) MkdataBaseView_DataSourceJoin_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceJoin: root of subtree no. 1 can not be made a dataBaseView_DataSourceJoin_name node ", 0, _coordref);
_currn->_desc2 = (_TSPdataBaseView_DataSourceJoin_joinOperator) MkdataBaseView_DataSourceJoin_joinOperator (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceJoin: root of subtree no. 2 can not be made a dataBaseView_DataSourceJoin_joinOperator node ", 0, _coordref);
_currn->_desc3 = (_TSPdataBaseView_DataSourceJoin_connection1) MkdataBaseView_DataSourceJoin_connection1 (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceJoin: root of subtree no. 3 can not be made a dataBaseView_DataSourceJoin_connection1 node ", 0, _coordref);
_currn->_desc4 = (_TSPdataBaseView_DataSourceJoin_connection2) MkdataBaseView_DataSourceJoin_connection2 (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceJoin: root of subtree no. 4 can not be made a dataBaseView_DataSourceJoin_connection2 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceJoin;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceJoin */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceJoin_name (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataSourceJoin_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataSourceJoin_name _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceJoin_name;
#else
_currn = (_TPPpdataBaseView_DataSourceJoin_name) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceJoin_name));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceJoin_name;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceJoin_name;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceJoin_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceJoin_joinOperator (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataSourceJoin_joinOperator (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataSourceJoin_joinOperator _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceJoin_joinOperator;
#else
_currn = (_TPPpdataBaseView_DataSourceJoin_joinOperator) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceJoin_joinOperator));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceJoin_joinOperator;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceJoin_joinOperator;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceJoin_joinOperator */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceJoin_from1 (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataSourceJoin_from1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataSourceJoin_from1 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceJoin_from1;
#else
_currn = (_TPPpdataBaseView_DataSourceJoin_from1) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceJoin_from1));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceJoin_from1;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceJoin_from1;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceJoin_from1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceJoin_dataSource1 (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataSourceJoin_dataSource1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataSourceJoin_dataSource1 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceJoin_dataSource1;
#else
_currn = (_TPPpdataBaseView_DataSourceJoin_dataSource1) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceJoin_dataSource1));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceJoin_dataSource1;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceJoin_dataSource1;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceJoin_dataSource1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceJoin_connection1 (POSITION *_coordref, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpdataBaseView_DataSourceJoin_connection1 (_coordref,_desc1,_desc2)
	POSITION *_coordref;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpdataBaseView_DataSourceJoin_connection1 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceJoin_connection1;
#else
_currn = (_TPPpdataBaseView_DataSourceJoin_connection1) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceJoin_connection1));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceJoin_connection1;
_currn->_desc1 = (_TSPdataBaseView_DataSourceJoin_from1) MkdataBaseView_DataSourceJoin_from1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceJoin_connection1: root of subtree no. 1 can not be made a dataBaseView_DataSourceJoin_from1 node ", 0, _coordref);
_currn->_desc2 = (_TSPdataBaseView_DataSourceJoin_dataSource1) MkdataBaseView_DataSourceJoin_dataSource1 (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceJoin_connection1: root of subtree no. 2 can not be made a dataBaseView_DataSourceJoin_dataSource1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceJoin_connection1;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceJoin_connection1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceJoin_from2 (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataSourceJoin_from2 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataSourceJoin_from2 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceJoin_from2;
#else
_currn = (_TPPpdataBaseView_DataSourceJoin_from2) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceJoin_from2));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceJoin_from2;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceJoin_from2;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceJoin_from2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceJoin_dataSource2 (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataSourceJoin_dataSource2 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataSourceJoin_dataSource2 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceJoin_dataSource2;
#else
_currn = (_TPPpdataBaseView_DataSourceJoin_dataSource2) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceJoin_dataSource2));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceJoin_dataSource2;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceJoin_dataSource2;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceJoin_dataSource2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceJoin_connection2 (POSITION *_coordref, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpdataBaseView_DataSourceJoin_connection2 (_coordref,_desc1,_desc2)
	POSITION *_coordref;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpdataBaseView_DataSourceJoin_connection2 _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceJoin_connection2;
#else
_currn = (_TPPpdataBaseView_DataSourceJoin_connection2) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceJoin_connection2));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceJoin_connection2;
_currn->_desc1 = (_TSPdataBaseView_DataSourceJoin_from2) MkdataBaseView_DataSourceJoin_from2 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceJoin_connection2: root of subtree no. 1 can not be made a dataBaseView_DataSourceJoin_from2 node ", 0, _coordref);
_currn->_desc2 = (_TSPdataBaseView_DataSourceJoin_dataSource2) MkdataBaseView_DataSourceJoin_dataSource2 (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceJoin_connection2: root of subtree no. 2 can not be made a dataBaseView_DataSourceJoin_dataSource2 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceJoin_connection2;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceJoin_connection2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceInsert (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4)
#else
NODEPTR MkpdataBaseView_DataSourceInsert (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
#endif
{	_TPPpdataBaseView_DataSourceInsert _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceInsert;
#else
_currn = (_TPPpdataBaseView_DataSourceInsert) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceInsert));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceInsert;
_currn->_desc1 = (_TSPdataBaseView_DataSourceInsert_name) MkdataBaseView_DataSourceInsert_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceInsert: root of subtree no. 1 can not be made a dataBaseView_DataSourceInsert_name node ", 0, _coordref);
_currn->_desc2 = (_TSPdataBaseView_DataSourceInsert_formalParams) MkdataBaseView_DataSourceInsert_formalParams (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceInsert: root of subtree no. 2 can not be made a dataBaseView_DataSourceInsert_formalParams node ", 0, _coordref);
_currn->_desc3 = (_TSPdataBaseView_DataSourceInsert_values) MkdataBaseView_DataSourceInsert_values (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceInsert: root of subtree no. 3 can not be made a dataBaseView_DataSourceInsert_values node ", 0, _coordref);
_currn->_desc4 = (_TSPdataBaseView_DataSourceInsert_connection) MkdataBaseView_DataSourceInsert_connection (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceInsert: root of subtree no. 4 can not be made a dataBaseView_DataSourceInsert_connection node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceInsert;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceInsert */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceInsert_name (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataSourceInsert_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataSourceInsert_name _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceInsert_name;
#else
_currn = (_TPPpdataBaseView_DataSourceInsert_name) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceInsert_name));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceInsert_name;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceInsert_name;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceInsert_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DataSourceFormalParampdataBaseView_DataSourceInsert_formalParams (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DataSourceFormalParampdataBaseView_DataSourceInsert_formalParams (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceInsert_formalParams _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceInsert_formalParams;
#else
	_new = (_TPPLST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceInsert_formalParams) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceInsert_formalParams));
#endif
	_new->_prod = RULELST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceInsert_formalParams;
	_new->_desc1 = (_TSPdataBaseView_DataSourceFormalParam)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceInsert_formalParams)_new;
	{_TPPLST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceInsert_formalParams _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceInsert_formalParams;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempdataBaseView_DataSourceInsert_formalParams (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempdataBaseView_DataSourceInsert_formalParams (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBdataBaseView_DataSourceFormalParam)) 
		return _MkdataBaseView_DataSourceFormalParampdataBaseView_DataSourceInsert_formalParams (_coordref, MkdataBaseView_DataSourceFormalParam (_coordref, t));
	message (DEADLY, "a pdataBaseView_DataSourceInsert_formalParams list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pdataBaseView_DataSourceInsert_formalParams (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pdataBaseView_DataSourceInsert_formalParams (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_dataBaseView_DataSourceInsert_formalParams)
	l =_ElempdataBaseView_DataSourceInsert_formalParams (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_dataBaseView_DataSourceInsert_formalParams)
	r = _ElempdataBaseView_DataSourceInsert_formalParams (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pdataBaseView_DataSourceInsert_formalParams)r)->_desc2);
((_TPPLST_0pdataBaseView_DataSourceInsert_formalParams)r)->_desc2 = ((_TPPLST_0pdataBaseView_DataSourceInsert_formalParams)l)->_desc2;
((_TPPLST_0pdataBaseView_DataSourceInsert_formalParams)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceInsert_formalParams (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpdataBaseView_DataSourceInsert_formalParams (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpdataBaseView_DataSourceInsert_formalParams root;
	_TPPLST_0pdataBaseView_DataSourceInsert_formalParams list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pdataBaseView_DataSourceInsert_formalParams)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_dataBaseView_DataSourceInsert_formalParams)
		t = _ElempdataBaseView_DataSourceInsert_formalParams (_coordref, t);
	remain = (_TPPLST_0pdataBaseView_DataSourceInsert_formalParams)(((_TPPLST_0pdataBaseView_DataSourceInsert_formalParams) t)->_desc2);
	((_TPPLST_0pdataBaseView_DataSourceInsert_formalParams) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pdataBaseView_DataSourceInsert_formalParams;
#else
list = (_TPPLST_0pdataBaseView_DataSourceInsert_formalParams) TreeNodeAlloc (sizeof (struct _TPLST_0pdataBaseView_DataSourceInsert_formalParams));
#endif
list->_prod = RULELST_0pdataBaseView_DataSourceInsert_formalParams;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pdataBaseView_DataSourceInsert_formalParams)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceInsert_formalParams)
	{_TPPLST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceInsert_formalParams _currn = _currn;
	 _TERMACT_LST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceInsert_formalParams;
	} this code has been moved to _ElempdataBaseView_DataSourceInsert_formalParams*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpdataBaseView_DataSourceInsert_formalParams;
#else
root = (_TPPpdataBaseView_DataSourceInsert_formalParams) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceInsert_formalParams));
#endif
root->_prod = RULEpdataBaseView_DataSourceInsert_formalParams;
root->_desc1 = (_TSPLST_dataBaseView_DataSourceInsert_formalParams)list;
{ _TPPpdataBaseView_DataSourceInsert_formalParams _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pdataBaseView_DataSourceInsert_formalParams;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DataSourceValuepdataBaseView_DataSourceInsert_values (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DataSourceValuepdataBaseView_DataSourceInsert_values (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DataSourceValuepdataBaseView_DataSourceInsert_values _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DataSourceValuepdataBaseView_DataSourceInsert_values;
#else
	_new = (_TPPLST_dataBaseView_DataSourceValuepdataBaseView_DataSourceInsert_values) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DataSourceValuepdataBaseView_DataSourceInsert_values));
#endif
	_new->_prod = RULELST_dataBaseView_DataSourceValuepdataBaseView_DataSourceInsert_values;
	_new->_desc1 = (_TSPdataBaseView_DataSourceValue)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceInsert_values)_new;
	{_TPPLST_dataBaseView_DataSourceValuepdataBaseView_DataSourceInsert_values _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DataSourceValuepdataBaseView_DataSourceInsert_values;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempdataBaseView_DataSourceInsert_values (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempdataBaseView_DataSourceInsert_values (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBdataBaseView_DataSourceValue)) 
		return _MkdataBaseView_DataSourceValuepdataBaseView_DataSourceInsert_values (_coordref, MkdataBaseView_DataSourceValue (_coordref, t));
	message (DEADLY, "a pdataBaseView_DataSourceInsert_values list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pdataBaseView_DataSourceInsert_values (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pdataBaseView_DataSourceInsert_values (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_dataBaseView_DataSourceInsert_values)
	l =_ElempdataBaseView_DataSourceInsert_values (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_dataBaseView_DataSourceInsert_values)
	r = _ElempdataBaseView_DataSourceInsert_values (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pdataBaseView_DataSourceInsert_values)r)->_desc2);
((_TPPLST_0pdataBaseView_DataSourceInsert_values)r)->_desc2 = ((_TPPLST_0pdataBaseView_DataSourceInsert_values)l)->_desc2;
((_TPPLST_0pdataBaseView_DataSourceInsert_values)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceInsert_values (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpdataBaseView_DataSourceInsert_values (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpdataBaseView_DataSourceInsert_values root;
	_TPPLST_0pdataBaseView_DataSourceInsert_values list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pdataBaseView_DataSourceInsert_values)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_dataBaseView_DataSourceInsert_values)
		t = _ElempdataBaseView_DataSourceInsert_values (_coordref, t);
	remain = (_TPPLST_0pdataBaseView_DataSourceInsert_values)(((_TPPLST_0pdataBaseView_DataSourceInsert_values) t)->_desc2);
	((_TPPLST_0pdataBaseView_DataSourceInsert_values) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pdataBaseView_DataSourceInsert_values;
#else
list = (_TPPLST_0pdataBaseView_DataSourceInsert_values) TreeNodeAlloc (sizeof (struct _TPLST_0pdataBaseView_DataSourceInsert_values));
#endif
list->_prod = RULELST_0pdataBaseView_DataSourceInsert_values;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pdataBaseView_DataSourceInsert_values)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_dataBaseView_DataSourceValuepdataBaseView_DataSourceInsert_values)
	{_TPPLST_dataBaseView_DataSourceValuepdataBaseView_DataSourceInsert_values _currn = _currn;
	 _TERMACT_LST_dataBaseView_DataSourceValuepdataBaseView_DataSourceInsert_values;
	} this code has been moved to _ElempdataBaseView_DataSourceInsert_values*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpdataBaseView_DataSourceInsert_values;
#else
root = (_TPPpdataBaseView_DataSourceInsert_values) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceInsert_values));
#endif
root->_prod = RULEpdataBaseView_DataSourceInsert_values;
root->_desc1 = (_TSPLST_dataBaseView_DataSourceInsert_values)list;
{ _TPPpdataBaseView_DataSourceInsert_values _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pdataBaseView_DataSourceInsert_values;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceInsert_from (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataSourceInsert_from (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataSourceInsert_from _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceInsert_from;
#else
_currn = (_TPPpdataBaseView_DataSourceInsert_from) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceInsert_from));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceInsert_from;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceInsert_from;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceInsert_from */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceInsert_targetDataSource (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataSourceInsert_targetDataSource (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataSourceInsert_targetDataSource _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceInsert_targetDataSource;
#else
_currn = (_TPPpdataBaseView_DataSourceInsert_targetDataSource) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceInsert_targetDataSource));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceInsert_targetDataSource;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceInsert_targetDataSource;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceInsert_targetDataSource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceInsert_connection (POSITION *_coordref, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpdataBaseView_DataSourceInsert_connection (_coordref,_desc1,_desc2)
	POSITION *_coordref;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpdataBaseView_DataSourceInsert_connection _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceInsert_connection;
#else
_currn = (_TPPpdataBaseView_DataSourceInsert_connection) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceInsert_connection));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceInsert_connection;
_currn->_desc1 = (_TSPdataBaseView_DataSourceInsert_from) MkdataBaseView_DataSourceInsert_from (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceInsert_connection: root of subtree no. 1 can not be made a dataBaseView_DataSourceInsert_from node ", 0, _coordref);
_currn->_desc2 = (_TSPdataBaseView_DataSourceInsert_targetDataSource) MkdataBaseView_DataSourceInsert_targetDataSource (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceInsert_connection: root of subtree no. 2 can not be made a dataBaseView_DataSourceInsert_targetDataSource node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceInsert_connection;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceInsert_connection */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceUpdate (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkpdataBaseView_DataSourceUpdate (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPpdataBaseView_DataSourceUpdate _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceUpdate;
#else
_currn = (_TPPpdataBaseView_DataSourceUpdate) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceUpdate));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceUpdate;
_currn->_desc1 = (_TSPdataBaseView_DataSourceUpdate_name) MkdataBaseView_DataSourceUpdate_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceUpdate: root of subtree no. 1 can not be made a dataBaseView_DataSourceUpdate_name node ", 0, _coordref);
_currn->_desc2 = (_TSPdataBaseView_DataSourceUpdate_formalParams) MkdataBaseView_DataSourceUpdate_formalParams (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceUpdate: root of subtree no. 2 can not be made a dataBaseView_DataSourceUpdate_formalParams node ", 0, _coordref);
_currn->_desc3 = (_TSPdataBaseView_DataSourceUpdate_values) MkdataBaseView_DataSourceUpdate_values (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceUpdate: root of subtree no. 3 can not be made a dataBaseView_DataSourceUpdate_values node ", 0, _coordref);
_currn->_desc4 = (_TSPdataBaseView_DataSourceUpdate_conditions) MkdataBaseView_DataSourceUpdate_conditions (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceUpdate: root of subtree no. 4 can not be made a dataBaseView_DataSourceUpdate_conditions node ", 0, _coordref);
_currn->_desc5 = (_TSPdataBaseView_DataSourceUpdate_connection) MkdataBaseView_DataSourceUpdate_connection (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceUpdate: root of subtree no. 5 can not be made a dataBaseView_DataSourceUpdate_connection node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceUpdate;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceUpdate */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceUpdate_name (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataSourceUpdate_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataSourceUpdate_name _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceUpdate_name;
#else
_currn = (_TPPpdataBaseView_DataSourceUpdate_name) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceUpdate_name));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceUpdate_name;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceUpdate_name;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceUpdate_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DataSourceFormalParampdataBaseView_DataSourceUpdate_formalParams (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DataSourceFormalParampdataBaseView_DataSourceUpdate_formalParams (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceUpdate_formalParams _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceUpdate_formalParams;
#else
	_new = (_TPPLST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceUpdate_formalParams) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceUpdate_formalParams));
#endif
	_new->_prod = RULELST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceUpdate_formalParams;
	_new->_desc1 = (_TSPdataBaseView_DataSourceFormalParam)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceUpdate_formalParams)_new;
	{_TPPLST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceUpdate_formalParams _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceUpdate_formalParams;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempdataBaseView_DataSourceUpdate_formalParams (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempdataBaseView_DataSourceUpdate_formalParams (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBdataBaseView_DataSourceFormalParam)) 
		return _MkdataBaseView_DataSourceFormalParampdataBaseView_DataSourceUpdate_formalParams (_coordref, MkdataBaseView_DataSourceFormalParam (_coordref, t));
	message (DEADLY, "a pdataBaseView_DataSourceUpdate_formalParams list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pdataBaseView_DataSourceUpdate_formalParams (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pdataBaseView_DataSourceUpdate_formalParams (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_dataBaseView_DataSourceUpdate_formalParams)
	l =_ElempdataBaseView_DataSourceUpdate_formalParams (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_dataBaseView_DataSourceUpdate_formalParams)
	r = _ElempdataBaseView_DataSourceUpdate_formalParams (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pdataBaseView_DataSourceUpdate_formalParams)r)->_desc2);
((_TPPLST_0pdataBaseView_DataSourceUpdate_formalParams)r)->_desc2 = ((_TPPLST_0pdataBaseView_DataSourceUpdate_formalParams)l)->_desc2;
((_TPPLST_0pdataBaseView_DataSourceUpdate_formalParams)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceUpdate_formalParams (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpdataBaseView_DataSourceUpdate_formalParams (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpdataBaseView_DataSourceUpdate_formalParams root;
	_TPPLST_0pdataBaseView_DataSourceUpdate_formalParams list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pdataBaseView_DataSourceUpdate_formalParams)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_dataBaseView_DataSourceUpdate_formalParams)
		t = _ElempdataBaseView_DataSourceUpdate_formalParams (_coordref, t);
	remain = (_TPPLST_0pdataBaseView_DataSourceUpdate_formalParams)(((_TPPLST_0pdataBaseView_DataSourceUpdate_formalParams) t)->_desc2);
	((_TPPLST_0pdataBaseView_DataSourceUpdate_formalParams) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pdataBaseView_DataSourceUpdate_formalParams;
#else
list = (_TPPLST_0pdataBaseView_DataSourceUpdate_formalParams) TreeNodeAlloc (sizeof (struct _TPLST_0pdataBaseView_DataSourceUpdate_formalParams));
#endif
list->_prod = RULELST_0pdataBaseView_DataSourceUpdate_formalParams;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pdataBaseView_DataSourceUpdate_formalParams)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceUpdate_formalParams)
	{_TPPLST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceUpdate_formalParams _currn = _currn;
	 _TERMACT_LST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceUpdate_formalParams;
	} this code has been moved to _ElempdataBaseView_DataSourceUpdate_formalParams*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpdataBaseView_DataSourceUpdate_formalParams;
#else
root = (_TPPpdataBaseView_DataSourceUpdate_formalParams) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceUpdate_formalParams));
#endif
root->_prod = RULEpdataBaseView_DataSourceUpdate_formalParams;
root->_desc1 = (_TSPLST_dataBaseView_DataSourceUpdate_formalParams)list;
{ _TPPpdataBaseView_DataSourceUpdate_formalParams _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pdataBaseView_DataSourceUpdate_formalParams;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DataSourceValuepdataBaseView_DataSourceUpdate_values (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DataSourceValuepdataBaseView_DataSourceUpdate_values (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DataSourceValuepdataBaseView_DataSourceUpdate_values _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DataSourceValuepdataBaseView_DataSourceUpdate_values;
#else
	_new = (_TPPLST_dataBaseView_DataSourceValuepdataBaseView_DataSourceUpdate_values) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DataSourceValuepdataBaseView_DataSourceUpdate_values));
#endif
	_new->_prod = RULELST_dataBaseView_DataSourceValuepdataBaseView_DataSourceUpdate_values;
	_new->_desc1 = (_TSPdataBaseView_DataSourceValue)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceUpdate_values)_new;
	{_TPPLST_dataBaseView_DataSourceValuepdataBaseView_DataSourceUpdate_values _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DataSourceValuepdataBaseView_DataSourceUpdate_values;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempdataBaseView_DataSourceUpdate_values (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempdataBaseView_DataSourceUpdate_values (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBdataBaseView_DataSourceValue)) 
		return _MkdataBaseView_DataSourceValuepdataBaseView_DataSourceUpdate_values (_coordref, MkdataBaseView_DataSourceValue (_coordref, t));
	message (DEADLY, "a pdataBaseView_DataSourceUpdate_values list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pdataBaseView_DataSourceUpdate_values (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pdataBaseView_DataSourceUpdate_values (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_dataBaseView_DataSourceUpdate_values)
	l =_ElempdataBaseView_DataSourceUpdate_values (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_dataBaseView_DataSourceUpdate_values)
	r = _ElempdataBaseView_DataSourceUpdate_values (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pdataBaseView_DataSourceUpdate_values)r)->_desc2);
((_TPPLST_0pdataBaseView_DataSourceUpdate_values)r)->_desc2 = ((_TPPLST_0pdataBaseView_DataSourceUpdate_values)l)->_desc2;
((_TPPLST_0pdataBaseView_DataSourceUpdate_values)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceUpdate_values (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpdataBaseView_DataSourceUpdate_values (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpdataBaseView_DataSourceUpdate_values root;
	_TPPLST_0pdataBaseView_DataSourceUpdate_values list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pdataBaseView_DataSourceUpdate_values)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_dataBaseView_DataSourceUpdate_values)
		t = _ElempdataBaseView_DataSourceUpdate_values (_coordref, t);
	remain = (_TPPLST_0pdataBaseView_DataSourceUpdate_values)(((_TPPLST_0pdataBaseView_DataSourceUpdate_values) t)->_desc2);
	((_TPPLST_0pdataBaseView_DataSourceUpdate_values) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pdataBaseView_DataSourceUpdate_values;
#else
list = (_TPPLST_0pdataBaseView_DataSourceUpdate_values) TreeNodeAlloc (sizeof (struct _TPLST_0pdataBaseView_DataSourceUpdate_values));
#endif
list->_prod = RULELST_0pdataBaseView_DataSourceUpdate_values;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pdataBaseView_DataSourceUpdate_values)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_dataBaseView_DataSourceValuepdataBaseView_DataSourceUpdate_values)
	{_TPPLST_dataBaseView_DataSourceValuepdataBaseView_DataSourceUpdate_values _currn = _currn;
	 _TERMACT_LST_dataBaseView_DataSourceValuepdataBaseView_DataSourceUpdate_values;
	} this code has been moved to _ElempdataBaseView_DataSourceUpdate_values*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpdataBaseView_DataSourceUpdate_values;
#else
root = (_TPPpdataBaseView_DataSourceUpdate_values) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceUpdate_values));
#endif
root->_prod = RULEpdataBaseView_DataSourceUpdate_values;
root->_desc1 = (_TSPLST_dataBaseView_DataSourceUpdate_values)list;
{ _TPPpdataBaseView_DataSourceUpdate_values _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pdataBaseView_DataSourceUpdate_values;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_BinaryExpressionpdataBaseView_DataSourceUpdate_conditions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_BinaryExpressionpdataBaseView_DataSourceUpdate_conditions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceUpdate_conditions _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceUpdate_conditions;
#else
	_new = (_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceUpdate_conditions) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceUpdate_conditions));
#endif
	_new->_prod = RULELST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceUpdate_conditions;
	_new->_desc1 = (_TSPdataBaseView_BinaryExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceUpdate_conditions)_new;
	{_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceUpdate_conditions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceUpdate_conditions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_UnaryExpressionpdataBaseView_DataSourceUpdate_conditions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_UnaryExpressionpdataBaseView_DataSourceUpdate_conditions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceUpdate_conditions _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceUpdate_conditions;
#else
	_new = (_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceUpdate_conditions) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceUpdate_conditions));
#endif
	_new->_prod = RULELST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceUpdate_conditions;
	_new->_desc1 = (_TSPdataBaseView_UnaryExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceUpdate_conditions)_new;
	{_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceUpdate_conditions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceUpdate_conditions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ConstantExpressionpdataBaseView_DataSourceUpdate_conditions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ConstantExpressionpdataBaseView_DataSourceUpdate_conditions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceUpdate_conditions _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceUpdate_conditions;
#else
	_new = (_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceUpdate_conditions) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceUpdate_conditions));
#endif
	_new->_prod = RULELST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceUpdate_conditions;
	_new->_desc1 = (_TSPdataBaseView_ConstantExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceUpdate_conditions)_new;
	{_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceUpdate_conditions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceUpdate_conditions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ColumnValueExpressionpdataBaseView_DataSourceUpdate_conditions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ColumnValueExpressionpdataBaseView_DataSourceUpdate_conditions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceUpdate_conditions _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceUpdate_conditions;
#else
	_new = (_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceUpdate_conditions) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceUpdate_conditions));
#endif
	_new->_prod = RULELST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceUpdate_conditions;
	_new->_desc1 = (_TSPdataBaseView_ColumnValueExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceUpdate_conditions)_new;
	{_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceUpdate_conditions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceUpdate_conditions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ParameterValueExpressionpdataBaseView_DataSourceUpdate_conditions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ParameterValueExpressionpdataBaseView_DataSourceUpdate_conditions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceUpdate_conditions _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceUpdate_conditions;
#else
	_new = (_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceUpdate_conditions) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceUpdate_conditions));
#endif
	_new->_prod = RULELST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceUpdate_conditions;
	_new->_desc1 = (_TSPdataBaseView_ParameterValueExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceUpdate_conditions)_new;
	{_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceUpdate_conditions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceUpdate_conditions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ComputationExpressionpdataBaseView_DataSourceUpdate_conditions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ComputationExpressionpdataBaseView_DataSourceUpdate_conditions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceUpdate_conditions _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceUpdate_conditions;
#else
	_new = (_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceUpdate_conditions) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceUpdate_conditions));
#endif
	_new->_prod = RULELST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceUpdate_conditions;
	_new->_desc1 = (_TSPdataBaseView_ComputationExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceUpdate_conditions)_new;
	{_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceUpdate_conditions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceUpdate_conditions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempdataBaseView_DataSourceUpdate_conditions (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempdataBaseView_DataSourceUpdate_conditions (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBdataBaseView_BinaryExpression)) 
		return _MkdataBaseView_BinaryExpressionpdataBaseView_DataSourceUpdate_conditions (_coordref, MkdataBaseView_BinaryExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_UnaryExpression)) 
		return _MkdataBaseView_UnaryExpressionpdataBaseView_DataSourceUpdate_conditions (_coordref, MkdataBaseView_UnaryExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ConstantExpression)) 
		return _MkdataBaseView_ConstantExpressionpdataBaseView_DataSourceUpdate_conditions (_coordref, MkdataBaseView_ConstantExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ColumnValueExpression)) 
		return _MkdataBaseView_ColumnValueExpressionpdataBaseView_DataSourceUpdate_conditions (_coordref, MkdataBaseView_ColumnValueExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ParameterValueExpression)) 
		return _MkdataBaseView_ParameterValueExpressionpdataBaseView_DataSourceUpdate_conditions (_coordref, MkdataBaseView_ParameterValueExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ComputationExpression)) 
		return _MkdataBaseView_ComputationExpressionpdataBaseView_DataSourceUpdate_conditions (_coordref, MkdataBaseView_ComputationExpression (_coordref, t));
	message (DEADLY, "a pdataBaseView_DataSourceUpdate_conditions list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pdataBaseView_DataSourceUpdate_conditions (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pdataBaseView_DataSourceUpdate_conditions (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_dataBaseView_DataSourceUpdate_conditions)
	l =_ElempdataBaseView_DataSourceUpdate_conditions (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_dataBaseView_DataSourceUpdate_conditions)
	r = _ElempdataBaseView_DataSourceUpdate_conditions (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pdataBaseView_DataSourceUpdate_conditions)r)->_desc2);
((_TPPLST_0pdataBaseView_DataSourceUpdate_conditions)r)->_desc2 = ((_TPPLST_0pdataBaseView_DataSourceUpdate_conditions)l)->_desc2;
((_TPPLST_0pdataBaseView_DataSourceUpdate_conditions)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceUpdate_conditions (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpdataBaseView_DataSourceUpdate_conditions (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpdataBaseView_DataSourceUpdate_conditions root;
	_TPPLST_0pdataBaseView_DataSourceUpdate_conditions list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pdataBaseView_DataSourceUpdate_conditions)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_dataBaseView_DataSourceUpdate_conditions)
		t = _ElempdataBaseView_DataSourceUpdate_conditions (_coordref, t);
	remain = (_TPPLST_0pdataBaseView_DataSourceUpdate_conditions)(((_TPPLST_0pdataBaseView_DataSourceUpdate_conditions) t)->_desc2);
	((_TPPLST_0pdataBaseView_DataSourceUpdate_conditions) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pdataBaseView_DataSourceUpdate_conditions;
#else
list = (_TPPLST_0pdataBaseView_DataSourceUpdate_conditions) TreeNodeAlloc (sizeof (struct _TPLST_0pdataBaseView_DataSourceUpdate_conditions));
#endif
list->_prod = RULELST_0pdataBaseView_DataSourceUpdate_conditions;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pdataBaseView_DataSourceUpdate_conditions)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceUpdate_conditions)
	{_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceUpdate_conditions _currn = _currn;
	 _TERMACT_LST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceUpdate_conditions;
	} this code has been moved to _ElempdataBaseView_DataSourceUpdate_conditions*/
	/*if (_currn->_prod == RULELST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceUpdate_conditions)
	{_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceUpdate_conditions _currn = _currn;
	 _TERMACT_LST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceUpdate_conditions;
	} this code has been moved to _ElempdataBaseView_DataSourceUpdate_conditions*/
	/*if (_currn->_prod == RULELST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceUpdate_conditions)
	{_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceUpdate_conditions _currn = _currn;
	 _TERMACT_LST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceUpdate_conditions;
	} this code has been moved to _ElempdataBaseView_DataSourceUpdate_conditions*/
	/*if (_currn->_prod == RULELST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceUpdate_conditions)
	{_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceUpdate_conditions _currn = _currn;
	 _TERMACT_LST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceUpdate_conditions;
	} this code has been moved to _ElempdataBaseView_DataSourceUpdate_conditions*/
	/*if (_currn->_prod == RULELST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceUpdate_conditions)
	{_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceUpdate_conditions _currn = _currn;
	 _TERMACT_LST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceUpdate_conditions;
	} this code has been moved to _ElempdataBaseView_DataSourceUpdate_conditions*/
	/*if (_currn->_prod == RULELST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceUpdate_conditions)
	{_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceUpdate_conditions _currn = _currn;
	 _TERMACT_LST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceUpdate_conditions;
	} this code has been moved to _ElempdataBaseView_DataSourceUpdate_conditions*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpdataBaseView_DataSourceUpdate_conditions;
#else
root = (_TPPpdataBaseView_DataSourceUpdate_conditions) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceUpdate_conditions));
#endif
root->_prod = RULEpdataBaseView_DataSourceUpdate_conditions;
root->_desc1 = (_TSPLST_dataBaseView_DataSourceUpdate_conditions)list;
{ _TPPpdataBaseView_DataSourceUpdate_conditions _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pdataBaseView_DataSourceUpdate_conditions;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceUpdate_from (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataSourceUpdate_from (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataSourceUpdate_from _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceUpdate_from;
#else
_currn = (_TPPpdataBaseView_DataSourceUpdate_from) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceUpdate_from));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceUpdate_from;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceUpdate_from;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceUpdate_from */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceUpdate_targetDataSource (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataSourceUpdate_targetDataSource (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataSourceUpdate_targetDataSource _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceUpdate_targetDataSource;
#else
_currn = (_TPPpdataBaseView_DataSourceUpdate_targetDataSource) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceUpdate_targetDataSource));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceUpdate_targetDataSource;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceUpdate_targetDataSource;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceUpdate_targetDataSource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceUpdate_connection (POSITION *_coordref, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpdataBaseView_DataSourceUpdate_connection (_coordref,_desc1,_desc2)
	POSITION *_coordref;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpdataBaseView_DataSourceUpdate_connection _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceUpdate_connection;
#else
_currn = (_TPPpdataBaseView_DataSourceUpdate_connection) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceUpdate_connection));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceUpdate_connection;
_currn->_desc1 = (_TSPdataBaseView_DataSourceUpdate_from) MkdataBaseView_DataSourceUpdate_from (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceUpdate_connection: root of subtree no. 1 can not be made a dataBaseView_DataSourceUpdate_from node ", 0, _coordref);
_currn->_desc2 = (_TSPdataBaseView_DataSourceUpdate_targetDataSource) MkdataBaseView_DataSourceUpdate_targetDataSource (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceUpdate_connection: root of subtree no. 2 can not be made a dataBaseView_DataSourceUpdate_targetDataSource node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceUpdate_connection;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceUpdate_connection */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceDelete (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4)
#else
NODEPTR MkpdataBaseView_DataSourceDelete (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
#endif
{	_TPPpdataBaseView_DataSourceDelete _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceDelete;
#else
_currn = (_TPPpdataBaseView_DataSourceDelete) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceDelete));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceDelete;
_currn->_desc1 = (_TSPdataBaseView_DataSourceDelete_name) MkdataBaseView_DataSourceDelete_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceDelete: root of subtree no. 1 can not be made a dataBaseView_DataSourceDelete_name node ", 0, _coordref);
_currn->_desc2 = (_TSPdataBaseView_DataSourceDelete_formalParams) MkdataBaseView_DataSourceDelete_formalParams (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceDelete: root of subtree no. 2 can not be made a dataBaseView_DataSourceDelete_formalParams node ", 0, _coordref);
_currn->_desc3 = (_TSPdataBaseView_DataSourceDelete_conditions) MkdataBaseView_DataSourceDelete_conditions (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceDelete: root of subtree no. 3 can not be made a dataBaseView_DataSourceDelete_conditions node ", 0, _coordref);
_currn->_desc4 = (_TSPdataBaseView_DataSourceDelete_connection) MkdataBaseView_DataSourceDelete_connection (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceDelete: root of subtree no. 4 can not be made a dataBaseView_DataSourceDelete_connection node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceDelete;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceDelete */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceDelete_name (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataSourceDelete_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataSourceDelete_name _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceDelete_name;
#else
_currn = (_TPPpdataBaseView_DataSourceDelete_name) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceDelete_name));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceDelete_name;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceDelete_name;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceDelete_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DataSourceFormalParampdataBaseView_DataSourceDelete_formalParams (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DataSourceFormalParampdataBaseView_DataSourceDelete_formalParams (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceDelete_formalParams _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceDelete_formalParams;
#else
	_new = (_TPPLST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceDelete_formalParams) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceDelete_formalParams));
#endif
	_new->_prod = RULELST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceDelete_formalParams;
	_new->_desc1 = (_TSPdataBaseView_DataSourceFormalParam)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceDelete_formalParams)_new;
	{_TPPLST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceDelete_formalParams _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceDelete_formalParams;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempdataBaseView_DataSourceDelete_formalParams (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempdataBaseView_DataSourceDelete_formalParams (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBdataBaseView_DataSourceFormalParam)) 
		return _MkdataBaseView_DataSourceFormalParampdataBaseView_DataSourceDelete_formalParams (_coordref, MkdataBaseView_DataSourceFormalParam (_coordref, t));
	message (DEADLY, "a pdataBaseView_DataSourceDelete_formalParams list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pdataBaseView_DataSourceDelete_formalParams (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pdataBaseView_DataSourceDelete_formalParams (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_dataBaseView_DataSourceDelete_formalParams)
	l =_ElempdataBaseView_DataSourceDelete_formalParams (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_dataBaseView_DataSourceDelete_formalParams)
	r = _ElempdataBaseView_DataSourceDelete_formalParams (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pdataBaseView_DataSourceDelete_formalParams)r)->_desc2);
((_TPPLST_0pdataBaseView_DataSourceDelete_formalParams)r)->_desc2 = ((_TPPLST_0pdataBaseView_DataSourceDelete_formalParams)l)->_desc2;
((_TPPLST_0pdataBaseView_DataSourceDelete_formalParams)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceDelete_formalParams (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpdataBaseView_DataSourceDelete_formalParams (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpdataBaseView_DataSourceDelete_formalParams root;
	_TPPLST_0pdataBaseView_DataSourceDelete_formalParams list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pdataBaseView_DataSourceDelete_formalParams)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_dataBaseView_DataSourceDelete_formalParams)
		t = _ElempdataBaseView_DataSourceDelete_formalParams (_coordref, t);
	remain = (_TPPLST_0pdataBaseView_DataSourceDelete_formalParams)(((_TPPLST_0pdataBaseView_DataSourceDelete_formalParams) t)->_desc2);
	((_TPPLST_0pdataBaseView_DataSourceDelete_formalParams) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pdataBaseView_DataSourceDelete_formalParams;
#else
list = (_TPPLST_0pdataBaseView_DataSourceDelete_formalParams) TreeNodeAlloc (sizeof (struct _TPLST_0pdataBaseView_DataSourceDelete_formalParams));
#endif
list->_prod = RULELST_0pdataBaseView_DataSourceDelete_formalParams;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pdataBaseView_DataSourceDelete_formalParams)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceDelete_formalParams)
	{_TPPLST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceDelete_formalParams _currn = _currn;
	 _TERMACT_LST_dataBaseView_DataSourceFormalParampdataBaseView_DataSourceDelete_formalParams;
	} this code has been moved to _ElempdataBaseView_DataSourceDelete_formalParams*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpdataBaseView_DataSourceDelete_formalParams;
#else
root = (_TPPpdataBaseView_DataSourceDelete_formalParams) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceDelete_formalParams));
#endif
root->_prod = RULEpdataBaseView_DataSourceDelete_formalParams;
root->_desc1 = (_TSPLST_dataBaseView_DataSourceDelete_formalParams)list;
{ _TPPpdataBaseView_DataSourceDelete_formalParams _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pdataBaseView_DataSourceDelete_formalParams;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_BinaryExpressionpdataBaseView_DataSourceDelete_conditions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_BinaryExpressionpdataBaseView_DataSourceDelete_conditions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceDelete_conditions _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceDelete_conditions;
#else
	_new = (_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceDelete_conditions) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceDelete_conditions));
#endif
	_new->_prod = RULELST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceDelete_conditions;
	_new->_desc1 = (_TSPdataBaseView_BinaryExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceDelete_conditions)_new;
	{_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceDelete_conditions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceDelete_conditions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_UnaryExpressionpdataBaseView_DataSourceDelete_conditions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_UnaryExpressionpdataBaseView_DataSourceDelete_conditions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceDelete_conditions _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceDelete_conditions;
#else
	_new = (_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceDelete_conditions) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceDelete_conditions));
#endif
	_new->_prod = RULELST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceDelete_conditions;
	_new->_desc1 = (_TSPdataBaseView_UnaryExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceDelete_conditions)_new;
	{_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceDelete_conditions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceDelete_conditions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ConstantExpressionpdataBaseView_DataSourceDelete_conditions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ConstantExpressionpdataBaseView_DataSourceDelete_conditions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceDelete_conditions _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceDelete_conditions;
#else
	_new = (_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceDelete_conditions) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceDelete_conditions));
#endif
	_new->_prod = RULELST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceDelete_conditions;
	_new->_desc1 = (_TSPdataBaseView_ConstantExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceDelete_conditions)_new;
	{_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceDelete_conditions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceDelete_conditions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ColumnValueExpressionpdataBaseView_DataSourceDelete_conditions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ColumnValueExpressionpdataBaseView_DataSourceDelete_conditions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceDelete_conditions _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceDelete_conditions;
#else
	_new = (_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceDelete_conditions) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceDelete_conditions));
#endif
	_new->_prod = RULELST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceDelete_conditions;
	_new->_desc1 = (_TSPdataBaseView_ColumnValueExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceDelete_conditions)_new;
	{_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceDelete_conditions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceDelete_conditions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ParameterValueExpressionpdataBaseView_DataSourceDelete_conditions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ParameterValueExpressionpdataBaseView_DataSourceDelete_conditions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceDelete_conditions _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceDelete_conditions;
#else
	_new = (_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceDelete_conditions) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceDelete_conditions));
#endif
	_new->_prod = RULELST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceDelete_conditions;
	_new->_desc1 = (_TSPdataBaseView_ParameterValueExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceDelete_conditions)_new;
	{_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceDelete_conditions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceDelete_conditions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_ComputationExpressionpdataBaseView_DataSourceDelete_conditions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_ComputationExpressionpdataBaseView_DataSourceDelete_conditions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceDelete_conditions _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceDelete_conditions;
#else
	_new = (_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceDelete_conditions) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceDelete_conditions));
#endif
	_new->_prod = RULELST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceDelete_conditions;
	_new->_desc1 = (_TSPdataBaseView_ComputationExpression)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceDelete_conditions)_new;
	{_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceDelete_conditions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceDelete_conditions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempdataBaseView_DataSourceDelete_conditions (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempdataBaseView_DataSourceDelete_conditions (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBdataBaseView_BinaryExpression)) 
		return _MkdataBaseView_BinaryExpressionpdataBaseView_DataSourceDelete_conditions (_coordref, MkdataBaseView_BinaryExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_UnaryExpression)) 
		return _MkdataBaseView_UnaryExpressionpdataBaseView_DataSourceDelete_conditions (_coordref, MkdataBaseView_UnaryExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ConstantExpression)) 
		return _MkdataBaseView_ConstantExpressionpdataBaseView_DataSourceDelete_conditions (_coordref, MkdataBaseView_ConstantExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ColumnValueExpression)) 
		return _MkdataBaseView_ColumnValueExpressionpdataBaseView_DataSourceDelete_conditions (_coordref, MkdataBaseView_ColumnValueExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ParameterValueExpression)) 
		return _MkdataBaseView_ParameterValueExpressionpdataBaseView_DataSourceDelete_conditions (_coordref, MkdataBaseView_ParameterValueExpression (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_ComputationExpression)) 
		return _MkdataBaseView_ComputationExpressionpdataBaseView_DataSourceDelete_conditions (_coordref, MkdataBaseView_ComputationExpression (_coordref, t));
	message (DEADLY, "a pdataBaseView_DataSourceDelete_conditions list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pdataBaseView_DataSourceDelete_conditions (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pdataBaseView_DataSourceDelete_conditions (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_dataBaseView_DataSourceDelete_conditions)
	l =_ElempdataBaseView_DataSourceDelete_conditions (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_dataBaseView_DataSourceDelete_conditions)
	r = _ElempdataBaseView_DataSourceDelete_conditions (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pdataBaseView_DataSourceDelete_conditions)r)->_desc2);
((_TPPLST_0pdataBaseView_DataSourceDelete_conditions)r)->_desc2 = ((_TPPLST_0pdataBaseView_DataSourceDelete_conditions)l)->_desc2;
((_TPPLST_0pdataBaseView_DataSourceDelete_conditions)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceDelete_conditions (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpdataBaseView_DataSourceDelete_conditions (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpdataBaseView_DataSourceDelete_conditions root;
	_TPPLST_0pdataBaseView_DataSourceDelete_conditions list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pdataBaseView_DataSourceDelete_conditions)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_dataBaseView_DataSourceDelete_conditions)
		t = _ElempdataBaseView_DataSourceDelete_conditions (_coordref, t);
	remain = (_TPPLST_0pdataBaseView_DataSourceDelete_conditions)(((_TPPLST_0pdataBaseView_DataSourceDelete_conditions) t)->_desc2);
	((_TPPLST_0pdataBaseView_DataSourceDelete_conditions) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pdataBaseView_DataSourceDelete_conditions;
#else
list = (_TPPLST_0pdataBaseView_DataSourceDelete_conditions) TreeNodeAlloc (sizeof (struct _TPLST_0pdataBaseView_DataSourceDelete_conditions));
#endif
list->_prod = RULELST_0pdataBaseView_DataSourceDelete_conditions;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pdataBaseView_DataSourceDelete_conditions)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceDelete_conditions)
	{_TPPLST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceDelete_conditions _currn = _currn;
	 _TERMACT_LST_dataBaseView_BinaryExpressionpdataBaseView_DataSourceDelete_conditions;
	} this code has been moved to _ElempdataBaseView_DataSourceDelete_conditions*/
	/*if (_currn->_prod == RULELST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceDelete_conditions)
	{_TPPLST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceDelete_conditions _currn = _currn;
	 _TERMACT_LST_dataBaseView_UnaryExpressionpdataBaseView_DataSourceDelete_conditions;
	} this code has been moved to _ElempdataBaseView_DataSourceDelete_conditions*/
	/*if (_currn->_prod == RULELST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceDelete_conditions)
	{_TPPLST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceDelete_conditions _currn = _currn;
	 _TERMACT_LST_dataBaseView_ConstantExpressionpdataBaseView_DataSourceDelete_conditions;
	} this code has been moved to _ElempdataBaseView_DataSourceDelete_conditions*/
	/*if (_currn->_prod == RULELST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceDelete_conditions)
	{_TPPLST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceDelete_conditions _currn = _currn;
	 _TERMACT_LST_dataBaseView_ColumnValueExpressionpdataBaseView_DataSourceDelete_conditions;
	} this code has been moved to _ElempdataBaseView_DataSourceDelete_conditions*/
	/*if (_currn->_prod == RULELST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceDelete_conditions)
	{_TPPLST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceDelete_conditions _currn = _currn;
	 _TERMACT_LST_dataBaseView_ParameterValueExpressionpdataBaseView_DataSourceDelete_conditions;
	} this code has been moved to _ElempdataBaseView_DataSourceDelete_conditions*/
	/*if (_currn->_prod == RULELST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceDelete_conditions)
	{_TPPLST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceDelete_conditions _currn = _currn;
	 _TERMACT_LST_dataBaseView_ComputationExpressionpdataBaseView_DataSourceDelete_conditions;
	} this code has been moved to _ElempdataBaseView_DataSourceDelete_conditions*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpdataBaseView_DataSourceDelete_conditions;
#else
root = (_TPPpdataBaseView_DataSourceDelete_conditions) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceDelete_conditions));
#endif
root->_prod = RULEpdataBaseView_DataSourceDelete_conditions;
root->_desc1 = (_TSPLST_dataBaseView_DataSourceDelete_conditions)list;
{ _TPPpdataBaseView_DataSourceDelete_conditions _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pdataBaseView_DataSourceDelete_conditions;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceDelete_from (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataSourceDelete_from (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataSourceDelete_from _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceDelete_from;
#else
_currn = (_TPPpdataBaseView_DataSourceDelete_from) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceDelete_from));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceDelete_from;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceDelete_from;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceDelete_from */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceDelete_targetDataSource (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataSourceDelete_targetDataSource (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataSourceDelete_targetDataSource _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceDelete_targetDataSource;
#else
_currn = (_TPPpdataBaseView_DataSourceDelete_targetDataSource) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceDelete_targetDataSource));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceDelete_targetDataSource;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceDelete_targetDataSource;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceDelete_targetDataSource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceDelete_connection (POSITION *_coordref, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpdataBaseView_DataSourceDelete_connection (_coordref,_desc1,_desc2)
	POSITION *_coordref;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpdataBaseView_DataSourceDelete_connection _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceDelete_connection;
#else
_currn = (_TPPpdataBaseView_DataSourceDelete_connection) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceDelete_connection));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceDelete_connection;
_currn->_desc1 = (_TSPdataBaseView_DataSourceDelete_from) MkdataBaseView_DataSourceDelete_from (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceDelete_connection: root of subtree no. 1 can not be made a dataBaseView_DataSourceDelete_from node ", 0, _coordref);
_currn->_desc2 = (_TSPdataBaseView_DataSourceDelete_targetDataSource) MkdataBaseView_DataSourceDelete_targetDataSource (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceDelete_connection: root of subtree no. 2 can not be made a dataBaseView_DataSourceDelete_targetDataSource node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceDelete_connection;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceDelete_connection */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceDB (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpdataBaseView_DataSourceDB (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpdataBaseView_DataSourceDB _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceDB;
#else
_currn = (_TPPpdataBaseView_DataSourceDB) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceDB));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceDB;
_currn->_desc1 = (_TSPdataBaseView_DataSourceDB_name) MkdataBaseView_DataSourceDB_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceDB: root of subtree no. 1 can not be made a dataBaseView_DataSourceDB_name node ", 0, _coordref);
_currn->_desc2 = (_TSPdataBaseView_DataSourceDB_position) MkdataBaseView_DataSourceDB_position (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceDB: root of subtree no. 2 can not be made a dataBaseView_DataSourceDB_position node ", 0, _coordref);
_currn->_desc3 = (_TSPdataBaseView_DataSourceDB_columns) MkdataBaseView_DataSourceDB_columns (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataSourceDB: root of subtree no. 3 can not be made a dataBaseView_DataSourceDB_columns node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceDB;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceDB */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceDB_name (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataSourceDB_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataSourceDB_name _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceDB_name;
#else
_currn = (_TPPpdataBaseView_DataSourceDB_name) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceDB_name));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceDB_name;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceDB_name;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceDB_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceDB_position (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataSourceDB_position (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataSourceDB_position _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataSourceDB_position;
#else
_currn = (_TPPpdataBaseView_DataSourceDB_position) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceDB_position));
#endif
_currn->_prod = RULEpdataBaseView_DataSourceDB_position;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataSourceDB_position;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataSourceDB_position */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DBColumnpdataBaseView_DataSourceDB_columns (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DBColumnpdataBaseView_DataSourceDB_columns (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DBColumnpdataBaseView_DataSourceDB_columns _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DBColumnpdataBaseView_DataSourceDB_columns;
#else
	_new = (_TPPLST_dataBaseView_DBColumnpdataBaseView_DataSourceDB_columns) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DBColumnpdataBaseView_DataSourceDB_columns));
#endif
	_new->_prod = RULELST_dataBaseView_DBColumnpdataBaseView_DataSourceDB_columns;
	_new->_desc1 = (_TSPdataBaseView_DBColumn)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataSourceDB_columns)_new;
	{_TPPLST_dataBaseView_DBColumnpdataBaseView_DataSourceDB_columns _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DBColumnpdataBaseView_DataSourceDB_columns;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempdataBaseView_DataSourceDB_columns (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempdataBaseView_DataSourceDB_columns (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBdataBaseView_DBColumn)) 
		return _MkdataBaseView_DBColumnpdataBaseView_DataSourceDB_columns (_coordref, MkdataBaseView_DBColumn (_coordref, t));
	message (DEADLY, "a pdataBaseView_DataSourceDB_columns list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pdataBaseView_DataSourceDB_columns (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pdataBaseView_DataSourceDB_columns (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_dataBaseView_DataSourceDB_columns)
	l =_ElempdataBaseView_DataSourceDB_columns (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_dataBaseView_DataSourceDB_columns)
	r = _ElempdataBaseView_DataSourceDB_columns (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pdataBaseView_DataSourceDB_columns)r)->_desc2);
((_TPPLST_0pdataBaseView_DataSourceDB_columns)r)->_desc2 = ((_TPPLST_0pdataBaseView_DataSourceDB_columns)l)->_desc2;
((_TPPLST_0pdataBaseView_DataSourceDB_columns)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataSourceDB_columns (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpdataBaseView_DataSourceDB_columns (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpdataBaseView_DataSourceDB_columns root;
	_TPPLST_0pdataBaseView_DataSourceDB_columns list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pdataBaseView_DataSourceDB_columns)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_dataBaseView_DataSourceDB_columns)
		t = _ElempdataBaseView_DataSourceDB_columns (_coordref, t);
	remain = (_TPPLST_0pdataBaseView_DataSourceDB_columns)(((_TPPLST_0pdataBaseView_DataSourceDB_columns) t)->_desc2);
	((_TPPLST_0pdataBaseView_DataSourceDB_columns) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pdataBaseView_DataSourceDB_columns;
#else
list = (_TPPLST_0pdataBaseView_DataSourceDB_columns) TreeNodeAlloc (sizeof (struct _TPLST_0pdataBaseView_DataSourceDB_columns));
#endif
list->_prod = RULELST_0pdataBaseView_DataSourceDB_columns;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pdataBaseView_DataSourceDB_columns)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_dataBaseView_DBColumnpdataBaseView_DataSourceDB_columns)
	{_TPPLST_dataBaseView_DBColumnpdataBaseView_DataSourceDB_columns _currn = _currn;
	 _TERMACT_LST_dataBaseView_DBColumnpdataBaseView_DataSourceDB_columns;
	} this code has been moved to _ElempdataBaseView_DataSourceDB_columns*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpdataBaseView_DataSourceDB_columns;
#else
root = (_TPPpdataBaseView_DataSourceDB_columns) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataSourceDB_columns));
#endif
root->_prod = RULEpdataBaseView_DataSourceDB_columns;
root->_desc1 = (_TSPLST_dataBaseView_DataSourceDB_columns)list;
{ _TPPpdataBaseView_DataSourceDB_columns _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pdataBaseView_DataSourceDB_columns;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataBase (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5, NODEPTR _desc6, NODEPTR _desc7, NODEPTR _desc8, NODEPTR _desc9, NODEPTR _desc10)
#else
NODEPTR MkpdataBaseView_DataBase (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5,_desc6,_desc7,_desc8,_desc9,_desc10)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
	NODEPTR _desc6;
	NODEPTR _desc7;
	NODEPTR _desc8;
	NODEPTR _desc9;
	NODEPTR _desc10;
#endif
{	_TPPpdataBaseView_DataBase _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataBase;
#else
_currn = (_TPPpdataBaseView_DataBase) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataBase));
#endif
_currn->_prod = RULEpdataBaseView_DataBase;
_currn->_desc1 = (_TSPdataBaseView_DataBase_name) MkdataBaseView_DataBase_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataBase: root of subtree no. 1 can not be made a dataBaseView_DataBase_name node ", 0, _coordref);
_currn->_desc2 = (_TSPdataBaseView_DataBase_userName) MkdataBaseView_DataBase_userName (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataBase: root of subtree no. 2 can not be made a dataBaseView_DataBase_userName node ", 0, _coordref);
_currn->_desc3 = (_TSPdataBaseView_DataBase_userPassword) MkdataBaseView_DataBase_userPassword (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataBase: root of subtree no. 3 can not be made a dataBaseView_DataBase_userPassword node ", 0, _coordref);
_currn->_desc4 = (_TSPdataBaseView_DataBase_hostName) MkdataBaseView_DataBase_hostName (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataBase: root of subtree no. 4 can not be made a dataBaseView_DataBase_hostName node ", 0, _coordref);
_currn->_desc5 = (_TSPdataBaseView_DataBase_port) MkdataBaseView_DataBase_port (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataBase: root of subtree no. 5 can not be made a dataBaseView_DataBase_port node ", 0, _coordref);
_currn->_desc6 = (_TSPdataBaseView_DataBase_dataBaseType) MkdataBaseView_DataBase_dataBaseType (_coordref, _desc6);	
if (((NODEPTR)_currn->_desc6) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataBase: root of subtree no. 6 can not be made a dataBaseView_DataBase_dataBaseType node ", 0, _coordref);
_currn->_desc7 = (_TSPdataBaseView_DataBase_dataSources) MkdataBaseView_DataBase_dataSources (_coordref, _desc7);	
if (((NODEPTR)_currn->_desc7) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataBase: root of subtree no. 7 can not be made a dataBaseView_DataBase_dataSources node ", 0, _coordref);
_currn->_desc8 = (_TSPdataBaseView_DataBase_dataSourceModifications) MkdataBaseView_DataBase_dataSourceModifications (_coordref, _desc8);	
if (((NODEPTR)_currn->_desc8) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataBase: root of subtree no. 8 can not be made a dataBaseView_DataBase_dataSourceModifications node ", 0, _coordref);
_currn->_desc9 = (_TSPdataBaseView_DataBase_setSize) MkdataBaseView_DataBase_setSize (_coordref, _desc9);	
if (((NODEPTR)_currn->_desc9) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataBase: root of subtree no. 9 can not be made a dataBaseView_DataBase_setSize node ", 0, _coordref);
_currn->_desc10 = (_TSPdataBaseView_DataBase_modifySetSize) MkdataBaseView_DataBase_modifySetSize (_coordref, _desc10);	
if (((NODEPTR)_currn->_desc10) == NULLNODEPTR)	
	message (DEADLY, "RULE pdataBaseView_DataBase: root of subtree no. 10 can not be made a dataBaseView_DataBase_modifySetSize node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataBase;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataBase */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataBase_name (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataBase_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataBase_name _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataBase_name;
#else
_currn = (_TPPpdataBaseView_DataBase_name) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataBase_name));
#endif
_currn->_prod = RULEpdataBaseView_DataBase_name;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataBase_name;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataBase_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataBase_userName (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataBase_userName (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataBase_userName _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataBase_userName;
#else
_currn = (_TPPpdataBaseView_DataBase_userName) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataBase_userName));
#endif
_currn->_prod = RULEpdataBaseView_DataBase_userName;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataBase_userName;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataBase_userName */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataBase_userPassword (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataBase_userPassword (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataBase_userPassword _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataBase_userPassword;
#else
_currn = (_TPPpdataBaseView_DataBase_userPassword) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataBase_userPassword));
#endif
_currn->_prod = RULEpdataBaseView_DataBase_userPassword;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataBase_userPassword;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataBase_userPassword */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataBase_hostName (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataBase_hostName (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataBase_hostName _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataBase_hostName;
#else
_currn = (_TPPpdataBaseView_DataBase_hostName) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataBase_hostName));
#endif
_currn->_prod = RULEpdataBaseView_DataBase_hostName;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataBase_hostName;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataBase_hostName */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataBase_port (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataBase_port (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataBase_port _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataBase_port;
#else
_currn = (_TPPpdataBaseView_DataBase_port) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataBase_port));
#endif
_currn->_prod = RULEpdataBaseView_DataBase_port;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataBase_port;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataBase_port */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataBase_dataBaseType (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataBase_dataBaseType (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataBase_dataBaseType _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataBase_dataBaseType;
#else
_currn = (_TPPpdataBaseView_DataBase_dataBaseType) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataBase_dataBaseType));
#endif
_currn->_prod = RULEpdataBaseView_DataBase_dataBaseType;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataBase_dataBaseType;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataBase_dataBaseType */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DataSourceDBpdataBaseView_DataBase_dataSources (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DataSourceDBpdataBaseView_DataBase_dataSources (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DataSourceDBpdataBaseView_DataBase_dataSources _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DataSourceDBpdataBaseView_DataBase_dataSources;
#else
	_new = (_TPPLST_dataBaseView_DataSourceDBpdataBaseView_DataBase_dataSources) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DataSourceDBpdataBaseView_DataBase_dataSources));
#endif
	_new->_prod = RULELST_dataBaseView_DataSourceDBpdataBaseView_DataBase_dataSources;
	_new->_desc1 = (_TSPdataBaseView_DataSourceDB)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataBase_dataSources)_new;
	{_TPPLST_dataBaseView_DataSourceDBpdataBaseView_DataBase_dataSources _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DataSourceDBpdataBaseView_DataBase_dataSources;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DataSourceProjectionFilterpdataBaseView_DataBase_dataSources (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DataSourceProjectionFilterpdataBaseView_DataBase_dataSources (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DataSourceProjectionFilterpdataBaseView_DataBase_dataSources _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DataSourceProjectionFilterpdataBaseView_DataBase_dataSources;
#else
	_new = (_TPPLST_dataBaseView_DataSourceProjectionFilterpdataBaseView_DataBase_dataSources) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DataSourceProjectionFilterpdataBaseView_DataBase_dataSources));
#endif
	_new->_prod = RULELST_dataBaseView_DataSourceProjectionFilterpdataBaseView_DataBase_dataSources;
	_new->_desc1 = (_TSPdataBaseView_DataSourceProjectionFilter)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataBase_dataSources)_new;
	{_TPPLST_dataBaseView_DataSourceProjectionFilterpdataBaseView_DataBase_dataSources _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DataSourceProjectionFilterpdataBaseView_DataBase_dataSources;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_LimitDataSourcepdataBaseView_DataBase_dataSources (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_LimitDataSourcepdataBaseView_DataBase_dataSources (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_LimitDataSourcepdataBaseView_DataBase_dataSources _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_LimitDataSourcepdataBaseView_DataBase_dataSources;
#else
	_new = (_TPPLST_dataBaseView_LimitDataSourcepdataBaseView_DataBase_dataSources) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_LimitDataSourcepdataBaseView_DataBase_dataSources));
#endif
	_new->_prod = RULELST_dataBaseView_LimitDataSourcepdataBaseView_DataBase_dataSources;
	_new->_desc1 = (_TSPdataBaseView_LimitDataSource)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataBase_dataSources)_new;
	{_TPPLST_dataBaseView_LimitDataSourcepdataBaseView_DataBase_dataSources _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_LimitDataSourcepdataBaseView_DataBase_dataSources;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DataSourceJoinpdataBaseView_DataBase_dataSources (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DataSourceJoinpdataBaseView_DataBase_dataSources (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DataSourceJoinpdataBaseView_DataBase_dataSources _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DataSourceJoinpdataBaseView_DataBase_dataSources;
#else
	_new = (_TPPLST_dataBaseView_DataSourceJoinpdataBaseView_DataBase_dataSources) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DataSourceJoinpdataBaseView_DataBase_dataSources));
#endif
	_new->_prod = RULELST_dataBaseView_DataSourceJoinpdataBaseView_DataBase_dataSources;
	_new->_desc1 = (_TSPdataBaseView_DataSourceJoin)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataBase_dataSources)_new;
	{_TPPLST_dataBaseView_DataSourceJoinpdataBaseView_DataBase_dataSources _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DataSourceJoinpdataBaseView_DataBase_dataSources;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempdataBaseView_DataBase_dataSources (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempdataBaseView_DataBase_dataSources (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBdataBaseView_DataSourceDB)) 
		return _MkdataBaseView_DataSourceDBpdataBaseView_DataBase_dataSources (_coordref, MkdataBaseView_DataSourceDB (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_DataSourceProjectionFilter)) 
		return _MkdataBaseView_DataSourceProjectionFilterpdataBaseView_DataBase_dataSources (_coordref, MkdataBaseView_DataSourceProjectionFilter (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_LimitDataSource)) 
		return _MkdataBaseView_LimitDataSourcepdataBaseView_DataBase_dataSources (_coordref, MkdataBaseView_LimitDataSource (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_DataSourceJoin)) 
		return _MkdataBaseView_DataSourceJoinpdataBaseView_DataBase_dataSources (_coordref, MkdataBaseView_DataSourceJoin (_coordref, t));
	message (DEADLY, "a pdataBaseView_DataBase_dataSources list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pdataBaseView_DataBase_dataSources (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pdataBaseView_DataBase_dataSources (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_dataBaseView_DataBase_dataSources)
	l =_ElempdataBaseView_DataBase_dataSources (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_dataBaseView_DataBase_dataSources)
	r = _ElempdataBaseView_DataBase_dataSources (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pdataBaseView_DataBase_dataSources)r)->_desc2);
((_TPPLST_0pdataBaseView_DataBase_dataSources)r)->_desc2 = ((_TPPLST_0pdataBaseView_DataBase_dataSources)l)->_desc2;
((_TPPLST_0pdataBaseView_DataBase_dataSources)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataBase_dataSources (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpdataBaseView_DataBase_dataSources (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpdataBaseView_DataBase_dataSources root;
	_TPPLST_0pdataBaseView_DataBase_dataSources list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pdataBaseView_DataBase_dataSources)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_dataBaseView_DataBase_dataSources)
		t = _ElempdataBaseView_DataBase_dataSources (_coordref, t);
	remain = (_TPPLST_0pdataBaseView_DataBase_dataSources)(((_TPPLST_0pdataBaseView_DataBase_dataSources) t)->_desc2);
	((_TPPLST_0pdataBaseView_DataBase_dataSources) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pdataBaseView_DataBase_dataSources;
#else
list = (_TPPLST_0pdataBaseView_DataBase_dataSources) TreeNodeAlloc (sizeof (struct _TPLST_0pdataBaseView_DataBase_dataSources));
#endif
list->_prod = RULELST_0pdataBaseView_DataBase_dataSources;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pdataBaseView_DataBase_dataSources)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_dataBaseView_DataSourceDBpdataBaseView_DataBase_dataSources)
	{_TPPLST_dataBaseView_DataSourceDBpdataBaseView_DataBase_dataSources _currn = _currn;
	 _TERMACT_LST_dataBaseView_DataSourceDBpdataBaseView_DataBase_dataSources;
	} this code has been moved to _ElempdataBaseView_DataBase_dataSources*/
	/*if (_currn->_prod == RULELST_dataBaseView_DataSourceProjectionFilterpdataBaseView_DataBase_dataSources)
	{_TPPLST_dataBaseView_DataSourceProjectionFilterpdataBaseView_DataBase_dataSources _currn = _currn;
	 _TERMACT_LST_dataBaseView_DataSourceProjectionFilterpdataBaseView_DataBase_dataSources;
	} this code has been moved to _ElempdataBaseView_DataBase_dataSources*/
	/*if (_currn->_prod == RULELST_dataBaseView_LimitDataSourcepdataBaseView_DataBase_dataSources)
	{_TPPLST_dataBaseView_LimitDataSourcepdataBaseView_DataBase_dataSources _currn = _currn;
	 _TERMACT_LST_dataBaseView_LimitDataSourcepdataBaseView_DataBase_dataSources;
	} this code has been moved to _ElempdataBaseView_DataBase_dataSources*/
	/*if (_currn->_prod == RULELST_dataBaseView_DataSourceJoinpdataBaseView_DataBase_dataSources)
	{_TPPLST_dataBaseView_DataSourceJoinpdataBaseView_DataBase_dataSources _currn = _currn;
	 _TERMACT_LST_dataBaseView_DataSourceJoinpdataBaseView_DataBase_dataSources;
	} this code has been moved to _ElempdataBaseView_DataBase_dataSources*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpdataBaseView_DataBase_dataSources;
#else
root = (_TPPpdataBaseView_DataBase_dataSources) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataBase_dataSources));
#endif
root->_prod = RULEpdataBaseView_DataBase_dataSources;
root->_desc1 = (_TSPLST_dataBaseView_DataBase_dataSources)list;
{ _TPPpdataBaseView_DataBase_dataSources _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pdataBaseView_DataBase_dataSources;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DataSourceInsertpdataBaseView_DataBase_dataSourceModifications (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DataSourceInsertpdataBaseView_DataBase_dataSourceModifications (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DataSourceInsertpdataBaseView_DataBase_dataSourceModifications _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DataSourceInsertpdataBaseView_DataBase_dataSourceModifications;
#else
	_new = (_TPPLST_dataBaseView_DataSourceInsertpdataBaseView_DataBase_dataSourceModifications) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DataSourceInsertpdataBaseView_DataBase_dataSourceModifications));
#endif
	_new->_prod = RULELST_dataBaseView_DataSourceInsertpdataBaseView_DataBase_dataSourceModifications;
	_new->_desc1 = (_TSPdataBaseView_DataSourceInsert)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataBase_dataSourceModifications)_new;
	{_TPPLST_dataBaseView_DataSourceInsertpdataBaseView_DataBase_dataSourceModifications _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DataSourceInsertpdataBaseView_DataBase_dataSourceModifications;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DataSourceUpdatepdataBaseView_DataBase_dataSourceModifications (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DataSourceUpdatepdataBaseView_DataBase_dataSourceModifications (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DataSourceUpdatepdataBaseView_DataBase_dataSourceModifications _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DataSourceUpdatepdataBaseView_DataBase_dataSourceModifications;
#else
	_new = (_TPPLST_dataBaseView_DataSourceUpdatepdataBaseView_DataBase_dataSourceModifications) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DataSourceUpdatepdataBaseView_DataBase_dataSourceModifications));
#endif
	_new->_prod = RULELST_dataBaseView_DataSourceUpdatepdataBaseView_DataBase_dataSourceModifications;
	_new->_desc1 = (_TSPdataBaseView_DataSourceUpdate)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataBase_dataSourceModifications)_new;
	{_TPPLST_dataBaseView_DataSourceUpdatepdataBaseView_DataBase_dataSourceModifications _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DataSourceUpdatepdataBaseView_DataBase_dataSourceModifications;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkdataBaseView_DataSourceDeletepdataBaseView_DataBase_dataSourceModifications (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkdataBaseView_DataSourceDeletepdataBaseView_DataBase_dataSourceModifications (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_dataBaseView_DataSourceDeletepdataBaseView_DataBase_dataSourceModifications _new;
#ifdef __cplusplus
	_new = new _TPLST_dataBaseView_DataSourceDeletepdataBaseView_DataBase_dataSourceModifications;
#else
	_new = (_TPPLST_dataBaseView_DataSourceDeletepdataBaseView_DataBase_dataSourceModifications) TreeNodeAlloc (sizeof (struct _TPLST_dataBaseView_DataSourceDeletepdataBaseView_DataBase_dataSourceModifications));
#endif
	_new->_prod = RULELST_dataBaseView_DataSourceDeletepdataBaseView_DataBase_dataSourceModifications;
	_new->_desc1 = (_TSPdataBaseView_DataSourceDelete)elem;
	_new->_desc2 = (_TSPLST_dataBaseView_DataBase_dataSourceModifications)_new;
	{_TPPLST_dataBaseView_DataSourceDeletepdataBaseView_DataBase_dataSourceModifications _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_dataBaseView_DataSourceDeletepdataBaseView_DataBase_dataSourceModifications;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempdataBaseView_DataBase_dataSourceModifications (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempdataBaseView_DataBase_dataSourceModifications (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBdataBaseView_DataSourceInsert)) 
		return _MkdataBaseView_DataSourceInsertpdataBaseView_DataBase_dataSourceModifications (_coordref, MkdataBaseView_DataSourceInsert (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_DataSourceUpdate)) 
		return _MkdataBaseView_DataSourceUpdatepdataBaseView_DataBase_dataSourceModifications (_coordref, MkdataBaseView_DataSourceUpdate (_coordref, t));
	if (IsSymb (t, SYMBdataBaseView_DataSourceDelete)) 
		return _MkdataBaseView_DataSourceDeletepdataBaseView_DataBase_dataSourceModifications (_coordref, MkdataBaseView_DataSourceDelete (_coordref, t));
	message (DEADLY, "a pdataBaseView_DataBase_dataSourceModifications list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pdataBaseView_DataBase_dataSourceModifications (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pdataBaseView_DataBase_dataSourceModifications (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_dataBaseView_DataBase_dataSourceModifications)
	l =_ElempdataBaseView_DataBase_dataSourceModifications (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_dataBaseView_DataBase_dataSourceModifications)
	r = _ElempdataBaseView_DataBase_dataSourceModifications (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pdataBaseView_DataBase_dataSourceModifications)r)->_desc2);
((_TPPLST_0pdataBaseView_DataBase_dataSourceModifications)r)->_desc2 = ((_TPPLST_0pdataBaseView_DataBase_dataSourceModifications)l)->_desc2;
((_TPPLST_0pdataBaseView_DataBase_dataSourceModifications)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataBase_dataSourceModifications (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpdataBaseView_DataBase_dataSourceModifications (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpdataBaseView_DataBase_dataSourceModifications root;
	_TPPLST_0pdataBaseView_DataBase_dataSourceModifications list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pdataBaseView_DataBase_dataSourceModifications)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_dataBaseView_DataBase_dataSourceModifications)
		t = _ElempdataBaseView_DataBase_dataSourceModifications (_coordref, t);
	remain = (_TPPLST_0pdataBaseView_DataBase_dataSourceModifications)(((_TPPLST_0pdataBaseView_DataBase_dataSourceModifications) t)->_desc2);
	((_TPPLST_0pdataBaseView_DataBase_dataSourceModifications) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pdataBaseView_DataBase_dataSourceModifications;
#else
list = (_TPPLST_0pdataBaseView_DataBase_dataSourceModifications) TreeNodeAlloc (sizeof (struct _TPLST_0pdataBaseView_DataBase_dataSourceModifications));
#endif
list->_prod = RULELST_0pdataBaseView_DataBase_dataSourceModifications;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pdataBaseView_DataBase_dataSourceModifications)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_dataBaseView_DataSourceInsertpdataBaseView_DataBase_dataSourceModifications)
	{_TPPLST_dataBaseView_DataSourceInsertpdataBaseView_DataBase_dataSourceModifications _currn = _currn;
	 _TERMACT_LST_dataBaseView_DataSourceInsertpdataBaseView_DataBase_dataSourceModifications;
	} this code has been moved to _ElempdataBaseView_DataBase_dataSourceModifications*/
	/*if (_currn->_prod == RULELST_dataBaseView_DataSourceUpdatepdataBaseView_DataBase_dataSourceModifications)
	{_TPPLST_dataBaseView_DataSourceUpdatepdataBaseView_DataBase_dataSourceModifications _currn = _currn;
	 _TERMACT_LST_dataBaseView_DataSourceUpdatepdataBaseView_DataBase_dataSourceModifications;
	} this code has been moved to _ElempdataBaseView_DataBase_dataSourceModifications*/
	/*if (_currn->_prod == RULELST_dataBaseView_DataSourceDeletepdataBaseView_DataBase_dataSourceModifications)
	{_TPPLST_dataBaseView_DataSourceDeletepdataBaseView_DataBase_dataSourceModifications _currn = _currn;
	 _TERMACT_LST_dataBaseView_DataSourceDeletepdataBaseView_DataBase_dataSourceModifications;
	} this code has been moved to _ElempdataBaseView_DataBase_dataSourceModifications*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpdataBaseView_DataBase_dataSourceModifications;
#else
root = (_TPPpdataBaseView_DataBase_dataSourceModifications) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataBase_dataSourceModifications));
#endif
root->_prod = RULEpdataBaseView_DataBase_dataSourceModifications;
root->_desc1 = (_TSPLST_dataBaseView_DataBase_dataSourceModifications)list;
{ _TPPpdataBaseView_DataBase_dataSourceModifications _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pdataBaseView_DataBase_dataSourceModifications;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataBase_setSize (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataBase_setSize (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataBase_setSize _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataBase_setSize;
#else
_currn = (_TPPpdataBaseView_DataBase_setSize) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataBase_setSize));
#endif
_currn->_prod = RULEpdataBaseView_DataBase_setSize;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataBase_setSize;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataBase_setSize */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpdataBaseView_DataBase_modifySetSize (POSITION *_coordref)
#else
NODEPTR MkpdataBaseView_DataBase_modifySetSize (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpdataBaseView_DataBase_modifySetSize _currn;
#ifdef __cplusplus
_currn = new _TPpdataBaseView_DataBase_modifySetSize;
#else
_currn = (_TPPpdataBaseView_DataBase_modifySetSize) TreeNodeAlloc (sizeof (struct _TPpdataBaseView_DataBase_modifySetSize));
#endif
_currn->_prod = RULEpdataBaseView_DataBase_modifySetSize;
_SETCOORD(_currn)
_TERMACT_pdataBaseView_DataBase_modifySetSize;
return ( (NODEPTR) _currn);
}/* MkpdataBaseView_DataBase_modifySetSize */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpVIEWROOT_fcDlg (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR MkpVIEWROOT_fcDlg (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPpVIEWROOT_fcDlg _currn;
#ifdef __cplusplus
_currn = new _TPpVIEWROOT_fcDlg;
#else
_currn = (_TPPpVIEWROOT_fcDlg) TreeNodeAlloc (sizeof (struct _TPpVIEWROOT_fcDlg));
#endif
_currn->_prod = RULEpVIEWROOT_fcDlg;
_currn->_desc1 = (_TSPfcDlg_FormularCondition) MkfcDlg_FormularCondition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pVIEWROOT_fcDlg: root of subtree no. 1 can not be made a fcDlg_FormularCondition node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_pVIEWROOT_fcDlg;
return ( (NODEPTR) _currn);
}/* MkpVIEWROOT_fcDlg */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpfcDlg_FormularCondition (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpfcDlg_FormularCondition (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpfcDlg_FormularCondition _currn;
#ifdef __cplusplus
_currn = new _TPpfcDlg_FormularCondition;
#else
_currn = (_TPPpfcDlg_FormularCondition) TreeNodeAlloc (sizeof (struct _TPpfcDlg_FormularCondition));
#endif
_currn->_prod = RULEpfcDlg_FormularCondition;
_currn->_desc1 = (_TSPfcDlg_FormularCondition_condition) MkfcDlg_FormularCondition_condition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pfcDlg_FormularCondition: root of subtree no. 1 can not be made a fcDlg_FormularCondition_condition node ", 0, _coordref);
_currn->_desc2 = (_TSPfcDlg_FormularCondition_params) MkfcDlg_FormularCondition_params (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pfcDlg_FormularCondition: root of subtree no. 2 can not be made a fcDlg_FormularCondition_params node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pfcDlg_FormularCondition;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpfcDlg_FormularCondition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpfcDlg_FormularCondition_condition (POSITION *_coordref)
#else
NODEPTR MkpfcDlg_FormularCondition_condition (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpfcDlg_FormularCondition_condition _currn;
#ifdef __cplusplus
_currn = new _TPpfcDlg_FormularCondition_condition;
#else
_currn = (_TPPpfcDlg_FormularCondition_condition) TreeNodeAlloc (sizeof (struct _TPpfcDlg_FormularCondition_condition));
#endif
_currn->_prod = RULEpfcDlg_FormularCondition_condition;
_SETCOORD(_currn)
_TERMACT_pfcDlg_FormularCondition_condition;
return ( (NODEPTR) _currn);
}/* MkpfcDlg_FormularCondition_condition */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkfcDlg_FormularConditionActualParampfcDlg_FormularCondition_params (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkfcDlg_FormularConditionActualParampfcDlg_FormularCondition_params (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_fcDlg_FormularConditionActualParampfcDlg_FormularCondition_params _new;
#ifdef __cplusplus
	_new = new _TPLST_fcDlg_FormularConditionActualParampfcDlg_FormularCondition_params;
#else
	_new = (_TPPLST_fcDlg_FormularConditionActualParampfcDlg_FormularCondition_params) TreeNodeAlloc (sizeof (struct _TPLST_fcDlg_FormularConditionActualParampfcDlg_FormularCondition_params));
#endif
	_new->_prod = RULELST_fcDlg_FormularConditionActualParampfcDlg_FormularCondition_params;
	_new->_desc1 = (_TSPfcDlg_FormularConditionActualParam)elem;
	_new->_desc2 = (_TSPLST_fcDlg_FormularCondition_params)_new;
	{_TPPLST_fcDlg_FormularConditionActualParampfcDlg_FormularCondition_params _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_fcDlg_FormularConditionActualParampfcDlg_FormularCondition_params;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempfcDlg_FormularCondition_params (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempfcDlg_FormularCondition_params (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBfcDlg_FormularConditionActualParam)) 
		return _MkfcDlg_FormularConditionActualParampfcDlg_FormularCondition_params (_coordref, MkfcDlg_FormularConditionActualParam (_coordref, t));
	message (DEADLY, "a pfcDlg_FormularCondition_params list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pfcDlg_FormularCondition_params (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pfcDlg_FormularCondition_params (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_fcDlg_FormularCondition_params)
	l =_ElempfcDlg_FormularCondition_params (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_fcDlg_FormularCondition_params)
	r = _ElempfcDlg_FormularCondition_params (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pfcDlg_FormularCondition_params)r)->_desc2);
((_TPPLST_0pfcDlg_FormularCondition_params)r)->_desc2 = ((_TPPLST_0pfcDlg_FormularCondition_params)l)->_desc2;
((_TPPLST_0pfcDlg_FormularCondition_params)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpfcDlg_FormularCondition_params (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpfcDlg_FormularCondition_params (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpfcDlg_FormularCondition_params root;
	_TPPLST_0pfcDlg_FormularCondition_params list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pfcDlg_FormularCondition_params)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_fcDlg_FormularCondition_params)
		t = _ElempfcDlg_FormularCondition_params (_coordref, t);
	remain = (_TPPLST_0pfcDlg_FormularCondition_params)(((_TPPLST_0pfcDlg_FormularCondition_params) t)->_desc2);
	((_TPPLST_0pfcDlg_FormularCondition_params) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pfcDlg_FormularCondition_params;
#else
list = (_TPPLST_0pfcDlg_FormularCondition_params) TreeNodeAlloc (sizeof (struct _TPLST_0pfcDlg_FormularCondition_params));
#endif
list->_prod = RULELST_0pfcDlg_FormularCondition_params;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pfcDlg_FormularCondition_params)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_fcDlg_FormularConditionActualParampfcDlg_FormularCondition_params)
	{_TPPLST_fcDlg_FormularConditionActualParampfcDlg_FormularCondition_params _currn = _currn;
	 _TERMACT_LST_fcDlg_FormularConditionActualParampfcDlg_FormularCondition_params;
	} this code has been moved to _ElempfcDlg_FormularCondition_params*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpfcDlg_FormularCondition_params;
#else
root = (_TPPpfcDlg_FormularCondition_params) TreeNodeAlloc (sizeof (struct _TPpfcDlg_FormularCondition_params));
#endif
root->_prod = RULEpfcDlg_FormularCondition_params;
root->_desc1 = (_TSPLST_fcDlg_FormularCondition_params)list;
{ _TPPpfcDlg_FormularCondition_params _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pfcDlg_FormularCondition_params;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpfcDlg_FormularConditionActualParam (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpfcDlg_FormularConditionActualParam (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpfcDlg_FormularConditionActualParam _currn;
#ifdef __cplusplus
_currn = new _TPpfcDlg_FormularConditionActualParam;
#else
_currn = (_TPPpfcDlg_FormularConditionActualParam) TreeNodeAlloc (sizeof (struct _TPpfcDlg_FormularConditionActualParam));
#endif
_currn->_prod = RULEpfcDlg_FormularConditionActualParam;
_currn->_desc1 = (_TSPfcDlg_FormularConditionActualParam_value) MkfcDlg_FormularConditionActualParam_value (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pfcDlg_FormularConditionActualParam: root of subtree no. 1 can not be made a fcDlg_FormularConditionActualParam_value node ", 0, _coordref);
_currn->_desc2 = (_TSPfcDlg_FormularConditionActualParam_formalparam) MkfcDlg_FormularConditionActualParam_formalparam (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pfcDlg_FormularConditionActualParam: root of subtree no. 2 can not be made a fcDlg_FormularConditionActualParam_formalparam node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pfcDlg_FormularConditionActualParam;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpfcDlg_FormularConditionActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpfcDlg_FormularConditionActualParam_value (POSITION *_coordref)
#else
NODEPTR MkpfcDlg_FormularConditionActualParam_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpfcDlg_FormularConditionActualParam_value _currn;
#ifdef __cplusplus
_currn = new _TPpfcDlg_FormularConditionActualParam_value;
#else
_currn = (_TPPpfcDlg_FormularConditionActualParam_value) TreeNodeAlloc (sizeof (struct _TPpfcDlg_FormularConditionActualParam_value));
#endif
_currn->_prod = RULEpfcDlg_FormularConditionActualParam_value;
_SETCOORD(_currn)
_TERMACT_pfcDlg_FormularConditionActualParam_value;
return ( (NODEPTR) _currn);
}/* MkpfcDlg_FormularConditionActualParam_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpfcDlg_FormularConditionActualParam_formalparam (POSITION *_coordref)
#else
NODEPTR MkpfcDlg_FormularConditionActualParam_formalparam (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpfcDlg_FormularConditionActualParam_formalparam _currn;
#ifdef __cplusplus
_currn = new _TPpfcDlg_FormularConditionActualParam_formalparam;
#else
_currn = (_TPPpfcDlg_FormularConditionActualParam_formalparam) TreeNodeAlloc (sizeof (struct _TPpfcDlg_FormularConditionActualParam_formalparam));
#endif
_currn->_prod = RULEpfcDlg_FormularConditionActualParam_formalparam;
_SETCOORD(_currn)
_TERMACT_pfcDlg_FormularConditionActualParam_formalparam;
return ( (NODEPTR) _currn);
}/* MkpfcDlg_FormularConditionActualParam_formalparam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpVIEWROOT_globalFrameView (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR MkpVIEWROOT_globalFrameView (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPpVIEWROOT_globalFrameView _currn;
#ifdef __cplusplus
_currn = new _TPpVIEWROOT_globalFrameView;
#else
_currn = (_TPPpVIEWROOT_globalFrameView) TreeNodeAlloc (sizeof (struct _TPpVIEWROOT_globalFrameView));
#endif
_currn->_prod = RULEpVIEWROOT_globalFrameView;
_currn->_desc1 = (_TSPglobalFrameView_GlobalFrame) MkglobalFrameView_GlobalFrame (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pVIEWROOT_globalFrameView: root of subtree no. 1 can not be made a globalFrameView_GlobalFrame node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_pVIEWROOT_globalFrameView;
return ( (NODEPTR) _currn);
}/* MkpVIEWROOT_globalFrameView */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularCheckboxElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4)
#else
NODEPTR MkpglobalFrameView_FormularCheckboxElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
#endif
{	_TPPpglobalFrameView_FormularCheckboxElement _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularCheckboxElement;
#else
_currn = (_TPPpglobalFrameView_FormularCheckboxElement) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularCheckboxElement));
#endif
_currn->_prod = RULEpglobalFrameView_FormularCheckboxElement;
_currn->_desc1 = (_TSPglobalFrameView_FormularCheckboxElement_label) MkglobalFrameView_FormularCheckboxElement_label (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularCheckboxElement: root of subtree no. 1 can not be made a globalFrameView_FormularCheckboxElement_label node ", 0, _coordref);
_currn->_desc2 = (_TSPglobalFrameView_FormularCheckboxElement_type) MkglobalFrameView_FormularCheckboxElement_type (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularCheckboxElement: root of subtree no. 2 can not be made a globalFrameView_FormularCheckboxElement_type node ", 0, _coordref);
_currn->_desc3 = (_TSPglobalFrameView_FormularCheckboxElement_condition) MkglobalFrameView_FormularCheckboxElement_condition (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularCheckboxElement: root of subtree no. 3 can not be made a globalFrameView_FormularCheckboxElement_condition node ", 0, _coordref);
_currn->_desc4 = (_TSPglobalFrameView_FormularCheckboxElement_description) MkglobalFrameView_FormularCheckboxElement_description (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularCheckboxElement: root of subtree no. 4 can not be made a globalFrameView_FormularCheckboxElement_description node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularCheckboxElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularCheckboxElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularCheckboxElement_label (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularCheckboxElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularCheckboxElement_label _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularCheckboxElement_label;
#else
_currn = (_TPPpglobalFrameView_FormularCheckboxElement_label) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularCheckboxElement_label));
#endif
_currn->_prod = RULEpglobalFrameView_FormularCheckboxElement_label;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularCheckboxElement_label;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularCheckboxElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularCheckboxElement_name (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularCheckboxElement_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularCheckboxElement_name _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularCheckboxElement_name;
#else
_currn = (_TPPpglobalFrameView_FormularCheckboxElement_name) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularCheckboxElement_name));
#endif
_currn->_prod = RULEpglobalFrameView_FormularCheckboxElement_name;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularCheckboxElement_name;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularCheckboxElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularCheckboxElement_type (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR MkpglobalFrameView_FormularCheckboxElement_type (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPpglobalFrameView_FormularCheckboxElement_type _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularCheckboxElement_type;
#else
_currn = (_TPPpglobalFrameView_FormularCheckboxElement_type) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularCheckboxElement_type));
#endif
_currn->_prod = RULEpglobalFrameView_FormularCheckboxElement_type;
_currn->_desc1 = (_TSPglobalFrameView_FormularCheckboxElement_name) MkglobalFrameView_FormularCheckboxElement_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularCheckboxElement_type: root of subtree no. 1 can not be made a globalFrameView_FormularCheckboxElement_name node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_pglobalFrameView_FormularCheckboxElement_type;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularCheckboxElement_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularCheckboxElement_description (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularCheckboxElement_description (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularCheckboxElement_description _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularCheckboxElement_description;
#else
_currn = (_TPPpglobalFrameView_FormularCheckboxElement_description) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularCheckboxElement_description));
#endif
_currn->_prod = RULEpglobalFrameView_FormularCheckboxElement_description;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularCheckboxElement_description;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularCheckboxElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularTextElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4)
#else
NODEPTR MkpglobalFrameView_FormularTextElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
#endif
{	_TPPpglobalFrameView_FormularTextElement _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularTextElement;
#else
_currn = (_TPPpglobalFrameView_FormularTextElement) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularTextElement));
#endif
_currn->_prod = RULEpglobalFrameView_FormularTextElement;
_currn->_desc1 = (_TSPglobalFrameView_FormularTextElement_label) MkglobalFrameView_FormularTextElement_label (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularTextElement: root of subtree no. 1 can not be made a globalFrameView_FormularTextElement_label node ", 0, _coordref);
_currn->_desc2 = (_TSPglobalFrameView_FormularTextElement_type) MkglobalFrameView_FormularTextElement_type (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularTextElement: root of subtree no. 2 can not be made a globalFrameView_FormularTextElement_type node ", 0, _coordref);
_currn->_desc3 = (_TSPglobalFrameView_FormularTextElement_condition) MkglobalFrameView_FormularTextElement_condition (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularTextElement: root of subtree no. 3 can not be made a globalFrameView_FormularTextElement_condition node ", 0, _coordref);
_currn->_desc4 = (_TSPglobalFrameView_FormularTextElement_description) MkglobalFrameView_FormularTextElement_description (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularTextElement: root of subtree no. 4 can not be made a globalFrameView_FormularTextElement_description node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularTextElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularTextElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularTextElement_label (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularTextElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularTextElement_label _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularTextElement_label;
#else
_currn = (_TPPpglobalFrameView_FormularTextElement_label) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularTextElement_label));
#endif
_currn->_prod = RULEpglobalFrameView_FormularTextElement_label;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularTextElement_label;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularTextElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularTextElement_type (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularTextElement_type (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularTextElement_type _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularTextElement_type;
#else
_currn = (_TPPpglobalFrameView_FormularTextElement_type) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularTextElement_type));
#endif
_currn->_prod = RULEpglobalFrameView_FormularTextElement_type;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularTextElement_type;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularTextElement_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularTextElement_description (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularTextElement_description (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularTextElement_description _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularTextElement_description;
#else
_currn = (_TPPpglobalFrameView_FormularTextElement_description) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularTextElement_description));
#endif
_currn->_prod = RULEpglobalFrameView_FormularTextElement_description;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularTextElement_description;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularTextElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularPasswordElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4)
#else
NODEPTR MkpglobalFrameView_FormularPasswordElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
#endif
{	_TPPpglobalFrameView_FormularPasswordElement _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularPasswordElement;
#else
_currn = (_TPPpglobalFrameView_FormularPasswordElement) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularPasswordElement));
#endif
_currn->_prod = RULEpglobalFrameView_FormularPasswordElement;
_currn->_desc1 = (_TSPglobalFrameView_FormularPasswordElement_label) MkglobalFrameView_FormularPasswordElement_label (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularPasswordElement: root of subtree no. 1 can not be made a globalFrameView_FormularPasswordElement_label node ", 0, _coordref);
_currn->_desc2 = (_TSPglobalFrameView_FormularPasswordElement_type) MkglobalFrameView_FormularPasswordElement_type (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularPasswordElement: root of subtree no. 2 can not be made a globalFrameView_FormularPasswordElement_type node ", 0, _coordref);
_currn->_desc3 = (_TSPglobalFrameView_FormularPasswordElement_condition) MkglobalFrameView_FormularPasswordElement_condition (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularPasswordElement: root of subtree no. 3 can not be made a globalFrameView_FormularPasswordElement_condition node ", 0, _coordref);
_currn->_desc4 = (_TSPglobalFrameView_FormularPasswordElement_description) MkglobalFrameView_FormularPasswordElement_description (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularPasswordElement: root of subtree no. 4 can not be made a globalFrameView_FormularPasswordElement_description node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularPasswordElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularPasswordElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularPasswordElement_label (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularPasswordElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularPasswordElement_label _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularPasswordElement_label;
#else
_currn = (_TPPpglobalFrameView_FormularPasswordElement_label) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularPasswordElement_label));
#endif
_currn->_prod = RULEpglobalFrameView_FormularPasswordElement_label;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularPasswordElement_label;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularPasswordElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularPasswordElement_type (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularPasswordElement_type (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularPasswordElement_type _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularPasswordElement_type;
#else
_currn = (_TPPpglobalFrameView_FormularPasswordElement_type) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularPasswordElement_type));
#endif
_currn->_prod = RULEpglobalFrameView_FormularPasswordElement_type;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularPasswordElement_type;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularPasswordElement_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularPasswordElement_description (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularPasswordElement_description (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularPasswordElement_description _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularPasswordElement_description;
#else
_currn = (_TPPpglobalFrameView_FormularPasswordElement_description) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularPasswordElement_description));
#endif
_currn->_prod = RULEpglobalFrameView_FormularPasswordElement_description;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularPasswordElement_description;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularPasswordElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularTextAreaElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4)
#else
NODEPTR MkpglobalFrameView_FormularTextAreaElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
#endif
{	_TPPpglobalFrameView_FormularTextAreaElement _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularTextAreaElement;
#else
_currn = (_TPPpglobalFrameView_FormularTextAreaElement) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularTextAreaElement));
#endif
_currn->_prod = RULEpglobalFrameView_FormularTextAreaElement;
_currn->_desc1 = (_TSPglobalFrameView_FormularTextAreaElement_label) MkglobalFrameView_FormularTextAreaElement_label (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularTextAreaElement: root of subtree no. 1 can not be made a globalFrameView_FormularTextAreaElement_label node ", 0, _coordref);
_currn->_desc2 = (_TSPglobalFrameView_FormularTextAreaElement_type) MkglobalFrameView_FormularTextAreaElement_type (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularTextAreaElement: root of subtree no. 2 can not be made a globalFrameView_FormularTextAreaElement_type node ", 0, _coordref);
_currn->_desc3 = (_TSPglobalFrameView_FormularTextAreaElement_condition) MkglobalFrameView_FormularTextAreaElement_condition (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularTextAreaElement: root of subtree no. 3 can not be made a globalFrameView_FormularTextAreaElement_condition node ", 0, _coordref);
_currn->_desc4 = (_TSPglobalFrameView_FormularTextAreaElement_description) MkglobalFrameView_FormularTextAreaElement_description (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularTextAreaElement: root of subtree no. 4 can not be made a globalFrameView_FormularTextAreaElement_description node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularTextAreaElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularTextAreaElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularTextAreaElement_label (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularTextAreaElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularTextAreaElement_label _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularTextAreaElement_label;
#else
_currn = (_TPPpglobalFrameView_FormularTextAreaElement_label) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularTextAreaElement_label));
#endif
_currn->_prod = RULEpglobalFrameView_FormularTextAreaElement_label;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularTextAreaElement_label;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularTextAreaElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularTextAreaElement_type (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularTextAreaElement_type (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularTextAreaElement_type _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularTextAreaElement_type;
#else
_currn = (_TPPpglobalFrameView_FormularTextAreaElement_type) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularTextAreaElement_type));
#endif
_currn->_prod = RULEpglobalFrameView_FormularTextAreaElement_type;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularTextAreaElement_type;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularTextAreaElement_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularTextAreaElement_description (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularTextAreaElement_description (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularTextAreaElement_description _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularTextAreaElement_description;
#else
_currn = (_TPPpglobalFrameView_FormularTextAreaElement_description) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularTextAreaElement_description));
#endif
_currn->_prod = RULEpglobalFrameView_FormularTextAreaElement_description;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularTextAreaElement_description;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularTextAreaElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularSelectElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4)
#else
NODEPTR MkpglobalFrameView_FormularSelectElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
#endif
{	_TPPpglobalFrameView_FormularSelectElement _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularSelectElement;
#else
_currn = (_TPPpglobalFrameView_FormularSelectElement) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularSelectElement));
#endif
_currn->_prod = RULEpglobalFrameView_FormularSelectElement;
_currn->_desc1 = (_TSPglobalFrameView_FormularSelectElement_label) MkglobalFrameView_FormularSelectElement_label (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularSelectElement: root of subtree no. 1 can not be made a globalFrameView_FormularSelectElement_label node ", 0, _coordref);
_currn->_desc2 = (_TSPglobalFrameView_FormularSelectElement_multiple) MkglobalFrameView_FormularSelectElement_multiple (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularSelectElement: root of subtree no. 2 can not be made a globalFrameView_FormularSelectElement_multiple node ", 0, _coordref);
_currn->_desc3 = (_TSPglobalFrameView_FormularSelectElement_options) MkglobalFrameView_FormularSelectElement_options (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularSelectElement: root of subtree no. 3 can not be made a globalFrameView_FormularSelectElement_options node ", 0, _coordref);
_currn->_desc4 = (_TSPglobalFrameView_FormularSelectElement_description) MkglobalFrameView_FormularSelectElement_description (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularSelectElement: root of subtree no. 4 can not be made a globalFrameView_FormularSelectElement_description node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularSelectElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularSelectElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularSelectElement_label (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularSelectElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularSelectElement_label _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularSelectElement_label;
#else
_currn = (_TPPpglobalFrameView_FormularSelectElement_label) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularSelectElement_label));
#endif
_currn->_prod = RULEpglobalFrameView_FormularSelectElement_label;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularSelectElement_label;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularSelectElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularSelectElement_multiple (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularSelectElement_multiple (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularSelectElement_multiple _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularSelectElement_multiple;
#else
_currn = (_TPPpglobalFrameView_FormularSelectElement_multiple) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularSelectElement_multiple));
#endif
_currn->_prod = RULEpglobalFrameView_FormularSelectElement_multiple;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularSelectElement_multiple;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularSelectElement_multiple */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_FormularConstantSelectOptionpglobalFrameView_FormularSelectElement_options (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_FormularConstantSelectOptionpglobalFrameView_FormularSelectElement_options (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_FormularConstantSelectOptionpglobalFrameView_FormularSelectElement_options _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_FormularConstantSelectOptionpglobalFrameView_FormularSelectElement_options;
#else
	_new = (_TPPLST_globalFrameView_FormularConstantSelectOptionpglobalFrameView_FormularSelectElement_options) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_FormularConstantSelectOptionpglobalFrameView_FormularSelectElement_options));
#endif
	_new->_prod = RULELST_globalFrameView_FormularConstantSelectOptionpglobalFrameView_FormularSelectElement_options;
	_new->_desc1 = (_TSPglobalFrameView_FormularConstantSelectOption)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_FormularSelectElement_options)_new;
	{_TPPLST_globalFrameView_FormularConstantSelectOptionpglobalFrameView_FormularSelectElement_options _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_FormularConstantSelectOptionpglobalFrameView_FormularSelectElement_options;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_FormularDataSourceSelectOptionpglobalFrameView_FormularSelectElement_options (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_FormularDataSourceSelectOptionpglobalFrameView_FormularSelectElement_options (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_FormularDataSourceSelectOptionpglobalFrameView_FormularSelectElement_options _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_FormularDataSourceSelectOptionpglobalFrameView_FormularSelectElement_options;
#else
	_new = (_TPPLST_globalFrameView_FormularDataSourceSelectOptionpglobalFrameView_FormularSelectElement_options) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_FormularDataSourceSelectOptionpglobalFrameView_FormularSelectElement_options));
#endif
	_new->_prod = RULELST_globalFrameView_FormularDataSourceSelectOptionpglobalFrameView_FormularSelectElement_options;
	_new->_desc1 = (_TSPglobalFrameView_FormularDataSourceSelectOption)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_FormularSelectElement_options)_new;
	{_TPPLST_globalFrameView_FormularDataSourceSelectOptionpglobalFrameView_FormularSelectElement_options _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_FormularDataSourceSelectOptionpglobalFrameView_FormularSelectElement_options;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_FormularPHPScriptSelectOptionpglobalFrameView_FormularSelectElement_options (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_FormularPHPScriptSelectOptionpglobalFrameView_FormularSelectElement_options (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_FormularPHPScriptSelectOptionpglobalFrameView_FormularSelectElement_options _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_FormularPHPScriptSelectOptionpglobalFrameView_FormularSelectElement_options;
#else
	_new = (_TPPLST_globalFrameView_FormularPHPScriptSelectOptionpglobalFrameView_FormularSelectElement_options) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_FormularPHPScriptSelectOptionpglobalFrameView_FormularSelectElement_options));
#endif
	_new->_prod = RULELST_globalFrameView_FormularPHPScriptSelectOptionpglobalFrameView_FormularSelectElement_options;
	_new->_desc1 = (_TSPglobalFrameView_FormularPHPScriptSelectOption)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_FormularSelectElement_options)_new;
	{_TPPLST_globalFrameView_FormularPHPScriptSelectOptionpglobalFrameView_FormularSelectElement_options _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_FormularPHPScriptSelectOptionpglobalFrameView_FormularSelectElement_options;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempglobalFrameView_FormularSelectElement_options (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempglobalFrameView_FormularSelectElement_options (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBglobalFrameView_FormularConstantSelectOption)) 
		return _MkglobalFrameView_FormularConstantSelectOptionpglobalFrameView_FormularSelectElement_options (_coordref, MkglobalFrameView_FormularConstantSelectOption (_coordref, t));
	if (IsSymb (t, SYMBglobalFrameView_FormularDataSourceSelectOption)) 
		return _MkglobalFrameView_FormularDataSourceSelectOptionpglobalFrameView_FormularSelectElement_options (_coordref, MkglobalFrameView_FormularDataSourceSelectOption (_coordref, t));
	if (IsSymb (t, SYMBglobalFrameView_FormularPHPScriptSelectOption)) 
		return _MkglobalFrameView_FormularPHPScriptSelectOptionpglobalFrameView_FormularSelectElement_options (_coordref, MkglobalFrameView_FormularPHPScriptSelectOption (_coordref, t));
	message (DEADLY, "a pglobalFrameView_FormularSelectElement_options list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pglobalFrameView_FormularSelectElement_options (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pglobalFrameView_FormularSelectElement_options (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_globalFrameView_FormularSelectElement_options)
	l =_ElempglobalFrameView_FormularSelectElement_options (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_globalFrameView_FormularSelectElement_options)
	r = _ElempglobalFrameView_FormularSelectElement_options (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pglobalFrameView_FormularSelectElement_options)r)->_desc2);
((_TPPLST_0pglobalFrameView_FormularSelectElement_options)r)->_desc2 = ((_TPPLST_0pglobalFrameView_FormularSelectElement_options)l)->_desc2;
((_TPPLST_0pglobalFrameView_FormularSelectElement_options)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularSelectElement_options (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpglobalFrameView_FormularSelectElement_options (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpglobalFrameView_FormularSelectElement_options root;
	_TPPLST_0pglobalFrameView_FormularSelectElement_options list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pglobalFrameView_FormularSelectElement_options)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_globalFrameView_FormularSelectElement_options)
		t = _ElempglobalFrameView_FormularSelectElement_options (_coordref, t);
	remain = (_TPPLST_0pglobalFrameView_FormularSelectElement_options)(((_TPPLST_0pglobalFrameView_FormularSelectElement_options) t)->_desc2);
	((_TPPLST_0pglobalFrameView_FormularSelectElement_options) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pglobalFrameView_FormularSelectElement_options;
#else
list = (_TPPLST_0pglobalFrameView_FormularSelectElement_options) TreeNodeAlloc (sizeof (struct _TPLST_0pglobalFrameView_FormularSelectElement_options));
#endif
list->_prod = RULELST_0pglobalFrameView_FormularSelectElement_options;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pglobalFrameView_FormularSelectElement_options)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_globalFrameView_FormularConstantSelectOptionpglobalFrameView_FormularSelectElement_options)
	{_TPPLST_globalFrameView_FormularConstantSelectOptionpglobalFrameView_FormularSelectElement_options _currn = _currn;
	 _TERMACT_LST_globalFrameView_FormularConstantSelectOptionpglobalFrameView_FormularSelectElement_options;
	} this code has been moved to _ElempglobalFrameView_FormularSelectElement_options*/
	/*if (_currn->_prod == RULELST_globalFrameView_FormularDataSourceSelectOptionpglobalFrameView_FormularSelectElement_options)
	{_TPPLST_globalFrameView_FormularDataSourceSelectOptionpglobalFrameView_FormularSelectElement_options _currn = _currn;
	 _TERMACT_LST_globalFrameView_FormularDataSourceSelectOptionpglobalFrameView_FormularSelectElement_options;
	} this code has been moved to _ElempglobalFrameView_FormularSelectElement_options*/
	/*if (_currn->_prod == RULELST_globalFrameView_FormularPHPScriptSelectOptionpglobalFrameView_FormularSelectElement_options)
	{_TPPLST_globalFrameView_FormularPHPScriptSelectOptionpglobalFrameView_FormularSelectElement_options _currn = _currn;
	 _TERMACT_LST_globalFrameView_FormularPHPScriptSelectOptionpglobalFrameView_FormularSelectElement_options;
	} this code has been moved to _ElempglobalFrameView_FormularSelectElement_options*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpglobalFrameView_FormularSelectElement_options;
#else
root = (_TPPpglobalFrameView_FormularSelectElement_options) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularSelectElement_options));
#endif
root->_prod = RULEpglobalFrameView_FormularSelectElement_options;
root->_desc1 = (_TSPLST_globalFrameView_FormularSelectElement_options)list;
{ _TPPpglobalFrameView_FormularSelectElement_options _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pglobalFrameView_FormularSelectElement_options;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularSelectElement_description (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularSelectElement_description (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularSelectElement_description _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularSelectElement_description;
#else
_currn = (_TPPpglobalFrameView_FormularSelectElement_description) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularSelectElement_description));
#endif
_currn->_prod = RULEpglobalFrameView_FormularSelectElement_description;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularSelectElement_description;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularSelectElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularRadioElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpglobalFrameView_FormularRadioElement (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpglobalFrameView_FormularRadioElement _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularRadioElement;
#else
_currn = (_TPPpglobalFrameView_FormularRadioElement) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularRadioElement));
#endif
_currn->_prod = RULEpglobalFrameView_FormularRadioElement;
_currn->_desc1 = (_TSPglobalFrameView_FormularRadioElement_type) MkglobalFrameView_FormularRadioElement_type (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularRadioElement: root of subtree no. 1 can not be made a globalFrameView_FormularRadioElement_type node ", 0, _coordref);
_currn->_desc2 = (_TSPglobalFrameView_FormularRadioElement_label) MkglobalFrameView_FormularRadioElement_label (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularRadioElement: root of subtree no. 2 can not be made a globalFrameView_FormularRadioElement_label node ", 0, _coordref);
_currn->_desc3 = (_TSPglobalFrameView_FormularRadioElement_name) MkglobalFrameView_FormularRadioElement_name (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularRadioElement: root of subtree no. 3 can not be made a globalFrameView_FormularRadioElement_name node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularRadioElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularRadioElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularRadioElement_type (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularRadioElement_type (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularRadioElement_type _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularRadioElement_type;
#else
_currn = (_TPPpglobalFrameView_FormularRadioElement_type) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularRadioElement_type));
#endif
_currn->_prod = RULEpglobalFrameView_FormularRadioElement_type;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularRadioElement_type;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularRadioElement_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularRadioElement_label (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularRadioElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularRadioElement_label _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularRadioElement_label;
#else
_currn = (_TPPpglobalFrameView_FormularRadioElement_label) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularRadioElement_label));
#endif
_currn->_prod = RULEpglobalFrameView_FormularRadioElement_label;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularRadioElement_label;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularRadioElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularRadioElement_name (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularRadioElement_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularRadioElement_name _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularRadioElement_name;
#else
_currn = (_TPPpglobalFrameView_FormularRadioElement_name) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularRadioElement_name));
#endif
_currn->_prod = RULEpglobalFrameView_FormularRadioElement_name;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularRadioElement_name;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularRadioElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_DataSourceActualParam (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpglobalFrameView_DataSourceActualParam (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpglobalFrameView_DataSourceActualParam _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_DataSourceActualParam;
#else
_currn = (_TPPpglobalFrameView_DataSourceActualParam) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_DataSourceActualParam));
#endif
_currn->_prod = RULEpglobalFrameView_DataSourceActualParam;
_currn->_desc1 = (_TSPglobalFrameView_DataSourceActualParam_formalParam) MkglobalFrameView_DataSourceActualParam_formalParam (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_DataSourceActualParam: root of subtree no. 1 can not be made a globalFrameView_DataSourceActualParam_formalParam node ", 0, _coordref);
_currn->_desc2 = (_TSPglobalFrameView_DataSourceActualParam_keyword1) MkglobalFrameView_DataSourceActualParam_keyword1 (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_DataSourceActualParam: root of subtree no. 2 can not be made a globalFrameView_DataSourceActualParam_keyword1 node ", 0, _coordref);
_currn->_desc3 = (_TSPglobalFrameView_DataSourceActualParam_expressions) MkglobalFrameView_DataSourceActualParam_expressions (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_DataSourceActualParam: root of subtree no. 3 can not be made a globalFrameView_DataSourceActualParam_expressions node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_DataSourceActualParam;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_DataSourceActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_DataSourceActualParam_formalParam (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_DataSourceActualParam_formalParam (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_DataSourceActualParam_formalParam _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_DataSourceActualParam_formalParam;
#else
_currn = (_TPPpglobalFrameView_DataSourceActualParam_formalParam) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_DataSourceActualParam_formalParam));
#endif
_currn->_prod = RULEpglobalFrameView_DataSourceActualParam_formalParam;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_DataSourceActualParam_formalParam;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_DataSourceActualParam_formalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_DataSourceActualParam_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_DataSourceActualParam_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_DataSourceActualParam_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_DataSourceActualParam_keyword1;
#else
_currn = (_TPPpglobalFrameView_DataSourceActualParam_keyword1) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_DataSourceActualParam_keyword1));
#endif
_currn->_prod = RULEpglobalFrameView_DataSourceActualParam_keyword1;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_DataSourceActualParam_keyword1;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_DataSourceActualParam_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_WebNodeConstantExpressionpglobalFrameView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_WebNodeConstantExpressionpglobalFrameView_DataSourceActualParam_expressions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_WebNodeConstantExpressionpglobalFrameView_DataSourceActualParam_expressions _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_WebNodeConstantExpressionpglobalFrameView_DataSourceActualParam_expressions;
#else
	_new = (_TPPLST_globalFrameView_WebNodeConstantExpressionpglobalFrameView_DataSourceActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_WebNodeConstantExpressionpglobalFrameView_DataSourceActualParam_expressions));
#endif
	_new->_prod = RULELST_globalFrameView_WebNodeConstantExpressionpglobalFrameView_DataSourceActualParam_expressions;
	_new->_desc1 = (_TSPglobalFrameView_WebNodeConstantExpression)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_DataSourceActualParam_expressions)_new;
	{_TPPLST_globalFrameView_WebNodeConstantExpressionpglobalFrameView_DataSourceActualParam_expressions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_WebNodeConstantExpressionpglobalFrameView_DataSourceActualParam_expressions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_WebNodeVariableExpressionpglobalFrameView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_WebNodeVariableExpressionpglobalFrameView_DataSourceActualParam_expressions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_WebNodeVariableExpressionpglobalFrameView_DataSourceActualParam_expressions _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_WebNodeVariableExpressionpglobalFrameView_DataSourceActualParam_expressions;
#else
	_new = (_TPPLST_globalFrameView_WebNodeVariableExpressionpglobalFrameView_DataSourceActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_WebNodeVariableExpressionpglobalFrameView_DataSourceActualParam_expressions));
#endif
	_new->_prod = RULELST_globalFrameView_WebNodeVariableExpressionpglobalFrameView_DataSourceActualParam_expressions;
	_new->_desc1 = (_TSPglobalFrameView_WebNodeVariableExpression)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_DataSourceActualParam_expressions)_new;
	{_TPPLST_globalFrameView_WebNodeVariableExpressionpglobalFrameView_DataSourceActualParam_expressions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_WebNodeVariableExpressionpglobalFrameView_DataSourceActualParam_expressions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_WebNodeColumnExpressionpglobalFrameView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_WebNodeColumnExpressionpglobalFrameView_DataSourceActualParam_expressions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_WebNodeColumnExpressionpglobalFrameView_DataSourceActualParam_expressions _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_WebNodeColumnExpressionpglobalFrameView_DataSourceActualParam_expressions;
#else
	_new = (_TPPLST_globalFrameView_WebNodeColumnExpressionpglobalFrameView_DataSourceActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_WebNodeColumnExpressionpglobalFrameView_DataSourceActualParam_expressions));
#endif
	_new->_prod = RULELST_globalFrameView_WebNodeColumnExpressionpglobalFrameView_DataSourceActualParam_expressions;
	_new->_desc1 = (_TSPglobalFrameView_WebNodeColumnExpression)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_DataSourceActualParam_expressions)_new;
	{_TPPLST_globalFrameView_WebNodeColumnExpressionpglobalFrameView_DataSourceActualParam_expressions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_WebNodeColumnExpressionpglobalFrameView_DataSourceActualParam_expressions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempglobalFrameView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempglobalFrameView_DataSourceActualParam_expressions (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBglobalFrameView_WebNodeConstantExpression)) 
		return _MkglobalFrameView_WebNodeConstantExpressionpglobalFrameView_DataSourceActualParam_expressions (_coordref, MkglobalFrameView_WebNodeConstantExpression (_coordref, t));
	if (IsSymb (t, SYMBglobalFrameView_WebNodeVariableExpression)) 
		return _MkglobalFrameView_WebNodeVariableExpressionpglobalFrameView_DataSourceActualParam_expressions (_coordref, MkglobalFrameView_WebNodeVariableExpression (_coordref, t));
	if (IsSymb (t, SYMBglobalFrameView_WebNodeColumnExpression)) 
		return _MkglobalFrameView_WebNodeColumnExpressionpglobalFrameView_DataSourceActualParam_expressions (_coordref, MkglobalFrameView_WebNodeColumnExpression (_coordref, t));
	message (DEADLY, "a pglobalFrameView_DataSourceActualParam_expressions list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pglobalFrameView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pglobalFrameView_DataSourceActualParam_expressions (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_globalFrameView_DataSourceActualParam_expressions)
	l =_ElempglobalFrameView_DataSourceActualParam_expressions (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_globalFrameView_DataSourceActualParam_expressions)
	r = _ElempglobalFrameView_DataSourceActualParam_expressions (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pglobalFrameView_DataSourceActualParam_expressions)r)->_desc2);
((_TPPLST_0pglobalFrameView_DataSourceActualParam_expressions)r)->_desc2 = ((_TPPLST_0pglobalFrameView_DataSourceActualParam_expressions)l)->_desc2;
((_TPPLST_0pglobalFrameView_DataSourceActualParam_expressions)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpglobalFrameView_DataSourceActualParam_expressions (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpglobalFrameView_DataSourceActualParam_expressions root;
	_TPPLST_0pglobalFrameView_DataSourceActualParam_expressions list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pglobalFrameView_DataSourceActualParam_expressions)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_globalFrameView_DataSourceActualParam_expressions)
		t = _ElempglobalFrameView_DataSourceActualParam_expressions (_coordref, t);
	remain = (_TPPLST_0pglobalFrameView_DataSourceActualParam_expressions)(((_TPPLST_0pglobalFrameView_DataSourceActualParam_expressions) t)->_desc2);
	((_TPPLST_0pglobalFrameView_DataSourceActualParam_expressions) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pglobalFrameView_DataSourceActualParam_expressions;
#else
list = (_TPPLST_0pglobalFrameView_DataSourceActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_0pglobalFrameView_DataSourceActualParam_expressions));
#endif
list->_prod = RULELST_0pglobalFrameView_DataSourceActualParam_expressions;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pglobalFrameView_DataSourceActualParam_expressions)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_globalFrameView_WebNodeConstantExpressionpglobalFrameView_DataSourceActualParam_expressions)
	{_TPPLST_globalFrameView_WebNodeConstantExpressionpglobalFrameView_DataSourceActualParam_expressions _currn = _currn;
	 _TERMACT_LST_globalFrameView_WebNodeConstantExpressionpglobalFrameView_DataSourceActualParam_expressions;
	} this code has been moved to _ElempglobalFrameView_DataSourceActualParam_expressions*/
	/*if (_currn->_prod == RULELST_globalFrameView_WebNodeVariableExpressionpglobalFrameView_DataSourceActualParam_expressions)
	{_TPPLST_globalFrameView_WebNodeVariableExpressionpglobalFrameView_DataSourceActualParam_expressions _currn = _currn;
	 _TERMACT_LST_globalFrameView_WebNodeVariableExpressionpglobalFrameView_DataSourceActualParam_expressions;
	} this code has been moved to _ElempglobalFrameView_DataSourceActualParam_expressions*/
	/*if (_currn->_prod == RULELST_globalFrameView_WebNodeColumnExpressionpglobalFrameView_DataSourceActualParam_expressions)
	{_TPPLST_globalFrameView_WebNodeColumnExpressionpglobalFrameView_DataSourceActualParam_expressions _currn = _currn;
	 _TERMACT_LST_globalFrameView_WebNodeColumnExpressionpglobalFrameView_DataSourceActualParam_expressions;
	} this code has been moved to _ElempglobalFrameView_DataSourceActualParam_expressions*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpglobalFrameView_DataSourceActualParam_expressions;
#else
root = (_TPPpglobalFrameView_DataSourceActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_DataSourceActualParam_expressions));
#endif
root->_prod = RULEpglobalFrameView_DataSourceActualParam_expressions;
root->_desc1 = (_TSPLST_globalFrameView_DataSourceActualParam_expressions)list;
{ _TPPpglobalFrameView_DataSourceActualParam_expressions _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pglobalFrameView_DataSourceActualParam_expressions;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_WebNodeActualParam (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpglobalFrameView_WebNodeActualParam (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpglobalFrameView_WebNodeActualParam _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_WebNodeActualParam;
#else
_currn = (_TPPpglobalFrameView_WebNodeActualParam) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_WebNodeActualParam));
#endif
_currn->_prod = RULEpglobalFrameView_WebNodeActualParam;
_currn->_desc1 = (_TSPglobalFrameView_WebNodeActualParam_formalParam) MkglobalFrameView_WebNodeActualParam_formalParam (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_WebNodeActualParam: root of subtree no. 1 can not be made a globalFrameView_WebNodeActualParam_formalParam node ", 0, _coordref);
_currn->_desc2 = (_TSPglobalFrameView_WebNodeActualParam_keyword1) MkglobalFrameView_WebNodeActualParam_keyword1 (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_WebNodeActualParam: root of subtree no. 2 can not be made a globalFrameView_WebNodeActualParam_keyword1 node ", 0, _coordref);
_currn->_desc3 = (_TSPglobalFrameView_WebNodeActualParam_expressions) MkglobalFrameView_WebNodeActualParam_expressions (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_WebNodeActualParam: root of subtree no. 3 can not be made a globalFrameView_WebNodeActualParam_expressions node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_WebNodeActualParam;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_WebNodeActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_WebNodeActualParam_formalParam (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_WebNodeActualParam_formalParam (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_WebNodeActualParam_formalParam _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_WebNodeActualParam_formalParam;
#else
_currn = (_TPPpglobalFrameView_WebNodeActualParam_formalParam) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_WebNodeActualParam_formalParam));
#endif
_currn->_prod = RULEpglobalFrameView_WebNodeActualParam_formalParam;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_WebNodeActualParam_formalParam;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_WebNodeActualParam_formalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_WebNodeActualParam_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_WebNodeActualParam_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_WebNodeActualParam_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_WebNodeActualParam_keyword1;
#else
_currn = (_TPPpglobalFrameView_WebNodeActualParam_keyword1) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_WebNodeActualParam_keyword1));
#endif
_currn->_prod = RULEpglobalFrameView_WebNodeActualParam_keyword1;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_WebNodeActualParam_keyword1;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_WebNodeActualParam_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_WebNodeConstantExpressionpglobalFrameView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_WebNodeConstantExpressionpglobalFrameView_WebNodeActualParam_expressions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_WebNodeConstantExpressionpglobalFrameView_WebNodeActualParam_expressions _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_WebNodeConstantExpressionpglobalFrameView_WebNodeActualParam_expressions;
#else
	_new = (_TPPLST_globalFrameView_WebNodeConstantExpressionpglobalFrameView_WebNodeActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_WebNodeConstantExpressionpglobalFrameView_WebNodeActualParam_expressions));
#endif
	_new->_prod = RULELST_globalFrameView_WebNodeConstantExpressionpglobalFrameView_WebNodeActualParam_expressions;
	_new->_desc1 = (_TSPglobalFrameView_WebNodeConstantExpression)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_WebNodeActualParam_expressions)_new;
	{_TPPLST_globalFrameView_WebNodeConstantExpressionpglobalFrameView_WebNodeActualParam_expressions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_WebNodeConstantExpressionpglobalFrameView_WebNodeActualParam_expressions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_WebNodeVariableExpressionpglobalFrameView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_WebNodeVariableExpressionpglobalFrameView_WebNodeActualParam_expressions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_WebNodeVariableExpressionpglobalFrameView_WebNodeActualParam_expressions _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_WebNodeVariableExpressionpglobalFrameView_WebNodeActualParam_expressions;
#else
	_new = (_TPPLST_globalFrameView_WebNodeVariableExpressionpglobalFrameView_WebNodeActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_WebNodeVariableExpressionpglobalFrameView_WebNodeActualParam_expressions));
#endif
	_new->_prod = RULELST_globalFrameView_WebNodeVariableExpressionpglobalFrameView_WebNodeActualParam_expressions;
	_new->_desc1 = (_TSPglobalFrameView_WebNodeVariableExpression)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_WebNodeActualParam_expressions)_new;
	{_TPPLST_globalFrameView_WebNodeVariableExpressionpglobalFrameView_WebNodeActualParam_expressions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_WebNodeVariableExpressionpglobalFrameView_WebNodeActualParam_expressions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_WebNodeColumnExpressionpglobalFrameView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_WebNodeColumnExpressionpglobalFrameView_WebNodeActualParam_expressions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_WebNodeColumnExpressionpglobalFrameView_WebNodeActualParam_expressions _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_WebNodeColumnExpressionpglobalFrameView_WebNodeActualParam_expressions;
#else
	_new = (_TPPLST_globalFrameView_WebNodeColumnExpressionpglobalFrameView_WebNodeActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_WebNodeColumnExpressionpglobalFrameView_WebNodeActualParam_expressions));
#endif
	_new->_prod = RULELST_globalFrameView_WebNodeColumnExpressionpglobalFrameView_WebNodeActualParam_expressions;
	_new->_desc1 = (_TSPglobalFrameView_WebNodeColumnExpression)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_WebNodeActualParam_expressions)_new;
	{_TPPLST_globalFrameView_WebNodeColumnExpressionpglobalFrameView_WebNodeActualParam_expressions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_WebNodeColumnExpressionpglobalFrameView_WebNodeActualParam_expressions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempglobalFrameView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempglobalFrameView_WebNodeActualParam_expressions (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBglobalFrameView_WebNodeConstantExpression)) 
		return _MkglobalFrameView_WebNodeConstantExpressionpglobalFrameView_WebNodeActualParam_expressions (_coordref, MkglobalFrameView_WebNodeConstantExpression (_coordref, t));
	if (IsSymb (t, SYMBglobalFrameView_WebNodeVariableExpression)) 
		return _MkglobalFrameView_WebNodeVariableExpressionpglobalFrameView_WebNodeActualParam_expressions (_coordref, MkglobalFrameView_WebNodeVariableExpression (_coordref, t));
	if (IsSymb (t, SYMBglobalFrameView_WebNodeColumnExpression)) 
		return _MkglobalFrameView_WebNodeColumnExpressionpglobalFrameView_WebNodeActualParam_expressions (_coordref, MkglobalFrameView_WebNodeColumnExpression (_coordref, t));
	message (DEADLY, "a pglobalFrameView_WebNodeActualParam_expressions list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pglobalFrameView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pglobalFrameView_WebNodeActualParam_expressions (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_globalFrameView_WebNodeActualParam_expressions)
	l =_ElempglobalFrameView_WebNodeActualParam_expressions (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_globalFrameView_WebNodeActualParam_expressions)
	r = _ElempglobalFrameView_WebNodeActualParam_expressions (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pglobalFrameView_WebNodeActualParam_expressions)r)->_desc2);
((_TPPLST_0pglobalFrameView_WebNodeActualParam_expressions)r)->_desc2 = ((_TPPLST_0pglobalFrameView_WebNodeActualParam_expressions)l)->_desc2;
((_TPPLST_0pglobalFrameView_WebNodeActualParam_expressions)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpglobalFrameView_WebNodeActualParam_expressions (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpglobalFrameView_WebNodeActualParam_expressions root;
	_TPPLST_0pglobalFrameView_WebNodeActualParam_expressions list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pglobalFrameView_WebNodeActualParam_expressions)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_globalFrameView_WebNodeActualParam_expressions)
		t = _ElempglobalFrameView_WebNodeActualParam_expressions (_coordref, t);
	remain = (_TPPLST_0pglobalFrameView_WebNodeActualParam_expressions)(((_TPPLST_0pglobalFrameView_WebNodeActualParam_expressions) t)->_desc2);
	((_TPPLST_0pglobalFrameView_WebNodeActualParam_expressions) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pglobalFrameView_WebNodeActualParam_expressions;
#else
list = (_TPPLST_0pglobalFrameView_WebNodeActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_0pglobalFrameView_WebNodeActualParam_expressions));
#endif
list->_prod = RULELST_0pglobalFrameView_WebNodeActualParam_expressions;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pglobalFrameView_WebNodeActualParam_expressions)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_globalFrameView_WebNodeConstantExpressionpglobalFrameView_WebNodeActualParam_expressions)
	{_TPPLST_globalFrameView_WebNodeConstantExpressionpglobalFrameView_WebNodeActualParam_expressions _currn = _currn;
	 _TERMACT_LST_globalFrameView_WebNodeConstantExpressionpglobalFrameView_WebNodeActualParam_expressions;
	} this code has been moved to _ElempglobalFrameView_WebNodeActualParam_expressions*/
	/*if (_currn->_prod == RULELST_globalFrameView_WebNodeVariableExpressionpglobalFrameView_WebNodeActualParam_expressions)
	{_TPPLST_globalFrameView_WebNodeVariableExpressionpglobalFrameView_WebNodeActualParam_expressions _currn = _currn;
	 _TERMACT_LST_globalFrameView_WebNodeVariableExpressionpglobalFrameView_WebNodeActualParam_expressions;
	} this code has been moved to _ElempglobalFrameView_WebNodeActualParam_expressions*/
	/*if (_currn->_prod == RULELST_globalFrameView_WebNodeColumnExpressionpglobalFrameView_WebNodeActualParam_expressions)
	{_TPPLST_globalFrameView_WebNodeColumnExpressionpglobalFrameView_WebNodeActualParam_expressions _currn = _currn;
	 _TERMACT_LST_globalFrameView_WebNodeColumnExpressionpglobalFrameView_WebNodeActualParam_expressions;
	} this code has been moved to _ElempglobalFrameView_WebNodeActualParam_expressions*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpglobalFrameView_WebNodeActualParam_expressions;
#else
root = (_TPPpglobalFrameView_WebNodeActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_WebNodeActualParam_expressions));
#endif
root->_prod = RULEpglobalFrameView_WebNodeActualParam_expressions;
root->_desc1 = (_TSPLST_globalFrameView_WebNodeActualParam_expressions)list;
{ _TPPpglobalFrameView_WebNodeActualParam_expressions _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pglobalFrameView_WebNodeActualParam_expressions;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_WebNodeConstantExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpglobalFrameView_WebNodeConstantExpression (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpglobalFrameView_WebNodeConstantExpression _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_WebNodeConstantExpression;
#else
_currn = (_TPPpglobalFrameView_WebNodeConstantExpression) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_WebNodeConstantExpression));
#endif
_currn->_prod = RULEpglobalFrameView_WebNodeConstantExpression;
_currn->_desc1 = (_TSPglobalFrameView_WebNodeConstantExpression_keyword1) MkglobalFrameView_WebNodeConstantExpression_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_WebNodeConstantExpression: root of subtree no. 1 can not be made a globalFrameView_WebNodeConstantExpression_keyword1 node ", 0, _coordref);
_currn->_desc2 = (_TSPglobalFrameView_WebNodeConstantExpression_value) MkglobalFrameView_WebNodeConstantExpression_value (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_WebNodeConstantExpression: root of subtree no. 2 can not be made a globalFrameView_WebNodeConstantExpression_value node ", 0, _coordref);
_currn->_desc3 = (_TSPglobalFrameView_WebNodeConstantExpression_keyword2) MkglobalFrameView_WebNodeConstantExpression_keyword2 (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_WebNodeConstantExpression: root of subtree no. 3 can not be made a globalFrameView_WebNodeConstantExpression_keyword2 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_WebNodeConstantExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_WebNodeConstantExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_WebNodeConstantExpression_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_WebNodeConstantExpression_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_WebNodeConstantExpression_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_WebNodeConstantExpression_keyword1;
#else
_currn = (_TPPpglobalFrameView_WebNodeConstantExpression_keyword1) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_WebNodeConstantExpression_keyword1));
#endif
_currn->_prod = RULEpglobalFrameView_WebNodeConstantExpression_keyword1;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_WebNodeConstantExpression_keyword1;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_WebNodeConstantExpression_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_WebNodeConstantExpression_value (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_WebNodeConstantExpression_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_WebNodeConstantExpression_value _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_WebNodeConstantExpression_value;
#else
_currn = (_TPPpglobalFrameView_WebNodeConstantExpression_value) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_WebNodeConstantExpression_value));
#endif
_currn->_prod = RULEpglobalFrameView_WebNodeConstantExpression_value;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_WebNodeConstantExpression_value;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_WebNodeConstantExpression_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_WebNodeConstantExpression_keyword2 (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_WebNodeConstantExpression_keyword2 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_WebNodeConstantExpression_keyword2 _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_WebNodeConstantExpression_keyword2;
#else
_currn = (_TPPpglobalFrameView_WebNodeConstantExpression_keyword2) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_WebNodeConstantExpression_keyword2));
#endif
_currn->_prod = RULEpglobalFrameView_WebNodeConstantExpression_keyword2;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_WebNodeConstantExpression_keyword2;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_WebNodeConstantExpression_keyword2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_WebNodeVariableExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpglobalFrameView_WebNodeVariableExpression (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpglobalFrameView_WebNodeVariableExpression _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_WebNodeVariableExpression;
#else
_currn = (_TPPpglobalFrameView_WebNodeVariableExpression) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_WebNodeVariableExpression));
#endif
_currn->_prod = RULEpglobalFrameView_WebNodeVariableExpression;
_currn->_desc1 = (_TSPglobalFrameView_WebNodeVariableExpression_content) MkglobalFrameView_WebNodeVariableExpression_content (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_WebNodeVariableExpression: root of subtree no. 1 can not be made a globalFrameView_WebNodeVariableExpression_content node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_WebNodeVariableExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_WebNodeVariableExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_WebNodeVariableExpression_content (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_WebNodeVariableExpression_content (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_WebNodeVariableExpression_content _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_WebNodeVariableExpression_content;
#else
_currn = (_TPPpglobalFrameView_WebNodeVariableExpression_content) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_WebNodeVariableExpression_content));
#endif
_currn->_prod = RULEpglobalFrameView_WebNodeVariableExpression_content;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_WebNodeVariableExpression_content;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_WebNodeVariableExpression_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_WebNodeColumnExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpglobalFrameView_WebNodeColumnExpression (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpglobalFrameView_WebNodeColumnExpression _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_WebNodeColumnExpression;
#else
_currn = (_TPPpglobalFrameView_WebNodeColumnExpression) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_WebNodeColumnExpression));
#endif
_currn->_prod = RULEpglobalFrameView_WebNodeColumnExpression;
_currn->_desc1 = (_TSPglobalFrameView_WebNodeColumnExpression_content) MkglobalFrameView_WebNodeColumnExpression_content (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_WebNodeColumnExpression: root of subtree no. 1 can not be made a globalFrameView_WebNodeColumnExpression_content node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_WebNodeColumnExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_WebNodeColumnExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_WebNodeColumnExpression_content (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_WebNodeColumnExpression_content (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_WebNodeColumnExpression_content _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_WebNodeColumnExpression_content;
#else
_currn = (_TPPpglobalFrameView_WebNodeColumnExpression_content) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_WebNodeColumnExpression_content));
#endif
_currn->_prod = RULEpglobalFrameView_WebNodeColumnExpression_content;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_WebNodeColumnExpression_content;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_WebNodeColumnExpression_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_Formular (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkpglobalFrameView_Formular (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPpglobalFrameView_Formular _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_Formular;
#else
_currn = (_TPPpglobalFrameView_Formular) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_Formular));
#endif
_currn->_prod = RULEpglobalFrameView_Formular;
_currn->_desc1 = (_TSPglobalFrameView_Formular_styleSheetBlock) MkglobalFrameView_Formular_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_Formular: root of subtree no. 1 can not be made a globalFrameView_Formular_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSPglobalFrameView_Formular_action) MkglobalFrameView_Formular_action (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_Formular: root of subtree no. 2 can not be made a globalFrameView_Formular_action node ", 0, _coordref);
_currn->_desc3 = (_TSPglobalFrameView_Formular_elements) MkglobalFrameView_Formular_elements (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_Formular: root of subtree no. 3 can not be made a globalFrameView_Formular_elements node ", 0, _coordref);
_currn->_desc4 = (_TSPglobalFrameView_Formular_datasource) MkglobalFrameView_Formular_datasource (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_Formular: root of subtree no. 4 can not be made a globalFrameView_Formular_datasource node ", 0, _coordref);
_currn->_desc5 = (_TSPglobalFrameView_Formular_actualParams) MkglobalFrameView_Formular_actualParams (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_Formular: root of subtree no. 5 can not be made a globalFrameView_Formular_actualParams node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_Formular;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_Formular */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_Formular_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_Formular_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_Formular_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_Formular_styleSheetBlock;
#else
_currn = (_TPPpglobalFrameView_Formular_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_Formular_styleSheetBlock));
#endif
_currn->_prod = RULEpglobalFrameView_Formular_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_Formular_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_Formular_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_Formular_action (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_Formular_action (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_Formular_action _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_Formular_action;
#else
_currn = (_TPPpglobalFrameView_Formular_action) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_Formular_action));
#endif
_currn->_prod = RULEpglobalFrameView_Formular_action;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_Formular_action;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_Formular_action */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_FormularRadioGroupElementpglobalFrameView_Formular_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_FormularRadioGroupElementpglobalFrameView_Formular_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_FormularRadioGroupElementpglobalFrameView_Formular_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_FormularRadioGroupElementpglobalFrameView_Formular_elements;
#else
	_new = (_TPPLST_globalFrameView_FormularRadioGroupElementpglobalFrameView_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_FormularRadioGroupElementpglobalFrameView_Formular_elements));
#endif
	_new->_prod = RULELST_globalFrameView_FormularRadioGroupElementpglobalFrameView_Formular_elements;
	_new->_desc1 = (_TSPglobalFrameView_FormularRadioGroupElement)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_Formular_elements)_new;
	{_TPPLST_globalFrameView_FormularRadioGroupElementpglobalFrameView_Formular_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_FormularRadioGroupElementpglobalFrameView_Formular_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_FormularTextElementpglobalFrameView_Formular_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_FormularTextElementpglobalFrameView_Formular_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_FormularTextElementpglobalFrameView_Formular_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_FormularTextElementpglobalFrameView_Formular_elements;
#else
	_new = (_TPPLST_globalFrameView_FormularTextElementpglobalFrameView_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_FormularTextElementpglobalFrameView_Formular_elements));
#endif
	_new->_prod = RULELST_globalFrameView_FormularTextElementpglobalFrameView_Formular_elements;
	_new->_desc1 = (_TSPglobalFrameView_FormularTextElement)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_Formular_elements)_new;
	{_TPPLST_globalFrameView_FormularTextElementpglobalFrameView_Formular_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_FormularTextElementpglobalFrameView_Formular_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_FormularTextAreaElementpglobalFrameView_Formular_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_FormularTextAreaElementpglobalFrameView_Formular_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_FormularTextAreaElementpglobalFrameView_Formular_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_FormularTextAreaElementpglobalFrameView_Formular_elements;
#else
	_new = (_TPPLST_globalFrameView_FormularTextAreaElementpglobalFrameView_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_FormularTextAreaElementpglobalFrameView_Formular_elements));
#endif
	_new->_prod = RULELST_globalFrameView_FormularTextAreaElementpglobalFrameView_Formular_elements;
	_new->_desc1 = (_TSPglobalFrameView_FormularTextAreaElement)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_Formular_elements)_new;
	{_TPPLST_globalFrameView_FormularTextAreaElementpglobalFrameView_Formular_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_FormularTextAreaElementpglobalFrameView_Formular_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_FormularPasswordElementpglobalFrameView_Formular_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_FormularPasswordElementpglobalFrameView_Formular_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_FormularPasswordElementpglobalFrameView_Formular_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_FormularPasswordElementpglobalFrameView_Formular_elements;
#else
	_new = (_TPPLST_globalFrameView_FormularPasswordElementpglobalFrameView_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_FormularPasswordElementpglobalFrameView_Formular_elements));
#endif
	_new->_prod = RULELST_globalFrameView_FormularPasswordElementpglobalFrameView_Formular_elements;
	_new->_desc1 = (_TSPglobalFrameView_FormularPasswordElement)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_Formular_elements)_new;
	{_TPPLST_globalFrameView_FormularPasswordElementpglobalFrameView_Formular_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_FormularPasswordElementpglobalFrameView_Formular_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_FormularCheckboxElementpglobalFrameView_Formular_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_FormularCheckboxElementpglobalFrameView_Formular_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_FormularCheckboxElementpglobalFrameView_Formular_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_FormularCheckboxElementpglobalFrameView_Formular_elements;
#else
	_new = (_TPPLST_globalFrameView_FormularCheckboxElementpglobalFrameView_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_FormularCheckboxElementpglobalFrameView_Formular_elements));
#endif
	_new->_prod = RULELST_globalFrameView_FormularCheckboxElementpglobalFrameView_Formular_elements;
	_new->_desc1 = (_TSPglobalFrameView_FormularCheckboxElement)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_Formular_elements)_new;
	{_TPPLST_globalFrameView_FormularCheckboxElementpglobalFrameView_Formular_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_FormularCheckboxElementpglobalFrameView_Formular_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_FormularSelectElementpglobalFrameView_Formular_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_FormularSelectElementpglobalFrameView_Formular_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_FormularSelectElementpglobalFrameView_Formular_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_FormularSelectElementpglobalFrameView_Formular_elements;
#else
	_new = (_TPPLST_globalFrameView_FormularSelectElementpglobalFrameView_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_FormularSelectElementpglobalFrameView_Formular_elements));
#endif
	_new->_prod = RULELST_globalFrameView_FormularSelectElementpglobalFrameView_Formular_elements;
	_new->_desc1 = (_TSPglobalFrameView_FormularSelectElement)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_Formular_elements)_new;
	{_TPPLST_globalFrameView_FormularSelectElementpglobalFrameView_Formular_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_FormularSelectElementpglobalFrameView_Formular_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempglobalFrameView_Formular_elements (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempglobalFrameView_Formular_elements (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBglobalFrameView_FormularRadioGroupElement)) 
		return _MkglobalFrameView_FormularRadioGroupElementpglobalFrameView_Formular_elements (_coordref, MkglobalFrameView_FormularRadioGroupElement (_coordref, t));
	if (IsSymb (t, SYMBglobalFrameView_FormularTextElement)) 
		return _MkglobalFrameView_FormularTextElementpglobalFrameView_Formular_elements (_coordref, MkglobalFrameView_FormularTextElement (_coordref, t));
	if (IsSymb (t, SYMBglobalFrameView_FormularTextAreaElement)) 
		return _MkglobalFrameView_FormularTextAreaElementpglobalFrameView_Formular_elements (_coordref, MkglobalFrameView_FormularTextAreaElement (_coordref, t));
	if (IsSymb (t, SYMBglobalFrameView_FormularPasswordElement)) 
		return _MkglobalFrameView_FormularPasswordElementpglobalFrameView_Formular_elements (_coordref, MkglobalFrameView_FormularPasswordElement (_coordref, t));
	if (IsSymb (t, SYMBglobalFrameView_FormularCheckboxElement)) 
		return _MkglobalFrameView_FormularCheckboxElementpglobalFrameView_Formular_elements (_coordref, MkglobalFrameView_FormularCheckboxElement (_coordref, t));
	if (IsSymb (t, SYMBglobalFrameView_FormularSelectElement)) 
		return _MkglobalFrameView_FormularSelectElementpglobalFrameView_Formular_elements (_coordref, MkglobalFrameView_FormularSelectElement (_coordref, t));
	message (DEADLY, "a pglobalFrameView_Formular_elements list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pglobalFrameView_Formular_elements (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pglobalFrameView_Formular_elements (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_globalFrameView_Formular_elements)
	l =_ElempglobalFrameView_Formular_elements (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_globalFrameView_Formular_elements)
	r = _ElempglobalFrameView_Formular_elements (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pglobalFrameView_Formular_elements)r)->_desc2);
((_TPPLST_0pglobalFrameView_Formular_elements)r)->_desc2 = ((_TPPLST_0pglobalFrameView_Formular_elements)l)->_desc2;
((_TPPLST_0pglobalFrameView_Formular_elements)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_Formular_elements (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpglobalFrameView_Formular_elements (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpglobalFrameView_Formular_elements root;
	_TPPLST_0pglobalFrameView_Formular_elements list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pglobalFrameView_Formular_elements)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_globalFrameView_Formular_elements)
		t = _ElempglobalFrameView_Formular_elements (_coordref, t);
	remain = (_TPPLST_0pglobalFrameView_Formular_elements)(((_TPPLST_0pglobalFrameView_Formular_elements) t)->_desc2);
	((_TPPLST_0pglobalFrameView_Formular_elements) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pglobalFrameView_Formular_elements;
#else
list = (_TPPLST_0pglobalFrameView_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_0pglobalFrameView_Formular_elements));
#endif
list->_prod = RULELST_0pglobalFrameView_Formular_elements;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pglobalFrameView_Formular_elements)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_globalFrameView_FormularRadioGroupElementpglobalFrameView_Formular_elements)
	{_TPPLST_globalFrameView_FormularRadioGroupElementpglobalFrameView_Formular_elements _currn = _currn;
	 _TERMACT_LST_globalFrameView_FormularRadioGroupElementpglobalFrameView_Formular_elements;
	} this code has been moved to _ElempglobalFrameView_Formular_elements*/
	/*if (_currn->_prod == RULELST_globalFrameView_FormularTextElementpglobalFrameView_Formular_elements)
	{_TPPLST_globalFrameView_FormularTextElementpglobalFrameView_Formular_elements _currn = _currn;
	 _TERMACT_LST_globalFrameView_FormularTextElementpglobalFrameView_Formular_elements;
	} this code has been moved to _ElempglobalFrameView_Formular_elements*/
	/*if (_currn->_prod == RULELST_globalFrameView_FormularTextAreaElementpglobalFrameView_Formular_elements)
	{_TPPLST_globalFrameView_FormularTextAreaElementpglobalFrameView_Formular_elements _currn = _currn;
	 _TERMACT_LST_globalFrameView_FormularTextAreaElementpglobalFrameView_Formular_elements;
	} this code has been moved to _ElempglobalFrameView_Formular_elements*/
	/*if (_currn->_prod == RULELST_globalFrameView_FormularPasswordElementpglobalFrameView_Formular_elements)
	{_TPPLST_globalFrameView_FormularPasswordElementpglobalFrameView_Formular_elements _currn = _currn;
	 _TERMACT_LST_globalFrameView_FormularPasswordElementpglobalFrameView_Formular_elements;
	} this code has been moved to _ElempglobalFrameView_Formular_elements*/
	/*if (_currn->_prod == RULELST_globalFrameView_FormularCheckboxElementpglobalFrameView_Formular_elements)
	{_TPPLST_globalFrameView_FormularCheckboxElementpglobalFrameView_Formular_elements _currn = _currn;
	 _TERMACT_LST_globalFrameView_FormularCheckboxElementpglobalFrameView_Formular_elements;
	} this code has been moved to _ElempglobalFrameView_Formular_elements*/
	/*if (_currn->_prod == RULELST_globalFrameView_FormularSelectElementpglobalFrameView_Formular_elements)
	{_TPPLST_globalFrameView_FormularSelectElementpglobalFrameView_Formular_elements _currn = _currn;
	 _TERMACT_LST_globalFrameView_FormularSelectElementpglobalFrameView_Formular_elements;
	} this code has been moved to _ElempglobalFrameView_Formular_elements*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpglobalFrameView_Formular_elements;
#else
root = (_TPPpglobalFrameView_Formular_elements) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_Formular_elements));
#endif
root->_prod = RULEpglobalFrameView_Formular_elements;
root->_desc1 = (_TSPLST_globalFrameView_Formular_elements)list;
{ _TPPpglobalFrameView_Formular_elements _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pglobalFrameView_Formular_elements;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_Formular_datasource (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_Formular_datasource (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_Formular_datasource _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_Formular_datasource;
#else
_currn = (_TPPpglobalFrameView_Formular_datasource) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_Formular_datasource));
#endif
_currn->_prod = RULEpglobalFrameView_Formular_datasource;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_Formular_datasource;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_Formular_datasource */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_WebNodeActualParampglobalFrameView_Formular_actualParams (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_WebNodeActualParampglobalFrameView_Formular_actualParams (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_WebNodeActualParampglobalFrameView_Formular_actualParams _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_WebNodeActualParampglobalFrameView_Formular_actualParams;
#else
	_new = (_TPPLST_globalFrameView_WebNodeActualParampglobalFrameView_Formular_actualParams) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_WebNodeActualParampglobalFrameView_Formular_actualParams));
#endif
	_new->_prod = RULELST_globalFrameView_WebNodeActualParampglobalFrameView_Formular_actualParams;
	_new->_desc1 = (_TSPglobalFrameView_WebNodeActualParam)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_Formular_actualParams)_new;
	{_TPPLST_globalFrameView_WebNodeActualParampglobalFrameView_Formular_actualParams _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_WebNodeActualParampglobalFrameView_Formular_actualParams;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempglobalFrameView_Formular_actualParams (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempglobalFrameView_Formular_actualParams (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBglobalFrameView_WebNodeActualParam)) 
		return _MkglobalFrameView_WebNodeActualParampglobalFrameView_Formular_actualParams (_coordref, MkglobalFrameView_WebNodeActualParam (_coordref, t));
	message (DEADLY, "a pglobalFrameView_Formular_actualParams list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pglobalFrameView_Formular_actualParams (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pglobalFrameView_Formular_actualParams (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_globalFrameView_Formular_actualParams)
	l =_ElempglobalFrameView_Formular_actualParams (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_globalFrameView_Formular_actualParams)
	r = _ElempglobalFrameView_Formular_actualParams (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pglobalFrameView_Formular_actualParams)r)->_desc2);
((_TPPLST_0pglobalFrameView_Formular_actualParams)r)->_desc2 = ((_TPPLST_0pglobalFrameView_Formular_actualParams)l)->_desc2;
((_TPPLST_0pglobalFrameView_Formular_actualParams)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_Formular_actualParams (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpglobalFrameView_Formular_actualParams (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpglobalFrameView_Formular_actualParams root;
	_TPPLST_0pglobalFrameView_Formular_actualParams list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pglobalFrameView_Formular_actualParams)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_globalFrameView_Formular_actualParams)
		t = _ElempglobalFrameView_Formular_actualParams (_coordref, t);
	remain = (_TPPLST_0pglobalFrameView_Formular_actualParams)(((_TPPLST_0pglobalFrameView_Formular_actualParams) t)->_desc2);
	((_TPPLST_0pglobalFrameView_Formular_actualParams) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pglobalFrameView_Formular_actualParams;
#else
list = (_TPPLST_0pglobalFrameView_Formular_actualParams) TreeNodeAlloc (sizeof (struct _TPLST_0pglobalFrameView_Formular_actualParams));
#endif
list->_prod = RULELST_0pglobalFrameView_Formular_actualParams;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pglobalFrameView_Formular_actualParams)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_globalFrameView_WebNodeActualParampglobalFrameView_Formular_actualParams)
	{_TPPLST_globalFrameView_WebNodeActualParampglobalFrameView_Formular_actualParams _currn = _currn;
	 _TERMACT_LST_globalFrameView_WebNodeActualParampglobalFrameView_Formular_actualParams;
	} this code has been moved to _ElempglobalFrameView_Formular_actualParams*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpglobalFrameView_Formular_actualParams;
#else
root = (_TPPpglobalFrameView_Formular_actualParams) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_Formular_actualParams));
#endif
root->_prod = RULEpglobalFrameView_Formular_actualParams;
root->_desc1 = (_TSPLST_globalFrameView_Formular_actualParams)list;
{ _TPPpglobalFrameView_Formular_actualParams _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pglobalFrameView_Formular_actualParams;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularRadioGroupElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5, NODEPTR _desc6, NODEPTR _desc7)
#else
NODEPTR MkpglobalFrameView_FormularRadioGroupElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5,_desc6,_desc7)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
	NODEPTR _desc6;
	NODEPTR _desc7;
#endif
{	_TPPpglobalFrameView_FormularRadioGroupElement _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularRadioGroupElement;
#else
_currn = (_TPPpglobalFrameView_FormularRadioGroupElement) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularRadioGroupElement));
#endif
_currn->_prod = RULEpglobalFrameView_FormularRadioGroupElement;
_currn->_desc1 = (_TSPglobalFrameView_FormularRadioGroupElement_name) MkglobalFrameView_FormularRadioGroupElement_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularRadioGroupElement: root of subtree no. 1 can not be made a globalFrameView_FormularRadioGroupElement_name node ", 0, _coordref);
_currn->_desc2 = (_TSPglobalFrameView_FormularRadioGroupElement_label) MkglobalFrameView_FormularRadioGroupElement_label (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularRadioGroupElement: root of subtree no. 2 can not be made a globalFrameView_FormularRadioGroupElement_label node ", 0, _coordref);
_currn->_desc3 = (_TSPglobalFrameView_FormularRadioGroupElement_description) MkglobalFrameView_FormularRadioGroupElement_description (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularRadioGroupElement: root of subtree no. 3 can not be made a globalFrameView_FormularRadioGroupElement_description node ", 0, _coordref);
_currn->_desc4 = (_TSPglobalFrameView_FormularRadioGroupElement_pvalue) MkglobalFrameView_FormularRadioGroupElement_pvalue (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularRadioGroupElement: root of subtree no. 4 can not be made a globalFrameView_FormularRadioGroupElement_pvalue node ", 0, _coordref);
_currn->_desc5 = (_TSPglobalFrameView_FormularRadioGroupElement_condition) MkglobalFrameView_FormularRadioGroupElement_condition (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularRadioGroupElement: root of subtree no. 5 can not be made a globalFrameView_FormularRadioGroupElement_condition node ", 0, _coordref);
_currn->_desc6 = (_TSPglobalFrameView_FormularRadioGroupElement_elements) MkglobalFrameView_FormularRadioGroupElement_elements (_coordref, _desc6);	
if (((NODEPTR)_currn->_desc6) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularRadioGroupElement: root of subtree no. 6 can not be made a globalFrameView_FormularRadioGroupElement_elements node ", 0, _coordref);
_currn->_desc7 = (_TSPglobalFrameView_FormularRadioGroupElement_checked) MkglobalFrameView_FormularRadioGroupElement_checked (_coordref, _desc7);	
if (((NODEPTR)_currn->_desc7) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularRadioGroupElement: root of subtree no. 7 can not be made a globalFrameView_FormularRadioGroupElement_checked node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularRadioGroupElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularRadioGroupElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularRadioGroupElement_name (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularRadioGroupElement_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularRadioGroupElement_name _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularRadioGroupElement_name;
#else
_currn = (_TPPpglobalFrameView_FormularRadioGroupElement_name) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularRadioGroupElement_name));
#endif
_currn->_prod = RULEpglobalFrameView_FormularRadioGroupElement_name;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularRadioGroupElement_name;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularRadioGroupElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularRadioGroupElement_label (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularRadioGroupElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularRadioGroupElement_label _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularRadioGroupElement_label;
#else
_currn = (_TPPpglobalFrameView_FormularRadioGroupElement_label) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularRadioGroupElement_label));
#endif
_currn->_prod = RULEpglobalFrameView_FormularRadioGroupElement_label;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularRadioGroupElement_label;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularRadioGroupElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularRadioGroupElement_description (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularRadioGroupElement_description (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularRadioGroupElement_description _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularRadioGroupElement_description;
#else
_currn = (_TPPpglobalFrameView_FormularRadioGroupElement_description) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularRadioGroupElement_description));
#endif
_currn->_prod = RULEpglobalFrameView_FormularRadioGroupElement_description;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularRadioGroupElement_description;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularRadioGroupElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularRadioGroupElement_pvalue (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularRadioGroupElement_pvalue (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularRadioGroupElement_pvalue _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularRadioGroupElement_pvalue;
#else
_currn = (_TPPpglobalFrameView_FormularRadioGroupElement_pvalue) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularRadioGroupElement_pvalue));
#endif
_currn->_prod = RULEpglobalFrameView_FormularRadioGroupElement_pvalue;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularRadioGroupElement_pvalue;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularRadioGroupElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_FormularRadioElementpglobalFrameView_FormularRadioGroupElement_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_FormularRadioElementpglobalFrameView_FormularRadioGroupElement_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_FormularRadioElementpglobalFrameView_FormularRadioGroupElement_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_FormularRadioElementpglobalFrameView_FormularRadioGroupElement_elements;
#else
	_new = (_TPPLST_globalFrameView_FormularRadioElementpglobalFrameView_FormularRadioGroupElement_elements) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_FormularRadioElementpglobalFrameView_FormularRadioGroupElement_elements));
#endif
	_new->_prod = RULELST_globalFrameView_FormularRadioElementpglobalFrameView_FormularRadioGroupElement_elements;
	_new->_desc1 = (_TSPglobalFrameView_FormularRadioElement)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_FormularRadioGroupElement_elements)_new;
	{_TPPLST_globalFrameView_FormularRadioElementpglobalFrameView_FormularRadioGroupElement_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_FormularRadioElementpglobalFrameView_FormularRadioGroupElement_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempglobalFrameView_FormularRadioGroupElement_elements (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempglobalFrameView_FormularRadioGroupElement_elements (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBglobalFrameView_FormularRadioElement)) 
		return _MkglobalFrameView_FormularRadioElementpglobalFrameView_FormularRadioGroupElement_elements (_coordref, MkglobalFrameView_FormularRadioElement (_coordref, t));
	message (DEADLY, "a pglobalFrameView_FormularRadioGroupElement_elements list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pglobalFrameView_FormularRadioGroupElement_elements (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pglobalFrameView_FormularRadioGroupElement_elements (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_globalFrameView_FormularRadioGroupElement_elements)
	l =_ElempglobalFrameView_FormularRadioGroupElement_elements (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_globalFrameView_FormularRadioGroupElement_elements)
	r = _ElempglobalFrameView_FormularRadioGroupElement_elements (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pglobalFrameView_FormularRadioGroupElement_elements)r)->_desc2);
((_TPPLST_0pglobalFrameView_FormularRadioGroupElement_elements)r)->_desc2 = ((_TPPLST_0pglobalFrameView_FormularRadioGroupElement_elements)l)->_desc2;
((_TPPLST_0pglobalFrameView_FormularRadioGroupElement_elements)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularRadioGroupElement_elements (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpglobalFrameView_FormularRadioGroupElement_elements (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpglobalFrameView_FormularRadioGroupElement_elements root;
	_TPPLST_0pglobalFrameView_FormularRadioGroupElement_elements list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pglobalFrameView_FormularRadioGroupElement_elements)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_globalFrameView_FormularRadioGroupElement_elements)
		t = _ElempglobalFrameView_FormularRadioGroupElement_elements (_coordref, t);
	remain = (_TPPLST_0pglobalFrameView_FormularRadioGroupElement_elements)(((_TPPLST_0pglobalFrameView_FormularRadioGroupElement_elements) t)->_desc2);
	((_TPPLST_0pglobalFrameView_FormularRadioGroupElement_elements) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pglobalFrameView_FormularRadioGroupElement_elements;
#else
list = (_TPPLST_0pglobalFrameView_FormularRadioGroupElement_elements) TreeNodeAlloc (sizeof (struct _TPLST_0pglobalFrameView_FormularRadioGroupElement_elements));
#endif
list->_prod = RULELST_0pglobalFrameView_FormularRadioGroupElement_elements;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pglobalFrameView_FormularRadioGroupElement_elements)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_globalFrameView_FormularRadioElementpglobalFrameView_FormularRadioGroupElement_elements)
	{_TPPLST_globalFrameView_FormularRadioElementpglobalFrameView_FormularRadioGroupElement_elements _currn = _currn;
	 _TERMACT_LST_globalFrameView_FormularRadioElementpglobalFrameView_FormularRadioGroupElement_elements;
	} this code has been moved to _ElempglobalFrameView_FormularRadioGroupElement_elements*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpglobalFrameView_FormularRadioGroupElement_elements;
#else
root = (_TPPpglobalFrameView_FormularRadioGroupElement_elements) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularRadioGroupElement_elements));
#endif
root->_prod = RULEpglobalFrameView_FormularRadioGroupElement_elements;
root->_desc1 = (_TSPLST_globalFrameView_FormularRadioGroupElement_elements)list;
{ _TPPpglobalFrameView_FormularRadioGroupElement_elements _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pglobalFrameView_FormularRadioGroupElement_elements;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularRadioGroupElement_checked (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularRadioGroupElement_checked (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularRadioGroupElement_checked _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularRadioGroupElement_checked;
#else
_currn = (_TPPpglobalFrameView_FormularRadioGroupElement_checked) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularRadioGroupElement_checked));
#endif
_currn->_prod = RULEpglobalFrameView_FormularRadioGroupElement_checked;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularRadioGroupElement_checked;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularRadioGroupElement_checked */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularConstantSelectOption (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpglobalFrameView_FormularConstantSelectOption (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpglobalFrameView_FormularConstantSelectOption _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularConstantSelectOption;
#else
_currn = (_TPPpglobalFrameView_FormularConstantSelectOption) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularConstantSelectOption));
#endif
_currn->_prod = RULEpglobalFrameView_FormularConstantSelectOption;
_currn->_desc1 = (_TSPglobalFrameView_FormularConstantSelectOption_values) MkglobalFrameView_FormularConstantSelectOption_values (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularConstantSelectOption: root of subtree no. 1 can not be made a globalFrameView_FormularConstantSelectOption_values node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularConstantSelectOption;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularConstantSelectOption */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularConstantSelectOption_values (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularConstantSelectOption_values (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularConstantSelectOption_values _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularConstantSelectOption_values;
#else
_currn = (_TPPpglobalFrameView_FormularConstantSelectOption_values) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularConstantSelectOption_values));
#endif
_currn->_prod = RULEpglobalFrameView_FormularConstantSelectOption_values;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularConstantSelectOption_values;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularConstantSelectOption_values */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularDataSourceSelectOption (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpglobalFrameView_FormularDataSourceSelectOption (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpglobalFrameView_FormularDataSourceSelectOption _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularDataSourceSelectOption;
#else
_currn = (_TPPpglobalFrameView_FormularDataSourceSelectOption) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularDataSourceSelectOption));
#endif
_currn->_prod = RULEpglobalFrameView_FormularDataSourceSelectOption;
_currn->_desc1 = (_TSPglobalFrameView_FormularDataSourceSelectOption_values) MkglobalFrameView_FormularDataSourceSelectOption_values (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularDataSourceSelectOption: root of subtree no. 1 can not be made a globalFrameView_FormularDataSourceSelectOption_values node ", 0, _coordref);
_currn->_desc2 = (_TSPglobalFrameView_FormularDataSourceSelectOption_params) MkglobalFrameView_FormularDataSourceSelectOption_params (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularDataSourceSelectOption: root of subtree no. 2 can not be made a globalFrameView_FormularDataSourceSelectOption_params node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularDataSourceSelectOption;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularDataSourceSelectOption */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularDataSourceSelectOption_values (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularDataSourceSelectOption_values (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularDataSourceSelectOption_values _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularDataSourceSelectOption_values;
#else
_currn = (_TPPpglobalFrameView_FormularDataSourceSelectOption_values) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularDataSourceSelectOption_values));
#endif
_currn->_prod = RULEpglobalFrameView_FormularDataSourceSelectOption_values;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularDataSourceSelectOption_values;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularDataSourceSelectOption_values */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_DataSourceActualParampglobalFrameView_FormularDataSourceSelectOption_params (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_DataSourceActualParampglobalFrameView_FormularDataSourceSelectOption_params (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_DataSourceActualParampglobalFrameView_FormularDataSourceSelectOption_params _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_DataSourceActualParampglobalFrameView_FormularDataSourceSelectOption_params;
#else
	_new = (_TPPLST_globalFrameView_DataSourceActualParampglobalFrameView_FormularDataSourceSelectOption_params) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_DataSourceActualParampglobalFrameView_FormularDataSourceSelectOption_params));
#endif
	_new->_prod = RULELST_globalFrameView_DataSourceActualParampglobalFrameView_FormularDataSourceSelectOption_params;
	_new->_desc1 = (_TSPglobalFrameView_DataSourceActualParam)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_FormularDataSourceSelectOption_params)_new;
	{_TPPLST_globalFrameView_DataSourceActualParampglobalFrameView_FormularDataSourceSelectOption_params _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_DataSourceActualParampglobalFrameView_FormularDataSourceSelectOption_params;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempglobalFrameView_FormularDataSourceSelectOption_params (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempglobalFrameView_FormularDataSourceSelectOption_params (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBglobalFrameView_DataSourceActualParam)) 
		return _MkglobalFrameView_DataSourceActualParampglobalFrameView_FormularDataSourceSelectOption_params (_coordref, MkglobalFrameView_DataSourceActualParam (_coordref, t));
	message (DEADLY, "a pglobalFrameView_FormularDataSourceSelectOption_params list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pglobalFrameView_FormularDataSourceSelectOption_params (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pglobalFrameView_FormularDataSourceSelectOption_params (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_globalFrameView_FormularDataSourceSelectOption_params)
	l =_ElempglobalFrameView_FormularDataSourceSelectOption_params (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_globalFrameView_FormularDataSourceSelectOption_params)
	r = _ElempglobalFrameView_FormularDataSourceSelectOption_params (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pglobalFrameView_FormularDataSourceSelectOption_params)r)->_desc2);
((_TPPLST_0pglobalFrameView_FormularDataSourceSelectOption_params)r)->_desc2 = ((_TPPLST_0pglobalFrameView_FormularDataSourceSelectOption_params)l)->_desc2;
((_TPPLST_0pglobalFrameView_FormularDataSourceSelectOption_params)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularDataSourceSelectOption_params (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpglobalFrameView_FormularDataSourceSelectOption_params (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpglobalFrameView_FormularDataSourceSelectOption_params root;
	_TPPLST_0pglobalFrameView_FormularDataSourceSelectOption_params list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pglobalFrameView_FormularDataSourceSelectOption_params)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_globalFrameView_FormularDataSourceSelectOption_params)
		t = _ElempglobalFrameView_FormularDataSourceSelectOption_params (_coordref, t);
	remain = (_TPPLST_0pglobalFrameView_FormularDataSourceSelectOption_params)(((_TPPLST_0pglobalFrameView_FormularDataSourceSelectOption_params) t)->_desc2);
	((_TPPLST_0pglobalFrameView_FormularDataSourceSelectOption_params) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pglobalFrameView_FormularDataSourceSelectOption_params;
#else
list = (_TPPLST_0pglobalFrameView_FormularDataSourceSelectOption_params) TreeNodeAlloc (sizeof (struct _TPLST_0pglobalFrameView_FormularDataSourceSelectOption_params));
#endif
list->_prod = RULELST_0pglobalFrameView_FormularDataSourceSelectOption_params;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pglobalFrameView_FormularDataSourceSelectOption_params)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_globalFrameView_DataSourceActualParampglobalFrameView_FormularDataSourceSelectOption_params)
	{_TPPLST_globalFrameView_DataSourceActualParampglobalFrameView_FormularDataSourceSelectOption_params _currn = _currn;
	 _TERMACT_LST_globalFrameView_DataSourceActualParampglobalFrameView_FormularDataSourceSelectOption_params;
	} this code has been moved to _ElempglobalFrameView_FormularDataSourceSelectOption_params*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpglobalFrameView_FormularDataSourceSelectOption_params;
#else
root = (_TPPpglobalFrameView_FormularDataSourceSelectOption_params) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularDataSourceSelectOption_params));
#endif
root->_prod = RULEpglobalFrameView_FormularDataSourceSelectOption_params;
root->_desc1 = (_TSPLST_globalFrameView_FormularDataSourceSelectOption_params)list;
{ _TPPpglobalFrameView_FormularDataSourceSelectOption_params _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pglobalFrameView_FormularDataSourceSelectOption_params;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularPHPScriptSelectOption (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpglobalFrameView_FormularPHPScriptSelectOption (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpglobalFrameView_FormularPHPScriptSelectOption _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularPHPScriptSelectOption;
#else
_currn = (_TPPpglobalFrameView_FormularPHPScriptSelectOption) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularPHPScriptSelectOption));
#endif
_currn->_prod = RULEpglobalFrameView_FormularPHPScriptSelectOption;
_currn->_desc1 = (_TSPglobalFrameView_FormularPHPScriptSelectOption_script) MkglobalFrameView_FormularPHPScriptSelectOption_script (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularPHPScriptSelectOption: root of subtree no. 1 can not be made a globalFrameView_FormularPHPScriptSelectOption_script node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularPHPScriptSelectOption;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularPHPScriptSelectOption */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularPHPScriptSelectOption_script (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularPHPScriptSelectOption_script (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularPHPScriptSelectOption_script _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularPHPScriptSelectOption_script;
#else
_currn = (_TPPpglobalFrameView_FormularPHPScriptSelectOption_script) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularPHPScriptSelectOption_script));
#endif
_currn->_prod = RULEpglobalFrameView_FormularPHPScriptSelectOption_script;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularPHPScriptSelectOption_script;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularPHPScriptSelectOption_script */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularCondition (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpglobalFrameView_FormularCondition (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpglobalFrameView_FormularCondition _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularCondition;
#else
_currn = (_TPPpglobalFrameView_FormularCondition) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularCondition));
#endif
_currn->_prod = RULEpglobalFrameView_FormularCondition;
_currn->_desc1 = (_TSPglobalFrameView_FormularCondition_condition) MkglobalFrameView_FormularCondition_condition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularCondition: root of subtree no. 1 can not be made a globalFrameView_FormularCondition_condition node ", 0, _coordref);
_currn->_desc2 = (_TSPglobalFrameView_FormularCondition_params) MkglobalFrameView_FormularCondition_params (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularCondition: root of subtree no. 2 can not be made a globalFrameView_FormularCondition_params node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularCondition;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularCondition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularCondition_condition (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularCondition_condition (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularCondition_condition _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularCondition_condition;
#else
_currn = (_TPPpglobalFrameView_FormularCondition_condition) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularCondition_condition));
#endif
_currn->_prod = RULEpglobalFrameView_FormularCondition_condition;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularCondition_condition;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularCondition_condition */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_FormularConditionActualParampglobalFrameView_FormularCondition_params (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_FormularConditionActualParampglobalFrameView_FormularCondition_params (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_FormularConditionActualParampglobalFrameView_FormularCondition_params _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_FormularConditionActualParampglobalFrameView_FormularCondition_params;
#else
	_new = (_TPPLST_globalFrameView_FormularConditionActualParampglobalFrameView_FormularCondition_params) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_FormularConditionActualParampglobalFrameView_FormularCondition_params));
#endif
	_new->_prod = RULELST_globalFrameView_FormularConditionActualParampglobalFrameView_FormularCondition_params;
	_new->_desc1 = (_TSPglobalFrameView_FormularConditionActualParam)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_FormularCondition_params)_new;
	{_TPPLST_globalFrameView_FormularConditionActualParampglobalFrameView_FormularCondition_params _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_FormularConditionActualParampglobalFrameView_FormularCondition_params;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempglobalFrameView_FormularCondition_params (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempglobalFrameView_FormularCondition_params (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBglobalFrameView_FormularConditionActualParam)) 
		return _MkglobalFrameView_FormularConditionActualParampglobalFrameView_FormularCondition_params (_coordref, MkglobalFrameView_FormularConditionActualParam (_coordref, t));
	message (DEADLY, "a pglobalFrameView_FormularCondition_params list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pglobalFrameView_FormularCondition_params (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pglobalFrameView_FormularCondition_params (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_globalFrameView_FormularCondition_params)
	l =_ElempglobalFrameView_FormularCondition_params (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_globalFrameView_FormularCondition_params)
	r = _ElempglobalFrameView_FormularCondition_params (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pglobalFrameView_FormularCondition_params)r)->_desc2);
((_TPPLST_0pglobalFrameView_FormularCondition_params)r)->_desc2 = ((_TPPLST_0pglobalFrameView_FormularCondition_params)l)->_desc2;
((_TPPLST_0pglobalFrameView_FormularCondition_params)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularCondition_params (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpglobalFrameView_FormularCondition_params (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpglobalFrameView_FormularCondition_params root;
	_TPPLST_0pglobalFrameView_FormularCondition_params list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pglobalFrameView_FormularCondition_params)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_globalFrameView_FormularCondition_params)
		t = _ElempglobalFrameView_FormularCondition_params (_coordref, t);
	remain = (_TPPLST_0pglobalFrameView_FormularCondition_params)(((_TPPLST_0pglobalFrameView_FormularCondition_params) t)->_desc2);
	((_TPPLST_0pglobalFrameView_FormularCondition_params) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pglobalFrameView_FormularCondition_params;
#else
list = (_TPPLST_0pglobalFrameView_FormularCondition_params) TreeNodeAlloc (sizeof (struct _TPLST_0pglobalFrameView_FormularCondition_params));
#endif
list->_prod = RULELST_0pglobalFrameView_FormularCondition_params;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pglobalFrameView_FormularCondition_params)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_globalFrameView_FormularConditionActualParampglobalFrameView_FormularCondition_params)
	{_TPPLST_globalFrameView_FormularConditionActualParampglobalFrameView_FormularCondition_params _currn = _currn;
	 _TERMACT_LST_globalFrameView_FormularConditionActualParampglobalFrameView_FormularCondition_params;
	} this code has been moved to _ElempglobalFrameView_FormularCondition_params*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpglobalFrameView_FormularCondition_params;
#else
root = (_TPPpglobalFrameView_FormularCondition_params) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularCondition_params));
#endif
root->_prod = RULEpglobalFrameView_FormularCondition_params;
root->_desc1 = (_TSPLST_globalFrameView_FormularCondition_params)list;
{ _TPPpglobalFrameView_FormularCondition_params _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pglobalFrameView_FormularCondition_params;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularConditionActualParam (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpglobalFrameView_FormularConditionActualParam (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpglobalFrameView_FormularConditionActualParam _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularConditionActualParam;
#else
_currn = (_TPPpglobalFrameView_FormularConditionActualParam) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularConditionActualParam));
#endif
_currn->_prod = RULEpglobalFrameView_FormularConditionActualParam;
_currn->_desc1 = (_TSPglobalFrameView_FormularConditionActualParam_value) MkglobalFrameView_FormularConditionActualParam_value (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularConditionActualParam: root of subtree no. 1 can not be made a globalFrameView_FormularConditionActualParam_value node ", 0, _coordref);
_currn->_desc2 = (_TSPglobalFrameView_FormularConditionActualParam_formalparam) MkglobalFrameView_FormularConditionActualParam_formalparam (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FormularConditionActualParam: root of subtree no. 2 can not be made a globalFrameView_FormularConditionActualParam_formalparam node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularConditionActualParam;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularConditionActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularConditionActualParam_value (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularConditionActualParam_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularConditionActualParam_value _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularConditionActualParam_value;
#else
_currn = (_TPPpglobalFrameView_FormularConditionActualParam_value) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularConditionActualParam_value));
#endif
_currn->_prod = RULEpglobalFrameView_FormularConditionActualParam_value;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularConditionActualParam_value;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularConditionActualParam_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FormularConditionActualParam_formalparam (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FormularConditionActualParam_formalparam (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FormularConditionActualParam_formalparam _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FormularConditionActualParam_formalparam;
#else
_currn = (_TPPpglobalFrameView_FormularConditionActualParam_formalparam) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FormularConditionActualParam_formalparam));
#endif
_currn->_prod = RULEpglobalFrameView_FormularConditionActualParam_formalparam;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FormularConditionActualParam_formalparam;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FormularConditionActualParam_formalparam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_Paragraph (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpglobalFrameView_Paragraph (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpglobalFrameView_Paragraph _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_Paragraph;
#else
_currn = (_TPPpglobalFrameView_Paragraph) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_Paragraph));
#endif
_currn->_prod = RULEpglobalFrameView_Paragraph;
_currn->_desc1 = (_TSPglobalFrameView_Paragraph_styleSheetBlock) MkglobalFrameView_Paragraph_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_Paragraph: root of subtree no. 1 can not be made a globalFrameView_Paragraph_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSPglobalFrameView_Paragraph_fragments) MkglobalFrameView_Paragraph_fragments (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_Paragraph: root of subtree no. 2 can not be made a globalFrameView_Paragraph_fragments node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_Paragraph;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_Paragraph */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_Paragraph_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_Paragraph_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_Paragraph_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_Paragraph_styleSheetBlock;
#else
_currn = (_TPPpglobalFrameView_Paragraph_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_Paragraph_styleSheetBlock));
#endif
_currn->_prod = RULEpglobalFrameView_Paragraph_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_Paragraph_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_Paragraph_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_TextFragmentpglobalFrameView_Paragraph_fragments (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_TextFragmentpglobalFrameView_Paragraph_fragments (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_TextFragmentpglobalFrameView_Paragraph_fragments _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_TextFragmentpglobalFrameView_Paragraph_fragments;
#else
	_new = (_TPPLST_globalFrameView_TextFragmentpglobalFrameView_Paragraph_fragments) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_TextFragmentpglobalFrameView_Paragraph_fragments));
#endif
	_new->_prod = RULELST_globalFrameView_TextFragmentpglobalFrameView_Paragraph_fragments;
	_new->_desc1 = (_TSPglobalFrameView_TextFragment)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_Paragraph_fragments)_new;
	{_TPPLST_globalFrameView_TextFragmentpglobalFrameView_Paragraph_fragments _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_TextFragmentpglobalFrameView_Paragraph_fragments;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_ImageFragmentpglobalFrameView_Paragraph_fragments (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_ImageFragmentpglobalFrameView_Paragraph_fragments (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_ImageFragmentpglobalFrameView_Paragraph_fragments _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_ImageFragmentpglobalFrameView_Paragraph_fragments;
#else
	_new = (_TPPLST_globalFrameView_ImageFragmentpglobalFrameView_Paragraph_fragments) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_ImageFragmentpglobalFrameView_Paragraph_fragments));
#endif
	_new->_prod = RULELST_globalFrameView_ImageFragmentpglobalFrameView_Paragraph_fragments;
	_new->_desc1 = (_TSPglobalFrameView_ImageFragment)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_Paragraph_fragments)_new;
	{_TPPLST_globalFrameView_ImageFragmentpglobalFrameView_Paragraph_fragments _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_ImageFragmentpglobalFrameView_Paragraph_fragments;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_OutputElementpglobalFrameView_Paragraph_fragments (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_OutputElementpglobalFrameView_Paragraph_fragments (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_OutputElementpglobalFrameView_Paragraph_fragments _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_OutputElementpglobalFrameView_Paragraph_fragments;
#else
	_new = (_TPPLST_globalFrameView_OutputElementpglobalFrameView_Paragraph_fragments) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_OutputElementpglobalFrameView_Paragraph_fragments));
#endif
	_new->_prod = RULELST_globalFrameView_OutputElementpglobalFrameView_Paragraph_fragments;
	_new->_desc1 = (_TSPglobalFrameView_OutputElement)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_Paragraph_fragments)_new;
	{_TPPLST_globalFrameView_OutputElementpglobalFrameView_Paragraph_fragments _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_OutputElementpglobalFrameView_Paragraph_fragments;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempglobalFrameView_Paragraph_fragments (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempglobalFrameView_Paragraph_fragments (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBglobalFrameView_TextFragment)) 
		return _MkglobalFrameView_TextFragmentpglobalFrameView_Paragraph_fragments (_coordref, MkglobalFrameView_TextFragment (_coordref, t));
	if (IsSymb (t, SYMBglobalFrameView_ImageFragment)) 
		return _MkglobalFrameView_ImageFragmentpglobalFrameView_Paragraph_fragments (_coordref, MkglobalFrameView_ImageFragment (_coordref, t));
	if (IsSymb (t, SYMBglobalFrameView_OutputElement)) 
		return _MkglobalFrameView_OutputElementpglobalFrameView_Paragraph_fragments (_coordref, MkglobalFrameView_OutputElement (_coordref, t));
	message (DEADLY, "a pglobalFrameView_Paragraph_fragments list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pglobalFrameView_Paragraph_fragments (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pglobalFrameView_Paragraph_fragments (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_globalFrameView_Paragraph_fragments)
	l =_ElempglobalFrameView_Paragraph_fragments (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_globalFrameView_Paragraph_fragments)
	r = _ElempglobalFrameView_Paragraph_fragments (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pglobalFrameView_Paragraph_fragments)r)->_desc2);
((_TPPLST_0pglobalFrameView_Paragraph_fragments)r)->_desc2 = ((_TPPLST_0pglobalFrameView_Paragraph_fragments)l)->_desc2;
((_TPPLST_0pglobalFrameView_Paragraph_fragments)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_Paragraph_fragments (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpglobalFrameView_Paragraph_fragments (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpglobalFrameView_Paragraph_fragments root;
	_TPPLST_0pglobalFrameView_Paragraph_fragments list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pglobalFrameView_Paragraph_fragments)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_globalFrameView_Paragraph_fragments)
		t = _ElempglobalFrameView_Paragraph_fragments (_coordref, t);
	remain = (_TPPLST_0pglobalFrameView_Paragraph_fragments)(((_TPPLST_0pglobalFrameView_Paragraph_fragments) t)->_desc2);
	((_TPPLST_0pglobalFrameView_Paragraph_fragments) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pglobalFrameView_Paragraph_fragments;
#else
list = (_TPPLST_0pglobalFrameView_Paragraph_fragments) TreeNodeAlloc (sizeof (struct _TPLST_0pglobalFrameView_Paragraph_fragments));
#endif
list->_prod = RULELST_0pglobalFrameView_Paragraph_fragments;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pglobalFrameView_Paragraph_fragments)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_globalFrameView_TextFragmentpglobalFrameView_Paragraph_fragments)
	{_TPPLST_globalFrameView_TextFragmentpglobalFrameView_Paragraph_fragments _currn = _currn;
	 _TERMACT_LST_globalFrameView_TextFragmentpglobalFrameView_Paragraph_fragments;
	} this code has been moved to _ElempglobalFrameView_Paragraph_fragments*/
	/*if (_currn->_prod == RULELST_globalFrameView_ImageFragmentpglobalFrameView_Paragraph_fragments)
	{_TPPLST_globalFrameView_ImageFragmentpglobalFrameView_Paragraph_fragments _currn = _currn;
	 _TERMACT_LST_globalFrameView_ImageFragmentpglobalFrameView_Paragraph_fragments;
	} this code has been moved to _ElempglobalFrameView_Paragraph_fragments*/
	/*if (_currn->_prod == RULELST_globalFrameView_OutputElementpglobalFrameView_Paragraph_fragments)
	{_TPPLST_globalFrameView_OutputElementpglobalFrameView_Paragraph_fragments _currn = _currn;
	 _TERMACT_LST_globalFrameView_OutputElementpglobalFrameView_Paragraph_fragments;
	} this code has been moved to _ElempglobalFrameView_Paragraph_fragments*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpglobalFrameView_Paragraph_fragments;
#else
root = (_TPPpglobalFrameView_Paragraph_fragments) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_Paragraph_fragments));
#endif
root->_prod = RULEpglobalFrameView_Paragraph_fragments;
root->_desc1 = (_TSPLST_globalFrameView_Paragraph_fragments)list;
{ _TPPpglobalFrameView_Paragraph_fragments _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pglobalFrameView_Paragraph_fragments;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_TextFragment (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpglobalFrameView_TextFragment (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpglobalFrameView_TextFragment _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_TextFragment;
#else
_currn = (_TPPpglobalFrameView_TextFragment) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_TextFragment));
#endif
_currn->_prod = RULEpglobalFrameView_TextFragment;
_currn->_desc1 = (_TSPglobalFrameView_TextFragment_link) MkglobalFrameView_TextFragment_link (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_TextFragment: root of subtree no. 1 can not be made a globalFrameView_TextFragment_link node ", 0, _coordref);
_currn->_desc2 = (_TSPglobalFrameView_TextFragment_value) MkglobalFrameView_TextFragment_value (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_TextFragment: root of subtree no. 2 can not be made a globalFrameView_TextFragment_value node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_TextFragment;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_TextFragment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_TextFragment_link (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_TextFragment_link (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_TextFragment_link _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_TextFragment_link;
#else
_currn = (_TPPpglobalFrameView_TextFragment_link) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_TextFragment_link));
#endif
_currn->_prod = RULEpglobalFrameView_TextFragment_link;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_TextFragment_link;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_TextFragment_link */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_TextFragment_value (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_TextFragment_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_TextFragment_value _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_TextFragment_value;
#else
_currn = (_TPPpglobalFrameView_TextFragment_value) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_TextFragment_value));
#endif
_currn->_prod = RULEpglobalFrameView_TextFragment_value;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_TextFragment_value;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_TextFragment_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_ImageFragment (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkpglobalFrameView_ImageFragment (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPpglobalFrameView_ImageFragment _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_ImageFragment;
#else
_currn = (_TPPpglobalFrameView_ImageFragment) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_ImageFragment));
#endif
_currn->_prod = RULEpglobalFrameView_ImageFragment;
_currn->_desc1 = (_TSPglobalFrameView_ImageFragment_link) MkglobalFrameView_ImageFragment_link (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_ImageFragment: root of subtree no. 1 can not be made a globalFrameView_ImageFragment_link node ", 0, _coordref);
_currn->_desc2 = (_TSPglobalFrameView_ImageFragment_width) MkglobalFrameView_ImageFragment_width (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_ImageFragment: root of subtree no. 2 can not be made a globalFrameView_ImageFragment_width node ", 0, _coordref);
_currn->_desc3 = (_TSPglobalFrameView_ImageFragment_height) MkglobalFrameView_ImageFragment_height (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_ImageFragment: root of subtree no. 3 can not be made a globalFrameView_ImageFragment_height node ", 0, _coordref);
_currn->_desc4 = (_TSPglobalFrameView_ImageFragment_text) MkglobalFrameView_ImageFragment_text (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_ImageFragment: root of subtree no. 4 can not be made a globalFrameView_ImageFragment_text node ", 0, _coordref);
_currn->_desc5 = (_TSPglobalFrameView_ImageFragment_imageName) MkglobalFrameView_ImageFragment_imageName (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_ImageFragment: root of subtree no. 5 can not be made a globalFrameView_ImageFragment_imageName node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_ImageFragment;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_ImageFragment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_ImageFragment_link (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_ImageFragment_link (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_ImageFragment_link _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_ImageFragment_link;
#else
_currn = (_TPPpglobalFrameView_ImageFragment_link) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_ImageFragment_link));
#endif
_currn->_prod = RULEpglobalFrameView_ImageFragment_link;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_ImageFragment_link;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_ImageFragment_link */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_ImageFragment_width (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_ImageFragment_width (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_ImageFragment_width _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_ImageFragment_width;
#else
_currn = (_TPPpglobalFrameView_ImageFragment_width) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_ImageFragment_width));
#endif
_currn->_prod = RULEpglobalFrameView_ImageFragment_width;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_ImageFragment_width;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_ImageFragment_width */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_ImageFragment_height (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_ImageFragment_height (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_ImageFragment_height _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_ImageFragment_height;
#else
_currn = (_TPPpglobalFrameView_ImageFragment_height) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_ImageFragment_height));
#endif
_currn->_prod = RULEpglobalFrameView_ImageFragment_height;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_ImageFragment_height;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_ImageFragment_height */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_ImageFragment_text (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_ImageFragment_text (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_ImageFragment_text _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_ImageFragment_text;
#else
_currn = (_TPPpglobalFrameView_ImageFragment_text) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_ImageFragment_text));
#endif
_currn->_prod = RULEpglobalFrameView_ImageFragment_text;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_ImageFragment_text;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_ImageFragment_text */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_ImageFragment_imageName (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_ImageFragment_imageName (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_ImageFragment_imageName _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_ImageFragment_imageName;
#else
_currn = (_TPPpglobalFrameView_ImageFragment_imageName) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_ImageFragment_imageName));
#endif
_currn->_prod = RULEpglobalFrameView_ImageFragment_imageName;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_ImageFragment_imageName;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_ImageFragment_imageName */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_Heading (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpglobalFrameView_Heading (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpglobalFrameView_Heading _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_Heading;
#else
_currn = (_TPPpglobalFrameView_Heading) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_Heading));
#endif
_currn->_prod = RULEpglobalFrameView_Heading;
_currn->_desc1 = (_TSPglobalFrameView_Heading_styleSheetBlock) MkglobalFrameView_Heading_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_Heading: root of subtree no. 1 can not be made a globalFrameView_Heading_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSPglobalFrameView_Heading_text) MkglobalFrameView_Heading_text (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_Heading: root of subtree no. 2 can not be made a globalFrameView_Heading_text node ", 0, _coordref);
_currn->_desc3 = (_TSPglobalFrameView_Heading_level) MkglobalFrameView_Heading_level (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_Heading: root of subtree no. 3 can not be made a globalFrameView_Heading_level node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_Heading;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_Heading */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_Heading_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_Heading_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_Heading_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_Heading_styleSheetBlock;
#else
_currn = (_TPPpglobalFrameView_Heading_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_Heading_styleSheetBlock));
#endif
_currn->_prod = RULEpglobalFrameView_Heading_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_Heading_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_Heading_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_Heading_text (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_Heading_text (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_Heading_text _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_Heading_text;
#else
_currn = (_TPPpglobalFrameView_Heading_text) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_Heading_text));
#endif
_currn->_prod = RULEpglobalFrameView_Heading_text;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_Heading_text;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_Heading_text */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_Heading_level (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_Heading_level (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_Heading_level _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_Heading_level;
#else
_currn = (_TPPpglobalFrameView_Heading_level) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_Heading_level));
#endif
_currn->_prod = RULEpglobalFrameView_Heading_level;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_Heading_level;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_Heading_level */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_Table (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpglobalFrameView_Table (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpglobalFrameView_Table _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_Table;
#else
_currn = (_TPPpglobalFrameView_Table) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_Table));
#endif
_currn->_prod = RULEpglobalFrameView_Table;
_currn->_desc1 = (_TSPglobalFrameView_Table_styleSheetBlock) MkglobalFrameView_Table_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_Table: root of subtree no. 1 can not be made a globalFrameView_Table_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSPglobalFrameView_Table_rows) MkglobalFrameView_Table_rows (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_Table: root of subtree no. 2 can not be made a globalFrameView_Table_rows node ", 0, _coordref);
_currn->_desc3 = (_TSPglobalFrameView_Table_columns) MkglobalFrameView_Table_columns (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_Table: root of subtree no. 3 can not be made a globalFrameView_Table_columns node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_Table;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_Table */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_Table_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_Table_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_Table_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_Table_styleSheetBlock;
#else
_currn = (_TPPpglobalFrameView_Table_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_Table_styleSheetBlock));
#endif
_currn->_prod = RULEpglobalFrameView_Table_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_Table_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_Table_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_TableRowpglobalFrameView_Table_rows (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_TableRowpglobalFrameView_Table_rows (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_TableRowpglobalFrameView_Table_rows _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_TableRowpglobalFrameView_Table_rows;
#else
	_new = (_TPPLST_globalFrameView_TableRowpglobalFrameView_Table_rows) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_TableRowpglobalFrameView_Table_rows));
#endif
	_new->_prod = RULELST_globalFrameView_TableRowpglobalFrameView_Table_rows;
	_new->_desc1 = (_TSPglobalFrameView_TableRow)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_Table_rows)_new;
	{_TPPLST_globalFrameView_TableRowpglobalFrameView_Table_rows _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_TableRowpglobalFrameView_Table_rows;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempglobalFrameView_Table_rows (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempglobalFrameView_Table_rows (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBglobalFrameView_TableRow)) 
		return _MkglobalFrameView_TableRowpglobalFrameView_Table_rows (_coordref, MkglobalFrameView_TableRow (_coordref, t));
	message (DEADLY, "a pglobalFrameView_Table_rows list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pglobalFrameView_Table_rows (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pglobalFrameView_Table_rows (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_globalFrameView_Table_rows)
	l =_ElempglobalFrameView_Table_rows (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_globalFrameView_Table_rows)
	r = _ElempglobalFrameView_Table_rows (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pglobalFrameView_Table_rows)r)->_desc2);
((_TPPLST_0pglobalFrameView_Table_rows)r)->_desc2 = ((_TPPLST_0pglobalFrameView_Table_rows)l)->_desc2;
((_TPPLST_0pglobalFrameView_Table_rows)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_Table_rows (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpglobalFrameView_Table_rows (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpglobalFrameView_Table_rows root;
	_TPPLST_0pglobalFrameView_Table_rows list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pglobalFrameView_Table_rows)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_globalFrameView_Table_rows)
		t = _ElempglobalFrameView_Table_rows (_coordref, t);
	remain = (_TPPLST_0pglobalFrameView_Table_rows)(((_TPPLST_0pglobalFrameView_Table_rows) t)->_desc2);
	((_TPPLST_0pglobalFrameView_Table_rows) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pglobalFrameView_Table_rows;
#else
list = (_TPPLST_0pglobalFrameView_Table_rows) TreeNodeAlloc (sizeof (struct _TPLST_0pglobalFrameView_Table_rows));
#endif
list->_prod = RULELST_0pglobalFrameView_Table_rows;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pglobalFrameView_Table_rows)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_globalFrameView_TableRowpglobalFrameView_Table_rows)
	{_TPPLST_globalFrameView_TableRowpglobalFrameView_Table_rows _currn = _currn;
	 _TERMACT_LST_globalFrameView_TableRowpglobalFrameView_Table_rows;
	} this code has been moved to _ElempglobalFrameView_Table_rows*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpglobalFrameView_Table_rows;
#else
root = (_TPPpglobalFrameView_Table_rows) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_Table_rows));
#endif
root->_prod = RULEpglobalFrameView_Table_rows;
root->_desc1 = (_TSPLST_globalFrameView_Table_rows)list;
{ _TPPpglobalFrameView_Table_rows _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pglobalFrameView_Table_rows;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_TableColumnpglobalFrameView_Table_columns (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_TableColumnpglobalFrameView_Table_columns (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_TableColumnpglobalFrameView_Table_columns _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_TableColumnpglobalFrameView_Table_columns;
#else
	_new = (_TPPLST_globalFrameView_TableColumnpglobalFrameView_Table_columns) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_TableColumnpglobalFrameView_Table_columns));
#endif
	_new->_prod = RULELST_globalFrameView_TableColumnpglobalFrameView_Table_columns;
	_new->_desc1 = (_TSPglobalFrameView_TableColumn)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_Table_columns)_new;
	{_TPPLST_globalFrameView_TableColumnpglobalFrameView_Table_columns _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_TableColumnpglobalFrameView_Table_columns;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempglobalFrameView_Table_columns (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempglobalFrameView_Table_columns (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBglobalFrameView_TableColumn)) 
		return _MkglobalFrameView_TableColumnpglobalFrameView_Table_columns (_coordref, MkglobalFrameView_TableColumn (_coordref, t));
	message (DEADLY, "a pglobalFrameView_Table_columns list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pglobalFrameView_Table_columns (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pglobalFrameView_Table_columns (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_globalFrameView_Table_columns)
	l =_ElempglobalFrameView_Table_columns (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_globalFrameView_Table_columns)
	r = _ElempglobalFrameView_Table_columns (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pglobalFrameView_Table_columns)r)->_desc2);
((_TPPLST_0pglobalFrameView_Table_columns)r)->_desc2 = ((_TPPLST_0pglobalFrameView_Table_columns)l)->_desc2;
((_TPPLST_0pglobalFrameView_Table_columns)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_Table_columns (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpglobalFrameView_Table_columns (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpglobalFrameView_Table_columns root;
	_TPPLST_0pglobalFrameView_Table_columns list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pglobalFrameView_Table_columns)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_globalFrameView_Table_columns)
		t = _ElempglobalFrameView_Table_columns (_coordref, t);
	remain = (_TPPLST_0pglobalFrameView_Table_columns)(((_TPPLST_0pglobalFrameView_Table_columns) t)->_desc2);
	((_TPPLST_0pglobalFrameView_Table_columns) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pglobalFrameView_Table_columns;
#else
list = (_TPPLST_0pglobalFrameView_Table_columns) TreeNodeAlloc (sizeof (struct _TPLST_0pglobalFrameView_Table_columns));
#endif
list->_prod = RULELST_0pglobalFrameView_Table_columns;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pglobalFrameView_Table_columns)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_globalFrameView_TableColumnpglobalFrameView_Table_columns)
	{_TPPLST_globalFrameView_TableColumnpglobalFrameView_Table_columns _currn = _currn;
	 _TERMACT_LST_globalFrameView_TableColumnpglobalFrameView_Table_columns;
	} this code has been moved to _ElempglobalFrameView_Table_columns*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpglobalFrameView_Table_columns;
#else
root = (_TPPpglobalFrameView_Table_columns) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_Table_columns));
#endif
root->_prod = RULEpglobalFrameView_Table_columns;
root->_desc1 = (_TSPLST_globalFrameView_Table_columns)list;
{ _TPPpglobalFrameView_Table_columns _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pglobalFrameView_Table_columns;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_TableRow (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpglobalFrameView_TableRow (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpglobalFrameView_TableRow _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_TableRow;
#else
_currn = (_TPPpglobalFrameView_TableRow) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_TableRow));
#endif
_currn->_prod = RULEpglobalFrameView_TableRow;
_currn->_desc1 = (_TSPglobalFrameView_TableRow_cells) MkglobalFrameView_TableRow_cells (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_TableRow: root of subtree no. 1 can not be made a globalFrameView_TableRow_cells node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_TableRow;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_TableRow */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_TableCellpglobalFrameView_TableRow_cells (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_TableCellpglobalFrameView_TableRow_cells (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_TableCellpglobalFrameView_TableRow_cells _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_TableCellpglobalFrameView_TableRow_cells;
#else
	_new = (_TPPLST_globalFrameView_TableCellpglobalFrameView_TableRow_cells) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_TableCellpglobalFrameView_TableRow_cells));
#endif
	_new->_prod = RULELST_globalFrameView_TableCellpglobalFrameView_TableRow_cells;
	_new->_desc1 = (_TSPglobalFrameView_TableCell)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_TableRow_cells)_new;
	{_TPPLST_globalFrameView_TableCellpglobalFrameView_TableRow_cells _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_TableCellpglobalFrameView_TableRow_cells;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempglobalFrameView_TableRow_cells (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempglobalFrameView_TableRow_cells (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBglobalFrameView_TableCell)) 
		return _MkglobalFrameView_TableCellpglobalFrameView_TableRow_cells (_coordref, MkglobalFrameView_TableCell (_coordref, t));
	message (DEADLY, "a pglobalFrameView_TableRow_cells list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pglobalFrameView_TableRow_cells (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pglobalFrameView_TableRow_cells (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_globalFrameView_TableRow_cells)
	l =_ElempglobalFrameView_TableRow_cells (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_globalFrameView_TableRow_cells)
	r = _ElempglobalFrameView_TableRow_cells (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pglobalFrameView_TableRow_cells)r)->_desc2);
((_TPPLST_0pglobalFrameView_TableRow_cells)r)->_desc2 = ((_TPPLST_0pglobalFrameView_TableRow_cells)l)->_desc2;
((_TPPLST_0pglobalFrameView_TableRow_cells)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_TableRow_cells (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpglobalFrameView_TableRow_cells (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpglobalFrameView_TableRow_cells root;
	_TPPLST_0pglobalFrameView_TableRow_cells list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pglobalFrameView_TableRow_cells)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_globalFrameView_TableRow_cells)
		t = _ElempglobalFrameView_TableRow_cells (_coordref, t);
	remain = (_TPPLST_0pglobalFrameView_TableRow_cells)(((_TPPLST_0pglobalFrameView_TableRow_cells) t)->_desc2);
	((_TPPLST_0pglobalFrameView_TableRow_cells) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pglobalFrameView_TableRow_cells;
#else
list = (_TPPLST_0pglobalFrameView_TableRow_cells) TreeNodeAlloc (sizeof (struct _TPLST_0pglobalFrameView_TableRow_cells));
#endif
list->_prod = RULELST_0pglobalFrameView_TableRow_cells;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pglobalFrameView_TableRow_cells)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_globalFrameView_TableCellpglobalFrameView_TableRow_cells)
	{_TPPLST_globalFrameView_TableCellpglobalFrameView_TableRow_cells _currn = _currn;
	 _TERMACT_LST_globalFrameView_TableCellpglobalFrameView_TableRow_cells;
	} this code has been moved to _ElempglobalFrameView_TableRow_cells*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpglobalFrameView_TableRow_cells;
#else
root = (_TPPpglobalFrameView_TableRow_cells) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_TableRow_cells));
#endif
root->_prod = RULEpglobalFrameView_TableRow_cells;
root->_desc1 = (_TSPLST_globalFrameView_TableRow_cells)list;
{ _TPPpglobalFrameView_TableRow_cells _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pglobalFrameView_TableRow_cells;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_TableColumn (POSITION *_coordref, int _TERM1)
#else
NODEPTR MkpglobalFrameView_TableColumn (_coordref, _TERM1)
	POSITION *_coordref;
	int _TERM1;
#endif
{	_TPPpglobalFrameView_TableColumn _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_TableColumn;
#else
_currn = (_TPPpglobalFrameView_TableColumn) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_TableColumn));
#endif
_currn->_prod = RULEpglobalFrameView_TableColumn;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_TableColumn;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_TableColumn */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_TableCell (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpglobalFrameView_TableCell (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpglobalFrameView_TableCell _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_TableCell;
#else
_currn = (_TPPpglobalFrameView_TableCell) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_TableCell));
#endif
_currn->_prod = RULEpglobalFrameView_TableCell;
_currn->_desc1 = (_TSPglobalFrameView_TableCell_columnRef) MkglobalFrameView_TableCell_columnRef (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_TableCell: root of subtree no. 1 can not be made a globalFrameView_TableCell_columnRef node ", 0, _coordref);
_currn->_desc2 = (_TSPglobalFrameView_TableCell_content) MkglobalFrameView_TableCell_content (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_TableCell: root of subtree no. 2 can not be made a globalFrameView_TableCell_content node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_TableCell;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_TableCell */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_TableCell_columnRef (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_TableCell_columnRef (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_TableCell_columnRef _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_TableCell_columnRef;
#else
_currn = (_TPPpglobalFrameView_TableCell_columnRef) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_TableCell_columnRef));
#endif
_currn->_prod = RULEpglobalFrameView_TableCell_columnRef;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_TableCell_columnRef;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_TableCell_columnRef */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_Frame (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpglobalFrameView_Frame (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpglobalFrameView_Frame _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_Frame;
#else
_currn = (_TPPpglobalFrameView_Frame) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_Frame));
#endif
_currn->_prod = RULEpglobalFrameView_Frame;
_currn->_desc1 = (_TSPglobalFrameView_Frame_components) MkglobalFrameView_Frame_components (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_Frame: root of subtree no. 1 can not be made a globalFrameView_Frame_components node ", 0, _coordref);
_currn->_desc2 = (_TSPglobalFrameView_Frame_styleSheetBlock) MkglobalFrameView_Frame_styleSheetBlock (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_Frame: root of subtree no. 2 can not be made a globalFrameView_Frame_styleSheetBlock node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_Frame;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_Frame */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_FormularpglobalFrameView_Frame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_FormularpglobalFrameView_Frame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_FormularpglobalFrameView_Frame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_FormularpglobalFrameView_Frame_components;
#else
	_new = (_TPPLST_globalFrameView_FormularpglobalFrameView_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_FormularpglobalFrameView_Frame_components));
#endif
	_new->_prod = RULELST_globalFrameView_FormularpglobalFrameView_Frame_components;
	_new->_desc1 = (_TSPglobalFrameView_Formular)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_Frame_components)_new;
	{_TPPLST_globalFrameView_FormularpglobalFrameView_Frame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_FormularpglobalFrameView_Frame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_ParagraphpglobalFrameView_Frame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_ParagraphpglobalFrameView_Frame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_ParagraphpglobalFrameView_Frame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_ParagraphpglobalFrameView_Frame_components;
#else
	_new = (_TPPLST_globalFrameView_ParagraphpglobalFrameView_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_ParagraphpglobalFrameView_Frame_components));
#endif
	_new->_prod = RULELST_globalFrameView_ParagraphpglobalFrameView_Frame_components;
	_new->_desc1 = (_TSPglobalFrameView_Paragraph)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_Frame_components)_new;
	{_TPPLST_globalFrameView_ParagraphpglobalFrameView_Frame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_ParagraphpglobalFrameView_Frame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_HeadingpglobalFrameView_Frame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_HeadingpglobalFrameView_Frame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_HeadingpglobalFrameView_Frame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_HeadingpglobalFrameView_Frame_components;
#else
	_new = (_TPPLST_globalFrameView_HeadingpglobalFrameView_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_HeadingpglobalFrameView_Frame_components));
#endif
	_new->_prod = RULELST_globalFrameView_HeadingpglobalFrameView_Frame_components;
	_new->_desc1 = (_TSPglobalFrameView_Heading)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_Frame_components)_new;
	{_TPPLST_globalFrameView_HeadingpglobalFrameView_Frame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_HeadingpglobalFrameView_Frame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_TablepglobalFrameView_Frame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_TablepglobalFrameView_Frame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_TablepglobalFrameView_Frame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_TablepglobalFrameView_Frame_components;
#else
	_new = (_TPPLST_globalFrameView_TablepglobalFrameView_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_TablepglobalFrameView_Frame_components));
#endif
	_new->_prod = RULELST_globalFrameView_TablepglobalFrameView_Frame_components;
	_new->_desc1 = (_TSPglobalFrameView_Table)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_Frame_components)_new;
	{_TPPLST_globalFrameView_TablepglobalFrameView_Frame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_TablepglobalFrameView_Frame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_FrameRefpglobalFrameView_Frame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_FrameRefpglobalFrameView_Frame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_FrameRefpglobalFrameView_Frame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_FrameRefpglobalFrameView_Frame_components;
#else
	_new = (_TPPLST_globalFrameView_FrameRefpglobalFrameView_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_FrameRefpglobalFrameView_Frame_components));
#endif
	_new->_prod = RULELST_globalFrameView_FrameRefpglobalFrameView_Frame_components;
	_new->_desc1 = (_TSPglobalFrameView_FrameRef)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_Frame_components)_new;
	{_TPPLST_globalFrameView_FrameRefpglobalFrameView_Frame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_FrameRefpglobalFrameView_Frame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_ListOutputElementpglobalFrameView_Frame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_ListOutputElementpglobalFrameView_Frame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_ListOutputElementpglobalFrameView_Frame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_ListOutputElementpglobalFrameView_Frame_components;
#else
	_new = (_TPPLST_globalFrameView_ListOutputElementpglobalFrameView_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_ListOutputElementpglobalFrameView_Frame_components));
#endif
	_new->_prod = RULELST_globalFrameView_ListOutputElementpglobalFrameView_Frame_components;
	_new->_desc1 = (_TSPglobalFrameView_ListOutputElement)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_Frame_components)_new;
	{_TPPLST_globalFrameView_ListOutputElementpglobalFrameView_Frame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_ListOutputElementpglobalFrameView_Frame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempglobalFrameView_Frame_components (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempglobalFrameView_Frame_components (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBglobalFrameView_Formular)) 
		return _MkglobalFrameView_FormularpglobalFrameView_Frame_components (_coordref, MkglobalFrameView_Formular (_coordref, t));
	if (IsSymb (t, SYMBglobalFrameView_Paragraph)) 
		return _MkglobalFrameView_ParagraphpglobalFrameView_Frame_components (_coordref, MkglobalFrameView_Paragraph (_coordref, t));
	if (IsSymb (t, SYMBglobalFrameView_Heading)) 
		return _MkglobalFrameView_HeadingpglobalFrameView_Frame_components (_coordref, MkglobalFrameView_Heading (_coordref, t));
	if (IsSymb (t, SYMBglobalFrameView_Table)) 
		return _MkglobalFrameView_TablepglobalFrameView_Frame_components (_coordref, MkglobalFrameView_Table (_coordref, t));
	if (IsSymb (t, SYMBglobalFrameView_FrameRef)) 
		return _MkglobalFrameView_FrameRefpglobalFrameView_Frame_components (_coordref, MkglobalFrameView_FrameRef (_coordref, t));
	if (IsSymb (t, SYMBglobalFrameView_ListOutputElement)) 
		return _MkglobalFrameView_ListOutputElementpglobalFrameView_Frame_components (_coordref, MkglobalFrameView_ListOutputElement (_coordref, t));
	message (DEADLY, "a pglobalFrameView_Frame_components list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pglobalFrameView_Frame_components (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pglobalFrameView_Frame_components (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_globalFrameView_Frame_components)
	l =_ElempglobalFrameView_Frame_components (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_globalFrameView_Frame_components)
	r = _ElempglobalFrameView_Frame_components (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pglobalFrameView_Frame_components)r)->_desc2);
((_TPPLST_0pglobalFrameView_Frame_components)r)->_desc2 = ((_TPPLST_0pglobalFrameView_Frame_components)l)->_desc2;
((_TPPLST_0pglobalFrameView_Frame_components)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_Frame_components (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpglobalFrameView_Frame_components (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpglobalFrameView_Frame_components root;
	_TPPLST_0pglobalFrameView_Frame_components list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pglobalFrameView_Frame_components)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_globalFrameView_Frame_components)
		t = _ElempglobalFrameView_Frame_components (_coordref, t);
	remain = (_TPPLST_0pglobalFrameView_Frame_components)(((_TPPLST_0pglobalFrameView_Frame_components) t)->_desc2);
	((_TPPLST_0pglobalFrameView_Frame_components) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pglobalFrameView_Frame_components;
#else
list = (_TPPLST_0pglobalFrameView_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_0pglobalFrameView_Frame_components));
#endif
list->_prod = RULELST_0pglobalFrameView_Frame_components;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pglobalFrameView_Frame_components)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_globalFrameView_FormularpglobalFrameView_Frame_components)
	{_TPPLST_globalFrameView_FormularpglobalFrameView_Frame_components _currn = _currn;
	 _TERMACT_LST_globalFrameView_FormularpglobalFrameView_Frame_components;
	} this code has been moved to _ElempglobalFrameView_Frame_components*/
	/*if (_currn->_prod == RULELST_globalFrameView_ParagraphpglobalFrameView_Frame_components)
	{_TPPLST_globalFrameView_ParagraphpglobalFrameView_Frame_components _currn = _currn;
	 _TERMACT_LST_globalFrameView_ParagraphpglobalFrameView_Frame_components;
	} this code has been moved to _ElempglobalFrameView_Frame_components*/
	/*if (_currn->_prod == RULELST_globalFrameView_HeadingpglobalFrameView_Frame_components)
	{_TPPLST_globalFrameView_HeadingpglobalFrameView_Frame_components _currn = _currn;
	 _TERMACT_LST_globalFrameView_HeadingpglobalFrameView_Frame_components;
	} this code has been moved to _ElempglobalFrameView_Frame_components*/
	/*if (_currn->_prod == RULELST_globalFrameView_TablepglobalFrameView_Frame_components)
	{_TPPLST_globalFrameView_TablepglobalFrameView_Frame_components _currn = _currn;
	 _TERMACT_LST_globalFrameView_TablepglobalFrameView_Frame_components;
	} this code has been moved to _ElempglobalFrameView_Frame_components*/
	/*if (_currn->_prod == RULELST_globalFrameView_FrameRefpglobalFrameView_Frame_components)
	{_TPPLST_globalFrameView_FrameRefpglobalFrameView_Frame_components _currn = _currn;
	 _TERMACT_LST_globalFrameView_FrameRefpglobalFrameView_Frame_components;
	} this code has been moved to _ElempglobalFrameView_Frame_components*/
	/*if (_currn->_prod == RULELST_globalFrameView_ListOutputElementpglobalFrameView_Frame_components)
	{_TPPLST_globalFrameView_ListOutputElementpglobalFrameView_Frame_components _currn = _currn;
	 _TERMACT_LST_globalFrameView_ListOutputElementpglobalFrameView_Frame_components;
	} this code has been moved to _ElempglobalFrameView_Frame_components*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpglobalFrameView_Frame_components;
#else
root = (_TPPpglobalFrameView_Frame_components) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_Frame_components));
#endif
root->_prod = RULEpglobalFrameView_Frame_components;
root->_desc1 = (_TSPLST_globalFrameView_Frame_components)list;
{ _TPPpglobalFrameView_Frame_components _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pglobalFrameView_Frame_components;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_Frame_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_Frame_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_Frame_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_Frame_styleSheetBlock;
#else
_currn = (_TPPpglobalFrameView_Frame_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_Frame_styleSheetBlock));
#endif
_currn->_prod = RULEpglobalFrameView_Frame_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_Frame_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_Frame_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_GlobalFrame (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpglobalFrameView_GlobalFrame (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpglobalFrameView_GlobalFrame _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_GlobalFrame;
#else
_currn = (_TPPpglobalFrameView_GlobalFrame) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_GlobalFrame));
#endif
_currn->_prod = RULEpglobalFrameView_GlobalFrame;
_currn->_desc1 = (_TSPglobalFrameView_GlobalFrame_name) MkglobalFrameView_GlobalFrame_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_GlobalFrame: root of subtree no. 1 can not be made a globalFrameView_GlobalFrame_name node ", 0, _coordref);
_currn->_desc2 = (_TSPglobalFrameView_GlobalFrame_components) MkglobalFrameView_GlobalFrame_components (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_GlobalFrame: root of subtree no. 2 can not be made a globalFrameView_GlobalFrame_components node ", 0, _coordref);
_currn->_desc3 = (_TSPglobalFrameView_GlobalFrame_setSize) MkglobalFrameView_GlobalFrame_setSize (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_GlobalFrame: root of subtree no. 3 can not be made a globalFrameView_GlobalFrame_setSize node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_GlobalFrame;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_GlobalFrame */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_GlobalFrame_name (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_GlobalFrame_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_GlobalFrame_name _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_GlobalFrame_name;
#else
_currn = (_TPPpglobalFrameView_GlobalFrame_name) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_GlobalFrame_name));
#endif
_currn->_prod = RULEpglobalFrameView_GlobalFrame_name;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_GlobalFrame_name;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_GlobalFrame_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_FormularpglobalFrameView_GlobalFrame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_FormularpglobalFrameView_GlobalFrame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_FormularpglobalFrameView_GlobalFrame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_FormularpglobalFrameView_GlobalFrame_components;
#else
	_new = (_TPPLST_globalFrameView_FormularpglobalFrameView_GlobalFrame_components) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_FormularpglobalFrameView_GlobalFrame_components));
#endif
	_new->_prod = RULELST_globalFrameView_FormularpglobalFrameView_GlobalFrame_components;
	_new->_desc1 = (_TSPglobalFrameView_Formular)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_GlobalFrame_components)_new;
	{_TPPLST_globalFrameView_FormularpglobalFrameView_GlobalFrame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_FormularpglobalFrameView_GlobalFrame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_ParagraphpglobalFrameView_GlobalFrame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_ParagraphpglobalFrameView_GlobalFrame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_ParagraphpglobalFrameView_GlobalFrame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_ParagraphpglobalFrameView_GlobalFrame_components;
#else
	_new = (_TPPLST_globalFrameView_ParagraphpglobalFrameView_GlobalFrame_components) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_ParagraphpglobalFrameView_GlobalFrame_components));
#endif
	_new->_prod = RULELST_globalFrameView_ParagraphpglobalFrameView_GlobalFrame_components;
	_new->_desc1 = (_TSPglobalFrameView_Paragraph)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_GlobalFrame_components)_new;
	{_TPPLST_globalFrameView_ParagraphpglobalFrameView_GlobalFrame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_ParagraphpglobalFrameView_GlobalFrame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_HeadingpglobalFrameView_GlobalFrame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_HeadingpglobalFrameView_GlobalFrame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_HeadingpglobalFrameView_GlobalFrame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_HeadingpglobalFrameView_GlobalFrame_components;
#else
	_new = (_TPPLST_globalFrameView_HeadingpglobalFrameView_GlobalFrame_components) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_HeadingpglobalFrameView_GlobalFrame_components));
#endif
	_new->_prod = RULELST_globalFrameView_HeadingpglobalFrameView_GlobalFrame_components;
	_new->_desc1 = (_TSPglobalFrameView_Heading)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_GlobalFrame_components)_new;
	{_TPPLST_globalFrameView_HeadingpglobalFrameView_GlobalFrame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_HeadingpglobalFrameView_GlobalFrame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_TablepglobalFrameView_GlobalFrame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_TablepglobalFrameView_GlobalFrame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_TablepglobalFrameView_GlobalFrame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_TablepglobalFrameView_GlobalFrame_components;
#else
	_new = (_TPPLST_globalFrameView_TablepglobalFrameView_GlobalFrame_components) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_TablepglobalFrameView_GlobalFrame_components));
#endif
	_new->_prod = RULELST_globalFrameView_TablepglobalFrameView_GlobalFrame_components;
	_new->_desc1 = (_TSPglobalFrameView_Table)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_GlobalFrame_components)_new;
	{_TPPLST_globalFrameView_TablepglobalFrameView_GlobalFrame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_TablepglobalFrameView_GlobalFrame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_FrameRefpglobalFrameView_GlobalFrame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_FrameRefpglobalFrameView_GlobalFrame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_FrameRefpglobalFrameView_GlobalFrame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_FrameRefpglobalFrameView_GlobalFrame_components;
#else
	_new = (_TPPLST_globalFrameView_FrameRefpglobalFrameView_GlobalFrame_components) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_FrameRefpglobalFrameView_GlobalFrame_components));
#endif
	_new->_prod = RULELST_globalFrameView_FrameRefpglobalFrameView_GlobalFrame_components;
	_new->_desc1 = (_TSPglobalFrameView_FrameRef)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_GlobalFrame_components)_new;
	{_TPPLST_globalFrameView_FrameRefpglobalFrameView_GlobalFrame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_FrameRefpglobalFrameView_GlobalFrame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_ListOutputElementpglobalFrameView_GlobalFrame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_ListOutputElementpglobalFrameView_GlobalFrame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_ListOutputElementpglobalFrameView_GlobalFrame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_ListOutputElementpglobalFrameView_GlobalFrame_components;
#else
	_new = (_TPPLST_globalFrameView_ListOutputElementpglobalFrameView_GlobalFrame_components) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_ListOutputElementpglobalFrameView_GlobalFrame_components));
#endif
	_new->_prod = RULELST_globalFrameView_ListOutputElementpglobalFrameView_GlobalFrame_components;
	_new->_desc1 = (_TSPglobalFrameView_ListOutputElement)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_GlobalFrame_components)_new;
	{_TPPLST_globalFrameView_ListOutputElementpglobalFrameView_GlobalFrame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_ListOutputElementpglobalFrameView_GlobalFrame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempglobalFrameView_GlobalFrame_components (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempglobalFrameView_GlobalFrame_components (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBglobalFrameView_Formular)) 
		return _MkglobalFrameView_FormularpglobalFrameView_GlobalFrame_components (_coordref, MkglobalFrameView_Formular (_coordref, t));
	if (IsSymb (t, SYMBglobalFrameView_Paragraph)) 
		return _MkglobalFrameView_ParagraphpglobalFrameView_GlobalFrame_components (_coordref, MkglobalFrameView_Paragraph (_coordref, t));
	if (IsSymb (t, SYMBglobalFrameView_Heading)) 
		return _MkglobalFrameView_HeadingpglobalFrameView_GlobalFrame_components (_coordref, MkglobalFrameView_Heading (_coordref, t));
	if (IsSymb (t, SYMBglobalFrameView_Table)) 
		return _MkglobalFrameView_TablepglobalFrameView_GlobalFrame_components (_coordref, MkglobalFrameView_Table (_coordref, t));
	if (IsSymb (t, SYMBglobalFrameView_FrameRef)) 
		return _MkglobalFrameView_FrameRefpglobalFrameView_GlobalFrame_components (_coordref, MkglobalFrameView_FrameRef (_coordref, t));
	if (IsSymb (t, SYMBglobalFrameView_ListOutputElement)) 
		return _MkglobalFrameView_ListOutputElementpglobalFrameView_GlobalFrame_components (_coordref, MkglobalFrameView_ListOutputElement (_coordref, t));
	message (DEADLY, "a pglobalFrameView_GlobalFrame_components list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pglobalFrameView_GlobalFrame_components (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pglobalFrameView_GlobalFrame_components (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_globalFrameView_GlobalFrame_components)
	l =_ElempglobalFrameView_GlobalFrame_components (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_globalFrameView_GlobalFrame_components)
	r = _ElempglobalFrameView_GlobalFrame_components (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pglobalFrameView_GlobalFrame_components)r)->_desc2);
((_TPPLST_0pglobalFrameView_GlobalFrame_components)r)->_desc2 = ((_TPPLST_0pglobalFrameView_GlobalFrame_components)l)->_desc2;
((_TPPLST_0pglobalFrameView_GlobalFrame_components)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_GlobalFrame_components (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpglobalFrameView_GlobalFrame_components (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpglobalFrameView_GlobalFrame_components root;
	_TPPLST_0pglobalFrameView_GlobalFrame_components list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pglobalFrameView_GlobalFrame_components)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_globalFrameView_GlobalFrame_components)
		t = _ElempglobalFrameView_GlobalFrame_components (_coordref, t);
	remain = (_TPPLST_0pglobalFrameView_GlobalFrame_components)(((_TPPLST_0pglobalFrameView_GlobalFrame_components) t)->_desc2);
	((_TPPLST_0pglobalFrameView_GlobalFrame_components) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pglobalFrameView_GlobalFrame_components;
#else
list = (_TPPLST_0pglobalFrameView_GlobalFrame_components) TreeNodeAlloc (sizeof (struct _TPLST_0pglobalFrameView_GlobalFrame_components));
#endif
list->_prod = RULELST_0pglobalFrameView_GlobalFrame_components;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pglobalFrameView_GlobalFrame_components)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_globalFrameView_FormularpglobalFrameView_GlobalFrame_components)
	{_TPPLST_globalFrameView_FormularpglobalFrameView_GlobalFrame_components _currn = _currn;
	 _TERMACT_LST_globalFrameView_FormularpglobalFrameView_GlobalFrame_components;
	} this code has been moved to _ElempglobalFrameView_GlobalFrame_components*/
	/*if (_currn->_prod == RULELST_globalFrameView_ParagraphpglobalFrameView_GlobalFrame_components)
	{_TPPLST_globalFrameView_ParagraphpglobalFrameView_GlobalFrame_components _currn = _currn;
	 _TERMACT_LST_globalFrameView_ParagraphpglobalFrameView_GlobalFrame_components;
	} this code has been moved to _ElempglobalFrameView_GlobalFrame_components*/
	/*if (_currn->_prod == RULELST_globalFrameView_HeadingpglobalFrameView_GlobalFrame_components)
	{_TPPLST_globalFrameView_HeadingpglobalFrameView_GlobalFrame_components _currn = _currn;
	 _TERMACT_LST_globalFrameView_HeadingpglobalFrameView_GlobalFrame_components;
	} this code has been moved to _ElempglobalFrameView_GlobalFrame_components*/
	/*if (_currn->_prod == RULELST_globalFrameView_TablepglobalFrameView_GlobalFrame_components)
	{_TPPLST_globalFrameView_TablepglobalFrameView_GlobalFrame_components _currn = _currn;
	 _TERMACT_LST_globalFrameView_TablepglobalFrameView_GlobalFrame_components;
	} this code has been moved to _ElempglobalFrameView_GlobalFrame_components*/
	/*if (_currn->_prod == RULELST_globalFrameView_FrameRefpglobalFrameView_GlobalFrame_components)
	{_TPPLST_globalFrameView_FrameRefpglobalFrameView_GlobalFrame_components _currn = _currn;
	 _TERMACT_LST_globalFrameView_FrameRefpglobalFrameView_GlobalFrame_components;
	} this code has been moved to _ElempglobalFrameView_GlobalFrame_components*/
	/*if (_currn->_prod == RULELST_globalFrameView_ListOutputElementpglobalFrameView_GlobalFrame_components)
	{_TPPLST_globalFrameView_ListOutputElementpglobalFrameView_GlobalFrame_components _currn = _currn;
	 _TERMACT_LST_globalFrameView_ListOutputElementpglobalFrameView_GlobalFrame_components;
	} this code has been moved to _ElempglobalFrameView_GlobalFrame_components*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpglobalFrameView_GlobalFrame_components;
#else
root = (_TPPpglobalFrameView_GlobalFrame_components) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_GlobalFrame_components));
#endif
root->_prod = RULEpglobalFrameView_GlobalFrame_components;
root->_desc1 = (_TSPLST_globalFrameView_GlobalFrame_components)list;
{ _TPPpglobalFrameView_GlobalFrame_components _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pglobalFrameView_GlobalFrame_components;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_GlobalFrame_setSize (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_GlobalFrame_setSize (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_GlobalFrame_setSize _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_GlobalFrame_setSize;
#else
_currn = (_TPPpglobalFrameView_GlobalFrame_setSize) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_GlobalFrame_setSize));
#endif
_currn->_prod = RULEpglobalFrameView_GlobalFrame_setSize;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_GlobalFrame_setSize;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_GlobalFrame_setSize */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FrameRef (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpglobalFrameView_FrameRef (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpglobalFrameView_FrameRef _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FrameRef;
#else
_currn = (_TPPpglobalFrameView_FrameRef) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FrameRef));
#endif
_currn->_prod = RULEpglobalFrameView_FrameRef;
_currn->_desc1 = (_TSPglobalFrameView_FrameRef_styleSheetBlock) MkglobalFrameView_FrameRef_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FrameRef: root of subtree no. 1 can not be made a globalFrameView_FrameRef_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSPglobalFrameView_FrameRef_frame) MkglobalFrameView_FrameRef_frame (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_FrameRef: root of subtree no. 2 can not be made a globalFrameView_FrameRef_frame node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FrameRef;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FrameRef */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FrameRef_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FrameRef_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FrameRef_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FrameRef_styleSheetBlock;
#else
_currn = (_TPPpglobalFrameView_FrameRef_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FrameRef_styleSheetBlock));
#endif
_currn->_prod = RULEpglobalFrameView_FrameRef_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FrameRef_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FrameRef_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_FrameRef_frame (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_FrameRef_frame (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_FrameRef_frame _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_FrameRef_frame;
#else
_currn = (_TPPpglobalFrameView_FrameRef_frame) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_FrameRef_frame));
#endif
_currn->_prod = RULEpglobalFrameView_FrameRef_frame;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_FrameRef_frame;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_FrameRef_frame */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_OutputElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpglobalFrameView_OutputElement (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpglobalFrameView_OutputElement _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_OutputElement;
#else
_currn = (_TPPpglobalFrameView_OutputElement) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_OutputElement));
#endif
_currn->_prod = RULEpglobalFrameView_OutputElement;
_currn->_desc1 = (_TSPglobalFrameView_OutputElement_link) MkglobalFrameView_OutputElement_link (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_OutputElement: root of subtree no. 1 can not be made a globalFrameView_OutputElement_link node ", 0, _coordref);
_currn->_desc2 = (_TSPglobalFrameView_OutputElement_content) MkglobalFrameView_OutputElement_content (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_OutputElement: root of subtree no. 2 can not be made a globalFrameView_OutputElement_content node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_OutputElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_OutputElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_OutputElement_link (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_OutputElement_link (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_OutputElement_link _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_OutputElement_link;
#else
_currn = (_TPPpglobalFrameView_OutputElement_link) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_OutputElement_link));
#endif
_currn->_prod = RULEpglobalFrameView_OutputElement_link;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_OutputElement_link;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_OutputElement_link */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_OutputElement_content (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_OutputElement_content (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_OutputElement_content _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_OutputElement_content;
#else
_currn = (_TPPpglobalFrameView_OutputElement_content) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_OutputElement_content));
#endif
_currn->_prod = RULEpglobalFrameView_OutputElement_content;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_OutputElement_content;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_OutputElement_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_ListOutputElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkpglobalFrameView_ListOutputElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPpglobalFrameView_ListOutputElement _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_ListOutputElement;
#else
_currn = (_TPPpglobalFrameView_ListOutputElement) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_ListOutputElement));
#endif
_currn->_prod = RULEpglobalFrameView_ListOutputElement;
_currn->_desc1 = (_TSPglobalFrameView_ListOutputElement_styleSheetBlock) MkglobalFrameView_ListOutputElement_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_ListOutputElement: root of subtree no. 1 can not be made a globalFrameView_ListOutputElement_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSPglobalFrameView_ListOutputElement_dataSourceActualParams) MkglobalFrameView_ListOutputElement_dataSourceActualParams (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_ListOutputElement: root of subtree no. 2 can not be made a globalFrameView_ListOutputElement_dataSourceActualParams node ", 0, _coordref);
_currn->_desc3 = (_TSPglobalFrameView_ListOutputElement_dataSource) MkglobalFrameView_ListOutputElement_dataSource (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_ListOutputElement: root of subtree no. 3 can not be made a globalFrameView_ListOutputElement_dataSource node ", 0, _coordref);
_currn->_desc4 = (_TSPglobalFrameView_ListOutputElement_maxNumberOfElements) MkglobalFrameView_ListOutputElement_maxNumberOfElements (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_ListOutputElement: root of subtree no. 4 can not be made a globalFrameView_ListOutputElement_maxNumberOfElements node ", 0, _coordref);
_currn->_desc5 = (_TSPglobalFrameView_ListOutputElement_navigationElement) MkglobalFrameView_ListOutputElement_navigationElement (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pglobalFrameView_ListOutputElement: root of subtree no. 5 can not be made a globalFrameView_ListOutputElement_navigationElement node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_ListOutputElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_ListOutputElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_ListOutputElement_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_ListOutputElement_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_ListOutputElement_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_ListOutputElement_styleSheetBlock;
#else
_currn = (_TPPpglobalFrameView_ListOutputElement_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_ListOutputElement_styleSheetBlock));
#endif
_currn->_prod = RULEpglobalFrameView_ListOutputElement_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_ListOutputElement_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_ListOutputElement_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkglobalFrameView_DataSourceActualParampglobalFrameView_ListOutputElement_dataSourceActualParams (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkglobalFrameView_DataSourceActualParampglobalFrameView_ListOutputElement_dataSourceActualParams (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_globalFrameView_DataSourceActualParampglobalFrameView_ListOutputElement_dataSourceActualParams _new;
#ifdef __cplusplus
	_new = new _TPLST_globalFrameView_DataSourceActualParampglobalFrameView_ListOutputElement_dataSourceActualParams;
#else
	_new = (_TPPLST_globalFrameView_DataSourceActualParampglobalFrameView_ListOutputElement_dataSourceActualParams) TreeNodeAlloc (sizeof (struct _TPLST_globalFrameView_DataSourceActualParampglobalFrameView_ListOutputElement_dataSourceActualParams));
#endif
	_new->_prod = RULELST_globalFrameView_DataSourceActualParampglobalFrameView_ListOutputElement_dataSourceActualParams;
	_new->_desc1 = (_TSPglobalFrameView_DataSourceActualParam)elem;
	_new->_desc2 = (_TSPLST_globalFrameView_ListOutputElement_dataSourceActualParams)_new;
	{_TPPLST_globalFrameView_DataSourceActualParampglobalFrameView_ListOutputElement_dataSourceActualParams _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_globalFrameView_DataSourceActualParampglobalFrameView_ListOutputElement_dataSourceActualParams;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempglobalFrameView_ListOutputElement_dataSourceActualParams (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempglobalFrameView_ListOutputElement_dataSourceActualParams (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBglobalFrameView_DataSourceActualParam)) 
		return _MkglobalFrameView_DataSourceActualParampglobalFrameView_ListOutputElement_dataSourceActualParams (_coordref, MkglobalFrameView_DataSourceActualParam (_coordref, t));
	message (DEADLY, "a pglobalFrameView_ListOutputElement_dataSourceActualParams list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pglobalFrameView_ListOutputElement_dataSourceActualParams (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pglobalFrameView_ListOutputElement_dataSourceActualParams (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_globalFrameView_ListOutputElement_dataSourceActualParams)
	l =_ElempglobalFrameView_ListOutputElement_dataSourceActualParams (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_globalFrameView_ListOutputElement_dataSourceActualParams)
	r = _ElempglobalFrameView_ListOutputElement_dataSourceActualParams (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pglobalFrameView_ListOutputElement_dataSourceActualParams)r)->_desc2);
((_TPPLST_0pglobalFrameView_ListOutputElement_dataSourceActualParams)r)->_desc2 = ((_TPPLST_0pglobalFrameView_ListOutputElement_dataSourceActualParams)l)->_desc2;
((_TPPLST_0pglobalFrameView_ListOutputElement_dataSourceActualParams)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_ListOutputElement_dataSourceActualParams (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpglobalFrameView_ListOutputElement_dataSourceActualParams (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpglobalFrameView_ListOutputElement_dataSourceActualParams root;
	_TPPLST_0pglobalFrameView_ListOutputElement_dataSourceActualParams list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pglobalFrameView_ListOutputElement_dataSourceActualParams)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_globalFrameView_ListOutputElement_dataSourceActualParams)
		t = _ElempglobalFrameView_ListOutputElement_dataSourceActualParams (_coordref, t);
	remain = (_TPPLST_0pglobalFrameView_ListOutputElement_dataSourceActualParams)(((_TPPLST_0pglobalFrameView_ListOutputElement_dataSourceActualParams) t)->_desc2);
	((_TPPLST_0pglobalFrameView_ListOutputElement_dataSourceActualParams) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pglobalFrameView_ListOutputElement_dataSourceActualParams;
#else
list = (_TPPLST_0pglobalFrameView_ListOutputElement_dataSourceActualParams) TreeNodeAlloc (sizeof (struct _TPLST_0pglobalFrameView_ListOutputElement_dataSourceActualParams));
#endif
list->_prod = RULELST_0pglobalFrameView_ListOutputElement_dataSourceActualParams;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pglobalFrameView_ListOutputElement_dataSourceActualParams)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_globalFrameView_DataSourceActualParampglobalFrameView_ListOutputElement_dataSourceActualParams)
	{_TPPLST_globalFrameView_DataSourceActualParampglobalFrameView_ListOutputElement_dataSourceActualParams _currn = _currn;
	 _TERMACT_LST_globalFrameView_DataSourceActualParampglobalFrameView_ListOutputElement_dataSourceActualParams;
	} this code has been moved to _ElempglobalFrameView_ListOutputElement_dataSourceActualParams*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpglobalFrameView_ListOutputElement_dataSourceActualParams;
#else
root = (_TPPpglobalFrameView_ListOutputElement_dataSourceActualParams) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_ListOutputElement_dataSourceActualParams));
#endif
root->_prod = RULEpglobalFrameView_ListOutputElement_dataSourceActualParams;
root->_desc1 = (_TSPLST_globalFrameView_ListOutputElement_dataSourceActualParams)list;
{ _TPPpglobalFrameView_ListOutputElement_dataSourceActualParams _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pglobalFrameView_ListOutputElement_dataSourceActualParams;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_ListOutputElement_dataSource (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_ListOutputElement_dataSource (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_ListOutputElement_dataSource _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_ListOutputElement_dataSource;
#else
_currn = (_TPPpglobalFrameView_ListOutputElement_dataSource) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_ListOutputElement_dataSource));
#endif
_currn->_prod = RULEpglobalFrameView_ListOutputElement_dataSource;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_ListOutputElement_dataSource;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_ListOutputElement_dataSource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_ListOutputElement_maxNumberOfElements (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_ListOutputElement_maxNumberOfElements (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_ListOutputElement_maxNumberOfElements _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_ListOutputElement_maxNumberOfElements;
#else
_currn = (_TPPpglobalFrameView_ListOutputElement_maxNumberOfElements) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_ListOutputElement_maxNumberOfElements));
#endif
_currn->_prod = RULEpglobalFrameView_ListOutputElement_maxNumberOfElements;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_ListOutputElement_maxNumberOfElements;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_ListOutputElement_maxNumberOfElements */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpglobalFrameView_ListOutputElement_navigationElement (POSITION *_coordref)
#else
NODEPTR MkpglobalFrameView_ListOutputElement_navigationElement (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpglobalFrameView_ListOutputElement_navigationElement _currn;
#ifdef __cplusplus
_currn = new _TPpglobalFrameView_ListOutputElement_navigationElement;
#else
_currn = (_TPPpglobalFrameView_ListOutputElement_navigationElement) TreeNodeAlloc (sizeof (struct _TPpglobalFrameView_ListOutputElement_navigationElement));
#endif
_currn->_prod = RULEpglobalFrameView_ListOutputElement_navigationElement;
_SETCOORD(_currn)
_TERMACT_pglobalFrameView_ListOutputElement_navigationElement;
return ( (NODEPTR) _currn);
}/* MkpglobalFrameView_ListOutputElement_navigationElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpVIEWROOT_metaInformation (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR MkpVIEWROOT_metaInformation (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPpVIEWROOT_metaInformation _currn;
#ifdef __cplusplus
_currn = new _TPpVIEWROOT_metaInformation;
#else
_currn = (_TPPpVIEWROOT_metaInformation) TreeNodeAlloc (sizeof (struct _TPpVIEWROOT_metaInformation));
#endif
_currn->_prod = RULEpVIEWROOT_metaInformation;
_currn->_desc1 = (_TSPmetaInformation_MetaInformation) MkmetaInformation_MetaInformation (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pVIEWROOT_metaInformation: root of subtree no. 1 can not be made a metaInformation_MetaInformation node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_pVIEWROOT_metaInformation;
return ( (NODEPTR) _currn);
}/* MkpVIEWROOT_metaInformation */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpmetaInformation_MetaInformation (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkpmetaInformation_MetaInformation (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPpmetaInformation_MetaInformation _currn;
#ifdef __cplusplus
_currn = new _TPpmetaInformation_MetaInformation;
#else
_currn = (_TPPpmetaInformation_MetaInformation) TreeNodeAlloc (sizeof (struct _TPpmetaInformation_MetaInformation));
#endif
_currn->_prod = RULEpmetaInformation_MetaInformation;
_currn->_desc1 = (_TSPmetaInformation_MetaInformation_keyword1) MkmetaInformation_MetaInformation_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pmetaInformation_MetaInformation: root of subtree no. 1 can not be made a metaInformation_MetaInformation_keyword1 node ", 0, _coordref);
_currn->_desc2 = (_TSPmetaInformation_MetaInformation_keyword2) MkmetaInformation_MetaInformation_keyword2 (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pmetaInformation_MetaInformation: root of subtree no. 2 can not be made a metaInformation_MetaInformation_keyword2 node ", 0, _coordref);
_currn->_desc3 = (_TSPmetaInformation_MetaInformation_keyword3) MkmetaInformation_MetaInformation_keyword3 (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pmetaInformation_MetaInformation: root of subtree no. 3 can not be made a metaInformation_MetaInformation_keyword3 node ", 0, _coordref);
_currn->_desc4 = (_TSPmetaInformation_MetaInformation_elements) MkmetaInformation_MetaInformation_elements (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pmetaInformation_MetaInformation: root of subtree no. 4 can not be made a metaInformation_MetaInformation_elements node ", 0, _coordref);
_currn->_desc5 = (_TSPmetaInformation_MetaInformation_keyword4) MkmetaInformation_MetaInformation_keyword4 (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pmetaInformation_MetaInformation: root of subtree no. 5 can not be made a metaInformation_MetaInformation_keyword4 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pmetaInformation_MetaInformation;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpmetaInformation_MetaInformation */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpmetaInformation_MetaInformation_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpmetaInformation_MetaInformation_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpmetaInformation_MetaInformation_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpmetaInformation_MetaInformation_keyword1;
#else
_currn = (_TPPpmetaInformation_MetaInformation_keyword1) TreeNodeAlloc (sizeof (struct _TPpmetaInformation_MetaInformation_keyword1));
#endif
_currn->_prod = RULEpmetaInformation_MetaInformation_keyword1;
_SETCOORD(_currn)
_TERMACT_pmetaInformation_MetaInformation_keyword1;
return ( (NODEPTR) _currn);
}/* MkpmetaInformation_MetaInformation_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpmetaInformation_MetaInformation_keyword2 (POSITION *_coordref)
#else
NODEPTR MkpmetaInformation_MetaInformation_keyword2 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpmetaInformation_MetaInformation_keyword2 _currn;
#ifdef __cplusplus
_currn = new _TPpmetaInformation_MetaInformation_keyword2;
#else
_currn = (_TPPpmetaInformation_MetaInformation_keyword2) TreeNodeAlloc (sizeof (struct _TPpmetaInformation_MetaInformation_keyword2));
#endif
_currn->_prod = RULEpmetaInformation_MetaInformation_keyword2;
_SETCOORD(_currn)
_TERMACT_pmetaInformation_MetaInformation_keyword2;
return ( (NODEPTR) _currn);
}/* MkpmetaInformation_MetaInformation_keyword2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpmetaInformation_MetaInformation_keyword3 (POSITION *_coordref)
#else
NODEPTR MkpmetaInformation_MetaInformation_keyword3 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpmetaInformation_MetaInformation_keyword3 _currn;
#ifdef __cplusplus
_currn = new _TPpmetaInformation_MetaInformation_keyword3;
#else
_currn = (_TPPpmetaInformation_MetaInformation_keyword3) TreeNodeAlloc (sizeof (struct _TPpmetaInformation_MetaInformation_keyword3));
#endif
_currn->_prod = RULEpmetaInformation_MetaInformation_keyword3;
_SETCOORD(_currn)
_TERMACT_pmetaInformation_MetaInformation_keyword3;
return ( (NODEPTR) _currn);
}/* MkpmetaInformation_MetaInformation_keyword3 */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkmetaInformation_MetaInformationElementpmetaInformation_MetaInformation_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkmetaInformation_MetaInformationElementpmetaInformation_MetaInformation_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_metaInformation_MetaInformationElementpmetaInformation_MetaInformation_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_metaInformation_MetaInformationElementpmetaInformation_MetaInformation_elements;
#else
	_new = (_TPPLST_metaInformation_MetaInformationElementpmetaInformation_MetaInformation_elements) TreeNodeAlloc (sizeof (struct _TPLST_metaInformation_MetaInformationElementpmetaInformation_MetaInformation_elements));
#endif
	_new->_prod = RULELST_metaInformation_MetaInformationElementpmetaInformation_MetaInformation_elements;
	_new->_desc1 = (_TSPmetaInformation_MetaInformationElement)elem;
	_new->_desc2 = (_TSPLST_metaInformation_MetaInformation_elements)_new;
	{_TPPLST_metaInformation_MetaInformationElementpmetaInformation_MetaInformation_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_metaInformation_MetaInformationElementpmetaInformation_MetaInformation_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempmetaInformation_MetaInformation_elements (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempmetaInformation_MetaInformation_elements (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBmetaInformation_MetaInformationElement)) 
		return _MkmetaInformation_MetaInformationElementpmetaInformation_MetaInformation_elements (_coordref, MkmetaInformation_MetaInformationElement (_coordref, t));
	message (DEADLY, "a pmetaInformation_MetaInformation_elements list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pmetaInformation_MetaInformation_elements (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pmetaInformation_MetaInformation_elements (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_metaInformation_MetaInformation_elements)
	l =_ElempmetaInformation_MetaInformation_elements (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_metaInformation_MetaInformation_elements)
	r = _ElempmetaInformation_MetaInformation_elements (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pmetaInformation_MetaInformation_elements)r)->_desc2);
((_TPPLST_0pmetaInformation_MetaInformation_elements)r)->_desc2 = ((_TPPLST_0pmetaInformation_MetaInformation_elements)l)->_desc2;
((_TPPLST_0pmetaInformation_MetaInformation_elements)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpmetaInformation_MetaInformation_elements (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpmetaInformation_MetaInformation_elements (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpmetaInformation_MetaInformation_elements root;
	_TPPLST_0pmetaInformation_MetaInformation_elements list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pmetaInformation_MetaInformation_elements)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_metaInformation_MetaInformation_elements)
		t = _ElempmetaInformation_MetaInformation_elements (_coordref, t);
	remain = (_TPPLST_0pmetaInformation_MetaInformation_elements)(((_TPPLST_0pmetaInformation_MetaInformation_elements) t)->_desc2);
	((_TPPLST_0pmetaInformation_MetaInformation_elements) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pmetaInformation_MetaInformation_elements;
#else
list = (_TPPLST_0pmetaInformation_MetaInformation_elements) TreeNodeAlloc (sizeof (struct _TPLST_0pmetaInformation_MetaInformation_elements));
#endif
list->_prod = RULELST_0pmetaInformation_MetaInformation_elements;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pmetaInformation_MetaInformation_elements)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_metaInformation_MetaInformationElementpmetaInformation_MetaInformation_elements)
	{_TPPLST_metaInformation_MetaInformationElementpmetaInformation_MetaInformation_elements _currn = _currn;
	 _TERMACT_LST_metaInformation_MetaInformationElementpmetaInformation_MetaInformation_elements;
	} this code has been moved to _ElempmetaInformation_MetaInformation_elements*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpmetaInformation_MetaInformation_elements;
#else
root = (_TPPpmetaInformation_MetaInformation_elements) TreeNodeAlloc (sizeof (struct _TPpmetaInformation_MetaInformation_elements));
#endif
root->_prod = RULEpmetaInformation_MetaInformation_elements;
root->_desc1 = (_TSPLST_metaInformation_MetaInformation_elements)list;
{ _TPPpmetaInformation_MetaInformation_elements _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pmetaInformation_MetaInformation_elements;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpmetaInformation_MetaInformation_keyword4 (POSITION *_coordref)
#else
NODEPTR MkpmetaInformation_MetaInformation_keyword4 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpmetaInformation_MetaInformation_keyword4 _currn;
#ifdef __cplusplus
_currn = new _TPpmetaInformation_MetaInformation_keyword4;
#else
_currn = (_TPPpmetaInformation_MetaInformation_keyword4) TreeNodeAlloc (sizeof (struct _TPpmetaInformation_MetaInformation_keyword4));
#endif
_currn->_prod = RULEpmetaInformation_MetaInformation_keyword4;
_SETCOORD(_currn)
_TERMACT_pmetaInformation_MetaInformation_keyword4;
return ( (NODEPTR) _currn);
}/* MkpmetaInformation_MetaInformation_keyword4 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpmetaInformation_MetaInformationElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4)
#else
NODEPTR MkpmetaInformation_MetaInformationElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
#endif
{	_TPPpmetaInformation_MetaInformationElement _currn;
#ifdef __cplusplus
_currn = new _TPpmetaInformation_MetaInformationElement;
#else
_currn = (_TPPpmetaInformation_MetaInformationElement) TreeNodeAlloc (sizeof (struct _TPpmetaInformation_MetaInformationElement));
#endif
_currn->_prod = RULEpmetaInformation_MetaInformationElement;
_currn->_desc1 = (_TSPmetaInformation_MetaInformationElement_property) MkmetaInformation_MetaInformationElement_property (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pmetaInformation_MetaInformationElement: root of subtree no. 1 can not be made a metaInformation_MetaInformationElement_property node ", 0, _coordref);
_currn->_desc2 = (_TSPmetaInformation_MetaInformationElement_keyword1) MkmetaInformation_MetaInformationElement_keyword1 (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pmetaInformation_MetaInformationElement: root of subtree no. 2 can not be made a metaInformation_MetaInformationElement_keyword1 node ", 0, _coordref);
_currn->_desc3 = (_TSPmetaInformation_MetaInformationElement_value) MkmetaInformation_MetaInformationElement_value (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pmetaInformation_MetaInformationElement: root of subtree no. 3 can not be made a metaInformation_MetaInformationElement_value node ", 0, _coordref);
_currn->_desc4 = (_TSPmetaInformation_MetaInformationElement_keyword2) MkmetaInformation_MetaInformationElement_keyword2 (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pmetaInformation_MetaInformationElement: root of subtree no. 4 can not be made a metaInformation_MetaInformationElement_keyword2 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pmetaInformation_MetaInformationElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpmetaInformation_MetaInformationElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpmetaInformation_MetaInformationElement_property (POSITION *_coordref)
#else
NODEPTR MkpmetaInformation_MetaInformationElement_property (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpmetaInformation_MetaInformationElement_property _currn;
#ifdef __cplusplus
_currn = new _TPpmetaInformation_MetaInformationElement_property;
#else
_currn = (_TPPpmetaInformation_MetaInformationElement_property) TreeNodeAlloc (sizeof (struct _TPpmetaInformation_MetaInformationElement_property));
#endif
_currn->_prod = RULEpmetaInformation_MetaInformationElement_property;
_SETCOORD(_currn)
_TERMACT_pmetaInformation_MetaInformationElement_property;
return ( (NODEPTR) _currn);
}/* MkpmetaInformation_MetaInformationElement_property */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpmetaInformation_MetaInformationElement_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpmetaInformation_MetaInformationElement_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpmetaInformation_MetaInformationElement_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpmetaInformation_MetaInformationElement_keyword1;
#else
_currn = (_TPPpmetaInformation_MetaInformationElement_keyword1) TreeNodeAlloc (sizeof (struct _TPpmetaInformation_MetaInformationElement_keyword1));
#endif
_currn->_prod = RULEpmetaInformation_MetaInformationElement_keyword1;
_SETCOORD(_currn)
_TERMACT_pmetaInformation_MetaInformationElement_keyword1;
return ( (NODEPTR) _currn);
}/* MkpmetaInformation_MetaInformationElement_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpmetaInformation_MetaInformationElement_value (POSITION *_coordref)
#else
NODEPTR MkpmetaInformation_MetaInformationElement_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpmetaInformation_MetaInformationElement_value _currn;
#ifdef __cplusplus
_currn = new _TPpmetaInformation_MetaInformationElement_value;
#else
_currn = (_TPPpmetaInformation_MetaInformationElement_value) TreeNodeAlloc (sizeof (struct _TPpmetaInformation_MetaInformationElement_value));
#endif
_currn->_prod = RULEpmetaInformation_MetaInformationElement_value;
_SETCOORD(_currn)
_TERMACT_pmetaInformation_MetaInformationElement_value;
return ( (NODEPTR) _currn);
}/* MkpmetaInformation_MetaInformationElement_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpmetaInformation_MetaInformationElement_keyword2 (POSITION *_coordref)
#else
NODEPTR MkpmetaInformation_MetaInformationElement_keyword2 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpmetaInformation_MetaInformationElement_keyword2 _currn;
#ifdef __cplusplus
_currn = new _TPpmetaInformation_MetaInformationElement_keyword2;
#else
_currn = (_TPPpmetaInformation_MetaInformationElement_keyword2) TreeNodeAlloc (sizeof (struct _TPpmetaInformation_MetaInformationElement_keyword2));
#endif
_currn->_prod = RULEpmetaInformation_MetaInformationElement_keyword2;
_SETCOORD(_currn)
_TERMACT_pmetaInformation_MetaInformationElement_keyword2;
return ( (NODEPTR) _currn);
}/* MkpmetaInformation_MetaInformationElement_keyword2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpVIEWROOT_navigationBarView (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR MkpVIEWROOT_navigationBarView (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPpVIEWROOT_navigationBarView _currn;
#ifdef __cplusplus
_currn = new _TPpVIEWROOT_navigationBarView;
#else
_currn = (_TPPpVIEWROOT_navigationBarView) TreeNodeAlloc (sizeof (struct _TPpVIEWROOT_navigationBarView));
#endif
_currn->_prod = RULEpVIEWROOT_navigationBarView;
_currn->_desc1 = (_TSPnavigationBarView_NavigationBar) MknavigationBarView_NavigationBar (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pVIEWROOT_navigationBarView: root of subtree no. 1 can not be made a navigationBarView_NavigationBar node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_pVIEWROOT_navigationBarView;
return ( (NODEPTR) _currn);
}/* MkpVIEWROOT_navigationBarView */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpnavigationBarView_NavigationBar (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpnavigationBarView_NavigationBar (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpnavigationBarView_NavigationBar _currn;
#ifdef __cplusplus
_currn = new _TPpnavigationBarView_NavigationBar;
#else
_currn = (_TPPpnavigationBarView_NavigationBar) TreeNodeAlloc (sizeof (struct _TPpnavigationBarView_NavigationBar));
#endif
_currn->_prod = RULEpnavigationBarView_NavigationBar;
_currn->_desc1 = (_TSPnavigationBarView_NavigationBar_styleSheetBlock) MknavigationBarView_NavigationBar_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pnavigationBarView_NavigationBar: root of subtree no. 1 can not be made a navigationBarView_NavigationBar_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSPnavigationBarView_NavigationBar_elements) MknavigationBarView_NavigationBar_elements (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pnavigationBarView_NavigationBar: root of subtree no. 2 can not be made a navigationBarView_NavigationBar_elements node ", 0, _coordref);
_currn->_desc3 = (_TSPnavigationBarView_NavigationBar_alignment) MknavigationBarView_NavigationBar_alignment (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pnavigationBarView_NavigationBar: root of subtree no. 3 can not be made a navigationBarView_NavigationBar_alignment node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pnavigationBarView_NavigationBar;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpnavigationBarView_NavigationBar */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpnavigationBarView_NavigationBar_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkpnavigationBarView_NavigationBar_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpnavigationBarView_NavigationBar_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpnavigationBarView_NavigationBar_styleSheetBlock;
#else
_currn = (_TPPpnavigationBarView_NavigationBar_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpnavigationBarView_NavigationBar_styleSheetBlock));
#endif
_currn->_prod = RULEpnavigationBarView_NavigationBar_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_pnavigationBarView_NavigationBar_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkpnavigationBarView_NavigationBar_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MknavigationBarView_NavigationBarElementpnavigationBarView_NavigationBar_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MknavigationBarView_NavigationBarElementpnavigationBarView_NavigationBar_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_navigationBarView_NavigationBarElementpnavigationBarView_NavigationBar_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_navigationBarView_NavigationBarElementpnavigationBarView_NavigationBar_elements;
#else
	_new = (_TPPLST_navigationBarView_NavigationBarElementpnavigationBarView_NavigationBar_elements) TreeNodeAlloc (sizeof (struct _TPLST_navigationBarView_NavigationBarElementpnavigationBarView_NavigationBar_elements));
#endif
	_new->_prod = RULELST_navigationBarView_NavigationBarElementpnavigationBarView_NavigationBar_elements;
	_new->_desc1 = (_TSPnavigationBarView_NavigationBarElement)elem;
	_new->_desc2 = (_TSPLST_navigationBarView_NavigationBar_elements)_new;
	{_TPPLST_navigationBarView_NavigationBarElementpnavigationBarView_NavigationBar_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_navigationBarView_NavigationBarElementpnavigationBarView_NavigationBar_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempnavigationBarView_NavigationBar_elements (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempnavigationBarView_NavigationBar_elements (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBnavigationBarView_NavigationBarElement)) 
		return _MknavigationBarView_NavigationBarElementpnavigationBarView_NavigationBar_elements (_coordref, MknavigationBarView_NavigationBarElement (_coordref, t));
	message (DEADLY, "a pnavigationBarView_NavigationBar_elements list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pnavigationBarView_NavigationBar_elements (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pnavigationBarView_NavigationBar_elements (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_navigationBarView_NavigationBar_elements)
	l =_ElempnavigationBarView_NavigationBar_elements (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_navigationBarView_NavigationBar_elements)
	r = _ElempnavigationBarView_NavigationBar_elements (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pnavigationBarView_NavigationBar_elements)r)->_desc2);
((_TPPLST_0pnavigationBarView_NavigationBar_elements)r)->_desc2 = ((_TPPLST_0pnavigationBarView_NavigationBar_elements)l)->_desc2;
((_TPPLST_0pnavigationBarView_NavigationBar_elements)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpnavigationBarView_NavigationBar_elements (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpnavigationBarView_NavigationBar_elements (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpnavigationBarView_NavigationBar_elements root;
	_TPPLST_0pnavigationBarView_NavigationBar_elements list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pnavigationBarView_NavigationBar_elements)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_navigationBarView_NavigationBar_elements)
		t = _ElempnavigationBarView_NavigationBar_elements (_coordref, t);
	remain = (_TPPLST_0pnavigationBarView_NavigationBar_elements)(((_TPPLST_0pnavigationBarView_NavigationBar_elements) t)->_desc2);
	((_TPPLST_0pnavigationBarView_NavigationBar_elements) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pnavigationBarView_NavigationBar_elements;
#else
list = (_TPPLST_0pnavigationBarView_NavigationBar_elements) TreeNodeAlloc (sizeof (struct _TPLST_0pnavigationBarView_NavigationBar_elements));
#endif
list->_prod = RULELST_0pnavigationBarView_NavigationBar_elements;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pnavigationBarView_NavigationBar_elements)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_navigationBarView_NavigationBarElementpnavigationBarView_NavigationBar_elements)
	{_TPPLST_navigationBarView_NavigationBarElementpnavigationBarView_NavigationBar_elements _currn = _currn;
	 _TERMACT_LST_navigationBarView_NavigationBarElementpnavigationBarView_NavigationBar_elements;
	} this code has been moved to _ElempnavigationBarView_NavigationBar_elements*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpnavigationBarView_NavigationBar_elements;
#else
root = (_TPPpnavigationBarView_NavigationBar_elements) TreeNodeAlloc (sizeof (struct _TPpnavigationBarView_NavigationBar_elements));
#endif
root->_prod = RULEpnavigationBarView_NavigationBar_elements;
root->_desc1 = (_TSPLST_navigationBarView_NavigationBar_elements)list;
{ _TPPpnavigationBarView_NavigationBar_elements _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pnavigationBarView_NavigationBar_elements;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpnavigationBarView_NavigationBar_alignment (POSITION *_coordref)
#else
NODEPTR MkpnavigationBarView_NavigationBar_alignment (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpnavigationBarView_NavigationBar_alignment _currn;
#ifdef __cplusplus
_currn = new _TPpnavigationBarView_NavigationBar_alignment;
#else
_currn = (_TPPpnavigationBarView_NavigationBar_alignment) TreeNodeAlloc (sizeof (struct _TPpnavigationBarView_NavigationBar_alignment));
#endif
_currn->_prod = RULEpnavigationBarView_NavigationBar_alignment;
_SETCOORD(_currn)
_TERMACT_pnavigationBarView_NavigationBar_alignment;
return ( (NODEPTR) _currn);
}/* MkpnavigationBarView_NavigationBar_alignment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpnavigationBarView_NavigationBarElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpnavigationBarView_NavigationBarElement (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpnavigationBarView_NavigationBarElement _currn;
#ifdef __cplusplus
_currn = new _TPpnavigationBarView_NavigationBarElement;
#else
_currn = (_TPPpnavigationBarView_NavigationBarElement) TreeNodeAlloc (sizeof (struct _TPpnavigationBarView_NavigationBarElement));
#endif
_currn->_prod = RULEpnavigationBarView_NavigationBarElement;
_currn->_desc1 = (_TSPnavigationBarView_NavigationBarElement_target) MknavigationBarView_NavigationBarElement_target (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pnavigationBarView_NavigationBarElement: root of subtree no. 1 can not be made a navigationBarView_NavigationBarElement_target node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pnavigationBarView_NavigationBarElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpnavigationBarView_NavigationBarElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpnavigationBarView_NavigationBarElement_target (POSITION *_coordref)
#else
NODEPTR MkpnavigationBarView_NavigationBarElement_target (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpnavigationBarView_NavigationBarElement_target _currn;
#ifdef __cplusplus
_currn = new _TPpnavigationBarView_NavigationBarElement_target;
#else
_currn = (_TPPpnavigationBarView_NavigationBarElement_target) TreeNodeAlloc (sizeof (struct _TPpnavigationBarView_NavigationBarElement_target));
#endif
_currn->_prod = RULEpnavigationBarView_NavigationBarElement_target;
_SETCOORD(_currn)
_TERMACT_pnavigationBarView_NavigationBarElement_target;
return ( (NODEPTR) _currn);
}/* MkpnavigationBarView_NavigationBarElement_target */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpVIEWROOT_rootView (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR MkpVIEWROOT_rootView (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPpVIEWROOT_rootView _currn;
#ifdef __cplusplus
_currn = new _TPpVIEWROOT_rootView;
#else
_currn = (_TPPpVIEWROOT_rootView) TreeNodeAlloc (sizeof (struct _TPpVIEWROOT_rootView));
#endif
_currn->_prod = RULEpVIEWROOT_rootView;
_currn->_desc1 = (_TSProotView_Root) MkrootView_Root (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pVIEWROOT_rootView: root of subtree no. 1 can not be made a rootView_Root node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_pVIEWROOT_rootView;
return ( (NODEPTR) _currn);
}/* MkpVIEWROOT_rootView */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebNodeFormalParam (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkprootView_WebNodeFormalParam (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPprootView_WebNodeFormalParam _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebNodeFormalParam;
#else
_currn = (_TPPprootView_WebNodeFormalParam) TreeNodeAlloc (sizeof (struct _TPprootView_WebNodeFormalParam));
#endif
_currn->_prod = RULEprootView_WebNodeFormalParam;
_currn->_desc1 = (_TSProotView_WebNodeFormalParam_name) MkrootView_WebNodeFormalParam_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebNodeFormalParam: root of subtree no. 1 can not be made a rootView_WebNodeFormalParam_name node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_WebNodeFormalParam_keyword1) MkrootView_WebNodeFormalParam_keyword1 (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebNodeFormalParam: root of subtree no. 2 can not be made a rootView_WebNodeFormalParam_keyword1 node ", 0, _coordref);
_currn->_desc3 = (_TSProotView_WebNodeFormalParam_type) MkrootView_WebNodeFormalParam_type (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebNodeFormalParam: root of subtree no. 3 can not be made a rootView_WebNodeFormalParam_type node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_WebNodeFormalParam;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_WebNodeFormalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebNodeFormalParam_name (POSITION *_coordref)
#else
NODEPTR MkprootView_WebNodeFormalParam_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_WebNodeFormalParam_name _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebNodeFormalParam_name;
#else
_currn = (_TPPprootView_WebNodeFormalParam_name) TreeNodeAlloc (sizeof (struct _TPprootView_WebNodeFormalParam_name));
#endif
_currn->_prod = RULEprootView_WebNodeFormalParam_name;
_SETCOORD(_currn)
_TERMACT_prootView_WebNodeFormalParam_name;
return ( (NODEPTR) _currn);
}/* MkprootView_WebNodeFormalParam_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebNodeFormalParam_keyword1 (POSITION *_coordref)
#else
NODEPTR MkprootView_WebNodeFormalParam_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_WebNodeFormalParam_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebNodeFormalParam_keyword1;
#else
_currn = (_TPPprootView_WebNodeFormalParam_keyword1) TreeNodeAlloc (sizeof (struct _TPprootView_WebNodeFormalParam_keyword1));
#endif
_currn->_prod = RULEprootView_WebNodeFormalParam_keyword1;
_SETCOORD(_currn)
_TERMACT_prootView_WebNodeFormalParam_keyword1;
return ( (NODEPTR) _currn);
}/* MkprootView_WebNodeFormalParam_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_DBStringprootView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_DBStringprootView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_DBStringprootView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_DBStringprootView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_rootView_DBStringprootView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_rootView_DBStringprootView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_rootView_DBStringprootView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSProotView_DBString)elem;
	_new->_desc2 = (_TSPLST_rootView_WebNodeFormalParam_type)_new;
	{_TPPLST_rootView_DBStringprootView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_DBStringprootView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_DBIntegerprootView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_DBIntegerprootView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_DBIntegerprootView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_DBIntegerprootView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_rootView_DBIntegerprootView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_rootView_DBIntegerprootView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_rootView_DBIntegerprootView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSProotView_DBInteger)elem;
	_new->_desc2 = (_TSPLST_rootView_WebNodeFormalParam_type)_new;
	{_TPPLST_rootView_DBIntegerprootView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_DBIntegerprootView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_DBFloatprootView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_DBFloatprootView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_DBFloatprootView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_DBFloatprootView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_rootView_DBFloatprootView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_rootView_DBFloatprootView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_rootView_DBFloatprootView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSProotView_DBFloat)elem;
	_new->_desc2 = (_TSPLST_rootView_WebNodeFormalParam_type)_new;
	{_TPPLST_rootView_DBFloatprootView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_DBFloatprootView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_DBDoubleprootView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_DBDoubleprootView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_DBDoubleprootView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_DBDoubleprootView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_rootView_DBDoubleprootView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_rootView_DBDoubleprootView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_rootView_DBDoubleprootView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSProotView_DBDouble)elem;
	_new->_desc2 = (_TSPLST_rootView_WebNodeFormalParam_type)_new;
	{_TPPLST_rootView_DBDoubleprootView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_DBDoubleprootView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_DBBigIntprootView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_DBBigIntprootView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_DBBigIntprootView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_DBBigIntprootView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_rootView_DBBigIntprootView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_rootView_DBBigIntprootView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_rootView_DBBigIntprootView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSProotView_DBBigInt)elem;
	_new->_desc2 = (_TSPLST_rootView_WebNodeFormalParam_type)_new;
	{_TPPLST_rootView_DBBigIntprootView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_DBBigIntprootView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_DBDateprootView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_DBDateprootView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_DBDateprootView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_DBDateprootView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_rootView_DBDateprootView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_rootView_DBDateprootView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_rootView_DBDateprootView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSProotView_DBDate)elem;
	_new->_desc2 = (_TSPLST_rootView_WebNodeFormalParam_type)_new;
	{_TPPLST_rootView_DBDateprootView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_DBDateprootView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_DBTimeStampprootView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_DBTimeStampprootView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_DBTimeStampprootView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_DBTimeStampprootView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_rootView_DBTimeStampprootView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_rootView_DBTimeStampprootView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_rootView_DBTimeStampprootView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSProotView_DBTimeStamp)elem;
	_new->_desc2 = (_TSPLST_rootView_WebNodeFormalParam_type)_new;
	{_TPPLST_rootView_DBTimeStampprootView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_DBTimeStampprootView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_DBBooleanprootView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_DBBooleanprootView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_DBBooleanprootView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_DBBooleanprootView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_rootView_DBBooleanprootView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_rootView_DBBooleanprootView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_rootView_DBBooleanprootView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSProotView_DBBoolean)elem;
	_new->_desc2 = (_TSPLST_rootView_WebNodeFormalParam_type)_new;
	{_TPPLST_rootView_DBBooleanprootView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_DBBooleanprootView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_DBBlobprootView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_DBBlobprootView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_DBBlobprootView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_DBBlobprootView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_rootView_DBBlobprootView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_rootView_DBBlobprootView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_rootView_DBBlobprootView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSProotView_DBBlob)elem;
	_new->_desc2 = (_TSPLST_rootView_WebNodeFormalParam_type)_new;
	{_TPPLST_rootView_DBBlobprootView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_DBBlobprootView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_WebNodeFormalParam_type (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_DBString)) 
		return _MkrootView_DBStringprootView_WebNodeFormalParam_type (_coordref, MkrootView_DBString (_coordref, t));
	if (IsSymb (t, SYMBrootView_DBInteger)) 
		return _MkrootView_DBIntegerprootView_WebNodeFormalParam_type (_coordref, MkrootView_DBInteger (_coordref, t));
	if (IsSymb (t, SYMBrootView_DBFloat)) 
		return _MkrootView_DBFloatprootView_WebNodeFormalParam_type (_coordref, MkrootView_DBFloat (_coordref, t));
	if (IsSymb (t, SYMBrootView_DBDouble)) 
		return _MkrootView_DBDoubleprootView_WebNodeFormalParam_type (_coordref, MkrootView_DBDouble (_coordref, t));
	if (IsSymb (t, SYMBrootView_DBBigInt)) 
		return _MkrootView_DBBigIntprootView_WebNodeFormalParam_type (_coordref, MkrootView_DBBigInt (_coordref, t));
	if (IsSymb (t, SYMBrootView_DBDate)) 
		return _MkrootView_DBDateprootView_WebNodeFormalParam_type (_coordref, MkrootView_DBDate (_coordref, t));
	if (IsSymb (t, SYMBrootView_DBTimeStamp)) 
		return _MkrootView_DBTimeStampprootView_WebNodeFormalParam_type (_coordref, MkrootView_DBTimeStamp (_coordref, t));
	if (IsSymb (t, SYMBrootView_DBBoolean)) 
		return _MkrootView_DBBooleanprootView_WebNodeFormalParam_type (_coordref, MkrootView_DBBoolean (_coordref, t));
	if (IsSymb (t, SYMBrootView_DBBlob)) 
		return _MkrootView_DBBlobprootView_WebNodeFormalParam_type (_coordref, MkrootView_DBBlob (_coordref, t));
	message (DEADLY, "a prootView_WebNodeFormalParam_type list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_WebNodeFormalParam_type (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_WebNodeFormalParam_type)
	l =_ElemprootView_WebNodeFormalParam_type (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_WebNodeFormalParam_type)
	r = _ElemprootView_WebNodeFormalParam_type (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_WebNodeFormalParam_type)r)->_desc2);
((_TPPLST_0prootView_WebNodeFormalParam_type)r)->_desc2 = ((_TPPLST_0prootView_WebNodeFormalParam_type)l)->_desc2;
((_TPPLST_0prootView_WebNodeFormalParam_type)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_WebNodeFormalParam_type (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_WebNodeFormalParam_type root;
	_TPPLST_0prootView_WebNodeFormalParam_type list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_WebNodeFormalParam_type)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_WebNodeFormalParam_type)
		t = _ElemprootView_WebNodeFormalParam_type (_coordref, t);
	remain = (_TPPLST_0prootView_WebNodeFormalParam_type)(((_TPPLST_0prootView_WebNodeFormalParam_type) t)->_desc2);
	((_TPPLST_0prootView_WebNodeFormalParam_type) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_WebNodeFormalParam_type;
#else
list = (_TPPLST_0prootView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_WebNodeFormalParam_type));
#endif
list->_prod = RULELST_0prootView_WebNodeFormalParam_type;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_WebNodeFormalParam_type)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_DBStringprootView_WebNodeFormalParam_type)
	{_TPPLST_rootView_DBStringprootView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_rootView_DBStringprootView_WebNodeFormalParam_type;
	} this code has been moved to _ElemprootView_WebNodeFormalParam_type*/
	/*if (_currn->_prod == RULELST_rootView_DBIntegerprootView_WebNodeFormalParam_type)
	{_TPPLST_rootView_DBIntegerprootView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_rootView_DBIntegerprootView_WebNodeFormalParam_type;
	} this code has been moved to _ElemprootView_WebNodeFormalParam_type*/
	/*if (_currn->_prod == RULELST_rootView_DBFloatprootView_WebNodeFormalParam_type)
	{_TPPLST_rootView_DBFloatprootView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_rootView_DBFloatprootView_WebNodeFormalParam_type;
	} this code has been moved to _ElemprootView_WebNodeFormalParam_type*/
	/*if (_currn->_prod == RULELST_rootView_DBDoubleprootView_WebNodeFormalParam_type)
	{_TPPLST_rootView_DBDoubleprootView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_rootView_DBDoubleprootView_WebNodeFormalParam_type;
	} this code has been moved to _ElemprootView_WebNodeFormalParam_type*/
	/*if (_currn->_prod == RULELST_rootView_DBBigIntprootView_WebNodeFormalParam_type)
	{_TPPLST_rootView_DBBigIntprootView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_rootView_DBBigIntprootView_WebNodeFormalParam_type;
	} this code has been moved to _ElemprootView_WebNodeFormalParam_type*/
	/*if (_currn->_prod == RULELST_rootView_DBDateprootView_WebNodeFormalParam_type)
	{_TPPLST_rootView_DBDateprootView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_rootView_DBDateprootView_WebNodeFormalParam_type;
	} this code has been moved to _ElemprootView_WebNodeFormalParam_type*/
	/*if (_currn->_prod == RULELST_rootView_DBTimeStampprootView_WebNodeFormalParam_type)
	{_TPPLST_rootView_DBTimeStampprootView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_rootView_DBTimeStampprootView_WebNodeFormalParam_type;
	} this code has been moved to _ElemprootView_WebNodeFormalParam_type*/
	/*if (_currn->_prod == RULELST_rootView_DBBooleanprootView_WebNodeFormalParam_type)
	{_TPPLST_rootView_DBBooleanprootView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_rootView_DBBooleanprootView_WebNodeFormalParam_type;
	} this code has been moved to _ElemprootView_WebNodeFormalParam_type*/
	/*if (_currn->_prod == RULELST_rootView_DBBlobprootView_WebNodeFormalParam_type)
	{_TPPLST_rootView_DBBlobprootView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_rootView_DBBlobprootView_WebNodeFormalParam_type;
	} this code has been moved to _ElemprootView_WebNodeFormalParam_type*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_WebNodeFormalParam_type;
#else
root = (_TPPprootView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPprootView_WebNodeFormalParam_type));
#endif
root->_prod = RULEprootView_WebNodeFormalParam_type;
root->_desc1 = (_TSPLST_rootView_WebNodeFormalParam_type)list;
{ _TPPprootView_WebNodeFormalParam_type _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_WebNodeFormalParam_type;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_DataBase (POSITION *_coordref, int _TERM1)
#else
NODEPTR MkprootView_DataBase (_coordref, _TERM1)
	POSITION *_coordref;
	int _TERM1;
#endif
{	_TPPprootView_DataBase _currn;
#ifdef __cplusplus
_currn = new _TPprootView_DataBase;
#else
_currn = (_TPPprootView_DataBase) TreeNodeAlloc (sizeof (struct _TPprootView_DataBase));
#endif
_currn->_prod = RULEprootView_DataBase;
_SETCOORD(_currn)
_TERMACT_prootView_DataBase;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_DataBase */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_DBString (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkprootView_DBString (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPprootView_DBString _currn;
#ifdef __cplusplus
_currn = new _TPprootView_DBString;
#else
_currn = (_TPPprootView_DBString) TreeNodeAlloc (sizeof (struct _TPprootView_DBString));
#endif
_currn->_prod = RULEprootView_DBString;
_currn->_desc1 = (_TSProotView_DBString_keyword1) MkrootView_DBString_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_DBString: root of subtree no. 1 can not be made a rootView_DBString_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_DBString;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_DBString */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_DBString_keyword1 (POSITION *_coordref)
#else
NODEPTR MkprootView_DBString_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_DBString_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPprootView_DBString_keyword1;
#else
_currn = (_TPPprootView_DBString_keyword1) TreeNodeAlloc (sizeof (struct _TPprootView_DBString_keyword1));
#endif
_currn->_prod = RULEprootView_DBString_keyword1;
_SETCOORD(_currn)
_TERMACT_prootView_DBString_keyword1;
return ( (NODEPTR) _currn);
}/* MkprootView_DBString_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_DBInteger (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkprootView_DBInteger (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPprootView_DBInteger _currn;
#ifdef __cplusplus
_currn = new _TPprootView_DBInteger;
#else
_currn = (_TPPprootView_DBInteger) TreeNodeAlloc (sizeof (struct _TPprootView_DBInteger));
#endif
_currn->_prod = RULEprootView_DBInteger;
_currn->_desc1 = (_TSProotView_DBInteger_keyword1) MkrootView_DBInteger_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_DBInteger: root of subtree no. 1 can not be made a rootView_DBInteger_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_DBInteger;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_DBInteger */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_DBInteger_keyword1 (POSITION *_coordref)
#else
NODEPTR MkprootView_DBInteger_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_DBInteger_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPprootView_DBInteger_keyword1;
#else
_currn = (_TPPprootView_DBInteger_keyword1) TreeNodeAlloc (sizeof (struct _TPprootView_DBInteger_keyword1));
#endif
_currn->_prod = RULEprootView_DBInteger_keyword1;
_SETCOORD(_currn)
_TERMACT_prootView_DBInteger_keyword1;
return ( (NODEPTR) _currn);
}/* MkprootView_DBInteger_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_DBFloat (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkprootView_DBFloat (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPprootView_DBFloat _currn;
#ifdef __cplusplus
_currn = new _TPprootView_DBFloat;
#else
_currn = (_TPPprootView_DBFloat) TreeNodeAlloc (sizeof (struct _TPprootView_DBFloat));
#endif
_currn->_prod = RULEprootView_DBFloat;
_currn->_desc1 = (_TSProotView_DBFloat_keyword1) MkrootView_DBFloat_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_DBFloat: root of subtree no. 1 can not be made a rootView_DBFloat_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_DBFloat;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_DBFloat */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_DBFloat_keyword1 (POSITION *_coordref)
#else
NODEPTR MkprootView_DBFloat_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_DBFloat_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPprootView_DBFloat_keyword1;
#else
_currn = (_TPPprootView_DBFloat_keyword1) TreeNodeAlloc (sizeof (struct _TPprootView_DBFloat_keyword1));
#endif
_currn->_prod = RULEprootView_DBFloat_keyword1;
_SETCOORD(_currn)
_TERMACT_prootView_DBFloat_keyword1;
return ( (NODEPTR) _currn);
}/* MkprootView_DBFloat_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_DBDouble (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkprootView_DBDouble (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPprootView_DBDouble _currn;
#ifdef __cplusplus
_currn = new _TPprootView_DBDouble;
#else
_currn = (_TPPprootView_DBDouble) TreeNodeAlloc (sizeof (struct _TPprootView_DBDouble));
#endif
_currn->_prod = RULEprootView_DBDouble;
_currn->_desc1 = (_TSProotView_DBDouble_keyword1) MkrootView_DBDouble_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_DBDouble: root of subtree no. 1 can not be made a rootView_DBDouble_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_DBDouble;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_DBDouble */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_DBDouble_keyword1 (POSITION *_coordref)
#else
NODEPTR MkprootView_DBDouble_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_DBDouble_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPprootView_DBDouble_keyword1;
#else
_currn = (_TPPprootView_DBDouble_keyword1) TreeNodeAlloc (sizeof (struct _TPprootView_DBDouble_keyword1));
#endif
_currn->_prod = RULEprootView_DBDouble_keyword1;
_SETCOORD(_currn)
_TERMACT_prootView_DBDouble_keyword1;
return ( (NODEPTR) _currn);
}/* MkprootView_DBDouble_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_DBBigInt (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkprootView_DBBigInt (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPprootView_DBBigInt _currn;
#ifdef __cplusplus
_currn = new _TPprootView_DBBigInt;
#else
_currn = (_TPPprootView_DBBigInt) TreeNodeAlloc (sizeof (struct _TPprootView_DBBigInt));
#endif
_currn->_prod = RULEprootView_DBBigInt;
_currn->_desc1 = (_TSProotView_DBBigInt_keyword1) MkrootView_DBBigInt_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_DBBigInt: root of subtree no. 1 can not be made a rootView_DBBigInt_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_DBBigInt;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_DBBigInt */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_DBBigInt_keyword1 (POSITION *_coordref)
#else
NODEPTR MkprootView_DBBigInt_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_DBBigInt_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPprootView_DBBigInt_keyword1;
#else
_currn = (_TPPprootView_DBBigInt_keyword1) TreeNodeAlloc (sizeof (struct _TPprootView_DBBigInt_keyword1));
#endif
_currn->_prod = RULEprootView_DBBigInt_keyword1;
_SETCOORD(_currn)
_TERMACT_prootView_DBBigInt_keyword1;
return ( (NODEPTR) _currn);
}/* MkprootView_DBBigInt_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_DBDate (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkprootView_DBDate (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPprootView_DBDate _currn;
#ifdef __cplusplus
_currn = new _TPprootView_DBDate;
#else
_currn = (_TPPprootView_DBDate) TreeNodeAlloc (sizeof (struct _TPprootView_DBDate));
#endif
_currn->_prod = RULEprootView_DBDate;
_currn->_desc1 = (_TSProotView_DBDate_keyword1) MkrootView_DBDate_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_DBDate: root of subtree no. 1 can not be made a rootView_DBDate_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_DBDate;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_DBDate */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_DBDate_keyword1 (POSITION *_coordref)
#else
NODEPTR MkprootView_DBDate_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_DBDate_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPprootView_DBDate_keyword1;
#else
_currn = (_TPPprootView_DBDate_keyword1) TreeNodeAlloc (sizeof (struct _TPprootView_DBDate_keyword1));
#endif
_currn->_prod = RULEprootView_DBDate_keyword1;
_SETCOORD(_currn)
_TERMACT_prootView_DBDate_keyword1;
return ( (NODEPTR) _currn);
}/* MkprootView_DBDate_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_DBTimeStamp (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkprootView_DBTimeStamp (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPprootView_DBTimeStamp _currn;
#ifdef __cplusplus
_currn = new _TPprootView_DBTimeStamp;
#else
_currn = (_TPPprootView_DBTimeStamp) TreeNodeAlloc (sizeof (struct _TPprootView_DBTimeStamp));
#endif
_currn->_prod = RULEprootView_DBTimeStamp;
_currn->_desc1 = (_TSProotView_DBTimeStamp_keyword1) MkrootView_DBTimeStamp_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_DBTimeStamp: root of subtree no. 1 can not be made a rootView_DBTimeStamp_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_DBTimeStamp;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_DBTimeStamp */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_DBTimeStamp_keyword1 (POSITION *_coordref)
#else
NODEPTR MkprootView_DBTimeStamp_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_DBTimeStamp_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPprootView_DBTimeStamp_keyword1;
#else
_currn = (_TPPprootView_DBTimeStamp_keyword1) TreeNodeAlloc (sizeof (struct _TPprootView_DBTimeStamp_keyword1));
#endif
_currn->_prod = RULEprootView_DBTimeStamp_keyword1;
_SETCOORD(_currn)
_TERMACT_prootView_DBTimeStamp_keyword1;
return ( (NODEPTR) _currn);
}/* MkprootView_DBTimeStamp_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_DBBoolean (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkprootView_DBBoolean (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPprootView_DBBoolean _currn;
#ifdef __cplusplus
_currn = new _TPprootView_DBBoolean;
#else
_currn = (_TPPprootView_DBBoolean) TreeNodeAlloc (sizeof (struct _TPprootView_DBBoolean));
#endif
_currn->_prod = RULEprootView_DBBoolean;
_currn->_desc1 = (_TSProotView_DBBoolean_keyword1) MkrootView_DBBoolean_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_DBBoolean: root of subtree no. 1 can not be made a rootView_DBBoolean_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_DBBoolean;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_DBBoolean */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_DBBoolean_keyword1 (POSITION *_coordref)
#else
NODEPTR MkprootView_DBBoolean_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_DBBoolean_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPprootView_DBBoolean_keyword1;
#else
_currn = (_TPPprootView_DBBoolean_keyword1) TreeNodeAlloc (sizeof (struct _TPprootView_DBBoolean_keyword1));
#endif
_currn->_prod = RULEprootView_DBBoolean_keyword1;
_SETCOORD(_currn)
_TERMACT_prootView_DBBoolean_keyword1;
return ( (NODEPTR) _currn);
}/* MkprootView_DBBoolean_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_DBBlob (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkprootView_DBBlob (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPprootView_DBBlob _currn;
#ifdef __cplusplus
_currn = new _TPprootView_DBBlob;
#else
_currn = (_TPPprootView_DBBlob) TreeNodeAlloc (sizeof (struct _TPprootView_DBBlob));
#endif
_currn->_prod = RULEprootView_DBBlob;
_currn->_desc1 = (_TSProotView_DBBlob_keyword1) MkrootView_DBBlob_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_DBBlob: root of subtree no. 1 can not be made a rootView_DBBlob_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_DBBlob;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_DBBlob */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_DBBlob_keyword1 (POSITION *_coordref)
#else
NODEPTR MkprootView_DBBlob_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_DBBlob_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPprootView_DBBlob_keyword1;
#else
_currn = (_TPPprootView_DBBlob_keyword1) TreeNodeAlloc (sizeof (struct _TPprootView_DBBlob_keyword1));
#endif
_currn->_prod = RULEprootView_DBBlob_keyword1;
_SETCOORD(_currn)
_TERMACT_prootView_DBBlob_keyword1;
return ( (NODEPTR) _currn);
}/* MkprootView_DBBlob_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_DataSourceActualParam (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkprootView_DataSourceActualParam (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPprootView_DataSourceActualParam _currn;
#ifdef __cplusplus
_currn = new _TPprootView_DataSourceActualParam;
#else
_currn = (_TPPprootView_DataSourceActualParam) TreeNodeAlloc (sizeof (struct _TPprootView_DataSourceActualParam));
#endif
_currn->_prod = RULEprootView_DataSourceActualParam;
_currn->_desc1 = (_TSProotView_DataSourceActualParam_formalParam) MkrootView_DataSourceActualParam_formalParam (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_DataSourceActualParam: root of subtree no. 1 can not be made a rootView_DataSourceActualParam_formalParam node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_DataSourceActualParam_expressions) MkrootView_DataSourceActualParam_expressions (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_DataSourceActualParam: root of subtree no. 2 can not be made a rootView_DataSourceActualParam_expressions node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_DataSourceActualParam;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_DataSourceActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_DataSourceActualParam_formalParam (POSITION *_coordref)
#else
NODEPTR MkprootView_DataSourceActualParam_formalParam (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_DataSourceActualParam_formalParam _currn;
#ifdef __cplusplus
_currn = new _TPprootView_DataSourceActualParam_formalParam;
#else
_currn = (_TPPprootView_DataSourceActualParam_formalParam) TreeNodeAlloc (sizeof (struct _TPprootView_DataSourceActualParam_formalParam));
#endif
_currn->_prod = RULEprootView_DataSourceActualParam_formalParam;
_SETCOORD(_currn)
_TERMACT_prootView_DataSourceActualParam_formalParam;
return ( (NODEPTR) _currn);
}/* MkprootView_DataSourceActualParam_formalParam */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_WebNodeConstantExpressionprootView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_WebNodeConstantExpressionprootView_DataSourceActualParam_expressions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_WebNodeConstantExpressionprootView_DataSourceActualParam_expressions _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_WebNodeConstantExpressionprootView_DataSourceActualParam_expressions;
#else
	_new = (_TPPLST_rootView_WebNodeConstantExpressionprootView_DataSourceActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_rootView_WebNodeConstantExpressionprootView_DataSourceActualParam_expressions));
#endif
	_new->_prod = RULELST_rootView_WebNodeConstantExpressionprootView_DataSourceActualParam_expressions;
	_new->_desc1 = (_TSProotView_WebNodeConstantExpression)elem;
	_new->_desc2 = (_TSPLST_rootView_DataSourceActualParam_expressions)_new;
	{_TPPLST_rootView_WebNodeConstantExpressionprootView_DataSourceActualParam_expressions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_WebNodeConstantExpressionprootView_DataSourceActualParam_expressions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_WebNodeVariableExpressionprootView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_WebNodeVariableExpressionprootView_DataSourceActualParam_expressions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_WebNodeVariableExpressionprootView_DataSourceActualParam_expressions _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_WebNodeVariableExpressionprootView_DataSourceActualParam_expressions;
#else
	_new = (_TPPLST_rootView_WebNodeVariableExpressionprootView_DataSourceActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_rootView_WebNodeVariableExpressionprootView_DataSourceActualParam_expressions));
#endif
	_new->_prod = RULELST_rootView_WebNodeVariableExpressionprootView_DataSourceActualParam_expressions;
	_new->_desc1 = (_TSProotView_WebNodeVariableExpression)elem;
	_new->_desc2 = (_TSPLST_rootView_DataSourceActualParam_expressions)_new;
	{_TPPLST_rootView_WebNodeVariableExpressionprootView_DataSourceActualParam_expressions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_WebNodeVariableExpressionprootView_DataSourceActualParam_expressions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_WebNodeColumnExpressionprootView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_WebNodeColumnExpressionprootView_DataSourceActualParam_expressions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_WebNodeColumnExpressionprootView_DataSourceActualParam_expressions _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_WebNodeColumnExpressionprootView_DataSourceActualParam_expressions;
#else
	_new = (_TPPLST_rootView_WebNodeColumnExpressionprootView_DataSourceActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_rootView_WebNodeColumnExpressionprootView_DataSourceActualParam_expressions));
#endif
	_new->_prod = RULELST_rootView_WebNodeColumnExpressionprootView_DataSourceActualParam_expressions;
	_new->_desc1 = (_TSProotView_WebNodeColumnExpression)elem;
	_new->_desc2 = (_TSPLST_rootView_DataSourceActualParam_expressions)_new;
	{_TPPLST_rootView_WebNodeColumnExpressionprootView_DataSourceActualParam_expressions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_WebNodeColumnExpressionprootView_DataSourceActualParam_expressions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_DataSourceActualParam_expressions (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_WebNodeConstantExpression)) 
		return _MkrootView_WebNodeConstantExpressionprootView_DataSourceActualParam_expressions (_coordref, MkrootView_WebNodeConstantExpression (_coordref, t));
	if (IsSymb (t, SYMBrootView_WebNodeVariableExpression)) 
		return _MkrootView_WebNodeVariableExpressionprootView_DataSourceActualParam_expressions (_coordref, MkrootView_WebNodeVariableExpression (_coordref, t));
	if (IsSymb (t, SYMBrootView_WebNodeColumnExpression)) 
		return _MkrootView_WebNodeColumnExpressionprootView_DataSourceActualParam_expressions (_coordref, MkrootView_WebNodeColumnExpression (_coordref, t));
	message (DEADLY, "a prootView_DataSourceActualParam_expressions list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_DataSourceActualParam_expressions (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_DataSourceActualParam_expressions)
	l =_ElemprootView_DataSourceActualParam_expressions (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_DataSourceActualParam_expressions)
	r = _ElemprootView_DataSourceActualParam_expressions (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_DataSourceActualParam_expressions)r)->_desc2);
((_TPPLST_0prootView_DataSourceActualParam_expressions)r)->_desc2 = ((_TPPLST_0prootView_DataSourceActualParam_expressions)l)->_desc2;
((_TPPLST_0prootView_DataSourceActualParam_expressions)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_DataSourceActualParam_expressions (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_DataSourceActualParam_expressions root;
	_TPPLST_0prootView_DataSourceActualParam_expressions list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_DataSourceActualParam_expressions)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_DataSourceActualParam_expressions)
		t = _ElemprootView_DataSourceActualParam_expressions (_coordref, t);
	remain = (_TPPLST_0prootView_DataSourceActualParam_expressions)(((_TPPLST_0prootView_DataSourceActualParam_expressions) t)->_desc2);
	((_TPPLST_0prootView_DataSourceActualParam_expressions) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_DataSourceActualParam_expressions;
#else
list = (_TPPLST_0prootView_DataSourceActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_DataSourceActualParam_expressions));
#endif
list->_prod = RULELST_0prootView_DataSourceActualParam_expressions;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_DataSourceActualParam_expressions)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_WebNodeConstantExpressionprootView_DataSourceActualParam_expressions)
	{_TPPLST_rootView_WebNodeConstantExpressionprootView_DataSourceActualParam_expressions _currn = _currn;
	 _TERMACT_LST_rootView_WebNodeConstantExpressionprootView_DataSourceActualParam_expressions;
	} this code has been moved to _ElemprootView_DataSourceActualParam_expressions*/
	/*if (_currn->_prod == RULELST_rootView_WebNodeVariableExpressionprootView_DataSourceActualParam_expressions)
	{_TPPLST_rootView_WebNodeVariableExpressionprootView_DataSourceActualParam_expressions _currn = _currn;
	 _TERMACT_LST_rootView_WebNodeVariableExpressionprootView_DataSourceActualParam_expressions;
	} this code has been moved to _ElemprootView_DataSourceActualParam_expressions*/
	/*if (_currn->_prod == RULELST_rootView_WebNodeColumnExpressionprootView_DataSourceActualParam_expressions)
	{_TPPLST_rootView_WebNodeColumnExpressionprootView_DataSourceActualParam_expressions _currn = _currn;
	 _TERMACT_LST_rootView_WebNodeColumnExpressionprootView_DataSourceActualParam_expressions;
	} this code has been moved to _ElemprootView_DataSourceActualParam_expressions*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_DataSourceActualParam_expressions;
#else
root = (_TPPprootView_DataSourceActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPprootView_DataSourceActualParam_expressions));
#endif
root->_prod = RULEprootView_DataSourceActualParam_expressions;
root->_desc1 = (_TSPLST_rootView_DataSourceActualParam_expressions)list;
{ _TPPprootView_DataSourceActualParam_expressions _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_DataSourceActualParam_expressions;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_Formular (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkprootView_Formular (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPprootView_Formular _currn;
#ifdef __cplusplus
_currn = new _TPprootView_Formular;
#else
_currn = (_TPPprootView_Formular) TreeNodeAlloc (sizeof (struct _TPprootView_Formular));
#endif
_currn->_prod = RULEprootView_Formular;
_currn->_desc1 = (_TSProotView_Formular_styleSheetBlock) MkrootView_Formular_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_Formular: root of subtree no. 1 can not be made a rootView_Formular_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_Formular_action) MkrootView_Formular_action (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_Formular: root of subtree no. 2 can not be made a rootView_Formular_action node ", 0, _coordref);
_currn->_desc3 = (_TSProotView_Formular_elements) MkrootView_Formular_elements (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_Formular: root of subtree no. 3 can not be made a rootView_Formular_elements node ", 0, _coordref);
_currn->_desc4 = (_TSProotView_Formular_datasource) MkrootView_Formular_datasource (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_Formular: root of subtree no. 4 can not be made a rootView_Formular_datasource node ", 0, _coordref);
_currn->_desc5 = (_TSProotView_Formular_actualParams) MkrootView_Formular_actualParams (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_Formular: root of subtree no. 5 can not be made a rootView_Formular_actualParams node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_Formular;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_Formular */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_Formular_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkprootView_Formular_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_Formular_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPprootView_Formular_styleSheetBlock;
#else
_currn = (_TPPprootView_Formular_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPprootView_Formular_styleSheetBlock));
#endif
_currn->_prod = RULEprootView_Formular_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_prootView_Formular_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkprootView_Formular_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_Formular_action (POSITION *_coordref)
#else
NODEPTR MkprootView_Formular_action (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_Formular_action _currn;
#ifdef __cplusplus
_currn = new _TPprootView_Formular_action;
#else
_currn = (_TPPprootView_Formular_action) TreeNodeAlloc (sizeof (struct _TPprootView_Formular_action));
#endif
_currn->_prod = RULEprootView_Formular_action;
_SETCOORD(_currn)
_TERMACT_prootView_Formular_action;
return ( (NODEPTR) _currn);
}/* MkprootView_Formular_action */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_FormularRadioGroupElementprootView_Formular_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_FormularRadioGroupElementprootView_Formular_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_FormularRadioGroupElementprootView_Formular_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_FormularRadioGroupElementprootView_Formular_elements;
#else
	_new = (_TPPLST_rootView_FormularRadioGroupElementprootView_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_rootView_FormularRadioGroupElementprootView_Formular_elements));
#endif
	_new->_prod = RULELST_rootView_FormularRadioGroupElementprootView_Formular_elements;
	_new->_desc1 = (_TSProotView_FormularRadioGroupElement)elem;
	_new->_desc2 = (_TSPLST_rootView_Formular_elements)_new;
	{_TPPLST_rootView_FormularRadioGroupElementprootView_Formular_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_FormularRadioGroupElementprootView_Formular_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_FormularTextElementprootView_Formular_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_FormularTextElementprootView_Formular_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_FormularTextElementprootView_Formular_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_FormularTextElementprootView_Formular_elements;
#else
	_new = (_TPPLST_rootView_FormularTextElementprootView_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_rootView_FormularTextElementprootView_Formular_elements));
#endif
	_new->_prod = RULELST_rootView_FormularTextElementprootView_Formular_elements;
	_new->_desc1 = (_TSProotView_FormularTextElement)elem;
	_new->_desc2 = (_TSPLST_rootView_Formular_elements)_new;
	{_TPPLST_rootView_FormularTextElementprootView_Formular_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_FormularTextElementprootView_Formular_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_FormularTextAreaElementprootView_Formular_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_FormularTextAreaElementprootView_Formular_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_FormularTextAreaElementprootView_Formular_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_FormularTextAreaElementprootView_Formular_elements;
#else
	_new = (_TPPLST_rootView_FormularTextAreaElementprootView_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_rootView_FormularTextAreaElementprootView_Formular_elements));
#endif
	_new->_prod = RULELST_rootView_FormularTextAreaElementprootView_Formular_elements;
	_new->_desc1 = (_TSProotView_FormularTextAreaElement)elem;
	_new->_desc2 = (_TSPLST_rootView_Formular_elements)_new;
	{_TPPLST_rootView_FormularTextAreaElementprootView_Formular_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_FormularTextAreaElementprootView_Formular_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_FormularPasswordElementprootView_Formular_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_FormularPasswordElementprootView_Formular_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_FormularPasswordElementprootView_Formular_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_FormularPasswordElementprootView_Formular_elements;
#else
	_new = (_TPPLST_rootView_FormularPasswordElementprootView_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_rootView_FormularPasswordElementprootView_Formular_elements));
#endif
	_new->_prod = RULELST_rootView_FormularPasswordElementprootView_Formular_elements;
	_new->_desc1 = (_TSProotView_FormularPasswordElement)elem;
	_new->_desc2 = (_TSPLST_rootView_Formular_elements)_new;
	{_TPPLST_rootView_FormularPasswordElementprootView_Formular_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_FormularPasswordElementprootView_Formular_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_FormularCheckboxElementprootView_Formular_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_FormularCheckboxElementprootView_Formular_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_FormularCheckboxElementprootView_Formular_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_FormularCheckboxElementprootView_Formular_elements;
#else
	_new = (_TPPLST_rootView_FormularCheckboxElementprootView_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_rootView_FormularCheckboxElementprootView_Formular_elements));
#endif
	_new->_prod = RULELST_rootView_FormularCheckboxElementprootView_Formular_elements;
	_new->_desc1 = (_TSProotView_FormularCheckboxElement)elem;
	_new->_desc2 = (_TSPLST_rootView_Formular_elements)_new;
	{_TPPLST_rootView_FormularCheckboxElementprootView_Formular_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_FormularCheckboxElementprootView_Formular_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_FormularSelectElementprootView_Formular_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_FormularSelectElementprootView_Formular_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_FormularSelectElementprootView_Formular_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_FormularSelectElementprootView_Formular_elements;
#else
	_new = (_TPPLST_rootView_FormularSelectElementprootView_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_rootView_FormularSelectElementprootView_Formular_elements));
#endif
	_new->_prod = RULELST_rootView_FormularSelectElementprootView_Formular_elements;
	_new->_desc1 = (_TSProotView_FormularSelectElement)elem;
	_new->_desc2 = (_TSPLST_rootView_Formular_elements)_new;
	{_TPPLST_rootView_FormularSelectElementprootView_Formular_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_FormularSelectElementprootView_Formular_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_Formular_elements (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_Formular_elements (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_FormularRadioGroupElement)) 
		return _MkrootView_FormularRadioGroupElementprootView_Formular_elements (_coordref, MkrootView_FormularRadioGroupElement (_coordref, t));
	if (IsSymb (t, SYMBrootView_FormularTextElement)) 
		return _MkrootView_FormularTextElementprootView_Formular_elements (_coordref, MkrootView_FormularTextElement (_coordref, t));
	if (IsSymb (t, SYMBrootView_FormularTextAreaElement)) 
		return _MkrootView_FormularTextAreaElementprootView_Formular_elements (_coordref, MkrootView_FormularTextAreaElement (_coordref, t));
	if (IsSymb (t, SYMBrootView_FormularPasswordElement)) 
		return _MkrootView_FormularPasswordElementprootView_Formular_elements (_coordref, MkrootView_FormularPasswordElement (_coordref, t));
	if (IsSymb (t, SYMBrootView_FormularCheckboxElement)) 
		return _MkrootView_FormularCheckboxElementprootView_Formular_elements (_coordref, MkrootView_FormularCheckboxElement (_coordref, t));
	if (IsSymb (t, SYMBrootView_FormularSelectElement)) 
		return _MkrootView_FormularSelectElementprootView_Formular_elements (_coordref, MkrootView_FormularSelectElement (_coordref, t));
	message (DEADLY, "a prootView_Formular_elements list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_Formular_elements (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_Formular_elements (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_Formular_elements)
	l =_ElemprootView_Formular_elements (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_Formular_elements)
	r = _ElemprootView_Formular_elements (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_Formular_elements)r)->_desc2);
((_TPPLST_0prootView_Formular_elements)r)->_desc2 = ((_TPPLST_0prootView_Formular_elements)l)->_desc2;
((_TPPLST_0prootView_Formular_elements)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_Formular_elements (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_Formular_elements (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_Formular_elements root;
	_TPPLST_0prootView_Formular_elements list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_Formular_elements)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_Formular_elements)
		t = _ElemprootView_Formular_elements (_coordref, t);
	remain = (_TPPLST_0prootView_Formular_elements)(((_TPPLST_0prootView_Formular_elements) t)->_desc2);
	((_TPPLST_0prootView_Formular_elements) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_Formular_elements;
#else
list = (_TPPLST_0prootView_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_Formular_elements));
#endif
list->_prod = RULELST_0prootView_Formular_elements;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_Formular_elements)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_FormularRadioGroupElementprootView_Formular_elements)
	{_TPPLST_rootView_FormularRadioGroupElementprootView_Formular_elements _currn = _currn;
	 _TERMACT_LST_rootView_FormularRadioGroupElementprootView_Formular_elements;
	} this code has been moved to _ElemprootView_Formular_elements*/
	/*if (_currn->_prod == RULELST_rootView_FormularTextElementprootView_Formular_elements)
	{_TPPLST_rootView_FormularTextElementprootView_Formular_elements _currn = _currn;
	 _TERMACT_LST_rootView_FormularTextElementprootView_Formular_elements;
	} this code has been moved to _ElemprootView_Formular_elements*/
	/*if (_currn->_prod == RULELST_rootView_FormularTextAreaElementprootView_Formular_elements)
	{_TPPLST_rootView_FormularTextAreaElementprootView_Formular_elements _currn = _currn;
	 _TERMACT_LST_rootView_FormularTextAreaElementprootView_Formular_elements;
	} this code has been moved to _ElemprootView_Formular_elements*/
	/*if (_currn->_prod == RULELST_rootView_FormularPasswordElementprootView_Formular_elements)
	{_TPPLST_rootView_FormularPasswordElementprootView_Formular_elements _currn = _currn;
	 _TERMACT_LST_rootView_FormularPasswordElementprootView_Formular_elements;
	} this code has been moved to _ElemprootView_Formular_elements*/
	/*if (_currn->_prod == RULELST_rootView_FormularCheckboxElementprootView_Formular_elements)
	{_TPPLST_rootView_FormularCheckboxElementprootView_Formular_elements _currn = _currn;
	 _TERMACT_LST_rootView_FormularCheckboxElementprootView_Formular_elements;
	} this code has been moved to _ElemprootView_Formular_elements*/
	/*if (_currn->_prod == RULELST_rootView_FormularSelectElementprootView_Formular_elements)
	{_TPPLST_rootView_FormularSelectElementprootView_Formular_elements _currn = _currn;
	 _TERMACT_LST_rootView_FormularSelectElementprootView_Formular_elements;
	} this code has been moved to _ElemprootView_Formular_elements*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_Formular_elements;
#else
root = (_TPPprootView_Formular_elements) TreeNodeAlloc (sizeof (struct _TPprootView_Formular_elements));
#endif
root->_prod = RULEprootView_Formular_elements;
root->_desc1 = (_TSPLST_rootView_Formular_elements)list;
{ _TPPprootView_Formular_elements _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_Formular_elements;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_Formular_datasource (POSITION *_coordref)
#else
NODEPTR MkprootView_Formular_datasource (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_Formular_datasource _currn;
#ifdef __cplusplus
_currn = new _TPprootView_Formular_datasource;
#else
_currn = (_TPPprootView_Formular_datasource) TreeNodeAlloc (sizeof (struct _TPprootView_Formular_datasource));
#endif
_currn->_prod = RULEprootView_Formular_datasource;
_SETCOORD(_currn)
_TERMACT_prootView_Formular_datasource;
return ( (NODEPTR) _currn);
}/* MkprootView_Formular_datasource */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_WebNodeActualParamprootView_Formular_actualParams (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_WebNodeActualParamprootView_Formular_actualParams (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_WebNodeActualParamprootView_Formular_actualParams _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_WebNodeActualParamprootView_Formular_actualParams;
#else
	_new = (_TPPLST_rootView_WebNodeActualParamprootView_Formular_actualParams) TreeNodeAlloc (sizeof (struct _TPLST_rootView_WebNodeActualParamprootView_Formular_actualParams));
#endif
	_new->_prod = RULELST_rootView_WebNodeActualParamprootView_Formular_actualParams;
	_new->_desc1 = (_TSProotView_WebNodeActualParam)elem;
	_new->_desc2 = (_TSPLST_rootView_Formular_actualParams)_new;
	{_TPPLST_rootView_WebNodeActualParamprootView_Formular_actualParams _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_WebNodeActualParamprootView_Formular_actualParams;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_Formular_actualParams (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_Formular_actualParams (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_WebNodeActualParam)) 
		return _MkrootView_WebNodeActualParamprootView_Formular_actualParams (_coordref, MkrootView_WebNodeActualParam (_coordref, t));
	message (DEADLY, "a prootView_Formular_actualParams list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_Formular_actualParams (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_Formular_actualParams (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_Formular_actualParams)
	l =_ElemprootView_Formular_actualParams (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_Formular_actualParams)
	r = _ElemprootView_Formular_actualParams (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_Formular_actualParams)r)->_desc2);
((_TPPLST_0prootView_Formular_actualParams)r)->_desc2 = ((_TPPLST_0prootView_Formular_actualParams)l)->_desc2;
((_TPPLST_0prootView_Formular_actualParams)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_Formular_actualParams (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_Formular_actualParams (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_Formular_actualParams root;
	_TPPLST_0prootView_Formular_actualParams list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_Formular_actualParams)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_Formular_actualParams)
		t = _ElemprootView_Formular_actualParams (_coordref, t);
	remain = (_TPPLST_0prootView_Formular_actualParams)(((_TPPLST_0prootView_Formular_actualParams) t)->_desc2);
	((_TPPLST_0prootView_Formular_actualParams) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_Formular_actualParams;
#else
list = (_TPPLST_0prootView_Formular_actualParams) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_Formular_actualParams));
#endif
list->_prod = RULELST_0prootView_Formular_actualParams;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_Formular_actualParams)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_WebNodeActualParamprootView_Formular_actualParams)
	{_TPPLST_rootView_WebNodeActualParamprootView_Formular_actualParams _currn = _currn;
	 _TERMACT_LST_rootView_WebNodeActualParamprootView_Formular_actualParams;
	} this code has been moved to _ElemprootView_Formular_actualParams*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_Formular_actualParams;
#else
root = (_TPPprootView_Formular_actualParams) TreeNodeAlloc (sizeof (struct _TPprootView_Formular_actualParams));
#endif
root->_prod = RULEprootView_Formular_actualParams;
root->_desc1 = (_TSPLST_rootView_Formular_actualParams)list;
{ _TPPprootView_Formular_actualParams _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_Formular_actualParams;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularRadioGroupElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5, NODEPTR _desc6, NODEPTR _desc7)
#else
NODEPTR MkprootView_FormularRadioGroupElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5,_desc6,_desc7)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
	NODEPTR _desc6;
	NODEPTR _desc7;
#endif
{	_TPPprootView_FormularRadioGroupElement _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularRadioGroupElement;
#else
_currn = (_TPPprootView_FormularRadioGroupElement) TreeNodeAlloc (sizeof (struct _TPprootView_FormularRadioGroupElement));
#endif
_currn->_prod = RULEprootView_FormularRadioGroupElement;
_currn->_desc1 = (_TSProotView_FormularRadioGroupElement_name) MkrootView_FormularRadioGroupElement_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularRadioGroupElement: root of subtree no. 1 can not be made a rootView_FormularRadioGroupElement_name node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_FormularRadioGroupElement_label) MkrootView_FormularRadioGroupElement_label (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularRadioGroupElement: root of subtree no. 2 can not be made a rootView_FormularRadioGroupElement_label node ", 0, _coordref);
_currn->_desc3 = (_TSProotView_FormularRadioGroupElement_description) MkrootView_FormularRadioGroupElement_description (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularRadioGroupElement: root of subtree no. 3 can not be made a rootView_FormularRadioGroupElement_description node ", 0, _coordref);
_currn->_desc4 = (_TSProotView_FormularRadioGroupElement_pvalue) MkrootView_FormularRadioGroupElement_pvalue (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularRadioGroupElement: root of subtree no. 4 can not be made a rootView_FormularRadioGroupElement_pvalue node ", 0, _coordref);
_currn->_desc5 = (_TSProotView_FormularRadioGroupElement_condition) MkrootView_FormularRadioGroupElement_condition (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularRadioGroupElement: root of subtree no. 5 can not be made a rootView_FormularRadioGroupElement_condition node ", 0, _coordref);
_currn->_desc6 = (_TSProotView_FormularRadioGroupElement_elements) MkrootView_FormularRadioGroupElement_elements (_coordref, _desc6);	
if (((NODEPTR)_currn->_desc6) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularRadioGroupElement: root of subtree no. 6 can not be made a rootView_FormularRadioGroupElement_elements node ", 0, _coordref);
_currn->_desc7 = (_TSProotView_FormularRadioGroupElement_checked) MkrootView_FormularRadioGroupElement_checked (_coordref, _desc7);	
if (((NODEPTR)_currn->_desc7) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularRadioGroupElement: root of subtree no. 7 can not be made a rootView_FormularRadioGroupElement_checked node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_FormularRadioGroupElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_FormularRadioGroupElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularRadioGroupElement_name (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularRadioGroupElement_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularRadioGroupElement_name _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularRadioGroupElement_name;
#else
_currn = (_TPPprootView_FormularRadioGroupElement_name) TreeNodeAlloc (sizeof (struct _TPprootView_FormularRadioGroupElement_name));
#endif
_currn->_prod = RULEprootView_FormularRadioGroupElement_name;
_SETCOORD(_currn)
_TERMACT_prootView_FormularRadioGroupElement_name;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularRadioGroupElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularRadioGroupElement_label (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularRadioGroupElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularRadioGroupElement_label _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularRadioGroupElement_label;
#else
_currn = (_TPPprootView_FormularRadioGroupElement_label) TreeNodeAlloc (sizeof (struct _TPprootView_FormularRadioGroupElement_label));
#endif
_currn->_prod = RULEprootView_FormularRadioGroupElement_label;
_SETCOORD(_currn)
_TERMACT_prootView_FormularRadioGroupElement_label;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularRadioGroupElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularRadioGroupElement_description (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularRadioGroupElement_description (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularRadioGroupElement_description _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularRadioGroupElement_description;
#else
_currn = (_TPPprootView_FormularRadioGroupElement_description) TreeNodeAlloc (sizeof (struct _TPprootView_FormularRadioGroupElement_description));
#endif
_currn->_prod = RULEprootView_FormularRadioGroupElement_description;
_SETCOORD(_currn)
_TERMACT_prootView_FormularRadioGroupElement_description;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularRadioGroupElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularRadioGroupElement_pvalue (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularRadioGroupElement_pvalue (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularRadioGroupElement_pvalue _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularRadioGroupElement_pvalue;
#else
_currn = (_TPPprootView_FormularRadioGroupElement_pvalue) TreeNodeAlloc (sizeof (struct _TPprootView_FormularRadioGroupElement_pvalue));
#endif
_currn->_prod = RULEprootView_FormularRadioGroupElement_pvalue;
_SETCOORD(_currn)
_TERMACT_prootView_FormularRadioGroupElement_pvalue;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularRadioGroupElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_FormularRadioElementprootView_FormularRadioGroupElement_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_FormularRadioElementprootView_FormularRadioGroupElement_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_FormularRadioElementprootView_FormularRadioGroupElement_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_FormularRadioElementprootView_FormularRadioGroupElement_elements;
#else
	_new = (_TPPLST_rootView_FormularRadioElementprootView_FormularRadioGroupElement_elements) TreeNodeAlloc (sizeof (struct _TPLST_rootView_FormularRadioElementprootView_FormularRadioGroupElement_elements));
#endif
	_new->_prod = RULELST_rootView_FormularRadioElementprootView_FormularRadioGroupElement_elements;
	_new->_desc1 = (_TSProotView_FormularRadioElement)elem;
	_new->_desc2 = (_TSPLST_rootView_FormularRadioGroupElement_elements)_new;
	{_TPPLST_rootView_FormularRadioElementprootView_FormularRadioGroupElement_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_FormularRadioElementprootView_FormularRadioGroupElement_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_FormularRadioGroupElement_elements (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_FormularRadioGroupElement_elements (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_FormularRadioElement)) 
		return _MkrootView_FormularRadioElementprootView_FormularRadioGroupElement_elements (_coordref, MkrootView_FormularRadioElement (_coordref, t));
	message (DEADLY, "a prootView_FormularRadioGroupElement_elements list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_FormularRadioGroupElement_elements (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_FormularRadioGroupElement_elements (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_FormularRadioGroupElement_elements)
	l =_ElemprootView_FormularRadioGroupElement_elements (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_FormularRadioGroupElement_elements)
	r = _ElemprootView_FormularRadioGroupElement_elements (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_FormularRadioGroupElement_elements)r)->_desc2);
((_TPPLST_0prootView_FormularRadioGroupElement_elements)r)->_desc2 = ((_TPPLST_0prootView_FormularRadioGroupElement_elements)l)->_desc2;
((_TPPLST_0prootView_FormularRadioGroupElement_elements)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularRadioGroupElement_elements (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_FormularRadioGroupElement_elements (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_FormularRadioGroupElement_elements root;
	_TPPLST_0prootView_FormularRadioGroupElement_elements list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_FormularRadioGroupElement_elements)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_FormularRadioGroupElement_elements)
		t = _ElemprootView_FormularRadioGroupElement_elements (_coordref, t);
	remain = (_TPPLST_0prootView_FormularRadioGroupElement_elements)(((_TPPLST_0prootView_FormularRadioGroupElement_elements) t)->_desc2);
	((_TPPLST_0prootView_FormularRadioGroupElement_elements) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_FormularRadioGroupElement_elements;
#else
list = (_TPPLST_0prootView_FormularRadioGroupElement_elements) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_FormularRadioGroupElement_elements));
#endif
list->_prod = RULELST_0prootView_FormularRadioGroupElement_elements;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_FormularRadioGroupElement_elements)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_FormularRadioElementprootView_FormularRadioGroupElement_elements)
	{_TPPLST_rootView_FormularRadioElementprootView_FormularRadioGroupElement_elements _currn = _currn;
	 _TERMACT_LST_rootView_FormularRadioElementprootView_FormularRadioGroupElement_elements;
	} this code has been moved to _ElemprootView_FormularRadioGroupElement_elements*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_FormularRadioGroupElement_elements;
#else
root = (_TPPprootView_FormularRadioGroupElement_elements) TreeNodeAlloc (sizeof (struct _TPprootView_FormularRadioGroupElement_elements));
#endif
root->_prod = RULEprootView_FormularRadioGroupElement_elements;
root->_desc1 = (_TSPLST_rootView_FormularRadioGroupElement_elements)list;
{ _TPPprootView_FormularRadioGroupElement_elements _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_FormularRadioGroupElement_elements;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularRadioGroupElement_checked (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularRadioGroupElement_checked (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularRadioGroupElement_checked _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularRadioGroupElement_checked;
#else
_currn = (_TPPprootView_FormularRadioGroupElement_checked) TreeNodeAlloc (sizeof (struct _TPprootView_FormularRadioGroupElement_checked));
#endif
_currn->_prod = RULEprootView_FormularRadioGroupElement_checked;
_SETCOORD(_currn)
_TERMACT_prootView_FormularRadioGroupElement_checked;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularRadioGroupElement_checked */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularTextElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkprootView_FormularTextElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPprootView_FormularTextElement _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularTextElement;
#else
_currn = (_TPPprootView_FormularTextElement) TreeNodeAlloc (sizeof (struct _TPprootView_FormularTextElement));
#endif
_currn->_prod = RULEprootView_FormularTextElement;
_currn->_desc1 = (_TSProotView_FormularTextElement_name) MkrootView_FormularTextElement_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularTextElement: root of subtree no. 1 can not be made a rootView_FormularTextElement_name node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_FormularTextElement_label) MkrootView_FormularTextElement_label (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularTextElement: root of subtree no. 2 can not be made a rootView_FormularTextElement_label node ", 0, _coordref);
_currn->_desc3 = (_TSProotView_FormularTextElement_description) MkrootView_FormularTextElement_description (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularTextElement: root of subtree no. 3 can not be made a rootView_FormularTextElement_description node ", 0, _coordref);
_currn->_desc4 = (_TSProotView_FormularTextElement_pvalue) MkrootView_FormularTextElement_pvalue (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularTextElement: root of subtree no. 4 can not be made a rootView_FormularTextElement_pvalue node ", 0, _coordref);
_currn->_desc5 = (_TSProotView_FormularTextElement_condition) MkrootView_FormularTextElement_condition (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularTextElement: root of subtree no. 5 can not be made a rootView_FormularTextElement_condition node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_FormularTextElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_FormularTextElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularTextElement_name (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularTextElement_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularTextElement_name _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularTextElement_name;
#else
_currn = (_TPPprootView_FormularTextElement_name) TreeNodeAlloc (sizeof (struct _TPprootView_FormularTextElement_name));
#endif
_currn->_prod = RULEprootView_FormularTextElement_name;
_SETCOORD(_currn)
_TERMACT_prootView_FormularTextElement_name;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularTextElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularTextElement_label (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularTextElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularTextElement_label _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularTextElement_label;
#else
_currn = (_TPPprootView_FormularTextElement_label) TreeNodeAlloc (sizeof (struct _TPprootView_FormularTextElement_label));
#endif
_currn->_prod = RULEprootView_FormularTextElement_label;
_SETCOORD(_currn)
_TERMACT_prootView_FormularTextElement_label;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularTextElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularTextElement_description (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularTextElement_description (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularTextElement_description _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularTextElement_description;
#else
_currn = (_TPPprootView_FormularTextElement_description) TreeNodeAlloc (sizeof (struct _TPprootView_FormularTextElement_description));
#endif
_currn->_prod = RULEprootView_FormularTextElement_description;
_SETCOORD(_currn)
_TERMACT_prootView_FormularTextElement_description;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularTextElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularTextElement_pvalue (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularTextElement_pvalue (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularTextElement_pvalue _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularTextElement_pvalue;
#else
_currn = (_TPPprootView_FormularTextElement_pvalue) TreeNodeAlloc (sizeof (struct _TPprootView_FormularTextElement_pvalue));
#endif
_currn->_prod = RULEprootView_FormularTextElement_pvalue;
_SETCOORD(_currn)
_TERMACT_prootView_FormularTextElement_pvalue;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularTextElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularTextAreaElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkprootView_FormularTextAreaElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPprootView_FormularTextAreaElement _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularTextAreaElement;
#else
_currn = (_TPPprootView_FormularTextAreaElement) TreeNodeAlloc (sizeof (struct _TPprootView_FormularTextAreaElement));
#endif
_currn->_prod = RULEprootView_FormularTextAreaElement;
_currn->_desc1 = (_TSProotView_FormularTextAreaElement_name) MkrootView_FormularTextAreaElement_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularTextAreaElement: root of subtree no. 1 can not be made a rootView_FormularTextAreaElement_name node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_FormularTextAreaElement_label) MkrootView_FormularTextAreaElement_label (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularTextAreaElement: root of subtree no. 2 can not be made a rootView_FormularTextAreaElement_label node ", 0, _coordref);
_currn->_desc3 = (_TSProotView_FormularTextAreaElement_description) MkrootView_FormularTextAreaElement_description (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularTextAreaElement: root of subtree no. 3 can not be made a rootView_FormularTextAreaElement_description node ", 0, _coordref);
_currn->_desc4 = (_TSProotView_FormularTextAreaElement_pvalue) MkrootView_FormularTextAreaElement_pvalue (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularTextAreaElement: root of subtree no. 4 can not be made a rootView_FormularTextAreaElement_pvalue node ", 0, _coordref);
_currn->_desc5 = (_TSProotView_FormularTextAreaElement_condition) MkrootView_FormularTextAreaElement_condition (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularTextAreaElement: root of subtree no. 5 can not be made a rootView_FormularTextAreaElement_condition node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_FormularTextAreaElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_FormularTextAreaElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularTextAreaElement_name (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularTextAreaElement_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularTextAreaElement_name _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularTextAreaElement_name;
#else
_currn = (_TPPprootView_FormularTextAreaElement_name) TreeNodeAlloc (sizeof (struct _TPprootView_FormularTextAreaElement_name));
#endif
_currn->_prod = RULEprootView_FormularTextAreaElement_name;
_SETCOORD(_currn)
_TERMACT_prootView_FormularTextAreaElement_name;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularTextAreaElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularTextAreaElement_label (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularTextAreaElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularTextAreaElement_label _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularTextAreaElement_label;
#else
_currn = (_TPPprootView_FormularTextAreaElement_label) TreeNodeAlloc (sizeof (struct _TPprootView_FormularTextAreaElement_label));
#endif
_currn->_prod = RULEprootView_FormularTextAreaElement_label;
_SETCOORD(_currn)
_TERMACT_prootView_FormularTextAreaElement_label;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularTextAreaElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularTextAreaElement_description (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularTextAreaElement_description (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularTextAreaElement_description _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularTextAreaElement_description;
#else
_currn = (_TPPprootView_FormularTextAreaElement_description) TreeNodeAlloc (sizeof (struct _TPprootView_FormularTextAreaElement_description));
#endif
_currn->_prod = RULEprootView_FormularTextAreaElement_description;
_SETCOORD(_currn)
_TERMACT_prootView_FormularTextAreaElement_description;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularTextAreaElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularTextAreaElement_pvalue (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularTextAreaElement_pvalue (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularTextAreaElement_pvalue _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularTextAreaElement_pvalue;
#else
_currn = (_TPPprootView_FormularTextAreaElement_pvalue) TreeNodeAlloc (sizeof (struct _TPprootView_FormularTextAreaElement_pvalue));
#endif
_currn->_prod = RULEprootView_FormularTextAreaElement_pvalue;
_SETCOORD(_currn)
_TERMACT_prootView_FormularTextAreaElement_pvalue;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularTextAreaElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularPasswordElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkprootView_FormularPasswordElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPprootView_FormularPasswordElement _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularPasswordElement;
#else
_currn = (_TPPprootView_FormularPasswordElement) TreeNodeAlloc (sizeof (struct _TPprootView_FormularPasswordElement));
#endif
_currn->_prod = RULEprootView_FormularPasswordElement;
_currn->_desc1 = (_TSProotView_FormularPasswordElement_name) MkrootView_FormularPasswordElement_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularPasswordElement: root of subtree no. 1 can not be made a rootView_FormularPasswordElement_name node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_FormularPasswordElement_label) MkrootView_FormularPasswordElement_label (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularPasswordElement: root of subtree no. 2 can not be made a rootView_FormularPasswordElement_label node ", 0, _coordref);
_currn->_desc3 = (_TSProotView_FormularPasswordElement_description) MkrootView_FormularPasswordElement_description (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularPasswordElement: root of subtree no. 3 can not be made a rootView_FormularPasswordElement_description node ", 0, _coordref);
_currn->_desc4 = (_TSProotView_FormularPasswordElement_pvalue) MkrootView_FormularPasswordElement_pvalue (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularPasswordElement: root of subtree no. 4 can not be made a rootView_FormularPasswordElement_pvalue node ", 0, _coordref);
_currn->_desc5 = (_TSProotView_FormularPasswordElement_condition) MkrootView_FormularPasswordElement_condition (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularPasswordElement: root of subtree no. 5 can not be made a rootView_FormularPasswordElement_condition node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_FormularPasswordElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_FormularPasswordElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularPasswordElement_name (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularPasswordElement_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularPasswordElement_name _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularPasswordElement_name;
#else
_currn = (_TPPprootView_FormularPasswordElement_name) TreeNodeAlloc (sizeof (struct _TPprootView_FormularPasswordElement_name));
#endif
_currn->_prod = RULEprootView_FormularPasswordElement_name;
_SETCOORD(_currn)
_TERMACT_prootView_FormularPasswordElement_name;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularPasswordElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularPasswordElement_label (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularPasswordElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularPasswordElement_label _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularPasswordElement_label;
#else
_currn = (_TPPprootView_FormularPasswordElement_label) TreeNodeAlloc (sizeof (struct _TPprootView_FormularPasswordElement_label));
#endif
_currn->_prod = RULEprootView_FormularPasswordElement_label;
_SETCOORD(_currn)
_TERMACT_prootView_FormularPasswordElement_label;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularPasswordElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularPasswordElement_description (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularPasswordElement_description (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularPasswordElement_description _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularPasswordElement_description;
#else
_currn = (_TPPprootView_FormularPasswordElement_description) TreeNodeAlloc (sizeof (struct _TPprootView_FormularPasswordElement_description));
#endif
_currn->_prod = RULEprootView_FormularPasswordElement_description;
_SETCOORD(_currn)
_TERMACT_prootView_FormularPasswordElement_description;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularPasswordElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularPasswordElement_pvalue (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularPasswordElement_pvalue (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularPasswordElement_pvalue _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularPasswordElement_pvalue;
#else
_currn = (_TPPprootView_FormularPasswordElement_pvalue) TreeNodeAlloc (sizeof (struct _TPprootView_FormularPasswordElement_pvalue));
#endif
_currn->_prod = RULEprootView_FormularPasswordElement_pvalue;
_SETCOORD(_currn)
_TERMACT_prootView_FormularPasswordElement_pvalue;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularPasswordElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularCheckboxElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5, NODEPTR _desc6, NODEPTR _desc7)
#else
NODEPTR MkprootView_FormularCheckboxElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5,_desc6,_desc7)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
	NODEPTR _desc6;
	NODEPTR _desc7;
#endif
{	_TPPprootView_FormularCheckboxElement _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularCheckboxElement;
#else
_currn = (_TPPprootView_FormularCheckboxElement) TreeNodeAlloc (sizeof (struct _TPprootView_FormularCheckboxElement));
#endif
_currn->_prod = RULEprootView_FormularCheckboxElement;
_currn->_desc1 = (_TSProotView_FormularCheckboxElement_name) MkrootView_FormularCheckboxElement_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularCheckboxElement: root of subtree no. 1 can not be made a rootView_FormularCheckboxElement_name node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_FormularCheckboxElement_label) MkrootView_FormularCheckboxElement_label (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularCheckboxElement: root of subtree no. 2 can not be made a rootView_FormularCheckboxElement_label node ", 0, _coordref);
_currn->_desc3 = (_TSProotView_FormularCheckboxElement_description) MkrootView_FormularCheckboxElement_description (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularCheckboxElement: root of subtree no. 3 can not be made a rootView_FormularCheckboxElement_description node ", 0, _coordref);
_currn->_desc4 = (_TSProotView_FormularCheckboxElement_pvalue) MkrootView_FormularCheckboxElement_pvalue (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularCheckboxElement: root of subtree no. 4 can not be made a rootView_FormularCheckboxElement_pvalue node ", 0, _coordref);
_currn->_desc5 = (_TSProotView_FormularCheckboxElement_condition) MkrootView_FormularCheckboxElement_condition (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularCheckboxElement: root of subtree no. 5 can not be made a rootView_FormularCheckboxElement_condition node ", 0, _coordref);
_currn->_desc6 = (_TSProotView_FormularCheckboxElement_value) MkrootView_FormularCheckboxElement_value (_coordref, _desc6);	
if (((NODEPTR)_currn->_desc6) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularCheckboxElement: root of subtree no. 6 can not be made a rootView_FormularCheckboxElement_value node ", 0, _coordref);
_currn->_desc7 = (_TSProotView_FormularCheckboxElement_checked) MkrootView_FormularCheckboxElement_checked (_coordref, _desc7);	
if (((NODEPTR)_currn->_desc7) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularCheckboxElement: root of subtree no. 7 can not be made a rootView_FormularCheckboxElement_checked node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_FormularCheckboxElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_FormularCheckboxElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularCheckboxElement_name (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularCheckboxElement_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularCheckboxElement_name _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularCheckboxElement_name;
#else
_currn = (_TPPprootView_FormularCheckboxElement_name) TreeNodeAlloc (sizeof (struct _TPprootView_FormularCheckboxElement_name));
#endif
_currn->_prod = RULEprootView_FormularCheckboxElement_name;
_SETCOORD(_currn)
_TERMACT_prootView_FormularCheckboxElement_name;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularCheckboxElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularCheckboxElement_label (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularCheckboxElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularCheckboxElement_label _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularCheckboxElement_label;
#else
_currn = (_TPPprootView_FormularCheckboxElement_label) TreeNodeAlloc (sizeof (struct _TPprootView_FormularCheckboxElement_label));
#endif
_currn->_prod = RULEprootView_FormularCheckboxElement_label;
_SETCOORD(_currn)
_TERMACT_prootView_FormularCheckboxElement_label;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularCheckboxElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularCheckboxElement_description (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularCheckboxElement_description (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularCheckboxElement_description _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularCheckboxElement_description;
#else
_currn = (_TPPprootView_FormularCheckboxElement_description) TreeNodeAlloc (sizeof (struct _TPprootView_FormularCheckboxElement_description));
#endif
_currn->_prod = RULEprootView_FormularCheckboxElement_description;
_SETCOORD(_currn)
_TERMACT_prootView_FormularCheckboxElement_description;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularCheckboxElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularCheckboxElement_pvalue (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularCheckboxElement_pvalue (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularCheckboxElement_pvalue _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularCheckboxElement_pvalue;
#else
_currn = (_TPPprootView_FormularCheckboxElement_pvalue) TreeNodeAlloc (sizeof (struct _TPprootView_FormularCheckboxElement_pvalue));
#endif
_currn->_prod = RULEprootView_FormularCheckboxElement_pvalue;
_SETCOORD(_currn)
_TERMACT_prootView_FormularCheckboxElement_pvalue;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularCheckboxElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularCheckboxElement_value (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularCheckboxElement_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularCheckboxElement_value _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularCheckboxElement_value;
#else
_currn = (_TPPprootView_FormularCheckboxElement_value) TreeNodeAlloc (sizeof (struct _TPprootView_FormularCheckboxElement_value));
#endif
_currn->_prod = RULEprootView_FormularCheckboxElement_value;
_SETCOORD(_currn)
_TERMACT_prootView_FormularCheckboxElement_value;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularCheckboxElement_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularCheckboxElement_checked (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularCheckboxElement_checked (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularCheckboxElement_checked _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularCheckboxElement_checked;
#else
_currn = (_TPPprootView_FormularCheckboxElement_checked) TreeNodeAlloc (sizeof (struct _TPprootView_FormularCheckboxElement_checked));
#endif
_currn->_prod = RULEprootView_FormularCheckboxElement_checked;
_SETCOORD(_currn)
_TERMACT_prootView_FormularCheckboxElement_checked;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularCheckboxElement_checked */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularSelectElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5, NODEPTR _desc6, NODEPTR _desc7, NODEPTR _desc8)
#else
NODEPTR MkprootView_FormularSelectElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5,_desc6,_desc7,_desc8)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
	NODEPTR _desc6;
	NODEPTR _desc7;
	NODEPTR _desc8;
#endif
{	_TPPprootView_FormularSelectElement _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularSelectElement;
#else
_currn = (_TPPprootView_FormularSelectElement) TreeNodeAlloc (sizeof (struct _TPprootView_FormularSelectElement));
#endif
_currn->_prod = RULEprootView_FormularSelectElement;
_currn->_desc1 = (_TSProotView_FormularSelectElement_name) MkrootView_FormularSelectElement_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularSelectElement: root of subtree no. 1 can not be made a rootView_FormularSelectElement_name node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_FormularSelectElement_label) MkrootView_FormularSelectElement_label (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularSelectElement: root of subtree no. 2 can not be made a rootView_FormularSelectElement_label node ", 0, _coordref);
_currn->_desc3 = (_TSProotView_FormularSelectElement_description) MkrootView_FormularSelectElement_description (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularSelectElement: root of subtree no. 3 can not be made a rootView_FormularSelectElement_description node ", 0, _coordref);
_currn->_desc4 = (_TSProotView_FormularSelectElement_pvalue) MkrootView_FormularSelectElement_pvalue (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularSelectElement: root of subtree no. 4 can not be made a rootView_FormularSelectElement_pvalue node ", 0, _coordref);
_currn->_desc5 = (_TSProotView_FormularSelectElement_condition) MkrootView_FormularSelectElement_condition (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularSelectElement: root of subtree no. 5 can not be made a rootView_FormularSelectElement_condition node ", 0, _coordref);
_currn->_desc6 = (_TSProotView_FormularSelectElement_options) MkrootView_FormularSelectElement_options (_coordref, _desc6);	
if (((NODEPTR)_currn->_desc6) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularSelectElement: root of subtree no. 6 can not be made a rootView_FormularSelectElement_options node ", 0, _coordref);
_currn->_desc7 = (_TSProotView_FormularSelectElement_size) MkrootView_FormularSelectElement_size (_coordref, _desc7);	
if (((NODEPTR)_currn->_desc7) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularSelectElement: root of subtree no. 7 can not be made a rootView_FormularSelectElement_size node ", 0, _coordref);
_currn->_desc8 = (_TSProotView_FormularSelectElement_multiple) MkrootView_FormularSelectElement_multiple (_coordref, _desc8);	
if (((NODEPTR)_currn->_desc8) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularSelectElement: root of subtree no. 8 can not be made a rootView_FormularSelectElement_multiple node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_FormularSelectElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_FormularSelectElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularSelectElement_name (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularSelectElement_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularSelectElement_name _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularSelectElement_name;
#else
_currn = (_TPPprootView_FormularSelectElement_name) TreeNodeAlloc (sizeof (struct _TPprootView_FormularSelectElement_name));
#endif
_currn->_prod = RULEprootView_FormularSelectElement_name;
_SETCOORD(_currn)
_TERMACT_prootView_FormularSelectElement_name;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularSelectElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularSelectElement_label (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularSelectElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularSelectElement_label _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularSelectElement_label;
#else
_currn = (_TPPprootView_FormularSelectElement_label) TreeNodeAlloc (sizeof (struct _TPprootView_FormularSelectElement_label));
#endif
_currn->_prod = RULEprootView_FormularSelectElement_label;
_SETCOORD(_currn)
_TERMACT_prootView_FormularSelectElement_label;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularSelectElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularSelectElement_description (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularSelectElement_description (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularSelectElement_description _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularSelectElement_description;
#else
_currn = (_TPPprootView_FormularSelectElement_description) TreeNodeAlloc (sizeof (struct _TPprootView_FormularSelectElement_description));
#endif
_currn->_prod = RULEprootView_FormularSelectElement_description;
_SETCOORD(_currn)
_TERMACT_prootView_FormularSelectElement_description;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularSelectElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularSelectElement_pvalue (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularSelectElement_pvalue (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularSelectElement_pvalue _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularSelectElement_pvalue;
#else
_currn = (_TPPprootView_FormularSelectElement_pvalue) TreeNodeAlloc (sizeof (struct _TPprootView_FormularSelectElement_pvalue));
#endif
_currn->_prod = RULEprootView_FormularSelectElement_pvalue;
_SETCOORD(_currn)
_TERMACT_prootView_FormularSelectElement_pvalue;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularSelectElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_FormularConstantSelectOptionprootView_FormularSelectElement_options (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_FormularConstantSelectOptionprootView_FormularSelectElement_options (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_FormularConstantSelectOptionprootView_FormularSelectElement_options _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_FormularConstantSelectOptionprootView_FormularSelectElement_options;
#else
	_new = (_TPPLST_rootView_FormularConstantSelectOptionprootView_FormularSelectElement_options) TreeNodeAlloc (sizeof (struct _TPLST_rootView_FormularConstantSelectOptionprootView_FormularSelectElement_options));
#endif
	_new->_prod = RULELST_rootView_FormularConstantSelectOptionprootView_FormularSelectElement_options;
	_new->_desc1 = (_TSProotView_FormularConstantSelectOption)elem;
	_new->_desc2 = (_TSPLST_rootView_FormularSelectElement_options)_new;
	{_TPPLST_rootView_FormularConstantSelectOptionprootView_FormularSelectElement_options _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_FormularConstantSelectOptionprootView_FormularSelectElement_options;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_FormularDataSourceSelectOptionprootView_FormularSelectElement_options (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_FormularDataSourceSelectOptionprootView_FormularSelectElement_options (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_FormularDataSourceSelectOptionprootView_FormularSelectElement_options _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_FormularDataSourceSelectOptionprootView_FormularSelectElement_options;
#else
	_new = (_TPPLST_rootView_FormularDataSourceSelectOptionprootView_FormularSelectElement_options) TreeNodeAlloc (sizeof (struct _TPLST_rootView_FormularDataSourceSelectOptionprootView_FormularSelectElement_options));
#endif
	_new->_prod = RULELST_rootView_FormularDataSourceSelectOptionprootView_FormularSelectElement_options;
	_new->_desc1 = (_TSProotView_FormularDataSourceSelectOption)elem;
	_new->_desc2 = (_TSPLST_rootView_FormularSelectElement_options)_new;
	{_TPPLST_rootView_FormularDataSourceSelectOptionprootView_FormularSelectElement_options _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_FormularDataSourceSelectOptionprootView_FormularSelectElement_options;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_FormularPHPScriptSelectOptionprootView_FormularSelectElement_options (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_FormularPHPScriptSelectOptionprootView_FormularSelectElement_options (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_FormularPHPScriptSelectOptionprootView_FormularSelectElement_options _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_FormularPHPScriptSelectOptionprootView_FormularSelectElement_options;
#else
	_new = (_TPPLST_rootView_FormularPHPScriptSelectOptionprootView_FormularSelectElement_options) TreeNodeAlloc (sizeof (struct _TPLST_rootView_FormularPHPScriptSelectOptionprootView_FormularSelectElement_options));
#endif
	_new->_prod = RULELST_rootView_FormularPHPScriptSelectOptionprootView_FormularSelectElement_options;
	_new->_desc1 = (_TSProotView_FormularPHPScriptSelectOption)elem;
	_new->_desc2 = (_TSPLST_rootView_FormularSelectElement_options)_new;
	{_TPPLST_rootView_FormularPHPScriptSelectOptionprootView_FormularSelectElement_options _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_FormularPHPScriptSelectOptionprootView_FormularSelectElement_options;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_FormularSelectElement_options (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_FormularSelectElement_options (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_FormularConstantSelectOption)) 
		return _MkrootView_FormularConstantSelectOptionprootView_FormularSelectElement_options (_coordref, MkrootView_FormularConstantSelectOption (_coordref, t));
	if (IsSymb (t, SYMBrootView_FormularDataSourceSelectOption)) 
		return _MkrootView_FormularDataSourceSelectOptionprootView_FormularSelectElement_options (_coordref, MkrootView_FormularDataSourceSelectOption (_coordref, t));
	if (IsSymb (t, SYMBrootView_FormularPHPScriptSelectOption)) 
		return _MkrootView_FormularPHPScriptSelectOptionprootView_FormularSelectElement_options (_coordref, MkrootView_FormularPHPScriptSelectOption (_coordref, t));
	message (DEADLY, "a prootView_FormularSelectElement_options list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_FormularSelectElement_options (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_FormularSelectElement_options (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_FormularSelectElement_options)
	l =_ElemprootView_FormularSelectElement_options (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_FormularSelectElement_options)
	r = _ElemprootView_FormularSelectElement_options (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_FormularSelectElement_options)r)->_desc2);
((_TPPLST_0prootView_FormularSelectElement_options)r)->_desc2 = ((_TPPLST_0prootView_FormularSelectElement_options)l)->_desc2;
((_TPPLST_0prootView_FormularSelectElement_options)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularSelectElement_options (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_FormularSelectElement_options (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_FormularSelectElement_options root;
	_TPPLST_0prootView_FormularSelectElement_options list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_FormularSelectElement_options)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_FormularSelectElement_options)
		t = _ElemprootView_FormularSelectElement_options (_coordref, t);
	remain = (_TPPLST_0prootView_FormularSelectElement_options)(((_TPPLST_0prootView_FormularSelectElement_options) t)->_desc2);
	((_TPPLST_0prootView_FormularSelectElement_options) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_FormularSelectElement_options;
#else
list = (_TPPLST_0prootView_FormularSelectElement_options) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_FormularSelectElement_options));
#endif
list->_prod = RULELST_0prootView_FormularSelectElement_options;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_FormularSelectElement_options)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_FormularConstantSelectOptionprootView_FormularSelectElement_options)
	{_TPPLST_rootView_FormularConstantSelectOptionprootView_FormularSelectElement_options _currn = _currn;
	 _TERMACT_LST_rootView_FormularConstantSelectOptionprootView_FormularSelectElement_options;
	} this code has been moved to _ElemprootView_FormularSelectElement_options*/
	/*if (_currn->_prod == RULELST_rootView_FormularDataSourceSelectOptionprootView_FormularSelectElement_options)
	{_TPPLST_rootView_FormularDataSourceSelectOptionprootView_FormularSelectElement_options _currn = _currn;
	 _TERMACT_LST_rootView_FormularDataSourceSelectOptionprootView_FormularSelectElement_options;
	} this code has been moved to _ElemprootView_FormularSelectElement_options*/
	/*if (_currn->_prod == RULELST_rootView_FormularPHPScriptSelectOptionprootView_FormularSelectElement_options)
	{_TPPLST_rootView_FormularPHPScriptSelectOptionprootView_FormularSelectElement_options _currn = _currn;
	 _TERMACT_LST_rootView_FormularPHPScriptSelectOptionprootView_FormularSelectElement_options;
	} this code has been moved to _ElemprootView_FormularSelectElement_options*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_FormularSelectElement_options;
#else
root = (_TPPprootView_FormularSelectElement_options) TreeNodeAlloc (sizeof (struct _TPprootView_FormularSelectElement_options));
#endif
root->_prod = RULEprootView_FormularSelectElement_options;
root->_desc1 = (_TSPLST_rootView_FormularSelectElement_options)list;
{ _TPPprootView_FormularSelectElement_options _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_FormularSelectElement_options;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularSelectElement_size (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularSelectElement_size (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularSelectElement_size _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularSelectElement_size;
#else
_currn = (_TPPprootView_FormularSelectElement_size) TreeNodeAlloc (sizeof (struct _TPprootView_FormularSelectElement_size));
#endif
_currn->_prod = RULEprootView_FormularSelectElement_size;
_SETCOORD(_currn)
_TERMACT_prootView_FormularSelectElement_size;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularSelectElement_size */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularSelectElement_multiple (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularSelectElement_multiple (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularSelectElement_multiple _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularSelectElement_multiple;
#else
_currn = (_TPPprootView_FormularSelectElement_multiple) TreeNodeAlloc (sizeof (struct _TPprootView_FormularSelectElement_multiple));
#endif
_currn->_prod = RULEprootView_FormularSelectElement_multiple;
_SETCOORD(_currn)
_TERMACT_prootView_FormularSelectElement_multiple;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularSelectElement_multiple */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularConstantSelectOption (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkprootView_FormularConstantSelectOption (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPprootView_FormularConstantSelectOption _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularConstantSelectOption;
#else
_currn = (_TPPprootView_FormularConstantSelectOption) TreeNodeAlloc (sizeof (struct _TPprootView_FormularConstantSelectOption));
#endif
_currn->_prod = RULEprootView_FormularConstantSelectOption;
_currn->_desc1 = (_TSProotView_FormularConstantSelectOption_values) MkrootView_FormularConstantSelectOption_values (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularConstantSelectOption: root of subtree no. 1 can not be made a rootView_FormularConstantSelectOption_values node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_FormularConstantSelectOption;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_FormularConstantSelectOption */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularConstantSelectOption_values (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularConstantSelectOption_values (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularConstantSelectOption_values _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularConstantSelectOption_values;
#else
_currn = (_TPPprootView_FormularConstantSelectOption_values) TreeNodeAlloc (sizeof (struct _TPprootView_FormularConstantSelectOption_values));
#endif
_currn->_prod = RULEprootView_FormularConstantSelectOption_values;
_SETCOORD(_currn)
_TERMACT_prootView_FormularConstantSelectOption_values;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularConstantSelectOption_values */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularDataSourceSelectOption (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkprootView_FormularDataSourceSelectOption (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPprootView_FormularDataSourceSelectOption _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularDataSourceSelectOption;
#else
_currn = (_TPPprootView_FormularDataSourceSelectOption) TreeNodeAlloc (sizeof (struct _TPprootView_FormularDataSourceSelectOption));
#endif
_currn->_prod = RULEprootView_FormularDataSourceSelectOption;
_currn->_desc1 = (_TSProotView_FormularDataSourceSelectOption_values) MkrootView_FormularDataSourceSelectOption_values (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularDataSourceSelectOption: root of subtree no. 1 can not be made a rootView_FormularDataSourceSelectOption_values node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_FormularDataSourceSelectOption_params) MkrootView_FormularDataSourceSelectOption_params (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularDataSourceSelectOption: root of subtree no. 2 can not be made a rootView_FormularDataSourceSelectOption_params node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_FormularDataSourceSelectOption;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_FormularDataSourceSelectOption */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularDataSourceSelectOption_values (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularDataSourceSelectOption_values (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularDataSourceSelectOption_values _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularDataSourceSelectOption_values;
#else
_currn = (_TPPprootView_FormularDataSourceSelectOption_values) TreeNodeAlloc (sizeof (struct _TPprootView_FormularDataSourceSelectOption_values));
#endif
_currn->_prod = RULEprootView_FormularDataSourceSelectOption_values;
_SETCOORD(_currn)
_TERMACT_prootView_FormularDataSourceSelectOption_values;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularDataSourceSelectOption_values */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_DataSourceActualParamprootView_FormularDataSourceSelectOption_params (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_DataSourceActualParamprootView_FormularDataSourceSelectOption_params (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_DataSourceActualParamprootView_FormularDataSourceSelectOption_params _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_DataSourceActualParamprootView_FormularDataSourceSelectOption_params;
#else
	_new = (_TPPLST_rootView_DataSourceActualParamprootView_FormularDataSourceSelectOption_params) TreeNodeAlloc (sizeof (struct _TPLST_rootView_DataSourceActualParamprootView_FormularDataSourceSelectOption_params));
#endif
	_new->_prod = RULELST_rootView_DataSourceActualParamprootView_FormularDataSourceSelectOption_params;
	_new->_desc1 = (_TSProotView_DataSourceActualParam)elem;
	_new->_desc2 = (_TSPLST_rootView_FormularDataSourceSelectOption_params)_new;
	{_TPPLST_rootView_DataSourceActualParamprootView_FormularDataSourceSelectOption_params _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_DataSourceActualParamprootView_FormularDataSourceSelectOption_params;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_FormularDataSourceSelectOption_params (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_FormularDataSourceSelectOption_params (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_DataSourceActualParam)) 
		return _MkrootView_DataSourceActualParamprootView_FormularDataSourceSelectOption_params (_coordref, MkrootView_DataSourceActualParam (_coordref, t));
	message (DEADLY, "a prootView_FormularDataSourceSelectOption_params list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_FormularDataSourceSelectOption_params (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_FormularDataSourceSelectOption_params (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_FormularDataSourceSelectOption_params)
	l =_ElemprootView_FormularDataSourceSelectOption_params (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_FormularDataSourceSelectOption_params)
	r = _ElemprootView_FormularDataSourceSelectOption_params (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_FormularDataSourceSelectOption_params)r)->_desc2);
((_TPPLST_0prootView_FormularDataSourceSelectOption_params)r)->_desc2 = ((_TPPLST_0prootView_FormularDataSourceSelectOption_params)l)->_desc2;
((_TPPLST_0prootView_FormularDataSourceSelectOption_params)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularDataSourceSelectOption_params (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_FormularDataSourceSelectOption_params (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_FormularDataSourceSelectOption_params root;
	_TPPLST_0prootView_FormularDataSourceSelectOption_params list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_FormularDataSourceSelectOption_params)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_FormularDataSourceSelectOption_params)
		t = _ElemprootView_FormularDataSourceSelectOption_params (_coordref, t);
	remain = (_TPPLST_0prootView_FormularDataSourceSelectOption_params)(((_TPPLST_0prootView_FormularDataSourceSelectOption_params) t)->_desc2);
	((_TPPLST_0prootView_FormularDataSourceSelectOption_params) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_FormularDataSourceSelectOption_params;
#else
list = (_TPPLST_0prootView_FormularDataSourceSelectOption_params) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_FormularDataSourceSelectOption_params));
#endif
list->_prod = RULELST_0prootView_FormularDataSourceSelectOption_params;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_FormularDataSourceSelectOption_params)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_DataSourceActualParamprootView_FormularDataSourceSelectOption_params)
	{_TPPLST_rootView_DataSourceActualParamprootView_FormularDataSourceSelectOption_params _currn = _currn;
	 _TERMACT_LST_rootView_DataSourceActualParamprootView_FormularDataSourceSelectOption_params;
	} this code has been moved to _ElemprootView_FormularDataSourceSelectOption_params*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_FormularDataSourceSelectOption_params;
#else
root = (_TPPprootView_FormularDataSourceSelectOption_params) TreeNodeAlloc (sizeof (struct _TPprootView_FormularDataSourceSelectOption_params));
#endif
root->_prod = RULEprootView_FormularDataSourceSelectOption_params;
root->_desc1 = (_TSPLST_rootView_FormularDataSourceSelectOption_params)list;
{ _TPPprootView_FormularDataSourceSelectOption_params _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_FormularDataSourceSelectOption_params;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularPHPScriptSelectOption (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkprootView_FormularPHPScriptSelectOption (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPprootView_FormularPHPScriptSelectOption _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularPHPScriptSelectOption;
#else
_currn = (_TPPprootView_FormularPHPScriptSelectOption) TreeNodeAlloc (sizeof (struct _TPprootView_FormularPHPScriptSelectOption));
#endif
_currn->_prod = RULEprootView_FormularPHPScriptSelectOption;
_currn->_desc1 = (_TSProotView_FormularPHPScriptSelectOption_script) MkrootView_FormularPHPScriptSelectOption_script (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularPHPScriptSelectOption: root of subtree no. 1 can not be made a rootView_FormularPHPScriptSelectOption_script node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_FormularPHPScriptSelectOption;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_FormularPHPScriptSelectOption */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularPHPScriptSelectOption_script (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularPHPScriptSelectOption_script (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularPHPScriptSelectOption_script _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularPHPScriptSelectOption_script;
#else
_currn = (_TPPprootView_FormularPHPScriptSelectOption_script) TreeNodeAlloc (sizeof (struct _TPprootView_FormularPHPScriptSelectOption_script));
#endif
_currn->_prod = RULEprootView_FormularPHPScriptSelectOption_script;
_SETCOORD(_currn)
_TERMACT_prootView_FormularPHPScriptSelectOption_script;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularPHPScriptSelectOption_script */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularRadioElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkprootView_FormularRadioElement (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPprootView_FormularRadioElement _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularRadioElement;
#else
_currn = (_TPPprootView_FormularRadioElement) TreeNodeAlloc (sizeof (struct _TPprootView_FormularRadioElement));
#endif
_currn->_prod = RULEprootView_FormularRadioElement;
_currn->_desc1 = (_TSProotView_FormularRadioElement_value) MkrootView_FormularRadioElement_value (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularRadioElement: root of subtree no. 1 can not be made a rootView_FormularRadioElement_value node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_FormularRadioElement_label) MkrootView_FormularRadioElement_label (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularRadioElement: root of subtree no. 2 can not be made a rootView_FormularRadioElement_label node ", 0, _coordref);
_currn->_desc3 = (_TSProotView_FormularRadioElement_name) MkrootView_FormularRadioElement_name (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularRadioElement: root of subtree no. 3 can not be made a rootView_FormularRadioElement_name node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_FormularRadioElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_FormularRadioElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularRadioElement_value (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularRadioElement_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularRadioElement_value _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularRadioElement_value;
#else
_currn = (_TPPprootView_FormularRadioElement_value) TreeNodeAlloc (sizeof (struct _TPprootView_FormularRadioElement_value));
#endif
_currn->_prod = RULEprootView_FormularRadioElement_value;
_SETCOORD(_currn)
_TERMACT_prootView_FormularRadioElement_value;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularRadioElement_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularRadioElement_label (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularRadioElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularRadioElement_label _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularRadioElement_label;
#else
_currn = (_TPPprootView_FormularRadioElement_label) TreeNodeAlloc (sizeof (struct _TPprootView_FormularRadioElement_label));
#endif
_currn->_prod = RULEprootView_FormularRadioElement_label;
_SETCOORD(_currn)
_TERMACT_prootView_FormularRadioElement_label;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularRadioElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularRadioElement_name (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularRadioElement_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularRadioElement_name _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularRadioElement_name;
#else
_currn = (_TPPprootView_FormularRadioElement_name) TreeNodeAlloc (sizeof (struct _TPprootView_FormularRadioElement_name));
#endif
_currn->_prod = RULEprootView_FormularRadioElement_name;
_SETCOORD(_currn)
_TERMACT_prootView_FormularRadioElement_name;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularRadioElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularCondition (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkprootView_FormularCondition (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPprootView_FormularCondition _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularCondition;
#else
_currn = (_TPPprootView_FormularCondition) TreeNodeAlloc (sizeof (struct _TPprootView_FormularCondition));
#endif
_currn->_prod = RULEprootView_FormularCondition;
_currn->_desc1 = (_TSProotView_FormularCondition_condition) MkrootView_FormularCondition_condition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularCondition: root of subtree no. 1 can not be made a rootView_FormularCondition_condition node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_FormularCondition_params) MkrootView_FormularCondition_params (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularCondition: root of subtree no. 2 can not be made a rootView_FormularCondition_params node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_FormularCondition;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_FormularCondition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularCondition_condition (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularCondition_condition (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularCondition_condition _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularCondition_condition;
#else
_currn = (_TPPprootView_FormularCondition_condition) TreeNodeAlloc (sizeof (struct _TPprootView_FormularCondition_condition));
#endif
_currn->_prod = RULEprootView_FormularCondition_condition;
_SETCOORD(_currn)
_TERMACT_prootView_FormularCondition_condition;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularCondition_condition */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_FormularConditionActualParamprootView_FormularCondition_params (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_FormularConditionActualParamprootView_FormularCondition_params (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_FormularConditionActualParamprootView_FormularCondition_params _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_FormularConditionActualParamprootView_FormularCondition_params;
#else
	_new = (_TPPLST_rootView_FormularConditionActualParamprootView_FormularCondition_params) TreeNodeAlloc (sizeof (struct _TPLST_rootView_FormularConditionActualParamprootView_FormularCondition_params));
#endif
	_new->_prod = RULELST_rootView_FormularConditionActualParamprootView_FormularCondition_params;
	_new->_desc1 = (_TSProotView_FormularConditionActualParam)elem;
	_new->_desc2 = (_TSPLST_rootView_FormularCondition_params)_new;
	{_TPPLST_rootView_FormularConditionActualParamprootView_FormularCondition_params _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_FormularConditionActualParamprootView_FormularCondition_params;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_FormularCondition_params (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_FormularCondition_params (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_FormularConditionActualParam)) 
		return _MkrootView_FormularConditionActualParamprootView_FormularCondition_params (_coordref, MkrootView_FormularConditionActualParam (_coordref, t));
	message (DEADLY, "a prootView_FormularCondition_params list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_FormularCondition_params (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_FormularCondition_params (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_FormularCondition_params)
	l =_ElemprootView_FormularCondition_params (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_FormularCondition_params)
	r = _ElemprootView_FormularCondition_params (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_FormularCondition_params)r)->_desc2);
((_TPPLST_0prootView_FormularCondition_params)r)->_desc2 = ((_TPPLST_0prootView_FormularCondition_params)l)->_desc2;
((_TPPLST_0prootView_FormularCondition_params)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularCondition_params (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_FormularCondition_params (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_FormularCondition_params root;
	_TPPLST_0prootView_FormularCondition_params list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_FormularCondition_params)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_FormularCondition_params)
		t = _ElemprootView_FormularCondition_params (_coordref, t);
	remain = (_TPPLST_0prootView_FormularCondition_params)(((_TPPLST_0prootView_FormularCondition_params) t)->_desc2);
	((_TPPLST_0prootView_FormularCondition_params) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_FormularCondition_params;
#else
list = (_TPPLST_0prootView_FormularCondition_params) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_FormularCondition_params));
#endif
list->_prod = RULELST_0prootView_FormularCondition_params;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_FormularCondition_params)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_FormularConditionActualParamprootView_FormularCondition_params)
	{_TPPLST_rootView_FormularConditionActualParamprootView_FormularCondition_params _currn = _currn;
	 _TERMACT_LST_rootView_FormularConditionActualParamprootView_FormularCondition_params;
	} this code has been moved to _ElemprootView_FormularCondition_params*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_FormularCondition_params;
#else
root = (_TPPprootView_FormularCondition_params) TreeNodeAlloc (sizeof (struct _TPprootView_FormularCondition_params));
#endif
root->_prod = RULEprootView_FormularCondition_params;
root->_desc1 = (_TSPLST_rootView_FormularCondition_params)list;
{ _TPPprootView_FormularCondition_params _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_FormularCondition_params;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularConditionDefinition (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkprootView_FormularConditionDefinition (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPprootView_FormularConditionDefinition _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularConditionDefinition;
#else
_currn = (_TPPprootView_FormularConditionDefinition) TreeNodeAlloc (sizeof (struct _TPprootView_FormularConditionDefinition));
#endif
_currn->_prod = RULEprootView_FormularConditionDefinition;
_currn->_desc1 = (_TSProotView_FormularConditionDefinition_name) MkrootView_FormularConditionDefinition_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularConditionDefinition: root of subtree no. 1 can not be made a rootView_FormularConditionDefinition_name node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_FormularConditionDefinition_params) MkrootView_FormularConditionDefinition_params (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularConditionDefinition: root of subtree no. 2 can not be made a rootView_FormularConditionDefinition_params node ", 0, _coordref);
_currn->_desc3 = (_TSProotView_FormularConditionDefinition_body) MkrootView_FormularConditionDefinition_body (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularConditionDefinition: root of subtree no. 3 can not be made a rootView_FormularConditionDefinition_body node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_FormularConditionDefinition;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_FormularConditionDefinition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularConditionDefinition_name (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularConditionDefinition_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularConditionDefinition_name _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularConditionDefinition_name;
#else
_currn = (_TPPprootView_FormularConditionDefinition_name) TreeNodeAlloc (sizeof (struct _TPprootView_FormularConditionDefinition_name));
#endif
_currn->_prod = RULEprootView_FormularConditionDefinition_name;
_SETCOORD(_currn)
_TERMACT_prootView_FormularConditionDefinition_name;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularConditionDefinition_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_FormularConditionFormalParamprootView_FormularConditionDefinition_params (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_FormularConditionFormalParamprootView_FormularConditionDefinition_params (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_FormularConditionFormalParamprootView_FormularConditionDefinition_params _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_FormularConditionFormalParamprootView_FormularConditionDefinition_params;
#else
	_new = (_TPPLST_rootView_FormularConditionFormalParamprootView_FormularConditionDefinition_params) TreeNodeAlloc (sizeof (struct _TPLST_rootView_FormularConditionFormalParamprootView_FormularConditionDefinition_params));
#endif
	_new->_prod = RULELST_rootView_FormularConditionFormalParamprootView_FormularConditionDefinition_params;
	_new->_desc1 = (_TSProotView_FormularConditionFormalParam)elem;
	_new->_desc2 = (_TSPLST_rootView_FormularConditionDefinition_params)_new;
	{_TPPLST_rootView_FormularConditionFormalParamprootView_FormularConditionDefinition_params _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_FormularConditionFormalParamprootView_FormularConditionDefinition_params;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_FormularConditionDefinition_params (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_FormularConditionDefinition_params (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_FormularConditionFormalParam)) 
		return _MkrootView_FormularConditionFormalParamprootView_FormularConditionDefinition_params (_coordref, MkrootView_FormularConditionFormalParam (_coordref, t));
	message (DEADLY, "a prootView_FormularConditionDefinition_params list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_FormularConditionDefinition_params (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_FormularConditionDefinition_params (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_FormularConditionDefinition_params)
	l =_ElemprootView_FormularConditionDefinition_params (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_FormularConditionDefinition_params)
	r = _ElemprootView_FormularConditionDefinition_params (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_FormularConditionDefinition_params)r)->_desc2);
((_TPPLST_0prootView_FormularConditionDefinition_params)r)->_desc2 = ((_TPPLST_0prootView_FormularConditionDefinition_params)l)->_desc2;
((_TPPLST_0prootView_FormularConditionDefinition_params)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularConditionDefinition_params (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_FormularConditionDefinition_params (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_FormularConditionDefinition_params root;
	_TPPLST_0prootView_FormularConditionDefinition_params list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_FormularConditionDefinition_params)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_FormularConditionDefinition_params)
		t = _ElemprootView_FormularConditionDefinition_params (_coordref, t);
	remain = (_TPPLST_0prootView_FormularConditionDefinition_params)(((_TPPLST_0prootView_FormularConditionDefinition_params) t)->_desc2);
	((_TPPLST_0prootView_FormularConditionDefinition_params) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_FormularConditionDefinition_params;
#else
list = (_TPPLST_0prootView_FormularConditionDefinition_params) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_FormularConditionDefinition_params));
#endif
list->_prod = RULELST_0prootView_FormularConditionDefinition_params;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_FormularConditionDefinition_params)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_FormularConditionFormalParamprootView_FormularConditionDefinition_params)
	{_TPPLST_rootView_FormularConditionFormalParamprootView_FormularConditionDefinition_params _currn = _currn;
	 _TERMACT_LST_rootView_FormularConditionFormalParamprootView_FormularConditionDefinition_params;
	} this code has been moved to _ElemprootView_FormularConditionDefinition_params*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_FormularConditionDefinition_params;
#else
root = (_TPPprootView_FormularConditionDefinition_params) TreeNodeAlloc (sizeof (struct _TPprootView_FormularConditionDefinition_params));
#endif
root->_prod = RULEprootView_FormularConditionDefinition_params;
root->_desc1 = (_TSPLST_rootView_FormularConditionDefinition_params)list;
{ _TPPprootView_FormularConditionDefinition_params _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_FormularConditionDefinition_params;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularConditionDefinition_body (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularConditionDefinition_body (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularConditionDefinition_body _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularConditionDefinition_body;
#else
_currn = (_TPPprootView_FormularConditionDefinition_body) TreeNodeAlloc (sizeof (struct _TPprootView_FormularConditionDefinition_body));
#endif
_currn->_prod = RULEprootView_FormularConditionDefinition_body;
_SETCOORD(_currn)
_TERMACT_prootView_FormularConditionDefinition_body;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularConditionDefinition_body */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularConditionFormalParam (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkprootView_FormularConditionFormalParam (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPprootView_FormularConditionFormalParam _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularConditionFormalParam;
#else
_currn = (_TPPprootView_FormularConditionFormalParam) TreeNodeAlloc (sizeof (struct _TPprootView_FormularConditionFormalParam));
#endif
_currn->_prod = RULEprootView_FormularConditionFormalParam;
_currn->_desc1 = (_TSProotView_FormularConditionFormalParam_name) MkrootView_FormularConditionFormalParam_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularConditionFormalParam: root of subtree no. 1 can not be made a rootView_FormularConditionFormalParam_name node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_FormularConditionFormalParam;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_FormularConditionFormalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularConditionFormalParam_name (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularConditionFormalParam_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularConditionFormalParam_name _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularConditionFormalParam_name;
#else
_currn = (_TPPprootView_FormularConditionFormalParam_name) TreeNodeAlloc (sizeof (struct _TPprootView_FormularConditionFormalParam_name));
#endif
_currn->_prod = RULEprootView_FormularConditionFormalParam_name;
_SETCOORD(_currn)
_TERMACT_prootView_FormularConditionFormalParam_name;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularConditionFormalParam_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularConditionActualParam (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkprootView_FormularConditionActualParam (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPprootView_FormularConditionActualParam _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularConditionActualParam;
#else
_currn = (_TPPprootView_FormularConditionActualParam) TreeNodeAlloc (sizeof (struct _TPprootView_FormularConditionActualParam));
#endif
_currn->_prod = RULEprootView_FormularConditionActualParam;
_currn->_desc1 = (_TSProotView_FormularConditionActualParam_value) MkrootView_FormularConditionActualParam_value (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularConditionActualParam: root of subtree no. 1 can not be made a rootView_FormularConditionActualParam_value node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_FormularConditionActualParam_formalparam) MkrootView_FormularConditionActualParam_formalparam (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FormularConditionActualParam: root of subtree no. 2 can not be made a rootView_FormularConditionActualParam_formalparam node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_FormularConditionActualParam;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_FormularConditionActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularConditionActualParam_value (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularConditionActualParam_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularConditionActualParam_value _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularConditionActualParam_value;
#else
_currn = (_TPPprootView_FormularConditionActualParam_value) TreeNodeAlloc (sizeof (struct _TPprootView_FormularConditionActualParam_value));
#endif
_currn->_prod = RULEprootView_FormularConditionActualParam_value;
_SETCOORD(_currn)
_TERMACT_prootView_FormularConditionActualParam_value;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularConditionActualParam_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FormularConditionActualParam_formalparam (POSITION *_coordref)
#else
NODEPTR MkprootView_FormularConditionActualParam_formalparam (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FormularConditionActualParam_formalparam _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FormularConditionActualParam_formalparam;
#else
_currn = (_TPPprootView_FormularConditionActualParam_formalparam) TreeNodeAlloc (sizeof (struct _TPprootView_FormularConditionActualParam_formalparam));
#endif
_currn->_prod = RULEprootView_FormularConditionActualParam_formalparam;
_SETCOORD(_currn)
_TERMACT_prootView_FormularConditionActualParam_formalparam;
return ( (NODEPTR) _currn);
}/* MkprootView_FormularConditionActualParam_formalparam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebNodeActualParam (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkprootView_WebNodeActualParam (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPprootView_WebNodeActualParam _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebNodeActualParam;
#else
_currn = (_TPPprootView_WebNodeActualParam) TreeNodeAlloc (sizeof (struct _TPprootView_WebNodeActualParam));
#endif
_currn->_prod = RULEprootView_WebNodeActualParam;
_currn->_desc1 = (_TSProotView_WebNodeActualParam_formalParam) MkrootView_WebNodeActualParam_formalParam (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebNodeActualParam: root of subtree no. 1 can not be made a rootView_WebNodeActualParam_formalParam node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_WebNodeActualParam_expressions) MkrootView_WebNodeActualParam_expressions (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebNodeActualParam: root of subtree no. 2 can not be made a rootView_WebNodeActualParam_expressions node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_WebNodeActualParam;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_WebNodeActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebNodeActualParam_formalParam (POSITION *_coordref)
#else
NODEPTR MkprootView_WebNodeActualParam_formalParam (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_WebNodeActualParam_formalParam _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebNodeActualParam_formalParam;
#else
_currn = (_TPPprootView_WebNodeActualParam_formalParam) TreeNodeAlloc (sizeof (struct _TPprootView_WebNodeActualParam_formalParam));
#endif
_currn->_prod = RULEprootView_WebNodeActualParam_formalParam;
_SETCOORD(_currn)
_TERMACT_prootView_WebNodeActualParam_formalParam;
return ( (NODEPTR) _currn);
}/* MkprootView_WebNodeActualParam_formalParam */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_WebNodeConstantExpressionprootView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_WebNodeConstantExpressionprootView_WebNodeActualParam_expressions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_WebNodeConstantExpressionprootView_WebNodeActualParam_expressions _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_WebNodeConstantExpressionprootView_WebNodeActualParam_expressions;
#else
	_new = (_TPPLST_rootView_WebNodeConstantExpressionprootView_WebNodeActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_rootView_WebNodeConstantExpressionprootView_WebNodeActualParam_expressions));
#endif
	_new->_prod = RULELST_rootView_WebNodeConstantExpressionprootView_WebNodeActualParam_expressions;
	_new->_desc1 = (_TSProotView_WebNodeConstantExpression)elem;
	_new->_desc2 = (_TSPLST_rootView_WebNodeActualParam_expressions)_new;
	{_TPPLST_rootView_WebNodeConstantExpressionprootView_WebNodeActualParam_expressions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_WebNodeConstantExpressionprootView_WebNodeActualParam_expressions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_WebNodeVariableExpressionprootView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_WebNodeVariableExpressionprootView_WebNodeActualParam_expressions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_WebNodeVariableExpressionprootView_WebNodeActualParam_expressions _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_WebNodeVariableExpressionprootView_WebNodeActualParam_expressions;
#else
	_new = (_TPPLST_rootView_WebNodeVariableExpressionprootView_WebNodeActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_rootView_WebNodeVariableExpressionprootView_WebNodeActualParam_expressions));
#endif
	_new->_prod = RULELST_rootView_WebNodeVariableExpressionprootView_WebNodeActualParam_expressions;
	_new->_desc1 = (_TSProotView_WebNodeVariableExpression)elem;
	_new->_desc2 = (_TSPLST_rootView_WebNodeActualParam_expressions)_new;
	{_TPPLST_rootView_WebNodeVariableExpressionprootView_WebNodeActualParam_expressions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_WebNodeVariableExpressionprootView_WebNodeActualParam_expressions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_WebNodeColumnExpressionprootView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_WebNodeColumnExpressionprootView_WebNodeActualParam_expressions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_WebNodeColumnExpressionprootView_WebNodeActualParam_expressions _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_WebNodeColumnExpressionprootView_WebNodeActualParam_expressions;
#else
	_new = (_TPPLST_rootView_WebNodeColumnExpressionprootView_WebNodeActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_rootView_WebNodeColumnExpressionprootView_WebNodeActualParam_expressions));
#endif
	_new->_prod = RULELST_rootView_WebNodeColumnExpressionprootView_WebNodeActualParam_expressions;
	_new->_desc1 = (_TSProotView_WebNodeColumnExpression)elem;
	_new->_desc2 = (_TSPLST_rootView_WebNodeActualParam_expressions)_new;
	{_TPPLST_rootView_WebNodeColumnExpressionprootView_WebNodeActualParam_expressions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_WebNodeColumnExpressionprootView_WebNodeActualParam_expressions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_WebNodeActualParam_expressions (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_WebNodeConstantExpression)) 
		return _MkrootView_WebNodeConstantExpressionprootView_WebNodeActualParam_expressions (_coordref, MkrootView_WebNodeConstantExpression (_coordref, t));
	if (IsSymb (t, SYMBrootView_WebNodeVariableExpression)) 
		return _MkrootView_WebNodeVariableExpressionprootView_WebNodeActualParam_expressions (_coordref, MkrootView_WebNodeVariableExpression (_coordref, t));
	if (IsSymb (t, SYMBrootView_WebNodeColumnExpression)) 
		return _MkrootView_WebNodeColumnExpressionprootView_WebNodeActualParam_expressions (_coordref, MkrootView_WebNodeColumnExpression (_coordref, t));
	message (DEADLY, "a prootView_WebNodeActualParam_expressions list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_WebNodeActualParam_expressions (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_WebNodeActualParam_expressions)
	l =_ElemprootView_WebNodeActualParam_expressions (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_WebNodeActualParam_expressions)
	r = _ElemprootView_WebNodeActualParam_expressions (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_WebNodeActualParam_expressions)r)->_desc2);
((_TPPLST_0prootView_WebNodeActualParam_expressions)r)->_desc2 = ((_TPPLST_0prootView_WebNodeActualParam_expressions)l)->_desc2;
((_TPPLST_0prootView_WebNodeActualParam_expressions)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_WebNodeActualParam_expressions (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_WebNodeActualParam_expressions root;
	_TPPLST_0prootView_WebNodeActualParam_expressions list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_WebNodeActualParam_expressions)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_WebNodeActualParam_expressions)
		t = _ElemprootView_WebNodeActualParam_expressions (_coordref, t);
	remain = (_TPPLST_0prootView_WebNodeActualParam_expressions)(((_TPPLST_0prootView_WebNodeActualParam_expressions) t)->_desc2);
	((_TPPLST_0prootView_WebNodeActualParam_expressions) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_WebNodeActualParam_expressions;
#else
list = (_TPPLST_0prootView_WebNodeActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_WebNodeActualParam_expressions));
#endif
list->_prod = RULELST_0prootView_WebNodeActualParam_expressions;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_WebNodeActualParam_expressions)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_WebNodeConstantExpressionprootView_WebNodeActualParam_expressions)
	{_TPPLST_rootView_WebNodeConstantExpressionprootView_WebNodeActualParam_expressions _currn = _currn;
	 _TERMACT_LST_rootView_WebNodeConstantExpressionprootView_WebNodeActualParam_expressions;
	} this code has been moved to _ElemprootView_WebNodeActualParam_expressions*/
	/*if (_currn->_prod == RULELST_rootView_WebNodeVariableExpressionprootView_WebNodeActualParam_expressions)
	{_TPPLST_rootView_WebNodeVariableExpressionprootView_WebNodeActualParam_expressions _currn = _currn;
	 _TERMACT_LST_rootView_WebNodeVariableExpressionprootView_WebNodeActualParam_expressions;
	} this code has been moved to _ElemprootView_WebNodeActualParam_expressions*/
	/*if (_currn->_prod == RULELST_rootView_WebNodeColumnExpressionprootView_WebNodeActualParam_expressions)
	{_TPPLST_rootView_WebNodeColumnExpressionprootView_WebNodeActualParam_expressions _currn = _currn;
	 _TERMACT_LST_rootView_WebNodeColumnExpressionprootView_WebNodeActualParam_expressions;
	} this code has been moved to _ElemprootView_WebNodeActualParam_expressions*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_WebNodeActualParam_expressions;
#else
root = (_TPPprootView_WebNodeActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPprootView_WebNodeActualParam_expressions));
#endif
root->_prod = RULEprootView_WebNodeActualParam_expressions;
root->_desc1 = (_TSPLST_rootView_WebNodeActualParam_expressions)list;
{ _TPPprootView_WebNodeActualParam_expressions _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_WebNodeActualParam_expressions;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebNodeConstantExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkprootView_WebNodeConstantExpression (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPprootView_WebNodeConstantExpression _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebNodeConstantExpression;
#else
_currn = (_TPPprootView_WebNodeConstantExpression) TreeNodeAlloc (sizeof (struct _TPprootView_WebNodeConstantExpression));
#endif
_currn->_prod = RULEprootView_WebNodeConstantExpression;
_currn->_desc1 = (_TSProotView_WebNodeConstantExpression_value) MkrootView_WebNodeConstantExpression_value (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebNodeConstantExpression: root of subtree no. 1 can not be made a rootView_WebNodeConstantExpression_value node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_WebNodeConstantExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_WebNodeConstantExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebNodeConstantExpression_value (POSITION *_coordref)
#else
NODEPTR MkprootView_WebNodeConstantExpression_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_WebNodeConstantExpression_value _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebNodeConstantExpression_value;
#else
_currn = (_TPPprootView_WebNodeConstantExpression_value) TreeNodeAlloc (sizeof (struct _TPprootView_WebNodeConstantExpression_value));
#endif
_currn->_prod = RULEprootView_WebNodeConstantExpression_value;
_SETCOORD(_currn)
_TERMACT_prootView_WebNodeConstantExpression_value;
return ( (NODEPTR) _currn);
}/* MkprootView_WebNodeConstantExpression_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebNodeVariableExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkprootView_WebNodeVariableExpression (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPprootView_WebNodeVariableExpression _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebNodeVariableExpression;
#else
_currn = (_TPPprootView_WebNodeVariableExpression) TreeNodeAlloc (sizeof (struct _TPprootView_WebNodeVariableExpression));
#endif
_currn->_prod = RULEprootView_WebNodeVariableExpression;
_currn->_desc1 = (_TSProotView_WebNodeVariableExpression_content) MkrootView_WebNodeVariableExpression_content (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebNodeVariableExpression: root of subtree no. 1 can not be made a rootView_WebNodeVariableExpression_content node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_WebNodeVariableExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_WebNodeVariableExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebNodeVariableExpression_content (POSITION *_coordref)
#else
NODEPTR MkprootView_WebNodeVariableExpression_content (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_WebNodeVariableExpression_content _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebNodeVariableExpression_content;
#else
_currn = (_TPPprootView_WebNodeVariableExpression_content) TreeNodeAlloc (sizeof (struct _TPprootView_WebNodeVariableExpression_content));
#endif
_currn->_prod = RULEprootView_WebNodeVariableExpression_content;
_SETCOORD(_currn)
_TERMACT_prootView_WebNodeVariableExpression_content;
return ( (NODEPTR) _currn);
}/* MkprootView_WebNodeVariableExpression_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebNodeColumnExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkprootView_WebNodeColumnExpression (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPprootView_WebNodeColumnExpression _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebNodeColumnExpression;
#else
_currn = (_TPPprootView_WebNodeColumnExpression) TreeNodeAlloc (sizeof (struct _TPprootView_WebNodeColumnExpression));
#endif
_currn->_prod = RULEprootView_WebNodeColumnExpression;
_currn->_desc1 = (_TSProotView_WebNodeColumnExpression_content) MkrootView_WebNodeColumnExpression_content (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebNodeColumnExpression: root of subtree no. 1 can not be made a rootView_WebNodeColumnExpression_content node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_WebNodeColumnExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_WebNodeColumnExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebNodeColumnExpression_content (POSITION *_coordref)
#else
NODEPTR MkprootView_WebNodeColumnExpression_content (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_WebNodeColumnExpression_content _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebNodeColumnExpression_content;
#else
_currn = (_TPPprootView_WebNodeColumnExpression_content) TreeNodeAlloc (sizeof (struct _TPprootView_WebNodeColumnExpression_content));
#endif
_currn->_prod = RULEprootView_WebNodeColumnExpression_content;
_SETCOORD(_currn)
_TERMACT_prootView_WebNodeColumnExpression_content;
return ( (NODEPTR) _currn);
}/* MkprootView_WebNodeColumnExpression_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_Paragraph (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkprootView_Paragraph (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPprootView_Paragraph _currn;
#ifdef __cplusplus
_currn = new _TPprootView_Paragraph;
#else
_currn = (_TPPprootView_Paragraph) TreeNodeAlloc (sizeof (struct _TPprootView_Paragraph));
#endif
_currn->_prod = RULEprootView_Paragraph;
_currn->_desc1 = (_TSProotView_Paragraph_styleSheetBlock) MkrootView_Paragraph_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_Paragraph: root of subtree no. 1 can not be made a rootView_Paragraph_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_Paragraph_fragments) MkrootView_Paragraph_fragments (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_Paragraph: root of subtree no. 2 can not be made a rootView_Paragraph_fragments node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_Paragraph;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_Paragraph */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_Paragraph_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkprootView_Paragraph_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_Paragraph_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPprootView_Paragraph_styleSheetBlock;
#else
_currn = (_TPPprootView_Paragraph_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPprootView_Paragraph_styleSheetBlock));
#endif
_currn->_prod = RULEprootView_Paragraph_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_prootView_Paragraph_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkprootView_Paragraph_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_TextFragmentprootView_Paragraph_fragments (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_TextFragmentprootView_Paragraph_fragments (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_TextFragmentprootView_Paragraph_fragments _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_TextFragmentprootView_Paragraph_fragments;
#else
	_new = (_TPPLST_rootView_TextFragmentprootView_Paragraph_fragments) TreeNodeAlloc (sizeof (struct _TPLST_rootView_TextFragmentprootView_Paragraph_fragments));
#endif
	_new->_prod = RULELST_rootView_TextFragmentprootView_Paragraph_fragments;
	_new->_desc1 = (_TSProotView_TextFragment)elem;
	_new->_desc2 = (_TSPLST_rootView_Paragraph_fragments)_new;
	{_TPPLST_rootView_TextFragmentprootView_Paragraph_fragments _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_TextFragmentprootView_Paragraph_fragments;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_ImageFragmentprootView_Paragraph_fragments (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_ImageFragmentprootView_Paragraph_fragments (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_ImageFragmentprootView_Paragraph_fragments _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_ImageFragmentprootView_Paragraph_fragments;
#else
	_new = (_TPPLST_rootView_ImageFragmentprootView_Paragraph_fragments) TreeNodeAlloc (sizeof (struct _TPLST_rootView_ImageFragmentprootView_Paragraph_fragments));
#endif
	_new->_prod = RULELST_rootView_ImageFragmentprootView_Paragraph_fragments;
	_new->_desc1 = (_TSProotView_ImageFragment)elem;
	_new->_desc2 = (_TSPLST_rootView_Paragraph_fragments)_new;
	{_TPPLST_rootView_ImageFragmentprootView_Paragraph_fragments _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_ImageFragmentprootView_Paragraph_fragments;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_OutputElementprootView_Paragraph_fragments (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_OutputElementprootView_Paragraph_fragments (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_OutputElementprootView_Paragraph_fragments _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_OutputElementprootView_Paragraph_fragments;
#else
	_new = (_TPPLST_rootView_OutputElementprootView_Paragraph_fragments) TreeNodeAlloc (sizeof (struct _TPLST_rootView_OutputElementprootView_Paragraph_fragments));
#endif
	_new->_prod = RULELST_rootView_OutputElementprootView_Paragraph_fragments;
	_new->_desc1 = (_TSProotView_OutputElement)elem;
	_new->_desc2 = (_TSPLST_rootView_Paragraph_fragments)_new;
	{_TPPLST_rootView_OutputElementprootView_Paragraph_fragments _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_OutputElementprootView_Paragraph_fragments;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_Paragraph_fragments (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_Paragraph_fragments (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_TextFragment)) 
		return _MkrootView_TextFragmentprootView_Paragraph_fragments (_coordref, MkrootView_TextFragment (_coordref, t));
	if (IsSymb (t, SYMBrootView_ImageFragment)) 
		return _MkrootView_ImageFragmentprootView_Paragraph_fragments (_coordref, MkrootView_ImageFragment (_coordref, t));
	if (IsSymb (t, SYMBrootView_OutputElement)) 
		return _MkrootView_OutputElementprootView_Paragraph_fragments (_coordref, MkrootView_OutputElement (_coordref, t));
	message (DEADLY, "a prootView_Paragraph_fragments list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_Paragraph_fragments (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_Paragraph_fragments (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_Paragraph_fragments)
	l =_ElemprootView_Paragraph_fragments (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_Paragraph_fragments)
	r = _ElemprootView_Paragraph_fragments (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_Paragraph_fragments)r)->_desc2);
((_TPPLST_0prootView_Paragraph_fragments)r)->_desc2 = ((_TPPLST_0prootView_Paragraph_fragments)l)->_desc2;
((_TPPLST_0prootView_Paragraph_fragments)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_Paragraph_fragments (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_Paragraph_fragments (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_Paragraph_fragments root;
	_TPPLST_0prootView_Paragraph_fragments list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_Paragraph_fragments)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_Paragraph_fragments)
		t = _ElemprootView_Paragraph_fragments (_coordref, t);
	remain = (_TPPLST_0prootView_Paragraph_fragments)(((_TPPLST_0prootView_Paragraph_fragments) t)->_desc2);
	((_TPPLST_0prootView_Paragraph_fragments) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_Paragraph_fragments;
#else
list = (_TPPLST_0prootView_Paragraph_fragments) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_Paragraph_fragments));
#endif
list->_prod = RULELST_0prootView_Paragraph_fragments;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_Paragraph_fragments)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_TextFragmentprootView_Paragraph_fragments)
	{_TPPLST_rootView_TextFragmentprootView_Paragraph_fragments _currn = _currn;
	 _TERMACT_LST_rootView_TextFragmentprootView_Paragraph_fragments;
	} this code has been moved to _ElemprootView_Paragraph_fragments*/
	/*if (_currn->_prod == RULELST_rootView_ImageFragmentprootView_Paragraph_fragments)
	{_TPPLST_rootView_ImageFragmentprootView_Paragraph_fragments _currn = _currn;
	 _TERMACT_LST_rootView_ImageFragmentprootView_Paragraph_fragments;
	} this code has been moved to _ElemprootView_Paragraph_fragments*/
	/*if (_currn->_prod == RULELST_rootView_OutputElementprootView_Paragraph_fragments)
	{_TPPLST_rootView_OutputElementprootView_Paragraph_fragments _currn = _currn;
	 _TERMACT_LST_rootView_OutputElementprootView_Paragraph_fragments;
	} this code has been moved to _ElemprootView_Paragraph_fragments*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_Paragraph_fragments;
#else
root = (_TPPprootView_Paragraph_fragments) TreeNodeAlloc (sizeof (struct _TPprootView_Paragraph_fragments));
#endif
root->_prod = RULEprootView_Paragraph_fragments;
root->_desc1 = (_TSPLST_rootView_Paragraph_fragments)list;
{ _TPPprootView_Paragraph_fragments _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_Paragraph_fragments;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_TextFragment (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkprootView_TextFragment (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPprootView_TextFragment _currn;
#ifdef __cplusplus
_currn = new _TPprootView_TextFragment;
#else
_currn = (_TPPprootView_TextFragment) TreeNodeAlloc (sizeof (struct _TPprootView_TextFragment));
#endif
_currn->_prod = RULEprootView_TextFragment;
_currn->_desc1 = (_TSProotView_TextFragment_link) MkrootView_TextFragment_link (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_TextFragment: root of subtree no. 1 can not be made a rootView_TextFragment_link node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_TextFragment_value) MkrootView_TextFragment_value (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_TextFragment: root of subtree no. 2 can not be made a rootView_TextFragment_value node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_TextFragment;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_TextFragment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_TextFragment_link (POSITION *_coordref)
#else
NODEPTR MkprootView_TextFragment_link (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_TextFragment_link _currn;
#ifdef __cplusplus
_currn = new _TPprootView_TextFragment_link;
#else
_currn = (_TPPprootView_TextFragment_link) TreeNodeAlloc (sizeof (struct _TPprootView_TextFragment_link));
#endif
_currn->_prod = RULEprootView_TextFragment_link;
_SETCOORD(_currn)
_TERMACT_prootView_TextFragment_link;
return ( (NODEPTR) _currn);
}/* MkprootView_TextFragment_link */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_TextFragment_value (POSITION *_coordref)
#else
NODEPTR MkprootView_TextFragment_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_TextFragment_value _currn;
#ifdef __cplusplus
_currn = new _TPprootView_TextFragment_value;
#else
_currn = (_TPPprootView_TextFragment_value) TreeNodeAlloc (sizeof (struct _TPprootView_TextFragment_value));
#endif
_currn->_prod = RULEprootView_TextFragment_value;
_SETCOORD(_currn)
_TERMACT_prootView_TextFragment_value;
return ( (NODEPTR) _currn);
}/* MkprootView_TextFragment_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_ImageFragment (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkprootView_ImageFragment (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPprootView_ImageFragment _currn;
#ifdef __cplusplus
_currn = new _TPprootView_ImageFragment;
#else
_currn = (_TPPprootView_ImageFragment) TreeNodeAlloc (sizeof (struct _TPprootView_ImageFragment));
#endif
_currn->_prod = RULEprootView_ImageFragment;
_currn->_desc1 = (_TSProotView_ImageFragment_link) MkrootView_ImageFragment_link (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_ImageFragment: root of subtree no. 1 can not be made a rootView_ImageFragment_link node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_ImageFragment_width) MkrootView_ImageFragment_width (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_ImageFragment: root of subtree no. 2 can not be made a rootView_ImageFragment_width node ", 0, _coordref);
_currn->_desc3 = (_TSProotView_ImageFragment_height) MkrootView_ImageFragment_height (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_ImageFragment: root of subtree no. 3 can not be made a rootView_ImageFragment_height node ", 0, _coordref);
_currn->_desc4 = (_TSProotView_ImageFragment_text) MkrootView_ImageFragment_text (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_ImageFragment: root of subtree no. 4 can not be made a rootView_ImageFragment_text node ", 0, _coordref);
_currn->_desc5 = (_TSProotView_ImageFragment_imageName) MkrootView_ImageFragment_imageName (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_ImageFragment: root of subtree no. 5 can not be made a rootView_ImageFragment_imageName node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_ImageFragment;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_ImageFragment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_ImageFragment_link (POSITION *_coordref)
#else
NODEPTR MkprootView_ImageFragment_link (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_ImageFragment_link _currn;
#ifdef __cplusplus
_currn = new _TPprootView_ImageFragment_link;
#else
_currn = (_TPPprootView_ImageFragment_link) TreeNodeAlloc (sizeof (struct _TPprootView_ImageFragment_link));
#endif
_currn->_prod = RULEprootView_ImageFragment_link;
_SETCOORD(_currn)
_TERMACT_prootView_ImageFragment_link;
return ( (NODEPTR) _currn);
}/* MkprootView_ImageFragment_link */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_ImageFragment_width (POSITION *_coordref)
#else
NODEPTR MkprootView_ImageFragment_width (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_ImageFragment_width _currn;
#ifdef __cplusplus
_currn = new _TPprootView_ImageFragment_width;
#else
_currn = (_TPPprootView_ImageFragment_width) TreeNodeAlloc (sizeof (struct _TPprootView_ImageFragment_width));
#endif
_currn->_prod = RULEprootView_ImageFragment_width;
_SETCOORD(_currn)
_TERMACT_prootView_ImageFragment_width;
return ( (NODEPTR) _currn);
}/* MkprootView_ImageFragment_width */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_ImageFragment_height (POSITION *_coordref)
#else
NODEPTR MkprootView_ImageFragment_height (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_ImageFragment_height _currn;
#ifdef __cplusplus
_currn = new _TPprootView_ImageFragment_height;
#else
_currn = (_TPPprootView_ImageFragment_height) TreeNodeAlloc (sizeof (struct _TPprootView_ImageFragment_height));
#endif
_currn->_prod = RULEprootView_ImageFragment_height;
_SETCOORD(_currn)
_TERMACT_prootView_ImageFragment_height;
return ( (NODEPTR) _currn);
}/* MkprootView_ImageFragment_height */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_ImageFragment_text (POSITION *_coordref)
#else
NODEPTR MkprootView_ImageFragment_text (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_ImageFragment_text _currn;
#ifdef __cplusplus
_currn = new _TPprootView_ImageFragment_text;
#else
_currn = (_TPPprootView_ImageFragment_text) TreeNodeAlloc (sizeof (struct _TPprootView_ImageFragment_text));
#endif
_currn->_prod = RULEprootView_ImageFragment_text;
_SETCOORD(_currn)
_TERMACT_prootView_ImageFragment_text;
return ( (NODEPTR) _currn);
}/* MkprootView_ImageFragment_text */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_ImageFragment_imageName (POSITION *_coordref)
#else
NODEPTR MkprootView_ImageFragment_imageName (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_ImageFragment_imageName _currn;
#ifdef __cplusplus
_currn = new _TPprootView_ImageFragment_imageName;
#else
_currn = (_TPPprootView_ImageFragment_imageName) TreeNodeAlloc (sizeof (struct _TPprootView_ImageFragment_imageName));
#endif
_currn->_prod = RULEprootView_ImageFragment_imageName;
_SETCOORD(_currn)
_TERMACT_prootView_ImageFragment_imageName;
return ( (NODEPTR) _currn);
}/* MkprootView_ImageFragment_imageName */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_Heading (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkprootView_Heading (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPprootView_Heading _currn;
#ifdef __cplusplus
_currn = new _TPprootView_Heading;
#else
_currn = (_TPPprootView_Heading) TreeNodeAlloc (sizeof (struct _TPprootView_Heading));
#endif
_currn->_prod = RULEprootView_Heading;
_currn->_desc1 = (_TSProotView_Heading_styleSheetBlock) MkrootView_Heading_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_Heading: root of subtree no. 1 can not be made a rootView_Heading_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_Heading_text) MkrootView_Heading_text (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_Heading: root of subtree no. 2 can not be made a rootView_Heading_text node ", 0, _coordref);
_currn->_desc3 = (_TSProotView_Heading_level) MkrootView_Heading_level (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_Heading: root of subtree no. 3 can not be made a rootView_Heading_level node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_Heading;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_Heading */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_Heading_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkprootView_Heading_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_Heading_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPprootView_Heading_styleSheetBlock;
#else
_currn = (_TPPprootView_Heading_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPprootView_Heading_styleSheetBlock));
#endif
_currn->_prod = RULEprootView_Heading_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_prootView_Heading_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkprootView_Heading_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_Heading_text (POSITION *_coordref)
#else
NODEPTR MkprootView_Heading_text (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_Heading_text _currn;
#ifdef __cplusplus
_currn = new _TPprootView_Heading_text;
#else
_currn = (_TPPprootView_Heading_text) TreeNodeAlloc (sizeof (struct _TPprootView_Heading_text));
#endif
_currn->_prod = RULEprootView_Heading_text;
_SETCOORD(_currn)
_TERMACT_prootView_Heading_text;
return ( (NODEPTR) _currn);
}/* MkprootView_Heading_text */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_Heading_level (POSITION *_coordref)
#else
NODEPTR MkprootView_Heading_level (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_Heading_level _currn;
#ifdef __cplusplus
_currn = new _TPprootView_Heading_level;
#else
_currn = (_TPPprootView_Heading_level) TreeNodeAlloc (sizeof (struct _TPprootView_Heading_level));
#endif
_currn->_prod = RULEprootView_Heading_level;
_SETCOORD(_currn)
_TERMACT_prootView_Heading_level;
return ( (NODEPTR) _currn);
}/* MkprootView_Heading_level */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_Table (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkprootView_Table (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPprootView_Table _currn;
#ifdef __cplusplus
_currn = new _TPprootView_Table;
#else
_currn = (_TPPprootView_Table) TreeNodeAlloc (sizeof (struct _TPprootView_Table));
#endif
_currn->_prod = RULEprootView_Table;
_currn->_desc1 = (_TSProotView_Table_styleSheetBlock) MkrootView_Table_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_Table: root of subtree no. 1 can not be made a rootView_Table_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_Table_rows) MkrootView_Table_rows (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_Table: root of subtree no. 2 can not be made a rootView_Table_rows node ", 0, _coordref);
_currn->_desc3 = (_TSProotView_Table_columns) MkrootView_Table_columns (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_Table: root of subtree no. 3 can not be made a rootView_Table_columns node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_Table;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_Table */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_Table_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkprootView_Table_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_Table_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPprootView_Table_styleSheetBlock;
#else
_currn = (_TPPprootView_Table_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPprootView_Table_styleSheetBlock));
#endif
_currn->_prod = RULEprootView_Table_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_prootView_Table_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkprootView_Table_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_TableRowprootView_Table_rows (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_TableRowprootView_Table_rows (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_TableRowprootView_Table_rows _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_TableRowprootView_Table_rows;
#else
	_new = (_TPPLST_rootView_TableRowprootView_Table_rows) TreeNodeAlloc (sizeof (struct _TPLST_rootView_TableRowprootView_Table_rows));
#endif
	_new->_prod = RULELST_rootView_TableRowprootView_Table_rows;
	_new->_desc1 = (_TSProotView_TableRow)elem;
	_new->_desc2 = (_TSPLST_rootView_Table_rows)_new;
	{_TPPLST_rootView_TableRowprootView_Table_rows _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_TableRowprootView_Table_rows;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_Table_rows (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_Table_rows (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_TableRow)) 
		return _MkrootView_TableRowprootView_Table_rows (_coordref, MkrootView_TableRow (_coordref, t));
	message (DEADLY, "a prootView_Table_rows list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_Table_rows (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_Table_rows (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_Table_rows)
	l =_ElemprootView_Table_rows (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_Table_rows)
	r = _ElemprootView_Table_rows (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_Table_rows)r)->_desc2);
((_TPPLST_0prootView_Table_rows)r)->_desc2 = ((_TPPLST_0prootView_Table_rows)l)->_desc2;
((_TPPLST_0prootView_Table_rows)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_Table_rows (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_Table_rows (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_Table_rows root;
	_TPPLST_0prootView_Table_rows list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_Table_rows)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_Table_rows)
		t = _ElemprootView_Table_rows (_coordref, t);
	remain = (_TPPLST_0prootView_Table_rows)(((_TPPLST_0prootView_Table_rows) t)->_desc2);
	((_TPPLST_0prootView_Table_rows) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_Table_rows;
#else
list = (_TPPLST_0prootView_Table_rows) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_Table_rows));
#endif
list->_prod = RULELST_0prootView_Table_rows;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_Table_rows)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_TableRowprootView_Table_rows)
	{_TPPLST_rootView_TableRowprootView_Table_rows _currn = _currn;
	 _TERMACT_LST_rootView_TableRowprootView_Table_rows;
	} this code has been moved to _ElemprootView_Table_rows*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_Table_rows;
#else
root = (_TPPprootView_Table_rows) TreeNodeAlloc (sizeof (struct _TPprootView_Table_rows));
#endif
root->_prod = RULEprootView_Table_rows;
root->_desc1 = (_TSPLST_rootView_Table_rows)list;
{ _TPPprootView_Table_rows _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_Table_rows;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_TableColumnprootView_Table_columns (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_TableColumnprootView_Table_columns (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_TableColumnprootView_Table_columns _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_TableColumnprootView_Table_columns;
#else
	_new = (_TPPLST_rootView_TableColumnprootView_Table_columns) TreeNodeAlloc (sizeof (struct _TPLST_rootView_TableColumnprootView_Table_columns));
#endif
	_new->_prod = RULELST_rootView_TableColumnprootView_Table_columns;
	_new->_desc1 = (_TSProotView_TableColumn)elem;
	_new->_desc2 = (_TSPLST_rootView_Table_columns)_new;
	{_TPPLST_rootView_TableColumnprootView_Table_columns _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_TableColumnprootView_Table_columns;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_Table_columns (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_Table_columns (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_TableColumn)) 
		return _MkrootView_TableColumnprootView_Table_columns (_coordref, MkrootView_TableColumn (_coordref, t));
	message (DEADLY, "a prootView_Table_columns list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_Table_columns (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_Table_columns (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_Table_columns)
	l =_ElemprootView_Table_columns (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_Table_columns)
	r = _ElemprootView_Table_columns (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_Table_columns)r)->_desc2);
((_TPPLST_0prootView_Table_columns)r)->_desc2 = ((_TPPLST_0prootView_Table_columns)l)->_desc2;
((_TPPLST_0prootView_Table_columns)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_Table_columns (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_Table_columns (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_Table_columns root;
	_TPPLST_0prootView_Table_columns list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_Table_columns)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_Table_columns)
		t = _ElemprootView_Table_columns (_coordref, t);
	remain = (_TPPLST_0prootView_Table_columns)(((_TPPLST_0prootView_Table_columns) t)->_desc2);
	((_TPPLST_0prootView_Table_columns) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_Table_columns;
#else
list = (_TPPLST_0prootView_Table_columns) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_Table_columns));
#endif
list->_prod = RULELST_0prootView_Table_columns;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_Table_columns)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_TableColumnprootView_Table_columns)
	{_TPPLST_rootView_TableColumnprootView_Table_columns _currn = _currn;
	 _TERMACT_LST_rootView_TableColumnprootView_Table_columns;
	} this code has been moved to _ElemprootView_Table_columns*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_Table_columns;
#else
root = (_TPPprootView_Table_columns) TreeNodeAlloc (sizeof (struct _TPprootView_Table_columns));
#endif
root->_prod = RULEprootView_Table_columns;
root->_desc1 = (_TSPLST_rootView_Table_columns)list;
{ _TPPprootView_Table_columns _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_Table_columns;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_TableRow (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkprootView_TableRow (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPprootView_TableRow _currn;
#ifdef __cplusplus
_currn = new _TPprootView_TableRow;
#else
_currn = (_TPPprootView_TableRow) TreeNodeAlloc (sizeof (struct _TPprootView_TableRow));
#endif
_currn->_prod = RULEprootView_TableRow;
_currn->_desc1 = (_TSProotView_TableRow_cells) MkrootView_TableRow_cells (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_TableRow: root of subtree no. 1 can not be made a rootView_TableRow_cells node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_TableRow;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_TableRow */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_TableCellprootView_TableRow_cells (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_TableCellprootView_TableRow_cells (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_TableCellprootView_TableRow_cells _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_TableCellprootView_TableRow_cells;
#else
	_new = (_TPPLST_rootView_TableCellprootView_TableRow_cells) TreeNodeAlloc (sizeof (struct _TPLST_rootView_TableCellprootView_TableRow_cells));
#endif
	_new->_prod = RULELST_rootView_TableCellprootView_TableRow_cells;
	_new->_desc1 = (_TSProotView_TableCell)elem;
	_new->_desc2 = (_TSPLST_rootView_TableRow_cells)_new;
	{_TPPLST_rootView_TableCellprootView_TableRow_cells _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_TableCellprootView_TableRow_cells;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_TableRow_cells (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_TableRow_cells (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_TableCell)) 
		return _MkrootView_TableCellprootView_TableRow_cells (_coordref, MkrootView_TableCell (_coordref, t));
	message (DEADLY, "a prootView_TableRow_cells list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_TableRow_cells (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_TableRow_cells (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_TableRow_cells)
	l =_ElemprootView_TableRow_cells (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_TableRow_cells)
	r = _ElemprootView_TableRow_cells (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_TableRow_cells)r)->_desc2);
((_TPPLST_0prootView_TableRow_cells)r)->_desc2 = ((_TPPLST_0prootView_TableRow_cells)l)->_desc2;
((_TPPLST_0prootView_TableRow_cells)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_TableRow_cells (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_TableRow_cells (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_TableRow_cells root;
	_TPPLST_0prootView_TableRow_cells list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_TableRow_cells)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_TableRow_cells)
		t = _ElemprootView_TableRow_cells (_coordref, t);
	remain = (_TPPLST_0prootView_TableRow_cells)(((_TPPLST_0prootView_TableRow_cells) t)->_desc2);
	((_TPPLST_0prootView_TableRow_cells) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_TableRow_cells;
#else
list = (_TPPLST_0prootView_TableRow_cells) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_TableRow_cells));
#endif
list->_prod = RULELST_0prootView_TableRow_cells;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_TableRow_cells)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_TableCellprootView_TableRow_cells)
	{_TPPLST_rootView_TableCellprootView_TableRow_cells _currn = _currn;
	 _TERMACT_LST_rootView_TableCellprootView_TableRow_cells;
	} this code has been moved to _ElemprootView_TableRow_cells*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_TableRow_cells;
#else
root = (_TPPprootView_TableRow_cells) TreeNodeAlloc (sizeof (struct _TPprootView_TableRow_cells));
#endif
root->_prod = RULEprootView_TableRow_cells;
root->_desc1 = (_TSPLST_rootView_TableRow_cells)list;
{ _TPPprootView_TableRow_cells _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_TableRow_cells;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_TableColumn (POSITION *_coordref, int _TERM1)
#else
NODEPTR MkprootView_TableColumn (_coordref, _TERM1)
	POSITION *_coordref;
	int _TERM1;
#endif
{	_TPPprootView_TableColumn _currn;
#ifdef __cplusplus
_currn = new _TPprootView_TableColumn;
#else
_currn = (_TPPprootView_TableColumn) TreeNodeAlloc (sizeof (struct _TPprootView_TableColumn));
#endif
_currn->_prod = RULEprootView_TableColumn;
_SETCOORD(_currn)
_TERMACT_prootView_TableColumn;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_TableColumn */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_TableCell (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkprootView_TableCell (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPprootView_TableCell _currn;
#ifdef __cplusplus
_currn = new _TPprootView_TableCell;
#else
_currn = (_TPPprootView_TableCell) TreeNodeAlloc (sizeof (struct _TPprootView_TableCell));
#endif
_currn->_prod = RULEprootView_TableCell;
_currn->_desc1 = (_TSProotView_TableCell_columnRef) MkrootView_TableCell_columnRef (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_TableCell: root of subtree no. 1 can not be made a rootView_TableCell_columnRef node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_TableCell_content) MkrootView_TableCell_content (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_TableCell: root of subtree no. 2 can not be made a rootView_TableCell_content node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_TableCell;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_TableCell */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_TableCell_columnRef (POSITION *_coordref)
#else
NODEPTR MkprootView_TableCell_columnRef (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_TableCell_columnRef _currn;
#ifdef __cplusplus
_currn = new _TPprootView_TableCell_columnRef;
#else
_currn = (_TPPprootView_TableCell_columnRef) TreeNodeAlloc (sizeof (struct _TPprootView_TableCell_columnRef));
#endif
_currn->_prod = RULEprootView_TableCell_columnRef;
_SETCOORD(_currn)
_TERMACT_prootView_TableCell_columnRef;
return ( (NODEPTR) _currn);
}/* MkprootView_TableCell_columnRef */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_LinkTargetExtern (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkprootView_LinkTargetExtern (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPprootView_LinkTargetExtern _currn;
#ifdef __cplusplus
_currn = new _TPprootView_LinkTargetExtern;
#else
_currn = (_TPPprootView_LinkTargetExtern) TreeNodeAlloc (sizeof (struct _TPprootView_LinkTargetExtern));
#endif
_currn->_prod = RULEprootView_LinkTargetExtern;
_currn->_desc1 = (_TSProotView_LinkTargetExtern_name) MkrootView_LinkTargetExtern_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_LinkTargetExtern: root of subtree no. 1 can not be made a rootView_LinkTargetExtern_name node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_LinkTargetExtern_url) MkrootView_LinkTargetExtern_url (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_LinkTargetExtern: root of subtree no. 2 can not be made a rootView_LinkTargetExtern_url node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_LinkTargetExtern;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_LinkTargetExtern */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_LinkTargetExtern_name (POSITION *_coordref)
#else
NODEPTR MkprootView_LinkTargetExtern_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_LinkTargetExtern_name _currn;
#ifdef __cplusplus
_currn = new _TPprootView_LinkTargetExtern_name;
#else
_currn = (_TPPprootView_LinkTargetExtern_name) TreeNodeAlloc (sizeof (struct _TPprootView_LinkTargetExtern_name));
#endif
_currn->_prod = RULEprootView_LinkTargetExtern_name;
_SETCOORD(_currn)
_TERMACT_prootView_LinkTargetExtern_name;
return ( (NODEPTR) _currn);
}/* MkprootView_LinkTargetExtern_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_LinkTargetExtern_url (POSITION *_coordref)
#else
NODEPTR MkprootView_LinkTargetExtern_url (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_LinkTargetExtern_url _currn;
#ifdef __cplusplus
_currn = new _TPprootView_LinkTargetExtern_url;
#else
_currn = (_TPPprootView_LinkTargetExtern_url) TreeNodeAlloc (sizeof (struct _TPprootView_LinkTargetExtern_url));
#endif
_currn->_prod = RULEprootView_LinkTargetExtern_url;
_SETCOORD(_currn)
_TERMACT_prootView_LinkTargetExtern_url;
return ( (NODEPTR) _currn);
}/* MkprootView_LinkTargetExtern_url */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_Root (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkprootView_Root (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPprootView_Root _currn;
#ifdef __cplusplus
_currn = new _TPprootView_Root;
#else
_currn = (_TPPprootView_Root) TreeNodeAlloc (sizeof (struct _TPprootView_Root));
#endif
_currn->_prod = RULEprootView_Root;
_currn->_desc1 = (_TSProotView_Root_webApp) MkrootView_Root_webApp (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_Root: root of subtree no. 1 can not be made a rootView_Root_webApp node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_Root_library) MkrootView_Root_library (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_Root: root of subtree no. 2 can not be made a rootView_Root_library node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_Root;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_Root */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_LibraryprootView_Root_library (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_LibraryprootView_Root_library (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_LibraryprootView_Root_library _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_LibraryprootView_Root_library;
#else
	_new = (_TPPLST_rootView_LibraryprootView_Root_library) TreeNodeAlloc (sizeof (struct _TPLST_rootView_LibraryprootView_Root_library));
#endif
	_new->_prod = RULELST_rootView_LibraryprootView_Root_library;
	_new->_desc1 = (_TSProotView_Library)elem;
	_new->_desc2 = (_TSPLST_rootView_Root_library)_new;
	{_TPPLST_rootView_LibraryprootView_Root_library _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_LibraryprootView_Root_library;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_Root_library (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_Root_library (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_Library)) 
		return _MkrootView_LibraryprootView_Root_library (_coordref, MkrootView_Library (_coordref, t));
	message (DEADLY, "a prootView_Root_library list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_Root_library (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_Root_library (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_Root_library)
	l =_ElemprootView_Root_library (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_Root_library)
	r = _ElemprootView_Root_library (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_Root_library)r)->_desc2);
((_TPPLST_0prootView_Root_library)r)->_desc2 = ((_TPPLST_0prootView_Root_library)l)->_desc2;
((_TPPLST_0prootView_Root_library)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_Root_library (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_Root_library (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_Root_library root;
	_TPPLST_0prootView_Root_library list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_Root_library)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_Root_library)
		t = _ElemprootView_Root_library (_coordref, t);
	remain = (_TPPLST_0prootView_Root_library)(((_TPPLST_0prootView_Root_library) t)->_desc2);
	((_TPPLST_0prootView_Root_library) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_Root_library;
#else
list = (_TPPLST_0prootView_Root_library) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_Root_library));
#endif
list->_prod = RULELST_0prootView_Root_library;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_Root_library)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_LibraryprootView_Root_library)
	{_TPPLST_rootView_LibraryprootView_Root_library _currn = _currn;
	 _TERMACT_LST_rootView_LibraryprootView_Root_library;
	} this code has been moved to _ElemprootView_Root_library*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_Root_library;
#else
root = (_TPPprootView_Root_library) TreeNodeAlloc (sizeof (struct _TPprootView_Root_library));
#endif
root->_prod = RULEprootView_Root_library;
root->_desc1 = (_TSPLST_rootView_Root_library)list;
{ _TPPprootView_Root_library _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_Root_library;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_Library (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkprootView_Library (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPprootView_Library _currn;
#ifdef __cplusplus
_currn = new _TPprootView_Library;
#else
_currn = (_TPPprootView_Library) TreeNodeAlloc (sizeof (struct _TPprootView_Library));
#endif
_currn->_prod = RULEprootView_Library;
_currn->_desc1 = (_TSProotView_Library_name) MkrootView_Library_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_Library: root of subtree no. 1 can not be made a rootView_Library_name node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_Library_formularconditions) MkrootView_Library_formularconditions (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_Library: root of subtree no. 2 can not be made a rootView_Library_formularconditions node ", 0, _coordref);
_currn->_desc3 = (_TSProotView_Library_navigationElement) MkrootView_Library_navigationElement (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_Library: root of subtree no. 3 can not be made a rootView_Library_navigationElement node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_Library;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_Library */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_Library_name (POSITION *_coordref)
#else
NODEPTR MkprootView_Library_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_Library_name _currn;
#ifdef __cplusplus
_currn = new _TPprootView_Library_name;
#else
_currn = (_TPPprootView_Library_name) TreeNodeAlloc (sizeof (struct _TPprootView_Library_name));
#endif
_currn->_prod = RULEprootView_Library_name;
_SETCOORD(_currn)
_TERMACT_prootView_Library_name;
return ( (NODEPTR) _currn);
}/* MkprootView_Library_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_FormularConditionDefinitionprootView_Library_formularconditions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_FormularConditionDefinitionprootView_Library_formularconditions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_FormularConditionDefinitionprootView_Library_formularconditions _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_FormularConditionDefinitionprootView_Library_formularconditions;
#else
	_new = (_TPPLST_rootView_FormularConditionDefinitionprootView_Library_formularconditions) TreeNodeAlloc (sizeof (struct _TPLST_rootView_FormularConditionDefinitionprootView_Library_formularconditions));
#endif
	_new->_prod = RULELST_rootView_FormularConditionDefinitionprootView_Library_formularconditions;
	_new->_desc1 = (_TSProotView_FormularConditionDefinition)elem;
	_new->_desc2 = (_TSPLST_rootView_Library_formularconditions)_new;
	{_TPPLST_rootView_FormularConditionDefinitionprootView_Library_formularconditions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_FormularConditionDefinitionprootView_Library_formularconditions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_Library_formularconditions (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_Library_formularconditions (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_FormularConditionDefinition)) 
		return _MkrootView_FormularConditionDefinitionprootView_Library_formularconditions (_coordref, MkrootView_FormularConditionDefinition (_coordref, t));
	message (DEADLY, "a prootView_Library_formularconditions list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_Library_formularconditions (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_Library_formularconditions (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_Library_formularconditions)
	l =_ElemprootView_Library_formularconditions (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_Library_formularconditions)
	r = _ElemprootView_Library_formularconditions (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_Library_formularconditions)r)->_desc2);
((_TPPLST_0prootView_Library_formularconditions)r)->_desc2 = ((_TPPLST_0prootView_Library_formularconditions)l)->_desc2;
((_TPPLST_0prootView_Library_formularconditions)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_Library_formularconditions (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_Library_formularconditions (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_Library_formularconditions root;
	_TPPLST_0prootView_Library_formularconditions list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_Library_formularconditions)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_Library_formularconditions)
		t = _ElemprootView_Library_formularconditions (_coordref, t);
	remain = (_TPPLST_0prootView_Library_formularconditions)(((_TPPLST_0prootView_Library_formularconditions) t)->_desc2);
	((_TPPLST_0prootView_Library_formularconditions) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_Library_formularconditions;
#else
list = (_TPPLST_0prootView_Library_formularconditions) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_Library_formularconditions));
#endif
list->_prod = RULELST_0prootView_Library_formularconditions;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_Library_formularconditions)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_FormularConditionDefinitionprootView_Library_formularconditions)
	{_TPPLST_rootView_FormularConditionDefinitionprootView_Library_formularconditions _currn = _currn;
	 _TERMACT_LST_rootView_FormularConditionDefinitionprootView_Library_formularconditions;
	} this code has been moved to _ElemprootView_Library_formularconditions*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_Library_formularconditions;
#else
root = (_TPPprootView_Library_formularconditions) TreeNodeAlloc (sizeof (struct _TPprootView_Library_formularconditions));
#endif
root->_prod = RULEprootView_Library_formularconditions;
root->_desc1 = (_TSPLST_rootView_Library_formularconditions)list;
{ _TPPprootView_Library_formularconditions _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_Library_formularconditions;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_NavigationElementprootView_Library_navigationElement (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_NavigationElementprootView_Library_navigationElement (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_NavigationElementprootView_Library_navigationElement _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_NavigationElementprootView_Library_navigationElement;
#else
	_new = (_TPPLST_rootView_NavigationElementprootView_Library_navigationElement) TreeNodeAlloc (sizeof (struct _TPLST_rootView_NavigationElementprootView_Library_navigationElement));
#endif
	_new->_prod = RULELST_rootView_NavigationElementprootView_Library_navigationElement;
	_new->_desc1 = (_TSProotView_NavigationElement)elem;
	_new->_desc2 = (_TSPLST_rootView_Library_navigationElement)_new;
	{_TPPLST_rootView_NavigationElementprootView_Library_navigationElement _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_NavigationElementprootView_Library_navigationElement;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_Library_navigationElement (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_Library_navigationElement (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_NavigationElement)) 
		return _MkrootView_NavigationElementprootView_Library_navigationElement (_coordref, MkrootView_NavigationElement (_coordref, t));
	message (DEADLY, "a prootView_Library_navigationElement list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_Library_navigationElement (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_Library_navigationElement (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_Library_navigationElement)
	l =_ElemprootView_Library_navigationElement (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_Library_navigationElement)
	r = _ElemprootView_Library_navigationElement (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_Library_navigationElement)r)->_desc2);
((_TPPLST_0prootView_Library_navigationElement)r)->_desc2 = ((_TPPLST_0prootView_Library_navigationElement)l)->_desc2;
((_TPPLST_0prootView_Library_navigationElement)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_Library_navigationElement (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_Library_navigationElement (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_Library_navigationElement root;
	_TPPLST_0prootView_Library_navigationElement list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_Library_navigationElement)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_Library_navigationElement)
		t = _ElemprootView_Library_navigationElement (_coordref, t);
	remain = (_TPPLST_0prootView_Library_navigationElement)(((_TPPLST_0prootView_Library_navigationElement) t)->_desc2);
	((_TPPLST_0prootView_Library_navigationElement) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_Library_navigationElement;
#else
list = (_TPPLST_0prootView_Library_navigationElement) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_Library_navigationElement));
#endif
list->_prod = RULELST_0prootView_Library_navigationElement;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_Library_navigationElement)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_NavigationElementprootView_Library_navigationElement)
	{_TPPLST_rootView_NavigationElementprootView_Library_navigationElement _currn = _currn;
	 _TERMACT_LST_rootView_NavigationElementprootView_Library_navigationElement;
	} this code has been moved to _ElemprootView_Library_navigationElement*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_Library_navigationElement;
#else
root = (_TPPprootView_Library_navigationElement) TreeNodeAlloc (sizeof (struct _TPprootView_Library_navigationElement));
#endif
root->_prod = RULEprootView_Library_navigationElement;
root->_desc1 = (_TSPLST_rootView_Library_navigationElement)list;
{ _TPPprootView_Library_navigationElement _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_Library_navigationElement;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_NavigationElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkprootView_NavigationElement (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPprootView_NavigationElement _currn;
#ifdef __cplusplus
_currn = new _TPprootView_NavigationElement;
#else
_currn = (_TPPprootView_NavigationElement) TreeNodeAlloc (sizeof (struct _TPprootView_NavigationElement));
#endif
_currn->_prod = RULEprootView_NavigationElement;
_currn->_desc1 = (_TSProotView_NavigationElement_code) MkrootView_NavigationElement_code (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_NavigationElement: root of subtree no. 1 can not be made a rootView_NavigationElement_code node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_NavigationElement_name) MkrootView_NavigationElement_name (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_NavigationElement: root of subtree no. 2 can not be made a rootView_NavigationElement_name node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_NavigationElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_NavigationElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_NavigationElement_code (POSITION *_coordref)
#else
NODEPTR MkprootView_NavigationElement_code (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_NavigationElement_code _currn;
#ifdef __cplusplus
_currn = new _TPprootView_NavigationElement_code;
#else
_currn = (_TPPprootView_NavigationElement_code) TreeNodeAlloc (sizeof (struct _TPprootView_NavigationElement_code));
#endif
_currn->_prod = RULEprootView_NavigationElement_code;
_SETCOORD(_currn)
_TERMACT_prootView_NavigationElement_code;
return ( (NODEPTR) _currn);
}/* MkprootView_NavigationElement_code */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_NavigationElement_name (POSITION *_coordref)
#else
NODEPTR MkprootView_NavigationElement_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_NavigationElement_name _currn;
#ifdef __cplusplus
_currn = new _TPprootView_NavigationElement_name;
#else
_currn = (_TPPprootView_NavigationElement_name) TreeNodeAlloc (sizeof (struct _TPprootView_NavigationElement_name));
#endif
_currn->_prod = RULEprootView_NavigationElement_name;
_SETCOORD(_currn)
_TERMACT_prootView_NavigationElement_name;
return ( (NODEPTR) _currn);
}/* MkprootView_NavigationElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebApplication (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5, NODEPTR _desc6, NODEPTR _desc7, NODEPTR _desc8, NODEPTR _desc9, NODEPTR _desc10)
#else
NODEPTR MkprootView_WebApplication (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5,_desc6,_desc7,_desc8,_desc9,_desc10)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
	NODEPTR _desc6;
	NODEPTR _desc7;
	NODEPTR _desc8;
	NODEPTR _desc9;
	NODEPTR _desc10;
#endif
{	_TPPprootView_WebApplication _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebApplication;
#else
_currn = (_TPPprootView_WebApplication) TreeNodeAlloc (sizeof (struct _TPprootView_WebApplication));
#endif
_currn->_prod = RULEprootView_WebApplication;
_currn->_desc1 = (_TSProotView_WebApplication_name) MkrootView_WebApplication_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebApplication: root of subtree no. 1 can not be made a rootView_WebApplication_name node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_WebApplication_webZone) MkrootView_WebApplication_webZone (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebApplication: root of subtree no. 2 can not be made a rootView_WebApplication_webZone node ", 0, _coordref);
_currn->_desc3 = (_TSProotView_WebApplication_navigationBar) MkrootView_WebApplication_navigationBar (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebApplication: root of subtree no. 3 can not be made a rootView_WebApplication_navigationBar node ", 0, _coordref);
_currn->_desc4 = (_TSProotView_WebApplication_styleSheet) MkrootView_WebApplication_styleSheet (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebApplication: root of subtree no. 4 can not be made a rootView_WebApplication_styleSheet node ", 0, _coordref);
_currn->_desc5 = (_TSProotView_WebApplication_frames) MkrootView_WebApplication_frames (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebApplication: root of subtree no. 5 can not be made a rootView_WebApplication_frames node ", 0, _coordref);
_currn->_desc6 = (_TSProotView_WebApplication_meta) MkrootView_WebApplication_meta (_coordref, _desc6);	
if (((NODEPTR)_currn->_desc6) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebApplication: root of subtree no. 6 can not be made a rootView_WebApplication_meta node ", 0, _coordref);
_currn->_desc7 = (_TSProotView_WebApplication_database) MkrootView_WebApplication_database (_coordref, _desc7);	
if (((NODEPTR)_currn->_desc7) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebApplication: root of subtree no. 7 can not be made a rootView_WebApplication_database node ", 0, _coordref);
_currn->_desc8 = (_TSProotView_WebApplication_roles) MkrootView_WebApplication_roles (_coordref, _desc8);	
if (((NODEPTR)_currn->_desc8) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebApplication: root of subtree no. 8 can not be made a rootView_WebApplication_roles node ", 0, _coordref);
_currn->_desc9 = (_TSProotView_WebApplication_linksExtern) MkrootView_WebApplication_linksExtern (_coordref, _desc9);	
if (((NODEPTR)_currn->_desc9) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebApplication: root of subtree no. 9 can not be made a rootView_WebApplication_linksExtern node ", 0, _coordref);
_currn->_desc10 = (_TSProotView_WebApplication_imagePath) MkrootView_WebApplication_imagePath (_coordref, _desc10);	
if (((NODEPTR)_currn->_desc10) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebApplication: root of subtree no. 10 can not be made a rootView_WebApplication_imagePath node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_WebApplication;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_WebApplication */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebApplication_name (POSITION *_coordref)
#else
NODEPTR MkprootView_WebApplication_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_WebApplication_name _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebApplication_name;
#else
_currn = (_TPPprootView_WebApplication_name) TreeNodeAlloc (sizeof (struct _TPprootView_WebApplication_name));
#endif
_currn->_prod = RULEprootView_WebApplication_name;
_SETCOORD(_currn)
_TERMACT_prootView_WebApplication_name;
return ( (NODEPTR) _currn);
}/* MkprootView_WebApplication_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_GlobalFrameprootView_WebApplication_frames (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_GlobalFrameprootView_WebApplication_frames (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_GlobalFrameprootView_WebApplication_frames _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_GlobalFrameprootView_WebApplication_frames;
#else
	_new = (_TPPLST_rootView_GlobalFrameprootView_WebApplication_frames) TreeNodeAlloc (sizeof (struct _TPLST_rootView_GlobalFrameprootView_WebApplication_frames));
#endif
	_new->_prod = RULELST_rootView_GlobalFrameprootView_WebApplication_frames;
	_new->_desc1 = (_TSProotView_GlobalFrame)elem;
	_new->_desc2 = (_TSPLST_rootView_WebApplication_frames)_new;
	{_TPPLST_rootView_GlobalFrameprootView_WebApplication_frames _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_GlobalFrameprootView_WebApplication_frames;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_WebApplication_frames (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_WebApplication_frames (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_GlobalFrame)) 
		return _MkrootView_GlobalFrameprootView_WebApplication_frames (_coordref, MkrootView_GlobalFrame (_coordref, t));
	message (DEADLY, "a prootView_WebApplication_frames list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_WebApplication_frames (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_WebApplication_frames (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_WebApplication_frames)
	l =_ElemprootView_WebApplication_frames (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_WebApplication_frames)
	r = _ElemprootView_WebApplication_frames (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_WebApplication_frames)r)->_desc2);
((_TPPLST_0prootView_WebApplication_frames)r)->_desc2 = ((_TPPLST_0prootView_WebApplication_frames)l)->_desc2;
((_TPPLST_0prootView_WebApplication_frames)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebApplication_frames (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_WebApplication_frames (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_WebApplication_frames root;
	_TPPLST_0prootView_WebApplication_frames list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_WebApplication_frames)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_WebApplication_frames)
		t = _ElemprootView_WebApplication_frames (_coordref, t);
	remain = (_TPPLST_0prootView_WebApplication_frames)(((_TPPLST_0prootView_WebApplication_frames) t)->_desc2);
	((_TPPLST_0prootView_WebApplication_frames) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_WebApplication_frames;
#else
list = (_TPPLST_0prootView_WebApplication_frames) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_WebApplication_frames));
#endif
list->_prod = RULELST_0prootView_WebApplication_frames;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_WebApplication_frames)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_GlobalFrameprootView_WebApplication_frames)
	{_TPPLST_rootView_GlobalFrameprootView_WebApplication_frames _currn = _currn;
	 _TERMACT_LST_rootView_GlobalFrameprootView_WebApplication_frames;
	} this code has been moved to _ElemprootView_WebApplication_frames*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_WebApplication_frames;
#else
root = (_TPPprootView_WebApplication_frames) TreeNodeAlloc (sizeof (struct _TPprootView_WebApplication_frames));
#endif
root->_prod = RULEprootView_WebApplication_frames;
root->_desc1 = (_TSPLST_rootView_WebApplication_frames)list;
{ _TPPprootView_WebApplication_frames _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_WebApplication_frames;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_WebRoleprootView_WebApplication_roles (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_WebRoleprootView_WebApplication_roles (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_WebRoleprootView_WebApplication_roles _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_WebRoleprootView_WebApplication_roles;
#else
	_new = (_TPPLST_rootView_WebRoleprootView_WebApplication_roles) TreeNodeAlloc (sizeof (struct _TPLST_rootView_WebRoleprootView_WebApplication_roles));
#endif
	_new->_prod = RULELST_rootView_WebRoleprootView_WebApplication_roles;
	_new->_desc1 = (_TSProotView_WebRole)elem;
	_new->_desc2 = (_TSPLST_rootView_WebApplication_roles)_new;
	{_TPPLST_rootView_WebRoleprootView_WebApplication_roles _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_WebRoleprootView_WebApplication_roles;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_WebApplication_roles (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_WebApplication_roles (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_WebRole)) 
		return _MkrootView_WebRoleprootView_WebApplication_roles (_coordref, MkrootView_WebRole (_coordref, t));
	message (DEADLY, "a prootView_WebApplication_roles list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_WebApplication_roles (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_WebApplication_roles (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_WebApplication_roles)
	l =_ElemprootView_WebApplication_roles (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_WebApplication_roles)
	r = _ElemprootView_WebApplication_roles (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_WebApplication_roles)r)->_desc2);
((_TPPLST_0prootView_WebApplication_roles)r)->_desc2 = ((_TPPLST_0prootView_WebApplication_roles)l)->_desc2;
((_TPPLST_0prootView_WebApplication_roles)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebApplication_roles (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_WebApplication_roles (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_WebApplication_roles root;
	_TPPLST_0prootView_WebApplication_roles list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_WebApplication_roles)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_WebApplication_roles)
		t = _ElemprootView_WebApplication_roles (_coordref, t);
	remain = (_TPPLST_0prootView_WebApplication_roles)(((_TPPLST_0prootView_WebApplication_roles) t)->_desc2);
	((_TPPLST_0prootView_WebApplication_roles) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_WebApplication_roles;
#else
list = (_TPPLST_0prootView_WebApplication_roles) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_WebApplication_roles));
#endif
list->_prod = RULELST_0prootView_WebApplication_roles;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_WebApplication_roles)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_WebRoleprootView_WebApplication_roles)
	{_TPPLST_rootView_WebRoleprootView_WebApplication_roles _currn = _currn;
	 _TERMACT_LST_rootView_WebRoleprootView_WebApplication_roles;
	} this code has been moved to _ElemprootView_WebApplication_roles*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_WebApplication_roles;
#else
root = (_TPPprootView_WebApplication_roles) TreeNodeAlloc (sizeof (struct _TPprootView_WebApplication_roles));
#endif
root->_prod = RULEprootView_WebApplication_roles;
root->_desc1 = (_TSPLST_rootView_WebApplication_roles)list;
{ _TPPprootView_WebApplication_roles _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_WebApplication_roles;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_LinkTargetExternprootView_WebApplication_linksExtern (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_LinkTargetExternprootView_WebApplication_linksExtern (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_LinkTargetExternprootView_WebApplication_linksExtern _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_LinkTargetExternprootView_WebApplication_linksExtern;
#else
	_new = (_TPPLST_rootView_LinkTargetExternprootView_WebApplication_linksExtern) TreeNodeAlloc (sizeof (struct _TPLST_rootView_LinkTargetExternprootView_WebApplication_linksExtern));
#endif
	_new->_prod = RULELST_rootView_LinkTargetExternprootView_WebApplication_linksExtern;
	_new->_desc1 = (_TSProotView_LinkTargetExtern)elem;
	_new->_desc2 = (_TSPLST_rootView_WebApplication_linksExtern)_new;
	{_TPPLST_rootView_LinkTargetExternprootView_WebApplication_linksExtern _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_LinkTargetExternprootView_WebApplication_linksExtern;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_WebApplication_linksExtern (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_WebApplication_linksExtern (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_LinkTargetExtern)) 
		return _MkrootView_LinkTargetExternprootView_WebApplication_linksExtern (_coordref, MkrootView_LinkTargetExtern (_coordref, t));
	message (DEADLY, "a prootView_WebApplication_linksExtern list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_WebApplication_linksExtern (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_WebApplication_linksExtern (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_WebApplication_linksExtern)
	l =_ElemprootView_WebApplication_linksExtern (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_WebApplication_linksExtern)
	r = _ElemprootView_WebApplication_linksExtern (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_WebApplication_linksExtern)r)->_desc2);
((_TPPLST_0prootView_WebApplication_linksExtern)r)->_desc2 = ((_TPPLST_0prootView_WebApplication_linksExtern)l)->_desc2;
((_TPPLST_0prootView_WebApplication_linksExtern)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebApplication_linksExtern (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_WebApplication_linksExtern (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_WebApplication_linksExtern root;
	_TPPLST_0prootView_WebApplication_linksExtern list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_WebApplication_linksExtern)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_WebApplication_linksExtern)
		t = _ElemprootView_WebApplication_linksExtern (_coordref, t);
	remain = (_TPPLST_0prootView_WebApplication_linksExtern)(((_TPPLST_0prootView_WebApplication_linksExtern) t)->_desc2);
	((_TPPLST_0prootView_WebApplication_linksExtern) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_WebApplication_linksExtern;
#else
list = (_TPPLST_0prootView_WebApplication_linksExtern) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_WebApplication_linksExtern));
#endif
list->_prod = RULELST_0prootView_WebApplication_linksExtern;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_WebApplication_linksExtern)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_LinkTargetExternprootView_WebApplication_linksExtern)
	{_TPPLST_rootView_LinkTargetExternprootView_WebApplication_linksExtern _currn = _currn;
	 _TERMACT_LST_rootView_LinkTargetExternprootView_WebApplication_linksExtern;
	} this code has been moved to _ElemprootView_WebApplication_linksExtern*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_WebApplication_linksExtern;
#else
root = (_TPPprootView_WebApplication_linksExtern) TreeNodeAlloc (sizeof (struct _TPprootView_WebApplication_linksExtern));
#endif
root->_prod = RULEprootView_WebApplication_linksExtern;
root->_desc1 = (_TSPLST_rootView_WebApplication_linksExtern)list;
{ _TPPprootView_WebApplication_linksExtern _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_WebApplication_linksExtern;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebApplication_imagePath (POSITION *_coordref)
#else
NODEPTR MkprootView_WebApplication_imagePath (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_WebApplication_imagePath _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebApplication_imagePath;
#else
_currn = (_TPPprootView_WebApplication_imagePath) TreeNodeAlloc (sizeof (struct _TPprootView_WebApplication_imagePath));
#endif
_currn->_prod = RULEprootView_WebApplication_imagePath;
_SETCOORD(_currn)
_TERMACT_prootView_WebApplication_imagePath;
return ( (NODEPTR) _currn);
}/* MkprootView_WebApplication_imagePath */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_RootWebZone (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4)
#else
NODEPTR MkprootView_RootWebZone (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
#endif
{	_TPPprootView_RootWebZone _currn;
#ifdef __cplusplus
_currn = new _TPprootView_RootWebZone;
#else
_currn = (_TPPprootView_RootWebZone) TreeNodeAlloc (sizeof (struct _TPprootView_RootWebZone));
#endif
_currn->_prod = RULEprootView_RootWebZone;
_currn->_desc1 = (_TSProotView_RootWebZone_webTargets) MkrootView_RootWebZone_webTargets (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_RootWebZone: root of subtree no. 1 can not be made a rootView_RootWebZone_webTargets node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_RootWebZone_errorLink) MkrootView_RootWebZone_errorLink (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_RootWebZone: root of subtree no. 2 can not be made a rootView_RootWebZone_errorLink node ", 0, _coordref);
_currn->_desc3 = (_TSProotView_RootWebZone_roles) MkrootView_RootWebZone_roles (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_RootWebZone: root of subtree no. 3 can not be made a rootView_RootWebZone_roles node ", 0, _coordref);
_currn->_desc4 = (_TSProotView_RootWebZone_setSize) MkrootView_RootWebZone_setSize (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_RootWebZone: root of subtree no. 4 can not be made a rootView_RootWebZone_setSize node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_RootWebZone;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_RootWebZone */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_SubWebZoneprootView_RootWebZone_webTargets (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_SubWebZoneprootView_RootWebZone_webTargets (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_SubWebZoneprootView_RootWebZone_webTargets _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_SubWebZoneprootView_RootWebZone_webTargets;
#else
	_new = (_TPPLST_rootView_SubWebZoneprootView_RootWebZone_webTargets) TreeNodeAlloc (sizeof (struct _TPLST_rootView_SubWebZoneprootView_RootWebZone_webTargets));
#endif
	_new->_prod = RULELST_rootView_SubWebZoneprootView_RootWebZone_webTargets;
	_new->_desc1 = (_TSProotView_SubWebZone)elem;
	_new->_desc2 = (_TSPLST_rootView_RootWebZone_webTargets)_new;
	{_TPPLST_rootView_SubWebZoneprootView_RootWebZone_webTargets _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_SubWebZoneprootView_RootWebZone_webTargets;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_WebPageprootView_RootWebZone_webTargets (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_WebPageprootView_RootWebZone_webTargets (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_WebPageprootView_RootWebZone_webTargets _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_WebPageprootView_RootWebZone_webTargets;
#else
	_new = (_TPPLST_rootView_WebPageprootView_RootWebZone_webTargets) TreeNodeAlloc (sizeof (struct _TPLST_rootView_WebPageprootView_RootWebZone_webTargets));
#endif
	_new->_prod = RULELST_rootView_WebPageprootView_RootWebZone_webTargets;
	_new->_desc1 = (_TSProotView_WebPage)elem;
	_new->_desc2 = (_TSPLST_rootView_RootWebZone_webTargets)_new;
	{_TPPLST_rootView_WebPageprootView_RootWebZone_webTargets _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_WebPageprootView_RootWebZone_webTargets;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_WebActionprootView_RootWebZone_webTargets (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_WebActionprootView_RootWebZone_webTargets (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_WebActionprootView_RootWebZone_webTargets _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_WebActionprootView_RootWebZone_webTargets;
#else
	_new = (_TPPLST_rootView_WebActionprootView_RootWebZone_webTargets) TreeNodeAlloc (sizeof (struct _TPLST_rootView_WebActionprootView_RootWebZone_webTargets));
#endif
	_new->_prod = RULELST_rootView_WebActionprootView_RootWebZone_webTargets;
	_new->_desc1 = (_TSProotView_WebAction)elem;
	_new->_desc2 = (_TSPLST_rootView_RootWebZone_webTargets)_new;
	{_TPPLST_rootView_WebActionprootView_RootWebZone_webTargets _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_WebActionprootView_RootWebZone_webTargets;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_RootWebZone_webTargets (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_RootWebZone_webTargets (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_SubWebZone)) 
		return _MkrootView_SubWebZoneprootView_RootWebZone_webTargets (_coordref, MkrootView_SubWebZone (_coordref, t));
	if (IsSymb (t, SYMBrootView_WebPage)) 
		return _MkrootView_WebPageprootView_RootWebZone_webTargets (_coordref, MkrootView_WebPage (_coordref, t));
	if (IsSymb (t, SYMBrootView_WebAction)) 
		return _MkrootView_WebActionprootView_RootWebZone_webTargets (_coordref, MkrootView_WebAction (_coordref, t));
	message (DEADLY, "a prootView_RootWebZone_webTargets list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_RootWebZone_webTargets (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_RootWebZone_webTargets (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_RootWebZone_webTargets)
	l =_ElemprootView_RootWebZone_webTargets (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_RootWebZone_webTargets)
	r = _ElemprootView_RootWebZone_webTargets (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_RootWebZone_webTargets)r)->_desc2);
((_TPPLST_0prootView_RootWebZone_webTargets)r)->_desc2 = ((_TPPLST_0prootView_RootWebZone_webTargets)l)->_desc2;
((_TPPLST_0prootView_RootWebZone_webTargets)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_RootWebZone_webTargets (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_RootWebZone_webTargets (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_RootWebZone_webTargets root;
	_TPPLST_0prootView_RootWebZone_webTargets list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_RootWebZone_webTargets)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_RootWebZone_webTargets)
		t = _ElemprootView_RootWebZone_webTargets (_coordref, t);
	remain = (_TPPLST_0prootView_RootWebZone_webTargets)(((_TPPLST_0prootView_RootWebZone_webTargets) t)->_desc2);
	((_TPPLST_0prootView_RootWebZone_webTargets) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_RootWebZone_webTargets;
#else
list = (_TPPLST_0prootView_RootWebZone_webTargets) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_RootWebZone_webTargets));
#endif
list->_prod = RULELST_0prootView_RootWebZone_webTargets;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_RootWebZone_webTargets)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_SubWebZoneprootView_RootWebZone_webTargets)
	{_TPPLST_rootView_SubWebZoneprootView_RootWebZone_webTargets _currn = _currn;
	 _TERMACT_LST_rootView_SubWebZoneprootView_RootWebZone_webTargets;
	} this code has been moved to _ElemprootView_RootWebZone_webTargets*/
	/*if (_currn->_prod == RULELST_rootView_WebPageprootView_RootWebZone_webTargets)
	{_TPPLST_rootView_WebPageprootView_RootWebZone_webTargets _currn = _currn;
	 _TERMACT_LST_rootView_WebPageprootView_RootWebZone_webTargets;
	} this code has been moved to _ElemprootView_RootWebZone_webTargets*/
	/*if (_currn->_prod == RULELST_rootView_WebActionprootView_RootWebZone_webTargets)
	{_TPPLST_rootView_WebActionprootView_RootWebZone_webTargets _currn = _currn;
	 _TERMACT_LST_rootView_WebActionprootView_RootWebZone_webTargets;
	} this code has been moved to _ElemprootView_RootWebZone_webTargets*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_RootWebZone_webTargets;
#else
root = (_TPPprootView_RootWebZone_webTargets) TreeNodeAlloc (sizeof (struct _TPprootView_RootWebZone_webTargets));
#endif
root->_prod = RULEprootView_RootWebZone_webTargets;
root->_desc1 = (_TSPLST_rootView_RootWebZone_webTargets)list;
{ _TPPprootView_RootWebZone_webTargets _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_RootWebZone_webTargets;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_RootWebZone_errorLink (POSITION *_coordref)
#else
NODEPTR MkprootView_RootWebZone_errorLink (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_RootWebZone_errorLink _currn;
#ifdef __cplusplus
_currn = new _TPprootView_RootWebZone_errorLink;
#else
_currn = (_TPPprootView_RootWebZone_errorLink) TreeNodeAlloc (sizeof (struct _TPprootView_RootWebZone_errorLink));
#endif
_currn->_prod = RULEprootView_RootWebZone_errorLink;
_SETCOORD(_currn)
_TERMACT_prootView_RootWebZone_errorLink;
return ( (NODEPTR) _currn);
}/* MkprootView_RootWebZone_errorLink */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_WebRoleRefprootView_RootWebZone_roles (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_WebRoleRefprootView_RootWebZone_roles (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_WebRoleRefprootView_RootWebZone_roles _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_WebRoleRefprootView_RootWebZone_roles;
#else
	_new = (_TPPLST_rootView_WebRoleRefprootView_RootWebZone_roles) TreeNodeAlloc (sizeof (struct _TPLST_rootView_WebRoleRefprootView_RootWebZone_roles));
#endif
	_new->_prod = RULELST_rootView_WebRoleRefprootView_RootWebZone_roles;
	_new->_desc1 = (_TSProotView_WebRoleRef)elem;
	_new->_desc2 = (_TSPLST_rootView_RootWebZone_roles)_new;
	{_TPPLST_rootView_WebRoleRefprootView_RootWebZone_roles _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_WebRoleRefprootView_RootWebZone_roles;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_RootWebZone_roles (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_RootWebZone_roles (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_WebRoleRef)) 
		return _MkrootView_WebRoleRefprootView_RootWebZone_roles (_coordref, MkrootView_WebRoleRef (_coordref, t));
	message (DEADLY, "a prootView_RootWebZone_roles list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_RootWebZone_roles (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_RootWebZone_roles (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_RootWebZone_roles)
	l =_ElemprootView_RootWebZone_roles (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_RootWebZone_roles)
	r = _ElemprootView_RootWebZone_roles (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_RootWebZone_roles)r)->_desc2);
((_TPPLST_0prootView_RootWebZone_roles)r)->_desc2 = ((_TPPLST_0prootView_RootWebZone_roles)l)->_desc2;
((_TPPLST_0prootView_RootWebZone_roles)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_RootWebZone_roles (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_RootWebZone_roles (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_RootWebZone_roles root;
	_TPPLST_0prootView_RootWebZone_roles list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_RootWebZone_roles)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_RootWebZone_roles)
		t = _ElemprootView_RootWebZone_roles (_coordref, t);
	remain = (_TPPLST_0prootView_RootWebZone_roles)(((_TPPLST_0prootView_RootWebZone_roles) t)->_desc2);
	((_TPPLST_0prootView_RootWebZone_roles) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_RootWebZone_roles;
#else
list = (_TPPLST_0prootView_RootWebZone_roles) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_RootWebZone_roles));
#endif
list->_prod = RULELST_0prootView_RootWebZone_roles;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_RootWebZone_roles)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_WebRoleRefprootView_RootWebZone_roles)
	{_TPPLST_rootView_WebRoleRefprootView_RootWebZone_roles _currn = _currn;
	 _TERMACT_LST_rootView_WebRoleRefprootView_RootWebZone_roles;
	} this code has been moved to _ElemprootView_RootWebZone_roles*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_RootWebZone_roles;
#else
root = (_TPPprootView_RootWebZone_roles) TreeNodeAlloc (sizeof (struct _TPprootView_RootWebZone_roles));
#endif
root->_prod = RULEprootView_RootWebZone_roles;
root->_desc1 = (_TSPLST_rootView_RootWebZone_roles)list;
{ _TPPprootView_RootWebZone_roles _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_RootWebZone_roles;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_RootWebZone_setSize (POSITION *_coordref)
#else
NODEPTR MkprootView_RootWebZone_setSize (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_RootWebZone_setSize _currn;
#ifdef __cplusplus
_currn = new _TPprootView_RootWebZone_setSize;
#else
_currn = (_TPPprootView_RootWebZone_setSize) TreeNodeAlloc (sizeof (struct _TPprootView_RootWebZone_setSize));
#endif
_currn->_prod = RULEprootView_RootWebZone_setSize;
_SETCOORD(_currn)
_TERMACT_prootView_RootWebZone_setSize;
return ( (NODEPTR) _currn);
}/* MkprootView_RootWebZone_setSize */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_SubWebZone (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5, NODEPTR _desc6, NODEPTR _desc7)
#else
NODEPTR MkprootView_SubWebZone (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5,_desc6,_desc7)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
	NODEPTR _desc6;
	NODEPTR _desc7;
#endif
{	_TPPprootView_SubWebZone _currn;
#ifdef __cplusplus
_currn = new _TPprootView_SubWebZone;
#else
_currn = (_TPPprootView_SubWebZone) TreeNodeAlloc (sizeof (struct _TPprootView_SubWebZone));
#endif
_currn->_prod = RULEprootView_SubWebZone;
_currn->_desc1 = (_TSProotView_SubWebZone_name) MkrootView_SubWebZone_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_SubWebZone: root of subtree no. 1 can not be made a rootView_SubWebZone_name node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_SubWebZone_webTargets) MkrootView_SubWebZone_webTargets (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_SubWebZone: root of subtree no. 2 can not be made a rootView_SubWebZone_webTargets node ", 0, _coordref);
_currn->_desc3 = (_TSProotView_SubWebZone_errorLink) MkrootView_SubWebZone_errorLink (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_SubWebZone: root of subtree no. 3 can not be made a rootView_SubWebZone_errorLink node ", 0, _coordref);
_currn->_desc4 = (_TSProotView_SubWebZone_globalparams) MkrootView_SubWebZone_globalparams (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_SubWebZone: root of subtree no. 4 can not be made a rootView_SubWebZone_globalparams node ", 0, _coordref);
_currn->_desc5 = (_TSProotView_SubWebZone_roles) MkrootView_SubWebZone_roles (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_SubWebZone: root of subtree no. 5 can not be made a rootView_SubWebZone_roles node ", 0, _coordref);
_currn->_desc6 = (_TSProotView_SubWebZone_setSize) MkrootView_SubWebZone_setSize (_coordref, _desc6);	
if (((NODEPTR)_currn->_desc6) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_SubWebZone: root of subtree no. 6 can not be made a rootView_SubWebZone_setSize node ", 0, _coordref);
_currn->_desc7 = (_TSProotView_SubWebZone_position) MkrootView_SubWebZone_position (_coordref, _desc7);	
if (((NODEPTR)_currn->_desc7) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_SubWebZone: root of subtree no. 7 can not be made a rootView_SubWebZone_position node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_SubWebZone;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_SubWebZone */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_SubWebZone_name (POSITION *_coordref)
#else
NODEPTR MkprootView_SubWebZone_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_SubWebZone_name _currn;
#ifdef __cplusplus
_currn = new _TPprootView_SubWebZone_name;
#else
_currn = (_TPPprootView_SubWebZone_name) TreeNodeAlloc (sizeof (struct _TPprootView_SubWebZone_name));
#endif
_currn->_prod = RULEprootView_SubWebZone_name;
_SETCOORD(_currn)
_TERMACT_prootView_SubWebZone_name;
return ( (NODEPTR) _currn);
}/* MkprootView_SubWebZone_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_SubWebZoneprootView_SubWebZone_webTargets (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_SubWebZoneprootView_SubWebZone_webTargets (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_SubWebZoneprootView_SubWebZone_webTargets _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_SubWebZoneprootView_SubWebZone_webTargets;
#else
	_new = (_TPPLST_rootView_SubWebZoneprootView_SubWebZone_webTargets) TreeNodeAlloc (sizeof (struct _TPLST_rootView_SubWebZoneprootView_SubWebZone_webTargets));
#endif
	_new->_prod = RULELST_rootView_SubWebZoneprootView_SubWebZone_webTargets;
	_new->_desc1 = (_TSProotView_SubWebZone)elem;
	_new->_desc2 = (_TSPLST_rootView_SubWebZone_webTargets)_new;
	{_TPPLST_rootView_SubWebZoneprootView_SubWebZone_webTargets _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_SubWebZoneprootView_SubWebZone_webTargets;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_WebPageprootView_SubWebZone_webTargets (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_WebPageprootView_SubWebZone_webTargets (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_WebPageprootView_SubWebZone_webTargets _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_WebPageprootView_SubWebZone_webTargets;
#else
	_new = (_TPPLST_rootView_WebPageprootView_SubWebZone_webTargets) TreeNodeAlloc (sizeof (struct _TPLST_rootView_WebPageprootView_SubWebZone_webTargets));
#endif
	_new->_prod = RULELST_rootView_WebPageprootView_SubWebZone_webTargets;
	_new->_desc1 = (_TSProotView_WebPage)elem;
	_new->_desc2 = (_TSPLST_rootView_SubWebZone_webTargets)_new;
	{_TPPLST_rootView_WebPageprootView_SubWebZone_webTargets _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_WebPageprootView_SubWebZone_webTargets;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_WebActionprootView_SubWebZone_webTargets (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_WebActionprootView_SubWebZone_webTargets (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_WebActionprootView_SubWebZone_webTargets _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_WebActionprootView_SubWebZone_webTargets;
#else
	_new = (_TPPLST_rootView_WebActionprootView_SubWebZone_webTargets) TreeNodeAlloc (sizeof (struct _TPLST_rootView_WebActionprootView_SubWebZone_webTargets));
#endif
	_new->_prod = RULELST_rootView_WebActionprootView_SubWebZone_webTargets;
	_new->_desc1 = (_TSProotView_WebAction)elem;
	_new->_desc2 = (_TSPLST_rootView_SubWebZone_webTargets)_new;
	{_TPPLST_rootView_WebActionprootView_SubWebZone_webTargets _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_WebActionprootView_SubWebZone_webTargets;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_SubWebZone_webTargets (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_SubWebZone_webTargets (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_SubWebZone)) 
		return _MkrootView_SubWebZoneprootView_SubWebZone_webTargets (_coordref, MkrootView_SubWebZone (_coordref, t));
	if (IsSymb (t, SYMBrootView_WebPage)) 
		return _MkrootView_WebPageprootView_SubWebZone_webTargets (_coordref, MkrootView_WebPage (_coordref, t));
	if (IsSymb (t, SYMBrootView_WebAction)) 
		return _MkrootView_WebActionprootView_SubWebZone_webTargets (_coordref, MkrootView_WebAction (_coordref, t));
	message (DEADLY, "a prootView_SubWebZone_webTargets list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_SubWebZone_webTargets (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_SubWebZone_webTargets (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_SubWebZone_webTargets)
	l =_ElemprootView_SubWebZone_webTargets (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_SubWebZone_webTargets)
	r = _ElemprootView_SubWebZone_webTargets (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_SubWebZone_webTargets)r)->_desc2);
((_TPPLST_0prootView_SubWebZone_webTargets)r)->_desc2 = ((_TPPLST_0prootView_SubWebZone_webTargets)l)->_desc2;
((_TPPLST_0prootView_SubWebZone_webTargets)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_SubWebZone_webTargets (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_SubWebZone_webTargets (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_SubWebZone_webTargets root;
	_TPPLST_0prootView_SubWebZone_webTargets list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_SubWebZone_webTargets)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_SubWebZone_webTargets)
		t = _ElemprootView_SubWebZone_webTargets (_coordref, t);
	remain = (_TPPLST_0prootView_SubWebZone_webTargets)(((_TPPLST_0prootView_SubWebZone_webTargets) t)->_desc2);
	((_TPPLST_0prootView_SubWebZone_webTargets) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_SubWebZone_webTargets;
#else
list = (_TPPLST_0prootView_SubWebZone_webTargets) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_SubWebZone_webTargets));
#endif
list->_prod = RULELST_0prootView_SubWebZone_webTargets;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_SubWebZone_webTargets)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_SubWebZoneprootView_SubWebZone_webTargets)
	{_TPPLST_rootView_SubWebZoneprootView_SubWebZone_webTargets _currn = _currn;
	 _TERMACT_LST_rootView_SubWebZoneprootView_SubWebZone_webTargets;
	} this code has been moved to _ElemprootView_SubWebZone_webTargets*/
	/*if (_currn->_prod == RULELST_rootView_WebPageprootView_SubWebZone_webTargets)
	{_TPPLST_rootView_WebPageprootView_SubWebZone_webTargets _currn = _currn;
	 _TERMACT_LST_rootView_WebPageprootView_SubWebZone_webTargets;
	} this code has been moved to _ElemprootView_SubWebZone_webTargets*/
	/*if (_currn->_prod == RULELST_rootView_WebActionprootView_SubWebZone_webTargets)
	{_TPPLST_rootView_WebActionprootView_SubWebZone_webTargets _currn = _currn;
	 _TERMACT_LST_rootView_WebActionprootView_SubWebZone_webTargets;
	} this code has been moved to _ElemprootView_SubWebZone_webTargets*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_SubWebZone_webTargets;
#else
root = (_TPPprootView_SubWebZone_webTargets) TreeNodeAlloc (sizeof (struct _TPprootView_SubWebZone_webTargets));
#endif
root->_prod = RULEprootView_SubWebZone_webTargets;
root->_desc1 = (_TSPLST_rootView_SubWebZone_webTargets)list;
{ _TPPprootView_SubWebZone_webTargets _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_SubWebZone_webTargets;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_SubWebZone_errorLink (POSITION *_coordref)
#else
NODEPTR MkprootView_SubWebZone_errorLink (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_SubWebZone_errorLink _currn;
#ifdef __cplusplus
_currn = new _TPprootView_SubWebZone_errorLink;
#else
_currn = (_TPPprootView_SubWebZone_errorLink) TreeNodeAlloc (sizeof (struct _TPprootView_SubWebZone_errorLink));
#endif
_currn->_prod = RULEprootView_SubWebZone_errorLink;
_SETCOORD(_currn)
_TERMACT_prootView_SubWebZone_errorLink;
return ( (NODEPTR) _currn);
}/* MkprootView_SubWebZone_errorLink */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_WebNodeFormalParamprootView_SubWebZone_globalparams (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_WebNodeFormalParamprootView_SubWebZone_globalparams (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_WebNodeFormalParamprootView_SubWebZone_globalparams _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_WebNodeFormalParamprootView_SubWebZone_globalparams;
#else
	_new = (_TPPLST_rootView_WebNodeFormalParamprootView_SubWebZone_globalparams) TreeNodeAlloc (sizeof (struct _TPLST_rootView_WebNodeFormalParamprootView_SubWebZone_globalparams));
#endif
	_new->_prod = RULELST_rootView_WebNodeFormalParamprootView_SubWebZone_globalparams;
	_new->_desc1 = (_TSProotView_WebNodeFormalParam)elem;
	_new->_desc2 = (_TSPLST_rootView_SubWebZone_globalparams)_new;
	{_TPPLST_rootView_WebNodeFormalParamprootView_SubWebZone_globalparams _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_WebNodeFormalParamprootView_SubWebZone_globalparams;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_SubWebZone_globalparams (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_SubWebZone_globalparams (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_WebNodeFormalParam)) 
		return _MkrootView_WebNodeFormalParamprootView_SubWebZone_globalparams (_coordref, MkrootView_WebNodeFormalParam (_coordref, t));
	message (DEADLY, "a prootView_SubWebZone_globalparams list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_SubWebZone_globalparams (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_SubWebZone_globalparams (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_SubWebZone_globalparams)
	l =_ElemprootView_SubWebZone_globalparams (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_SubWebZone_globalparams)
	r = _ElemprootView_SubWebZone_globalparams (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_SubWebZone_globalparams)r)->_desc2);
((_TPPLST_0prootView_SubWebZone_globalparams)r)->_desc2 = ((_TPPLST_0prootView_SubWebZone_globalparams)l)->_desc2;
((_TPPLST_0prootView_SubWebZone_globalparams)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_SubWebZone_globalparams (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_SubWebZone_globalparams (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_SubWebZone_globalparams root;
	_TPPLST_0prootView_SubWebZone_globalparams list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_SubWebZone_globalparams)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_SubWebZone_globalparams)
		t = _ElemprootView_SubWebZone_globalparams (_coordref, t);
	remain = (_TPPLST_0prootView_SubWebZone_globalparams)(((_TPPLST_0prootView_SubWebZone_globalparams) t)->_desc2);
	((_TPPLST_0prootView_SubWebZone_globalparams) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_SubWebZone_globalparams;
#else
list = (_TPPLST_0prootView_SubWebZone_globalparams) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_SubWebZone_globalparams));
#endif
list->_prod = RULELST_0prootView_SubWebZone_globalparams;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_SubWebZone_globalparams)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_WebNodeFormalParamprootView_SubWebZone_globalparams)
	{_TPPLST_rootView_WebNodeFormalParamprootView_SubWebZone_globalparams _currn = _currn;
	 _TERMACT_LST_rootView_WebNodeFormalParamprootView_SubWebZone_globalparams;
	} this code has been moved to _ElemprootView_SubWebZone_globalparams*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_SubWebZone_globalparams;
#else
root = (_TPPprootView_SubWebZone_globalparams) TreeNodeAlloc (sizeof (struct _TPprootView_SubWebZone_globalparams));
#endif
root->_prod = RULEprootView_SubWebZone_globalparams;
root->_desc1 = (_TSPLST_rootView_SubWebZone_globalparams)list;
{ _TPPprootView_SubWebZone_globalparams _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_SubWebZone_globalparams;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_WebRoleRefprootView_SubWebZone_roles (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_WebRoleRefprootView_SubWebZone_roles (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_WebRoleRefprootView_SubWebZone_roles _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_WebRoleRefprootView_SubWebZone_roles;
#else
	_new = (_TPPLST_rootView_WebRoleRefprootView_SubWebZone_roles) TreeNodeAlloc (sizeof (struct _TPLST_rootView_WebRoleRefprootView_SubWebZone_roles));
#endif
	_new->_prod = RULELST_rootView_WebRoleRefprootView_SubWebZone_roles;
	_new->_desc1 = (_TSProotView_WebRoleRef)elem;
	_new->_desc2 = (_TSPLST_rootView_SubWebZone_roles)_new;
	{_TPPLST_rootView_WebRoleRefprootView_SubWebZone_roles _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_WebRoleRefprootView_SubWebZone_roles;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_SubWebZone_roles (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_SubWebZone_roles (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_WebRoleRef)) 
		return _MkrootView_WebRoleRefprootView_SubWebZone_roles (_coordref, MkrootView_WebRoleRef (_coordref, t));
	message (DEADLY, "a prootView_SubWebZone_roles list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_SubWebZone_roles (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_SubWebZone_roles (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_SubWebZone_roles)
	l =_ElemprootView_SubWebZone_roles (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_SubWebZone_roles)
	r = _ElemprootView_SubWebZone_roles (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_SubWebZone_roles)r)->_desc2);
((_TPPLST_0prootView_SubWebZone_roles)r)->_desc2 = ((_TPPLST_0prootView_SubWebZone_roles)l)->_desc2;
((_TPPLST_0prootView_SubWebZone_roles)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_SubWebZone_roles (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_SubWebZone_roles (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_SubWebZone_roles root;
	_TPPLST_0prootView_SubWebZone_roles list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_SubWebZone_roles)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_SubWebZone_roles)
		t = _ElemprootView_SubWebZone_roles (_coordref, t);
	remain = (_TPPLST_0prootView_SubWebZone_roles)(((_TPPLST_0prootView_SubWebZone_roles) t)->_desc2);
	((_TPPLST_0prootView_SubWebZone_roles) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_SubWebZone_roles;
#else
list = (_TPPLST_0prootView_SubWebZone_roles) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_SubWebZone_roles));
#endif
list->_prod = RULELST_0prootView_SubWebZone_roles;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_SubWebZone_roles)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_WebRoleRefprootView_SubWebZone_roles)
	{_TPPLST_rootView_WebRoleRefprootView_SubWebZone_roles _currn = _currn;
	 _TERMACT_LST_rootView_WebRoleRefprootView_SubWebZone_roles;
	} this code has been moved to _ElemprootView_SubWebZone_roles*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_SubWebZone_roles;
#else
root = (_TPPprootView_SubWebZone_roles) TreeNodeAlloc (sizeof (struct _TPprootView_SubWebZone_roles));
#endif
root->_prod = RULEprootView_SubWebZone_roles;
root->_desc1 = (_TSPLST_rootView_SubWebZone_roles)list;
{ _TPPprootView_SubWebZone_roles _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_SubWebZone_roles;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_SubWebZone_setSize (POSITION *_coordref)
#else
NODEPTR MkprootView_SubWebZone_setSize (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_SubWebZone_setSize _currn;
#ifdef __cplusplus
_currn = new _TPprootView_SubWebZone_setSize;
#else
_currn = (_TPPprootView_SubWebZone_setSize) TreeNodeAlloc (sizeof (struct _TPprootView_SubWebZone_setSize));
#endif
_currn->_prod = RULEprootView_SubWebZone_setSize;
_SETCOORD(_currn)
_TERMACT_prootView_SubWebZone_setSize;
return ( (NODEPTR) _currn);
}/* MkprootView_SubWebZone_setSize */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_SubWebZone_position (POSITION *_coordref)
#else
NODEPTR MkprootView_SubWebZone_position (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_SubWebZone_position _currn;
#ifdef __cplusplus
_currn = new _TPprootView_SubWebZone_position;
#else
_currn = (_TPPprootView_SubWebZone_position) TreeNodeAlloc (sizeof (struct _TPprootView_SubWebZone_position));
#endif
_currn->_prod = RULEprootView_SubWebZone_position;
_SETCOORD(_currn)
_TERMACT_prootView_SubWebZone_position;
return ( (NODEPTR) _currn);
}/* MkprootView_SubWebZone_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebRole (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkprootView_WebRole (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPprootView_WebRole _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebRole;
#else
_currn = (_TPPprootView_WebRole) TreeNodeAlloc (sizeof (struct _TPprootView_WebRole));
#endif
_currn->_prod = RULEprootView_WebRole;
_currn->_desc1 = (_TSProotView_WebRole_name) MkrootView_WebRole_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebRole: root of subtree no. 1 can not be made a rootView_WebRole_name node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_WebRole_color) MkrootView_WebRole_color (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebRole: root of subtree no. 2 can not be made a rootView_WebRole_color node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_WebRole;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_WebRole */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebRole_name (POSITION *_coordref)
#else
NODEPTR MkprootView_WebRole_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_WebRole_name _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebRole_name;
#else
_currn = (_TPPprootView_WebRole_name) TreeNodeAlloc (sizeof (struct _TPprootView_WebRole_name));
#endif
_currn->_prod = RULEprootView_WebRole_name;
_SETCOORD(_currn)
_TERMACT_prootView_WebRole_name;
return ( (NODEPTR) _currn);
}/* MkprootView_WebRole_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebRole_color (POSITION *_coordref)
#else
NODEPTR MkprootView_WebRole_color (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_WebRole_color _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebRole_color;
#else
_currn = (_TPPprootView_WebRole_color) TreeNodeAlloc (sizeof (struct _TPprootView_WebRole_color));
#endif
_currn->_prod = RULEprootView_WebRole_color;
_SETCOORD(_currn)
_TERMACT_prootView_WebRole_color;
return ( (NODEPTR) _currn);
}/* MkprootView_WebRole_color */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebRoleRef (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkprootView_WebRoleRef (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPprootView_WebRoleRef _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebRoleRef;
#else
_currn = (_TPPprootView_WebRoleRef) TreeNodeAlloc (sizeof (struct _TPprootView_WebRoleRef));
#endif
_currn->_prod = RULEprootView_WebRoleRef;
_currn->_desc1 = (_TSProotView_WebRoleRef_role) MkrootView_WebRoleRef_role (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebRoleRef: root of subtree no. 1 can not be made a rootView_WebRoleRef_role node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_WebRoleRef;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_WebRoleRef */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebRoleRef_role (POSITION *_coordref)
#else
NODEPTR MkprootView_WebRoleRef_role (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_WebRoleRef_role _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebRoleRef_role;
#else
_currn = (_TPPprootView_WebRoleRef_role) TreeNodeAlloc (sizeof (struct _TPprootView_WebRoleRef_role));
#endif
_currn->_prod = RULEprootView_WebRoleRef_role;
_SETCOORD(_currn)
_TERMACT_prootView_WebRoleRef_role;
return ( (NODEPTR) _currn);
}/* MkprootView_WebRoleRef_role */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebPage (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5, NODEPTR _desc6)
#else
NODEPTR MkprootView_WebPage (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5,_desc6)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
	NODEPTR _desc6;
#endif
{	_TPPprootView_WebPage _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebPage;
#else
_currn = (_TPPprootView_WebPage) TreeNodeAlloc (sizeof (struct _TPprootView_WebPage));
#endif
_currn->_prod = RULEprootView_WebPage;
_currn->_desc1 = (_TSProotView_WebPage_name) MkrootView_WebPage_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebPage: root of subtree no. 1 can not be made a rootView_WebPage_name node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_WebPage_formalParameter) MkrootView_WebPage_formalParameter (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebPage: root of subtree no. 2 can not be made a rootView_WebPage_formalParameter node ", 0, _coordref);
_currn->_desc3 = (_TSProotView_WebPage_title) MkrootView_WebPage_title (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebPage: root of subtree no. 3 can not be made a rootView_WebPage_title node ", 0, _coordref);
_currn->_desc4 = (_TSProotView_WebPage_frame) MkrootView_WebPage_frame (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebPage: root of subtree no. 4 can not be made a rootView_WebPage_frame node ", 0, _coordref);
_currn->_desc5 = (_TSProotView_WebPage_styleSheetBlock) MkrootView_WebPage_styleSheetBlock (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebPage: root of subtree no. 5 can not be made a rootView_WebPage_styleSheetBlock node ", 0, _coordref);
_currn->_desc6 = (_TSProotView_WebPage_position) MkrootView_WebPage_position (_coordref, _desc6);	
if (((NODEPTR)_currn->_desc6) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebPage: root of subtree no. 6 can not be made a rootView_WebPage_position node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_WebPage;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_WebPage */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebPage_name (POSITION *_coordref)
#else
NODEPTR MkprootView_WebPage_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_WebPage_name _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebPage_name;
#else
_currn = (_TPPprootView_WebPage_name) TreeNodeAlloc (sizeof (struct _TPprootView_WebPage_name));
#endif
_currn->_prod = RULEprootView_WebPage_name;
_SETCOORD(_currn)
_TERMACT_prootView_WebPage_name;
return ( (NODEPTR) _currn);
}/* MkprootView_WebPage_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_WebNodeFormalParamprootView_WebPage_formalParameter (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_WebNodeFormalParamprootView_WebPage_formalParameter (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_WebNodeFormalParamprootView_WebPage_formalParameter _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_WebNodeFormalParamprootView_WebPage_formalParameter;
#else
	_new = (_TPPLST_rootView_WebNodeFormalParamprootView_WebPage_formalParameter) TreeNodeAlloc (sizeof (struct _TPLST_rootView_WebNodeFormalParamprootView_WebPage_formalParameter));
#endif
	_new->_prod = RULELST_rootView_WebNodeFormalParamprootView_WebPage_formalParameter;
	_new->_desc1 = (_TSProotView_WebNodeFormalParam)elem;
	_new->_desc2 = (_TSPLST_rootView_WebPage_formalParameter)_new;
	{_TPPLST_rootView_WebNodeFormalParamprootView_WebPage_formalParameter _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_WebNodeFormalParamprootView_WebPage_formalParameter;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_WebPage_formalParameter (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_WebPage_formalParameter (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_WebNodeFormalParam)) 
		return _MkrootView_WebNodeFormalParamprootView_WebPage_formalParameter (_coordref, MkrootView_WebNodeFormalParam (_coordref, t));
	message (DEADLY, "a prootView_WebPage_formalParameter list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_WebPage_formalParameter (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_WebPage_formalParameter (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_WebPage_formalParameter)
	l =_ElemprootView_WebPage_formalParameter (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_WebPage_formalParameter)
	r = _ElemprootView_WebPage_formalParameter (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_WebPage_formalParameter)r)->_desc2);
((_TPPLST_0prootView_WebPage_formalParameter)r)->_desc2 = ((_TPPLST_0prootView_WebPage_formalParameter)l)->_desc2;
((_TPPLST_0prootView_WebPage_formalParameter)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebPage_formalParameter (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_WebPage_formalParameter (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_WebPage_formalParameter root;
	_TPPLST_0prootView_WebPage_formalParameter list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_WebPage_formalParameter)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_WebPage_formalParameter)
		t = _ElemprootView_WebPage_formalParameter (_coordref, t);
	remain = (_TPPLST_0prootView_WebPage_formalParameter)(((_TPPLST_0prootView_WebPage_formalParameter) t)->_desc2);
	((_TPPLST_0prootView_WebPage_formalParameter) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_WebPage_formalParameter;
#else
list = (_TPPLST_0prootView_WebPage_formalParameter) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_WebPage_formalParameter));
#endif
list->_prod = RULELST_0prootView_WebPage_formalParameter;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_WebPage_formalParameter)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_WebNodeFormalParamprootView_WebPage_formalParameter)
	{_TPPLST_rootView_WebNodeFormalParamprootView_WebPage_formalParameter _currn = _currn;
	 _TERMACT_LST_rootView_WebNodeFormalParamprootView_WebPage_formalParameter;
	} this code has been moved to _ElemprootView_WebPage_formalParameter*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_WebPage_formalParameter;
#else
root = (_TPPprootView_WebPage_formalParameter) TreeNodeAlloc (sizeof (struct _TPprootView_WebPage_formalParameter));
#endif
root->_prod = RULEprootView_WebPage_formalParameter;
root->_desc1 = (_TSPLST_rootView_WebPage_formalParameter)list;
{ _TPPprootView_WebPage_formalParameter _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_WebPage_formalParameter;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebPage_title (POSITION *_coordref)
#else
NODEPTR MkprootView_WebPage_title (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_WebPage_title _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebPage_title;
#else
_currn = (_TPPprootView_WebPage_title) TreeNodeAlloc (sizeof (struct _TPprootView_WebPage_title));
#endif
_currn->_prod = RULEprootView_WebPage_title;
_SETCOORD(_currn)
_TERMACT_prootView_WebPage_title;
return ( (NODEPTR) _currn);
}/* MkprootView_WebPage_title */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebPage_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkprootView_WebPage_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_WebPage_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebPage_styleSheetBlock;
#else
_currn = (_TPPprootView_WebPage_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPprootView_WebPage_styleSheetBlock));
#endif
_currn->_prod = RULEprootView_WebPage_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_prootView_WebPage_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkprootView_WebPage_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebPage_position (POSITION *_coordref)
#else
NODEPTR MkprootView_WebPage_position (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_WebPage_position _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebPage_position;
#else
_currn = (_TPPprootView_WebPage_position) TreeNodeAlloc (sizeof (struct _TPprootView_WebPage_position));
#endif
_currn->_prod = RULEprootView_WebPage_position;
_SETCOORD(_currn)
_TERMACT_prootView_WebPage_position;
return ( (NODEPTR) _currn);
}/* MkprootView_WebPage_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_Frame (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkprootView_Frame (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPprootView_Frame _currn;
#ifdef __cplusplus
_currn = new _TPprootView_Frame;
#else
_currn = (_TPPprootView_Frame) TreeNodeAlloc (sizeof (struct _TPprootView_Frame));
#endif
_currn->_prod = RULEprootView_Frame;
_currn->_desc1 = (_TSProotView_Frame_components) MkrootView_Frame_components (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_Frame: root of subtree no. 1 can not be made a rootView_Frame_components node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_Frame_styleSheetBlock) MkrootView_Frame_styleSheetBlock (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_Frame: root of subtree no. 2 can not be made a rootView_Frame_styleSheetBlock node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_Frame;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_Frame */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_FormularprootView_Frame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_FormularprootView_Frame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_FormularprootView_Frame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_FormularprootView_Frame_components;
#else
	_new = (_TPPLST_rootView_FormularprootView_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_rootView_FormularprootView_Frame_components));
#endif
	_new->_prod = RULELST_rootView_FormularprootView_Frame_components;
	_new->_desc1 = (_TSProotView_Formular)elem;
	_new->_desc2 = (_TSPLST_rootView_Frame_components)_new;
	{_TPPLST_rootView_FormularprootView_Frame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_FormularprootView_Frame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_ParagraphprootView_Frame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_ParagraphprootView_Frame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_ParagraphprootView_Frame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_ParagraphprootView_Frame_components;
#else
	_new = (_TPPLST_rootView_ParagraphprootView_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_rootView_ParagraphprootView_Frame_components));
#endif
	_new->_prod = RULELST_rootView_ParagraphprootView_Frame_components;
	_new->_desc1 = (_TSProotView_Paragraph)elem;
	_new->_desc2 = (_TSPLST_rootView_Frame_components)_new;
	{_TPPLST_rootView_ParagraphprootView_Frame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_ParagraphprootView_Frame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_HeadingprootView_Frame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_HeadingprootView_Frame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_HeadingprootView_Frame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_HeadingprootView_Frame_components;
#else
	_new = (_TPPLST_rootView_HeadingprootView_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_rootView_HeadingprootView_Frame_components));
#endif
	_new->_prod = RULELST_rootView_HeadingprootView_Frame_components;
	_new->_desc1 = (_TSProotView_Heading)elem;
	_new->_desc2 = (_TSPLST_rootView_Frame_components)_new;
	{_TPPLST_rootView_HeadingprootView_Frame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_HeadingprootView_Frame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_TableprootView_Frame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_TableprootView_Frame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_TableprootView_Frame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_TableprootView_Frame_components;
#else
	_new = (_TPPLST_rootView_TableprootView_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_rootView_TableprootView_Frame_components));
#endif
	_new->_prod = RULELST_rootView_TableprootView_Frame_components;
	_new->_desc1 = (_TSProotView_Table)elem;
	_new->_desc2 = (_TSPLST_rootView_Frame_components)_new;
	{_TPPLST_rootView_TableprootView_Frame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_TableprootView_Frame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_FrameRefprootView_Frame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_FrameRefprootView_Frame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_FrameRefprootView_Frame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_FrameRefprootView_Frame_components;
#else
	_new = (_TPPLST_rootView_FrameRefprootView_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_rootView_FrameRefprootView_Frame_components));
#endif
	_new->_prod = RULELST_rootView_FrameRefprootView_Frame_components;
	_new->_desc1 = (_TSProotView_FrameRef)elem;
	_new->_desc2 = (_TSPLST_rootView_Frame_components)_new;
	{_TPPLST_rootView_FrameRefprootView_Frame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_FrameRefprootView_Frame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_ListOutputElementprootView_Frame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_ListOutputElementprootView_Frame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_ListOutputElementprootView_Frame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_ListOutputElementprootView_Frame_components;
#else
	_new = (_TPPLST_rootView_ListOutputElementprootView_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_rootView_ListOutputElementprootView_Frame_components));
#endif
	_new->_prod = RULELST_rootView_ListOutputElementprootView_Frame_components;
	_new->_desc1 = (_TSProotView_ListOutputElement)elem;
	_new->_desc2 = (_TSPLST_rootView_Frame_components)_new;
	{_TPPLST_rootView_ListOutputElementprootView_Frame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_ListOutputElementprootView_Frame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_Frame_components (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_Frame_components (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_Formular)) 
		return _MkrootView_FormularprootView_Frame_components (_coordref, MkrootView_Formular (_coordref, t));
	if (IsSymb (t, SYMBrootView_Paragraph)) 
		return _MkrootView_ParagraphprootView_Frame_components (_coordref, MkrootView_Paragraph (_coordref, t));
	if (IsSymb (t, SYMBrootView_Heading)) 
		return _MkrootView_HeadingprootView_Frame_components (_coordref, MkrootView_Heading (_coordref, t));
	if (IsSymb (t, SYMBrootView_Table)) 
		return _MkrootView_TableprootView_Frame_components (_coordref, MkrootView_Table (_coordref, t));
	if (IsSymb (t, SYMBrootView_FrameRef)) 
		return _MkrootView_FrameRefprootView_Frame_components (_coordref, MkrootView_FrameRef (_coordref, t));
	if (IsSymb (t, SYMBrootView_ListOutputElement)) 
		return _MkrootView_ListOutputElementprootView_Frame_components (_coordref, MkrootView_ListOutputElement (_coordref, t));
	message (DEADLY, "a prootView_Frame_components list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_Frame_components (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_Frame_components (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_Frame_components)
	l =_ElemprootView_Frame_components (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_Frame_components)
	r = _ElemprootView_Frame_components (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_Frame_components)r)->_desc2);
((_TPPLST_0prootView_Frame_components)r)->_desc2 = ((_TPPLST_0prootView_Frame_components)l)->_desc2;
((_TPPLST_0prootView_Frame_components)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_Frame_components (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_Frame_components (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_Frame_components root;
	_TPPLST_0prootView_Frame_components list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_Frame_components)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_Frame_components)
		t = _ElemprootView_Frame_components (_coordref, t);
	remain = (_TPPLST_0prootView_Frame_components)(((_TPPLST_0prootView_Frame_components) t)->_desc2);
	((_TPPLST_0prootView_Frame_components) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_Frame_components;
#else
list = (_TPPLST_0prootView_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_Frame_components));
#endif
list->_prod = RULELST_0prootView_Frame_components;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_Frame_components)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_FormularprootView_Frame_components)
	{_TPPLST_rootView_FormularprootView_Frame_components _currn = _currn;
	 _TERMACT_LST_rootView_FormularprootView_Frame_components;
	} this code has been moved to _ElemprootView_Frame_components*/
	/*if (_currn->_prod == RULELST_rootView_ParagraphprootView_Frame_components)
	{_TPPLST_rootView_ParagraphprootView_Frame_components _currn = _currn;
	 _TERMACT_LST_rootView_ParagraphprootView_Frame_components;
	} this code has been moved to _ElemprootView_Frame_components*/
	/*if (_currn->_prod == RULELST_rootView_HeadingprootView_Frame_components)
	{_TPPLST_rootView_HeadingprootView_Frame_components _currn = _currn;
	 _TERMACT_LST_rootView_HeadingprootView_Frame_components;
	} this code has been moved to _ElemprootView_Frame_components*/
	/*if (_currn->_prod == RULELST_rootView_TableprootView_Frame_components)
	{_TPPLST_rootView_TableprootView_Frame_components _currn = _currn;
	 _TERMACT_LST_rootView_TableprootView_Frame_components;
	} this code has been moved to _ElemprootView_Frame_components*/
	/*if (_currn->_prod == RULELST_rootView_FrameRefprootView_Frame_components)
	{_TPPLST_rootView_FrameRefprootView_Frame_components _currn = _currn;
	 _TERMACT_LST_rootView_FrameRefprootView_Frame_components;
	} this code has been moved to _ElemprootView_Frame_components*/
	/*if (_currn->_prod == RULELST_rootView_ListOutputElementprootView_Frame_components)
	{_TPPLST_rootView_ListOutputElementprootView_Frame_components _currn = _currn;
	 _TERMACT_LST_rootView_ListOutputElementprootView_Frame_components;
	} this code has been moved to _ElemprootView_Frame_components*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_Frame_components;
#else
root = (_TPPprootView_Frame_components) TreeNodeAlloc (sizeof (struct _TPprootView_Frame_components));
#endif
root->_prod = RULEprootView_Frame_components;
root->_desc1 = (_TSPLST_rootView_Frame_components)list;
{ _TPPprootView_Frame_components _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_Frame_components;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_Frame_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkprootView_Frame_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_Frame_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPprootView_Frame_styleSheetBlock;
#else
_currn = (_TPPprootView_Frame_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPprootView_Frame_styleSheetBlock));
#endif
_currn->_prod = RULEprootView_Frame_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_prootView_Frame_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkprootView_Frame_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_GlobalFrame (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkprootView_GlobalFrame (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPprootView_GlobalFrame _currn;
#ifdef __cplusplus
_currn = new _TPprootView_GlobalFrame;
#else
_currn = (_TPPprootView_GlobalFrame) TreeNodeAlloc (sizeof (struct _TPprootView_GlobalFrame));
#endif
_currn->_prod = RULEprootView_GlobalFrame;
_currn->_desc1 = (_TSProotView_GlobalFrame_name) MkrootView_GlobalFrame_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_GlobalFrame: root of subtree no. 1 can not be made a rootView_GlobalFrame_name node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_GlobalFrame_components) MkrootView_GlobalFrame_components (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_GlobalFrame: root of subtree no. 2 can not be made a rootView_GlobalFrame_components node ", 0, _coordref);
_currn->_desc3 = (_TSProotView_GlobalFrame_setSize) MkrootView_GlobalFrame_setSize (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_GlobalFrame: root of subtree no. 3 can not be made a rootView_GlobalFrame_setSize node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_GlobalFrame;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_GlobalFrame */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_GlobalFrame_name (POSITION *_coordref)
#else
NODEPTR MkprootView_GlobalFrame_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_GlobalFrame_name _currn;
#ifdef __cplusplus
_currn = new _TPprootView_GlobalFrame_name;
#else
_currn = (_TPPprootView_GlobalFrame_name) TreeNodeAlloc (sizeof (struct _TPprootView_GlobalFrame_name));
#endif
_currn->_prod = RULEprootView_GlobalFrame_name;
_SETCOORD(_currn)
_TERMACT_prootView_GlobalFrame_name;
return ( (NODEPTR) _currn);
}/* MkprootView_GlobalFrame_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_FormularprootView_GlobalFrame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_FormularprootView_GlobalFrame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_FormularprootView_GlobalFrame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_FormularprootView_GlobalFrame_components;
#else
	_new = (_TPPLST_rootView_FormularprootView_GlobalFrame_components) TreeNodeAlloc (sizeof (struct _TPLST_rootView_FormularprootView_GlobalFrame_components));
#endif
	_new->_prod = RULELST_rootView_FormularprootView_GlobalFrame_components;
	_new->_desc1 = (_TSProotView_Formular)elem;
	_new->_desc2 = (_TSPLST_rootView_GlobalFrame_components)_new;
	{_TPPLST_rootView_FormularprootView_GlobalFrame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_FormularprootView_GlobalFrame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_ParagraphprootView_GlobalFrame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_ParagraphprootView_GlobalFrame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_ParagraphprootView_GlobalFrame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_ParagraphprootView_GlobalFrame_components;
#else
	_new = (_TPPLST_rootView_ParagraphprootView_GlobalFrame_components) TreeNodeAlloc (sizeof (struct _TPLST_rootView_ParagraphprootView_GlobalFrame_components));
#endif
	_new->_prod = RULELST_rootView_ParagraphprootView_GlobalFrame_components;
	_new->_desc1 = (_TSProotView_Paragraph)elem;
	_new->_desc2 = (_TSPLST_rootView_GlobalFrame_components)_new;
	{_TPPLST_rootView_ParagraphprootView_GlobalFrame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_ParagraphprootView_GlobalFrame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_HeadingprootView_GlobalFrame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_HeadingprootView_GlobalFrame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_HeadingprootView_GlobalFrame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_HeadingprootView_GlobalFrame_components;
#else
	_new = (_TPPLST_rootView_HeadingprootView_GlobalFrame_components) TreeNodeAlloc (sizeof (struct _TPLST_rootView_HeadingprootView_GlobalFrame_components));
#endif
	_new->_prod = RULELST_rootView_HeadingprootView_GlobalFrame_components;
	_new->_desc1 = (_TSProotView_Heading)elem;
	_new->_desc2 = (_TSPLST_rootView_GlobalFrame_components)_new;
	{_TPPLST_rootView_HeadingprootView_GlobalFrame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_HeadingprootView_GlobalFrame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_TableprootView_GlobalFrame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_TableprootView_GlobalFrame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_TableprootView_GlobalFrame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_TableprootView_GlobalFrame_components;
#else
	_new = (_TPPLST_rootView_TableprootView_GlobalFrame_components) TreeNodeAlloc (sizeof (struct _TPLST_rootView_TableprootView_GlobalFrame_components));
#endif
	_new->_prod = RULELST_rootView_TableprootView_GlobalFrame_components;
	_new->_desc1 = (_TSProotView_Table)elem;
	_new->_desc2 = (_TSPLST_rootView_GlobalFrame_components)_new;
	{_TPPLST_rootView_TableprootView_GlobalFrame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_TableprootView_GlobalFrame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_FrameRefprootView_GlobalFrame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_FrameRefprootView_GlobalFrame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_FrameRefprootView_GlobalFrame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_FrameRefprootView_GlobalFrame_components;
#else
	_new = (_TPPLST_rootView_FrameRefprootView_GlobalFrame_components) TreeNodeAlloc (sizeof (struct _TPLST_rootView_FrameRefprootView_GlobalFrame_components));
#endif
	_new->_prod = RULELST_rootView_FrameRefprootView_GlobalFrame_components;
	_new->_desc1 = (_TSProotView_FrameRef)elem;
	_new->_desc2 = (_TSPLST_rootView_GlobalFrame_components)_new;
	{_TPPLST_rootView_FrameRefprootView_GlobalFrame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_FrameRefprootView_GlobalFrame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_ListOutputElementprootView_GlobalFrame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_ListOutputElementprootView_GlobalFrame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_ListOutputElementprootView_GlobalFrame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_ListOutputElementprootView_GlobalFrame_components;
#else
	_new = (_TPPLST_rootView_ListOutputElementprootView_GlobalFrame_components) TreeNodeAlloc (sizeof (struct _TPLST_rootView_ListOutputElementprootView_GlobalFrame_components));
#endif
	_new->_prod = RULELST_rootView_ListOutputElementprootView_GlobalFrame_components;
	_new->_desc1 = (_TSProotView_ListOutputElement)elem;
	_new->_desc2 = (_TSPLST_rootView_GlobalFrame_components)_new;
	{_TPPLST_rootView_ListOutputElementprootView_GlobalFrame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_ListOutputElementprootView_GlobalFrame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_GlobalFrame_components (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_GlobalFrame_components (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_Formular)) 
		return _MkrootView_FormularprootView_GlobalFrame_components (_coordref, MkrootView_Formular (_coordref, t));
	if (IsSymb (t, SYMBrootView_Paragraph)) 
		return _MkrootView_ParagraphprootView_GlobalFrame_components (_coordref, MkrootView_Paragraph (_coordref, t));
	if (IsSymb (t, SYMBrootView_Heading)) 
		return _MkrootView_HeadingprootView_GlobalFrame_components (_coordref, MkrootView_Heading (_coordref, t));
	if (IsSymb (t, SYMBrootView_Table)) 
		return _MkrootView_TableprootView_GlobalFrame_components (_coordref, MkrootView_Table (_coordref, t));
	if (IsSymb (t, SYMBrootView_FrameRef)) 
		return _MkrootView_FrameRefprootView_GlobalFrame_components (_coordref, MkrootView_FrameRef (_coordref, t));
	if (IsSymb (t, SYMBrootView_ListOutputElement)) 
		return _MkrootView_ListOutputElementprootView_GlobalFrame_components (_coordref, MkrootView_ListOutputElement (_coordref, t));
	message (DEADLY, "a prootView_GlobalFrame_components list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_GlobalFrame_components (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_GlobalFrame_components (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_GlobalFrame_components)
	l =_ElemprootView_GlobalFrame_components (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_GlobalFrame_components)
	r = _ElemprootView_GlobalFrame_components (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_GlobalFrame_components)r)->_desc2);
((_TPPLST_0prootView_GlobalFrame_components)r)->_desc2 = ((_TPPLST_0prootView_GlobalFrame_components)l)->_desc2;
((_TPPLST_0prootView_GlobalFrame_components)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_GlobalFrame_components (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_GlobalFrame_components (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_GlobalFrame_components root;
	_TPPLST_0prootView_GlobalFrame_components list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_GlobalFrame_components)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_GlobalFrame_components)
		t = _ElemprootView_GlobalFrame_components (_coordref, t);
	remain = (_TPPLST_0prootView_GlobalFrame_components)(((_TPPLST_0prootView_GlobalFrame_components) t)->_desc2);
	((_TPPLST_0prootView_GlobalFrame_components) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_GlobalFrame_components;
#else
list = (_TPPLST_0prootView_GlobalFrame_components) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_GlobalFrame_components));
#endif
list->_prod = RULELST_0prootView_GlobalFrame_components;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_GlobalFrame_components)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_FormularprootView_GlobalFrame_components)
	{_TPPLST_rootView_FormularprootView_GlobalFrame_components _currn = _currn;
	 _TERMACT_LST_rootView_FormularprootView_GlobalFrame_components;
	} this code has been moved to _ElemprootView_GlobalFrame_components*/
	/*if (_currn->_prod == RULELST_rootView_ParagraphprootView_GlobalFrame_components)
	{_TPPLST_rootView_ParagraphprootView_GlobalFrame_components _currn = _currn;
	 _TERMACT_LST_rootView_ParagraphprootView_GlobalFrame_components;
	} this code has been moved to _ElemprootView_GlobalFrame_components*/
	/*if (_currn->_prod == RULELST_rootView_HeadingprootView_GlobalFrame_components)
	{_TPPLST_rootView_HeadingprootView_GlobalFrame_components _currn = _currn;
	 _TERMACT_LST_rootView_HeadingprootView_GlobalFrame_components;
	} this code has been moved to _ElemprootView_GlobalFrame_components*/
	/*if (_currn->_prod == RULELST_rootView_TableprootView_GlobalFrame_components)
	{_TPPLST_rootView_TableprootView_GlobalFrame_components _currn = _currn;
	 _TERMACT_LST_rootView_TableprootView_GlobalFrame_components;
	} this code has been moved to _ElemprootView_GlobalFrame_components*/
	/*if (_currn->_prod == RULELST_rootView_FrameRefprootView_GlobalFrame_components)
	{_TPPLST_rootView_FrameRefprootView_GlobalFrame_components _currn = _currn;
	 _TERMACT_LST_rootView_FrameRefprootView_GlobalFrame_components;
	} this code has been moved to _ElemprootView_GlobalFrame_components*/
	/*if (_currn->_prod == RULELST_rootView_ListOutputElementprootView_GlobalFrame_components)
	{_TPPLST_rootView_ListOutputElementprootView_GlobalFrame_components _currn = _currn;
	 _TERMACT_LST_rootView_ListOutputElementprootView_GlobalFrame_components;
	} this code has been moved to _ElemprootView_GlobalFrame_components*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_GlobalFrame_components;
#else
root = (_TPPprootView_GlobalFrame_components) TreeNodeAlloc (sizeof (struct _TPprootView_GlobalFrame_components));
#endif
root->_prod = RULEprootView_GlobalFrame_components;
root->_desc1 = (_TSPLST_rootView_GlobalFrame_components)list;
{ _TPPprootView_GlobalFrame_components _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_GlobalFrame_components;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_GlobalFrame_setSize (POSITION *_coordref)
#else
NODEPTR MkprootView_GlobalFrame_setSize (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_GlobalFrame_setSize _currn;
#ifdef __cplusplus
_currn = new _TPprootView_GlobalFrame_setSize;
#else
_currn = (_TPPprootView_GlobalFrame_setSize) TreeNodeAlloc (sizeof (struct _TPprootView_GlobalFrame_setSize));
#endif
_currn->_prod = RULEprootView_GlobalFrame_setSize;
_SETCOORD(_currn)
_TERMACT_prootView_GlobalFrame_setSize;
return ( (NODEPTR) _currn);
}/* MkprootView_GlobalFrame_setSize */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_NavigationBar (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkprootView_NavigationBar (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPprootView_NavigationBar _currn;
#ifdef __cplusplus
_currn = new _TPprootView_NavigationBar;
#else
_currn = (_TPPprootView_NavigationBar) TreeNodeAlloc (sizeof (struct _TPprootView_NavigationBar));
#endif
_currn->_prod = RULEprootView_NavigationBar;
_currn->_desc1 = (_TSProotView_NavigationBar_styleSheetBlock) MkrootView_NavigationBar_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_NavigationBar: root of subtree no. 1 can not be made a rootView_NavigationBar_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_NavigationBar_elements) MkrootView_NavigationBar_elements (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_NavigationBar: root of subtree no. 2 can not be made a rootView_NavigationBar_elements node ", 0, _coordref);
_currn->_desc3 = (_TSProotView_NavigationBar_alignment) MkrootView_NavigationBar_alignment (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_NavigationBar: root of subtree no. 3 can not be made a rootView_NavigationBar_alignment node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_NavigationBar;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_NavigationBar */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_NavigationBar_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkprootView_NavigationBar_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_NavigationBar_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPprootView_NavigationBar_styleSheetBlock;
#else
_currn = (_TPPprootView_NavigationBar_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPprootView_NavigationBar_styleSheetBlock));
#endif
_currn->_prod = RULEprootView_NavigationBar_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_prootView_NavigationBar_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkprootView_NavigationBar_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_NavigationBarElementprootView_NavigationBar_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_NavigationBarElementprootView_NavigationBar_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_NavigationBarElementprootView_NavigationBar_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_NavigationBarElementprootView_NavigationBar_elements;
#else
	_new = (_TPPLST_rootView_NavigationBarElementprootView_NavigationBar_elements) TreeNodeAlloc (sizeof (struct _TPLST_rootView_NavigationBarElementprootView_NavigationBar_elements));
#endif
	_new->_prod = RULELST_rootView_NavigationBarElementprootView_NavigationBar_elements;
	_new->_desc1 = (_TSProotView_NavigationBarElement)elem;
	_new->_desc2 = (_TSPLST_rootView_NavigationBar_elements)_new;
	{_TPPLST_rootView_NavigationBarElementprootView_NavigationBar_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_NavigationBarElementprootView_NavigationBar_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_NavigationBar_elements (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_NavigationBar_elements (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_NavigationBarElement)) 
		return _MkrootView_NavigationBarElementprootView_NavigationBar_elements (_coordref, MkrootView_NavigationBarElement (_coordref, t));
	message (DEADLY, "a prootView_NavigationBar_elements list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_NavigationBar_elements (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_NavigationBar_elements (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_NavigationBar_elements)
	l =_ElemprootView_NavigationBar_elements (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_NavigationBar_elements)
	r = _ElemprootView_NavigationBar_elements (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_NavigationBar_elements)r)->_desc2);
((_TPPLST_0prootView_NavigationBar_elements)r)->_desc2 = ((_TPPLST_0prootView_NavigationBar_elements)l)->_desc2;
((_TPPLST_0prootView_NavigationBar_elements)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_NavigationBar_elements (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_NavigationBar_elements (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_NavigationBar_elements root;
	_TPPLST_0prootView_NavigationBar_elements list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_NavigationBar_elements)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_NavigationBar_elements)
		t = _ElemprootView_NavigationBar_elements (_coordref, t);
	remain = (_TPPLST_0prootView_NavigationBar_elements)(((_TPPLST_0prootView_NavigationBar_elements) t)->_desc2);
	((_TPPLST_0prootView_NavigationBar_elements) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_NavigationBar_elements;
#else
list = (_TPPLST_0prootView_NavigationBar_elements) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_NavigationBar_elements));
#endif
list->_prod = RULELST_0prootView_NavigationBar_elements;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_NavigationBar_elements)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_NavigationBarElementprootView_NavigationBar_elements)
	{_TPPLST_rootView_NavigationBarElementprootView_NavigationBar_elements _currn = _currn;
	 _TERMACT_LST_rootView_NavigationBarElementprootView_NavigationBar_elements;
	} this code has been moved to _ElemprootView_NavigationBar_elements*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_NavigationBar_elements;
#else
root = (_TPPprootView_NavigationBar_elements) TreeNodeAlloc (sizeof (struct _TPprootView_NavigationBar_elements));
#endif
root->_prod = RULEprootView_NavigationBar_elements;
root->_desc1 = (_TSPLST_rootView_NavigationBar_elements)list;
{ _TPPprootView_NavigationBar_elements _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_NavigationBar_elements;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_NavigationBar_alignment (POSITION *_coordref)
#else
NODEPTR MkprootView_NavigationBar_alignment (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_NavigationBar_alignment _currn;
#ifdef __cplusplus
_currn = new _TPprootView_NavigationBar_alignment;
#else
_currn = (_TPPprootView_NavigationBar_alignment) TreeNodeAlloc (sizeof (struct _TPprootView_NavigationBar_alignment));
#endif
_currn->_prod = RULEprootView_NavigationBar_alignment;
_SETCOORD(_currn)
_TERMACT_prootView_NavigationBar_alignment;
return ( (NODEPTR) _currn);
}/* MkprootView_NavigationBar_alignment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_NavigationBarElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkprootView_NavigationBarElement (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPprootView_NavigationBarElement _currn;
#ifdef __cplusplus
_currn = new _TPprootView_NavigationBarElement;
#else
_currn = (_TPPprootView_NavigationBarElement) TreeNodeAlloc (sizeof (struct _TPprootView_NavigationBarElement));
#endif
_currn->_prod = RULEprootView_NavigationBarElement;
_currn->_desc1 = (_TSProotView_NavigationBarElement_target) MkrootView_NavigationBarElement_target (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_NavigationBarElement: root of subtree no. 1 can not be made a rootView_NavigationBarElement_target node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_NavigationBarElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_NavigationBarElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_NavigationBarElement_target (POSITION *_coordref)
#else
NODEPTR MkprootView_NavigationBarElement_target (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_NavigationBarElement_target _currn;
#ifdef __cplusplus
_currn = new _TPprootView_NavigationBarElement_target;
#else
_currn = (_TPPprootView_NavigationBarElement_target) TreeNodeAlloc (sizeof (struct _TPprootView_NavigationBarElement_target));
#endif
_currn->_prod = RULEprootView_NavigationBarElement_target;
_SETCOORD(_currn)
_TERMACT_prootView_NavigationBarElement_target;
return ( (NODEPTR) _currn);
}/* MkprootView_NavigationBarElement_target */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FrameRef (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkprootView_FrameRef (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPprootView_FrameRef _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FrameRef;
#else
_currn = (_TPPprootView_FrameRef) TreeNodeAlloc (sizeof (struct _TPprootView_FrameRef));
#endif
_currn->_prod = RULEprootView_FrameRef;
_currn->_desc1 = (_TSProotView_FrameRef_styleSheetBlock) MkrootView_FrameRef_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FrameRef: root of subtree no. 1 can not be made a rootView_FrameRef_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_FrameRef_frame) MkrootView_FrameRef_frame (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FrameRef: root of subtree no. 2 can not be made a rootView_FrameRef_frame node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_FrameRef;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_FrameRef */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FrameRef_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkprootView_FrameRef_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FrameRef_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FrameRef_styleSheetBlock;
#else
_currn = (_TPPprootView_FrameRef_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPprootView_FrameRef_styleSheetBlock));
#endif
_currn->_prod = RULEprootView_FrameRef_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_prootView_FrameRef_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkprootView_FrameRef_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FrameRef_frame (POSITION *_coordref)
#else
NODEPTR MkprootView_FrameRef_frame (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FrameRef_frame _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FrameRef_frame;
#else
_currn = (_TPPprootView_FrameRef_frame) TreeNodeAlloc (sizeof (struct _TPprootView_FrameRef_frame));
#endif
_currn->_prod = RULEprootView_FrameRef_frame;
_SETCOORD(_currn)
_TERMACT_prootView_FrameRef_frame;
return ( (NODEPTR) _currn);
}/* MkprootView_FrameRef_frame */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_MetaInformation (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkprootView_MetaInformation (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPprootView_MetaInformation _currn;
#ifdef __cplusplus
_currn = new _TPprootView_MetaInformation;
#else
_currn = (_TPPprootView_MetaInformation) TreeNodeAlloc (sizeof (struct _TPprootView_MetaInformation));
#endif
_currn->_prod = RULEprootView_MetaInformation;
_currn->_desc1 = (_TSProotView_MetaInformation_name) MkrootView_MetaInformation_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_MetaInformation: root of subtree no. 1 can not be made a rootView_MetaInformation_name node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_MetaInformation_elements) MkrootView_MetaInformation_elements (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_MetaInformation: root of subtree no. 2 can not be made a rootView_MetaInformation_elements node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_MetaInformation;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_MetaInformation */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_MetaInformation_name (POSITION *_coordref)
#else
NODEPTR MkprootView_MetaInformation_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_MetaInformation_name _currn;
#ifdef __cplusplus
_currn = new _TPprootView_MetaInformation_name;
#else
_currn = (_TPPprootView_MetaInformation_name) TreeNodeAlloc (sizeof (struct _TPprootView_MetaInformation_name));
#endif
_currn->_prod = RULEprootView_MetaInformation_name;
_SETCOORD(_currn)
_TERMACT_prootView_MetaInformation_name;
return ( (NODEPTR) _currn);
}/* MkprootView_MetaInformation_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_MetaInformationElementprootView_MetaInformation_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_MetaInformationElementprootView_MetaInformation_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_MetaInformationElementprootView_MetaInformation_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_MetaInformationElementprootView_MetaInformation_elements;
#else
	_new = (_TPPLST_rootView_MetaInformationElementprootView_MetaInformation_elements) TreeNodeAlloc (sizeof (struct _TPLST_rootView_MetaInformationElementprootView_MetaInformation_elements));
#endif
	_new->_prod = RULELST_rootView_MetaInformationElementprootView_MetaInformation_elements;
	_new->_desc1 = (_TSProotView_MetaInformationElement)elem;
	_new->_desc2 = (_TSPLST_rootView_MetaInformation_elements)_new;
	{_TPPLST_rootView_MetaInformationElementprootView_MetaInformation_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_MetaInformationElementprootView_MetaInformation_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_MetaInformation_elements (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_MetaInformation_elements (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_MetaInformationElement)) 
		return _MkrootView_MetaInformationElementprootView_MetaInformation_elements (_coordref, MkrootView_MetaInformationElement (_coordref, t));
	message (DEADLY, "a prootView_MetaInformation_elements list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_MetaInformation_elements (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_MetaInformation_elements (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_MetaInformation_elements)
	l =_ElemprootView_MetaInformation_elements (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_MetaInformation_elements)
	r = _ElemprootView_MetaInformation_elements (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_MetaInformation_elements)r)->_desc2);
((_TPPLST_0prootView_MetaInformation_elements)r)->_desc2 = ((_TPPLST_0prootView_MetaInformation_elements)l)->_desc2;
((_TPPLST_0prootView_MetaInformation_elements)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_MetaInformation_elements (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_MetaInformation_elements (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_MetaInformation_elements root;
	_TPPLST_0prootView_MetaInformation_elements list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_MetaInformation_elements)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_MetaInformation_elements)
		t = _ElemprootView_MetaInformation_elements (_coordref, t);
	remain = (_TPPLST_0prootView_MetaInformation_elements)(((_TPPLST_0prootView_MetaInformation_elements) t)->_desc2);
	((_TPPLST_0prootView_MetaInformation_elements) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_MetaInformation_elements;
#else
list = (_TPPLST_0prootView_MetaInformation_elements) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_MetaInformation_elements));
#endif
list->_prod = RULELST_0prootView_MetaInformation_elements;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_MetaInformation_elements)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_MetaInformationElementprootView_MetaInformation_elements)
	{_TPPLST_rootView_MetaInformationElementprootView_MetaInformation_elements _currn = _currn;
	 _TERMACT_LST_rootView_MetaInformationElementprootView_MetaInformation_elements;
	} this code has been moved to _ElemprootView_MetaInformation_elements*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_MetaInformation_elements;
#else
root = (_TPPprootView_MetaInformation_elements) TreeNodeAlloc (sizeof (struct _TPprootView_MetaInformation_elements));
#endif
root->_prod = RULEprootView_MetaInformation_elements;
root->_desc1 = (_TSPLST_rootView_MetaInformation_elements)list;
{ _TPPprootView_MetaInformation_elements _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_MetaInformation_elements;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_MetaInformationElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkprootView_MetaInformationElement (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPprootView_MetaInformationElement _currn;
#ifdef __cplusplus
_currn = new _TPprootView_MetaInformationElement;
#else
_currn = (_TPPprootView_MetaInformationElement) TreeNodeAlloc (sizeof (struct _TPprootView_MetaInformationElement));
#endif
_currn->_prod = RULEprootView_MetaInformationElement;
_currn->_desc1 = (_TSProotView_MetaInformationElement_property) MkrootView_MetaInformationElement_property (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_MetaInformationElement: root of subtree no. 1 can not be made a rootView_MetaInformationElement_property node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_MetaInformationElement_value) MkrootView_MetaInformationElement_value (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_MetaInformationElement: root of subtree no. 2 can not be made a rootView_MetaInformationElement_value node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_MetaInformationElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_MetaInformationElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_MetaInformationElement_property (POSITION *_coordref)
#else
NODEPTR MkprootView_MetaInformationElement_property (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_MetaInformationElement_property _currn;
#ifdef __cplusplus
_currn = new _TPprootView_MetaInformationElement_property;
#else
_currn = (_TPPprootView_MetaInformationElement_property) TreeNodeAlloc (sizeof (struct _TPprootView_MetaInformationElement_property));
#endif
_currn->_prod = RULEprootView_MetaInformationElement_property;
_SETCOORD(_currn)
_TERMACT_prootView_MetaInformationElement_property;
return ( (NODEPTR) _currn);
}/* MkprootView_MetaInformationElement_property */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_MetaInformationElement_value (POSITION *_coordref)
#else
NODEPTR MkprootView_MetaInformationElement_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_MetaInformationElement_value _currn;
#ifdef __cplusplus
_currn = new _TPprootView_MetaInformationElement_value;
#else
_currn = (_TPPprootView_MetaInformationElement_value) TreeNodeAlloc (sizeof (struct _TPprootView_MetaInformationElement_value));
#endif
_currn->_prod = RULEprootView_MetaInformationElement_value;
_SETCOORD(_currn)
_TERMACT_prootView_MetaInformationElement_value;
return ( (NODEPTR) _currn);
}/* MkprootView_MetaInformationElement_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_OutputElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkprootView_OutputElement (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPprootView_OutputElement _currn;
#ifdef __cplusplus
_currn = new _TPprootView_OutputElement;
#else
_currn = (_TPPprootView_OutputElement) TreeNodeAlloc (sizeof (struct _TPprootView_OutputElement));
#endif
_currn->_prod = RULEprootView_OutputElement;
_currn->_desc1 = (_TSProotView_OutputElement_link) MkrootView_OutputElement_link (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_OutputElement: root of subtree no. 1 can not be made a rootView_OutputElement_link node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_OutputElement_content) MkrootView_OutputElement_content (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_OutputElement: root of subtree no. 2 can not be made a rootView_OutputElement_content node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_OutputElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_OutputElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_OutputElement_link (POSITION *_coordref)
#else
NODEPTR MkprootView_OutputElement_link (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_OutputElement_link _currn;
#ifdef __cplusplus
_currn = new _TPprootView_OutputElement_link;
#else
_currn = (_TPPprootView_OutputElement_link) TreeNodeAlloc (sizeof (struct _TPprootView_OutputElement_link));
#endif
_currn->_prod = RULEprootView_OutputElement_link;
_SETCOORD(_currn)
_TERMACT_prootView_OutputElement_link;
return ( (NODEPTR) _currn);
}/* MkprootView_OutputElement_link */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_OutputElement_content (POSITION *_coordref)
#else
NODEPTR MkprootView_OutputElement_content (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_OutputElement_content _currn;
#ifdef __cplusplus
_currn = new _TPprootView_OutputElement_content;
#else
_currn = (_TPPprootView_OutputElement_content) TreeNodeAlloc (sizeof (struct _TPprootView_OutputElement_content));
#endif
_currn->_prod = RULEprootView_OutputElement_content;
_SETCOORD(_currn)
_TERMACT_prootView_OutputElement_content;
return ( (NODEPTR) _currn);
}/* MkprootView_OutputElement_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_ListOutputElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkprootView_ListOutputElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPprootView_ListOutputElement _currn;
#ifdef __cplusplus
_currn = new _TPprootView_ListOutputElement;
#else
_currn = (_TPPprootView_ListOutputElement) TreeNodeAlloc (sizeof (struct _TPprootView_ListOutputElement));
#endif
_currn->_prod = RULEprootView_ListOutputElement;
_currn->_desc1 = (_TSProotView_ListOutputElement_styleSheetBlock) MkrootView_ListOutputElement_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_ListOutputElement: root of subtree no. 1 can not be made a rootView_ListOutputElement_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_ListOutputElement_dataSourceActualParams) MkrootView_ListOutputElement_dataSourceActualParams (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_ListOutputElement: root of subtree no. 2 can not be made a rootView_ListOutputElement_dataSourceActualParams node ", 0, _coordref);
_currn->_desc3 = (_TSProotView_ListOutputElement_dataSource) MkrootView_ListOutputElement_dataSource (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_ListOutputElement: root of subtree no. 3 can not be made a rootView_ListOutputElement_dataSource node ", 0, _coordref);
_currn->_desc4 = (_TSProotView_ListOutputElement_maxNumberOfElements) MkrootView_ListOutputElement_maxNumberOfElements (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_ListOutputElement: root of subtree no. 4 can not be made a rootView_ListOutputElement_maxNumberOfElements node ", 0, _coordref);
_currn->_desc5 = (_TSProotView_ListOutputElement_navigationElement) MkrootView_ListOutputElement_navigationElement (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_ListOutputElement: root of subtree no. 5 can not be made a rootView_ListOutputElement_navigationElement node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_ListOutputElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_ListOutputElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_ListOutputElement_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkprootView_ListOutputElement_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_ListOutputElement_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPprootView_ListOutputElement_styleSheetBlock;
#else
_currn = (_TPPprootView_ListOutputElement_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPprootView_ListOutputElement_styleSheetBlock));
#endif
_currn->_prod = RULEprootView_ListOutputElement_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_prootView_ListOutputElement_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkprootView_ListOutputElement_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_DataSourceActualParamprootView_ListOutputElement_dataSourceActualParams (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_DataSourceActualParamprootView_ListOutputElement_dataSourceActualParams (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_DataSourceActualParamprootView_ListOutputElement_dataSourceActualParams _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_DataSourceActualParamprootView_ListOutputElement_dataSourceActualParams;
#else
	_new = (_TPPLST_rootView_DataSourceActualParamprootView_ListOutputElement_dataSourceActualParams) TreeNodeAlloc (sizeof (struct _TPLST_rootView_DataSourceActualParamprootView_ListOutputElement_dataSourceActualParams));
#endif
	_new->_prod = RULELST_rootView_DataSourceActualParamprootView_ListOutputElement_dataSourceActualParams;
	_new->_desc1 = (_TSProotView_DataSourceActualParam)elem;
	_new->_desc2 = (_TSPLST_rootView_ListOutputElement_dataSourceActualParams)_new;
	{_TPPLST_rootView_DataSourceActualParamprootView_ListOutputElement_dataSourceActualParams _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_DataSourceActualParamprootView_ListOutputElement_dataSourceActualParams;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_ListOutputElement_dataSourceActualParams (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_ListOutputElement_dataSourceActualParams (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_DataSourceActualParam)) 
		return _MkrootView_DataSourceActualParamprootView_ListOutputElement_dataSourceActualParams (_coordref, MkrootView_DataSourceActualParam (_coordref, t));
	message (DEADLY, "a prootView_ListOutputElement_dataSourceActualParams list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_ListOutputElement_dataSourceActualParams (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_ListOutputElement_dataSourceActualParams (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_ListOutputElement_dataSourceActualParams)
	l =_ElemprootView_ListOutputElement_dataSourceActualParams (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_ListOutputElement_dataSourceActualParams)
	r = _ElemprootView_ListOutputElement_dataSourceActualParams (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_ListOutputElement_dataSourceActualParams)r)->_desc2);
((_TPPLST_0prootView_ListOutputElement_dataSourceActualParams)r)->_desc2 = ((_TPPLST_0prootView_ListOutputElement_dataSourceActualParams)l)->_desc2;
((_TPPLST_0prootView_ListOutputElement_dataSourceActualParams)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_ListOutputElement_dataSourceActualParams (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_ListOutputElement_dataSourceActualParams (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_ListOutputElement_dataSourceActualParams root;
	_TPPLST_0prootView_ListOutputElement_dataSourceActualParams list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_ListOutputElement_dataSourceActualParams)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_ListOutputElement_dataSourceActualParams)
		t = _ElemprootView_ListOutputElement_dataSourceActualParams (_coordref, t);
	remain = (_TPPLST_0prootView_ListOutputElement_dataSourceActualParams)(((_TPPLST_0prootView_ListOutputElement_dataSourceActualParams) t)->_desc2);
	((_TPPLST_0prootView_ListOutputElement_dataSourceActualParams) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_ListOutputElement_dataSourceActualParams;
#else
list = (_TPPLST_0prootView_ListOutputElement_dataSourceActualParams) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_ListOutputElement_dataSourceActualParams));
#endif
list->_prod = RULELST_0prootView_ListOutputElement_dataSourceActualParams;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_ListOutputElement_dataSourceActualParams)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_DataSourceActualParamprootView_ListOutputElement_dataSourceActualParams)
	{_TPPLST_rootView_DataSourceActualParamprootView_ListOutputElement_dataSourceActualParams _currn = _currn;
	 _TERMACT_LST_rootView_DataSourceActualParamprootView_ListOutputElement_dataSourceActualParams;
	} this code has been moved to _ElemprootView_ListOutputElement_dataSourceActualParams*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_ListOutputElement_dataSourceActualParams;
#else
root = (_TPPprootView_ListOutputElement_dataSourceActualParams) TreeNodeAlloc (sizeof (struct _TPprootView_ListOutputElement_dataSourceActualParams));
#endif
root->_prod = RULEprootView_ListOutputElement_dataSourceActualParams;
root->_desc1 = (_TSPLST_rootView_ListOutputElement_dataSourceActualParams)list;
{ _TPPprootView_ListOutputElement_dataSourceActualParams _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_ListOutputElement_dataSourceActualParams;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_ListOutputElement_dataSource (POSITION *_coordref)
#else
NODEPTR MkprootView_ListOutputElement_dataSource (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_ListOutputElement_dataSource _currn;
#ifdef __cplusplus
_currn = new _TPprootView_ListOutputElement_dataSource;
#else
_currn = (_TPPprootView_ListOutputElement_dataSource) TreeNodeAlloc (sizeof (struct _TPprootView_ListOutputElement_dataSource));
#endif
_currn->_prod = RULEprootView_ListOutputElement_dataSource;
_SETCOORD(_currn)
_TERMACT_prootView_ListOutputElement_dataSource;
return ( (NODEPTR) _currn);
}/* MkprootView_ListOutputElement_dataSource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_ListOutputElement_maxNumberOfElements (POSITION *_coordref)
#else
NODEPTR MkprootView_ListOutputElement_maxNumberOfElements (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_ListOutputElement_maxNumberOfElements _currn;
#ifdef __cplusplus
_currn = new _TPprootView_ListOutputElement_maxNumberOfElements;
#else
_currn = (_TPPprootView_ListOutputElement_maxNumberOfElements) TreeNodeAlloc (sizeof (struct _TPprootView_ListOutputElement_maxNumberOfElements));
#endif
_currn->_prod = RULEprootView_ListOutputElement_maxNumberOfElements;
_SETCOORD(_currn)
_TERMACT_prootView_ListOutputElement_maxNumberOfElements;
return ( (NODEPTR) _currn);
}/* MkprootView_ListOutputElement_maxNumberOfElements */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_ListOutputElement_navigationElement (POSITION *_coordref)
#else
NODEPTR MkprootView_ListOutputElement_navigationElement (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_ListOutputElement_navigationElement _currn;
#ifdef __cplusplus
_currn = new _TPprootView_ListOutputElement_navigationElement;
#else
_currn = (_TPPprootView_ListOutputElement_navigationElement) TreeNodeAlloc (sizeof (struct _TPprootView_ListOutputElement_navigationElement));
#endif
_currn->_prod = RULEprootView_ListOutputElement_navigationElement;
_SETCOORD(_currn)
_TERMACT_prootView_ListOutputElement_navigationElement;
return ( (NODEPTR) _currn);
}/* MkprootView_ListOutputElement_navigationElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_StyleSheet (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkprootView_StyleSheet (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPprootView_StyleSheet _currn;
#ifdef __cplusplus
_currn = new _TPprootView_StyleSheet;
#else
_currn = (_TPPprootView_StyleSheet) TreeNodeAlloc (sizeof (struct _TPprootView_StyleSheet));
#endif
_currn->_prod = RULEprootView_StyleSheet;
_currn->_desc1 = (_TSProotView_StyleSheet_name) MkrootView_StyleSheet_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_StyleSheet: root of subtree no. 1 can not be made a rootView_StyleSheet_name node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_StyleSheet_custom) MkrootView_StyleSheet_custom (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_StyleSheet: root of subtree no. 2 can not be made a rootView_StyleSheet_custom node ", 0, _coordref);
_currn->_desc3 = (_TSProotView_StyleSheet_default) MkrootView_StyleSheet_default (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_StyleSheet: root of subtree no. 3 can not be made a rootView_StyleSheet_default node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_StyleSheet;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_StyleSheet */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_StyleSheet_name (POSITION *_coordref)
#else
NODEPTR MkprootView_StyleSheet_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_StyleSheet_name _currn;
#ifdef __cplusplus
_currn = new _TPprootView_StyleSheet_name;
#else
_currn = (_TPPprootView_StyleSheet_name) TreeNodeAlloc (sizeof (struct _TPprootView_StyleSheet_name));
#endif
_currn->_prod = RULEprootView_StyleSheet_name;
_SETCOORD(_currn)
_TERMACT_prootView_StyleSheet_name;
return ( (NODEPTR) _currn);
}/* MkprootView_StyleSheet_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_CustomStyleSheetBlockprootView_StyleSheet_custom (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_CustomStyleSheetBlockprootView_StyleSheet_custom (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_CustomStyleSheetBlockprootView_StyleSheet_custom _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_CustomStyleSheetBlockprootView_StyleSheet_custom;
#else
	_new = (_TPPLST_rootView_CustomStyleSheetBlockprootView_StyleSheet_custom) TreeNodeAlloc (sizeof (struct _TPLST_rootView_CustomStyleSheetBlockprootView_StyleSheet_custom));
#endif
	_new->_prod = RULELST_rootView_CustomStyleSheetBlockprootView_StyleSheet_custom;
	_new->_desc1 = (_TSProotView_CustomStyleSheetBlock)elem;
	_new->_desc2 = (_TSPLST_rootView_StyleSheet_custom)_new;
	{_TPPLST_rootView_CustomStyleSheetBlockprootView_StyleSheet_custom _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_CustomStyleSheetBlockprootView_StyleSheet_custom;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_StyleSheet_custom (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_StyleSheet_custom (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_CustomStyleSheetBlock)) 
		return _MkrootView_CustomStyleSheetBlockprootView_StyleSheet_custom (_coordref, MkrootView_CustomStyleSheetBlock (_coordref, t));
	message (DEADLY, "a prootView_StyleSheet_custom list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_StyleSheet_custom (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_StyleSheet_custom (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_StyleSheet_custom)
	l =_ElemprootView_StyleSheet_custom (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_StyleSheet_custom)
	r = _ElemprootView_StyleSheet_custom (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_StyleSheet_custom)r)->_desc2);
((_TPPLST_0prootView_StyleSheet_custom)r)->_desc2 = ((_TPPLST_0prootView_StyleSheet_custom)l)->_desc2;
((_TPPLST_0prootView_StyleSheet_custom)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_StyleSheet_custom (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_StyleSheet_custom (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_StyleSheet_custom root;
	_TPPLST_0prootView_StyleSheet_custom list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_StyleSheet_custom)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_StyleSheet_custom)
		t = _ElemprootView_StyleSheet_custom (_coordref, t);
	remain = (_TPPLST_0prootView_StyleSheet_custom)(((_TPPLST_0prootView_StyleSheet_custom) t)->_desc2);
	((_TPPLST_0prootView_StyleSheet_custom) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_StyleSheet_custom;
#else
list = (_TPPLST_0prootView_StyleSheet_custom) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_StyleSheet_custom));
#endif
list->_prod = RULELST_0prootView_StyleSheet_custom;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_StyleSheet_custom)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_CustomStyleSheetBlockprootView_StyleSheet_custom)
	{_TPPLST_rootView_CustomStyleSheetBlockprootView_StyleSheet_custom _currn = _currn;
	 _TERMACT_LST_rootView_CustomStyleSheetBlockprootView_StyleSheet_custom;
	} this code has been moved to _ElemprootView_StyleSheet_custom*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_StyleSheet_custom;
#else
root = (_TPPprootView_StyleSheet_custom) TreeNodeAlloc (sizeof (struct _TPprootView_StyleSheet_custom));
#endif
root->_prod = RULEprootView_StyleSheet_custom;
root->_desc1 = (_TSPLST_rootView_StyleSheet_custom)list;
{ _TPPprootView_StyleSheet_custom _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_StyleSheet_custom;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_HtmlStyleSheetBlockprootView_StyleSheet_default (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_HtmlStyleSheetBlockprootView_StyleSheet_default (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_HtmlStyleSheetBlockprootView_StyleSheet_default _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_HtmlStyleSheetBlockprootView_StyleSheet_default;
#else
	_new = (_TPPLST_rootView_HtmlStyleSheetBlockprootView_StyleSheet_default) TreeNodeAlloc (sizeof (struct _TPLST_rootView_HtmlStyleSheetBlockprootView_StyleSheet_default));
#endif
	_new->_prod = RULELST_rootView_HtmlStyleSheetBlockprootView_StyleSheet_default;
	_new->_desc1 = (_TSProotView_HtmlStyleSheetBlock)elem;
	_new->_desc2 = (_TSPLST_rootView_StyleSheet_default)_new;
	{_TPPLST_rootView_HtmlStyleSheetBlockprootView_StyleSheet_default _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_HtmlStyleSheetBlockprootView_StyleSheet_default;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_StyleSheet_default (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_StyleSheet_default (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_HtmlStyleSheetBlock)) 
		return _MkrootView_HtmlStyleSheetBlockprootView_StyleSheet_default (_coordref, MkrootView_HtmlStyleSheetBlock (_coordref, t));
	message (DEADLY, "a prootView_StyleSheet_default list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_StyleSheet_default (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_StyleSheet_default (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_StyleSheet_default)
	l =_ElemprootView_StyleSheet_default (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_StyleSheet_default)
	r = _ElemprootView_StyleSheet_default (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_StyleSheet_default)r)->_desc2);
((_TPPLST_0prootView_StyleSheet_default)r)->_desc2 = ((_TPPLST_0prootView_StyleSheet_default)l)->_desc2;
((_TPPLST_0prootView_StyleSheet_default)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_StyleSheet_default (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_StyleSheet_default (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_StyleSheet_default root;
	_TPPLST_0prootView_StyleSheet_default list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_StyleSheet_default)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_StyleSheet_default)
		t = _ElemprootView_StyleSheet_default (_coordref, t);
	remain = (_TPPLST_0prootView_StyleSheet_default)(((_TPPLST_0prootView_StyleSheet_default) t)->_desc2);
	((_TPPLST_0prootView_StyleSheet_default) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_StyleSheet_default;
#else
list = (_TPPLST_0prootView_StyleSheet_default) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_StyleSheet_default));
#endif
list->_prod = RULELST_0prootView_StyleSheet_default;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_StyleSheet_default)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_HtmlStyleSheetBlockprootView_StyleSheet_default)
	{_TPPLST_rootView_HtmlStyleSheetBlockprootView_StyleSheet_default _currn = _currn;
	 _TERMACT_LST_rootView_HtmlStyleSheetBlockprootView_StyleSheet_default;
	} this code has been moved to _ElemprootView_StyleSheet_default*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_StyleSheet_default;
#else
root = (_TPPprootView_StyleSheet_default) TreeNodeAlloc (sizeof (struct _TPprootView_StyleSheet_default));
#endif
root->_prod = RULEprootView_StyleSheet_default;
root->_desc1 = (_TSPLST_rootView_StyleSheet_default)list;
{ _TPPprootView_StyleSheet_default _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_StyleSheet_default;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_CustomStyleSheetBlock (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkprootView_CustomStyleSheetBlock (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPprootView_CustomStyleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPprootView_CustomStyleSheetBlock;
#else
_currn = (_TPPprootView_CustomStyleSheetBlock) TreeNodeAlloc (sizeof (struct _TPprootView_CustomStyleSheetBlock));
#endif
_currn->_prod = RULEprootView_CustomStyleSheetBlock;
_currn->_desc1 = (_TSProotView_CustomStyleSheetBlock_pairs) MkrootView_CustomStyleSheetBlock_pairs (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_CustomStyleSheetBlock: root of subtree no. 1 can not be made a rootView_CustomStyleSheetBlock_pairs node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_CustomStyleSheetBlock_name) MkrootView_CustomStyleSheetBlock_name (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_CustomStyleSheetBlock: root of subtree no. 2 can not be made a rootView_CustomStyleSheetBlock_name node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_CustomStyleSheetBlock;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_CustomStyleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_StyleSheetPairprootView_CustomStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_StyleSheetPairprootView_CustomStyleSheetBlock_pairs (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_StyleSheetPairprootView_CustomStyleSheetBlock_pairs _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_StyleSheetPairprootView_CustomStyleSheetBlock_pairs;
#else
	_new = (_TPPLST_rootView_StyleSheetPairprootView_CustomStyleSheetBlock_pairs) TreeNodeAlloc (sizeof (struct _TPLST_rootView_StyleSheetPairprootView_CustomStyleSheetBlock_pairs));
#endif
	_new->_prod = RULELST_rootView_StyleSheetPairprootView_CustomStyleSheetBlock_pairs;
	_new->_desc1 = (_TSProotView_StyleSheetPair)elem;
	_new->_desc2 = (_TSPLST_rootView_CustomStyleSheetBlock_pairs)_new;
	{_TPPLST_rootView_StyleSheetPairprootView_CustomStyleSheetBlock_pairs _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_StyleSheetPairprootView_CustomStyleSheetBlock_pairs;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_CustomStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_CustomStyleSheetBlock_pairs (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_StyleSheetPair)) 
		return _MkrootView_StyleSheetPairprootView_CustomStyleSheetBlock_pairs (_coordref, MkrootView_StyleSheetPair (_coordref, t));
	message (DEADLY, "a prootView_CustomStyleSheetBlock_pairs list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_CustomStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_CustomStyleSheetBlock_pairs (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_CustomStyleSheetBlock_pairs)
	l =_ElemprootView_CustomStyleSheetBlock_pairs (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_CustomStyleSheetBlock_pairs)
	r = _ElemprootView_CustomStyleSheetBlock_pairs (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_CustomStyleSheetBlock_pairs)r)->_desc2);
((_TPPLST_0prootView_CustomStyleSheetBlock_pairs)r)->_desc2 = ((_TPPLST_0prootView_CustomStyleSheetBlock_pairs)l)->_desc2;
((_TPPLST_0prootView_CustomStyleSheetBlock_pairs)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_CustomStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_CustomStyleSheetBlock_pairs (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_CustomStyleSheetBlock_pairs root;
	_TPPLST_0prootView_CustomStyleSheetBlock_pairs list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_CustomStyleSheetBlock_pairs)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_CustomStyleSheetBlock_pairs)
		t = _ElemprootView_CustomStyleSheetBlock_pairs (_coordref, t);
	remain = (_TPPLST_0prootView_CustomStyleSheetBlock_pairs)(((_TPPLST_0prootView_CustomStyleSheetBlock_pairs) t)->_desc2);
	((_TPPLST_0prootView_CustomStyleSheetBlock_pairs) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_CustomStyleSheetBlock_pairs;
#else
list = (_TPPLST_0prootView_CustomStyleSheetBlock_pairs) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_CustomStyleSheetBlock_pairs));
#endif
list->_prod = RULELST_0prootView_CustomStyleSheetBlock_pairs;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_CustomStyleSheetBlock_pairs)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_StyleSheetPairprootView_CustomStyleSheetBlock_pairs)
	{_TPPLST_rootView_StyleSheetPairprootView_CustomStyleSheetBlock_pairs _currn = _currn;
	 _TERMACT_LST_rootView_StyleSheetPairprootView_CustomStyleSheetBlock_pairs;
	} this code has been moved to _ElemprootView_CustomStyleSheetBlock_pairs*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_CustomStyleSheetBlock_pairs;
#else
root = (_TPPprootView_CustomStyleSheetBlock_pairs) TreeNodeAlloc (sizeof (struct _TPprootView_CustomStyleSheetBlock_pairs));
#endif
root->_prod = RULEprootView_CustomStyleSheetBlock_pairs;
root->_desc1 = (_TSPLST_rootView_CustomStyleSheetBlock_pairs)list;
{ _TPPprootView_CustomStyleSheetBlock_pairs _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_CustomStyleSheetBlock_pairs;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_CustomStyleSheetBlock_name (POSITION *_coordref)
#else
NODEPTR MkprootView_CustomStyleSheetBlock_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_CustomStyleSheetBlock_name _currn;
#ifdef __cplusplus
_currn = new _TPprootView_CustomStyleSheetBlock_name;
#else
_currn = (_TPPprootView_CustomStyleSheetBlock_name) TreeNodeAlloc (sizeof (struct _TPprootView_CustomStyleSheetBlock_name));
#endif
_currn->_prod = RULEprootView_CustomStyleSheetBlock_name;
_SETCOORD(_currn)
_TERMACT_prootView_CustomStyleSheetBlock_name;
return ( (NODEPTR) _currn);
}/* MkprootView_CustomStyleSheetBlock_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_StyleSheetPair (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4)
#else
NODEPTR MkprootView_StyleSheetPair (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
#endif
{	_TPPprootView_StyleSheetPair _currn;
#ifdef __cplusplus
_currn = new _TPprootView_StyleSheetPair;
#else
_currn = (_TPPprootView_StyleSheetPair) TreeNodeAlloc (sizeof (struct _TPprootView_StyleSheetPair));
#endif
_currn->_prod = RULEprootView_StyleSheetPair;
_currn->_desc1 = (_TSProotView_StyleSheetPair_category) MkrootView_StyleSheetPair_category (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_StyleSheetPair: root of subtree no. 1 can not be made a rootView_StyleSheetPair_category node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_StyleSheetPair_attribute) MkrootView_StyleSheetPair_attribute (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_StyleSheetPair: root of subtree no. 2 can not be made a rootView_StyleSheetPair_attribute node ", 0, _coordref);
_currn->_desc3 = (_TSProotView_StyleSheetPair_valueCategory) MkrootView_StyleSheetPair_valueCategory (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_StyleSheetPair: root of subtree no. 3 can not be made a rootView_StyleSheetPair_valueCategory node ", 0, _coordref);
_currn->_desc4 = (_TSProotView_StyleSheetPair_value) MkrootView_StyleSheetPair_value (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_StyleSheetPair: root of subtree no. 4 can not be made a rootView_StyleSheetPair_value node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_StyleSheetPair;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_StyleSheetPair */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_StyleSheetPair_category (POSITION *_coordref)
#else
NODEPTR MkprootView_StyleSheetPair_category (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_StyleSheetPair_category _currn;
#ifdef __cplusplus
_currn = new _TPprootView_StyleSheetPair_category;
#else
_currn = (_TPPprootView_StyleSheetPair_category) TreeNodeAlloc (sizeof (struct _TPprootView_StyleSheetPair_category));
#endif
_currn->_prod = RULEprootView_StyleSheetPair_category;
_SETCOORD(_currn)
_TERMACT_prootView_StyleSheetPair_category;
return ( (NODEPTR) _currn);
}/* MkprootView_StyleSheetPair_category */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_StyleSheetPair_attribute (POSITION *_coordref)
#else
NODEPTR MkprootView_StyleSheetPair_attribute (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_StyleSheetPair_attribute _currn;
#ifdef __cplusplus
_currn = new _TPprootView_StyleSheetPair_attribute;
#else
_currn = (_TPPprootView_StyleSheetPair_attribute) TreeNodeAlloc (sizeof (struct _TPprootView_StyleSheetPair_attribute));
#endif
_currn->_prod = RULEprootView_StyleSheetPair_attribute;
_SETCOORD(_currn)
_TERMACT_prootView_StyleSheetPair_attribute;
return ( (NODEPTR) _currn);
}/* MkprootView_StyleSheetPair_attribute */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_StyleSheetPair_valueCategory (POSITION *_coordref)
#else
NODEPTR MkprootView_StyleSheetPair_valueCategory (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_StyleSheetPair_valueCategory _currn;
#ifdef __cplusplus
_currn = new _TPprootView_StyleSheetPair_valueCategory;
#else
_currn = (_TPPprootView_StyleSheetPair_valueCategory) TreeNodeAlloc (sizeof (struct _TPprootView_StyleSheetPair_valueCategory));
#endif
_currn->_prod = RULEprootView_StyleSheetPair_valueCategory;
_SETCOORD(_currn)
_TERMACT_prootView_StyleSheetPair_valueCategory;
return ( (NODEPTR) _currn);
}/* MkprootView_StyleSheetPair_valueCategory */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_SizeValueprootView_StyleSheetPair_value (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_SizeValueprootView_StyleSheetPair_value (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_SizeValueprootView_StyleSheetPair_value _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_SizeValueprootView_StyleSheetPair_value;
#else
	_new = (_TPPLST_rootView_SizeValueprootView_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_rootView_SizeValueprootView_StyleSheetPair_value));
#endif
	_new->_prod = RULELST_rootView_SizeValueprootView_StyleSheetPair_value;
	_new->_desc1 = (_TSProotView_SizeValue)elem;
	_new->_desc2 = (_TSPLST_rootView_StyleSheetPair_value)_new;
	{_TPPLST_rootView_SizeValueprootView_StyleSheetPair_value _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_SizeValueprootView_StyleSheetPair_value;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_ColorValueprootView_StyleSheetPair_value (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_ColorValueprootView_StyleSheetPair_value (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_ColorValueprootView_StyleSheetPair_value _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_ColorValueprootView_StyleSheetPair_value;
#else
	_new = (_TPPLST_rootView_ColorValueprootView_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_rootView_ColorValueprootView_StyleSheetPair_value));
#endif
	_new->_prod = RULELST_rootView_ColorValueprootView_StyleSheetPair_value;
	_new->_desc1 = (_TSProotView_ColorValue)elem;
	_new->_desc2 = (_TSPLST_rootView_StyleSheetPair_value)_new;
	{_TPPLST_rootView_ColorValueprootView_StyleSheetPair_value _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_ColorValueprootView_StyleSheetPair_value;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_FamilyValueprootView_StyleSheetPair_value (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_FamilyValueprootView_StyleSheetPair_value (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_FamilyValueprootView_StyleSheetPair_value _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_FamilyValueprootView_StyleSheetPair_value;
#else
	_new = (_TPPLST_rootView_FamilyValueprootView_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_rootView_FamilyValueprootView_StyleSheetPair_value));
#endif
	_new->_prod = RULELST_rootView_FamilyValueprootView_StyleSheetPair_value;
	_new->_desc1 = (_TSProotView_FamilyValue)elem;
	_new->_desc2 = (_TSPLST_rootView_StyleSheetPair_value)_new;
	{_TPPLST_rootView_FamilyValueprootView_StyleSheetPair_value _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_FamilyValueprootView_StyleSheetPair_value;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_StyleValueprootView_StyleSheetPair_value (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_StyleValueprootView_StyleSheetPair_value (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_StyleValueprootView_StyleSheetPair_value _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_StyleValueprootView_StyleSheetPair_value;
#else
	_new = (_TPPLST_rootView_StyleValueprootView_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_rootView_StyleValueprootView_StyleSheetPair_value));
#endif
	_new->_prod = RULELST_rootView_StyleValueprootView_StyleSheetPair_value;
	_new->_desc1 = (_TSProotView_StyleValue)elem;
	_new->_desc2 = (_TSPLST_rootView_StyleSheetPair_value)_new;
	{_TPPLST_rootView_StyleValueprootView_StyleSheetPair_value _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_StyleValueprootView_StyleSheetPair_value;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_HeadingValueprootView_StyleSheetPair_value (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_HeadingValueprootView_StyleSheetPair_value (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_HeadingValueprootView_StyleSheetPair_value _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_HeadingValueprootView_StyleSheetPair_value;
#else
	_new = (_TPPLST_rootView_HeadingValueprootView_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_rootView_HeadingValueprootView_StyleSheetPair_value));
#endif
	_new->_prod = RULELST_rootView_HeadingValueprootView_StyleSheetPair_value;
	_new->_desc1 = (_TSProotView_HeadingValue)elem;
	_new->_desc2 = (_TSPLST_rootView_StyleSheetPair_value)_new;
	{_TPPLST_rootView_HeadingValueprootView_StyleSheetPair_value _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_HeadingValueprootView_StyleSheetPair_value;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_AlignmentValueprootView_StyleSheetPair_value (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_AlignmentValueprootView_StyleSheetPair_value (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_AlignmentValueprootView_StyleSheetPair_value _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_AlignmentValueprootView_StyleSheetPair_value;
#else
	_new = (_TPPLST_rootView_AlignmentValueprootView_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_rootView_AlignmentValueprootView_StyleSheetPair_value));
#endif
	_new->_prod = RULELST_rootView_AlignmentValueprootView_StyleSheetPair_value;
	_new->_desc1 = (_TSProotView_AlignmentValue)elem;
	_new->_desc2 = (_TSPLST_rootView_StyleSheetPair_value)_new;
	{_TPPLST_rootView_AlignmentValueprootView_StyleSheetPair_value _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_AlignmentValueprootView_StyleSheetPair_value;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_ListValueprootView_StyleSheetPair_value (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_ListValueprootView_StyleSheetPair_value (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_ListValueprootView_StyleSheetPair_value _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_ListValueprootView_StyleSheetPair_value;
#else
	_new = (_TPPLST_rootView_ListValueprootView_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_rootView_ListValueprootView_StyleSheetPair_value));
#endif
	_new->_prod = RULELST_rootView_ListValueprootView_StyleSheetPair_value;
	_new->_desc1 = (_TSProotView_ListValue)elem;
	_new->_desc2 = (_TSPLST_rootView_StyleSheetPair_value)_new;
	{_TPPLST_rootView_ListValueprootView_StyleSheetPair_value _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_ListValueprootView_StyleSheetPair_value;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_CustomValueprootView_StyleSheetPair_value (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_CustomValueprootView_StyleSheetPair_value (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_CustomValueprootView_StyleSheetPair_value _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_CustomValueprootView_StyleSheetPair_value;
#else
	_new = (_TPPLST_rootView_CustomValueprootView_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_rootView_CustomValueprootView_StyleSheetPair_value));
#endif
	_new->_prod = RULELST_rootView_CustomValueprootView_StyleSheetPair_value;
	_new->_desc1 = (_TSProotView_CustomValue)elem;
	_new->_desc2 = (_TSPLST_rootView_StyleSheetPair_value)_new;
	{_TPPLST_rootView_CustomValueprootView_StyleSheetPair_value _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_CustomValueprootView_StyleSheetPair_value;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_StyleSheetPair_value (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_StyleSheetPair_value (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_SizeValue)) 
		return _MkrootView_SizeValueprootView_StyleSheetPair_value (_coordref, MkrootView_SizeValue (_coordref, t));
	if (IsSymb (t, SYMBrootView_ColorValue)) 
		return _MkrootView_ColorValueprootView_StyleSheetPair_value (_coordref, MkrootView_ColorValue (_coordref, t));
	if (IsSymb (t, SYMBrootView_FamilyValue)) 
		return _MkrootView_FamilyValueprootView_StyleSheetPair_value (_coordref, MkrootView_FamilyValue (_coordref, t));
	if (IsSymb (t, SYMBrootView_StyleValue)) 
		return _MkrootView_StyleValueprootView_StyleSheetPair_value (_coordref, MkrootView_StyleValue (_coordref, t));
	if (IsSymb (t, SYMBrootView_HeadingValue)) 
		return _MkrootView_HeadingValueprootView_StyleSheetPair_value (_coordref, MkrootView_HeadingValue (_coordref, t));
	if (IsSymb (t, SYMBrootView_AlignmentValue)) 
		return _MkrootView_AlignmentValueprootView_StyleSheetPair_value (_coordref, MkrootView_AlignmentValue (_coordref, t));
	if (IsSymb (t, SYMBrootView_ListValue)) 
		return _MkrootView_ListValueprootView_StyleSheetPair_value (_coordref, MkrootView_ListValue (_coordref, t));
	if (IsSymb (t, SYMBrootView_CustomValue)) 
		return _MkrootView_CustomValueprootView_StyleSheetPair_value (_coordref, MkrootView_CustomValue (_coordref, t));
	message (DEADLY, "a prootView_StyleSheetPair_value list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_StyleSheetPair_value (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_StyleSheetPair_value (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_StyleSheetPair_value)
	l =_ElemprootView_StyleSheetPair_value (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_StyleSheetPair_value)
	r = _ElemprootView_StyleSheetPair_value (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_StyleSheetPair_value)r)->_desc2);
((_TPPLST_0prootView_StyleSheetPair_value)r)->_desc2 = ((_TPPLST_0prootView_StyleSheetPair_value)l)->_desc2;
((_TPPLST_0prootView_StyleSheetPair_value)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_StyleSheetPair_value (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_StyleSheetPair_value (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_StyleSheetPair_value root;
	_TPPLST_0prootView_StyleSheetPair_value list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_StyleSheetPair_value)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_StyleSheetPair_value)
		t = _ElemprootView_StyleSheetPair_value (_coordref, t);
	remain = (_TPPLST_0prootView_StyleSheetPair_value)(((_TPPLST_0prootView_StyleSheetPair_value) t)->_desc2);
	((_TPPLST_0prootView_StyleSheetPair_value) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_StyleSheetPair_value;
#else
list = (_TPPLST_0prootView_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_StyleSheetPair_value));
#endif
list->_prod = RULELST_0prootView_StyleSheetPair_value;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_StyleSheetPair_value)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_SizeValueprootView_StyleSheetPair_value)
	{_TPPLST_rootView_SizeValueprootView_StyleSheetPair_value _currn = _currn;
	 _TERMACT_LST_rootView_SizeValueprootView_StyleSheetPair_value;
	} this code has been moved to _ElemprootView_StyleSheetPair_value*/
	/*if (_currn->_prod == RULELST_rootView_ColorValueprootView_StyleSheetPair_value)
	{_TPPLST_rootView_ColorValueprootView_StyleSheetPair_value _currn = _currn;
	 _TERMACT_LST_rootView_ColorValueprootView_StyleSheetPair_value;
	} this code has been moved to _ElemprootView_StyleSheetPair_value*/
	/*if (_currn->_prod == RULELST_rootView_FamilyValueprootView_StyleSheetPair_value)
	{_TPPLST_rootView_FamilyValueprootView_StyleSheetPair_value _currn = _currn;
	 _TERMACT_LST_rootView_FamilyValueprootView_StyleSheetPair_value;
	} this code has been moved to _ElemprootView_StyleSheetPair_value*/
	/*if (_currn->_prod == RULELST_rootView_StyleValueprootView_StyleSheetPair_value)
	{_TPPLST_rootView_StyleValueprootView_StyleSheetPair_value _currn = _currn;
	 _TERMACT_LST_rootView_StyleValueprootView_StyleSheetPair_value;
	} this code has been moved to _ElemprootView_StyleSheetPair_value*/
	/*if (_currn->_prod == RULELST_rootView_HeadingValueprootView_StyleSheetPair_value)
	{_TPPLST_rootView_HeadingValueprootView_StyleSheetPair_value _currn = _currn;
	 _TERMACT_LST_rootView_HeadingValueprootView_StyleSheetPair_value;
	} this code has been moved to _ElemprootView_StyleSheetPair_value*/
	/*if (_currn->_prod == RULELST_rootView_AlignmentValueprootView_StyleSheetPair_value)
	{_TPPLST_rootView_AlignmentValueprootView_StyleSheetPair_value _currn = _currn;
	 _TERMACT_LST_rootView_AlignmentValueprootView_StyleSheetPair_value;
	} this code has been moved to _ElemprootView_StyleSheetPair_value*/
	/*if (_currn->_prod == RULELST_rootView_ListValueprootView_StyleSheetPair_value)
	{_TPPLST_rootView_ListValueprootView_StyleSheetPair_value _currn = _currn;
	 _TERMACT_LST_rootView_ListValueprootView_StyleSheetPair_value;
	} this code has been moved to _ElemprootView_StyleSheetPair_value*/
	/*if (_currn->_prod == RULELST_rootView_CustomValueprootView_StyleSheetPair_value)
	{_TPPLST_rootView_CustomValueprootView_StyleSheetPair_value _currn = _currn;
	 _TERMACT_LST_rootView_CustomValueprootView_StyleSheetPair_value;
	} this code has been moved to _ElemprootView_StyleSheetPair_value*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_StyleSheetPair_value;
#else
root = (_TPPprootView_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPprootView_StyleSheetPair_value));
#endif
root->_prod = RULEprootView_StyleSheetPair_value;
root->_desc1 = (_TSPLST_rootView_StyleSheetPair_value)list;
{ _TPPprootView_StyleSheetPair_value _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_StyleSheetPair_value;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_SizeValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkprootView_SizeValue (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPprootView_SizeValue _currn;
#ifdef __cplusplus
_currn = new _TPprootView_SizeValue;
#else
_currn = (_TPPprootView_SizeValue) TreeNodeAlloc (sizeof (struct _TPprootView_SizeValue));
#endif
_currn->_prod = RULEprootView_SizeValue;
_currn->_desc1 = (_TSProotView_SizeValue_measure) MkrootView_SizeValue_measure (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_SizeValue: root of subtree no. 1 can not be made a rootView_SizeValue_measure node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_SizeValue_value) MkrootView_SizeValue_value (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_SizeValue: root of subtree no. 2 can not be made a rootView_SizeValue_value node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_SizeValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_SizeValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_SizeValue_measure (POSITION *_coordref)
#else
NODEPTR MkprootView_SizeValue_measure (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_SizeValue_measure _currn;
#ifdef __cplusplus
_currn = new _TPprootView_SizeValue_measure;
#else
_currn = (_TPPprootView_SizeValue_measure) TreeNodeAlloc (sizeof (struct _TPprootView_SizeValue_measure));
#endif
_currn->_prod = RULEprootView_SizeValue_measure;
_SETCOORD(_currn)
_TERMACT_prootView_SizeValue_measure;
return ( (NODEPTR) _currn);
}/* MkprootView_SizeValue_measure */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_SizeValue_value (POSITION *_coordref)
#else
NODEPTR MkprootView_SizeValue_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_SizeValue_value _currn;
#ifdef __cplusplus
_currn = new _TPprootView_SizeValue_value;
#else
_currn = (_TPPprootView_SizeValue_value) TreeNodeAlloc (sizeof (struct _TPprootView_SizeValue_value));
#endif
_currn->_prod = RULEprootView_SizeValue_value;
_SETCOORD(_currn)
_TERMACT_prootView_SizeValue_value;
return ( (NODEPTR) _currn);
}/* MkprootView_SizeValue_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_ColorValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkprootView_ColorValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPprootView_ColorValue _currn;
#ifdef __cplusplus
_currn = new _TPprootView_ColorValue;
#else
_currn = (_TPPprootView_ColorValue) TreeNodeAlloc (sizeof (struct _TPprootView_ColorValue));
#endif
_currn->_prod = RULEprootView_ColorValue;
_currn->_desc1 = (_TSProotView_ColorValue_color) MkrootView_ColorValue_color (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_ColorValue: root of subtree no. 1 can not be made a rootView_ColorValue_color node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_ColorValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_ColorValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_ColorValue_color (POSITION *_coordref)
#else
NODEPTR MkprootView_ColorValue_color (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_ColorValue_color _currn;
#ifdef __cplusplus
_currn = new _TPprootView_ColorValue_color;
#else
_currn = (_TPPprootView_ColorValue_color) TreeNodeAlloc (sizeof (struct _TPprootView_ColorValue_color));
#endif
_currn->_prod = RULEprootView_ColorValue_color;
_SETCOORD(_currn)
_TERMACT_prootView_ColorValue_color;
return ( (NODEPTR) _currn);
}/* MkprootView_ColorValue_color */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FamilyValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkprootView_FamilyValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPprootView_FamilyValue _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FamilyValue;
#else
_currn = (_TPPprootView_FamilyValue) TreeNodeAlloc (sizeof (struct _TPprootView_FamilyValue));
#endif
_currn->_prod = RULEprootView_FamilyValue;
_currn->_desc1 = (_TSProotView_FamilyValue_family) MkrootView_FamilyValue_family (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_FamilyValue: root of subtree no. 1 can not be made a rootView_FamilyValue_family node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_FamilyValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_FamilyValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_FamilyValue_family (POSITION *_coordref)
#else
NODEPTR MkprootView_FamilyValue_family (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_FamilyValue_family _currn;
#ifdef __cplusplus
_currn = new _TPprootView_FamilyValue_family;
#else
_currn = (_TPPprootView_FamilyValue_family) TreeNodeAlloc (sizeof (struct _TPprootView_FamilyValue_family));
#endif
_currn->_prod = RULEprootView_FamilyValue_family;
_SETCOORD(_currn)
_TERMACT_prootView_FamilyValue_family;
return ( (NODEPTR) _currn);
}/* MkprootView_FamilyValue_family */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_StyleValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkprootView_StyleValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPprootView_StyleValue _currn;
#ifdef __cplusplus
_currn = new _TPprootView_StyleValue;
#else
_currn = (_TPPprootView_StyleValue) TreeNodeAlloc (sizeof (struct _TPprootView_StyleValue));
#endif
_currn->_prod = RULEprootView_StyleValue;
_currn->_desc1 = (_TSProotView_StyleValue_style) MkrootView_StyleValue_style (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_StyleValue: root of subtree no. 1 can not be made a rootView_StyleValue_style node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_StyleValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_StyleValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_StyleValue_style (POSITION *_coordref)
#else
NODEPTR MkprootView_StyleValue_style (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_StyleValue_style _currn;
#ifdef __cplusplus
_currn = new _TPprootView_StyleValue_style;
#else
_currn = (_TPPprootView_StyleValue_style) TreeNodeAlloc (sizeof (struct _TPprootView_StyleValue_style));
#endif
_currn->_prod = RULEprootView_StyleValue_style;
_SETCOORD(_currn)
_TERMACT_prootView_StyleValue_style;
return ( (NODEPTR) _currn);
}/* MkprootView_StyleValue_style */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_HeadingValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkprootView_HeadingValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPprootView_HeadingValue _currn;
#ifdef __cplusplus
_currn = new _TPprootView_HeadingValue;
#else
_currn = (_TPPprootView_HeadingValue) TreeNodeAlloc (sizeof (struct _TPprootView_HeadingValue));
#endif
_currn->_prod = RULEprootView_HeadingValue;
_currn->_desc1 = (_TSProotView_HeadingValue_heading) MkrootView_HeadingValue_heading (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_HeadingValue: root of subtree no. 1 can not be made a rootView_HeadingValue_heading node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_HeadingValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_HeadingValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_HeadingValue_heading (POSITION *_coordref)
#else
NODEPTR MkprootView_HeadingValue_heading (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_HeadingValue_heading _currn;
#ifdef __cplusplus
_currn = new _TPprootView_HeadingValue_heading;
#else
_currn = (_TPPprootView_HeadingValue_heading) TreeNodeAlloc (sizeof (struct _TPprootView_HeadingValue_heading));
#endif
_currn->_prod = RULEprootView_HeadingValue_heading;
_SETCOORD(_currn)
_TERMACT_prootView_HeadingValue_heading;
return ( (NODEPTR) _currn);
}/* MkprootView_HeadingValue_heading */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_AlignmentValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkprootView_AlignmentValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPprootView_AlignmentValue _currn;
#ifdef __cplusplus
_currn = new _TPprootView_AlignmentValue;
#else
_currn = (_TPPprootView_AlignmentValue) TreeNodeAlloc (sizeof (struct _TPprootView_AlignmentValue));
#endif
_currn->_prod = RULEprootView_AlignmentValue;
_currn->_desc1 = (_TSProotView_AlignmentValue_alignment) MkrootView_AlignmentValue_alignment (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_AlignmentValue: root of subtree no. 1 can not be made a rootView_AlignmentValue_alignment node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_AlignmentValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_AlignmentValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_AlignmentValue_alignment (POSITION *_coordref)
#else
NODEPTR MkprootView_AlignmentValue_alignment (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_AlignmentValue_alignment _currn;
#ifdef __cplusplus
_currn = new _TPprootView_AlignmentValue_alignment;
#else
_currn = (_TPPprootView_AlignmentValue_alignment) TreeNodeAlloc (sizeof (struct _TPprootView_AlignmentValue_alignment));
#endif
_currn->_prod = RULEprootView_AlignmentValue_alignment;
_SETCOORD(_currn)
_TERMACT_prootView_AlignmentValue_alignment;
return ( (NODEPTR) _currn);
}/* MkprootView_AlignmentValue_alignment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_ListValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkprootView_ListValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPprootView_ListValue _currn;
#ifdef __cplusplus
_currn = new _TPprootView_ListValue;
#else
_currn = (_TPPprootView_ListValue) TreeNodeAlloc (sizeof (struct _TPprootView_ListValue));
#endif
_currn->_prod = RULEprootView_ListValue;
_currn->_desc1 = (_TSProotView_ListValue_listed) MkrootView_ListValue_listed (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_ListValue: root of subtree no. 1 can not be made a rootView_ListValue_listed node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_ListValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_ListValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_ListValue_listed (POSITION *_coordref)
#else
NODEPTR MkprootView_ListValue_listed (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_ListValue_listed _currn;
#ifdef __cplusplus
_currn = new _TPprootView_ListValue_listed;
#else
_currn = (_TPPprootView_ListValue_listed) TreeNodeAlloc (sizeof (struct _TPprootView_ListValue_listed));
#endif
_currn->_prod = RULEprootView_ListValue_listed;
_SETCOORD(_currn)
_TERMACT_prootView_ListValue_listed;
return ( (NODEPTR) _currn);
}/* MkprootView_ListValue_listed */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_CustomValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkprootView_CustomValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPprootView_CustomValue _currn;
#ifdef __cplusplus
_currn = new _TPprootView_CustomValue;
#else
_currn = (_TPPprootView_CustomValue) TreeNodeAlloc (sizeof (struct _TPprootView_CustomValue));
#endif
_currn->_prod = RULEprootView_CustomValue;
_currn->_desc1 = (_TSProotView_CustomValue_value) MkrootView_CustomValue_value (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_CustomValue: root of subtree no. 1 can not be made a rootView_CustomValue_value node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_CustomValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_CustomValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_CustomValue_value (POSITION *_coordref)
#else
NODEPTR MkprootView_CustomValue_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_CustomValue_value _currn;
#ifdef __cplusplus
_currn = new _TPprootView_CustomValue_value;
#else
_currn = (_TPPprootView_CustomValue_value) TreeNodeAlloc (sizeof (struct _TPprootView_CustomValue_value));
#endif
_currn->_prod = RULEprootView_CustomValue_value;
_SETCOORD(_currn)
_TERMACT_prootView_CustomValue_value;
return ( (NODEPTR) _currn);
}/* MkprootView_CustomValue_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_HtmlStyleSheetBlock (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkprootView_HtmlStyleSheetBlock (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPprootView_HtmlStyleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPprootView_HtmlStyleSheetBlock;
#else
_currn = (_TPPprootView_HtmlStyleSheetBlock) TreeNodeAlloc (sizeof (struct _TPprootView_HtmlStyleSheetBlock));
#endif
_currn->_prod = RULEprootView_HtmlStyleSheetBlock;
_currn->_desc1 = (_TSProotView_HtmlStyleSheetBlock_pairs) MkrootView_HtmlStyleSheetBlock_pairs (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_HtmlStyleSheetBlock: root of subtree no. 1 can not be made a rootView_HtmlStyleSheetBlock_pairs node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_HtmlStyleSheetBlock_name) MkrootView_HtmlStyleSheetBlock_name (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_HtmlStyleSheetBlock: root of subtree no. 2 can not be made a rootView_HtmlStyleSheetBlock_name node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_HtmlStyleSheetBlock;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_HtmlStyleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_StyleSheetPairprootView_HtmlStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_StyleSheetPairprootView_HtmlStyleSheetBlock_pairs (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_StyleSheetPairprootView_HtmlStyleSheetBlock_pairs _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_StyleSheetPairprootView_HtmlStyleSheetBlock_pairs;
#else
	_new = (_TPPLST_rootView_StyleSheetPairprootView_HtmlStyleSheetBlock_pairs) TreeNodeAlloc (sizeof (struct _TPLST_rootView_StyleSheetPairprootView_HtmlStyleSheetBlock_pairs));
#endif
	_new->_prod = RULELST_rootView_StyleSheetPairprootView_HtmlStyleSheetBlock_pairs;
	_new->_desc1 = (_TSProotView_StyleSheetPair)elem;
	_new->_desc2 = (_TSPLST_rootView_HtmlStyleSheetBlock_pairs)_new;
	{_TPPLST_rootView_StyleSheetPairprootView_HtmlStyleSheetBlock_pairs _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_StyleSheetPairprootView_HtmlStyleSheetBlock_pairs;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_HtmlStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_HtmlStyleSheetBlock_pairs (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_StyleSheetPair)) 
		return _MkrootView_StyleSheetPairprootView_HtmlStyleSheetBlock_pairs (_coordref, MkrootView_StyleSheetPair (_coordref, t));
	message (DEADLY, "a prootView_HtmlStyleSheetBlock_pairs list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_HtmlStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_HtmlStyleSheetBlock_pairs (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_HtmlStyleSheetBlock_pairs)
	l =_ElemprootView_HtmlStyleSheetBlock_pairs (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_HtmlStyleSheetBlock_pairs)
	r = _ElemprootView_HtmlStyleSheetBlock_pairs (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_HtmlStyleSheetBlock_pairs)r)->_desc2);
((_TPPLST_0prootView_HtmlStyleSheetBlock_pairs)r)->_desc2 = ((_TPPLST_0prootView_HtmlStyleSheetBlock_pairs)l)->_desc2;
((_TPPLST_0prootView_HtmlStyleSheetBlock_pairs)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_HtmlStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_HtmlStyleSheetBlock_pairs (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_HtmlStyleSheetBlock_pairs root;
	_TPPLST_0prootView_HtmlStyleSheetBlock_pairs list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_HtmlStyleSheetBlock_pairs)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_HtmlStyleSheetBlock_pairs)
		t = _ElemprootView_HtmlStyleSheetBlock_pairs (_coordref, t);
	remain = (_TPPLST_0prootView_HtmlStyleSheetBlock_pairs)(((_TPPLST_0prootView_HtmlStyleSheetBlock_pairs) t)->_desc2);
	((_TPPLST_0prootView_HtmlStyleSheetBlock_pairs) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_HtmlStyleSheetBlock_pairs;
#else
list = (_TPPLST_0prootView_HtmlStyleSheetBlock_pairs) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_HtmlStyleSheetBlock_pairs));
#endif
list->_prod = RULELST_0prootView_HtmlStyleSheetBlock_pairs;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_HtmlStyleSheetBlock_pairs)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_StyleSheetPairprootView_HtmlStyleSheetBlock_pairs)
	{_TPPLST_rootView_StyleSheetPairprootView_HtmlStyleSheetBlock_pairs _currn = _currn;
	 _TERMACT_LST_rootView_StyleSheetPairprootView_HtmlStyleSheetBlock_pairs;
	} this code has been moved to _ElemprootView_HtmlStyleSheetBlock_pairs*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_HtmlStyleSheetBlock_pairs;
#else
root = (_TPPprootView_HtmlStyleSheetBlock_pairs) TreeNodeAlloc (sizeof (struct _TPprootView_HtmlStyleSheetBlock_pairs));
#endif
root->_prod = RULEprootView_HtmlStyleSheetBlock_pairs;
root->_desc1 = (_TSPLST_rootView_HtmlStyleSheetBlock_pairs)list;
{ _TPPprootView_HtmlStyleSheetBlock_pairs _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_HtmlStyleSheetBlock_pairs;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_HtmlStyleSheetBlock_name (POSITION *_coordref)
#else
NODEPTR MkprootView_HtmlStyleSheetBlock_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_HtmlStyleSheetBlock_name _currn;
#ifdef __cplusplus
_currn = new _TPprootView_HtmlStyleSheetBlock_name;
#else
_currn = (_TPPprootView_HtmlStyleSheetBlock_name) TreeNodeAlloc (sizeof (struct _TPprootView_HtmlStyleSheetBlock_name));
#endif
_currn->_prod = RULEprootView_HtmlStyleSheetBlock_name;
_SETCOORD(_currn)
_TERMACT_prootView_HtmlStyleSheetBlock_name;
return ( (NODEPTR) _currn);
}/* MkprootView_HtmlStyleSheetBlock_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_ActionHandler (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkprootView_ActionHandler (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPprootView_ActionHandler _currn;
#ifdef __cplusplus
_currn = new _TPprootView_ActionHandler;
#else
_currn = (_TPPprootView_ActionHandler) TreeNodeAlloc (sizeof (struct _TPprootView_ActionHandler));
#endif
_currn->_prod = RULEprootView_ActionHandler;
_currn->_desc1 = (_TSProotView_ActionHandler_targetpage) MkrootView_ActionHandler_targetpage (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_ActionHandler: root of subtree no. 1 can not be made a rootView_ActionHandler_targetpage node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_ActionHandler_params) MkrootView_ActionHandler_params (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_ActionHandler: root of subtree no. 2 can not be made a rootView_ActionHandler_params node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_ActionHandler;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_ActionHandler */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_ActionHandler_targetpage (POSITION *_coordref)
#else
NODEPTR MkprootView_ActionHandler_targetpage (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_ActionHandler_targetpage _currn;
#ifdef __cplusplus
_currn = new _TPprootView_ActionHandler_targetpage;
#else
_currn = (_TPPprootView_ActionHandler_targetpage) TreeNodeAlloc (sizeof (struct _TPprootView_ActionHandler_targetpage));
#endif
_currn->_prod = RULEprootView_ActionHandler_targetpage;
_SETCOORD(_currn)
_TERMACT_prootView_ActionHandler_targetpage;
return ( (NODEPTR) _currn);
}/* MkprootView_ActionHandler_targetpage */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_WebNodeActualParamprootView_ActionHandler_params (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_WebNodeActualParamprootView_ActionHandler_params (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_WebNodeActualParamprootView_ActionHandler_params _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_WebNodeActualParamprootView_ActionHandler_params;
#else
	_new = (_TPPLST_rootView_WebNodeActualParamprootView_ActionHandler_params) TreeNodeAlloc (sizeof (struct _TPLST_rootView_WebNodeActualParamprootView_ActionHandler_params));
#endif
	_new->_prod = RULELST_rootView_WebNodeActualParamprootView_ActionHandler_params;
	_new->_desc1 = (_TSProotView_WebNodeActualParam)elem;
	_new->_desc2 = (_TSPLST_rootView_ActionHandler_params)_new;
	{_TPPLST_rootView_WebNodeActualParamprootView_ActionHandler_params _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_WebNodeActualParamprootView_ActionHandler_params;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_ActionHandler_params (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_ActionHandler_params (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_WebNodeActualParam)) 
		return _MkrootView_WebNodeActualParamprootView_ActionHandler_params (_coordref, MkrootView_WebNodeActualParam (_coordref, t));
	message (DEADLY, "a prootView_ActionHandler_params list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_ActionHandler_params (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_ActionHandler_params (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_ActionHandler_params)
	l =_ElemprootView_ActionHandler_params (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_ActionHandler_params)
	r = _ElemprootView_ActionHandler_params (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_ActionHandler_params)r)->_desc2);
((_TPPLST_0prootView_ActionHandler_params)r)->_desc2 = ((_TPPLST_0prootView_ActionHandler_params)l)->_desc2;
((_TPPLST_0prootView_ActionHandler_params)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_ActionHandler_params (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_ActionHandler_params (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_ActionHandler_params root;
	_TPPLST_0prootView_ActionHandler_params list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_ActionHandler_params)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_ActionHandler_params)
		t = _ElemprootView_ActionHandler_params (_coordref, t);
	remain = (_TPPLST_0prootView_ActionHandler_params)(((_TPPLST_0prootView_ActionHandler_params) t)->_desc2);
	((_TPPLST_0prootView_ActionHandler_params) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_ActionHandler_params;
#else
list = (_TPPLST_0prootView_ActionHandler_params) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_ActionHandler_params));
#endif
list->_prod = RULELST_0prootView_ActionHandler_params;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_ActionHandler_params)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_WebNodeActualParamprootView_ActionHandler_params)
	{_TPPLST_rootView_WebNodeActualParamprootView_ActionHandler_params _currn = _currn;
	 _TERMACT_LST_rootView_WebNodeActualParamprootView_ActionHandler_params;
	} this code has been moved to _ElemprootView_ActionHandler_params*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_ActionHandler_params;
#else
root = (_TPPprootView_ActionHandler_params) TreeNodeAlloc (sizeof (struct _TPprootView_ActionHandler_params));
#endif
root->_prod = RULEprootView_ActionHandler_params;
root->_desc1 = (_TSPLST_rootView_ActionHandler_params)list;
{ _TPPprootView_ActionHandler_params _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_ActionHandler_params;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebAction (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkprootView_WebAction (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPprootView_WebAction _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebAction;
#else
_currn = (_TPPprootView_WebAction) TreeNodeAlloc (sizeof (struct _TPprootView_WebAction));
#endif
_currn->_prod = RULEprootView_WebAction;
_currn->_desc1 = (_TSProotView_WebAction_name) MkrootView_WebAction_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebAction: root of subtree no. 1 can not be made a rootView_WebAction_name node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_WebAction_formalParameter) MkrootView_WebAction_formalParameter (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebAction: root of subtree no. 2 can not be made a rootView_WebAction_formalParameter node ", 0, _coordref);
_currn->_desc3 = (_TSProotView_WebAction_successpage) MkrootView_WebAction_successpage (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebAction: root of subtree no. 3 can not be made a rootView_WebAction_successpage node ", 0, _coordref);
_currn->_desc4 = (_TSProotView_WebAction_operations) MkrootView_WebAction_operations (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebAction: root of subtree no. 4 can not be made a rootView_WebAction_operations node ", 0, _coordref);
_currn->_desc5 = (_TSProotView_WebAction_position) MkrootView_WebAction_position (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_WebAction: root of subtree no. 5 can not be made a rootView_WebAction_position node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_WebAction;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_WebAction */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebAction_name (POSITION *_coordref)
#else
NODEPTR MkprootView_WebAction_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_WebAction_name _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebAction_name;
#else
_currn = (_TPPprootView_WebAction_name) TreeNodeAlloc (sizeof (struct _TPprootView_WebAction_name));
#endif
_currn->_prod = RULEprootView_WebAction_name;
_SETCOORD(_currn)
_TERMACT_prootView_WebAction_name;
return ( (NODEPTR) _currn);
}/* MkprootView_WebAction_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_WebNodeFormalParamprootView_WebAction_formalParameter (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_WebNodeFormalParamprootView_WebAction_formalParameter (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_WebNodeFormalParamprootView_WebAction_formalParameter _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_WebNodeFormalParamprootView_WebAction_formalParameter;
#else
	_new = (_TPPLST_rootView_WebNodeFormalParamprootView_WebAction_formalParameter) TreeNodeAlloc (sizeof (struct _TPLST_rootView_WebNodeFormalParamprootView_WebAction_formalParameter));
#endif
	_new->_prod = RULELST_rootView_WebNodeFormalParamprootView_WebAction_formalParameter;
	_new->_desc1 = (_TSProotView_WebNodeFormalParam)elem;
	_new->_desc2 = (_TSPLST_rootView_WebAction_formalParameter)_new;
	{_TPPLST_rootView_WebNodeFormalParamprootView_WebAction_formalParameter _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_WebNodeFormalParamprootView_WebAction_formalParameter;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_WebAction_formalParameter (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_WebAction_formalParameter (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_WebNodeFormalParam)) 
		return _MkrootView_WebNodeFormalParamprootView_WebAction_formalParameter (_coordref, MkrootView_WebNodeFormalParam (_coordref, t));
	message (DEADLY, "a prootView_WebAction_formalParameter list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_WebAction_formalParameter (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_WebAction_formalParameter (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_WebAction_formalParameter)
	l =_ElemprootView_WebAction_formalParameter (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_WebAction_formalParameter)
	r = _ElemprootView_WebAction_formalParameter (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_WebAction_formalParameter)r)->_desc2);
((_TPPLST_0prootView_WebAction_formalParameter)r)->_desc2 = ((_TPPLST_0prootView_WebAction_formalParameter)l)->_desc2;
((_TPPLST_0prootView_WebAction_formalParameter)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebAction_formalParameter (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_WebAction_formalParameter (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_WebAction_formalParameter root;
	_TPPLST_0prootView_WebAction_formalParameter list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_WebAction_formalParameter)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_WebAction_formalParameter)
		t = _ElemprootView_WebAction_formalParameter (_coordref, t);
	remain = (_TPPLST_0prootView_WebAction_formalParameter)(((_TPPLST_0prootView_WebAction_formalParameter) t)->_desc2);
	((_TPPLST_0prootView_WebAction_formalParameter) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_WebAction_formalParameter;
#else
list = (_TPPLST_0prootView_WebAction_formalParameter) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_WebAction_formalParameter));
#endif
list->_prod = RULELST_0prootView_WebAction_formalParameter;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_WebAction_formalParameter)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_WebNodeFormalParamprootView_WebAction_formalParameter)
	{_TPPLST_rootView_WebNodeFormalParamprootView_WebAction_formalParameter _currn = _currn;
	 _TERMACT_LST_rootView_WebNodeFormalParamprootView_WebAction_formalParameter;
	} this code has been moved to _ElemprootView_WebAction_formalParameter*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_WebAction_formalParameter;
#else
root = (_TPPprootView_WebAction_formalParameter) TreeNodeAlloc (sizeof (struct _TPprootView_WebAction_formalParameter));
#endif
root->_prod = RULEprootView_WebAction_formalParameter;
root->_desc1 = (_TSPLST_rootView_WebAction_formalParameter)list;
{ _TPPprootView_WebAction_formalParameter _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_WebAction_formalParameter;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_PHPOperationprootView_WebAction_operations (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_PHPOperationprootView_WebAction_operations (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_PHPOperationprootView_WebAction_operations _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_PHPOperationprootView_WebAction_operations;
#else
	_new = (_TPPLST_rootView_PHPOperationprootView_WebAction_operations) TreeNodeAlloc (sizeof (struct _TPLST_rootView_PHPOperationprootView_WebAction_operations));
#endif
	_new->_prod = RULELST_rootView_PHPOperationprootView_WebAction_operations;
	_new->_desc1 = (_TSProotView_PHPOperation)elem;
	_new->_desc2 = (_TSPLST_rootView_WebAction_operations)_new;
	{_TPPLST_rootView_PHPOperationprootView_WebAction_operations _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_PHPOperationprootView_WebAction_operations;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_DataBaseOperationprootView_WebAction_operations (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_DataBaseOperationprootView_WebAction_operations (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_DataBaseOperationprootView_WebAction_operations _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_DataBaseOperationprootView_WebAction_operations;
#else
	_new = (_TPPLST_rootView_DataBaseOperationprootView_WebAction_operations) TreeNodeAlloc (sizeof (struct _TPLST_rootView_DataBaseOperationprootView_WebAction_operations));
#endif
	_new->_prod = RULELST_rootView_DataBaseOperationprootView_WebAction_operations;
	_new->_desc1 = (_TSProotView_DataBaseOperation)elem;
	_new->_desc2 = (_TSPLST_rootView_WebAction_operations)_new;
	{_TPPLST_rootView_DataBaseOperationprootView_WebAction_operations _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_DataBaseOperationprootView_WebAction_operations;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_WebAction_operations (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_WebAction_operations (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_PHPOperation)) 
		return _MkrootView_PHPOperationprootView_WebAction_operations (_coordref, MkrootView_PHPOperation (_coordref, t));
	if (IsSymb (t, SYMBrootView_DataBaseOperation)) 
		return _MkrootView_DataBaseOperationprootView_WebAction_operations (_coordref, MkrootView_DataBaseOperation (_coordref, t));
	message (DEADLY, "a prootView_WebAction_operations list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_WebAction_operations (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_WebAction_operations (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_WebAction_operations)
	l =_ElemprootView_WebAction_operations (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_WebAction_operations)
	r = _ElemprootView_WebAction_operations (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_WebAction_operations)r)->_desc2);
((_TPPLST_0prootView_WebAction_operations)r)->_desc2 = ((_TPPLST_0prootView_WebAction_operations)l)->_desc2;
((_TPPLST_0prootView_WebAction_operations)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebAction_operations (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_WebAction_operations (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_WebAction_operations root;
	_TPPLST_0prootView_WebAction_operations list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_WebAction_operations)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_WebAction_operations)
		t = _ElemprootView_WebAction_operations (_coordref, t);
	remain = (_TPPLST_0prootView_WebAction_operations)(((_TPPLST_0prootView_WebAction_operations) t)->_desc2);
	((_TPPLST_0prootView_WebAction_operations) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_WebAction_operations;
#else
list = (_TPPLST_0prootView_WebAction_operations) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_WebAction_operations));
#endif
list->_prod = RULELST_0prootView_WebAction_operations;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_WebAction_operations)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_PHPOperationprootView_WebAction_operations)
	{_TPPLST_rootView_PHPOperationprootView_WebAction_operations _currn = _currn;
	 _TERMACT_LST_rootView_PHPOperationprootView_WebAction_operations;
	} this code has been moved to _ElemprootView_WebAction_operations*/
	/*if (_currn->_prod == RULELST_rootView_DataBaseOperationprootView_WebAction_operations)
	{_TPPLST_rootView_DataBaseOperationprootView_WebAction_operations _currn = _currn;
	 _TERMACT_LST_rootView_DataBaseOperationprootView_WebAction_operations;
	} this code has been moved to _ElemprootView_WebAction_operations*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_WebAction_operations;
#else
root = (_TPPprootView_WebAction_operations) TreeNodeAlloc (sizeof (struct _TPprootView_WebAction_operations));
#endif
root->_prod = RULEprootView_WebAction_operations;
root->_desc1 = (_TSPLST_rootView_WebAction_operations)list;
{ _TPPprootView_WebAction_operations _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_WebAction_operations;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_WebAction_position (POSITION *_coordref)
#else
NODEPTR MkprootView_WebAction_position (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_WebAction_position _currn;
#ifdef __cplusplus
_currn = new _TPprootView_WebAction_position;
#else
_currn = (_TPPprootView_WebAction_position) TreeNodeAlloc (sizeof (struct _TPprootView_WebAction_position));
#endif
_currn->_prod = RULEprootView_WebAction_position;
_SETCOORD(_currn)
_TERMACT_prootView_WebAction_position;
return ( (NODEPTR) _currn);
}/* MkprootView_WebAction_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_PHPOperation (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkprootView_PHPOperation (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPprootView_PHPOperation _currn;
#ifdef __cplusplus
_currn = new _TPprootView_PHPOperation;
#else
_currn = (_TPPprootView_PHPOperation) TreeNodeAlloc (sizeof (struct _TPprootView_PHPOperation));
#endif
_currn->_prod = RULEprootView_PHPOperation;
_currn->_desc1 = (_TSProotView_PHPOperation_errorhandler) MkrootView_PHPOperation_errorhandler (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_PHPOperation: root of subtree no. 1 can not be made a rootView_PHPOperation_errorhandler node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_PHPOperation_content) MkrootView_PHPOperation_content (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_PHPOperation: root of subtree no. 2 can not be made a rootView_PHPOperation_content node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_PHPOperation;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_PHPOperation */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_ActionHandlerprootView_PHPOperation_errorhandler (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_ActionHandlerprootView_PHPOperation_errorhandler (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_ActionHandlerprootView_PHPOperation_errorhandler _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_ActionHandlerprootView_PHPOperation_errorhandler;
#else
	_new = (_TPPLST_rootView_ActionHandlerprootView_PHPOperation_errorhandler) TreeNodeAlloc (sizeof (struct _TPLST_rootView_ActionHandlerprootView_PHPOperation_errorhandler));
#endif
	_new->_prod = RULELST_rootView_ActionHandlerprootView_PHPOperation_errorhandler;
	_new->_desc1 = (_TSProotView_ActionHandler)elem;
	_new->_desc2 = (_TSPLST_rootView_PHPOperation_errorhandler)_new;
	{_TPPLST_rootView_ActionHandlerprootView_PHPOperation_errorhandler _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_ActionHandlerprootView_PHPOperation_errorhandler;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_PHPOperation_errorhandler (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_PHPOperation_errorhandler (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_ActionHandler)) 
		return _MkrootView_ActionHandlerprootView_PHPOperation_errorhandler (_coordref, MkrootView_ActionHandler (_coordref, t));
	message (DEADLY, "a prootView_PHPOperation_errorhandler list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_PHPOperation_errorhandler (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_PHPOperation_errorhandler (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_PHPOperation_errorhandler)
	l =_ElemprootView_PHPOperation_errorhandler (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_PHPOperation_errorhandler)
	r = _ElemprootView_PHPOperation_errorhandler (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_PHPOperation_errorhandler)r)->_desc2);
((_TPPLST_0prootView_PHPOperation_errorhandler)r)->_desc2 = ((_TPPLST_0prootView_PHPOperation_errorhandler)l)->_desc2;
((_TPPLST_0prootView_PHPOperation_errorhandler)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_PHPOperation_errorhandler (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_PHPOperation_errorhandler (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_PHPOperation_errorhandler root;
	_TPPLST_0prootView_PHPOperation_errorhandler list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_PHPOperation_errorhandler)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_PHPOperation_errorhandler)
		t = _ElemprootView_PHPOperation_errorhandler (_coordref, t);
	remain = (_TPPLST_0prootView_PHPOperation_errorhandler)(((_TPPLST_0prootView_PHPOperation_errorhandler) t)->_desc2);
	((_TPPLST_0prootView_PHPOperation_errorhandler) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_PHPOperation_errorhandler;
#else
list = (_TPPLST_0prootView_PHPOperation_errorhandler) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_PHPOperation_errorhandler));
#endif
list->_prod = RULELST_0prootView_PHPOperation_errorhandler;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_PHPOperation_errorhandler)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_ActionHandlerprootView_PHPOperation_errorhandler)
	{_TPPLST_rootView_ActionHandlerprootView_PHPOperation_errorhandler _currn = _currn;
	 _TERMACT_LST_rootView_ActionHandlerprootView_PHPOperation_errorhandler;
	} this code has been moved to _ElemprootView_PHPOperation_errorhandler*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_PHPOperation_errorhandler;
#else
root = (_TPPprootView_PHPOperation_errorhandler) TreeNodeAlloc (sizeof (struct _TPprootView_PHPOperation_errorhandler));
#endif
root->_prod = RULEprootView_PHPOperation_errorhandler;
root->_desc1 = (_TSPLST_rootView_PHPOperation_errorhandler)list;
{ _TPPprootView_PHPOperation_errorhandler _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_PHPOperation_errorhandler;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_PHPOperation_content (POSITION *_coordref)
#else
NODEPTR MkprootView_PHPOperation_content (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_PHPOperation_content _currn;
#ifdef __cplusplus
_currn = new _TPprootView_PHPOperation_content;
#else
_currn = (_TPPprootView_PHPOperation_content) TreeNodeAlloc (sizeof (struct _TPprootView_PHPOperation_content));
#endif
_currn->_prod = RULEprootView_PHPOperation_content;
_SETCOORD(_currn)
_TERMACT_prootView_PHPOperation_content;
return ( (NODEPTR) _currn);
}/* MkprootView_PHPOperation_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_DataBaseOperation (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkprootView_DataBaseOperation (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPprootView_DataBaseOperation _currn;
#ifdef __cplusplus
_currn = new _TPprootView_DataBaseOperation;
#else
_currn = (_TPPprootView_DataBaseOperation) TreeNodeAlloc (sizeof (struct _TPprootView_DataBaseOperation));
#endif
_currn->_prod = RULEprootView_DataBaseOperation;
_currn->_desc1 = (_TSProotView_DataBaseOperation_errorhandler) MkrootView_DataBaseOperation_errorhandler (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_DataBaseOperation: root of subtree no. 1 can not be made a rootView_DataBaseOperation_errorhandler node ", 0, _coordref);
_currn->_desc2 = (_TSProotView_DataBaseOperation_filter) MkrootView_DataBaseOperation_filter (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_DataBaseOperation: root of subtree no. 2 can not be made a rootView_DataBaseOperation_filter node ", 0, _coordref);
_currn->_desc3 = (_TSProotView_DataBaseOperation_params) MkrootView_DataBaseOperation_params (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE prootView_DataBaseOperation: root of subtree no. 3 can not be made a rootView_DataBaseOperation_params node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_prootView_DataBaseOperation;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkprootView_DataBaseOperation */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_ActionHandlerprootView_DataBaseOperation_errorhandler (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_ActionHandlerprootView_DataBaseOperation_errorhandler (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_ActionHandlerprootView_DataBaseOperation_errorhandler _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_ActionHandlerprootView_DataBaseOperation_errorhandler;
#else
	_new = (_TPPLST_rootView_ActionHandlerprootView_DataBaseOperation_errorhandler) TreeNodeAlloc (sizeof (struct _TPLST_rootView_ActionHandlerprootView_DataBaseOperation_errorhandler));
#endif
	_new->_prod = RULELST_rootView_ActionHandlerprootView_DataBaseOperation_errorhandler;
	_new->_desc1 = (_TSProotView_ActionHandler)elem;
	_new->_desc2 = (_TSPLST_rootView_DataBaseOperation_errorhandler)_new;
	{_TPPLST_rootView_ActionHandlerprootView_DataBaseOperation_errorhandler _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_ActionHandlerprootView_DataBaseOperation_errorhandler;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_DataBaseOperation_errorhandler (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_DataBaseOperation_errorhandler (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_ActionHandler)) 
		return _MkrootView_ActionHandlerprootView_DataBaseOperation_errorhandler (_coordref, MkrootView_ActionHandler (_coordref, t));
	message (DEADLY, "a prootView_DataBaseOperation_errorhandler list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_DataBaseOperation_errorhandler (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_DataBaseOperation_errorhandler (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_DataBaseOperation_errorhandler)
	l =_ElemprootView_DataBaseOperation_errorhandler (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_DataBaseOperation_errorhandler)
	r = _ElemprootView_DataBaseOperation_errorhandler (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_DataBaseOperation_errorhandler)r)->_desc2);
((_TPPLST_0prootView_DataBaseOperation_errorhandler)r)->_desc2 = ((_TPPLST_0prootView_DataBaseOperation_errorhandler)l)->_desc2;
((_TPPLST_0prootView_DataBaseOperation_errorhandler)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_DataBaseOperation_errorhandler (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_DataBaseOperation_errorhandler (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_DataBaseOperation_errorhandler root;
	_TPPLST_0prootView_DataBaseOperation_errorhandler list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_DataBaseOperation_errorhandler)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_DataBaseOperation_errorhandler)
		t = _ElemprootView_DataBaseOperation_errorhandler (_coordref, t);
	remain = (_TPPLST_0prootView_DataBaseOperation_errorhandler)(((_TPPLST_0prootView_DataBaseOperation_errorhandler) t)->_desc2);
	((_TPPLST_0prootView_DataBaseOperation_errorhandler) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_DataBaseOperation_errorhandler;
#else
list = (_TPPLST_0prootView_DataBaseOperation_errorhandler) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_DataBaseOperation_errorhandler));
#endif
list->_prod = RULELST_0prootView_DataBaseOperation_errorhandler;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_DataBaseOperation_errorhandler)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_ActionHandlerprootView_DataBaseOperation_errorhandler)
	{_TPPLST_rootView_ActionHandlerprootView_DataBaseOperation_errorhandler _currn = _currn;
	 _TERMACT_LST_rootView_ActionHandlerprootView_DataBaseOperation_errorhandler;
	} this code has been moved to _ElemprootView_DataBaseOperation_errorhandler*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_DataBaseOperation_errorhandler;
#else
root = (_TPPprootView_DataBaseOperation_errorhandler) TreeNodeAlloc (sizeof (struct _TPprootView_DataBaseOperation_errorhandler));
#endif
root->_prod = RULEprootView_DataBaseOperation_errorhandler;
root->_desc1 = (_TSPLST_rootView_DataBaseOperation_errorhandler)list;
{ _TPPprootView_DataBaseOperation_errorhandler _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_DataBaseOperation_errorhandler;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_DataBaseOperation_filter (POSITION *_coordref)
#else
NODEPTR MkprootView_DataBaseOperation_filter (_coordref)
	POSITION *_coordref;
#endif
{	_TPPprootView_DataBaseOperation_filter _currn;
#ifdef __cplusplus
_currn = new _TPprootView_DataBaseOperation_filter;
#else
_currn = (_TPPprootView_DataBaseOperation_filter) TreeNodeAlloc (sizeof (struct _TPprootView_DataBaseOperation_filter));
#endif
_currn->_prod = RULEprootView_DataBaseOperation_filter;
_SETCOORD(_currn)
_TERMACT_prootView_DataBaseOperation_filter;
return ( (NODEPTR) _currn);
}/* MkprootView_DataBaseOperation_filter */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkrootView_DataSourceActualParamprootView_DataBaseOperation_params (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkrootView_DataSourceActualParamprootView_DataBaseOperation_params (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_rootView_DataSourceActualParamprootView_DataBaseOperation_params _new;
#ifdef __cplusplus
	_new = new _TPLST_rootView_DataSourceActualParamprootView_DataBaseOperation_params;
#else
	_new = (_TPPLST_rootView_DataSourceActualParamprootView_DataBaseOperation_params) TreeNodeAlloc (sizeof (struct _TPLST_rootView_DataSourceActualParamprootView_DataBaseOperation_params));
#endif
	_new->_prod = RULELST_rootView_DataSourceActualParamprootView_DataBaseOperation_params;
	_new->_desc1 = (_TSProotView_DataSourceActualParam)elem;
	_new->_desc2 = (_TSPLST_rootView_DataBaseOperation_params)_new;
	{_TPPLST_rootView_DataSourceActualParamprootView_DataBaseOperation_params _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_rootView_DataSourceActualParamprootView_DataBaseOperation_params;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElemprootView_DataBaseOperation_params (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElemprootView_DataBaseOperation_params (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBrootView_DataSourceActualParam)) 
		return _MkrootView_DataSourceActualParamprootView_DataBaseOperation_params (_coordref, MkrootView_DataSourceActualParam (_coordref, t));
	message (DEADLY, "a prootView_DataBaseOperation_params list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2prootView_DataBaseOperation_params (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2prootView_DataBaseOperation_params (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_rootView_DataBaseOperation_params)
	l =_ElemprootView_DataBaseOperation_params (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_rootView_DataBaseOperation_params)
	r = _ElemprootView_DataBaseOperation_params (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0prootView_DataBaseOperation_params)r)->_desc2);
((_TPPLST_0prootView_DataBaseOperation_params)r)->_desc2 = ((_TPPLST_0prootView_DataBaseOperation_params)l)->_desc2;
((_TPPLST_0prootView_DataBaseOperation_params)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkprootView_DataBaseOperation_params (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkprootView_DataBaseOperation_params (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPprootView_DataBaseOperation_params root;
	_TPPLST_0prootView_DataBaseOperation_params list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0prootView_DataBaseOperation_params)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_rootView_DataBaseOperation_params)
		t = _ElemprootView_DataBaseOperation_params (_coordref, t);
	remain = (_TPPLST_0prootView_DataBaseOperation_params)(((_TPPLST_0prootView_DataBaseOperation_params) t)->_desc2);
	((_TPPLST_0prootView_DataBaseOperation_params) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0prootView_DataBaseOperation_params;
#else
list = (_TPPLST_0prootView_DataBaseOperation_params) TreeNodeAlloc (sizeof (struct _TPLST_0prootView_DataBaseOperation_params));
#endif
list->_prod = RULELST_0prootView_DataBaseOperation_params;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0prootView_DataBaseOperation_params)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_rootView_DataSourceActualParamprootView_DataBaseOperation_params)
	{_TPPLST_rootView_DataSourceActualParamprootView_DataBaseOperation_params _currn = _currn;
	 _TERMACT_LST_rootView_DataSourceActualParamprootView_DataBaseOperation_params;
	} this code has been moved to _ElemprootView_DataBaseOperation_params*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPprootView_DataBaseOperation_params;
#else
root = (_TPPprootView_DataBaseOperation_params) TreeNodeAlloc (sizeof (struct _TPprootView_DataBaseOperation_params));
#endif
root->_prod = RULEprootView_DataBaseOperation_params;
root->_desc1 = (_TSPLST_rootView_DataBaseOperation_params)list;
{ _TPPprootView_DataBaseOperation_params _currn = root;
  _SETCOORD(_currn)
  _TERMACT_prootView_DataBaseOperation_params;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpVIEWROOT_sspDlg (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR MkpVIEWROOT_sspDlg (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPpVIEWROOT_sspDlg _currn;
#ifdef __cplusplus
_currn = new _TPpVIEWROOT_sspDlg;
#else
_currn = (_TPPpVIEWROOT_sspDlg) TreeNodeAlloc (sizeof (struct _TPpVIEWROOT_sspDlg));
#endif
_currn->_prod = RULEpVIEWROOT_sspDlg;
_currn->_desc1 = (_TSPsspDlg_StyleSheetPair) MksspDlg_StyleSheetPair (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pVIEWROOT_sspDlg: root of subtree no. 1 can not be made a sspDlg_StyleSheetPair node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_pVIEWROOT_sspDlg;
return ( (NODEPTR) _currn);
}/* MkpVIEWROOT_sspDlg */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpsspDlg_StyleSheetPair (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4)
#else
NODEPTR MkpsspDlg_StyleSheetPair (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
#endif
{	_TPPpsspDlg_StyleSheetPair _currn;
#ifdef __cplusplus
_currn = new _TPpsspDlg_StyleSheetPair;
#else
_currn = (_TPPpsspDlg_StyleSheetPair) TreeNodeAlloc (sizeof (struct _TPpsspDlg_StyleSheetPair));
#endif
_currn->_prod = RULEpsspDlg_StyleSheetPair;
_currn->_desc1 = (_TSPsspDlg_StyleSheetPair_category) MksspDlg_StyleSheetPair_category (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE psspDlg_StyleSheetPair: root of subtree no. 1 can not be made a sspDlg_StyleSheetPair_category node ", 0, _coordref);
_currn->_desc2 = (_TSPsspDlg_StyleSheetPair_attribute) MksspDlg_StyleSheetPair_attribute (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE psspDlg_StyleSheetPair: root of subtree no. 2 can not be made a sspDlg_StyleSheetPair_attribute node ", 0, _coordref);
_currn->_desc3 = (_TSPsspDlg_StyleSheetPair_valueCategory) MksspDlg_StyleSheetPair_valueCategory (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE psspDlg_StyleSheetPair: root of subtree no. 3 can not be made a sspDlg_StyleSheetPair_valueCategory node ", 0, _coordref);
_currn->_desc4 = (_TSPsspDlg_StyleSheetPair_value) MksspDlg_StyleSheetPair_value (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE psspDlg_StyleSheetPair: root of subtree no. 4 can not be made a sspDlg_StyleSheetPair_value node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_psspDlg_StyleSheetPair;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpsspDlg_StyleSheetPair */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpsspDlg_StyleSheetPair_category (POSITION *_coordref)
#else
NODEPTR MkpsspDlg_StyleSheetPair_category (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpsspDlg_StyleSheetPair_category _currn;
#ifdef __cplusplus
_currn = new _TPpsspDlg_StyleSheetPair_category;
#else
_currn = (_TPPpsspDlg_StyleSheetPair_category) TreeNodeAlloc (sizeof (struct _TPpsspDlg_StyleSheetPair_category));
#endif
_currn->_prod = RULEpsspDlg_StyleSheetPair_category;
_SETCOORD(_currn)
_TERMACT_psspDlg_StyleSheetPair_category;
return ( (NODEPTR) _currn);
}/* MkpsspDlg_StyleSheetPair_category */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpsspDlg_StyleSheetPair_attribute (POSITION *_coordref)
#else
NODEPTR MkpsspDlg_StyleSheetPair_attribute (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpsspDlg_StyleSheetPair_attribute _currn;
#ifdef __cplusplus
_currn = new _TPpsspDlg_StyleSheetPair_attribute;
#else
_currn = (_TPPpsspDlg_StyleSheetPair_attribute) TreeNodeAlloc (sizeof (struct _TPpsspDlg_StyleSheetPair_attribute));
#endif
_currn->_prod = RULEpsspDlg_StyleSheetPair_attribute;
_SETCOORD(_currn)
_TERMACT_psspDlg_StyleSheetPair_attribute;
return ( (NODEPTR) _currn);
}/* MkpsspDlg_StyleSheetPair_attribute */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpsspDlg_StyleSheetPair_valueCategory (POSITION *_coordref)
#else
NODEPTR MkpsspDlg_StyleSheetPair_valueCategory (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpsspDlg_StyleSheetPair_valueCategory _currn;
#ifdef __cplusplus
_currn = new _TPpsspDlg_StyleSheetPair_valueCategory;
#else
_currn = (_TPPpsspDlg_StyleSheetPair_valueCategory) TreeNodeAlloc (sizeof (struct _TPpsspDlg_StyleSheetPair_valueCategory));
#endif
_currn->_prod = RULEpsspDlg_StyleSheetPair_valueCategory;
_SETCOORD(_currn)
_TERMACT_psspDlg_StyleSheetPair_valueCategory;
return ( (NODEPTR) _currn);
}/* MkpsspDlg_StyleSheetPair_valueCategory */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MksspDlg_SizeValuepsspDlg_StyleSheetPair_value (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MksspDlg_SizeValuepsspDlg_StyleSheetPair_value (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_sspDlg_SizeValuepsspDlg_StyleSheetPair_value _new;
#ifdef __cplusplus
	_new = new _TPLST_sspDlg_SizeValuepsspDlg_StyleSheetPair_value;
#else
	_new = (_TPPLST_sspDlg_SizeValuepsspDlg_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_sspDlg_SizeValuepsspDlg_StyleSheetPair_value));
#endif
	_new->_prod = RULELST_sspDlg_SizeValuepsspDlg_StyleSheetPair_value;
	_new->_desc1 = (_TSPsspDlg_SizeValue)elem;
	_new->_desc2 = (_TSPLST_sspDlg_StyleSheetPair_value)_new;
	{_TPPLST_sspDlg_SizeValuepsspDlg_StyleSheetPair_value _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_sspDlg_SizeValuepsspDlg_StyleSheetPair_value;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MksspDlg_ColorValuepsspDlg_StyleSheetPair_value (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MksspDlg_ColorValuepsspDlg_StyleSheetPair_value (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_sspDlg_ColorValuepsspDlg_StyleSheetPair_value _new;
#ifdef __cplusplus
	_new = new _TPLST_sspDlg_ColorValuepsspDlg_StyleSheetPair_value;
#else
	_new = (_TPPLST_sspDlg_ColorValuepsspDlg_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_sspDlg_ColorValuepsspDlg_StyleSheetPair_value));
#endif
	_new->_prod = RULELST_sspDlg_ColorValuepsspDlg_StyleSheetPair_value;
	_new->_desc1 = (_TSPsspDlg_ColorValue)elem;
	_new->_desc2 = (_TSPLST_sspDlg_StyleSheetPair_value)_new;
	{_TPPLST_sspDlg_ColorValuepsspDlg_StyleSheetPair_value _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_sspDlg_ColorValuepsspDlg_StyleSheetPair_value;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MksspDlg_FamilyValuepsspDlg_StyleSheetPair_value (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MksspDlg_FamilyValuepsspDlg_StyleSheetPair_value (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_sspDlg_FamilyValuepsspDlg_StyleSheetPair_value _new;
#ifdef __cplusplus
	_new = new _TPLST_sspDlg_FamilyValuepsspDlg_StyleSheetPair_value;
#else
	_new = (_TPPLST_sspDlg_FamilyValuepsspDlg_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_sspDlg_FamilyValuepsspDlg_StyleSheetPair_value));
#endif
	_new->_prod = RULELST_sspDlg_FamilyValuepsspDlg_StyleSheetPair_value;
	_new->_desc1 = (_TSPsspDlg_FamilyValue)elem;
	_new->_desc2 = (_TSPLST_sspDlg_StyleSheetPair_value)_new;
	{_TPPLST_sspDlg_FamilyValuepsspDlg_StyleSheetPair_value _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_sspDlg_FamilyValuepsspDlg_StyleSheetPair_value;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MksspDlg_StyleValuepsspDlg_StyleSheetPair_value (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MksspDlg_StyleValuepsspDlg_StyleSheetPair_value (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_sspDlg_StyleValuepsspDlg_StyleSheetPair_value _new;
#ifdef __cplusplus
	_new = new _TPLST_sspDlg_StyleValuepsspDlg_StyleSheetPair_value;
#else
	_new = (_TPPLST_sspDlg_StyleValuepsspDlg_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_sspDlg_StyleValuepsspDlg_StyleSheetPair_value));
#endif
	_new->_prod = RULELST_sspDlg_StyleValuepsspDlg_StyleSheetPair_value;
	_new->_desc1 = (_TSPsspDlg_StyleValue)elem;
	_new->_desc2 = (_TSPLST_sspDlg_StyleSheetPair_value)_new;
	{_TPPLST_sspDlg_StyleValuepsspDlg_StyleSheetPair_value _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_sspDlg_StyleValuepsspDlg_StyleSheetPair_value;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MksspDlg_HeadingValuepsspDlg_StyleSheetPair_value (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MksspDlg_HeadingValuepsspDlg_StyleSheetPair_value (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_sspDlg_HeadingValuepsspDlg_StyleSheetPair_value _new;
#ifdef __cplusplus
	_new = new _TPLST_sspDlg_HeadingValuepsspDlg_StyleSheetPair_value;
#else
	_new = (_TPPLST_sspDlg_HeadingValuepsspDlg_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_sspDlg_HeadingValuepsspDlg_StyleSheetPair_value));
#endif
	_new->_prod = RULELST_sspDlg_HeadingValuepsspDlg_StyleSheetPair_value;
	_new->_desc1 = (_TSPsspDlg_HeadingValue)elem;
	_new->_desc2 = (_TSPLST_sspDlg_StyleSheetPair_value)_new;
	{_TPPLST_sspDlg_HeadingValuepsspDlg_StyleSheetPair_value _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_sspDlg_HeadingValuepsspDlg_StyleSheetPair_value;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MksspDlg_AlignmentValuepsspDlg_StyleSheetPair_value (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MksspDlg_AlignmentValuepsspDlg_StyleSheetPair_value (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_sspDlg_AlignmentValuepsspDlg_StyleSheetPair_value _new;
#ifdef __cplusplus
	_new = new _TPLST_sspDlg_AlignmentValuepsspDlg_StyleSheetPair_value;
#else
	_new = (_TPPLST_sspDlg_AlignmentValuepsspDlg_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_sspDlg_AlignmentValuepsspDlg_StyleSheetPair_value));
#endif
	_new->_prod = RULELST_sspDlg_AlignmentValuepsspDlg_StyleSheetPair_value;
	_new->_desc1 = (_TSPsspDlg_AlignmentValue)elem;
	_new->_desc2 = (_TSPLST_sspDlg_StyleSheetPair_value)_new;
	{_TPPLST_sspDlg_AlignmentValuepsspDlg_StyleSheetPair_value _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_sspDlg_AlignmentValuepsspDlg_StyleSheetPair_value;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MksspDlg_ListValuepsspDlg_StyleSheetPair_value (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MksspDlg_ListValuepsspDlg_StyleSheetPair_value (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_sspDlg_ListValuepsspDlg_StyleSheetPair_value _new;
#ifdef __cplusplus
	_new = new _TPLST_sspDlg_ListValuepsspDlg_StyleSheetPair_value;
#else
	_new = (_TPPLST_sspDlg_ListValuepsspDlg_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_sspDlg_ListValuepsspDlg_StyleSheetPair_value));
#endif
	_new->_prod = RULELST_sspDlg_ListValuepsspDlg_StyleSheetPair_value;
	_new->_desc1 = (_TSPsspDlg_ListValue)elem;
	_new->_desc2 = (_TSPLST_sspDlg_StyleSheetPair_value)_new;
	{_TPPLST_sspDlg_ListValuepsspDlg_StyleSheetPair_value _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_sspDlg_ListValuepsspDlg_StyleSheetPair_value;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MksspDlg_CustomValuepsspDlg_StyleSheetPair_value (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MksspDlg_CustomValuepsspDlg_StyleSheetPair_value (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_sspDlg_CustomValuepsspDlg_StyleSheetPair_value _new;
#ifdef __cplusplus
	_new = new _TPLST_sspDlg_CustomValuepsspDlg_StyleSheetPair_value;
#else
	_new = (_TPPLST_sspDlg_CustomValuepsspDlg_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_sspDlg_CustomValuepsspDlg_StyleSheetPair_value));
#endif
	_new->_prod = RULELST_sspDlg_CustomValuepsspDlg_StyleSheetPair_value;
	_new->_desc1 = (_TSPsspDlg_CustomValue)elem;
	_new->_desc2 = (_TSPLST_sspDlg_StyleSheetPair_value)_new;
	{_TPPLST_sspDlg_CustomValuepsspDlg_StyleSheetPair_value _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_sspDlg_CustomValuepsspDlg_StyleSheetPair_value;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempsspDlg_StyleSheetPair_value (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempsspDlg_StyleSheetPair_value (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBsspDlg_SizeValue)) 
		return _MksspDlg_SizeValuepsspDlg_StyleSheetPair_value (_coordref, MksspDlg_SizeValue (_coordref, t));
	if (IsSymb (t, SYMBsspDlg_ColorValue)) 
		return _MksspDlg_ColorValuepsspDlg_StyleSheetPair_value (_coordref, MksspDlg_ColorValue (_coordref, t));
	if (IsSymb (t, SYMBsspDlg_FamilyValue)) 
		return _MksspDlg_FamilyValuepsspDlg_StyleSheetPair_value (_coordref, MksspDlg_FamilyValue (_coordref, t));
	if (IsSymb (t, SYMBsspDlg_StyleValue)) 
		return _MksspDlg_StyleValuepsspDlg_StyleSheetPair_value (_coordref, MksspDlg_StyleValue (_coordref, t));
	if (IsSymb (t, SYMBsspDlg_HeadingValue)) 
		return _MksspDlg_HeadingValuepsspDlg_StyleSheetPair_value (_coordref, MksspDlg_HeadingValue (_coordref, t));
	if (IsSymb (t, SYMBsspDlg_AlignmentValue)) 
		return _MksspDlg_AlignmentValuepsspDlg_StyleSheetPair_value (_coordref, MksspDlg_AlignmentValue (_coordref, t));
	if (IsSymb (t, SYMBsspDlg_ListValue)) 
		return _MksspDlg_ListValuepsspDlg_StyleSheetPair_value (_coordref, MksspDlg_ListValue (_coordref, t));
	if (IsSymb (t, SYMBsspDlg_CustomValue)) 
		return _MksspDlg_CustomValuepsspDlg_StyleSheetPair_value (_coordref, MksspDlg_CustomValue (_coordref, t));
	message (DEADLY, "a psspDlg_StyleSheetPair_value list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2psspDlg_StyleSheetPair_value (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2psspDlg_StyleSheetPair_value (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_sspDlg_StyleSheetPair_value)
	l =_ElempsspDlg_StyleSheetPair_value (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_sspDlg_StyleSheetPair_value)
	r = _ElempsspDlg_StyleSheetPair_value (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0psspDlg_StyleSheetPair_value)r)->_desc2);
((_TPPLST_0psspDlg_StyleSheetPair_value)r)->_desc2 = ((_TPPLST_0psspDlg_StyleSheetPair_value)l)->_desc2;
((_TPPLST_0psspDlg_StyleSheetPair_value)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpsspDlg_StyleSheetPair_value (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpsspDlg_StyleSheetPair_value (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpsspDlg_StyleSheetPair_value root;
	_TPPLST_0psspDlg_StyleSheetPair_value list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0psspDlg_StyleSheetPair_value)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_sspDlg_StyleSheetPair_value)
		t = _ElempsspDlg_StyleSheetPair_value (_coordref, t);
	remain = (_TPPLST_0psspDlg_StyleSheetPair_value)(((_TPPLST_0psspDlg_StyleSheetPair_value) t)->_desc2);
	((_TPPLST_0psspDlg_StyleSheetPair_value) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0psspDlg_StyleSheetPair_value;
#else
list = (_TPPLST_0psspDlg_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_0psspDlg_StyleSheetPair_value));
#endif
list->_prod = RULELST_0psspDlg_StyleSheetPair_value;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0psspDlg_StyleSheetPair_value)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_sspDlg_SizeValuepsspDlg_StyleSheetPair_value)
	{_TPPLST_sspDlg_SizeValuepsspDlg_StyleSheetPair_value _currn = _currn;
	 _TERMACT_LST_sspDlg_SizeValuepsspDlg_StyleSheetPair_value;
	} this code has been moved to _ElempsspDlg_StyleSheetPair_value*/
	/*if (_currn->_prod == RULELST_sspDlg_ColorValuepsspDlg_StyleSheetPair_value)
	{_TPPLST_sspDlg_ColorValuepsspDlg_StyleSheetPair_value _currn = _currn;
	 _TERMACT_LST_sspDlg_ColorValuepsspDlg_StyleSheetPair_value;
	} this code has been moved to _ElempsspDlg_StyleSheetPair_value*/
	/*if (_currn->_prod == RULELST_sspDlg_FamilyValuepsspDlg_StyleSheetPair_value)
	{_TPPLST_sspDlg_FamilyValuepsspDlg_StyleSheetPair_value _currn = _currn;
	 _TERMACT_LST_sspDlg_FamilyValuepsspDlg_StyleSheetPair_value;
	} this code has been moved to _ElempsspDlg_StyleSheetPair_value*/
	/*if (_currn->_prod == RULELST_sspDlg_StyleValuepsspDlg_StyleSheetPair_value)
	{_TPPLST_sspDlg_StyleValuepsspDlg_StyleSheetPair_value _currn = _currn;
	 _TERMACT_LST_sspDlg_StyleValuepsspDlg_StyleSheetPair_value;
	} this code has been moved to _ElempsspDlg_StyleSheetPair_value*/
	/*if (_currn->_prod == RULELST_sspDlg_HeadingValuepsspDlg_StyleSheetPair_value)
	{_TPPLST_sspDlg_HeadingValuepsspDlg_StyleSheetPair_value _currn = _currn;
	 _TERMACT_LST_sspDlg_HeadingValuepsspDlg_StyleSheetPair_value;
	} this code has been moved to _ElempsspDlg_StyleSheetPair_value*/
	/*if (_currn->_prod == RULELST_sspDlg_AlignmentValuepsspDlg_StyleSheetPair_value)
	{_TPPLST_sspDlg_AlignmentValuepsspDlg_StyleSheetPair_value _currn = _currn;
	 _TERMACT_LST_sspDlg_AlignmentValuepsspDlg_StyleSheetPair_value;
	} this code has been moved to _ElempsspDlg_StyleSheetPair_value*/
	/*if (_currn->_prod == RULELST_sspDlg_ListValuepsspDlg_StyleSheetPair_value)
	{_TPPLST_sspDlg_ListValuepsspDlg_StyleSheetPair_value _currn = _currn;
	 _TERMACT_LST_sspDlg_ListValuepsspDlg_StyleSheetPair_value;
	} this code has been moved to _ElempsspDlg_StyleSheetPair_value*/
	/*if (_currn->_prod == RULELST_sspDlg_CustomValuepsspDlg_StyleSheetPair_value)
	{_TPPLST_sspDlg_CustomValuepsspDlg_StyleSheetPair_value _currn = _currn;
	 _TERMACT_LST_sspDlg_CustomValuepsspDlg_StyleSheetPair_value;
	} this code has been moved to _ElempsspDlg_StyleSheetPair_value*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpsspDlg_StyleSheetPair_value;
#else
root = (_TPPpsspDlg_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPpsspDlg_StyleSheetPair_value));
#endif
root->_prod = RULEpsspDlg_StyleSheetPair_value;
root->_desc1 = (_TSPLST_sspDlg_StyleSheetPair_value)list;
{ _TPPpsspDlg_StyleSheetPair_value _currn = root;
  _SETCOORD(_currn)
  _TERMACT_psspDlg_StyleSheetPair_value;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpsspDlg_SizeValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpsspDlg_SizeValue (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpsspDlg_SizeValue _currn;
#ifdef __cplusplus
_currn = new _TPpsspDlg_SizeValue;
#else
_currn = (_TPPpsspDlg_SizeValue) TreeNodeAlloc (sizeof (struct _TPpsspDlg_SizeValue));
#endif
_currn->_prod = RULEpsspDlg_SizeValue;
_currn->_desc1 = (_TSPsspDlg_SizeValue_measure) MksspDlg_SizeValue_measure (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE psspDlg_SizeValue: root of subtree no. 1 can not be made a sspDlg_SizeValue_measure node ", 0, _coordref);
_currn->_desc2 = (_TSPsspDlg_SizeValue_value) MksspDlg_SizeValue_value (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE psspDlg_SizeValue: root of subtree no. 2 can not be made a sspDlg_SizeValue_value node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_psspDlg_SizeValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpsspDlg_SizeValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpsspDlg_SizeValue_measure (POSITION *_coordref)
#else
NODEPTR MkpsspDlg_SizeValue_measure (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpsspDlg_SizeValue_measure _currn;
#ifdef __cplusplus
_currn = new _TPpsspDlg_SizeValue_measure;
#else
_currn = (_TPPpsspDlg_SizeValue_measure) TreeNodeAlloc (sizeof (struct _TPpsspDlg_SizeValue_measure));
#endif
_currn->_prod = RULEpsspDlg_SizeValue_measure;
_SETCOORD(_currn)
_TERMACT_psspDlg_SizeValue_measure;
return ( (NODEPTR) _currn);
}/* MkpsspDlg_SizeValue_measure */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpsspDlg_SizeValue_value (POSITION *_coordref)
#else
NODEPTR MkpsspDlg_SizeValue_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpsspDlg_SizeValue_value _currn;
#ifdef __cplusplus
_currn = new _TPpsspDlg_SizeValue_value;
#else
_currn = (_TPPpsspDlg_SizeValue_value) TreeNodeAlloc (sizeof (struct _TPpsspDlg_SizeValue_value));
#endif
_currn->_prod = RULEpsspDlg_SizeValue_value;
_SETCOORD(_currn)
_TERMACT_psspDlg_SizeValue_value;
return ( (NODEPTR) _currn);
}/* MkpsspDlg_SizeValue_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpsspDlg_ColorValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpsspDlg_ColorValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpsspDlg_ColorValue _currn;
#ifdef __cplusplus
_currn = new _TPpsspDlg_ColorValue;
#else
_currn = (_TPPpsspDlg_ColorValue) TreeNodeAlloc (sizeof (struct _TPpsspDlg_ColorValue));
#endif
_currn->_prod = RULEpsspDlg_ColorValue;
_currn->_desc1 = (_TSPsspDlg_ColorValue_color) MksspDlg_ColorValue_color (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE psspDlg_ColorValue: root of subtree no. 1 can not be made a sspDlg_ColorValue_color node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_psspDlg_ColorValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpsspDlg_ColorValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpsspDlg_ColorValue_color (POSITION *_coordref)
#else
NODEPTR MkpsspDlg_ColorValue_color (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpsspDlg_ColorValue_color _currn;
#ifdef __cplusplus
_currn = new _TPpsspDlg_ColorValue_color;
#else
_currn = (_TPPpsspDlg_ColorValue_color) TreeNodeAlloc (sizeof (struct _TPpsspDlg_ColorValue_color));
#endif
_currn->_prod = RULEpsspDlg_ColorValue_color;
_SETCOORD(_currn)
_TERMACT_psspDlg_ColorValue_color;
return ( (NODEPTR) _currn);
}/* MkpsspDlg_ColorValue_color */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpsspDlg_FamilyValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpsspDlg_FamilyValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpsspDlg_FamilyValue _currn;
#ifdef __cplusplus
_currn = new _TPpsspDlg_FamilyValue;
#else
_currn = (_TPPpsspDlg_FamilyValue) TreeNodeAlloc (sizeof (struct _TPpsspDlg_FamilyValue));
#endif
_currn->_prod = RULEpsspDlg_FamilyValue;
_currn->_desc1 = (_TSPsspDlg_FamilyValue_family) MksspDlg_FamilyValue_family (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE psspDlg_FamilyValue: root of subtree no. 1 can not be made a sspDlg_FamilyValue_family node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_psspDlg_FamilyValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpsspDlg_FamilyValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpsspDlg_FamilyValue_family (POSITION *_coordref)
#else
NODEPTR MkpsspDlg_FamilyValue_family (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpsspDlg_FamilyValue_family _currn;
#ifdef __cplusplus
_currn = new _TPpsspDlg_FamilyValue_family;
#else
_currn = (_TPPpsspDlg_FamilyValue_family) TreeNodeAlloc (sizeof (struct _TPpsspDlg_FamilyValue_family));
#endif
_currn->_prod = RULEpsspDlg_FamilyValue_family;
_SETCOORD(_currn)
_TERMACT_psspDlg_FamilyValue_family;
return ( (NODEPTR) _currn);
}/* MkpsspDlg_FamilyValue_family */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpsspDlg_StyleValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpsspDlg_StyleValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpsspDlg_StyleValue _currn;
#ifdef __cplusplus
_currn = new _TPpsspDlg_StyleValue;
#else
_currn = (_TPPpsspDlg_StyleValue) TreeNodeAlloc (sizeof (struct _TPpsspDlg_StyleValue));
#endif
_currn->_prod = RULEpsspDlg_StyleValue;
_currn->_desc1 = (_TSPsspDlg_StyleValue_style) MksspDlg_StyleValue_style (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE psspDlg_StyleValue: root of subtree no. 1 can not be made a sspDlg_StyleValue_style node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_psspDlg_StyleValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpsspDlg_StyleValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpsspDlg_StyleValue_style (POSITION *_coordref)
#else
NODEPTR MkpsspDlg_StyleValue_style (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpsspDlg_StyleValue_style _currn;
#ifdef __cplusplus
_currn = new _TPpsspDlg_StyleValue_style;
#else
_currn = (_TPPpsspDlg_StyleValue_style) TreeNodeAlloc (sizeof (struct _TPpsspDlg_StyleValue_style));
#endif
_currn->_prod = RULEpsspDlg_StyleValue_style;
_SETCOORD(_currn)
_TERMACT_psspDlg_StyleValue_style;
return ( (NODEPTR) _currn);
}/* MkpsspDlg_StyleValue_style */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpsspDlg_HeadingValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpsspDlg_HeadingValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpsspDlg_HeadingValue _currn;
#ifdef __cplusplus
_currn = new _TPpsspDlg_HeadingValue;
#else
_currn = (_TPPpsspDlg_HeadingValue) TreeNodeAlloc (sizeof (struct _TPpsspDlg_HeadingValue));
#endif
_currn->_prod = RULEpsspDlg_HeadingValue;
_currn->_desc1 = (_TSPsspDlg_HeadingValue_heading) MksspDlg_HeadingValue_heading (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE psspDlg_HeadingValue: root of subtree no. 1 can not be made a sspDlg_HeadingValue_heading node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_psspDlg_HeadingValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpsspDlg_HeadingValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpsspDlg_HeadingValue_heading (POSITION *_coordref)
#else
NODEPTR MkpsspDlg_HeadingValue_heading (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpsspDlg_HeadingValue_heading _currn;
#ifdef __cplusplus
_currn = new _TPpsspDlg_HeadingValue_heading;
#else
_currn = (_TPPpsspDlg_HeadingValue_heading) TreeNodeAlloc (sizeof (struct _TPpsspDlg_HeadingValue_heading));
#endif
_currn->_prod = RULEpsspDlg_HeadingValue_heading;
_SETCOORD(_currn)
_TERMACT_psspDlg_HeadingValue_heading;
return ( (NODEPTR) _currn);
}/* MkpsspDlg_HeadingValue_heading */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpsspDlg_AlignmentValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpsspDlg_AlignmentValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpsspDlg_AlignmentValue _currn;
#ifdef __cplusplus
_currn = new _TPpsspDlg_AlignmentValue;
#else
_currn = (_TPPpsspDlg_AlignmentValue) TreeNodeAlloc (sizeof (struct _TPpsspDlg_AlignmentValue));
#endif
_currn->_prod = RULEpsspDlg_AlignmentValue;
_currn->_desc1 = (_TSPsspDlg_AlignmentValue_alignment) MksspDlg_AlignmentValue_alignment (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE psspDlg_AlignmentValue: root of subtree no. 1 can not be made a sspDlg_AlignmentValue_alignment node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_psspDlg_AlignmentValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpsspDlg_AlignmentValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpsspDlg_AlignmentValue_alignment (POSITION *_coordref)
#else
NODEPTR MkpsspDlg_AlignmentValue_alignment (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpsspDlg_AlignmentValue_alignment _currn;
#ifdef __cplusplus
_currn = new _TPpsspDlg_AlignmentValue_alignment;
#else
_currn = (_TPPpsspDlg_AlignmentValue_alignment) TreeNodeAlloc (sizeof (struct _TPpsspDlg_AlignmentValue_alignment));
#endif
_currn->_prod = RULEpsspDlg_AlignmentValue_alignment;
_SETCOORD(_currn)
_TERMACT_psspDlg_AlignmentValue_alignment;
return ( (NODEPTR) _currn);
}/* MkpsspDlg_AlignmentValue_alignment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpsspDlg_ListValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpsspDlg_ListValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpsspDlg_ListValue _currn;
#ifdef __cplusplus
_currn = new _TPpsspDlg_ListValue;
#else
_currn = (_TPPpsspDlg_ListValue) TreeNodeAlloc (sizeof (struct _TPpsspDlg_ListValue));
#endif
_currn->_prod = RULEpsspDlg_ListValue;
_currn->_desc1 = (_TSPsspDlg_ListValue_listed) MksspDlg_ListValue_listed (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE psspDlg_ListValue: root of subtree no. 1 can not be made a sspDlg_ListValue_listed node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_psspDlg_ListValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpsspDlg_ListValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpsspDlg_ListValue_listed (POSITION *_coordref)
#else
NODEPTR MkpsspDlg_ListValue_listed (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpsspDlg_ListValue_listed _currn;
#ifdef __cplusplus
_currn = new _TPpsspDlg_ListValue_listed;
#else
_currn = (_TPPpsspDlg_ListValue_listed) TreeNodeAlloc (sizeof (struct _TPpsspDlg_ListValue_listed));
#endif
_currn->_prod = RULEpsspDlg_ListValue_listed;
_SETCOORD(_currn)
_TERMACT_psspDlg_ListValue_listed;
return ( (NODEPTR) _currn);
}/* MkpsspDlg_ListValue_listed */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpsspDlg_CustomValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpsspDlg_CustomValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpsspDlg_CustomValue _currn;
#ifdef __cplusplus
_currn = new _TPpsspDlg_CustomValue;
#else
_currn = (_TPPpsspDlg_CustomValue) TreeNodeAlloc (sizeof (struct _TPpsspDlg_CustomValue));
#endif
_currn->_prod = RULEpsspDlg_CustomValue;
_currn->_desc1 = (_TSPsspDlg_CustomValue_value) MksspDlg_CustomValue_value (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE psspDlg_CustomValue: root of subtree no. 1 can not be made a sspDlg_CustomValue_value node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_psspDlg_CustomValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpsspDlg_CustomValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpsspDlg_CustomValue_value (POSITION *_coordref)
#else
NODEPTR MkpsspDlg_CustomValue_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpsspDlg_CustomValue_value _currn;
#ifdef __cplusplus
_currn = new _TPpsspDlg_CustomValue_value;
#else
_currn = (_TPPpsspDlg_CustomValue_value) TreeNodeAlloc (sizeof (struct _TPpsspDlg_CustomValue_value));
#endif
_currn->_prod = RULEpsspDlg_CustomValue_value;
_SETCOORD(_currn)
_TERMACT_psspDlg_CustomValue_value;
return ( (NODEPTR) _currn);
}/* MkpsspDlg_CustomValue_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpVIEWROOT_stylesheetView (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR MkpVIEWROOT_stylesheetView (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPpVIEWROOT_stylesheetView _currn;
#ifdef __cplusplus
_currn = new _TPpVIEWROOT_stylesheetView;
#else
_currn = (_TPPpVIEWROOT_stylesheetView) TreeNodeAlloc (sizeof (struct _TPpVIEWROOT_stylesheetView));
#endif
_currn->_prod = RULEpVIEWROOT_stylesheetView;
_currn->_desc1 = (_TSPstylesheetView_StyleSheet) MkstylesheetView_StyleSheet (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pVIEWROOT_stylesheetView: root of subtree no. 1 can not be made a stylesheetView_StyleSheet node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_pVIEWROOT_stylesheetView;
return ( (NODEPTR) _currn);
}/* MkpVIEWROOT_stylesheetView */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_StyleSheet (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpstylesheetView_StyleSheet (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpstylesheetView_StyleSheet _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_StyleSheet;
#else
_currn = (_TPPpstylesheetView_StyleSheet) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_StyleSheet));
#endif
_currn->_prod = RULEpstylesheetView_StyleSheet;
_currn->_desc1 = (_TSPstylesheetView_StyleSheet_name) MkstylesheetView_StyleSheet_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pstylesheetView_StyleSheet: root of subtree no. 1 can not be made a stylesheetView_StyleSheet_name node ", 0, _coordref);
_currn->_desc2 = (_TSPstylesheetView_StyleSheet_custom) MkstylesheetView_StyleSheet_custom (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pstylesheetView_StyleSheet: root of subtree no. 2 can not be made a stylesheetView_StyleSheet_custom node ", 0, _coordref);
_currn->_desc3 = (_TSPstylesheetView_StyleSheet_default) MkstylesheetView_StyleSheet_default (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pstylesheetView_StyleSheet: root of subtree no. 3 can not be made a stylesheetView_StyleSheet_default node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pstylesheetView_StyleSheet;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpstylesheetView_StyleSheet */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_StyleSheet_name (POSITION *_coordref)
#else
NODEPTR MkpstylesheetView_StyleSheet_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpstylesheetView_StyleSheet_name _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_StyleSheet_name;
#else
_currn = (_TPPpstylesheetView_StyleSheet_name) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_StyleSheet_name));
#endif
_currn->_prod = RULEpstylesheetView_StyleSheet_name;
_SETCOORD(_currn)
_TERMACT_pstylesheetView_StyleSheet_name;
return ( (NODEPTR) _currn);
}/* MkpstylesheetView_StyleSheet_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkstylesheetView_CustomStyleSheetBlockpstylesheetView_StyleSheet_custom (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkstylesheetView_CustomStyleSheetBlockpstylesheetView_StyleSheet_custom (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_stylesheetView_CustomStyleSheetBlockpstylesheetView_StyleSheet_custom _new;
#ifdef __cplusplus
	_new = new _TPLST_stylesheetView_CustomStyleSheetBlockpstylesheetView_StyleSheet_custom;
#else
	_new = (_TPPLST_stylesheetView_CustomStyleSheetBlockpstylesheetView_StyleSheet_custom) TreeNodeAlloc (sizeof (struct _TPLST_stylesheetView_CustomStyleSheetBlockpstylesheetView_StyleSheet_custom));
#endif
	_new->_prod = RULELST_stylesheetView_CustomStyleSheetBlockpstylesheetView_StyleSheet_custom;
	_new->_desc1 = (_TSPstylesheetView_CustomStyleSheetBlock)elem;
	_new->_desc2 = (_TSPLST_stylesheetView_StyleSheet_custom)_new;
	{_TPPLST_stylesheetView_CustomStyleSheetBlockpstylesheetView_StyleSheet_custom _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_stylesheetView_CustomStyleSheetBlockpstylesheetView_StyleSheet_custom;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempstylesheetView_StyleSheet_custom (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempstylesheetView_StyleSheet_custom (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBstylesheetView_CustomStyleSheetBlock)) 
		return _MkstylesheetView_CustomStyleSheetBlockpstylesheetView_StyleSheet_custom (_coordref, MkstylesheetView_CustomStyleSheetBlock (_coordref, t));
	message (DEADLY, "a pstylesheetView_StyleSheet_custom list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pstylesheetView_StyleSheet_custom (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pstylesheetView_StyleSheet_custom (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_stylesheetView_StyleSheet_custom)
	l =_ElempstylesheetView_StyleSheet_custom (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_stylesheetView_StyleSheet_custom)
	r = _ElempstylesheetView_StyleSheet_custom (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pstylesheetView_StyleSheet_custom)r)->_desc2);
((_TPPLST_0pstylesheetView_StyleSheet_custom)r)->_desc2 = ((_TPPLST_0pstylesheetView_StyleSheet_custom)l)->_desc2;
((_TPPLST_0pstylesheetView_StyleSheet_custom)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_StyleSheet_custom (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpstylesheetView_StyleSheet_custom (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpstylesheetView_StyleSheet_custom root;
	_TPPLST_0pstylesheetView_StyleSheet_custom list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pstylesheetView_StyleSheet_custom)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_stylesheetView_StyleSheet_custom)
		t = _ElempstylesheetView_StyleSheet_custom (_coordref, t);
	remain = (_TPPLST_0pstylesheetView_StyleSheet_custom)(((_TPPLST_0pstylesheetView_StyleSheet_custom) t)->_desc2);
	((_TPPLST_0pstylesheetView_StyleSheet_custom) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pstylesheetView_StyleSheet_custom;
#else
list = (_TPPLST_0pstylesheetView_StyleSheet_custom) TreeNodeAlloc (sizeof (struct _TPLST_0pstylesheetView_StyleSheet_custom));
#endif
list->_prod = RULELST_0pstylesheetView_StyleSheet_custom;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pstylesheetView_StyleSheet_custom)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_stylesheetView_CustomStyleSheetBlockpstylesheetView_StyleSheet_custom)
	{_TPPLST_stylesheetView_CustomStyleSheetBlockpstylesheetView_StyleSheet_custom _currn = _currn;
	 _TERMACT_LST_stylesheetView_CustomStyleSheetBlockpstylesheetView_StyleSheet_custom;
	} this code has been moved to _ElempstylesheetView_StyleSheet_custom*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpstylesheetView_StyleSheet_custom;
#else
root = (_TPPpstylesheetView_StyleSheet_custom) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_StyleSheet_custom));
#endif
root->_prod = RULEpstylesheetView_StyleSheet_custom;
root->_desc1 = (_TSPLST_stylesheetView_StyleSheet_custom)list;
{ _TPPpstylesheetView_StyleSheet_custom _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pstylesheetView_StyleSheet_custom;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkstylesheetView_HtmlStyleSheetBlockpstylesheetView_StyleSheet_default (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkstylesheetView_HtmlStyleSheetBlockpstylesheetView_StyleSheet_default (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_stylesheetView_HtmlStyleSheetBlockpstylesheetView_StyleSheet_default _new;
#ifdef __cplusplus
	_new = new _TPLST_stylesheetView_HtmlStyleSheetBlockpstylesheetView_StyleSheet_default;
#else
	_new = (_TPPLST_stylesheetView_HtmlStyleSheetBlockpstylesheetView_StyleSheet_default) TreeNodeAlloc (sizeof (struct _TPLST_stylesheetView_HtmlStyleSheetBlockpstylesheetView_StyleSheet_default));
#endif
	_new->_prod = RULELST_stylesheetView_HtmlStyleSheetBlockpstylesheetView_StyleSheet_default;
	_new->_desc1 = (_TSPstylesheetView_HtmlStyleSheetBlock)elem;
	_new->_desc2 = (_TSPLST_stylesheetView_StyleSheet_default)_new;
	{_TPPLST_stylesheetView_HtmlStyleSheetBlockpstylesheetView_StyleSheet_default _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_stylesheetView_HtmlStyleSheetBlockpstylesheetView_StyleSheet_default;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempstylesheetView_StyleSheet_default (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempstylesheetView_StyleSheet_default (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBstylesheetView_HtmlStyleSheetBlock)) 
		return _MkstylesheetView_HtmlStyleSheetBlockpstylesheetView_StyleSheet_default (_coordref, MkstylesheetView_HtmlStyleSheetBlock (_coordref, t));
	message (DEADLY, "a pstylesheetView_StyleSheet_default list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pstylesheetView_StyleSheet_default (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pstylesheetView_StyleSheet_default (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_stylesheetView_StyleSheet_default)
	l =_ElempstylesheetView_StyleSheet_default (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_stylesheetView_StyleSheet_default)
	r = _ElempstylesheetView_StyleSheet_default (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pstylesheetView_StyleSheet_default)r)->_desc2);
((_TPPLST_0pstylesheetView_StyleSheet_default)r)->_desc2 = ((_TPPLST_0pstylesheetView_StyleSheet_default)l)->_desc2;
((_TPPLST_0pstylesheetView_StyleSheet_default)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_StyleSheet_default (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpstylesheetView_StyleSheet_default (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpstylesheetView_StyleSheet_default root;
	_TPPLST_0pstylesheetView_StyleSheet_default list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pstylesheetView_StyleSheet_default)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_stylesheetView_StyleSheet_default)
		t = _ElempstylesheetView_StyleSheet_default (_coordref, t);
	remain = (_TPPLST_0pstylesheetView_StyleSheet_default)(((_TPPLST_0pstylesheetView_StyleSheet_default) t)->_desc2);
	((_TPPLST_0pstylesheetView_StyleSheet_default) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pstylesheetView_StyleSheet_default;
#else
list = (_TPPLST_0pstylesheetView_StyleSheet_default) TreeNodeAlloc (sizeof (struct _TPLST_0pstylesheetView_StyleSheet_default));
#endif
list->_prod = RULELST_0pstylesheetView_StyleSheet_default;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pstylesheetView_StyleSheet_default)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_stylesheetView_HtmlStyleSheetBlockpstylesheetView_StyleSheet_default)
	{_TPPLST_stylesheetView_HtmlStyleSheetBlockpstylesheetView_StyleSheet_default _currn = _currn;
	 _TERMACT_LST_stylesheetView_HtmlStyleSheetBlockpstylesheetView_StyleSheet_default;
	} this code has been moved to _ElempstylesheetView_StyleSheet_default*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpstylesheetView_StyleSheet_default;
#else
root = (_TPPpstylesheetView_StyleSheet_default) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_StyleSheet_default));
#endif
root->_prod = RULEpstylesheetView_StyleSheet_default;
root->_desc1 = (_TSPLST_stylesheetView_StyleSheet_default)list;
{ _TPPpstylesheetView_StyleSheet_default _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pstylesheetView_StyleSheet_default;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_CustomStyleSheetBlock (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpstylesheetView_CustomStyleSheetBlock (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpstylesheetView_CustomStyleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_CustomStyleSheetBlock;
#else
_currn = (_TPPpstylesheetView_CustomStyleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_CustomStyleSheetBlock));
#endif
_currn->_prod = RULEpstylesheetView_CustomStyleSheetBlock;
_currn->_desc1 = (_TSPstylesheetView_CustomStyleSheetBlock_pairs) MkstylesheetView_CustomStyleSheetBlock_pairs (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pstylesheetView_CustomStyleSheetBlock: root of subtree no. 1 can not be made a stylesheetView_CustomStyleSheetBlock_pairs node ", 0, _coordref);
_currn->_desc2 = (_TSPstylesheetView_CustomStyleSheetBlock_name) MkstylesheetView_CustomStyleSheetBlock_name (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pstylesheetView_CustomStyleSheetBlock: root of subtree no. 2 can not be made a stylesheetView_CustomStyleSheetBlock_name node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pstylesheetView_CustomStyleSheetBlock;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpstylesheetView_CustomStyleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkstylesheetView_StyleSheetPairpstylesheetView_CustomStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkstylesheetView_StyleSheetPairpstylesheetView_CustomStyleSheetBlock_pairs (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_stylesheetView_StyleSheetPairpstylesheetView_CustomStyleSheetBlock_pairs _new;
#ifdef __cplusplus
	_new = new _TPLST_stylesheetView_StyleSheetPairpstylesheetView_CustomStyleSheetBlock_pairs;
#else
	_new = (_TPPLST_stylesheetView_StyleSheetPairpstylesheetView_CustomStyleSheetBlock_pairs) TreeNodeAlloc (sizeof (struct _TPLST_stylesheetView_StyleSheetPairpstylesheetView_CustomStyleSheetBlock_pairs));
#endif
	_new->_prod = RULELST_stylesheetView_StyleSheetPairpstylesheetView_CustomStyleSheetBlock_pairs;
	_new->_desc1 = (_TSPstylesheetView_StyleSheetPair)elem;
	_new->_desc2 = (_TSPLST_stylesheetView_CustomStyleSheetBlock_pairs)_new;
	{_TPPLST_stylesheetView_StyleSheetPairpstylesheetView_CustomStyleSheetBlock_pairs _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_stylesheetView_StyleSheetPairpstylesheetView_CustomStyleSheetBlock_pairs;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempstylesheetView_CustomStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempstylesheetView_CustomStyleSheetBlock_pairs (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBstylesheetView_StyleSheetPair)) 
		return _MkstylesheetView_StyleSheetPairpstylesheetView_CustomStyleSheetBlock_pairs (_coordref, MkstylesheetView_StyleSheetPair (_coordref, t));
	message (DEADLY, "a pstylesheetView_CustomStyleSheetBlock_pairs list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pstylesheetView_CustomStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pstylesheetView_CustomStyleSheetBlock_pairs (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_stylesheetView_CustomStyleSheetBlock_pairs)
	l =_ElempstylesheetView_CustomStyleSheetBlock_pairs (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_stylesheetView_CustomStyleSheetBlock_pairs)
	r = _ElempstylesheetView_CustomStyleSheetBlock_pairs (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pstylesheetView_CustomStyleSheetBlock_pairs)r)->_desc2);
((_TPPLST_0pstylesheetView_CustomStyleSheetBlock_pairs)r)->_desc2 = ((_TPPLST_0pstylesheetView_CustomStyleSheetBlock_pairs)l)->_desc2;
((_TPPLST_0pstylesheetView_CustomStyleSheetBlock_pairs)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_CustomStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpstylesheetView_CustomStyleSheetBlock_pairs (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpstylesheetView_CustomStyleSheetBlock_pairs root;
	_TPPLST_0pstylesheetView_CustomStyleSheetBlock_pairs list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pstylesheetView_CustomStyleSheetBlock_pairs)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_stylesheetView_CustomStyleSheetBlock_pairs)
		t = _ElempstylesheetView_CustomStyleSheetBlock_pairs (_coordref, t);
	remain = (_TPPLST_0pstylesheetView_CustomStyleSheetBlock_pairs)(((_TPPLST_0pstylesheetView_CustomStyleSheetBlock_pairs) t)->_desc2);
	((_TPPLST_0pstylesheetView_CustomStyleSheetBlock_pairs) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pstylesheetView_CustomStyleSheetBlock_pairs;
#else
list = (_TPPLST_0pstylesheetView_CustomStyleSheetBlock_pairs) TreeNodeAlloc (sizeof (struct _TPLST_0pstylesheetView_CustomStyleSheetBlock_pairs));
#endif
list->_prod = RULELST_0pstylesheetView_CustomStyleSheetBlock_pairs;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pstylesheetView_CustomStyleSheetBlock_pairs)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_stylesheetView_StyleSheetPairpstylesheetView_CustomStyleSheetBlock_pairs)
	{_TPPLST_stylesheetView_StyleSheetPairpstylesheetView_CustomStyleSheetBlock_pairs _currn = _currn;
	 _TERMACT_LST_stylesheetView_StyleSheetPairpstylesheetView_CustomStyleSheetBlock_pairs;
	} this code has been moved to _ElempstylesheetView_CustomStyleSheetBlock_pairs*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpstylesheetView_CustomStyleSheetBlock_pairs;
#else
root = (_TPPpstylesheetView_CustomStyleSheetBlock_pairs) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_CustomStyleSheetBlock_pairs));
#endif
root->_prod = RULEpstylesheetView_CustomStyleSheetBlock_pairs;
root->_desc1 = (_TSPLST_stylesheetView_CustomStyleSheetBlock_pairs)list;
{ _TPPpstylesheetView_CustomStyleSheetBlock_pairs _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pstylesheetView_CustomStyleSheetBlock_pairs;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_CustomStyleSheetBlock_name (POSITION *_coordref)
#else
NODEPTR MkpstylesheetView_CustomStyleSheetBlock_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpstylesheetView_CustomStyleSheetBlock_name _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_CustomStyleSheetBlock_name;
#else
_currn = (_TPPpstylesheetView_CustomStyleSheetBlock_name) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_CustomStyleSheetBlock_name));
#endif
_currn->_prod = RULEpstylesheetView_CustomStyleSheetBlock_name;
_SETCOORD(_currn)
_TERMACT_pstylesheetView_CustomStyleSheetBlock_name;
return ( (NODEPTR) _currn);
}/* MkpstylesheetView_CustomStyleSheetBlock_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_StyleSheetPair (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4)
#else
NODEPTR MkpstylesheetView_StyleSheetPair (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
#endif
{	_TPPpstylesheetView_StyleSheetPair _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_StyleSheetPair;
#else
_currn = (_TPPpstylesheetView_StyleSheetPair) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_StyleSheetPair));
#endif
_currn->_prod = RULEpstylesheetView_StyleSheetPair;
_currn->_desc1 = (_TSPstylesheetView_StyleSheetPair_category) MkstylesheetView_StyleSheetPair_category (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pstylesheetView_StyleSheetPair: root of subtree no. 1 can not be made a stylesheetView_StyleSheetPair_category node ", 0, _coordref);
_currn->_desc2 = (_TSPstylesheetView_StyleSheetPair_attribute) MkstylesheetView_StyleSheetPair_attribute (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pstylesheetView_StyleSheetPair: root of subtree no. 2 can not be made a stylesheetView_StyleSheetPair_attribute node ", 0, _coordref);
_currn->_desc3 = (_TSPstylesheetView_StyleSheetPair_valueCategory) MkstylesheetView_StyleSheetPair_valueCategory (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pstylesheetView_StyleSheetPair: root of subtree no. 3 can not be made a stylesheetView_StyleSheetPair_valueCategory node ", 0, _coordref);
_currn->_desc4 = (_TSPstylesheetView_StyleSheetPair_value) MkstylesheetView_StyleSheetPair_value (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pstylesheetView_StyleSheetPair: root of subtree no. 4 can not be made a stylesheetView_StyleSheetPair_value node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pstylesheetView_StyleSheetPair;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpstylesheetView_StyleSheetPair */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_StyleSheetPair_category (POSITION *_coordref)
#else
NODEPTR MkpstylesheetView_StyleSheetPair_category (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpstylesheetView_StyleSheetPair_category _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_StyleSheetPair_category;
#else
_currn = (_TPPpstylesheetView_StyleSheetPair_category) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_StyleSheetPair_category));
#endif
_currn->_prod = RULEpstylesheetView_StyleSheetPair_category;
_SETCOORD(_currn)
_TERMACT_pstylesheetView_StyleSheetPair_category;
return ( (NODEPTR) _currn);
}/* MkpstylesheetView_StyleSheetPair_category */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_StyleSheetPair_attribute (POSITION *_coordref)
#else
NODEPTR MkpstylesheetView_StyleSheetPair_attribute (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpstylesheetView_StyleSheetPair_attribute _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_StyleSheetPair_attribute;
#else
_currn = (_TPPpstylesheetView_StyleSheetPair_attribute) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_StyleSheetPair_attribute));
#endif
_currn->_prod = RULEpstylesheetView_StyleSheetPair_attribute;
_SETCOORD(_currn)
_TERMACT_pstylesheetView_StyleSheetPair_attribute;
return ( (NODEPTR) _currn);
}/* MkpstylesheetView_StyleSheetPair_attribute */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_StyleSheetPair_valueCategory (POSITION *_coordref)
#else
NODEPTR MkpstylesheetView_StyleSheetPair_valueCategory (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpstylesheetView_StyleSheetPair_valueCategory _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_StyleSheetPair_valueCategory;
#else
_currn = (_TPPpstylesheetView_StyleSheetPair_valueCategory) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_StyleSheetPair_valueCategory));
#endif
_currn->_prod = RULEpstylesheetView_StyleSheetPair_valueCategory;
_SETCOORD(_currn)
_TERMACT_pstylesheetView_StyleSheetPair_valueCategory;
return ( (NODEPTR) _currn);
}/* MkpstylesheetView_StyleSheetPair_valueCategory */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkstylesheetView_SizeValuepstylesheetView_StyleSheetPair_value (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkstylesheetView_SizeValuepstylesheetView_StyleSheetPair_value (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_stylesheetView_SizeValuepstylesheetView_StyleSheetPair_value _new;
#ifdef __cplusplus
	_new = new _TPLST_stylesheetView_SizeValuepstylesheetView_StyleSheetPair_value;
#else
	_new = (_TPPLST_stylesheetView_SizeValuepstylesheetView_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_stylesheetView_SizeValuepstylesheetView_StyleSheetPair_value));
#endif
	_new->_prod = RULELST_stylesheetView_SizeValuepstylesheetView_StyleSheetPair_value;
	_new->_desc1 = (_TSPstylesheetView_SizeValue)elem;
	_new->_desc2 = (_TSPLST_stylesheetView_StyleSheetPair_value)_new;
	{_TPPLST_stylesheetView_SizeValuepstylesheetView_StyleSheetPair_value _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_stylesheetView_SizeValuepstylesheetView_StyleSheetPair_value;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkstylesheetView_ColorValuepstylesheetView_StyleSheetPair_value (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkstylesheetView_ColorValuepstylesheetView_StyleSheetPair_value (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_stylesheetView_ColorValuepstylesheetView_StyleSheetPair_value _new;
#ifdef __cplusplus
	_new = new _TPLST_stylesheetView_ColorValuepstylesheetView_StyleSheetPair_value;
#else
	_new = (_TPPLST_stylesheetView_ColorValuepstylesheetView_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_stylesheetView_ColorValuepstylesheetView_StyleSheetPair_value));
#endif
	_new->_prod = RULELST_stylesheetView_ColorValuepstylesheetView_StyleSheetPair_value;
	_new->_desc1 = (_TSPstylesheetView_ColorValue)elem;
	_new->_desc2 = (_TSPLST_stylesheetView_StyleSheetPair_value)_new;
	{_TPPLST_stylesheetView_ColorValuepstylesheetView_StyleSheetPair_value _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_stylesheetView_ColorValuepstylesheetView_StyleSheetPair_value;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkstylesheetView_FamilyValuepstylesheetView_StyleSheetPair_value (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkstylesheetView_FamilyValuepstylesheetView_StyleSheetPair_value (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_stylesheetView_FamilyValuepstylesheetView_StyleSheetPair_value _new;
#ifdef __cplusplus
	_new = new _TPLST_stylesheetView_FamilyValuepstylesheetView_StyleSheetPair_value;
#else
	_new = (_TPPLST_stylesheetView_FamilyValuepstylesheetView_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_stylesheetView_FamilyValuepstylesheetView_StyleSheetPair_value));
#endif
	_new->_prod = RULELST_stylesheetView_FamilyValuepstylesheetView_StyleSheetPair_value;
	_new->_desc1 = (_TSPstylesheetView_FamilyValue)elem;
	_new->_desc2 = (_TSPLST_stylesheetView_StyleSheetPair_value)_new;
	{_TPPLST_stylesheetView_FamilyValuepstylesheetView_StyleSheetPair_value _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_stylesheetView_FamilyValuepstylesheetView_StyleSheetPair_value;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkstylesheetView_StyleValuepstylesheetView_StyleSheetPair_value (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkstylesheetView_StyleValuepstylesheetView_StyleSheetPair_value (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_stylesheetView_StyleValuepstylesheetView_StyleSheetPair_value _new;
#ifdef __cplusplus
	_new = new _TPLST_stylesheetView_StyleValuepstylesheetView_StyleSheetPair_value;
#else
	_new = (_TPPLST_stylesheetView_StyleValuepstylesheetView_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_stylesheetView_StyleValuepstylesheetView_StyleSheetPair_value));
#endif
	_new->_prod = RULELST_stylesheetView_StyleValuepstylesheetView_StyleSheetPair_value;
	_new->_desc1 = (_TSPstylesheetView_StyleValue)elem;
	_new->_desc2 = (_TSPLST_stylesheetView_StyleSheetPair_value)_new;
	{_TPPLST_stylesheetView_StyleValuepstylesheetView_StyleSheetPair_value _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_stylesheetView_StyleValuepstylesheetView_StyleSheetPair_value;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkstylesheetView_HeadingValuepstylesheetView_StyleSheetPair_value (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkstylesheetView_HeadingValuepstylesheetView_StyleSheetPair_value (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_stylesheetView_HeadingValuepstylesheetView_StyleSheetPair_value _new;
#ifdef __cplusplus
	_new = new _TPLST_stylesheetView_HeadingValuepstylesheetView_StyleSheetPair_value;
#else
	_new = (_TPPLST_stylesheetView_HeadingValuepstylesheetView_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_stylesheetView_HeadingValuepstylesheetView_StyleSheetPair_value));
#endif
	_new->_prod = RULELST_stylesheetView_HeadingValuepstylesheetView_StyleSheetPair_value;
	_new->_desc1 = (_TSPstylesheetView_HeadingValue)elem;
	_new->_desc2 = (_TSPLST_stylesheetView_StyleSheetPair_value)_new;
	{_TPPLST_stylesheetView_HeadingValuepstylesheetView_StyleSheetPair_value _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_stylesheetView_HeadingValuepstylesheetView_StyleSheetPair_value;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkstylesheetView_AlignmentValuepstylesheetView_StyleSheetPair_value (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkstylesheetView_AlignmentValuepstylesheetView_StyleSheetPair_value (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_stylesheetView_AlignmentValuepstylesheetView_StyleSheetPair_value _new;
#ifdef __cplusplus
	_new = new _TPLST_stylesheetView_AlignmentValuepstylesheetView_StyleSheetPair_value;
#else
	_new = (_TPPLST_stylesheetView_AlignmentValuepstylesheetView_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_stylesheetView_AlignmentValuepstylesheetView_StyleSheetPair_value));
#endif
	_new->_prod = RULELST_stylesheetView_AlignmentValuepstylesheetView_StyleSheetPair_value;
	_new->_desc1 = (_TSPstylesheetView_AlignmentValue)elem;
	_new->_desc2 = (_TSPLST_stylesheetView_StyleSheetPair_value)_new;
	{_TPPLST_stylesheetView_AlignmentValuepstylesheetView_StyleSheetPair_value _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_stylesheetView_AlignmentValuepstylesheetView_StyleSheetPair_value;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkstylesheetView_ListValuepstylesheetView_StyleSheetPair_value (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkstylesheetView_ListValuepstylesheetView_StyleSheetPair_value (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_stylesheetView_ListValuepstylesheetView_StyleSheetPair_value _new;
#ifdef __cplusplus
	_new = new _TPLST_stylesheetView_ListValuepstylesheetView_StyleSheetPair_value;
#else
	_new = (_TPPLST_stylesheetView_ListValuepstylesheetView_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_stylesheetView_ListValuepstylesheetView_StyleSheetPair_value));
#endif
	_new->_prod = RULELST_stylesheetView_ListValuepstylesheetView_StyleSheetPair_value;
	_new->_desc1 = (_TSPstylesheetView_ListValue)elem;
	_new->_desc2 = (_TSPLST_stylesheetView_StyleSheetPair_value)_new;
	{_TPPLST_stylesheetView_ListValuepstylesheetView_StyleSheetPair_value _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_stylesheetView_ListValuepstylesheetView_StyleSheetPair_value;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkstylesheetView_CustomValuepstylesheetView_StyleSheetPair_value (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkstylesheetView_CustomValuepstylesheetView_StyleSheetPair_value (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_stylesheetView_CustomValuepstylesheetView_StyleSheetPair_value _new;
#ifdef __cplusplus
	_new = new _TPLST_stylesheetView_CustomValuepstylesheetView_StyleSheetPair_value;
#else
	_new = (_TPPLST_stylesheetView_CustomValuepstylesheetView_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_stylesheetView_CustomValuepstylesheetView_StyleSheetPair_value));
#endif
	_new->_prod = RULELST_stylesheetView_CustomValuepstylesheetView_StyleSheetPair_value;
	_new->_desc1 = (_TSPstylesheetView_CustomValue)elem;
	_new->_desc2 = (_TSPLST_stylesheetView_StyleSheetPair_value)_new;
	{_TPPLST_stylesheetView_CustomValuepstylesheetView_StyleSheetPair_value _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_stylesheetView_CustomValuepstylesheetView_StyleSheetPair_value;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempstylesheetView_StyleSheetPair_value (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempstylesheetView_StyleSheetPair_value (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBstylesheetView_SizeValue)) 
		return _MkstylesheetView_SizeValuepstylesheetView_StyleSheetPair_value (_coordref, MkstylesheetView_SizeValue (_coordref, t));
	if (IsSymb (t, SYMBstylesheetView_ColorValue)) 
		return _MkstylesheetView_ColorValuepstylesheetView_StyleSheetPair_value (_coordref, MkstylesheetView_ColorValue (_coordref, t));
	if (IsSymb (t, SYMBstylesheetView_FamilyValue)) 
		return _MkstylesheetView_FamilyValuepstylesheetView_StyleSheetPair_value (_coordref, MkstylesheetView_FamilyValue (_coordref, t));
	if (IsSymb (t, SYMBstylesheetView_StyleValue)) 
		return _MkstylesheetView_StyleValuepstylesheetView_StyleSheetPair_value (_coordref, MkstylesheetView_StyleValue (_coordref, t));
	if (IsSymb (t, SYMBstylesheetView_HeadingValue)) 
		return _MkstylesheetView_HeadingValuepstylesheetView_StyleSheetPair_value (_coordref, MkstylesheetView_HeadingValue (_coordref, t));
	if (IsSymb (t, SYMBstylesheetView_AlignmentValue)) 
		return _MkstylesheetView_AlignmentValuepstylesheetView_StyleSheetPair_value (_coordref, MkstylesheetView_AlignmentValue (_coordref, t));
	if (IsSymb (t, SYMBstylesheetView_ListValue)) 
		return _MkstylesheetView_ListValuepstylesheetView_StyleSheetPair_value (_coordref, MkstylesheetView_ListValue (_coordref, t));
	if (IsSymb (t, SYMBstylesheetView_CustomValue)) 
		return _MkstylesheetView_CustomValuepstylesheetView_StyleSheetPair_value (_coordref, MkstylesheetView_CustomValue (_coordref, t));
	message (DEADLY, "a pstylesheetView_StyleSheetPair_value list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pstylesheetView_StyleSheetPair_value (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pstylesheetView_StyleSheetPair_value (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_stylesheetView_StyleSheetPair_value)
	l =_ElempstylesheetView_StyleSheetPair_value (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_stylesheetView_StyleSheetPair_value)
	r = _ElempstylesheetView_StyleSheetPair_value (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pstylesheetView_StyleSheetPair_value)r)->_desc2);
((_TPPLST_0pstylesheetView_StyleSheetPair_value)r)->_desc2 = ((_TPPLST_0pstylesheetView_StyleSheetPair_value)l)->_desc2;
((_TPPLST_0pstylesheetView_StyleSheetPair_value)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_StyleSheetPair_value (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpstylesheetView_StyleSheetPair_value (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpstylesheetView_StyleSheetPair_value root;
	_TPPLST_0pstylesheetView_StyleSheetPair_value list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pstylesheetView_StyleSheetPair_value)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_stylesheetView_StyleSheetPair_value)
		t = _ElempstylesheetView_StyleSheetPair_value (_coordref, t);
	remain = (_TPPLST_0pstylesheetView_StyleSheetPair_value)(((_TPPLST_0pstylesheetView_StyleSheetPair_value) t)->_desc2);
	((_TPPLST_0pstylesheetView_StyleSheetPair_value) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pstylesheetView_StyleSheetPair_value;
#else
list = (_TPPLST_0pstylesheetView_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPLST_0pstylesheetView_StyleSheetPair_value));
#endif
list->_prod = RULELST_0pstylesheetView_StyleSheetPair_value;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pstylesheetView_StyleSheetPair_value)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_stylesheetView_SizeValuepstylesheetView_StyleSheetPair_value)
	{_TPPLST_stylesheetView_SizeValuepstylesheetView_StyleSheetPair_value _currn = _currn;
	 _TERMACT_LST_stylesheetView_SizeValuepstylesheetView_StyleSheetPair_value;
	} this code has been moved to _ElempstylesheetView_StyleSheetPair_value*/
	/*if (_currn->_prod == RULELST_stylesheetView_ColorValuepstylesheetView_StyleSheetPair_value)
	{_TPPLST_stylesheetView_ColorValuepstylesheetView_StyleSheetPair_value _currn = _currn;
	 _TERMACT_LST_stylesheetView_ColorValuepstylesheetView_StyleSheetPair_value;
	} this code has been moved to _ElempstylesheetView_StyleSheetPair_value*/
	/*if (_currn->_prod == RULELST_stylesheetView_FamilyValuepstylesheetView_StyleSheetPair_value)
	{_TPPLST_stylesheetView_FamilyValuepstylesheetView_StyleSheetPair_value _currn = _currn;
	 _TERMACT_LST_stylesheetView_FamilyValuepstylesheetView_StyleSheetPair_value;
	} this code has been moved to _ElempstylesheetView_StyleSheetPair_value*/
	/*if (_currn->_prod == RULELST_stylesheetView_StyleValuepstylesheetView_StyleSheetPair_value)
	{_TPPLST_stylesheetView_StyleValuepstylesheetView_StyleSheetPair_value _currn = _currn;
	 _TERMACT_LST_stylesheetView_StyleValuepstylesheetView_StyleSheetPair_value;
	} this code has been moved to _ElempstylesheetView_StyleSheetPair_value*/
	/*if (_currn->_prod == RULELST_stylesheetView_HeadingValuepstylesheetView_StyleSheetPair_value)
	{_TPPLST_stylesheetView_HeadingValuepstylesheetView_StyleSheetPair_value _currn = _currn;
	 _TERMACT_LST_stylesheetView_HeadingValuepstylesheetView_StyleSheetPair_value;
	} this code has been moved to _ElempstylesheetView_StyleSheetPair_value*/
	/*if (_currn->_prod == RULELST_stylesheetView_AlignmentValuepstylesheetView_StyleSheetPair_value)
	{_TPPLST_stylesheetView_AlignmentValuepstylesheetView_StyleSheetPair_value _currn = _currn;
	 _TERMACT_LST_stylesheetView_AlignmentValuepstylesheetView_StyleSheetPair_value;
	} this code has been moved to _ElempstylesheetView_StyleSheetPair_value*/
	/*if (_currn->_prod == RULELST_stylesheetView_ListValuepstylesheetView_StyleSheetPair_value)
	{_TPPLST_stylesheetView_ListValuepstylesheetView_StyleSheetPair_value _currn = _currn;
	 _TERMACT_LST_stylesheetView_ListValuepstylesheetView_StyleSheetPair_value;
	} this code has been moved to _ElempstylesheetView_StyleSheetPair_value*/
	/*if (_currn->_prod == RULELST_stylesheetView_CustomValuepstylesheetView_StyleSheetPair_value)
	{_TPPLST_stylesheetView_CustomValuepstylesheetView_StyleSheetPair_value _currn = _currn;
	 _TERMACT_LST_stylesheetView_CustomValuepstylesheetView_StyleSheetPair_value;
	} this code has been moved to _ElempstylesheetView_StyleSheetPair_value*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpstylesheetView_StyleSheetPair_value;
#else
root = (_TPPpstylesheetView_StyleSheetPair_value) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_StyleSheetPair_value));
#endif
root->_prod = RULEpstylesheetView_StyleSheetPair_value;
root->_desc1 = (_TSPLST_stylesheetView_StyleSheetPair_value)list;
{ _TPPpstylesheetView_StyleSheetPair_value _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pstylesheetView_StyleSheetPair_value;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_SizeValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpstylesheetView_SizeValue (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpstylesheetView_SizeValue _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_SizeValue;
#else
_currn = (_TPPpstylesheetView_SizeValue) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_SizeValue));
#endif
_currn->_prod = RULEpstylesheetView_SizeValue;
_currn->_desc1 = (_TSPstylesheetView_SizeValue_measure) MkstylesheetView_SizeValue_measure (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pstylesheetView_SizeValue: root of subtree no. 1 can not be made a stylesheetView_SizeValue_measure node ", 0, _coordref);
_currn->_desc2 = (_TSPstylesheetView_SizeValue_value) MkstylesheetView_SizeValue_value (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pstylesheetView_SizeValue: root of subtree no. 2 can not be made a stylesheetView_SizeValue_value node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pstylesheetView_SizeValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpstylesheetView_SizeValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_SizeValue_measure (POSITION *_coordref)
#else
NODEPTR MkpstylesheetView_SizeValue_measure (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpstylesheetView_SizeValue_measure _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_SizeValue_measure;
#else
_currn = (_TPPpstylesheetView_SizeValue_measure) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_SizeValue_measure));
#endif
_currn->_prod = RULEpstylesheetView_SizeValue_measure;
_SETCOORD(_currn)
_TERMACT_pstylesheetView_SizeValue_measure;
return ( (NODEPTR) _currn);
}/* MkpstylesheetView_SizeValue_measure */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_SizeValue_value (POSITION *_coordref)
#else
NODEPTR MkpstylesheetView_SizeValue_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpstylesheetView_SizeValue_value _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_SizeValue_value;
#else
_currn = (_TPPpstylesheetView_SizeValue_value) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_SizeValue_value));
#endif
_currn->_prod = RULEpstylesheetView_SizeValue_value;
_SETCOORD(_currn)
_TERMACT_pstylesheetView_SizeValue_value;
return ( (NODEPTR) _currn);
}/* MkpstylesheetView_SizeValue_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_ColorValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpstylesheetView_ColorValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpstylesheetView_ColorValue _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_ColorValue;
#else
_currn = (_TPPpstylesheetView_ColorValue) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_ColorValue));
#endif
_currn->_prod = RULEpstylesheetView_ColorValue;
_currn->_desc1 = (_TSPstylesheetView_ColorValue_color) MkstylesheetView_ColorValue_color (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pstylesheetView_ColorValue: root of subtree no. 1 can not be made a stylesheetView_ColorValue_color node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pstylesheetView_ColorValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpstylesheetView_ColorValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_ColorValue_color (POSITION *_coordref)
#else
NODEPTR MkpstylesheetView_ColorValue_color (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpstylesheetView_ColorValue_color _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_ColorValue_color;
#else
_currn = (_TPPpstylesheetView_ColorValue_color) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_ColorValue_color));
#endif
_currn->_prod = RULEpstylesheetView_ColorValue_color;
_SETCOORD(_currn)
_TERMACT_pstylesheetView_ColorValue_color;
return ( (NODEPTR) _currn);
}/* MkpstylesheetView_ColorValue_color */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_FamilyValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpstylesheetView_FamilyValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpstylesheetView_FamilyValue _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_FamilyValue;
#else
_currn = (_TPPpstylesheetView_FamilyValue) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_FamilyValue));
#endif
_currn->_prod = RULEpstylesheetView_FamilyValue;
_currn->_desc1 = (_TSPstylesheetView_FamilyValue_family) MkstylesheetView_FamilyValue_family (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pstylesheetView_FamilyValue: root of subtree no. 1 can not be made a stylesheetView_FamilyValue_family node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pstylesheetView_FamilyValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpstylesheetView_FamilyValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_FamilyValue_family (POSITION *_coordref)
#else
NODEPTR MkpstylesheetView_FamilyValue_family (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpstylesheetView_FamilyValue_family _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_FamilyValue_family;
#else
_currn = (_TPPpstylesheetView_FamilyValue_family) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_FamilyValue_family));
#endif
_currn->_prod = RULEpstylesheetView_FamilyValue_family;
_SETCOORD(_currn)
_TERMACT_pstylesheetView_FamilyValue_family;
return ( (NODEPTR) _currn);
}/* MkpstylesheetView_FamilyValue_family */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_StyleValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpstylesheetView_StyleValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpstylesheetView_StyleValue _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_StyleValue;
#else
_currn = (_TPPpstylesheetView_StyleValue) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_StyleValue));
#endif
_currn->_prod = RULEpstylesheetView_StyleValue;
_currn->_desc1 = (_TSPstylesheetView_StyleValue_style) MkstylesheetView_StyleValue_style (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pstylesheetView_StyleValue: root of subtree no. 1 can not be made a stylesheetView_StyleValue_style node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pstylesheetView_StyleValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpstylesheetView_StyleValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_StyleValue_style (POSITION *_coordref)
#else
NODEPTR MkpstylesheetView_StyleValue_style (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpstylesheetView_StyleValue_style _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_StyleValue_style;
#else
_currn = (_TPPpstylesheetView_StyleValue_style) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_StyleValue_style));
#endif
_currn->_prod = RULEpstylesheetView_StyleValue_style;
_SETCOORD(_currn)
_TERMACT_pstylesheetView_StyleValue_style;
return ( (NODEPTR) _currn);
}/* MkpstylesheetView_StyleValue_style */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_HeadingValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpstylesheetView_HeadingValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpstylesheetView_HeadingValue _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_HeadingValue;
#else
_currn = (_TPPpstylesheetView_HeadingValue) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_HeadingValue));
#endif
_currn->_prod = RULEpstylesheetView_HeadingValue;
_currn->_desc1 = (_TSPstylesheetView_HeadingValue_heading) MkstylesheetView_HeadingValue_heading (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pstylesheetView_HeadingValue: root of subtree no. 1 can not be made a stylesheetView_HeadingValue_heading node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pstylesheetView_HeadingValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpstylesheetView_HeadingValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_HeadingValue_heading (POSITION *_coordref)
#else
NODEPTR MkpstylesheetView_HeadingValue_heading (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpstylesheetView_HeadingValue_heading _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_HeadingValue_heading;
#else
_currn = (_TPPpstylesheetView_HeadingValue_heading) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_HeadingValue_heading));
#endif
_currn->_prod = RULEpstylesheetView_HeadingValue_heading;
_SETCOORD(_currn)
_TERMACT_pstylesheetView_HeadingValue_heading;
return ( (NODEPTR) _currn);
}/* MkpstylesheetView_HeadingValue_heading */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_AlignmentValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpstylesheetView_AlignmentValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpstylesheetView_AlignmentValue _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_AlignmentValue;
#else
_currn = (_TPPpstylesheetView_AlignmentValue) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_AlignmentValue));
#endif
_currn->_prod = RULEpstylesheetView_AlignmentValue;
_currn->_desc1 = (_TSPstylesheetView_AlignmentValue_alignment) MkstylesheetView_AlignmentValue_alignment (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pstylesheetView_AlignmentValue: root of subtree no. 1 can not be made a stylesheetView_AlignmentValue_alignment node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pstylesheetView_AlignmentValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpstylesheetView_AlignmentValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_AlignmentValue_alignment (POSITION *_coordref)
#else
NODEPTR MkpstylesheetView_AlignmentValue_alignment (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpstylesheetView_AlignmentValue_alignment _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_AlignmentValue_alignment;
#else
_currn = (_TPPpstylesheetView_AlignmentValue_alignment) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_AlignmentValue_alignment));
#endif
_currn->_prod = RULEpstylesheetView_AlignmentValue_alignment;
_SETCOORD(_currn)
_TERMACT_pstylesheetView_AlignmentValue_alignment;
return ( (NODEPTR) _currn);
}/* MkpstylesheetView_AlignmentValue_alignment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_ListValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpstylesheetView_ListValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpstylesheetView_ListValue _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_ListValue;
#else
_currn = (_TPPpstylesheetView_ListValue) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_ListValue));
#endif
_currn->_prod = RULEpstylesheetView_ListValue;
_currn->_desc1 = (_TSPstylesheetView_ListValue_listed) MkstylesheetView_ListValue_listed (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pstylesheetView_ListValue: root of subtree no. 1 can not be made a stylesheetView_ListValue_listed node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pstylesheetView_ListValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpstylesheetView_ListValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_ListValue_listed (POSITION *_coordref)
#else
NODEPTR MkpstylesheetView_ListValue_listed (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpstylesheetView_ListValue_listed _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_ListValue_listed;
#else
_currn = (_TPPpstylesheetView_ListValue_listed) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_ListValue_listed));
#endif
_currn->_prod = RULEpstylesheetView_ListValue_listed;
_SETCOORD(_currn)
_TERMACT_pstylesheetView_ListValue_listed;
return ( (NODEPTR) _currn);
}/* MkpstylesheetView_ListValue_listed */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_CustomValue (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpstylesheetView_CustomValue (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpstylesheetView_CustomValue _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_CustomValue;
#else
_currn = (_TPPpstylesheetView_CustomValue) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_CustomValue));
#endif
_currn->_prod = RULEpstylesheetView_CustomValue;
_currn->_desc1 = (_TSPstylesheetView_CustomValue_value) MkstylesheetView_CustomValue_value (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pstylesheetView_CustomValue: root of subtree no. 1 can not be made a stylesheetView_CustomValue_value node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pstylesheetView_CustomValue;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpstylesheetView_CustomValue */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_CustomValue_value (POSITION *_coordref)
#else
NODEPTR MkpstylesheetView_CustomValue_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpstylesheetView_CustomValue_value _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_CustomValue_value;
#else
_currn = (_TPPpstylesheetView_CustomValue_value) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_CustomValue_value));
#endif
_currn->_prod = RULEpstylesheetView_CustomValue_value;
_SETCOORD(_currn)
_TERMACT_pstylesheetView_CustomValue_value;
return ( (NODEPTR) _currn);
}/* MkpstylesheetView_CustomValue_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_HtmlStyleSheetBlock (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpstylesheetView_HtmlStyleSheetBlock (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpstylesheetView_HtmlStyleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_HtmlStyleSheetBlock;
#else
_currn = (_TPPpstylesheetView_HtmlStyleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_HtmlStyleSheetBlock));
#endif
_currn->_prod = RULEpstylesheetView_HtmlStyleSheetBlock;
_currn->_desc1 = (_TSPstylesheetView_HtmlStyleSheetBlock_pairs) MkstylesheetView_HtmlStyleSheetBlock_pairs (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pstylesheetView_HtmlStyleSheetBlock: root of subtree no. 1 can not be made a stylesheetView_HtmlStyleSheetBlock_pairs node ", 0, _coordref);
_currn->_desc2 = (_TSPstylesheetView_HtmlStyleSheetBlock_name) MkstylesheetView_HtmlStyleSheetBlock_name (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pstylesheetView_HtmlStyleSheetBlock: root of subtree no. 2 can not be made a stylesheetView_HtmlStyleSheetBlock_name node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pstylesheetView_HtmlStyleSheetBlock;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpstylesheetView_HtmlStyleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkstylesheetView_StyleSheetPairpstylesheetView_HtmlStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkstylesheetView_StyleSheetPairpstylesheetView_HtmlStyleSheetBlock_pairs (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_stylesheetView_StyleSheetPairpstylesheetView_HtmlStyleSheetBlock_pairs _new;
#ifdef __cplusplus
	_new = new _TPLST_stylesheetView_StyleSheetPairpstylesheetView_HtmlStyleSheetBlock_pairs;
#else
	_new = (_TPPLST_stylesheetView_StyleSheetPairpstylesheetView_HtmlStyleSheetBlock_pairs) TreeNodeAlloc (sizeof (struct _TPLST_stylesheetView_StyleSheetPairpstylesheetView_HtmlStyleSheetBlock_pairs));
#endif
	_new->_prod = RULELST_stylesheetView_StyleSheetPairpstylesheetView_HtmlStyleSheetBlock_pairs;
	_new->_desc1 = (_TSPstylesheetView_StyleSheetPair)elem;
	_new->_desc2 = (_TSPLST_stylesheetView_HtmlStyleSheetBlock_pairs)_new;
	{_TPPLST_stylesheetView_StyleSheetPairpstylesheetView_HtmlStyleSheetBlock_pairs _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_stylesheetView_StyleSheetPairpstylesheetView_HtmlStyleSheetBlock_pairs;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempstylesheetView_HtmlStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempstylesheetView_HtmlStyleSheetBlock_pairs (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBstylesheetView_StyleSheetPair)) 
		return _MkstylesheetView_StyleSheetPairpstylesheetView_HtmlStyleSheetBlock_pairs (_coordref, MkstylesheetView_StyleSheetPair (_coordref, t));
	message (DEADLY, "a pstylesheetView_HtmlStyleSheetBlock_pairs list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pstylesheetView_HtmlStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pstylesheetView_HtmlStyleSheetBlock_pairs (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_stylesheetView_HtmlStyleSheetBlock_pairs)
	l =_ElempstylesheetView_HtmlStyleSheetBlock_pairs (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_stylesheetView_HtmlStyleSheetBlock_pairs)
	r = _ElempstylesheetView_HtmlStyleSheetBlock_pairs (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pstylesheetView_HtmlStyleSheetBlock_pairs)r)->_desc2);
((_TPPLST_0pstylesheetView_HtmlStyleSheetBlock_pairs)r)->_desc2 = ((_TPPLST_0pstylesheetView_HtmlStyleSheetBlock_pairs)l)->_desc2;
((_TPPLST_0pstylesheetView_HtmlStyleSheetBlock_pairs)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_HtmlStyleSheetBlock_pairs (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpstylesheetView_HtmlStyleSheetBlock_pairs (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpstylesheetView_HtmlStyleSheetBlock_pairs root;
	_TPPLST_0pstylesheetView_HtmlStyleSheetBlock_pairs list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pstylesheetView_HtmlStyleSheetBlock_pairs)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_stylesheetView_HtmlStyleSheetBlock_pairs)
		t = _ElempstylesheetView_HtmlStyleSheetBlock_pairs (_coordref, t);
	remain = (_TPPLST_0pstylesheetView_HtmlStyleSheetBlock_pairs)(((_TPPLST_0pstylesheetView_HtmlStyleSheetBlock_pairs) t)->_desc2);
	((_TPPLST_0pstylesheetView_HtmlStyleSheetBlock_pairs) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pstylesheetView_HtmlStyleSheetBlock_pairs;
#else
list = (_TPPLST_0pstylesheetView_HtmlStyleSheetBlock_pairs) TreeNodeAlloc (sizeof (struct _TPLST_0pstylesheetView_HtmlStyleSheetBlock_pairs));
#endif
list->_prod = RULELST_0pstylesheetView_HtmlStyleSheetBlock_pairs;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pstylesheetView_HtmlStyleSheetBlock_pairs)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_stylesheetView_StyleSheetPairpstylesheetView_HtmlStyleSheetBlock_pairs)
	{_TPPLST_stylesheetView_StyleSheetPairpstylesheetView_HtmlStyleSheetBlock_pairs _currn = _currn;
	 _TERMACT_LST_stylesheetView_StyleSheetPairpstylesheetView_HtmlStyleSheetBlock_pairs;
	} this code has been moved to _ElempstylesheetView_HtmlStyleSheetBlock_pairs*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpstylesheetView_HtmlStyleSheetBlock_pairs;
#else
root = (_TPPpstylesheetView_HtmlStyleSheetBlock_pairs) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_HtmlStyleSheetBlock_pairs));
#endif
root->_prod = RULEpstylesheetView_HtmlStyleSheetBlock_pairs;
root->_desc1 = (_TSPLST_stylesheetView_HtmlStyleSheetBlock_pairs)list;
{ _TPPpstylesheetView_HtmlStyleSheetBlock_pairs _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pstylesheetView_HtmlStyleSheetBlock_pairs;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpstylesheetView_HtmlStyleSheetBlock_name (POSITION *_coordref)
#else
NODEPTR MkpstylesheetView_HtmlStyleSheetBlock_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpstylesheetView_HtmlStyleSheetBlock_name _currn;
#ifdef __cplusplus
_currn = new _TPpstylesheetView_HtmlStyleSheetBlock_name;
#else
_currn = (_TPPpstylesheetView_HtmlStyleSheetBlock_name) TreeNodeAlloc (sizeof (struct _TPpstylesheetView_HtmlStyleSheetBlock_name));
#endif
_currn->_prod = RULEpstylesheetView_HtmlStyleSheetBlock_name;
_SETCOORD(_currn)
_TERMACT_pstylesheetView_HtmlStyleSheetBlock_name;
return ( (NODEPTR) _currn);
}/* MkpstylesheetView_HtmlStyleSheetBlock_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpVIEWROOT_webActionView (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR MkpVIEWROOT_webActionView (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPpVIEWROOT_webActionView _currn;
#ifdef __cplusplus
_currn = new _TPpVIEWROOT_webActionView;
#else
_currn = (_TPPpVIEWROOT_webActionView) TreeNodeAlloc (sizeof (struct _TPpVIEWROOT_webActionView));
#endif
_currn->_prod = RULEpVIEWROOT_webActionView;
_currn->_desc1 = (_TSPwebActionView_WebAction) MkwebActionView_WebAction (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pVIEWROOT_webActionView: root of subtree no. 1 can not be made a webActionView_WebAction node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_pVIEWROOT_webActionView;
return ( (NODEPTR) _currn);
}/* MkpVIEWROOT_webActionView */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_WebNodeFormalParam (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpwebActionView_WebNodeFormalParam (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpwebActionView_WebNodeFormalParam _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_WebNodeFormalParam;
#else
_currn = (_TPPpwebActionView_WebNodeFormalParam) TreeNodeAlloc (sizeof (struct _TPpwebActionView_WebNodeFormalParam));
#endif
_currn->_prod = RULEpwebActionView_WebNodeFormalParam;
_currn->_desc1 = (_TSPwebActionView_WebNodeFormalParam_name) MkwebActionView_WebNodeFormalParam_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_WebNodeFormalParam: root of subtree no. 1 can not be made a webActionView_WebNodeFormalParam_name node ", 0, _coordref);
_currn->_desc2 = (_TSPwebActionView_WebNodeFormalParam_keyword1) MkwebActionView_WebNodeFormalParam_keyword1 (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_WebNodeFormalParam: root of subtree no. 2 can not be made a webActionView_WebNodeFormalParam_keyword1 node ", 0, _coordref);
_currn->_desc3 = (_TSPwebActionView_WebNodeFormalParam_type) MkwebActionView_WebNodeFormalParam_type (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_WebNodeFormalParam: root of subtree no. 3 can not be made a webActionView_WebNodeFormalParam_type node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebActionView_WebNodeFormalParam;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebActionView_WebNodeFormalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_WebNodeFormalParam_name (POSITION *_coordref)
#else
NODEPTR MkpwebActionView_WebNodeFormalParam_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebActionView_WebNodeFormalParam_name _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_WebNodeFormalParam_name;
#else
_currn = (_TPPpwebActionView_WebNodeFormalParam_name) TreeNodeAlloc (sizeof (struct _TPpwebActionView_WebNodeFormalParam_name));
#endif
_currn->_prod = RULEpwebActionView_WebNodeFormalParam_name;
_SETCOORD(_currn)
_TERMACT_pwebActionView_WebNodeFormalParam_name;
return ( (NODEPTR) _currn);
}/* MkpwebActionView_WebNodeFormalParam_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_WebNodeFormalParam_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpwebActionView_WebNodeFormalParam_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebActionView_WebNodeFormalParam_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_WebNodeFormalParam_keyword1;
#else
_currn = (_TPPpwebActionView_WebNodeFormalParam_keyword1) TreeNodeAlloc (sizeof (struct _TPpwebActionView_WebNodeFormalParam_keyword1));
#endif
_currn->_prod = RULEpwebActionView_WebNodeFormalParam_keyword1;
_SETCOORD(_currn)
_TERMACT_pwebActionView_WebNodeFormalParam_keyword1;
return ( (NODEPTR) _currn);
}/* MkpwebActionView_WebNodeFormalParam_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebActionView_DBStringpwebActionView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebActionView_DBStringpwebActionView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webActionView_DBStringpwebActionView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_webActionView_DBStringpwebActionView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_webActionView_DBStringpwebActionView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_webActionView_DBStringpwebActionView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_webActionView_DBStringpwebActionView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSPwebActionView_DBString)elem;
	_new->_desc2 = (_TSPLST_webActionView_WebNodeFormalParam_type)_new;
	{_TPPLST_webActionView_DBStringpwebActionView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webActionView_DBStringpwebActionView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebActionView_DBIntegerpwebActionView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebActionView_DBIntegerpwebActionView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webActionView_DBIntegerpwebActionView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_webActionView_DBIntegerpwebActionView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_webActionView_DBIntegerpwebActionView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_webActionView_DBIntegerpwebActionView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_webActionView_DBIntegerpwebActionView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSPwebActionView_DBInteger)elem;
	_new->_desc2 = (_TSPLST_webActionView_WebNodeFormalParam_type)_new;
	{_TPPLST_webActionView_DBIntegerpwebActionView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webActionView_DBIntegerpwebActionView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebActionView_DBFloatpwebActionView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebActionView_DBFloatpwebActionView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webActionView_DBFloatpwebActionView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_webActionView_DBFloatpwebActionView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_webActionView_DBFloatpwebActionView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_webActionView_DBFloatpwebActionView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_webActionView_DBFloatpwebActionView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSPwebActionView_DBFloat)elem;
	_new->_desc2 = (_TSPLST_webActionView_WebNodeFormalParam_type)_new;
	{_TPPLST_webActionView_DBFloatpwebActionView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webActionView_DBFloatpwebActionView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebActionView_DBDoublepwebActionView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebActionView_DBDoublepwebActionView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webActionView_DBDoublepwebActionView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_webActionView_DBDoublepwebActionView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_webActionView_DBDoublepwebActionView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_webActionView_DBDoublepwebActionView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_webActionView_DBDoublepwebActionView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSPwebActionView_DBDouble)elem;
	_new->_desc2 = (_TSPLST_webActionView_WebNodeFormalParam_type)_new;
	{_TPPLST_webActionView_DBDoublepwebActionView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webActionView_DBDoublepwebActionView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebActionView_DBBigIntpwebActionView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebActionView_DBBigIntpwebActionView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webActionView_DBBigIntpwebActionView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_webActionView_DBBigIntpwebActionView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_webActionView_DBBigIntpwebActionView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_webActionView_DBBigIntpwebActionView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_webActionView_DBBigIntpwebActionView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSPwebActionView_DBBigInt)elem;
	_new->_desc2 = (_TSPLST_webActionView_WebNodeFormalParam_type)_new;
	{_TPPLST_webActionView_DBBigIntpwebActionView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webActionView_DBBigIntpwebActionView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebActionView_DBDatepwebActionView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebActionView_DBDatepwebActionView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webActionView_DBDatepwebActionView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_webActionView_DBDatepwebActionView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_webActionView_DBDatepwebActionView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_webActionView_DBDatepwebActionView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_webActionView_DBDatepwebActionView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSPwebActionView_DBDate)elem;
	_new->_desc2 = (_TSPLST_webActionView_WebNodeFormalParam_type)_new;
	{_TPPLST_webActionView_DBDatepwebActionView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webActionView_DBDatepwebActionView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebActionView_DBTimeStamppwebActionView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebActionView_DBTimeStamppwebActionView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webActionView_DBTimeStamppwebActionView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_webActionView_DBTimeStamppwebActionView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_webActionView_DBTimeStamppwebActionView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_webActionView_DBTimeStamppwebActionView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_webActionView_DBTimeStamppwebActionView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSPwebActionView_DBTimeStamp)elem;
	_new->_desc2 = (_TSPLST_webActionView_WebNodeFormalParam_type)_new;
	{_TPPLST_webActionView_DBTimeStamppwebActionView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webActionView_DBTimeStamppwebActionView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebActionView_DBBooleanpwebActionView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebActionView_DBBooleanpwebActionView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webActionView_DBBooleanpwebActionView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_webActionView_DBBooleanpwebActionView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_webActionView_DBBooleanpwebActionView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_webActionView_DBBooleanpwebActionView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_webActionView_DBBooleanpwebActionView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSPwebActionView_DBBoolean)elem;
	_new->_desc2 = (_TSPLST_webActionView_WebNodeFormalParam_type)_new;
	{_TPPLST_webActionView_DBBooleanpwebActionView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webActionView_DBBooleanpwebActionView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebActionView_DBBlobpwebActionView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebActionView_DBBlobpwebActionView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webActionView_DBBlobpwebActionView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_webActionView_DBBlobpwebActionView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_webActionView_DBBlobpwebActionView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_webActionView_DBBlobpwebActionView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_webActionView_DBBlobpwebActionView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSPwebActionView_DBBlob)elem;
	_new->_desc2 = (_TSPLST_webActionView_WebNodeFormalParam_type)_new;
	{_TPPLST_webActionView_DBBlobpwebActionView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webActionView_DBBlobpwebActionView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempwebActionView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempwebActionView_WebNodeFormalParam_type (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBwebActionView_DBString)) 
		return _MkwebActionView_DBStringpwebActionView_WebNodeFormalParam_type (_coordref, MkwebActionView_DBString (_coordref, t));
	if (IsSymb (t, SYMBwebActionView_DBInteger)) 
		return _MkwebActionView_DBIntegerpwebActionView_WebNodeFormalParam_type (_coordref, MkwebActionView_DBInteger (_coordref, t));
	if (IsSymb (t, SYMBwebActionView_DBFloat)) 
		return _MkwebActionView_DBFloatpwebActionView_WebNodeFormalParam_type (_coordref, MkwebActionView_DBFloat (_coordref, t));
	if (IsSymb (t, SYMBwebActionView_DBDouble)) 
		return _MkwebActionView_DBDoublepwebActionView_WebNodeFormalParam_type (_coordref, MkwebActionView_DBDouble (_coordref, t));
	if (IsSymb (t, SYMBwebActionView_DBBigInt)) 
		return _MkwebActionView_DBBigIntpwebActionView_WebNodeFormalParam_type (_coordref, MkwebActionView_DBBigInt (_coordref, t));
	if (IsSymb (t, SYMBwebActionView_DBDate)) 
		return _MkwebActionView_DBDatepwebActionView_WebNodeFormalParam_type (_coordref, MkwebActionView_DBDate (_coordref, t));
	if (IsSymb (t, SYMBwebActionView_DBTimeStamp)) 
		return _MkwebActionView_DBTimeStamppwebActionView_WebNodeFormalParam_type (_coordref, MkwebActionView_DBTimeStamp (_coordref, t));
	if (IsSymb (t, SYMBwebActionView_DBBoolean)) 
		return _MkwebActionView_DBBooleanpwebActionView_WebNodeFormalParam_type (_coordref, MkwebActionView_DBBoolean (_coordref, t));
	if (IsSymb (t, SYMBwebActionView_DBBlob)) 
		return _MkwebActionView_DBBlobpwebActionView_WebNodeFormalParam_type (_coordref, MkwebActionView_DBBlob (_coordref, t));
	message (DEADLY, "a pwebActionView_WebNodeFormalParam_type list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pwebActionView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pwebActionView_WebNodeFormalParam_type (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_webActionView_WebNodeFormalParam_type)
	l =_ElempwebActionView_WebNodeFormalParam_type (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_webActionView_WebNodeFormalParam_type)
	r = _ElempwebActionView_WebNodeFormalParam_type (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pwebActionView_WebNodeFormalParam_type)r)->_desc2);
((_TPPLST_0pwebActionView_WebNodeFormalParam_type)r)->_desc2 = ((_TPPLST_0pwebActionView_WebNodeFormalParam_type)l)->_desc2;
((_TPPLST_0pwebActionView_WebNodeFormalParam_type)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpwebActionView_WebNodeFormalParam_type (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpwebActionView_WebNodeFormalParam_type root;
	_TPPLST_0pwebActionView_WebNodeFormalParam_type list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pwebActionView_WebNodeFormalParam_type)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_webActionView_WebNodeFormalParam_type)
		t = _ElempwebActionView_WebNodeFormalParam_type (_coordref, t);
	remain = (_TPPLST_0pwebActionView_WebNodeFormalParam_type)(((_TPPLST_0pwebActionView_WebNodeFormalParam_type) t)->_desc2);
	((_TPPLST_0pwebActionView_WebNodeFormalParam_type) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pwebActionView_WebNodeFormalParam_type;
#else
list = (_TPPLST_0pwebActionView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_0pwebActionView_WebNodeFormalParam_type));
#endif
list->_prod = RULELST_0pwebActionView_WebNodeFormalParam_type;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pwebActionView_WebNodeFormalParam_type)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_webActionView_DBStringpwebActionView_WebNodeFormalParam_type)
	{_TPPLST_webActionView_DBStringpwebActionView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_webActionView_DBStringpwebActionView_WebNodeFormalParam_type;
	} this code has been moved to _ElempwebActionView_WebNodeFormalParam_type*/
	/*if (_currn->_prod == RULELST_webActionView_DBIntegerpwebActionView_WebNodeFormalParam_type)
	{_TPPLST_webActionView_DBIntegerpwebActionView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_webActionView_DBIntegerpwebActionView_WebNodeFormalParam_type;
	} this code has been moved to _ElempwebActionView_WebNodeFormalParam_type*/
	/*if (_currn->_prod == RULELST_webActionView_DBFloatpwebActionView_WebNodeFormalParam_type)
	{_TPPLST_webActionView_DBFloatpwebActionView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_webActionView_DBFloatpwebActionView_WebNodeFormalParam_type;
	} this code has been moved to _ElempwebActionView_WebNodeFormalParam_type*/
	/*if (_currn->_prod == RULELST_webActionView_DBDoublepwebActionView_WebNodeFormalParam_type)
	{_TPPLST_webActionView_DBDoublepwebActionView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_webActionView_DBDoublepwebActionView_WebNodeFormalParam_type;
	} this code has been moved to _ElempwebActionView_WebNodeFormalParam_type*/
	/*if (_currn->_prod == RULELST_webActionView_DBBigIntpwebActionView_WebNodeFormalParam_type)
	{_TPPLST_webActionView_DBBigIntpwebActionView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_webActionView_DBBigIntpwebActionView_WebNodeFormalParam_type;
	} this code has been moved to _ElempwebActionView_WebNodeFormalParam_type*/
	/*if (_currn->_prod == RULELST_webActionView_DBDatepwebActionView_WebNodeFormalParam_type)
	{_TPPLST_webActionView_DBDatepwebActionView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_webActionView_DBDatepwebActionView_WebNodeFormalParam_type;
	} this code has been moved to _ElempwebActionView_WebNodeFormalParam_type*/
	/*if (_currn->_prod == RULELST_webActionView_DBTimeStamppwebActionView_WebNodeFormalParam_type)
	{_TPPLST_webActionView_DBTimeStamppwebActionView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_webActionView_DBTimeStamppwebActionView_WebNodeFormalParam_type;
	} this code has been moved to _ElempwebActionView_WebNodeFormalParam_type*/
	/*if (_currn->_prod == RULELST_webActionView_DBBooleanpwebActionView_WebNodeFormalParam_type)
	{_TPPLST_webActionView_DBBooleanpwebActionView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_webActionView_DBBooleanpwebActionView_WebNodeFormalParam_type;
	} this code has been moved to _ElempwebActionView_WebNodeFormalParam_type*/
	/*if (_currn->_prod == RULELST_webActionView_DBBlobpwebActionView_WebNodeFormalParam_type)
	{_TPPLST_webActionView_DBBlobpwebActionView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_webActionView_DBBlobpwebActionView_WebNodeFormalParam_type;
	} this code has been moved to _ElempwebActionView_WebNodeFormalParam_type*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpwebActionView_WebNodeFormalParam_type;
#else
root = (_TPPpwebActionView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPpwebActionView_WebNodeFormalParam_type));
#endif
root->_prod = RULEpwebActionView_WebNodeFormalParam_type;
root->_desc1 = (_TSPLST_webActionView_WebNodeFormalParam_type)list;
{ _TPPpwebActionView_WebNodeFormalParam_type _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pwebActionView_WebNodeFormalParam_type;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_WebNodeActualParam (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpwebActionView_WebNodeActualParam (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpwebActionView_WebNodeActualParam _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_WebNodeActualParam;
#else
_currn = (_TPPpwebActionView_WebNodeActualParam) TreeNodeAlloc (sizeof (struct _TPpwebActionView_WebNodeActualParam));
#endif
_currn->_prod = RULEpwebActionView_WebNodeActualParam;
_currn->_desc1 = (_TSPwebActionView_WebNodeActualParam_formalParam) MkwebActionView_WebNodeActualParam_formalParam (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_WebNodeActualParam: root of subtree no. 1 can not be made a webActionView_WebNodeActualParam_formalParam node ", 0, _coordref);
_currn->_desc2 = (_TSPwebActionView_WebNodeActualParam_keyword1) MkwebActionView_WebNodeActualParam_keyword1 (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_WebNodeActualParam: root of subtree no. 2 can not be made a webActionView_WebNodeActualParam_keyword1 node ", 0, _coordref);
_currn->_desc3 = (_TSPwebActionView_WebNodeActualParam_expressions) MkwebActionView_WebNodeActualParam_expressions (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_WebNodeActualParam: root of subtree no. 3 can not be made a webActionView_WebNodeActualParam_expressions node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebActionView_WebNodeActualParam;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebActionView_WebNodeActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_WebNodeActualParam_formalParam (POSITION *_coordref)
#else
NODEPTR MkpwebActionView_WebNodeActualParam_formalParam (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebActionView_WebNodeActualParam_formalParam _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_WebNodeActualParam_formalParam;
#else
_currn = (_TPPpwebActionView_WebNodeActualParam_formalParam) TreeNodeAlloc (sizeof (struct _TPpwebActionView_WebNodeActualParam_formalParam));
#endif
_currn->_prod = RULEpwebActionView_WebNodeActualParam_formalParam;
_SETCOORD(_currn)
_TERMACT_pwebActionView_WebNodeActualParam_formalParam;
return ( (NODEPTR) _currn);
}/* MkpwebActionView_WebNodeActualParam_formalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_WebNodeActualParam_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpwebActionView_WebNodeActualParam_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebActionView_WebNodeActualParam_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_WebNodeActualParam_keyword1;
#else
_currn = (_TPPpwebActionView_WebNodeActualParam_keyword1) TreeNodeAlloc (sizeof (struct _TPpwebActionView_WebNodeActualParam_keyword1));
#endif
_currn->_prod = RULEpwebActionView_WebNodeActualParam_keyword1;
_SETCOORD(_currn)
_TERMACT_pwebActionView_WebNodeActualParam_keyword1;
return ( (NODEPTR) _currn);
}/* MkpwebActionView_WebNodeActualParam_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebActionView_WebNodeConstantExpressionpwebActionView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebActionView_WebNodeConstantExpressionpwebActionView_WebNodeActualParam_expressions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webActionView_WebNodeConstantExpressionpwebActionView_WebNodeActualParam_expressions _new;
#ifdef __cplusplus
	_new = new _TPLST_webActionView_WebNodeConstantExpressionpwebActionView_WebNodeActualParam_expressions;
#else
	_new = (_TPPLST_webActionView_WebNodeConstantExpressionpwebActionView_WebNodeActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_webActionView_WebNodeConstantExpressionpwebActionView_WebNodeActualParam_expressions));
#endif
	_new->_prod = RULELST_webActionView_WebNodeConstantExpressionpwebActionView_WebNodeActualParam_expressions;
	_new->_desc1 = (_TSPwebActionView_WebNodeConstantExpression)elem;
	_new->_desc2 = (_TSPLST_webActionView_WebNodeActualParam_expressions)_new;
	{_TPPLST_webActionView_WebNodeConstantExpressionpwebActionView_WebNodeActualParam_expressions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webActionView_WebNodeConstantExpressionpwebActionView_WebNodeActualParam_expressions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebActionView_WebNodeVariableExpressionpwebActionView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebActionView_WebNodeVariableExpressionpwebActionView_WebNodeActualParam_expressions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webActionView_WebNodeVariableExpressionpwebActionView_WebNodeActualParam_expressions _new;
#ifdef __cplusplus
	_new = new _TPLST_webActionView_WebNodeVariableExpressionpwebActionView_WebNodeActualParam_expressions;
#else
	_new = (_TPPLST_webActionView_WebNodeVariableExpressionpwebActionView_WebNodeActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_webActionView_WebNodeVariableExpressionpwebActionView_WebNodeActualParam_expressions));
#endif
	_new->_prod = RULELST_webActionView_WebNodeVariableExpressionpwebActionView_WebNodeActualParam_expressions;
	_new->_desc1 = (_TSPwebActionView_WebNodeVariableExpression)elem;
	_new->_desc2 = (_TSPLST_webActionView_WebNodeActualParam_expressions)_new;
	{_TPPLST_webActionView_WebNodeVariableExpressionpwebActionView_WebNodeActualParam_expressions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webActionView_WebNodeVariableExpressionpwebActionView_WebNodeActualParam_expressions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebActionView_WebNodeColumnExpressionpwebActionView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebActionView_WebNodeColumnExpressionpwebActionView_WebNodeActualParam_expressions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webActionView_WebNodeColumnExpressionpwebActionView_WebNodeActualParam_expressions _new;
#ifdef __cplusplus
	_new = new _TPLST_webActionView_WebNodeColumnExpressionpwebActionView_WebNodeActualParam_expressions;
#else
	_new = (_TPPLST_webActionView_WebNodeColumnExpressionpwebActionView_WebNodeActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_webActionView_WebNodeColumnExpressionpwebActionView_WebNodeActualParam_expressions));
#endif
	_new->_prod = RULELST_webActionView_WebNodeColumnExpressionpwebActionView_WebNodeActualParam_expressions;
	_new->_desc1 = (_TSPwebActionView_WebNodeColumnExpression)elem;
	_new->_desc2 = (_TSPLST_webActionView_WebNodeActualParam_expressions)_new;
	{_TPPLST_webActionView_WebNodeColumnExpressionpwebActionView_WebNodeActualParam_expressions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webActionView_WebNodeColumnExpressionpwebActionView_WebNodeActualParam_expressions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempwebActionView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempwebActionView_WebNodeActualParam_expressions (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBwebActionView_WebNodeConstantExpression)) 
		return _MkwebActionView_WebNodeConstantExpressionpwebActionView_WebNodeActualParam_expressions (_coordref, MkwebActionView_WebNodeConstantExpression (_coordref, t));
	if (IsSymb (t, SYMBwebActionView_WebNodeVariableExpression)) 
		return _MkwebActionView_WebNodeVariableExpressionpwebActionView_WebNodeActualParam_expressions (_coordref, MkwebActionView_WebNodeVariableExpression (_coordref, t));
	if (IsSymb (t, SYMBwebActionView_WebNodeColumnExpression)) 
		return _MkwebActionView_WebNodeColumnExpressionpwebActionView_WebNodeActualParam_expressions (_coordref, MkwebActionView_WebNodeColumnExpression (_coordref, t));
	message (DEADLY, "a pwebActionView_WebNodeActualParam_expressions list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pwebActionView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pwebActionView_WebNodeActualParam_expressions (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_webActionView_WebNodeActualParam_expressions)
	l =_ElempwebActionView_WebNodeActualParam_expressions (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_webActionView_WebNodeActualParam_expressions)
	r = _ElempwebActionView_WebNodeActualParam_expressions (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pwebActionView_WebNodeActualParam_expressions)r)->_desc2);
((_TPPLST_0pwebActionView_WebNodeActualParam_expressions)r)->_desc2 = ((_TPPLST_0pwebActionView_WebNodeActualParam_expressions)l)->_desc2;
((_TPPLST_0pwebActionView_WebNodeActualParam_expressions)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpwebActionView_WebNodeActualParam_expressions (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpwebActionView_WebNodeActualParam_expressions root;
	_TPPLST_0pwebActionView_WebNodeActualParam_expressions list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pwebActionView_WebNodeActualParam_expressions)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_webActionView_WebNodeActualParam_expressions)
		t = _ElempwebActionView_WebNodeActualParam_expressions (_coordref, t);
	remain = (_TPPLST_0pwebActionView_WebNodeActualParam_expressions)(((_TPPLST_0pwebActionView_WebNodeActualParam_expressions) t)->_desc2);
	((_TPPLST_0pwebActionView_WebNodeActualParam_expressions) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pwebActionView_WebNodeActualParam_expressions;
#else
list = (_TPPLST_0pwebActionView_WebNodeActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_0pwebActionView_WebNodeActualParam_expressions));
#endif
list->_prod = RULELST_0pwebActionView_WebNodeActualParam_expressions;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pwebActionView_WebNodeActualParam_expressions)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_webActionView_WebNodeConstantExpressionpwebActionView_WebNodeActualParam_expressions)
	{_TPPLST_webActionView_WebNodeConstantExpressionpwebActionView_WebNodeActualParam_expressions _currn = _currn;
	 _TERMACT_LST_webActionView_WebNodeConstantExpressionpwebActionView_WebNodeActualParam_expressions;
	} this code has been moved to _ElempwebActionView_WebNodeActualParam_expressions*/
	/*if (_currn->_prod == RULELST_webActionView_WebNodeVariableExpressionpwebActionView_WebNodeActualParam_expressions)
	{_TPPLST_webActionView_WebNodeVariableExpressionpwebActionView_WebNodeActualParam_expressions _currn = _currn;
	 _TERMACT_LST_webActionView_WebNodeVariableExpressionpwebActionView_WebNodeActualParam_expressions;
	} this code has been moved to _ElempwebActionView_WebNodeActualParam_expressions*/
	/*if (_currn->_prod == RULELST_webActionView_WebNodeColumnExpressionpwebActionView_WebNodeActualParam_expressions)
	{_TPPLST_webActionView_WebNodeColumnExpressionpwebActionView_WebNodeActualParam_expressions _currn = _currn;
	 _TERMACT_LST_webActionView_WebNodeColumnExpressionpwebActionView_WebNodeActualParam_expressions;
	} this code has been moved to _ElempwebActionView_WebNodeActualParam_expressions*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpwebActionView_WebNodeActualParam_expressions;
#else
root = (_TPPpwebActionView_WebNodeActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPpwebActionView_WebNodeActualParam_expressions));
#endif
root->_prod = RULEpwebActionView_WebNodeActualParam_expressions;
root->_desc1 = (_TSPLST_webActionView_WebNodeActualParam_expressions)list;
{ _TPPpwebActionView_WebNodeActualParam_expressions _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pwebActionView_WebNodeActualParam_expressions;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_WebNodeConstantExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpwebActionView_WebNodeConstantExpression (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpwebActionView_WebNodeConstantExpression _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_WebNodeConstantExpression;
#else
_currn = (_TPPpwebActionView_WebNodeConstantExpression) TreeNodeAlloc (sizeof (struct _TPpwebActionView_WebNodeConstantExpression));
#endif
_currn->_prod = RULEpwebActionView_WebNodeConstantExpression;
_currn->_desc1 = (_TSPwebActionView_WebNodeConstantExpression_keyword1) MkwebActionView_WebNodeConstantExpression_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_WebNodeConstantExpression: root of subtree no. 1 can not be made a webActionView_WebNodeConstantExpression_keyword1 node ", 0, _coordref);
_currn->_desc2 = (_TSPwebActionView_WebNodeConstantExpression_value) MkwebActionView_WebNodeConstantExpression_value (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_WebNodeConstantExpression: root of subtree no. 2 can not be made a webActionView_WebNodeConstantExpression_value node ", 0, _coordref);
_currn->_desc3 = (_TSPwebActionView_WebNodeConstantExpression_keyword2) MkwebActionView_WebNodeConstantExpression_keyword2 (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_WebNodeConstantExpression: root of subtree no. 3 can not be made a webActionView_WebNodeConstantExpression_keyword2 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebActionView_WebNodeConstantExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebActionView_WebNodeConstantExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_WebNodeConstantExpression_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpwebActionView_WebNodeConstantExpression_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebActionView_WebNodeConstantExpression_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_WebNodeConstantExpression_keyword1;
#else
_currn = (_TPPpwebActionView_WebNodeConstantExpression_keyword1) TreeNodeAlloc (sizeof (struct _TPpwebActionView_WebNodeConstantExpression_keyword1));
#endif
_currn->_prod = RULEpwebActionView_WebNodeConstantExpression_keyword1;
_SETCOORD(_currn)
_TERMACT_pwebActionView_WebNodeConstantExpression_keyword1;
return ( (NODEPTR) _currn);
}/* MkpwebActionView_WebNodeConstantExpression_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_WebNodeConstantExpression_value (POSITION *_coordref)
#else
NODEPTR MkpwebActionView_WebNodeConstantExpression_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebActionView_WebNodeConstantExpression_value _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_WebNodeConstantExpression_value;
#else
_currn = (_TPPpwebActionView_WebNodeConstantExpression_value) TreeNodeAlloc (sizeof (struct _TPpwebActionView_WebNodeConstantExpression_value));
#endif
_currn->_prod = RULEpwebActionView_WebNodeConstantExpression_value;
_SETCOORD(_currn)
_TERMACT_pwebActionView_WebNodeConstantExpression_value;
return ( (NODEPTR) _currn);
}/* MkpwebActionView_WebNodeConstantExpression_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_WebNodeConstantExpression_keyword2 (POSITION *_coordref)
#else
NODEPTR MkpwebActionView_WebNodeConstantExpression_keyword2 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebActionView_WebNodeConstantExpression_keyword2 _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_WebNodeConstantExpression_keyword2;
#else
_currn = (_TPPpwebActionView_WebNodeConstantExpression_keyword2) TreeNodeAlloc (sizeof (struct _TPpwebActionView_WebNodeConstantExpression_keyword2));
#endif
_currn->_prod = RULEpwebActionView_WebNodeConstantExpression_keyword2;
_SETCOORD(_currn)
_TERMACT_pwebActionView_WebNodeConstantExpression_keyword2;
return ( (NODEPTR) _currn);
}/* MkpwebActionView_WebNodeConstantExpression_keyword2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_WebNodeVariableExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpwebActionView_WebNodeVariableExpression (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpwebActionView_WebNodeVariableExpression _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_WebNodeVariableExpression;
#else
_currn = (_TPPpwebActionView_WebNodeVariableExpression) TreeNodeAlloc (sizeof (struct _TPpwebActionView_WebNodeVariableExpression));
#endif
_currn->_prod = RULEpwebActionView_WebNodeVariableExpression;
_currn->_desc1 = (_TSPwebActionView_WebNodeVariableExpression_content) MkwebActionView_WebNodeVariableExpression_content (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_WebNodeVariableExpression: root of subtree no. 1 can not be made a webActionView_WebNodeVariableExpression_content node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebActionView_WebNodeVariableExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebActionView_WebNodeVariableExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_WebNodeVariableExpression_content (POSITION *_coordref)
#else
NODEPTR MkpwebActionView_WebNodeVariableExpression_content (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebActionView_WebNodeVariableExpression_content _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_WebNodeVariableExpression_content;
#else
_currn = (_TPPpwebActionView_WebNodeVariableExpression_content) TreeNodeAlloc (sizeof (struct _TPpwebActionView_WebNodeVariableExpression_content));
#endif
_currn->_prod = RULEpwebActionView_WebNodeVariableExpression_content;
_SETCOORD(_currn)
_TERMACT_pwebActionView_WebNodeVariableExpression_content;
return ( (NODEPTR) _currn);
}/* MkpwebActionView_WebNodeVariableExpression_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_WebNodeColumnExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpwebActionView_WebNodeColumnExpression (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpwebActionView_WebNodeColumnExpression _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_WebNodeColumnExpression;
#else
_currn = (_TPPpwebActionView_WebNodeColumnExpression) TreeNodeAlloc (sizeof (struct _TPpwebActionView_WebNodeColumnExpression));
#endif
_currn->_prod = RULEpwebActionView_WebNodeColumnExpression;
_currn->_desc1 = (_TSPwebActionView_WebNodeColumnExpression_content) MkwebActionView_WebNodeColumnExpression_content (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_WebNodeColumnExpression: root of subtree no. 1 can not be made a webActionView_WebNodeColumnExpression_content node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebActionView_WebNodeColumnExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebActionView_WebNodeColumnExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_WebNodeColumnExpression_content (POSITION *_coordref)
#else
NODEPTR MkpwebActionView_WebNodeColumnExpression_content (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebActionView_WebNodeColumnExpression_content _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_WebNodeColumnExpression_content;
#else
_currn = (_TPPpwebActionView_WebNodeColumnExpression_content) TreeNodeAlloc (sizeof (struct _TPpwebActionView_WebNodeColumnExpression_content));
#endif
_currn->_prod = RULEpwebActionView_WebNodeColumnExpression_content;
_SETCOORD(_currn)
_TERMACT_pwebActionView_WebNodeColumnExpression_content;
return ( (NODEPTR) _currn);
}/* MkpwebActionView_WebNodeColumnExpression_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_DBString (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpwebActionView_DBString (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpwebActionView_DBString _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_DBString;
#else
_currn = (_TPPpwebActionView_DBString) TreeNodeAlloc (sizeof (struct _TPpwebActionView_DBString));
#endif
_currn->_prod = RULEpwebActionView_DBString;
_currn->_desc1 = (_TSPwebActionView_DBString_keyword1) MkwebActionView_DBString_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_DBString: root of subtree no. 1 can not be made a webActionView_DBString_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebActionView_DBString;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebActionView_DBString */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_DBString_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpwebActionView_DBString_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebActionView_DBString_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_DBString_keyword1;
#else
_currn = (_TPPpwebActionView_DBString_keyword1) TreeNodeAlloc (sizeof (struct _TPpwebActionView_DBString_keyword1));
#endif
_currn->_prod = RULEpwebActionView_DBString_keyword1;
_SETCOORD(_currn)
_TERMACT_pwebActionView_DBString_keyword1;
return ( (NODEPTR) _currn);
}/* MkpwebActionView_DBString_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_DBInteger (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpwebActionView_DBInteger (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpwebActionView_DBInteger _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_DBInteger;
#else
_currn = (_TPPpwebActionView_DBInteger) TreeNodeAlloc (sizeof (struct _TPpwebActionView_DBInteger));
#endif
_currn->_prod = RULEpwebActionView_DBInteger;
_currn->_desc1 = (_TSPwebActionView_DBInteger_keyword1) MkwebActionView_DBInteger_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_DBInteger: root of subtree no. 1 can not be made a webActionView_DBInteger_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebActionView_DBInteger;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebActionView_DBInteger */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_DBInteger_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpwebActionView_DBInteger_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebActionView_DBInteger_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_DBInteger_keyword1;
#else
_currn = (_TPPpwebActionView_DBInteger_keyword1) TreeNodeAlloc (sizeof (struct _TPpwebActionView_DBInteger_keyword1));
#endif
_currn->_prod = RULEpwebActionView_DBInteger_keyword1;
_SETCOORD(_currn)
_TERMACT_pwebActionView_DBInteger_keyword1;
return ( (NODEPTR) _currn);
}/* MkpwebActionView_DBInteger_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_DBFloat (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpwebActionView_DBFloat (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpwebActionView_DBFloat _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_DBFloat;
#else
_currn = (_TPPpwebActionView_DBFloat) TreeNodeAlloc (sizeof (struct _TPpwebActionView_DBFloat));
#endif
_currn->_prod = RULEpwebActionView_DBFloat;
_currn->_desc1 = (_TSPwebActionView_DBFloat_keyword1) MkwebActionView_DBFloat_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_DBFloat: root of subtree no. 1 can not be made a webActionView_DBFloat_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebActionView_DBFloat;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebActionView_DBFloat */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_DBFloat_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpwebActionView_DBFloat_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebActionView_DBFloat_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_DBFloat_keyword1;
#else
_currn = (_TPPpwebActionView_DBFloat_keyword1) TreeNodeAlloc (sizeof (struct _TPpwebActionView_DBFloat_keyword1));
#endif
_currn->_prod = RULEpwebActionView_DBFloat_keyword1;
_SETCOORD(_currn)
_TERMACT_pwebActionView_DBFloat_keyword1;
return ( (NODEPTR) _currn);
}/* MkpwebActionView_DBFloat_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_DBDouble (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpwebActionView_DBDouble (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpwebActionView_DBDouble _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_DBDouble;
#else
_currn = (_TPPpwebActionView_DBDouble) TreeNodeAlloc (sizeof (struct _TPpwebActionView_DBDouble));
#endif
_currn->_prod = RULEpwebActionView_DBDouble;
_currn->_desc1 = (_TSPwebActionView_DBDouble_keyword1) MkwebActionView_DBDouble_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_DBDouble: root of subtree no. 1 can not be made a webActionView_DBDouble_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebActionView_DBDouble;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebActionView_DBDouble */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_DBDouble_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpwebActionView_DBDouble_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebActionView_DBDouble_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_DBDouble_keyword1;
#else
_currn = (_TPPpwebActionView_DBDouble_keyword1) TreeNodeAlloc (sizeof (struct _TPpwebActionView_DBDouble_keyword1));
#endif
_currn->_prod = RULEpwebActionView_DBDouble_keyword1;
_SETCOORD(_currn)
_TERMACT_pwebActionView_DBDouble_keyword1;
return ( (NODEPTR) _currn);
}/* MkpwebActionView_DBDouble_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_DBBigInt (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpwebActionView_DBBigInt (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpwebActionView_DBBigInt _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_DBBigInt;
#else
_currn = (_TPPpwebActionView_DBBigInt) TreeNodeAlloc (sizeof (struct _TPpwebActionView_DBBigInt));
#endif
_currn->_prod = RULEpwebActionView_DBBigInt;
_currn->_desc1 = (_TSPwebActionView_DBBigInt_keyword1) MkwebActionView_DBBigInt_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_DBBigInt: root of subtree no. 1 can not be made a webActionView_DBBigInt_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebActionView_DBBigInt;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebActionView_DBBigInt */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_DBBigInt_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpwebActionView_DBBigInt_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebActionView_DBBigInt_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_DBBigInt_keyword1;
#else
_currn = (_TPPpwebActionView_DBBigInt_keyword1) TreeNodeAlloc (sizeof (struct _TPpwebActionView_DBBigInt_keyword1));
#endif
_currn->_prod = RULEpwebActionView_DBBigInt_keyword1;
_SETCOORD(_currn)
_TERMACT_pwebActionView_DBBigInt_keyword1;
return ( (NODEPTR) _currn);
}/* MkpwebActionView_DBBigInt_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_DBDate (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpwebActionView_DBDate (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpwebActionView_DBDate _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_DBDate;
#else
_currn = (_TPPpwebActionView_DBDate) TreeNodeAlloc (sizeof (struct _TPpwebActionView_DBDate));
#endif
_currn->_prod = RULEpwebActionView_DBDate;
_currn->_desc1 = (_TSPwebActionView_DBDate_keyword1) MkwebActionView_DBDate_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_DBDate: root of subtree no. 1 can not be made a webActionView_DBDate_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebActionView_DBDate;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebActionView_DBDate */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_DBDate_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpwebActionView_DBDate_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebActionView_DBDate_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_DBDate_keyword1;
#else
_currn = (_TPPpwebActionView_DBDate_keyword1) TreeNodeAlloc (sizeof (struct _TPpwebActionView_DBDate_keyword1));
#endif
_currn->_prod = RULEpwebActionView_DBDate_keyword1;
_SETCOORD(_currn)
_TERMACT_pwebActionView_DBDate_keyword1;
return ( (NODEPTR) _currn);
}/* MkpwebActionView_DBDate_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_DBTimeStamp (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpwebActionView_DBTimeStamp (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpwebActionView_DBTimeStamp _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_DBTimeStamp;
#else
_currn = (_TPPpwebActionView_DBTimeStamp) TreeNodeAlloc (sizeof (struct _TPpwebActionView_DBTimeStamp));
#endif
_currn->_prod = RULEpwebActionView_DBTimeStamp;
_currn->_desc1 = (_TSPwebActionView_DBTimeStamp_keyword1) MkwebActionView_DBTimeStamp_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_DBTimeStamp: root of subtree no. 1 can not be made a webActionView_DBTimeStamp_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebActionView_DBTimeStamp;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebActionView_DBTimeStamp */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_DBTimeStamp_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpwebActionView_DBTimeStamp_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebActionView_DBTimeStamp_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_DBTimeStamp_keyword1;
#else
_currn = (_TPPpwebActionView_DBTimeStamp_keyword1) TreeNodeAlloc (sizeof (struct _TPpwebActionView_DBTimeStamp_keyword1));
#endif
_currn->_prod = RULEpwebActionView_DBTimeStamp_keyword1;
_SETCOORD(_currn)
_TERMACT_pwebActionView_DBTimeStamp_keyword1;
return ( (NODEPTR) _currn);
}/* MkpwebActionView_DBTimeStamp_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_DBBoolean (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpwebActionView_DBBoolean (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpwebActionView_DBBoolean _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_DBBoolean;
#else
_currn = (_TPPpwebActionView_DBBoolean) TreeNodeAlloc (sizeof (struct _TPpwebActionView_DBBoolean));
#endif
_currn->_prod = RULEpwebActionView_DBBoolean;
_currn->_desc1 = (_TSPwebActionView_DBBoolean_keyword1) MkwebActionView_DBBoolean_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_DBBoolean: root of subtree no. 1 can not be made a webActionView_DBBoolean_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebActionView_DBBoolean;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebActionView_DBBoolean */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_DBBoolean_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpwebActionView_DBBoolean_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebActionView_DBBoolean_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_DBBoolean_keyword1;
#else
_currn = (_TPPpwebActionView_DBBoolean_keyword1) TreeNodeAlloc (sizeof (struct _TPpwebActionView_DBBoolean_keyword1));
#endif
_currn->_prod = RULEpwebActionView_DBBoolean_keyword1;
_SETCOORD(_currn)
_TERMACT_pwebActionView_DBBoolean_keyword1;
return ( (NODEPTR) _currn);
}/* MkpwebActionView_DBBoolean_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_DBBlob (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpwebActionView_DBBlob (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpwebActionView_DBBlob _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_DBBlob;
#else
_currn = (_TPPpwebActionView_DBBlob) TreeNodeAlloc (sizeof (struct _TPpwebActionView_DBBlob));
#endif
_currn->_prod = RULEpwebActionView_DBBlob;
_currn->_desc1 = (_TSPwebActionView_DBBlob_keyword1) MkwebActionView_DBBlob_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_DBBlob: root of subtree no. 1 can not be made a webActionView_DBBlob_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebActionView_DBBlob;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebActionView_DBBlob */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_DBBlob_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpwebActionView_DBBlob_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebActionView_DBBlob_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_DBBlob_keyword1;
#else
_currn = (_TPPpwebActionView_DBBlob_keyword1) TreeNodeAlloc (sizeof (struct _TPpwebActionView_DBBlob_keyword1));
#endif
_currn->_prod = RULEpwebActionView_DBBlob_keyword1;
_SETCOORD(_currn)
_TERMACT_pwebActionView_DBBlob_keyword1;
return ( (NODEPTR) _currn);
}/* MkpwebActionView_DBBlob_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_DataSourceActualParam (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpwebActionView_DataSourceActualParam (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpwebActionView_DataSourceActualParam _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_DataSourceActualParam;
#else
_currn = (_TPPpwebActionView_DataSourceActualParam) TreeNodeAlloc (sizeof (struct _TPpwebActionView_DataSourceActualParam));
#endif
_currn->_prod = RULEpwebActionView_DataSourceActualParam;
_currn->_desc1 = (_TSPwebActionView_DataSourceActualParam_formalParam) MkwebActionView_DataSourceActualParam_formalParam (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_DataSourceActualParam: root of subtree no. 1 can not be made a webActionView_DataSourceActualParam_formalParam node ", 0, _coordref);
_currn->_desc2 = (_TSPwebActionView_DataSourceActualParam_keyword1) MkwebActionView_DataSourceActualParam_keyword1 (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_DataSourceActualParam: root of subtree no. 2 can not be made a webActionView_DataSourceActualParam_keyword1 node ", 0, _coordref);
_currn->_desc3 = (_TSPwebActionView_DataSourceActualParam_expressions) MkwebActionView_DataSourceActualParam_expressions (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_DataSourceActualParam: root of subtree no. 3 can not be made a webActionView_DataSourceActualParam_expressions node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebActionView_DataSourceActualParam;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebActionView_DataSourceActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_DataSourceActualParam_formalParam (POSITION *_coordref)
#else
NODEPTR MkpwebActionView_DataSourceActualParam_formalParam (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebActionView_DataSourceActualParam_formalParam _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_DataSourceActualParam_formalParam;
#else
_currn = (_TPPpwebActionView_DataSourceActualParam_formalParam) TreeNodeAlloc (sizeof (struct _TPpwebActionView_DataSourceActualParam_formalParam));
#endif
_currn->_prod = RULEpwebActionView_DataSourceActualParam_formalParam;
_SETCOORD(_currn)
_TERMACT_pwebActionView_DataSourceActualParam_formalParam;
return ( (NODEPTR) _currn);
}/* MkpwebActionView_DataSourceActualParam_formalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_DataSourceActualParam_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpwebActionView_DataSourceActualParam_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebActionView_DataSourceActualParam_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_DataSourceActualParam_keyword1;
#else
_currn = (_TPPpwebActionView_DataSourceActualParam_keyword1) TreeNodeAlloc (sizeof (struct _TPpwebActionView_DataSourceActualParam_keyword1));
#endif
_currn->_prod = RULEpwebActionView_DataSourceActualParam_keyword1;
_SETCOORD(_currn)
_TERMACT_pwebActionView_DataSourceActualParam_keyword1;
return ( (NODEPTR) _currn);
}/* MkpwebActionView_DataSourceActualParam_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebActionView_WebNodeConstantExpressionpwebActionView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebActionView_WebNodeConstantExpressionpwebActionView_DataSourceActualParam_expressions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webActionView_WebNodeConstantExpressionpwebActionView_DataSourceActualParam_expressions _new;
#ifdef __cplusplus
	_new = new _TPLST_webActionView_WebNodeConstantExpressionpwebActionView_DataSourceActualParam_expressions;
#else
	_new = (_TPPLST_webActionView_WebNodeConstantExpressionpwebActionView_DataSourceActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_webActionView_WebNodeConstantExpressionpwebActionView_DataSourceActualParam_expressions));
#endif
	_new->_prod = RULELST_webActionView_WebNodeConstantExpressionpwebActionView_DataSourceActualParam_expressions;
	_new->_desc1 = (_TSPwebActionView_WebNodeConstantExpression)elem;
	_new->_desc2 = (_TSPLST_webActionView_DataSourceActualParam_expressions)_new;
	{_TPPLST_webActionView_WebNodeConstantExpressionpwebActionView_DataSourceActualParam_expressions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webActionView_WebNodeConstantExpressionpwebActionView_DataSourceActualParam_expressions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebActionView_WebNodeVariableExpressionpwebActionView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebActionView_WebNodeVariableExpressionpwebActionView_DataSourceActualParam_expressions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webActionView_WebNodeVariableExpressionpwebActionView_DataSourceActualParam_expressions _new;
#ifdef __cplusplus
	_new = new _TPLST_webActionView_WebNodeVariableExpressionpwebActionView_DataSourceActualParam_expressions;
#else
	_new = (_TPPLST_webActionView_WebNodeVariableExpressionpwebActionView_DataSourceActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_webActionView_WebNodeVariableExpressionpwebActionView_DataSourceActualParam_expressions));
#endif
	_new->_prod = RULELST_webActionView_WebNodeVariableExpressionpwebActionView_DataSourceActualParam_expressions;
	_new->_desc1 = (_TSPwebActionView_WebNodeVariableExpression)elem;
	_new->_desc2 = (_TSPLST_webActionView_DataSourceActualParam_expressions)_new;
	{_TPPLST_webActionView_WebNodeVariableExpressionpwebActionView_DataSourceActualParam_expressions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webActionView_WebNodeVariableExpressionpwebActionView_DataSourceActualParam_expressions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebActionView_WebNodeColumnExpressionpwebActionView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebActionView_WebNodeColumnExpressionpwebActionView_DataSourceActualParam_expressions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webActionView_WebNodeColumnExpressionpwebActionView_DataSourceActualParam_expressions _new;
#ifdef __cplusplus
	_new = new _TPLST_webActionView_WebNodeColumnExpressionpwebActionView_DataSourceActualParam_expressions;
#else
	_new = (_TPPLST_webActionView_WebNodeColumnExpressionpwebActionView_DataSourceActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_webActionView_WebNodeColumnExpressionpwebActionView_DataSourceActualParam_expressions));
#endif
	_new->_prod = RULELST_webActionView_WebNodeColumnExpressionpwebActionView_DataSourceActualParam_expressions;
	_new->_desc1 = (_TSPwebActionView_WebNodeColumnExpression)elem;
	_new->_desc2 = (_TSPLST_webActionView_DataSourceActualParam_expressions)_new;
	{_TPPLST_webActionView_WebNodeColumnExpressionpwebActionView_DataSourceActualParam_expressions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webActionView_WebNodeColumnExpressionpwebActionView_DataSourceActualParam_expressions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempwebActionView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempwebActionView_DataSourceActualParam_expressions (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBwebActionView_WebNodeConstantExpression)) 
		return _MkwebActionView_WebNodeConstantExpressionpwebActionView_DataSourceActualParam_expressions (_coordref, MkwebActionView_WebNodeConstantExpression (_coordref, t));
	if (IsSymb (t, SYMBwebActionView_WebNodeVariableExpression)) 
		return _MkwebActionView_WebNodeVariableExpressionpwebActionView_DataSourceActualParam_expressions (_coordref, MkwebActionView_WebNodeVariableExpression (_coordref, t));
	if (IsSymb (t, SYMBwebActionView_WebNodeColumnExpression)) 
		return _MkwebActionView_WebNodeColumnExpressionpwebActionView_DataSourceActualParam_expressions (_coordref, MkwebActionView_WebNodeColumnExpression (_coordref, t));
	message (DEADLY, "a pwebActionView_DataSourceActualParam_expressions list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pwebActionView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pwebActionView_DataSourceActualParam_expressions (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_webActionView_DataSourceActualParam_expressions)
	l =_ElempwebActionView_DataSourceActualParam_expressions (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_webActionView_DataSourceActualParam_expressions)
	r = _ElempwebActionView_DataSourceActualParam_expressions (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pwebActionView_DataSourceActualParam_expressions)r)->_desc2);
((_TPPLST_0pwebActionView_DataSourceActualParam_expressions)r)->_desc2 = ((_TPPLST_0pwebActionView_DataSourceActualParam_expressions)l)->_desc2;
((_TPPLST_0pwebActionView_DataSourceActualParam_expressions)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpwebActionView_DataSourceActualParam_expressions (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpwebActionView_DataSourceActualParam_expressions root;
	_TPPLST_0pwebActionView_DataSourceActualParam_expressions list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pwebActionView_DataSourceActualParam_expressions)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_webActionView_DataSourceActualParam_expressions)
		t = _ElempwebActionView_DataSourceActualParam_expressions (_coordref, t);
	remain = (_TPPLST_0pwebActionView_DataSourceActualParam_expressions)(((_TPPLST_0pwebActionView_DataSourceActualParam_expressions) t)->_desc2);
	((_TPPLST_0pwebActionView_DataSourceActualParam_expressions) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pwebActionView_DataSourceActualParam_expressions;
#else
list = (_TPPLST_0pwebActionView_DataSourceActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_0pwebActionView_DataSourceActualParam_expressions));
#endif
list->_prod = RULELST_0pwebActionView_DataSourceActualParam_expressions;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pwebActionView_DataSourceActualParam_expressions)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_webActionView_WebNodeConstantExpressionpwebActionView_DataSourceActualParam_expressions)
	{_TPPLST_webActionView_WebNodeConstantExpressionpwebActionView_DataSourceActualParam_expressions _currn = _currn;
	 _TERMACT_LST_webActionView_WebNodeConstantExpressionpwebActionView_DataSourceActualParam_expressions;
	} this code has been moved to _ElempwebActionView_DataSourceActualParam_expressions*/
	/*if (_currn->_prod == RULELST_webActionView_WebNodeVariableExpressionpwebActionView_DataSourceActualParam_expressions)
	{_TPPLST_webActionView_WebNodeVariableExpressionpwebActionView_DataSourceActualParam_expressions _currn = _currn;
	 _TERMACT_LST_webActionView_WebNodeVariableExpressionpwebActionView_DataSourceActualParam_expressions;
	} this code has been moved to _ElempwebActionView_DataSourceActualParam_expressions*/
	/*if (_currn->_prod == RULELST_webActionView_WebNodeColumnExpressionpwebActionView_DataSourceActualParam_expressions)
	{_TPPLST_webActionView_WebNodeColumnExpressionpwebActionView_DataSourceActualParam_expressions _currn = _currn;
	 _TERMACT_LST_webActionView_WebNodeColumnExpressionpwebActionView_DataSourceActualParam_expressions;
	} this code has been moved to _ElempwebActionView_DataSourceActualParam_expressions*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpwebActionView_DataSourceActualParam_expressions;
#else
root = (_TPPpwebActionView_DataSourceActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPpwebActionView_DataSourceActualParam_expressions));
#endif
root->_prod = RULEpwebActionView_DataSourceActualParam_expressions;
root->_desc1 = (_TSPLST_webActionView_DataSourceActualParam_expressions)list;
{ _TPPpwebActionView_DataSourceActualParam_expressions _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pwebActionView_DataSourceActualParam_expressions;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_ActionHandler (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpwebActionView_ActionHandler (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpwebActionView_ActionHandler _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_ActionHandler;
#else
_currn = (_TPPpwebActionView_ActionHandler) TreeNodeAlloc (sizeof (struct _TPpwebActionView_ActionHandler));
#endif
_currn->_prod = RULEpwebActionView_ActionHandler;
_currn->_desc1 = (_TSPwebActionView_ActionHandler_targetpage) MkwebActionView_ActionHandler_targetpage (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_ActionHandler: root of subtree no. 1 can not be made a webActionView_ActionHandler_targetpage node ", 0, _coordref);
_currn->_desc2 = (_TSPwebActionView_ActionHandler_params) MkwebActionView_ActionHandler_params (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_ActionHandler: root of subtree no. 2 can not be made a webActionView_ActionHandler_params node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebActionView_ActionHandler;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebActionView_ActionHandler */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_ActionHandler_targetpage (POSITION *_coordref)
#else
NODEPTR MkpwebActionView_ActionHandler_targetpage (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebActionView_ActionHandler_targetpage _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_ActionHandler_targetpage;
#else
_currn = (_TPPpwebActionView_ActionHandler_targetpage) TreeNodeAlloc (sizeof (struct _TPpwebActionView_ActionHandler_targetpage));
#endif
_currn->_prod = RULEpwebActionView_ActionHandler_targetpage;
_SETCOORD(_currn)
_TERMACT_pwebActionView_ActionHandler_targetpage;
return ( (NODEPTR) _currn);
}/* MkpwebActionView_ActionHandler_targetpage */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebActionView_WebNodeActualParampwebActionView_ActionHandler_params (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebActionView_WebNodeActualParampwebActionView_ActionHandler_params (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webActionView_WebNodeActualParampwebActionView_ActionHandler_params _new;
#ifdef __cplusplus
	_new = new _TPLST_webActionView_WebNodeActualParampwebActionView_ActionHandler_params;
#else
	_new = (_TPPLST_webActionView_WebNodeActualParampwebActionView_ActionHandler_params) TreeNodeAlloc (sizeof (struct _TPLST_webActionView_WebNodeActualParampwebActionView_ActionHandler_params));
#endif
	_new->_prod = RULELST_webActionView_WebNodeActualParampwebActionView_ActionHandler_params;
	_new->_desc1 = (_TSPwebActionView_WebNodeActualParam)elem;
	_new->_desc2 = (_TSPLST_webActionView_ActionHandler_params)_new;
	{_TPPLST_webActionView_WebNodeActualParampwebActionView_ActionHandler_params _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webActionView_WebNodeActualParampwebActionView_ActionHandler_params;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempwebActionView_ActionHandler_params (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempwebActionView_ActionHandler_params (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBwebActionView_WebNodeActualParam)) 
		return _MkwebActionView_WebNodeActualParampwebActionView_ActionHandler_params (_coordref, MkwebActionView_WebNodeActualParam (_coordref, t));
	message (DEADLY, "a pwebActionView_ActionHandler_params list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pwebActionView_ActionHandler_params (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pwebActionView_ActionHandler_params (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_webActionView_ActionHandler_params)
	l =_ElempwebActionView_ActionHandler_params (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_webActionView_ActionHandler_params)
	r = _ElempwebActionView_ActionHandler_params (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pwebActionView_ActionHandler_params)r)->_desc2);
((_TPPLST_0pwebActionView_ActionHandler_params)r)->_desc2 = ((_TPPLST_0pwebActionView_ActionHandler_params)l)->_desc2;
((_TPPLST_0pwebActionView_ActionHandler_params)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_ActionHandler_params (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpwebActionView_ActionHandler_params (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpwebActionView_ActionHandler_params root;
	_TPPLST_0pwebActionView_ActionHandler_params list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pwebActionView_ActionHandler_params)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_webActionView_ActionHandler_params)
		t = _ElempwebActionView_ActionHandler_params (_coordref, t);
	remain = (_TPPLST_0pwebActionView_ActionHandler_params)(((_TPPLST_0pwebActionView_ActionHandler_params) t)->_desc2);
	((_TPPLST_0pwebActionView_ActionHandler_params) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pwebActionView_ActionHandler_params;
#else
list = (_TPPLST_0pwebActionView_ActionHandler_params) TreeNodeAlloc (sizeof (struct _TPLST_0pwebActionView_ActionHandler_params));
#endif
list->_prod = RULELST_0pwebActionView_ActionHandler_params;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pwebActionView_ActionHandler_params)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_webActionView_WebNodeActualParampwebActionView_ActionHandler_params)
	{_TPPLST_webActionView_WebNodeActualParampwebActionView_ActionHandler_params _currn = _currn;
	 _TERMACT_LST_webActionView_WebNodeActualParampwebActionView_ActionHandler_params;
	} this code has been moved to _ElempwebActionView_ActionHandler_params*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpwebActionView_ActionHandler_params;
#else
root = (_TPPpwebActionView_ActionHandler_params) TreeNodeAlloc (sizeof (struct _TPpwebActionView_ActionHandler_params));
#endif
root->_prod = RULEpwebActionView_ActionHandler_params;
root->_desc1 = (_TSPLST_webActionView_ActionHandler_params)list;
{ _TPPpwebActionView_ActionHandler_params _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pwebActionView_ActionHandler_params;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_WebAction (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkpwebActionView_WebAction (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPpwebActionView_WebAction _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_WebAction;
#else
_currn = (_TPPpwebActionView_WebAction) TreeNodeAlloc (sizeof (struct _TPpwebActionView_WebAction));
#endif
_currn->_prod = RULEpwebActionView_WebAction;
_currn->_desc1 = (_TSPwebActionView_WebAction_name) MkwebActionView_WebAction_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_WebAction: root of subtree no. 1 can not be made a webActionView_WebAction_name node ", 0, _coordref);
_currn->_desc2 = (_TSPwebActionView_WebAction_formalParameter) MkwebActionView_WebAction_formalParameter (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_WebAction: root of subtree no. 2 can not be made a webActionView_WebAction_formalParameter node ", 0, _coordref);
_currn->_desc3 = (_TSPwebActionView_WebAction_successpage) MkwebActionView_WebAction_successpage (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_WebAction: root of subtree no. 3 can not be made a webActionView_WebAction_successpage node ", 0, _coordref);
_currn->_desc4 = (_TSPwebActionView_WebAction_operations) MkwebActionView_WebAction_operations (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_WebAction: root of subtree no. 4 can not be made a webActionView_WebAction_operations node ", 0, _coordref);
_currn->_desc5 = (_TSPwebActionView_WebAction_position) MkwebActionView_WebAction_position (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_WebAction: root of subtree no. 5 can not be made a webActionView_WebAction_position node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebActionView_WebAction;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebActionView_WebAction */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_WebAction_name (POSITION *_coordref)
#else
NODEPTR MkpwebActionView_WebAction_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebActionView_WebAction_name _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_WebAction_name;
#else
_currn = (_TPPpwebActionView_WebAction_name) TreeNodeAlloc (sizeof (struct _TPpwebActionView_WebAction_name));
#endif
_currn->_prod = RULEpwebActionView_WebAction_name;
_SETCOORD(_currn)
_TERMACT_pwebActionView_WebAction_name;
return ( (NODEPTR) _currn);
}/* MkpwebActionView_WebAction_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebActionView_WebNodeFormalParampwebActionView_WebAction_formalParameter (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebActionView_WebNodeFormalParampwebActionView_WebAction_formalParameter (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webActionView_WebNodeFormalParampwebActionView_WebAction_formalParameter _new;
#ifdef __cplusplus
	_new = new _TPLST_webActionView_WebNodeFormalParampwebActionView_WebAction_formalParameter;
#else
	_new = (_TPPLST_webActionView_WebNodeFormalParampwebActionView_WebAction_formalParameter) TreeNodeAlloc (sizeof (struct _TPLST_webActionView_WebNodeFormalParampwebActionView_WebAction_formalParameter));
#endif
	_new->_prod = RULELST_webActionView_WebNodeFormalParampwebActionView_WebAction_formalParameter;
	_new->_desc1 = (_TSPwebActionView_WebNodeFormalParam)elem;
	_new->_desc2 = (_TSPLST_webActionView_WebAction_formalParameter)_new;
	{_TPPLST_webActionView_WebNodeFormalParampwebActionView_WebAction_formalParameter _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webActionView_WebNodeFormalParampwebActionView_WebAction_formalParameter;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempwebActionView_WebAction_formalParameter (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempwebActionView_WebAction_formalParameter (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBwebActionView_WebNodeFormalParam)) 
		return _MkwebActionView_WebNodeFormalParampwebActionView_WebAction_formalParameter (_coordref, MkwebActionView_WebNodeFormalParam (_coordref, t));
	message (DEADLY, "a pwebActionView_WebAction_formalParameter list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pwebActionView_WebAction_formalParameter (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pwebActionView_WebAction_formalParameter (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_webActionView_WebAction_formalParameter)
	l =_ElempwebActionView_WebAction_formalParameter (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_webActionView_WebAction_formalParameter)
	r = _ElempwebActionView_WebAction_formalParameter (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pwebActionView_WebAction_formalParameter)r)->_desc2);
((_TPPLST_0pwebActionView_WebAction_formalParameter)r)->_desc2 = ((_TPPLST_0pwebActionView_WebAction_formalParameter)l)->_desc2;
((_TPPLST_0pwebActionView_WebAction_formalParameter)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_WebAction_formalParameter (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpwebActionView_WebAction_formalParameter (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpwebActionView_WebAction_formalParameter root;
	_TPPLST_0pwebActionView_WebAction_formalParameter list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pwebActionView_WebAction_formalParameter)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_webActionView_WebAction_formalParameter)
		t = _ElempwebActionView_WebAction_formalParameter (_coordref, t);
	remain = (_TPPLST_0pwebActionView_WebAction_formalParameter)(((_TPPLST_0pwebActionView_WebAction_formalParameter) t)->_desc2);
	((_TPPLST_0pwebActionView_WebAction_formalParameter) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pwebActionView_WebAction_formalParameter;
#else
list = (_TPPLST_0pwebActionView_WebAction_formalParameter) TreeNodeAlloc (sizeof (struct _TPLST_0pwebActionView_WebAction_formalParameter));
#endif
list->_prod = RULELST_0pwebActionView_WebAction_formalParameter;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pwebActionView_WebAction_formalParameter)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_webActionView_WebNodeFormalParampwebActionView_WebAction_formalParameter)
	{_TPPLST_webActionView_WebNodeFormalParampwebActionView_WebAction_formalParameter _currn = _currn;
	 _TERMACT_LST_webActionView_WebNodeFormalParampwebActionView_WebAction_formalParameter;
	} this code has been moved to _ElempwebActionView_WebAction_formalParameter*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpwebActionView_WebAction_formalParameter;
#else
root = (_TPPpwebActionView_WebAction_formalParameter) TreeNodeAlloc (sizeof (struct _TPpwebActionView_WebAction_formalParameter));
#endif
root->_prod = RULEpwebActionView_WebAction_formalParameter;
root->_desc1 = (_TSPLST_webActionView_WebAction_formalParameter)list;
{ _TPPpwebActionView_WebAction_formalParameter _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pwebActionView_WebAction_formalParameter;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebActionView_PHPOperationpwebActionView_WebAction_operations (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebActionView_PHPOperationpwebActionView_WebAction_operations (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webActionView_PHPOperationpwebActionView_WebAction_operations _new;
#ifdef __cplusplus
	_new = new _TPLST_webActionView_PHPOperationpwebActionView_WebAction_operations;
#else
	_new = (_TPPLST_webActionView_PHPOperationpwebActionView_WebAction_operations) TreeNodeAlloc (sizeof (struct _TPLST_webActionView_PHPOperationpwebActionView_WebAction_operations));
#endif
	_new->_prod = RULELST_webActionView_PHPOperationpwebActionView_WebAction_operations;
	_new->_desc1 = (_TSPwebActionView_PHPOperation)elem;
	_new->_desc2 = (_TSPLST_webActionView_WebAction_operations)_new;
	{_TPPLST_webActionView_PHPOperationpwebActionView_WebAction_operations _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webActionView_PHPOperationpwebActionView_WebAction_operations;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebActionView_DataBaseOperationpwebActionView_WebAction_operations (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebActionView_DataBaseOperationpwebActionView_WebAction_operations (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webActionView_DataBaseOperationpwebActionView_WebAction_operations _new;
#ifdef __cplusplus
	_new = new _TPLST_webActionView_DataBaseOperationpwebActionView_WebAction_operations;
#else
	_new = (_TPPLST_webActionView_DataBaseOperationpwebActionView_WebAction_operations) TreeNodeAlloc (sizeof (struct _TPLST_webActionView_DataBaseOperationpwebActionView_WebAction_operations));
#endif
	_new->_prod = RULELST_webActionView_DataBaseOperationpwebActionView_WebAction_operations;
	_new->_desc1 = (_TSPwebActionView_DataBaseOperation)elem;
	_new->_desc2 = (_TSPLST_webActionView_WebAction_operations)_new;
	{_TPPLST_webActionView_DataBaseOperationpwebActionView_WebAction_operations _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webActionView_DataBaseOperationpwebActionView_WebAction_operations;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempwebActionView_WebAction_operations (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempwebActionView_WebAction_operations (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBwebActionView_PHPOperation)) 
		return _MkwebActionView_PHPOperationpwebActionView_WebAction_operations (_coordref, MkwebActionView_PHPOperation (_coordref, t));
	if (IsSymb (t, SYMBwebActionView_DataBaseOperation)) 
		return _MkwebActionView_DataBaseOperationpwebActionView_WebAction_operations (_coordref, MkwebActionView_DataBaseOperation (_coordref, t));
	message (DEADLY, "a pwebActionView_WebAction_operations list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pwebActionView_WebAction_operations (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pwebActionView_WebAction_operations (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_webActionView_WebAction_operations)
	l =_ElempwebActionView_WebAction_operations (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_webActionView_WebAction_operations)
	r = _ElempwebActionView_WebAction_operations (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pwebActionView_WebAction_operations)r)->_desc2);
((_TPPLST_0pwebActionView_WebAction_operations)r)->_desc2 = ((_TPPLST_0pwebActionView_WebAction_operations)l)->_desc2;
((_TPPLST_0pwebActionView_WebAction_operations)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_WebAction_operations (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpwebActionView_WebAction_operations (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpwebActionView_WebAction_operations root;
	_TPPLST_0pwebActionView_WebAction_operations list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pwebActionView_WebAction_operations)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_webActionView_WebAction_operations)
		t = _ElempwebActionView_WebAction_operations (_coordref, t);
	remain = (_TPPLST_0pwebActionView_WebAction_operations)(((_TPPLST_0pwebActionView_WebAction_operations) t)->_desc2);
	((_TPPLST_0pwebActionView_WebAction_operations) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pwebActionView_WebAction_operations;
#else
list = (_TPPLST_0pwebActionView_WebAction_operations) TreeNodeAlloc (sizeof (struct _TPLST_0pwebActionView_WebAction_operations));
#endif
list->_prod = RULELST_0pwebActionView_WebAction_operations;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pwebActionView_WebAction_operations)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_webActionView_PHPOperationpwebActionView_WebAction_operations)
	{_TPPLST_webActionView_PHPOperationpwebActionView_WebAction_operations _currn = _currn;
	 _TERMACT_LST_webActionView_PHPOperationpwebActionView_WebAction_operations;
	} this code has been moved to _ElempwebActionView_WebAction_operations*/
	/*if (_currn->_prod == RULELST_webActionView_DataBaseOperationpwebActionView_WebAction_operations)
	{_TPPLST_webActionView_DataBaseOperationpwebActionView_WebAction_operations _currn = _currn;
	 _TERMACT_LST_webActionView_DataBaseOperationpwebActionView_WebAction_operations;
	} this code has been moved to _ElempwebActionView_WebAction_operations*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpwebActionView_WebAction_operations;
#else
root = (_TPPpwebActionView_WebAction_operations) TreeNodeAlloc (sizeof (struct _TPpwebActionView_WebAction_operations));
#endif
root->_prod = RULEpwebActionView_WebAction_operations;
root->_desc1 = (_TSPLST_webActionView_WebAction_operations)list;
{ _TPPpwebActionView_WebAction_operations _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pwebActionView_WebAction_operations;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_WebAction_position (POSITION *_coordref)
#else
NODEPTR MkpwebActionView_WebAction_position (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebActionView_WebAction_position _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_WebAction_position;
#else
_currn = (_TPPpwebActionView_WebAction_position) TreeNodeAlloc (sizeof (struct _TPpwebActionView_WebAction_position));
#endif
_currn->_prod = RULEpwebActionView_WebAction_position;
_SETCOORD(_currn)
_TERMACT_pwebActionView_WebAction_position;
return ( (NODEPTR) _currn);
}/* MkpwebActionView_WebAction_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_PHPOperation (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpwebActionView_PHPOperation (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpwebActionView_PHPOperation _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_PHPOperation;
#else
_currn = (_TPPpwebActionView_PHPOperation) TreeNodeAlloc (sizeof (struct _TPpwebActionView_PHPOperation));
#endif
_currn->_prod = RULEpwebActionView_PHPOperation;
_currn->_desc1 = (_TSPwebActionView_PHPOperation_errorhandler) MkwebActionView_PHPOperation_errorhandler (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_PHPOperation: root of subtree no. 1 can not be made a webActionView_PHPOperation_errorhandler node ", 0, _coordref);
_currn->_desc2 = (_TSPwebActionView_PHPOperation_content) MkwebActionView_PHPOperation_content (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_PHPOperation: root of subtree no. 2 can not be made a webActionView_PHPOperation_content node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebActionView_PHPOperation;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebActionView_PHPOperation */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebActionView_ActionHandlerpwebActionView_PHPOperation_errorhandler (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebActionView_ActionHandlerpwebActionView_PHPOperation_errorhandler (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webActionView_ActionHandlerpwebActionView_PHPOperation_errorhandler _new;
#ifdef __cplusplus
	_new = new _TPLST_webActionView_ActionHandlerpwebActionView_PHPOperation_errorhandler;
#else
	_new = (_TPPLST_webActionView_ActionHandlerpwebActionView_PHPOperation_errorhandler) TreeNodeAlloc (sizeof (struct _TPLST_webActionView_ActionHandlerpwebActionView_PHPOperation_errorhandler));
#endif
	_new->_prod = RULELST_webActionView_ActionHandlerpwebActionView_PHPOperation_errorhandler;
	_new->_desc1 = (_TSPwebActionView_ActionHandler)elem;
	_new->_desc2 = (_TSPLST_webActionView_PHPOperation_errorhandler)_new;
	{_TPPLST_webActionView_ActionHandlerpwebActionView_PHPOperation_errorhandler _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webActionView_ActionHandlerpwebActionView_PHPOperation_errorhandler;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempwebActionView_PHPOperation_errorhandler (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempwebActionView_PHPOperation_errorhandler (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBwebActionView_ActionHandler)) 
		return _MkwebActionView_ActionHandlerpwebActionView_PHPOperation_errorhandler (_coordref, MkwebActionView_ActionHandler (_coordref, t));
	message (DEADLY, "a pwebActionView_PHPOperation_errorhandler list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pwebActionView_PHPOperation_errorhandler (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pwebActionView_PHPOperation_errorhandler (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_webActionView_PHPOperation_errorhandler)
	l =_ElempwebActionView_PHPOperation_errorhandler (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_webActionView_PHPOperation_errorhandler)
	r = _ElempwebActionView_PHPOperation_errorhandler (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pwebActionView_PHPOperation_errorhandler)r)->_desc2);
((_TPPLST_0pwebActionView_PHPOperation_errorhandler)r)->_desc2 = ((_TPPLST_0pwebActionView_PHPOperation_errorhandler)l)->_desc2;
((_TPPLST_0pwebActionView_PHPOperation_errorhandler)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_PHPOperation_errorhandler (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpwebActionView_PHPOperation_errorhandler (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpwebActionView_PHPOperation_errorhandler root;
	_TPPLST_0pwebActionView_PHPOperation_errorhandler list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pwebActionView_PHPOperation_errorhandler)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_webActionView_PHPOperation_errorhandler)
		t = _ElempwebActionView_PHPOperation_errorhandler (_coordref, t);
	remain = (_TPPLST_0pwebActionView_PHPOperation_errorhandler)(((_TPPLST_0pwebActionView_PHPOperation_errorhandler) t)->_desc2);
	((_TPPLST_0pwebActionView_PHPOperation_errorhandler) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pwebActionView_PHPOperation_errorhandler;
#else
list = (_TPPLST_0pwebActionView_PHPOperation_errorhandler) TreeNodeAlloc (sizeof (struct _TPLST_0pwebActionView_PHPOperation_errorhandler));
#endif
list->_prod = RULELST_0pwebActionView_PHPOperation_errorhandler;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pwebActionView_PHPOperation_errorhandler)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_webActionView_ActionHandlerpwebActionView_PHPOperation_errorhandler)
	{_TPPLST_webActionView_ActionHandlerpwebActionView_PHPOperation_errorhandler _currn = _currn;
	 _TERMACT_LST_webActionView_ActionHandlerpwebActionView_PHPOperation_errorhandler;
	} this code has been moved to _ElempwebActionView_PHPOperation_errorhandler*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpwebActionView_PHPOperation_errorhandler;
#else
root = (_TPPpwebActionView_PHPOperation_errorhandler) TreeNodeAlloc (sizeof (struct _TPpwebActionView_PHPOperation_errorhandler));
#endif
root->_prod = RULEpwebActionView_PHPOperation_errorhandler;
root->_desc1 = (_TSPLST_webActionView_PHPOperation_errorhandler)list;
{ _TPPpwebActionView_PHPOperation_errorhandler _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pwebActionView_PHPOperation_errorhandler;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_PHPOperation_content (POSITION *_coordref)
#else
NODEPTR MkpwebActionView_PHPOperation_content (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebActionView_PHPOperation_content _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_PHPOperation_content;
#else
_currn = (_TPPpwebActionView_PHPOperation_content) TreeNodeAlloc (sizeof (struct _TPpwebActionView_PHPOperation_content));
#endif
_currn->_prod = RULEpwebActionView_PHPOperation_content;
_SETCOORD(_currn)
_TERMACT_pwebActionView_PHPOperation_content;
return ( (NODEPTR) _currn);
}/* MkpwebActionView_PHPOperation_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_DataBaseOperation (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpwebActionView_DataBaseOperation (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpwebActionView_DataBaseOperation _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_DataBaseOperation;
#else
_currn = (_TPPpwebActionView_DataBaseOperation) TreeNodeAlloc (sizeof (struct _TPpwebActionView_DataBaseOperation));
#endif
_currn->_prod = RULEpwebActionView_DataBaseOperation;
_currn->_desc1 = (_TSPwebActionView_DataBaseOperation_errorhandler) MkwebActionView_DataBaseOperation_errorhandler (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_DataBaseOperation: root of subtree no. 1 can not be made a webActionView_DataBaseOperation_errorhandler node ", 0, _coordref);
_currn->_desc2 = (_TSPwebActionView_DataBaseOperation_filter) MkwebActionView_DataBaseOperation_filter (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_DataBaseOperation: root of subtree no. 2 can not be made a webActionView_DataBaseOperation_filter node ", 0, _coordref);
_currn->_desc3 = (_TSPwebActionView_DataBaseOperation_params) MkwebActionView_DataBaseOperation_params (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebActionView_DataBaseOperation: root of subtree no. 3 can not be made a webActionView_DataBaseOperation_params node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebActionView_DataBaseOperation;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebActionView_DataBaseOperation */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebActionView_ActionHandlerpwebActionView_DataBaseOperation_errorhandler (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebActionView_ActionHandlerpwebActionView_DataBaseOperation_errorhandler (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webActionView_ActionHandlerpwebActionView_DataBaseOperation_errorhandler _new;
#ifdef __cplusplus
	_new = new _TPLST_webActionView_ActionHandlerpwebActionView_DataBaseOperation_errorhandler;
#else
	_new = (_TPPLST_webActionView_ActionHandlerpwebActionView_DataBaseOperation_errorhandler) TreeNodeAlloc (sizeof (struct _TPLST_webActionView_ActionHandlerpwebActionView_DataBaseOperation_errorhandler));
#endif
	_new->_prod = RULELST_webActionView_ActionHandlerpwebActionView_DataBaseOperation_errorhandler;
	_new->_desc1 = (_TSPwebActionView_ActionHandler)elem;
	_new->_desc2 = (_TSPLST_webActionView_DataBaseOperation_errorhandler)_new;
	{_TPPLST_webActionView_ActionHandlerpwebActionView_DataBaseOperation_errorhandler _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webActionView_ActionHandlerpwebActionView_DataBaseOperation_errorhandler;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempwebActionView_DataBaseOperation_errorhandler (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempwebActionView_DataBaseOperation_errorhandler (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBwebActionView_ActionHandler)) 
		return _MkwebActionView_ActionHandlerpwebActionView_DataBaseOperation_errorhandler (_coordref, MkwebActionView_ActionHandler (_coordref, t));
	message (DEADLY, "a pwebActionView_DataBaseOperation_errorhandler list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pwebActionView_DataBaseOperation_errorhandler (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pwebActionView_DataBaseOperation_errorhandler (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_webActionView_DataBaseOperation_errorhandler)
	l =_ElempwebActionView_DataBaseOperation_errorhandler (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_webActionView_DataBaseOperation_errorhandler)
	r = _ElempwebActionView_DataBaseOperation_errorhandler (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pwebActionView_DataBaseOperation_errorhandler)r)->_desc2);
((_TPPLST_0pwebActionView_DataBaseOperation_errorhandler)r)->_desc2 = ((_TPPLST_0pwebActionView_DataBaseOperation_errorhandler)l)->_desc2;
((_TPPLST_0pwebActionView_DataBaseOperation_errorhandler)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_DataBaseOperation_errorhandler (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpwebActionView_DataBaseOperation_errorhandler (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpwebActionView_DataBaseOperation_errorhandler root;
	_TPPLST_0pwebActionView_DataBaseOperation_errorhandler list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pwebActionView_DataBaseOperation_errorhandler)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_webActionView_DataBaseOperation_errorhandler)
		t = _ElempwebActionView_DataBaseOperation_errorhandler (_coordref, t);
	remain = (_TPPLST_0pwebActionView_DataBaseOperation_errorhandler)(((_TPPLST_0pwebActionView_DataBaseOperation_errorhandler) t)->_desc2);
	((_TPPLST_0pwebActionView_DataBaseOperation_errorhandler) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pwebActionView_DataBaseOperation_errorhandler;
#else
list = (_TPPLST_0pwebActionView_DataBaseOperation_errorhandler) TreeNodeAlloc (sizeof (struct _TPLST_0pwebActionView_DataBaseOperation_errorhandler));
#endif
list->_prod = RULELST_0pwebActionView_DataBaseOperation_errorhandler;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pwebActionView_DataBaseOperation_errorhandler)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_webActionView_ActionHandlerpwebActionView_DataBaseOperation_errorhandler)
	{_TPPLST_webActionView_ActionHandlerpwebActionView_DataBaseOperation_errorhandler _currn = _currn;
	 _TERMACT_LST_webActionView_ActionHandlerpwebActionView_DataBaseOperation_errorhandler;
	} this code has been moved to _ElempwebActionView_DataBaseOperation_errorhandler*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpwebActionView_DataBaseOperation_errorhandler;
#else
root = (_TPPpwebActionView_DataBaseOperation_errorhandler) TreeNodeAlloc (sizeof (struct _TPpwebActionView_DataBaseOperation_errorhandler));
#endif
root->_prod = RULEpwebActionView_DataBaseOperation_errorhandler;
root->_desc1 = (_TSPLST_webActionView_DataBaseOperation_errorhandler)list;
{ _TPPpwebActionView_DataBaseOperation_errorhandler _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pwebActionView_DataBaseOperation_errorhandler;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_DataBaseOperation_filter (POSITION *_coordref)
#else
NODEPTR MkpwebActionView_DataBaseOperation_filter (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebActionView_DataBaseOperation_filter _currn;
#ifdef __cplusplus
_currn = new _TPpwebActionView_DataBaseOperation_filter;
#else
_currn = (_TPPpwebActionView_DataBaseOperation_filter) TreeNodeAlloc (sizeof (struct _TPpwebActionView_DataBaseOperation_filter));
#endif
_currn->_prod = RULEpwebActionView_DataBaseOperation_filter;
_SETCOORD(_currn)
_TERMACT_pwebActionView_DataBaseOperation_filter;
return ( (NODEPTR) _currn);
}/* MkpwebActionView_DataBaseOperation_filter */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebActionView_DataSourceActualParampwebActionView_DataBaseOperation_params (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebActionView_DataSourceActualParampwebActionView_DataBaseOperation_params (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webActionView_DataSourceActualParampwebActionView_DataBaseOperation_params _new;
#ifdef __cplusplus
	_new = new _TPLST_webActionView_DataSourceActualParampwebActionView_DataBaseOperation_params;
#else
	_new = (_TPPLST_webActionView_DataSourceActualParampwebActionView_DataBaseOperation_params) TreeNodeAlloc (sizeof (struct _TPLST_webActionView_DataSourceActualParampwebActionView_DataBaseOperation_params));
#endif
	_new->_prod = RULELST_webActionView_DataSourceActualParampwebActionView_DataBaseOperation_params;
	_new->_desc1 = (_TSPwebActionView_DataSourceActualParam)elem;
	_new->_desc2 = (_TSPLST_webActionView_DataBaseOperation_params)_new;
	{_TPPLST_webActionView_DataSourceActualParampwebActionView_DataBaseOperation_params _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webActionView_DataSourceActualParampwebActionView_DataBaseOperation_params;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempwebActionView_DataBaseOperation_params (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempwebActionView_DataBaseOperation_params (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBwebActionView_DataSourceActualParam)) 
		return _MkwebActionView_DataSourceActualParampwebActionView_DataBaseOperation_params (_coordref, MkwebActionView_DataSourceActualParam (_coordref, t));
	message (DEADLY, "a pwebActionView_DataBaseOperation_params list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pwebActionView_DataBaseOperation_params (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pwebActionView_DataBaseOperation_params (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_webActionView_DataBaseOperation_params)
	l =_ElempwebActionView_DataBaseOperation_params (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_webActionView_DataBaseOperation_params)
	r = _ElempwebActionView_DataBaseOperation_params (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pwebActionView_DataBaseOperation_params)r)->_desc2);
((_TPPLST_0pwebActionView_DataBaseOperation_params)r)->_desc2 = ((_TPPLST_0pwebActionView_DataBaseOperation_params)l)->_desc2;
((_TPPLST_0pwebActionView_DataBaseOperation_params)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebActionView_DataBaseOperation_params (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpwebActionView_DataBaseOperation_params (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpwebActionView_DataBaseOperation_params root;
	_TPPLST_0pwebActionView_DataBaseOperation_params list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pwebActionView_DataBaseOperation_params)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_webActionView_DataBaseOperation_params)
		t = _ElempwebActionView_DataBaseOperation_params (_coordref, t);
	remain = (_TPPLST_0pwebActionView_DataBaseOperation_params)(((_TPPLST_0pwebActionView_DataBaseOperation_params) t)->_desc2);
	((_TPPLST_0pwebActionView_DataBaseOperation_params) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pwebActionView_DataBaseOperation_params;
#else
list = (_TPPLST_0pwebActionView_DataBaseOperation_params) TreeNodeAlloc (sizeof (struct _TPLST_0pwebActionView_DataBaseOperation_params));
#endif
list->_prod = RULELST_0pwebActionView_DataBaseOperation_params;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pwebActionView_DataBaseOperation_params)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_webActionView_DataSourceActualParampwebActionView_DataBaseOperation_params)
	{_TPPLST_webActionView_DataSourceActualParampwebActionView_DataBaseOperation_params _currn = _currn;
	 _TERMACT_LST_webActionView_DataSourceActualParampwebActionView_DataBaseOperation_params;
	} this code has been moved to _ElempwebActionView_DataBaseOperation_params*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpwebActionView_DataBaseOperation_params;
#else
root = (_TPPpwebActionView_DataBaseOperation_params) TreeNodeAlloc (sizeof (struct _TPpwebActionView_DataBaseOperation_params));
#endif
root->_prod = RULEpwebActionView_DataBaseOperation_params;
root->_desc1 = (_TSPLST_webActionView_DataBaseOperation_params)list;
{ _TPPpwebActionView_DataBaseOperation_params _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pwebActionView_DataBaseOperation_params;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpVIEWROOT_webPageView (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR MkpVIEWROOT_webPageView (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPpVIEWROOT_webPageView _currn;
#ifdef __cplusplus
_currn = new _TPpVIEWROOT_webPageView;
#else
_currn = (_TPPpVIEWROOT_webPageView) TreeNodeAlloc (sizeof (struct _TPpVIEWROOT_webPageView));
#endif
_currn->_prod = RULEpVIEWROOT_webPageView;
_currn->_desc1 = (_TSPwebPageView_WebPage) MkwebPageView_WebPage (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pVIEWROOT_webPageView: root of subtree no. 1 can not be made a webPageView_WebPage node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_pVIEWROOT_webPageView;
return ( (NODEPTR) _currn);
}/* MkpVIEWROOT_webPageView */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularCheckboxElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4)
#else
NODEPTR MkpwebPageView_FormularCheckboxElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
#endif
{	_TPPpwebPageView_FormularCheckboxElement _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularCheckboxElement;
#else
_currn = (_TPPpwebPageView_FormularCheckboxElement) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularCheckboxElement));
#endif
_currn->_prod = RULEpwebPageView_FormularCheckboxElement;
_currn->_desc1 = (_TSPwebPageView_FormularCheckboxElement_label) MkwebPageView_FormularCheckboxElement_label (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularCheckboxElement: root of subtree no. 1 can not be made a webPageView_FormularCheckboxElement_label node ", 0, _coordref);
_currn->_desc2 = (_TSPwebPageView_FormularCheckboxElement_type) MkwebPageView_FormularCheckboxElement_type (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularCheckboxElement: root of subtree no. 2 can not be made a webPageView_FormularCheckboxElement_type node ", 0, _coordref);
_currn->_desc3 = (_TSPwebPageView_FormularCheckboxElement_condition) MkwebPageView_FormularCheckboxElement_condition (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularCheckboxElement: root of subtree no. 3 can not be made a webPageView_FormularCheckboxElement_condition node ", 0, _coordref);
_currn->_desc4 = (_TSPwebPageView_FormularCheckboxElement_description) MkwebPageView_FormularCheckboxElement_description (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularCheckboxElement: root of subtree no. 4 can not be made a webPageView_FormularCheckboxElement_description node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularCheckboxElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularCheckboxElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularCheckboxElement_label (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularCheckboxElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularCheckboxElement_label _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularCheckboxElement_label;
#else
_currn = (_TPPpwebPageView_FormularCheckboxElement_label) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularCheckboxElement_label));
#endif
_currn->_prod = RULEpwebPageView_FormularCheckboxElement_label;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularCheckboxElement_label;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularCheckboxElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularCheckboxElement_name (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularCheckboxElement_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularCheckboxElement_name _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularCheckboxElement_name;
#else
_currn = (_TPPpwebPageView_FormularCheckboxElement_name) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularCheckboxElement_name));
#endif
_currn->_prod = RULEpwebPageView_FormularCheckboxElement_name;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularCheckboxElement_name;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularCheckboxElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularCheckboxElement_type (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR MkpwebPageView_FormularCheckboxElement_type (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPpwebPageView_FormularCheckboxElement_type _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularCheckboxElement_type;
#else
_currn = (_TPPpwebPageView_FormularCheckboxElement_type) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularCheckboxElement_type));
#endif
_currn->_prod = RULEpwebPageView_FormularCheckboxElement_type;
_currn->_desc1 = (_TSPwebPageView_FormularCheckboxElement_name) MkwebPageView_FormularCheckboxElement_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularCheckboxElement_type: root of subtree no. 1 can not be made a webPageView_FormularCheckboxElement_name node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_pwebPageView_FormularCheckboxElement_type;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularCheckboxElement_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularCheckboxElement_description (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularCheckboxElement_description (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularCheckboxElement_description _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularCheckboxElement_description;
#else
_currn = (_TPPpwebPageView_FormularCheckboxElement_description) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularCheckboxElement_description));
#endif
_currn->_prod = RULEpwebPageView_FormularCheckboxElement_description;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularCheckboxElement_description;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularCheckboxElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularTextElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4)
#else
NODEPTR MkpwebPageView_FormularTextElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
#endif
{	_TPPpwebPageView_FormularTextElement _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularTextElement;
#else
_currn = (_TPPpwebPageView_FormularTextElement) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularTextElement));
#endif
_currn->_prod = RULEpwebPageView_FormularTextElement;
_currn->_desc1 = (_TSPwebPageView_FormularTextElement_label) MkwebPageView_FormularTextElement_label (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularTextElement: root of subtree no. 1 can not be made a webPageView_FormularTextElement_label node ", 0, _coordref);
_currn->_desc2 = (_TSPwebPageView_FormularTextElement_type) MkwebPageView_FormularTextElement_type (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularTextElement: root of subtree no. 2 can not be made a webPageView_FormularTextElement_type node ", 0, _coordref);
_currn->_desc3 = (_TSPwebPageView_FormularTextElement_condition) MkwebPageView_FormularTextElement_condition (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularTextElement: root of subtree no. 3 can not be made a webPageView_FormularTextElement_condition node ", 0, _coordref);
_currn->_desc4 = (_TSPwebPageView_FormularTextElement_description) MkwebPageView_FormularTextElement_description (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularTextElement: root of subtree no. 4 can not be made a webPageView_FormularTextElement_description node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularTextElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularTextElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularTextElement_label (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularTextElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularTextElement_label _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularTextElement_label;
#else
_currn = (_TPPpwebPageView_FormularTextElement_label) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularTextElement_label));
#endif
_currn->_prod = RULEpwebPageView_FormularTextElement_label;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularTextElement_label;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularTextElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularTextElement_type (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularTextElement_type (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularTextElement_type _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularTextElement_type;
#else
_currn = (_TPPpwebPageView_FormularTextElement_type) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularTextElement_type));
#endif
_currn->_prod = RULEpwebPageView_FormularTextElement_type;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularTextElement_type;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularTextElement_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularTextElement_description (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularTextElement_description (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularTextElement_description _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularTextElement_description;
#else
_currn = (_TPPpwebPageView_FormularTextElement_description) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularTextElement_description));
#endif
_currn->_prod = RULEpwebPageView_FormularTextElement_description;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularTextElement_description;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularTextElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularPasswordElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4)
#else
NODEPTR MkpwebPageView_FormularPasswordElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
#endif
{	_TPPpwebPageView_FormularPasswordElement _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularPasswordElement;
#else
_currn = (_TPPpwebPageView_FormularPasswordElement) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularPasswordElement));
#endif
_currn->_prod = RULEpwebPageView_FormularPasswordElement;
_currn->_desc1 = (_TSPwebPageView_FormularPasswordElement_label) MkwebPageView_FormularPasswordElement_label (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularPasswordElement: root of subtree no. 1 can not be made a webPageView_FormularPasswordElement_label node ", 0, _coordref);
_currn->_desc2 = (_TSPwebPageView_FormularPasswordElement_type) MkwebPageView_FormularPasswordElement_type (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularPasswordElement: root of subtree no. 2 can not be made a webPageView_FormularPasswordElement_type node ", 0, _coordref);
_currn->_desc3 = (_TSPwebPageView_FormularPasswordElement_condition) MkwebPageView_FormularPasswordElement_condition (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularPasswordElement: root of subtree no. 3 can not be made a webPageView_FormularPasswordElement_condition node ", 0, _coordref);
_currn->_desc4 = (_TSPwebPageView_FormularPasswordElement_description) MkwebPageView_FormularPasswordElement_description (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularPasswordElement: root of subtree no. 4 can not be made a webPageView_FormularPasswordElement_description node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularPasswordElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularPasswordElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularPasswordElement_label (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularPasswordElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularPasswordElement_label _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularPasswordElement_label;
#else
_currn = (_TPPpwebPageView_FormularPasswordElement_label) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularPasswordElement_label));
#endif
_currn->_prod = RULEpwebPageView_FormularPasswordElement_label;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularPasswordElement_label;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularPasswordElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularPasswordElement_type (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularPasswordElement_type (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularPasswordElement_type _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularPasswordElement_type;
#else
_currn = (_TPPpwebPageView_FormularPasswordElement_type) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularPasswordElement_type));
#endif
_currn->_prod = RULEpwebPageView_FormularPasswordElement_type;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularPasswordElement_type;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularPasswordElement_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularPasswordElement_description (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularPasswordElement_description (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularPasswordElement_description _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularPasswordElement_description;
#else
_currn = (_TPPpwebPageView_FormularPasswordElement_description) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularPasswordElement_description));
#endif
_currn->_prod = RULEpwebPageView_FormularPasswordElement_description;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularPasswordElement_description;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularPasswordElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularTextAreaElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4)
#else
NODEPTR MkpwebPageView_FormularTextAreaElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
#endif
{	_TPPpwebPageView_FormularTextAreaElement _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularTextAreaElement;
#else
_currn = (_TPPpwebPageView_FormularTextAreaElement) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularTextAreaElement));
#endif
_currn->_prod = RULEpwebPageView_FormularTextAreaElement;
_currn->_desc1 = (_TSPwebPageView_FormularTextAreaElement_label) MkwebPageView_FormularTextAreaElement_label (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularTextAreaElement: root of subtree no. 1 can not be made a webPageView_FormularTextAreaElement_label node ", 0, _coordref);
_currn->_desc2 = (_TSPwebPageView_FormularTextAreaElement_type) MkwebPageView_FormularTextAreaElement_type (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularTextAreaElement: root of subtree no. 2 can not be made a webPageView_FormularTextAreaElement_type node ", 0, _coordref);
_currn->_desc3 = (_TSPwebPageView_FormularTextAreaElement_condition) MkwebPageView_FormularTextAreaElement_condition (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularTextAreaElement: root of subtree no. 3 can not be made a webPageView_FormularTextAreaElement_condition node ", 0, _coordref);
_currn->_desc4 = (_TSPwebPageView_FormularTextAreaElement_description) MkwebPageView_FormularTextAreaElement_description (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularTextAreaElement: root of subtree no. 4 can not be made a webPageView_FormularTextAreaElement_description node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularTextAreaElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularTextAreaElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularTextAreaElement_label (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularTextAreaElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularTextAreaElement_label _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularTextAreaElement_label;
#else
_currn = (_TPPpwebPageView_FormularTextAreaElement_label) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularTextAreaElement_label));
#endif
_currn->_prod = RULEpwebPageView_FormularTextAreaElement_label;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularTextAreaElement_label;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularTextAreaElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularTextAreaElement_type (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularTextAreaElement_type (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularTextAreaElement_type _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularTextAreaElement_type;
#else
_currn = (_TPPpwebPageView_FormularTextAreaElement_type) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularTextAreaElement_type));
#endif
_currn->_prod = RULEpwebPageView_FormularTextAreaElement_type;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularTextAreaElement_type;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularTextAreaElement_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularTextAreaElement_description (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularTextAreaElement_description (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularTextAreaElement_description _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularTextAreaElement_description;
#else
_currn = (_TPPpwebPageView_FormularTextAreaElement_description) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularTextAreaElement_description));
#endif
_currn->_prod = RULEpwebPageView_FormularTextAreaElement_description;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularTextAreaElement_description;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularTextAreaElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularSelectElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4)
#else
NODEPTR MkpwebPageView_FormularSelectElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
#endif
{	_TPPpwebPageView_FormularSelectElement _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularSelectElement;
#else
_currn = (_TPPpwebPageView_FormularSelectElement) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularSelectElement));
#endif
_currn->_prod = RULEpwebPageView_FormularSelectElement;
_currn->_desc1 = (_TSPwebPageView_FormularSelectElement_label) MkwebPageView_FormularSelectElement_label (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularSelectElement: root of subtree no. 1 can not be made a webPageView_FormularSelectElement_label node ", 0, _coordref);
_currn->_desc2 = (_TSPwebPageView_FormularSelectElement_multiple) MkwebPageView_FormularSelectElement_multiple (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularSelectElement: root of subtree no. 2 can not be made a webPageView_FormularSelectElement_multiple node ", 0, _coordref);
_currn->_desc3 = (_TSPwebPageView_FormularSelectElement_options) MkwebPageView_FormularSelectElement_options (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularSelectElement: root of subtree no. 3 can not be made a webPageView_FormularSelectElement_options node ", 0, _coordref);
_currn->_desc4 = (_TSPwebPageView_FormularSelectElement_description) MkwebPageView_FormularSelectElement_description (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularSelectElement: root of subtree no. 4 can not be made a webPageView_FormularSelectElement_description node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularSelectElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularSelectElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularSelectElement_label (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularSelectElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularSelectElement_label _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularSelectElement_label;
#else
_currn = (_TPPpwebPageView_FormularSelectElement_label) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularSelectElement_label));
#endif
_currn->_prod = RULEpwebPageView_FormularSelectElement_label;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularSelectElement_label;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularSelectElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularSelectElement_multiple (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularSelectElement_multiple (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularSelectElement_multiple _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularSelectElement_multiple;
#else
_currn = (_TPPpwebPageView_FormularSelectElement_multiple) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularSelectElement_multiple));
#endif
_currn->_prod = RULEpwebPageView_FormularSelectElement_multiple;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularSelectElement_multiple;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularSelectElement_multiple */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_FormularConstantSelectOptionpwebPageView_FormularSelectElement_options (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_FormularConstantSelectOptionpwebPageView_FormularSelectElement_options (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_FormularConstantSelectOptionpwebPageView_FormularSelectElement_options _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_FormularConstantSelectOptionpwebPageView_FormularSelectElement_options;
#else
	_new = (_TPPLST_webPageView_FormularConstantSelectOptionpwebPageView_FormularSelectElement_options) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_FormularConstantSelectOptionpwebPageView_FormularSelectElement_options));
#endif
	_new->_prod = RULELST_webPageView_FormularConstantSelectOptionpwebPageView_FormularSelectElement_options;
	_new->_desc1 = (_TSPwebPageView_FormularConstantSelectOption)elem;
	_new->_desc2 = (_TSPLST_webPageView_FormularSelectElement_options)_new;
	{_TPPLST_webPageView_FormularConstantSelectOptionpwebPageView_FormularSelectElement_options _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_FormularConstantSelectOptionpwebPageView_FormularSelectElement_options;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_FormularDataSourceSelectOptionpwebPageView_FormularSelectElement_options (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_FormularDataSourceSelectOptionpwebPageView_FormularSelectElement_options (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_FormularDataSourceSelectOptionpwebPageView_FormularSelectElement_options _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_FormularDataSourceSelectOptionpwebPageView_FormularSelectElement_options;
#else
	_new = (_TPPLST_webPageView_FormularDataSourceSelectOptionpwebPageView_FormularSelectElement_options) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_FormularDataSourceSelectOptionpwebPageView_FormularSelectElement_options));
#endif
	_new->_prod = RULELST_webPageView_FormularDataSourceSelectOptionpwebPageView_FormularSelectElement_options;
	_new->_desc1 = (_TSPwebPageView_FormularDataSourceSelectOption)elem;
	_new->_desc2 = (_TSPLST_webPageView_FormularSelectElement_options)_new;
	{_TPPLST_webPageView_FormularDataSourceSelectOptionpwebPageView_FormularSelectElement_options _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_FormularDataSourceSelectOptionpwebPageView_FormularSelectElement_options;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_FormularPHPScriptSelectOptionpwebPageView_FormularSelectElement_options (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_FormularPHPScriptSelectOptionpwebPageView_FormularSelectElement_options (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_FormularPHPScriptSelectOptionpwebPageView_FormularSelectElement_options _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_FormularPHPScriptSelectOptionpwebPageView_FormularSelectElement_options;
#else
	_new = (_TPPLST_webPageView_FormularPHPScriptSelectOptionpwebPageView_FormularSelectElement_options) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_FormularPHPScriptSelectOptionpwebPageView_FormularSelectElement_options));
#endif
	_new->_prod = RULELST_webPageView_FormularPHPScriptSelectOptionpwebPageView_FormularSelectElement_options;
	_new->_desc1 = (_TSPwebPageView_FormularPHPScriptSelectOption)elem;
	_new->_desc2 = (_TSPLST_webPageView_FormularSelectElement_options)_new;
	{_TPPLST_webPageView_FormularPHPScriptSelectOptionpwebPageView_FormularSelectElement_options _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_FormularPHPScriptSelectOptionpwebPageView_FormularSelectElement_options;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempwebPageView_FormularSelectElement_options (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempwebPageView_FormularSelectElement_options (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBwebPageView_FormularConstantSelectOption)) 
		return _MkwebPageView_FormularConstantSelectOptionpwebPageView_FormularSelectElement_options (_coordref, MkwebPageView_FormularConstantSelectOption (_coordref, t));
	if (IsSymb (t, SYMBwebPageView_FormularDataSourceSelectOption)) 
		return _MkwebPageView_FormularDataSourceSelectOptionpwebPageView_FormularSelectElement_options (_coordref, MkwebPageView_FormularDataSourceSelectOption (_coordref, t));
	if (IsSymb (t, SYMBwebPageView_FormularPHPScriptSelectOption)) 
		return _MkwebPageView_FormularPHPScriptSelectOptionpwebPageView_FormularSelectElement_options (_coordref, MkwebPageView_FormularPHPScriptSelectOption (_coordref, t));
	message (DEADLY, "a pwebPageView_FormularSelectElement_options list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pwebPageView_FormularSelectElement_options (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pwebPageView_FormularSelectElement_options (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_webPageView_FormularSelectElement_options)
	l =_ElempwebPageView_FormularSelectElement_options (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_webPageView_FormularSelectElement_options)
	r = _ElempwebPageView_FormularSelectElement_options (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pwebPageView_FormularSelectElement_options)r)->_desc2);
((_TPPLST_0pwebPageView_FormularSelectElement_options)r)->_desc2 = ((_TPPLST_0pwebPageView_FormularSelectElement_options)l)->_desc2;
((_TPPLST_0pwebPageView_FormularSelectElement_options)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularSelectElement_options (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpwebPageView_FormularSelectElement_options (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpwebPageView_FormularSelectElement_options root;
	_TPPLST_0pwebPageView_FormularSelectElement_options list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pwebPageView_FormularSelectElement_options)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_webPageView_FormularSelectElement_options)
		t = _ElempwebPageView_FormularSelectElement_options (_coordref, t);
	remain = (_TPPLST_0pwebPageView_FormularSelectElement_options)(((_TPPLST_0pwebPageView_FormularSelectElement_options) t)->_desc2);
	((_TPPLST_0pwebPageView_FormularSelectElement_options) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pwebPageView_FormularSelectElement_options;
#else
list = (_TPPLST_0pwebPageView_FormularSelectElement_options) TreeNodeAlloc (sizeof (struct _TPLST_0pwebPageView_FormularSelectElement_options));
#endif
list->_prod = RULELST_0pwebPageView_FormularSelectElement_options;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pwebPageView_FormularSelectElement_options)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_webPageView_FormularConstantSelectOptionpwebPageView_FormularSelectElement_options)
	{_TPPLST_webPageView_FormularConstantSelectOptionpwebPageView_FormularSelectElement_options _currn = _currn;
	 _TERMACT_LST_webPageView_FormularConstantSelectOptionpwebPageView_FormularSelectElement_options;
	} this code has been moved to _ElempwebPageView_FormularSelectElement_options*/
	/*if (_currn->_prod == RULELST_webPageView_FormularDataSourceSelectOptionpwebPageView_FormularSelectElement_options)
	{_TPPLST_webPageView_FormularDataSourceSelectOptionpwebPageView_FormularSelectElement_options _currn = _currn;
	 _TERMACT_LST_webPageView_FormularDataSourceSelectOptionpwebPageView_FormularSelectElement_options;
	} this code has been moved to _ElempwebPageView_FormularSelectElement_options*/
	/*if (_currn->_prod == RULELST_webPageView_FormularPHPScriptSelectOptionpwebPageView_FormularSelectElement_options)
	{_TPPLST_webPageView_FormularPHPScriptSelectOptionpwebPageView_FormularSelectElement_options _currn = _currn;
	 _TERMACT_LST_webPageView_FormularPHPScriptSelectOptionpwebPageView_FormularSelectElement_options;
	} this code has been moved to _ElempwebPageView_FormularSelectElement_options*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpwebPageView_FormularSelectElement_options;
#else
root = (_TPPpwebPageView_FormularSelectElement_options) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularSelectElement_options));
#endif
root->_prod = RULEpwebPageView_FormularSelectElement_options;
root->_desc1 = (_TSPLST_webPageView_FormularSelectElement_options)list;
{ _TPPpwebPageView_FormularSelectElement_options _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pwebPageView_FormularSelectElement_options;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularSelectElement_description (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularSelectElement_description (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularSelectElement_description _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularSelectElement_description;
#else
_currn = (_TPPpwebPageView_FormularSelectElement_description) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularSelectElement_description));
#endif
_currn->_prod = RULEpwebPageView_FormularSelectElement_description;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularSelectElement_description;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularSelectElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularRadioElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpwebPageView_FormularRadioElement (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpwebPageView_FormularRadioElement _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularRadioElement;
#else
_currn = (_TPPpwebPageView_FormularRadioElement) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularRadioElement));
#endif
_currn->_prod = RULEpwebPageView_FormularRadioElement;
_currn->_desc1 = (_TSPwebPageView_FormularRadioElement_type) MkwebPageView_FormularRadioElement_type (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularRadioElement: root of subtree no. 1 can not be made a webPageView_FormularRadioElement_type node ", 0, _coordref);
_currn->_desc2 = (_TSPwebPageView_FormularRadioElement_label) MkwebPageView_FormularRadioElement_label (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularRadioElement: root of subtree no. 2 can not be made a webPageView_FormularRadioElement_label node ", 0, _coordref);
_currn->_desc3 = (_TSPwebPageView_FormularRadioElement_name) MkwebPageView_FormularRadioElement_name (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularRadioElement: root of subtree no. 3 can not be made a webPageView_FormularRadioElement_name node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularRadioElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularRadioElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularRadioElement_type (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularRadioElement_type (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularRadioElement_type _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularRadioElement_type;
#else
_currn = (_TPPpwebPageView_FormularRadioElement_type) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularRadioElement_type));
#endif
_currn->_prod = RULEpwebPageView_FormularRadioElement_type;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularRadioElement_type;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularRadioElement_type */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularRadioElement_label (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularRadioElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularRadioElement_label _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularRadioElement_label;
#else
_currn = (_TPPpwebPageView_FormularRadioElement_label) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularRadioElement_label));
#endif
_currn->_prod = RULEpwebPageView_FormularRadioElement_label;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularRadioElement_label;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularRadioElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularRadioElement_name (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularRadioElement_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularRadioElement_name _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularRadioElement_name;
#else
_currn = (_TPPpwebPageView_FormularRadioElement_name) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularRadioElement_name));
#endif
_currn->_prod = RULEpwebPageView_FormularRadioElement_name;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularRadioElement_name;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularRadioElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_DataSourceActualParam (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpwebPageView_DataSourceActualParam (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpwebPageView_DataSourceActualParam _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_DataSourceActualParam;
#else
_currn = (_TPPpwebPageView_DataSourceActualParam) TreeNodeAlloc (sizeof (struct _TPpwebPageView_DataSourceActualParam));
#endif
_currn->_prod = RULEpwebPageView_DataSourceActualParam;
_currn->_desc1 = (_TSPwebPageView_DataSourceActualParam_formalParam) MkwebPageView_DataSourceActualParam_formalParam (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_DataSourceActualParam: root of subtree no. 1 can not be made a webPageView_DataSourceActualParam_formalParam node ", 0, _coordref);
_currn->_desc2 = (_TSPwebPageView_DataSourceActualParam_keyword1) MkwebPageView_DataSourceActualParam_keyword1 (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_DataSourceActualParam: root of subtree no. 2 can not be made a webPageView_DataSourceActualParam_keyword1 node ", 0, _coordref);
_currn->_desc3 = (_TSPwebPageView_DataSourceActualParam_expressions) MkwebPageView_DataSourceActualParam_expressions (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_DataSourceActualParam: root of subtree no. 3 can not be made a webPageView_DataSourceActualParam_expressions node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_DataSourceActualParam;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_DataSourceActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_DataSourceActualParam_formalParam (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_DataSourceActualParam_formalParam (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_DataSourceActualParam_formalParam _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_DataSourceActualParam_formalParam;
#else
_currn = (_TPPpwebPageView_DataSourceActualParam_formalParam) TreeNodeAlloc (sizeof (struct _TPpwebPageView_DataSourceActualParam_formalParam));
#endif
_currn->_prod = RULEpwebPageView_DataSourceActualParam_formalParam;
_SETCOORD(_currn)
_TERMACT_pwebPageView_DataSourceActualParam_formalParam;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_DataSourceActualParam_formalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_DataSourceActualParam_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_DataSourceActualParam_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_DataSourceActualParam_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_DataSourceActualParam_keyword1;
#else
_currn = (_TPPpwebPageView_DataSourceActualParam_keyword1) TreeNodeAlloc (sizeof (struct _TPpwebPageView_DataSourceActualParam_keyword1));
#endif
_currn->_prod = RULEpwebPageView_DataSourceActualParam_keyword1;
_SETCOORD(_currn)
_TERMACT_pwebPageView_DataSourceActualParam_keyword1;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_DataSourceActualParam_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_WebNodeConstantExpressionpwebPageView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_WebNodeConstantExpressionpwebPageView_DataSourceActualParam_expressions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_WebNodeConstantExpressionpwebPageView_DataSourceActualParam_expressions _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_WebNodeConstantExpressionpwebPageView_DataSourceActualParam_expressions;
#else
	_new = (_TPPLST_webPageView_WebNodeConstantExpressionpwebPageView_DataSourceActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_WebNodeConstantExpressionpwebPageView_DataSourceActualParam_expressions));
#endif
	_new->_prod = RULELST_webPageView_WebNodeConstantExpressionpwebPageView_DataSourceActualParam_expressions;
	_new->_desc1 = (_TSPwebPageView_WebNodeConstantExpression)elem;
	_new->_desc2 = (_TSPLST_webPageView_DataSourceActualParam_expressions)_new;
	{_TPPLST_webPageView_WebNodeConstantExpressionpwebPageView_DataSourceActualParam_expressions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_WebNodeConstantExpressionpwebPageView_DataSourceActualParam_expressions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_WebNodeVariableExpressionpwebPageView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_WebNodeVariableExpressionpwebPageView_DataSourceActualParam_expressions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_WebNodeVariableExpressionpwebPageView_DataSourceActualParam_expressions _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_WebNodeVariableExpressionpwebPageView_DataSourceActualParam_expressions;
#else
	_new = (_TPPLST_webPageView_WebNodeVariableExpressionpwebPageView_DataSourceActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_WebNodeVariableExpressionpwebPageView_DataSourceActualParam_expressions));
#endif
	_new->_prod = RULELST_webPageView_WebNodeVariableExpressionpwebPageView_DataSourceActualParam_expressions;
	_new->_desc1 = (_TSPwebPageView_WebNodeVariableExpression)elem;
	_new->_desc2 = (_TSPLST_webPageView_DataSourceActualParam_expressions)_new;
	{_TPPLST_webPageView_WebNodeVariableExpressionpwebPageView_DataSourceActualParam_expressions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_WebNodeVariableExpressionpwebPageView_DataSourceActualParam_expressions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_WebNodeColumnExpressionpwebPageView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_WebNodeColumnExpressionpwebPageView_DataSourceActualParam_expressions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_WebNodeColumnExpressionpwebPageView_DataSourceActualParam_expressions _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_WebNodeColumnExpressionpwebPageView_DataSourceActualParam_expressions;
#else
	_new = (_TPPLST_webPageView_WebNodeColumnExpressionpwebPageView_DataSourceActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_WebNodeColumnExpressionpwebPageView_DataSourceActualParam_expressions));
#endif
	_new->_prod = RULELST_webPageView_WebNodeColumnExpressionpwebPageView_DataSourceActualParam_expressions;
	_new->_desc1 = (_TSPwebPageView_WebNodeColumnExpression)elem;
	_new->_desc2 = (_TSPLST_webPageView_DataSourceActualParam_expressions)_new;
	{_TPPLST_webPageView_WebNodeColumnExpressionpwebPageView_DataSourceActualParam_expressions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_WebNodeColumnExpressionpwebPageView_DataSourceActualParam_expressions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempwebPageView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempwebPageView_DataSourceActualParam_expressions (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBwebPageView_WebNodeConstantExpression)) 
		return _MkwebPageView_WebNodeConstantExpressionpwebPageView_DataSourceActualParam_expressions (_coordref, MkwebPageView_WebNodeConstantExpression (_coordref, t));
	if (IsSymb (t, SYMBwebPageView_WebNodeVariableExpression)) 
		return _MkwebPageView_WebNodeVariableExpressionpwebPageView_DataSourceActualParam_expressions (_coordref, MkwebPageView_WebNodeVariableExpression (_coordref, t));
	if (IsSymb (t, SYMBwebPageView_WebNodeColumnExpression)) 
		return _MkwebPageView_WebNodeColumnExpressionpwebPageView_DataSourceActualParam_expressions (_coordref, MkwebPageView_WebNodeColumnExpression (_coordref, t));
	message (DEADLY, "a pwebPageView_DataSourceActualParam_expressions list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pwebPageView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pwebPageView_DataSourceActualParam_expressions (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_webPageView_DataSourceActualParam_expressions)
	l =_ElempwebPageView_DataSourceActualParam_expressions (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_webPageView_DataSourceActualParam_expressions)
	r = _ElempwebPageView_DataSourceActualParam_expressions (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pwebPageView_DataSourceActualParam_expressions)r)->_desc2);
((_TPPLST_0pwebPageView_DataSourceActualParam_expressions)r)->_desc2 = ((_TPPLST_0pwebPageView_DataSourceActualParam_expressions)l)->_desc2;
((_TPPLST_0pwebPageView_DataSourceActualParam_expressions)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_DataSourceActualParam_expressions (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpwebPageView_DataSourceActualParam_expressions (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpwebPageView_DataSourceActualParam_expressions root;
	_TPPLST_0pwebPageView_DataSourceActualParam_expressions list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pwebPageView_DataSourceActualParam_expressions)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_webPageView_DataSourceActualParam_expressions)
		t = _ElempwebPageView_DataSourceActualParam_expressions (_coordref, t);
	remain = (_TPPLST_0pwebPageView_DataSourceActualParam_expressions)(((_TPPLST_0pwebPageView_DataSourceActualParam_expressions) t)->_desc2);
	((_TPPLST_0pwebPageView_DataSourceActualParam_expressions) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pwebPageView_DataSourceActualParam_expressions;
#else
list = (_TPPLST_0pwebPageView_DataSourceActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_0pwebPageView_DataSourceActualParam_expressions));
#endif
list->_prod = RULELST_0pwebPageView_DataSourceActualParam_expressions;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pwebPageView_DataSourceActualParam_expressions)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_webPageView_WebNodeConstantExpressionpwebPageView_DataSourceActualParam_expressions)
	{_TPPLST_webPageView_WebNodeConstantExpressionpwebPageView_DataSourceActualParam_expressions _currn = _currn;
	 _TERMACT_LST_webPageView_WebNodeConstantExpressionpwebPageView_DataSourceActualParam_expressions;
	} this code has been moved to _ElempwebPageView_DataSourceActualParam_expressions*/
	/*if (_currn->_prod == RULELST_webPageView_WebNodeVariableExpressionpwebPageView_DataSourceActualParam_expressions)
	{_TPPLST_webPageView_WebNodeVariableExpressionpwebPageView_DataSourceActualParam_expressions _currn = _currn;
	 _TERMACT_LST_webPageView_WebNodeVariableExpressionpwebPageView_DataSourceActualParam_expressions;
	} this code has been moved to _ElempwebPageView_DataSourceActualParam_expressions*/
	/*if (_currn->_prod == RULELST_webPageView_WebNodeColumnExpressionpwebPageView_DataSourceActualParam_expressions)
	{_TPPLST_webPageView_WebNodeColumnExpressionpwebPageView_DataSourceActualParam_expressions _currn = _currn;
	 _TERMACT_LST_webPageView_WebNodeColumnExpressionpwebPageView_DataSourceActualParam_expressions;
	} this code has been moved to _ElempwebPageView_DataSourceActualParam_expressions*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpwebPageView_DataSourceActualParam_expressions;
#else
root = (_TPPpwebPageView_DataSourceActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPpwebPageView_DataSourceActualParam_expressions));
#endif
root->_prod = RULEpwebPageView_DataSourceActualParam_expressions;
root->_desc1 = (_TSPLST_webPageView_DataSourceActualParam_expressions)list;
{ _TPPpwebPageView_DataSourceActualParam_expressions _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pwebPageView_DataSourceActualParam_expressions;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_WebNodeActualParam (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpwebPageView_WebNodeActualParam (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpwebPageView_WebNodeActualParam _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_WebNodeActualParam;
#else
_currn = (_TPPpwebPageView_WebNodeActualParam) TreeNodeAlloc (sizeof (struct _TPpwebPageView_WebNodeActualParam));
#endif
_currn->_prod = RULEpwebPageView_WebNodeActualParam;
_currn->_desc1 = (_TSPwebPageView_WebNodeActualParam_formalParam) MkwebPageView_WebNodeActualParam_formalParam (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_WebNodeActualParam: root of subtree no. 1 can not be made a webPageView_WebNodeActualParam_formalParam node ", 0, _coordref);
_currn->_desc2 = (_TSPwebPageView_WebNodeActualParam_keyword1) MkwebPageView_WebNodeActualParam_keyword1 (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_WebNodeActualParam: root of subtree no. 2 can not be made a webPageView_WebNodeActualParam_keyword1 node ", 0, _coordref);
_currn->_desc3 = (_TSPwebPageView_WebNodeActualParam_expressions) MkwebPageView_WebNodeActualParam_expressions (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_WebNodeActualParam: root of subtree no. 3 can not be made a webPageView_WebNodeActualParam_expressions node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_WebNodeActualParam;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_WebNodeActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_WebNodeActualParam_formalParam (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_WebNodeActualParam_formalParam (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_WebNodeActualParam_formalParam _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_WebNodeActualParam_formalParam;
#else
_currn = (_TPPpwebPageView_WebNodeActualParam_formalParam) TreeNodeAlloc (sizeof (struct _TPpwebPageView_WebNodeActualParam_formalParam));
#endif
_currn->_prod = RULEpwebPageView_WebNodeActualParam_formalParam;
_SETCOORD(_currn)
_TERMACT_pwebPageView_WebNodeActualParam_formalParam;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_WebNodeActualParam_formalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_WebNodeActualParam_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_WebNodeActualParam_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_WebNodeActualParam_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_WebNodeActualParam_keyword1;
#else
_currn = (_TPPpwebPageView_WebNodeActualParam_keyword1) TreeNodeAlloc (sizeof (struct _TPpwebPageView_WebNodeActualParam_keyword1));
#endif
_currn->_prod = RULEpwebPageView_WebNodeActualParam_keyword1;
_SETCOORD(_currn)
_TERMACT_pwebPageView_WebNodeActualParam_keyword1;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_WebNodeActualParam_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_WebNodeConstantExpressionpwebPageView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_WebNodeConstantExpressionpwebPageView_WebNodeActualParam_expressions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_WebNodeConstantExpressionpwebPageView_WebNodeActualParam_expressions _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_WebNodeConstantExpressionpwebPageView_WebNodeActualParam_expressions;
#else
	_new = (_TPPLST_webPageView_WebNodeConstantExpressionpwebPageView_WebNodeActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_WebNodeConstantExpressionpwebPageView_WebNodeActualParam_expressions));
#endif
	_new->_prod = RULELST_webPageView_WebNodeConstantExpressionpwebPageView_WebNodeActualParam_expressions;
	_new->_desc1 = (_TSPwebPageView_WebNodeConstantExpression)elem;
	_new->_desc2 = (_TSPLST_webPageView_WebNodeActualParam_expressions)_new;
	{_TPPLST_webPageView_WebNodeConstantExpressionpwebPageView_WebNodeActualParam_expressions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_WebNodeConstantExpressionpwebPageView_WebNodeActualParam_expressions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_WebNodeVariableExpressionpwebPageView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_WebNodeVariableExpressionpwebPageView_WebNodeActualParam_expressions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_WebNodeVariableExpressionpwebPageView_WebNodeActualParam_expressions _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_WebNodeVariableExpressionpwebPageView_WebNodeActualParam_expressions;
#else
	_new = (_TPPLST_webPageView_WebNodeVariableExpressionpwebPageView_WebNodeActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_WebNodeVariableExpressionpwebPageView_WebNodeActualParam_expressions));
#endif
	_new->_prod = RULELST_webPageView_WebNodeVariableExpressionpwebPageView_WebNodeActualParam_expressions;
	_new->_desc1 = (_TSPwebPageView_WebNodeVariableExpression)elem;
	_new->_desc2 = (_TSPLST_webPageView_WebNodeActualParam_expressions)_new;
	{_TPPLST_webPageView_WebNodeVariableExpressionpwebPageView_WebNodeActualParam_expressions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_WebNodeVariableExpressionpwebPageView_WebNodeActualParam_expressions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_WebNodeColumnExpressionpwebPageView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_WebNodeColumnExpressionpwebPageView_WebNodeActualParam_expressions (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_WebNodeColumnExpressionpwebPageView_WebNodeActualParam_expressions _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_WebNodeColumnExpressionpwebPageView_WebNodeActualParam_expressions;
#else
	_new = (_TPPLST_webPageView_WebNodeColumnExpressionpwebPageView_WebNodeActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_WebNodeColumnExpressionpwebPageView_WebNodeActualParam_expressions));
#endif
	_new->_prod = RULELST_webPageView_WebNodeColumnExpressionpwebPageView_WebNodeActualParam_expressions;
	_new->_desc1 = (_TSPwebPageView_WebNodeColumnExpression)elem;
	_new->_desc2 = (_TSPLST_webPageView_WebNodeActualParam_expressions)_new;
	{_TPPLST_webPageView_WebNodeColumnExpressionpwebPageView_WebNodeActualParam_expressions _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_WebNodeColumnExpressionpwebPageView_WebNodeActualParam_expressions;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempwebPageView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempwebPageView_WebNodeActualParam_expressions (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBwebPageView_WebNodeConstantExpression)) 
		return _MkwebPageView_WebNodeConstantExpressionpwebPageView_WebNodeActualParam_expressions (_coordref, MkwebPageView_WebNodeConstantExpression (_coordref, t));
	if (IsSymb (t, SYMBwebPageView_WebNodeVariableExpression)) 
		return _MkwebPageView_WebNodeVariableExpressionpwebPageView_WebNodeActualParam_expressions (_coordref, MkwebPageView_WebNodeVariableExpression (_coordref, t));
	if (IsSymb (t, SYMBwebPageView_WebNodeColumnExpression)) 
		return _MkwebPageView_WebNodeColumnExpressionpwebPageView_WebNodeActualParam_expressions (_coordref, MkwebPageView_WebNodeColumnExpression (_coordref, t));
	message (DEADLY, "a pwebPageView_WebNodeActualParam_expressions list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pwebPageView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pwebPageView_WebNodeActualParam_expressions (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_webPageView_WebNodeActualParam_expressions)
	l =_ElempwebPageView_WebNodeActualParam_expressions (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_webPageView_WebNodeActualParam_expressions)
	r = _ElempwebPageView_WebNodeActualParam_expressions (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pwebPageView_WebNodeActualParam_expressions)r)->_desc2);
((_TPPLST_0pwebPageView_WebNodeActualParam_expressions)r)->_desc2 = ((_TPPLST_0pwebPageView_WebNodeActualParam_expressions)l)->_desc2;
((_TPPLST_0pwebPageView_WebNodeActualParam_expressions)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_WebNodeActualParam_expressions (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpwebPageView_WebNodeActualParam_expressions (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpwebPageView_WebNodeActualParam_expressions root;
	_TPPLST_0pwebPageView_WebNodeActualParam_expressions list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pwebPageView_WebNodeActualParam_expressions)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_webPageView_WebNodeActualParam_expressions)
		t = _ElempwebPageView_WebNodeActualParam_expressions (_coordref, t);
	remain = (_TPPLST_0pwebPageView_WebNodeActualParam_expressions)(((_TPPLST_0pwebPageView_WebNodeActualParam_expressions) t)->_desc2);
	((_TPPLST_0pwebPageView_WebNodeActualParam_expressions) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pwebPageView_WebNodeActualParam_expressions;
#else
list = (_TPPLST_0pwebPageView_WebNodeActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPLST_0pwebPageView_WebNodeActualParam_expressions));
#endif
list->_prod = RULELST_0pwebPageView_WebNodeActualParam_expressions;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pwebPageView_WebNodeActualParam_expressions)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_webPageView_WebNodeConstantExpressionpwebPageView_WebNodeActualParam_expressions)
	{_TPPLST_webPageView_WebNodeConstantExpressionpwebPageView_WebNodeActualParam_expressions _currn = _currn;
	 _TERMACT_LST_webPageView_WebNodeConstantExpressionpwebPageView_WebNodeActualParam_expressions;
	} this code has been moved to _ElempwebPageView_WebNodeActualParam_expressions*/
	/*if (_currn->_prod == RULELST_webPageView_WebNodeVariableExpressionpwebPageView_WebNodeActualParam_expressions)
	{_TPPLST_webPageView_WebNodeVariableExpressionpwebPageView_WebNodeActualParam_expressions _currn = _currn;
	 _TERMACT_LST_webPageView_WebNodeVariableExpressionpwebPageView_WebNodeActualParam_expressions;
	} this code has been moved to _ElempwebPageView_WebNodeActualParam_expressions*/
	/*if (_currn->_prod == RULELST_webPageView_WebNodeColumnExpressionpwebPageView_WebNodeActualParam_expressions)
	{_TPPLST_webPageView_WebNodeColumnExpressionpwebPageView_WebNodeActualParam_expressions _currn = _currn;
	 _TERMACT_LST_webPageView_WebNodeColumnExpressionpwebPageView_WebNodeActualParam_expressions;
	} this code has been moved to _ElempwebPageView_WebNodeActualParam_expressions*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpwebPageView_WebNodeActualParam_expressions;
#else
root = (_TPPpwebPageView_WebNodeActualParam_expressions) TreeNodeAlloc (sizeof (struct _TPpwebPageView_WebNodeActualParam_expressions));
#endif
root->_prod = RULEpwebPageView_WebNodeActualParam_expressions;
root->_desc1 = (_TSPLST_webPageView_WebNodeActualParam_expressions)list;
{ _TPPpwebPageView_WebNodeActualParam_expressions _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pwebPageView_WebNodeActualParam_expressions;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_WebNodeConstantExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpwebPageView_WebNodeConstantExpression (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpwebPageView_WebNodeConstantExpression _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_WebNodeConstantExpression;
#else
_currn = (_TPPpwebPageView_WebNodeConstantExpression) TreeNodeAlloc (sizeof (struct _TPpwebPageView_WebNodeConstantExpression));
#endif
_currn->_prod = RULEpwebPageView_WebNodeConstantExpression;
_currn->_desc1 = (_TSPwebPageView_WebNodeConstantExpression_keyword1) MkwebPageView_WebNodeConstantExpression_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_WebNodeConstantExpression: root of subtree no. 1 can not be made a webPageView_WebNodeConstantExpression_keyword1 node ", 0, _coordref);
_currn->_desc2 = (_TSPwebPageView_WebNodeConstantExpression_value) MkwebPageView_WebNodeConstantExpression_value (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_WebNodeConstantExpression: root of subtree no. 2 can not be made a webPageView_WebNodeConstantExpression_value node ", 0, _coordref);
_currn->_desc3 = (_TSPwebPageView_WebNodeConstantExpression_keyword2) MkwebPageView_WebNodeConstantExpression_keyword2 (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_WebNodeConstantExpression: root of subtree no. 3 can not be made a webPageView_WebNodeConstantExpression_keyword2 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_WebNodeConstantExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_WebNodeConstantExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_WebNodeConstantExpression_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_WebNodeConstantExpression_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_WebNodeConstantExpression_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_WebNodeConstantExpression_keyword1;
#else
_currn = (_TPPpwebPageView_WebNodeConstantExpression_keyword1) TreeNodeAlloc (sizeof (struct _TPpwebPageView_WebNodeConstantExpression_keyword1));
#endif
_currn->_prod = RULEpwebPageView_WebNodeConstantExpression_keyword1;
_SETCOORD(_currn)
_TERMACT_pwebPageView_WebNodeConstantExpression_keyword1;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_WebNodeConstantExpression_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_WebNodeConstantExpression_value (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_WebNodeConstantExpression_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_WebNodeConstantExpression_value _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_WebNodeConstantExpression_value;
#else
_currn = (_TPPpwebPageView_WebNodeConstantExpression_value) TreeNodeAlloc (sizeof (struct _TPpwebPageView_WebNodeConstantExpression_value));
#endif
_currn->_prod = RULEpwebPageView_WebNodeConstantExpression_value;
_SETCOORD(_currn)
_TERMACT_pwebPageView_WebNodeConstantExpression_value;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_WebNodeConstantExpression_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_WebNodeConstantExpression_keyword2 (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_WebNodeConstantExpression_keyword2 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_WebNodeConstantExpression_keyword2 _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_WebNodeConstantExpression_keyword2;
#else
_currn = (_TPPpwebPageView_WebNodeConstantExpression_keyword2) TreeNodeAlloc (sizeof (struct _TPpwebPageView_WebNodeConstantExpression_keyword2));
#endif
_currn->_prod = RULEpwebPageView_WebNodeConstantExpression_keyword2;
_SETCOORD(_currn)
_TERMACT_pwebPageView_WebNodeConstantExpression_keyword2;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_WebNodeConstantExpression_keyword2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_WebNodeVariableExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpwebPageView_WebNodeVariableExpression (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpwebPageView_WebNodeVariableExpression _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_WebNodeVariableExpression;
#else
_currn = (_TPPpwebPageView_WebNodeVariableExpression) TreeNodeAlloc (sizeof (struct _TPpwebPageView_WebNodeVariableExpression));
#endif
_currn->_prod = RULEpwebPageView_WebNodeVariableExpression;
_currn->_desc1 = (_TSPwebPageView_WebNodeVariableExpression_content) MkwebPageView_WebNodeVariableExpression_content (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_WebNodeVariableExpression: root of subtree no. 1 can not be made a webPageView_WebNodeVariableExpression_content node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_WebNodeVariableExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_WebNodeVariableExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_WebNodeVariableExpression_content (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_WebNodeVariableExpression_content (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_WebNodeVariableExpression_content _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_WebNodeVariableExpression_content;
#else
_currn = (_TPPpwebPageView_WebNodeVariableExpression_content) TreeNodeAlloc (sizeof (struct _TPpwebPageView_WebNodeVariableExpression_content));
#endif
_currn->_prod = RULEpwebPageView_WebNodeVariableExpression_content;
_SETCOORD(_currn)
_TERMACT_pwebPageView_WebNodeVariableExpression_content;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_WebNodeVariableExpression_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_WebNodeColumnExpression (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpwebPageView_WebNodeColumnExpression (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpwebPageView_WebNodeColumnExpression _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_WebNodeColumnExpression;
#else
_currn = (_TPPpwebPageView_WebNodeColumnExpression) TreeNodeAlloc (sizeof (struct _TPpwebPageView_WebNodeColumnExpression));
#endif
_currn->_prod = RULEpwebPageView_WebNodeColumnExpression;
_currn->_desc1 = (_TSPwebPageView_WebNodeColumnExpression_content) MkwebPageView_WebNodeColumnExpression_content (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_WebNodeColumnExpression: root of subtree no. 1 can not be made a webPageView_WebNodeColumnExpression_content node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_WebNodeColumnExpression;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_WebNodeColumnExpression */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_WebNodeColumnExpression_content (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_WebNodeColumnExpression_content (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_WebNodeColumnExpression_content _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_WebNodeColumnExpression_content;
#else
_currn = (_TPPpwebPageView_WebNodeColumnExpression_content) TreeNodeAlloc (sizeof (struct _TPpwebPageView_WebNodeColumnExpression_content));
#endif
_currn->_prod = RULEpwebPageView_WebNodeColumnExpression_content;
_SETCOORD(_currn)
_TERMACT_pwebPageView_WebNodeColumnExpression_content;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_WebNodeColumnExpression_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_WebNodeFormalParam (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpwebPageView_WebNodeFormalParam (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpwebPageView_WebNodeFormalParam _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_WebNodeFormalParam;
#else
_currn = (_TPPpwebPageView_WebNodeFormalParam) TreeNodeAlloc (sizeof (struct _TPpwebPageView_WebNodeFormalParam));
#endif
_currn->_prod = RULEpwebPageView_WebNodeFormalParam;
_currn->_desc1 = (_TSPwebPageView_WebNodeFormalParam_name) MkwebPageView_WebNodeFormalParam_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_WebNodeFormalParam: root of subtree no. 1 can not be made a webPageView_WebNodeFormalParam_name node ", 0, _coordref);
_currn->_desc2 = (_TSPwebPageView_WebNodeFormalParam_keyword1) MkwebPageView_WebNodeFormalParam_keyword1 (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_WebNodeFormalParam: root of subtree no. 2 can not be made a webPageView_WebNodeFormalParam_keyword1 node ", 0, _coordref);
_currn->_desc3 = (_TSPwebPageView_WebNodeFormalParam_type) MkwebPageView_WebNodeFormalParam_type (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_WebNodeFormalParam: root of subtree no. 3 can not be made a webPageView_WebNodeFormalParam_type node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_WebNodeFormalParam;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_WebNodeFormalParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_WebNodeFormalParam_name (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_WebNodeFormalParam_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_WebNodeFormalParam_name _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_WebNodeFormalParam_name;
#else
_currn = (_TPPpwebPageView_WebNodeFormalParam_name) TreeNodeAlloc (sizeof (struct _TPpwebPageView_WebNodeFormalParam_name));
#endif
_currn->_prod = RULEpwebPageView_WebNodeFormalParam_name;
_SETCOORD(_currn)
_TERMACT_pwebPageView_WebNodeFormalParam_name;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_WebNodeFormalParam_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_WebNodeFormalParam_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_WebNodeFormalParam_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_WebNodeFormalParam_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_WebNodeFormalParam_keyword1;
#else
_currn = (_TPPpwebPageView_WebNodeFormalParam_keyword1) TreeNodeAlloc (sizeof (struct _TPpwebPageView_WebNodeFormalParam_keyword1));
#endif
_currn->_prod = RULEpwebPageView_WebNodeFormalParam_keyword1;
_SETCOORD(_currn)
_TERMACT_pwebPageView_WebNodeFormalParam_keyword1;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_WebNodeFormalParam_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_DBStringpwebPageView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_DBStringpwebPageView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_DBStringpwebPageView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_DBStringpwebPageView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_webPageView_DBStringpwebPageView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_DBStringpwebPageView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_webPageView_DBStringpwebPageView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSPwebPageView_DBString)elem;
	_new->_desc2 = (_TSPLST_webPageView_WebNodeFormalParam_type)_new;
	{_TPPLST_webPageView_DBStringpwebPageView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_DBStringpwebPageView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_DBIntegerpwebPageView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_DBIntegerpwebPageView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_DBIntegerpwebPageView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_DBIntegerpwebPageView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_webPageView_DBIntegerpwebPageView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_DBIntegerpwebPageView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_webPageView_DBIntegerpwebPageView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSPwebPageView_DBInteger)elem;
	_new->_desc2 = (_TSPLST_webPageView_WebNodeFormalParam_type)_new;
	{_TPPLST_webPageView_DBIntegerpwebPageView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_DBIntegerpwebPageView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_DBFloatpwebPageView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_DBFloatpwebPageView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_DBFloatpwebPageView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_DBFloatpwebPageView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_webPageView_DBFloatpwebPageView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_DBFloatpwebPageView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_webPageView_DBFloatpwebPageView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSPwebPageView_DBFloat)elem;
	_new->_desc2 = (_TSPLST_webPageView_WebNodeFormalParam_type)_new;
	{_TPPLST_webPageView_DBFloatpwebPageView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_DBFloatpwebPageView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_DBDoublepwebPageView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_DBDoublepwebPageView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_DBDoublepwebPageView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_DBDoublepwebPageView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_webPageView_DBDoublepwebPageView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_DBDoublepwebPageView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_webPageView_DBDoublepwebPageView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSPwebPageView_DBDouble)elem;
	_new->_desc2 = (_TSPLST_webPageView_WebNodeFormalParam_type)_new;
	{_TPPLST_webPageView_DBDoublepwebPageView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_DBDoublepwebPageView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_DBBigIntpwebPageView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_DBBigIntpwebPageView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_DBBigIntpwebPageView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_DBBigIntpwebPageView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_webPageView_DBBigIntpwebPageView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_DBBigIntpwebPageView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_webPageView_DBBigIntpwebPageView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSPwebPageView_DBBigInt)elem;
	_new->_desc2 = (_TSPLST_webPageView_WebNodeFormalParam_type)_new;
	{_TPPLST_webPageView_DBBigIntpwebPageView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_DBBigIntpwebPageView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_DBDatepwebPageView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_DBDatepwebPageView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_DBDatepwebPageView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_DBDatepwebPageView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_webPageView_DBDatepwebPageView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_DBDatepwebPageView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_webPageView_DBDatepwebPageView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSPwebPageView_DBDate)elem;
	_new->_desc2 = (_TSPLST_webPageView_WebNodeFormalParam_type)_new;
	{_TPPLST_webPageView_DBDatepwebPageView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_DBDatepwebPageView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_DBTimeStamppwebPageView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_DBTimeStamppwebPageView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_DBTimeStamppwebPageView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_DBTimeStamppwebPageView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_webPageView_DBTimeStamppwebPageView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_DBTimeStamppwebPageView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_webPageView_DBTimeStamppwebPageView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSPwebPageView_DBTimeStamp)elem;
	_new->_desc2 = (_TSPLST_webPageView_WebNodeFormalParam_type)_new;
	{_TPPLST_webPageView_DBTimeStamppwebPageView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_DBTimeStamppwebPageView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_DBBooleanpwebPageView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_DBBooleanpwebPageView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_DBBooleanpwebPageView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_DBBooleanpwebPageView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_webPageView_DBBooleanpwebPageView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_DBBooleanpwebPageView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_webPageView_DBBooleanpwebPageView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSPwebPageView_DBBoolean)elem;
	_new->_desc2 = (_TSPLST_webPageView_WebNodeFormalParam_type)_new;
	{_TPPLST_webPageView_DBBooleanpwebPageView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_DBBooleanpwebPageView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_DBBlobpwebPageView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_DBBlobpwebPageView_WebNodeFormalParam_type (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_DBBlobpwebPageView_WebNodeFormalParam_type _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_DBBlobpwebPageView_WebNodeFormalParam_type;
#else
	_new = (_TPPLST_webPageView_DBBlobpwebPageView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_DBBlobpwebPageView_WebNodeFormalParam_type));
#endif
	_new->_prod = RULELST_webPageView_DBBlobpwebPageView_WebNodeFormalParam_type;
	_new->_desc1 = (_TSPwebPageView_DBBlob)elem;
	_new->_desc2 = (_TSPLST_webPageView_WebNodeFormalParam_type)_new;
	{_TPPLST_webPageView_DBBlobpwebPageView_WebNodeFormalParam_type _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_DBBlobpwebPageView_WebNodeFormalParam_type;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempwebPageView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempwebPageView_WebNodeFormalParam_type (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBwebPageView_DBString)) 
		return _MkwebPageView_DBStringpwebPageView_WebNodeFormalParam_type (_coordref, MkwebPageView_DBString (_coordref, t));
	if (IsSymb (t, SYMBwebPageView_DBInteger)) 
		return _MkwebPageView_DBIntegerpwebPageView_WebNodeFormalParam_type (_coordref, MkwebPageView_DBInteger (_coordref, t));
	if (IsSymb (t, SYMBwebPageView_DBFloat)) 
		return _MkwebPageView_DBFloatpwebPageView_WebNodeFormalParam_type (_coordref, MkwebPageView_DBFloat (_coordref, t));
	if (IsSymb (t, SYMBwebPageView_DBDouble)) 
		return _MkwebPageView_DBDoublepwebPageView_WebNodeFormalParam_type (_coordref, MkwebPageView_DBDouble (_coordref, t));
	if (IsSymb (t, SYMBwebPageView_DBBigInt)) 
		return _MkwebPageView_DBBigIntpwebPageView_WebNodeFormalParam_type (_coordref, MkwebPageView_DBBigInt (_coordref, t));
	if (IsSymb (t, SYMBwebPageView_DBDate)) 
		return _MkwebPageView_DBDatepwebPageView_WebNodeFormalParam_type (_coordref, MkwebPageView_DBDate (_coordref, t));
	if (IsSymb (t, SYMBwebPageView_DBTimeStamp)) 
		return _MkwebPageView_DBTimeStamppwebPageView_WebNodeFormalParam_type (_coordref, MkwebPageView_DBTimeStamp (_coordref, t));
	if (IsSymb (t, SYMBwebPageView_DBBoolean)) 
		return _MkwebPageView_DBBooleanpwebPageView_WebNodeFormalParam_type (_coordref, MkwebPageView_DBBoolean (_coordref, t));
	if (IsSymb (t, SYMBwebPageView_DBBlob)) 
		return _MkwebPageView_DBBlobpwebPageView_WebNodeFormalParam_type (_coordref, MkwebPageView_DBBlob (_coordref, t));
	message (DEADLY, "a pwebPageView_WebNodeFormalParam_type list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pwebPageView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pwebPageView_WebNodeFormalParam_type (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_webPageView_WebNodeFormalParam_type)
	l =_ElempwebPageView_WebNodeFormalParam_type (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_webPageView_WebNodeFormalParam_type)
	r = _ElempwebPageView_WebNodeFormalParam_type (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pwebPageView_WebNodeFormalParam_type)r)->_desc2);
((_TPPLST_0pwebPageView_WebNodeFormalParam_type)r)->_desc2 = ((_TPPLST_0pwebPageView_WebNodeFormalParam_type)l)->_desc2;
((_TPPLST_0pwebPageView_WebNodeFormalParam_type)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_WebNodeFormalParam_type (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpwebPageView_WebNodeFormalParam_type (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpwebPageView_WebNodeFormalParam_type root;
	_TPPLST_0pwebPageView_WebNodeFormalParam_type list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pwebPageView_WebNodeFormalParam_type)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_webPageView_WebNodeFormalParam_type)
		t = _ElempwebPageView_WebNodeFormalParam_type (_coordref, t);
	remain = (_TPPLST_0pwebPageView_WebNodeFormalParam_type)(((_TPPLST_0pwebPageView_WebNodeFormalParam_type) t)->_desc2);
	((_TPPLST_0pwebPageView_WebNodeFormalParam_type) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pwebPageView_WebNodeFormalParam_type;
#else
list = (_TPPLST_0pwebPageView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPLST_0pwebPageView_WebNodeFormalParam_type));
#endif
list->_prod = RULELST_0pwebPageView_WebNodeFormalParam_type;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pwebPageView_WebNodeFormalParam_type)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_webPageView_DBStringpwebPageView_WebNodeFormalParam_type)
	{_TPPLST_webPageView_DBStringpwebPageView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_webPageView_DBStringpwebPageView_WebNodeFormalParam_type;
	} this code has been moved to _ElempwebPageView_WebNodeFormalParam_type*/
	/*if (_currn->_prod == RULELST_webPageView_DBIntegerpwebPageView_WebNodeFormalParam_type)
	{_TPPLST_webPageView_DBIntegerpwebPageView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_webPageView_DBIntegerpwebPageView_WebNodeFormalParam_type;
	} this code has been moved to _ElempwebPageView_WebNodeFormalParam_type*/
	/*if (_currn->_prod == RULELST_webPageView_DBFloatpwebPageView_WebNodeFormalParam_type)
	{_TPPLST_webPageView_DBFloatpwebPageView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_webPageView_DBFloatpwebPageView_WebNodeFormalParam_type;
	} this code has been moved to _ElempwebPageView_WebNodeFormalParam_type*/
	/*if (_currn->_prod == RULELST_webPageView_DBDoublepwebPageView_WebNodeFormalParam_type)
	{_TPPLST_webPageView_DBDoublepwebPageView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_webPageView_DBDoublepwebPageView_WebNodeFormalParam_type;
	} this code has been moved to _ElempwebPageView_WebNodeFormalParam_type*/
	/*if (_currn->_prod == RULELST_webPageView_DBBigIntpwebPageView_WebNodeFormalParam_type)
	{_TPPLST_webPageView_DBBigIntpwebPageView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_webPageView_DBBigIntpwebPageView_WebNodeFormalParam_type;
	} this code has been moved to _ElempwebPageView_WebNodeFormalParam_type*/
	/*if (_currn->_prod == RULELST_webPageView_DBDatepwebPageView_WebNodeFormalParam_type)
	{_TPPLST_webPageView_DBDatepwebPageView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_webPageView_DBDatepwebPageView_WebNodeFormalParam_type;
	} this code has been moved to _ElempwebPageView_WebNodeFormalParam_type*/
	/*if (_currn->_prod == RULELST_webPageView_DBTimeStamppwebPageView_WebNodeFormalParam_type)
	{_TPPLST_webPageView_DBTimeStamppwebPageView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_webPageView_DBTimeStamppwebPageView_WebNodeFormalParam_type;
	} this code has been moved to _ElempwebPageView_WebNodeFormalParam_type*/
	/*if (_currn->_prod == RULELST_webPageView_DBBooleanpwebPageView_WebNodeFormalParam_type)
	{_TPPLST_webPageView_DBBooleanpwebPageView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_webPageView_DBBooleanpwebPageView_WebNodeFormalParam_type;
	} this code has been moved to _ElempwebPageView_WebNodeFormalParam_type*/
	/*if (_currn->_prod == RULELST_webPageView_DBBlobpwebPageView_WebNodeFormalParam_type)
	{_TPPLST_webPageView_DBBlobpwebPageView_WebNodeFormalParam_type _currn = _currn;
	 _TERMACT_LST_webPageView_DBBlobpwebPageView_WebNodeFormalParam_type;
	} this code has been moved to _ElempwebPageView_WebNodeFormalParam_type*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpwebPageView_WebNodeFormalParam_type;
#else
root = (_TPPpwebPageView_WebNodeFormalParam_type) TreeNodeAlloc (sizeof (struct _TPpwebPageView_WebNodeFormalParam_type));
#endif
root->_prod = RULEpwebPageView_WebNodeFormalParam_type;
root->_desc1 = (_TSPLST_webPageView_WebNodeFormalParam_type)list;
{ _TPPpwebPageView_WebNodeFormalParam_type _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pwebPageView_WebNodeFormalParam_type;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_DBString (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpwebPageView_DBString (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpwebPageView_DBString _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_DBString;
#else
_currn = (_TPPpwebPageView_DBString) TreeNodeAlloc (sizeof (struct _TPpwebPageView_DBString));
#endif
_currn->_prod = RULEpwebPageView_DBString;
_currn->_desc1 = (_TSPwebPageView_DBString_keyword1) MkwebPageView_DBString_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_DBString: root of subtree no. 1 can not be made a webPageView_DBString_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_DBString;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_DBString */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_DBString_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_DBString_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_DBString_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_DBString_keyword1;
#else
_currn = (_TPPpwebPageView_DBString_keyword1) TreeNodeAlloc (sizeof (struct _TPpwebPageView_DBString_keyword1));
#endif
_currn->_prod = RULEpwebPageView_DBString_keyword1;
_SETCOORD(_currn)
_TERMACT_pwebPageView_DBString_keyword1;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_DBString_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_DBInteger (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpwebPageView_DBInteger (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpwebPageView_DBInteger _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_DBInteger;
#else
_currn = (_TPPpwebPageView_DBInteger) TreeNodeAlloc (sizeof (struct _TPpwebPageView_DBInteger));
#endif
_currn->_prod = RULEpwebPageView_DBInteger;
_currn->_desc1 = (_TSPwebPageView_DBInteger_keyword1) MkwebPageView_DBInteger_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_DBInteger: root of subtree no. 1 can not be made a webPageView_DBInteger_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_DBInteger;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_DBInteger */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_DBInteger_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_DBInteger_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_DBInteger_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_DBInteger_keyword1;
#else
_currn = (_TPPpwebPageView_DBInteger_keyword1) TreeNodeAlloc (sizeof (struct _TPpwebPageView_DBInteger_keyword1));
#endif
_currn->_prod = RULEpwebPageView_DBInteger_keyword1;
_SETCOORD(_currn)
_TERMACT_pwebPageView_DBInteger_keyword1;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_DBInteger_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_DBFloat (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpwebPageView_DBFloat (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpwebPageView_DBFloat _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_DBFloat;
#else
_currn = (_TPPpwebPageView_DBFloat) TreeNodeAlloc (sizeof (struct _TPpwebPageView_DBFloat));
#endif
_currn->_prod = RULEpwebPageView_DBFloat;
_currn->_desc1 = (_TSPwebPageView_DBFloat_keyword1) MkwebPageView_DBFloat_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_DBFloat: root of subtree no. 1 can not be made a webPageView_DBFloat_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_DBFloat;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_DBFloat */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_DBFloat_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_DBFloat_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_DBFloat_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_DBFloat_keyword1;
#else
_currn = (_TPPpwebPageView_DBFloat_keyword1) TreeNodeAlloc (sizeof (struct _TPpwebPageView_DBFloat_keyword1));
#endif
_currn->_prod = RULEpwebPageView_DBFloat_keyword1;
_SETCOORD(_currn)
_TERMACT_pwebPageView_DBFloat_keyword1;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_DBFloat_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_DBDouble (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpwebPageView_DBDouble (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpwebPageView_DBDouble _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_DBDouble;
#else
_currn = (_TPPpwebPageView_DBDouble) TreeNodeAlloc (sizeof (struct _TPpwebPageView_DBDouble));
#endif
_currn->_prod = RULEpwebPageView_DBDouble;
_currn->_desc1 = (_TSPwebPageView_DBDouble_keyword1) MkwebPageView_DBDouble_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_DBDouble: root of subtree no. 1 can not be made a webPageView_DBDouble_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_DBDouble;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_DBDouble */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_DBDouble_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_DBDouble_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_DBDouble_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_DBDouble_keyword1;
#else
_currn = (_TPPpwebPageView_DBDouble_keyword1) TreeNodeAlloc (sizeof (struct _TPpwebPageView_DBDouble_keyword1));
#endif
_currn->_prod = RULEpwebPageView_DBDouble_keyword1;
_SETCOORD(_currn)
_TERMACT_pwebPageView_DBDouble_keyword1;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_DBDouble_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_DBBigInt (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpwebPageView_DBBigInt (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpwebPageView_DBBigInt _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_DBBigInt;
#else
_currn = (_TPPpwebPageView_DBBigInt) TreeNodeAlloc (sizeof (struct _TPpwebPageView_DBBigInt));
#endif
_currn->_prod = RULEpwebPageView_DBBigInt;
_currn->_desc1 = (_TSPwebPageView_DBBigInt_keyword1) MkwebPageView_DBBigInt_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_DBBigInt: root of subtree no. 1 can not be made a webPageView_DBBigInt_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_DBBigInt;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_DBBigInt */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_DBBigInt_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_DBBigInt_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_DBBigInt_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_DBBigInt_keyword1;
#else
_currn = (_TPPpwebPageView_DBBigInt_keyword1) TreeNodeAlloc (sizeof (struct _TPpwebPageView_DBBigInt_keyword1));
#endif
_currn->_prod = RULEpwebPageView_DBBigInt_keyword1;
_SETCOORD(_currn)
_TERMACT_pwebPageView_DBBigInt_keyword1;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_DBBigInt_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_DBDate (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpwebPageView_DBDate (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpwebPageView_DBDate _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_DBDate;
#else
_currn = (_TPPpwebPageView_DBDate) TreeNodeAlloc (sizeof (struct _TPpwebPageView_DBDate));
#endif
_currn->_prod = RULEpwebPageView_DBDate;
_currn->_desc1 = (_TSPwebPageView_DBDate_keyword1) MkwebPageView_DBDate_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_DBDate: root of subtree no. 1 can not be made a webPageView_DBDate_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_DBDate;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_DBDate */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_DBDate_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_DBDate_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_DBDate_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_DBDate_keyword1;
#else
_currn = (_TPPpwebPageView_DBDate_keyword1) TreeNodeAlloc (sizeof (struct _TPpwebPageView_DBDate_keyword1));
#endif
_currn->_prod = RULEpwebPageView_DBDate_keyword1;
_SETCOORD(_currn)
_TERMACT_pwebPageView_DBDate_keyword1;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_DBDate_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_DBTimeStamp (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpwebPageView_DBTimeStamp (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpwebPageView_DBTimeStamp _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_DBTimeStamp;
#else
_currn = (_TPPpwebPageView_DBTimeStamp) TreeNodeAlloc (sizeof (struct _TPpwebPageView_DBTimeStamp));
#endif
_currn->_prod = RULEpwebPageView_DBTimeStamp;
_currn->_desc1 = (_TSPwebPageView_DBTimeStamp_keyword1) MkwebPageView_DBTimeStamp_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_DBTimeStamp: root of subtree no. 1 can not be made a webPageView_DBTimeStamp_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_DBTimeStamp;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_DBTimeStamp */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_DBTimeStamp_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_DBTimeStamp_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_DBTimeStamp_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_DBTimeStamp_keyword1;
#else
_currn = (_TPPpwebPageView_DBTimeStamp_keyword1) TreeNodeAlloc (sizeof (struct _TPpwebPageView_DBTimeStamp_keyword1));
#endif
_currn->_prod = RULEpwebPageView_DBTimeStamp_keyword1;
_SETCOORD(_currn)
_TERMACT_pwebPageView_DBTimeStamp_keyword1;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_DBTimeStamp_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_DBBoolean (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpwebPageView_DBBoolean (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpwebPageView_DBBoolean _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_DBBoolean;
#else
_currn = (_TPPpwebPageView_DBBoolean) TreeNodeAlloc (sizeof (struct _TPpwebPageView_DBBoolean));
#endif
_currn->_prod = RULEpwebPageView_DBBoolean;
_currn->_desc1 = (_TSPwebPageView_DBBoolean_keyword1) MkwebPageView_DBBoolean_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_DBBoolean: root of subtree no. 1 can not be made a webPageView_DBBoolean_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_DBBoolean;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_DBBoolean */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_DBBoolean_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_DBBoolean_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_DBBoolean_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_DBBoolean_keyword1;
#else
_currn = (_TPPpwebPageView_DBBoolean_keyword1) TreeNodeAlloc (sizeof (struct _TPpwebPageView_DBBoolean_keyword1));
#endif
_currn->_prod = RULEpwebPageView_DBBoolean_keyword1;
_SETCOORD(_currn)
_TERMACT_pwebPageView_DBBoolean_keyword1;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_DBBoolean_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_DBBlob (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpwebPageView_DBBlob (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpwebPageView_DBBlob _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_DBBlob;
#else
_currn = (_TPPpwebPageView_DBBlob) TreeNodeAlloc (sizeof (struct _TPpwebPageView_DBBlob));
#endif
_currn->_prod = RULEpwebPageView_DBBlob;
_currn->_desc1 = (_TSPwebPageView_DBBlob_keyword1) MkwebPageView_DBBlob_keyword1 (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_DBBlob: root of subtree no. 1 can not be made a webPageView_DBBlob_keyword1 node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_DBBlob;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_DBBlob */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_DBBlob_keyword1 (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_DBBlob_keyword1 (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_DBBlob_keyword1 _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_DBBlob_keyword1;
#else
_currn = (_TPPpwebPageView_DBBlob_keyword1) TreeNodeAlloc (sizeof (struct _TPpwebPageView_DBBlob_keyword1));
#endif
_currn->_prod = RULEpwebPageView_DBBlob_keyword1;
_SETCOORD(_currn)
_TERMACT_pwebPageView_DBBlob_keyword1;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_DBBlob_keyword1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_Formular (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkpwebPageView_Formular (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPpwebPageView_Formular _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_Formular;
#else
_currn = (_TPPpwebPageView_Formular) TreeNodeAlloc (sizeof (struct _TPpwebPageView_Formular));
#endif
_currn->_prod = RULEpwebPageView_Formular;
_currn->_desc1 = (_TSPwebPageView_Formular_styleSheetBlock) MkwebPageView_Formular_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_Formular: root of subtree no. 1 can not be made a webPageView_Formular_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSPwebPageView_Formular_action) MkwebPageView_Formular_action (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_Formular: root of subtree no. 2 can not be made a webPageView_Formular_action node ", 0, _coordref);
_currn->_desc3 = (_TSPwebPageView_Formular_elements) MkwebPageView_Formular_elements (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_Formular: root of subtree no. 3 can not be made a webPageView_Formular_elements node ", 0, _coordref);
_currn->_desc4 = (_TSPwebPageView_Formular_datasource) MkwebPageView_Formular_datasource (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_Formular: root of subtree no. 4 can not be made a webPageView_Formular_datasource node ", 0, _coordref);
_currn->_desc5 = (_TSPwebPageView_Formular_actualParams) MkwebPageView_Formular_actualParams (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_Formular: root of subtree no. 5 can not be made a webPageView_Formular_actualParams node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_Formular;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_Formular */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_Formular_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_Formular_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_Formular_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_Formular_styleSheetBlock;
#else
_currn = (_TPPpwebPageView_Formular_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpwebPageView_Formular_styleSheetBlock));
#endif
_currn->_prod = RULEpwebPageView_Formular_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_pwebPageView_Formular_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_Formular_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_Formular_action (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_Formular_action (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_Formular_action _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_Formular_action;
#else
_currn = (_TPPpwebPageView_Formular_action) TreeNodeAlloc (sizeof (struct _TPpwebPageView_Formular_action));
#endif
_currn->_prod = RULEpwebPageView_Formular_action;
_SETCOORD(_currn)
_TERMACT_pwebPageView_Formular_action;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_Formular_action */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_FormularRadioGroupElementpwebPageView_Formular_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_FormularRadioGroupElementpwebPageView_Formular_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_FormularRadioGroupElementpwebPageView_Formular_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_FormularRadioGroupElementpwebPageView_Formular_elements;
#else
	_new = (_TPPLST_webPageView_FormularRadioGroupElementpwebPageView_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_FormularRadioGroupElementpwebPageView_Formular_elements));
#endif
	_new->_prod = RULELST_webPageView_FormularRadioGroupElementpwebPageView_Formular_elements;
	_new->_desc1 = (_TSPwebPageView_FormularRadioGroupElement)elem;
	_new->_desc2 = (_TSPLST_webPageView_Formular_elements)_new;
	{_TPPLST_webPageView_FormularRadioGroupElementpwebPageView_Formular_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_FormularRadioGroupElementpwebPageView_Formular_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_FormularTextElementpwebPageView_Formular_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_FormularTextElementpwebPageView_Formular_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_FormularTextElementpwebPageView_Formular_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_FormularTextElementpwebPageView_Formular_elements;
#else
	_new = (_TPPLST_webPageView_FormularTextElementpwebPageView_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_FormularTextElementpwebPageView_Formular_elements));
#endif
	_new->_prod = RULELST_webPageView_FormularTextElementpwebPageView_Formular_elements;
	_new->_desc1 = (_TSPwebPageView_FormularTextElement)elem;
	_new->_desc2 = (_TSPLST_webPageView_Formular_elements)_new;
	{_TPPLST_webPageView_FormularTextElementpwebPageView_Formular_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_FormularTextElementpwebPageView_Formular_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_FormularTextAreaElementpwebPageView_Formular_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_FormularTextAreaElementpwebPageView_Formular_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_FormularTextAreaElementpwebPageView_Formular_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_FormularTextAreaElementpwebPageView_Formular_elements;
#else
	_new = (_TPPLST_webPageView_FormularTextAreaElementpwebPageView_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_FormularTextAreaElementpwebPageView_Formular_elements));
#endif
	_new->_prod = RULELST_webPageView_FormularTextAreaElementpwebPageView_Formular_elements;
	_new->_desc1 = (_TSPwebPageView_FormularTextAreaElement)elem;
	_new->_desc2 = (_TSPLST_webPageView_Formular_elements)_new;
	{_TPPLST_webPageView_FormularTextAreaElementpwebPageView_Formular_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_FormularTextAreaElementpwebPageView_Formular_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_FormularPasswordElementpwebPageView_Formular_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_FormularPasswordElementpwebPageView_Formular_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_FormularPasswordElementpwebPageView_Formular_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_FormularPasswordElementpwebPageView_Formular_elements;
#else
	_new = (_TPPLST_webPageView_FormularPasswordElementpwebPageView_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_FormularPasswordElementpwebPageView_Formular_elements));
#endif
	_new->_prod = RULELST_webPageView_FormularPasswordElementpwebPageView_Formular_elements;
	_new->_desc1 = (_TSPwebPageView_FormularPasswordElement)elem;
	_new->_desc2 = (_TSPLST_webPageView_Formular_elements)_new;
	{_TPPLST_webPageView_FormularPasswordElementpwebPageView_Formular_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_FormularPasswordElementpwebPageView_Formular_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_FormularCheckboxElementpwebPageView_Formular_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_FormularCheckboxElementpwebPageView_Formular_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_FormularCheckboxElementpwebPageView_Formular_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_FormularCheckboxElementpwebPageView_Formular_elements;
#else
	_new = (_TPPLST_webPageView_FormularCheckboxElementpwebPageView_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_FormularCheckboxElementpwebPageView_Formular_elements));
#endif
	_new->_prod = RULELST_webPageView_FormularCheckboxElementpwebPageView_Formular_elements;
	_new->_desc1 = (_TSPwebPageView_FormularCheckboxElement)elem;
	_new->_desc2 = (_TSPLST_webPageView_Formular_elements)_new;
	{_TPPLST_webPageView_FormularCheckboxElementpwebPageView_Formular_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_FormularCheckboxElementpwebPageView_Formular_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_FormularSelectElementpwebPageView_Formular_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_FormularSelectElementpwebPageView_Formular_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_FormularSelectElementpwebPageView_Formular_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_FormularSelectElementpwebPageView_Formular_elements;
#else
	_new = (_TPPLST_webPageView_FormularSelectElementpwebPageView_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_FormularSelectElementpwebPageView_Formular_elements));
#endif
	_new->_prod = RULELST_webPageView_FormularSelectElementpwebPageView_Formular_elements;
	_new->_desc1 = (_TSPwebPageView_FormularSelectElement)elem;
	_new->_desc2 = (_TSPLST_webPageView_Formular_elements)_new;
	{_TPPLST_webPageView_FormularSelectElementpwebPageView_Formular_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_FormularSelectElementpwebPageView_Formular_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempwebPageView_Formular_elements (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempwebPageView_Formular_elements (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBwebPageView_FormularRadioGroupElement)) 
		return _MkwebPageView_FormularRadioGroupElementpwebPageView_Formular_elements (_coordref, MkwebPageView_FormularRadioGroupElement (_coordref, t));
	if (IsSymb (t, SYMBwebPageView_FormularTextElement)) 
		return _MkwebPageView_FormularTextElementpwebPageView_Formular_elements (_coordref, MkwebPageView_FormularTextElement (_coordref, t));
	if (IsSymb (t, SYMBwebPageView_FormularTextAreaElement)) 
		return _MkwebPageView_FormularTextAreaElementpwebPageView_Formular_elements (_coordref, MkwebPageView_FormularTextAreaElement (_coordref, t));
	if (IsSymb (t, SYMBwebPageView_FormularPasswordElement)) 
		return _MkwebPageView_FormularPasswordElementpwebPageView_Formular_elements (_coordref, MkwebPageView_FormularPasswordElement (_coordref, t));
	if (IsSymb (t, SYMBwebPageView_FormularCheckboxElement)) 
		return _MkwebPageView_FormularCheckboxElementpwebPageView_Formular_elements (_coordref, MkwebPageView_FormularCheckboxElement (_coordref, t));
	if (IsSymb (t, SYMBwebPageView_FormularSelectElement)) 
		return _MkwebPageView_FormularSelectElementpwebPageView_Formular_elements (_coordref, MkwebPageView_FormularSelectElement (_coordref, t));
	message (DEADLY, "a pwebPageView_Formular_elements list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pwebPageView_Formular_elements (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pwebPageView_Formular_elements (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_webPageView_Formular_elements)
	l =_ElempwebPageView_Formular_elements (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_webPageView_Formular_elements)
	r = _ElempwebPageView_Formular_elements (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pwebPageView_Formular_elements)r)->_desc2);
((_TPPLST_0pwebPageView_Formular_elements)r)->_desc2 = ((_TPPLST_0pwebPageView_Formular_elements)l)->_desc2;
((_TPPLST_0pwebPageView_Formular_elements)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_Formular_elements (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpwebPageView_Formular_elements (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpwebPageView_Formular_elements root;
	_TPPLST_0pwebPageView_Formular_elements list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pwebPageView_Formular_elements)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_webPageView_Formular_elements)
		t = _ElempwebPageView_Formular_elements (_coordref, t);
	remain = (_TPPLST_0pwebPageView_Formular_elements)(((_TPPLST_0pwebPageView_Formular_elements) t)->_desc2);
	((_TPPLST_0pwebPageView_Formular_elements) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pwebPageView_Formular_elements;
#else
list = (_TPPLST_0pwebPageView_Formular_elements) TreeNodeAlloc (sizeof (struct _TPLST_0pwebPageView_Formular_elements));
#endif
list->_prod = RULELST_0pwebPageView_Formular_elements;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pwebPageView_Formular_elements)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_webPageView_FormularRadioGroupElementpwebPageView_Formular_elements)
	{_TPPLST_webPageView_FormularRadioGroupElementpwebPageView_Formular_elements _currn = _currn;
	 _TERMACT_LST_webPageView_FormularRadioGroupElementpwebPageView_Formular_elements;
	} this code has been moved to _ElempwebPageView_Formular_elements*/
	/*if (_currn->_prod == RULELST_webPageView_FormularTextElementpwebPageView_Formular_elements)
	{_TPPLST_webPageView_FormularTextElementpwebPageView_Formular_elements _currn = _currn;
	 _TERMACT_LST_webPageView_FormularTextElementpwebPageView_Formular_elements;
	} this code has been moved to _ElempwebPageView_Formular_elements*/
	/*if (_currn->_prod == RULELST_webPageView_FormularTextAreaElementpwebPageView_Formular_elements)
	{_TPPLST_webPageView_FormularTextAreaElementpwebPageView_Formular_elements _currn = _currn;
	 _TERMACT_LST_webPageView_FormularTextAreaElementpwebPageView_Formular_elements;
	} this code has been moved to _ElempwebPageView_Formular_elements*/
	/*if (_currn->_prod == RULELST_webPageView_FormularPasswordElementpwebPageView_Formular_elements)
	{_TPPLST_webPageView_FormularPasswordElementpwebPageView_Formular_elements _currn = _currn;
	 _TERMACT_LST_webPageView_FormularPasswordElementpwebPageView_Formular_elements;
	} this code has been moved to _ElempwebPageView_Formular_elements*/
	/*if (_currn->_prod == RULELST_webPageView_FormularCheckboxElementpwebPageView_Formular_elements)
	{_TPPLST_webPageView_FormularCheckboxElementpwebPageView_Formular_elements _currn = _currn;
	 _TERMACT_LST_webPageView_FormularCheckboxElementpwebPageView_Formular_elements;
	} this code has been moved to _ElempwebPageView_Formular_elements*/
	/*if (_currn->_prod == RULELST_webPageView_FormularSelectElementpwebPageView_Formular_elements)
	{_TPPLST_webPageView_FormularSelectElementpwebPageView_Formular_elements _currn = _currn;
	 _TERMACT_LST_webPageView_FormularSelectElementpwebPageView_Formular_elements;
	} this code has been moved to _ElempwebPageView_Formular_elements*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpwebPageView_Formular_elements;
#else
root = (_TPPpwebPageView_Formular_elements) TreeNodeAlloc (sizeof (struct _TPpwebPageView_Formular_elements));
#endif
root->_prod = RULEpwebPageView_Formular_elements;
root->_desc1 = (_TSPLST_webPageView_Formular_elements)list;
{ _TPPpwebPageView_Formular_elements _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pwebPageView_Formular_elements;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_Formular_datasource (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_Formular_datasource (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_Formular_datasource _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_Formular_datasource;
#else
_currn = (_TPPpwebPageView_Formular_datasource) TreeNodeAlloc (sizeof (struct _TPpwebPageView_Formular_datasource));
#endif
_currn->_prod = RULEpwebPageView_Formular_datasource;
_SETCOORD(_currn)
_TERMACT_pwebPageView_Formular_datasource;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_Formular_datasource */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_WebNodeActualParampwebPageView_Formular_actualParams (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_WebNodeActualParampwebPageView_Formular_actualParams (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_WebNodeActualParampwebPageView_Formular_actualParams _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_WebNodeActualParampwebPageView_Formular_actualParams;
#else
	_new = (_TPPLST_webPageView_WebNodeActualParampwebPageView_Formular_actualParams) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_WebNodeActualParampwebPageView_Formular_actualParams));
#endif
	_new->_prod = RULELST_webPageView_WebNodeActualParampwebPageView_Formular_actualParams;
	_new->_desc1 = (_TSPwebPageView_WebNodeActualParam)elem;
	_new->_desc2 = (_TSPLST_webPageView_Formular_actualParams)_new;
	{_TPPLST_webPageView_WebNodeActualParampwebPageView_Formular_actualParams _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_WebNodeActualParampwebPageView_Formular_actualParams;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempwebPageView_Formular_actualParams (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempwebPageView_Formular_actualParams (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBwebPageView_WebNodeActualParam)) 
		return _MkwebPageView_WebNodeActualParampwebPageView_Formular_actualParams (_coordref, MkwebPageView_WebNodeActualParam (_coordref, t));
	message (DEADLY, "a pwebPageView_Formular_actualParams list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pwebPageView_Formular_actualParams (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pwebPageView_Formular_actualParams (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_webPageView_Formular_actualParams)
	l =_ElempwebPageView_Formular_actualParams (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_webPageView_Formular_actualParams)
	r = _ElempwebPageView_Formular_actualParams (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pwebPageView_Formular_actualParams)r)->_desc2);
((_TPPLST_0pwebPageView_Formular_actualParams)r)->_desc2 = ((_TPPLST_0pwebPageView_Formular_actualParams)l)->_desc2;
((_TPPLST_0pwebPageView_Formular_actualParams)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_Formular_actualParams (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpwebPageView_Formular_actualParams (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpwebPageView_Formular_actualParams root;
	_TPPLST_0pwebPageView_Formular_actualParams list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pwebPageView_Formular_actualParams)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_webPageView_Formular_actualParams)
		t = _ElempwebPageView_Formular_actualParams (_coordref, t);
	remain = (_TPPLST_0pwebPageView_Formular_actualParams)(((_TPPLST_0pwebPageView_Formular_actualParams) t)->_desc2);
	((_TPPLST_0pwebPageView_Formular_actualParams) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pwebPageView_Formular_actualParams;
#else
list = (_TPPLST_0pwebPageView_Formular_actualParams) TreeNodeAlloc (sizeof (struct _TPLST_0pwebPageView_Formular_actualParams));
#endif
list->_prod = RULELST_0pwebPageView_Formular_actualParams;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pwebPageView_Formular_actualParams)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_webPageView_WebNodeActualParampwebPageView_Formular_actualParams)
	{_TPPLST_webPageView_WebNodeActualParampwebPageView_Formular_actualParams _currn = _currn;
	 _TERMACT_LST_webPageView_WebNodeActualParampwebPageView_Formular_actualParams;
	} this code has been moved to _ElempwebPageView_Formular_actualParams*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpwebPageView_Formular_actualParams;
#else
root = (_TPPpwebPageView_Formular_actualParams) TreeNodeAlloc (sizeof (struct _TPpwebPageView_Formular_actualParams));
#endif
root->_prod = RULEpwebPageView_Formular_actualParams;
root->_desc1 = (_TSPLST_webPageView_Formular_actualParams)list;
{ _TPPpwebPageView_Formular_actualParams _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pwebPageView_Formular_actualParams;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularRadioGroupElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5, NODEPTR _desc6, NODEPTR _desc7)
#else
NODEPTR MkpwebPageView_FormularRadioGroupElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5,_desc6,_desc7)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
	NODEPTR _desc6;
	NODEPTR _desc7;
#endif
{	_TPPpwebPageView_FormularRadioGroupElement _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularRadioGroupElement;
#else
_currn = (_TPPpwebPageView_FormularRadioGroupElement) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularRadioGroupElement));
#endif
_currn->_prod = RULEpwebPageView_FormularRadioGroupElement;
_currn->_desc1 = (_TSPwebPageView_FormularRadioGroupElement_name) MkwebPageView_FormularRadioGroupElement_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularRadioGroupElement: root of subtree no. 1 can not be made a webPageView_FormularRadioGroupElement_name node ", 0, _coordref);
_currn->_desc2 = (_TSPwebPageView_FormularRadioGroupElement_label) MkwebPageView_FormularRadioGroupElement_label (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularRadioGroupElement: root of subtree no. 2 can not be made a webPageView_FormularRadioGroupElement_label node ", 0, _coordref);
_currn->_desc3 = (_TSPwebPageView_FormularRadioGroupElement_description) MkwebPageView_FormularRadioGroupElement_description (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularRadioGroupElement: root of subtree no. 3 can not be made a webPageView_FormularRadioGroupElement_description node ", 0, _coordref);
_currn->_desc4 = (_TSPwebPageView_FormularRadioGroupElement_pvalue) MkwebPageView_FormularRadioGroupElement_pvalue (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularRadioGroupElement: root of subtree no. 4 can not be made a webPageView_FormularRadioGroupElement_pvalue node ", 0, _coordref);
_currn->_desc5 = (_TSPwebPageView_FormularRadioGroupElement_condition) MkwebPageView_FormularRadioGroupElement_condition (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularRadioGroupElement: root of subtree no. 5 can not be made a webPageView_FormularRadioGroupElement_condition node ", 0, _coordref);
_currn->_desc6 = (_TSPwebPageView_FormularRadioGroupElement_elements) MkwebPageView_FormularRadioGroupElement_elements (_coordref, _desc6);	
if (((NODEPTR)_currn->_desc6) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularRadioGroupElement: root of subtree no. 6 can not be made a webPageView_FormularRadioGroupElement_elements node ", 0, _coordref);
_currn->_desc7 = (_TSPwebPageView_FormularRadioGroupElement_checked) MkwebPageView_FormularRadioGroupElement_checked (_coordref, _desc7);	
if (((NODEPTR)_currn->_desc7) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularRadioGroupElement: root of subtree no. 7 can not be made a webPageView_FormularRadioGroupElement_checked node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularRadioGroupElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularRadioGroupElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularRadioGroupElement_name (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularRadioGroupElement_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularRadioGroupElement_name _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularRadioGroupElement_name;
#else
_currn = (_TPPpwebPageView_FormularRadioGroupElement_name) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularRadioGroupElement_name));
#endif
_currn->_prod = RULEpwebPageView_FormularRadioGroupElement_name;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularRadioGroupElement_name;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularRadioGroupElement_name */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularRadioGroupElement_label (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularRadioGroupElement_label (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularRadioGroupElement_label _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularRadioGroupElement_label;
#else
_currn = (_TPPpwebPageView_FormularRadioGroupElement_label) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularRadioGroupElement_label));
#endif
_currn->_prod = RULEpwebPageView_FormularRadioGroupElement_label;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularRadioGroupElement_label;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularRadioGroupElement_label */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularRadioGroupElement_description (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularRadioGroupElement_description (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularRadioGroupElement_description _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularRadioGroupElement_description;
#else
_currn = (_TPPpwebPageView_FormularRadioGroupElement_description) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularRadioGroupElement_description));
#endif
_currn->_prod = RULEpwebPageView_FormularRadioGroupElement_description;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularRadioGroupElement_description;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularRadioGroupElement_description */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularRadioGroupElement_pvalue (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularRadioGroupElement_pvalue (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularRadioGroupElement_pvalue _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularRadioGroupElement_pvalue;
#else
_currn = (_TPPpwebPageView_FormularRadioGroupElement_pvalue) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularRadioGroupElement_pvalue));
#endif
_currn->_prod = RULEpwebPageView_FormularRadioGroupElement_pvalue;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularRadioGroupElement_pvalue;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularRadioGroupElement_pvalue */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_FormularRadioElementpwebPageView_FormularRadioGroupElement_elements (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_FormularRadioElementpwebPageView_FormularRadioGroupElement_elements (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_FormularRadioElementpwebPageView_FormularRadioGroupElement_elements _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_FormularRadioElementpwebPageView_FormularRadioGroupElement_elements;
#else
	_new = (_TPPLST_webPageView_FormularRadioElementpwebPageView_FormularRadioGroupElement_elements) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_FormularRadioElementpwebPageView_FormularRadioGroupElement_elements));
#endif
	_new->_prod = RULELST_webPageView_FormularRadioElementpwebPageView_FormularRadioGroupElement_elements;
	_new->_desc1 = (_TSPwebPageView_FormularRadioElement)elem;
	_new->_desc2 = (_TSPLST_webPageView_FormularRadioGroupElement_elements)_new;
	{_TPPLST_webPageView_FormularRadioElementpwebPageView_FormularRadioGroupElement_elements _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_FormularRadioElementpwebPageView_FormularRadioGroupElement_elements;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempwebPageView_FormularRadioGroupElement_elements (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempwebPageView_FormularRadioGroupElement_elements (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBwebPageView_FormularRadioElement)) 
		return _MkwebPageView_FormularRadioElementpwebPageView_FormularRadioGroupElement_elements (_coordref, MkwebPageView_FormularRadioElement (_coordref, t));
	message (DEADLY, "a pwebPageView_FormularRadioGroupElement_elements list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pwebPageView_FormularRadioGroupElement_elements (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pwebPageView_FormularRadioGroupElement_elements (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_webPageView_FormularRadioGroupElement_elements)
	l =_ElempwebPageView_FormularRadioGroupElement_elements (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_webPageView_FormularRadioGroupElement_elements)
	r = _ElempwebPageView_FormularRadioGroupElement_elements (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pwebPageView_FormularRadioGroupElement_elements)r)->_desc2);
((_TPPLST_0pwebPageView_FormularRadioGroupElement_elements)r)->_desc2 = ((_TPPLST_0pwebPageView_FormularRadioGroupElement_elements)l)->_desc2;
((_TPPLST_0pwebPageView_FormularRadioGroupElement_elements)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularRadioGroupElement_elements (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpwebPageView_FormularRadioGroupElement_elements (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpwebPageView_FormularRadioGroupElement_elements root;
	_TPPLST_0pwebPageView_FormularRadioGroupElement_elements list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pwebPageView_FormularRadioGroupElement_elements)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_webPageView_FormularRadioGroupElement_elements)
		t = _ElempwebPageView_FormularRadioGroupElement_elements (_coordref, t);
	remain = (_TPPLST_0pwebPageView_FormularRadioGroupElement_elements)(((_TPPLST_0pwebPageView_FormularRadioGroupElement_elements) t)->_desc2);
	((_TPPLST_0pwebPageView_FormularRadioGroupElement_elements) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pwebPageView_FormularRadioGroupElement_elements;
#else
list = (_TPPLST_0pwebPageView_FormularRadioGroupElement_elements) TreeNodeAlloc (sizeof (struct _TPLST_0pwebPageView_FormularRadioGroupElement_elements));
#endif
list->_prod = RULELST_0pwebPageView_FormularRadioGroupElement_elements;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pwebPageView_FormularRadioGroupElement_elements)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_webPageView_FormularRadioElementpwebPageView_FormularRadioGroupElement_elements)
	{_TPPLST_webPageView_FormularRadioElementpwebPageView_FormularRadioGroupElement_elements _currn = _currn;
	 _TERMACT_LST_webPageView_FormularRadioElementpwebPageView_FormularRadioGroupElement_elements;
	} this code has been moved to _ElempwebPageView_FormularRadioGroupElement_elements*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpwebPageView_FormularRadioGroupElement_elements;
#else
root = (_TPPpwebPageView_FormularRadioGroupElement_elements) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularRadioGroupElement_elements));
#endif
root->_prod = RULEpwebPageView_FormularRadioGroupElement_elements;
root->_desc1 = (_TSPLST_webPageView_FormularRadioGroupElement_elements)list;
{ _TPPpwebPageView_FormularRadioGroupElement_elements _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pwebPageView_FormularRadioGroupElement_elements;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularRadioGroupElement_checked (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularRadioGroupElement_checked (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularRadioGroupElement_checked _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularRadioGroupElement_checked;
#else
_currn = (_TPPpwebPageView_FormularRadioGroupElement_checked) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularRadioGroupElement_checked));
#endif
_currn->_prod = RULEpwebPageView_FormularRadioGroupElement_checked;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularRadioGroupElement_checked;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularRadioGroupElement_checked */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularConstantSelectOption (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpwebPageView_FormularConstantSelectOption (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpwebPageView_FormularConstantSelectOption _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularConstantSelectOption;
#else
_currn = (_TPPpwebPageView_FormularConstantSelectOption) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularConstantSelectOption));
#endif
_currn->_prod = RULEpwebPageView_FormularConstantSelectOption;
_currn->_desc1 = (_TSPwebPageView_FormularConstantSelectOption_values) MkwebPageView_FormularConstantSelectOption_values (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularConstantSelectOption: root of subtree no. 1 can not be made a webPageView_FormularConstantSelectOption_values node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularConstantSelectOption;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularConstantSelectOption */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularConstantSelectOption_values (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularConstantSelectOption_values (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularConstantSelectOption_values _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularConstantSelectOption_values;
#else
_currn = (_TPPpwebPageView_FormularConstantSelectOption_values) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularConstantSelectOption_values));
#endif
_currn->_prod = RULEpwebPageView_FormularConstantSelectOption_values;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularConstantSelectOption_values;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularConstantSelectOption_values */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularDataSourceSelectOption (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpwebPageView_FormularDataSourceSelectOption (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpwebPageView_FormularDataSourceSelectOption _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularDataSourceSelectOption;
#else
_currn = (_TPPpwebPageView_FormularDataSourceSelectOption) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularDataSourceSelectOption));
#endif
_currn->_prod = RULEpwebPageView_FormularDataSourceSelectOption;
_currn->_desc1 = (_TSPwebPageView_FormularDataSourceSelectOption_values) MkwebPageView_FormularDataSourceSelectOption_values (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularDataSourceSelectOption: root of subtree no. 1 can not be made a webPageView_FormularDataSourceSelectOption_values node ", 0, _coordref);
_currn->_desc2 = (_TSPwebPageView_FormularDataSourceSelectOption_params) MkwebPageView_FormularDataSourceSelectOption_params (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularDataSourceSelectOption: root of subtree no. 2 can not be made a webPageView_FormularDataSourceSelectOption_params node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularDataSourceSelectOption;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularDataSourceSelectOption */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularDataSourceSelectOption_values (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularDataSourceSelectOption_values (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularDataSourceSelectOption_values _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularDataSourceSelectOption_values;
#else
_currn = (_TPPpwebPageView_FormularDataSourceSelectOption_values) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularDataSourceSelectOption_values));
#endif
_currn->_prod = RULEpwebPageView_FormularDataSourceSelectOption_values;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularDataSourceSelectOption_values;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularDataSourceSelectOption_values */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_DataSourceActualParampwebPageView_FormularDataSourceSelectOption_params (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_DataSourceActualParampwebPageView_FormularDataSourceSelectOption_params (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_DataSourceActualParampwebPageView_FormularDataSourceSelectOption_params _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_DataSourceActualParampwebPageView_FormularDataSourceSelectOption_params;
#else
	_new = (_TPPLST_webPageView_DataSourceActualParampwebPageView_FormularDataSourceSelectOption_params) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_DataSourceActualParampwebPageView_FormularDataSourceSelectOption_params));
#endif
	_new->_prod = RULELST_webPageView_DataSourceActualParampwebPageView_FormularDataSourceSelectOption_params;
	_new->_desc1 = (_TSPwebPageView_DataSourceActualParam)elem;
	_new->_desc2 = (_TSPLST_webPageView_FormularDataSourceSelectOption_params)_new;
	{_TPPLST_webPageView_DataSourceActualParampwebPageView_FormularDataSourceSelectOption_params _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_DataSourceActualParampwebPageView_FormularDataSourceSelectOption_params;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempwebPageView_FormularDataSourceSelectOption_params (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempwebPageView_FormularDataSourceSelectOption_params (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBwebPageView_DataSourceActualParam)) 
		return _MkwebPageView_DataSourceActualParampwebPageView_FormularDataSourceSelectOption_params (_coordref, MkwebPageView_DataSourceActualParam (_coordref, t));
	message (DEADLY, "a pwebPageView_FormularDataSourceSelectOption_params list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pwebPageView_FormularDataSourceSelectOption_params (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pwebPageView_FormularDataSourceSelectOption_params (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_webPageView_FormularDataSourceSelectOption_params)
	l =_ElempwebPageView_FormularDataSourceSelectOption_params (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_webPageView_FormularDataSourceSelectOption_params)
	r = _ElempwebPageView_FormularDataSourceSelectOption_params (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pwebPageView_FormularDataSourceSelectOption_params)r)->_desc2);
((_TPPLST_0pwebPageView_FormularDataSourceSelectOption_params)r)->_desc2 = ((_TPPLST_0pwebPageView_FormularDataSourceSelectOption_params)l)->_desc2;
((_TPPLST_0pwebPageView_FormularDataSourceSelectOption_params)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularDataSourceSelectOption_params (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpwebPageView_FormularDataSourceSelectOption_params (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpwebPageView_FormularDataSourceSelectOption_params root;
	_TPPLST_0pwebPageView_FormularDataSourceSelectOption_params list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pwebPageView_FormularDataSourceSelectOption_params)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_webPageView_FormularDataSourceSelectOption_params)
		t = _ElempwebPageView_FormularDataSourceSelectOption_params (_coordref, t);
	remain = (_TPPLST_0pwebPageView_FormularDataSourceSelectOption_params)(((_TPPLST_0pwebPageView_FormularDataSourceSelectOption_params) t)->_desc2);
	((_TPPLST_0pwebPageView_FormularDataSourceSelectOption_params) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pwebPageView_FormularDataSourceSelectOption_params;
#else
list = (_TPPLST_0pwebPageView_FormularDataSourceSelectOption_params) TreeNodeAlloc (sizeof (struct _TPLST_0pwebPageView_FormularDataSourceSelectOption_params));
#endif
list->_prod = RULELST_0pwebPageView_FormularDataSourceSelectOption_params;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pwebPageView_FormularDataSourceSelectOption_params)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_webPageView_DataSourceActualParampwebPageView_FormularDataSourceSelectOption_params)
	{_TPPLST_webPageView_DataSourceActualParampwebPageView_FormularDataSourceSelectOption_params _currn = _currn;
	 _TERMACT_LST_webPageView_DataSourceActualParampwebPageView_FormularDataSourceSelectOption_params;
	} this code has been moved to _ElempwebPageView_FormularDataSourceSelectOption_params*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpwebPageView_FormularDataSourceSelectOption_params;
#else
root = (_TPPpwebPageView_FormularDataSourceSelectOption_params) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularDataSourceSelectOption_params));
#endif
root->_prod = RULEpwebPageView_FormularDataSourceSelectOption_params;
root->_desc1 = (_TSPLST_webPageView_FormularDataSourceSelectOption_params)list;
{ _TPPpwebPageView_FormularDataSourceSelectOption_params _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pwebPageView_FormularDataSourceSelectOption_params;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularPHPScriptSelectOption (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpwebPageView_FormularPHPScriptSelectOption (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpwebPageView_FormularPHPScriptSelectOption _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularPHPScriptSelectOption;
#else
_currn = (_TPPpwebPageView_FormularPHPScriptSelectOption) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularPHPScriptSelectOption));
#endif
_currn->_prod = RULEpwebPageView_FormularPHPScriptSelectOption;
_currn->_desc1 = (_TSPwebPageView_FormularPHPScriptSelectOption_script) MkwebPageView_FormularPHPScriptSelectOption_script (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularPHPScriptSelectOption: root of subtree no. 1 can not be made a webPageView_FormularPHPScriptSelectOption_script node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularPHPScriptSelectOption;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularPHPScriptSelectOption */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularPHPScriptSelectOption_script (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularPHPScriptSelectOption_script (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularPHPScriptSelectOption_script _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularPHPScriptSelectOption_script;
#else
_currn = (_TPPpwebPageView_FormularPHPScriptSelectOption_script) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularPHPScriptSelectOption_script));
#endif
_currn->_prod = RULEpwebPageView_FormularPHPScriptSelectOption_script;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularPHPScriptSelectOption_script;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularPHPScriptSelectOption_script */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularCondition (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpwebPageView_FormularCondition (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpwebPageView_FormularCondition _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularCondition;
#else
_currn = (_TPPpwebPageView_FormularCondition) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularCondition));
#endif
_currn->_prod = RULEpwebPageView_FormularCondition;
_currn->_desc1 = (_TSPwebPageView_FormularCondition_condition) MkwebPageView_FormularCondition_condition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularCondition: root of subtree no. 1 can not be made a webPageView_FormularCondition_condition node ", 0, _coordref);
_currn->_desc2 = (_TSPwebPageView_FormularCondition_params) MkwebPageView_FormularCondition_params (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularCondition: root of subtree no. 2 can not be made a webPageView_FormularCondition_params node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularCondition;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularCondition */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularCondition_condition (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularCondition_condition (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularCondition_condition _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularCondition_condition;
#else
_currn = (_TPPpwebPageView_FormularCondition_condition) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularCondition_condition));
#endif
_currn->_prod = RULEpwebPageView_FormularCondition_condition;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularCondition_condition;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularCondition_condition */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_FormularConditionActualParampwebPageView_FormularCondition_params (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_FormularConditionActualParampwebPageView_FormularCondition_params (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_FormularConditionActualParampwebPageView_FormularCondition_params _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_FormularConditionActualParampwebPageView_FormularCondition_params;
#else
	_new = (_TPPLST_webPageView_FormularConditionActualParampwebPageView_FormularCondition_params) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_FormularConditionActualParampwebPageView_FormularCondition_params));
#endif
	_new->_prod = RULELST_webPageView_FormularConditionActualParampwebPageView_FormularCondition_params;
	_new->_desc1 = (_TSPwebPageView_FormularConditionActualParam)elem;
	_new->_desc2 = (_TSPLST_webPageView_FormularCondition_params)_new;
	{_TPPLST_webPageView_FormularConditionActualParampwebPageView_FormularCondition_params _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_FormularConditionActualParampwebPageView_FormularCondition_params;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempwebPageView_FormularCondition_params (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempwebPageView_FormularCondition_params (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBwebPageView_FormularConditionActualParam)) 
		return _MkwebPageView_FormularConditionActualParampwebPageView_FormularCondition_params (_coordref, MkwebPageView_FormularConditionActualParam (_coordref, t));
	message (DEADLY, "a pwebPageView_FormularCondition_params list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pwebPageView_FormularCondition_params (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pwebPageView_FormularCondition_params (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_webPageView_FormularCondition_params)
	l =_ElempwebPageView_FormularCondition_params (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_webPageView_FormularCondition_params)
	r = _ElempwebPageView_FormularCondition_params (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pwebPageView_FormularCondition_params)r)->_desc2);
((_TPPLST_0pwebPageView_FormularCondition_params)r)->_desc2 = ((_TPPLST_0pwebPageView_FormularCondition_params)l)->_desc2;
((_TPPLST_0pwebPageView_FormularCondition_params)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularCondition_params (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpwebPageView_FormularCondition_params (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpwebPageView_FormularCondition_params root;
	_TPPLST_0pwebPageView_FormularCondition_params list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pwebPageView_FormularCondition_params)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_webPageView_FormularCondition_params)
		t = _ElempwebPageView_FormularCondition_params (_coordref, t);
	remain = (_TPPLST_0pwebPageView_FormularCondition_params)(((_TPPLST_0pwebPageView_FormularCondition_params) t)->_desc2);
	((_TPPLST_0pwebPageView_FormularCondition_params) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pwebPageView_FormularCondition_params;
#else
list = (_TPPLST_0pwebPageView_FormularCondition_params) TreeNodeAlloc (sizeof (struct _TPLST_0pwebPageView_FormularCondition_params));
#endif
list->_prod = RULELST_0pwebPageView_FormularCondition_params;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pwebPageView_FormularCondition_params)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_webPageView_FormularConditionActualParampwebPageView_FormularCondition_params)
	{_TPPLST_webPageView_FormularConditionActualParampwebPageView_FormularCondition_params _currn = _currn;
	 _TERMACT_LST_webPageView_FormularConditionActualParampwebPageView_FormularCondition_params;
	} this code has been moved to _ElempwebPageView_FormularCondition_params*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpwebPageView_FormularCondition_params;
#else
root = (_TPPpwebPageView_FormularCondition_params) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularCondition_params));
#endif
root->_prod = RULEpwebPageView_FormularCondition_params;
root->_desc1 = (_TSPLST_webPageView_FormularCondition_params)list;
{ _TPPpwebPageView_FormularCondition_params _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pwebPageView_FormularCondition_params;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularConditionActualParam (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpwebPageView_FormularConditionActualParam (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpwebPageView_FormularConditionActualParam _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularConditionActualParam;
#else
_currn = (_TPPpwebPageView_FormularConditionActualParam) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularConditionActualParam));
#endif
_currn->_prod = RULEpwebPageView_FormularConditionActualParam;
_currn->_desc1 = (_TSPwebPageView_FormularConditionActualParam_value) MkwebPageView_FormularConditionActualParam_value (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularConditionActualParam: root of subtree no. 1 can not be made a webPageView_FormularConditionActualParam_value node ", 0, _coordref);
_currn->_desc2 = (_TSPwebPageView_FormularConditionActualParam_formalparam) MkwebPageView_FormularConditionActualParam_formalparam (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FormularConditionActualParam: root of subtree no. 2 can not be made a webPageView_FormularConditionActualParam_formalparam node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularConditionActualParam;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularConditionActualParam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularConditionActualParam_value (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularConditionActualParam_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularConditionActualParam_value _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularConditionActualParam_value;
#else
_currn = (_TPPpwebPageView_FormularConditionActualParam_value) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularConditionActualParam_value));
#endif
_currn->_prod = RULEpwebPageView_FormularConditionActualParam_value;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularConditionActualParam_value;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularConditionActualParam_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FormularConditionActualParam_formalparam (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FormularConditionActualParam_formalparam (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FormularConditionActualParam_formalparam _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FormularConditionActualParam_formalparam;
#else
_currn = (_TPPpwebPageView_FormularConditionActualParam_formalparam) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FormularConditionActualParam_formalparam));
#endif
_currn->_prod = RULEpwebPageView_FormularConditionActualParam_formalparam;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FormularConditionActualParam_formalparam;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FormularConditionActualParam_formalparam */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_Paragraph (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpwebPageView_Paragraph (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpwebPageView_Paragraph _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_Paragraph;
#else
_currn = (_TPPpwebPageView_Paragraph) TreeNodeAlloc (sizeof (struct _TPpwebPageView_Paragraph));
#endif
_currn->_prod = RULEpwebPageView_Paragraph;
_currn->_desc1 = (_TSPwebPageView_Paragraph_styleSheetBlock) MkwebPageView_Paragraph_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_Paragraph: root of subtree no. 1 can not be made a webPageView_Paragraph_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSPwebPageView_Paragraph_fragments) MkwebPageView_Paragraph_fragments (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_Paragraph: root of subtree no. 2 can not be made a webPageView_Paragraph_fragments node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_Paragraph;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_Paragraph */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_Paragraph_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_Paragraph_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_Paragraph_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_Paragraph_styleSheetBlock;
#else
_currn = (_TPPpwebPageView_Paragraph_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpwebPageView_Paragraph_styleSheetBlock));
#endif
_currn->_prod = RULEpwebPageView_Paragraph_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_pwebPageView_Paragraph_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_Paragraph_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_TextFragmentpwebPageView_Paragraph_fragments (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_TextFragmentpwebPageView_Paragraph_fragments (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_TextFragmentpwebPageView_Paragraph_fragments _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_TextFragmentpwebPageView_Paragraph_fragments;
#else
	_new = (_TPPLST_webPageView_TextFragmentpwebPageView_Paragraph_fragments) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_TextFragmentpwebPageView_Paragraph_fragments));
#endif
	_new->_prod = RULELST_webPageView_TextFragmentpwebPageView_Paragraph_fragments;
	_new->_desc1 = (_TSPwebPageView_TextFragment)elem;
	_new->_desc2 = (_TSPLST_webPageView_Paragraph_fragments)_new;
	{_TPPLST_webPageView_TextFragmentpwebPageView_Paragraph_fragments _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_TextFragmentpwebPageView_Paragraph_fragments;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_ImageFragmentpwebPageView_Paragraph_fragments (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_ImageFragmentpwebPageView_Paragraph_fragments (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_ImageFragmentpwebPageView_Paragraph_fragments _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_ImageFragmentpwebPageView_Paragraph_fragments;
#else
	_new = (_TPPLST_webPageView_ImageFragmentpwebPageView_Paragraph_fragments) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_ImageFragmentpwebPageView_Paragraph_fragments));
#endif
	_new->_prod = RULELST_webPageView_ImageFragmentpwebPageView_Paragraph_fragments;
	_new->_desc1 = (_TSPwebPageView_ImageFragment)elem;
	_new->_desc2 = (_TSPLST_webPageView_Paragraph_fragments)_new;
	{_TPPLST_webPageView_ImageFragmentpwebPageView_Paragraph_fragments _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_ImageFragmentpwebPageView_Paragraph_fragments;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_OutputElementpwebPageView_Paragraph_fragments (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_OutputElementpwebPageView_Paragraph_fragments (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_OutputElementpwebPageView_Paragraph_fragments _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_OutputElementpwebPageView_Paragraph_fragments;
#else
	_new = (_TPPLST_webPageView_OutputElementpwebPageView_Paragraph_fragments) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_OutputElementpwebPageView_Paragraph_fragments));
#endif
	_new->_prod = RULELST_webPageView_OutputElementpwebPageView_Paragraph_fragments;
	_new->_desc1 = (_TSPwebPageView_OutputElement)elem;
	_new->_desc2 = (_TSPLST_webPageView_Paragraph_fragments)_new;
	{_TPPLST_webPageView_OutputElementpwebPageView_Paragraph_fragments _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_OutputElementpwebPageView_Paragraph_fragments;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempwebPageView_Paragraph_fragments (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempwebPageView_Paragraph_fragments (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBwebPageView_TextFragment)) 
		return _MkwebPageView_TextFragmentpwebPageView_Paragraph_fragments (_coordref, MkwebPageView_TextFragment (_coordref, t));
	if (IsSymb (t, SYMBwebPageView_ImageFragment)) 
		return _MkwebPageView_ImageFragmentpwebPageView_Paragraph_fragments (_coordref, MkwebPageView_ImageFragment (_coordref, t));
	if (IsSymb (t, SYMBwebPageView_OutputElement)) 
		return _MkwebPageView_OutputElementpwebPageView_Paragraph_fragments (_coordref, MkwebPageView_OutputElement (_coordref, t));
	message (DEADLY, "a pwebPageView_Paragraph_fragments list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pwebPageView_Paragraph_fragments (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pwebPageView_Paragraph_fragments (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_webPageView_Paragraph_fragments)
	l =_ElempwebPageView_Paragraph_fragments (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_webPageView_Paragraph_fragments)
	r = _ElempwebPageView_Paragraph_fragments (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pwebPageView_Paragraph_fragments)r)->_desc2);
((_TPPLST_0pwebPageView_Paragraph_fragments)r)->_desc2 = ((_TPPLST_0pwebPageView_Paragraph_fragments)l)->_desc2;
((_TPPLST_0pwebPageView_Paragraph_fragments)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_Paragraph_fragments (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpwebPageView_Paragraph_fragments (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpwebPageView_Paragraph_fragments root;
	_TPPLST_0pwebPageView_Paragraph_fragments list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pwebPageView_Paragraph_fragments)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_webPageView_Paragraph_fragments)
		t = _ElempwebPageView_Paragraph_fragments (_coordref, t);
	remain = (_TPPLST_0pwebPageView_Paragraph_fragments)(((_TPPLST_0pwebPageView_Paragraph_fragments) t)->_desc2);
	((_TPPLST_0pwebPageView_Paragraph_fragments) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pwebPageView_Paragraph_fragments;
#else
list = (_TPPLST_0pwebPageView_Paragraph_fragments) TreeNodeAlloc (sizeof (struct _TPLST_0pwebPageView_Paragraph_fragments));
#endif
list->_prod = RULELST_0pwebPageView_Paragraph_fragments;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pwebPageView_Paragraph_fragments)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_webPageView_TextFragmentpwebPageView_Paragraph_fragments)
	{_TPPLST_webPageView_TextFragmentpwebPageView_Paragraph_fragments _currn = _currn;
	 _TERMACT_LST_webPageView_TextFragmentpwebPageView_Paragraph_fragments;
	} this code has been moved to _ElempwebPageView_Paragraph_fragments*/
	/*if (_currn->_prod == RULELST_webPageView_ImageFragmentpwebPageView_Paragraph_fragments)
	{_TPPLST_webPageView_ImageFragmentpwebPageView_Paragraph_fragments _currn = _currn;
	 _TERMACT_LST_webPageView_ImageFragmentpwebPageView_Paragraph_fragments;
	} this code has been moved to _ElempwebPageView_Paragraph_fragments*/
	/*if (_currn->_prod == RULELST_webPageView_OutputElementpwebPageView_Paragraph_fragments)
	{_TPPLST_webPageView_OutputElementpwebPageView_Paragraph_fragments _currn = _currn;
	 _TERMACT_LST_webPageView_OutputElementpwebPageView_Paragraph_fragments;
	} this code has been moved to _ElempwebPageView_Paragraph_fragments*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpwebPageView_Paragraph_fragments;
#else
root = (_TPPpwebPageView_Paragraph_fragments) TreeNodeAlloc (sizeof (struct _TPpwebPageView_Paragraph_fragments));
#endif
root->_prod = RULEpwebPageView_Paragraph_fragments;
root->_desc1 = (_TSPLST_webPageView_Paragraph_fragments)list;
{ _TPPpwebPageView_Paragraph_fragments _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pwebPageView_Paragraph_fragments;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_TextFragment (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpwebPageView_TextFragment (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpwebPageView_TextFragment _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_TextFragment;
#else
_currn = (_TPPpwebPageView_TextFragment) TreeNodeAlloc (sizeof (struct _TPpwebPageView_TextFragment));
#endif
_currn->_prod = RULEpwebPageView_TextFragment;
_currn->_desc1 = (_TSPwebPageView_TextFragment_link) MkwebPageView_TextFragment_link (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_TextFragment: root of subtree no. 1 can not be made a webPageView_TextFragment_link node ", 0, _coordref);
_currn->_desc2 = (_TSPwebPageView_TextFragment_value) MkwebPageView_TextFragment_value (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_TextFragment: root of subtree no. 2 can not be made a webPageView_TextFragment_value node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_TextFragment;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_TextFragment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_TextFragment_link (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_TextFragment_link (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_TextFragment_link _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_TextFragment_link;
#else
_currn = (_TPPpwebPageView_TextFragment_link) TreeNodeAlloc (sizeof (struct _TPpwebPageView_TextFragment_link));
#endif
_currn->_prod = RULEpwebPageView_TextFragment_link;
_SETCOORD(_currn)
_TERMACT_pwebPageView_TextFragment_link;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_TextFragment_link */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_TextFragment_value (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_TextFragment_value (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_TextFragment_value _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_TextFragment_value;
#else
_currn = (_TPPpwebPageView_TextFragment_value) TreeNodeAlloc (sizeof (struct _TPpwebPageView_TextFragment_value));
#endif
_currn->_prod = RULEpwebPageView_TextFragment_value;
_SETCOORD(_currn)
_TERMACT_pwebPageView_TextFragment_value;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_TextFragment_value */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_ImageFragment (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkpwebPageView_ImageFragment (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPpwebPageView_ImageFragment _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_ImageFragment;
#else
_currn = (_TPPpwebPageView_ImageFragment) TreeNodeAlloc (sizeof (struct _TPpwebPageView_ImageFragment));
#endif
_currn->_prod = RULEpwebPageView_ImageFragment;
_currn->_desc1 = (_TSPwebPageView_ImageFragment_link) MkwebPageView_ImageFragment_link (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_ImageFragment: root of subtree no. 1 can not be made a webPageView_ImageFragment_link node ", 0, _coordref);
_currn->_desc2 = (_TSPwebPageView_ImageFragment_width) MkwebPageView_ImageFragment_width (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_ImageFragment: root of subtree no. 2 can not be made a webPageView_ImageFragment_width node ", 0, _coordref);
_currn->_desc3 = (_TSPwebPageView_ImageFragment_height) MkwebPageView_ImageFragment_height (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_ImageFragment: root of subtree no. 3 can not be made a webPageView_ImageFragment_height node ", 0, _coordref);
_currn->_desc4 = (_TSPwebPageView_ImageFragment_text) MkwebPageView_ImageFragment_text (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_ImageFragment: root of subtree no. 4 can not be made a webPageView_ImageFragment_text node ", 0, _coordref);
_currn->_desc5 = (_TSPwebPageView_ImageFragment_imageName) MkwebPageView_ImageFragment_imageName (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_ImageFragment: root of subtree no. 5 can not be made a webPageView_ImageFragment_imageName node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_ImageFragment;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_ImageFragment */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_ImageFragment_link (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_ImageFragment_link (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_ImageFragment_link _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_ImageFragment_link;
#else
_currn = (_TPPpwebPageView_ImageFragment_link) TreeNodeAlloc (sizeof (struct _TPpwebPageView_ImageFragment_link));
#endif
_currn->_prod = RULEpwebPageView_ImageFragment_link;
_SETCOORD(_currn)
_TERMACT_pwebPageView_ImageFragment_link;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_ImageFragment_link */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_ImageFragment_width (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_ImageFragment_width (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_ImageFragment_width _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_ImageFragment_width;
#else
_currn = (_TPPpwebPageView_ImageFragment_width) TreeNodeAlloc (sizeof (struct _TPpwebPageView_ImageFragment_width));
#endif
_currn->_prod = RULEpwebPageView_ImageFragment_width;
_SETCOORD(_currn)
_TERMACT_pwebPageView_ImageFragment_width;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_ImageFragment_width */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_ImageFragment_height (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_ImageFragment_height (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_ImageFragment_height _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_ImageFragment_height;
#else
_currn = (_TPPpwebPageView_ImageFragment_height) TreeNodeAlloc (sizeof (struct _TPpwebPageView_ImageFragment_height));
#endif
_currn->_prod = RULEpwebPageView_ImageFragment_height;
_SETCOORD(_currn)
_TERMACT_pwebPageView_ImageFragment_height;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_ImageFragment_height */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_ImageFragment_text (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_ImageFragment_text (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_ImageFragment_text _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_ImageFragment_text;
#else
_currn = (_TPPpwebPageView_ImageFragment_text) TreeNodeAlloc (sizeof (struct _TPpwebPageView_ImageFragment_text));
#endif
_currn->_prod = RULEpwebPageView_ImageFragment_text;
_SETCOORD(_currn)
_TERMACT_pwebPageView_ImageFragment_text;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_ImageFragment_text */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_ImageFragment_imageName (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_ImageFragment_imageName (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_ImageFragment_imageName _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_ImageFragment_imageName;
#else
_currn = (_TPPpwebPageView_ImageFragment_imageName) TreeNodeAlloc (sizeof (struct _TPpwebPageView_ImageFragment_imageName));
#endif
_currn->_prod = RULEpwebPageView_ImageFragment_imageName;
_SETCOORD(_currn)
_TERMACT_pwebPageView_ImageFragment_imageName;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_ImageFragment_imageName */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_Heading (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpwebPageView_Heading (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpwebPageView_Heading _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_Heading;
#else
_currn = (_TPPpwebPageView_Heading) TreeNodeAlloc (sizeof (struct _TPpwebPageView_Heading));
#endif
_currn->_prod = RULEpwebPageView_Heading;
_currn->_desc1 = (_TSPwebPageView_Heading_styleSheetBlock) MkwebPageView_Heading_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_Heading: root of subtree no. 1 can not be made a webPageView_Heading_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSPwebPageView_Heading_text) MkwebPageView_Heading_text (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_Heading: root of subtree no. 2 can not be made a webPageView_Heading_text node ", 0, _coordref);
_currn->_desc3 = (_TSPwebPageView_Heading_level) MkwebPageView_Heading_level (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_Heading: root of subtree no. 3 can not be made a webPageView_Heading_level node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_Heading;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_Heading */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_Heading_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_Heading_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_Heading_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_Heading_styleSheetBlock;
#else
_currn = (_TPPpwebPageView_Heading_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpwebPageView_Heading_styleSheetBlock));
#endif
_currn->_prod = RULEpwebPageView_Heading_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_pwebPageView_Heading_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_Heading_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_Heading_text (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_Heading_text (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_Heading_text _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_Heading_text;
#else
_currn = (_TPPpwebPageView_Heading_text) TreeNodeAlloc (sizeof (struct _TPpwebPageView_Heading_text));
#endif
_currn->_prod = RULEpwebPageView_Heading_text;
_SETCOORD(_currn)
_TERMACT_pwebPageView_Heading_text;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_Heading_text */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_Heading_level (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_Heading_level (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_Heading_level _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_Heading_level;
#else
_currn = (_TPPpwebPageView_Heading_level) TreeNodeAlloc (sizeof (struct _TPpwebPageView_Heading_level));
#endif
_currn->_prod = RULEpwebPageView_Heading_level;
_SETCOORD(_currn)
_TERMACT_pwebPageView_Heading_level;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_Heading_level */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_Table (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3)
#else
NODEPTR MkpwebPageView_Table (_coordref, _TERM1,_desc1,_desc2,_desc3)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
#endif
{	_TPPpwebPageView_Table _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_Table;
#else
_currn = (_TPPpwebPageView_Table) TreeNodeAlloc (sizeof (struct _TPpwebPageView_Table));
#endif
_currn->_prod = RULEpwebPageView_Table;
_currn->_desc1 = (_TSPwebPageView_Table_styleSheetBlock) MkwebPageView_Table_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_Table: root of subtree no. 1 can not be made a webPageView_Table_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSPwebPageView_Table_rows) MkwebPageView_Table_rows (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_Table: root of subtree no. 2 can not be made a webPageView_Table_rows node ", 0, _coordref);
_currn->_desc3 = (_TSPwebPageView_Table_columns) MkwebPageView_Table_columns (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_Table: root of subtree no. 3 can not be made a webPageView_Table_columns node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_Table;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_Table */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_Table_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_Table_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_Table_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_Table_styleSheetBlock;
#else
_currn = (_TPPpwebPageView_Table_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpwebPageView_Table_styleSheetBlock));
#endif
_currn->_prod = RULEpwebPageView_Table_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_pwebPageView_Table_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_Table_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_TableRowpwebPageView_Table_rows (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_TableRowpwebPageView_Table_rows (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_TableRowpwebPageView_Table_rows _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_TableRowpwebPageView_Table_rows;
#else
	_new = (_TPPLST_webPageView_TableRowpwebPageView_Table_rows) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_TableRowpwebPageView_Table_rows));
#endif
	_new->_prod = RULELST_webPageView_TableRowpwebPageView_Table_rows;
	_new->_desc1 = (_TSPwebPageView_TableRow)elem;
	_new->_desc2 = (_TSPLST_webPageView_Table_rows)_new;
	{_TPPLST_webPageView_TableRowpwebPageView_Table_rows _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_TableRowpwebPageView_Table_rows;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempwebPageView_Table_rows (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempwebPageView_Table_rows (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBwebPageView_TableRow)) 
		return _MkwebPageView_TableRowpwebPageView_Table_rows (_coordref, MkwebPageView_TableRow (_coordref, t));
	message (DEADLY, "a pwebPageView_Table_rows list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pwebPageView_Table_rows (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pwebPageView_Table_rows (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_webPageView_Table_rows)
	l =_ElempwebPageView_Table_rows (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_webPageView_Table_rows)
	r = _ElempwebPageView_Table_rows (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pwebPageView_Table_rows)r)->_desc2);
((_TPPLST_0pwebPageView_Table_rows)r)->_desc2 = ((_TPPLST_0pwebPageView_Table_rows)l)->_desc2;
((_TPPLST_0pwebPageView_Table_rows)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_Table_rows (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpwebPageView_Table_rows (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpwebPageView_Table_rows root;
	_TPPLST_0pwebPageView_Table_rows list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pwebPageView_Table_rows)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_webPageView_Table_rows)
		t = _ElempwebPageView_Table_rows (_coordref, t);
	remain = (_TPPLST_0pwebPageView_Table_rows)(((_TPPLST_0pwebPageView_Table_rows) t)->_desc2);
	((_TPPLST_0pwebPageView_Table_rows) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pwebPageView_Table_rows;
#else
list = (_TPPLST_0pwebPageView_Table_rows) TreeNodeAlloc (sizeof (struct _TPLST_0pwebPageView_Table_rows));
#endif
list->_prod = RULELST_0pwebPageView_Table_rows;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pwebPageView_Table_rows)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_webPageView_TableRowpwebPageView_Table_rows)
	{_TPPLST_webPageView_TableRowpwebPageView_Table_rows _currn = _currn;
	 _TERMACT_LST_webPageView_TableRowpwebPageView_Table_rows;
	} this code has been moved to _ElempwebPageView_Table_rows*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpwebPageView_Table_rows;
#else
root = (_TPPpwebPageView_Table_rows) TreeNodeAlloc (sizeof (struct _TPpwebPageView_Table_rows));
#endif
root->_prod = RULEpwebPageView_Table_rows;
root->_desc1 = (_TSPLST_webPageView_Table_rows)list;
{ _TPPpwebPageView_Table_rows _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pwebPageView_Table_rows;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_TableColumnpwebPageView_Table_columns (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_TableColumnpwebPageView_Table_columns (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_TableColumnpwebPageView_Table_columns _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_TableColumnpwebPageView_Table_columns;
#else
	_new = (_TPPLST_webPageView_TableColumnpwebPageView_Table_columns) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_TableColumnpwebPageView_Table_columns));
#endif
	_new->_prod = RULELST_webPageView_TableColumnpwebPageView_Table_columns;
	_new->_desc1 = (_TSPwebPageView_TableColumn)elem;
	_new->_desc2 = (_TSPLST_webPageView_Table_columns)_new;
	{_TPPLST_webPageView_TableColumnpwebPageView_Table_columns _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_TableColumnpwebPageView_Table_columns;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempwebPageView_Table_columns (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempwebPageView_Table_columns (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBwebPageView_TableColumn)) 
		return _MkwebPageView_TableColumnpwebPageView_Table_columns (_coordref, MkwebPageView_TableColumn (_coordref, t));
	message (DEADLY, "a pwebPageView_Table_columns list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pwebPageView_Table_columns (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pwebPageView_Table_columns (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_webPageView_Table_columns)
	l =_ElempwebPageView_Table_columns (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_webPageView_Table_columns)
	r = _ElempwebPageView_Table_columns (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pwebPageView_Table_columns)r)->_desc2);
((_TPPLST_0pwebPageView_Table_columns)r)->_desc2 = ((_TPPLST_0pwebPageView_Table_columns)l)->_desc2;
((_TPPLST_0pwebPageView_Table_columns)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_Table_columns (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpwebPageView_Table_columns (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpwebPageView_Table_columns root;
	_TPPLST_0pwebPageView_Table_columns list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pwebPageView_Table_columns)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_webPageView_Table_columns)
		t = _ElempwebPageView_Table_columns (_coordref, t);
	remain = (_TPPLST_0pwebPageView_Table_columns)(((_TPPLST_0pwebPageView_Table_columns) t)->_desc2);
	((_TPPLST_0pwebPageView_Table_columns) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pwebPageView_Table_columns;
#else
list = (_TPPLST_0pwebPageView_Table_columns) TreeNodeAlloc (sizeof (struct _TPLST_0pwebPageView_Table_columns));
#endif
list->_prod = RULELST_0pwebPageView_Table_columns;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pwebPageView_Table_columns)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_webPageView_TableColumnpwebPageView_Table_columns)
	{_TPPLST_webPageView_TableColumnpwebPageView_Table_columns _currn = _currn;
	 _TERMACT_LST_webPageView_TableColumnpwebPageView_Table_columns;
	} this code has been moved to _ElempwebPageView_Table_columns*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpwebPageView_Table_columns;
#else
root = (_TPPpwebPageView_Table_columns) TreeNodeAlloc (sizeof (struct _TPpwebPageView_Table_columns));
#endif
root->_prod = RULEpwebPageView_Table_columns;
root->_desc1 = (_TSPLST_webPageView_Table_columns)list;
{ _TPPpwebPageView_Table_columns _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pwebPageView_Table_columns;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_TableRow (POSITION *_coordref, int _TERM1, NODEPTR _desc1)
#else
NODEPTR MkpwebPageView_TableRow (_coordref, _TERM1,_desc1)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
#endif
{	_TPPpwebPageView_TableRow _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_TableRow;
#else
_currn = (_TPPpwebPageView_TableRow) TreeNodeAlloc (sizeof (struct _TPpwebPageView_TableRow));
#endif
_currn->_prod = RULEpwebPageView_TableRow;
_currn->_desc1 = (_TSPwebPageView_TableRow_cells) MkwebPageView_TableRow_cells (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_TableRow: root of subtree no. 1 can not be made a webPageView_TableRow_cells node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_TableRow;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_TableRow */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_TableCellpwebPageView_TableRow_cells (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_TableCellpwebPageView_TableRow_cells (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_TableCellpwebPageView_TableRow_cells _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_TableCellpwebPageView_TableRow_cells;
#else
	_new = (_TPPLST_webPageView_TableCellpwebPageView_TableRow_cells) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_TableCellpwebPageView_TableRow_cells));
#endif
	_new->_prod = RULELST_webPageView_TableCellpwebPageView_TableRow_cells;
	_new->_desc1 = (_TSPwebPageView_TableCell)elem;
	_new->_desc2 = (_TSPLST_webPageView_TableRow_cells)_new;
	{_TPPLST_webPageView_TableCellpwebPageView_TableRow_cells _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_TableCellpwebPageView_TableRow_cells;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempwebPageView_TableRow_cells (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempwebPageView_TableRow_cells (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBwebPageView_TableCell)) 
		return _MkwebPageView_TableCellpwebPageView_TableRow_cells (_coordref, MkwebPageView_TableCell (_coordref, t));
	message (DEADLY, "a pwebPageView_TableRow_cells list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pwebPageView_TableRow_cells (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pwebPageView_TableRow_cells (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_webPageView_TableRow_cells)
	l =_ElempwebPageView_TableRow_cells (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_webPageView_TableRow_cells)
	r = _ElempwebPageView_TableRow_cells (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pwebPageView_TableRow_cells)r)->_desc2);
((_TPPLST_0pwebPageView_TableRow_cells)r)->_desc2 = ((_TPPLST_0pwebPageView_TableRow_cells)l)->_desc2;
((_TPPLST_0pwebPageView_TableRow_cells)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_TableRow_cells (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpwebPageView_TableRow_cells (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpwebPageView_TableRow_cells root;
	_TPPLST_0pwebPageView_TableRow_cells list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pwebPageView_TableRow_cells)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_webPageView_TableRow_cells)
		t = _ElempwebPageView_TableRow_cells (_coordref, t);
	remain = (_TPPLST_0pwebPageView_TableRow_cells)(((_TPPLST_0pwebPageView_TableRow_cells) t)->_desc2);
	((_TPPLST_0pwebPageView_TableRow_cells) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pwebPageView_TableRow_cells;
#else
list = (_TPPLST_0pwebPageView_TableRow_cells) TreeNodeAlloc (sizeof (struct _TPLST_0pwebPageView_TableRow_cells));
#endif
list->_prod = RULELST_0pwebPageView_TableRow_cells;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pwebPageView_TableRow_cells)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_webPageView_TableCellpwebPageView_TableRow_cells)
	{_TPPLST_webPageView_TableCellpwebPageView_TableRow_cells _currn = _currn;
	 _TERMACT_LST_webPageView_TableCellpwebPageView_TableRow_cells;
	} this code has been moved to _ElempwebPageView_TableRow_cells*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpwebPageView_TableRow_cells;
#else
root = (_TPPpwebPageView_TableRow_cells) TreeNodeAlloc (sizeof (struct _TPpwebPageView_TableRow_cells));
#endif
root->_prod = RULEpwebPageView_TableRow_cells;
root->_desc1 = (_TSPLST_webPageView_TableRow_cells)list;
{ _TPPpwebPageView_TableRow_cells _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pwebPageView_TableRow_cells;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_TableColumn (POSITION *_coordref, int _TERM1)
#else
NODEPTR MkpwebPageView_TableColumn (_coordref, _TERM1)
	POSITION *_coordref;
	int _TERM1;
#endif
{	_TPPpwebPageView_TableColumn _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_TableColumn;
#else
_currn = (_TPPpwebPageView_TableColumn) TreeNodeAlloc (sizeof (struct _TPpwebPageView_TableColumn));
#endif
_currn->_prod = RULEpwebPageView_TableColumn;
_SETCOORD(_currn)
_TERMACT_pwebPageView_TableColumn;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_TableColumn */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_TableCell (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpwebPageView_TableCell (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpwebPageView_TableCell _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_TableCell;
#else
_currn = (_TPPpwebPageView_TableCell) TreeNodeAlloc (sizeof (struct _TPpwebPageView_TableCell));
#endif
_currn->_prod = RULEpwebPageView_TableCell;
_currn->_desc1 = (_TSPwebPageView_TableCell_columnRef) MkwebPageView_TableCell_columnRef (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_TableCell: root of subtree no. 1 can not be made a webPageView_TableCell_columnRef node ", 0, _coordref);
_currn->_desc2 = (_TSPwebPageView_TableCell_content) MkwebPageView_TableCell_content (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_TableCell: root of subtree no. 2 can not be made a webPageView_TableCell_content node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_TableCell;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_TableCell */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_TableCell_columnRef (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_TableCell_columnRef (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_TableCell_columnRef _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_TableCell_columnRef;
#else
_currn = (_TPPpwebPageView_TableCell_columnRef) TreeNodeAlloc (sizeof (struct _TPpwebPageView_TableCell_columnRef));
#endif
_currn->_prod = RULEpwebPageView_TableCell_columnRef;
_SETCOORD(_currn)
_TERMACT_pwebPageView_TableCell_columnRef;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_TableCell_columnRef */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_WebPage (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5, NODEPTR _desc6)
#else
NODEPTR MkpwebPageView_WebPage (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5,_desc6)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
	NODEPTR _desc6;
#endif
{	_TPPpwebPageView_WebPage _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_WebPage;
#else
_currn = (_TPPpwebPageView_WebPage) TreeNodeAlloc (sizeof (struct _TPpwebPageView_WebPage));
#endif
_currn->_prod = RULEpwebPageView_WebPage;
_currn->_desc1 = (_TSPwebPageView_WebPage_name) MkwebPageView_WebPage_name (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_WebPage: root of subtree no. 1 can not be made a webPageView_WebPage_name node ", 0, _coordref);
_currn->_desc2 = (_TSPwebPageView_WebPage_formalParameter) MkwebPageView_WebPage_formalParameter (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_WebPage: root of subtree no. 2 can not be made a webPageView_WebPage_formalParameter node ", 0, _coordref);
_currn->_desc3 = (_TSPwebPageView_WebPage_title) MkwebPageView_WebPage_title (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_WebPage: root of subtree no. 3 can not be made a webPageView_WebPage_title node ", 0, _coordref);
_currn->_desc4 = (_TSPwebPageView_WebPage_frame) MkwebPageView_WebPage_frame (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_WebPage: root of subtree no. 4 can not be made a webPageView_WebPage_frame node ", 0, _coordref);
_currn->_desc5 = (_TSPwebPageView_WebPage_styleSheetBlock) MkwebPageView_WebPage_styleSheetBlock (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_WebPage: root of subtree no. 5 can not be made a webPageView_WebPage_styleSheetBlock node ", 0, _coordref);
_currn->_desc6 = (_TSPwebPageView_WebPage_position) MkwebPageView_WebPage_position (_coordref, _desc6);	
if (((NODEPTR)_currn->_desc6) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_WebPage: root of subtree no. 6 can not be made a webPageView_WebPage_position node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_WebPage;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_WebPage */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_WebPage_name (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_WebPage_name (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_WebPage_name _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_WebPage_name;
#else
_currn = (_TPPpwebPageView_WebPage_name) TreeNodeAlloc (sizeof (struct _TPpwebPageView_WebPage_name));
#endif
_currn->_prod = RULEpwebPageView_WebPage_name;
_SETCOORD(_currn)
_TERMACT_pwebPageView_WebPage_name;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_WebPage_name */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_WebNodeFormalParampwebPageView_WebPage_formalParameter (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_WebNodeFormalParampwebPageView_WebPage_formalParameter (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_WebNodeFormalParampwebPageView_WebPage_formalParameter _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_WebNodeFormalParampwebPageView_WebPage_formalParameter;
#else
	_new = (_TPPLST_webPageView_WebNodeFormalParampwebPageView_WebPage_formalParameter) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_WebNodeFormalParampwebPageView_WebPage_formalParameter));
#endif
	_new->_prod = RULELST_webPageView_WebNodeFormalParampwebPageView_WebPage_formalParameter;
	_new->_desc1 = (_TSPwebPageView_WebNodeFormalParam)elem;
	_new->_desc2 = (_TSPLST_webPageView_WebPage_formalParameter)_new;
	{_TPPLST_webPageView_WebNodeFormalParampwebPageView_WebPage_formalParameter _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_WebNodeFormalParampwebPageView_WebPage_formalParameter;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempwebPageView_WebPage_formalParameter (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempwebPageView_WebPage_formalParameter (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBwebPageView_WebNodeFormalParam)) 
		return _MkwebPageView_WebNodeFormalParampwebPageView_WebPage_formalParameter (_coordref, MkwebPageView_WebNodeFormalParam (_coordref, t));
	message (DEADLY, "a pwebPageView_WebPage_formalParameter list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pwebPageView_WebPage_formalParameter (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pwebPageView_WebPage_formalParameter (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_webPageView_WebPage_formalParameter)
	l =_ElempwebPageView_WebPage_formalParameter (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_webPageView_WebPage_formalParameter)
	r = _ElempwebPageView_WebPage_formalParameter (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pwebPageView_WebPage_formalParameter)r)->_desc2);
((_TPPLST_0pwebPageView_WebPage_formalParameter)r)->_desc2 = ((_TPPLST_0pwebPageView_WebPage_formalParameter)l)->_desc2;
((_TPPLST_0pwebPageView_WebPage_formalParameter)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_WebPage_formalParameter (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpwebPageView_WebPage_formalParameter (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpwebPageView_WebPage_formalParameter root;
	_TPPLST_0pwebPageView_WebPage_formalParameter list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pwebPageView_WebPage_formalParameter)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_webPageView_WebPage_formalParameter)
		t = _ElempwebPageView_WebPage_formalParameter (_coordref, t);
	remain = (_TPPLST_0pwebPageView_WebPage_formalParameter)(((_TPPLST_0pwebPageView_WebPage_formalParameter) t)->_desc2);
	((_TPPLST_0pwebPageView_WebPage_formalParameter) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pwebPageView_WebPage_formalParameter;
#else
list = (_TPPLST_0pwebPageView_WebPage_formalParameter) TreeNodeAlloc (sizeof (struct _TPLST_0pwebPageView_WebPage_formalParameter));
#endif
list->_prod = RULELST_0pwebPageView_WebPage_formalParameter;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pwebPageView_WebPage_formalParameter)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_webPageView_WebNodeFormalParampwebPageView_WebPage_formalParameter)
	{_TPPLST_webPageView_WebNodeFormalParampwebPageView_WebPage_formalParameter _currn = _currn;
	 _TERMACT_LST_webPageView_WebNodeFormalParampwebPageView_WebPage_formalParameter;
	} this code has been moved to _ElempwebPageView_WebPage_formalParameter*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpwebPageView_WebPage_formalParameter;
#else
root = (_TPPpwebPageView_WebPage_formalParameter) TreeNodeAlloc (sizeof (struct _TPpwebPageView_WebPage_formalParameter));
#endif
root->_prod = RULEpwebPageView_WebPage_formalParameter;
root->_desc1 = (_TSPLST_webPageView_WebPage_formalParameter)list;
{ _TPPpwebPageView_WebPage_formalParameter _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pwebPageView_WebPage_formalParameter;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_WebPage_title (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_WebPage_title (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_WebPage_title _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_WebPage_title;
#else
_currn = (_TPPpwebPageView_WebPage_title) TreeNodeAlloc (sizeof (struct _TPpwebPageView_WebPage_title));
#endif
_currn->_prod = RULEpwebPageView_WebPage_title;
_SETCOORD(_currn)
_TERMACT_pwebPageView_WebPage_title;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_WebPage_title */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_WebPage_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_WebPage_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_WebPage_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_WebPage_styleSheetBlock;
#else
_currn = (_TPPpwebPageView_WebPage_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpwebPageView_WebPage_styleSheetBlock));
#endif
_currn->_prod = RULEpwebPageView_WebPage_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_pwebPageView_WebPage_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_WebPage_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_WebPage_position (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_WebPage_position (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_WebPage_position _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_WebPage_position;
#else
_currn = (_TPPpwebPageView_WebPage_position) TreeNodeAlloc (sizeof (struct _TPpwebPageView_WebPage_position));
#endif
_currn->_prod = RULEpwebPageView_WebPage_position;
_SETCOORD(_currn)
_TERMACT_pwebPageView_WebPage_position;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_WebPage_position */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_Frame (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpwebPageView_Frame (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpwebPageView_Frame _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_Frame;
#else
_currn = (_TPPpwebPageView_Frame) TreeNodeAlloc (sizeof (struct _TPpwebPageView_Frame));
#endif
_currn->_prod = RULEpwebPageView_Frame;
_currn->_desc1 = (_TSPwebPageView_Frame_components) MkwebPageView_Frame_components (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_Frame: root of subtree no. 1 can not be made a webPageView_Frame_components node ", 0, _coordref);
_currn->_desc2 = (_TSPwebPageView_Frame_styleSheetBlock) MkwebPageView_Frame_styleSheetBlock (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_Frame: root of subtree no. 2 can not be made a webPageView_Frame_styleSheetBlock node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_Frame;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_Frame */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_FormularpwebPageView_Frame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_FormularpwebPageView_Frame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_FormularpwebPageView_Frame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_FormularpwebPageView_Frame_components;
#else
	_new = (_TPPLST_webPageView_FormularpwebPageView_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_FormularpwebPageView_Frame_components));
#endif
	_new->_prod = RULELST_webPageView_FormularpwebPageView_Frame_components;
	_new->_desc1 = (_TSPwebPageView_Formular)elem;
	_new->_desc2 = (_TSPLST_webPageView_Frame_components)_new;
	{_TPPLST_webPageView_FormularpwebPageView_Frame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_FormularpwebPageView_Frame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_ParagraphpwebPageView_Frame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_ParagraphpwebPageView_Frame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_ParagraphpwebPageView_Frame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_ParagraphpwebPageView_Frame_components;
#else
	_new = (_TPPLST_webPageView_ParagraphpwebPageView_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_ParagraphpwebPageView_Frame_components));
#endif
	_new->_prod = RULELST_webPageView_ParagraphpwebPageView_Frame_components;
	_new->_desc1 = (_TSPwebPageView_Paragraph)elem;
	_new->_desc2 = (_TSPLST_webPageView_Frame_components)_new;
	{_TPPLST_webPageView_ParagraphpwebPageView_Frame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_ParagraphpwebPageView_Frame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_HeadingpwebPageView_Frame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_HeadingpwebPageView_Frame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_HeadingpwebPageView_Frame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_HeadingpwebPageView_Frame_components;
#else
	_new = (_TPPLST_webPageView_HeadingpwebPageView_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_HeadingpwebPageView_Frame_components));
#endif
	_new->_prod = RULELST_webPageView_HeadingpwebPageView_Frame_components;
	_new->_desc1 = (_TSPwebPageView_Heading)elem;
	_new->_desc2 = (_TSPLST_webPageView_Frame_components)_new;
	{_TPPLST_webPageView_HeadingpwebPageView_Frame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_HeadingpwebPageView_Frame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_TablepwebPageView_Frame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_TablepwebPageView_Frame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_TablepwebPageView_Frame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_TablepwebPageView_Frame_components;
#else
	_new = (_TPPLST_webPageView_TablepwebPageView_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_TablepwebPageView_Frame_components));
#endif
	_new->_prod = RULELST_webPageView_TablepwebPageView_Frame_components;
	_new->_desc1 = (_TSPwebPageView_Table)elem;
	_new->_desc2 = (_TSPLST_webPageView_Frame_components)_new;
	{_TPPLST_webPageView_TablepwebPageView_Frame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_TablepwebPageView_Frame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_FrameRefpwebPageView_Frame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_FrameRefpwebPageView_Frame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_FrameRefpwebPageView_Frame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_FrameRefpwebPageView_Frame_components;
#else
	_new = (_TPPLST_webPageView_FrameRefpwebPageView_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_FrameRefpwebPageView_Frame_components));
#endif
	_new->_prod = RULELST_webPageView_FrameRefpwebPageView_Frame_components;
	_new->_desc1 = (_TSPwebPageView_FrameRef)elem;
	_new->_desc2 = (_TSPLST_webPageView_Frame_components)_new;
	{_TPPLST_webPageView_FrameRefpwebPageView_Frame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_FrameRefpwebPageView_Frame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_ListOutputElementpwebPageView_Frame_components (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_ListOutputElementpwebPageView_Frame_components (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_ListOutputElementpwebPageView_Frame_components _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_ListOutputElementpwebPageView_Frame_components;
#else
	_new = (_TPPLST_webPageView_ListOutputElementpwebPageView_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_ListOutputElementpwebPageView_Frame_components));
#endif
	_new->_prod = RULELST_webPageView_ListOutputElementpwebPageView_Frame_components;
	_new->_desc1 = (_TSPwebPageView_ListOutputElement)elem;
	_new->_desc2 = (_TSPLST_webPageView_Frame_components)_new;
	{_TPPLST_webPageView_ListOutputElementpwebPageView_Frame_components _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_ListOutputElementpwebPageView_Frame_components;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempwebPageView_Frame_components (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempwebPageView_Frame_components (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBwebPageView_Formular)) 
		return _MkwebPageView_FormularpwebPageView_Frame_components (_coordref, MkwebPageView_Formular (_coordref, t));
	if (IsSymb (t, SYMBwebPageView_Paragraph)) 
		return _MkwebPageView_ParagraphpwebPageView_Frame_components (_coordref, MkwebPageView_Paragraph (_coordref, t));
	if (IsSymb (t, SYMBwebPageView_Heading)) 
		return _MkwebPageView_HeadingpwebPageView_Frame_components (_coordref, MkwebPageView_Heading (_coordref, t));
	if (IsSymb (t, SYMBwebPageView_Table)) 
		return _MkwebPageView_TablepwebPageView_Frame_components (_coordref, MkwebPageView_Table (_coordref, t));
	if (IsSymb (t, SYMBwebPageView_FrameRef)) 
		return _MkwebPageView_FrameRefpwebPageView_Frame_components (_coordref, MkwebPageView_FrameRef (_coordref, t));
	if (IsSymb (t, SYMBwebPageView_ListOutputElement)) 
		return _MkwebPageView_ListOutputElementpwebPageView_Frame_components (_coordref, MkwebPageView_ListOutputElement (_coordref, t));
	message (DEADLY, "a pwebPageView_Frame_components list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pwebPageView_Frame_components (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pwebPageView_Frame_components (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_webPageView_Frame_components)
	l =_ElempwebPageView_Frame_components (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_webPageView_Frame_components)
	r = _ElempwebPageView_Frame_components (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pwebPageView_Frame_components)r)->_desc2);
((_TPPLST_0pwebPageView_Frame_components)r)->_desc2 = ((_TPPLST_0pwebPageView_Frame_components)l)->_desc2;
((_TPPLST_0pwebPageView_Frame_components)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_Frame_components (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpwebPageView_Frame_components (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpwebPageView_Frame_components root;
	_TPPLST_0pwebPageView_Frame_components list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pwebPageView_Frame_components)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_webPageView_Frame_components)
		t = _ElempwebPageView_Frame_components (_coordref, t);
	remain = (_TPPLST_0pwebPageView_Frame_components)(((_TPPLST_0pwebPageView_Frame_components) t)->_desc2);
	((_TPPLST_0pwebPageView_Frame_components) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pwebPageView_Frame_components;
#else
list = (_TPPLST_0pwebPageView_Frame_components) TreeNodeAlloc (sizeof (struct _TPLST_0pwebPageView_Frame_components));
#endif
list->_prod = RULELST_0pwebPageView_Frame_components;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pwebPageView_Frame_components)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_webPageView_FormularpwebPageView_Frame_components)
	{_TPPLST_webPageView_FormularpwebPageView_Frame_components _currn = _currn;
	 _TERMACT_LST_webPageView_FormularpwebPageView_Frame_components;
	} this code has been moved to _ElempwebPageView_Frame_components*/
	/*if (_currn->_prod == RULELST_webPageView_ParagraphpwebPageView_Frame_components)
	{_TPPLST_webPageView_ParagraphpwebPageView_Frame_components _currn = _currn;
	 _TERMACT_LST_webPageView_ParagraphpwebPageView_Frame_components;
	} this code has been moved to _ElempwebPageView_Frame_components*/
	/*if (_currn->_prod == RULELST_webPageView_HeadingpwebPageView_Frame_components)
	{_TPPLST_webPageView_HeadingpwebPageView_Frame_components _currn = _currn;
	 _TERMACT_LST_webPageView_HeadingpwebPageView_Frame_components;
	} this code has been moved to _ElempwebPageView_Frame_components*/
	/*if (_currn->_prod == RULELST_webPageView_TablepwebPageView_Frame_components)
	{_TPPLST_webPageView_TablepwebPageView_Frame_components _currn = _currn;
	 _TERMACT_LST_webPageView_TablepwebPageView_Frame_components;
	} this code has been moved to _ElempwebPageView_Frame_components*/
	/*if (_currn->_prod == RULELST_webPageView_FrameRefpwebPageView_Frame_components)
	{_TPPLST_webPageView_FrameRefpwebPageView_Frame_components _currn = _currn;
	 _TERMACT_LST_webPageView_FrameRefpwebPageView_Frame_components;
	} this code has been moved to _ElempwebPageView_Frame_components*/
	/*if (_currn->_prod == RULELST_webPageView_ListOutputElementpwebPageView_Frame_components)
	{_TPPLST_webPageView_ListOutputElementpwebPageView_Frame_components _currn = _currn;
	 _TERMACT_LST_webPageView_ListOutputElementpwebPageView_Frame_components;
	} this code has been moved to _ElempwebPageView_Frame_components*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpwebPageView_Frame_components;
#else
root = (_TPPpwebPageView_Frame_components) TreeNodeAlloc (sizeof (struct _TPpwebPageView_Frame_components));
#endif
root->_prod = RULEpwebPageView_Frame_components;
root->_desc1 = (_TSPLST_webPageView_Frame_components)list;
{ _TPPpwebPageView_Frame_components _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pwebPageView_Frame_components;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_Frame_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_Frame_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_Frame_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_Frame_styleSheetBlock;
#else
_currn = (_TPPpwebPageView_Frame_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpwebPageView_Frame_styleSheetBlock));
#endif
_currn->_prod = RULEpwebPageView_Frame_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_pwebPageView_Frame_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_Frame_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FrameRef (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpwebPageView_FrameRef (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpwebPageView_FrameRef _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FrameRef;
#else
_currn = (_TPPpwebPageView_FrameRef) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FrameRef));
#endif
_currn->_prod = RULEpwebPageView_FrameRef;
_currn->_desc1 = (_TSPwebPageView_FrameRef_styleSheetBlock) MkwebPageView_FrameRef_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FrameRef: root of subtree no. 1 can not be made a webPageView_FrameRef_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSPwebPageView_FrameRef_frame) MkwebPageView_FrameRef_frame (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_FrameRef: root of subtree no. 2 can not be made a webPageView_FrameRef_frame node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_FrameRef;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_FrameRef */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FrameRef_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FrameRef_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FrameRef_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FrameRef_styleSheetBlock;
#else
_currn = (_TPPpwebPageView_FrameRef_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FrameRef_styleSheetBlock));
#endif
_currn->_prod = RULEpwebPageView_FrameRef_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FrameRef_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FrameRef_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_FrameRef_frame (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_FrameRef_frame (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_FrameRef_frame _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_FrameRef_frame;
#else
_currn = (_TPPpwebPageView_FrameRef_frame) TreeNodeAlloc (sizeof (struct _TPpwebPageView_FrameRef_frame));
#endif
_currn->_prod = RULEpwebPageView_FrameRef_frame;
_SETCOORD(_currn)
_TERMACT_pwebPageView_FrameRef_frame;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_FrameRef_frame */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_OutputElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2)
#else
NODEPTR MkpwebPageView_OutputElement (_coordref, _TERM1,_desc1,_desc2)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
#endif
{	_TPPpwebPageView_OutputElement _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_OutputElement;
#else
_currn = (_TPPpwebPageView_OutputElement) TreeNodeAlloc (sizeof (struct _TPpwebPageView_OutputElement));
#endif
_currn->_prod = RULEpwebPageView_OutputElement;
_currn->_desc1 = (_TSPwebPageView_OutputElement_link) MkwebPageView_OutputElement_link (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_OutputElement: root of subtree no. 1 can not be made a webPageView_OutputElement_link node ", 0, _coordref);
_currn->_desc2 = (_TSPwebPageView_OutputElement_content) MkwebPageView_OutputElement_content (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_OutputElement: root of subtree no. 2 can not be made a webPageView_OutputElement_content node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_OutputElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_OutputElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_OutputElement_link (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_OutputElement_link (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_OutputElement_link _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_OutputElement_link;
#else
_currn = (_TPPpwebPageView_OutputElement_link) TreeNodeAlloc (sizeof (struct _TPpwebPageView_OutputElement_link));
#endif
_currn->_prod = RULEpwebPageView_OutputElement_link;
_SETCOORD(_currn)
_TERMACT_pwebPageView_OutputElement_link;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_OutputElement_link */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_OutputElement_content (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_OutputElement_content (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_OutputElement_content _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_OutputElement_content;
#else
_currn = (_TPPpwebPageView_OutputElement_content) TreeNodeAlloc (sizeof (struct _TPpwebPageView_OutputElement_content));
#endif
_currn->_prod = RULEpwebPageView_OutputElement_content;
_SETCOORD(_currn)
_TERMACT_pwebPageView_OutputElement_content;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_OutputElement_content */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_ListOutputElement (POSITION *_coordref, int _TERM1, NODEPTR _desc1, NODEPTR _desc2, NODEPTR _desc3, NODEPTR _desc4, NODEPTR _desc5)
#else
NODEPTR MkpwebPageView_ListOutputElement (_coordref, _TERM1,_desc1,_desc2,_desc3,_desc4,_desc5)
	POSITION *_coordref;
	int _TERM1;
	NODEPTR _desc1;
	NODEPTR _desc2;
	NODEPTR _desc3;
	NODEPTR _desc4;
	NODEPTR _desc5;
#endif
{	_TPPpwebPageView_ListOutputElement _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_ListOutputElement;
#else
_currn = (_TPPpwebPageView_ListOutputElement) TreeNodeAlloc (sizeof (struct _TPpwebPageView_ListOutputElement));
#endif
_currn->_prod = RULEpwebPageView_ListOutputElement;
_currn->_desc1 = (_TSPwebPageView_ListOutputElement_styleSheetBlock) MkwebPageView_ListOutputElement_styleSheetBlock (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_ListOutputElement: root of subtree no. 1 can not be made a webPageView_ListOutputElement_styleSheetBlock node ", 0, _coordref);
_currn->_desc2 = (_TSPwebPageView_ListOutputElement_dataSourceActualParams) MkwebPageView_ListOutputElement_dataSourceActualParams (_coordref, _desc2);	
if (((NODEPTR)_currn->_desc2) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_ListOutputElement: root of subtree no. 2 can not be made a webPageView_ListOutputElement_dataSourceActualParams node ", 0, _coordref);
_currn->_desc3 = (_TSPwebPageView_ListOutputElement_dataSource) MkwebPageView_ListOutputElement_dataSource (_coordref, _desc3);	
if (((NODEPTR)_currn->_desc3) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_ListOutputElement: root of subtree no. 3 can not be made a webPageView_ListOutputElement_dataSource node ", 0, _coordref);
_currn->_desc4 = (_TSPwebPageView_ListOutputElement_maxNumberOfElements) MkwebPageView_ListOutputElement_maxNumberOfElements (_coordref, _desc4);	
if (((NODEPTR)_currn->_desc4) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_ListOutputElement: root of subtree no. 4 can not be made a webPageView_ListOutputElement_maxNumberOfElements node ", 0, _coordref);
_currn->_desc5 = (_TSPwebPageView_ListOutputElement_navigationElement) MkwebPageView_ListOutputElement_navigationElement (_coordref, _desc5);	
if (((NODEPTR)_currn->_desc5) == NULLNODEPTR)	
	message (DEADLY, "RULE pwebPageView_ListOutputElement: root of subtree no. 5 can not be made a webPageView_ListOutputElement_navigationElement node ", 0, _coordref);
_SETCOORD(_currn)
_TERMACT_pwebPageView_ListOutputElement;
#ifdef MONITOR
_dapto_term_int(((NODEPTR)_currn), "ID", _TERM1);
#endif

return ( (NODEPTR) _currn);
}/* MkpwebPageView_ListOutputElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_ListOutputElement_styleSheetBlock (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_ListOutputElement_styleSheetBlock (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_ListOutputElement_styleSheetBlock _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_ListOutputElement_styleSheetBlock;
#else
_currn = (_TPPpwebPageView_ListOutputElement_styleSheetBlock) TreeNodeAlloc (sizeof (struct _TPpwebPageView_ListOutputElement_styleSheetBlock));
#endif
_currn->_prod = RULEpwebPageView_ListOutputElement_styleSheetBlock;
_SETCOORD(_currn)
_TERMACT_pwebPageView_ListOutputElement_styleSheetBlock;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_ListOutputElement_styleSheetBlock */

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _MkwebPageView_DataSourceActualParampwebPageView_ListOutputElement_dataSourceActualParams (POSITION *_coordref, NODEPTR elem)
#else
static NODEPTR _MkwebPageView_DataSourceActualParampwebPageView_ListOutputElement_dataSourceActualParams (_coordref, elem)
	POSITION *_coordref; NODEPTR elem;
#endif
{
	_TPPLST_webPageView_DataSourceActualParampwebPageView_ListOutputElement_dataSourceActualParams _new;
#ifdef __cplusplus
	_new = new _TPLST_webPageView_DataSourceActualParampwebPageView_ListOutputElement_dataSourceActualParams;
#else
	_new = (_TPPLST_webPageView_DataSourceActualParampwebPageView_ListOutputElement_dataSourceActualParams) TreeNodeAlloc (sizeof (struct _TPLST_webPageView_DataSourceActualParampwebPageView_ListOutputElement_dataSourceActualParams));
#endif
	_new->_prod = RULELST_webPageView_DataSourceActualParampwebPageView_ListOutputElement_dataSourceActualParams;
	_new->_desc1 = (_TSPwebPageView_DataSourceActualParam)elem;
	_new->_desc2 = (_TSPLST_webPageView_ListOutputElement_dataSourceActualParams)_new;
	{_TPPLST_webPageView_DataSourceActualParampwebPageView_ListOutputElement_dataSourceActualParams _currn = _new;
	 _SETCOORD(_currn)
	 _TERMACT_LST_webPageView_DataSourceActualParampwebPageView_ListOutputElement_dataSourceActualParams;
	}
	return (NODEPTR)_new;
}

#if defined(__STDC__) || defined(__cplusplus)
static NODEPTR _ElempwebPageView_ListOutputElement_dataSourceActualParams (POSITION *_coordref, NODEPTR t)
#else
static NODEPTR _ElempwebPageView_ListOutputElement_dataSourceActualParams (_coordref, t)
	POSITION *_coordref; NODEPTR t;
#endif
/* coerces t to a list element and returns a single circular list */
{	NODEPTR elem;
	/* check first without then with coercion: */
	if (IsSymb (t, SYMBwebPageView_DataSourceActualParam)) 
		return _MkwebPageView_DataSourceActualParampwebPageView_ListOutputElement_dataSourceActualParams (_coordref, MkwebPageView_DataSourceActualParam (_coordref, t));
	message (DEADLY, "a pwebPageView_ListOutputElement_dataSourceActualParams list element node can not be generated ",
		0, _coordref);
	return t; /* never reached */
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mk2pwebPageView_ListOutputElement_dataSourceActualParams (POSITION *_coordref, NODEPTR l, NODEPTR r)
#else
NODEPTR Mk2pwebPageView_ListOutputElement_dataSourceActualParams (_coordref, l, r) POSITION *_coordref; NODEPTR l, r;
#endif
/* make a list of l and r, which may be null, elements, or lists */
{	NODEPTR last;
if (l == NULLNODEPTR) return r;
if (r == NULLNODEPTR) return l;
/* ensure l and r to be lists: */
if (LHSMAP[l->_prod] != SYMBLST_webPageView_ListOutputElement_dataSourceActualParams)
	l =_ElempwebPageView_ListOutputElement_dataSourceActualParams (_coordref, l);
if (LHSMAP[r->_prod] != SYMBLST_webPageView_ListOutputElement_dataSourceActualParams)
	r = _ElempwebPageView_ListOutputElement_dataSourceActualParams (_coordref, r);
/* concatenate two lists: */
last = ((NODEPTR)((_TPPLST_0pwebPageView_ListOutputElement_dataSourceActualParams)r)->_desc2);
((_TPPLST_0pwebPageView_ListOutputElement_dataSourceActualParams)r)->_desc2 = ((_TPPLST_0pwebPageView_ListOutputElement_dataSourceActualParams)l)->_desc2;
((_TPPLST_0pwebPageView_ListOutputElement_dataSourceActualParams)l)->_desc2 = last;
return l;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_ListOutputElement_dataSourceActualParams (POSITION *_coordref, NODEPTR t)
#else
NODEPTR MkpwebPageView_ListOutputElement_dataSourceActualParams (_coordref, t) POSITION *_coordref; NODEPTR t;
#endif
{	NODEPTR _currn; _TPPpwebPageView_ListOutputElement_dataSourceActualParams root;
	_TPPLST_0pwebPageView_ListOutputElement_dataSourceActualParams list, remain, p;
/* argument t may be empty, a list element, or a list: */
if (t == NULLNODEPTR)
	remain = (_TPPLST_0pwebPageView_ListOutputElement_dataSourceActualParams)NULLNODEPTR;
else {	if (LHSMAP[t->_prod] != SYMBLST_webPageView_ListOutputElement_dataSourceActualParams)
		t = _ElempwebPageView_ListOutputElement_dataSourceActualParams (_coordref, t);
	remain = (_TPPLST_0pwebPageView_ListOutputElement_dataSourceActualParams)(((_TPPLST_0pwebPageView_ListOutputElement_dataSourceActualParams) t)->_desc2);
	((_TPPLST_0pwebPageView_ListOutputElement_dataSourceActualParams) t)->_desc2 = NULLNODEPTR;
}
/* remain is a possibly empty list in reverse order */
/* finalize the list by an empty context: */
#ifdef __cplusplus
list = new _TPLST_0pwebPageView_ListOutputElement_dataSourceActualParams;
#else
list = (_TPPLST_0pwebPageView_ListOutputElement_dataSourceActualParams) TreeNodeAlloc (sizeof (struct _TPLST_0pwebPageView_ListOutputElement_dataSourceActualParams));
#endif
list->_prod = RULELST_0pwebPageView_ListOutputElement_dataSourceActualParams;
list->_desc1 = NULLNODEPTR;
list->_desc2 = NULLNODEPTR;
_SETCOORD(list)
_currn = (NODEPTR)list;
/* reverse the list and visit its nodes: */
while (remain) {
	p = (_TPPLST_0pwebPageView_ListOutputElement_dataSourceActualParams)(remain->_desc2);
	remain->_desc2 = (NODEPTR)list;
	list = remain;
	remain = p;
	_currn = (NODEPTR)list;
	/*if (_currn->_prod == RULELST_webPageView_DataSourceActualParampwebPageView_ListOutputElement_dataSourceActualParams)
	{_TPPLST_webPageView_DataSourceActualParampwebPageView_ListOutputElement_dataSourceActualParams _currn = _currn;
	 _TERMACT_LST_webPageView_DataSourceActualParampwebPageView_ListOutputElement_dataSourceActualParams;
	} this code has been moved to _ElempwebPageView_ListOutputElement_dataSourceActualParams*/
}
/* make root node and visit it: */
#ifdef __cplusplus
root = new _TPpwebPageView_ListOutputElement_dataSourceActualParams;
#else
root = (_TPPpwebPageView_ListOutputElement_dataSourceActualParams) TreeNodeAlloc (sizeof (struct _TPpwebPageView_ListOutputElement_dataSourceActualParams));
#endif
root->_prod = RULEpwebPageView_ListOutputElement_dataSourceActualParams;
root->_desc1 = (_TSPLST_webPageView_ListOutputElement_dataSourceActualParams)list;
{ _TPPpwebPageView_ListOutputElement_dataSourceActualParams _currn = root;
  _SETCOORD(_currn)
  _TERMACT_pwebPageView_ListOutputElement_dataSourceActualParams;
}
_currn = (NODEPTR)root;
return (NODEPTR)_currn;
}

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_ListOutputElement_dataSource (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_ListOutputElement_dataSource (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_ListOutputElement_dataSource _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_ListOutputElement_dataSource;
#else
_currn = (_TPPpwebPageView_ListOutputElement_dataSource) TreeNodeAlloc (sizeof (struct _TPpwebPageView_ListOutputElement_dataSource));
#endif
_currn->_prod = RULEpwebPageView_ListOutputElement_dataSource;
_SETCOORD(_currn)
_TERMACT_pwebPageView_ListOutputElement_dataSource;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_ListOutputElement_dataSource */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_ListOutputElement_maxNumberOfElements (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_ListOutputElement_maxNumberOfElements (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_ListOutputElement_maxNumberOfElements _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_ListOutputElement_maxNumberOfElements;
#else
_currn = (_TPPpwebPageView_ListOutputElement_maxNumberOfElements) TreeNodeAlloc (sizeof (struct _TPpwebPageView_ListOutputElement_maxNumberOfElements));
#endif
_currn->_prod = RULEpwebPageView_ListOutputElement_maxNumberOfElements;
_SETCOORD(_currn)
_TERMACT_pwebPageView_ListOutputElement_maxNumberOfElements;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_ListOutputElement_maxNumberOfElements */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR MkpwebPageView_ListOutputElement_navigationElement (POSITION *_coordref)
#else
NODEPTR MkpwebPageView_ListOutputElement_navigationElement (_coordref)
	POSITION *_coordref;
#endif
{	_TPPpwebPageView_ListOutputElement_navigationElement _currn;
#ifdef __cplusplus
_currn = new _TPpwebPageView_ListOutputElement_navigationElement;
#else
_currn = (_TPPpwebPageView_ListOutputElement_navigationElement) TreeNodeAlloc (sizeof (struct _TPpwebPageView_ListOutputElement_navigationElement));
#endif
_currn->_prod = RULEpwebPageView_ListOutputElement_navigationElement;
_SETCOORD(_currn)
_TERMACT_pwebPageView_ListOutputElement_navigationElement;
return ( (NODEPTR) _currn);
}/* MkpwebPageView_ListOutputElement_navigationElement */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_1 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_1 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_1 _currn;
#ifdef __cplusplus
_currn = new _TPrule_1;
#else
_currn = (_TPPrule_1) TreeNodeAlloc (sizeof (struct _TPrule_1));
#endif
_currn->_prod = RULErule_1;
_currn->_desc1 = (_TSPcodeGen_DBBlob) MkcodeGen_DBBlob (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_1: root of subtree no. 1 can not be made a codeGen_DBBlob node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_1;
return ( (NODEPTR) _currn);
}/* Mkrule_1 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_2 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_2 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_2 _currn;
#ifdef __cplusplus
_currn = new _TPrule_2;
#else
_currn = (_TPPrule_2) TreeNodeAlloc (sizeof (struct _TPrule_2));
#endif
_currn->_prod = RULErule_2;
_currn->_desc1 = (_TSPcodeGen_DBBoolean) MkcodeGen_DBBoolean (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_2: root of subtree no. 1 can not be made a codeGen_DBBoolean node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_2;
return ( (NODEPTR) _currn);
}/* Mkrule_2 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_3 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_3 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_3 _currn;
#ifdef __cplusplus
_currn = new _TPrule_3;
#else
_currn = (_TPPrule_3) TreeNodeAlloc (sizeof (struct _TPrule_3));
#endif
_currn->_prod = RULErule_3;
_currn->_desc1 = (_TSPcodeGen_DBTimeStamp) MkcodeGen_DBTimeStamp (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_3: root of subtree no. 1 can not be made a codeGen_DBTimeStamp node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_3;
return ( (NODEPTR) _currn);
}/* Mkrule_3 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_4 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_4 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_4 _currn;
#ifdef __cplusplus
_currn = new _TPrule_4;
#else
_currn = (_TPPrule_4) TreeNodeAlloc (sizeof (struct _TPrule_4));
#endif
_currn->_prod = RULErule_4;
_currn->_desc1 = (_TSPcodeGen_DBDate) MkcodeGen_DBDate (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_4: root of subtree no. 1 can not be made a codeGen_DBDate node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_4;
return ( (NODEPTR) _currn);
}/* Mkrule_4 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_5 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_5 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_5 _currn;
#ifdef __cplusplus
_currn = new _TPrule_5;
#else
_currn = (_TPPrule_5) TreeNodeAlloc (sizeof (struct _TPrule_5));
#endif
_currn->_prod = RULErule_5;
_currn->_desc1 = (_TSPcodeGen_DBBigInt) MkcodeGen_DBBigInt (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_5: root of subtree no. 1 can not be made a codeGen_DBBigInt node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_5;
return ( (NODEPTR) _currn);
}/* Mkrule_5 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_6 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_6 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_6 _currn;
#ifdef __cplusplus
_currn = new _TPrule_6;
#else
_currn = (_TPPrule_6) TreeNodeAlloc (sizeof (struct _TPrule_6));
#endif
_currn->_prod = RULErule_6;
_currn->_desc1 = (_TSPcodeGen_DBDouble) MkcodeGen_DBDouble (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_6: root of subtree no. 1 can not be made a codeGen_DBDouble node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_6;
return ( (NODEPTR) _currn);
}/* Mkrule_6 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_7 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_7 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_7 _currn;
#ifdef __cplusplus
_currn = new _TPrule_7;
#else
_currn = (_TPPrule_7) TreeNodeAlloc (sizeof (struct _TPrule_7));
#endif
_currn->_prod = RULErule_7;
_currn->_desc1 = (_TSPcodeGen_DBFloat) MkcodeGen_DBFloat (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_7: root of subtree no. 1 can not be made a codeGen_DBFloat node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_7;
return ( (NODEPTR) _currn);
}/* Mkrule_7 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_8 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_8 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_8 _currn;
#ifdef __cplusplus
_currn = new _TPrule_8;
#else
_currn = (_TPPrule_8) TreeNodeAlloc (sizeof (struct _TPrule_8));
#endif
_currn->_prod = RULErule_8;
_currn->_desc1 = (_TSPcodeGen_DBInteger) MkcodeGen_DBInteger (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_8: root of subtree no. 1 can not be made a codeGen_DBInteger node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_8;
return ( (NODEPTR) _currn);
}/* Mkrule_8 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_9 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_9 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_9 _currn;
#ifdef __cplusplus
_currn = new _TPrule_9;
#else
_currn = (_TPPrule_9) TreeNodeAlloc (sizeof (struct _TPrule_9));
#endif
_currn->_prod = RULErule_9;
_currn->_desc1 = (_TSPcodeGen_DBString) MkcodeGen_DBString (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_9: root of subtree no. 1 can not be made a codeGen_DBString node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_9;
return ( (NODEPTR) _currn);
}/* Mkrule_9 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_10 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_10 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_10 _currn;
#ifdef __cplusplus
_currn = new _TPrule_10;
#else
_currn = (_TPPrule_10) TreeNodeAlloc (sizeof (struct _TPrule_10));
#endif
_currn->_prod = RULErule_10;
_currn->_desc1 = (_TSPcodeGen_ComputationExpression) MkcodeGen_ComputationExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_10: root of subtree no. 1 can not be made a codeGen_ComputationExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_10;
return ( (NODEPTR) _currn);
}/* Mkrule_10 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_11 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_11 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_11 _currn;
#ifdef __cplusplus
_currn = new _TPrule_11;
#else
_currn = (_TPPrule_11) TreeNodeAlloc (sizeof (struct _TPrule_11));
#endif
_currn->_prod = RULErule_11;
_currn->_desc1 = (_TSPcodeGen_ParameterValueExpression) MkcodeGen_ParameterValueExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_11: root of subtree no. 1 can not be made a codeGen_ParameterValueExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_11;
return ( (NODEPTR) _currn);
}/* Mkrule_11 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_12 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_12 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_12 _currn;
#ifdef __cplusplus
_currn = new _TPrule_12;
#else
_currn = (_TPPrule_12) TreeNodeAlloc (sizeof (struct _TPrule_12));
#endif
_currn->_prod = RULErule_12;
_currn->_desc1 = (_TSPcodeGen_ColumnValueExpression) MkcodeGen_ColumnValueExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_12: root of subtree no. 1 can not be made a codeGen_ColumnValueExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_12;
return ( (NODEPTR) _currn);
}/* Mkrule_12 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_13 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_13 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_13 _currn;
#ifdef __cplusplus
_currn = new _TPrule_13;
#else
_currn = (_TPPrule_13) TreeNodeAlloc (sizeof (struct _TPrule_13));
#endif
_currn->_prod = RULErule_13;
_currn->_desc1 = (_TSPcodeGen_ConstantExpression) MkcodeGen_ConstantExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_13: root of subtree no. 1 can not be made a codeGen_ConstantExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_13;
return ( (NODEPTR) _currn);
}/* Mkrule_13 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_14 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_14 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_14 _currn;
#ifdef __cplusplus
_currn = new _TPrule_14;
#else
_currn = (_TPPrule_14) TreeNodeAlloc (sizeof (struct _TPrule_14));
#endif
_currn->_prod = RULErule_14;
_currn->_desc1 = (_TSPcodeGen_UnaryExpression) MkcodeGen_UnaryExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_14: root of subtree no. 1 can not be made a codeGen_UnaryExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_14;
return ( (NODEPTR) _currn);
}/* Mkrule_14 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_15 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_15 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_15 _currn;
#ifdef __cplusplus
_currn = new _TPrule_15;
#else
_currn = (_TPPrule_15) TreeNodeAlloc (sizeof (struct _TPrule_15));
#endif
_currn->_prod = RULErule_15;
_currn->_desc1 = (_TSPcodeGen_BinaryExpression) MkcodeGen_BinaryExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_15: root of subtree no. 1 can not be made a codeGen_BinaryExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_15;
return ( (NODEPTR) _currn);
}/* Mkrule_15 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_16 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_16 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_16 _currn;
#ifdef __cplusplus
_currn = new _TPrule_16;
#else
_currn = (_TPPrule_16) TreeNodeAlloc (sizeof (struct _TPrule_16));
#endif
_currn->_prod = RULErule_16;
_currn->_desc1 = (_TSPcodeGen_ComputationExpression) MkcodeGen_ComputationExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_16: root of subtree no. 1 can not be made a codeGen_ComputationExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_16;
return ( (NODEPTR) _currn);
}/* Mkrule_16 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_17 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_17 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_17 _currn;
#ifdef __cplusplus
_currn = new _TPrule_17;
#else
_currn = (_TPPrule_17) TreeNodeAlloc (sizeof (struct _TPrule_17));
#endif
_currn->_prod = RULErule_17;
_currn->_desc1 = (_TSPcodeGen_ParameterValueExpression) MkcodeGen_ParameterValueExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_17: root of subtree no. 1 can not be made a codeGen_ParameterValueExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_17;
return ( (NODEPTR) _currn);
}/* Mkrule_17 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_18 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_18 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_18 _currn;
#ifdef __cplusplus
_currn = new _TPrule_18;
#else
_currn = (_TPPrule_18) TreeNodeAlloc (sizeof (struct _TPrule_18));
#endif
_currn->_prod = RULErule_18;
_currn->_desc1 = (_TSPcodeGen_ColumnValueExpression) MkcodeGen_ColumnValueExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_18: root of subtree no. 1 can not be made a codeGen_ColumnValueExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_18;
return ( (NODEPTR) _currn);
}/* Mkrule_18 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_19 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_19 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_19 _currn;
#ifdef __cplusplus
_currn = new _TPrule_19;
#else
_currn = (_TPPrule_19) TreeNodeAlloc (sizeof (struct _TPrule_19));
#endif
_currn->_prod = RULErule_19;
_currn->_desc1 = (_TSPcodeGen_ConstantExpression) MkcodeGen_ConstantExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_19: root of subtree no. 1 can not be made a codeGen_ConstantExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_19;
return ( (NODEPTR) _currn);
}/* Mkrule_19 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_20 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_20 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_20 _currn;
#ifdef __cplusplus
_currn = new _TPrule_20;
#else
_currn = (_TPPrule_20) TreeNodeAlloc (sizeof (struct _TPrule_20));
#endif
_currn->_prod = RULErule_20;
_currn->_desc1 = (_TSPcodeGen_UnaryExpression) MkcodeGen_UnaryExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_20: root of subtree no. 1 can not be made a codeGen_UnaryExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_20;
return ( (NODEPTR) _currn);
}/* Mkrule_20 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_21 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_21 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_21 _currn;
#ifdef __cplusplus
_currn = new _TPrule_21;
#else
_currn = (_TPPrule_21) TreeNodeAlloc (sizeof (struct _TPrule_21));
#endif
_currn->_prod = RULErule_21;
_currn->_desc1 = (_TSPcodeGen_BinaryExpression) MkcodeGen_BinaryExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_21: root of subtree no. 1 can not be made a codeGen_BinaryExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_21;
return ( (NODEPTR) _currn);
}/* Mkrule_21 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_22 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_22 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_22 _currn;
#ifdef __cplusplus
_currn = new _TPrule_22;
#else
_currn = (_TPPrule_22) TreeNodeAlloc (sizeof (struct _TPrule_22));
#endif
_currn->_prod = RULErule_22;
_currn->_desc1 = (_TSPcodeGen_ComputationExpression) MkcodeGen_ComputationExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_22: root of subtree no. 1 can not be made a codeGen_ComputationExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_22;
return ( (NODEPTR) _currn);
}/* Mkrule_22 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_23 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_23 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_23 _currn;
#ifdef __cplusplus
_currn = new _TPrule_23;
#else
_currn = (_TPPrule_23) TreeNodeAlloc (sizeof (struct _TPrule_23));
#endif
_currn->_prod = RULErule_23;
_currn->_desc1 = (_TSPcodeGen_ParameterValueExpression) MkcodeGen_ParameterValueExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_23: root of subtree no. 1 can not be made a codeGen_ParameterValueExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_23;
return ( (NODEPTR) _currn);
}/* Mkrule_23 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_24 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_24 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_24 _currn;
#ifdef __cplusplus
_currn = new _TPrule_24;
#else
_currn = (_TPPrule_24) TreeNodeAlloc (sizeof (struct _TPrule_24));
#endif
_currn->_prod = RULErule_24;
_currn->_desc1 = (_TSPcodeGen_ColumnValueExpression) MkcodeGen_ColumnValueExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_24: root of subtree no. 1 can not be made a codeGen_ColumnValueExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_24;
return ( (NODEPTR) _currn);
}/* Mkrule_24 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_25 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_25 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_25 _currn;
#ifdef __cplusplus
_currn = new _TPrule_25;
#else
_currn = (_TPPrule_25) TreeNodeAlloc (sizeof (struct _TPrule_25));
#endif
_currn->_prod = RULErule_25;
_currn->_desc1 = (_TSPcodeGen_ConstantExpression) MkcodeGen_ConstantExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_25: root of subtree no. 1 can not be made a codeGen_ConstantExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_25;
return ( (NODEPTR) _currn);
}/* Mkrule_25 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_26 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_26 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_26 _currn;
#ifdef __cplusplus
_currn = new _TPrule_26;
#else
_currn = (_TPPrule_26) TreeNodeAlloc (sizeof (struct _TPrule_26));
#endif
_currn->_prod = RULErule_26;
_currn->_desc1 = (_TSPcodeGen_UnaryExpression) MkcodeGen_UnaryExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_26: root of subtree no. 1 can not be made a codeGen_UnaryExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_26;
return ( (NODEPTR) _currn);
}/* Mkrule_26 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_27 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_27 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_27 _currn;
#ifdef __cplusplus
_currn = new _TPrule_27;
#else
_currn = (_TPPrule_27) TreeNodeAlloc (sizeof (struct _TPrule_27));
#endif
_currn->_prod = RULErule_27;
_currn->_desc1 = (_TSPcodeGen_BinaryExpression) MkcodeGen_BinaryExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_27: root of subtree no. 1 can not be made a codeGen_BinaryExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_27;
return ( (NODEPTR) _currn);
}/* Mkrule_27 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_28 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_28 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_28 _currn;
#ifdef __cplusplus
_currn = new _TPrule_28;
#else
_currn = (_TPPrule_28) TreeNodeAlloc (sizeof (struct _TPrule_28));
#endif
_currn->_prod = RULErule_28;
_currn->_desc1 = (_TSPcodeGen_ComputationExpression) MkcodeGen_ComputationExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_28: root of subtree no. 1 can not be made a codeGen_ComputationExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_28;
return ( (NODEPTR) _currn);
}/* Mkrule_28 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_29 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_29 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_29 _currn;
#ifdef __cplusplus
_currn = new _TPrule_29;
#else
_currn = (_TPPrule_29) TreeNodeAlloc (sizeof (struct _TPrule_29));
#endif
_currn->_prod = RULErule_29;
_currn->_desc1 = (_TSPcodeGen_ParameterValueExpression) MkcodeGen_ParameterValueExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_29: root of subtree no. 1 can not be made a codeGen_ParameterValueExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_29;
return ( (NODEPTR) _currn);
}/* Mkrule_29 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_30 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_30 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_30 _currn;
#ifdef __cplusplus
_currn = new _TPrule_30;
#else
_currn = (_TPPrule_30) TreeNodeAlloc (sizeof (struct _TPrule_30));
#endif
_currn->_prod = RULErule_30;
_currn->_desc1 = (_TSPcodeGen_ColumnValueExpression) MkcodeGen_ColumnValueExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_30: root of subtree no. 1 can not be made a codeGen_ColumnValueExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_30;
return ( (NODEPTR) _currn);
}/* Mkrule_30 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_31 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_31 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_31 _currn;
#ifdef __cplusplus
_currn = new _TPrule_31;
#else
_currn = (_TPPrule_31) TreeNodeAlloc (sizeof (struct _TPrule_31));
#endif
_currn->_prod = RULErule_31;
_currn->_desc1 = (_TSPcodeGen_ConstantExpression) MkcodeGen_ConstantExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_31: root of subtree no. 1 can not be made a codeGen_ConstantExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_31;
return ( (NODEPTR) _currn);
}/* Mkrule_31 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_32 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_32 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_32 _currn;
#ifdef __cplusplus
_currn = new _TPrule_32;
#else
_currn = (_TPPrule_32) TreeNodeAlloc (sizeof (struct _TPrule_32));
#endif
_currn->_prod = RULErule_32;
_currn->_desc1 = (_TSPcodeGen_UnaryExpression) MkcodeGen_UnaryExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_32: root of subtree no. 1 can not be made a codeGen_UnaryExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_32;
return ( (NODEPTR) _currn);
}/* Mkrule_32 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_33 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_33 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_33 _currn;
#ifdef __cplusplus
_currn = new _TPrule_33;
#else
_currn = (_TPPrule_33) TreeNodeAlloc (sizeof (struct _TPrule_33));
#endif
_currn->_prod = RULErule_33;
_currn->_desc1 = (_TSPcodeGen_BinaryExpression) MkcodeGen_BinaryExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_33: root of subtree no. 1 can not be made a codeGen_BinaryExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_33;
return ( (NODEPTR) _currn);
}/* Mkrule_33 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_34 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_34 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_34 _currn;
#ifdef __cplusplus
_currn = new _TPrule_34;
#else
_currn = (_TPPrule_34) TreeNodeAlloc (sizeof (struct _TPrule_34));
#endif
_currn->_prod = RULErule_34;
_currn->_desc1 = (_TSPcodeGen_ComputationExpression) MkcodeGen_ComputationExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_34: root of subtree no. 1 can not be made a codeGen_ComputationExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_34;
return ( (NODEPTR) _currn);
}/* Mkrule_34 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_35 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_35 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_35 _currn;
#ifdef __cplusplus
_currn = new _TPrule_35;
#else
_currn = (_TPPrule_35) TreeNodeAlloc (sizeof (struct _TPrule_35));
#endif
_currn->_prod = RULErule_35;
_currn->_desc1 = (_TSPcodeGen_ParameterValueExpression) MkcodeGen_ParameterValueExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_35: root of subtree no. 1 can not be made a codeGen_ParameterValueExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_35;
return ( (NODEPTR) _currn);
}/* Mkrule_35 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_36 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_36 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_36 _currn;
#ifdef __cplusplus
_currn = new _TPrule_36;
#else
_currn = (_TPPrule_36) TreeNodeAlloc (sizeof (struct _TPrule_36));
#endif
_currn->_prod = RULErule_36;
_currn->_desc1 = (_TSPcodeGen_ColumnValueExpression) MkcodeGen_ColumnValueExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_36: root of subtree no. 1 can not be made a codeGen_ColumnValueExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_36;
return ( (NODEPTR) _currn);
}/* Mkrule_36 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_37 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_37 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_37 _currn;
#ifdef __cplusplus
_currn = new _TPrule_37;
#else
_currn = (_TPPrule_37) TreeNodeAlloc (sizeof (struct _TPrule_37));
#endif
_currn->_prod = RULErule_37;
_currn->_desc1 = (_TSPcodeGen_ConstantExpression) MkcodeGen_ConstantExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_37: root of subtree no. 1 can not be made a codeGen_ConstantExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_37;
return ( (NODEPTR) _currn);
}/* Mkrule_37 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_38 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_38 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_38 _currn;
#ifdef __cplusplus
_currn = new _TPrule_38;
#else
_currn = (_TPPrule_38) TreeNodeAlloc (sizeof (struct _TPrule_38));
#endif
_currn->_prod = RULErule_38;
_currn->_desc1 = (_TSPcodeGen_UnaryExpression) MkcodeGen_UnaryExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_38: root of subtree no. 1 can not be made a codeGen_UnaryExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_38;
return ( (NODEPTR) _currn);
}/* Mkrule_38 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_39 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_39 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_39 _currn;
#ifdef __cplusplus
_currn = new _TPrule_39;
#else
_currn = (_TPPrule_39) TreeNodeAlloc (sizeof (struct _TPrule_39));
#endif
_currn->_prod = RULErule_39;
_currn->_desc1 = (_TSPcodeGen_BinaryExpression) MkcodeGen_BinaryExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_39: root of subtree no. 1 can not be made a codeGen_BinaryExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_39;
return ( (NODEPTR) _currn);
}/* Mkrule_39 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_40 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_40 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_40 _currn;
#ifdef __cplusplus
_currn = new _TPrule_40;
#else
_currn = (_TPPrule_40) TreeNodeAlloc (sizeof (struct _TPrule_40));
#endif
_currn->_prod = RULErule_40;
_currn->_desc1 = (_TSPcodeGen_ComputationExpression) MkcodeGen_ComputationExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_40: root of subtree no. 1 can not be made a codeGen_ComputationExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_40;
return ( (NODEPTR) _currn);
}/* Mkrule_40 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_41 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_41 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_41 _currn;
#ifdef __cplusplus
_currn = new _TPrule_41;
#else
_currn = (_TPPrule_41) TreeNodeAlloc (sizeof (struct _TPrule_41));
#endif
_currn->_prod = RULErule_41;
_currn->_desc1 = (_TSPcodeGen_ParameterValueExpression) MkcodeGen_ParameterValueExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_41: root of subtree no. 1 can not be made a codeGen_ParameterValueExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_41;
return ( (NODEPTR) _currn);
}/* Mkrule_41 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_42 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_42 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_42 _currn;
#ifdef __cplusplus
_currn = new _TPrule_42;
#else
_currn = (_TPPrule_42) TreeNodeAlloc (sizeof (struct _TPrule_42));
#endif
_currn->_prod = RULErule_42;
_currn->_desc1 = (_TSPcodeGen_ColumnValueExpression) MkcodeGen_ColumnValueExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_42: root of subtree no. 1 can not be made a codeGen_ColumnValueExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_42;
return ( (NODEPTR) _currn);
}/* Mkrule_42 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_43 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_43 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_43 _currn;
#ifdef __cplusplus
_currn = new _TPrule_43;
#else
_currn = (_TPPrule_43) TreeNodeAlloc (sizeof (struct _TPrule_43));
#endif
_currn->_prod = RULErule_43;
_currn->_desc1 = (_TSPcodeGen_ConstantExpression) MkcodeGen_ConstantExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_43: root of subtree no. 1 can not be made a codeGen_ConstantExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_43;
return ( (NODEPTR) _currn);
}/* Mkrule_43 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_44 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_44 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_44 _currn;
#ifdef __cplusplus
_currn = new _TPrule_44;
#else
_currn = (_TPPrule_44) TreeNodeAlloc (sizeof (struct _TPrule_44));
#endif
_currn->_prod = RULErule_44;
_currn->_desc1 = (_TSPcodeGen_UnaryExpression) MkcodeGen_UnaryExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_44: root of subtree no. 1 can not be made a codeGen_UnaryExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_44;
return ( (NODEPTR) _currn);
}/* Mkrule_44 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_45 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_45 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_45 _currn;
#ifdef __cplusplus
_currn = new _TPrule_45;
#else
_currn = (_TPPrule_45) TreeNodeAlloc (sizeof (struct _TPrule_45));
#endif
_currn->_prod = RULErule_45;
_currn->_desc1 = (_TSPcodeGen_BinaryExpression) MkcodeGen_BinaryExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_45: root of subtree no. 1 can not be made a codeGen_BinaryExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_45;
return ( (NODEPTR) _currn);
}/* Mkrule_45 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_46 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_46 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_46 _currn;
#ifdef __cplusplus
_currn = new _TPrule_46;
#else
_currn = (_TPPrule_46) TreeNodeAlloc (sizeof (struct _TPrule_46));
#endif
_currn->_prod = RULErule_46;
_currn->_desc1 = (_TSPcodeGen_DBBlob) MkcodeGen_DBBlob (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_46: root of subtree no. 1 can not be made a codeGen_DBBlob node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_46;
return ( (NODEPTR) _currn);
}/* Mkrule_46 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_47 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_47 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_47 _currn;
#ifdef __cplusplus
_currn = new _TPrule_47;
#else
_currn = (_TPPrule_47) TreeNodeAlloc (sizeof (struct _TPrule_47));
#endif
_currn->_prod = RULErule_47;
_currn->_desc1 = (_TSPcodeGen_DBBoolean) MkcodeGen_DBBoolean (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_47: root of subtree no. 1 can not be made a codeGen_DBBoolean node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_47;
return ( (NODEPTR) _currn);
}/* Mkrule_47 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_48 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_48 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_48 _currn;
#ifdef __cplusplus
_currn = new _TPrule_48;
#else
_currn = (_TPPrule_48) TreeNodeAlloc (sizeof (struct _TPrule_48));
#endif
_currn->_prod = RULErule_48;
_currn->_desc1 = (_TSPcodeGen_DBTimeStamp) MkcodeGen_DBTimeStamp (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_48: root of subtree no. 1 can not be made a codeGen_DBTimeStamp node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_48;
return ( (NODEPTR) _currn);
}/* Mkrule_48 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_49 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_49 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_49 _currn;
#ifdef __cplusplus
_currn = new _TPrule_49;
#else
_currn = (_TPPrule_49) TreeNodeAlloc (sizeof (struct _TPrule_49));
#endif
_currn->_prod = RULErule_49;
_currn->_desc1 = (_TSPcodeGen_DBDate) MkcodeGen_DBDate (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_49: root of subtree no. 1 can not be made a codeGen_DBDate node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_49;
return ( (NODEPTR) _currn);
}/* Mkrule_49 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_50 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_50 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_50 _currn;
#ifdef __cplusplus
_currn = new _TPrule_50;
#else
_currn = (_TPPrule_50) TreeNodeAlloc (sizeof (struct _TPrule_50));
#endif
_currn->_prod = RULErule_50;
_currn->_desc1 = (_TSPcodeGen_DBBigInt) MkcodeGen_DBBigInt (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_50: root of subtree no. 1 can not be made a codeGen_DBBigInt node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_50;
return ( (NODEPTR) _currn);
}/* Mkrule_50 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_51 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_51 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_51 _currn;
#ifdef __cplusplus
_currn = new _TPrule_51;
#else
_currn = (_TPPrule_51) TreeNodeAlloc (sizeof (struct _TPrule_51));
#endif
_currn->_prod = RULErule_51;
_currn->_desc1 = (_TSPcodeGen_DBDouble) MkcodeGen_DBDouble (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_51: root of subtree no. 1 can not be made a codeGen_DBDouble node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_51;
return ( (NODEPTR) _currn);
}/* Mkrule_51 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_52 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_52 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_52 _currn;
#ifdef __cplusplus
_currn = new _TPrule_52;
#else
_currn = (_TPPrule_52) TreeNodeAlloc (sizeof (struct _TPrule_52));
#endif
_currn->_prod = RULErule_52;
_currn->_desc1 = (_TSPcodeGen_DBFloat) MkcodeGen_DBFloat (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_52: root of subtree no. 1 can not be made a codeGen_DBFloat node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_52;
return ( (NODEPTR) _currn);
}/* Mkrule_52 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_53 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_53 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_53 _currn;
#ifdef __cplusplus
_currn = new _TPrule_53;
#else
_currn = (_TPPrule_53) TreeNodeAlloc (sizeof (struct _TPrule_53));
#endif
_currn->_prod = RULErule_53;
_currn->_desc1 = (_TSPcodeGen_DBInteger) MkcodeGen_DBInteger (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_53: root of subtree no. 1 can not be made a codeGen_DBInteger node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_53;
return ( (NODEPTR) _currn);
}/* Mkrule_53 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_54 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_54 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_54 _currn;
#ifdef __cplusplus
_currn = new _TPrule_54;
#else
_currn = (_TPPrule_54) TreeNodeAlloc (sizeof (struct _TPrule_54));
#endif
_currn->_prod = RULErule_54;
_currn->_desc1 = (_TSPcodeGen_DBString) MkcodeGen_DBString (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_54: root of subtree no. 1 can not be made a codeGen_DBString node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_54;
return ( (NODEPTR) _currn);
}/* Mkrule_54 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_55 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_55 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_55 _currn;
#ifdef __cplusplus
_currn = new _TPrule_55;
#else
_currn = (_TPPrule_55) TreeNodeAlloc (sizeof (struct _TPrule_55));
#endif
_currn->_prod = RULErule_55;
_currn->_desc1 = (_TSPcodeGen_ComputationExpression) MkcodeGen_ComputationExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_55: root of subtree no. 1 can not be made a codeGen_ComputationExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_55;
return ( (NODEPTR) _currn);
}/* Mkrule_55 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_56 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_56 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_56 _currn;
#ifdef __cplusplus
_currn = new _TPrule_56;
#else
_currn = (_TPPrule_56) TreeNodeAlloc (sizeof (struct _TPrule_56));
#endif
_currn->_prod = RULErule_56;
_currn->_desc1 = (_TSPcodeGen_ParameterValueExpression) MkcodeGen_ParameterValueExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_56: root of subtree no. 1 can not be made a codeGen_ParameterValueExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_56;
return ( (NODEPTR) _currn);
}/* Mkrule_56 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_57 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_57 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_57 _currn;
#ifdef __cplusplus
_currn = new _TPrule_57;
#else
_currn = (_TPPrule_57) TreeNodeAlloc (sizeof (struct _TPrule_57));
#endif
_currn->_prod = RULErule_57;
_currn->_desc1 = (_TSPcodeGen_ColumnValueExpression) MkcodeGen_ColumnValueExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_57: root of subtree no. 1 can not be made a codeGen_ColumnValueExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_57;
return ( (NODEPTR) _currn);
}/* Mkrule_57 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_58 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_58 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_58 _currn;
#ifdef __cplusplus
_currn = new _TPrule_58;
#else
_currn = (_TPPrule_58) TreeNodeAlloc (sizeof (struct _TPrule_58));
#endif
_currn->_prod = RULErule_58;
_currn->_desc1 = (_TSPcodeGen_ConstantExpression) MkcodeGen_ConstantExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_58: root of subtree no. 1 can not be made a codeGen_ConstantExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_58;
return ( (NODEPTR) _currn);
}/* Mkrule_58 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_59 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_59 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_59 _currn;
#ifdef __cplusplus
_currn = new _TPrule_59;
#else
_currn = (_TPPrule_59) TreeNodeAlloc (sizeof (struct _TPrule_59));
#endif
_currn->_prod = RULErule_59;
_currn->_desc1 = (_TSPcodeGen_UnaryExpression) MkcodeGen_UnaryExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_59: root of subtree no. 1 can not be made a codeGen_UnaryExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_59;
return ( (NODEPTR) _currn);
}/* Mkrule_59 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_60 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_60 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_60 _currn;
#ifdef __cplusplus
_currn = new _TPrule_60;
#else
_currn = (_TPPrule_60) TreeNodeAlloc (sizeof (struct _TPrule_60));
#endif
_currn->_prod = RULErule_60;
_currn->_desc1 = (_TSPcodeGen_BinaryExpression) MkcodeGen_BinaryExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_60: root of subtree no. 1 can not be made a codeGen_BinaryExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_60;
return ( (NODEPTR) _currn);
}/* Mkrule_60 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_61 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_61 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_61 _currn;
#ifdef __cplusplus
_currn = new _TPrule_61;
#else
_currn = (_TPPrule_61) TreeNodeAlloc (sizeof (struct _TPrule_61));
#endif
_currn->_prod = RULErule_61;
_currn->_desc1 = (_TSPcodeGen_ComputationExpression) MkcodeGen_ComputationExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_61: root of subtree no. 1 can not be made a codeGen_ComputationExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_61;
return ( (NODEPTR) _currn);
}/* Mkrule_61 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_62 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_62 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_62 _currn;
#ifdef __cplusplus
_currn = new _TPrule_62;
#else
_currn = (_TPPrule_62) TreeNodeAlloc (sizeof (struct _TPrule_62));
#endif
_currn->_prod = RULErule_62;
_currn->_desc1 = (_TSPcodeGen_ParameterValueExpression) MkcodeGen_ParameterValueExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_62: root of subtree no. 1 can not be made a codeGen_ParameterValueExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_62;
return ( (NODEPTR) _currn);
}/* Mkrule_62 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_63 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_63 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_63 _currn;
#ifdef __cplusplus
_currn = new _TPrule_63;
#else
_currn = (_TPPrule_63) TreeNodeAlloc (sizeof (struct _TPrule_63));
#endif
_currn->_prod = RULErule_63;
_currn->_desc1 = (_TSPcodeGen_ColumnValueExpression) MkcodeGen_ColumnValueExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_63: root of subtree no. 1 can not be made a codeGen_ColumnValueExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_63;
return ( (NODEPTR) _currn);
}/* Mkrule_63 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_64 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_64 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_64 _currn;
#ifdef __cplusplus
_currn = new _TPrule_64;
#else
_currn = (_TPPrule_64) TreeNodeAlloc (sizeof (struct _TPrule_64));
#endif
_currn->_prod = RULErule_64;
_currn->_desc1 = (_TSPcodeGen_ConstantExpression) MkcodeGen_ConstantExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_64: root of subtree no. 1 can not be made a codeGen_ConstantExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_64;
return ( (NODEPTR) _currn);
}/* Mkrule_64 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_65 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_65 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_65 _currn;
#ifdef __cplusplus
_currn = new _TPrule_65;
#else
_currn = (_TPPrule_65) TreeNodeAlloc (sizeof (struct _TPrule_65));
#endif
_currn->_prod = RULErule_65;
_currn->_desc1 = (_TSPcodeGen_UnaryExpression) MkcodeGen_UnaryExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_65: root of subtree no. 1 can not be made a codeGen_UnaryExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_65;
return ( (NODEPTR) _currn);
}/* Mkrule_65 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_66 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_66 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_66 _currn;
#ifdef __cplusplus
_currn = new _TPrule_66;
#else
_currn = (_TPPrule_66) TreeNodeAlloc (sizeof (struct _TPrule_66));
#endif
_currn->_prod = RULErule_66;
_currn->_desc1 = (_TSPcodeGen_BinaryExpression) MkcodeGen_BinaryExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_66: root of subtree no. 1 can not be made a codeGen_BinaryExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_66;
return ( (NODEPTR) _currn);
}/* Mkrule_66 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_67 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_67 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_67 _currn;
#ifdef __cplusplus
_currn = new _TPrule_67;
#else
_currn = (_TPPrule_67) TreeNodeAlloc (sizeof (struct _TPrule_67));
#endif
_currn->_prod = RULErule_67;
_currn->_desc1 = (_TSPcodeGen_ComputationExpression) MkcodeGen_ComputationExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_67: root of subtree no. 1 can not be made a codeGen_ComputationExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_67;
return ( (NODEPTR) _currn);
}/* Mkrule_67 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_68 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_68 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_68 _currn;
#ifdef __cplusplus
_currn = new _TPrule_68;
#else
_currn = (_TPPrule_68) TreeNodeAlloc (sizeof (struct _TPrule_68));
#endif
_currn->_prod = RULErule_68;
_currn->_desc1 = (_TSPcodeGen_ParameterValueExpression) MkcodeGen_ParameterValueExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_68: root of subtree no. 1 can not be made a codeGen_ParameterValueExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_68;
return ( (NODEPTR) _currn);
}/* Mkrule_68 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_69 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_69 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_69 _currn;
#ifdef __cplusplus
_currn = new _TPrule_69;
#else
_currn = (_TPPrule_69) TreeNodeAlloc (sizeof (struct _TPrule_69));
#endif
_currn->_prod = RULErule_69;
_currn->_desc1 = (_TSPcodeGen_ColumnValueExpression) MkcodeGen_ColumnValueExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_69: root of subtree no. 1 can not be made a codeGen_ColumnValueExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_69;
return ( (NODEPTR) _currn);
}/* Mkrule_69 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_70 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_70 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_70 _currn;
#ifdef __cplusplus
_currn = new _TPrule_70;
#else
_currn = (_TPPrule_70) TreeNodeAlloc (sizeof (struct _TPrule_70));
#endif
_currn->_prod = RULErule_70;
_currn->_desc1 = (_TSPcodeGen_ConstantExpression) MkcodeGen_ConstantExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_70: root of subtree no. 1 can not be made a codeGen_ConstantExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_70;
return ( (NODEPTR) _currn);
}/* Mkrule_70 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_71 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_71 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_71 _currn;
#ifdef __cplusplus
_currn = new _TPrule_71;
#else
_currn = (_TPPrule_71) TreeNodeAlloc (sizeof (struct _TPrule_71));
#endif
_currn->_prod = RULErule_71;
_currn->_desc1 = (_TSPcodeGen_UnaryExpression) MkcodeGen_UnaryExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_71: root of subtree no. 1 can not be made a codeGen_UnaryExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_71;
return ( (NODEPTR) _currn);
}/* Mkrule_71 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_72 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_72 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_72 _currn;
#ifdef __cplusplus
_currn = new _TPrule_72;
#else
_currn = (_TPPrule_72) TreeNodeAlloc (sizeof (struct _TPrule_72));
#endif
_currn->_prod = RULErule_72;
_currn->_desc1 = (_TSPcodeGen_BinaryExpression) MkcodeGen_BinaryExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_72: root of subtree no. 1 can not be made a codeGen_BinaryExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_72;
return ( (NODEPTR) _currn);
}/* Mkrule_72 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_73 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_73 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_73 _currn;
#ifdef __cplusplus
_currn = new _TPrule_73;
#else
_currn = (_TPPrule_73) TreeNodeAlloc (sizeof (struct _TPrule_73));
#endif
_currn->_prod = RULErule_73;
_currn->_desc1 = (_TSPcodeGen_WebNodeColumnExpression) MkcodeGen_WebNodeColumnExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_73: root of subtree no. 1 can not be made a codeGen_WebNodeColumnExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_73;
return ( (NODEPTR) _currn);
}/* Mkrule_73 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_74 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_74 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_74 _currn;
#ifdef __cplusplus
_currn = new _TPrule_74;
#else
_currn = (_TPPrule_74) TreeNodeAlloc (sizeof (struct _TPrule_74));
#endif
_currn->_prod = RULErule_74;
_currn->_desc1 = (_TSPcodeGen_WebNodeVariableExpression) MkcodeGen_WebNodeVariableExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_74: root of subtree no. 1 can not be made a codeGen_WebNodeVariableExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_74;
return ( (NODEPTR) _currn);
}/* Mkrule_74 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_75 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_75 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_75 _currn;
#ifdef __cplusplus
_currn = new _TPrule_75;
#else
_currn = (_TPPrule_75) TreeNodeAlloc (sizeof (struct _TPrule_75));
#endif
_currn->_prod = RULErule_75;
_currn->_desc1 = (_TSPcodeGen_WebNodeConstantExpression) MkcodeGen_WebNodeConstantExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_75: root of subtree no. 1 can not be made a codeGen_WebNodeConstantExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_75;
return ( (NODEPTR) _currn);
}/* Mkrule_75 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_76 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_76 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_76 _currn;
#ifdef __cplusplus
_currn = new _TPrule_76;
#else
_currn = (_TPPrule_76) TreeNodeAlloc (sizeof (struct _TPrule_76));
#endif
_currn->_prod = RULErule_76;
_currn->_desc1 = (_TSPcodeGen_DBBlob) MkcodeGen_DBBlob (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_76: root of subtree no. 1 can not be made a codeGen_DBBlob node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_76;
return ( (NODEPTR) _currn);
}/* Mkrule_76 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_77 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_77 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_77 _currn;
#ifdef __cplusplus
_currn = new _TPrule_77;
#else
_currn = (_TPPrule_77) TreeNodeAlloc (sizeof (struct _TPrule_77));
#endif
_currn->_prod = RULErule_77;
_currn->_desc1 = (_TSPcodeGen_DBBoolean) MkcodeGen_DBBoolean (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_77: root of subtree no. 1 can not be made a codeGen_DBBoolean node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_77;
return ( (NODEPTR) _currn);
}/* Mkrule_77 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_78 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_78 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_78 _currn;
#ifdef __cplusplus
_currn = new _TPrule_78;
#else
_currn = (_TPPrule_78) TreeNodeAlloc (sizeof (struct _TPrule_78));
#endif
_currn->_prod = RULErule_78;
_currn->_desc1 = (_TSPcodeGen_DBTimeStamp) MkcodeGen_DBTimeStamp (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_78: root of subtree no. 1 can not be made a codeGen_DBTimeStamp node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_78;
return ( (NODEPTR) _currn);
}/* Mkrule_78 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_79 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_79 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_79 _currn;
#ifdef __cplusplus
_currn = new _TPrule_79;
#else
_currn = (_TPPrule_79) TreeNodeAlloc (sizeof (struct _TPrule_79));
#endif
_currn->_prod = RULErule_79;
_currn->_desc1 = (_TSPcodeGen_DBDate) MkcodeGen_DBDate (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_79: root of subtree no. 1 can not be made a codeGen_DBDate node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_79;
return ( (NODEPTR) _currn);
}/* Mkrule_79 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_80 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_80 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_80 _currn;
#ifdef __cplusplus
_currn = new _TPrule_80;
#else
_currn = (_TPPrule_80) TreeNodeAlloc (sizeof (struct _TPrule_80));
#endif
_currn->_prod = RULErule_80;
_currn->_desc1 = (_TSPcodeGen_DBBigInt) MkcodeGen_DBBigInt (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_80: root of subtree no. 1 can not be made a codeGen_DBBigInt node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_80;
return ( (NODEPTR) _currn);
}/* Mkrule_80 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_81 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_81 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_81 _currn;
#ifdef __cplusplus
_currn = new _TPrule_81;
#else
_currn = (_TPPrule_81) TreeNodeAlloc (sizeof (struct _TPrule_81));
#endif
_currn->_prod = RULErule_81;
_currn->_desc1 = (_TSPcodeGen_DBDouble) MkcodeGen_DBDouble (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_81: root of subtree no. 1 can not be made a codeGen_DBDouble node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_81;
return ( (NODEPTR) _currn);
}/* Mkrule_81 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_82 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_82 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_82 _currn;
#ifdef __cplusplus
_currn = new _TPrule_82;
#else
_currn = (_TPPrule_82) TreeNodeAlloc (sizeof (struct _TPrule_82));
#endif
_currn->_prod = RULErule_82;
_currn->_desc1 = (_TSPcodeGen_DBFloat) MkcodeGen_DBFloat (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_82: root of subtree no. 1 can not be made a codeGen_DBFloat node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_82;
return ( (NODEPTR) _currn);
}/* Mkrule_82 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_83 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_83 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_83 _currn;
#ifdef __cplusplus
_currn = new _TPrule_83;
#else
_currn = (_TPPrule_83) TreeNodeAlloc (sizeof (struct _TPrule_83));
#endif
_currn->_prod = RULErule_83;
_currn->_desc1 = (_TSPcodeGen_DBInteger) MkcodeGen_DBInteger (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_83: root of subtree no. 1 can not be made a codeGen_DBInteger node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_83;
return ( (NODEPTR) _currn);
}/* Mkrule_83 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_84 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_84 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_84 _currn;
#ifdef __cplusplus
_currn = new _TPrule_84;
#else
_currn = (_TPPrule_84) TreeNodeAlloc (sizeof (struct _TPrule_84));
#endif
_currn->_prod = RULErule_84;
_currn->_desc1 = (_TSPcodeGen_DBString) MkcodeGen_DBString (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_84: root of subtree no. 1 can not be made a codeGen_DBString node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_84;
return ( (NODEPTR) _currn);
}/* Mkrule_84 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_85 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_85 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_85 _currn;
#ifdef __cplusplus
_currn = new _TPrule_85;
#else
_currn = (_TPPrule_85) TreeNodeAlloc (sizeof (struct _TPrule_85));
#endif
_currn->_prod = RULErule_85;
_currn->_desc1 = (_TSPcodeGen_FormularCondition) MkcodeGen_FormularCondition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_85: root of subtree no. 1 can not be made a codeGen_FormularCondition node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_85;
return ( (NODEPTR) _currn);
}/* Mkrule_85 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_86 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_86 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_86 _currn;
#ifdef __cplusplus
_currn = new _TPrule_86;
#else
_currn = (_TPPrule_86) TreeNodeAlloc (sizeof (struct _TPrule_86));
#endif
_currn->_prod = RULErule_86;
_currn->_desc1 = (_TSPcodeGen_FormularCondition) MkcodeGen_FormularCondition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_86: root of subtree no. 1 can not be made a codeGen_FormularCondition node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_86;
return ( (NODEPTR) _currn);
}/* Mkrule_86 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_87 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_87 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_87 _currn;
#ifdef __cplusplus
_currn = new _TPrule_87;
#else
_currn = (_TPPrule_87) TreeNodeAlloc (sizeof (struct _TPrule_87));
#endif
_currn->_prod = RULErule_87;
_currn->_desc1 = (_TSPcodeGen_FormularCondition) MkcodeGen_FormularCondition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_87: root of subtree no. 1 can not be made a codeGen_FormularCondition node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_87;
return ( (NODEPTR) _currn);
}/* Mkrule_87 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_88 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_88 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_88 _currn;
#ifdef __cplusplus
_currn = new _TPrule_88;
#else
_currn = (_TPPrule_88) TreeNodeAlloc (sizeof (struct _TPrule_88));
#endif
_currn->_prod = RULErule_88;
_currn->_desc1 = (_TSPcodeGen_FormularCondition) MkcodeGen_FormularCondition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_88: root of subtree no. 1 can not be made a codeGen_FormularCondition node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_88;
return ( (NODEPTR) _currn);
}/* Mkrule_88 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_89 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_89 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_89 _currn;
#ifdef __cplusplus
_currn = new _TPrule_89;
#else
_currn = (_TPPrule_89) TreeNodeAlloc (sizeof (struct _TPrule_89));
#endif
_currn->_prod = RULErule_89;
_currn->_desc1 = (_TSPcodeGen_FormularCondition) MkcodeGen_FormularCondition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_89: root of subtree no. 1 can not be made a codeGen_FormularCondition node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_89;
return ( (NODEPTR) _currn);
}/* Mkrule_89 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_90 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_90 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_90 _currn;
#ifdef __cplusplus
_currn = new _TPrule_90;
#else
_currn = (_TPPrule_90) TreeNodeAlloc (sizeof (struct _TPrule_90));
#endif
_currn->_prod = RULErule_90;
_currn->_desc1 = (_TSPcodeGen_FormularCondition) MkcodeGen_FormularCondition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_90: root of subtree no. 1 can not be made a codeGen_FormularCondition node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_90;
return ( (NODEPTR) _currn);
}/* Mkrule_90 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_91 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_91 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_91 _currn;
#ifdef __cplusplus
_currn = new _TPrule_91;
#else
_currn = (_TPPrule_91) TreeNodeAlloc (sizeof (struct _TPrule_91));
#endif
_currn->_prod = RULErule_91;
_currn->_desc1 = (_TSPcodeGen_FormularPHPScriptSelectOption) MkcodeGen_FormularPHPScriptSelectOption (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_91: root of subtree no. 1 can not be made a codeGen_FormularPHPScriptSelectOption node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_91;
return ( (NODEPTR) _currn);
}/* Mkrule_91 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_92 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_92 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_92 _currn;
#ifdef __cplusplus
_currn = new _TPrule_92;
#else
_currn = (_TPPrule_92) TreeNodeAlloc (sizeof (struct _TPrule_92));
#endif
_currn->_prod = RULErule_92;
_currn->_desc1 = (_TSPcodeGen_FormularDataSourceSelectOption) MkcodeGen_FormularDataSourceSelectOption (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_92: root of subtree no. 1 can not be made a codeGen_FormularDataSourceSelectOption node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_92;
return ( (NODEPTR) _currn);
}/* Mkrule_92 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_93 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_93 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_93 _currn;
#ifdef __cplusplus
_currn = new _TPrule_93;
#else
_currn = (_TPPrule_93) TreeNodeAlloc (sizeof (struct _TPrule_93));
#endif
_currn->_prod = RULErule_93;
_currn->_desc1 = (_TSPcodeGen_FormularConstantSelectOption) MkcodeGen_FormularConstantSelectOption (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_93: root of subtree no. 1 can not be made a codeGen_FormularConstantSelectOption node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_93;
return ( (NODEPTR) _currn);
}/* Mkrule_93 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_94 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_94 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_94 _currn;
#ifdef __cplusplus
_currn = new _TPrule_94;
#else
_currn = (_TPPrule_94) TreeNodeAlloc (sizeof (struct _TPrule_94));
#endif
_currn->_prod = RULErule_94;
_currn->_desc1 = (_TSPcodeGen_WebNodeColumnExpression) MkcodeGen_WebNodeColumnExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_94: root of subtree no. 1 can not be made a codeGen_WebNodeColumnExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_94;
return ( (NODEPTR) _currn);
}/* Mkrule_94 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_95 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_95 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_95 _currn;
#ifdef __cplusplus
_currn = new _TPrule_95;
#else
_currn = (_TPPrule_95) TreeNodeAlloc (sizeof (struct _TPrule_95));
#endif
_currn->_prod = RULErule_95;
_currn->_desc1 = (_TSPcodeGen_WebNodeVariableExpression) MkcodeGen_WebNodeVariableExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_95: root of subtree no. 1 can not be made a codeGen_WebNodeVariableExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_95;
return ( (NODEPTR) _currn);
}/* Mkrule_95 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_96 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_96 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_96 _currn;
#ifdef __cplusplus
_currn = new _TPrule_96;
#else
_currn = (_TPPrule_96) TreeNodeAlloc (sizeof (struct _TPrule_96));
#endif
_currn->_prod = RULErule_96;
_currn->_desc1 = (_TSPcodeGen_WebNodeConstantExpression) MkcodeGen_WebNodeConstantExpression (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_96: root of subtree no. 1 can not be made a codeGen_WebNodeConstantExpression node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_96;
return ( (NODEPTR) _currn);
}/* Mkrule_96 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_97 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_97 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_97 _currn;
#ifdef __cplusplus
_currn = new _TPrule_97;
#else
_currn = (_TPPrule_97) TreeNodeAlloc (sizeof (struct _TPrule_97));
#endif
_currn->_prod = RULErule_97;
_currn->_desc1 = (_TSPcodeGen_DBBlob) MkcodeGen_DBBlob (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_97: root of subtree no. 1 can not be made a codeGen_DBBlob node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_97;
return ( (NODEPTR) _currn);
}/* Mkrule_97 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_98 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_98 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_98 _currn;
#ifdef __cplusplus
_currn = new _TPrule_98;
#else
_currn = (_TPPrule_98) TreeNodeAlloc (sizeof (struct _TPrule_98));
#endif
_currn->_prod = RULErule_98;
_currn->_desc1 = (_TSPcodeGen_DBBoolean) MkcodeGen_DBBoolean (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_98: root of subtree no. 1 can not be made a codeGen_DBBoolean node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_98;
return ( (NODEPTR) _currn);
}/* Mkrule_98 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_99 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_99 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_99 _currn;
#ifdef __cplusplus
_currn = new _TPrule_99;
#else
_currn = (_TPPrule_99) TreeNodeAlloc (sizeof (struct _TPrule_99));
#endif
_currn->_prod = RULErule_99;
_currn->_desc1 = (_TSPcodeGen_DBTimeStamp) MkcodeGen_DBTimeStamp (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_99: root of subtree no. 1 can not be made a codeGen_DBTimeStamp node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_99;
return ( (NODEPTR) _currn);
}/* Mkrule_99 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_100 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_100 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_100 _currn;
#ifdef __cplusplus
_currn = new _TPrule_100;
#else
_currn = (_TPPrule_100) TreeNodeAlloc (sizeof (struct _TPrule_100));
#endif
_currn->_prod = RULErule_100;
_currn->_desc1 = (_TSPcodeGen_DBDate) MkcodeGen_DBDate (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_100: root of subtree no. 1 can not be made a codeGen_DBDate node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_100;
return ( (NODEPTR) _currn);
}/* Mkrule_100 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_101 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_101 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_101 _currn;
#ifdef __cplusplus
_currn = new _TPrule_101;
#else
_currn = (_TPPrule_101) TreeNodeAlloc (sizeof (struct _TPrule_101));
#endif
_currn->_prod = RULErule_101;
_currn->_desc1 = (_TSPcodeGen_DBBigInt) MkcodeGen_DBBigInt (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_101: root of subtree no. 1 can not be made a codeGen_DBBigInt node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_101;
return ( (NODEPTR) _currn);
}/* Mkrule_101 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_102 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_102 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_102 _currn;
#ifdef __cplusplus
_currn = new _TPrule_102;
#else
_currn = (_TPPrule_102) TreeNodeAlloc (sizeof (struct _TPrule_102));
#endif
_currn->_prod = RULErule_102;
_currn->_desc1 = (_TSPcodeGen_DBDouble) MkcodeGen_DBDouble (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_102: root of subtree no. 1 can not be made a codeGen_DBDouble node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_102;
return ( (NODEPTR) _currn);
}/* Mkrule_102 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_103 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_103 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_103 _currn;
#ifdef __cplusplus
_currn = new _TPrule_103;
#else
_currn = (_TPPrule_103) TreeNodeAlloc (sizeof (struct _TPrule_103));
#endif
_currn->_prod = RULErule_103;
_currn->_desc1 = (_TSPcodeGen_DBFloat) MkcodeGen_DBFloat (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_103: root of subtree no. 1 can not be made a codeGen_DBFloat node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_103;
return ( (NODEPTR) _currn);
}/* Mkrule_103 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_104 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_104 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_104 _currn;
#ifdef __cplusplus
_currn = new _TPrule_104;
#else
_currn = (_TPPrule_104) TreeNodeAlloc (sizeof (struct _TPrule_104));
#endif
_currn->_prod = RULErule_104;
_currn->_desc1 = (_TSPcodeGen_DBInteger) MkcodeGen_DBInteger (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_104: root of subtree no. 1 can not be made a codeGen_DBInteger node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_104;
return ( (NODEPTR) _currn);
}/* Mkrule_104 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_105 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_105 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_105 _currn;
#ifdef __cplusplus
_currn = new _TPrule_105;
#else
_currn = (_TPPrule_105) TreeNodeAlloc (sizeof (struct _TPrule_105));
#endif
_currn->_prod = RULErule_105;
_currn->_desc1 = (_TSPcodeGen_DBString) MkcodeGen_DBString (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_105: root of subtree no. 1 can not be made a codeGen_DBString node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_105;
return ( (NODEPTR) _currn);
}/* Mkrule_105 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_106 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_106 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_106 _currn;
#ifdef __cplusplus
_currn = new _TPrule_106;
#else
_currn = (_TPPrule_106) TreeNodeAlloc (sizeof (struct _TPrule_106));
#endif
_currn->_prod = RULErule_106;
_currn->_desc1 = (_TSPcodeGen_Frame) MkcodeGen_Frame (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_106: root of subtree no. 1 can not be made a codeGen_Frame node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_106;
return ( (NODEPTR) _currn);
}/* Mkrule_106 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_107 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_107 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_107 _currn;
#ifdef __cplusplus
_currn = new _TPrule_107;
#else
_currn = (_TPPrule_107) TreeNodeAlloc (sizeof (struct _TPrule_107));
#endif
_currn->_prod = RULErule_107;
_currn->_desc1 = (_TSPcodeGen_WebApplication) MkcodeGen_WebApplication (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_107: root of subtree no. 1 can not be made a codeGen_WebApplication node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_107;
return ( (NODEPTR) _currn);
}/* Mkrule_107 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_108 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_108 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_108 _currn;
#ifdef __cplusplus
_currn = new _TPrule_108;
#else
_currn = (_TPPrule_108) TreeNodeAlloc (sizeof (struct _TPrule_108));
#endif
_currn->_prod = RULErule_108;
_currn->_desc1 = (_TSPcodeGen_RootWebZone) MkcodeGen_RootWebZone (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_108: root of subtree no. 1 can not be made a codeGen_RootWebZone node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_108;
return ( (NODEPTR) _currn);
}/* Mkrule_108 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_109 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_109 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_109 _currn;
#ifdef __cplusplus
_currn = new _TPrule_109;
#else
_currn = (_TPPrule_109) TreeNodeAlloc (sizeof (struct _TPrule_109));
#endif
_currn->_prod = RULErule_109;
_currn->_desc1 = (_TSPcodeGen_NavigationBar) MkcodeGen_NavigationBar (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_109: root of subtree no. 1 can not be made a codeGen_NavigationBar node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_109;
return ( (NODEPTR) _currn);
}/* Mkrule_109 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_110 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_110 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_110 _currn;
#ifdef __cplusplus
_currn = new _TPrule_110;
#else
_currn = (_TPPrule_110) TreeNodeAlloc (sizeof (struct _TPrule_110));
#endif
_currn->_prod = RULErule_110;
_currn->_desc1 = (_TSPcodeGen_StyleSheet) MkcodeGen_StyleSheet (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_110: root of subtree no. 1 can not be made a codeGen_StyleSheet node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_110;
return ( (NODEPTR) _currn);
}/* Mkrule_110 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_111 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_111 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_111 _currn;
#ifdef __cplusplus
_currn = new _TPrule_111;
#else
_currn = (_TPPrule_111) TreeNodeAlloc (sizeof (struct _TPrule_111));
#endif
_currn->_prod = RULErule_111;
_currn->_desc1 = (_TSPcodeGen_MetaInformation) MkcodeGen_MetaInformation (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_111: root of subtree no. 1 can not be made a codeGen_MetaInformation node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_111;
return ( (NODEPTR) _currn);
}/* Mkrule_111 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_112 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_112 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_112 _currn;
#ifdef __cplusplus
_currn = new _TPrule_112;
#else
_currn = (_TPPrule_112) TreeNodeAlloc (sizeof (struct _TPrule_112));
#endif
_currn->_prod = RULErule_112;
_currn->_desc1 = (_TSPcodeGen_DataBase) MkcodeGen_DataBase (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_112: root of subtree no. 1 can not be made a codeGen_DataBase node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_112;
return ( (NODEPTR) _currn);
}/* Mkrule_112 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_113 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_113 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_113 _currn;
#ifdef __cplusplus
_currn = new _TPrule_113;
#else
_currn = (_TPPrule_113) TreeNodeAlloc (sizeof (struct _TPrule_113));
#endif
_currn->_prod = RULErule_113;
_currn->_desc1 = (_TSPcodeGen_Frame) MkcodeGen_Frame (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_113: root of subtree no. 1 can not be made a codeGen_Frame node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_113;
return ( (NODEPTR) _currn);
}/* Mkrule_113 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_114 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_114 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_114 _currn;
#ifdef __cplusplus
_currn = new _TPrule_114;
#else
_currn = (_TPPrule_114) TreeNodeAlloc (sizeof (struct _TPrule_114));
#endif
_currn->_prod = RULErule_114;
_currn->_desc1 = (_TSPcodeGen_CustomValue) MkcodeGen_CustomValue (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_114: root of subtree no. 1 can not be made a codeGen_CustomValue node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_114;
return ( (NODEPTR) _currn);
}/* Mkrule_114 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_115 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_115 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_115 _currn;
#ifdef __cplusplus
_currn = new _TPrule_115;
#else
_currn = (_TPPrule_115) TreeNodeAlloc (sizeof (struct _TPrule_115));
#endif
_currn->_prod = RULErule_115;
_currn->_desc1 = (_TSPcodeGen_ListValue) MkcodeGen_ListValue (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_115: root of subtree no. 1 can not be made a codeGen_ListValue node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_115;
return ( (NODEPTR) _currn);
}/* Mkrule_115 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_116 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_116 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_116 _currn;
#ifdef __cplusplus
_currn = new _TPrule_116;
#else
_currn = (_TPPrule_116) TreeNodeAlloc (sizeof (struct _TPrule_116));
#endif
_currn->_prod = RULErule_116;
_currn->_desc1 = (_TSPcodeGen_AlignmentValue) MkcodeGen_AlignmentValue (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_116: root of subtree no. 1 can not be made a codeGen_AlignmentValue node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_116;
return ( (NODEPTR) _currn);
}/* Mkrule_116 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_117 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_117 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_117 _currn;
#ifdef __cplusplus
_currn = new _TPrule_117;
#else
_currn = (_TPPrule_117) TreeNodeAlloc (sizeof (struct _TPrule_117));
#endif
_currn->_prod = RULErule_117;
_currn->_desc1 = (_TSPcodeGen_HeadingValue) MkcodeGen_HeadingValue (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_117: root of subtree no. 1 can not be made a codeGen_HeadingValue node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_117;
return ( (NODEPTR) _currn);
}/* Mkrule_117 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_118 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_118 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_118 _currn;
#ifdef __cplusplus
_currn = new _TPrule_118;
#else
_currn = (_TPPrule_118) TreeNodeAlloc (sizeof (struct _TPrule_118));
#endif
_currn->_prod = RULErule_118;
_currn->_desc1 = (_TSPcodeGen_StyleValue) MkcodeGen_StyleValue (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_118: root of subtree no. 1 can not be made a codeGen_StyleValue node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_118;
return ( (NODEPTR) _currn);
}/* Mkrule_118 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_119 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_119 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_119 _currn;
#ifdef __cplusplus
_currn = new _TPrule_119;
#else
_currn = (_TPPrule_119) TreeNodeAlloc (sizeof (struct _TPrule_119));
#endif
_currn->_prod = RULErule_119;
_currn->_desc1 = (_TSPcodeGen_FamilyValue) MkcodeGen_FamilyValue (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_119: root of subtree no. 1 can not be made a codeGen_FamilyValue node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_119;
return ( (NODEPTR) _currn);
}/* Mkrule_119 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_120 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_120 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_120 _currn;
#ifdef __cplusplus
_currn = new _TPrule_120;
#else
_currn = (_TPPrule_120) TreeNodeAlloc (sizeof (struct _TPrule_120));
#endif
_currn->_prod = RULErule_120;
_currn->_desc1 = (_TSPcodeGen_ColorValue) MkcodeGen_ColorValue (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_120: root of subtree no. 1 can not be made a codeGen_ColorValue node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_120;
return ( (NODEPTR) _currn);
}/* Mkrule_120 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_121 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_121 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_121 _currn;
#ifdef __cplusplus
_currn = new _TPrule_121;
#else
_currn = (_TPPrule_121) TreeNodeAlloc (sizeof (struct _TPrule_121));
#endif
_currn->_prod = RULErule_121;
_currn->_desc1 = (_TSPcodeGen_SizeValue) MkcodeGen_SizeValue (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_121: root of subtree no. 1 can not be made a codeGen_SizeValue node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_121;
return ( (NODEPTR) _currn);
}/* Mkrule_121 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_122 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_122 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_122 _currn;
#ifdef __cplusplus
_currn = new _TPrule_122;
#else
_currn = (_TPPrule_122) TreeNodeAlloc (sizeof (struct _TPrule_122));
#endif
_currn->_prod = RULErule_122;
_currn->_desc1 = (_TSPcodeGen_ActionHandler) MkcodeGen_ActionHandler (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_122: root of subtree no. 1 can not be made a codeGen_ActionHandler node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_122;
return ( (NODEPTR) _currn);
}/* Mkrule_122 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_123 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_123 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_123 _currn;
#ifdef __cplusplus
_currn = new _TPrule_123;
#else
_currn = (_TPPrule_123) TreeNodeAlloc (sizeof (struct _TPrule_123));
#endif
_currn->_prod = RULErule_123;
_currn->_desc1 = (_TSPglobalFrameView_FormularCondition) MkglobalFrameView_FormularCondition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_123: root of subtree no. 1 can not be made a globalFrameView_FormularCondition node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_123;
return ( (NODEPTR) _currn);
}/* Mkrule_123 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_124 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_124 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_124 _currn;
#ifdef __cplusplus
_currn = new _TPrule_124;
#else
_currn = (_TPPrule_124) TreeNodeAlloc (sizeof (struct _TPrule_124));
#endif
_currn->_prod = RULErule_124;
_currn->_desc1 = (_TSPglobalFrameView_FormularCondition) MkglobalFrameView_FormularCondition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_124: root of subtree no. 1 can not be made a globalFrameView_FormularCondition node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_124;
return ( (NODEPTR) _currn);
}/* Mkrule_124 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_125 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_125 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_125 _currn;
#ifdef __cplusplus
_currn = new _TPrule_125;
#else
_currn = (_TPPrule_125) TreeNodeAlloc (sizeof (struct _TPrule_125));
#endif
_currn->_prod = RULErule_125;
_currn->_desc1 = (_TSPglobalFrameView_FormularCondition) MkglobalFrameView_FormularCondition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_125: root of subtree no. 1 can not be made a globalFrameView_FormularCondition node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_125;
return ( (NODEPTR) _currn);
}/* Mkrule_125 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_126 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_126 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_126 _currn;
#ifdef __cplusplus
_currn = new _TPrule_126;
#else
_currn = (_TPPrule_126) TreeNodeAlloc (sizeof (struct _TPrule_126));
#endif
_currn->_prod = RULErule_126;
_currn->_desc1 = (_TSPglobalFrameView_FormularCondition) MkglobalFrameView_FormularCondition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_126: root of subtree no. 1 can not be made a globalFrameView_FormularCondition node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_126;
return ( (NODEPTR) _currn);
}/* Mkrule_126 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_127 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_127 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_127 _currn;
#ifdef __cplusplus
_currn = new _TPrule_127;
#else
_currn = (_TPPrule_127) TreeNodeAlloc (sizeof (struct _TPrule_127));
#endif
_currn->_prod = RULErule_127;
_currn->_desc1 = (_TSPglobalFrameView_FormularCondition) MkglobalFrameView_FormularCondition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_127: root of subtree no. 1 can not be made a globalFrameView_FormularCondition node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_127;
return ( (NODEPTR) _currn);
}/* Mkrule_127 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_128 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_128 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_128 _currn;
#ifdef __cplusplus
_currn = new _TPrule_128;
#else
_currn = (_TPPrule_128) TreeNodeAlloc (sizeof (struct _TPrule_128));
#endif
_currn->_prod = RULErule_128;
_currn->_desc1 = (_TSPglobalFrameView_Frame) MkglobalFrameView_Frame (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_128: root of subtree no. 1 can not be made a globalFrameView_Frame node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_128;
return ( (NODEPTR) _currn);
}/* Mkrule_128 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_129 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_129 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_129 _currn;
#ifdef __cplusplus
_currn = new _TPrule_129;
#else
_currn = (_TPPrule_129) TreeNodeAlloc (sizeof (struct _TPrule_129));
#endif
_currn->_prod = RULErule_129;
_currn->_desc1 = (_TSProotView_FormularCondition) MkrootView_FormularCondition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_129: root of subtree no. 1 can not be made a rootView_FormularCondition node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_129;
return ( (NODEPTR) _currn);
}/* Mkrule_129 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_130 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_130 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_130 _currn;
#ifdef __cplusplus
_currn = new _TPrule_130;
#else
_currn = (_TPPrule_130) TreeNodeAlloc (sizeof (struct _TPrule_130));
#endif
_currn->_prod = RULErule_130;
_currn->_desc1 = (_TSProotView_FormularCondition) MkrootView_FormularCondition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_130: root of subtree no. 1 can not be made a rootView_FormularCondition node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_130;
return ( (NODEPTR) _currn);
}/* Mkrule_130 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_131 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_131 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_131 _currn;
#ifdef __cplusplus
_currn = new _TPrule_131;
#else
_currn = (_TPPrule_131) TreeNodeAlloc (sizeof (struct _TPrule_131));
#endif
_currn->_prod = RULErule_131;
_currn->_desc1 = (_TSProotView_FormularCondition) MkrootView_FormularCondition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_131: root of subtree no. 1 can not be made a rootView_FormularCondition node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_131;
return ( (NODEPTR) _currn);
}/* Mkrule_131 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_132 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_132 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_132 _currn;
#ifdef __cplusplus
_currn = new _TPrule_132;
#else
_currn = (_TPPrule_132) TreeNodeAlloc (sizeof (struct _TPrule_132));
#endif
_currn->_prod = RULErule_132;
_currn->_desc1 = (_TSProotView_FormularCondition) MkrootView_FormularCondition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_132: root of subtree no. 1 can not be made a rootView_FormularCondition node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_132;
return ( (NODEPTR) _currn);
}/* Mkrule_132 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_133 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_133 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_133 _currn;
#ifdef __cplusplus
_currn = new _TPrule_133;
#else
_currn = (_TPPrule_133) TreeNodeAlloc (sizeof (struct _TPrule_133));
#endif
_currn->_prod = RULErule_133;
_currn->_desc1 = (_TSProotView_FormularCondition) MkrootView_FormularCondition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_133: root of subtree no. 1 can not be made a rootView_FormularCondition node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_133;
return ( (NODEPTR) _currn);
}/* Mkrule_133 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_134 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_134 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_134 _currn;
#ifdef __cplusplus
_currn = new _TPrule_134;
#else
_currn = (_TPPrule_134) TreeNodeAlloc (sizeof (struct _TPrule_134));
#endif
_currn->_prod = RULErule_134;
_currn->_desc1 = (_TSProotView_FormularCondition) MkrootView_FormularCondition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_134: root of subtree no. 1 can not be made a rootView_FormularCondition node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_134;
return ( (NODEPTR) _currn);
}/* Mkrule_134 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_135 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_135 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_135 _currn;
#ifdef __cplusplus
_currn = new _TPrule_135;
#else
_currn = (_TPPrule_135) TreeNodeAlloc (sizeof (struct _TPrule_135));
#endif
_currn->_prod = RULErule_135;
_currn->_desc1 = (_TSProotView_Frame) MkrootView_Frame (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_135: root of subtree no. 1 can not be made a rootView_Frame node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_135;
return ( (NODEPTR) _currn);
}/* Mkrule_135 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_136 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_136 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_136 _currn;
#ifdef __cplusplus
_currn = new _TPrule_136;
#else
_currn = (_TPPrule_136) TreeNodeAlloc (sizeof (struct _TPrule_136));
#endif
_currn->_prod = RULErule_136;
_currn->_desc1 = (_TSProotView_WebApplication) MkrootView_WebApplication (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_136: root of subtree no. 1 can not be made a rootView_WebApplication node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_136;
return ( (NODEPTR) _currn);
}/* Mkrule_136 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_137 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_137 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_137 _currn;
#ifdef __cplusplus
_currn = new _TPrule_137;
#else
_currn = (_TPPrule_137) TreeNodeAlloc (sizeof (struct _TPrule_137));
#endif
_currn->_prod = RULErule_137;
_currn->_desc1 = (_TSProotView_RootWebZone) MkrootView_RootWebZone (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_137: root of subtree no. 1 can not be made a rootView_RootWebZone node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_137;
return ( (NODEPTR) _currn);
}/* Mkrule_137 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_138 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_138 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_138 _currn;
#ifdef __cplusplus
_currn = new _TPrule_138;
#else
_currn = (_TPPrule_138) TreeNodeAlloc (sizeof (struct _TPrule_138));
#endif
_currn->_prod = RULErule_138;
_currn->_desc1 = (_TSProotView_NavigationBar) MkrootView_NavigationBar (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_138: root of subtree no. 1 can not be made a rootView_NavigationBar node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_138;
return ( (NODEPTR) _currn);
}/* Mkrule_138 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_139 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_139 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_139 _currn;
#ifdef __cplusplus
_currn = new _TPrule_139;
#else
_currn = (_TPPrule_139) TreeNodeAlloc (sizeof (struct _TPrule_139));
#endif
_currn->_prod = RULErule_139;
_currn->_desc1 = (_TSProotView_StyleSheet) MkrootView_StyleSheet (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_139: root of subtree no. 1 can not be made a rootView_StyleSheet node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_139;
return ( (NODEPTR) _currn);
}/* Mkrule_139 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_140 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_140 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_140 _currn;
#ifdef __cplusplus
_currn = new _TPrule_140;
#else
_currn = (_TPPrule_140) TreeNodeAlloc (sizeof (struct _TPrule_140));
#endif
_currn->_prod = RULErule_140;
_currn->_desc1 = (_TSProotView_MetaInformation) MkrootView_MetaInformation (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_140: root of subtree no. 1 can not be made a rootView_MetaInformation node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_140;
return ( (NODEPTR) _currn);
}/* Mkrule_140 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_141 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_141 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_141 _currn;
#ifdef __cplusplus
_currn = new _TPrule_141;
#else
_currn = (_TPPrule_141) TreeNodeAlloc (sizeof (struct _TPrule_141));
#endif
_currn->_prod = RULErule_141;
_currn->_desc1 = (_TSProotView_DataBase) MkrootView_DataBase (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_141: root of subtree no. 1 can not be made a rootView_DataBase node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_141;
return ( (NODEPTR) _currn);
}/* Mkrule_141 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_142 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_142 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_142 _currn;
#ifdef __cplusplus
_currn = new _TPrule_142;
#else
_currn = (_TPPrule_142) TreeNodeAlloc (sizeof (struct _TPrule_142));
#endif
_currn->_prod = RULErule_142;
_currn->_desc1 = (_TSProotView_Frame) MkrootView_Frame (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_142: root of subtree no. 1 can not be made a rootView_Frame node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_142;
return ( (NODEPTR) _currn);
}/* Mkrule_142 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_143 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_143 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_143 _currn;
#ifdef __cplusplus
_currn = new _TPrule_143;
#else
_currn = (_TPPrule_143) TreeNodeAlloc (sizeof (struct _TPrule_143));
#endif
_currn->_prod = RULErule_143;
_currn->_desc1 = (_TSProotView_ActionHandler) MkrootView_ActionHandler (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_143: root of subtree no. 1 can not be made a rootView_ActionHandler node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_143;
return ( (NODEPTR) _currn);
}/* Mkrule_143 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_144 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_144 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_144 _currn;
#ifdef __cplusplus
_currn = new _TPrule_144;
#else
_currn = (_TPPrule_144) TreeNodeAlloc (sizeof (struct _TPrule_144));
#endif
_currn->_prod = RULErule_144;
_currn->_desc1 = (_TSPwebActionView_ActionHandler) MkwebActionView_ActionHandler (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_144: root of subtree no. 1 can not be made a webActionView_ActionHandler node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_144;
return ( (NODEPTR) _currn);
}/* Mkrule_144 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_145 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_145 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_145 _currn;
#ifdef __cplusplus
_currn = new _TPrule_145;
#else
_currn = (_TPPrule_145) TreeNodeAlloc (sizeof (struct _TPrule_145));
#endif
_currn->_prod = RULErule_145;
_currn->_desc1 = (_TSPwebPageView_FormularCondition) MkwebPageView_FormularCondition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_145: root of subtree no. 1 can not be made a webPageView_FormularCondition node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_145;
return ( (NODEPTR) _currn);
}/* Mkrule_145 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_146 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_146 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_146 _currn;
#ifdef __cplusplus
_currn = new _TPrule_146;
#else
_currn = (_TPPrule_146) TreeNodeAlloc (sizeof (struct _TPrule_146));
#endif
_currn->_prod = RULErule_146;
_currn->_desc1 = (_TSPwebPageView_FormularCondition) MkwebPageView_FormularCondition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_146: root of subtree no. 1 can not be made a webPageView_FormularCondition node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_146;
return ( (NODEPTR) _currn);
}/* Mkrule_146 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_147 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_147 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_147 _currn;
#ifdef __cplusplus
_currn = new _TPrule_147;
#else
_currn = (_TPPrule_147) TreeNodeAlloc (sizeof (struct _TPrule_147));
#endif
_currn->_prod = RULErule_147;
_currn->_desc1 = (_TSPwebPageView_FormularCondition) MkwebPageView_FormularCondition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_147: root of subtree no. 1 can not be made a webPageView_FormularCondition node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_147;
return ( (NODEPTR) _currn);
}/* Mkrule_147 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_148 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_148 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_148 _currn;
#ifdef __cplusplus
_currn = new _TPrule_148;
#else
_currn = (_TPPrule_148) TreeNodeAlloc (sizeof (struct _TPrule_148));
#endif
_currn->_prod = RULErule_148;
_currn->_desc1 = (_TSPwebPageView_FormularCondition) MkwebPageView_FormularCondition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_148: root of subtree no. 1 can not be made a webPageView_FormularCondition node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_148;
return ( (NODEPTR) _currn);
}/* Mkrule_148 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_149 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_149 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_149 _currn;
#ifdef __cplusplus
_currn = new _TPrule_149;
#else
_currn = (_TPPrule_149) TreeNodeAlloc (sizeof (struct _TPrule_149));
#endif
_currn->_prod = RULErule_149;
_currn->_desc1 = (_TSPwebPageView_FormularCondition) MkwebPageView_FormularCondition (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_149: root of subtree no. 1 can not be made a webPageView_FormularCondition node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_149;
return ( (NODEPTR) _currn);
}/* Mkrule_149 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_150 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_150 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_150 _currn;
#ifdef __cplusplus
_currn = new _TPrule_150;
#else
_currn = (_TPPrule_150) TreeNodeAlloc (sizeof (struct _TPrule_150));
#endif
_currn->_prod = RULErule_150;
_currn->_desc1 = (_TSPwebPageView_Frame) MkwebPageView_Frame (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_150: root of subtree no. 1 can not be made a webPageView_Frame node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_150;
return ( (NODEPTR) _currn);
}/* Mkrule_150 */

#if defined(__STDC__) || defined(__cplusplus)
NODEPTR Mkrule_151 (POSITION *_coordref, NODEPTR _desc1)
#else
NODEPTR Mkrule_151 (_coordref,_desc1)
	POSITION *_coordref;
	NODEPTR _desc1;
#endif
{	_TPPrule_151 _currn;
#ifdef __cplusplus
_currn = new _TPrule_151;
#else
_currn = (_TPPrule_151) TreeNodeAlloc (sizeof (struct _TPrule_151));
#endif
_currn->_prod = RULErule_151;
_currn->_desc1 = (_TSPwebPageView_Frame) MkwebPageView_Frame (_coordref, _desc1);	
if (((NODEPTR)_currn->_desc1) == NULLNODEPTR)	
	message (DEADLY, "RULE rule_151: root of subtree no. 1 can not be made a webPageView_Frame node ", 0, _coordref);
_COPYCOORD(_currn)
_TERMACT_rule_151;
return ( (NODEPTR) _currn);
}/* Mkrule_151 */
